/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "driver/driver_wrapper.h"
#include "apsh_instance.h"

#include "os/windows/windows.h"

#define APSH_FUNC_DEFINE(x) x

uint8_t
apsh_get_address_width(apsh_instance_t apsh)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return 0;
#endif

	switch (apsh->page_mode) {
	case APSH_PM_AARCH64:
	case APSH_PM_IA32E:
		return 8;
	case APSH_PM_AARCH32:
	case APSH_PM_LEGACY:
	case APSH_PM_PAE:
		return 4;
	default:
		return 0;
	}
}

os_t
apsh_get_ostype(apsh_instance_t apsh)
{
	return
#ifdef ENABLE_SAFETY_CHECKS
	    (apsh == NULL) ? APSH_FUNC_DEFINE(APSH_OS_UNKNOWN) :
#endif
			  APSH_FUNC_DEFINE(apsh->os_type);
}

bool apsh_get_windows_build_info(apsh_instance_t apsh, win_build_info_t *info);
win_ver_t apsh_get_winver(apsh_instance_t apsh);

uint16_t
apsh_get_win_buildnumber(apsh_instance_t apsh)
{
#ifndef ENABLE_WINDOWS
	print_error("**LibAPSH wasn't compiled with Windows support!\n");
	return 0;
#else

#ifndef ENABLE_JSON_PROFILES
	print_error("**LibAPSH wasn't compiled with JSON profiles support!\n");
	return 0;
#endif

	windows_instance_t windows_instance = NULL;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return 0;

	if (apsh->os_type != APSH_OS_WIN)
		return 0;

	if (!apsh->os_data)
		return 0;

	if (!json_profile(apsh)) {
		print_error("** LibAPSH wasn't initialized with JSON profile!\n");
		return 0;
	}
#endif

	windows_instance = apsh->os_data;

	return windows_instance->build;
#endif
}

const char *
apsh_get_winver_str(apsh_instance_t apsh)
{

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return "APSH_OS_WIN_NONE";
#endif

	win_ver_t ver = apsh_get_winver(apsh);

	switch (ver) {
	case APSH_OS_WIN_NONE:
		return "APSH_OS_WIN_NONE";
	case APSH_OS_WIN_UNKNOWN:
		return "APSH_OS_WIN_UNKNOWN";
	case APSH_OS_WIN_2000:
		return "APSH_OS_WIN_2000";
	case APSH_OS_WIN_XP:
		return "APSH_OS_WIN_XP";
	case APSH_OS_WIN_2003:
		return "APSH_OS_WIN_2003";
	case APSH_OS_WIN_VISTA:
		return "APSH_OS_WIN_VISTA";
	case APSH_OS_WIN_2008:
		return "APSH_OS_WIN_2008";
	case APSH_OS_WIN_7:
		return "APSH_OS_WIN_7";
	default:
		return "<Illegal value for Windows version>";
	} // switch
}

win_ver_t
apsh_get_winver_manual(apsh_instance_t apsh, addr_t kdbg_pa)
{
#ifdef ENABLE_WINDOWS
	return find_windows_version(apsh, kdbg_pa);
#else
	print_error("**LibAPSH wasn't compiled with Windows support!\n");
	return APSH_OS_WIN_NONE;
#endif
}

status_t
apsh_get_offset(apsh_instance_t apsh, const char *offset_name, addr_t *offset)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return APSH_F;

	if (!apsh->os_interface || !apsh->os_interface->os_get_offset)
		return APSH_F;
#endif

	return apsh->os_interface->os_get_offset(apsh, offset_name, offset);
}

status_t
apsh_get_kernel_struct_offset(apsh_instance_t apsh, const char *symbol, const char *member,
			     addr_t *addr)
{

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !addr)
		return 0;
#endif

	return apsh->os_interface->os_get_kernel_struct_offset(apsh, symbol, member, addr);
}

uint64_t
apsh_get_memsize(apsh_instance_t apsh)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return 0;
#endif

	if (driver_get_memsize(apsh, &apsh->allocated_ram_size, &apsh->max_physical_address) == APSH_F)
		return 0;

	return apsh->allocated_ram_size;
}

addr_t
apsh_get_max_physical_address(apsh_instance_t apsh)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return 0;
#endif

	if (driver_get_memsize(apsh, &apsh->allocated_ram_size, &apsh->max_physical_address) == APSH_F)
		return 0;

	return apsh->max_physical_address;
}

unsigned int
apsh_get_num_vcpus(apsh_instance_t apsh)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return 0;
#endif

	return apsh->num_vcpus;
}

char *
apsh_get_name(apsh_instance_t apsh)
{
	char *name = NULL;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;
#endif

	if (driver_get_name(apsh, &name) == APSH_F)
		return NULL;
	return name;
}

status_t
apsh_translate_ksym2v(apsh_instance_t apsh, const char *symbol, addr_t *vaddr)
{
	status_t status = APSH_F;
	addr_t address = 0;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !symbol || !vaddr)
		return APSH_F;
#endif

	status = sym_cache_get(apsh, 0, 0, symbol, &address);

	if (status == APSH_F) {
		if (apsh->os_interface && apsh->os_interface->os_ksym2v) {
			addr_t _base_vaddr;

			status = apsh->os_interface->os_ksym2v(apsh, symbol, &_base_vaddr, &address);
			if (status == APSH_S) {
				address = canonical_addr(address);
				sym_cache_set(apsh, 0, 0, symbol, address);
			}
		}
	}

	*vaddr = address;
	return status;
}

status_t
apsh_translate_sym2v(apsh_instance_t apsh, const access_context_t *ctx, const char *symbol,
		    addr_t *vaddr);

const char *
apsh_translate_v2sym(apsh_instance_t apsh, const access_context_t *ctx, addr_t rva)
{
	char *ret = NULL;
	addr_t dtb = 0;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !ctx)
		return NULL;
#endif

	switch (ctx->translate_mechanism) {
	case APSH_TM_PROCESS_PID:
		if (apsh_pid_to_dtb(apsh, ctx->pid, &dtb) == APSH_F)
			return NULL;
		break;
	case APSH_TM_PROCESS_DTB:
		dtb = ctx->dtb;
		break;
	default:
		debug_print(APSH_DEBUG_MISC, "v2sym only supported in a virtual context!\n");
		return NULL;
	};

	if (rva_cache_get(apsh, ctx->addr, dtb, rva, &ret) == APSH_F) {
		if (apsh->os_interface && apsh->os_interface->os_v2sym)
			ret = apsh->os_interface->os_v2sym(apsh, rva, ctx);

		if (ret)
			rva_cache_set(apsh, ctx->addr, dtb, rva, ret);
	}

	return ret;
}


const char *
apsh_translate_v2ksym(apsh_instance_t apsh, const access_context_t *ctx, addr_t va);


status_t
apsh_pid_to_dtb(apsh_instance_t apsh, apsh_pid_t pid, addr_t *dtb)
{
	status_t ret = APSH_F;
	addr_t _dtb = 0;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !dtb)
		return APSH_F;

	if (!apsh->os_interface)
		return APSH_F;
#endif

	if (!pid) {
		*dtb = apsh->kpgd;
		return APSH_S;
	}

	ret = pid_cache_get(apsh, pid, &_dtb);
	if (ret == APSH_F) {
		if (apsh->os_interface->os_pid_to_pgd)
			ret = apsh->os_interface->os_pid_to_pgd(apsh, pid, &_dtb);

		if (ret == APSH_S)
			pid_cache_set(apsh, pid, _dtb);
	}

	*dtb = _dtb;
	return ret;
}


status_t
apsh_dtb_to_pid(apsh_instance_t apsh, addr_t dtb, apsh_pid_t *pid)
{
	status_t ret = APSH_F;
	apsh_pid_t _pid = -1;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !pid)
		return APSH_F;
#endif

	if (apsh->os_interface && apsh->os_interface->os_pgd_to_pid)
		ret = apsh->os_interface->os_pgd_to_pid(apsh, dtb, &_pid);

	*pid = _pid;
	return ret;
}

void *
apsh_read_page(apsh_instance_t apsh, addr_t frame_num)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;
#endif

	return driver_read_page(apsh, frame_num);
}

void *
apsh_read_memory(apsh_instance_t apsh, addr_t paddr, uint32_t length)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;
#endif

	return driver_read_memory(apsh, paddr, length);
}

GSList *
apsh_get_va_pages(apsh_instance_t apsh, addr_t dtb)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;

	if (!apsh->arch_interface.get_pages[apsh->page_mode]) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "Invalid or not supported paging mode during get_va_pages\n");
		return NULL;
	}
#endif

	return apsh->arch_interface.get_pages[apsh->page_mode](apsh, 0, 0, dtb);
}

GSList *
apsh_get_nested_va_pages(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t pt, page_mode_t pm)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;

	if (valid_npm(npm) && !npt)
		return NULL;

	if (!valid_pm(pm)) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "Invalid or not supported paging mode during get_va_pages\n");
		return NULL;
	}
#endif

	return apsh->arch_interface.get_pages[pm](apsh, npt, npm, pt);
}

status_t
apsh_pagetable_lookup(apsh_instance_t apsh, addr_t pt, addr_t vaddr, addr_t *paddr)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !paddr)
		return APSH_F;
#endif

	return apsh_pagetable_lookup_cache(apsh, pt, vaddr, paddr);
}

status_t
apsh_nested_pagetable_lookup(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t pt,
			    page_mode_t pm, addr_t vaddr, addr_t *paddr, addr_t *naddr)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !paddr)
		return APSH_F;

	if (valid_npm(npm) && !naddr)
		return APSH_F;

	if (!valid_pm(pm)) {
		print_error("Invalid paging mode during apsh_nested_pagetable_lookup: %i\n", pm);
		return APSH_F;
	}
#endif


	if (v2p_cache_get(apsh, vaddr, pt, npt, paddr) == APSH_S) {
		uint8_t value = 0;

		if (apsh_read_8_pa(apsh, *paddr, &value) == APSH_S) {
			if (valid_npm(npm)) {
				*naddr = *paddr;
				*paddr = ~0ull;
			}

			return APSH_S;
		}
	}

	page_info_t info;

	if (apsh->arch_interface.lookup[pm](apsh, npt, npm, pt, vaddr, &info) == APSH_F)
		return APSH_F;

	*paddr = info.paddr;

	if (valid_npm(npm)) {
		*naddr = info.naddr;
		v2p_cache_set(apsh, vaddr, pt, npt, info.naddr);
		return APSH_S;
	}

	v2p_cache_set(apsh, vaddr, pt, 0, info.paddr);
	return APSH_S;
}

/*
 * Return a status when page_info is not needed, but also use the cache,
 * which apsh_pagetable_lookup_extended() does not do.
 *
 */
status_t
apsh_pagetable_lookup_cache(apsh_instance_t apsh, addr_t pt, addr_t vaddr, addr_t *paddr)
{
	status_t ret = APSH_F;
	page_info_t info = {.vaddr = vaddr, .pt = pt, .pm = apsh->page_mode};

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !paddr)
		return ret;

	if (!valid_pm(apsh->page_mode))
		return ret;
#endif

	*paddr = 0;

	if (v2p_cache_get(apsh, vaddr, pt, 0, paddr) == APSH_S) {
		uint8_t value = 0;

		if (apsh_read_8_pa(apsh, *paddr, &value) == APSH_S)
			return APSH_S;

		if (v2p_cache_del(apsh, vaddr, 0, pt) == APSH_F)
			return APSH_F;
	}

	if (apsh->arch_interface.lookup[apsh->page_mode]) {
		ret = apsh->arch_interface.lookup[apsh->page_mode](apsh, 0, 0, pt, vaddr, &info);
	} else {
		print_error("Invalid paging mode during apsh_pagetable_lookup\n");
		ret = APSH_F;
	}

	if (ret == APSH_S) {
		*paddr = info.paddr;
		v2p_cache_set(apsh, vaddr, pt, 0, info.paddr);
	}
	return ret;
}

status_t
apsh_pagetable_lookup_extended(apsh_instance_t apsh, addr_t pt, addr_t vaddr, page_info_t *info)
{
	status_t ret = APSH_F;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !info)
		return ret;

	if (!valid_pm(apsh->page_mode))
		return ret;
#endif

	memset(info, 0, sizeof(page_info_t));
	info->vaddr = vaddr;
	info->pt = pt;
	info->pm = apsh->page_mode;

	if (apsh->arch_interface.lookup[apsh->page_mode])
		ret = apsh->arch_interface.lookup[apsh->page_mode](apsh, 0, 0, pt, vaddr, info);
	else
		print_error("Invalid paging mode during apsh_pagetable_lookup\n");

	if (ret == APSH_S)
		v2p_cache_set(apsh, vaddr, pt, 0, info->paddr);

	return ret;
}


status_t
apsh_translate_kv2p(apsh_instance_t apsh, addr_t virt_address, addr_t *paddr)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !paddr)
		return APSH_F;

	if (!apsh->kpgd) {
		debug_print(APSH_DEBUG_PTLOOKUP, "--early bail on v2p lookup because the kernel page "
					    "global directory is unknown\n");
		return APSH_F;
	}
#endif

	addr_t tmp;

	if (apsh->EPT)
		return apsh_nested_pagetable_lookup(apsh, apsh->EPT, APSH_PM_EPT_4L, apsh->kpgd, apsh->page_mode, virt_address, paddr, &tmp);

	return apsh_pagetable_lookup(apsh, apsh->kpgd, virt_address, paddr);
}

status_t
apsh_translate_uv2p(apsh_instance_t apsh, addr_t virt_address, apsh_pid_t pid, addr_t *paddr)
{
	status_t ret = APSH_F;
	addr_t dtb = 0;
	addr_t hpa;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !paddr)
		return APSH_F;
#endif

	if (apsh_pid_to_dtb(apsh, pid, &dtb) == APSH_F || !dtb) {
		debug_print(APSH_DEBUG_PTLOOKUP, "--early bail on v2p lookup because dtb not found\n");
		return APSH_F;
	}

	if (apsh->EPT)
		ret = apsh_nested_pagetable_lookup(apsh, apsh->EPT, APSH_PM_EPT_4L, dtb, apsh->page_mode,
						  virt_address, paddr, &hpa);
	else
		ret = apsh_pagetable_lookup_cache(apsh, dtb, virt_address, paddr);

	if (ret == APSH_F) {
		if (pid_cache_del(apsh, pid) == APSH_F)
			return APSH_F;

		ret = apsh_pid_to_dtb(apsh, pid, &dtb);
		if (ret == APSH_S) {
			page_info_t info = {0};

			ret = apsh_pagetable_lookup_extended(apsh, dtb, virt_address, &info);

			if (ret == APSH_S)
				*paddr = info.paddr;
		}
	}

	return ret;
}

status_t
apsh_reinit_linux(apsh_instance_t apsh)
{
	return refind_kpgd(apsh);
}

const char *
apsh_get_linux_sysmap(apsh_instance_t apsh)
{
	linux_instance_t linux_instance = NULL;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;

	if (apsh->os_type != APSH_OS_LINUX)
		return NULL;

	if (!apsh->os_data)
		return NULL;
#endif

	linux_instance = apsh->os_data;

	return linux_instance->sysmap;
}

const char *
apsh_get_rekall_path(apsh_instance_t apsh)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;
#endif

#ifdef ENABLE_JSON_PROFILES
	return apsh->json.path;
#else
	return NULL;
#endif
}

const char *
apsh_get_os_profile_path(apsh_instance_t apsh)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return NULL;
#endif

#ifdef ENABLE_JSON_PROFILES
	if (apsh->json.path)
		return apsh->json.path;
#endif

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh->os_data)
		return NULL;
#endif

	switch (apsh->os_type) {
	case APSH_OS_LINUX: {
		linux_instance_t linux_instance = apsh->os_data;

		return linux_instance->sysmap;
	}
	default:
		break;
	};

	return NULL;
}

void
apsh_pidcache_add(apsh_instance_t apsh, apsh_pid_t pid, addr_t dtb)
{
	if (!apsh)
		return;

	return pid_cache_set(apsh, pid, dtb);
}

void
apsh_pidcache_flush(apsh_instance_t apsh)
{
	if (!apsh)
		return;

	return pid_cache_flush(apsh);
}

void
apsh_symcache_add(apsh_instance_t apsh, addr_t base_addr, apsh_pid_t pid, char *sym, addr_t va)
{
	if (!apsh)
		return;

	return sym_cache_set(apsh, base_addr, pid, sym, va);
}

void
apsh_symcache_flush(apsh_instance_t apsh)
{
	if (!apsh)
		return;

	return sym_cache_flush(apsh);
}

void
apsh_rvacache_add(apsh_instance_t apsh, addr_t base_addr, apsh_pid_t pid, addr_t rva, char *sym)
{
	addr_t pgd = 0;

	if (!apsh)
		return;

	if (apsh_pid_to_dtb(apsh, pid, &pgd) != APSH_S) {
		debug_print(APSH_DEBUG_SYMCACHE, "--SYM cache failed to find base for PID %u\n", pid);
		return;
	}

	return rva_cache_set(apsh, base_addr, pgd, rva, sym);
}

void
apsh_rvacache_flush(apsh_instance_t apsh)
{
	if (!apsh)
		return;

	return rva_cache_flush(apsh);
}

void
apsh_v2pcache_add(apsh_instance_t apsh, addr_t va, addr_t pt, addr_t pa)
{
	if (!apsh)
		return;

	return v2p_cache_set(apsh, va, pt, 0, pa);
}

void
apsh_v2pcache_nested_add(apsh_instance_t apsh, addr_t va, addr_t pt, addr_t npt, addr_t pa)
{
	if (!apsh)
		return;

	return v2p_cache_set(apsh, va, pt, npt, pa);
}

void
apsh_v2pcache_flush(apsh_instance_t apsh, addr_t pt)
{
	if (!apsh)
		return;

	return v2p_cache_flush(apsh, pt, 0);
}

void
apsh_v2pcache_nested_flush(apsh_instance_t apsh, addr_t pt, addr_t npt)
{
	if (!apsh)
		return;

	return v2p_cache_flush(apsh, pt, npt);
}

bool
apsh_get_windows_build_info(apsh_instance_t apsh, win_build_info_t *info)
{
#ifndef ENABLE_WINDOWS
	print_error("**LibAPSH wasn't compiled with Windows support!\n");
	return false;
#else
	APSH_FUNC_DEFINE(windows_instance_t windows_instance = NULL);

#ifdef ENABLE_SAFETY_CHECKS
	if (!info || !apsh) {
		print_error("**%s:%d Error: null pointer!\n", __FUNCTION__, __LINE__);
		return false;
	}

	if (apsh->os_type != APSH_OS_WIN)
		return false;

	if (!apsh->os_data)
		return false;
#endif

	windows_instance = apsh->os_data;

	if (!windows_instance->version || windows_instance->version == APSH_OS_WIN_UNKNOWN) {
		addr_t kdbg = windows_instance->ntoskrnl + windows_instance->kdbg_offset;

		windows_instance->version = find_windows_version(apsh, kdbg);
	}

	APSH_FUNC_DEFINE(info->version = windows_instance->version);
	APSH_FUNC_DEFINE(info->buildnumber = windows_instance->build);
	APSH_FUNC_DEFINE(info->major = windows_instance->major);
	APSH_FUNC_DEFINE(info->minor = windows_instance->minor);

	return true;
#endif
}

win_ver_t
apsh_get_winver(apsh_instance_t apsh)
{
#ifndef ENABLE_WINDOWS
	print_error("**LibAPSH wasn't compiled with Windows support!\n");
	return APSH_OS_WIN_NONE;
#else
	windows_instance_t windows_instance = NULL;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh)
		return APSH_OS_WIN_NONE;

	if (apsh->os_type != APSH_OS_WIN)
		return APSH_OS_WIN_NONE;

	if (!apsh->os_data)
		return APSH_OS_WIN_NONE;
#endif

	windows_instance = apsh->os_data;

	if (!windows_instance->version || windows_instance->version == APSH_OS_WIN_UNKNOWN) {
		addr_t kdbg = windows_instance->ntoskrnl + windows_instance->kdbg_offset;

		windows_instance->version = find_windows_version(apsh, kdbg);
	}
	return windows_instance->version;
#endif
}

status_t
apsh_translate_sym2v(apsh_instance_t apsh, const access_context_t *ctx, const char *symbol,
		    addr_t *vaddr)
{
	status_t status;
	addr_t rva = 0;
	addr_t address = 0;
	addr_t dtb = 0;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !ctx || !symbol || !vaddr)
		return APSH_F;
#endif

	switch (ctx->translate_mechanism) {
	case APSH_TM_PROCESS_PID:
		if (apsh_pid_to_dtb(apsh, ctx->pid, &dtb) == APSH_F)
			return APSH_F;
		break;
	case APSH_TM_PROCESS_DTB:
		dtb = ctx->dtb;
		break;
	default:
		debug_print(APSH_DEBUG_MISC, "sym2v only supported in a virtual context!\n");
		return APSH_F;
	};

	status = sym_cache_get(apsh, ctx->addr, dtb, symbol, &address);
	if (status == APSH_F) {
		if (apsh->os_interface && apsh->os_interface->os_usym2rva) {
			status = apsh->os_interface->os_usym2rva(apsh, ctx, symbol, &rva);
			if (status == APSH_S) {
				address = canonical_addr(ctx->addr + rva);
				sym_cache_set(apsh, ctx->addr, dtb, symbol, address);
			}
		}
	}

	*vaddr = address;
	return status;
}

const char *
apsh_translate_v2ksym(apsh_instance_t apsh, const access_context_t *ctx, addr_t va)
{
	char *ret = NULL;
	addr_t dtb = 0;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh || !ctx)
		return NULL;
#endif

	switch (ctx->translate_mechanism) {
	case APSH_TM_PROCESS_PID:
		if (apsh_pid_to_dtb(apsh, ctx->pid, &dtb) == APSH_F)
			return NULL;
		break;
	case APSH_TM_PROCESS_DTB:
		dtb = ctx->dtb;
		break;
	default:
		debug_print(APSH_DEBUG_MISC, "v2ksym only supported in a virtual context!\n");
		return NULL;
	};

	if (rva_cache_get(apsh, ctx->addr, dtb, va, &ret) == APSH_F) {
		if (apsh->os_interface && apsh->os_interface->os_v2ksym)
			ret = apsh->os_interface->os_v2ksym(apsh, va, ctx);

		if (ret)
			rva_cache_set(apsh, ctx->addr, dtb, va, ret);
	}

	return ret;
}
