/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef LIBAPSH_H
#define LIBAPSH_H

#include "apsh_const.h"
#include "apsh_page.h"
#include <doca_dev.h>
#include <glib.h>

#ifdef __cplusplus
extern "C" {
#define NOEXCEPT noexcept
#else
#define NOEXCEPT
#endif

#define APSH_TM_NONE              0
#define APSH_TM_PROCESS_DTB       1
#define APSH_TM_PROCESS_PT        1
#define APSH_TM_PROCESS_PID       2
#define APSH_TM_KERNEL_SYMBOL     3

#define ACCESS_CONTEXT_VERSION 0x1u
#define ACCESS_CONTEXT(C, ...)                  \
    access_context_t C = {                      \
        .version = ACCESS_CONTEXT_VERSION,      \
        __VA_ARGS__                             \
    }

typedef uint32_t translation_mechanism_t;

typedef enum arch {
    APSH_ARCH_UNKNOWN,
    APSH_ARCH_X86,
    APSH_ARCH_X86_64,
    APSH_ARCH_ARM32,
    APSH_ARCH_ARM64
} apsh_arch_t;

typedef struct {
    uint32_t version;
    uint32_t _pad;

    union {
        page_mode_t pm;
        page_mode_t page_mode;
    };
    union {
        page_mode_t npm;
        page_mode_t nested_page_mode;
    };
    union {
        addr_t npt;
        addr_t nested_page_table;
    };
    union {
        translation_mechanism_t tm;
        translation_mechanism_t translate_mechanism;
    };

    uint32_t _pad2;

    union {
        struct {
            addr_t addr;

            union {
                addr_t pt;
                addr_t page_table;
                addr_t dtb;
                apsh_pid_t pid;
            };
        };

        /* APSH_TM_KERNEL_SYMBOL */
        const char *ksym;
    };
} access_context_t;

#define APSH_GET_BIT(reg, bit) (!!(reg & (1ULL<<bit)))

#define APSH_BIT_MASK(a, b) (((unsigned long long) -1 >> (63 - (b))) & ~((1ULL << (a)) - 1))

typedef struct _ustring {

    size_t length;

    uint8_t *contents;

    const char *encoding;
} unicode_string_t;

typedef struct apsh_instance *apsh_instance_t;



status_t apsh_init(
    apsh_instance_t *apsh,
    apsh_mode_t mode,
    const void *domain,
    uint64_t init_flags,
    apsh_init_data_t *init_data,
    addr_t vf_id,
    const char* mem_region_json,
    const char* kpgd_file,
    const char* mlx_device,
    struct doca_dev *regex_dev,
    struct doca_dev_rep *system_dev,
    apsh_init_error_t *error) NOEXCEPT;

status_t apsh_init_complete(
    apsh_instance_t *apsh,
    const void *domain,
    uint64_t init_flags,
    apsh_init_data_t *init_data,
    apsh_config_t config_mode,
    void *config,
    addr_t EPT,
    addr_t vf_id,
    const char* mem_region_json,
    const char* kpgd_file,
    const char* mlx_device,
    struct doca_dev *regex_dev,
    struct doca_dev_rep *system_dev,
    apsh_init_error_t *error) NOEXCEPT;

page_mode_t apsh_init_paging(
    apsh_instance_t apsh,
    uint64_t flags) NOEXCEPT;

status_t apsh_set_paging(
    apsh_instance_t apsh,
    page_mode_t page_mode) NOEXCEPT;

status_t apsh_set_limits_table(
    apsh_instance_t apsh,
    GHashTable *limits_table) NOEXCEPT;

os_t apsh_init_os(
    apsh_instance_t apsh,
    apsh_config_t config_mode,
    void *config,
    addr_t EPT,
    const char* kpgd_file,
    apsh_init_error_t *error) NOEXCEPT;

status_t apsh_destroy(
    apsh_instance_t apsh) NOEXCEPT;

apsh_arch_t apsh_get_library_arch(void) NOEXCEPT;



status_t apsh_translate_kv2p(
    apsh_instance_t apsh,
    addr_t vaddr,
    addr_t *paddr) NOEXCEPT;

status_t apsh_translate_uv2p(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    addr_t *paddr) NOEXCEPT;

status_t apsh_translate_ksym2v(
    apsh_instance_t apsh,
    const char *symbol,
    addr_t *vaddr) NOEXCEPT;

status_t apsh_translate_sym2v(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    const char *symbol,
    addr_t *vaddr) NOEXCEPT;

const char* apsh_translate_v2sym(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    addr_t rva) NOEXCEPT;

const char* apsh_translate_v2ksym(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    addr_t va) NOEXCEPT;

status_t apsh_pid_to_dtb(
    apsh_instance_t apsh,
    apsh_pid_t pid,
    addr_t *dtb) NOEXCEPT;

status_t apsh_dtb_to_pid(
    apsh_instance_t apsh,
    addr_t dtb,
    apsh_pid_t *pid) NOEXCEPT;


status_t apsh_pagetable_lookup (
    apsh_instance_t apsh,
    addr_t pt,
    addr_t vaddr,
    addr_t *paddr) NOEXCEPT;

status_t apsh_pagetable_lookup_extended(
    apsh_instance_t apsh,
    addr_t pt,
    addr_t vaddr,
    page_info_t *info) NOEXCEPT;

status_t apsh_nested_pagetable_lookup (
    apsh_instance_t apsh,
    addr_t npt,
    page_mode_t npm,
    addr_t pt,
    page_mode_t pm,
    addr_t vaddr,
    addr_t *paddr,
    addr_t *naddr) NOEXCEPT;

status_t apsh_nested_pagetable_lookup_extended(
    apsh_instance_t apsh,
    addr_t npt,
    page_mode_t npm,
    addr_t pt,
    page_mode_t pm,
    addr_t vaddr,
    page_info_t *info) NOEXCEPT;



status_t apsh_read(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    size_t count,
    void *buf,
    size_t *bytes_read) NOEXCEPT;

status_t apsh_read_8(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    uint8_t * value) NOEXCEPT;

status_t apsh_read_16(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    uint16_t * value) NOEXCEPT;

status_t apsh_read_32(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    uint32_t * value) NOEXCEPT;

status_t apsh_read_64(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    uint64_t * value) NOEXCEPT;

status_t apsh_read_addr(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    addr_t *value) NOEXCEPT;

char *apsh_read_str(
    apsh_instance_t apsh,
    const access_context_t *ctx) NOEXCEPT;

unicode_string_t *apsh_read_unicode_str(
    apsh_instance_t apsh,
    const access_context_t *ctx) NOEXCEPT;

unicode_string_t *apsh_read_unicode_str_pm(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    page_mode_t mode ) NOEXCEPT;

status_t apsh_read_ksym(
    apsh_instance_t apsh,
    const char *sym,
    size_t count,
    void *buf,
    size_t *bytes_read
) NOEXCEPT;

status_t apsh_read_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    size_t count,
    void *buf,
    size_t *bytes_read
) NOEXCEPT;

status_t apsh_mmap_guest(
    apsh_instance_t apsh,
    const access_context_t *ctx,
    size_t num_pages,
    void **access_ptrs
) NOEXCEPT;

status_t apsh_read_pa(
    apsh_instance_t apsh,
    addr_t paddr,
    size_t count,
    void *buf,
    size_t *bytes_read
) NOEXCEPT;

status_t apsh_read_8_ksym(
    apsh_instance_t apsh,
    const char *sym,
    uint8_t * value) NOEXCEPT;

status_t apsh_read_16_ksym(
    apsh_instance_t apsh,
    const char *sym,
    uint16_t * value) NOEXCEPT;

status_t apsh_read_32_ksym(
    apsh_instance_t apsh,
    const char *sym,
    uint32_t * value) NOEXCEPT;

status_t apsh_read_64_ksym(
    apsh_instance_t apsh,
    const char *sym,
    uint64_t * value) NOEXCEPT;

status_t apsh_read_addr_ksym(
    apsh_instance_t apsh,
    const char *sym,
    addr_t *value) NOEXCEPT;

char *apsh_read_str_ksym(
    apsh_instance_t apsh,
    const char *sym) NOEXCEPT;

status_t apsh_read_8_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    uint8_t * value) NOEXCEPT;

status_t apsh_read_16_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    uint16_t * value) NOEXCEPT;

status_t apsh_read_32_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    uint32_t * value) NOEXCEPT;

status_t apsh_read_64_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    uint64_t * value) NOEXCEPT;

status_t apsh_read_addr_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    addr_t *value) NOEXCEPT;

char *apsh_read_str_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid) NOEXCEPT;

unicode_string_t *apsh_read_unicode_str_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid) NOEXCEPT;

status_t apsh_convert_str_encoding(
    const unicode_string_t *in,
    unicode_string_t *out,
    const char *outencoding) NOEXCEPT;

void apsh_free_unicode_str(
    unicode_string_t *p_us) NOEXCEPT;

status_t apsh_read_8_pa(
    apsh_instance_t apsh,
    addr_t paddr,
    uint8_t * value) NOEXCEPT;

status_t apsh_read_16_pa(
    apsh_instance_t apsh,
    addr_t paddr,
    uint16_t * value) NOEXCEPT;

status_t apsh_read_32_pa(
    apsh_instance_t apsh,
    addr_t paddr,
    uint32_t * value) NOEXCEPT;

status_t apsh_read_64_pa(
    apsh_instance_t apsh,
    addr_t paddr,
    uint64_t * value) NOEXCEPT;

status_t apsh_read_addr_pa(
    apsh_instance_t apsh,
    addr_t paddr,
    addr_t *value) NOEXCEPT;

char *apsh_read_str_pa(
    apsh_instance_t apsh,
    addr_t paddr) NOEXCEPT;



void apsh_print_hex(
    const unsigned char *data,
    unsigned long length) NOEXCEPT;

void apsh_print_hex_ksym(
    apsh_instance_t apsh,
    const char *sym,
    size_t length) NOEXCEPT;

void apsh_print_hex_va(
    apsh_instance_t apsh,
    addr_t vaddr,
    apsh_pid_t pid,
    size_t length) NOEXCEPT;

void apsh_print_hex_pa(
    apsh_instance_t apsh,
    addr_t paddr,
    size_t length) NOEXCEPT;

char *apsh_get_name(
    apsh_instance_t apsh) NOEXCEPT;

uint64_t apsh_get_apshd(
    apsh_instance_t apsh) NOEXCEPT;

status_t apsh_get_access_mode(
    apsh_instance_t apsh,
    const void *domain,
    uint64_t init_flags,
    apsh_init_data_t *init_data,
    apsh_mode_t *mode) NOEXCEPT;

page_mode_t apsh_get_page_mode(
    apsh_instance_t apsh,
    unsigned long vcpu) NOEXCEPT;

uint8_t apsh_get_address_width(
    apsh_instance_t apsh) NOEXCEPT;

os_t apsh_get_ostype(
    apsh_instance_t apsh) NOEXCEPT;

bool apsh_get_windows_build_info(
    apsh_instance_t apsh, win_build_info_t* info) NOEXCEPT;

win_ver_t apsh_get_winver(
    apsh_instance_t apsh) NOEXCEPT;

uint16_t apsh_get_win_buildnumber(
    apsh_instance_t apsh) NOEXCEPT;

const char *apsh_get_winver_str(
    apsh_instance_t apsh) NOEXCEPT;

win_ver_t apsh_get_winver_manual(
    apsh_instance_t apsh,
    addr_t kdvb_pa) NOEXCEPT;

status_t apsh_get_offset(
    apsh_instance_t apsh,
    const char *offset_name,
    addr_t *offset) NOEXCEPT;

status_t apsh_get_kernel_struct_offset(
    apsh_instance_t apsh,
    const char* struct_name,
    const char* member,
    addr_t *addr) NOEXCEPT;

status_t apsh_get_xsave_info(
    apsh_instance_t apsh,
    unsigned long vcpu,
    xsave_area_t *xsave_info) NOEXCEPT;

uint64_t apsh_get_memsize(
    apsh_instance_t apsh) NOEXCEPT;

addr_t apsh_get_max_physical_address(
    apsh_instance_t apsh) NOEXCEPT;

unsigned int apsh_get_num_vcpus (
    apsh_instance_t apsh) NOEXCEPT;

status_t
apsh_request_page_fault(
    apsh_instance_t apsh,
    unsigned long vcpu,
    uint64_t virtual_address,
    uint32_t error_code) NOEXCEPT;

status_t
apsh_get_tsc_info(
    apsh_instance_t apsh,
    uint32_t *tsc_mode,
    uint64_t *elapsed_nsec,
    uint32_t *gtsc_khz,
    uint32_t *incarnation) NOEXCEPT;

void apsh_v2pcache_flush(
    apsh_instance_t apsh,
    addr_t pt) NOEXCEPT;

void apsh_v2pcache_nested_flush(
    apsh_instance_t apsh,
    addr_t pt,
    addr_t npt) NOEXCEPT;

void apsh_v2pcache_add(
    apsh_instance_t apsh,
    addr_t va,
    addr_t pt,
    addr_t pa) NOEXCEPT;

void apsh_v2pcache_nested_add(
    apsh_instance_t apsh,
    addr_t va,
    addr_t pt,
    addr_t npt,
    addr_t pa) NOEXCEPT;

void apsh_symcache_flush(
    apsh_instance_t apsh) NOEXCEPT;

void apsh_symcache_add(
    apsh_instance_t apsh,
    addr_t base_addr,
    apsh_pid_t pid,
    char *sym,
    addr_t va) NOEXCEPT;

void apsh_rvacache_flush(
    apsh_instance_t apsh) NOEXCEPT;

void apsh_rvacache_add(
    apsh_instance_t apsh,
    addr_t base_addr,
    apsh_pid_t pid,
    addr_t rva,
    char *sym) NOEXCEPT;

void apsh_pidcache_flush(
    apsh_instance_t apsh) NOEXCEPT;

void apsh_pidcache_add(
    apsh_instance_t apsh,
    apsh_pid_t pid,
    addr_t dtb) NOEXCEPT;

void apsh_pagecache_flush(
    apsh_instance_t apsh) NOEXCEPT;

status_t apsh_reinit_linux(apsh_instance_t apsh) NOEXCEPT;

const char * apsh_get_linux_sysmap(apsh_instance_t apsh) NOEXCEPT;

const char * apsh_get_freebsd_sysmap(apsh_instance_t apsh) NOEXCEPT;

const char *apsh_get_rekall_path(
    apsh_instance_t apsh) NOEXCEPT;

const char *apsh_get_os_profile_path(
    apsh_instance_t apsh) NOEXCEPT;

#ifdef __cplusplus
}
#endif

#endif /* LIBAPSH_H */
