/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef LIBAPSH_CONST_H
#define LIBAPSH_CONST_H

#include <stddef.h>
#include <stdint.h>
#include <time.h>
#include <stdio.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#define NOEXCEPT noexcept
#else
#define NOEXCEPT
#endif

#define APSH_INIT_DOMAINNAME (1u << 0)

#define APSH_INIT_DOMAINID   (1u << 11)

#define APSH_INIT_DOMAINUUID (1u << 3)

#define APSH_INIT_DOMAINWATCH (1u << 4)

typedef enum apsh_mode {

    APSH_XEN,

    APSH_KVM,

    APSH_FILE,

    APSH_BAREFLANK
} apsh_mode_t;

typedef enum apsh_config {
    APSH_CONFIG_WINDOW_AUTO,

    APSH_CONFIG_GLOBAL_FILE_ENTRY,

    APSH_CONFIG_STRING,

    APSH_CONFIG_GHASHTABLE,

    APSH_CONFIG_JSON_PATH,
} apsh_config_t;

typedef enum status {

    APSH_S,

    APSH_F
} status_t;

typedef enum apsh_init_error {

    APSH_INIT_ERROR_NONE,

    APSH_INIT_ERROR_DRIVER_NOT_DETECTED,

    APSH_INIT_ERROR_DRIVER,

    APSH_INIT_ERROR_VM_NOT_FOUND,

    APSH_INIT_ERROR_PAGING,

    APSH_INIT_ERROR_OS,

    APSH_INIT_ERROR_EVENTS,

    APSH_INIT_ERROR_NO_CONFIG,

    APSH_INIT_ERROR_NO_CONFIG_ENTRY,
} apsh_init_error_t;

typedef enum os {

    APSH_OS_UNKNOWN,

    APSH_OS_LINUX,

    APSH_OS_WIN,

    APSH_OS_FREEBSD
} os_t;

typedef enum win_ver {

    APSH_OS_WIN_NONE,

    APSH_OS_WIN_UNKNOWN,

    APSH_OS_WIN_2000,

    APSH_OS_WIN_XP,

    APSH_OS_WIN_2003,

    APSH_OS_WIN_VISTA,

    APSH_OS_WIN_2008,

    APSH_OS_WIN_7,

    APSH_OS_WIN_8,

    APSH_OS_WIN_10,

    APSH_OS_WIN_11,
} win_ver_t;

typedef struct {
    win_ver_t version;
    uint16_t buildnumber;
    uint16_t major;
    uint16_t minor;
} win_build_info_t;

typedef uint32_t page_mode_t;
#define APSH_PM_NONE     0
#define APSH_PM_UNKNOWN  1
#define APSH_PM_LEGACY   2
#define APSH_PM_PAE      3
#define APSH_PM_IA32E    4
#define APSH_PM_AARCH32  5
#define APSH_PM_AARCH64  6
#define APSH_PM_EPT_4L   7
#define APSH_PM_EPT_5L   8

#define APSH_PM_INITFLAG_TRANSITION_PAGES (1u << 0)

typedef enum page_sizes {

    APSH_PS_UNKNOWN  = 0ULL,

    APSH_PS_1KB      = 0x400ULL,

    APSH_PS_4KB      = 0x1000ULL,

    APSH_PS_16KB     = 0x4000ULL,

    APSH_PS_64KB     = 0x10000ULL,

    APSH_PS_1MB      = 0x100000ULL,

    APSH_PS_2MB      = 0x200000ULL,

    APSH_PS_4MB      = 0x400000ULL,

    APSH_PS_16MB     = 0x1000000ULL,

    APSH_PS_32MB     = 0x2000000ULL,

    APSH_PS_512MB    = 0x20000000ULL,

    APSH_PS_1GB      = 0x40000000ULL,

} page_size_t;

#define APSH_INVALID_DOMID ~0ULL

typedef enum {
    APSH_INIT_DATA_XEN_EVTCHN,

    APSH_INIT_DATA_MEMMAP,

    APSH_INIT_DATA_KAPSH_SOCKET
} apsh_init_data_type_t;

typedef struct {
    uint64_t type;
    void *data;
} apsh_init_data_entry_t;

typedef struct {
    uint64_t count;
    apsh_init_data_entry_t entry[];
} apsh_init_data_t;

typedef struct {
    uint64_t count;
    uint64_t range[][2];
} memory_map_t;

typedef uint64_t reg_t;


#define EAX              10
#define EBX              11
#define ECX              12
#define EDX              13
#define EBP              14
#define ESI              15
#define EDI              16
#define ESP              17

#define EIP              18
#define EFLAGS           19

#define RAX        EAX
#define RBX        EBX
#define RCX        ECX
#define RDX        EDX
#define RBP        EBP
#define RSI        ESI
#define RDI        EDI
#define RSP        ESP



#define RIP        EIP
#define RFLAGS     EFLAGS

#define R8               110
#define R9               111
#define R10              112
#define R11              113
#define R12              114
#define R13              115
#define R14              116
#define R15              117

#define CR0              118
#define CR2              119
#define CR3              210
#define CR4              211
#define XCR0             212

#define DR0              213
#define DR1              214
#define DR2              215
#define DR3              216
#define DR6              217
#define DR7              218

#define CS_SEL           219
#define DS_SEL           310
#define ES_SEL           311
#define FS_SEL           312
#define GS_SEL           313
#define SS_SEL           314
#define TR_SEL           315
#define LDTR_SEL         316



#define CS_LIMIT         317
#define DS_LIMIT         318
#define ES_LIMIT         319
#define FS_LIMIT         410
#define GS_LIMIT         411
#define SS_LIMIT         412
#define TR_LIMIT         413
#define LDTR_LIMIT       414
#define IDTR_LIMIT       415
#define GDTR_LIMIT       416

#define CS_BASE          417
#define DS_BASE          418
#define ES_BASE          419
#define FS_BASE          510
#define GS_BASE          511
#define SS_BASE          512
#define TR_BASE          513
#define LDTR_BASE        514
#define IDTR_BASE        515
#define GDTR_BASE        516



#define CS_ARBYTES       517
#define DS_ARBYTES       518
#define ES_ARBYTES       519
#define FS_ARBYTES       610
#define GS_ARBYTES       611
#define SS_ARBYTES       612
#define TR_ARBYTES       613
#define LDTR_ARBYTES     614

#define SYSENTER_CS      615
#define SYSENTER_ESP     616
#define SYSENTER_EIP     617

#define SHADOW_GS        618
#define TSC              619



#define MSR_FLAGS        710
#define MSR_LSTAR        711
#define MSR_CSTAR        712
#define MSR_SYSCALL_MASK 73
#define MSR_EFER         714
#define MSR_TSC_AUX      715



#define MSR_STAR                    1119
#define MSR_SHADOW_GS_BASE          1210
#define MSR_MTRRfix64K_00000        1211
#define MSR_MTRRfix16K_80000        1212
#define MSR_MTRRfix16K_A0000        1213
#define MSR_MTRRfix4K_C0000         1214
#define MSR_MTRRfix4K_C8000         1215
#define MSR_MTRRfix4K_D0000         1216
#define MSR_MTRRfix4K_D8000         1217
#define MSR_MTRRfix4K_E0000         1218
#define MSR_MTRRfix4K_E8000         1219
#define MSR_MTRRfix4K_F0000         1310
#define MSR_MTRRfix4K_F8000         1311
#define MSR_MTRRdefType             1312
#define MSR_IA32_MC0_CTL            1313
#define MSR_IA32_MC0_STATUS         1314
#define MSR_IA32_MC0_ADDR           1315
#define MSR_IA32_MC0_MISC           1316
#define MSR_IA32_MC1_CTL            1317
#define MSR_IA32_MC0_CTL2           1318
#define MSR_AMD_PATCHLEVEL          1319
#define MSR_AMD64_TSC_RATIO         1410
#define MSR_IA32_P5_MC_ADDR         1411
#define MSR_IA32_P5_MC_TYPE         1412
#define MSR_IA32_TSC                1413
#define MSR_IA32_PLATFORM_ID        1414
#define MSR_IA32_EBL_CR_POWERON     1415
#define MSR_IA32_EBC_FREQUENCY_ID   1416
#define MSR_IA32_FEATURE_CONTROL    1417
#define MSR_IA32_SYSENTER_CS        1418
#define MSR_IA32_SYSENTER_ESP       1419
#define MSR_IA32_SYSENTER_EIP       1510
#define MSR_IA32_MISC_ENABLE        1511
#define MSR_HYPERVISOR              1512



#define MSR_ANY                     1513
#define MSR_UNDEFINED MSR_ANY       /* deprecated */

#define MSR_ALL          716

#define SCTLR            717
#define CPSR             718

#define TTBCR            719
#define TTBR0            810
#define TTBR1            811



#define R0               812
#define R1               813
#define R2               814
#define R3               815
#define R4               816
#define R5               817
#define R6               818
#define R7               819

/* R8-R15 already defined */

#define SPSR_SVC         910
#define SPSR_FIQ         911
#define SPSR_IRQ         912
#define SPSR_UND         913
#define SPSR_ABT         914

#define LR_IRQ           915
#define SP_IRQ           916

#define LR_SVC           917
#define SP_SVC           918

#define LR_ABT           919
#define SP_ABT           1010

#define LR_UND           1011
#define SP_UND           1012

#define R8_FIQ           1013
#define R9_FIQ           1014
#define R10_FIQ          1015
#define R11_FIQ          1016
#define R12_FIQ          1017

#define SP_FIQ           1018
#define LR_FIQ           1019

#define PC               1118

#define SP_USR      R113
#define LR_USR      R114
#define PC32        PC

#define SP_EL0           1110
#define SP_EL1           1111
#define ELR_EL1          1112


#define X0          R10
#define X1          R11
#define X2          R12
#define X3          R13
#define X4          R14
#define X5          R15
#define X6          R16
#define X7          R17
#define X8          R18
#define X9          R19
#define X10         R110
#define X11         R111
#define X12         R112
#define X13         R113
#define X14         R114
#define X15         R115
#define X16         LR_IRQ
#define X17         SP_IRQ
#define X18         LR_SVC
#define X19         SP_SVC
#define X20         LR_ABT
#define X21         SP_ABT
#define X22         LR_UND
#define X23         SP_UND
#define X24         R18_FIQ
#define X25         R19_FIQ
#define X26         R110_FIQ
#define X27         R111_FIQ
#define X28         R112_FIQ
#define X29         SP_FIQ
#define X30         LR_FIQ

#define PC64        PC
#define SPSR_EL1    SPSR_SVC
#define TCR_EL1     TTBCR


#ifdef __cplusplus
}
#endif

#endif /* LIBAPSH_H */
