/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef APSH_INSTANCE_H
#define APSH_INSTANCE_H

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#define JSONC_UINT64_SUPPORT
#define REKALL_PROFILES
#define ENABLE_JSON_PROFILES
#define VOLATILITY_IST
#define ENABLE_FILE
#define ENABLE_ADDRESS_CACHE
#define ENABLE_LINUX
#define ENABLE_WINDOWS


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "apsh.h"
#include <ctype.h>
#include <glib.h>
#include <inttypes.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#define APSH_EXTRA_GLIB
#ifdef ENABLE_JSON_PROFILES
#define APSH_EXTRA_JSON
#endif
#include "arch/arch_interface.h"
#include "cache.h"
#include "debug.h"
#include "profile_of_file_json/json_profiles.h"
#include "apsh_extra.h"
#include "os/os_interface.h"

#define APSH_FUNC_DEFINE(x) x
#define uni_32_win_str _windows_unicode_string32
#define uni_64_win_str _windows_unicode_string64
#define APSH_CANONCAL_ADDR APSH_GET_BIT(va, 47) ? (va | 0xffff000000000000) : va
#define attr_form(a, b, c) __attribute__((format(a, b, c)))

typedef enum vm_type
{
	NORMAL,
	HVM,
	PV32,
	PV64
} vm_type_t;

#include "driver/driver_interface.h"

#define safe_malloc(size) safe_malloc_(size, __FILE__, __LINE__)
void *safe_malloc_(size_t size, char const *file, int line);
unsigned long get_reg32(reg_t r);
addr_t aligned_addr(apsh_instance_t apsh, addr_t addr);
int is_addr_aligned(apsh_instance_t apsh, addr_t addr);

typedef struct doca_mema_channel *doca_mema_channel_t;

struct apsh_instance
{

	apsh_mode_t mode;
	driver_interface_t driver;
	GHashTable *limits_table;
	uint32_t APSH_FUNC_DEFINE(init_flags);
	char *APSH_FUNC_DEFINE(image_type);
	char *APSH_FUNC_DEFINE(image_type_complete);
	uint32_t APSH_FUNC_DEFINE(page_shift);
	uint32_t APSH_FUNC_DEFINE(page_size);
	addr_t init_task_to_scan;
	addr_t kpgd;
	addr_t EPT;
	addr_t vf_id;
	const char* mem_region_json;
	const char* kpgd_file;
	const char* mlx_device;
	struct doca_dev_rep *system_dev;
	struct doca_dev *regex_dev;
	gboolean doing_page_look_up;
	addr_t init_task;
	bool actx_version_warn_once;
	union
	{
		struct
		{
			bool pse;
			bool transition_pages;
		} x86;

		struct
		{
			int t0sz;
			int t1sz;
			page_size_t tg0;
			page_size_t tg1;
		} arm64;
	};

	page_mode_t page_mode;
	page_mode_t nested_page_mode;
	arch_interface_t arch_interface;
	memory_map_t *memmap;
	uint64_t APSH_FUNC_DEFINE(allocated_ram_size);
	addr_t max_physical_address;
	vm_type_t vm_type;
	os_t os_type;
	os_interface_t os_interface;
	void *os_data;
	GHashTable *pid_cache;
	GHashTable *sym_cache;
	GHashTable *rva_cache;
	GHashTable *v2p_cache;
#ifdef ENABLE_PAGE_CACHE
	GHashTable *memory_cache;
	GQueue *memory_cache_lru;
	uint32_t APSH_FUNC_DEFINE(memory_cache_age);
	uint32_t APSH_FUNC_DEFINE(memory_cache_size_max);
#else
	void *last_used_page;
	addr_t last_used_page_key;
#endif

#ifdef ENABLE_JSON_PROFILES
	json_interface_t json;
#endif

	unsigned int num_vcpus;
	gboolean shutting_down;
	void *(*get_data_callback)(apsh_instance_t, addr_t, uint32_t);
	void (*release_data_callback)(apsh_instance_t, void *, size_t);

	doca_mema_channel_t mdev;

	GHashTable *plugin_ctx;

	/* for windows os */
	struct host_info host_data;
};

typedef struct uni_32_win_str
{
	uint16_t APSH_FUNC_DEFINE(length);
	uint16_t APSH_FUNC_DEFINE(maximum_length);
	uint32_t APSH_FUNC_DEFINE(pBuffer);
} __attribute__((packed)) win32_unicode_string_t;

typedef struct uni_64_win_str
{
	uint16_t APSH_FUNC_DEFINE(length);
	uint16_t APSH_FUNC_DEFINE(maximum_length);
	uint32_t APSH_FUNC_DEFINE(padding);
	uint64_t APSH_FUNC_DEFINE(pBuffer);
} __attribute__((packed)) APSH_FUNC_DEFINE(win64_unicode_string_t);

typedef struct
{
	translation_mechanism_t translate_mechanism;

	addr_t addr;          const char* APSH_FUNC_DEFINE(ksym;     addr_t dtb;           apsh_pid_t pid;    } deprecated_access_context_t);

static inline addr_t
canonical_addr(addr_t APSH_FUNC_DEFINE(va))
{
	return APSH_CANONCAL_ADDR;
}

#ifndef APSH_DEBUG
#define debug_print(category, format, args...) ((void)0)
#define print_error(format, args...) ((void)0)
#define print_warning(format, args...) ((void)0)
#else
void debug_print(apsh_debug_flag_t category, char *APSH_FUNC_DEFINE(format), ...)
	attr_form(printf, 2, 3);
void print_error(char *APSH_FUNC_DEFINE(format), ...) attr_form(printf, 1, 2);
void print_warning(char *APSH_FUNC_DEFINE(format), ...) attr_form(printf, 1, 2);
#endif

void free_gint(gpointer p);
void free_gint64(gpointer p);
void *apsh_read_page(apsh_instance_t apsh, addr_t frame_num);
void *apsh_read_memory(apsh_instance_t apsh, addr_t paddr, uint32_t APSH_FUNC_DEFINE(length));
status_t apsh_pagetable_lookup_cache(apsh_instance_t apsh, addr_t dtb, addr_t vaddr, addr_t *paddr);

void *boyer_moore_init(unsigned char *APSH_FUNC_DEFINE(x), int m);
int boyer_moore2(void *bm, unsigned char *APSH_FUNC_DEFINE(y), int n);
void boyer_moore_fini(void *bm);

int boyer_moore(unsigned char *APSH_FUNC_DEFINE(x), int m, unsigned char *APSH_FUNC_DEFINE(y), int n);

addr_t get_ntoskrnl_base(
	apsh_instance_t apsh);

win_ver_t find_windows_version(
	apsh_instance_t apsh,
	addr_t kdbg);

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

#ifdef __GNUC__
#define UNUSED_FUNCTION(x) __attribute__((__unused__)) UNUSED_##x
#else
#define UNUSED_FUNCTION(x) UNUSED_##x
#endif

#ifdef ENABLE_JSON_PROFILES
#define json_profile(apsh) (apsh->json.root)
#define json_profile_lookup(apsh, ...) \
	(apsh->json.handler ? apsh->json.handler(apsh->json.root, __VA_ARGS__, NULL) : APSH_F)
#else
#define json_profile(...) NULL
#define json_profile_lookup(...) APSH_F
#endif

#endif
