/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <glib.h>
#include <stdlib.h>
#include <sys/mman.h>

#include "arch/amd64.h"
#include "driver/driver_wrapper.h"
#include "apsh_instance.h"
#include "x86.h"

/* PML4 Table  */
static inline addr_t
get_pml4_index(addr_t vaddr)
{
	return (vaddr & APSH_BIT_MASK(39, 47)) >> 36;
}

static inline status_t
get_pml4e(apsh_instance_t apsh, access_context_t *ctx, page_info_t *amd64_inf_page)
{
	ctx->addr = (amd64_inf_page->pt & APSH_BIT_MASK(12, 51)) |
		    get_pml4_index(amd64_inf_page->vaddr);

	if (apsh_read_64(apsh, ctx, &amd64_inf_page->x86_ia32e.pml4e_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: error reading pml4e_location = 0x%.16" PRIx64 "\n",
			    ctx->addr);
		return APSH_F;
	}

	amd64_inf_page->x86_ia32e.pml4e_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pml4e_location = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pml4e_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pml4e_value = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pml4e_value);
	return APSH_S;
}

static inline addr_t
get_pdpt_index_ia32e(addr_t vaddr)
{
	return (vaddr & APSH_BIT_MASK(30, 38)) >> 27;
}

static inline status_t
get_pdpte_ia32e(apsh_instance_t apsh, access_context_t *ctx, page_info_t *amd64_inf_page)
{
	ctx->addr = (amd64_inf_page->x86_ia32e.pml4e_value & APSH_BIT_MASK(12, 51)) |
		    get_pdpt_index_ia32e(amd64_inf_page->vaddr);

	if (apsh_read_64(apsh, ctx, &amd64_inf_page->x86_ia32e.pdpte_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pdpte_location = 0x%.16" PRIx64 "\n",
			    ctx->addr);
		return APSH_F;
	}

	amd64_inf_page->x86_ia32e.pdpte_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pdpte_location = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pdpte_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pdpte_value = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pdpte_value);
	return APSH_S;
}

static inline uint64_t
get_pd_index_ia32e(addr_t vaddr)
{
	return (vaddr & APSH_BIT_MASK(21, 29)) >> 18;
}

static inline status_t
get_pde_ia32e(apsh_instance_t apsh, access_context_t *ctx, page_info_t *amd64_inf_page)
{
	ctx->addr = (amd64_inf_page->x86_ia32e.pdpte_value & APSH_BIT_MASK(12, 51)) |
		    get_pd_index_ia32e(amd64_inf_page->vaddr);

	if (apsh_read_64(apsh, ctx, &amd64_inf_page->x86_ia32e.pgd_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pde_location = 0x%.16" PRIx64 "\n",
			    ctx->addr);
		return APSH_F;
	}

	amd64_inf_page->x86_ia32e.pgd_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pde_location = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pgd_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pde_value= 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pgd_value);

	return APSH_S;
}

static inline uint64_t
get_pt_index_ia32e(addr_t vaddr)
{
	return (vaddr & APSH_BIT_MASK(12, 20)) >> 9;
}

static inline status_t
get_pte_ia32e(apsh_instance_t apsh, access_context_t *ctx, page_info_t *amd64_inf_page)
{
	ctx->addr = (amd64_inf_page->x86_ia32e.pgd_value & APSH_BIT_MASK(12, 51)) |
		    get_pt_index_ia32e(amd64_inf_page->vaddr);

	if (apsh_read_64(apsh, ctx, &amd64_inf_page->x86_ia32e.pte_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pte_location = 0x%.16" PRIx64 "\n",
			    ctx->addr);
		return APSH_F;
	}

	amd64_inf_page->x86_ia32e.pte_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pte_location = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pte_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pte_value = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->x86_ia32e.pte_value);

	return APSH_S;
}

static inline uint64_t
get_paddr_ia32e(addr_t vaddr, uint64_t pte)
{
	return (pte & APSH_BIT_MASK(12, 51)) | (vaddr & APSH_BIT_MASK(0, 11));
}

static inline uint64_t
get_gigpage_ia32e(addr_t vaddr, uint64_t pdpte)
{
	return (pdpte & APSH_BIT_MASK(30, 51)) | (vaddr & APSH_BIT_MASK(0, 29));
}

static inline uint64_t
get_2megpage_ia32e(addr_t vaddr, uint64_t pde)
{
	return (pde & APSH_BIT_MASK(21, 51)) | (vaddr & APSH_BIT_MASK(0, 20));
}

status_t
v2p_ia32e(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t pt, addr_t vaddr,
	  page_info_t *amd64_inf_page)
{
	status_t status;

	ACCESS_CONTEXT(ctx, .npt = npt, .npm = npm);

	amd64_inf_page->npt = npt;
	amd64_inf_page->npm = npm;
	amd64_inf_page->pt = pt;
	amd64_inf_page->pm = APSH_PM_IA32E;
	amd64_inf_page->vaddr = vaddr;

	apsh->doing_page_look_up = true;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: lookup addr = 0x%.16" PRIx64 "\n", vaddr);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: npt = 0x%.16" PRIx64 " npm = %" PRIu32 "\n",
		    npt, npm);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pt = 0x%.16" PRIx64 "\n", pt);

	status = get_pml4e(apsh, &ctx, amd64_inf_page);
	if (status != APSH_S)
		goto done;

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, amd64_inf_page->x86_ia32e.pml4e_value)) {
		status = APSH_F;
		goto done;
	}

	status = get_pdpte_ia32e(apsh, &ctx, amd64_inf_page);
	if (status != APSH_S)
		goto done;

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, amd64_inf_page->x86_ia32e.pdpte_value)) {
		status = APSH_F;
		goto done;
	}

	if (PAGE_SIZE(amd64_inf_page->x86_ia32e.pdpte_value)) { /* pdpte maps a 1GB page */
		amd64_inf_page->size = APSH_PS_1GB;
		amd64_inf_page->paddr =
			get_gigpage_ia32e(vaddr, amd64_inf_page->x86_ia32e.pdpte_value);
		status = APSH_S;
		debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: 1GB page\n");
		goto done;
	}

	status = get_pde_ia32e(apsh, &ctx, amd64_inf_page);
	if (status != APSH_S)
		goto done;

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, amd64_inf_page->x86_ia32e.pgd_value)) {
		status = APSH_F;
		goto done;
	}

	if (PAGE_SIZE(amd64_inf_page->x86_ia32e.pgd_value)) { /* pde maps a 2MB page */
		amd64_inf_page->size = APSH_PS_2MB;
		amd64_inf_page->paddr =
			get_2megpage_ia32e(vaddr, amd64_inf_page->x86_ia32e.pgd_value);
		status = APSH_S;
		debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: 2MB page\n");
		goto done;
	}

	status = get_pte_ia32e(apsh, &ctx, amd64_inf_page);
	if (status != APSH_S)
		goto done;

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, amd64_inf_page->x86_ia32e.pte_value)) {
		status = APSH_F;
		goto done;
	}

	amd64_inf_page->size = APSH_PS_4KB;
	amd64_inf_page->paddr = get_paddr_ia32e(vaddr, amd64_inf_page->x86_ia32e.pte_value);
	status = APSH_S;

done:
	apsh->doing_page_look_up = false;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: paddr = 0x%.16" PRIx64 "\n",
		    amd64_inf_page->paddr);

	if (valid_npm(npm) &&
	    apsh_nested_pagetable_lookup(apsh, 0, 0, npt, npm, amd64_inf_page->paddr,
						   &amd64_inf_page->naddr, NULL) == APSH_F)
		return APSH_F;

	return status;
}

GSList *
get_pages_ia32e(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t dtb)
{

	GSList *ret = NULL;
	uint8_t entry_size = 0x8;

	/* 0x1000/0x8 */
#define IA32E_ENTRIES_PER_PAGE 0x200

	uint64_t *pml4_page = g_malloc(APSH_PS_4KB);
	uint64_t *pdpt_page = g_try_malloc0(APSH_PS_4KB);
	uint64_t *pgd_page = g_try_malloc0(APSH_PS_4KB);
	uint64_t *pt_page = g_try_malloc0(APSH_PS_4KB);

	if (!pml4_page || !pdpt_page || !pgd_page || !pt_page)
		goto done;

	addr_t pml4e_location = dtb & APSH_BIT_MASK(12, 51);

	ACCESS_CONTEXT(ctx, .npt = npt, .npm = npm);

	ctx.addr = pml4e_location;
	if (apsh_read(apsh, &ctx, APSH_PS_4KB, pml4_page, NULL) == APSH_F)
		goto done;

	uint64_t pml4e_index;

	for (pml4e_index = 0; pml4e_index < IA32E_ENTRIES_PER_PAGE;
	     pml4e_index++, pml4e_location += entry_size) {

		uint64_t pml4e_value = pml4_page[pml4e_index];

		if (!ENTRY_PRESENT(apsh->x86.transition_pages, pml4e_value))
			continue;


		uint64_t pdpte_location = pml4e_value & APSH_BIT_MASK(12, 51);

		ctx.addr = pdpte_location;
		if (apsh_read(apsh, &ctx, APSH_PS_4KB, pdpt_page, NULL) == APSH_F)
			goto done;

		uint64_t pdpte_index;

		for (pdpte_index = 0; pdpte_index < IA32E_ENTRIES_PER_PAGE;
		     pdpte_index++, pdpte_location++) {

			uint64_t pdpte_value = pdpt_page[pdpte_index];

			if (!ENTRY_PRESENT(apsh->x86.transition_pages, pdpte_value))
				continue;

			if (PAGE_SIZE(pdpte_value)) {
				page_info_t *amd64_inf_page = g_try_malloc0(sizeof(page_info_t));

				if (!amd64_inf_page)
					goto done;

				amd64_inf_page->vaddr =
					canonical_addr((pml4e_index << 39) | (pdpte_index << 30));
				amd64_inf_page->pt = dtb;
				amd64_inf_page->paddr =
					get_gigpage_ia32e(amd64_inf_page->vaddr, pdpte_value);
				amd64_inf_page->size = APSH_PS_1GB;
				amd64_inf_page->x86_ia32e.pml4e_location = pml4e_location;
				amd64_inf_page->x86_ia32e.pml4e_value = pml4e_value;
				amd64_inf_page->x86_ia32e.pdpte_location = pdpte_location;
				amd64_inf_page->x86_ia32e.pdpte_value = pdpte_value;
				ret = g_slist_prepend(ret, amd64_inf_page);
				continue;
			}

			uint64_t pgd_location = pdpte_value & APSH_BIT_MASK(12, 51);

			ctx.addr = pgd_location;
			if (apsh_read(apsh, &ctx, APSH_PS_4KB, pgd_page, NULL) == APSH_F)
				goto done;

			uint64_t pgde_index;

			for (pgde_index = 0; pgde_index < IA32E_ENTRIES_PER_PAGE;
			     pgde_index++, pgd_location += entry_size) {

				uint64_t pgd_value = pgd_page[pgde_index];

				if (ENTRY_PRESENT(apsh->os_type, pgd_value)) {

					if (PAGE_SIZE(pgd_value)) {
						page_info_t *amd64_inf_page =
							g_try_malloc0(sizeof(page_info_t));
						if (!amd64_inf_page)
							goto done;

						amd64_inf_page->vaddr = canonical_addr(
							(pml4e_index << 39) | (pdpte_index << 30) |
							(pgde_index << 21));
						amd64_inf_page->pt = dtb;
						amd64_inf_page->paddr = get_2megpage_ia32e(
							amd64_inf_page->vaddr, pgd_value);
						amd64_inf_page->size = APSH_PS_2MB;
						amd64_inf_page->x86_ia32e.pml4e_location =
							pml4e_location;
						amd64_inf_page->x86_ia32e.pml4e_value = pml4e_value;
						amd64_inf_page->x86_ia32e.pdpte_location =
							pdpte_location;
						amd64_inf_page->x86_ia32e.pdpte_value = pdpte_value;
						amd64_inf_page->x86_ia32e.pgd_location =
							pgd_location;
						amd64_inf_page->x86_ia32e.pgd_value = pgd_value;
						ret = g_slist_prepend(ret, amd64_inf_page);
						continue;
					}

					uint64_t pte_location = (pgd_value & APSH_BIT_MASK(12, 51));

					ctx.addr = pte_location;
					if (APSH_F ==
					    apsh_read(apsh, &ctx, APSH_PS_4KB, pt_page, NULL))
						goto done;

					uint64_t pte_index;
					for (pte_index = 0; pte_index < IA32E_ENTRIES_PER_PAGE;
					     pte_index++, pte_location += entry_size) {
						uint64_t pte_value = pt_page[pte_index];

						if (ENTRY_PRESENT(apsh->os_type, pte_value)) {
							page_info_t *amd64_inf_page =
								g_try_malloc0(sizeof(page_info_t));
							if (!amd64_inf_page)
								goto done;

							amd64_inf_page->vaddr =
								canonical_addr((pml4e_index << 39) |
									       (pdpte_index << 30) |
									       (pgde_index << 21) |
									       (pte_index << 12));
							amd64_inf_page->pt = dtb;
							amd64_inf_page->paddr = get_paddr_ia32e(
								amd64_inf_page->vaddr, pte_value);
							amd64_inf_page->size = APSH_PS_4KB;
							amd64_inf_page->x86_ia32e.pml4e_location =
								pml4e_location;
							amd64_inf_page->x86_ia32e.pml4e_value =
								pml4e_value;
							amd64_inf_page->x86_ia32e.pdpte_location =
								pdpte_location;
							amd64_inf_page->x86_ia32e.pdpte_value =
								pdpte_value;
							amd64_inf_page->x86_ia32e.pgd_location =
								pgd_location;
							amd64_inf_page->x86_ia32e.pgd_value =
								pgd_value;
							amd64_inf_page->x86_ia32e.pte_location =
								pte_location;
							amd64_inf_page->x86_ia32e.pte_value =
								pte_value;
							ret = g_slist_prepend(ret, amd64_inf_page);
							continue;
						}
					}
				}
			}
		}
	}

done:
	g_free(pt_page);
	g_free(pgd_page);
	g_free(pdpt_page);
	g_free(pml4_page);

	return ret;
}
