/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <glib.h>
#include <stdlib.h>
#include <sys/mman.h>

#include "arch/arm_aarch32.h"
#include "driver/driver_wrapper.h"
#include "apsh_instance.h"

static inline void get_fine_second_level_descriptor(apsh_instance_t apsh, uint32_t vaddr,
						    page_info_t *info);

static inline uint32_t coarse_second_level_table_index(uint32_t vaddr);

static inline uint32_t
first_level_table_index(uint32_t vaddr)
{
	return (vaddr >> 20);
}

static inline void
get_first_level_descriptor(apsh_instance_t apsh, uint32_t dtb, uint32_t vaddr, page_info_t *info)
{
	info->arm_aarch32.fld_location =
		(dtb & APSH_BIT_MASK(14, 31)) | (first_level_table_index(vaddr) << 2);
	uint32_t fld_v;

	if (apsh_read_32_pa(apsh, info->arm_aarch32.fld_location, &fld_v) == APSH_S)
		info->arm_aarch32.fld_value = fld_v;
}

static inline void
get_coarse_second_level_descriptor(apsh_instance_t apsh, uint32_t vaddr, page_info_t *info)
{
	info->arm_aarch32.sld_location = (info->arm_aarch32.fld_value & APSH_BIT_MASK(10, 31)) |
					 (coarse_second_level_table_index(vaddr) << 2);
	uint32_t sld_v;

	if (apsh_read_32_pa(apsh, info->arm_aarch32.sld_location, &sld_v) == APSH_S)
		info->arm_aarch32.sld_value = sld_v;
}

static inline uint32_t
fine_second_level_table_index(uint32_t vaddr)
{
	return (vaddr >> 10) & APSH_BIT_MASK(0, 9);
}

status_t
v2p_aarch32(apsh_instance_t apsh, addr_t UNUSED(npt), page_mode_t UNUSED(npm), addr_t pt,
	    addr_t vaddr, page_info_t *info)
{
	status_t status = APSH_F;

	apsh->doing_page_look_up = true;

	debug_print(APSH_DEBUG_PTLOOKUP,
		    "--ARM AArch32 PTLookup: vaddr = 0x%.16" PRIx64 ", pt = 0x%.16" PRIx64 "\n",
		    vaddr, pt);

	get_first_level_descriptor(apsh, pt, vaddr, info);

	debug_print(APSH_DEBUG_PTLOOKUP, "--ARM AArch32 PTLookup: fld_location = 0x%" PRIx32 "\n",
		    info->arm_aarch32.fld_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--ARM AArch32 PTLookup: fld_value = 0x%" PRIx32 "\n",
		    info->arm_aarch32.fld_value);

	switch (info->arm_aarch32.fld_value & APSH_BIT_MASK(0, 1)) {

	case 0b01: {

		debug_print(
			APSH_DEBUG_PTLOOKUP,
			"--ARM AArch32 PTLookup: the entry gives the physical address of a coarse second-level table\n");

		get_coarse_second_level_descriptor(apsh, vaddr, info);

		debug_print(APSH_DEBUG_PTLOOKUP, "--ARM AArch32 PTLookup: l2d = 0x%" PRIx32 "\n",
			    info->arm_aarch32.sld_value);

		switch (info->arm_aarch32.sld_value & APSH_BIT_MASK(0, 1)) {
		case 0b01:
			info->size = APSH_PS_64KB;
			info->paddr = (info->arm_aarch32.sld_value & APSH_BIT_MASK(16, 31)) |
				      (vaddr & APSH_BIT_MASK(0, 15));
			status = APSH_S;
			break;
		case 0b10:
		case 0b11:
			info->size = APSH_PS_4KB;
			info->paddr = (info->arm_aarch32.sld_value & APSH_BIT_MASK(12, 31)) |
				      (vaddr & APSH_BIT_MASK(0, 11));
			status = APSH_S;
		default:
			break;
		}

		break;
	}

	case 0b10: {

		if (!APSH_GET_BIT(info->arm_aarch32.fld_value, 18)) {
			debug_print(
				APSH_DEBUG_PTLOOKUP,
				"--ARM AArch32 PTLookup: the entry is a section descriptor for its associated modified virtual addresses\n");
			info->size = APSH_PS_1MB;
			info->paddr = (info->arm_aarch32.fld_value & APSH_BIT_MASK(20, 31)) |
				      (vaddr & APSH_BIT_MASK(0, 19));
		} else {
			debug_print(
				APSH_DEBUG_PTLOOKUP,
				"--ARM AArch32 PTLookup: the entry is a supersection descriptor for its associated modified virtual addresses\n");
			info->size = APSH_PS_16MB;
			info->paddr = (info->arm_aarch32.fld_value & APSH_BIT_MASK(24, 31)) |
				      (vaddr & APSH_BIT_MASK(0, 23));
		}
		status = APSH_S;

		break;
	}

	case 0b11: {

		debug_print(
			APSH_DEBUG_PTLOOKUP,
			"--ARM AArch32 PTLookup: the entry gives the physical address of a fine second-level table\n");

		get_fine_second_level_descriptor(apsh, vaddr, info);

		debug_print(APSH_DEBUG_PTLOOKUP, "--ARM AArch32 PTLookup: sld = 0x%" PRIx32 "\n",
			    info->arm_aarch32.sld_value);

		switch (info->arm_aarch32.sld_value & APSH_BIT_MASK(0, 1)) {
		case 0b01:
			info->size = APSH_PS_64KB;
			info->paddr = (info->arm_aarch32.sld_value & APSH_BIT_MASK(16, 31)) |
				      (vaddr & APSH_BIT_MASK(0, 15));
			status = APSH_S;
			break;
		case 0b10:
			info->size = APSH_PS_4KB;
			info->paddr = (info->arm_aarch32.sld_value & APSH_BIT_MASK(12, 31)) |
				      (vaddr & APSH_BIT_MASK(0, 11));
			status = APSH_S;
			break;
		case 0b11:
			info->size = APSH_PS_1KB;
			info->paddr = (info->arm_aarch32.sld_value & APSH_BIT_MASK(10, 31)) |
				      (vaddr & APSH_BIT_MASK(0, 9));
			status = APSH_S;
			break;
		default:
			break;
		}

		break;
	}

	default:
		break;
	}

	apsh->doing_page_look_up = false;
	debug_print(APSH_DEBUG_PTLOOKUP, "--ARM PTLookup: PA = 0x%" PRIx64 "\n", info->paddr);

	return status;
}

GSList *
get_va_pages_aarch32(apsh_instance_t UNUSED(apsh), addr_t UNUSED(dtb))
{
	return NULL;
}

static inline void
get_fine_second_level_descriptor(apsh_instance_t apsh, uint32_t vaddr, page_info_t *info)
{
	info->arm_aarch32.sld_location = (info->arm_aarch32.fld_value & APSH_BIT_MASK(12, 31)) |
					 (fine_second_level_table_index(vaddr) << 2);
	uint32_t sld_v;

	if (apsh_read_32_pa(apsh, info->arm_aarch32.sld_location, &sld_v) == APSH_S)
		info->arm_aarch32.sld_value = sld_v;
}

static inline uint32_t
coarse_second_level_table_index(uint32_t vaddr)
{
	return (vaddr >> 12) & APSH_BIT_MASK(0, 7);
}
