/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <glib.h>
#include <stdlib.h>
#include <sys/mman.h>

#include "arch/arm_aarch64.h"
#include "driver/driver_wrapper.h"
#include "apsh_instance.h"

static inline uint64_t
zero_level_4kb_table_index(uint64_t vaddr)
{
	return (vaddr >> 39) & APSH_BIT_MASK(0, 8);
}

static inline void
get_zero_level_4kb_descriptor(apsh_instance_t apsh, uint64_t dtb, uint64_t vaddr, page_info_t *info)
{
	info->arm_aarch64.zld_location =
		(dtb & APSH_BIT_MASK(12, 47)) | (zero_level_4kb_table_index(vaddr) << 3);
	uint64_t zld_v;

	if (apsh_read_64_pa(apsh, info->arm_aarch64.zld_location, &zld_v) == APSH_S)
		info->arm_aarch64.zld_value = zld_v;
}

static inline uint64_t
first_level_4kb_table_index(uint64_t vaddr)
{
	return (vaddr >> 30) & APSH_BIT_MASK(0, 8);
}

static inline void get_first_level_4kb_descriptor(apsh_instance_t apsh, uint64_t dtb,
						  uint64_t vaddr, page_info_t *info);

static inline uint64_t
first_level_64kb_table_index(uint64_t vaddr)
{
	return (vaddr >> 42) & APSH_BIT_MASK(0, 5);
}

static inline void get_first_level_64kb_descriptor(apsh_instance_t apsh, uint64_t dtb,
						   uint64_t vaddr, page_info_t *info);

static inline uint64_t second_level_4kb_table_index(uint64_t vaddr);

static inline void get_second_level_4kb_descriptor(apsh_instance_t apsh, uint64_t dtb,
						   uint64_t vaddr, page_info_t *info);

static inline uint64_t
second_level_64kb_table_index(uint64_t vaddr)
{
	return (vaddr >> 29) & APSH_BIT_MASK(0, 12);
}

static inline void get_second_level_64kb_descriptor(apsh_instance_t apsh, uint64_t dtb,
						    uint64_t vaddr, page_info_t *info);

static inline uint64_t
third_level_4kb_table_index(uint64_t vaddr)
{
	return (vaddr >> 12) & APSH_BIT_MASK(0, 8);
}

static inline void
get_third_level_4kb_descriptor(apsh_instance_t apsh, uint64_t vaddr, page_info_t *info)
{
	info->arm_aarch64.tld_location = (info->arm_aarch64.sld_value & APSH_BIT_MASK(12, 47)) |
					 (third_level_4kb_table_index(vaddr) << 3);
	uint64_t tld_v;

	if (apsh_read_64_pa(apsh, info->arm_aarch64.tld_location, &tld_v) == APSH_S)
		info->arm_aarch64.tld_value = tld_v;
}

static inline uint64_t
third_level_64kb_table_index(uint64_t vaddr)
{
	return (vaddr >> 16) & APSH_BIT_MASK(0, 12);
}

static inline void get_third_level_64kb_descriptor(apsh_instance_t apsh, uint64_t vaddr,
						   page_info_t *info);

status_t
v2p_aarch64(apsh_instance_t apsh, addr_t UNUSED(npt), page_mode_t UNUSED(npm), addr_t pt,
	    addr_t vaddr, page_info_t *info)
{
	status_t status = APSH_F;

	apsh->doing_page_look_up = true;

	debug_print(APSH_DEBUG_PTLOOKUP,
		    "--ARM AArch64 PTLookup: vaddr = 0x%.16" PRIx64 ", pt = 0x%.16" PRIx64 "\n",
		    vaddr, pt);

	bool is_pt_ttbr1 = false;
	page_size_t ps;
	uint8_t levels;
	uint8_t va_width;

	if (pt == apsh->kpgd)
		is_pt_ttbr1 = true;

	if (is_pt_ttbr1) {
		ps = apsh->arm64.tg1;
		va_width = 64 - apsh->arm64.t1sz;
	} else {
		ps = apsh->arm64.tg0;
		va_width = 64 - apsh->arm64.t0sz;
	}

	if (ps == APSH_PS_4KB)
		levels = va_width == 39 ? 3 : 4;
	else if (ps == APSH_PS_64KB)
		levels = va_width == 42 ? 2 : 3;
	else {
		print_error("16KB granule size ARM64 lookups are not yet implemented\n");
		goto done;
	}

	if (levels == 4) {
		get_zero_level_4kb_descriptor(apsh, pt, vaddr, info);
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--ARM AArch64 PTLookup: zld_value = 0x%" PRIx64 "\n",
			    info->arm_aarch64.zld_value);

		if ((info->arm_aarch64.zld_value & APSH_BIT_MASK(0, 1)) != 0b11)
			goto done;

		pt = info->arm_aarch64.zld_value & APSH_BIT_MASK(12, 47);
		--levels;
	}

	if (levels == 3) {
		if (ps == APSH_PS_4KB) {
			get_first_level_4kb_descriptor(apsh, pt, vaddr, info);
			debug_print(APSH_DEBUG_PTLOOKUP,
				    "--ARM AArch64 4kb PTLookup: fld_value = 0x%" PRIx64 "\n",
				    info->arm_aarch64.fld_value);

			switch (info->arm_aarch64.fld_value & APSH_BIT_MASK(0, 1)) {
			case 0b11:
				pt = info->arm_aarch64.fld_value & APSH_BIT_MASK(12, 47);
				--levels;
				break;
			case 0b01:
				info->size = APSH_PS_1GB;
				info->paddr =
					(info->arm_aarch64.fld_value & APSH_BIT_MASK(30, 47)) |
					(vaddr & APSH_BIT_MASK(0, 29));
				status = APSH_S;
				goto done;
			default:
				goto done;
			}
		}
		if (ps == APSH_PS_64KB) {
			get_first_level_64kb_descriptor(apsh, pt, vaddr, info);
			debug_print(APSH_DEBUG_PTLOOKUP,
				    "--ARM AArch64 64kb PTLookup: fld_value = 0x%" PRIx64 "\n",
				    info->arm_aarch64.fld_value);

			switch (info->arm_aarch64.fld_value & APSH_BIT_MASK(0, 1)) {
			case 0b11:
				pt = info->arm_aarch64.fld_value & APSH_BIT_MASK(16, 47);
				--levels;
				break;
			default:
				goto done;
			}
		}
	}

	if (levels == 2) {
		if (ps == APSH_PS_4KB) {
			get_second_level_4kb_descriptor(apsh, pt, vaddr, info);
			debug_print(APSH_DEBUG_PTLOOKUP,
				    "--ARM AArch64 4kb PTLookup: sld_value = 0x%" PRIx64 "\n",
				    info->arm_aarch64.sld_value);

			switch (info->arm_aarch64.sld_value & APSH_BIT_MASK(0, 1)) {
			case 0b11:
				get_third_level_4kb_descriptor(apsh, vaddr, info);
				debug_print(APSH_DEBUG_PTLOOKUP,
					    "--ARM AArch64 4kb PTLookup: tld_value = 0x%" PRIx64
					    "\n",
					    info->arm_aarch64.tld_value);

				info->size = APSH_PS_4KB;
				info->paddr =
					(info->arm_aarch64.tld_value & APSH_BIT_MASK(12, 47)) |
					(vaddr & APSH_BIT_MASK(0, 11));
				status = APSH_S;
				break;
			case 0b01:
				info->size = APSH_PS_2MB;
				info->paddr =
					(info->arm_aarch64.sld_value & APSH_BIT_MASK(21, 47)) |
					(vaddr & APSH_BIT_MASK(0, 20));
				status = APSH_S;
				goto done;
			default:
				goto done;
			}
		}
		if (ps == APSH_PS_64KB) {
			get_second_level_64kb_descriptor(apsh, pt, vaddr, info);
			debug_print(APSH_DEBUG_PTLOOKUP,
				    "--ARM AArch64 64kb PTLookup: sld_value = 0x%" PRIx64 "\n",
				    info->arm_aarch64.sld_value);

			switch (info->arm_aarch64.sld_value & APSH_BIT_MASK(0, 1)) {
			case 0b11:
				get_third_level_64kb_descriptor(apsh, vaddr, info);
				debug_print(APSH_DEBUG_PTLOOKUP,
					    "--ARM AArch64 64kb PTLookup: tld_value = 0x%" PRIx64
					    "\n",
					    info->arm_aarch64.tld_value);

				info->size = APSH_PS_4KB;
				info->paddr =
					(info->arm_aarch64.tld_value & APSH_BIT_MASK(16, 47)) |
					(vaddr & APSH_BIT_MASK(0, 15));
				status = APSH_S;
				goto done;
			case 0b01:
				info->size = APSH_PS_512MB;
				info->paddr =
					(info->arm_aarch64.sld_value & APSH_BIT_MASK(29, 47)) |
					(vaddr & APSH_BIT_MASK(0, 28));
				status = APSH_S;
				goto done;
			default:
				goto done;
			}
		}
	}

done:
	apsh->doing_page_look_up = false;
	debug_print(APSH_DEBUG_PTLOOKUP, "--ARM PTLookup: PA = 0x%" PRIx64 "\n", info->paddr);
	return status;
}

GSList *
get_va_pages_aarch64(apsh_instance_t UNUSED(apsh), addr_t UNUSED(dtb))
{
	return NULL;
}

static inline void
get_first_level_4kb_descriptor(apsh_instance_t apsh, uint64_t dtb, uint64_t vaddr,
			       page_info_t *info)
{
	info->arm_aarch64.fld_location =
		(dtb & APSH_BIT_MASK(12, 47)) | (first_level_4kb_table_index(vaddr) << 3);
	uint64_t fld_v;

	if (apsh_read_64_pa(apsh, info->arm_aarch64.fld_location, &fld_v) == APSH_S)
		info->arm_aarch64.fld_value = fld_v;
}

static inline void
get_second_level_64kb_descriptor(apsh_instance_t apsh, uint64_t dtb, uint64_t vaddr,
				 page_info_t *info)
{
	info->arm_aarch64.sld_location =
		(dtb & APSH_BIT_MASK(16, 47)) | (second_level_64kb_table_index(vaddr) << 3);
	uint64_t sld_v;

	if (apsh_read_64_pa(apsh, info->arm_aarch64.sld_location, &sld_v) == APSH_S)
		info->arm_aarch64.sld_value = sld_v;
}

static inline void
get_third_level_64kb_descriptor(apsh_instance_t apsh, uint64_t vaddr, page_info_t *info)
{
	info->arm_aarch64.tld_location = (info->arm_aarch64.sld_value & APSH_BIT_MASK(16, 47)) |
					 (third_level_64kb_table_index(vaddr) << 3);
	uint64_t tld_v;

	if (apsh_read_64_pa(apsh, info->arm_aarch64.tld_location, &tld_v) == APSH_S)
		info->arm_aarch64.tld_value = tld_v;
}

static inline void
get_second_level_4kb_descriptor(apsh_instance_t apsh, uint64_t dtb, uint64_t vaddr,
				page_info_t *info)
{
	info->arm_aarch64.sld_location =
		(dtb & APSH_BIT_MASK(12, 47)) | (second_level_4kb_table_index(vaddr) << 3);
	uint64_t sld_v;

	if (apsh_read_64_pa(apsh, info->arm_aarch64.sld_location, &sld_v) == APSH_S)
		info->arm_aarch64.sld_value = sld_v;
}

static inline uint64_t
second_level_4kb_table_index(uint64_t vaddr)
{
	return (vaddr >> 21) & APSH_BIT_MASK(0, 8);
}

static inline void
get_first_level_64kb_descriptor(apsh_instance_t apsh, uint64_t dtb, uint64_t vaddr,
				page_info_t *info)
{
	info->arm_aarch64.fld_location =
		(dtb & APSH_BIT_MASK(9, 47)) | (first_level_64kb_table_index(vaddr) << 3);
	uint64_t fld_v;

	if (apsh_read_64_pa(apsh, info->arm_aarch64.fld_location, &fld_v) == APSH_S)
		info->arm_aarch64.fld_value = fld_v;
}
