/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <glib.h>
#include <stdlib.h>
#include <sys/mman.h>

#include "arch/intel.h"
#include "driver/driver_wrapper.h"
#include "apsh_instance.h"
#include "x86.h"

#define APSH_FUNC_DEFINE(x) x
#define uit_32 uint32_t
#define uit_64 uint64_t

/* page directory pointer table */
static inline uit_64
get_pdptb(uit_64 pdpr)
{
	return pdpr & APSH_BIT_MASK(5, 63);
}

static inline uit_32
pdpi_index(uit_32 pdpi)
{
	return (pdpi >> 30) * sizeof(uit_64);
}

static inline status_t
get_pdpi(apsh_instance_t instance, access_context_t *ctx, page_info_t *info)
{
	ctx->addr = get_pdptb(info->pt) + pdpi_index(info->vaddr);

	if (apsh_read_64(instance, ctx, &info->x86_pae.pdpe_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pdpi_location = 0x%.16" PRIx64 "\n",
			    ctx->addr);
		return APSH_F;
	}

	info->x86_pae.pdpe_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pdpe_location = 0x%.16" PRIx64 "\n",
		    info->x86_pae.pdpe_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pdpe_value = 0x%.16" PRIx64 "\n",
		    info->x86_pae.pdpe_value);

	return APSH_S;
}

/* page directory */
static inline uit_32
pgd_index_nopae(uit_32 address)
{
	return (((address) >> 22) & APSH_BIT_MASK(0, 9)) * sizeof(uit_32);
}

static inline uit_32
pgd_index_pae(uit_32 address)
{
	return (((address) >> 21) & APSH_BIT_MASK(0, 8)) * sizeof(uit_64);
}

static inline uit_32
pdba_base_nopae(uit_32 pdpe)
{
	return pdpe & APSH_BIT_MASK(12, 31);
}

static inline uit_64
pdba_base_pae(uit_64 pdpe)
{
	return pdpe & APSH_BIT_MASK(12, 51);
}

static inline status_t
get_pgd_nopae(apsh_instance_t instance, access_context_t *ctx, page_info_t *info)
{
	ctx->addr = pdba_base_nopae(ctx->pt) + pgd_index_nopae(info->vaddr);
	info->x86_legacy.pgd_value = 0;

	if (apsh_read_32(instance, ctx, (uit_32 *)&info->x86_legacy.pgd_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pgd_location at = 0x%.8" PRIx64 "\n",
			    ctx->addr);
		return APSH_F;
	}

	info->x86_legacy.pgd_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pgd_entry = 0x%.8" PRIx64 "\n",
		    info->x86_legacy.pgd_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pgd_entry = 0x%.8" PRIx64 "\n",
		    info->x86_legacy.pgd_value);
	return APSH_S;
}

static inline status_t
get_pgd_pae(apsh_instance_t instance, access_context_t *ctx, page_info_t *info)
{
	ctx->addr = pdba_base_pae(info->x86_pae.pdpe_value) + pgd_index_pae(info->vaddr);

	if (apsh_read_64(instance, ctx, &info->x86_pae.pgd_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pgd_entry = 0x%.8" PRIx64 "\n", ctx->addr);
		return APSH_F;
	}

	info->x86_pae.pgd_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pgd_location = 0x%.8" PRIx64 "\n",
		    info->x86_pae.pgd_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pgd_value = 0x%.8" PRIx64 "\n",
		    info->x86_pae.pgd_value);
	return APSH_S;
}

/* page table */
static inline uit_32
pte_index_nopae(uit_32 address)
{
	return (((address) >> 12) & APSH_BIT_MASK(0, 9)) * sizeof(uit_32);
}

static inline uit_32
pte_index_pae(uit_32 address)
{
	return (((address) >> 12) & APSH_BIT_MASK(0, 8)) * sizeof(uit_64);
}

static inline uit_32
ptba_base_nopae(uit_32 pde)
{
	return pde & APSH_BIT_MASK(12, 31);
}

static inline uit_64
ptba_base_pae(uit_64 pde)
{
	return pde & APSH_BIT_MASK(12, 35);
}

static inline status_t
get_pte_nopae(apsh_instance_t instance, access_context_t *ctx, page_info_t *info)
{
	ctx->addr = ptba_base_nopae(info->x86_legacy.pgd_value) + pte_index_nopae(info->vaddr);
	info->x86_legacy.pte_value = 0;

	if (apsh_read_32(instance, ctx, (uit_32 *)&info->x86_legacy.pte_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pte_entry = 0x%.8" PRIx64 "\n", ctx->addr);
		return APSH_F;
	}

	info->x86_legacy.pte_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pte_location = 0x%.8" PRIx64 "\n",
		    info->x86_legacy.pte_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pte_value = 0x%.8" PRIx64 "\n",
		    info->x86_legacy.pte_value);
	return APSH_S;
}

static inline status_t
get_pte_pae(apsh_instance_t instance, access_context_t *ctx, page_info_t *info)
{
	ctx->addr = ptba_base_pae(info->x86_pae.pgd_value) + pte_index_pae(info->vaddr);

	if (apsh_read_64(instance, ctx, &info->x86_pae.pte_value) == APSH_F) {
		debug_print(APSH_DEBUG_PTLOOKUP,
			    "--PTLookup: failed to read pte_entry = 0x%.8" PRIx64 "\n", ctx->addr);
		return APSH_F;
	}

	info->x86_pae.pte_location = ctx->addr;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pte_location = 0x%.8" PRIx64 "\n",
		    info->x86_pae.pte_location);
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pte_value = 0x%.8" PRIx64 "\n",
		    info->x86_pae.pte_value);
	return APSH_S;
}

/* page */
static inline uit_32
pte_pfn_nopae(uit_32 pte)
{
	return pte & APSH_BIT_MASK(12, 31);
}

static inline uit_64
pte_pfn_pae(uit_64 pte)
{
	return pte & APSH_BIT_MASK(12, 35);
}

static inline uit_32
get_paddr_nopae(uit_32 vaddr, uit_32 pte)
{
	return pte_pfn_nopae(pte) | (vaddr & APSH_BIT_MASK(0, 11));
}

static inline uit_64
get_paddr_pae(uit_32 vaddr, uit_64 pte)
{
	return pte_pfn_pae(pte) | (vaddr & APSH_BIT_MASK(0, 11));
}

static inline uit_32
get_large_paddr_nopae(uit_32 vaddr, uit_32 pgd_entry)
{
	return (pgd_entry & APSH_BIT_MASK(22, 31)) | (vaddr & APSH_BIT_MASK(0, 21));
}

static inline uit_32
get_large_paddr_pae(uit_32 vaddr, uit_32 pgd_entry)
{
	return (pgd_entry & APSH_BIT_MASK(21, 31)) | (vaddr & APSH_BIT_MASK(0, 20));
}

void
buffalo_nopae(apsh_instance_t instance, uit_32 entry, int pde)
{
	if (instance->os_type != APSH_OS_WIN)
		return;

	if (!TRANSITION(entry) && !PROTOTYPE(entry)) {
		uit_32 pfnum = (entry >> 1) & APSH_BIT_MASK(0, 3);
		uit_32 pfframe = entry & APSH_BIT_MASK(12, 31);

		if (pfnum != 0 && pfframe != 0) {
			debug_print(APSH_DEBUG_PTLOOKUP,
				    "--Buffalo: page file = %d, frame = 0x%.8x\n", pfnum, pfframe);
		} else if (pfnum == 0 && pfframe == 0) {
			debug_print(APSH_DEBUG_PTLOOKUP, "--Buffalo: demand zero page\n");
		}
	}

	else if (TRANSITION(entry) && !PROTOTYPE(entry)) {
		debug_print(APSH_DEBUG_PTLOOKUP, "--Buffalo: page in transition\n");
	}

	else if (!pde && PROTOTYPE(entry)) {
		debug_print(APSH_DEBUG_PTLOOKUP, "--Buffalo: prototype entry\n");
	}

	else if (entry == 0) {
		debug_print(APSH_DEBUG_PTLOOKUP, "--Buffalo: entry is zero\n");
	}

	else {
		debug_print(APSH_DEBUG_PTLOOKUP, "--Buffalo: unknown\n");
	}
}

status_t
v2p_nopae(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t pt, addr_t vaddr,
	  page_info_t *info)
{
	status_t status;

	apsh->doing_page_look_up = true;

	ACCESS_CONTEXT(ctx, .npt = npt, .npm = npm);

	info->npt = npt;
	info->npm = npm;
	info->pt = pt;
	info->pm = APSH_PM_LEGACY;
	info->vaddr = vaddr;

	debug_print(APSH_DEBUG_PTLOOKUP,
		    "--PTLookup: lookup vaddr = 0x%.16" PRIx64 ", pt = 0x%.16" PRIx64 "\n", vaddr,
		    pt);
	status = get_pgd_nopae(apsh, &ctx, info);
	if (status != APSH_S)
		goto done;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pgd = 0x%.8" PRIx64 "\n",
		    info->x86_legacy.pgd_value);

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, info->x86_legacy.pgd_value)) {
		status = APSH_F;
		buffalo_nopae(apsh, info->x86_legacy.pgd_value, 0);
		goto done;
	} else {
		status = APSH_S;
	}

	if (PAGE_SIZE(info->x86_legacy.pgd_value) && (apsh->mode == APSH_FILE || apsh->x86.pse)) {
		info->size = APSH_PS_4MB;
		info->paddr = get_large_paddr_nopae(vaddr, info->x86_legacy.pgd_value);
		debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: 4MB page 0x%" PRIx64 "\n",
			    info->x86_legacy.pgd_value);
		status = APSH_S;
		goto done;
	}

	status = get_pte_nopae(apsh, &ctx, info);
	if (status != APSH_S)
		goto done;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pte = 0x%.8" PRIx64 "\n",
		    info->x86_legacy.pte_value);

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, info->x86_legacy.pte_value)) {
		buffalo_nopae(apsh, info->x86_legacy.pte_value, 1);
		status = APSH_F;
		goto done;
	}

	info->size = APSH_PS_4KB;
	info->paddr = get_paddr_nopae(vaddr, info->x86_legacy.pte_value);
	status = APSH_S;

done:
	apsh->doing_page_look_up = false;
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: paddr = 0x%.16" PRIx64 "\n", info->paddr);

	if (valid_npm(npm)) {
		if (apsh_nested_pagetable_lookup(apsh, 0, 0, npt, npm, info->paddr,
							   &info->naddr, NULL) == APSH_F)
			return APSH_F;

		debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: naddr = 0x%.16" PRIx64 "\n",
			    info->naddr);
	}

	return status;
}

status_t
v2p_pae(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t pt, addr_t vaddr,
	page_info_t *info)
{
	status_t status;

	apsh->doing_page_look_up = true;
	ACCESS_CONTEXT(ctx, .npt = npt, .npm = npm);

	info->vaddr = vaddr;
	info->pm = APSH_PM_PAE;
	info->pt = pt;
	info->npt = npt;
	info->npm = npm;

	debug_print(APSH_DEBUG_PTLOOKUP,
		    "--PTLookup: lookup vaddr = 0x%.16" PRIx64 " dtb = 0x%.16" PRIx64 "\n", vaddr,
		    pt);
	status = get_pdpi(apsh, &ctx, info);

	if (status != APSH_S)
		goto done;

	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: pdpe = 0x%" PRIx64 "\n",
		    info->x86_pae.pdpe_value);

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, info->x86_pae.pdpe_value))
		goto done;

	status = get_pgd_pae(apsh, &ctx, info);
	if (status != APSH_S)
		goto done;

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, info->x86_pae.pgd_value)) {
		status = APSH_F;
		goto done;
	}

	if (PAGE_SIZE(info->x86_pae.pgd_value)) {
		info->paddr = get_large_paddr_pae(vaddr, info->x86_pae.pgd_value);
		info->size = APSH_PS_2MB;
		status = APSH_S;
		debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: 2MB page\n");
		goto done;
	}

	status = get_pte_pae(apsh, &ctx, info);
	if (status != APSH_S)
		goto done;

	if (!ENTRY_PRESENT(apsh->x86.transition_pages, info->x86_pae.pte_value)) {
		status = APSH_F;
		goto done;
	}

	info->size = APSH_PS_4KB;
	info->paddr = get_paddr_pae(vaddr, info->x86_pae.pte_value);
	status = APSH_S;

done:
	apsh->doing_page_look_up = false;
	debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: paddr = 0x%.16" PRIx64 "\n", info->paddr);

	if (valid_npm(npm)) {
		if (apsh_nested_pagetable_lookup(apsh, 0, 0, npt, npm, info->paddr,
							   &info->naddr, NULL) == APSH_F)
			return APSH_F;

		debug_print(APSH_DEBUG_PTLOOKUP, "--PTLookup: naddr = 0x%.16" PRIx64 "\n",
			    info->naddr);
	}

	return status;
}

GSList *
get_pages_nopae(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t dtb)
{

	addr_t pgd_location = dtb;
	uint8_t entry_size = 0x4;

	GSList *ret = NULL;
	uit_32 pg = PTRS_PER_NOPAE_PGD;

	uit_32 *pgd_page = g_try_malloc0(APSH_PS_4KB);
	uit_32 *pt_page = g_try_malloc0(entry_size * pg);

	ACCESS_CONTEXT(ctx, .npt = npt, .npm = npm);

	ctx.addr = dtb;
	if (apsh_read(apsh, &ctx, APSH_PS_4KB, pgd_page, NULL) == APSH_F)
		goto done;

	uit_32 i;

	APSH_FUNC_DEFINE();
	for (i = 0; i < pg; i++, pgd_location += entry_size) {
		APSH_FUNC_DEFINE()
		uit_64 pgd_base_vaddr = i * pg * PTRS_PER_NOPAE_PTE * entry_size;

		APSH_FUNC_DEFINE()
		uit_32 pe = pgd_page[i];

		APSH_FUNC_DEFINE()

		APSH_FUNC_DEFINE()
		if (ENTRY_PRESENT(apsh->os_type, pe)) {
			APSH_FUNC_DEFINE()
			if (PAGE_SIZE(pe) && (apsh->mode == APSH_FILE || apsh->x86.pse)) {
				page_info_t *p = g_try_malloc0(sizeof(page_info_t));

				if (!p)
					goto done;
				APSH_FUNC_DEFINE(p->vaddr = pgd_base_vaddr);
				APSH_FUNC_DEFINE(p->pt = dtb);
				APSH_FUNC_DEFINE(
					p->paddr = get_large_paddr_nopae(p->vaddr, pgd_base_vaddr));
				APSH_FUNC_DEFINE(p->size = APSH_PS_4MB);
				APSH_FUNC_DEFINE(p->x86_legacy.pgd_location = pgd_location);
				APSH_FUNC_DEFINE(p->x86_legacy.pgd_value = pe);
				ret = g_slist_prepend(ret, p);
				continue;
			}
			APSH_FUNC_DEFINE()

			uit_32 pte_location = ptba_base_nopae(pe);

			ctx.addr = pte_location;
			if (apsh_read(apsh, &ctx, APSH_PS_4KB, pt_page, NULL) == APSH_F)
				goto done;
			APSH_FUNC_DEFINE()

			uit_32 pte_index;

			for (pte_index = 0; pte_index < PTRS_PER_NOPAE_PTE;
			     pte_index++, pte_location += entry_size) {
				uit_32 pte_entry = pt_page[pte_index];

				if (ENTRY_PRESENT(apsh->os_type, pte_entry)) {
					page_info_t *p = g_try_malloc0(sizeof(page_info_t));

					if (!p)
						goto done;

					APSH_FUNC_DEFINE(p->vaddr = pgd_base_vaddr +
								    pte_index * APSH_PS_4KB);
					APSH_FUNC_DEFINE(p->pt = dtb);
					APSH_FUNC_DEFINE(
						p->paddr = get_paddr_nopae(p->vaddr, pte_entry));
					APSH_FUNC_DEFINE(p->size = APSH_PS_4KB);
					APSH_FUNC_DEFINE(p->x86_legacy.pgd_location = pgd_location);
					APSH_FUNC_DEFINE(p->x86_legacy.pgd_value = pe);
					APSH_FUNC_DEFINE(p->x86_legacy.pte_location = pte_location);
					APSH_FUNC_DEFINE(p->x86_legacy.pte_value = pte_entry);
					ret = g_slist_prepend(ret, p);
				}
			}
		}
	}

done:
	g_free(pt_page);
	g_free(pgd_page);

	return ret;
}

GSList *
get_pages_pae(apsh_instance_t apsh, addr_t npt, page_mode_t npm, addr_t dtb)
{

	uit_32 pdpi_base = get_pdptb(dtb);
	uint8_t entry_size = 0x8;

	GSList *ret = NULL;

	uit_64 pdpi_table[PTRS_PER_PDPI];
	uit_64 *page_directory = NULL;
	uit_64 *page_table = NULL;

	ACCESS_CONTEXT(ctx, .npt = npt, .npm = npm);

	ctx.addr = pdpi_base;
	APSH_FUNC_DEFINE()
	if (apsh_read(apsh, &ctx, sizeof(pdpi_table), pdpi_table, NULL) == APSH_F)
		return ret;

	page_directory = g_malloc(APSH_PS_4KB);
	APSH_FUNC_DEFINE()
	if (!page_directory)
		goto done;

	page_table = g_malloc(APSH_PS_4KB);
	if (!page_table)
		goto done;

	uit_32 pdp_index = 0;
	uit_64 pdpi_location = pdpi_base;

	APSH_FUNC_DEFINE()
	for (pdp_index = 0; pdp_index < PTRS_PER_PDPI; pdp_index++, pdpi_location += entry_size) {

		uit_64 pdp_base_va = pdp_index * PTRS_PER_PAE_PGD * PTRS_PER_PAE_PGD *
				     PTRS_PER_PAE_PTE * entry_size;
		uit_64 pdp_entry = pdpi_table[pdp_index];

		APSH_FUNC_DEFINE()

		if (!ENTRY_PRESENT(apsh->x86.transition_pages, pdp_entry)) {
			continue;
		}

		uit_64 pde_location = pdba_base_pae(pdp_entry);

		APSH_FUNC_DEFINE()

		ctx.addr = pde_location;
		if (apsh_read(apsh, &ctx, APSH_PS_4KB, page_directory, NULL) == APSH_F)
			goto done;

		uit_32 pd_index = 0;

		for (pd_index = 0; pd_index < PTRS_PER_PAE_PGD;
		     pd_index++, pde_location += entry_size) {
			uit_64 pd_base_va = pdp_base_va + (pd_index * PTRS_PER_PAE_PGD *
							   PTRS_PER_PAE_PTE * entry_size);
			APSH_FUNC_DEFINE()

			uit_64 pd_entry = page_directory[pd_index];

			if (ENTRY_PRESENT(apsh->os_type, pd_entry)) {
				APSH_FUNC_DEFINE()

				if (PAGE_SIZE(pd_entry)) {
					page_info_t *new_page_info =
						g_try_malloc0(sizeof(page_info_t));
					if (!new_page_info)
						goto done;
					APSH_FUNC_DEFINE()

					new_page_info->vaddr = pd_base_va;
					new_page_info->pt = dtb;
					new_page_info->paddr =
						get_large_paddr_pae(new_page_info->vaddr, pd_entry);
					new_page_info->size = APSH_PS_2MB;
					new_page_info->x86_pae.pdpe_location = pdpi_location;
					new_page_info->x86_pae.pdpe_value = pdp_entry;
					new_page_info->x86_pae.pgd_location = pde_location;
					new_page_info->x86_pae.pgd_value = pd_entry;
					ret = g_slist_prepend(ret, new_page_info);
					continue;
				}

				uit_64 pte_location = ptba_base_pae(pd_entry);

				ctx.addr = pte_location;
				if (apsh_read(apsh, &ctx, APSH_PS_4KB, page_table, NULL) == APSH_F)
					goto done;

				APSH_FUNC_DEFINE()

				uit_32 pt_index;

				for (pt_index = 0; pt_index < PTRS_PER_PAE_PTE;
				     pt_index++, pte_location += entry_size) {
					uit_64 pte_entry = page_table[pt_index];

					APSH_FUNC_DEFINE()

					if (ENTRY_PRESENT(apsh->os_type, pte_entry)) {
						page_info_t *new_page_info =
							g_try_malloc0(sizeof(page_info_t));
						if (!new_page_info)
							goto done;

						new_page_info->vaddr =
							pd_base_va + pt_index * APSH_PS_4KB;
						new_page_info->pt = dtb;
						new_page_info->paddr = get_paddr_pae(
							new_page_info->vaddr, pte_entry);
						new_page_info->size = APSH_PS_4KB;
						new_page_info->x86_pae.pdpe_location =
							pdpi_location;
						new_page_info->x86_pae.pdpe_value = pdp_entry;
						new_page_info->x86_pae.pgd_location = pde_location;
						new_page_info->x86_pae.pgd_value = pd_entry;
						new_page_info->x86_pae.pte_location = pte_location;
						new_page_info->x86_pae.pte_value = pte_entry;
						ret = g_slist_prepend(ret, new_page_info);
					}
				}
			}
		}
	}

done:
	g_free(page_directory);
	g_free(page_table);

	return ret;
}
