/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <glib.h>
#include <string.h>
#include <time.h>

#include "glib_compat.h"
#include "apsh.h"
#include "apsh_instance.h"

#ifdef __clang_analyzer__
#define g_free free
#endif

#define st_stat() status_t
#define apsh_int64 uint64_t

apsh_int64
hash128to64(apsh_int64 low, apsh_int64 high)
{
	apsh_int64 kMul = 0x9ddfea08eb382d69ULL;
	apsh_int64 a = (low ^ high) * kMul;

	a ^= (a >> 47);
	apsh_int64 b = (high ^ a) * kMul;

	b ^= (b >> 47);
	b *= kMul;
	return b;
}

guint
key_128_hash(gconstpointer key)
{
	const key_128_t cache_key = (const key_128_t)key;
	const apsh_int64 hash64 = hash128to64(cache_key->low, cache_key->high);

	return g_int64_hash(&hash64);
}

gboolean
key_128_equals(gconstpointer key1, gconstpointer key2)
{
	const key_128_t cache_key1 = (const key_128_t)key1;
	const key_128_t cache_key2 = (const key_128_t)key2;

	return cache_key1->low == cache_key2->low && cache_key1->high == cache_key2->high;
}

static inline void key_128_init(key_128_t key, apsh_int64 low, apsh_int64 high);

static inline key_128_t key_128_build(apsh_int64 low, apsh_int64 high);

struct pid_cache_entry {
	apsh_pid_t pid;
	addr_t dtb;
};
typedef struct pid_cache_entry *pid_cache_entry_t;

static void pid_cache_key_free(gpointer data);

static void pid_cache_entry_free(gpointer data);

static pid_cache_entry_t pid_cache_entry_create(apsh_pid_t pid, addr_t dtb);

void
pid_cache_init(apsh_instance_t apsh)
{
	apsh->pid_cache = g_hash_table_new_full(g_int_hash, g_int_equal, pid_cache_key_free,
					       pid_cache_entry_free);
}

void
pid_cache_destroy(apsh_instance_t apsh)
{
	if (apsh->pid_cache)
		g_hash_table_destroy(apsh->pid_cache);
}

st_stat()
pid_cache_get(apsh_instance_t apsh, apsh_pid_t pid, addr_t *dtb)
{
	pid_cache_entry_t table_entry_line = NULL;
	gint key = (gint)pid;

	table_entry_line = g_hash_table_lookup(apsh->pid_cache, &key);
	if (table_entry_line != NULL) {
		*dtb = table_entry_line->dtb;
		debug_print(APSH_DEBUG_PIDCACHE, "--PID cache hit %d -- 0x%.16" PRIx64 "\n", pid,
			    *dtb);
		return APSH_S;
	}

	return APSH_F;
}

void
pid_cache_set(apsh_instance_t apsh, apsh_pid_t pid, addr_t dtb)
{
	pid_cache_entry_t table_entry_line = NULL;
	gint *key = (gint *)g_malloc(sizeof(gint));

	if (!key)
		goto cleanup;

	*key = pid;

	table_entry_line = pid_cache_entry_create(pid, dtb);
	if (!table_entry_line)
		goto cleanup;

	(void)g_hash_table_insert_compat(apsh->pid_cache, key, table_entry_line);
	debug_print(APSH_DEBUG_PIDCACHE, "--PID cache set %d -- 0x%.16" PRIx64 "\n", pid, dtb);
	return;

cleanup:
	g_free(key);
	g_free(table_entry_line);
}

st_stat()
pid_cache_del(apsh_instance_t apsh, apsh_pid_t pid)
{
	gint key = (gint)pid;

	debug_print(APSH_DEBUG_PIDCACHE, "--PID cache del %d\n", pid);
	if (g_hash_table_remove(apsh->pid_cache, &key) == TRUE)
		return APSH_S;
	else
		return APSH_F;
}

void
pid_cache_flush(apsh_instance_t apsh)
{
	g_hash_table_remove_all(apsh->pid_cache);
	debug_print(APSH_DEBUG_PIDCACHE, "--PID cache flushed\n");
}

struct sym_cache_entry {
	char *sym;
	addr_t va;
	addr_t base_addr;
	apsh_pid_t pid;
};
typedef struct sym_cache_entry *sym_cache_entry_t;

static void
sym_cache_entry_free(gpointer data)
{
	sym_cache_entry_t table_entry_line = (sym_cache_entry_t)data;

	if (table_entry_line) {
		g_free(table_entry_line->sym);
		g_free(table_entry_line);
	}
}

static sym_cache_entry_t
sym_cache_entry_create(const char *sym, addr_t va, addr_t base_addr, apsh_pid_t pid)
{
	sym_cache_entry_t table_entry_line = (sym_cache_entry_t)g_malloc(sizeof(struct sym_cache_entry));

	if (!table_entry_line)
		goto cleanup;

	table_entry_line->sym = g_strdup(sym);
	if (!table_entry_line->sym)
		goto cleanup;

	table_entry_line->va = va;
	table_entry_line->base_addr = base_addr;
	table_entry_line->pid = pid;
	return table_entry_line;

cleanup:
	if (table_entry_line) {
		g_free(table_entry_line->sym);
		g_free(table_entry_line);
	}
	return NULL;
}

void
sym_cache_init(apsh_instance_t apsh)
{
	apsh->sym_cache = g_hash_table_new_full((GHashFunc)key_128_hash, key_128_equals, g_free,
					       (GDestroyNotify)g_hash_table_destroy);
}

void
sym_cache_destroy(apsh_instance_t apsh)
{
	if (apsh->sym_cache)
		g_hash_table_destroy(apsh->sym_cache);
}

st_stat()
sym_cache_get(apsh_instance_t apsh, addr_t base_addr, apsh_pid_t pid, const char *sym, addr_t *va)
{

	st_stat() ret = APSH_F;

	GHashTable *sym_apsh_tab = NULL;
	sym_cache_entry_t table_entry_line = NULL;

	struct key_128 local_key;
	key_128_t key = &local_key;

	key_128_init(key, (apsh_int64)base_addr, (apsh_int64)pid);

	sym_apsh_tab = g_hash_table_lookup(apsh->sym_cache, key);
	if (sym_apsh_tab == NULL)
		return ret;

	table_entry_line = g_hash_table_lookup(sym_apsh_tab, sym);
	if (table_entry_line != NULL) {
		*va = table_entry_line->va;
		debug_print(APSH_DEBUG_SYMCACHE,
			    "--SYM cache hit %u:0x%.16" PRIx64 ":%s -- 0x%.16" PRIx64 "\n", pid,
			    base_addr, sym, *va);
		ret = APSH_S;
	}

	return ret;
}

void
sym_cache_set(apsh_instance_t apsh, addr_t base_addr, apsh_pid_t pid, const char *sym, addr_t va)
{
	GHashTable *sym_apsh_tab = NULL;
	char *sym_dup = NULL;
	gboolean new_symbol_table = FALSE;
	sym_cache_entry_t table_entry_line = NULL;

	key_128_t key = key_128_build((apsh_int64)base_addr, (apsh_int64)pid);

	if (!key)
		goto cleanup;

	sym_apsh_tab = g_hash_table_lookup(apsh->sym_cache, key);
	if (!sym_apsh_tab) {
		sym_apsh_tab =
		    g_hash_table_new_full(g_str_hash, g_str_equal, g_free, sym_cache_entry_free);
		if (!sym_apsh_tab)
			goto cleanup;

		new_symbol_table = TRUE;

		(void)g_hash_table_insert_compat(apsh->sym_cache, key, sym_apsh_tab);
	} else {
		g_free(key);
		key = NULL;
	}

	table_entry_line = sym_cache_entry_create(sym, va, base_addr, pid);
	if (!table_entry_line)
		goto cleanup;

	sym_dup = g_strndup(sym, 100);
	if (!sym_dup)
		goto cleanup;

	(void)g_hash_table_insert_compat(sym_apsh_tab, sym_dup, table_entry_line);
	debug_print(APSH_DEBUG_SYMCACHE, "--SYM cache set %s -- 0x%.16" PRIx64 "\n", sym, va);
	return;

cleanup:
	g_free(sym_dup);
	g_free(table_entry_line);

	if (new_symbol_table) {
		// destroys key and value
		g_hash_table_remove(apsh->sym_cache, key);
		key = NULL;
		sym_apsh_tab = NULL;
	}

	g_free(key);
}

st_stat()
sym_cache_del(apsh_instance_t apsh, addr_t base_addr, apsh_pid_t pid, char *sym)
{
	st_stat() ret = APSH_F;
	GHashTable *sym_apsh_tab = NULL;
	struct key_128 local_key;
	key_128_t key = &local_key;

	key_128_init(key, (apsh_int64)base_addr, (apsh_int64)pid);

	sym_apsh_tab = g_hash_table_lookup(apsh->sym_cache, key);
	if (sym_apsh_tab == NULL)
		return ret;

	debug_print(APSH_DEBUG_SYMCACHE, "--SYM cache del %u:0x%.16" PRIx64 ":%s\n", pid, base_addr,
		    sym);

	if (g_hash_table_remove(sym_apsh_tab, sym) == TRUE) {
		ret = APSH_S;

		if (!g_hash_table_size(sym_apsh_tab))
			g_hash_table_remove(apsh->sym_cache, key);
	}

	return ret;
}

void
sym_cache_flush(apsh_instance_t apsh)
{
	g_hash_table_remove_all(apsh->sym_cache);
	debug_print(APSH_DEBUG_SYMCACHE, "--SYM cache flushed\n");
}

void
rva_cache_init(apsh_instance_t apsh)
{
	apsh->rva_cache = g_hash_table_new_full((GHashFunc)key_128_hash, key_128_equals, g_free,
					       (GDestroyNotify)g_hash_table_destroy);
}

void
rva_cache_destroy(apsh_instance_t apsh)
{
	if (apsh->rva_cache)
		g_hash_table_destroy(apsh->rva_cache);
}

st_stat()
rva_cache_get(apsh_instance_t apsh, addr_t base_addr, addr_t dtb, addr_t rva, char **sym)
{
	st_stat() ret = APSH_F;

	GHashTable *apsh_rva_hash_table = NULL;
	sym_cache_entry_t table_entry_line = NULL;

	struct key_128 local_key;
	key_128_t key = &local_key;

	key_128_init(key, (apsh_int64)base_addr, (apsh_int64)dtb);

	apsh_rva_hash_table = g_hash_table_lookup(apsh->rva_cache, key);
	if (apsh_rva_hash_table == NULL)
		return ret;

	table_entry_line = g_hash_table_lookup(apsh_rva_hash_table, GUINT_TO_POINTER(rva));
	if (table_entry_line != NULL) {
		*sym = table_entry_line->sym;
		debug_print(APSH_DEBUG_RVACACHE,
			    "--RVA cache hit 0x%.16" PRIx64 ":0x%.16" PRIx64 ":%s -- 0x%.16" PRIx64
			    "\n",
			    dtb, base_addr, *sym, rva);
		ret = APSH_S;
	}

	return ret;
}

void
rva_cache_set(apsh_instance_t apsh, addr_t base_addr, addr_t dtb, addr_t rva, char *sym)
{
	GHashTable *apsh_rva_hash_table = NULL;
	sym_cache_entry_t table_entry_line = NULL;

	key_128_t key = key_128_build((apsh_int64)base_addr, (apsh_int64)dtb);

	if (!key)
		goto cleanup;

	table_entry_line = sym_cache_entry_create(sym, rva, base_addr, dtb);
	if (!table_entry_line)
		goto cleanup;

	apsh_rva_hash_table = g_hash_table_lookup(apsh->rva_cache, key);
	if (apsh_rva_hash_table == NULL) {
		apsh_rva_hash_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL,
						  sym_cache_entry_free);
		if (!apsh_rva_hash_table)
			goto cleanup;

		(void)g_hash_table_insert_compat(apsh->rva_cache, GUINT_TO_POINTER(key), apsh_rva_hash_table);
	} else {
		g_free(key);
	}

	(void)g_hash_table_insert_compat(apsh_rva_hash_table, GUINT_TO_POINTER(rva), table_entry_line);
	debug_print(APSH_DEBUG_RVACACHE, "--RVA cache set %s -- 0x%.16" PRIx64 "\n", sym, rva);
	return;

cleanup:
	g_free(table_entry_line);
	g_free(key);
}

st_stat()
rva_cache_del(apsh_instance_t apsh, addr_t base_addr, addr_t dtb, addr_t rva)
{
	st_stat() ret = APSH_F;
	GHashTable *apsh_rva_hash_table = NULL;
	struct key_128 local_key;
	key_128_t key = &local_key;

	key_128_init(key, (apsh_int64)base_addr, (apsh_int64)dtb);

	apsh_rva_hash_table = g_hash_table_lookup(apsh->rva_cache, key);
	if (apsh_rva_hash_table == NULL)
		return ret;

	debug_print(APSH_DEBUG_RVACACHE,
		    "--RVA cache del 0x%.16" PRIx64 ":0x%.16" PRIx64 ":0x%.16" PRIx64 "\n", dtb,
		    base_addr, rva);

	if (g_hash_table_remove(apsh_rva_hash_table, GUINT_TO_POINTER(rva)) == TRUE) {
		ret = APSH_S;

		if (!g_hash_table_size(apsh_rva_hash_table))
			g_hash_table_remove(apsh->rva_cache, key);
	}

	return ret;
}

void
rva_cache_flush(apsh_instance_t apsh)
{
	g_hash_table_remove_all(apsh->rva_cache);
	debug_print(APSH_DEBUG_RVACACHE, "--RVA cache flushed\n");
}

void
v2p_cache_init(apsh_instance_t apsh)
{
	apsh->v2p_cache = g_hash_table_new_full((GHashFunc)key_128_hash, key_128_equals, g_free,
					       (GDestroyNotify)g_hash_table_destroy);
}

void
v2p_cache_destroy(apsh_instance_t apsh)
{
	if (apsh->v2p_cache)
		g_hash_table_destroy(apsh->v2p_cache);
}

st_stat()
v2p_cache_get(apsh_instance_t apsh, addr_t va, addr_t pt, addr_t npt, addr_t *pa)
{
	struct key_128 local_key;
	key_128_t key = &local_key;

	key_128_init(key, pt, npt);

	GHashTable *v = g_hash_table_lookup(apsh->v2p_cache, key);

	if (!v) {
		debug_print(APSH_DEBUG_V2PCACHE,
			    "--V2P cache miss (no address space) 0x%.16" PRIx64 " 0x%.16" PRIx64
			    "\n",
			    pt, npt);
		return APSH_F;
	}

	addr_t offset = APSH_BIT_MASK(0, 11) & va;

	va = (va >> 12) << 12;

	gpointer _pa = g_hash_table_lookup(v, &va);

	if (!_pa) {
		debug_print(APSH_DEBUG_V2PCACHE, "--V2P cache miss (no page) 0x%.16" PRIx64 "\n",
			    va);
		return APSH_F;
	}

	*pa = GPOINTER_TO_SIZE(_pa);
	*pa |= offset;
	debug_print(APSH_DEBUG_V2PCACHE, "--V2P cache hit 0x%.16" PRIx64 " -- 0x%.16" PRIx64 "\n",
		    va | offset, *pa);

	return APSH_S;
}

void
v2p_cache_set(apsh_instance_t apsh, addr_t va, addr_t pt, addr_t npt, addr_t pa)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!va || !pt || !pa)
		return;
#endif

	key_128_t key = key_128_build(pt, npt);

	if (!key)
		return;

	GHashTable *v = g_hash_table_lookup(apsh->v2p_cache, key);
	gboolean new_process_space = FALSE;
	addr_t *_va = NULL;

	if (!v) {
		new_process_space = TRUE;

		v = g_hash_table_new_full(g_int64_hash, g_int64_equal, g_free, NULL);
		if (!v)
			goto cleanup;

		(void)g_hash_table_insert_compat(apsh->v2p_cache, key, v);
	} else {
		g_free(key);
		key = NULL;
	}

	/* bundle the cache entries per page */
	va = (va >> 12) << 12;
	pa = (pa >> 12) << 12;

	_va = g_memdup(&va, sizeof(addr_t));
	if (!_va)
		goto cleanup;

	(void)g_hash_table_insert_compat(v, _va, GSIZE_TO_POINTER(pa));

	debug_print(APSH_DEBUG_V2PCACHE,
		    "--V2P cache set for page 0x%.16" PRIx64 " -- 0x%.16" PRIx64 "\n", va, pa);
	return;

cleanup:
	if (new_process_space)
		g_hash_table_remove(apsh->v2p_cache, key);
	g_free(key);
}

st_stat()
v2p_cache_del(apsh_instance_t apsh, addr_t va, addr_t pt, addr_t npt)
{
	struct key_128 local_key;
	key_128_t key = &local_key;

	key_128_init(key, npt, pt);

	GHashTable *v = g_hash_table_lookup(apsh->v2p_cache, key);

	if (!v)
		return APSH_S;

	va = (va >> 12) << 12;
	(void)g_hash_table_remove(v, &va);

	if (!g_hash_table_size(v))
		g_hash_table_remove(apsh->v2p_cache, key);

	debug_print(APSH_DEBUG_V2PCACHE, "--V2P cache del 0x%.16" PRIx64 "\n", va);

	return APSH_S;
}

void
v2p_cache_flush(apsh_instance_t apsh, addr_t pt, addr_t npt)
{
	if (~0ull == pt)
		g_hash_table_remove_all(apsh->v2p_cache);
	else {
		struct key_128 local_key;
		key_128_t key = &local_key;

		key_128_init(key, npt, pt);
		(void)g_hash_table_remove(apsh->v2p_cache, key);
	}
	debug_print(APSH_DEBUG_V2PCACHE, "--V2P cache flushed\n");
}

static inline void
key_128_init(key_128_t key, apsh_int64 low, apsh_int64 high)
{
	key->low = low;
	key->high = high;
}

static inline key_128_t
key_128_build(apsh_int64 low, apsh_int64 high)
{
	key_128_t key = (key_128_t)g_malloc(sizeof(struct key_128));

	if (key)
		key_128_init(key, low, high);
	return key;
}

static void
pid_cache_key_free(gpointer data)
{
	g_free(data);
}

static void
pid_cache_entry_free(gpointer data)
{
	pid_cache_entry_t table_entry_line = (pid_cache_entry_t)data;

	g_free(table_entry_line);
}

static pid_cache_entry_t
pid_cache_entry_create(apsh_pid_t pid, addr_t dtb)
{
	pid_cache_entry_t table_entry_line = (pid_cache_entry_t)g_malloc(sizeof(struct pid_cache_entry));

	if (!table_entry_line)
		return NULL;

	table_entry_line->pid = pid;
	table_entry_line->dtb = dtb;
	return table_entry_line;
}
