/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <glib.h>
#include <iconv.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "apsh_instance.h"

#define defined_char char

#ifdef ENABLE_BAREFLANK
#define PRINT_FILE_DESC stdout
#else
#define PRINT_FILE_DESC stderr
#endif

#ifdef ENV_DEBUG
#define ENV_DEBUG_COND getenv("APSH_DEBUG")
#else
#define ENV_DEBUG_COND true
#endif

#ifndef APSH_DEBUG
/* Nothing */
#else
void
debug_print(apsh_debug_flag_t category, defined_char *format, ...)
{
	if ((category & APSH_DEBUG) && ENV_DEBUG_COND) {
		va_list args;

		va_start(args, format);
		vfprintf(stderr, format, args);
		va_end(args);
	}
}

void
print_error(defined_char *format, ...)
{
	va_list args;

	fprintf(PRINT_FILE_DESC, "APSH_ERROR: ");
	va_start(args, format);
	vfprintf(PRINT_FILE_DESC, format, args);
	va_end(args);
}

void
print_warning(defined_char *format, ...)
{
	va_list args;

	fprintf(PRINT_FILE_DESC, "APSH_WARNING: ");
	va_start(args, format);
	vfprintf(PRINT_FILE_DESC, format, args);
	va_end(args);
}
#endif

void *safe_malloc_(size_t size, defined_char const *file, int line);

unsigned long
get_reg32(reg_t r)
{
	return (unsigned long)r;
}

addr_t
aligned_addr(apsh_instance_t apsh, addr_t addr)
{
	addr_t mask = ~((addr_t)apsh->page_size - 1);
	addr_t aligned = (addr_t)addr & (addr_t)mask;

	return aligned;
}

int
is_addr_aligned(apsh_instance_t apsh, addr_t addr)
{
	return (addr == aligned_addr(apsh, addr));
}

status_t
apsh_convert_str_encoding(const unicode_string_t *string_in, unicode_string_t *string_out,
			 const defined_char *outencoding)
{
	iconv_t cd = 0;
	size_t iconv_val = 0;
	size_t inlen;
	size_t outlen;
	defined_char *incurr;

	if (!string_in || !string_out)
		return APSH_F;

	inlen = string_in->length;
	outlen = 2 * (inlen + 1);

	incurr = (defined_char *)string_in->contents;

	memset(string_out, 0, sizeof(*string_out));
	string_out->contents = safe_malloc(outlen);
	memset(string_out->contents, 0, outlen);

	defined_char *outstart = (defined_char *)string_out->contents;
	defined_char *outcurr = (defined_char *)string_out->contents;

	string_out->encoding = outencoding;

	cd = iconv_open(string_out->encoding, string_in->encoding);
	if (cd == (iconv_t)(-1)) {
		if (errno == EINVAL) {
			debug_print(APSH_DEBUG_READ,
				    "%s: conversion from '%s' to '%s' not supported\n",
				    __FUNCTION__, string_in->encoding, string_out->encoding);
		} else {
			debug_print(APSH_DEBUG_READ, "%s: Initializiation failure: %s\n",
				    __FUNCTION__, strerror(errno));
		}
		goto fail;
	}

	iconv_val = iconv(cd, &incurr, &inlen, &outcurr, &outlen);
	if ((size_t)-1 == iconv_val) {
		debug_print(APSH_DEBUG_READ,
			    "%s: iconv failed, in string '%s' length %zu, "
			    "out string '%s' length %zu\n",
			    __FUNCTION__, string_in->contents, string_in->length,
			    string_out->contents, outlen);
		switch (errno) {
		case EILSEQ:
			debug_print(APSH_DEBUG_READ, "invalid multibyte sequence");
			break;
		case EINVAL:
			debug_print(APSH_DEBUG_READ, "incomplete multibyte sequence");
			break;
		case E2BIG:
			debug_print(APSH_DEBUG_READ, "no more room");
			break;
		default:
			debug_print(APSH_DEBUG_READ, "error: %s\n", strerror(errno));
			break;
		}
		goto fail;
	}

	string_out->length = (size_t)(outcurr - outstart);
	(void)iconv_close(cd);
	return APSH_S;

fail:
	if (string_out->contents)
		free(string_out->contents);

	memset(string_out, 0, sizeof(*string_out));

	if (cd != (iconv_t)(-1))
		(void)iconv_close(cd);

	return APSH_F;
}

void apsh_free_unicode_str(unicode_string_t *p_us);

void free_gint(gpointer p);

void
free_gint64(gpointer p)
{
	g_slice_free(gint64, p);
}

void *
safe_malloc_(size_t size, defined_char const *file, int line)
{
	void *p = malloc(size);

	if (p == NULL) {
		print_error("malloc %lu bytes failed at %s:%d\n", (unsigned long)size, file, line);
		exit(EXIT_FAILURE);
	}
	return p;
}

void
apsh_free_unicode_str(unicode_string_t *p_us)
{
	if (!p_us)
		return;

	if (p_us->contents)
		free(p_us->contents);
	memset((void *)p_us, 0, sizeof(*p_us));
	free(p_us);
}

void
free_gint(gpointer p)
{
	g_slice_free(gint, p);
}
