/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <fnmatch.h>
#include <limits.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>

#include "driver/driver_wrapper.h"
#include "os/linux/linux.h"
#include "os/os_interface.h"
#include "os/windows/windows.h"
#include "plugin_context.h"
#include "apsh_instance.h"

#define st_stat() status_t
#define APSH_FUNC_DEFINE(x) x

#ifndef ENABLE_CONFIGFILEdasdsa
static inline st_stat()
	read_config_string(apsh_instance_t UNUSED(apsh), const char *UNUSED(config),
			   GHashTable **UNUSED(_config), apsh_init_error_t *UNUSED(error))
{
	return APSH_F;
}

static inline st_stat()
	read_config_file_entry(apsh_instance_t UNUSED(apsh), GHashTable **UNUSED(config),
			       apsh_init_error_t *UNUSED(error))
{
	return APSH_F;
}
#else

#include "config/config_parser.h"

extern FILE *yyin;

static FILE *
open_config_file()
{
	FILE *f = NULL;
	gchar *location;
	char *sudo_user = NULL;
	struct passwd *pw_entry = NULL;
	char cwd[1024] = {0};

	if (getcwd(cwd, sizeof(cwd)) != NULL) {
		location = g_strconcat(cwd, "/APSH.conf", NULL);
		debug_print(APSH_DEBUG_CORE, "--looking for config file at %s\n", location);

		f = fopen(location, "r");
		g_free(location);

		if (f)
			return f;
	}

	sudo_user = getenv("SUDO_USER");
	if (sudo_user != NULL) {
		pw_entry = getpwnam(sudo_user);
		if (pw_entry != NULL) {
			location = g_strconcat(pw_entry->pw_dir, "/etc/APSH.conf", NULL);
			debug_print(APSH_DEBUG_CORE, "--looking for config file at %s\n", location);

			f = fopen(location, "r");
			g_free(location);

			if (f)
				return f;
		}
	}

	location = g_strconcat(getenv("HOME"), "/etc/APSH.conf", NULL);
	debug_print(APSH_DEBUG_CORE, "--looking for config file at %s\n", location);

	f = fopen(location, "r");
	g_free(location);

	if (f)
		return f;

	debug_print(APSH_DEBUG_CORE, "--looking for config file at /etc/APSH.conf\n");
	f = fopen("/etc/APSH.conf", "r");

	if (f)
		return f;

	return NULL;
}

static st_stat() read_config_file(apsh_instance_t apsh, FILE *config_file, GHashTable **config,
				  apsh_init_error_t *error)
{
	st_stat() ret = APSH_F;

	yyin = config_file;

	if (apsh_parse_config(apsh->image_type) != 0) {
		if (error)
			*error = APSH_INIT_ERROR_NO_CONFIG;

		print_error("Failed to read config file.\n");
		goto error_exit;
	}

	*config = apsh_get_config();

	if (*config == NULL) {
		if (error)
			*error = APSH_INIT_ERROR_NO_CONFIG_ENTRY;

		print_error("No entry in config file for %s.\n", apsh->image_type);
		goto error_exit;
	}

	ret = APSH_S;

error_exit:
	if (config_file)
		fclose(config_file);

	return ret;
}

st_stat() read_config_string(apsh_instance_t apsh, const char *config, GHashTable **_config,
			     apsh_init_error_t *error)
{
	st_stat() ret = APSH_S;
	FILE *config_file = NULL;

	if (config == NULL) {
		if (error)
			*error = APSH_INIT_ERROR_NO_CONFIG;

		print_error("APSH_ERROR: NULL string passed for APSH_CONFIG_STRING\n");
		return APSH_F;
	}

	APSH_FUNC_DEFINE(gchar *config_str = g_strconcat(apsh->image_type, " ", config, NULL));

	config_file = fmemopen(APSH_FUNC_DEFINE(config_str), strlen(config_str) + 1, "r");
	ret = read_config_file(apsh, config_file, _config, error);

	APSH_FUNC_DEFINE(g_free(config_str));

	return ret;
}

static st_stat() read_config_file_entry(apsh_instance_t apsh, GHashTable **APSH_FUNC_DEFINE(config),
					apsh_init_error_t *error)
{
	FILE *config_file = open_config_file();

	if (config_file == NULL) {
		if (error)
			APSH_FUNC_DEFINE(*error = APSH_INIT_ERROR_NO_CONFIG);

		APSH_FUNC_DEFINE(fprintf(stderr, "ERROR: config file not found.\n"));
		return APSH_F;
	}

	return read_config_file(apsh, config_file, config, error);
}

#endif

st_stat() set_os_type_from_config(apsh_instance_t apsh, GHashTable *configtbl)
{
	st_stat() ret = APSH_F;
	const char *ostype = NULL;

	APSH_FUNC_DEFINE(apsh->os_type = APSH_OS_UNKNOWN);
	if (apsh->os_data) {
		APSH_FUNC_DEFINE(free(apsh->os_data));
		APSH_FUNC_DEFINE(apsh->os_data = NULL);
	}

#ifdef ENABLE_JSON_PROFILES
	const char *json_path = g_hash_table_lookup(configtbl, "volatility_ist");

	if (!json_path)
		APSH_FUNC_DEFINE(json_path = g_hash_table_lookup(configtbl, "rekall_profile"));

	if (json_path && json_profile_init(apsh, json_path))
		APSH_FUNC_DEFINE(ostype = apsh->json.get_os_type(apsh));
#endif

	if (!ostype) {
		ostype = g_hash_table_lookup(configtbl, "ostype");
		if (ostype == NULL)
			ostype = g_hash_table_lookup(configtbl, "os_type");
	}

	if (!ostype) {
		print_error("Undefined OS type!\n");
		return APSH_F;
	}

	if (!strcmp(ostype, "Linux")) {
		apsh->os_type = APSH_OS_LINUX;
		ret = APSH_S;
	} else if (!strcmp(ostype, "Windows")) {
		apsh->os_type = APSH_OS_WIN;
		ret = APSH_S;
	} else {
		print_error("APSH_ERROR: Unknown OS type: %s!\n", ostype);
		ret = APSH_F;
	}

#ifdef APSH_D
	if (apsh->os_type == APSH_OS_LINUX)
		debug_print(APSH_DEBUG_CORE, "**set os_type to Linux.\n");
	else if (apsh->os_type == APSH_OS_WIN)
		debug_print(APSH_DEBUG_CORE, "**set os_type to Windows.\n");
	else
		debug_print(APSH_DEBUG_CORE, "**set os_type to unknown.\n");
#endif

	return ret;
}

static inline st_stat() init_page_offset(apsh_instance_t apsh)
{
	apsh->page_shift = 12;
	apsh->page_size = APSH_PS_4KB;

	return APSH_S;
}

static inline void
set_image_type_for_file(apsh_instance_t apsh, const char *name)
{
	const char *ptr = NULL;

	ptr = strrchr(name, '/');
	if (ptr == NULL)
		ptr = name;
	else
		ptr++;

	apsh->image_type = strndup(ptr, 500);
	apsh->image_type_complete = strndup(name, 500);
}

static st_stat() set_id_and_name(apsh_instance_t apsh, const void *domain)
{
	const char *name = NULL;
	uint64_t id = 0;
	const char *uuid = 0;
	unsigned int init_flags_count = 0;

	if (apsh->init_flags & APSH_INIT_DOMAINNAME) {
		name = (const char *)domain;
		init_flags_count++;
	}

	if (init_flags_count != 1) {
		print_error("Specifying none or more than one init params is not valid!\n");
		return APSH_F;
	}

	if (apsh->mode == APSH_FILE) {
		if (name) {
			set_image_type_for_file(apsh, name);
			driver_set_name(apsh, name);
			goto done;
		}

		print_error("Must specify name for file mode.\n");
		return APSH_F;
	}

	if (apsh->mode == APSH_XEN) {
		char *idstring = NULL;

		if (asprintf(&idstring, "domid-%" PRIu64, id) > 0) {
			apsh->image_type = idstring;
			goto done;
		}
	}

	return APSH_F;

done:
	debug_print(APSH_DEBUG_CORE, "**set image_type = %s\n", apsh->image_type);
	return APSH_S;
}

page_mode_t
apsh_get_page_mode(apsh_instance_t apsh, unsigned long vcpu)
{
	page_mode_t pm = APSH_PM_UNKNOWN;

	if (!apsh)
		return pm;

	if (apsh->mode == APSH_FILE)
		return apsh->page_mode;

	return pm;
}

st_stat() apsh_get_access_mode(apsh_instance_t apsh, const void *domain, uint64_t init_flags,
			       apsh_init_data_t *init_data, apsh_mode_t *mode)
{
	if (apsh) {
		*mode = apsh->mode;
		return APSH_S;
	}

	const char *name = NULL;
	uint64_t id = APSH_INVALID_DOMID;

	if (init_flags & APSH_INIT_DOMAINNAME)
		name = (const char *)domain;

	return driver_init_mode(name, id, init_flags, init_data, mode);
}

static inline st_stat() driver_sanity_check(apsh_mode_t mode)
{
	switch (mode) {
	case APSH_XEN:
#ifndef ENABLE_XEN
		return APSH_F;
#else
		break;
#endif
	case APSH_KVM:
#ifndef ENABLE_KVM
		return APSH_F;
#else
		break;
#endif
	case APSH_FILE:
#ifndef ENABLE_FILE
		return APSH_F;
#else
		break;
#endif
	case APSH_BAREFLANK:
#ifndef ENABLE_BAREFLANK
		return APSH_F;
#else
		break;
#endif
	default:
		return APSH_F;
	};

	return APSH_S;
}

st_stat() apsh_init(apsh_instance_t *apsh, apsh_mode_t mode, const void *domain,
		    uint64_t init_flags, apsh_init_data_t *init_data, addr_t vf_id,
		    const char *mem_region_json, const char *kpgd_file, const char *mlx_device,
		    struct doca_dev *regex_dev, struct doca_dev_rep *system_dev, apsh_init_error_t *error)
{
	if (driver_sanity_check(mode) == APSH_F) {
		print_error("The selected LibAPSH mode is not available!\n");
		return APSH_F;
	}

	st_stat() status = APSH_F;

	apsh_instance_t _apsh = (apsh_instance_t)g_try_malloc0(sizeof(struct apsh_instance));

	if (!_apsh)
		return APSH_F;

	debug_print(APSH_DEBUG_CORE, "LibAPSH Version %s\n", PACKAGE_VERSION);

	_apsh->mode = mode;
	debug_print(APSH_DEBUG_CORE, "LibAPSH Driver Mode %d\n", _apsh->mode);

	_apsh->init_flags = init_flags;
	_apsh->page_mode = APSH_PM_UNKNOWN;
	_apsh->vf_id = vf_id;
	_apsh->mem_region_json = mem_region_json;
	_apsh->kpgd_file = kpgd_file;
	_apsh->mlx_device = mlx_device;
	_apsh->regex_dev = regex_dev;
	_apsh->system_dev = system_dev;

	_apsh->plugin_ctx = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, plugin_ctx_destroy);
	if (_apsh->plugin_ctx == NULL)
		goto error_exit;

	arch_init_lookup_tables(_apsh);

	if (init_data && init_data->count) {
		uint64_t i;

		for (i = 0; i < init_data->count; i++) {
			switch (init_data->entry[i].type) {
			case APSH_INIT_DATA_MEMMAP:
				_apsh->memmap = (memory_map_t *)g_memdup(init_data->entry[i].data,
									 sizeof(memory_map_t));
				if (!_apsh->memmap)
					goto error_exit;
				break;
			default:
				break;
			};
		}
	}

	if (init_page_offset(_apsh) == APSH_F) {
		if (error)
			*error = APSH_INIT_ERROR_DRIVER;

		goto error_exit;
	}
	if (driver_init(_apsh, init_flags, init_data) == APSH_F) {
		if (error)
			*error = APSH_INIT_ERROR_DRIVER;

		goto error_exit;
	}
	debug_print(APSH_DEBUG_CORE, "--completed driver init.\n");

	if (set_id_and_name(_apsh, domain) == APSH_F) {
		if (error)
			*error = APSH_INIT_ERROR_VM_NOT_FOUND;

		goto error_exit;
	}

	if (driver_init_apsh(_apsh, init_flags, init_data) == APSH_F) {
		if (error)
			*error = APSH_INIT_ERROR_DRIVER;

		goto error_exit;
	}

	debug_print(APSH_DEBUG_CORE, "--completed APSH init.\n");

	if (driver_get_memsize(_apsh, &_apsh->allocated_ram_size, &_apsh->max_physical_address) ==
	    APSH_F) {
		if (error)
			*error = APSH_INIT_ERROR_DRIVER;

		goto error_exit;
	}

	pid_cache_init(_apsh);
	sym_cache_init(_apsh);
	rva_cache_init(_apsh);
	v2p_cache_init(_apsh);

	status = APSH_S;

	debug_print(APSH_DEBUG_CORE,
		    "**set allocated_ram_size = %" PRIx64 ", "
		    "max_physical_address = 0x%" PRIx64 "\n",
		    _apsh->allocated_ram_size, _apsh->max_physical_address);

error_exit:
	if (status == APSH_F) {
		apsh_destroy(_apsh);
		*apsh = NULL;
	} else
		*apsh = _apsh;

	return status;
}

page_mode_t
apsh_init_paging(apsh_instance_t apsh, uint64_t flags)
{
	if (!apsh)
		return APSH_PM_UNKNOWN;

	apsh->page_mode = APSH_PM_UNKNOWN;

	if (arch_init(apsh) == APSH_F)
		return APSH_PM_UNKNOWN;

	if (flags) {
		switch (apsh->page_mode) {
		case APSH_PM_LEGACY:
		case APSH_PM_PAE:
		case APSH_PM_IA32E:
			if (flags & APSH_PM_INITFLAG_TRANSITION_PAGES)
				apsh->x86.transition_pages = true;
			break;
		default:
			break;
		};
	}

	return apsh->page_mode;
}

st_stat() apsh_set_paging(apsh_instance_t apsh, page_mode_t page_mode)
{

	apsh->page_mode = page_mode;

	return APSH_S;
}

status_t apsh_set_limits_table(apsh_instance_t apsh, GHashTable *limits_table)
{
	apsh->limits_table = limits_table;
	return APSH_S;
}

GHashTable *
init_config(apsh_instance_t apsh, apsh_config_t config_mode, void *config, addr_t EPT,
	    apsh_init_error_t *error)
{
	GHashTable *_config = NULL;
	addr_t kernbase;
	uint8_t pe[APSH_PS_4KB];
	char *config_cpy;

	apsh->EPT = EPT;
	switch (config_mode) {
	case APSH_CONFIG_WINDOW_AUTO:
		kernbase = get_ntoskrnl_base(apsh);
		if (get_windows_profile(apsh, kernbase, pe) == APSH_S) {
			asprintf(&config, "/etc/pdb_profiles/%s.json", apsh->host_data.pdb_guid);
			debug_print(APSH_DEBUG_MISC, "Set profile to : %s\n", (char *)config);
			if (access(config, F_OK) != 0) {
				char *command;

				asprintf(&command,
					 "pdbparser --p %s -g %s -o /etc/pdb_profiles/%s.json",
					 apsh->host_data.kernel_name, apsh->host_data.pdb_guid,
					 apsh->host_data.pdb_guid);
				system(command);
				free(command);
			}
			if (error)
				*error = APSH_INIT_ERROR_NO_CONFIG;
			_config = g_hash_table_new(g_str_hash, g_str_equal);
			if (_config == NULL) {
				free(config);
				return NULL;
			}
			config_cpy = g_strdup(config);
			free(config);
			if (config_cpy == NULL) {
				g_hash_table_destroy(_config);
				return NULL;
			}
			g_hash_table_insert(_config, "volatility_ist", config_cpy);
			break;
		} else {
			return NULL;
		}
	case APSH_CONFIG_STRING:
		if (read_config_string(apsh, (const char *)config, &_config, error) == APSH_F)
			return NULL;
		break;
	case APSH_CONFIG_GLOBAL_FILE_ENTRY:
		if (read_config_file_entry(apsh, &_config, error) == APSH_F)
			return NULL;
		break;
	case APSH_CONFIG_GHASHTABLE:
		if (!config) {

			if (error)
				*error = APSH_INIT_ERROR_NO_CONFIG;

			return NULL;
		}
		_config = (GHashTable *)config;
		break;
	case APSH_CONFIG_JSON_PATH:
		if (!config) {

			if (error)
				*error = APSH_INIT_ERROR_NO_CONFIG;

			return NULL;
		}
		_config = g_hash_table_new(g_str_hash, g_str_equal);
		if (_config == NULL)
			return NULL;
		config_cpy = g_strdup(config);
		if (config_cpy == NULL) {
			g_hash_table_destroy(_config);
			return NULL;
		}
		g_hash_table_insert(_config, "volatility_ist", config_cpy);
		break;
	default:
		return NULL;
	}

	if (set_os_type_from_config(apsh, _config) == APSH_F) {
		if (error)
			*error = APSH_INIT_ERROR_NO_CONFIG_ENTRY;

		debug_print(APSH_DEBUG_CORE, "--failed to determine os type from config\n");
		return NULL;
	}

	return _config;
}

os_t
apsh_init_profile(apsh_instance_t apsh, apsh_config_t config_mode, void *config, addr_t EPT)
{
	GHashTable *_config = init_config(apsh, config_mode, config, EPT, NULL);

	if (!_config)
		return APSH_OS_UNKNOWN;

	return apsh->os_type;
}

os_t
apsh_init_os(apsh_instance_t apsh, apsh_config_t config_mode, void *config, addr_t EPT,
	     const char *kpgd_file, apsh_init_error_t *error)
{
	if (!apsh)
		return APSH_OS_UNKNOWN;

	apsh->kpgd_file = kpgd_file;
	apsh->os_type = APSH_OS_UNKNOWN;
	GHashTable *_config = init_config(apsh, config_mode, config, EPT, error);

	if (!_config)
		goto error_exit;

	if (apsh->mode != APSH_FILE && apsh->page_mode == APSH_PM_UNKNOWN &&
	    apsh_init_paging(apsh, apsh->os_type == APSH_OS_WIN ? APSH_PM_INITFLAG_TRANSITION_PAGES
								: 0) == APSH_PM_UNKNOWN) {
		apsh->os_type = APSH_OS_UNKNOWN;
		if (error)
			*error = APSH_INIT_ERROR_PAGING;

		goto error_exit;
	}

	switch (apsh->os_type) {
#ifdef ENABLE_LINUX
	case APSH_OS_LINUX:
		if (linux_init(apsh, _config) == APSH_F) {
			apsh->os_type = APSH_OS_UNKNOWN;
			if (error)
				*error = APSH_INIT_ERROR_OS;

			goto error_exit;
		}
		break;
#endif
#ifdef ENABLE_WINDOWS
	case APSH_OS_WIN:
		if (windows_init(apsh, _config, EPT) == APSH_F) {
			apsh->os_type = APSH_OS_UNKNOWN;
			if (error)
				*error = APSH_INIT_ERROR_OS;

			goto error_exit;
		}
		break;
#endif
	default:
		apsh->os_type = APSH_OS_UNKNOWN;
		if (error)
			*error = APSH_INIT_ERROR_OS;

		goto error_exit;
	};

error_exit:
	if (config_mode == APSH_CONFIG_JSON_PATH || config_mode == APSH_CONFIG_WINDOW_AUTO) {
		g_free(g_hash_table_lookup(_config, "volatility_ist"));
		g_hash_table_destroy(_config);
	}

	return apsh->os_type;
}

st_stat() apsh_init_complete(apsh_instance_t *apsh, const void *domain, uint64_t init_flags,
			     apsh_init_data_t *init_data, apsh_config_t config_mode, void *config,
			     addr_t EPT, addr_t vf_id, const char *mem_region_json, const char *kpgd_file,
			     const char *mlx_device, struct doca_dev *regex_dev,
			     struct doca_dev_rep *system_dev, apsh_init_error_t *error)
{
	apsh_instance_t _apsh = NULL;
	apsh_mode_t mode;
	if (error) {
		*error = APSH_INIT_ERROR_NONE;
	}
	if (apsh_get_access_mode(_apsh, domain, init_flags, init_data, &mode) == APSH_F) {
		if (error)
			*error = APSH_INIT_ERROR_DRIVER_NOT_DETECTED;

		return APSH_F;
	}

	if (apsh_init(&_apsh, mode, domain, init_flags, init_data, vf_id, mem_region_json,
			kpgd_file, mlx_device, regex_dev, system_dev, error) == APSH_F)
		return APSH_F;

	if (APSH_FILE != mode && APSH_PM_UNKNOWN == apsh_init_paging(_apsh, 0)) {
		if (error)
			*error = APSH_INIT_ERROR_PAGING;

		goto error_exit;
	}

	if (apsh_init_os(_apsh, config_mode, config, EPT, kpgd_file, error) == APSH_OS_UNKNOWN)
		goto error_exit;

	debug_print(APSH_DEBUG_CORE, "--OS Init Completed.\n");
	*apsh = _apsh;
	return APSH_S;

error_exit:
	apsh_destroy(_apsh);
	return APSH_F;
}

st_stat() apsh_destroy(apsh_instance_t apsh)
{
	if (!apsh)
		return APSH_F;

	apsh->shutting_down = TRUE;

	driver_destroy(apsh);

	if (apsh->os_interface)
		os_destroy(apsh);

	free(apsh->os_data);
	apsh->os_data = NULL;

#ifdef ENABLE_JSON_PROFILES
	g_free((char *)apsh->json.path);
	if (apsh->json.root)
		json_object_put(apsh->json.root);
	jsonout_deinit();
#endif

	pid_cache_destroy(apsh);
	sym_cache_destroy(apsh);
	rva_cache_destroy(apsh);
	v2p_cache_destroy(apsh);

	if (apsh->plugin_ctx)
		g_hash_table_destroy(apsh->plugin_ctx);

	if (apsh->image_type)
		free(apsh->image_type);
	if (apsh->image_type_complete)
		free(apsh->image_type_complete);
	g_free(apsh->memmap);
	g_free(apsh);
	return APSH_S;
}

apsh_arch_t
apsh_get_library_arch(void)
{
#ifdef I386
	return APSH_ARCH_X86;
#elif defined(X86_64)
	return APSH_ARCH_X86_64;
#elif defined(ARM32)
	return APSH_ARCH_ARM32;
#elif defined(ARM64)
	return APSH_ARCH_ARM64;
#endif

	return APSH_ARCH_UNKNOWN;
}
