/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DRIVER_WRAPPER_H
#define DRIVER_WRAPPER_H

#include "apsh_instance.h"

/*
 * The following functions are safety-wrappers that should be used internally
 * instead of calling the functions directly on the driver.
 */

static inline void
driver_destroy(
    apsh_instance_t apsh)
{
#ifdef ENABLE_SAFETY_CHECKS
    if (apsh->driver.initialized && apsh->driver.destroy_ptr)
#endif
        apsh->driver.destroy_ptr(apsh);

    memset(&apsh->driver, 0, sizeof(driver_interface_t));
}

static inline status_t
driver_get_name(
    apsh_instance_t apsh,
    char **name)
{
#ifdef ENABLE_SAFETY_CHECKS
    if (!apsh->driver.initialized || !apsh->driver.get_name_ptr) {
        debug_print(APSH_DEBUG_DRIVER, "WARNING: driver_get_name function not implemented.\n");
        return APSH_F;
    }
#endif

    return apsh->driver.get_name_ptr(apsh, name);
}

static inline void
driver_set_name(
    apsh_instance_t apsh,
    const char *name)
{
#ifdef ENABLE_SAFETY_CHECKS
    if (!apsh->driver.initialized || !apsh->driver.set_name_ptr) {
        debug_print(APSH_DEBUG_DRIVER, "WARNING: driver_set_name function not implemented.\n");
        return;
    }
#endif

    return apsh->driver.set_name_ptr(apsh, name);
}

static inline status_t
driver_get_memsize(
    apsh_instance_t apsh,
    uint64_t *allocated_ram_size,
    addr_t *max_physical_address)
{
#ifdef ENABLE_SAFETY_CHECKS
    if (!apsh->driver.initialized || !apsh->driver.get_memsize_ptr) {
        debug_print(APSH_DEBUG_DRIVER, "WARNING: driver_get_memsize function not implemented.\n");
        return APSH_F;
    }
#endif

    return apsh->driver.get_memsize_ptr(apsh, allocated_ram_size, max_physical_address);
}

static inline void *
driver_read_page(
    apsh_instance_t apsh,
    addr_t page)
{
#ifdef ENABLE_SAFETY_CHECKS
    if (!apsh->driver.initialized || !apsh->driver.read_page_ptr) {
        debug_print(APSH_DEBUG_DRIVER, "WARNING: driver_read_page function not implemented.\n");
        return NULL;
    }
#endif

    return apsh->driver.read_page_ptr(apsh, page);
}

static inline void *
driver_read_memory(
    apsh_instance_t apsh,
    addr_t paddr,
    uint32_t length)
{
#ifdef ENABLE_SAFETY_CHECKS
    if (!apsh->driver.initialized || !apsh->driver.read_memory_ptr) {
        debug_print(APSH_DEBUG_DRIVER, "WARNING: read_memory_ptr function not implemented.\n");
        return NULL;
    }
#endif

    return apsh->driver.read_memory_ptr(apsh, paddr, length);
}



#endif /* DRIVER_WRAPPER_H */

