/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "apsh_instance.h"
#include "driver/handle_files/file.h"
#include "driver/handle_files/file_private.h"
#include "driver/driver_interface.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <string.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include "doca_mema.h"


// Use mmap() if this evaluates to true; otherwise, use a file pointer with
// seek/read
#define USE_MMAP 0

// Avoid errors on systems that don't have MAP_POPULATE defined
#ifndef MAP_POPULATE
#define MAP_POPULATE 0
#endif

//----------------------------------------------------------------------------
// File-Specific Interface Functions


void *
file_get_memory(
    apsh_instance_t apsh,
    addr_t paddr,
    uint32_t length)
{
    char *buffer = g_try_malloc0(length);
    char *source;

    if (!buffer) {
      debug_print(APSH_DEBUG_WRITE, "%s: failed to allocate memory\n",
                  __FUNCTION__);
      return NULL;
    }

    if (!pf_mem_is_valid(apsh->mdev->mmem_config, paddr,length)){
        goto error_print;
    }

    source = doca_mema_chnl_dma_read(apsh->mdev, paddr, length);
    if (!source) {
        goto error_print;
    }
    memcpy(buffer, source, length);

    return buffer;

error_print:
    debug_print(APSH_DEBUG_WRITE, "%s: failed to read %d bytes at "
                             "PA (offset) 0x%.16" PRIx64 " [VM size 0x%.16" PRIx64 "]\n",
            __FUNCTION__,
            length, paddr, apsh->allocated_ram_size);

    g_free(buffer);
    return NULL;
}

void file_release_memory(
    apsh_instance_t UNUSED(apsh),
    void *memory,
    size_t UNUSED(length))
{
    if (memory)
        free(memory);
}

//----------------------------------------------------------------------------
// General Interface Functions (1-1 mapping to driver_* function)

status_t
file_init(
    apsh_instance_t apsh,
    uint32_t UNUSED(init_flags),
    apsh_init_data_t *UNUSED(init_data))
{
    apsh->driver.driver_data = g_try_malloc0(sizeof(file_instance_t));
    return APSH_S;
}

status_t
file_init_apsh(
    apsh_instance_t apsh,
    uint32_t UNUSED(init_flags),
    apsh_init_data_t *UNUSED(init_data))
{

    int err = 0;
    err = doca_mema_create_channel(&apsh->mdev, apsh->mlx_device, apsh->vf_id);
    if (err)
    {
        printf("failed to create channel, err=%d\n", err);
        return APSH_F;
    }

    apsh->mdev->mmem_config = create_machine_mem_regions(apsh->mem_region_json);
    apsh->vm_type = NORMAL;
    return APSH_S;
}

void file_destroy(apsh_instance_t apsh)
{
	if (apsh->driver.driver_data != NULL)
		g_free(apsh->driver.driver_data);
	doca_mema_destroy_channel(&apsh->mdev);
}

status_t
file_get_name()
{
    return APSH_S;
}

void file_set_name()
{
}

status_t
file_get_memsize(
    apsh_instance_t apsh,
    uint64_t *allocated_ram_size,
    addr_t *max_physical_address)
{
    status_t ret = APSH_F;

    if(!apsh || !apsh->mdev)
        return ret;

    if (!apsh->mdev->mmem_config)
        return ret;

    *allocated_ram_size = 0;
    *max_physical_address = 0;

    mr_t* mem_regions = apsh->mdev->mmem_config->mrs;
    int mem_regions_cnt = apsh->mdev->mmem_config->mrs_count;

    for(int i = 0; i< mem_regions_cnt; i++){
        uint64_t last_address = mem_regions[i].start_address + mem_regions[i].length;
        if (*max_physical_address < last_address){
            *max_physical_address = last_address;
            *allocated_ram_size = last_address;
        }
        ret = APSH_S;
    }

    return ret;
}

void *
file_read_page(
    apsh_instance_t apsh,
    addr_t page)
{
    addr_t paddr = page << apsh->page_shift;
    addr_t paddr_aligned = paddr & ~(((addr_t) apsh->page_size) - 1);

    if (paddr != paddr_aligned) {
        print_error("Request for non-aligned page\n");
        return NULL;
    }

    return file_get_memory(apsh, paddr, apsh->page_size);
}

status_t file_test()
{
    return APSH_S;
}
