/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "apsh_instance.h"
#include "config/config_parser.h"
#include "driver/driver_wrapper.h"
#include "os/linux/linux.h"

#define INDIRECT(X) #X
#define STRINGIFY(X) INDIRECT(X)

#define MAX_LINE_LEN 256

#define st_stat() status_t
#define LSTA linux_symbol_to_address
#define copy_string_func strncmp

static void
linux_read_config_ghashtable_entries(char *key, gpointer value, apsh_instance_t apsh);

static st_stat()
init_task_kaslr_test(apsh_instance_t apsh, addr_t page_vaddr);

static bool
is_x86_64_pd(apsh_instance_t apsh, addr_t pa);

static st_stat()
init_kaslr(apsh_instance_t apsh);

static st_stat()
brute_force_find_kern_mem(apsh_instance_t apsh);

static st_stat()
verify_linux_paging(apsh_instance_t apsh);

static st_stat()
find_kpgd_with_file(apsh_instance_t apsh);

static st_stat()
find_page_directories(apsh_instance_t apsh)
{
	st_stat() rc = APSH_F;
	addr_t candidate;

	if (apsh_get_page_mode(apsh, 0) != APSH_PM_IA32E)
		goto exit;

	for (candidate = 0x1000; candidate < apsh_get_max_physical_address(apsh); candidate += APSH_PS_4KB) {
		if (is_x86_64_pd(apsh, candidate)) {
			apsh->kpgd = candidate;
			if (verify_linux_paging(apsh) == APSH_S) {
				rc = APSH_S;
				break;
			}
		}
	}

exit:
	return rc;
}

static st_stat()
linux_filemode_32bit_init(apsh_instance_t apsh, addr_t swapper_pg_dir, addr_t boundary, addr_t pa, addr_t va)
{
	addr_t test = 0;

	apsh->page_mode = APSH_PM_LEGACY;
	if (arch_init(apsh) == APSH_S) {
		if (apsh_pagetable_lookup(apsh, swapper_pg_dir - boundary, va, &test) == APSH_S && test == pa) {
			apsh->kpgd = swapper_pg_dir - boundary;
			return APSH_S;
		}
	}

	apsh->page_mode = APSH_PM_PAE;
	if (arch_init(apsh) == APSH_S) {
		if (apsh_pagetable_lookup(apsh, swapper_pg_dir - boundary, va, &test) == APSH_S && test == pa) {
			apsh->kpgd = swapper_pg_dir - boundary;
			return APSH_S;
		}
	}

	apsh->page_mode = APSH_PM_AARCH32;
	if (arch_init(apsh) == APSH_S) {
		if (apsh_pagetable_lookup(apsh, swapper_pg_dir - boundary, va, &test) == APSH_S && test == pa) {
			apsh->kpgd = swapper_pg_dir - boundary;
			return APSH_S;
		}
	}

	return APSH_F;
}

static st_stat()
init_paging(apsh_instance_t apsh)
{
	st_stat() rc;
	linux_instance_t linux_instance = apsh->os_data;

	switch (apsh->page_mode) {
	case APSH_PM_AARCH64:
	case APSH_PM_IA32E:
		LSTA(apsh, "phys_startup_64", NULL, &linux_instance->phys_start);
		LSTA(apsh, "startup_64", NULL, &linux_instance->virt_start);
		break;
	case APSH_PM_AARCH32:
	case APSH_PM_LEGACY:
	case APSH_PM_PAE:
		LSTA(apsh, "phys_startup_32", NULL, &linux_instance->phys_start);
		LSTA(apsh, "startup_32", NULL, &linux_instance->virt_start);
		break;
	case APSH_PM_UNKNOWN:
		LSTA(apsh, "phys_startup_64", NULL, &linux_instance->phys_start);
		LSTA(apsh, "startup_64", NULL, &linux_instance->virt_start);

		if (linux_instance->phys_start && linux_instance->virt_start)
			break;
		linux_instance->phys_start = linux_instance->virt_start = 0;

		LSTA(apsh, "phys_startup_32", NULL, &linux_instance->phys_start);
		LSTA(apsh, "startup_32", NULL, &linux_instance->virt_start);
		break;
	}

	linux_instance->virt_start = canonical_addr(linux_instance->virt_start);

	if (linux_instance->phys_start && linux_instance->virt_start &&
	    linux_instance->phys_start < linux_instance->virt_start) {
		linux_instance->boundary = linux_instance->virt_start - linux_instance->phys_start;
		debug_print(APSH_DEBUG_MISC, "--got kernel boundary (0x%.16" PRIx64 ").\n", linux_instance->boundary);
	}

	rc = LSTA(apsh, "swapper_pg_dir", NULL, &linux_instance->swapper_pg_dir);

	if (rc == APSH_S) {

		debug_print(APSH_DEBUG_MISC, "--got vaddr for swapper_pg_dir (0x%.16" PRIx64 ").\n",
			    linux_instance->swapper_pg_dir);

		linux_instance->swapper_pg_dir = canonical_addr(linux_instance->swapper_pg_dir);

		if (linux_instance->boundary) {
			rc = linux_filemode_32bit_init(apsh, linux_instance->swapper_pg_dir, linux_instance->boundary,
						       linux_instance->phys_start, linux_instance->virt_start);
			if (rc == APSH_S)
				goto done;
		}

		linux_instance->boundary = 0xC0000000;
		debug_print(APSH_DEBUG_MISC, "--trying boundary 0x%.16" PRIx64 ".\n", linux_instance->boundary);
		rc = linux_filemode_32bit_init(apsh, linux_instance->swapper_pg_dir, linux_instance->boundary,
					       linux_instance->swapper_pg_dir - linux_instance->boundary,
					       linux_instance->swapper_pg_dir);
		if (rc == APSH_S)
			goto done;

		linux_instance->boundary = 0x80000000;
		debug_print(APSH_DEBUG_MISC, "--trying boundary 0x%.16" PRIx64 ".\n", linux_instance->boundary);
		rc = linux_filemode_32bit_init(apsh, linux_instance->swapper_pg_dir, linux_instance->boundary,
					       linux_instance->swapper_pg_dir - linux_instance->boundary,
					       linux_instance->swapper_pg_dir);
		if (rc == APSH_S)
			goto done;

		linux_instance->boundary = 0x40000000;
		debug_print(APSH_DEBUG_MISC, "--trying boundary 0x%.16" PRIx64 ".\n", linux_instance->boundary);
		rc = linux_filemode_32bit_init(apsh, linux_instance->swapper_pg_dir, linux_instance->boundary,
					       linux_instance->swapper_pg_dir - linux_instance->boundary,
					       linux_instance->swapper_pg_dir);
		goto done;
	}

	rc = LSTA(apsh, "init_level4_pgt", NULL, &linux_instance->kernel_pgt);
	if (rc == APSH_F)
		rc = LSTA(apsh, "init_top_pgt", NULL, &linux_instance->kernel_pgt);

	if (rc == APSH_F)
		goto done;

	debug_print(APSH_DEBUG_MISC, "--got vaddr for kernel pagetable (0x%.16" PRIx64 ").\n",
		    linux_instance->kernel_pgt);

	linux_instance->kernel_pgt = canonical_addr(linux_instance->kernel_pgt);

	if (!linux_instance->boundary)
		return APSH_F;

	apsh->page_mode = APSH_PM_IA32E;

	rc = arch_init(apsh);
	if (rc == APSH_F)
		return APSH_F;

done:
	return rc;
}

static st_stat()
linux_filemode_init(apsh_instance_t apsh)
{
	st_stat() rc;
	linux_instance_t linux_instance = apsh->os_data;

	addr_t test = 0;

	rc = apsh_pagetable_lookup(apsh, linux_instance->kernel_pgt - linux_instance->boundary,
				   linux_instance->virt_start, &test);
	if (rc == APSH_S && test == linux_instance->phys_start) {
		apsh->kpgd = linux_instance->kernel_pgt - linux_instance->boundary;
		rc = verify_linux_paging(apsh);
	}

	if (rc == APSH_F)
		rc = brute_force_find_kern_mem(apsh);

	return rc;
}

static st_stat()
init_from_json_profile(apsh_instance_t apsh)
{

	st_stat() ret = APSH_F;
	linux_instance_t linux_instance = apsh->os_data;

	if (!linux_instance->tasks_offset) {
		if (json_profile_lookup(apsh, "task_struct", "tasks", &linux_instance->tasks_offset) == APSH_F)
			goto done;
	}
	if (!linux_instance->mm_offset) {
		if (json_profile_lookup(apsh, "task_struct", "mm", &linux_instance->mm_offset) == APSH_F)
			goto done;
	}
	if (!linux_instance->pid_offset) {
		if (json_profile_lookup(apsh, "task_struct", "pid", &linux_instance->pid_offset) == APSH_F)
			goto done;
	}
	if (!linux_instance->name_offset) {
		if (json_profile_lookup(apsh, "task_struct", "comm", &linux_instance->name_offset) == APSH_F)
			goto done;
	}
	if (!linux_instance->pgd_offset) {
		if (json_profile_lookup(apsh, "mm_struct", "pgd", &linux_instance->pgd_offset) == APSH_F)
			goto done;
	}
	if (!apsh->init_task) {
		if (json_profile_lookup(apsh, "init_task", NULL, &apsh->init_task) == APSH_F)
			goto done;
	}

	ret = APSH_S;

done:
	return ret;
}

static st_stat()
init_task_kaslr_test(apsh_instance_t apsh, addr_t page_vaddr)
{
	st_stat() ret = APSH_F;
	uint32_t pid = -1;
	addr_t addr = ~0;
	addr_t init_task = page_vaddr + (apsh->init_task & APSH_BIT_MASK(0, 11));
	linux_instance_t linux_instance = apsh->os_data;

	ACCESS_CONTEXT(ctx, .pm = apsh->page_mode, .translate_mechanism = APSH_TM_PROCESS_PT, .pt = apsh->kpgd);

	ctx.addr = init_task + linux_instance->pid_offset;
	if (apsh_read_32(apsh, &ctx, &pid) == APSH_F)
		return ret;

	if (pid != 0)
		return ret;

	ctx.addr = init_task + linux_instance->mm_offset;
	if (apsh_read_addr(apsh, &ctx, &addr) == APSH_F)
		return ret;

	if (addr != 0)
		return ret;

	ctx.addr = init_task + linux_instance->tasks_offset;
	if (apsh_read_addr(apsh, &ctx, &addr) == APSH_F)
		return ret;

	ctx.addr = addr;
	if (apsh_read_addr(apsh, &ctx, &addr) == APSH_F)
		return ret;

	ctx.addr = init_task + linux_instance->name_offset;
	char *init_task_name = apsh_read_str(apsh, &ctx);

	if (init_task_name && !copy_string_func("swapper", init_task_name, 7))
		ret = APSH_S;

	free(init_task_name);
	return ret;
}

static st_stat()
get_kaslr_offset_ia32e(apsh_instance_t apsh)
{
	addr_t va, pa;
	addr_t kernel_text_start = 0xffffffff81000000;
	addr_t kernel_text_end = kernel_text_start + (1024 * 1024 * 1024);

	linux_instance_t linux_instance = apsh->os_data;

	apsh->init_task = linux_instance->init_task_fixed;

	for (va = kernel_text_start; va < kernel_text_end; va += 0x200000) {
		if (apsh_translate_kv2p(apsh, va, &pa) == APSH_S) {
			linux_instance->kaslr_offset = va - kernel_text_start;
			apsh->init_task += linux_instance->kaslr_offset;
			debug_print(APSH_DEBUG_MISC, "**calculated KASLR offset in 64-bit mode: 0x%" PRIx64 "\n",
				    linux_instance->kaslr_offset);
			return APSH_S;
		}
	}
	return APSH_F;
}

static st_stat()
init_kaslr(apsh_instance_t apsh)
{
	uint32_t test;
	linux_instance_t linux_instance = apsh->os_data;
	addr_t init_task_symbol_addr;

	if (LSTA(apsh, "init_task", NULL, &init_task_symbol_addr) == APSH_F)
		return APSH_F;

	ACCESS_CONTEXT(ctx, .translate_mechanism = APSH_TM_PROCESS_DTB, .pt = apsh->kpgd, .addr = apsh->init_task);

	if (apsh_read_32(apsh, &ctx, &test) == APSH_S) {

		linux_instance->kaslr_offset = apsh->init_task - init_task_symbol_addr;
		debug_print(APSH_DEBUG_MISC,
			    "**calculated KASLR offset from pre-defined init_task addr: 0x%" PRIx64 "\n",
			    linux_instance->kaslr_offset);
		return APSH_S;
	}

	apsh->init_task = init_task_symbol_addr;

	st_stat() ret = APSH_F;
	if (apsh->page_mode == APSH_PM_IA32E) {
		if (get_kaslr_offset_ia32e(apsh) == APSH_S)
			return APSH_S;
	} else {
		GSList *loop, *pages = apsh_get_va_pages(apsh, apsh->kpgd);

		loop = pages;
		while (loop) {
			page_info_t *info = loop->data;

			if (!linux_instance->kaslr_offset) {
				switch (apsh->page_mode) {
				case APSH_PM_AARCH64:
				case APSH_PM_IA32E:
					if (APSH_GET_BIT(info->vaddr, 47))
						ret = init_task_kaslr_test(apsh, info->vaddr);
					break;
				default:
					ret = init_task_kaslr_test(apsh, info->vaddr);
					break;
				}

				if (ret == APSH_S) {
					linux_instance->kaslr_offset =
						info->vaddr - (apsh->init_task & ~APSH_BIT_MASK(0, 11));
					apsh->init_task =
						linux_instance->init_task_fixed + linux_instance->kaslr_offset;
					debug_print(APSH_DEBUG_MISC, "**calculated KASLR offset: 0x%" PRIx64 "\n",
						    linux_instance->kaslr_offset);
					break;
				}
			}

			g_free(info);
			loop = loop->next;
		}
		g_slist_free(pages);
	}
	return ret;
}

static st_stat()
verify_linux_paging(apsh_instance_t apsh)
{
	if (init_kaslr(apsh) == APSH_F)
		return APSH_F;

	return init_task_kaslr_test(apsh, apsh->init_task & ~APSH_BIT_MASK(0, 11));
}

static st_stat()
brute_force_find_kern_mem(apsh_instance_t apsh)
{
	st_stat() rc = APSH_F;

	if (find_page_directories(apsh) == APSH_S) {
		rc = APSH_S;
		goto exit;
	}

	print_warning("Looking for kernel PGD and KASLR with slowest available technique\n");

	for (apsh->kpgd = 0; apsh->kpgd < apsh_get_max_physical_address(apsh); apsh->kpgd += APSH_PS_4KB) {
		if (verify_linux_paging(apsh) == APSH_S) {
			rc = APSH_S;
			break;
		}
	}

exit:
	if (rc == APSH_S) {
		debug_print(APSH_DEBUG_MISC, "**found kernel PGD: 0x%" PRIx64 " and init task: 0x%" PRIx64 "\n",
			    apsh->kpgd, apsh->init_task);
	} else {
		print_error("Brute force search failed to find kernel PDB and KASLR offset\n");
	}
	return rc;
}

static st_stat() find_kpgd_with_file(apsh_instance_t apsh)
{
	debug_print(APSH_DEBUG_MISC, "**Trying to aquire kpgd from kpgd_file\n");

	if (apsh->kpgd_file == NULL) {
		debug_print(APSH_DEBUG_MISC, "**kpgd_file is NULL\n");
		return APSH_F;
	}

	char buff[MAX_LINE_LEN + 1];
	FILE *fp = fopen(apsh->kpgd_file, "r");

	if (fp == NULL) {
		debug_print(APSH_DEBUG_MISC, "**couldn't open kpgd_file\n");
		return APSH_F;
	}

	if (fscanf(fp, "%"STRINGIFY(MAX_LINE_LEN)"s", buff) != EOF) {
		char *endptr;
		addr_t value = (addr_t)strtoull(buff, &endptr, 16);

		if (endptr != buff) {
			apsh->kpgd = value;
			debug_print(APSH_DEBUG_MISC, "**set apsh->kpgd from file (0x%.16" PRIx64 ").\n",
				    apsh->kpgd);
		}

		if (fscanf(fp, "%"STRINGIFY(MAX_LINE_LEN)"s", buff) != EOF) {
			value = (addr_t)strtoull(buff, &endptr, 16);

			if (endptr != buff) {
				apsh->init_task = value;
				debug_print(APSH_DEBUG_MISC, "**set apsh->init_task from file (0x%.16" PRIx64 ").\n",
					    apsh->init_task);
			}
		}
	}

	fclose(fp);
	return APSH_S;

}

st_stat() refind_kpgd(apsh_instance_t apsh)
{
	debug_print(APSH_DEBUG_MISC, "**Trying to search for kpgd from saved candidates list\n");
	debug_print(APSH_DEBUG_MISC, "**Scanning with apsh->init_task (0x%.16" PRIx64 ").\n", apsh->init_task);

	FILE *fp = fopen("/etc/APSH_KPGD_CANDIDATES.conf", "r");
	char buff[MAX_LINE_LEN + 1];

	if (fp != NULL) {
		while (fscanf(fp, "%"STRINGIFY(MAX_LINE_LEN)"s", buff) == 1) {
			if (apsh->init_task != 0) {
				addr_t candidate = 0;
				char *endptr;
				addr_t value = (addr_t)strtoull(buff, &endptr, 16);

				if (endptr != buff)
					candidate = value;

				if (candidate != 0) {
					if (is_x86_64_pd(apsh, candidate)) {
						apsh->kpgd = candidate;
						if (verify_linux_paging(apsh) == APSH_S) {
							debug_print(APSH_DEBUG_MISC,
								"*****Found KPGD from correct init_task and KASLR\n");
							fclose(fp);

							/* rewrite correct value to load */
							fp = fopen("/etc/APSH_LINUX_KPGD.conf", "w");
							if (fp != NULL) {
								fprintf(fp, "%.16" PRIx64 "\n%.16" PRIx64 "", apsh->kpgd,
									apsh->init_task);
								fclose(fp);
							}

							return APSH_S;
						}
					}
				}
			}
		}
		fclose(fp);
	}

	debug_print(APSH_DEBUG_MISC, "*****Candidate list failed...\n");
	return APSH_F;
}

st_stat()
linux_init(apsh_instance_t apsh, GHashTable *config)
{
	st_stat() rc;
	os_interface_t int_os = NULL;
	bool shouldSaveNewCandidates = false;

	if (!config) {
		print_error("No config table found\n");
		return APSH_F;
	}

	if (apsh->os_data != NULL) {
		print_error("os data already initialized, reinitializing\n");
		g_free(apsh->os_data);
	}

	apsh->os_data = g_try_malloc0(sizeof(struct linux_instance));
	if (!apsh->os_data)
		return APSH_F;

	linux_instance_t linux_instance = apsh->os_data;

	g_hash_table_foreach(config, (GHFunc)linux_read_config_ghashtable_entries, apsh);

	rc = init_from_json_profile(apsh);

	if (rc == APSH_F && !apsh->init_task)
		rc = LSTA(apsh, "init_task", NULL, &apsh->init_task);
	else
		rc = APSH_S;

	if (rc == APSH_F) {
		print_error("Failed to determine init_task!\n");
		goto _exit;
	}

	apsh->init_task = canonical_addr(apsh->init_task);
	linux_instance->init_task_fixed = apsh->init_task;

	init_paging(apsh);

	if (!apsh->kpgd) {
		addr_t original_init_task = apsh->init_task;

		find_kpgd_with_file(apsh);
		/* validate KPGD... */
		if (apsh->kpgd) {
			if (!linux_instance->kaslr_offset) {
				if (verify_linux_paging(apsh) == APSH_F) {
					apsh->kpgd |= 0x1000ull;
					if (verify_linux_paging(apsh) == APSH_F) {
						debug_print(APSH_DEBUG_MISC, "**KPGD invalid, recalcing kpgd and kaslar\n");
						apsh->kpgd = 0;
						apsh->init_task = original_init_task;
					}
				}
			}
		}
	}

	FILE *fp = NULL;
	FILE *fp_find = NULL;
	FILE *fp_exists = NULL;

	if (!apsh->kpgd) {
		char buff[MAX_LINE_LEN + 1];

		fp = fopen("/etc/APSH_LINUX_KPGD.conf", "r");
		if (fp != NULL) {
			if (fscanf(fp, "%"STRINGIFY(MAX_LINE_LEN)"s", buff) != EOF) {
				char *endptr;
				addr_t value = (addr_t)strtoull(buff, &endptr, 16);

				if (endptr != buff) {
					apsh->kpgd = value;
					debug_print(APSH_DEBUG_MISC, "**set apsh->kpgd from file (0x%.16" PRIx64 ").\n",
								apsh->kpgd);
				}

				if (fscanf(fp, "%"STRINGIFY(MAX_LINE_LEN)"s", buff) != EOF) {
					char *endptr;
					addr_t value = (addr_t)strtoull(buff, &endptr, 16);

					if (endptr != buff) {
						apsh->init_task = value;
						debug_print(APSH_DEBUG_MISC, "**set apsh->init_task from file (0x%.16" PRIx64 ").\n",
									apsh->init_task);
					}
				}
			}
			fclose(fp);
		}
		if (!apsh->kpgd)
			rc = APSH_F;
	}

	/* validate KPGD... */
	if (apsh->kpgd) {
		if (!linux_instance->kaslr_offset) {
			if (verify_linux_paging(apsh) == APSH_F) {
				apsh->kpgd |= 0x1000ull;
				if (verify_linux_paging(apsh) == APSH_F) {
					debug_print(APSH_DEBUG_MISC, "**KPGD invalid, recalcing kpgd and kaslar\n");
					if (refind_kpgd(apsh) == APSH_F)
						rc = APSH_F;
				}
			}
		}
	}

	if (rc == APSH_F && linux_filemode_init(apsh) == APSH_F)
		goto _exit;

	if (!linux_instance->kaslr_offset) {
		if (verify_linux_paging(apsh) == APSH_F) {
			apsh->kpgd |= 0x1000ull;
			if (verify_linux_paging(apsh) == APSH_F) {
				debug_print(APSH_DEBUG_MISC, "**failed to determine KASLR offset\n");
				goto _exit;
			}
		}
	}

	fp = fopen("/etc/APSH_LINUX_KPGD.conf", "w");
	if (fp == NULL)
		goto _exit;

	fprintf(fp, "%.16" PRIx64 "\n%.16" PRIx64 "", apsh->kpgd, apsh->init_task);
	fclose(fp);

	debug_print(APSH_DEBUG_MISC, "**set apsh->kpgd (0x%.16" PRIx64 ").\n", apsh->kpgd);
	debug_print(APSH_DEBUG_MISC, "**set apsh->init_task (0x%.16" PRIx64 ").\n", apsh->init_task);

	// Save a list of good kpgd candidates inside of memory with swapper!
	fp_find = fopen("/etc/APSH_KPGD_FIND_CANDIDATES.conf", "r");
	fp_exists = fopen("/etc/APSH_KPGD_CANDIDATES.conf", "r");

	if (fp_find != NULL && fp_exists == NULL)
		shouldSaveNewCandidates = true;

	if (fp_find != NULL)
		fclose(fp_find);

	if (fp_exists != NULL)
		fclose(fp_exists);

	if (shouldSaveNewCandidates) {
		addr_t original_kpgd = apsh->kpgd;

		fp = fopen("/etc/APSH_KPGD_CANDIDATES.conf", "w");
		if (fp != NULL) {
			debug_print(APSH_DEBUG_MISC, "**Saving all init tasks with apsh->init (0x%.16" PRIx64 ").\n",
				    apsh->init_task);
			int kpgd_count = 0;
			int max_candidates = 100;

			if (apsh->init_task != 0) {
				addr_t candidate;

				for (candidate = 0x1000; candidate < apsh_get_max_physical_address(apsh);
					candidate += APSH_PS_4KB) {
					if (is_x86_64_pd(apsh, candidate)) {
						apsh->kpgd = candidate;
						if (verify_linux_paging(apsh) == APSH_S) {
							/* debug_print(APSH_DEBUG_MISC, "**Found KPGD from correct init_task and
							 * KASLR %.16" PRIx64 "\n", candidate);
							 */
							fprintf(fp, "%.16" PRIx64 "\n", candidate);
							kpgd_count++;

							if (kpgd_count >= max_candidates)
								break;
						}
					}
				}
			}
			debug_print(APSH_DEBUG_MISC, "*Total KPGD Located: %d\n", kpgd_count);
			fclose(fp);
		}
		apsh->kpgd = original_kpgd;
	}

	int_os = g_malloc(sizeof(struct os_interface));
	if (!int_os)
		goto _exit;

	memset(int_os, 0, sizeof(struct os_interface));
	int_os->os_get_offset = linux_get_offset;
	int_os->os_get_kernel_struct_offset = linux_get_kernel_struct_offset;
	int_os->os_pid_to_pgd = linux_pid_to_pgd;
	int_os->os_pgd_to_pid = linux_pgd_to_pid;
	int_os->os_ksym2v = LSTA;
	int_os->os_usym2rva = NULL;
	int_os->os_v2sym = NULL;
	int_os->os_v2ksym = linux_system_map_address_to_symbol;
	int_os->os_read_unicode_struct = NULL;
	int_os->os_teardown = linux_teardown;

	apsh->os_interface = int_os;

	return APSH_S;

_exit:
	linux_teardown(apsh);
	return APSH_F;
}

void
linux_read_config_ghashtable_entries(char *key, gpointer value, apsh_instance_t apsh);

st_stat()
linux_get_kernel_struct_offset(apsh_instance_t apsh, const char *symbol, const char *member, addr_t *addr)
{
	return json_profile_lookup(apsh, symbol, member, addr);
}

#define linux_get_offset_check(x) (copy_string_func(offset_name, x, max_length) == 0)

st_stat()
linux_get_offset(apsh_instance_t apsh, const char *offset_name, addr_t *offset)
{
	const size_t max_length = 100;
	linux_instance_t linux_instance = apsh->os_data;

	if (linux_instance == NULL) {
		print_error("APSH_ERROR: OS instance not initialized\n");
		return APSH_F;
	}

	if (linux_get_offset_check("linux_tasks")) {
		*offset = linux_instance->tasks_offset;
		return APSH_S;
	} else if (linux_get_offset_check("linux_mm")) {
		*offset = linux_instance->mm_offset;
		return APSH_S;
	} else if (linux_get_offset_check("linux_pid")) {
		*offset = linux_instance->pid_offset;
		return APSH_S;
	} else if (linux_get_offset_check("linux_name")) {
		*offset = linux_instance->name_offset;
		return APSH_S;
	} else if (linux_get_offset_check("linux_pgd")) {
		*offset = linux_instance->pgd_offset;
		return APSH_S;
	} else if (linux_get_offset_check("linux_kaslr")) {
		*offset = linux_instance->kaslr_offset;
		return APSH_S;
	} else if (linux_get_offset_check("linux_init_task")) {
		*offset = apsh->init_task;
		return APSH_S;
	} else if (linux_get_offset_check("kpgd")) {
		*offset = apsh->kpgd;
		return APSH_S;
	}

	print_warning("Invalid offset name in linux_get_offset (%s).\n", offset_name);
	return APSH_F;
}

st_stat()
linux_teardown(apsh_instance_t apsh)
{
	linux_instance_t linux_instance = apsh->os_data;

	if (apsh->os_data == NULL)
		return APSH_S;

	free(linux_instance->sysmap);
	g_free(linux_instance);

	apsh->os_data = NULL;
	apsh->kpgd = 0;

	return APSH_S;
}

static bool
is_x86_64_pd(apsh_instance_t apsh, addr_t pa)
{
	bool rc = false;

	st_stat() status = APSH_F;
	size_t i = 0;

#define PD_ENTRIES (APSH_PS_4KB / sizeof(uint64_t))
	uint64_t pdes[PD_ENTRIES];
	addr_t maxframe = apsh_get_max_physical_address(apsh) >> 12;

	status = apsh_read_pa(apsh, pa, sizeof(pdes), (void *)pdes, NULL);
	if (status == APSH_F)
		goto exit;

	for (i = 0; i < PD_ENTRIES; ++i) {
		uint64_t pde = pdes[i];

		if (APSH_GET_BIT(pde, 7) || APSH_GET_BIT(pde, 63)) {
			rc = false;
			goto exit;
		}

		if (!(APSH_GET_BIT(pde, 0)))
			continue;

		addr_t gfn = pde >> 12;

		if (gfn == 0 || gfn > maxframe) {
			rc = false;
			goto exit;
		}

		rc = true;
	}

exit:
	return rc;
}

void
linux_read_config_ghashtable_entries(char *key, gpointer value, apsh_instance_t apsh)
{

	linux_instance_t linux_instance = apsh->os_data;

	if (copy_string_func(key, "sysmap", CONFIG_STR_LENGTH) == 0) {
		linux_instance->sysmap = strdup((char *)value);
		goto _done;
	}

	if (copy_string_func(key, "linux_tasks", CONFIG_STR_LENGTH) == 0) {
		linux_instance->tasks_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "linux_mm", CONFIG_STR_LENGTH) == 0) {
		linux_instance->mm_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "linux_pid", CONFIG_STR_LENGTH) == 0) {
		linux_instance->pid_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "linux_name", CONFIG_STR_LENGTH) == 0) {
		linux_instance->name_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "linux_pgd", CONFIG_STR_LENGTH) == 0) {
		linux_instance->pgd_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "linux_init_task", CONFIG_STR_LENGTH) == 0) {
		apsh->init_task = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "linux_kaslr", CONFIG_STR_LENGTH) == 0) {
		linux_instance->kaslr_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "kpgd", CONFIG_STR_LENGTH) == 0) {
		apsh->kpgd = *(addr_t *)value;
		goto _done;
	}

_done:
	return;
}
