/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "driver/driver_wrapper.h"
#include "os/linux/linux.h"
#include "apsh_instance.h"
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>

#define address_to_t addr_t

static address_to_t
linux_get_taskstruct_addr_from_pid(apsh_instance_t apsh, apsh_pid_t pid)
{
	address_to_t list_head = 0, curr_entry = 0;
	address_to_t curr_proc = 0;
	apsh_pid_t task_pid = -1;
	linux_instance_t linux_instance = NULL;
	int pid_offset = 0;
	int tasks_offset = 0;

	if (apsh->os_data == NULL) {
		print_error("APSH_ERROR: No os_data initialized\n");
		return 0;
	}

	linux_instance = apsh->os_data;

	pid_offset = linux_instance->pid_offset;
	tasks_offset = linux_instance->tasks_offset;

	curr_proc = apsh->init_task;
	curr_entry = curr_proc + tasks_offset;

	list_head = curr_entry;

	do {
		apsh_read_32_va(apsh, curr_proc + pid_offset, 0, (uint32_t *)&task_pid);

		if (task_pid == pid)
			return curr_proc;

		apsh_read_addr_va(apsh, curr_entry, 0, &curr_entry);
		curr_proc = curr_entry - tasks_offset;

	} while (list_head != curr_entry);

	return 0;
}

static address_to_t
linux_get_taskstruct_addr_from_pgd(apsh_instance_t apsh, address_to_t pgd);

status_t
linux_pid_to_pgd(apsh_instance_t apsh, apsh_pid_t pid, address_to_t *pgd)
{
	address_to_t ts_addr = 0, ptr = 0;
	uint8_t width = 0;
	status_t rc = APSH_F;
	linux_instance_t linux_instance = NULL;
	int mm_offset = 0;
	int pgd_offset = 0;

	if (!apsh->os_data) {
		print_error("APSH_ERROR: No os_data initialized\n");
		return APSH_F;
	}

	linux_instance = apsh->os_data;

	mm_offset = linux_instance->mm_offset;
	pgd_offset = linux_instance->pgd_offset;

	ts_addr = linux_get_taskstruct_addr_from_pid(apsh, pid);
	if (!ts_addr) {
		print_error("Could not find task struct for pid = %d.\n", pid);
		return APSH_F;
	}

	rc = apsh_read_addr_va(apsh, ts_addr + mm_offset, 0, &ptr);
	if (rc == APSH_F)
		return APSH_F;

	if (!ptr) {
		switch (apsh->page_mode) {
		case APSH_PM_AARCH64:
		case APSH_PM_IA32E:
			width = 8;
			break;
		case APSH_PM_AARCH32:
		case APSH_PM_LEGACY:
		case APSH_PM_PAE:
			width = 4;
			break;
		default:
			return APSH_S;
		};

		rc = apsh_read_addr_va(apsh, ts_addr + mm_offset + width, 0, &ptr);

		if (rc == APSH_F || !ptr)
			return rc;
	}

	rc = apsh_read_addr_va(apsh, ptr + pgd_offset, 0, pgd);
	if (rc == APSH_F)
		return rc;

	return apsh_translate_kv2p(apsh, *pgd, pgd);
}

status_t
linux_pgd_to_pid(apsh_instance_t apsh, address_to_t pgd, apsh_pid_t *pid)
{
	address_to_t ts_addr = 0;
	linux_instance_t linux_instance = NULL;
	int pid_offset = 0;

	if (!apsh->os_data) {
		print_error("APSH_ERROR: No os_data initialized\n");
		return APSH_F;
	}

	linux_instance = apsh->os_data;
	pid_offset = linux_instance->pid_offset;

	ts_addr = linux_get_taskstruct_addr_from_pgd(apsh, pgd);
	if (!ts_addr) {
		print_error("Could not find task struct for pgd = 0x%" PRIx64 ".\n", pgd);
		return APSH_F;
	}

	return apsh_read_32_va(apsh, ts_addr + pid_offset, 0, (uint32_t *)pid);
}

static address_to_t
linux_get_taskstruct_addr_from_pgd(apsh_instance_t apsh, address_to_t pgd)
{
	address_to_t list_head = 0, next_process = 0;
	address_to_t task_pgd = 0;
	int tasks_offset = 0;
	int mm_offset = 0;
	int pgd_offset = 0;
	linux_instance_t os = NULL;

	if (apsh->os_data == NULL) {
		print_error("APSH_ERROR: No os_data initialized\n");
		return 0;
	}

	os = apsh->os_data;

	tasks_offset = os->tasks_offset;
	mm_offset = os->mm_offset;
	pgd_offset = os->pgd_offset;

	next_process = apsh->init_task;
	list_head = next_process;

	do {
		address_to_t ptr = 0;

		apsh_read_addr_va(apsh, next_process + mm_offset, 0, &ptr);

		apsh_read_addr_va(apsh, ptr + pgd_offset, 0, &task_pgd);

		apsh_translate_kv2p(apsh, task_pgd, &task_pgd);
		if (task_pgd == pgd || task_pgd == (pgd & ~0x1fff))
			return next_process;

		apsh_read_addr_va(apsh, next_process + tasks_offset, 0, &next_process);
		next_process -= tasks_offset;

	} while (list_head != next_process);

	return 0;
}
