/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "apsh_instance.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "os/linux/linux.h"
#include <ctype.h>
#include <stdio.h>
#include <string.h>

#define MAX_ROW_LENGTH 500
#define string char *

static int get_symbol_row(FILE *f, string row, const string symbol, int position);

static status_t
linux_system_map_symbol_to_address(apsh_instance_t apsh, const string symbol, addr_t *addr)
{
	FILE *f = NULL;
	string row = NULL;
	status_t ret = APSH_F;

	linux_instance_t system_of_linux_kernel = apsh->os_data;

	if (system_of_linux_kernel == NULL) {
		print_error("APSH_ERROR: OS instance not initialized\n");
		goto done;
	}

	if ((system_of_linux_kernel->sysmap == NULL) ||
	    (strlen(system_of_linux_kernel->sysmap) == 0)) {
		print_error("APSH_WARNING: No linux sysmap configured\n");
		goto done;
	}

	row = g_try_malloc0(MAX_ROW_LENGTH);
	if (!row)
		goto done;

	f = fopen(system_of_linux_kernel->sysmap, "r");
	if (f == NULL) {
		addr = 0;
		fprintf(stderr,
			"APSH_ERROR: could not find System.map file after checking:\n\t%s\nTo fix this problem, add the correct sysmap entry to /etc/APSH.conf\n",
			system_of_linux_kernel->sysmap);

		goto done;
	}
	if (get_symbol_row(f, row, symbol, 2) == APSH_F) {
		addr = 0;
		goto done;
	}

	(*addr) = (addr_t)strtoull(row, NULL, 16);

	ret = APSH_S;

done:
	if (row)
		free(row);
	if (f)
		fclose(f);
	return ret;
}

static int
get_symbol_row(FILE *f, string row, const string symbol, int position)
{
	int ret = APSH_F;

	while (fgets(row, MAX_ROW_LENGTH, f) != NULL) {
		string token = NULL;

		int curpos = 0;
		int position_copy = position;

		while (position_copy > 0 && curpos < MAX_ROW_LENGTH) {
			if (isspace(row[curpos])) {
				while (isspace(row[curpos])) {
					row[curpos] = '\0';
					++curpos;
				}
				--position_copy;
				continue;
			}
			++curpos;
		}
		if (position_copy == 0) {
			token = row + curpos;
			while (curpos < MAX_ROW_LENGTH) {
				if (isspace(row[curpos]))
					row[curpos] = '\0';

				++curpos;
			}
		} else {
			goto error_exit;
		}

		if (strncmp(token, symbol, MAX_ROW_LENGTH) == 0) {
			ret = APSH_S;
			break;
		}
	}

error_exit:
	if (ret == APSH_F)
		memset(row, 0, MAX_ROW_LENGTH);

	return ret;
}

string
linux_system_map_address_to_symbol(apsh_instance_t apsh, addr_t addr,
				   const access_context_t *context)
{
	FILE *f = NULL;
	string symbol = NULL;
	int size_of_sprint = 0;
	string row = NULL;
	string it = NULL;
	linux_instance_t system_of_linux_kernel = apsh->os_data;
	string string_of_the_address = NULL;

#ifdef ENABLE_SAFETY_CHECKS
	if (!system_of_linux_kernel) {
		print_error("APSH_ERROR: OS instance not initialized\n");
		goto done;
	}
#endif

	addr -= system_of_linux_kernel->kaslr_offset;

	switch (context->translate_mechanism) {
	case APSH_TM_PROCESS_DTB:
		if (context->dtb != apsh->kpgd)
			goto err;
		break;
	case APSH_TM_PROCESS_PID:
		if (context->pid != 0)
			goto err;
		break;
	default:
		goto err;
	};

	if ((system_of_linux_kernel->sysmap == NULL) ||
	    (strlen(system_of_linux_kernel->sysmap) == 0)) {
		print_error("APSH_WARNING: No configured linux sysmap\n");
		goto done;
	}

	row = g_try_malloc0(MAX_ROW_LENGTH);
	if (!row)
		goto done;

	f = fopen(system_of_linux_kernel->sysmap, "r");
	if (f == NULL) {
		fprintf(stderr,
			"APSH ERROR: could not find System.map file after checking:\n\t%s\nTo fix this problem, add the correct sysmap entry to /etc/APSH.conf\n",
			system_of_linux_kernel->sysmap);
		goto done;
	}
	size_of_sprint = snprintf(NULL, 0, "%" PRIx64 "", addr) + 1;
	string_of_the_address = g_try_malloc0(size_of_sprint);
	if (!string_of_the_address)
		goto done;

	snprintf(string_of_the_address, size_of_sprint, "%" PRIx64 "", addr);
	if (get_symbol_row(f, row, string_of_the_address, 0) == APSH_F)
		goto done;

	for (it = row; *it != 0; it++)
		;
	for (it++; *it != 0; it++)
		;
	it++;

	symbol = strdup(it);

done:
	if (row)
		free(row);
	if (f)
		fclose(f);
	if (string_of_the_address)
		free(string_of_the_address);
	return symbol;

err:
	print_error("APSH_WARNING: Lookup is implemented for kernel symbols only\n");
	return NULL;
}

status_t
linux_symbol_to_address(apsh_instance_t apsh, const string symbol, addr_t *UNUSED(__unused),
			addr_t *addr)
{
	status_t ret = APSH_F;
	linux_instance_t system_of_linux_kernel = apsh->os_data;

	if (system_of_linux_kernel == NULL) {
		print_error("APSH_ERROR: OS instance not initialized\n");
		goto done;
	}

	if (!system_of_linux_kernel->sysmap && !json_profile(apsh)) {
		print_error("APSH_WARNING: No linux sysmap and Rekall profile configured\n");
		goto done;
	}

	if (system_of_linux_kernel->sysmap)
		ret = linux_system_map_symbol_to_address(apsh, symbol, addr);
	else
		ret = json_profile_lookup(apsh, symbol, NULL, addr);

	if (ret == APSH_S)
		*addr += system_of_linux_kernel->kaslr_offset;

done:
	return ret;
}
