/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "config/config_parser.h"
#include "driver/driver_wrapper.h"
#include "peparse.h"
#include "os/windows/windows.h"
#include "regexoob.h"
#include "apsh_instance.h"
#include "ntoskrnl_compiled_regex_rules.h"
#include <stdlib.h>

#define st_stat() status_t
#define copy_string_func strncmp
#define APSH_FUNC_DEFINE(x) x
#define GB_2 (1UL<<31)
#define MB_8 (1UL<<23)
#define COMMAND_MAX_LEN 256
#define MACHINE_I386 0x014c
#define MACHINE_ARM 0x01c0
#define MACHINE_IA64 0x0200
#define MACHINE_AMD64 0x8664
#define MACHINE_EBC 0x0ebc

extern size_t binary_data_len;
extern uint8_t *binary_data;

static inline win_ver_t ntbuild2version(uint16_t ntbuildnumber);

static inline win_ver_t
pe2version(apsh_instance_t apsh, addr_t kernbase_pa, uint16_t *major, uint16_t *minor)
{
	*major = 0;
	*minor = 0;

	uint16_t optional_header_type = 0;
	struct optional_header_pe32 *optional_header_pe32 = NULL;
	struct optional_header_pe32plus *optional_header_pe32plus = NULL;
	uint8_t pe[APSH_PS_4KB];

	ACCESS_CONTEXT(ctx, .addr = kernbase_pa);

	if (peparse_get_image(apsh, &ctx, APSH_PS_4KB, pe) == APSH_F)
		return APSH_OS_WIN_NONE;

	peparse_assign_headers(pe, NULL, NULL, &optional_header_type, NULL, &optional_header_pe32,
			       &optional_header_pe32plus);

	switch (optional_header_type) {
	case IMAGE_PE32_MAGIC:
		*major = optional_header_pe32->major_os_version;
		*minor = optional_header_pe32->minor_os_version;
		break;
	case IMAGE_PE32_PLUS_MAGIC:
		*major = optional_header_pe32plus->major_os_version;
		*minor = optional_header_pe32plus->minor_os_version;
		break;
	default:
		return APSH_OS_WIN_NONE;
	};

	switch (*major) {
	case 3:
	case 4:
		return APSH_OS_WIN_NONE;
	case 5:
		switch (*minor) {
		case 0:
			return APSH_OS_WIN_2000;
		case 1:
			return APSH_OS_WIN_XP;
		case 2:
			return APSH_OS_WIN_2003;
		};
		break;
	case 6:
		switch (*minor) {
		case 0:
			return APSH_OS_WIN_VISTA;
		case 1:
			return APSH_OS_WIN_7;
		case 2:
			return APSH_OS_WIN_8;
		};
		break;
	case 10:
		switch (*minor) {
		case 0:
			return APSH_OS_WIN_10;
		};
		break;
	};

	return APSH_OS_WIN_NONE;
}

static inline st_stat() check_pdbase_offset(apsh_instance_t apsh)
{
	st_stat() ret = APSH_F;
	windows_instance_t windows = apsh->os_data;

	if (!windows->pdbase_offset) {
		if (json_profile_lookup(apsh, "_KPROCESS", "DirectoryTableBase",
						 &windows->pdbase_offset) == APSH_F) {
			debug_print(APSH_DEBUG_MISC, "--win_pdbase is undefined\n");
			goto done;
		}
	}

	ret = APSH_S;

done:
	return ret;
}

st_stat() check_sections(apsh_instance_t apsh, addr_t image_base_p, uint8_t *pe)
{
	struct pe_header *pe_h = NULL;
	struct dos_header *dosh = NULL;
	uint16_t optional_header_type = 0;

	peparse_assign_headers(pe, &dosh, &pe_h, &optional_header_type, NULL, NULL, NULL);
	uint32_t c;

	for (c = 0; c < pe_h->number_of_sections; c++) {
		struct section_header section;
		addr_t section_t = image_base_p + dosh->offset_to_pe + sizeof(struct pe_header) +
				   pe_h->size_of_optional_header +
				   c * sizeof(struct section_header);
		if (apsh_read_pa(apsh, section_t, sizeof(struct section_header),
					 (uint8_t *)&section, NULL) == APSH_F)
			return APSH_F;
		if (!copy_string_func(section.short_name, "INITKDBG", 8))
			return APSH_S;
	}

	return APSH_F;
}

uint32_t
get_image_size(uint16_t optional_header_type, void *optional_pe_header)
{
	if (optional_header_type != IMAGE_PE32_MAGIC &&
	    optional_header_type != IMAGE_PE32_PLUS_MAGIC)
		return 0;
	if (optional_header_type == IMAGE_PE32_MAGIC) {
		struct optional_header_pe32 *oh_pe32 =
		    (struct optional_header_pe32 *)optional_pe_header;
		return oh_pe32->size_of_image;
	} else if (optional_header_type == IMAGE_PE32_PLUS_MAGIC) {
		struct optional_header_pe32plus *oh_pe32plus =
		    (struct optional_header_pe32plus *)optional_pe_header;
		return oh_pe32plus->size_of_image;
	} else {
		return 0;
	}
}

static addr_t
get_ntoskrnl_base_regex(apsh_instance_t apsh)
{
	regexoob_ctx_t rgx_ctx;
	regexoob_result_t *rgx_res;

	if (regexoob_init(binary_data, binary_data_len, apsh->system_dev, apsh->regex_dev, apsh->mem_region_json, &rgx_ctx) != APSH_S) {
		print_error("couldn't start regexoob_ctx\n");
		return 0;
	}

	addr_t ret_addr = 0;
	addr_t regex_current_addr, current_addr;

	ACCESS_CONTEXT(ctx);

	for (regex_current_addr = 0 ; regex_current_addr < (GB_2) ; regex_current_addr += MB_8) {
		if (regexoob_scan(rgx_ctx, regex_current_addr, MB_8, &rgx_res) == APSH_F) {
			print_error("couldn't scan memory with regex\n");
			goto regex_end;
		}

		if (rgx_res != NULL) {

			for (addr_t current_addr_offset = 0 ; current_addr_offset < (rgx_res->phy_add - regex_current_addr) ; current_addr_offset += APSH_PS_4KB) {
				current_addr = regex_current_addr + current_addr_offset;
				{
					ctx.addr = current_addr;
				}
				uint8_t page[APSH_PS_4KB];

				if (peparse_get_image(apsh, &ctx, APSH_PS_4KB, page) == APSH_F)
					continue;

				st_stat() ret = APSH_F;

				void *optional_pe_header = NULL;
				uint16_t optional_header_type = 0;
				struct export_table et;

				peparse_assign_headers(page, NULL, NULL, &optional_header_type, &optional_pe_header,
						NULL, NULL);
				addr_t export_header_offset =
				peparse_get_idd_rva(IMAGE_DIRECTORY_ENTRY_EXPORT, &optional_header_type,
							optional_pe_header, NULL, NULL);
				size_t export_header_size =
				peparse_get_idd_size(IMAGE_DIRECTORY_ENTRY_EXPORT, &optional_header_type,
							optional_pe_header, NULL, NULL);

				if (ctx.addr + export_header_offset <
				ctx.addr + get_image_size(optional_header_type, optional_pe_header)) {
					if (apsh_read_pa(apsh, ctx.addr + export_header_offset,
								sizeof(struct export_table), &et, NULL) == APSH_S &&
					!(et.export_flags || !et.name) && et.name >= export_header_offset &&
					et.name <= export_header_offset + export_header_size) {
						char *name = apsh_read_str_pa(apsh, ctx.addr + et.name);

						if (name != NULL) {
							if (strcmp("ntoskrnl.exe", name) == 0) {
								debug_print(APSH_DEBUG_MISC, "--Starting scan.\n");
								free(name);
								ret_addr = ctx.addr;
								regexoob_result_destroy(rgx_res);
								goto regex_end;
							}

							free(name);
						}
					}
				}

			}
		}

		regexoob_result_destroy(rgx_res);


	}
regex_end:
	regexoob_destroy(rgx_ctx);
	return ret_addr;
}

static bool check_version_block(apsh_instance_t apsh, uint64_t ntoskrnl_pa, uint64_t kd_version_block_offset,
				uint64_t kern_base_offset, uint64_t machine_type_offset)
{
	uint64_t ntoskrnl_va;
	uint16_t machine_type;

	if (apsh_read_64_pa(apsh, ntoskrnl_pa + kd_version_block_offset + kern_base_offset, &ntoskrnl_va) == APSH_F)
		return false;
	if (apsh_read_16_pa(apsh, ntoskrnl_pa + kd_version_block_offset + machine_type_offset, &machine_type) == APSH_F)
		return false;

	if (ntoskrnl_va == 0 ||
	(machine_type != MACHINE_I386 && machine_type != MACHINE_ARM && machine_type != MACHINE_IA64 &&
	 machine_type != MACHINE_AMD64 && machine_type != MACHINE_EBC))
		return false;

	return true;
}

addr_t
get_ntoskrnl_base(apsh_instance_t apsh)
{
	addr_t regex_res = get_ntoskrnl_base_regex(apsh);

	if (regex_res != 0)
		return regex_res;

	addr_t ret_addr = 0;
	addr_t current_addr;
	uint64_t kd_version_block_offset, kern_base_offset, machine_type_offset;

	ACCESS_CONTEXT(ctx);

	if (json_profile_lookup(apsh, "KdVersionBlock", NULL, &kd_version_block_offset) == APSH_F ||
	    json_profile_lookup(apsh, "_DBGKD_GET_VERSION64", "KernBase", &kern_base_offset) == APSH_F ||
	    json_profile_lookup(apsh, "_DBGKD_GET_VERSION64", "MachineType", &machine_type_offset) == APSH_F) {
		print_error("Failed to find field offsets of _DBGKD_GET_VERSION64\n");
		return 0;
	}

	for (current_addr = 0; current_addr < apsh->max_physical_address;
	     current_addr += APSH_PS_4KB) {
		{
			ctx.addr = current_addr;
		}

		uint8_t page[APSH_PS_4KB];

		if (peparse_get_image(apsh, &ctx, APSH_PS_4KB, page) == APSH_F)
			continue;

		st_stat() ret = APSH_F;

		void *optional_pe_header = NULL;
		uint16_t optional_header_type = 0;
		struct export_table et;

		peparse_assign_headers(page, NULL, NULL, &optional_header_type, &optional_pe_header,
				       NULL, NULL);
		addr_t export_header_offset =
		    peparse_get_idd_rva(IMAGE_DIRECTORY_ENTRY_EXPORT, &optional_header_type,
					optional_pe_header, NULL, NULL);
		size_t export_header_size =
		    peparse_get_idd_size(IMAGE_DIRECTORY_ENTRY_EXPORT, &optional_header_type,
					 optional_pe_header, NULL, NULL);

		if (ctx.addr + export_header_offset <
		    ctx.addr + get_image_size(optional_header_type, optional_pe_header)) {
			if (apsh_read_pa(apsh, ctx.addr + export_header_offset,
						 sizeof(struct export_table), &et, NULL) == APSH_S &&
			    !(et.export_flags || !et.name) && et.name >= export_header_offset &&
			    et.name <= export_header_offset + export_header_size) {
				char *name = apsh_read_str_pa(apsh, ctx.addr + et.name);

				if (name != NULL) {
					if (strcmp("ntoskrnl.exe", name) == 0 &&
					    check_version_block(apsh, ctx.addr, kd_version_block_offset,
					    kern_base_offset, machine_type_offset)) {
						debug_print(APSH_DEBUG_MISC, "--Starting scan.\n");
						free(name);
						return ctx.addr;
					}

					free(name);
				}
			}
		}

		if (ret == APSH_F) {
			if (check_sections(apsh, ctx.addr, page) == APSH_S &&
			    check_version_block(apsh, ctx.addr, kd_version_block_offset, kern_base_offset,
			    machine_type_offset))
				return ctx.addr;
		}
	}

	return ret_addr;
}

bool
kernel_debug_search(apsh_instance_t apsh, struct cv_info_pdb70 *pdb_header)
{
	addr_t i;
	addr_t max_mem = apsh_get_max_physical_address(apsh);

	for (i = 0; i < max_mem; i += APSH_PS_4KB) {
		uint8_t pe[APSH_PS_4KB];

		if (apsh_read_pa(apsh, i, APSH_PS_4KB, pe, NULL) == APSH_F)
			continue;

		uint32_t c;

		for (c = 0; c < APSH_PS_4KB - PDB_FILENAME_LENGTH; c++) {
			if (!copy_string_func((char *)&pe[c], "ntkrnlmp.pdb",
					      PDB_FILENAME_LENGTH) ||
			    !copy_string_func((char *)&pe[c], "ntoskrnl.pdb",
					      PDB_FILENAME_LENGTH) ||
			    !copy_string_func((char *)&pe[c], "ntkrnlpa.pdb",
					      PDB_FILENAME_LENGTH) ||
			    !copy_string_func((char *)&pe[c], "ntkrpamp.pdb",
					      PDB_FILENAME_LENGTH)) {
				if (APSH_F ==
				    apsh_read_pa(apsh,
						i + c - 2 * sizeof(uint32_t) - sizeof(struct guid),
						sizeof(struct cv_info_pdb70) + PDB_FILENAME_LENGTH,
						pdb_header, NULL))
					continue;

				if (pdb_header->cv_signature != RSDS)
					continue;
				else
					return 1;
			}
		}
	}
	return 0;
}

st_stat() get_windows_profile(apsh_instance_t apsh, addr_t kernel_base_p, uint8_t *pe)
{
	debug_print(APSH_DEBUG_MISC, "--Kernel base PA@ 0x%" PRIx64 "\n", kernel_base_p);
	ACCESS_CONTEXT(ctx, .addr = kernel_base_p);

	if (peparse_get_image(apsh, &ctx, APSH_PS_4KB, pe) == APSH_F)
		return APSH_F;

	uint32_t image_size;

	bool debug_directory_valid = 0;
	struct pe_header *pe_header = APSH_FUNC_DEFINE(NULL);
	uint16_t optional_header_type = 0;
	struct optional_header_pe32 *optional_header_pe32 = NULL;
	struct optional_header_pe32plus *APSH_FUNC_DEFINE(optional_header_pe32plus = NULL);

	APSH_FUNC_DEFINE(peparse_assign_headers(pe, NULL, &pe_header,
					       APSH_FUNC_DEFINE(&optional_header_type), NULL,
					       &optional_header_pe32, &optional_header_pe32plus));
	addr_t debug_offset =
	    APSH_FUNC_DEFINE(peparse_get_idd_rva(IMAGE_DIRECTORY_ENTRY_DEBUG, NULL, NULL,
						optional_header_pe32, optional_header_pe32plus));

	switch (optional_header_type) {
	case APSH_FUNC_DEFINE(IMAGE_PE32_MAGIC):
		APSH_FUNC_DEFINE(image_size = optional_header_pe32->size_of_image);
		break;
	case APSH_FUNC_DEFINE(IMAGE_PE32_PLUS_MAGIC):
		APSH_FUNC_DEFINE(image_size = optional_header_pe32plus->size_of_image);
		break;
	default:
		return APSH_F;
	}

	APSH_FUNC_DEFINE(struct image_debug_directory debug_directory = {0});
	APSH_FUNC_DEFINE(struct cv_info_pdb70 *pdb_header =
			    g_try_malloc0(sizeof(struct cv_info_pdb70) + PDB_FILENAME_LENGTH + 1));
	if (apsh_read_pa(apsh, kernel_base_p + debug_offset,
				 sizeof(struct image_debug_directory), (uint8_t *)&debug_directory,
				 NULL) == APSH_F) {
		APSH_FUNC_DEFINE(g_free(pdb_header));
		return APSH_F;
	}

	APSH_FUNC_DEFINE(debug_print(APSH_DEBUG_MISC, "\tPE GUID: %.8x%.5x\n",
				    pe_header->time_date_stamp, image_size));

	switch (debug_directory.type) {
	case IMAGE_DEBUG_TYPE_CODEVIEW:

		debug_directory_valid = 1;
		break;
	case IMAGE_DEBUG_TYPE_MISC:
		APSH_FUNC_DEFINE(debug_print(APSH_DEBUG_MISC,
					    "This operating system uses .dbg instead of .pdb\n"));
		goto done;
	default:
		break;
	}

	if (debug_directory_valid) {
		if (debug_directory.size_of_data > APSH_PS_4KB / 4) {
			APSH_FUNC_DEFINE(debug_print(APSH_DEBUG_MISC,
						    "The size of the debug directory is huge, "
						    "something might be wrong.\n"));
			goto done;
		}

		if (apsh_read_pa(apsh, kernel_base_p + debug_directory.address_of_raw_data,
					 sizeof(struct cv_info_pdb70) + PDB_FILENAME_LENGTH,
					 APSH_FUNC_DEFINE(pdb_header), NULL) == APSH_F)
			goto done;

		if (pdb_header->cv_signature != RSDS) {
			APSH_FUNC_DEFINE(debug_print(
			    APSH_DEBUG_MISC, "The CodeView debug information has to be in PDB "
					    "7.0 for the kernel!\n"));
			goto done;
		}
	}

	APSH_FUNC_DEFINE(sprintf(apsh->host_data.pdb_guid, "%.8x%.4x%.4x", pdb_header->signature.data1,
				 pdb_header->signature.data2, pdb_header->signature.data3));

	for (int c = 0; c < 8; c++)
		sprintf((apsh->host_data.pdb_guid + 16 + 2*c), "%.2x", pdb_header->signature.data4[c]);

	sprintf(apsh->host_data.pdb_guid + 32, "%.1x", pdb_header->age & 0xf);
	debug_print(APSH_DEBUG_MISC, "%s\n", apsh->host_data.pdb_guid);

	asprintf(&apsh->host_data.kernel_name, "%s", (char *)pdb_header->pdb_file_name);
	debug_print(APSH_DEBUG_MISC, "\tKernel filename: %s\n", apsh->host_data.kernel_name);

	g_free(pdb_header);
	return APSH_S;

done:
	g_free(pdb_header);
	return APSH_F;
}

st_stat() get_page_mode_pdb(apsh_instance_t apsh, addr_t kernel_base_p, uint8_t *pe)
{

	ACCESS_CONTEXT(ctx, .addr = kernel_base_p);

	if (peparse_get_image(apsh, &ctx, APSH_PS_4KB, pe) == APSH_F)
		return APSH_F;

	uint16_t optional_header_type = 0;
	uint32_t image_size;

	bool debug_directory_valid = 0;
	struct optional_header_pe32 *optional_header_pe32 = APSH_FUNC_DEFINE(NULL);

	APSH_FUNC_DEFINE(struct optional_header_pe32plus *optional_header_pe32plus = NULL);
	APSH_FUNC_DEFINE(struct image_debug_directory debug_directory = {0});
	APSH_FUNC_DEFINE(struct pe_header *pe_header = APSH_FUNC_DEFINE(NULL));

	APSH_FUNC_DEFINE(peparse_assign_headers(
	    pe, NULL, &pe_header, &APSH_FUNC_DEFINE(optional_header_type), NULL,
	    &APSH_FUNC_DEFINE(optional_header_pe32), &optional_header_pe32plus));
	addr_t debug_offset =
	    APSH_FUNC_DEFINE(peparse_get_idd_rva(IMAGE_DIRECTORY_ENTRY_DEBUG, NULL, NULL,
						optional_header_pe32, optional_header_pe32plus));

	switch (APSH_FUNC_DEFINE(optional_header_type)) {
	case APSH_FUNC_DEFINE(IMAGE_PE32_MAGIC):
		APSH_FUNC_DEFINE(image_size = optional_header_pe32->size_of_image);
		break;
	case APSH_FUNC_DEFINE(IMAGE_PE32_PLUS_MAGIC):
		APSH_FUNC_DEFINE(image_size = optional_header_pe32plus->size_of_image);
		break;
	default:
		return APSH_F;
	}

	struct cv_info_pdb70 *pdb_header =
	    g_try_malloc0(sizeof(struct cv_info_pdb70) + APSH_FUNC_DEFINE(PDB_FILENAME_LENGTH) + 1);
	if (apsh_read_pa(apsh, kernel_base_p + debug_offset,
				 sizeof(struct image_debug_directory), (uint8_t *)&debug_directory,
				 NULL) == APSH_F) {
		g_free(pdb_header);
		return APSH_F;
	}

	APSH_FUNC_DEFINE(debug_print(APSH_DEBUG_MISC, "\tPE GUID: %.8x%.5x\n",
				    pe_header->time_date_stamp, image_size));

	switch (debug_directory.type) {
	case IMAGE_DEBUG_TYPE_CODEVIEW:

		debug_directory_valid = 1;
		break;
	case IMAGE_DEBUG_TYPE_MISC:
		debug_print(APSH_DEBUG_MISC, "This operating system uses .dbg instead of .pdb\n");
		goto done;
	default:
		break;
	}

	if (debug_directory_valid) {
		if (debug_directory.size_of_data > APSH_PS_4KB / 4) {

			debug_print(
			    APSH_DEBUG_MISC,
			    "The size of the debug directory is huge, something might be wrong.\n");
			goto done;
		}

		if (apsh_read_pa(apsh, kernel_base_p + debug_directory.address_of_raw_data,
					 sizeof(struct cv_info_pdb70) + PDB_FILENAME_LENGTH,
					 pdb_header, NULL) == APSH_F)
			goto done;

		if (pdb_header->cv_signature != RSDS) {
			debug_print(APSH_DEBUG_MISC,
				    "The CodeView debug information has to be in PDB "
				    "7.0 for the kernel!\n");
			goto done;
		}
	} else {
		if (!kernel_debug_search(apsh, pdb_header))
			goto done;
	}

	debug_print(APSH_DEBUG_MISC, "\tPDB GUID: ");
	debug_print(APSH_DEBUG_MISC, "%.8x", pdb_header->signature.data1);
	debug_print(APSH_DEBUG_MISC, "%.4x", pdb_header->signature.data2);
	debug_print(APSH_DEBUG_MISC, "%.4x", pdb_header->signature.data3);

	int c;

	for (c = 0; c < 8; c++)
		debug_print(APSH_DEBUG_MISC, "%.2x", pdb_header->signature.data4[c]);

	debug_print(APSH_DEBUG_MISC, "%.1x", pdb_header->age & 0xf);
	debug_print(APSH_DEBUG_MISC, "\n");
	debug_print(APSH_DEBUG_MISC, "\tKernel filename: %s\n", (char *)pdb_header->pdb_file_name);

	if (!strcmp("ntoskrnl.pdb", (char *)pdb_header->pdb_file_name)) {
		debug_print(APSH_DEBUG_MISC, "\tSingle-processor without PAE\n");
		apsh->page_mode = APSH_PM_LEGACY;
		g_free(pdb_header);
		return APSH_S;
	} else if (!strcmp("ntkrnlmp.pdb", (char *)pdb_header->pdb_file_name)) {
		debug_print(APSH_DEBUG_MISC, "\tMulti-processor without PAE\n");
		apsh->page_mode = APSH_PM_IA32E;
		g_free(pdb_header);
		return APSH_S;
	} else if (!strcmp("ntkrnlpa.pdb", (char *)pdb_header->pdb_file_name)) {
		debug_print(APSH_DEBUG_MISC,
			    "\tSingle-processor with PAE (version 5.0 and higher)\n");
		apsh->page_mode = APSH_PM_PAE;
		g_free(pdb_header);
		return APSH_S;
	} else if (!strcmp("ntkrpamp.pdb", (char *)pdb_header->pdb_file_name)) {
		debug_print(APSH_DEBUG_MISC,
			    "\tMulti-processor with PAE (version 5.0 and higher)\n");
		apsh->page_mode = APSH_PM_PAE;
		g_free(pdb_header);
		return APSH_S;
	}

done:
	g_free(pdb_header);
	return APSH_F;
}

/* Tries to determine the page mode based on the kpgd found via heuristics */
static st_stat() find_page_mode_unknown(apsh_instance_t apsh)
{
	windows_instance_t windows = apsh->os_data;
	uint8_t pe[APSH_PS_4KB];
	addr_t base = get_ntoskrnl_base(apsh);

	debug_print(APSH_DEBUG_MISC, "--Kernel base PA@ 0x%" PRIx64 "\n", base);
	if (get_page_mode_pdb(apsh, base, pe) == APSH_S)
		return APSH_S;

	return APSH_F;
}

static st_stat() find_page_mode(apsh_instance_t apsh)
{
	windows_instance_t windows = apsh->os_data;

	st_stat() ret = APSH_F;
	uint32_t mask = ~0;

	if (!windows) {
		print_error("Windows functions not initialized in %s\n", __FUNCTION__);
		return APSH_F;
	}

	if (!windows->ntoskrnl || !windows->ntoskrnl_va) {
		print_error("Windows kernel virtual and physical address required for determining "
			    "page mode\n");
		return APSH_F;
	}

	if (!apsh->kpgd) {
		print_error(
		    "Windows kernel directory table base not set, can't determine page mode\n");
		return APSH_F;
	}

	debug_print(APSH_DEBUG_MISC, "--trying APSH_PM_LEGACY\n");
	apsh->page_mode = APSH_PM_LEGACY;

	if (arch_init(apsh) == APSH_S) {
		addr_t test = 0;

		if (apsh_pagetable_lookup(apsh, (apsh->kpgd & mask), windows->ntoskrnl_va,
						  &test) == APSH_S &&
		    test == windows->ntoskrnl) {
			apsh->kpgd &= mask;
			debug_print(APSH_DEBUG_MISC, "--Page mode set to: APSH_PM_LEGACY\n");
			goto found_pm;
		}
	}

	debug_print(APSH_DEBUG_MISC, "--trying APSH_PM_PAE\n");
	apsh->page_mode = APSH_PM_PAE;

	if (arch_init(apsh) == APSH_S) {
		addr_t test = 0;

		if (apsh_pagetable_lookup(apsh, (apsh->kpgd & mask), windows->ntoskrnl_va,
						  &test) == APSH_S &&
		    test == windows->ntoskrnl) {
			apsh->kpgd &= mask;
			debug_print(APSH_DEBUG_MISC, "--Page mode set to: APSH_PM_PAE\n");
			goto found_pm;
		}
	}

	debug_print(APSH_DEBUG_MISC, "--trying APSH_PM_IA32E\n");
	apsh->page_mode = APSH_PM_IA32E;

	if (arch_init(apsh) == APSH_S) {
		addr_t test = 0;

		if (apsh_pagetable_lookup(apsh, apsh->kpgd, windows->ntoskrnl_va, &test) == APSH_S &&
		    test == windows->ntoskrnl) {
			debug_print(APSH_DEBUG_MISC, "--Page mode set to: APSH_PM_IA32E\n");
			goto found_pm;
		}
	}

	goto done;

found_pm:
	ret = APSH_S;

done:
	return ret;
}

static st_stat() dtb_finder(apsh_instance_t apsh)
{
	addr_t dtb;
	windows_instance_t windows = apsh->os_data;
	addr_t lookup;
	addr_t hpa = 0;

	for (dtb = 0x1a0000; dtb < 0x1b0000; dtb += APSH_PS_4KB) {
		if (apsh->EPT) {
			if (apsh_nested_pagetable_lookup(
					 apsh, apsh->EPT, APSH_PM_EPT_4L, dtb, apsh->page_mode,
					 windows->ntoskrnl_va, &lookup, &hpa) == APSH_S) {
				if (lookup == windows->ntoskrnl) {
					debug_print(APSH_DEBUG_MISC,
						    "--DTB Found at @ (0x%.16" PRIx64 "). \n", dtb);
					apsh->kpgd = dtb;
					return APSH_S;
				}
			}
		} else {
			if (APSH_S ==
			    apsh_pagetable_lookup(apsh, dtb, windows->ntoskrnl_va, &lookup)) {
				if (lookup == windows->ntoskrnl) {
					debug_print(APSH_DEBUG_MISC,
						    "--DTB Found at @ (0x%.16" PRIx64 "). \n", dtb);
					apsh->kpgd = dtb;
					return APSH_S;
				}
			}
		}
	}


	for (dtb = 0x0; dtb < 0x1000000; dtb += APSH_PS_4KB) {
		if (APSH_S == apsh_pagetable_lookup(apsh, dtb, windows->ntoskrnl_va, &lookup)) {
			if (lookup == windows->ntoskrnl) {
				debug_print(APSH_DEBUG_MISC, "--DTB Found at @ (0x%.16" PRIx64 ").\n", dtb);
				apsh->kpgd = dtb;
				return APSH_S;
			}
		}
	}

	return APSH_F;
}

static st_stat() get_kpgd_method2(apsh_instance_t apsh)
{
	addr_t sysproc = 0;
	windows_instance_t windows = NULL;

	if (apsh->os_data == NULL) {
		print_error("APSH_ERROR: No OS data initialized\n");
		return APSH_F;
	}

	windows = apsh->os_data;
	sysproc = windows->sysproc;

	if (!sysproc) {
		sysproc = windows_find_eprocess(apsh, "System");
		if (sysproc == 0) {
			debug_print(APSH_DEBUG_MISC, "--failed to find System process.\n");
			goto error_exit;
		}
		printf("LibAPSH Suggestion: set win_sysproc=0x%" PRIx64
		       " in libapsh.conf for faster startup.\n",
		       sysproc);
	}
	debug_print(APSH_DEBUG_MISC, "--got PA to PsInitialSystemProcess (0x%.16" PRIx64 ").\n",
		    sysproc);

	if (apsh_read_64_pa(apsh, sysproc + windows->pdbase_offset, &apsh->kpgd) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to resolve PD for System process\n");
		goto error_exit;
	}

	if (!apsh->kpgd) {
		debug_print(APSH_DEBUG_MISC, "--kpgd was zero\n");
		goto error_exit;
	}

	if (apsh_read_64_pa(apsh, sysproc + windows->tasks_offset, &apsh->init_task) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to resolve address of System process\n");
		goto error_exit;
	}

	apsh->init_task -= windows->tasks_offset;

	switch (apsh->page_mode) {
	case APSH_PM_LEGACY:
	case APSH_PM_PAE: {
		uint32_t mask = ~0;

		apsh->kpgd &= mask;
		apsh->init_task &= mask;
		break;
	}
	default:
		break;
	}

	debug_print(APSH_DEBUG_MISC, "**set kpgd2 (0x%.16" PRIx64 ").\n", apsh->kpgd);
	debug_print(APSH_DEBUG_MISC, "**set init_task (0x%.16" PRIx64 ").\n", apsh->init_task);

	return APSH_S;

error_exit:
	return APSH_F;
}

addr_t
windows_find_cr3(apsh_instance_t apsh)
{
	get_kpgd_method2(apsh);
	return apsh->kpgd;
}

static st_stat() get_kpgd_method1(apsh_instance_t apsh)
{
	addr_t sysproc = 0;
	windows_instance_t windows = NULL;

	if (apsh->os_data == NULL) {
		print_error("APSH_ERROR: No OS data initialized\n");
		return APSH_F;
	}

	windows = apsh->os_data;

	if (apsh_read_addr_ksym(apsh, "PsInitialSystemProcess", &sysproc) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to read pointer for system process\n");
		goto error_exit;
	}

	if (apsh_translate_kv2p(apsh, sysproc, &sysproc) == APSH_F)
		goto error_exit;

	debug_print(APSH_DEBUG_MISC, "--got PA to PsInitialSystemProcess (0x%.16" PRIx64 ").\n",
		    sysproc);

	if (apsh_read_addr_pa(apsh, sysproc + windows->pdbase_offset, &apsh->kpgd) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to resolve pointer for system process\n");
		goto error_exit;
	}

	if (!apsh->kpgd) {
		debug_print(APSH_DEBUG_MISC, "--kpgd was zero\n");
		goto error_exit;
	}
	debug_print(APSH_DEBUG_MISC, "**set kpgd1 (0x%.16" PRIx64 ").\n", apsh->kpgd);

	if (apsh_read_addr_pa(apsh, sysproc + windows->tasks_offset, &apsh->init_task) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to resolve address of System process\n");
		goto error_exit;
	}
	apsh->init_task -= windows->tasks_offset;
	debug_print(APSH_DEBUG_MISC, "**set init_task (0x%.16" PRIx64 ").\n", apsh->init_task);

	return APSH_S;

error_exit:
	return APSH_F;
}

static st_stat() get_kpgd_method0(apsh_instance_t apsh)
{
	addr_t sysproc_va = 0;
	addr_t sysproc_pa = 0;
	addr_t active_process_head = 0;
	windows_instance_t windows = NULL;
	apsh_pid_t pid = 4;
	size_t len = sizeof(apsh_pid_t);
	addr_t kpgd = 0;

	if (apsh->os_data == NULL) {
		print_error("APSH_ERROR: No OS data initialized\n");
		return APSH_F;
	}

	windows = apsh->os_data;

	if (apsh_read_addr_ksym(apsh, "PsActiveProcessHead", &active_process_head) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to resolve PsActiveProcessHead\n");
		goto error_exit;
	}

	debug_print(APSH_DEBUG_MISC,
		    "--starting search from PsActiveProcessHead (0x%.16" PRIx64
		    ") using kpgd (0x%.16" PRIx64 ").\n",
		    active_process_head, apsh->kpgd);

	sysproc_va = eprocess_list_search(apsh, active_process_head - windows->tasks_offset,
					  windows->pid_offset, len, &pid);

	if (sysproc_va == 0) {
		debug_print(APSH_DEBUG_MISC, "--failed to find system process with pid 4\n");
		goto error_exit;
	}

	sysproc_va -= windows->tasks_offset;
	debug_print(APSH_DEBUG_MISC, "--Found System process at %lx\n", sysproc_va);
	if (apsh_translate_kv2p(apsh, sysproc_va, &sysproc_pa) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to translate System process\n");
		goto error_exit;
	}

	debug_print(APSH_DEBUG_MISC, "--Found System process physical address at %lx\n", sysproc_pa);

	if (apsh_read_addr_pa(apsh, sysproc_pa + windows->pdbase_offset, &kpgd) == APSH_F) {
		debug_print(APSH_DEBUG_MISC, "--failed to resolve pointer for system process\n");
		goto error_exit;
	}

	if (!kpgd) {
		debug_print(APSH_DEBUG_MISC, "--kpgd was zero\n");
		goto error_exit;
	}
	apsh->kpgd = kpgd;
	debug_print(APSH_DEBUG_MISC, "**set kpgd0 (0x%.16" PRIx64 ").\n", apsh->kpgd);

	apsh->init_task = sysproc_va;
	debug_print(APSH_DEBUG_MISC, "**set init_task (0x%.16" PRIx64 ").\n", apsh->init_task);

	return APSH_S;

error_exit:
	return APSH_F;
}

st_stat() windows_get_kernel_struct_offset(apsh_instance_t apsh, const char *symbol,
					   const char *member, addr_t *addr)
{
	return json_profile_lookup(apsh, symbol, member, addr);
}

#define windows_get_offset_check(x) (strncmp(offset_name, x, max_length) == 0)

st_stat() windows_get_offset(apsh_instance_t apsh, const char *offset_name, addr_t *offset)
{
	const size_t max_length = 100;
	windows_instance_t windows = apsh->os_data;

	if (windows == NULL) {
		print_error("APSH_ERROR: OS instance not initialized\n");
		return APSH_F;
	}

	if (windows_get_offset_check("win_ntoskrnl")) {
		*offset = windows->ntoskrnl;
		return APSH_S;
	} else if (windows_get_offset_check("win_ntoskrnl_va")) {
		*offset = windows->ntoskrnl_va;
		return APSH_S;
	} else if (windows_get_offset_check("win_tasks")) {
		*offset = windows->tasks_offset;
		return APSH_S;
	} else if (windows_get_offset_check("win_pdbase")) {
		*offset = windows->pdbase_offset;
		return APSH_S;
	} else if (windows_get_offset_check("win_pid")) {
		*offset = windows->pid_offset;
		return APSH_S;
	} else if (windows_get_offset_check("win_pname")) {
		if (windows->pname_offset == 0) {
			windows->pname_offset = find_pname_offset(apsh, NULL);
			if (windows->pname_offset == 0) {
				debug_print(APSH_DEBUG_MISC, "--failed to find pname_offset\n");
				return APSH_F;
			}
		}

		*offset = windows->pname_offset;
		return APSH_S;
	} else if (windows_get_offset_check("win_kdvb")) {
		*offset = windows->kdbg_va;
		return APSH_S;
	} else if (windows_get_offset_check("win_sysproc")) {
		*offset = windows->sysproc;
		return APSH_S;
	} else if (windows_get_offset_check("win_kpcr")) {
		*offset = windows->kpcr_offset;
		return APSH_S;
	} else if (windows_get_offset_check("win_kdbg")) {
		*offset = windows->kdbg_offset;
		return APSH_S;
	} else if (windows_get_offset_check("kpgd")) {
		*offset = apsh->kpgd;
		return APSH_S;
	}

	print_warning("Invalid offset name in windows_get_offset (%s).\n", offset_name);
	return APSH_F;
}

void
windows_read_config_ghashtable_entries(char *key, gpointer value, apsh_instance_t apsh)
{

	windows_instance_t windows_instance = apsh->os_data;

	if (copy_string_func(key, "win_ntoskrnl", CONFIG_STR_LENGTH) == 0) {
		windows_instance->ntoskrnl = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_ntoskrnl_va", CONFIG_STR_LENGTH) == 0) {
		windows_instance->ntoskrnl_va = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_tasks", CONFIG_STR_LENGTH) == 0) {
		windows_instance->tasks_offset = *(int *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_pdbase", CONFIG_STR_LENGTH) == 0) {
		windows_instance->pdbase_offset = *(int *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_pid", CONFIG_STR_LENGTH) == 0) {
		windows_instance->pid_offset = *(int *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_pname", CONFIG_STR_LENGTH) == 0) {
		windows_instance->pname_offset = *(int *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_kdvb", CONFIG_STR_LENGTH) == 0) {
		windows_instance->kdbg_va = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_sysproc", CONFIG_STR_LENGTH) == 0) {
		windows_instance->sysproc = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_kpcr", CONFIG_STR_LENGTH) == 0) {
		windows_instance->kpcr_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "win_kdbg", CONFIG_STR_LENGTH) == 0) {
		windows_instance->kdbg_offset = *(addr_t *)value;
		goto _done;
	}

	if (copy_string_func(key, "kpgd", CONFIG_STR_LENGTH) == 0) {
		apsh->kpgd = *(addr_t *)value;
		goto _done;
	}

_done:
	return;
}

static st_stat() get_kpgd_from_json_profile(apsh_instance_t apsh)
{
	st_stat() ret = APSH_F;
	windows_instance_t windows = apsh->os_data;
	addr_t sysproc_pointer_rva = 0;
	addr_t sysproc_pdbase_addr = 0;

	if (!json_profile(apsh) || !windows->ntoskrnl || !windows->pdbase_offset || !apsh->kpgd)
		return ret;

	if (apsh->kpgd && windows->sysproc)
		return APSH_S;

	debug_print(APSH_DEBUG_MISC, "**Getting kernel page directory from JSON profile\n");

	if (!windows->sysproc) {
		ret =
		    json_profile_lookup(apsh, "PsInitialSystemProcess", NULL, &sysproc_pointer_rva);
		if (ret == APSH_F)
			return ret;

		if (apsh_pagetable_lookup(apsh, apsh->kpgd,
						  windows->ntoskrnl_va + sysproc_pointer_rva,
						  &sysproc_pdbase_addr) == APSH_F)
			return ret;

		ret = apsh_read_addr_pa(apsh, sysproc_pdbase_addr, &windows->sysproc);
		if (ret == APSH_F)
			return ret;

		debug_print(APSH_DEBUG_MISC, "**Found PsInitialSystemProcess at 0x%lx\n",
			    windows->sysproc);
	}

	if (apsh_pagetable_lookup(apsh, apsh->kpgd, windows->sysproc + windows->pdbase_offset,
					  &sysproc_pdbase_addr) == APSH_F)
		return APSH_F;

	ret = apsh_read_addr_pa(apsh, sysproc_pdbase_addr, &apsh->kpgd);
	if (ret == APSH_S && apsh->kpgd)
		return APSH_S;

	return APSH_F;
}

static st_stat() find_windows_version_from_json_profile(apsh_instance_t apsh)
{
	st_stat() ret = APSH_F;
	windows_instance_t windows = apsh->os_data;
	addr_t ntbuildnumber_rva;
	uint16_t ntbuildnumber = 0;

	if (!windows->ntoskrnl)
		goto done;

	if (json_profile_lookup(apsh, "NtBuildNumber", NULL, &ntbuildnumber_rva) == APSH_F ||
	    apsh_read_16_pa(apsh, windows->ntoskrnl + ntbuildnumber_rva, &ntbuildnumber) == APSH_F)
		goto done;

	windows->version = ntbuild2version(ntbuildnumber);
	windows->build = ntbuildnumber;

	if (windows->version == APSH_OS_WIN_UNKNOWN) {
		windows->version =
		    pe2version(apsh, windows->ntoskrnl, &windows->major, &windows->minor);
		if (windows->version == APSH_OS_WIN_NONE) {
			debug_print(APSH_DEBUG_MISC,
				    "Failed to find a known version of Windows, "
				    "the JSON Profile may be incorrect for this VM or "
				    "the version of Windows is not supported!\n");
			goto done;
		}
	}

	debug_print(APSH_DEBUG_MISC, "---Profile Windows version from profile: %u\n",
		    windows->version);

	ret = APSH_S;

done:
	return ret;
}

static st_stat() kpcr_find1(apsh_instance_t apsh, windows_instance_t windows, reg_t kpcr_reg)
{
	debug_print(APSH_DEBUG_MISC, "** Trying kpcr_find1\n");

	addr_t kpcr_rva;

	if (json_profile_lookup(apsh, "KiInitialPCR", NULL, &kpcr_rva) == APSH_F)
		return APSH_F;

	if (kpcr_reg < kpcr_rva) {
		debug_print(APSH_DEBUG_MISC,
			    "**vCPU0 doesn't seem to have KiInitialPCR mapped,"
			    " (kpcr < kpcr_rva) (KiInitialPCR) can't init from JSON profile. "
			    "Kpcr=0x%" PRIx64 "kpcr_rva=0x%" PRIx64 "\n",
			    kpcr_reg, kpcr_rva);
		return APSH_F;
	}

	if (apsh->page_mode == APSH_PM_IA32E && kpcr_reg < 0xffff800000000000) {
		debug_print(APSH_DEBUG_MISC,
			    "**Error while init from JSON profile. Getting KPCR from "
			    "user mode or just after syscall before 'swapgs'.\n");
		debug_print(APSH_DEBUG_MISC,
			    "**vCPU0 doesn't seem to have KiInitialPCR mapped,"
			    " can't init from JSON profile. Kpcr=0x%" PRIx64 ", kpcr_rva=0x%" PRIx64
			    "\n",
			    kpcr_reg, kpcr_rva);
		return APSH_F;
	}

	windows->ntoskrnl_va = kpcr_reg - kpcr_rva;

	return APSH_S;
}

static st_stat() kpcr_find4(apsh_instance_t apsh, windows_instance_t windows, reg_t kpcr_reg)
{
	debug_print(APSH_DEBUG_MISC, "** Trying kpcr_find4\n");

	if (kpcr_reg != 0x00000000ffdff000)
		return APSH_F;

	addr_t kdvb = 0, kdvb_offset = 0, kernbase_offset = 0;

	if (json_profile_lookup(apsh, "_KPCR", "KdVersionBlock", &kdvb_offset) == APSH_F)
		return APSH_F;
	if (json_profile_lookup(apsh, "_DBGKD_GET_VERSION64", "KernBase", &kernbase_offset) == APSH_F)
		return APSH_F;
	if (apsh_read_addr_va(apsh, kpcr_reg + kdvb_offset, 0, &kdvb) == APSH_F)
		return APSH_F;
	if (apsh_read_addr_va(apsh, kdvb + kernbase_offset, 0, &windows->ntoskrnl_va) == APSH_F)
		return APSH_F;

	return APSH_S;
}

static st_stat() init_from_json_profile_real(apsh_instance_t apsh, reg_t kpcr_register_to_use)
{
	st_stat() ret = APSH_F;
	windows_instance_t system = apsh->os_data;

	debug_print(APSH_DEBUG_MISC, "**Trying to init from JSON profile\n");

	if (kpcr_register_to_use) {
		reg_t kpcr_reg = 0;

		if (kpcr_find1(apsh, system, kpcr_reg) == APSH_S) {
		} else if (kpcr_find4(apsh, system, kpcr_reg) == APSH_S) {
		} else
			goto done;

		if (APSH_FUNC_DEFINE(APSH_F) ==
			apsh_translate_kv2p(apsh, system->ntoskrnl_va, &system->ntoskrnl) ||
		    !system->ntoskrnl) {
			windows_find_cr3(apsh);
			if (APSH_F == APSH_FUNC_DEFINE(apsh_translate_kv2p(apsh, system->ntoskrnl_va,
									&system->ntoskrnl)))
				goto done;
		}

		APSH_FUNC_DEFINE(
		    debug_print(APSH_DEBUG_MISC, "**KernBase PA=0x%" PRIx64 "\n", system->ntoskrnl));
	}

	if (!system->ntoskrnl || !system->ntoskrnl_va) {
		if (!system->ntoskrnl)
			APSH_FUNC_DEFINE(system->ntoskrnl = get_ntoskrnl_base(apsh));

		addr_t kdvb = APSH_FUNC_DEFINE(0), kernbase_offset = APSH_FUNC_DEFINE(0);

		if (json_profile_lookup(apsh, "KdVersionBlock", NULL, &kdvb) == APSH_F)
			goto done;
		if (APSH_F ==
		    json_profile_lookup(apsh, "_DBGKD_GET_VERSION64", "KernBase", &kernbase_offset))
			goto done;

		APSH_FUNC_DEFINE(debug_print(APSH_DEBUG_MISC,
					    "**KdVersionBlock RVA 0x%lx. KernBase RVA: 0x%lx\n",
					    kdvb, kernbase_offset));
		APSH_FUNC_DEFINE(
		    debug_print(APSH_DEBUG_MISC, "**KernBase PA=0x%" PRIx64 "\n", system->ntoskrnl));

		if (system->ntoskrnl && kdvb && kernbase_offset) {
			if (apsh_read_addr_pa(apsh, system->ntoskrnl + kdvb + kernbase_offset,
						      &system->ntoskrnl_va) == APSH_F)
				goto done;

			if (!system->ntoskrnl_va) {
				if (APSH_F ==
				    apsh_read_32_pa(apsh, system->ntoskrnl + kdvb + kernbase_offset,
						   (uint32_t *)&system->ntoskrnl_va))
					goto done;
			}

			if (!system->ntoskrnl_va) {
				APSH_FUNC_DEFINE(debug_print(
				    APSH_DEBUG_MISC,
				    "**failed to find Windows kernel VA via KdVersionBlock\n"));
				goto done;
			}
		} else {
			APSH_FUNC_DEFINE(debug_print(
			    APSH_DEBUG_MISC,
			    "**Failed to find required offsets and/or kernel base PA\n"));
			goto done;
		}
	}

	debug_print(APSH_DEBUG_MISC, "**KernBase VA=0x%" PRIx64 "\n", system->ntoskrnl_va);

	if (!system->pdbase_offset) {
		if (APSH_F == json_profile_lookup(apsh, "_KPROCESS", "DirectoryTableBase",
						 &system->pdbase_offset)) {
			goto done;
		}
	}
	if (!system->tasks_offset) {
		if (APSH_F == json_profile_lookup(apsh, "_EPROCESS", "ActiveProcessLinks",
						 &system->tasks_offset)) {
			goto done;
		}
	}
	if (!system->pid_offset) {
		if (APSH_F ==
		    json_profile_lookup(apsh, "_EPROCESS", "UniqueProcessId", &system->pid_offset)) {
			goto done;
		}
	}
	if (!system->pname_offset) {
		if (APSH_F ==
		    json_profile_lookup(apsh, "_EPROCESS", "ImageFileName", &system->pname_offset)) {
			goto done;
		}
	}

	ret = APSH_S;
	debug_print(APSH_DEBUG_MISC, "**init from JSON profile success\n");

done:
	return ret;
}

static st_stat() init_from_json_profile(apsh_instance_t apsh)
{
	st_stat() ret = APSH_F;
	windows_instance_t windows = apsh->os_data;

	if (windows->ntoskrnl && windows->ntoskrnl_va && apsh->mode != APSH_FILE) {
		ret = APSH_S;
		goto done;
	}

	if (apsh->mode != APSH_FILE) {
		switch (apsh->page_mode) {
		case APSH_PM_IA32E: {

			const reg_t kpcr_registers_to_try[] = {GS_BASE, SHADOW_GS,
							       MSR_SHADOW_GS_BASE, FS_BASE};
#ifdef APSH_DEBUG
			const char *kpcr_registers_to_try_names[] = {
			    "GS_BASE", "SHADOW_GS", "MSR_SHADOW_GS_BASE", "FS_BASE"};
#endif
			debug_print(APSH_DEBUG_MISC,
				    "** (apsh->page_mode == APSH_PM_IA32E) Entering KPCR "
				    "register selection loop...\n");
			{
				size_t i = 0;

				for (i = 0; i < sizeof(kpcr_registers_to_try) /
						    sizeof(kpcr_registers_to_try[0]);
				     ++i) {
					debug_print(APSH_DEBUG_MISC,
						    "** (apsh->page_mode == APSH_PM_IA32E) => Using "
						    "kpcr_register_to_use=%s.\n",
						    kpcr_registers_to_try[i]);

					ret = init_from_json_profile_real(apsh,
									  kpcr_registers_to_try[i]);
					if (ret != APSH_F)
						goto done;

					debug_print(APSH_DEBUG_MISC, "** Trying %s failed.\n",
						    kpcr_registers_to_try_names[i]);
				}
			}

			debug_print(APSH_DEBUG_MISC, "** KPCR register selection loop ended.\n");
			goto done;
		}

		case APSH_PM_LEGACY:
		case APSH_PM_PAE:
			debug_print(APSH_DEBUG_MISC,
				    "** apsh->page_mode in {APSH_PM_LEGACY, APSH_PM_PAE} "
				    "=> Trying FS_BASE.\n");
			ret = init_from_json_profile_real(apsh, FS_BASE);
			goto done;

		default:
			debug_print(APSH_DEBUG_MISC,
				    "** apsh->page_mode is unhandled, no KPCR init.\n");
			goto done;
		};
	} else {
		debug_print(APSH_DEBUG_MISC, "** Not retrieving KPCR via 'switch', setting "
					    "kpcr_register_to_use to unused = 0 .\n");

		reg_t unused = 0;

		ret = init_from_json_profile_real(apsh, unused);
		goto done;
	}

done:
	if (ret == APSH_S)
		ret = find_windows_version_from_json_profile(apsh);

	return ret;
}

static st_stat() init_core(apsh_instance_t apsh)
{
	st_stat() ret = init_from_json_profile(apsh);

	if (ret == APSH_S)
		goto done;

	ret = init_from_kdbg(apsh);
	if (ret == APSH_F)
		goto done;

	debug_print(APSH_DEBUG_MISC, "** Retrying init from Rekall profile.\n");
	ret = init_from_json_profile_real(apsh, (reg_t)0);

done:
	return ret;
}

st_stat() windows_init(apsh_instance_t apsh, GHashTable *config, addr_t EPT)
{
	st_stat() status = APSH_F;
	windows_instance_t windows = NULL;
	os_interface_t int_os = NULL;

	st_stat() real_kpgd_found = APSH_S;

	if (EPT && EPT != 0)
		apsh->EPT = EPT;

	if (!config) {
		print_error("APSH_ERROR: No config table found\n");
		return APSH_F;
	}

	if (apsh->os_data != NULL) {
		print_error("APSH_ERROR: os data already initialized, resetting\n");
		memset(apsh->os_data, 0, sizeof(struct windows_instance));
	} else {
		apsh->os_data = g_try_malloc0(sizeof(struct windows_instance));
		if (!apsh->os_data)
			return APSH_F;
	}

	windows = apsh->os_data;
	windows->version = APSH_OS_WIN_UNKNOWN;

	g_hash_table_foreach(config, (GHFunc)windows_read_config_ghashtable_entries, apsh);

	int_os = g_try_malloc0(sizeof(struct os_interface));
	if (!int_os)
		goto done;

	memset(int_os, 0, sizeof(struct os_interface));
	int_os->os_get_kernel_struct_offset = windows_get_kernel_struct_offset;
	int_os->os_get_offset = windows_get_offset;
	int_os->os_pid_to_pgd = windows_pid_to_pgd;
	int_os->os_pgd_to_pid = windows_pgd_to_pid;
	int_os->os_ksym2v = windows_kernel_symbol_to_address;
	int_os->os_usym2rva = windows_export_to_rva;
	int_os->os_v2sym = windows_rva_to_export;
	int_os->os_v2ksym = NULL;
	int_os->os_read_unicode_struct = windows_read_unicode_struct;
	int_os->os_read_unicode_struct_pm = windows_read_unicode_struct_pm;
	int_os->os_teardown = windows_teardown;

	apsh->os_interface = int_os;

	if (check_pdbase_offset(apsh) == APSH_F)
		goto done;

	if (apsh->page_mode == APSH_PM_UNKNOWN) {

		if (!windows->ntoskrnl) {
			if (find_page_mode_unknown(apsh) == APSH_F) {
				print_error("Failed to find correct page mode.\n");
				goto done;
			}
		} else {
			uint8_t pe[APSH_PS_4KB];

			if (get_page_mode_pdb(apsh, windows->ntoskrnl, pe) == APSH_F) {
				print_error("Failed to find correct page mode given kernel pa.\n");
				goto done;
			}
		}
	}

	if (init_core(apsh) == APSH_F)
		goto done;

	if (dtb_finder(apsh) == APSH_S) {
		debug_print(APSH_DEBUG_MISC, "--Found to find DTB the fast way\n");
		status = APSH_S;
		debug_print(APSH_DEBUG_MISC, "**KernBase VA=0x%" PRIx64 "\n", windows->ntoskrnl_va);
		goto done;
	}

	if (get_kpgd_from_json_profile(apsh) == APSH_S) {
		debug_print(APSH_DEBUG_MISC, "--kpgd from JSON profile success\n");
		status = APSH_S;
		goto done;
	}

	if (get_kpgd_method0(apsh) == APSH_S) {
		debug_print(APSH_DEBUG_MISC, "--kpgd method0 success\n");
		status = APSH_S;
		goto done;
	}
	if (get_kpgd_method1(apsh) == APSH_S) {
		debug_print(APSH_DEBUG_MISC, "--kpgd method1 success\n");
		status = APSH_S;
		goto done;
	}

	if (get_kpgd_method2(apsh) == APSH_S) {
		debug_print(APSH_DEBUG_MISC, "--kpgd method2 success\n");
		status = APSH_S;
		goto done;
	}

	apsh->kpgd = 0;
	print_error("Failed to find kernel page directory.\n");

done:
	if (status == APSH_F)
		windows_teardown(apsh);
	else
		apsh->x86.transition_pages = true;

	return status;
}

st_stat() windows_teardown(apsh_instance_t apsh)
{

	st_stat() ret = APSH_S;
	windows_instance_t windows = apsh->os_data;

	if (!windows)
		goto done;

	g_free(apsh->os_data);
	apsh->os_data = NULL;

	if (apsh->host_data.kernel_name)
		free(apsh->host_data.kernel_name);


done:
	return ret;
}

static inline win_ver_t
ntbuild2version(uint16_t ntbuildnumber)
{
	switch (ntbuildnumber) {
	case 2195:
		return APSH_OS_WIN_2000;
	case 2600:
	case 3790:
		return APSH_OS_WIN_XP;
	case 6000:
	case 6001:
	case 6002:
		return APSH_OS_WIN_VISTA;
	case 7600:
	case 7601:
		return APSH_OS_WIN_7;
	case 9200:
	case 9600:
		return APSH_OS_WIN_8;
	case 10240:
	case 10586:
	case 14393:
	case 18432:
	case 19041:
		return APSH_OS_WIN_10;
	case 22000:
	case 22621:
		return APSH_OS_WIN_11;
	default:
		break;
	}
	return APSH_OS_WIN_UNKNOWN;
};
