/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "apsh_instance.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "driver/driver_wrapper.h"
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>

#define address_to_t addr_t

status_t
windows_kernel_symbol_to_address(apsh_instance_t apsh, const char *symbol,
				 address_to_t *kernel_base_address, address_to_t *address)
{
	status_t ret = APSH_F;
	address_to_t rva = 0;
	windows_instance_t windows = apsh->os_data;

	if (windows == NULL || !windows->ntoskrnl_va)
		goto exit;

	debug_print(APSH_DEBUG_MISC, "--windows symbol lookup (%s)\n", symbol);

	if (json_profile_lookup(apsh, symbol, NULL, &rva) == APSH_S) {
		*address = windows->ntoskrnl_va + rva;
		debug_print(APSH_DEBUG_MISC,
			    "--got symbol from JSON profile (%s --> 0x%.16" PRIx64 ").\n", symbol,
			    *address);
		ret = APSH_S;
		goto success;
	}

	if (windows_kdbg_lookup(apsh, symbol, address) == APSH_S) {
		debug_print(APSH_DEBUG_MISC, "--got symbol from kdbg (%s --> 0x%" PRIx64 ").\n",
			    symbol, *address);
		ret = APSH_S;
		goto success;
	}

	debug_print(APSH_DEBUG_MISC, "--kdbg lookup failed\n");
	debug_print(APSH_DEBUG_MISC, "--trying kernel PE export table\n");

	ACCESS_CONTEXT(ctx, .pm = apsh->page_mode, .translate_mechanism = APSH_TM_PROCESS_PID,
		       .addr = windows->ntoskrnl_va);

	if (windows_export_to_rva(apsh, &ctx, symbol, &rva) == APSH_S) {
		*address = windows->ntoskrnl_va + rva;
		debug_print(APSH_DEBUG_MISC,
			    "--got symbol from PE export table (%s --> 0x%.16" PRIx64 ").\n",
			    symbol, *address);
		ret = APSH_S;
		goto success;
	}

	debug_print(APSH_DEBUG_MISC, "--kernel PE export table failed\n");

	goto exit;

success:
	if (kernel_base_address)
		*kernel_base_address = windows->ntoskrnl_va;

exit:
	return ret;
}

status_t
windows_pgd_to_pid(apsh_instance_t apsh, address_to_t pgd, apsh_pid_t *pid)
{
	status_t ret = APSH_F;
	address_to_t eprocess = 0;
	int tasks_offset = 0;
	int pid_offset = 0;
	windows_instance_t windows = apsh->os_data;

	if (!apsh->os_data)
		return APSH_F;

	tasks_offset = windows->tasks_offset;
	pid_offset = windows->pid_offset;

	eprocess = windows_find_eprocess_list_pgd(apsh, pgd);
	if (!eprocess) {
		print_error("Could not find EPROCESS struct for pgd = 0x%" PRIx64 ".\n", pgd);
		goto error_exit;
	}

	ret = apsh_read_32_va(apsh, eprocess + pid_offset - tasks_offset, 0, (uint32_t *)pid);

error_exit:
	return ret;
}

status_t
windows_pid_to_pgd(apsh_instance_t apsh, apsh_pid_t pid, address_to_t *dtb)
{
	status_t ret = APSH_F;
	address_to_t eprocess = 0;
	int tasks_offset = 0;
	int pdbase_offset = 0;
	windows_instance_t windows = apsh->os_data;

	if (!apsh->os_data)
		return APSH_F;

	tasks_offset = windows->tasks_offset;
	pdbase_offset = windows->pdbase_offset;

	eprocess = windows_find_eprocess_list_pid(apsh, pid);
	if (!eprocess) {
		debug_print(APSH_DEBUG_MISC, "Could not find EPROCESS struct for pid = %d.\n", pid);
		goto error_exit;
	}

	ret = apsh_read_addr_va(apsh, eprocess + pdbase_offset - tasks_offset, 0, dtb);

error_exit:
	return ret;
}
