/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <string.h>

#include "peparse.h"
#include "apsh_instance.h"

#define st_stat() status_t
#define address_to_t addr_t

void
dump_exports(apsh_instance_t apsh, struct export_table *export_table_inst, const access_context_t *context)
{
	access_context_t _ctx = *context;
	address_to_t base_addr = context->addr;
	address_to_t base1 = base_addr + export_table_inst->address_of_names;
	address_to_t base2 = base_addr + export_table_inst->address_of_name_ordinals;
	address_to_t base3 = base_addr + export_table_inst->address_of_functions;
	uint32_t i = 0;

	for (; i < export_table_inst->number_of_names; ++i) {
		uint32_t rva = 0;
		uint16_t ordinal = 0;
		uint32_t loc = 0;
		char *str = NULL;

		_ctx.addr = base1 + i * sizeof(uint32_t);
		if (apsh_read_32(apsh, &_ctx, &rva) == APSH_F)
			continue;

		if (rva) {
			_ctx.addr = base_addr + rva;
			str = apsh_read_str(apsh, &_ctx);
			if (str) {
				_ctx.addr = base2 + i * sizeof(uint16_t);
				if (apsh_read_16(apsh, &_ctx, &ordinal) == APSH_F) {
					free(str);
					continue;
				}

				_ctx.addr = base3 + ordinal * sizeof(uint32_t);
				if (apsh_read_32(apsh, &_ctx, &loc) == APSH_F) {
					free(str);
					continue;
				}

				printf("%s:%d:0x%" PRIx64 "\n", str, ordinal, base_addr + loc);
				free(str);
			}
		}
	}
}

st_stat() get_export_rva(apsh_instance_t apsh, address_to_t *rva, int aof_index,
			 struct export_table *export_table_inst, const access_context_t *context)
{
	access_context_t _ctx = *context;

	_ctx.addr += export_table_inst->address_of_functions + aof_index * sizeof(uint32_t);

	uint32_t tmp = 0;

	if (apsh_read_32(apsh, &_ctx, &tmp) == APSH_S) {
		*rva = (address_to_t)tmp;
		return APSH_S;
	}

	return APSH_F;
}

int get_aof_index(apsh_instance_t apsh, int aon_index, struct export_table *export_table_inst,
		  const access_context_t *context);

int
get_aon_index_linear(apsh_instance_t apsh, const char *symbol, struct export_table *export_table_inst,
		     const access_context_t *context)
{
	access_context_t _ctx = *context;
	uint32_t i = 0;

	for (; i < export_table_inst->number_of_names; ++i) {
		_ctx.addr = context->addr + export_table_inst->address_of_names + i * sizeof(uint32_t);
		uint32_t str_rva = 0;

		if (apsh_read_32(apsh, &_ctx, &str_rva) == APSH_S && str_rva) {
			_ctx.addr = context->addr + str_rva;
			char *rva = apsh_read_str(apsh, &_ctx);

			if (rva != NULL) {
				if (strncmp(rva, symbol, strlen(rva)) == 0) {
					free(rva);
					return (int)i;
				}
			}
			free(rva);
		}
	}

	return -1;
}

static int find_aon_idx_bin(apsh_instance_t apsh, const char *symbol, address_to_t aon_base_va,
			    int low, int high, const access_context_t *context);

int
get_aon_index_binary(apsh_instance_t apsh, const char *symbol, struct export_table *export_table_inst,
		     const access_context_t *context)
{
	address_to_t aon_base_addr = context->addr + export_table_inst->address_of_names;
	int name_ct = export_table_inst->number_of_names;

	return find_aon_idx_bin(apsh, symbol, aon_base_addr, 0, name_ct - 1, context);
}

int get_aon_index(apsh_instance_t apsh, const char *symbol, struct export_table *export_table_inst,
		  const access_context_t *context);

st_stat() peparse_get_image(apsh_instance_t apsh, const access_context_t *context, size_t len,
			    const uint8_t *const image_array)
{
	if (apsh_read(apsh, context, len, (void *)image_array, NULL) == APSH_F) {
		if (context->addr == 0x208960)
			debug_print(APSH_DEBUG_MISC, "--PEPARSE: failed to read PE header\n");
		return APSH_F;
	}

	if (peparse_validate_pe_image(image_array, len) != APSH_S) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEPARSE: failed to validate PE header(s)\n");
		return APSH_F;
	}

	return APSH_S;
}

void peparse_assign_headers(const uint8_t *const image_array, struct dos_header **dos_header,
			    struct pe_header **pe_header, uint16_t *optional_header_type,
			    void **optional_pe_header, struct optional_header_pe32 **pe32_of_the_oh,
			    struct optional_header_pe32plus **pe32_of_the_oh_plus);

address_to_t
peparse_get_idd_rva(uint32_t entry_id, uint16_t *optional_header_type, void *header_that_is_optional,
		    struct optional_header_pe32 *pe32_of_the_oh,
		    struct optional_header_pe32plus *pe32_of_the_oh_plus)
{

	address_to_t rva = 0;

	if (optional_header_type == NULL) {

		if (pe32_of_the_oh != NULL && pe32_of_the_oh->number_of_rva_and_sizes >= entry_id) {
			rva = pe32_of_the_oh->idd[entry_id].virtual_address;
			goto done;
		}

		if (pe32_of_the_oh_plus != NULL &&
		    pe32_of_the_oh_plus->number_of_rva_and_sizes >= entry_id) {
			rva = pe32_of_the_oh_plus->idd[entry_id].virtual_address;
			goto done;
		}
	} else if (header_that_is_optional) {

		switch (*optional_header_type) {

		case IMAGE_PE32_PLUS_MAGIC: {
			struct optional_header_pe32plus *oh_pe32plus_t =
			    (struct optional_header_pe32plus *)header_that_is_optional;
			if (oh_pe32plus_t->number_of_rva_and_sizes >= entry_id)
				rva = oh_pe32plus_t->idd[entry_id].virtual_address;

			break;
		}

		case IMAGE_PE32_MAGIC: {
			struct optional_header_pe32 *oh_pe32_t =
			    (struct optional_header_pe32 *)header_that_is_optional;
			if (oh_pe32_t->number_of_rva_and_sizes >= entry_id)
				rva = oh_pe32_t->idd[entry_id].virtual_address;

			break;
		}

		default:
			break;
		}
	}

done:
	if (rva == 0)
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: Image data directory RVA is 0\n");

	return rva;
}

size_t peparse_get_idd_size(uint32_t entry_id, uint16_t *optional_header_type,
			    void *header_that_is_optional, struct optional_header_pe32 *pe32_of_the_oh,
			    struct optional_header_pe32plus *pe32_of_the_oh_plus);
void peparse_assign_headers(const uint8_t *const image_array, struct dos_header **dos_header,
			    struct pe_header **pe_header, uint16_t *optional_header_type,
			    void **optional_pe_header, struct optional_header_pe32 **pe32_of_the_oh,
			    struct optional_header_pe32plus **pe32_of_the_oh_plus);

size_t peparse_get_idd_size(uint32_t entry_id, uint16_t *optional_header_type,
			    void *header_that_is_optional, struct optional_header_pe32 *pe32_of_the_oh,
			    struct optional_header_pe32plus *pe32_of_the_oh_plus);

st_stat() peparse_get_export_table(apsh_instance_t apsh, const access_context_t *context,
				   struct export_table *export_table_inst, address_to_t *rva_export_of_the_table,
				   size_t *export_table_size)
{

	access_context_t _ctx = *context;
	address_to_t rva_export_of_the_header = 0;
	size_t export_header_size = 0;

#define MAX_HEADER_BYTES 1024
	uint8_t image_array[MAX_HEADER_BYTES];

	if (peparse_get_image(apsh, context, MAX_HEADER_BYTES, image_array) == APSH_F)
		return APSH_F;

	void *header_that_is_optional = NULL;
	uint16_t magic = 0;

	peparse_assign_headers(image_array, NULL, NULL, &magic, &header_that_is_optional, NULL, NULL);
	rva_export_of_the_header =
	    peparse_get_idd_rva(IMAGE_DIRECTORY_ENTRY_EXPORT, &magic, header_that_is_optional, NULL, NULL);
	export_header_size =
	    peparse_get_idd_size(IMAGE_DIRECTORY_ENTRY_EXPORT, &magic, header_that_is_optional, NULL, NULL);

	if (rva_export_of_the_table)
		*rva_export_of_the_table = rva_export_of_the_header;

	if (export_table_size)
		*export_table_size = export_header_size;

	debug_print(APSH_DEBUG_PEPARSE,
		    "--PEParse: DLL base 0x%.16" PRIx64 ". Export header [RVA] 0x%.16" PRIx64
		    ". Size %" PRIu64 ".\n",
		    context->addr, rva_export_of_the_header, export_header_size);

	_ctx.addr = context->addr + rva_export_of_the_header;
	if (apsh_read(apsh, &_ctx, sizeof(struct export_table), export_table_inst, NULL) == APSH_F) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: failed to map export header\n");

		if (!((_ctx.addr + 4) & 0xfff)) {
			debug_print(APSH_DEBUG_PEPARSE,
				    "--PEParse: export table is mapped on page boundary\n");
			_ctx.addr += 4;
			if (apsh_read(apsh, &_ctx, sizeof(struct export_table) - 4,
					      (void *)((char *)export_table_inst + 4), NULL) == APSH_F) {
				debug_print(APSH_DEBUG_PEPARSE,
					    "--PEParse: still failed to map export header\n");
				return APSH_F;
			}

			export_table_inst->export_flags = 0;
		} else {
			return APSH_F;
		}
	}

	if (export_table_inst->export_flags || !export_table_inst->name) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: bad export directory table\n");
		return APSH_F;
	}

	return APSH_S;
}

size_t
peparse_get_idd_size(uint32_t entry_id, uint16_t *optional_header_type, void *header_that_is_optional,
		     struct optional_header_pe32 *pe32_of_the_oh,
		     struct optional_header_pe32plus *pe32_of_the_oh_plus)
{

	size_t size = 0;

	if (optional_header_type == NULL) {

		if (pe32_of_the_oh != NULL) {
			size = pe32_of_the_oh->idd[entry_id].size;
			goto done;
		}

		if (pe32_of_the_oh_plus != NULL) {
			size = pe32_of_the_oh_plus->idd[entry_id].size;
			goto done;
		}
	} else if (header_that_is_optional != NULL) {

		if (*optional_header_type == IMAGE_PE32_PLUS_MAGIC) {
			struct optional_header_pe32plus *oh_pe32plus_t =
			    (struct optional_header_pe32plus *)header_that_is_optional;
			size = oh_pe32plus_t->idd[entry_id].size;
			goto done;
		}

		if (*optional_header_type == IMAGE_PE32_MAGIC) {
			struct optional_header_pe32 *oh_pe32_t =
			    (struct optional_header_pe32 *)header_that_is_optional;
			size = oh_pe32_t->idd[entry_id].size;
			goto done;
		}
	}

done:
	return size;
};

st_stat() windows_export_to_rva(apsh_instance_t apsh, const access_context_t *context, const char *symbol,
				address_to_t *rva)
{
	struct export_table export_table_inst;
	address_to_t address_of_rva;
	size_t size_of_et;
	int aon_index = -1;
	int aof_index = -1;

	if (peparse_get_export_table(apsh, context, &export_table_inst, &address_of_rva, &size_of_et) != APSH_S) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: failed to get export table\n");
		return APSH_F;
	}

	aon_index = get_aon_index(apsh, symbol, &export_table_inst, context);
	if (-1 == aon_index) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: failed to get aon index\n");
		return APSH_F;
	}

	aof_index = get_aof_index(apsh, aon_index, &export_table_inst, context);
	if (-1 == aof_index) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: failed to get aof index\n");
		return APSH_F;
	}

	if (get_export_rva(apsh, rva, aof_index, &export_table_inst, context) == APSH_S) {
		if (*rva >= address_of_rva && *rva < address_of_rva + size_of_et) {
			debug_print(APSH_DEBUG_PEPARSE, "--PEParse: %s @ %p is forwarded\n",
				    symbol, context);
			return APSH_F;
		} else {
			return APSH_S;
		}
	} else {
		return APSH_F;
	}
}

char *
windows_rva_to_export(apsh_instance_t apsh, address_to_t rva, const access_context_t *context)
{
	access_context_t _ctx = *context;
	struct export_table export_table_inst;
	address_to_t address_of_rva;
	size_t size_of_et;

	if (peparse_get_export_table(apsh, context, &export_table_inst, &address_of_rva, &size_of_et) != APSH_S) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: failed to get export table\n");
		return NULL;
	}

	if (rva >= address_of_rva && rva < address_of_rva + size_of_et) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEParse: symbol @ 0x%" PRIx64 " is forwarded\n",
			    context->addr + rva);
		return NULL;
	}

	address_to_t base1 = context->addr + export_table_inst.address_of_names;
	address_to_t base2 = context->addr + export_table_inst.address_of_name_ordinals;
	address_to_t base3 = context->addr + export_table_inst.address_of_functions;
	uint32_t i = 0;

	for (; i < export_table_inst.number_of_functions; ++i) {
		uint32_t name_rva = 0;
		uint16_t ordinal = 0;
		uint32_t loc = 0;

		_ctx.addr = base2 + i * sizeof(uint16_t);
		if (apsh_read_16(apsh, &_ctx, &ordinal) == APSH_F)
			continue;

		_ctx.addr = base3 + ordinal * sizeof(uint32_t);
		if (apsh_read_32(apsh, &_ctx, &loc) == APSH_F)
			continue;

		if (loc == rva) {

			_ctx.addr = base1 + i * sizeof(uint32_t);
			if (i < export_table_inst.number_of_names && APSH_S == apsh_read_32(apsh, &_ctx, &name_rva) &&
			    name_rva) {
				_ctx.addr = context->addr + name_rva;
				return apsh_read_str(apsh, &_ctx);
			}

			debug_print(APSH_DEBUG_PEPARSE,
				    "--PEParse: symbol @ 0x%" PRIx64
				    " is exported by ordinal only\n",
				    context->addr + rva);
			break;
		}
	}

	return NULL;
}

void
peparse_assign_headers(const uint8_t *const image_array, struct dos_header **dos_header,
		       struct pe_header **pe_header, uint16_t *optional_header_type,
		       void **optional_pe_header, struct optional_header_pe32 **pe32_of_the_oh,
		       struct optional_header_pe32plus **pe32_of_the_oh_plus)
{

	struct dos_header *dos_h_t_address = (struct dos_header *)image_array;

	if (dos_header != NULL)
		*dos_header = dos_h_t_address;

	struct pe_header *pe_h_t = (struct pe_header *)(image_array + dos_h_t_address->offset_to_pe);

	if (pe_header != NULL)
		*pe_header = pe_h_t;

	void *op_h_t_address = (void *)((uint8_t *)pe_h_t + sizeof(struct pe_header));

	if (optional_pe_header != NULL)
		*optional_pe_header = op_h_t_address;

	uint16_t magic = *((uint16_t *)op_h_t_address);

	if (optional_header_type != NULL)
		*optional_header_type = magic;

	debug_print(APSH_DEBUG_PEPARSE, "--PEParse: magic is 0x%" PRIx16 "\n", magic);

	if (magic == IMAGE_PE32_MAGIC && pe32_of_the_oh != NULL)
		*pe32_of_the_oh = (struct optional_header_pe32 *)op_h_t_address;
	else if (magic == IMAGE_PE32_PLUS_MAGIC && pe32_of_the_oh_plus != NULL)
		*pe32_of_the_oh_plus = (struct optional_header_pe32plus *)op_h_t_address;
}

static int
find_aon_idx_bin(apsh_instance_t apsh, const char *symbol, address_to_t aon_base_va, int low,
		 int high, const access_context_t *context)
{
	access_context_t _ctx = *context;
	int mid, cmp;
	uint32_t str_rva = 0;
	char *name = 0;

	if (high < low)
		goto not_found;

	mid = (low + high) / 2;

	_ctx.addr = aon_base_va + mid * sizeof(uint32_t);
	if (apsh_read_32(apsh, &_ctx, &str_rva) == APSH_F || !str_rva)
		goto not_found;

	_ctx.addr = context->addr + str_rva;
	name = apsh_read_str(apsh, &_ctx);
	if (!name)
		goto not_found;

	cmp = strcmp(symbol, name);
	free(name);

	if (cmp < 0)
		return find_aon_idx_bin(apsh, symbol, aon_base_va, low, mid - 1, context);
	else if (cmp > 0)
		return find_aon_idx_bin(apsh, symbol, aon_base_va, mid + 1, high, context);
	else
		return mid;

not_found:
	return -1;
}

int
get_aon_index(apsh_instance_t apsh, const char *symbol, struct export_table *export_table_inst,
	      const access_context_t *context)
{
	int index = get_aon_index_binary(apsh, symbol, export_table_inst, context);

	if (-1 == index) {
		debug_print(APSH_DEBUG_PEPARSE,
			    "--PEParse: Falling back to linear search for aon index\n");
		index = get_aon_index_linear(apsh, symbol, export_table_inst, context);
	}
	return index;
}

int
get_aof_index(apsh_instance_t apsh, int aon_index, struct export_table *export_table_inst,
	      const access_context_t *context)
{
	access_context_t _ctx = *context;

	_ctx.addr += export_table_inst->address_of_name_ordinals + aon_index * sizeof(uint16_t);

	uint16_t aof_index;

	if (apsh_read_16(apsh, &_ctx, &aof_index) == APSH_S)
		return (int)aof_index;
	else
		return -1;
};

status_t peparse_validate_pe_image(const uint8_t *const image_array, size_t len)
{
	struct dos_header *header_of_dos = (struct dos_header *)image_array;

	if (header_of_dos->signature != IMAGE_DOS_HEADER) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEPARSE: DOS header signature not found\n");
		return APSH_F;
	}

	uint32_t pe_from_ofs = header_of_dos->offset_to_pe;

	if (pe_from_ofs + sizeof(struct pe_header) + sizeof(struct optional_header_pe32) > len) {
		debug_print(APSH_DEBUG_PEPARSE,
			    "--PEPARSE: offset to PE would be outside buffer length\n");
		return APSH_F;
	}

	struct pe_header *header_of_pa = (struct pe_header *)(image_array + pe_from_ofs);

	if (header_of_pa->signature != IMAGE_NT_SIGNATURE) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEPARSE: PE header signature invalid\n");
		return APSH_F;
	}

	struct optional_header_pe32 *pe_opt_header =
	    (struct optional_header_pe32 *)((uint8_t *)header_of_pa + sizeof(struct pe_header));

	if (pe_opt_header->magic != IMAGE_PE32_MAGIC &&
	    pe_opt_header->magic != IMAGE_PE32_PLUS_MAGIC) {
		debug_print(APSH_DEBUG_PEPARSE, "--PEPARSE: Optional header magic value unknown\n");
		return APSH_F;
	}

	return APSH_S;
};
