/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>

#include "apsh_instance.h"

#define MAGIC1 0x1b0003
#define MAGIC2 0x200003
#define MAGIC3 0x300003
#define MAGIC4 0x580003
#define MAGIC5 0x260003
#define MAGIC_WIN10 0xB80003
#define MAGIC_WIN2019 0xB60003
#define MAGIC_WIN2012 0xB20003

#define WIN_MAX_PROCNAME 16
#define address_to_t addr_t
#define string char *

string windows_get_eprocess_name(apsh_instance_t apsh, address_to_t paddr);

static inline int
check_magic_2k(uint32_t a)
{
	return (a == MAGIC1);
}

static inline int
check_magic_vista(uint32_t a)
{
	return (a == MAGIC2 || a == MAGIC3);
}

static inline int
check_magic_7(uint32_t a)
{
	return (a == MAGIC4 || a == MAGIC5);
}

static inline int
check_magic_unknown(uint32_t a)
{
	return (a == MAGIC1 || a == MAGIC2 || a == MAGIC3 || a == MAGIC4 || a == MAGIC5 ||
		a == MAGIC_WIN10 || a == MAGIC_WIN2019 || a == MAGIC_WIN2012);
}

static check_magic_func get_check_magic_func(apsh_instance_t apsh);

int
find_pname_offset(apsh_instance_t apsh, check_magic_func check)
{
	address_to_t pa_address_block = 0;
	address_to_t offset = 0;
	uint32_t value = 0;
	void *bm = 0;

	bm = boyer_moore_init((unsigned string) "Idle", 4);

#define BLOCK_SIZE (1024 * 1024 * 1)
	unsigned char block_buffer[BLOCK_SIZE];

	if (check == NULL)
		check = get_check_magic_func(apsh);

	for (pa_address_block = 4096; pa_address_block + BLOCK_SIZE < apsh->max_physical_address;
	     pa_address_block += BLOCK_SIZE) {
		if (apsh_read_pa(apsh, pa_address_block, BLOCK_SIZE, block_buffer, NULL) == APSH_F)
			continue;

		for (offset = 0; offset < BLOCK_SIZE; offset += 8) {
			memcpy(&value, block_buffer + offset, 4);

			if (check(value)) { // look for specific magic #
				debug_print(APSH_DEBUG_MISC,
					    "--%s: found magic value 0x%.8" PRIx32
					    " @ offset 0x%.8" PRIx64 "\n",
					    __FUNCTION__, value, pa_address_block + offset);

				unsigned char haystack[0x500];

				if (apsh_read_pa(apsh, pa_address_block + offset, 0x500, haystack, NULL) == APSH_F)
					continue;

				int i = boyer_moore2(bm, haystack, 0x500);

				if (-1 == i) {
					continue;
				} else {
					apsh->init_task = pa_address_block + offset;
					debug_print(APSH_DEBUG_MISC,
						    "--%s: found Idle process at 0x%.8" PRIx64
						    " + 0x%x\n",
						    __FUNCTION__, pa_address_block + offset, i);
					boyer_moore_fini(bm);
					return i;
				}
			}
		}
	}
	boyer_moore_fini(bm);
	return 0;
}

static check_magic_func
get_check_magic_func(apsh_instance_t apsh)
{
	check_magic_func rtn = NULL;

	if (apsh->os_data == NULL)
		return &check_magic_unknown;

	switch (((windows_instance_t)apsh->os_data)->version) {
	case APSH_OS_WIN_2000:
	case APSH_OS_WIN_XP:
	case APSH_OS_WIN_2003:
		rtn = &check_magic_2k;
		break;
	case APSH_OS_WIN_VISTA:
		rtn = &check_magic_vista;
		break;
	case APSH_OS_WIN_7:
		rtn = &check_magic_7;
		break;
	case APSH_OS_WIN_2008:
	case APSH_OS_WIN_UNKNOWN:
		rtn = &check_magic_unknown;
		break;
	default:
		rtn = &check_magic_unknown;
		debug_print(APSH_DEBUG_MISC,
			    "--%s: illegal value in apsh->os.windows_instance.version\n",
			    __FUNCTION__);
		break;
	}

	return rtn;
}

static address_to_t
find_process_by_name(apsh_instance_t apsh, check_magic_func check, address_to_t start_address,
		     const string name)
{

	debug_print(APSH_DEBUG_MISC, "--searching for process by name: %s\n", name);

	address_to_t pa_address_block = 0;
	address_to_t offset = 0;
	uint32_t value = 0;

	/*
	 * Some compilers don't align this large buffer. Force alignment so
	 * the offsets match between physical mem and the buffer.
	 */
	unsigned char block_buffer[APSH_PS_4KB];

	if (check == NULL)
		check = get_check_magic_func(apsh);

	for (pa_address_block = start_address; pa_address_block + APSH_PS_4KB < apsh->max_physical_address;
	     pa_address_block += APSH_PS_4KB) {
		if (apsh_read_pa(apsh, pa_address_block, APSH_PS_4KB, block_buffer, NULL) == APSH_F)
			continue;

		for (offset = 0; offset < APSH_PS_4KB; offset += 8) {
			memcpy(&value, block_buffer + offset, 4);

			if (check(value)) { /* look for specific magic # */

				string procname = windows_get_eprocess_name(apsh, pa_address_block + offset);

				if (procname) {
					if (strncmp(procname, name, WIN_MAX_PROCNAME) == 0) {
						free(procname);
						return pa_address_block + offset;
					}
					free(procname);
				}
			}
		}
	}
	return 0;
}

address_to_t
windows_find_eprocess(apsh_instance_t apsh, const string name)
{

	address_to_t start_address = 0;
	windows_instance_t os_win = apsh->os_data;
	check_magic_func check = get_check_magic_func(apsh);

	if (os_win == NULL)
		return 0;

	debug_print(APSH_DEBUG_MISC, "--Windows version %u\n", os_win->version);
	debug_print(APSH_DEBUG_MISC, "--Windows Build %hu\n", os_win->build);

	if (!os_win->pname_offset) {
		if (APSH_F ==
		    json_profile_lookup(apsh, "_EPROCESS", "ImageFileName", &os_win->pname_offset))
			os_win->pname_offset = find_pname_offset(apsh, check);

		if (!os_win->pname_offset) {
			debug_print(APSH_DEBUG_MISC, "--failed to find pname_offset\n");
			return 0;
		} else {
			debug_print(APSH_DEBUG_MISC,
				    "**set os.windows_instance.pname_offset (0x%" PRIx64 ")\n",
				    os_win->pname_offset);
		}
	}

	if (apsh->init_task)
		start_address = apsh->init_task;

	return find_process_by_name(apsh, check, start_address, name);
}

address_to_t
eprocess_list_search(apsh_instance_t apsh, address_to_t list_head, int offset, size_t len,
		     void *value)
{
	address_to_t next_process = 0;
	address_to_t tasks_offset = 0;
	address_to_t rtnval = 0;
	void *buf = g_try_malloc0(len);

	if (!buf)
		goto exit;

	if (apsh_get_offset(apsh, "win_tasks", &tasks_offset) == APSH_F)
		goto exit;

	if (apsh_read_addr_va(apsh, list_head + tasks_offset, 0, &next_process) == APSH_F)
		goto exit;

	if (apsh_read_va(apsh, list_head + offset, 0, len, buf, NULL) == APSH_F)
		goto exit;

	if (memcmp(buf, value, len) == 0) {
		rtnval = list_head + tasks_offset;
		goto exit;
	}
	list_head = next_process;

	while (1) {
		address_to_t tmp_next = 0;

		if (apsh_read_addr_va(apsh, next_process, 0, &tmp_next) == APSH_F)
			goto exit;

		if (list_head == tmp_next)
			break;

		if (apsh_read_va(apsh, next_process + offset - tasks_offset, 0, len, buf, NULL) == APSH_F)
			goto exit;

		if (memcmp(buf, value, len) == 0) {
			rtnval = next_process;
			goto exit;
		}
		next_process = tmp_next;
	}

exit:
	g_free(buf);
	return rtnval;
}

static address_to_t
find_eprocess(apsh_instance_t apsh, int offset, size_t len, void *value)
{
	address_to_t eprocess = 0;
	address_to_t list_head = ((windows_instance_t)apsh->os_data)->sysproc;

	if (list_head)
		eprocess = eprocess_list_search(apsh, list_head, offset, len, value);

	if (!eprocess && APSH_F == apsh_read_addr_ksym(apsh, "PsInitialSystemProcess", &list_head))
		return 0;

	return eprocess_list_search(apsh, list_head, offset, len, value);
}

address_to_t
windows_find_eprocess_list_pid(apsh_instance_t apsh, apsh_pid_t pid)
{
	size_t len = sizeof(apsh_pid_t);
	int pid_offset = 0;

	if (!apsh->os_data)
		return 0;

	pid_offset = ((windows_instance_t)apsh->os_data)->pid_offset;

	return find_eprocess(apsh, pid_offset, len, &pid);
}

address_to_t
windows_find_eprocess_list_pgd(apsh_instance_t apsh, address_to_t pgd)
{
	int pdbase_offset = 0;
	size_t len = 0;

	if (!apsh->os_data)
		return 0;

	pdbase_offset = ((windows_instance_t)apsh->os_data)->pdbase_offset;

	if (apsh->page_mode == APSH_PM_LEGACY || apsh->page_mode == APSH_PM_PAE)
		len = sizeof(uint32_t);
	else
		len = sizeof(address_to_t);

	return find_eprocess(apsh, pdbase_offset, len, &pgd);
}

string
windows_get_eprocess_name(apsh_instance_t apsh, address_to_t paddr)
{
	size_t name_length = WIN_MAX_PROCNAME;
	windows_instance_t os_win = apsh->os_data;

	if (os_win == NULL)
		return NULL;

	address_to_t name_paddr = paddr + os_win->pname_offset;
	string name = (string)g_try_malloc0(name_length);

	if (!name)
		return NULL;

	if (apsh_read_pa(apsh, name_paddr, name_length, name, NULL) == APSH_S)
		return name;

	free(name);
	return NULL;
}
