/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <iconv.h>
#include <string.h>
#include <wchar.h>

#include "apsh_instance.h"

unicode_string_t *
windows_read_unicode_struct_pm(apsh_instance_t apsh, const access_context_t *ctx,
			       page_mode_t page_mode)
{

	addr_t va_of_the_buffer = 0;
	uint16_t length_of_the_buff = 0;
	access_context_t _ctx = *ctx;
	unicode_string_t *us = 0;
	size_t the_size_of_struct = 0;

	if (page_mode == APSH_PM_IA32E) {
		win64_unicode_string_t us64 = {0};

		the_size_of_struct = sizeof(us64);
		if (apsh_read(apsh, ctx, the_size_of_struct, &us64, NULL) == APSH_F) {
			debug_print(APSH_DEBUG_READ, "--%s: failed to read UNICODE_STRING\n",
				    __FUNCTION__);
			goto out_error;
		}

		length_of_the_buff = us64.length;
		va_of_the_buffer = us64.pBuffer;
	} else {
		win32_unicode_string_t us32 = {0};

		the_size_of_struct = sizeof(us32);
		if (apsh_read(apsh, ctx, the_size_of_struct, &us32, NULL) == APSH_F) {
			debug_print(APSH_DEBUG_READ, "--%s: failed to read UNICODE_STRING\n",
				    __FUNCTION__);
			goto out_error;
		}

		length_of_the_buff = us32.length;

		va_of_the_buffer = us32.pBuffer;
	}

	if (length_of_the_buff > APSH_PS_4KB) {
		debug_print(APSH_DEBUG_READ,
			    "--%s: the length of %" PRIu16 " in the UNICODE_STRING at 0x%" PRIx64
			    " is excessive, bailing out.\n",
			    __FUNCTION__, length_of_the_buff, ctx->addr);
		return NULL;
	}

	us = g_try_malloc0(sizeof(unicode_string_t));
	if (!us)
		return NULL;

	us->contents = g_try_malloc0(sizeof(uint8_t) * (length_of_the_buff + 2));
	us->length = length_of_the_buff;

	if (!us->contents)
		goto out_error;

	_ctx.addr = va_of_the_buffer;
	if (apsh_read(apsh, &_ctx, us->length, us->contents, NULL) == APSH_F) {
		debug_print(APSH_DEBUG_READ, "--%s: failed to read UNICODE_STRING buffer\n",
			    __FUNCTION__);
		goto out_error;
	}

	us->contents[length_of_the_buff] = 0;
	us->contents[length_of_the_buff + 1] = 0;

	us->encoding = "UTF-16";

	return us;

out_error:
	if (us)
		g_free(us->contents);
	g_free(us);
	return 0;
}

unicode_string_t *
windows_read_unicode_struct(apsh_instance_t apsh, const access_context_t *ctx)
{
	return windows_read_unicode_struct_pm(apsh, ctx, apsh->page_mode);
}
