/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef LIB_APSH_PEPARSE_H
#define LIB_APSH_PEPARSE_H

#include "apsh.h"

#ifdef __cplusplus
extern "C" {
#define NOEXCEPT noexcept
#else
#define NOEXCEPT
#endif

#define int_32 uint32_t
#define int_16 uint16_t

struct pe_header {
	int32_t signature;
	int_16 machine;
	int_16 number_of_sections;
	int_32 time_date_stamp;
	int_32 pointer_to_symbol_table;
	int_32 number_of_symbols;
	int_16 size_of_optional_header;
	int_16 characteristics;
} __attribute__((packed));

#define IMAGE_DIRECTORY_ENTRY_CERTIFICATE 4
#define IMAGE_DIRECTORY_ENTRY_BASERELOC 5

#define IMAGE_DOS_HEADER 0x5A4D
#define IMAGE_NT_SIGNATURE 0x00004550

#define IMAGE_PE32_MAGIC 0x10b
#define IMAGE_PE32_PLUS_MAGIC 0x20b

#define IMAGE_DIRECTORY_ENTRY_DEBUG 6
#define IMAGE_DIRECTORY_ENTRY_ARCHITECTURE 7
#define IMAGE_DIRECTORY_ENTRY_GLOBALPTR 8

struct image_data_directory {
	int_32 virtual_address;
	int_32 size;
} __attribute__((packed));

#define IMAGE_DIRECTORY_ENTRY_IAT 12
#define IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT 13
#define IMAGE_DIRECTORY_ENTRY_CLR_RUNTIME 14


#define IMAGE_DIRECTORY_ENTRY_EXPORT 0
#define IMAGE_DIRECTORY_ENTRY_IMPORT 1
#define IMAGE_DIRECTORY_ENTRY_RESOURCE 2
#define IMAGE_DIRECTORY_ENTRY_EXCEPTION 3

struct dos_header {
	int_16 signature;
	int_16 bytes_in_last_block;
	int_16 blocks_in_file;
	int_16 reloc_ct;
	int_16 header_size;
	int_16 min_mem;
	int_16 max_mem;
	int_16 ss;
	int_16 sp;
	int_16 chksum;
	int_16 ip;
	int_16 cs;
	int_16 reloc_tbl_ofs;
	int_16 overlay;
	uint8_t reserved[32];
	int_32 offset_to_pe;
} __attribute__((packed));

#define IMAGE_DIRECTORY_ENTRY_TLS 9
#define IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG 10
#define IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT 11
#define IMAGE_DIRECTORY_ENTRY_RESERVED 15



struct optional_header_pe32 {
	int_16 magic;
	uint8_t major_linker_version;
	uint8_t minor_linker_version;
	int_32 size_of_code;
	int_32 size_of_initialized_data;
	int_32 size_of_uninitialized_data;
	int_32 address_of_entry_point;
	int_32 base_of_code;
	int_32 base_of_data;
	int_32 image_base;
	int_32 section_alignment;
	int_32 file_alignment;
	int_16 major_os_version;
	int_16 minor_os_version;
	int_16 major_image_version;
	int_16 minor_image_version;
	int_16 major_subsystem_version;
	int_16 minor_subsystem_version;
	int_32 win32_version_value;
	int_32 size_of_image;
	int_32 size_of_headers;
	int_32 checksum;
	int_16 subsystem;
	int_16 dll_characteristics;
	int_32 size_of_stack_reserve;
	int_32 size_of_stack_commit;
	int_32 size_of_heap_reserve;
	int_32 size_of_heap_commit;
	int_32 loader_flags;
	int_32 number_of_rva_and_sizes;
	struct image_data_directory idd[16];
} __attribute__((packed));

struct optional_header_pe32plus {
	int_16 magic;
	uint8_t major_linker_version;
	uint8_t minor_linker_version;
	int_32 size_of_code;
	int_32 size_of_initialized_data;
	int_32 size_of_uninitialized_data;
	int_32 address_of_entry_point;
	int_32 base_of_code;
	uint64_t image_base;
	int_32 section_alignment;
	int_32 file_alignment;
	int_16 major_os_version;
	int_16 minor_os_version;
	int_16 major_image_version;
	int_16 minor_image_version;
	int_16 major_subsystem_version;
	int_16 minor_subsystem_version;
	int_32 win32_version_value;
	int_32 size_of_image;
	int_32 size_of_headers;
	int_32 checksum;
	int_16 subsystem;
	int_16 dll_characteristics;
	uint64_t size_of_stack_reserve;
	uint64_t size_of_stack_commit;
	uint64_t size_of_heap_reserve;
	uint64_t size_of_heap_commit;
	int_32 loader_flags;
	int_32 number_of_rva_and_sizes;
	struct image_data_directory idd[16];
} __attribute__((packed));

struct section_header {
	char short_name[8];
	union {
		int_32 physical_address;
		int_32 virtual_size;
	} a;
	int_32 virtual_address;
	int_32 size_of_raw_data;
	int_32 pointer_to_raw_data;
	int_32 pointer_to_relocations;
	int_32 pointer_to_line_numbers;
	int_16 number_of_relocations;
	int_16 number_of_line_numbers;
	int_32 characteristics;
} __attribute__((packed));

struct export_table {
	int_32 export_flags;
	int_32 time_date_stamp;
	int_16 major_version;
	int_16 minor_version;
	int_32 name;
	int_32 base;
	int_32 number_of_functions;
	int_32 number_of_names;
	int_32 address_of_functions;
	int_32 address_of_names;
	int_32 address_of_name_ordinals;
} __attribute__((packed));

status_t peparse_validate_pe_image(const uint8_t *const image, size_t len) NOEXCEPT;

status_t peparse_get_image(apsh_instance_t apsh, const access_context_t *ctx, size_t len,
			   const uint8_t *const image) NOEXCEPT;

void peparse_assign_headers(const uint8_t *const image, struct dos_header **dos_header,
			    struct pe_header **pe_header, int_16 *optional_header_type,
			    void **optional_pe_header, struct optional_header_pe32 **oh_pe32,
			    struct optional_header_pe32plus **oh_pe32plus) NOEXCEPT;

addr_t peparse_get_idd_rva(int_32 entry_id, int_16 *optional_header_type, void *optional_header,
			   struct optional_header_pe32 *oh_pe32,
			   struct optional_header_pe32plus *oh_pe32plus) NOEXCEPT;

size_t peparse_get_idd_size(int_32 entry_id, int_16 *optional_header_type, void *optional_header,
			    struct optional_header_pe32 *oh_pe32,
			    struct optional_header_pe32plus *oh_pe32plus) NOEXCEPT;

void dump_exports(apsh_instance_t apsh, struct export_table *et,
		  const access_context_t *ctx) NOEXCEPT;

status_t peparse_get_export_table(apsh_instance_t apsh, const access_context_t *ctx,
				  struct export_table *et, addr_t *export_table_rva,
				  size_t *export_table_size) NOEXCEPT;

#ifdef __cplusplus
}
#endif

#endif
