/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdbool.h>

#include "profile_of_file_json/rekall.h"
#include "profile_of_file_json/volatility_ist.h"
#include "json_profiles.h"
#include "apsh_instance.h"

typedef enum json_profile_type
{
	JPT_INVALID,
	JPT_REKALL_PROFILE,
	JPT_VOLATILITY_IST
} json_profile_type_t;

bool
json_profile_init(apsh_instance_t apsh, const char *path);

json_object *
apsh_get_kernel_json(apsh_instance_t apsh)
{
	return apsh->json.root;
}

status_t
apsh_get_symbol_addr_from_json(apsh_instance_t apsh, json_object *json, const char *symbol,
			      addr_t *addr)
{
	if (!apsh->json.handler)
		return APSH_F;

	return apsh->json.handler(json, symbol, NULL, addr, NULL);
}

status_t
apsh_get_struct_size_from_json(apsh_instance_t apsh, json_object *json, const char *struct_name,
			      size_t *size)
{
	if (!apsh->json.handler)
		return APSH_F;

	return apsh->json.handler(json, struct_name, NULL, NULL, size);
}

status_t
apsh_get_struct_member_offset_from_json(apsh_instance_t apsh, json_object *json,
				       const char *struct_name, const char *struct_member,
				       addr_t *offset)
{
	if (!apsh->json.handler)
		return APSH_F;

	return apsh->json.handler(json, struct_name, struct_member, offset, NULL);
}

status_t
apsh_get_bitfield_offset_and_size_from_json(apsh_instance_t apsh, json_object *json,
					   const char *struct_name, const char *struct_member,
					   addr_t *offset, size_t *start_bit, size_t *end_bit)
{
	if (!apsh->json.bitfield_offset_and_size)
		return APSH_F;

	return apsh->json.bitfield_offset_and_size(json, struct_name, struct_member, offset,
						  start_bit, end_bit);
}

status_t
apsh_get_symbol_addr_from_apsh_json(apsh_instance_t apsh, const char *symbol, addr_t *addr)
{
	if (!apsh->json.handler)
		return APSH_F;

	return apsh->json.handler(apsh->json.root, symbol, NULL, addr, NULL);
}

status_t
apsh_get_struct_size_from_apsh_json(apsh_instance_t apsh, const char *struct_name, size_t *size)
{
	if (!apsh->json.handler)
		return APSH_F;

	return apsh->json.handler(apsh->json.root, struct_name, NULL, NULL, size);
}

status_t
apsh_get_struct_member_offset_from_apsh_json(apsh_instance_t apsh, const char *struct_name,
					   const char *struct_member, addr_t *offset)
{
	if (!apsh->json.handler)
		return APSH_F;

	return apsh->json.handler(apsh->json.root, struct_name, struct_member, offset, NULL);
}

status_t
apsh_get_bitfield_offset_and_size_from_apsh_json(apsh_instance_t apsh, const char *struct_name,
					       const char *struct_member, addr_t *offset,
					       size_t *start_bit, size_t *end_bit)
{
	if (!apsh->json.bitfield_offset_and_size)
		return APSH_F;

	return apsh->json.bitfield_offset_and_size(apsh->json.root, struct_name, struct_member,
						  offset, start_bit, end_bit);
}

bool
json_profile_init(apsh_instance_t apsh, const char *path)
{
	json_interface_t *json = &apsh->json;
	json_profile_type_t type = JPT_INVALID;
	json_object *metadata = NULL;

	if (json->path) {
		print_error("Duplicate JSON profile detected: %s\n", path);
		return false;
	}

	json->path = g_strdup(path);
	json->root = json_object_from_file(json->path);

	if (!json->root) {
		json->path = NULL;
		g_free((char *)json->path);
		print_error("JSON at %s couldn't be opened!\n", path);
		return false;
	}

	if (json_object_object_get_ex(apsh->json.root, "metadata", &metadata))
		type = JPT_VOLATILITY_IST;
	else if (json_object_object_get_ex(apsh->json.root, "$METADATA", &metadata))
		type = JPT_REKALL_PROFILE;

	switch (type) {
	case JPT_REKALL_PROFILE:
		json->handler = rekall_profile_symbol_to_rva;
		json->bitfield_offset_and_size = rekall_profile_bitfield_offset_and_size;
		json->get_os_type = rekall_get_os_type;
		break;
	case JPT_VOLATILITY_IST:
		json->handler = volatility_ist_symbol_to_rva;
		json->bitfield_offset_and_size = volatility_profile_bitfield_offset_and_size;
		json->get_os_type = volatility_get_os_type;
		break;
	default:
		return false;
	};

	return true;
};
