/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "apsh_instance.h"
#include <json-c/json.h>
#include <stdio.h>

#define string char *
#define st_stat() status_t
#define address_to_t addr_t

static st_stat() rekall_find_offset(json_object *json, const string symbol, const string sub_of_sym,
				    address_to_t *address_of_rva);

st_stat() rekall_profile_symbol_to_rva(json_object *json, const string symbol,
				       const string sub_of_sym, address_to_t *address_of_rva, size_t *size)
{
	st_stat() ret = APSH_F;
	if (!json || !symbol)
		return ret;

	if (!sub_of_sym && !size) {
		json_object *constants = NULL, *functions = NULL, *jsymbol = NULL;

		if (json_object_object_get_ex(json, "$CONSTANTS", &constants)) {
			if (!json_object_object_get_ex(constants, symbol, &jsymbol)) {
				debug_print(APSH_DEBUG_MISC, "Rekall: symbol '%s' not found in $CONSTANTS\n", symbol);
			} else {
				*address_of_rva = json_object_get_int64(jsymbol);
				ret = APSH_S;
				goto exit;
			}
		} else {
			debug_print(APSH_DEBUG_MISC,
				    "Rekall: no $CONSTANTS section found\n");
		}

		if (json_object_object_get_ex(json, "$FUNCTIONS", &functions)) {
			if (!json_object_object_get_ex(functions, symbol, &jsymbol)) {
				debug_print(APSH_DEBUG_MISC,
					    "Rekall: symbol '%s' not found in $FUNCTIONS\n",
					    symbol);
			} else {
				*address_of_rva = json_object_get_int64(jsymbol);
				ret = APSH_S;
				goto exit;
			}
		} else {
			debug_print(APSH_DEBUG_MISC,
				    "Rekall: no $FUNCTIONS section found\n");
		}
	} else {
		json_object *structs = NULL, *json_of_structs = NULL;

		if (!json_object_object_get_ex(json, "$STRUCTS", &structs)) {
			debug_print(APSH_DEBUG_MISC, "profile of the rekall message no $STRUCTS section found\n");
			goto exit;
		}
		if (!json_object_object_get_ex(structs, symbol, &json_of_structs)) {
			debug_print(APSH_DEBUG_MISC, "profile of the rekall message no %s found\n", symbol);
			goto exit;
		}

		if (size) {
			json_object *jsize = json_object_array_get_idx(json_of_structs, 0);
			*size = json_object_get_int64(jsize);

			ret = APSH_S;
			goto exit;
		}
		*address_of_rva = 0;
		ret = rekall_find_offset(json, symbol, sub_of_sym, address_of_rva);
	}

exit:
	debug_print(APSH_DEBUG_MISC, "Rekall profile lookup %s %s: 0x%lx\n", symbol ?: NULL,
		    sub_of_sym ?: NULL, address_of_rva ? *address_of_rva : 0);

	return ret;
}

static st_stat() rekall_find_offset(json_object *json, const string symbol, const string sub_of_sym,
				    address_to_t *address_of_rva)
{
	st_stat() ret = APSH_F;
	json_object *structs = NULL, *json_of_structs = NULL, *json_of_structs2 = NULL, *jmember = NULL,
		    *json_of_value = NULL;
	struct json_object_iterator iter, iend;

	if (!json_object_object_get_ex(json, "$STRUCTS", &structs)) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message no $STRUCTS section found\n");
		goto exit;
	}
	if (!json_object_object_get_ex(structs, symbol, &json_of_structs)) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message no %s found\n", symbol);
		goto exit;
	}

	json_of_structs2 = json_object_array_get_idx(json_of_structs, 1);
	if (!json_of_structs2) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message struct %s has no second element\n",
			    symbol);
		goto exit;
	}

	if (json_object_object_get_ex(json_of_structs2, sub_of_sym, &jmember)) {
		json_of_value = json_object_array_get_idx(jmember, 0);
		if (!json_of_value) {
			debug_print(APSH_DEBUG_MISC,
				    "profile of the rekall message %s.%s exists but has no RVA defined\n", symbol,
				    sub_of_sym);
			goto exit;
		}
		ret = APSH_S;
		*address_of_rva += json_object_get_int64(json_of_value);
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s.%s @ offset 0x%lx\n", symbol,
			    sub_of_sym, *address_of_rva);
		goto exit;
	}

	iter = json_object_iter_begin(json_of_structs2);
	iend = json_object_iter_end(json_of_structs2);

	while (!json_object_iter_equal(&iter, &iend)) {
		json_object *subval = NULL, *subval2 = NULL, *subval3 = NULL;
		const string subname1 = NULL;
		const string embedded = NULL;

		subval = json_object_iter_peek_value(&iter);
		if (subval == NULL)
			goto next;

		subname1 = json_object_iter_peek_name(&iter);
		(void)subname1;
		subval2 = json_object_array_get_idx(subval, 1);
		subval3 = json_object_array_get_idx(subval2, 0);
		if (!subval3)
			goto next;

		embedded = json_object_get_string(subval3);
		if (!embedded)
			goto next;

		ret = rekall_find_offset(json, embedded, sub_of_sym, address_of_rva);
		if (ret == APSH_S) {
			json_object *jofs = NULL;

			jofs = json_object_array_get_idx(subval, 0);
			if (!jofs) {
				ret = APSH_F;
				debug_print(APSH_DEBUG_MISC,
					    "profile of the rekall message anonymous/embedded struct %s has no "
					    "offset in %s\n",
					    subname1, symbol);
				goto exit;
			}

			*address_of_rva += json_object_get_int64(jofs);
			debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s.%s @ offset 0x%lx\n",
				    symbol, subname1, *address_of_rva);
			goto exit;
		}

next:
		json_object_iter_next(&iter);
	}

exit:
	return ret;
}

st_stat() rekall_profile_bitfield_offset_and_size(json_object *json, const string symbol,
						  const string sub_of_sym, address_to_t *address_of_rva,
						  size_t *start_bit, size_t *end_bit)
{
	st_stat() ret = APSH_F;
	if (!json || !symbol)
		return ret;

	json_object *structs = NULL, *json_of_structs = NULL, *json_of_structs2 = NULL, *json_of_structs3 = NULL,
		    *jmember = NULL, *json_of_value = NULL;
	if (!json_object_object_get_ex(json, "$STRUCTS", &structs)) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message no $STRUCTS section found\n");
		goto exit;
	}
	if (!json_object_object_get_ex(structs, symbol, &json_of_structs)) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message no %s found\n", symbol);
		goto exit;
	}

	json_of_structs2 = json_object_array_get_idx(json_of_structs, 1);
	if (!json_of_structs2) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message struct %s has no second element\n",
			    symbol);
		goto exit;
	}

	if (!json_object_object_get_ex(json_of_structs2, sub_of_sym, &jmember)) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s has no %s member\n", symbol,
			    sub_of_sym);
		goto exit;
	}

	json_of_value = json_object_array_get_idx(jmember, 0);
	if (!json_of_value) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s.%s has no RVA defined\n", symbol,
			    sub_of_sym);
		goto exit;
	}

	*address_of_rva = json_object_get_int64(json_of_value);

	json_of_value = json_object_array_get_idx(jmember, 1);
	if (!json_of_value) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s.%s has no BitField declaration\n",
			    symbol, sub_of_sym);
		goto exit;
	}

	json_of_structs3 = json_object_array_get_idx(json_of_value, 1);
	if (!json_of_value) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s.%s has no BitField definition\n",
			    symbol, sub_of_sym);
		goto exit;
	}

	if (!json_object_object_get_ex(json_of_structs3, "start_bit", &json_of_value)) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s.%s has no member start_bit\n",
			    symbol, sub_of_sym);
		goto exit;
	}
	*start_bit = json_object_get_int64(json_of_value);

	if (!json_object_object_get_ex(json_of_structs3, "end_bit", &json_of_value)) {
		debug_print(APSH_DEBUG_MISC, "profile of the rekall message %s.%s has no member end_bit\n", symbol,
			    sub_of_sym);
		goto exit;
	}
	*end_bit = json_object_get_int64(json_of_value);

	ret = APSH_S;

exit:
	debug_print(APSH_DEBUG_MISC,
		    "Rekall profile lookup %s %s: offset 0x%lx, start bit %ld, end bit %ld\n",
		    symbol ?: NULL, sub_of_sym ?: NULL, *address_of_rva, *start_bit, *end_bit);

	return ret;
}

const string
rekall_get_os_type(apsh_instance_t apsh)
{
	json_object *metadata = NULL, *os = NULL;

	if (!json_object_object_get_ex(apsh->json.root, "$METADATA", &metadata))
		return NULL;
	if (!json_object_object_get_ex(metadata, "ProfileClass", &os))
		return NULL;

	if (!strcmp("Linux", json_object_get_string(os)))
		return "Linux";
	else
		return "Windows";

	return NULL;
}
