/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "apsh_instance.h"
#include <stdio.h>
#include <json-c/json.h>

#define st_stat() status_t
#define address_to_t addr_t

static st_stat() volatility_ist_find_offset(json_object *object_with_json_format, const char *symbol,
					    const char *subsymbol, address_to_t *rva);

st_stat() volatility_ist_symbol_to_rva(json_object *object_with_json_format, const char *symbol, const char *subsymbol,
				       address_to_t *rva, size_t *size);

const char *
volatility_get_os_type(apsh_instance_t apsh)
{
	json_object *metadata = NULL, *os = NULL;

	if (!json_object_object_get_ex(apsh->json.root, "metadata", &metadata)) {
		debug_print(APSH_DEBUG_MISC, "APSH volat IST no metadata section found\n");
		return NULL;
	}

	if (json_object_object_get_ex(metadata, "windows", &os))
		return "Windows";

	return "Linux";
}

st_stat() volatility_profile_bitfield_offset_and_size(json_object *object_with_json_format, const char *symbol,
						      const char *subsymbol, address_to_t *rva,
						      size_t *start_bit, size_t *end_bit);

st_stat() volatility_ist_symbol_to_rva(json_object *object_with_json_format, const char *symbol, const char *subsymbol,
				       address_to_t *rva, size_t *size)
{
	st_stat() ret = APSH_F;

	if (!object_with_json_format || !symbol)
		return ret;

	if (!subsymbol && !size) {
		json_object *json_object_of_symbols = NULL, *json_object_of_symbol = NULL, *address = NULL;

		if (!json_object_object_get_ex(object_with_json_format, "symbols", &json_object_of_symbols)) {
			debug_print(APSH_DEBUG_MISC,
				    "APSH volat IST no symbols section found\n");
			goto exit;
		}
		if (!json_object_object_get_ex(json_object_of_symbols, symbol, &json_object_of_symbol)) {
			debug_print(APSH_DEBUG_MISC,
				    "APSH volat IST: symbol '%s' not found in json_object_of_symbols\n", symbol);
			goto exit;
		}
		if (!json_object_object_get_ex(json_object_of_symbol, "address", &address)) {
			debug_print(APSH_DEBUG_MISC, "APSH volat IST: no address found for %s\n",
				    symbol);
			goto exit;
		}

#ifdef JSONC_UINT64_SUPPORT
		*rva = json_object_get_uint64(address);
#else
		*rva = json_object_get_uint64(address);
#endif
		ret = APSH_S;
	} else {
		json_object *user_types = NULL, *json_struct_object = NULL;

		if (!json_object_object_get_ex(object_with_json_format, "user_types", &user_types)) {
			debug_print(APSH_DEBUG_MISC,
				    "APSH volat IST no user_types section found\n");
			goto exit;
		}
		if (!json_object_object_get_ex(user_types, symbol, &json_struct_object)) {
			debug_print(APSH_DEBUG_MISC, "APSH volat IST no %s found\n",
				    symbol);
			goto exit;
		}

		if (size) {
			json_object *json_size_of_object = NULL;

			if (!json_object_object_get_ex(json_struct_object, "size", &json_size_of_object)) {
				debug_print(APSH_DEBUG_MISC,
					    "APSH volat IST Struct '%s' size not found\n",
					    symbol);
				goto exit;
			}

#ifdef JSONC_UINT64_SUPPORT
			*size = json_object_get_uint64(json_size_of_object);
#else
			*size = json_object_get_uint64(json_size_of_object);
#endif
			ret = APSH_S;
			goto exit;
		}
		*rva = 0;
		ret = volatility_ist_find_offset(object_with_json_format, symbol, subsymbol, rva);
	}

exit:
	debug_print(APSH_DEBUG_MISC, "APSH volat IST profile lookup %s %s: 0x%lx\n", symbol ?: NULL,
		    subsymbol ?: NULL, rva ? *rva : 0);
	return ret;
}

#define apsh_json_get json_object_object_get_ex
#define debug_exit(x...) {debug_print(APSH_DEBUG_MISC, x);goto exit;}

st_stat() volatility_profile_bitfield_offset_and_size(json_object *jobject_format, const char *symbol,
						      const char *subsymbol, address_to_t *rva,
						      size_t *start_bit, size_t *end_bit)
{
	st_stat() ret = APSH_F;
	size_t bit_length;
	json_object *utypes = NULL, *jobject = NULL, *fields = NULL, *type = NULL,
		    *jmember = NULL, *jvalue = NULL;
	if (!jobject_format || !symbol)
		return ret;

	if (!apsh_json_get(jobject_format, "user_types", &utypes))
		debug_exit("APSH volat no utypes section found\n");
	if (!apsh_json_get(utypes, symbol, &jobject))
		debug_exit("APSH volat no %s found\n", symbol);
	if (!apsh_json_get(jobject, "fields", &fields))
		debug_print(APSH_DEBUG_MISC, "APSH volat no fields found\n");
	if (!apsh_json_get(fields, subsymbol, &jmember))
		debug_exit("APSH volat %s has no %s member\n", symbol, subsymbol);
	if (!apsh_json_get(jmember, "offset", &jvalue))
		debug_exit("APSH volat IST %s.%s has no offset defined\n", symbol, subsymbol);

	*rva = json_object_get_uint64(jvalue);

	if (!apsh_json_get(jmember, "type", &type))
		debug_exit("APSH volat no type found\n");

	if (!apsh_json_get(type, "bit_position", &jvalue))
		debug_exit("APSH volat %s.%s has no member bit_position\n", symbol, subsymbol);

	*start_bit = json_object_get_uint64(jvalue);

	if (!apsh_json_get(type, "bit_length", &jvalue))
		debug_exit("APSH volat %s.%s has no member bit_length\n",symbol, subsymbol);

	bit_length = json_object_get_uint64(jvalue);

	*end_bit = *start_bit + bit_length;

	ret = APSH_S;

exit:
	debug_print(APSH_DEBUG_MISC,
		    "APSH volat profile lookup %s %s: offset 0x%lx, start bit %ld, end bit %ld\n",
		    symbol ?: NULL, subsymbol ?: NULL, *rva, *start_bit, *end_bit);

	return ret;
}

static st_stat() volatility_ist_find_offset(json_object *object_with_json_format, const char *symbol,
					    const char *subsymbol, address_to_t *rva)
{
	st_stat() ret = APSH_F;
	json_object *user_types = NULL, *json_struct_object = NULL, *jstruct2 = NULL, *jmember = NULL,
		    *jvalue = NULL;
	struct json_object_iterator iter, iend;

	if (!apsh_json_get(object_with_json_format, "user_types", &user_types))
		debug_exit("APSH volat IST no user_types section found\n");
	if (!apsh_json_get(user_types, symbol, &json_struct_object)) 
		debug_exit("APSH volat IST no %s found\n", symbol);

	if (!apsh_json_get(json_struct_object, "fields", &jstruct2))
		debug_exit("APSH volat IST struct %s has no fields element\n", symbol);

	if (apsh_json_get(jstruct2, subsymbol, &jmember)) {
		if (!apsh_json_get(jmember, "offset", &jvalue))
			debug_exit("APSH volat IST %s.%s has no offset\n", symbol, subsymbol);
		ret = APSH_S;
#ifdef JSONC_UINT64_SUPPORT
		*rva += json_object_get_uint64(jvalue);
#else
		*rva += json_object_get_uint64(jvalue);
#endif
		goto exit;
	}

	iter = json_object_iter_begin(jstruct2);
	iend = json_object_iter_end(jstruct2);

	while (!json_object_iter_equal(&iter, &iend)) {
		json_object *subval = NULL, *subval2 = NULL, *subval3 = NULL;
		const char *subname1 = NULL;
		const char *embedded = NULL;

		subval = json_object_iter_peek_value(&iter);
		subname1 = json_object_iter_peek_name(&iter);
		(void)subname1;

		if (!apsh_json_get(subval, "type", &subval2))
			goto next;

		if (!apsh_json_get(subval2, "name", &subval3))
			goto next;

		embedded = json_object_get_string(subval3);
		if (!embedded)
			goto next;

		debug_print(APSH_DEBUG_MISC,
			    "APSH volat IST exploring anonymous/embedded struct %s (%s) "
			    "for offset for %s\n",
			    subname1, embedded, subsymbol);
		ret = volatility_ist_find_offset(object_with_json_format, embedded, subsymbol, rva);
		if (ret == APSH_S) {
			json_object *jofs = NULL;

			if (!apsh_json_get(subval, "offset", &jofs)) {
				ret = APSH_F;
				debug_print(APSH_DEBUG_MISC,
					    "APSH volat IST anonymous struct %s has no "
					    "offset in %s\n",
					    subname1, symbol);
				goto exit;
			}

#ifdef JSONC_UINT64_SUPPORT
			*rva += json_object_get_uint64(jofs);
#else
			*rva += json_object_get_uint64(jofs);
#endif
			debug_print(APSH_DEBUG_MISC, "APSH volat IST %s.%s @ offset %ld\n",
				    symbol, subname1, *rva);
			goto exit;
		}

next:
		json_object_iter_next(&iter);
	}

exit:
	return ret;
}

