/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <string.h>
#include <wchar.h>

#include "doca_mema.h"
#include "driver/driver_wrapper.h"
#include "apsh_instance.h"
#include "doca_apsh_attr.h"

#define  DOCA_APSH_DEFAULT_STR_LEN_LIMIT 10000

#define st_stat() status_t

st_stat() apsh_read(apsh_instance_t apsh, const access_context_t *ctx, size_t count, void *buf,
		    size_t *bytes_read)
{
	st_stat() ret = APSH_F;
	size_t buf_offset = 0;
	addr_t start_addr;
	addr_t paddr;
	addr_t naddr;
	addr_t pfn;
	addr_t offset;
	addr_t pt;
	page_mode_t pm;
	addr_t npt;
	page_mode_t npm;

#ifdef ENABLE_SAFETY_CHECKS
	if (apsh == NULL) {
		debug_print(APSH_DEBUG_READ, "--%s: apsh passed as NULL, returning without read\n",
			    __FUNCTION__);
		goto done;
	}

	if (ctx == NULL) {
		debug_print(APSH_DEBUG_READ, "--%s: ctx passed as NULL, returning without read\n",
			    __FUNCTION__);
		goto done;
	}

	if (buf == NULL) {
		debug_print(APSH_DEBUG_READ, "--%s: buf passed as NULL, returning without read\n",
			    __FUNCTION__);
		goto done;
	}

	if (ctx->version != ACCESS_CONTEXT_VERSION) {
		if (!apsh->actx_version_warn_once)
			print_error(
				"--%s: access context version mismatch, please update your code\n",
				__FUNCTION__);
		apsh->actx_version_warn_once = 1;
	}
#endif

	pt = ctx->pt;
	pm = ctx->pm;
	npt = ctx->npt;
	npm = ctx->npm;
	start_addr = ctx->addr;

	switch (ctx->tm) {
	case APSH_TM_NONE:
		pm = APSH_PM_NONE;
		pt = 0;
		break;
	case APSH_TM_KERNEL_SYMBOL:
#ifdef ENABLE_SAFETY_CHECKS
		if (!apsh->os_interface || !apsh->kpgd)
			goto done;
#endif
		if (apsh_translate_ksym2v(apsh, ctx->ksym, &start_addr) == APSH_F)
			goto done;

		pt = apsh->kpgd;
		if (!pm)
			pm = apsh->page_mode;

		break;
	case APSH_TM_PROCESS_PID:
#ifdef ENABLE_SAFETY_CHECKS
		if (!apsh->os_interface)
			goto done;
#endif

		if (!ctx->pid)
			pt = apsh->kpgd;
		else if (ctx->pid > 0) {
			if (apsh_pid_to_dtb(apsh, ctx->pid, &pt) == APSH_F)
				goto done;
		}
		if (!pm)
			pm = apsh->page_mode;
		if (!pt)
			goto done;
		break;
	case APSH_TM_PROCESS_DTB:
		if (!pm)
			pm = apsh->page_mode;
		break;
	default:
		print_error("%s error: translation mechanism is not defined. Type: %d\n",
			    __FUNCTION__, ctx->tm);
		goto done;
	}

	if (apsh->EPT && !apsh->doing_page_look_up && !valid_npm(npm)) {
		npm = APSH_PM_EPT_4L;
		npt = apsh->EPT;
	}

#ifdef ENABLE_SAFETY_CHECKS
	if (pt && !valid_pm(pm)) {
		debug_print(APSH_DEBUG_READ, "--%s: pagetable specified with no page mode\n",
			    __FUNCTION__);
		goto done;
	}

	if (npt && !valid_npm(npm)) {
		debug_print(APSH_DEBUG_READ,
			    "--%s: nested pagetable specified with no nested page mode\n",
			    __FUNCTION__);
	}
#endif

	while (count > 0) {
		size_t read_len = 0;

		if (valid_pm(pm)) {
			if (apsh_nested_pagetable_lookup(apsh, npt, npm, pt, pm,
								   start_addr + buf_offset, &paddr,
								   &naddr) != APSH_S)
				goto done;

			if (valid_npm(npm)) {
				debug_print(APSH_DEBUG_READ,
					    "--Setting paddr to nested address 0x%lx\n", naddr);
				paddr = naddr;
			}
		} else {
			paddr = start_addr + buf_offset;

			if (valid_npm(npm) &&
			    apsh_nested_pagetable_lookup(apsh, 0, 0, npt, npm, paddr,
								   &paddr, NULL) != APSH_S)
				goto done;
		}

		pfn = paddr >> apsh->page_shift;
		debug_print(APSH_DEBUG_READ, "--Reading pfn 0x%lx\n", pfn);

		offset = (apsh->page_size - 1) & paddr;

		if ((offset + count) > apsh->page_size)
			read_len = apsh->page_size - offset;
		else
			read_len = count;

		void *buffer = apsh_read_memory(apsh, (pfn << apsh->page_shift) + (addr_t)offset,
						read_len);

		if (buffer == NULL) {
			goto done;
		}

		memcpy(((char *)buf) + (addr_t)buf_offset, buffer, read_len);

		count -= read_len;
		buf_offset += read_len;
		g_free(buffer);
	}

	ret = APSH_S;

done:
	if (bytes_read)
		*bytes_read = buf_offset;

	return ret;
}

st_stat() apsh_read_pa(apsh_instance_t apsh, addr_t paddr, size_t count, void *buf,
		       size_t *bytes_read)
{
	ACCESS_CONTEXT(ctx, .addr = paddr);
	return apsh_read(apsh, &ctx, count, buf, bytes_read);
}

st_stat() apsh_read_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid, size_t count, void *buf,
		       size_t *bytes_read)
{
	ACCESS_CONTEXT(ctx, .translate_mechanism = APSH_TM_PROCESS_PID, .addr = vaddr, .pid = pid);

	return apsh_read(apsh, &ctx, count, buf, bytes_read);
}

st_stat() apsh_read_ksym(apsh_instance_t apsh, const char *sym, size_t count, void *buf,
			 size_t *bytes_read)
{
	ACCESS_CONTEXT(ctx, .translate_mechanism = APSH_TM_KERNEL_SYMBOL, .pm = apsh->page_mode,
		       .ksym = sym);

	return apsh_read(apsh, &ctx, count, buf, bytes_read);
}

st_stat() apsh_read_8(apsh_instance_t apsh, const access_context_t *ctx, uint8_t *value)
{
	return apsh_read(apsh, ctx, 1, value, NULL);
}

st_stat() apsh_read_16(apsh_instance_t apsh, const access_context_t *ctx, uint16_t *value)
{
	return apsh_read(apsh, ctx, 2, value, NULL);
}

st_stat() apsh_read_32(apsh_instance_t apsh, const access_context_t *ctx, uint32_t *value)
{
	return apsh_read(apsh, ctx, 4, value, NULL);
}

st_stat() apsh_read_64(apsh_instance_t apsh, const access_context_t *ctx, uint64_t *value)
{
	return apsh_read(apsh, ctx, 8, value, NULL);
}

st_stat() apsh_read_addr(apsh_instance_t apsh, const access_context_t *ctx, addr_t *value)
{
	st_stat() ret = APSH_F;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh) {
		debug_print(APSH_DEBUG_READ, "--%s: apsh passed as NULL", __FUNCTION__);
		return APSH_F;
	}
#endif

	switch (apsh->page_mode) {
	case APSH_PM_AARCH64:
	case APSH_PM_IA32E:
		ret = apsh_read(apsh, ctx, 8, value, NULL);
		break;
	case APSH_PM_AARCH32:
	case APSH_PM_LEGACY:
	case APSH_PM_PAE: {
		uint32_t tmp = 0;

		ret = apsh_read(apsh, ctx, 4, &tmp, NULL);
		*value = 0;
		*value = (addr_t)tmp;
		break;
	}
	default:
		debug_print(APSH_DEBUG_READ,
			    "--%s: unknown page mode, can't read addr as width is unknown",
			    __FUNCTION__);
		break;
	}

	return ret;
}

char *
apsh_read_str(apsh_instance_t apsh, const access_context_t *ctx)
{
	int limit = DOCA_APSH_DEFAULT_STR_LEN_LIMIT;

	if (apsh->limits_table != NULL)
		limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_STRING_LIMIT)));

	access_context_t _ctx = *ctx;
	addr_t len = 0;
	uint8_t buf[APSH_PS_4KB];
	size_t bytes_read;
	bool read_more = 1;
	char *ret = NULL;

	do {
		size_t offset = _ctx.addr & APSH_BIT_MASK(0, 11);
		size_t read_size = APSH_PS_4KB - offset;

		debug_print(APSH_DEBUG_READ,
			    "--start to read string from 0x%lx, page offset 0x%lx, read: %lu\n",
			    _ctx.addr, offset, read_size);

		if (apsh_read(apsh, &_ctx, read_size, (void *)&buf, &bytes_read) == APSH_F &&
		    !bytes_read) {
			return ret;
		}

		size_t read_len = 0;

		for (read_len = 0; read_len < bytes_read; read_len++) {
			if (buf[read_len] == '\0') {
				read_more = 0;
				break;
			}
		}

		/*
		 * Realloc, tack on the '\0' in case of errors and
		 * get ready to read the next page.
		 */
		char *_ret = realloc(ret, len + read_len + 1);

		if (!_ret)
			return ret;

		ret = _ret;
		memcpy(&ret[len], &buf, read_len);
		len += read_len;
		ret[len] = '\0';
		_ctx.addr += offset;
	} while (read_more && len < limit);

	return ret;
}

unicode_string_t *
apsh_read_unicode_str(apsh_instance_t apsh, const access_context_t *ctx)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh) {
		debug_print(APSH_DEBUG_READ, "--%s: apsh passed as NULL, returning without read",
			    __FUNCTION__);
		return NULL;
	}
	if (!apsh->os_interface || !apsh->os_interface->os_read_unicode_struct)
		return NULL;
#endif

	return apsh->os_interface->os_read_unicode_struct(apsh, ctx);
}

unicode_string_t *
apsh_read_unicode_str_pm(apsh_instance_t apsh, const access_context_t *ctx, page_mode_t mode)
{
#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh) {
		debug_print(APSH_DEBUG_READ, "--%s: apsh passed as NULL, returning without read",
			    __FUNCTION__);
		return NULL;
	}
	if (!apsh->os_interface || !apsh->os_interface->os_read_unicode_struct_pm)
		return NULL;
#endif

	return apsh->os_interface->os_read_unicode_struct_pm(apsh, ctx, mode);
}

st_stat() apsh_read_8_pa(apsh_instance_t apsh, addr_t paddr, uint8_t *value)
{
	return apsh_read_pa(apsh, paddr, 1, value, NULL);
}

st_stat() apsh_read_16_pa(apsh_instance_t apsh, addr_t paddr, uint16_t *value)
{
	return apsh_read_pa(apsh, paddr, 2, value, NULL);
}

st_stat() apsh_read_32_pa(apsh_instance_t apsh, addr_t paddr, uint32_t *value)
{
	return apsh_read_pa(apsh, paddr, 4, value, NULL);
}

st_stat() apsh_read_64_pa(apsh_instance_t apsh, addr_t paddr, uint64_t *value)
{
	return apsh_read_pa(apsh, paddr, 8, value, NULL);
}

st_stat() apsh_read_addr_pa(apsh_instance_t apsh, addr_t paddr, addr_t *value)
{
	st_stat() ret = APSH_F;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh) {
		debug_print(APSH_DEBUG_MISC, "--%s: apsh passed as NULL, returning without read",
			    __FUNCTION__);
		return APSH_F;
	}
#endif

	switch (apsh->page_mode) {
	case APSH_PM_AARCH64:
	case APSH_PM_IA32E:
		ret = apsh_read_pa(apsh, paddr, 8, value, NULL);
		break;
	case APSH_PM_AARCH32:
	case APSH_PM_LEGACY:
	case APSH_PM_PAE: {
		uint32_t tmp = 0;

		ret = apsh_read_pa(apsh, paddr, 4, &tmp, NULL);
		*value = 0;
		*value = (addr_t)tmp;
		break;
	}
	default:
		debug_print(APSH_DEBUG_MISC,
			    "--%s: unknown page mode, can't read addr as width is unknown",
			    __FUNCTION__);
		break;
	}

	return ret;
}

char *
apsh_read_str_pa(apsh_instance_t apsh, addr_t paddr)
{
	ACCESS_CONTEXT(ctx, .addr = paddr);
	return apsh_read_str(apsh, &ctx);
}

st_stat() apsh_read_8_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid, uint8_t *value)
{
	return apsh_read_va(apsh, vaddr, pid, 1, value, NULL);
}

st_stat() apsh_read_16_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid, uint16_t *value)
{
	return apsh_read_va(apsh, vaddr, pid, 2, value, NULL);
}

st_stat() apsh_read_32_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid, uint32_t *value)
{
	return apsh_read_va(apsh, vaddr, pid, 4, value, NULL);
}

st_stat() apsh_read_64_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid, uint64_t *value)
{
	return apsh_read_va(apsh, vaddr, pid, 8, value, NULL);
}

st_stat() apsh_read_addr_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid, addr_t *value)
{
	st_stat() ret = APSH_F;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh) {
		debug_print(APSH_DEBUG_READ, "--%s: apsh passed as NULL, returning without read",
			    __FUNCTION__);
		return APSH_F;
	}
#endif

	switch (apsh->page_mode) {
	case APSH_PM_AARCH64:
	case APSH_PM_IA32E:
		ret = apsh_read_va(apsh, vaddr, pid, 8, value, NULL);
		break;
	case APSH_PM_AARCH32:
	case APSH_PM_LEGACY:
	case APSH_PM_PAE: {
		uint32_t tmp = 0;

		ret = apsh_read_va(apsh, vaddr, pid, 4, &tmp, NULL);
		*value = 0;
		*value = (addr_t)tmp;
		break;
	}
	default:
		debug_print(APSH_DEBUG_READ,
			    "--%s: unknown page mode, can't read addr as width is unknown",
			    __FUNCTION__);
		break;
	}

	return ret;
}

char *
apsh_read_str_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid)
{
	ACCESS_CONTEXT(ctx, .translate_mechanism = APSH_TM_PROCESS_PID, .addr = vaddr, .pid = pid);

	return apsh_read_str(apsh, &ctx);
}

unicode_string_t *
apsh_read_unicode_str_va(apsh_instance_t apsh, addr_t vaddr, apsh_pid_t pid)
{
	ACCESS_CONTEXT(ctx, .translate_mechanism = APSH_TM_PROCESS_PID, .addr = vaddr, .pid = pid);

	return apsh_read_unicode_str(apsh, &ctx);
}

st_stat() apsh_read_8_ksym(apsh_instance_t apsh, const char *sym, uint8_t *value)
{
	return apsh_read_ksym(apsh, sym, 1, value, NULL);
}

st_stat() apsh_read_16_ksym(apsh_instance_t apsh, const char *sym, uint16_t *value)
{
	return apsh_read_ksym(apsh, sym, 2, value, NULL);
}

st_stat() apsh_read_32_ksym(apsh_instance_t apsh, const char *sym, uint32_t *value)
{
	return apsh_read_ksym(apsh, sym, 4, value, NULL);
}

st_stat() apsh_read_64_ksym(apsh_instance_t apsh, const char *sym, uint64_t *value)
{
	return apsh_read_ksym(apsh, sym, 8, value, NULL);
}

st_stat() apsh_read_addr_ksym(apsh_instance_t apsh, const char *sym, addr_t *value)
{
	st_stat() ret = APSH_F;

#ifdef ENABLE_SAFETY_CHECKS
	if (!apsh) {
		debug_print(APSH_DEBUG_READ, "--%s: apsh passed as NULL, returning without read",
			    __FUNCTION__);
		return APSH_F;
	}
#endif

	switch (apsh->page_mode) {
	case APSH_PM_AARCH64:
	case APSH_PM_IA32E:
		ret = apsh_read_ksym(apsh, sym, 8, value, NULL);
		break;
	case APSH_PM_AARCH32:
	case APSH_PM_LEGACY:
	case APSH_PM_PAE: {
		uint32_t tmp = 0;

		ret = apsh_read_ksym(apsh, sym, 4, &tmp, NULL);
		*value = 0;
		*value = (addr_t)tmp;
		break;
	}
	default:
		debug_print(APSH_DEBUG_READ,
			    "--%s: unknown page mode, can't read addr as width is unknown",
			    __FUNCTION__);
		break;
	}

	return ret;
}

char *
apsh_read_str_ksym(apsh_instance_t apsh, const char *sym)
{
	addr_t vaddr = 0;

	if (apsh_translate_ksym2v(apsh, sym, &vaddr) == APSH_F)
		return NULL;

	return apsh_read_str_va(apsh, vaddr, 0);
}
