/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <string.h>
#include <time.h>

#include <doca_buf.h>
#include <doca_ctx.h>
#include <doca_buf_inventory.h>
#include <doca_dev.h>
#include <doca_error.h>
#include <doca_log.h>
#include <doca_mmap.h>
#include <doca_regex.h>
#include <doca_regex_mempool.h>
#include <priv_doca_buf.h>
#include "regexoob.h"

#define MEMPOOL_SIZE 8
#define CHUNCK_NUM 1
#define CHUNCK_LEN (1<<23)
#define MINIMUM_CHUNCK_LEN (1<<12)
#define SLEEP_IN_NANOS (10 * 1000)

DOCA_LOG_REGISTER(APSH::LIBS::REGEX::OOB);


struct regexoob_ctx {
	uint8_t *rules_buffer;				/* Rules buffer */
	size_t rules_buffer_len;				/* Length of rules buffer */
	memp_t *mem_regions;				/* Allowed memory regions */
	int chunk_len;					/* size of chunk to send to regex */
	struct doca_buf *buf;				/* active job buffer */
	struct doca_buf_inventory *buf_inv;		/* Pool of doca_buf objects */
	struct doca_dev *dev;				/* DOCA device */
	struct doca_dev_rep *host_rep;			/* DOCA device */
	struct doca_mmap *mmap;				/* DOCA Memory orchestration */
	struct doca_regex *doca_regex;			/* DOCA RegEx interface */
	struct doca_workq *workq;			/* DOCA work queue */
	struct doca_regex_search_result *results;	/* Pointer to array of result objects */
};

static void
add_regex_result(regexoob_result_t **regex_result, uint64_t addr, uint32_t len, int rule_id)
{
	regexoob_result_t *new_regex_result = calloc(1, sizeof(regexoob_result_t));

	if (new_regex_result == NULL)
		return;
	new_regex_result->phy_add = addr;
	new_regex_result->len = len;
	new_regex_result->rule_id = rule_id;
	new_regex_result->next = *regex_result;
	*regex_result = new_regex_result;
}

static int32_t
regex_scan_enq_job(struct regexoob_ctx *regex_oob, struct doca_regex_job_search *job_request,
		   uint64_t const read_offset, uint32_t const read_len)
{
	doca_error_t result;
	uint32_t nb_free = 0;

	doca_buf_inventory_get_num_free_elements(regex_oob->buf_inv, &nb_free);

	if (nb_free != 0) {
		struct doca_buf *buf;
		void *mbuf_data;
		uint32_t const job_size = read_len;

		if (doca_buf_inventory_buf_by_addr(regex_oob->buf_inv, regex_oob->mmap,
						   (void *)(read_offset), job_size,
						   &buf) != DOCA_SUCCESS)
			return -1;

		doca_buf_get_data(buf, &mbuf_data);
		doca_buf_set_data(buf, mbuf_data, job_size);

		regex_oob->buf = buf;
		job_request->buffer = buf;
		job_request->result = regex_oob->results;
		job_request->allow_batching = false;
		result = doca_workq_submit(regex_oob->workq, (struct doca_job *)job_request);
		if (result == DOCA_ERROR_NO_MEMORY) {
			doca_buf_refcount_rm(buf, NULL);
			return 0; /* QP is full, try to dequeue. */
		}
		if (result != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Unable to enqueue job.");
			return -1;
		}
	}

	return 1;
}

static int32_t
regex_scan_deq_job(struct regexoob_ctx *regex_oob, uint64_t offset, regexoob_result_t **regex_result)
{
	doca_error_t result;
	struct doca_event event = {0};
	struct timespec ts;

	do {
		result = doca_workq_progress_retrieve(regex_oob->workq, &event, DOCA_WORKQ_RETRIEVE_FLAGS_NONE);
		if (result == DOCA_SUCCESS) {
			doca_buf_refcount_rm(regex_oob->buf, NULL);
			struct doca_regex_search_result *search_result = (struct doca_regex_search_result *)event.result.ptr;

			if (search_result->num_matches == 0)
				return 1;
			struct doca_regex_match *ptr = search_result->matches;

			while (ptr != NULL) {
				add_regex_result(regex_result, offset + ptr->match_start, ptr->length, ptr->rule_id);
				struct doca_regex_search_result *const to_release_match = ptr;

				ptr = ptr->next;
				doca_regex_mempool_put_obj(search_result->matches_mempool, to_release_match);
			}
		} else if (result == DOCA_ERROR_AGAIN) {
			/* Wait for the job to complete */
			ts.tv_sec = 0;
			ts.tv_nsec = SLEEP_IN_NANOS;
			nanosleep(&ts, &ts);
		} else {
			DOCA_DLOG_ERR("Failed to dequeue results.");
			return -1;
		}

	} while (result == DOCA_ERROR_AGAIN);

	return 1;
}


status_t
regexoob_init(uint8_t *rules_buffer, size_t rules_buffer_len, struct doca_dev_rep *host_rep, struct doca_dev *regex_dev, const char *mem_regions, regexoob_ctx_t *regex_oob)
{
	if (regex_dev == NULL) {
		DOCA_DLOG_INFO("init failed, regex_dev is NULL.");
		return APSH_F;
	}

	if (rules_buffer == NULL || rules_buffer_len <= 0 || host_rep == NULL || mem_regions == NULL || regex_oob == NULL) {
		DOCA_DLOG_ERR("init failed, one of the arguments is NULL.");
		return APSH_F;
	}

	struct doca_devinfo *regex_info = doca_dev_as_devinfo(regex_dev);

	if (regex_info == NULL) {
		DOCA_DLOG_ERR("failed to get regex_dev info");
		return APSH_F;
	}

	if (doca_regex_get_hardware_supported(regex_info) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("regex_dev doesn't support regex offload");
		return APSH_F;
	}

	struct regexoob_ctx *new_regexoob = calloc(1, sizeof(struct regexoob_ctx));

	if (new_regexoob == NULL) {
		DOCA_DLOG_ERR("failed to allocale regex_oob");
		return APSH_F;
	}

	int ret;
	doca_error_t result;
	bool regex_started = false, inventory_started = false;

	*regex_oob = new_regexoob;
	new_regexoob->dev = regex_dev;
	new_regexoob->host_rep = host_rep;

	new_regexoob->mem_regions = create_machine_mem_regions(mem_regions);
	if (new_regexoob->mem_regions == NULL) {
		DOCA_DLOG_ERR("failed to read memory regions file");
		goto error_exit;
	}

	new_regexoob->rules_buffer = calloc(rules_buffer_len, sizeof(uint8_t));

	if (new_regexoob->rules_buffer == NULL) {
		DOCA_DLOG_ERR("failed to read rules file");
		goto error_exit;
	}

	memcpy(new_regexoob->rules_buffer, rules_buffer, rules_buffer_len);
	new_regexoob->rules_buffer_len = rules_buffer_len;

	ret = doca_regex_create(&(new_regexoob->doca_regex));
	if (ret < 0) {
		DOCA_DLOG_ERR("Unable to create RegEx device.");
		goto error_exit;
	}

	result = doca_ctx_dev_add(doca_regex_as_ctx(new_regexoob->doca_regex), new_regexoob->dev);
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to set RegEx device. [%s]", strerror(abs(ret)));
		goto error_exit;
	}

	result = doca_regex_set_workq_matches_memory_pool_size(new_regexoob->doca_regex, MEMPOOL_SIZE);
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable set matches mempool size. Reason: %s", doca_get_error_string(result));
		goto error_exit;
	}

	result = doca_regex_set_hardware_compiled_rules(new_regexoob->doca_regex, (uint8_t *)new_regexoob->rules_buffer,
			new_regexoob->rules_buffer_len);
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to program rules file. Reason: %s", doca_get_error_string(result));
		goto error_exit;
	}

	result = doca_ctx_start(doca_regex_as_ctx(new_regexoob->doca_regex));
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to start DOCA RegEx. [%s]", doca_get_error_string(result));
		goto error_exit;
	}
	regex_started = true;

	result = doca_workq_create(CHUNCK_NUM, &(new_regexoob->workq));
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create work queue. Reason: %s", doca_get_error_string(result));
		goto error_exit;
	}

	result = doca_ctx_workq_add(doca_regex_as_ctx(new_regexoob->doca_regex), new_regexoob->workq);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to attach work queue to RegEx. Reason: %s", doca_get_error_string(result));
		goto error_exit;
	}

	result = doca_buf_inventory_create(NULL, 1, DOCA_BUF_EXTENSION_NONE, &new_regexoob->buf_inv);
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create buffer inventory. Reason: %s", doca_get_error_string(result));
		goto error_exit;
	}

	result = doca_buf_inventory_start(new_regexoob->buf_inv);
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to start buffer inventory. Reason: %s", doca_get_error_string(result));
		goto error_exit;
	}
	inventory_started = true;

	result = priv_doca_mmap_create_from_rep(NULL, new_regexoob->host_rep, &new_regexoob->mmap);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create memory map. Reason: %s", doca_get_error_string(result));
		goto error_exit;
	}

	new_regexoob->results = calloc(CHUNCK_NUM, sizeof(struct doca_regex_search_result));
	if (new_regexoob->results == NULL) {
		DOCA_LOG_ERR("Unable to add allocate results storage");
		goto error_exit;
	}

	return APSH_S;

error_exit:

	if (new_regexoob->mem_regions != NULL)
		destroy_pf_mem_regions(new_regexoob->mem_regions);
	if (new_regexoob->rules_buffer != NULL)
		free(new_regexoob->rules_buffer);
	if (new_regexoob->workq != NULL) {
		doca_ctx_workq_rm(doca_regex_as_ctx(new_regexoob->doca_regex), new_regexoob->workq);
		doca_workq_destroy(new_regexoob->workq);
	}
	if (regex_started)
		doca_ctx_stop(doca_regex_as_ctx(new_regexoob->doca_regex));
	if (new_regexoob->doca_regex != NULL)
		doca_regex_destroy(new_regexoob->doca_regex);
	if (new_regexoob->mmap != NULL)
		doca_mmap_destroy(new_regexoob->mmap);
	if (inventory_started)
		doca_buf_inventory_stop(new_regexoob->buf_inv);
	if (new_regexoob->buf_inv != NULL)
		doca_buf_inventory_destroy(new_regexoob->buf_inv);
	if (new_regexoob->results != NULL)
		free(new_regexoob->results);
	free(new_regexoob);
	return APSH_F;
}

status_t
regexoob_scan(regexoob_ctx_t regex_oob, uint64_t start_phy_add, uint32_t len, regexoob_result_t **regex_result)
{
	if (regex_result == NULL) {
		DOCA_DLOG_ERR("scan falied, one of the arguments is NULL");
		return APSH_F;
	}

	*regex_result = NULL;
	uint64_t current_addr = start_phy_add;
	struct doca_regex_job_search job_request = {0};
	uint32_t delta_len;

	job_request.base.type = DOCA_REGEX_JOB_SEARCH;
	job_request.rule_group_ids[0] = 1;
	job_request.base.ctx = doca_regex_as_ctx(regex_oob->doca_regex);
	while (current_addr < (start_phy_add + len)) {
		if (pf_mem_is_valid(regex_oob->mem_regions, current_addr, CHUNCK_LEN) == 1) {
			delta_len = CHUNCK_LEN;
		} else {
			delta_len = MINIMUM_CHUNCK_LEN;
			if (pf_mem_is_valid(regex_oob->mem_regions, current_addr, MINIMUM_CHUNCK_LEN) != 1) {
				current_addr += delta_len;
				continue;
			}
		}

		if (regex_scan_enq_job(regex_oob, &job_request, current_addr, delta_len) < 0)
			return APSH_F;

		if (regex_scan_deq_job(regex_oob, current_addr, regex_result) < 0)
			return APSH_F;

		current_addr += delta_len;
	}

	return APSH_S;
}


void
regexoob_destroy(regexoob_ctx_t regex_oob)
{
	if (regex_oob == NULL)
		return;

	struct regexoob_ctx *to_delete = regex_oob;

	destroy_pf_mem_regions(to_delete->mem_regions);
	free(to_delete->rules_buffer);
	doca_ctx_workq_rm(doca_regex_as_ctx(to_delete->doca_regex), to_delete->workq);
	doca_workq_destroy(to_delete->workq);
	doca_ctx_stop(doca_regex_as_ctx(to_delete->doca_regex));
	doca_regex_destroy(to_delete->doca_regex);
	doca_mmap_destroy(to_delete->mmap);
	doca_buf_inventory_stop(to_delete->buf_inv);
	doca_buf_inventory_destroy(to_delete->buf_inv);
	free(to_delete->results);

	free(to_delete);
}

void
regexoob_result_destroy(regexoob_result_t *regex_result)
{
	regexoob_result_t *p_next;

	while (regex_result) {
		p_next = regex_result->next;
		free(regex_result);
		regex_result = p_next;
	}
}

