/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>

#include "private/common_defines.h"

#include "apsh_instance.h"

#define ASIZE 256

#define number_1 m
#define number_2 n

static void
preBmBc(unsigned char *x, int number_1, int bmBc[])
{
	int index_1;

	for (index_1 = 0; index_1 < ASIZE; ++index_1)
		bmBc[index_1] = number_1;

	for (index_1 = 0; index_1 < number_1 - 1; ++index_1)
		bmBc[x[index_1]] = number_1 - index_1 - 1;
}

static void
suffixes(unsigned char *x, int number_1, int *suff)
{
	int f = 0, g, index_1;

	suff[number_1 - 1] = number_1;
	g = number_1 - 1;
	for (index_1 = number_1 - 2; index_1 >= 0; --index_1) {
		if (index_1 > g && suff[index_1 + number_1 - 1 - f] < index_1 - g) {
			suff[index_1] = suff[index_1 + number_1 - 1 - f];
		} else {
			if (index_1 < g)
				g = index_1;

			f = index_1;
			while (g >= 0 && x[g] == x[g + number_1 - 1 - f])
				--g;

			suff[index_1] = f - g;
		}
	}
}

static void
preBmGs(unsigned char *x, int number_1, int bmGs[])
{
	int index_1, index_2;
	int *suff = safe_malloc(number_1 * sizeof(int));

	suffixes(x, number_1, suff);

	for (index_1 = 0; index_1 < number_1; ++index_1)
		bmGs[index_1] = number_1;

	index_2 = 0;
	for (index_1 = number_1 - 1; index_1 >= 0; --index_1) {
		if (suff[index_1] == index_1 + 1) {
			for (; index_2 < number_1 - 1 - index_1; ++index_2) {
				if (bmGs[index_2] == number_1)
					bmGs[index_2] = number_1 - 1 - index_1;
			}
		}
	}
	for (index_1 = 0; index_1 <= number_1 - 2; ++index_1)
		bmGs[number_1 - 1 - suff[index_1]] = number_1 - 1 - index_1;

	free(suff);
}

int
boyer_moore(unsigned char *x, int number_1, unsigned char *y, int number_2)
{
	int index_1, index_2, bmBc[ASIZE];
	int *bmGs = safe_malloc(number_1 * sizeof(int));

	/* Preprocessing */
	preBmGs(x, number_1, bmGs);
	preBmBc(x, number_1, bmBc);

	/* Searching */
	index_2 = 0;
	while (index_2 <= number_2 - number_1) {
		for (index_1 = number_1 - 1; index_1 >= 0 && x[index_1] == y[index_1 + index_2]; --index_1)
			;
		if (index_1 < 0) {
			free(bmGs);
			return index_2;

		} else {
			index_2 += MAX(bmGs[index_1], bmBc[y[index_1 + index_2]] - number_1 + 1 + index_1);
		}
	}

	free(bmGs);
	return -1;
}

struct boyer_moore_data_t {
	unsigned char *x;
	int number_1;
	int bmBc[ASIZE];
	int *bmGs;
};

void *
boyer_moore_init(unsigned char *x, int number_1)
{
	struct boyer_moore_data_t *bm = safe_malloc(sizeof(struct boyer_moore_data_t));

	bm->x = safe_malloc(number_1 * sizeof(*x));
	memcpy(bm->x, x, number_1 * sizeof(*x));

	bm->number_1 = number_1;

	bm->bmGs = safe_malloc(number_1 * sizeof(int));

	preBmGs(x, number_1, bm->bmGs);
	preBmBc(x, number_1, bm->bmBc);

	return (void *)bm;
}

void
boyer_moore_fini(void *bm)
{
	struct boyer_moore_data_t *_bm = (struct boyer_moore_data_t *)bm;

	free(_bm->x);
	free(_bm->bmGs);
	free(_bm);
}

int
boyer_moore2(void *bm, unsigned char *y, int number_2)
{
	int index_1, index_2;
	struct boyer_moore_data_t *_bm = (struct boyer_moore_data_t *)bm;
	int number_1 = _bm->number_1;
	unsigned char *x = _bm->x;

	index_2 = 0;
	while (index_2 <= number_2 - number_1) {
		for (index_1 = number_1 - 1; index_1 >= 0 && x[index_1] == y[index_1 + index_2]; --index_1)
			;
		if (index_1 < 0) {
			return index_2;

		} else {
			index_2 += MAX(_bm->bmGs[index_1],
				       _bm->bmBc[y[index_1 + index_2]] - number_1 + 1 + index_1);
		}
	}

	return -1;
}
