/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <infiniband/mlx5dv.h>
#include <infiniband/verbs.h>
#include <json-c/json.h>
#include <stdint.h>
#include <stdio.h>

#include "doca_mema.h"
#include "maoob_device.h"
#include "mem_pool.h"
#include "mema.h"
#include "mlx5_ifc.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(MEMA);

/**
 * struct sample_dma_request - An example of the DMA request
 * @data: data buffer to read data from host or to write data to host. It must
 *        reside in memory registered by the ibv_reg_mr()
 * @chnl: pointer to the controller
 * @comp: dma completion that will be passed to the snap_dma_q_read() or
 *        snap_dma_q_write()

struct sample_dma_request {
	void                    *data;
	struct maoob_channel    *ctrl;
	struct snap_dma_completion  comp;
};
*/

/* number of requests that can be used for async DMA operations */

#define MEMA_CHANNEL_N_REQUESTS 1
#define MEMA_CHANNEL_HUGE_PAGE_N_REQUESTS 1

#define SNAP_8KB_MEM_BUF 1
#define SNAP_32MB_MEM_BUF 1

/**
 * struct memma - Controller implementation for the simple emulated
 *                            device
 * @sctx: snap library context
 * @sdev: snap library 'emulated device'
 *


 * @dma_rkey:    host memory key. Must be obtained by get_introspection_mkey
 * @dma_membase: points to the block of memory registered with ibv_reg_mr()
 *               The memory is then split among dma requests. Usually it is
 *               somewhat better to have one big pool of registered memory.
 * @dma_reqs:    dma requests
 * @mr:     holds @dma_membase registration and lkey (local memory key)
 * @ib_ctx: ib verbs context. Represents device on which to create @dma_q
 * @pd:     protection domain created on @ib_ctx. Used by @dma_q
 *
 * @last_cmd: last command that was written to the command register
 * @enabled: controller status
 */

static void
mema_chnl_destroy_requests(struct doca_mema_channel *mdev)
{
	ibv_dereg_mr(mdev->mr);
	free(mdev->dma_membase);
	ibv_dereg_mr(mdev->hp_mr);
}

// allocate registeration buffer
static int
mema_chnl_allocate_dma_buf(struct doca_mema_channel *mdev)
{
	int i;

	mdev->dma_membase = calloc(MEMA_CHANNEL_N_REQUESTS, SNAP_SAMPLE_DEV_MAX_DATA_SIZE);

	if (!mdev->dma_membase) {
		DOCA_DLOG_ERR("Failed allocate requests memory %d\n", 0x10);
		return -1;
	}

	mdev->mr = ibv_reg_mr(mdev->pd, mdev->dma_membase,
			      MEMA_CHANNEL_N_REQUESTS * SNAP_SAMPLE_DEV_MAX_DATA_SIZE,
			      IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE |
				      IBV_ACCESS_REMOTE_READ);

	if (!mdev->mr) {
		DOCA_DLOG_ERR("Failed to register requests memory %d\n", 0x12);
		goto free_mem;
	}

	return 0;

free_mem:
	free(mdev->dma_membase);
	return -1;
}

static int
mema_chnl_allocate_hp_dma_buf(struct doca_mema_channel *mdev, enum huge_page p_size)
{
	int i;
	uint32_t hp_num = 0;

	alloc_hp_buf(&hp_num, &(mdev->hp_dma_membase), p_size);

	if (!mdev->hp_dma_membase) {
		DOCA_DLOG_ERR("Failed allocate requests memory %d", 0x10); // FIXME: provide a meaningful error code
		goto free_mem;
	}

	mdev->hp_mr = ibv_reg_mr(mdev->pd, mdev->hp_dma_membase, p_size,
				 IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE |
					 IBV_ACCESS_REMOTE_READ);

	if (!mdev->hp_mr) {
		DOCA_DLOG_ERR("Failed to register huge memory %d\n", 0x12); // FIXME: provide a meaningful error code
		goto free_mem;
	}
	mdev->hp_num = hp_num;

	return 0;

free_mem:
	return -1;
}

static int
mema_chnl_mem_pool_init(struct doca_mema_channel *mdev)
{
	int ret = 0;

	ret = mema_chnl_allocate_dma_buf(mdev);
	if (ret)
		goto err_mem;

	if (HUGE_PAGE_USE_ENABLE)
		ret = mema_chnl_allocate_hp_dma_buf(mdev, HUGE_PAGE_2MB);

	if (ret)
		goto err_mem;

	return 0;
err_mem:
	return -1;
}

static struct ibv_device *
find_device(struct ibv_device **dev_list, int n, char *dev_name)
{
	int i;

	for (i = 0; i < n; i++) {
		if (strcmp(ibv_get_device_name(dev_list[i]), dev_name) == 0)
			return dev_list[i];
	}
	return NULL;
}

/**
 * mema_chnl_init() - Initialize sample device controller
 * @dev_name:  device name (e.x mlx5_0) that has emulation manager
 *                 capabilities
 * @function_id: function_id of the machine we are inspecting e.g. VF_ID, PF_ID
 *
 * The function initializes resources required by the controller.
 *
 * Note: on Bluefield1 and Bluefield2 emulation manager device is different
 * from the rdma one. In fact emulation manager is not capable of doing RDMA.
 *
 * Return: pointer to the new controller or NULL on failure
 */
struct doca_mema_channel *
mema_chnl_init(char *dev_name, uint16_t pcie_function_id)
{
	// struct snap_device_attr snap_attr = {};
	// struct mlx5dv_context_attr rdma_attr = {};
	struct doca_mema_channel *mdev;
	struct ibv_device **dev_list, *dev;
	struct mema_device_attr attr;
	int n_dev;
	int rc = 0;

	mdev = calloc(sizeof(*mdev), 1);
	if (!mdev) {
		DOCA_DLOG_ERR("Failed to allocate controller");
		return NULL;
	}

	dev_list = ibv_get_device_list(&n_dev);
	if (!dev_list) {
		DOCA_DLOG_ERR("Failed to get device list");
		goto free_device;
	}

	/* Open device on the given mlx device */
	dev = find_device(dev_list, n_dev, dev_name);
	if (!dev) {
		DOCA_DLOG_ERR("The given device %s is not available", dev_name);
		goto free_dev_list;
	}

	mdev->ib_ctx = mema_open(dev);
	if (!mdev->ib_ctx) {
		DOCA_DLOG_ERR("Failed to open device context on %s", dev_name);
		goto free_dev_list;
	}

	attr.pcie_function_id = pcie_function_id;
	mdev->sdev = mema_open_device(mdev->ib_ctx, &attr);
	if (!mdev->sdev) {
		DOCA_DLOG_ERR("Failed open device on pf, errno=%d", errno);
		goto doca_mema_close;
	}

	mdev->pd = ibv_alloc_pd(mdev->ib_ctx);
	if (!mdev->pd) {
		DOCA_DLOG_ERR("Failed to create PD");
		goto close_ib_ctx;
	}

	rc = mema_chnl_mem_pool_init(mdev);
	if (rc)
		goto free_pd;

	ibv_free_device_list(dev_list);
	return mdev;

free_pd:
	ibv_dealloc_pd(mdev->pd);
close_ib_ctx:
	ibv_close_device(mdev->ib_ctx);
// snap_close_dev:
//	snap_close_device(ctrl->sdev);
doca_mema_close:
//	doca_chnl_close(ctrl->sctx);
free_dev_list:
	ibv_free_device_list(dev_list);
free_device:
	free(mdev);
	return NULL;
}

int
t_modify_qp(struct doca_mema_channel *mdev)
{
	struct ibv_qp_attr qp_attr = {};
	int ret;

	qp_attr.qp_state = IBV_QPS_RTR;
	qp_attr.path_mtu = IBV_MTU_1024;
	qp_attr.dest_qp_num = mdev->qp->qp_num;
	qp_attr.rq_psn = 0;
	qp_attr.max_dest_rd_atomic = 1;
	qp_attr.min_rnr_timer = 12;
	qp_attr.ah_attr.is_global = 0;
	qp_attr.ah_attr.src_path_bits = 0;
	qp_attr.ah_attr.port_num = 1;

	ret = ibv_modify_qp(mdev->qp, &qp_attr,
			    IBV_QP_STATE | IBV_QP_PATH_MTU | IBV_QP_DEST_QPN | IBV_QP_RQ_PSN |
				    IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER | IBV_QP_AV);
	if (ret) {
		DOCA_DLOG_ERR("AH: error: failed modify QP to RTR. ret=%d.\n", ret);
		return 1;
	}

	qp_attr.qp_state = IBV_QPS_RTS;
	qp_attr.timeout = 12;
	qp_attr.retry_cnt = 3;
	qp_attr.rnr_retry = 3;
	qp_attr.sq_psn = 0;
	qp_attr.max_rd_atomic = 1;
	ret = ibv_modify_qp(mdev->qp, &qp_attr,
			    IBV_QP_STATE | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT | IBV_QP_RNR_RETRY |
				    IBV_QP_SQ_PSN | IBV_QP_MAX_QP_RD_ATOMIC);
	if (ret) {
		DOCA_DLOG_ERR("AH: error: faield modify QP to RTS. ret=%d.\n", ret);
		return 1;
	}

	return 0;
}

static inline uint32_t
u32log2(uint32_t x)
{
	if (x == 0) {
		/* log(0) is undefined */
		return 0;
	}
	return 31u - __builtin_clz(x);
}

static int
modify_qp_to_init(struct ibv_qp *qp, struct ibv_qp_attr *qp_attr, int attr_mask)
{
	uint8_t in[DEVX_ST_SZ_BYTES(rst2init_qp_in)] = {0};
	uint8_t out[DEVX_ST_SZ_BYTES(rst2init_qp_out)] = {0};
	void *qpc = DEVX_ADDR_OF(rst2init_qp_in, in, qpc);

	DEVX_SET(rst2init_qp_in, in, opcode, MLX5_CMD_OP_RST2INIT_QP);
	DEVX_SET(rst2init_qp_in, in, qpn, qp->qp_num);
	DEVX_SET(qpc, qpc, pm_state, MLX5_QPC_PM_STATE_MIGRATED);

	if (attr_mask & IBV_QP_PKEY_INDEX)
		DEVX_SET(qpc, qpc, primary_address_path.pkey_index, qp_attr->pkey_index);

	if (attr_mask & IBV_QP_PORT)
		DEVX_SET(qpc, qpc, primary_address_path.vhca_port_num, qp_attr->port_num);

	if (attr_mask & IBV_QP_ACCESS_FLAGS) {
		if (qp_attr->qp_access_flags & IBV_ACCESS_REMOTE_READ)
			DEVX_SET(qpc, qpc, rre, 1);
		if (qp_attr->qp_access_flags & IBV_ACCESS_REMOTE_WRITE)
			DEVX_SET(qpc, qpc, rwe, 1);
	}

	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
}

static int
modify_qp_to_rtr(struct ibv_qp *qp, struct ibv_qp_attr *qp_attr, struct mlx5dv_ah *dv_ah,
		 int attr_mask)
{
	uint8_t in[DEVX_ST_SZ_BYTES(init2rtr_qp_in)] = {0};
	uint8_t out[DEVX_ST_SZ_BYTES(init2rtr_qp_out)] = {0};
	void *qpc = DEVX_ADDR_OF(init2rtr_qp_in, in, qpc);

	DEVX_SET(init2rtr_qp_in, in, opcode, MLX5_CMD_OP_INIT2RTR_QP);
	DEVX_SET(init2rtr_qp_in, in, qpn, qp->qp_num);

	/* 30 is the maximum value for Infiniband QPs*/
	DEVX_SET(qpc, qpc, log_msg_max, 30);

	if (attr_mask & IBV_QP_PATH_MTU)
		DEVX_SET(qpc, qpc, mtu, qp_attr->path_mtu);
	if (attr_mask & IBV_QP_DEST_QPN)
		DEVX_SET(qpc, qpc, remote_qpn, qp_attr->dest_qp_num);
	if (attr_mask & IBV_QP_RQ_PSN)
		DEVX_SET(qpc, qpc, next_rcv_psn, qp_attr->rq_psn & 0xffffff);
	if (attr_mask & IBV_QP_TIMEOUT)
		DEVX_SET(qpc, qpc, primary_address_path.ack_timeout, qp_attr->timeout);
	if (attr_mask & IBV_QP_PKEY_INDEX)
		DEVX_SET(qpc, qpc, primary_address_path.pkey_index, qp_attr->pkey_index);
	if (attr_mask & IBV_QP_PORT)
		DEVX_SET(qpc, qpc, primary_address_path.vhca_port_num, qp_attr->port_num);
	if (attr_mask & IBV_QP_MAX_DEST_RD_ATOMIC)
		DEVX_SET(qpc, qpc, log_rra_max, u32log2(qp_attr->max_dest_rd_atomic));
	if (attr_mask & IBV_QP_MIN_RNR_TIMER)
		DEVX_SET(qpc, qpc, min_rnr_nak, qp_attr->min_rnr_timer);
	if (attr_mask & IBV_QP_AV) {
		DEVX_SET(qpc, qpc, primary_address_path.rlid, qp_attr->ah_attr.dlid);
		DEVX_SET(qpc, qpc, primary_address_path.grh, 0);
		if (qp_attr->ah_attr.sl & 0xf)
			DEVX_SET(qpc, qpc, primary_address_path.sl, qp_attr->ah_attr.sl & 0xf);
	}

	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
}

static int
modify_qp_to_rts(struct ibv_qp *qp, struct ibv_qp_attr *qp_attr, int attr_mask)
{
	uint32_t in[DEVX_ST_SZ_DW(rtr2rts_qp_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(rtr2rts_qp_out)] = {0};
	void *qpc = DEVX_ADDR_OF(rtr2rts_qp_in, in, qpc);

	DEVX_SET(rtr2rts_qp_in, in, opcode, MLX5_CMD_OP_RTR2RTS_QP);
	DEVX_SET(rtr2rts_qp_in, in, qpn, qp->qp_num);

	if (attr_mask & IBV_QP_TIMEOUT)
		DEVX_SET(qpc, qpc, primary_address_path.ack_timeout, qp_attr->timeout);
	if (attr_mask & IBV_QP_RETRY_CNT)
		DEVX_SET(qpc, qpc, retry_count, qp_attr->retry_cnt);
	if (attr_mask & IBV_QP_SQ_PSN)
		DEVX_SET(qpc, qpc, next_send_psn, qp_attr->sq_psn & 0xffffff);
	if (attr_mask & IBV_QP_RNR_RETRY)
		DEVX_SET(qpc, qpc, rnr_retry, qp_attr->rnr_retry);
	if (attr_mask & IBV_QP_MAX_QP_RD_ATOMIC)
		DEVX_SET(qpc, qpc, log_sra_max, u32log2(qp_attr->max_rd_atomic));

	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
}

static int
qp_enable_mmo(struct ibv_qp *qp)
{
	uint32_t in[DEVX_ST_SZ_DW(init2init_qp_in)] = {};
	uint32_t out[DEVX_ST_SZ_DW(init2init_qp_out)] = {};
	void *qpce = DEVX_ADDR_OF(init2init_qp_in, in, qpc_data_extension);

	DEVX_SET(init2init_qp_in, in, opcode, MLX5_CMD_OP_INIT2INIT_QP);
	DEVX_SET(init2init_qp_in, in, qpc_ext, 1);
	DEVX_SET(init2init_qp_in, in, qpn, qp->qp_num);
	DEVX_SET64(init2init_qp_in, in, opt_param_mask_95_32, MLX5_QPC_OPT_MASK_32_INIT2INIT_MMO);

	DEVX_SET(qpc_ext, qpce, mmo, 1);

	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
}

static int
modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *qp_attr, struct mlx5dv_ah *dv_ah, int attr_mask)
{
	int ret;

	/* state mask is a must for modifying QP */
	if (!(attr_mask & IBV_QP_STATE))
		return -EINVAL;

	switch (qp_attr->qp_state) {
	case IBV_QPS_INIT:
		ret = modify_qp_to_init(qp, qp_attr, attr_mask);
		break;
	case IBV_QPS_RTR:
		ret = modify_qp_to_rtr(qp, qp_attr, dv_ah, attr_mask);
		break;
	case IBV_QPS_RTS:
		ret = modify_qp_to_rts(qp, qp_attr, attr_mask);
		break;
	default:
		ret = -EINVAL;
	}

	return ret;
}

/**
 * mema_open_chnl() - open a channel between ARM and host memory based on
 * @chnl:  sample channel
 *
 * The function initializes device and creates submission and completion
 * queues.
 *
 * Return: 0 on success or -1 on failure
 */
static int
mema_open_chnl(struct doca_mema_channel *mdev, uint32_t pcie_function_id)
{
	struct ibv_cq_init_attr_ex cq_init_attr_ex = {};

	struct ibv_qp_init_attr_ex qp_init_attr_ex = {};
	struct ibv_qp_cap qp_cap = {
		.max_send_wr = 6400,
		.max_recv_wr = 6400,
		.max_send_sge = 1,
		.max_recv_sge = 1,
		.max_inline_data = 64,
	};
	struct mlx5dv_qp_init_attr dv_qp_init_attr = {};

	struct ibv_qp_attr qp_attr = {};
	int rc = 0;

	if (mdev == NULL) {
		DOCA_DLOG_ERR("mdev is NULL");
		return -1;
	}

	if (mdev->enabled) {
		DOCA_DLOG_ERR("device already started");
		return -1;
	}

	cq_init_attr_ex.cqe = CQ_MAX_SIZE * 200;
	cq_init_attr_ex.cq_context = NULL;
	cq_init_attr_ex.channel = NULL;
	cq_init_attr_ex.comp_vector = 0;

	mdev->cq_ex = ibv_create_cq_ex(mdev->ib_ctx, &cq_init_attr_ex);
	if (!mdev->cq_ex) {
		DOCA_DLOG_ERR("AH: error: can't create cq_ex. err=%d.\n", errno);
		return 1;
	}

	DOCA_DLOG_INFO("AH: CQ handle = %u.\n", mdev->cq_ex->handle);

	qp_init_attr_ex.sq_sig_all = 0;
	qp_init_attr_ex.send_cq = ibv_cq_ex_to_cq(mdev->cq_ex);
	qp_init_attr_ex.recv_cq = ibv_cq_ex_to_cq(mdev->cq_ex);
	qp_init_attr_ex.cap = qp_cap;
	qp_init_attr_ex.qp_type = IBV_QPT_RC;
	qp_init_attr_ex.comp_mask = IBV_QP_INIT_ATTR_PD | IBV_QP_INIT_ATTR_SEND_OPS_FLAGS;
	qp_init_attr_ex.pd = mdev->pd;
	qp_init_attr_ex.send_ops_flags =
		IBV_QP_EX_WITH_RDMA_WRITE | IBV_QP_EX_WITH_SEND | IBV_QP_EX_WITH_RDMA_READ;

	dv_qp_init_attr.comp_mask = MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS;
	dv_qp_init_attr.send_ops_flags = MLX5DV_QP_EX_WITH_MEMCPY;

	mdev->qp = mlx5dv_create_qp(mdev->ib_ctx, &qp_init_attr_ex, &dv_qp_init_attr);
	if (!mdev->qp) {
		DOCA_DLOG_ERR("AH: error: can't create qp. err=%d.\n", errno);
		return 1;
	}

	mdev->qp_ex = ibv_qp_to_qp_ex(mdev->qp);
	if (!mdev->qp_ex) {
		DOCA_DLOG_ERR("AH: error: failed to get qp_ex.\n");
		goto err_qp;
	}

	mdev->mqp_ex = mlx5dv_qp_ex_from_ibv_qp_ex(mdev->qp_ex);
	if (!mdev->mqp_ex) {
		DOCA_DLOG_ERR("AH: error: failed to get mqp_ex.\n");
		goto err_qp;
	}

	// DOCA_DLOG_INFO("AH: after create qp. mdev->qp_ex->wr_start=%p.\n",
	// mdev->qp_ex->wr_start);

	// modify to init

	qp_attr.qp_state = IBV_QPS_INIT;
	qp_attr.pkey_index = 0;
	qp_attr.port_num = 1;
	qp_attr.qp_access_flags =
		IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ;

	// rc = ibv_modify_qp(mdev->qp, &qp_attr, IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_PORT |
	// IBV_QP_ACCESS_FLAGS); ret = ibv_modify_qp(mdev->qp, &qp_attr, IBV_QP_STATE |
	// IBV_QP_PORT);
	rc = modify_qp(mdev->qp, &qp_attr, NULL,
		       IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_PORT | IBV_QP_ACCESS_FLAGS);
	if (rc) {
		DOCA_DLOG_ERR("AH: error: can't modify qp to init. err=%d.\n", rc);
		return 1;
	}

	rc = qp_enable_mmo(mdev->qp);
	if (rc) {
		DOCA_DLOG_ERR("AH: error: can't enable mmo. err=%d.\n", rc);
		return 1;
	}

	qp_attr.qp_state = IBV_QPS_RTR;
	qp_attr.path_mtu = IBV_MTU_1024;
	qp_attr.dest_qp_num = mdev->qp->qp_num;
	qp_attr.rq_psn = 0;
	qp_attr.max_dest_rd_atomic = 1;
	qp_attr.min_rnr_timer = 12;
	qp_attr.ah_attr.is_global = 0;
	qp_attr.ah_attr.src_path_bits = 0;
	qp_attr.ah_attr.port_num = 1;

	rc = modify_qp(mdev->qp, &qp_attr, NULL,
		       IBV_QP_STATE | IBV_QP_PATH_MTU | IBV_QP_DEST_QPN | IBV_QP_RQ_PSN |
			       IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER | IBV_QP_AV);
	if (rc) {
		DOCA_DLOG_ERR("AH: error: failed modify QP to RTR. ret=%d.\n", rc);
		return 1;
	}

	qp_attr.qp_state = IBV_QPS_RTS;
	qp_attr.timeout = 12;
	qp_attr.retry_cnt = 3;
	qp_attr.rnr_retry = 3;
	qp_attr.sq_psn = 0;
	qp_attr.max_rd_atomic = 1;
	rc = modify_qp(mdev->qp, &qp_attr, NULL,
		       IBV_QP_STATE | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT | IBV_QP_RNR_RETRY |
			       IBV_QP_SQ_PSN | IBV_QP_MAX_QP_RD_ATOMIC);
	if (rc) {
		DOCA_DLOG_ERR("AH: error: failed modify QP to RTS. rc=%d.", rc);
		return 1;
	}

	mdev->sdev->mkey = mema_create_cross_mkey(mdev->pd, mdev->sdev);
	if (!mdev->sdev->mkey) {
		DOCA_DLOG_ERR("failed to create crossing mkey");
		goto free_cq;
	}
	DOCA_DLOG_INFO("dma key is 0x%x", mdev->sdev->mkey->mkey);
	mdev->dma_rkey = mdev->sdev->mkey->mkey;

	return 0;

err_qp:
	rc = ibv_destroy_qp(mdev->qp);
	if (rc)
		DOCA_DLOG_ERR("error: can't destroy QP. ret=%d.\n", rc);
	mdev->qp = NULL;

free_cq:
	rc = ibv_destroy_cq(ibv_cq_ex_to_cq(mdev->cq_ex));
	if (rc)
		DOCA_DLOG_ERR("error: can't destroy CQ. ret=%d.\n", rc);
	mdev->cq_ex = NULL;
teardown_device:
	mdev->enabled = false;
	return -1;
}

int
mema_poll_cq(struct doca_mema_channel *mdev, size_t num_to_poll)
{
	struct ibv_wc wc;
	int num_polled;

	while (num_to_poll > 0) {
		num_polled = ibv_poll_cq(ibv_cq_ex_to_cq(mdev->cq_ex), 1, &wc);
		if (num_polled < 0) {
			DOCA_DLOG_ERR("error: failed to poll cq. num_polled=%d.", num_polled);
			return 1;
		} else if (num_polled == 1) {
			if (wc.status != IBV_WC_SUCCESS) {
				DOCA_DLOG_ERR("error: polled cq with error status. status=%d.",
					       wc.status);
				DOCA_DLOG_ERR("AH: vendor_err = 0x%x.", wc.vendor_err);
				return 1;
			}
			num_to_poll--;
		}
	}

	return 0;
}

static int
mema_chnl_rx_read(struct doca_mema_channel *mdev, uint64_t addr, uint32_t len)
{

	int rc;

	ibv_wr_start(mdev->qp_ex);
	mdev->qp_ex->wr_id = MEMCPY_WR_ID;
	mdev->qp_ex->wr_flags = IBV_SEND_SIGNALED | IBV_SEND_FENCE;

	mlx5dv_wr_memcpy(mdev->mqp_ex, mdev->mr->lkey, (uintptr_t)mdev->dma_membase, mdev->dma_rkey,
			 (uintptr_t)addr, len);

	rc = ibv_wr_complete(mdev->qp_ex);
	if (rc) {
		DOCA_DLOG_ERR("error: failed ibv_wr_complete. rc=%d.", rc);
		return rc;
	}

	return mema_poll_cq(mdev, 1);
}

static int
mema_chnl_rx_read_list(struct doca_mema_channel *mdev, struct mema_req *phyaddr_lst)
{
	if (phyaddr_lst == NULL)
		return -1;

	int rc = 0;
	int list_size = phyaddr_lst->list_size - 1;

	mdev->qp_ex->wr_id = MEMCPY_WR_ID;
	mdev->qp_ex->wr_flags = 0; // IBV_SEND_SIGNALED | IBV_SEND_FENCE;
	ibv_wr_start(mdev->qp_ex);

	for (int i = 0; i < (phyaddr_lst->list_size - 1); i++) {

		// DOCA_DLOG_INFO("opcode = %d, lkey = %d, local = %p
		// \n",phyaddr_lst->opcode,mdev->hp_mr->lkey,phyaddr_lst->local_phy_addr_list[i]);
		mlx5dv_wr_memcpy(mdev->mqp_ex,
				 phyaddr_lst->opcode != MEMA_OP_USE_HUGE_PAGES ? mdev->mr->lkey
									       : mdev->hp_mr->lkey,
				 (uintptr_t)phyaddr_lst->local_phy_addr_list[i], mdev->dma_rkey,
				 (uintptr_t)(phyaddr_lst->remote_phy_addr_list[i]),
				 phyaddr_lst->len_list[i]);
	}

	mdev->qp_ex->wr_flags = IBV_SEND_SIGNALED; // | IBV_SEND_FENCE;
	mlx5dv_wr_memcpy(mdev->mqp_ex,
			 phyaddr_lst->opcode != MEMA_OP_USE_HUGE_PAGES ? mdev->mr->lkey
								       : mdev->hp_mr->lkey,
			 (uintptr_t)phyaddr_lst->local_phy_addr_list[list_size], mdev->dma_rkey,
			 (uintptr_t)(phyaddr_lst->remote_phy_addr_list[list_size]),
			 phyaddr_lst->len_list[list_size]);

	rc = ibv_wr_complete(mdev->qp_ex);
	if (rc) {
		DOCA_DLOG_ERR("error: failed ibv_wr_complete. rc=%d.", rc);
		return rc;
	}

	return mema_poll_cq(mdev, 1);
}

char *
doca_mema_chnl_dma_read(struct doca_mema_channel *mdev, uint64_t phy_addr, uint32_t len)
{

	int err = 0;

	/* if memory region is not valid keep the dma_buf's content as zeros */
#ifndef MEMORY_REGION_BYPASS
	if (!pf_mem_is_valid(mdev->mmem_config, phy_addr, len))
		return NULL;
#endif

	err = mema_chnl_rx_read(mdev, phy_addr, sizeof(char) * len);
	if (err)
		return NULL;

	return mdev->dma_membase;
}

int
doca_mema_chnl_dma_read_pointers_list(struct doca_mema_channel *mdev, struct mema_req *phyaddr_lst)
{

	for (int i = 0; i < phyaddr_lst->list_size; i++) {
		phyaddr_lst->local_phy_addr_list[i] =
			(uintptr_t)mdev->dma_membase + i * PONTERS_LEN;
	}

	return mema_chnl_rx_read_list(mdev, phyaddr_lst);
}

int
doca_mema_chnl_dma_read_pages_list(struct doca_mema_channel *mdev, struct mema_req *phyaddr_lst)
{
	for (int i = 0; i < phyaddr_lst->list_size; i++) {
		phyaddr_lst->local_phy_addr_list[i] =
			(uintptr_t)mdev->hp_dma_membase + i * phyaddr_lst->len_list[i];
	}
	return mema_chnl_rx_read_list(mdev, phyaddr_lst);
}

/**
 * doca_mema_destroy_chnl() - close open connection
 * @chnl:  chnl to vm/host memory
 *
 * The function stops mema device. Submission and completion queues,
 * DMA QP are destroyed.
 *
 * Return: 0 on success or errno on failure
 */
int
doca_mema_destroy_channel(struct doca_mema_channel **mdev)
{
	int last_ret = 0; // return the last error encountered, but try to clean up everything
	int ret = 0;

	if (!mdev)
		return 0;

	struct doca_mema_channel *mdev_ = *mdev;

	if (!mdev_)
		return 0;

	/* Wait until all outstanding operations are completed */
	// snap_dma_q_flush(chnl->dma_q);
	destroy_pf_mem_regions(mdev_->mmem_config);

	ret = mema_destroy_cross_mkey(mdev_->sdev->mkey);
	if (ret) {
		DOCA_DLOG_ERR("Failed to destroy crossing mkey");
		last_ret = ret;
	}

	ret = ibv_destroy_qp(mdev_->qp);
	if (ret) {
		DOCA_DLOG_ERR("error: can't destroy QP. ret=%d.", ret);
		last_ret = ret;
	}

	if (mdev_->cq_ex) {
		ret = ibv_destroy_cq(ibv_cq_ex_to_cq(mdev_->cq_ex));
		if (ret) {
			DOCA_DLOG_ERR("error: can't destroy CQ. ret=%d.", ret);
			last_ret = ret;
		}
	}

	mema_chnl_destroy_requests(mdev_);

	ret = ibv_dealloc_pd(mdev_->pd);
	if (ret) {
		DOCA_DLOG_ERR("error: can't dealloc parent domain. ret=%d.", ret);
		last_ret = ret;
	}

	ret = ibv_close_device(mdev_->ib_ctx);
	if (ret) {
		DOCA_DLOG_ERR("error: can't close device. ret=%d.", ret);
		last_ret = ret;
	}

	mema_close_device(mdev_->sdev);

	mdev_->enabled = false;
	free(mdev_);
	return last_ret;
}

int
doca_mema_create_channel(struct doca_mema_channel **mdev, char *dev_name, uint16_t pcie_function_id)
{
	if (!mdev)
		return -1;
	int err = 0;

	*mdev = mema_chnl_init(dev_name, pcie_function_id);
	if (*mdev == NULL)
		return -1;

	err = mema_open_chnl(*mdev, pcie_function_id);
	if (err) {
		mema_chnl_destroy_requests(*mdev);
		*mdev = NULL;
		return -1;
	}

	return 0;
}
