/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <json.h>

#include <stdint.h>
#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "maoob_device.h"
#include "mem_pool.h"

#include <doca_log.h>

#ifndef SHM_HUGETLB
	#define SHM_HUGETLB 04000
#endif


DOCA_LOG_REGISTER(MEMA::POOL);

/* This functions builds the valid memory regions map from a file
 * returns 1 on sucess, NULL on failure.
 */
memp_t *
create_machine_mem_regions(const char *mem_region_path)
{
	uint64_t count = 0;
	FILE *fp = NULL;
	char *file_buffer = NULL;		    // reading the json file to this buffer
	char counter_buff[10] = {0};		    // an array to conver an int counter to a string
	struct json_object *parsed_json_file;	    // an object holding the whole parsed file
	struct json_object *parsed_allowed_regions; // an object holding all the parsed allowed
						    // region in the file
	struct json_object *parsed_allowed_region;  // an object holding one parsed allowed region
	struct json_object *parsed_region_length;   // an object holding a region's length
	struct json_object *parsed_region_start;    // an object holding a region's start address
	int i = 0, err = 0, size = 0;
	memp_t *ret = NULL, *pf_mem = (memp_t *)calloc(1, sizeof(memp_t));

	if (pf_mem == NULL)
		return ret;

	fp = fopen(mem_region_path, "r");
	if (fp == NULL) {
		DOCA_DLOG_ERR(
			"%s: Error while opening the file, make sure that mem_regions.json resides in current working directory fp = %p.\n",
			__func__, fp);
		goto error_free_pf_mem;
	}

	err = fseek(fp, 0, SEEK_END);
	if (err) {
		DOCA_DLOG_ERR("%s: error in fseek\n", __func__);
		goto error_close_file;
	}
	size = ftell(fp); // size of file in bytes
	if (size < 0) {
		DOCA_DLOG_ERR("%s: error in receiving file size\n", __func__);
		goto error_close_file;
	}
	err = fseek(fp, 0, SEEK_SET); // reset the file pointer
	if (err) {
		DOCA_DLOG_ERR("%s: error in fseek\n", __func__);
		goto error_close_file;
	}
	file_buffer = (char *)calloc((size + 100), sizeof(char)); // +100 to be on the safe side
	if (file_buffer == NULL) {
		DOCA_DLOG_ERR("%s: Error while allocating the read buffer\n", __func__);
		goto error_close_file;
	}

	fread(file_buffer, sizeof(char), size, fp);

	parsed_json_file = json_tokener_parse(file_buffer);
	if (parsed_json_file == NULL)
		goto error_free_file_buffer;

	json_object_object_get_ex(parsed_json_file, "allowed_regions", &parsed_allowed_regions);

	if (json_object_get_type(parsed_allowed_regions) != json_type_object) {
		DOCA_DLOG_ERR("%s: Error while reading the file's contents.", __func__);
		goto error_free_json;
	}

	pf_mem->mrs_count = json_object_object_length(parsed_allowed_regions);

	if (pf_mem->mrs_count <= 0) {
		DOCA_DLOG_ERR("%s: Error while reading the file's contents. %d", __func__,
			       pf_mem->mrs_count);
		goto error_free_json;
	}
	pf_mem->mrs = (mr_t *)calloc(pf_mem->mrs_count, sizeof(mr_t));

	if (pf_mem->mrs == NULL) {
		DOCA_DLOG_ERR("build_mem_regions: calloc error\n");
		goto error_free_json;
	}

	for (i = 0; i < pf_mem->mrs_count; i++) {
		sprintf(counter_buff, "%d", i);
		json_object_object_get_ex(parsed_allowed_regions, counter_buff,
					  &parsed_allowed_region);
		json_object_object_get_ex(parsed_allowed_region, "length", &parsed_region_length);
		json_object_object_get_ex(parsed_allowed_region, "start", &parsed_region_start);
		pf_mem->mrs[i].start_address = json_object_get_uint64(parsed_region_start);
		pf_mem->mrs[i].length = json_object_get_uint64(parsed_region_length);
		DOCA_DLOG_INFO("region %d, start: %llu, length: %llu\n", i,
			       pf_mem->mrs[i].start_address, pf_mem->mrs[i].length);
		memset(counter_buff, 0, 10);
	}
	ret = pf_mem;

error_free_json:
	json_object_put(parsed_json_file);
error_free_file_buffer:
	free(file_buffer);
error_close_file:
	fclose(fp);
error_free_pf_mem:
	if (ret == NULL)
		free(pf_mem);
	return ret;
}

int
destroy_pf_mem_regions(memp_t *mem_p)
{
	if (!mem_p)
		return 0;

	free(mem_p->mrs);
	free(mem_p);
	return 1;
}

/*
 * This functions gets a memory address and a length (together - an address region)
 * checks if this region can be accessed
 */
int
pf_mem_is_valid(memp_t *mem_p, uint64_t phy_addr, uint32_t len)
{

	int i = 0;

	/* we need to build the memory regions map*/
	if ((mem_p == NULL) || (mem_p->mrs_count == 0)) {
		if (mem_p != NULL)
			DOCA_DLOG_ERR("%s: error building valid memory regions %d", __func__,
				       mem_p->mrs_count);
		else
			DOCA_DLOG_ERR("%s: error building valid memory regions", __func__);

		return 0;
	}

	for (i = 0; i < mem_p->mrs_count; i++) {
		if (phy_addr >= mem_p->mrs[i].start_address &&
		    (phy_addr + len) <= (mem_p->mrs[i].start_address + mem_p->mrs[i].length)) {
			return 1;
		}
	}
	return 0;
}

int
virt2phy(const void *virt_addr, char **ret_phy_adrs)
{
	if (virt_addr == NULL) {
		DOCA_DLOG_ERR("NULL virtaddr");
		return -1;
	}
	if (ret_phy_adrs == NULL) {
		DOCA_DLOG_ERR("NULL ret_phy_adrs");
		return -1;
	}

	int err = 0;
	int page_size = getpagesize();
	uint64_t page, physaddr;
	unsigned long virt_pfn;
	off_t offset = 0;
	off_t offset_seek = 0;
	ssize_t read_size = 0;

	int fd = open("/proc/self/pagemap", O_RDONLY);

	if (fd < 0) {
		DOCA_DLOG_ERR("Open /proc/self/pagemap failed");
		return -1;
	}
	check_rc(fd < 0, "Open /proc/self/pagemap failed");

	virt_pfn = (unsigned long)virt_addr / page_size; // The number of pages before me
	offset = sizeof(uint64_t) * virt_pfn; // Each page address is 8B * pages before me

	offset_seek = lseek(fd, offset, SEEK_SET);
	check_rc(offset_seek == (off_t)-1, "Seek error in /proc/self/pagemap");

	*(int *)virt_addr = 0; // Magic line
	read_size = read(fd, &page, sizeof(uint64_t));
	check_rc(read_size < 0, "Cannot read /proc/self/pagemap");

	physaddr =
		((page & 0x7fffffffffffffULL) * page_size) + ((unsigned long)virt_addr % page_size);
	*ret_phy_adrs = (char *)((page & 0x7fffffffffffffULL) * page_size) +
			((unsigned long)virt_addr % page_size);

	close(fd);

	return 0;

bail:
	if (fd) {
		close(fd);
	};

	return 1;
}

int
alloc_hp_buf(int *hp_num, char **ret_buf, enum huge_page hp_size)
{
	int err = 1;
	uint64_t len = hp_size;
	int shmid;
	char *addr;

	check_rc(ret_buf == NULL, "NULL ret buffer");
	DOCA_DLOG_INFO("\tAllocating HP\n");

	shmid = shmget(IPC_PRIVATE, len, SHM_HUGETLB | IPC_CREAT | IPC_EXCL | SHM_R | SHM_W);
	*hp_num = shmid;
	check_rc(shmid < 0, "shmget failed");

	DOCA_DLOG_INFO("\thmid: 0x%x\n", shmid);

	addr = shmat(shmid, NULL, 0);
	/* no matter what was the status, mark memory to be deleted at the end */
	shmctl(shmid, IPC_RMID, NULL);
	check_rc(addr == (char *)-1, "shmat failed");
	DOCA_DLOG_INFO("\tHuge page virtual address(%p), size(0x%lx)\n", addr, len);
	*ret_buf = addr;

	return 0;
bail:
	return 1;
}
