/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <infiniband/mlx5dv.h>
#include <infiniband/verbs.h>
#include <json-c/json.h>
#include <stdint.h>
#include <stdio.h>

#include "mema.h"
#include "mlx5_ifc.h"
#include <doca_log.h>

DOCA_LOG_REGISTER(MEMA::CORE);

/**
 * mema_open() - Opens and create a new apsh context
 * @ibdev:       RDMA device
 *
 * Return: Returns apsh_context in case of success, NULL otherwise and errno
 * will be set to indicate the failure reason.
 */

struct ibv_context *
mema_open(struct ibv_device *ibdev)
{
	struct mlx5dv_context_attr attrs = {};
	struct ibv_context *context;
	int rc;

	if (!mlx5dv_is_supported(ibdev)) {
		errno = ENOTSUP;
		goto out_err;
	}

	attrs.flags = MLX5DV_CONTEXT_FLAGS_DEVX;
	context = mlx5dv_open_device(ibdev, &attrs);
	if (!context) {
		errno = EAGAIN;
		goto out_err;
	}
	return context;

out_err:
	return NULL;
}

/**
 * doca_mema_open_device() - Create a new apsh memory acqusition device from apsh context
 * @sctx:       apsh context
 * @attr:       apsh memory acqusition device attributes
 *
 * Allocates a new apsh memory acqusition device that will be associated to the given apsh
 * context according to the requested attributes.
 *
 * Return: Returns a apsh memory acqusition device in case of success, NULL otherwise and
 * errno will be set to indicate the failure reason. The caller is responsible for
 * freeing the mema device.
 */
struct mema_device *
mema_open_device(struct ibv_context *ib_ctx, struct mema_device_attr *attr)
{
	struct mema_device *sdev;
	struct ibv_device_attr_ex attr_ex;
	struct mlx5dv_port info;
	int ret;

	sdev = calloc(1, sizeof(*sdev));
	if (!sdev) {
		errno = ENOMEM;
		goto out_err;
	}

	ret = ibv_query_device_ex(ib_ctx, NULL, &attr_ex);
	if (ret)
		goto out_free_sdev;

	sdev->phy_ports_cnt_ex = attr_ex.phys_port_cnt_ex;
	for (int i = 1; i < sdev->phy_ports_cnt_ex; i++) {
		ret = mlx5dv_query_port(ib_ctx, i, &info);
		if (ret)
			goto out_free_sdev;

		if (info.flags & (MLX5DV_QUERY_PORT_VPORT | MLX5DV_QUERY_PORT_VPORT_VHCA_ID) &&
		    attr->pcie_function_id == info.vport_vhca_id) {
			DOCA_DLOG_INFO("i: %d, vport:%d!!", i, info.vport);
			sdev->vport_vhca_id = info.vport_vhca_id;
			sdev->enabled = true;
			break;
		}
	}

	if (sdev->enabled == false) {
		DOCA_DLOG_ERR("VF function doesn't exist");
		errno = ENODEV;
		goto out_free_sdev;
	}

	ret = mema_get_introspection_mkey(ib_ctx, sdev->vport_vhca_id, info.vport, &sdev->introspection_mkey);
	if (ret) {
		DOCA_DLOG_ERR("mema_get_introspection_mkey: failed, ret=%d strerr=%s", ret, strerror(-ret));
		errno = ENODEV;
		goto out_free_sdev;
	}

	return sdev;

out_free_sdev:
	free(sdev);
out_err:
	return NULL;
}

/**
 * mema_close_device() - Destroy a mema device
 * @sdev:       mema device
 *
 * Destroy and free a mema device.
 */
void
mema_close_device(struct mema_device *sdev)
{
	free(sdev);
}

static bool
mema_introspection_mkey_access_allowed(struct ibv_context *context, uint16_t function_id, uint16_t vport)
{
	uint8_t in[DEVX_ST_SZ_BYTES(query_hca_cap_in)] = {};
	uint8_t out[DEVX_ST_SZ_BYTES(query_hca_cap_out)] = {};
	uint32_t allowed_obj_types_mask;
	int ret;

	DEVX_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	DEVX_SET(query_hca_cap_in, in, op_mod, MLX5_SET_HCA_CAP_OP_MOD_GENERAL_DEVICE2);

	DEVX_SET(query_hca_cap_in, in, other_function, MLX5_SET_HCA_CAP_OTHER_FUNCTION);

	DEVX_SET(query_hca_cap_in, in, function_id, vport);

	ret = mlx5dv_devx_general_cmd(context, in, sizeof(in), out, sizeof(out));
	if (ret)
		return false;

	allowed_obj_types_mask = DEVX_GET(
		query_hca_cap_out, out, capability.cmd_hca_cap2.introspection_mkey_access_allowed);

	if ((!DEVX_GET(query_hca_cap_out, out, capability.cmd_hca_cap.dma_mmo_qp)) &&
	    (!DEVX_GET(query_hca_cap_out, out, capability.cmd_hca_cap.dma_mmo_sq))) {
		DOCA_DLOG_INFO("DMA MMO not available");
		// NOTE: this can fail but MMO will still operate
	}

	if (allowed_obj_types_mask != 1)
		return false;

	return true;
}

// returns negative errno
int
mema_get_introspection_mkey(struct ibv_context *context, uint16_t function_id, uint16_t vport, uint32_t *mkey)
{
	int ret = 0;
	uint8_t in[DEVX_ST_SZ_BYTES(query_hca_cap_in)] = {};
	uint8_t out[DEVX_ST_SZ_BYTES(query_hca_cap_out)] = {};

	if (!mema_introspection_mkey_access_allowed(context, function_id, vport)) {
		DOCA_DLOG_ERR("mema_introspection_mkey_access_allowed is not allowed");
		return -ENOTSUP;
	}

	DEVX_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	DEVX_SET(query_hca_cap_in, in, op_mod, MLX5_SET_HCA_CAP_OP_MOD_GENERAL_DEVICE2);

	DEVX_SET(query_hca_cap_in, in, other_function, MLX5_SET_HCA_CAP_OTHER_FUNCTION);

	DEVX_SET(query_hca_cap_in, in, function_id, vport);

	ret = mlx5dv_devx_general_cmd(context, in, sizeof(in), out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to allow other vhca access, ret=%d", ret);
		return -ENOTSUP;
	}

	*mkey = DEVX_GET(query_hca_cap_out, out, capability.cmd_hca_cap2.introspection_mkey);

	DOCA_DLOG_INFO("introspection mkey 0x%x", *mkey);
	// DOCA_DLOG_INFO("Other VHCA access is allowed for object 0x%x\n", introspection_mkey);
	return 0;
}

static int
get_pd_id(struct ibv_pd *pd, uint32_t *pd_id)
{
	int ret = 0;
	struct mlx5dv_pd pd_info;
	struct mlx5dv_obj obj;

	if (!pd)
		return -EINVAL;
	obj.pd.in = pd;
	obj.pd.out = &pd_info;
	ret = mlx5dv_init_obj(&obj, MLX5DV_OBJ_PD);
	if (ret)
		return ret;
	*pd_id = pd_info.pdn;
	return 0;
}

uint16_t
mema_get_vhca_id(struct mema_device *sdev)
{
	return sdev->vport_vhca_id;
}

struct mema_cross_mkey *
mema_create_cross_mkey(struct ibv_pd *pd, struct mema_device *target_sdev)
{
	uint8_t in[DEVX_ST_SZ_BYTES(create_mkey_in)] = {0};
	uint8_t out[DEVX_ST_SZ_BYTES(create_mkey_out)] = {0};
	void *mkc = DEVX_ADDR_OF(create_mkey_in, in, memory_key_mkey_entry);
	struct ibv_context *ctx = pd->context;
	struct mema_cross_mkey *cmkey;
	uint32_t pd_id;
	int ret;

	cmkey = calloc(1, sizeof(*cmkey));
	if (!cmkey) {
		DOCA_DLOG_ERR("failed to alloc cross_mkey for pd: 0x%x\n, err: %m\n", pd->handle);
		return NULL;
	}

	ret = get_pd_id(pd, &pd_id);
	if (ret) {
		DOCA_DLOG_ERR("get_pd_id failed, ret=%d", ret);
		goto out_err;
	}

	DEVX_SET(create_mkey_in, in, opcode, MLX5_CMD_OP_CREATE_MKEY);
	DEVX_SET(mkc, mkc, access_mode_1_0, MLX5_MKC_ACCESS_MODE_CROSSING_VHCA_MKEY & 0x3);
	DEVX_SET(mkc, mkc, access_mode_4_2, (MLX5_MKC_ACCESS_MODE_CROSSING_VHCA_MKEY >> 2) & 0x7);
	DEVX_SET(mkc, mkc, a, 1);
	DEVX_SET(mkc, mkc, rw, 1);
	DEVX_SET(mkc, mkc, rr, 1);
	DEVX_SET(mkc, mkc, lw, 1);
	DEVX_SET(mkc, mkc, lr, 1);
	DEVX_SET(mkc, mkc, pd, pd_id);
	DEVX_SET(mkc, mkc, qpn, 0xffffff);
	DEVX_SET(mkc, mkc, length64, 1);
	DEVX_SET(mkc, mkc, mkey_7_0, 0x42);
	DEVX_SET(mkc, mkc, crossing_target_vhca_id, mema_get_vhca_id(target_sdev));
	DEVX_SET(mkc, mkc, translations_octword_size_crossing_target_mkey,
		 target_sdev->introspection_mkey);

	cmkey->devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!cmkey->devx_obj) {
		DOCA_DLOG_ERR("AH: error in mlx5dv_devx_obj_create. err = %d.\n", errno);
		goto out_err;
	}

	cmkey->mkey = DEVX_GET(create_mkey_out, out, mkey_index) << 8 | 0x42;

	return cmkey;

out_err:
	free(cmkey);
	return NULL;
}

int
mema_destroy_cross_mkey(struct mema_cross_mkey *mkey)
{
	int ret = 0;

	if (mkey == NULL)
		return 0;

	if (mkey->devx_obj)
		ret = mlx5dv_devx_obj_destroy(mkey->devx_obj);

	free(mkey);

	return ret;
}
