/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "doca_log.h"
#include "openssl/sha.h"
#include <dirent.h>
#include <errno.h>
#include <getopt.h>
#include <gmodule.h>
#include <inttypes.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zip.h>
#include "private/common_defines.h"

DOCA_LOG_REGISTER(APSH::LINUX::HASHBUILD);

#define PAGE_SIZE 0x1000
#define MAX_DIR_LEN 2048
#define ERROR(x) \
	{ \
		DOCA_DLOG_ERR("error in %d\n", __LINE__); \
		return x; \
	}
#define ERROR_GO(x) \
	{ \
		DOCA_DLOG_ERR("error in %d\n", __LINE__); \
		goto x; \
	}

static int
string_sort_function(gconstpointer a, gconstpointer b)
{
	const char *str_a = (const char *)a;
	const char *str_b = (const char *)b;

	return strcmp(str_a, str_b);
}

static int
int_sort_function(gconstpointer a, gconstpointer b)
{
	int int_a = GPOINTER_TO_INT(a);
	int int_b = GPOINTER_TO_INT(b);

	return int_a - int_b;
}

static int
is_file_relevant(const char *file_name)
{
	return 1;
	/* In linux the files can have no extension, if there are issues uncomment those lines
	 * const char *extensions[] = {".axf", ".bin", ".o", ".elf", ".prx", ".puff", ".ko", ".mod", ".so"};
	 * const int len = ARRAY_SIZE(extensions);
	 * const int file_len = strlen(file_name);
	 * int i;
	 *
	 * for (i = 0; i < len; ++i)
	 * if (!strncmp(extensions[i], file_name + file_len - strlen(extensions[i]), strlen(extensions[i])))
	 *     return 1;
	 * return 0;
	 */
}

static int
show_dir_content(const char *path, GHashTable **path_table_p)
{
	char d_path[MAX_DIR_LEN], init_path[MAX_DIR_LEN];
	struct dirent *dir;

	DIR *d = opendir(path);

	if (d == NULL)
		ERROR(-1);
	sprintf(init_path, path[strlen(path) - 1] != '/' ? "%s/" : "%s", path);

	if (*path_table_p == NULL)
		*path_table_p =
			g_hash_table_new_full(g_str_hash, g_str_equal, free, (GDestroyNotify)g_hash_table_destroy);
	GHashTable *path_table = *path_table_p;

	if (path_table == NULL)
		ERROR_GO(close_file);

	while ((dir = readdir(d)) != NULL) {
		sprintf(d_path, "%s%s", init_path, dir->d_name);
		if (dir->d_type == DT_REG) {
			if (!is_file_relevant(dir->d_name))
				continue;
			GHashTable *name_table = g_hash_table_lookup(path_table, dir->d_name);

			if (name_table == NULL) {
				name_table = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
				if (name_table == NULL)
					ERROR_GO(free_table);
				char *d_name_dup = strdup(dir->d_name);

				if (d_name_dup == NULL)
					ERROR_GO(free_table);
				g_hash_table_insert(path_table, d_name_dup, name_table);
			}
			if (g_hash_table_lookup(name_table, d_path) == NULL) {
				char *d_path_dup = strdup(d_path);

				if (d_path_dup == NULL)
					ERROR_GO(free_table);
				g_hash_table_insert(name_table, d_path_dup, "full");
			}
		} else if (dir->d_type == DT_DIR && strcmp(dir->d_name, ".") != 0 && strcmp(dir->d_name, "..") != 0) {
			int ret = show_dir_content(d_path, path_table_p);

			if (ret) {
				closedir(d);
				ERROR(ret);
			}
		}
	}
	closedir(d);
	return 0;
free_table:
	g_hash_table_destroy(path_table);
close_file:
	closedir(d);
	ERROR(-1);
}

struct elf_segment_t {
	uint64_t paddr;
	uint64_t vaddr;
	uint64_t psize;
	uint64_t vsize;
	uint32_t type;
	uint32_t perm;
};

struct elf_section_t {
	uint8_t *name;
	uint64_t vaddr;
	uint64_t psize;
	uint64_t paddr;
	uint32_t type;
};

struct elf_t {
	uint64_t type;
	uint64_t is32;
	uint64_t num_section_headers;
	uint64_t section_header_size;
	uint64_t sections_header_table_address;
	uint64_t sections_header_table_size;
	uint64_t num_segment_headers;
	uint64_t segment_header_size;
	uint64_t segment_header_table_address;
	uint64_t segments_header_table_size;
	uint64_t names_section_index;
	uint64_t file_header_size;
	struct elf_section_t *sections;
	struct elf_segment_t *segments;
};

struct file_elf_t {
	char *name;
	uint8_t *bytes;
	GHashTable *alterations;
	GHashTable *pages;
	uint8_t *perms;
	uint8_t *virtual;
	int virtual_size;
	int perms_size;
	struct elf_t elf;
};

struct file_elf_t
create_file_elf()
{
	return (struct file_elf_t){
		NULL, /*path*/
		NULL, /*name*/
		NULL, /*bytes*/
		NULL, /*alterations*/
		NULL, /*pages*/
		NULL, /*perms*/
		NULL, /*virtual*/
		0,    /*virtual_size*/
		0,    /*perms_size*/
		0     /*pe_binary*/
	};
}

void
free_file_elf(struct file_elf_t *file)
{
	if (file->elf.sections != NULL)
		free(file->elf.sections);
	if (file->elf.segments != NULL)
		free(file->elf.segments);
	if (file->perms != NULL)
		free(file->perms);
	if (file->bytes != NULL)
		free(file->bytes);
	if (file->virtual != NULL)
		free(file->virtual);
	if (file->alterations != NULL)
		g_hash_table_destroy(file->alterations);
	if (file->pages != NULL)
		g_hash_table_destroy(file->pages);
}

int
elf_parse_segments(struct file_elf_t *file)
{
	size_t i;

	file->elf.segments =
		(struct elf_segment_t *)calloc(file->elf.num_segment_headers, sizeof(struct elf_segment_t));
	if (file->elf.segments == NULL)
		return -1;
	int is32 = file->elf.is32;

	for (i = 0; i < file->elf.num_segment_headers; i++) {
		uint64_t addr =
			file->bytes + file->elf.segment_header_table_address + i * file->elf.segment_header_size;
		struct elf_segment_t *segment = file->elf.segments + i;

		segment->paddr = is32 ? *(uint32_t *)(addr + 0x04) : *(uint64_t *)(addr + 0x08);
		segment->vaddr = is32 ? *(uint32_t *)(addr + 0x08) : *(uint64_t *)(addr + 0x10);
		segment->psize = is32 ? *(uint32_t *)(addr + 0x10) : *(uint64_t *)(addr + 0x20);
		segment->vsize = is32 ? *(uint32_t *)(addr + 0x14) : *(uint64_t *)(addr + 0x28);
		segment->type = *(uint32_t *)(addr);
		uint32_t flag = *(uint32_t *)(addr + (is32 ? 0x18 : 0x4));

		segment->perm = 0; // Readable
		if (flag & 0x1)
			segment->perm = 2; // Executable
		if (flag & 0x2)
			segment->perm = 1; // Writable
	}
	return 0;
}

int
elf_parse_sections(struct file_elf_t *file)
{
	size_t i;
	uint64_t names_section_addr =
		file->elf.sections_header_table_address + file->elf.names_section_index * file->elf.section_header_size;

	file->elf.sections =
		(struct elf_section_t *)calloc(file->elf.num_section_headers, sizeof(struct elf_section_t));
	if (file->elf.sections == NULL)
		return -1;

	int is32 = file->elf.is32;
	uint64_t names = is32 ? *(uint32_t *)(file->bytes + names_section_addr + 0x10)
			      : *(uint64_t *)(file->bytes + names_section_addr + 0x18);

	for (i = 0; i < file->elf.num_section_headers; i++) {
		uint64_t addr =
			file->bytes + file->elf.sections_header_table_address + i * file->elf.section_header_size;
		struct elf_section_t *section = file->elf.sections + i;

		section->paddr = is32 ? *(uint32_t *)(addr + 0x10) : *(uint64_t *)(addr + 0x18);
		section->vaddr = is32 ? *(uint32_t *)(addr + 0x0C) : *(uint64_t *)(addr + 0x10);
		section->psize = is32 ? *(uint32_t *)(addr + 0x14) : *(uint64_t *)(addr + 0x20);
		section->type = *(uint32_t *)(addr + 0x4);
		section->name = (char *)(file->bytes + names + *(uint32_t *)(addr));
	}
	return 0;
}

int
elf_parse(struct file_elf_t *file)
{
	char hex[4] = {0x7F, 0x45, 0x4C, 0x46};

	if (strncmp(file->bytes, hex, 4))
		return -1;
	file->elf.type = *(uint32_t *)(file->bytes + 0x4);
	int is32 = file->elf.type == 1;

	file->elf.is32 = is32;
	file->elf.num_section_headers = *(uint16_t *)(file->bytes + (is32 ? 0x30 : 0x3C));
	file->elf.section_header_size = *(uint16_t *)(file->bytes + (is32 ? 0x2E : 0x3A));
	file->elf.sections_header_table_address = *(uint64_t *)(file->bytes + (is32 ? 0x20 : 0x28));
	file->elf.sections_header_table_size =
		file->elf.sections_header_table_address + file->elf.section_header_size * file->elf.num_section_headers;
	file->elf.num_segment_headers = *(uint16_t *)(file->bytes + (is32 ? 0x2C : 0x38));
	file->elf.segment_header_size = *(uint16_t *)(file->bytes + (is32 ? 0x2A : 0x36));
	file->elf.segment_header_table_address =
		is32 ? *(uint32_t *)(file->bytes + 0x1C) : *(uint64_t *)(file->bytes + 0x20);
	file->elf.segments_header_table_size =
		file->elf.segment_header_table_address + file->elf.segment_header_size * file->elf.num_segment_headers;
	file->elf.names_section_index = *(uint16_t *)(file->bytes + (is32 ? 0x32 : 0x3E));
	file->elf.file_header_size = is32 ? 52 : 64;
	return elf_parse_segments(file) || elf_parse_sections(file);
}

static int
expand_section(const struct file_elf_t *file, const struct elf_segment_t *segment, uint8_t **expanded, int *e_size)
{
	uint32_t size = segment->vsize;

	if ((size % PAGE_SIZE) != 0)
		size += (PAGE_SIZE - (size % PAGE_SIZE));

	*expanded = (uint8_t *)calloc(size, sizeof(uint8_t));
	if (*expanded == NULL)
		ERROR(-1);
	memcpy(*expanded, file->bytes + segment->paddr - (segment->paddr % 0x1000), size);
	*e_size = size;
	return 0;
}

static int
expand_header(struct file_elf_t *file)
{
	file->virtual_size = 0;
	file->virtual = (uint8_t *)calloc(file->virtual_size, sizeof(uint8_t));
	if (file->virtual == NULL)
		ERROR(-1);
	return 0;
}

static int
append(struct file_elf_t *file, uint8_t *expanded, int e_size, int vaddr)
{
	uint8_t *virtual = (uint8_t *)calloc(vaddr + e_size, sizeof(uint8_t));

	if (virtual == NULL)
		ERROR(-1);
	int size = vaddr < file->virtual_size ? vaddr : file->virtual_size;

	memcpy(virtual, file->virtual, size);
	free(file->virtual);
	file->virtual = virtual;
	file->virtual_size = vaddr + e_size;
	memcpy(file->virtual + vaddr, expanded, e_size);
	return 0;
}

static int
expand(struct file_elf_t *file)
{
	int ret = -1;

	if (expand_header(file))
		ERROR(ret);

	int perms_s = file->virtual_size / PAGE_SIZE;
	uint8_t *prev_perms, *perms = (uint8_t *)calloc(perms_s + 1, sizeof(uint8_t));

	if (perms == NULL)
		ERROR(ret);

	uint8_t prev_perm = 0;
	uint8_t *expanded = NULL;
	size_t i;

	for (i = 0; i < file->elf.num_segment_headers; ++i) {
		struct elf_segment_t *segment = file->elf.segments + i;
		/* type Loadable segment and writable, or read only after relocation */
		if (!((segment->type == 0x1 && segment->perm != 1) || segment->type == 0x6474e552))
			continue;
		int vaddr = segment->vaddr;
		int size = (vaddr - file->virtual_size) / PAGE_SIZE;
		int e_size;

		perms = realloc(prev_perms = perms, (perms_s + size) * sizeof(uint8_t) + 1);
		if (perms == NULL)
			ERROR_GO(error);
		if (memset(perms + perms_s, prev_perm, size) == NULL)
			ERROR_GO(error);
		perms_s += size;
		if (expand_section(file, segment, &expanded, &e_size))
			ERROR_GO(error);
		if (e_size / PAGE_SIZE == 0) {
			free(expanded);
			expanded = NULL;
			continue;
		}
		if (append(file, expanded, e_size, vaddr - vaddr % 0x1000))
			ERROR_GO(error);
		free(expanded);
		expanded = NULL;
		e_size = e_size / PAGE_SIZE;
		perms = realloc(prev_perms = perms, (perms_s + e_size) * sizeof(uint8_t) + 1);
		if (perms == NULL)
			ERROR_GO(error);
		if (memset(perms + perms_s, segment->perm, e_size) == NULL)
			ERROR_GO(error);
		perms_s += e_size;
		prev_perm = segment->perm;
	}
	file->perms = perms;
	file->perms_size = perms_s;
	return 0;
error:
	if (perms == NULL)
		free(prev_perms);
	else
		free(perms);
	if (expanded)
		free(expanded);
	return ret;
}

static uint8_t *
file_bytes(uint8_t *path)
{
	FILE *fileptr = fopen(path, "rb");

	if (fileptr == NULL)
		ERROR(NULL);
	fseek(fileptr, 0, SEEK_END);
	long filelen = ftell(fileptr);

	rewind(fileptr);
	uint8_t *bytes = (uint8_t *)calloc(filelen, sizeof(uint8_t));

	if (bytes == NULL) {
		fclose(fileptr);
		ERROR(NULL);
	}
	fread(bytes, filelen, 1, fileptr);
	fclose(fileptr);
	return bytes;
}

static void
insert_zero(GHashTable *zeroes, uint32_t addr)
{
	g_hash_table_insert(zeroes, GINT_TO_POINTER(addr), "full");
}

static GHashTable *
look_zeroes(GHashTable *alterations, uint32_t vaddr)
{
	GHashTable *zeroes = g_hash_table_lookup(alterations, GINT_TO_POINTER(vaddr));

	if (zeroes != NULL)
		return zeroes;
	zeroes = g_hash_table_new_full(NULL, NULL, NULL, NULL);
	if (zeroes == NULL)
		ERROR(NULL);
	g_hash_table_insert(alterations, GINT_TO_POINTER(vaddr), zeroes);
	return zeroes;
}

static int
reloc_sections(struct file_elf_t *file, GHashTable *alterations)
{
	size_t i;
	int is32 = file->elf.is32;

	for (i = 0; i < file->elf.num_section_headers; i++) {
		struct elf_section_t *section = file->elf.sections + i;
		// Relocation entries with addends or Relocation entries no addends
		if (!(section->type == 0x4 || section->type == 0x9))
			continue;
		uint64_t offset;
		uint64_t address_size = is32 ? 4 : 8;
		uint64_t entry_size = is32 ? 8 : 16;

		if (section->type != 0x9)
			entry_size += entry_size / 2;
		for (offset = 0; offset < section->psize; offset += entry_size) {
			uint64_t reloc_addr = file->bytes + section->paddr + offset;

			reloc_addr = is32 ? *(uint32_t *)(reloc_addr) : *(uint64_t *)(reloc_addr);
			uint64_t reloc_addr_base = reloc_addr - (reloc_addr % 0x1000);
			GHashTable *zeroes = look_zeroes(alterations, reloc_addr_base);

			if (zeroes == NULL)
				ERROR(-1);
			insert_zero(zeroes, reloc_addr - reloc_addr_base);
			if (!is32)
				insert_zero(zeroes, reloc_addr - reloc_addr_base + 4);
		}
	}
	return 0;
}

static int
dynamic_sections(struct file_elf_t *file, GHashTable *alterations)
{
	size_t i;
	int is32 = file->elf.is32;

	for (i = 0; i < file->elf.num_section_headers; i++) {
		struct elf_section_t *section = file->elf.sections + i;
		// Dynamic linking information
		if (!(section->type == 0x6 || !strcmp(section->name, ".data.rel.ro")))
			continue;
		uint64_t offset;

		for (offset = 0; offset < section->psize; offset += 4) {
			uint64_t reloc_addr = section->vaddr + offset;
			uint64_t reloc_addr_base = reloc_addr - (reloc_addr % 0x1000);
			GHashTable *zeroes = look_zeroes(alterations, reloc_addr_base);

			if (zeroes == NULL)
				ERROR(-1);
			insert_zero(zeroes, reloc_addr - reloc_addr_base);
		}
	}
	return 0;
}

static uint8_t *
look_page(GHashTable *pages, uint32_t vaddr)
{
	uint8_t *page = g_hash_table_lookup(pages, GINT_TO_POINTER(vaddr));

	if (page != NULL)
		return page;
	page = (uint8_t *)calloc(PAGE_SIZE, sizeof(uint8_t));
	if (page == NULL)
		ERROR(NULL);
	g_hash_table_insert(pages, GINT_TO_POINTER(vaddr), page);
	return page;
}

static int
parse_pages(struct file_elf_t *file)
{
	file->pages = g_hash_table_new_full(NULL, NULL, NULL, free);
	if (file->pages == NULL)
		ERROR(-1);
	uint32_t vaddr = 0;
	uint32_t unapplied = 0;

	while (vaddr < file->virtual_size) {
		GHashTable *zeroes = g_hash_table_lookup(file->alterations, GINT_TO_POINTER(vaddr));
		uint32_t page_s = 0;
		uint8_t *page = look_page(file->pages, vaddr);

		if (page == NULL)
			ERROR(-1);
		if (zeroes == NULL) {
			memcpy(page, file->virtual + vaddr, PAGE_SIZE);
			vaddr += PAGE_SIZE;
			continue;
		}
		uint32_t offset = 0;

		if (unapplied > 0) {
			offset = unapplied;
			page_s += unapplied;
			insert_zero(zeroes, -(4 - unapplied));
			unapplied = 0;
		}
		GList *zeroes_head = g_hash_table_get_keys(zeroes);
		GList *zeroes_p = g_list_sort(zeroes_head, int_sort_function);
		GList *zeroes_p_const = zeroes_p;

		while (zeroes_p) {
			int32_t zero = GPOINTER_TO_UINT(zeroes_p->data);

			if (zero < 0 || zero < offset) {
				zeroes_p = zeroes_p->next;
				continue;
			}
			memcpy(page + page_s, file->virtual + vaddr + offset, zero - offset);
			page_s += zero - offset;
			if (zero <= PAGE_SIZE - 4) {
				offset = zero + 4;
				page_s += 4;
			} else {
				unapplied = 4 - (PAGE_SIZE - zero);
				offset = PAGE_SIZE;
				page_s += PAGE_SIZE - zero;
			}
			zeroes_p = zeroes_p->next;
		}
		g_list_free(zeroes_p_const);
		if (offset < PAGE_SIZE)
			memcpy(page + page_s, file->virtual + vaddr + offset, PAGE_SIZE - offset);
		vaddr += PAGE_SIZE;
	}
	return 0;
}

static int
parse_alterations(struct file_elf_t *file)
{
	file->alterations = g_hash_table_new_full(NULL, NULL, NULL, (GDestroyNotify)g_hash_table_destroy);
	if (file->alterations == NULL)
		ERROR(-1);
	if (reloc_sections(file, file->alterations))
		ERROR(-1);
	if (dynamic_sections(file, file->alterations))
		ERROR(-1);
	return 0;
}

static void
hashbuild_sha256h_string(uint8_t *string, unsigned int length, unsigned char *hash)
{
	struct timespec ts1, ts2;
	SHA256_CTX sha256;
	unsigned char hash_b[SHA256_DIGEST_LENGTH];

	SHA256_Init(&sha256);
	SHA256_Update(&sha256, string, length);
	SHA256_Final(hash_b, &sha256);
	int i;

	for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
		sprintf(hash + (i * 2), "%02x", hash_b[i]);
}

static int
output(struct file_elf_t *file, GHashTable *outputs)
{
	uint32_t offset = 0;
	uint32_t max_offset_digits = floor(log10(file->virtual_size / PAGE_SIZE)) + 1;

	while (offset < file->virtual_size) {
		uint8_t *page = g_hash_table_lookup(file->pages, GINT_TO_POINTER(offset));

		if (page == NULL)
			ERROR(-1);
		unsigned char hash[SHA256_DIGEST_LENGTH * 2 + 1] = {0};

		hashbuild_sha256h_string(page, PAGE_SIZE, hash);
		GHashTable *zeroes = g_hash_table_lookup(file->alterations, GINT_TO_POINTER(offset));
		GList *zeroes_head = zeroes == NULL ? NULL : g_hash_table_get_keys(zeroes);
		GList *zeroes_p = g_list_sort(zeroes_head, int_sort_function);
		GList *zeroes_p_const = zeroes_p;
		uint32_t max_zero_digits = 0;

		if (zeroes_p)
			max_zero_digits = floor(log10(GPOINTER_TO_UINT(g_list_last(zeroes_p)->data))) + 1;
		uint32_t zero_size = g_list_length(zeroes_p) * (max_zero_digits + 1) + 1;
		char *zero_s = (char *)calloc(zero_size, sizeof(char));

		if (zero_s == NULL)
			ERROR(-1);
		uint32_t zero_len = 0;

		while (zeroes_p) {
			uint32_t zero = GPOINTER_TO_UINT(zeroes_p->data);
			int ret = sprintf(zero_s + zero_len, zero_len ? " %d" : "%d", zero);

			if (ret < 0) {
				free(zero_s);
				g_list_free(zeroes_p_const);
				ERROR(ret);
			}
			zero_len += ret;
			if (zero_size <= zero_len) {
				free(zero_s);
				g_list_free(zeroes_p_const);
				ERROR(-1);
			}
			zeroes_p = zeroes_p->next;
		}
		g_list_free(zeroes_p_const);
		uint32_t output_len = strlen(file->name) + 1 + max_offset_digits + 1 + (SHA256_DIGEST_LENGTH * 2 + 1) +
				      1 + 1 + 1 + zero_len + 1 + 1;
		char *output = (char *)calloc(output_len, sizeof(char));

		if (output == NULL) {
			free(zero_s);
			ERROR(-1);
		}
		int ret = sprintf(output, "%s,%d,%s,%d,%s\n", file->name, offset / PAGE_SIZE, hash,
				  file->perms[offset / PAGE_SIZE], zero_s);
		free(zero_s);
		if (ret < 0) {
			free(output);
			ERROR(ret);
		}

		g_hash_table_insert(outputs, output, "full");
		offset += PAGE_SIZE;
	}
	return 0;
}

static int
process(char *content, char *name, GHashTable *outputs)
{
	int ret = -1;
	struct file_elf_t file = create_file_elf();

	file.name = name;
	file.bytes = content;
	if (file.bytes == NULL)
		ERROR_GO(error);
	if (elf_parse(&file))
		ERROR_GO(error);
	if (expand(&file))
		ERROR_GO(error);
	if (parse_alterations(&file))
		ERROR_GO(error);
	if (parse_pages(&file))
		ERROR_GO(error);
	if (output(&file, outputs))
		ERROR_GO(error);
	ret = 0;
error:
	free_file_elf(&file);
	return ret;
}

static int
output_sort_function(gconstpointer a, gconstpointer b)
{
	const char *str_a = (const char *)a;
	const char *str_b = (const char *)b;
	const char *p_a = strchr(str_a, ',');
	const char *p_b = strchr(str_b, ',');
	int len = p_a - str_a > p_b - str_b ? p_a - str_a : p_b - str_b;
	int ret = strncmp(str_a, str_b, len);

	if (ret)
		return ret;
	return atoi(p_a + 1) > atoi(p_b + 1);
}

static int
write_outputs(GHashTable *outputs, char *irregular_files, char **blob)
{
	int ret = -1;
	*blob = (char *)calloc(1, sizeof(char));
	if (*blob == NULL)
		return -1;
	char *prev_blob = *blob;
	uint32_t blob_s = 0;

	GList *outputs_head = g_hash_table_get_keys(outputs);
	GList *outputs_p = g_list_sort(outputs_head, output_sort_function);
	GList *outputs_p_const = outputs_p;

	while (outputs_p) {
		char *output = outputs_p->data;
		uint32_t len = strlen(output);
		*blob = realloc(prev_blob = (*blob), (blob_s + len) * sizeof(char) + 1);
		if ((*blob) == NULL)
			ERROR_GO(error);
		memcpy((*blob) + blob_s, output, len);
		blob_s += len;
		outputs_p = outputs_p->next;
	}
	if (irregular_files != NULL) {
		uint32_t irregular_len = strlen(irregular_files);
		*blob = realloc(prev_blob = (*blob), (blob_s + irregular_len) * sizeof(char) + 1);
		if ((*blob) == NULL)
			ERROR_GO(error);
		memcpy((*blob) + blob_s, irregular_files, irregular_len);
		blob_s += irregular_len;
	}
	*((*blob) + blob_s) = '\0';
	g_list_free(outputs_p_const);
	return 0;
error:
	g_list_free(outputs_p_const);
	free(prev_blob);
	return -1;
}

int
apsh_linux_zip_to_json(char *path_to_zip, char **output)
{
	int ret = -1;
	int err = 0;
	struct zip *zip_file = zip_open(path_to_zip, 0, &err);
	int64_t i, entries = zip_get_num_entries(zip_file, 0);

	if (zip_file == NULL || entries < 0)
		ERROR(ret);
	GHashTable *outputs = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
	char *irregular_files_string = NULL;

	if (outputs == NULL)
		ERROR_GO(close_zip);
	for (i = 0; i < entries; i++) {
		struct zip_stat st;

		zip_stat_init(&st);
		if (zip_stat_index(zip_file, i, 0, &st))
			ERROR_GO(destroy_outputs);
		if (st.size == 0)
			continue;
		if (!is_file_relevant(st.name))
			continue;
		struct zip_file *file = zip_fopen_index(zip_file, i, 0);

		if (file == NULL)
			goto destroy_outputs;
		uint8_t *content = (uint8_t *)calloc(st.size, sizeof(uint8_t));

		if (content == NULL)
			ERROR_GO(destroy_outputs);
		zip_fread(file, content, st.size);
		zip_fclose(file);
		if (strcmp(st.name, "doca_apsh_irregular_files.txt") != 0) {
			ret = process(content, st.name, outputs);
			if (ret)
				ERROR_GO(destroy_outputs);
		} else {
			irregular_files_string = content;
		}
	}
	ret = write_outputs(outputs, irregular_files_string, output);
destroy_outputs:
	if (irregular_files_string != NULL)
		free(irregular_files_string);
	g_hash_table_destroy(outputs);
close_zip:
	zip_close(zip_file);
	return ret;
}
