/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "hashtest.h"
#include "../../doca_apsh_attr.h"
#include "doca_log.h"
#include "jsonout.h"
#include "openssl/sha.h"
#include <glib.h>
#include <stdbool.h>
#include "../../core/APSH/plugin_context.h"
#include "../../../core/APSH/apsh_instance.h"

#define PROTECTION_FLAGS 3
#define HASHTEST_LINUX_RETRY_INIT_TIMES 3
#define CONTAINS_WRITE_EXEC 1
#define ADDED_WRITE 2
#define ADDED_EXECUTE 4

DOCA_LOG_REGISTER(APSH::LINUX::HASHTEST);

#define PAGE_SIZE 0x1000

static const struct protection_t {
	char name;
	uint32_t mask;
} linux_protect_att[PROTECTION_FLAGS] = {
	{'r', 0x00000001},
	{'w', 0x00000002},
	{'x', 0x00000004},
};

struct hashtest_context {

	struct _task_struct {
		addr_t tasks;
		addr_t comm;
		addr_t pid;
		addr_t mm;
	} task_struct;

	struct _mm_struct {
		addr_t mmap;	    /* list of memory areas */
		addr_t start_brk;   /* start address of heap */
		addr_t brk;	    /* end address of heap */
		addr_t start_stack; /* start address of stack */
		addr_t start_code;  /* start address of code */
		addr_t end_code;    /* end address of code */
		addr_t start_data;  /* start address of data */
		addr_t end_data;    /* end address of data */
		addr_t vdso;	    /* vdso address */
	} mm_struct;

	struct _vm_area_struct {
		addr_t vm_pgoff;
		addr_t vm_start;
		addr_t vm_end;
		addr_t vm_file;
		addr_t vm_page_prot;
		addr_t vm_flags;
		addr_t vm_next;
	} vm_area_struct;

	struct _file_struct {
		addr_t f_inode;
		addr_t f_owner;
		addr_t f_path;
	} file_struct;

	struct _path_struct {
		addr_t dentry;
	} path_struct;

	struct _dentry_struct {
		addr_t d_name;
	} dentry_struct;

	struct _qstr_struct {
		addr_t name;
	} qstr_struct;

	struct PsList {
		addr_t head;
		addr_t current_entry;
		addr_t next_entry;
	} ps_list;
};


struct zero_entry {
	/* offset from the begginning of the page where zero should be applied */
	unsigned int entry_offset;
	struct zero_entry *next;
};

struct page_entry {
	int index;
	unsigned char hash[SHA256_DIGEST_LENGTH];
	int permission;
	struct page_entry *next;
	struct zero_entry *zeroes;
};

struct file_entry {
	unsigned char file_name[MAX_FILENAME];
	bool is_irregular_file;
	struct page_entry *pages;
	struct file_entry *next;
};

struct page_permission {
	GHashTable *permissions_hash;
};

struct area_result {
	int pages_present;
	int matching_hashes;
};

struct analyze_memory_ctx {
	char prev_path[MAX_FILENAME+1];
	uint64_t start_address;
};

static bool
is_file_irregular(char *file_name, struct file_entry *file_hashes) {
	struct file_entry *files_ptr = file_hashes;

	while (files_ptr != NULL) {
		if (strcmp(files_ptr->file_name, file_name) == 0)
			return files_ptr->is_irregular_file;
		files_ptr = files_ptr->next;
	}

	return false;
}

static status_t
check_permissions(GHashTable **permissions_hash, addr_t vma_start, addr_t vma_end, addr_t vma_flags, uint16_t *perm_res)
{
	if (permissions_hash == NULL || vma_start >= vma_end)
		return APSH_F;

	if (*permissions_hash == NULL) {
		*permissions_hash = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, NULL);
		if (*permissions_hash == NULL)
			return APSH_F;
	}
	*perm_res = 0;

	if (((vma_flags & linux_protect_att[1].mask) != 0) && ((vma_flags & linux_protect_att[2].mask) != 0))
		(*perm_res) |= CONTAINS_WRITE_EXEC;

	for (addr_t curr_page = vma_start; curr_page < vma_end; curr_page += PAGE_SIZE) {
		addr_t prev_flags = g_hash_table_lookup(*permissions_hash, curr_page);

		if (prev_flags == 0) {
			g_hash_table_insert(*permissions_hash, curr_page, vma_flags);
		} else {
			if (((vma_flags & linux_protect_att[1].mask) != 0) &&
			    ((prev_flags & linux_protect_att[1].mask) == 0))
				(*perm_res) |= ADDED_WRITE;

			if (((vma_flags & linux_protect_att[2].mask) != 0) &&
			    ((prev_flags & linux_protect_att[2].mask) == 0))
				(*perm_res) |= ADDED_EXECUTE;
		}
	}

	return APSH_S;
}

static bool
are_equal(const char *s1, const char *s2, int length)
{
	for (int i = 0; i < length; i++) {
		if (s1[i] != s2[i])
			return false; // They are different
	}
	return true; // They must be the same
}

static void
sha256h_string(char *string, unsigned int length, unsigned char *hash)
{
	SHA256_CTX sha256;

	SHA256_Init(&sha256);
	SHA256_Update(&sha256, string, length);
	SHA256_Final(hash, &sha256);
}

static void
free_zero_nodes(struct page_entry *page_node)
{
	struct zero_entry *zero_node = page_node->zeroes;
	struct zero_entry *zero_node_next;

	while (zero_node != NULL) {
		zero_node_next = zero_node->next;
		free(zero_node);
		zero_node = zero_node_next;
	}
}

static void
free_page_nodes(struct file_entry *file_node)
{
	struct page_entry *page_node = file_node->pages;
	struct page_entry *page_node_next;

	while (page_node != NULL) {
		page_node_next = page_node->next;
		free_zero_nodes(page_node);
		free(page_node);
		page_node = page_node_next;
	}
}

static void
free_file_nodes(struct file_entry *file_head)
{
	struct file_entry *file_node = file_head;
	struct file_entry *file_node_next;

	while (file_node != NULL) {
		file_node_next = file_node->next;
		free_page_nodes(file_node);
		free(file_node);
		file_node = file_node_next;
	}
}

static int
print_pages(apsh_instance_t apsh, addr_t start_addr, addr_t end_addr, apsh_pid_t pid)
{

	int committed = 0;
	unsigned long length = end_addr - start_addr;

	if (length < 0xFFFFFFFF && length > 0) {
		while (start_addr < end_addr) {
			char buffer[PAGE_SIZE] = {0};

			if (apsh_read_va(apsh, start_addr, pid, PAGE_SIZE, &buffer, NULL) == APSH_S) {
				++committed;

				char log_buffer[0x2001] = {0};

				DOCA_DLOG_INFO("\n");
				for (int i = 0; i < PAGE_SIZE; i++)
					sprintf(log_buffer + 2*i, "%02x", buffer[i]);
				DOCA_DLOG_INFO("%s", log_buffer);
			}
			start_addr += PAGE_SIZE;
		}
	}
	return committed;
}

static status_t
parse_line(char *line, size_t line_len, struct file_entry **file_hashes_head)
{

	char *zero_line, *ptr;
	char *file_name;
	struct file_entry *file_node, *new_file_node;
	struct page_entry *page_node, *new_page_node;
	struct zero_entry *zero_node, *new_zero_node;
	unsigned char input_hash[SHA256_DIGEST_LENGTH * 2 + 1] = {0};

	if (line_len <= 0 || line == NULL) {
		DOCA_DLOG_ERR("%s: error while prsing the line", __func__);
		return APSH_F;
	}
	if (line[0] == '#') { //irregular file
		new_file_node = calloc(1, sizeof(struct file_entry));
		if (!new_file_node) {
			DOCA_DLOG_ERR("%s: allocation error", __func__);
			goto error_exit;
		}
		new_file_node->next = NULL;
		new_file_node->pages = NULL;
		new_file_node->is_irregular_file = true;
		strncpy(new_file_node->file_name, line + 1, MAX_FILENAME);

		file_node = *file_hashes_head;
		while (file_node->next != NULL)
			file_node = file_node->next;

		if (*file_hashes_head == NULL)
			*file_hashes_head = new_file_node;
		else
			file_node->next = new_file_node;
		return APSH_S;
	}

	file_name = strtok(line, ",");
	file_node = *file_hashes_head;
	bool found = false;

	while (file_node) {
		if (strcmp(file_node->file_name, file_name) == 0) {
			found = true;
			break;
		}
		if (!file_node->next)
			break;
		file_node = file_node->next;
	}

	if (!found) {

		new_file_node = malloc(sizeof(struct file_entry));
		if (!new_file_node) {
			DOCA_DLOG_ERR("%s: allocation error", __func__);
			goto error_exit;
		}
		new_file_node->next = NULL;
		new_file_node->pages = NULL;
		new_file_node->is_irregular_file = false;
		memset(new_file_node->file_name, 0, MAX_FILENAME);
		strncpy(new_file_node->file_name, file_name, MAX_FILENAME);
		if (*file_hashes_head == NULL)
			file_node = *file_hashes_head = new_file_node;
		else {
			file_node->next = new_file_node;
			file_node = file_node->next;
		}
	}

	page_node = file_node->pages;
	while (page_node && page_node->next)
		page_node = page_node->next;

	new_page_node = malloc(sizeof(struct page_entry));
	if (!new_page_node) {
		DOCA_DLOG_ERR("%s: allocation error", __func__);
		goto error_exit;
	}
	new_page_node->zeroes = NULL;
	new_page_node->next = NULL;
	new_page_node->index = atoi(strtok(NULL, ","));
	memset(input_hash, 0, SHA256_DIGEST_LENGTH * 2);
	strncpy(input_hash, strtok(NULL, ","), SHA256_DIGEST_LENGTH * 2);

	for (int i = 0; i < SHA256_DIGEST_LENGTH; i++) {
		char base[3] = {0};

		base[0] = input_hash[2 * i];
		base[1] = input_hash[2 * i + 1];
		new_page_node->hash[i] = strtoul(base, NULL, 16);
	}

	new_page_node->permission = atoi(strtok(NULL, ","));
	if (!file_node->pages)
		file_node->pages = new_page_node;
	else
		page_node->next = new_page_node;

	zero_line = strtok(NULL, ",");
	if ((zero_line != NULL) && (strlen(zero_line) > 2)) {
		zero_line = strtok(zero_line, " ");
		while (zero_line != NULL) {
			new_zero_node = malloc(sizeof(struct zero_entry));
			if (!new_zero_node) {
				DOCA_DLOG_ERR("%s: allocation error", __func__);
				goto error_exit;
			}
			new_zero_node->next = NULL;
			new_zero_node->entry_offset = atoi(zero_line);
			if (!new_page_node->zeroes) {
				zero_node = new_page_node->zeroes = new_zero_node;
			} else {
				zero_node->next = new_zero_node;
				zero_node = zero_node->next;
			}
			zero_line = strtok(NULL, " ");
		}
	}
	return APSH_S;

error_exit:
	return APSH_F;
}

static status_t
build_indexes_zip(const char *hash_file_path, struct file_entry **file_hashes)
{
	char *output;
	int ret = apsh_linux_zip_to_json(hash_file_path, &output);

	if (ret) {
		DOCA_DLOG_ERR("parse zip error");
		return APSH_F;
	}
	char *cur_line = output;

	while (cur_line) {
		char *next_line = strchr(cur_line, '\n');
		if (next_line) *next_line = '\0';
		else break;
		char* line = strdup(cur_line);

		if (line == NULL) {
			DOCA_DLOG_ERR("strdup failed");
			free(output);
			return APSH_F;
		}
		if (parse_line(line, next_line - cur_line, file_hashes) != APSH_S) {
			DOCA_DLOG_ERR("parse_line error");
			free(line);
			free(output);
			return APSH_F;
		}
		free(line);
		if (next_line) *next_line = '\n';
		cur_line = next_line + 1;
	}
	free(output);
   	return APSH_S;
}

static status_t
build_indexes_json(const char *hash_file_path, struct file_entry **file_hashes)
{
	FILE *fp = NULL;
	char *buffer = NULL;
	size_t bufsize = 1000, line_len = 0;
	status_t ret = APSH_F;

	if (!hash_file_path) {
		DOCA_DLOG_ERR("hashes file path is null");
		goto error_exit;
	}

	buffer = (char *)malloc(bufsize * sizeof(char));
	if (buffer == NULL) {
		DOCA_DLOG_ERR("Unable to allocate buffer");
		goto error_exit;
	}

	/* Just a standard way to open files. */
	fp = fopen(hash_file_path, "r");
	/* If file does not exist, exit. */
	if (!fp) {
		DOCA_DLOG_ERR("error opening hashes file path");
		goto error_exit;
	}

	/* Read file line by line. */
	while ((line_len = getline(&buffer, &bufsize, fp)) != -1) {
		if (parse_line(buffer, line_len, file_hashes) != APSH_S) {
			DOCA_DLOG_ERR("parse_line error");
			goto error_exit;
		}
	}

	fclose(fp);
	free(buffer);
	return APSH_S;

error_exit:
	if (fp != NULL)
		fclose(fp);
	free(buffer);
	return APSH_F;
}

static status_t
build_indexes(char *hash_file_path, struct file_entry **file_hashes){
	const char *zip_ending = ".zip";
	char *file_ending = hash_file_path+strlen(hash_file_path)-strlen(zip_ending);
	if(!strncmp(zip_ending, file_ending, strlen(zip_ending)))
		return build_indexes_zip(hash_file_path, file_hashes);
	else
		return build_indexes_json(hash_file_path, file_hashes);
}

static void
print_page_node(struct page_entry *page_node)
{

	if (!page_node)
		return;
	char log_buffer[2048] = {0};
	int buffer_offset = 0;

	buffer_offset += sprintf(log_buffer + buffer_offset, "    index: %d, permission: %d, hash: ", page_node->index, page_node->permission);
	for (int i = 0; i < SHA256_DIGEST_LENGTH; i++)
		buffer_offset += sprintf(log_buffer + buffer_offset, "%x", page_node->hash[i]);

	if (page_node->zeroes) {
		buffer_offset += sprintf(log_buffer + buffer_offset, ", zeroes: ");
		for (struct zero_entry *zero_node = page_node->zeroes; zero_node != NULL;
		    zero_node = zero_node->next) {
			buffer_offset += sprintf(log_buffer + buffer_offset, "%u ", zero_node->entry_offset);
		}
	}
	DOCA_DLOG_INFO("%s", log_buffer);
}

static void
print_file_node(struct file_entry *file_node)
{

	if (!file_node)
		return;

	DOCA_DLOG_DBG("file name: %s", file_node->file_name);
	for (struct page_entry *page_node = file_node->pages; page_node != NULL;
	     page_node = page_node->next) {
		print_page_node(page_node);
	}
	return;
}

/* Checks if a node exists with the same name
 * if not, it might be a library injection
 */
static bool
file_nodes_exist(const char *file_name, struct file_entry *file_hashes)
{

	if (!file_name)
		return false;


	struct file_entry *file_node;

	file_node = file_hashes;

	while (file_node) {
		if (strcmp(file_node->file_name, file_name) == 0)
			return true;
		file_node = file_node->next;
	}
	return false;
}

static struct page_entry *
find_page_node(const char *file_name, int page_index, struct file_entry *file_hashes)
{

	struct file_entry *file_node;
	struct page_entry *page_node;

	file_node = file_hashes;

	while (file_node) {
		if (strcmp(file_node->file_name, file_name) == 0) {
			page_node = file_node->pages;
			while (page_node) {
				if (page_node->index == page_index)
					return page_node;
				page_node = page_node->next;
			}
			return NULL;
		}
		file_node = file_node->next;
	}
	return NULL;
}

static void
zero_page(char *page_content, int page_size, struct zero_entry *zero_indexes)
{

	unsigned int offset = 0;

	for (struct zero_entry *zero_node = zero_indexes; zero_node != NULL; zero_node = zero_node->next) {

		offset = zero_node->entry_offset;
		if (offset <= page_size - 4) { /* zero 4 bytes from offset inclusive*/
			page_content[offset] = 0;
			page_content[offset + 1] = 0;
			page_content[offset + 2] = 0;
			page_content[offset + 3] = 0;
		}
	}
}

static void
print_hash256(const char *hash)
{
	if (!hash)
		return;

	char log_buffer[1024] = {0};
	int buffer_offset = 0;

	buffer_offset += sprintf(log_buffer + buffer_offset, "SHA256:");
	for (int i = 0; i < SHA256_DIGEST_LENGTH; i++)
		buffer_offset += sprintf(log_buffer + buffer_offset, "%x", hash[i]);
	DOCA_DLOG_INFO("%s", log_buffer);
	return;
}

static status_t
analyze_area_pages(apsh_instance_t apsh, addr_t start_addr, addr_t end_addr, apsh_pid_t pid,
		   const char *path, struct file_entry *file_hashes, struct area_result *result, struct analyze_memory_ctx *ctx)
{
	char page_content[PAGE_SIZE];
	char page_content_copy[PAGE_SIZE];
	unsigned char page_hash[SHA256_DIGEST_LENGTH];
	static struct page_entry *page_node;

	if (path == NULL || result == NULL) {
		DOCA_DLOG_ERR("null pointers received");
		return APSH_F;
	}

	if (strcmp(ctx->prev_path, path) != 0) {
		ctx->start_address = start_addr;
		strncpy(ctx->prev_path, path, MAX_FILENAME);
	}

	result->pages_present = 0;
	result->matching_hashes = 0;

	for (addr_t page_addr = start_addr; page_addr < end_addr; page_addr += PAGE_SIZE) {
		if (apsh_read_va(apsh, page_addr, pid, PAGE_SIZE, &page_content, NULL) == APSH_S) {
			++result->pages_present;
			page_node = find_page_node(path, (page_addr - ctx->start_address) / PAGE_SIZE, file_hashes);
			if (page_node == NULL)
				continue;
			if (page_node->zeroes != NULL)
				memcpy(page_content_copy, page_content, PAGE_SIZE);

			zero_page(page_content, PAGE_SIZE, page_node->zeroes);
			sha256h_string(page_content, PAGE_SIZE, page_hash);
			if (are_equal(page_hash, page_node->hash, SHA256_DIGEST_LENGTH)) {
				++result->matching_hashes;
				if (page_node->zeroes != NULL) {
					free_zero_nodes(page_node);
					page_node->zeroes = NULL;
					sha256h_string(page_content_copy, PAGE_SIZE, page_node->hash);
				}
			}
		}
	}

	return APSH_S;
}

static int
area_pages_present(apsh_instance_t apsh, addr_t start_addr, addr_t end_addr, apsh_pid_t pid)
{
	addr_t paddr = 0;
	int pages_present = 0;

	for (addr_t page_addr = start_addr; page_addr < end_addr; page_addr += PAGE_SIZE) {
		if (apsh_translate_uv2p(apsh, page_addr, pid, &paddr) == APSH_S)
			pages_present++;
	}
	return pages_present;
}

static void
free_hashtest_hashtable(GHashTable *hash_table)
{
	char *process_name;
	char *path;
	char *protection;

	process_name = g_hash_table_lookup(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_COMM));
	path = g_hash_table_lookup(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PATH_OF_MEMORY_AREA));
	protection = g_hash_table_lookup(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PROTECTION));

	if (process_name != NULL)
		g_free(process_name);

	if (path != NULL)
		g_free(path);

	if (protection != NULL)
		g_free(protection);

	g_hash_table_destroy(hash_table);
}

void
linux_free_hashtest_list(GSList *results_list)
{
	if (results_list != NULL) {
		g_slist_foreach(results_list, free_hashtest_hashtable, NULL);
		g_slist_free(results_list);
	}
}

void
linux_free_hashtest_results(linux_hashtest_t *results)
{
	if (results != NULL) {
		if (results->hash_file != NULL)
			free(results->hash_file);
		if (results->file_hashes != NULL)
			free_file_nodes(results->file_hashes);
		if (results->page_permissions != NULL) {
			if (results->page_permissions->permissions_hash != NULL)
				g_hash_table_destroy(results->page_permissions->permissions_hash);
			free(results->page_permissions);
		}
		linux_free_hashtest_list(results->results_list);
		free(results);
	}
}

static void
print_hash_table_list(GSList *hash_table_list)
{
	uint32_t pid;
	char *process_name;
	char *path;
	char *protection;
	addr_t start_address;
	addr_t end_address;
	int pages_number;
	int pages_present;
	int matching_hashes;
	bool hash_file_present;
	GHashTable *hash_table;
	int i = 0;

	if (!hash_table_list)
		return;

	while (1) {
		hash_table = g_slist_nth_data(hash_table_list, i);
		if (hash_table == NULL)
			break;

		pid = g_hash_table_lookup(hash_table,
					  GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PID));
		process_name = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_COMM));
		path = g_hash_table_lookup(
			hash_table,
			GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PATH_OF_MEMORY_AREA));
		protection = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PROTECTION));
		start_address = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_START_ADDRESS));
		end_address = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_END_ADDRESS));
		pages_number = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_NUMBER));
		pages_present = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_PRESENT));
		matching_hashes = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_MATCHING_HASHES));
		hash_file_present = g_hash_table_lookup(
			hash_table,
			GINT_TO_POINTER(DOCA_APSH_ATTESTATION_HASH_DATA_IS_PRESENT));

		DOCA_DLOG_INFO("[Pid : %d] [Process Name : %s] [Protection : %s] [Start: %lx] [End: %lx] [Path: %s] [Pages Number: %d] [present Pages: %d] [Hash Matches: %d] [Hash Present: %d]\n",
		       pid, process_name, protection, start_address, end_address, path,
		       (end_address - start_address) / PAGE_SIZE, pages_present, matching_hashes,
		       hash_file_present);
		i++;
	}
	return;
}

static status_t
insert_result(GSList **hash_table_list, uint32_t pid, char *process_name, char *path,
	      char *protection, addr_t start_address, addr_t end_address, int pages_number,
	      int pages_present, int matching_hashes, bool hash_file_present)
{

	gboolean result;

	GHashTable *hash_table = g_hash_table_new(g_direct_hash, g_direct_equal);

	if (hash_table == NULL)
		return APSH_F;

	char *proc_name_copy = g_strndup(process_name, strlen(process_name) + 1);
	char *path_copy = g_strndup(path, strlen(path) + 1);
	char *protection_copy = g_strndup(protection, strlen(protection) + 1);

	result = g_hash_table_insert(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PID),
				     GINT_TO_POINTER(pid));
	result &= g_hash_table_insert(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_COMM),
				      proc_name_copy);
	result &= g_hash_table_insert(
		hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PATH_OF_MEMORY_AREA),
		path_copy);
	result &= g_hash_table_insert(hash_table,
				      GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PROTECTION),
				      protection_copy);
	result &= g_hash_table_insert(hash_table,
				      GINT_TO_POINTER(DOCA_APSH_ATTESTATION_START_ADDRESS),
				      GUINT_TO_POINTER(start_address));
	result &= g_hash_table_insert(hash_table,
				      GINT_TO_POINTER(DOCA_APSH_ATTESTATION_END_ADDRESS),
				      GUINT_TO_POINTER(end_address));
	result &= g_hash_table_insert(hash_table,
				      GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_NUMBER),
				      GINT_TO_POINTER(pages_number));
	result &= g_hash_table_insert(hash_table,
				      GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_PRESENT),
				      GINT_TO_POINTER(pages_present));
	result &= g_hash_table_insert(hash_table,
				      GINT_TO_POINTER(DOCA_APSH_ATTESTATION_MATCHING_HASHES),
				      GINT_TO_POINTER(matching_hashes));
	result &= g_hash_table_insert(
		hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_HASH_DATA_IS_PRESENT),
		GINT_TO_POINTER(hash_file_present));

	if (result) {
		*hash_table_list = g_slist_append(*hash_table_list, hash_table);
		return APSH_S;
	}

	if (proc_name_copy != NULL)
		g_free(proc_name_copy);

	if (path_copy != NULL)
		g_free(path_copy);

	if (protection_copy != NULL)
		g_free(protection_copy);

	g_hash_table_destroy(hash_table);
	return APSH_F;
}

static status_t
analyze_memory_map(apsh_instance_t apsh, uint32_t pid, char *process_name, addr_t memory_map, struct file_entry *file_hashes,
		GHashTable **permissions_hash, struct hashtest_context *hashtest_ctx, GSList **hash_table_list)
{

	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_HASHTEST_LIMIT)));

	addr_t vm_area_struct_head_addr = 0;
	addr_t vm_area_struct_addr = 0;
	addr_t vm_file_addr = 0;
	addr_t dentry_path = 0;
	char *path = NULL;
	bool path_was_allocated;
	struct analyze_memory_ctx ctx= {0};

	if (apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.mmap, pid, &vm_area_struct_addr) == APSH_S) {
		vm_area_struct_head_addr = vm_area_struct_addr;

		addr_t brk;
		addr_t start_brk;
		addr_t start_stack;

		addr_t start_code;
		addr_t end_code;
		addr_t start_data;
		addr_t end_data;
		addr_t vdso_addr = NULL;
		uint64_t count = 0;

		apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.brk, pid, &brk);
		apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.start_brk, pid, &start_brk);
		apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.start_stack, pid, &start_stack);

		apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.start_code, pid, &start_code);
		apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.end_code, pid, &end_code);
		apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.start_data, pid, &start_data);
		apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.end_data, pid, &end_data);

		if (hashtest_ctx->mm_struct.vdso)
			apsh_read_addr_va(apsh, memory_map + hashtest_ctx->mm_struct.vdso, pid, &vdso_addr);

		while (count++ < limit) {
			path_was_allocated = false;
			addr_t vma_start;
			addr_t vma_end;
			addr_t vma_flags;

			apsh_read_addr_va(apsh, vm_area_struct_addr + hashtest_ctx->vm_area_struct.vm_flags, pid,
					  &vma_flags);
			apsh_read_addr_va(apsh, vm_area_struct_addr + hashtest_ctx->vm_area_struct.vm_start, pid,
					  &vma_start);
			apsh_read_addr_va(apsh, vm_area_struct_addr + hashtest_ctx->vm_area_struct.vm_end, pid,
					  &vma_end);

			char protection_flags[PROTECTION_FLAGS + 1] = {0};

			for (int i = 0; i < PROTECTION_FLAGS; i++) {
				protection_flags[i] = '-';
				if ((vma_flags)&linux_protect_att[i].mask)
					protection_flags[i] = linux_protect_att[i].name;
			}

			path = NULL;
			if (apsh_read_addr_va(apsh, vm_area_struct_addr + hashtest_ctx->vm_area_struct.vm_file, pid, &vm_file_addr) == APSH_S) {
				if (apsh_read_addr_va(apsh, vm_file_addr + hashtest_ctx->file_struct.f_path + hashtest_ctx->path_struct.dentry, pid,
										&dentry_path) == APSH_S) {
					path = apsh_read_str_va(apsh, dentry_path + hashtest_ctx->dentry_struct.d_name + hashtest_ctx->qstr_struct.name + 16, pid);
					if (path != NULL)
						path_was_allocated = true;
				}
			}

			if (!path) {
				if (vma_start <= start_brk && vma_end >= brk) {
					path = "Heap";
				} else if (vma_start <= start_stack && vma_end >= start_stack) {
					path = "Stack";
				} else if (vma_start == 0xffffffffff600000 &&
					   vma_end == 0xffffffffffdfffff) {
					path = "vsyscall";
				} else if (vdso_addr && vdso_addr == vma_start) {
					path = "vdso";
				} else
					path = "Anonymous Mapping";
			}
			uint16_t perm_res;

			if (check_permissions(permissions_hash, vma_start, vma_end, vma_flags, &perm_res) != APSH_S) {
				if (hash_table_list && *hash_table_list)
					linux_free_hashtest_list(*hash_table_list);

				if (path_was_allocated)
					free(path);
				return APSH_F;
			}
			/* need to change how to report this in the future*/
			if (perm_res != 0) {
				char tmp_path[MAX_FILENAME + strlen("_permissions") + 1];
				int len = strlen(path);

				strcpy(tmp_path, path);
				strcpy(tmp_path + len, "_permissions");

				insert_result(hash_table_list, pid, process_name, tmp_path,
						      protection_flags, vma_start, vma_end,
						      (vma_end - vma_start) / 0x1000,
						      1, 0, file_nodes_exist(path, file_hashes));
			}

			if (is_file_irregular(path, file_hashes)) {
				;
			} else if (strcmp(path, "Anonymous Mapping") == 0) { /* handle anonymous sections */
				struct area_result result = {0};

				/* iterate on all pages of section and compare read only pages to zero page */
				for (addr_t page_addr = vma_start; page_addr < vma_end; page_addr += PAGE_SIZE) {
					static unsigned char page_content[PAGE_SIZE] = {0};

					if (apsh_read_va(apsh, page_addr, pid, PAGE_SIZE, &page_content, NULL) == APSH_S) {
						++result.pages_present;
						/* hash read only sections and compare to zero */
						if ((vma_flags&linux_protect_att[2].mask) == 0 && (vma_flags&linux_protect_att[1].mask) == 0) {
							static const unsigned char zero_page_content[PAGE_SIZE] = {0};

							if (are_equal(page_content, zero_page_content, PAGE_SIZE))
								++result.matching_hashes;
						}
					}
				}
				if (result.pages_present > 0 && !!(vma_flags&linux_protect_att[2].mask)) { /* in case of executable section */
					insert_result(hash_table_list, pid, process_name, path, protection_flags, vma_start, vma_end,
										(vma_end - vma_start) / PAGE_SIZE, result.pages_present, 0, false);
				} else if (result.pages_present > 0 && !(vma_flags&linux_protect_att[1].mask)) { /* in case of read only section */
					insert_result(hash_table_list, pid, process_name, path, protection_flags, vma_start, vma_end,
										(vma_end - vma_start) / PAGE_SIZE, result.pages_present, result.matching_hashes, true);
				}
			} else if ((((vma_flags)&linux_protect_att[1].mask) == 0)) {
				struct area_result result = {0};

				if (analyze_area_pages(apsh, vma_start, vma_end, pid, path, file_hashes, &result, &ctx) != APSH_S) {
					if (hash_table_list && *hash_table_list)
						linux_free_hashtest_list(*hash_table_list);

					if (path_was_allocated)
						free(path);
					return APSH_F; // there was an error allocating structures
				}
				/* NOT supporting anonymous mapping hash testing for now, needs to be added in
				 * the future!
				 */
				if (result.pages_present > 0 && strcmp(path, "Anonymous Mapping") != 0) {
					insert_result(hash_table_list, pid, process_name, path,
						      protection_flags, vma_start, vma_end,
						      (vma_end - vma_start) / PAGE_SIZE,
						      result.pages_present, result.matching_hashes,
						      file_nodes_exist(path, file_hashes));
				}
			}
			/* Segment is writeable*/
			else {
				int pages_present = 0;
				// check if also executable and has present pages
				if ((vma_flags)&linux_protect_att[2].mask) {
					pages_present = area_pages_present(apsh, vma_start, vma_end, pid);
					if (pages_present > 0)
						insert_result(hash_table_list, pid, process_name, path, protection_flags, vma_start, vma_end,
										(vma_end - vma_start) / PAGE_SIZE, pages_present, 0, false);
				}
			}
			/* follow the next pointer */
			if (APSH_F ==
			    apsh_read_addr_va(apsh, vm_area_struct_addr + hashtest_ctx->vm_area_struct.vm_next,
					      pid, &vm_area_struct_addr)) {
				if (hash_table_list && *hash_table_list)
					linux_free_hashtest_list(*hash_table_list);

				if (path_was_allocated)
					free(path);
				return APSH_F;
			}

			if (vm_area_struct_addr == NULL) {
				//	print_hash_table_list(*hash_table_list);
				if (path_was_allocated)
					free(path);
				return APSH_S; // we are done
			}
			if (path_was_allocated)
				free(path);
		}
		return APSH_F;
	} else {
		if (hash_table_list && *hash_table_list)
			linux_free_hashtest_list(*hash_table_list);

		return APSH_F;
	}
}

static status_t
init_plugin(apsh_instance_t apsh, const char *hash_file_path, struct file_entry **file_hashes,
	    struct hashtest_context **hashtest_ctx)
{
	if ((*file_hashes == NULL) && build_indexes(hash_file_path, file_hashes) != APSH_S) {
		DOCA_DLOG_ERR("Failed in building the indexes from hash file");
		return APSH_F;
	}

	struct hashtest_context *hashtest_ctx_ptr = (struct hashtest_context *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_HASHTEST);

	if (hashtest_ctx_ptr != NULL) {
		*hashtest_ctx = hashtest_ctx_ptr;
		return APSH_S;
	}

	hashtest_ctx_ptr = malloc(sizeof(*hashtest_ctx_ptr));

	if (hashtest_ctx_ptr == NULL) {
		DOCA_DLOG_ERR("Failed to allocate plugin context memory");
		return APSH_F;
	}

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "tasks",
							 &(hashtest_ctx_ptr->task_struct.tasks)) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "pid",
							 &hashtest_ctx_ptr->task_struct.pid) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "comm",
							 &hashtest_ctx_ptr->task_struct.comm) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "mm",
							 &(hashtest_ctx_ptr->task_struct.mm)) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "mmap",
							 &hashtest_ctx_ptr->mm_struct.mmap) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_brk",
							 &hashtest_ctx_ptr->mm_struct.start_brk) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "brk",
							 &hashtest_ctx_ptr->mm_struct.brk) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_stack",
							 &hashtest_ctx_ptr->mm_struct.start_stack) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_code",
							 &hashtest_ctx_ptr->mm_struct.start_code) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "end_code",
							 &hashtest_ctx_ptr->mm_struct.end_code) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_data",
							 &hashtest_ctx_ptr->mm_struct.start_data) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "end_data",
							 &hashtest_ctx_ptr->mm_struct.end_data) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_pgoff",
							 &hashtest_ctx_ptr->vm_area_struct.vm_pgoff) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_start",
							 &hashtest_ctx_ptr->vm_area_struct.vm_start) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_end",
							 &hashtest_ctx_ptr->vm_area_struct.vm_end) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_file",
							 &hashtest_ctx_ptr->vm_area_struct.vm_file) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_page_prot",
							 &hashtest_ctx_ptr->vm_area_struct.vm_page_prot) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_flags",
							 &hashtest_ctx_ptr->vm_area_struct.vm_flags) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_next",
							 &hashtest_ctx_ptr->vm_area_struct.vm_next) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_inode",
							 &hashtest_ctx_ptr->file_struct.f_inode) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_owner",
							 &hashtest_ctx_ptr->file_struct.f_owner) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_path",
							 &hashtest_ctx_ptr->file_struct.f_path) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "path", "dentry",
							 &hashtest_ctx_ptr->path_struct.dentry) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "dentry", "d_name",
							 &hashtest_ctx_ptr->dentry_struct.d_name) == APSH_F)
		goto json_error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "qstr", "name", &hashtest_ctx_ptr->qstr_struct.name) ==
	    APSH_F)
		goto json_error_exit;

	// profile might or might not have this part
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "vdso", &hashtest_ctx_ptr->mm_struct.vdso) ==
	    APSH_F)
		goto json_error_exit;

	if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_HASHTEST, hashtest_ctx_ptr, free)) {
		free(hashtest_ctx_ptr);
		return APSH_F;
	}

	*hashtest_ctx = hashtest_ctx_ptr;
	return APSH_S;

json_error_exit:
	free(hashtest_ctx_ptr);
	DOCA_DLOG_ERR("Failed while retrieving info from system map json file");
	return APSH_F;
}

status_t
linux_hashtest_internal(apsh_instance_t apsh, linux_processes_list_t *process, uint32_t pid, const char *hash_file,
			GHashTable *permissions_hash, struct file_entry *file_hashes_head, linux_hashtest_t **results_list)
{
	status_t result;
	linux_processes_list_t *psl = NULL;
	GSList *temp_results_list = NULL;
	linux_processes_list_t *process_p;
	struct file_entry *file_hashes = file_hashes_head;
	GHashTable *prev_permissions_hash = permissions_hash;

	if (hash_file == NULL) {
		DOCA_DLOG_ERR("No hash file provided");
		goto error_exit;
	}

	struct hashtest_context *hashtest_ctx;

	if (init_plugin(apsh, hash_file, &file_hashes, &hashtest_ctx) == APSH_F) {
		DOCA_DLOG_ERR("Failed in plugin init");
		goto error_exit;
	}

	if (process == NULL) {
		/* search for process using pid */
		bool process_found = false;

		if (linux_pslist_processes_list(apsh, &psl) == APSH_F) {
			DOCA_DLOG_ERR("Failed to retrieve process list");
			goto error_exit;
		}
		linux_processes_list_t *psl_iterator = psl;

		while (psl_iterator) {
			if (pid == psl_iterator->pid) {
				process_found = true;
				process_p = psl_iterator;
				break;
			}
			psl_iterator = psl_iterator->next;
		}
		if (process_found == false) {
			DOCA_DLOG_ERR("Could not find a process with PID:%d", pid);
			goto error_exit;
		}
	} else {
		if (process->pid != pid) {
			DOCA_DLOG_ERR("process pid %d doesn't match pid provided %d", process->pid, pid);
			goto error_exit;
		}
		process_p = process;
	}

	addr_t memory_map = 0;

	result = apsh_read_addr_va(apsh, process_p->address + hashtest_ctx->task_struct.mm, 0, &memory_map);
	if (result == APSH_S && memory_map != 0) {
		result = analyze_memory_map(apsh, process_p->pid, process_p->name, memory_map, file_hashes, &permissions_hash,
		hashtest_ctx, &temp_results_list);
		if (result != APSH_S)
			goto error_exit;

		if (results_list == NULL) {
			g_hash_table_destroy(permissions_hash);
			print_hash_table_list(temp_results_list);
		} else {
			linux_hashtest_t *result_struct = malloc(sizeof(*result_struct));

			if (result_struct == NULL) {
				DOCA_DLOG_ERR("Failed to allocate memory to hashtest result");
				goto error_exit;
			}

			result_struct->page_permissions = calloc(1, sizeof(struct page_permission));

			if (result_struct->page_permissions == NULL) {
				DOCA_DLOG_ERR("Failed to allocate memory to page permissions");
				free(result_struct);
				goto error_exit;
			}

			result_struct->file_hashes = file_hashes;
			result_struct->page_permissions->permissions_hash = permissions_hash;
			result_struct->results_list = temp_results_list;
			result_struct->pid = pid;
			result_struct->hash_file = hash_file;
			*results_list = result_struct;
		}
		if (psl != NULL)
			linux_free_process_list(psl);
		return APSH_S;
	}
	DOCA_DLOG_ERR("cannot retrieve memory map for process %d", process_p->pid);
error_exit:
	if (prev_permissions_hash == NULL && permissions_hash != NULL)
		g_hash_table_destroy(permissions_hash);
	if (psl != NULL)
		linux_free_process_list(psl);
	return APSH_F;
}

status_t
linux_hashtest_main(apsh_instance_t apsh, linux_processes_list_t *process, uint32_t pid, const char *hash_file,
		GHashTable *permissions_hash, struct file_entry *file_hashes_head, linux_hashtest_t **results_list)
{
	uint32_t i;
	for (i = 0; i < HASHTEST_LINUX_RETRY_INIT_TIMES; i++){
		status_t res = linux_hashtest_internal(apsh, process, pid, hash_file, permissions_hash,
							 file_hashes_head, results_list);
		if(res == APSH_S)
			return APSH_S;
		apsh_reinit_linux(apsh);
	}
	return APSH_F;
}

status_t
linux_hashtest(apsh_instance_t apsh, linux_processes_list_t *process, uint32_t pid,
	       const char *hash_file, linux_hashtest_t **results_list)
{
	return linux_hashtest_main(apsh, process, pid, strndup(hash_file, MAX_FILENAME), NULL, NULL, results_list);
}

status_t
linux_hashtest_refresh(apsh_instance_t apsh, linux_processes_list_t *process, linux_hashtest_t **results_list)
{
	status_t ret;
	linux_hashtest_t *new_results_list;

	ret = linux_hashtest_main(apsh, process, (*results_list)->pid, (*results_list)->hash_file,
		(*results_list)->page_permissions->permissions_hash, (*results_list)->file_hashes, &new_results_list);
	if (ret != APSH_S)
		return ret;
	(*results_list)->hash_file = NULL;
	(*results_list)->file_hashes = NULL;
	free((*results_list)->page_permissions);
	(*results_list)->page_permissions = NULL;
	*results_list = new_results_list;
	return APSH_S;
}
