/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "libs.h"
#include "doca_apsh_attr.h"
#include <string.h>
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::LINUX::LIBS);


struct protection_t {
	char name[32];
	uint32_t mask;
};

static const struct protection_t protect[] = {
	{"r", 0x00000001},
	{"w", 0x00000002},
	{"x", 0x00000004}
};

struct libs_context {
	struct _task_struct {
		addr_t tasks;
		addr_t comm;
		addr_t pid;
		addr_t mm;
	} task_struct;

	struct _mm_struct {
		addr_t mmap;
		addr_t start_brk;
		addr_t brk;
		addr_t start_stack;
		addr_t vdso;
	} mm_struct;

	struct _vm_area_struct {
		addr_t vm_pgoff;
		addr_t vm_start;
		addr_t vm_end;
		addr_t vm_file;
		addr_t vm_page_prot;
		addr_t vm_next;
		addr_t vm_mm;
		addr_t vm_flags;
	} vm_area_struct;

	struct _file_struct {
		addr_t f_inode;
		addr_t f_owner;
		addr_t f_path;
	} file_struct;

	struct _path_struct {
		addr_t dentry;
	} path_struct;

	struct _dentry_struct {
		addr_t d_name;
	} dentry_struct;

	struct _qstr_struct {
		addr_t name;
	} qstr_struct;
};

static status_t
load_offsets(apsh_instance_t apsh, struct libs_context **libs_ctx)
{
	struct libs_context *libs_ctx_ptr = (struct libs_context *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_LIBS);

	if (libs_ctx_ptr != NULL) {
		*libs_ctx = libs_ctx_ptr;
		return APSH_S;
	}

	libs_ctx_ptr = malloc(sizeof(*libs_ctx_ptr));

	if (libs_ctx_ptr == NULL) {
		DOCA_DLOG_ERR("Failed to allocate plugin context memory");
		return APSH_F;
	}


	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "tasks",
								&libs_ctx_ptr->task_struct.tasks) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "pid",
								&libs_ctx_ptr->task_struct.pid) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "comm",
								&libs_ctx_ptr->task_struct.comm) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "mm",
								&libs_ctx_ptr->task_struct.mm) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "mmap",
								&libs_ctx_ptr->mm_struct.mmap) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "brk",
								&libs_ctx_ptr->mm_struct.brk) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_brk",
								&libs_ctx_ptr->mm_struct.start_brk) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_stack",
								&libs_ctx_ptr->mm_struct.start_stack) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_pgoff", &libs_ctx_ptr->vm_area_struct.vm_pgoff) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_start", &libs_ctx_ptr->vm_area_struct.vm_start) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_end",
								&libs_ctx_ptr->vm_area_struct.vm_end) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_file",
								&libs_ctx_ptr->vm_area_struct.vm_file) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_page_prot", &libs_ctx_ptr->vm_area_struct.vm_page_prot) ==
		APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_flags", &libs_ctx_ptr->vm_area_struct.vm_flags) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_next",
								&libs_ctx_ptr->vm_area_struct.vm_next) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_inode",
								&libs_ctx_ptr->file_struct.f_inode) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_owner",
								&libs_ctx_ptr->file_struct.f_owner) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_path",
								&libs_ctx_ptr->file_struct.f_path) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "path", "dentry",
								&libs_ctx_ptr->path_struct.dentry) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "dentry", "d_name",
								&libs_ctx_ptr->dentry_struct.d_name) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "qstr", "name",
								&libs_ctx_ptr->qstr_struct.name) == APSH_F)
		return APSH_F;

	// profile might or might not have this part
	apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_mm",
								&libs_ctx_ptr->vm_area_struct.vm_mm);
	apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "vdso",
								&libs_ctx_ptr->mm_struct.vdso);

	if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_LIBS, libs_ctx_ptr, free)) {
		free(libs_ctx_ptr);
		return APSH_F;
	}

	*libs_ctx = libs_ctx_ptr;

	return APSH_S;
}

static status_t
parse_libs(apsh_instance_t apsh, linux_processes_list_t *process, struct libs_context *libs_ctx, linux_libs_t **libs_head)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_LIBS_LIMIT)));

	addr_t mm = 0;
	addr_t vm_area_struct_head_addr = 0;
	addr_t vm_area_struct_addr = 0;
	addr_t vm_file_addr = 0;
	addr_t start;
	addr_t dentry_path = 0;
	addr_t vdso_addr;

	GHashTable *visited_libs = NULL;
	linux_libs_t *libs_prev = NULL;
	linux_libs_t *libs_first = NULL;
	uint64_t count = 0;

	if (apsh_read_addr_va(apsh, process->address + libs_ctx->task_struct.mm, 0, &mm) == APSH_F)
		return APSH_F;

	if (mm == 0)
		return APSH_S;
	if (apsh_read_addr_va(apsh, mm + libs_ctx->mm_struct.mmap, 0, &vm_area_struct_addr) == APSH_S) {
		if (libs_ctx->mm_struct.vdso) {
			apsh_read_addr_va(apsh, mm + libs_ctx->mm_struct.vdso, 0,
						&vdso_addr);
		}

		vm_area_struct_head_addr = vm_area_struct_addr;
		visited_libs = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, NULL);

		if (visited_libs == NULL)
			goto error_exit;

		while (count++ < limit) {
			char *path = NULL;

			if (apsh_read_addr_va(apsh, vm_area_struct_addr + libs_ctx->vm_area_struct.vm_file, 0,
						&vm_file_addr) == APSH_S) {
				if (apsh_read_addr_va(apsh, vm_file_addr + libs_ctx->file_struct.f_path + libs_ctx->path_struct.dentry,
							0, &dentry_path) == APSH_S) {
					path = apsh_read_str_va(apsh, dentry_path + libs_ctx->dentry_struct.d_name + libs_ctx->qstr_struct.name + 16,
								0);
				}
			}

			if (apsh_read_addr_va(apsh, vm_area_struct_addr + libs_ctx->vm_area_struct.vm_start, 0, &start) != APSH_S)
				goto error_exit;

			if (path != NULL && g_hash_table_lookup(visited_libs, path) == NULL && strstr(path, ".so") != NULL) {
				g_hash_table_insert(visited_libs, path, GINT_TO_POINTER(1));
				// allocate
				linux_libs_t *libs = malloc(sizeof(linux_libs_t));

				if (libs_first == NULL)
					libs_first = libs;

				if (libs_prev)
					libs_prev->next = libs;

				libs->pid = process->pid;
				libs->address = start;
				libs->path = path;
				libs->next = NULL;
				char *lib_name = strrchr(path, '/');

				if (lib_name == NULL)
					libs->lib_name = libs->path;
				else
					libs->lib_name = lib_name + 1;

				libs->data = g_hash_table_new(NULL, NULL);
				if (libs->data == NULL)
					goto error_exit;
				int result;

				result = g_hash_table_insert(
					libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_PID),
					GINT_TO_POINTER(libs->pid));
				char *name_cpy = g_strndup(libs->lib_name, strlen(libs->lib_name) + 1);

				result &= g_hash_table_insert(
					libs->data,
					GINT_TO_POINTER(DOCA_APSH_LIB_COMM),
					name_cpy);
				char *path_cpy = g_strndup(libs->path, strlen(libs->path) + 1);

				result &= g_hash_table_insert(
					libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_LIBRARY_PATH),
					path_cpy);
				result &= g_hash_table_insert(
					libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_LINUX_LOAD_ADRESS),
					GINT_TO_POINTER(libs->address));

				if (!result)
					goto error_exit;

				// set current to prev
				libs_prev = libs;
			} else if (path != NULL)
				free(path);

			/* follow the next pointer */
			if (apsh_read_addr_va(apsh, vm_area_struct_addr + libs_ctx->vm_area_struct.vm_next,
							0, &vm_area_struct_addr) == APSH_F) {
				if (libs_prev != NULL)
					libs_prev->next = NULL;
				break;
			}

			if (vm_area_struct_addr == vm_area_struct_head_addr ||
				vm_area_struct_addr == 0) {
				if (libs_prev != NULL)
					libs_prev->next = NULL;
				break;
			}
		}
	}
	if (visited_libs != NULL)
		g_hash_table_destroy(visited_libs);
	if (*libs_head == NULL)
		*libs_head = libs_first;
	else {
		linux_libs_t *lib_iterator = *libs_head;

		while (lib_iterator->next)
			lib_iterator = lib_iterator->next;
		lib_iterator->next = libs_first;
	}
	return APSH_S;

error_exit:
	if (visited_libs != NULL)
		g_hash_table_destroy(visited_libs);
	if (libs_first != NULL)
		linux_free_libs_list(libs_first);

	return APSH_F;
}

status_t
linux_get_libs(apsh_instance_t apsh, linux_processes_list_t *process_list, addr_t pid,
	linux_libs_t **libs_list)
{
	linux_processes_list_t *psl = NULL, *psl_allocation = NULL;
	struct libs_context *libs_ctx;

	// load up all the offset once
	if (load_offsets(apsh, &libs_ctx) == APSH_F)
		return APSH_F;

	// get process list
	if (process_list == NULL) {
		if (linux_pslist_processes_list(apsh, &psl_allocation))
			return APSH_F;
		psl = psl_allocation;
	} else
		psl = process_list;

	linux_libs_t *libs_head = NULL;

	/* walk the process list */
	while (psl) {
		if (pid == 0 || pid == psl->pid)
			parse_libs(apsh, psl, libs_ctx, &libs_head);
next_process:
		/* follow the next pointer */
		psl = psl->next;
	};
	*libs_list = libs_head;
	if (psl_allocation != NULL)
		linux_free_process_list(psl_allocation);
	return APSH_S;
}

status_t
linux_libs(apsh_instance_t apsh, linux_processes_list_t *process_list, addr_t pid,
	  json_object **json_obj, bool printout)
{
	status_t retcode = APSH_F;

	// get libs list
	linux_libs_t *libs_list = NULL;
	linux_libs_t *head = NULL;

	// print out titles
	char *titles[] = {"PID", "Name", "Path", "Address", "\0"};
	json_object *jout = jsonout_new("libs", titles);

	if (linux_get_libs(apsh, process_list, pid, &libs_list) == APSH_F)
		goto exit;

	head = libs_list;
	while (libs_list) {
		json_object *jdata = jsonout_data_element_new(jout);

		jsonout_data_element_value_sprintf(jdata, "%d", libs_list->pid);
		jsonout_data_element_value_add(jdata, libs_list->lib_name);
		jsonout_data_element_value_add(jdata, libs_list->path);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", libs_list->address);
		libs_list = libs_list->next;
	}

	// convert jsonout to string
	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	retcode = APSH_S;
exit:
	linux_free_libs_list(head);
	return retcode;
}

void
linux_free_libs_list(linux_libs_t *libs_list)
{
	linux_libs_t *libs = libs_list;

	while (libs) {
		linux_libs_t *_next = libs->next;

		if (libs->data) {
			char *comm = g_hash_table_lookup(libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_COMM));
			char *path = g_hash_table_lookup(libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_LIBRARY_PATH));

			if (comm)
				g_free(comm);
			if (path)
				g_free(path);
			g_hash_table_destroy(libs->data);
		}

		free(libs->path);
		free(libs);
		libs = _next;
	}
}
