/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "modules.h"
#include "doca_apsh_attr.h"
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::LINUX::LIBS);

struct MODULES_OFFSET {
	addr_t name;
	addr_t init_layout;
	addr_t init_size;
	addr_t core_layout;
	addr_t core_size;
};

status_t
linux_modules(apsh_instance_t apsh, json_object **json_obj, bool printout)
{
	status_t retval = APSH_F;
	linux_module_list_t *module_list;

	if (linux_get_modules(apsh, &module_list) == APSH_F)
		goto error_exit;

	// create new jsonout object
	char *titles[] = {"Offset", "Name", "Size", "\0"};
	json_object *jout = jsonout_new("linux_modules", titles);

	// walk through module list
	linux_module_list_t *_ml = module_list;

	while (_ml) {
		json_object *jdata = jsonout_data_element_new(jout);

		jsonout_data_element_value_sprintf(jdata, "0x%lx", _ml->entry);
		// dll name
		if (_ml->name)
			jsonout_data_element_value_add(jdata, _ml->name);
		// size
		jsonout_data_element_value_sprintf(jdata, "%d", _ml->size);

		_ml = _ml->next;
	}

	// convert jsonout to string
	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	retval = APSH_S;
error_exit:
	linux_free_module_list(module_list);
	return retval;
}

static char *
linux_modules_get_name(apsh_instance_t apsh, addr_t base)
{
	return apsh_read_str_va(apsh, base + 16, 0);
}

static uint64_t
linux_modules_get_size(apsh_instance_t apsh, addr_t base, struct MODULES_OFFSET *modules_offsets)
{
	uint64_t init = 0;
	uint64_t core = 0;

	if (modules_offsets->init_layout)
		apsh_read_64_va(apsh, base + modules_offsets->init_layout, 0, &init);
	else
		apsh_read_64_va(apsh, base + modules_offsets->init_size, 0, &init);

	if (modules_offsets->core_layout)
		apsh_read_64_va(apsh, base + modules_offsets->core_layout, 0, &core);
	else
		apsh_read_64_va(apsh, base + modules_offsets->core_size, 0, &core);

	return init + core;
}

status_t
modules_offsets_get(apsh_instance_t apsh, struct MODULES_OFFSET **modules_offsets)
{
	struct MODULES_OFFSET *modules_offsets_ptr = (struct MODULES_OFFSET *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_MODULES);

	if (modules_offsets_ptr != NULL) {
		*modules_offsets = modules_offsets_ptr;
		return APSH_S;
	}

	modules_offsets_ptr = malloc(sizeof(*modules_offsets_ptr));

	if (modules_offsets_ptr == NULL) {
		DOCA_DLOG_ERR("Failed to allocate plugin context memory");
		return APSH_F;
	}

	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "module", "init_layout", &modules_offsets_ptr->init_layout) == APSH_F)
		if (apsh_get_struct_member_offset_from_apsh_json(
				apsh, "module", "init_size", &modules_offsets_ptr->init_size) ==
			APSH_F)
			goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "module", "core_layout", &modules_offsets_ptr->core_layout) == APSH_F)
		if (apsh_get_struct_member_offset_from_apsh_json(
				apsh, "module", "core_size", &modules_offsets_ptr->core_size) ==
			APSH_F)
			goto error_exit;

	if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_MODULES, modules_offsets_ptr, free))
		goto error_exit;

	*modules_offsets = modules_offsets_ptr;
	return APSH_S;

error_exit:
	free(modules_offsets_ptr);
	return APSH_F;
}

status_t
linux_get_modules(apsh_instance_t apsh, linux_module_list_t **modules_list)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_MODULES_LIMIT)));

	status_t retval = APSH_F;

	addr_t current_module;
	addr_t name_offset;
	addr_t size_offset;
	addr_t args_offset;

	*modules_list = NULL;

	//  get modules_addr from global symbol "modules_addr"
	if (apsh_read_addr_ksym(apsh, "modules", &current_module) == APSH_F)
		return APSH_F;

	// walk through list
	addr_t head = current_module;
	linux_module_list_t *_ml_header = NULL;
	linux_module_list_t *_prev = NULL;
	uint64_t count = 0;

	/* get modules structs offsets */
	struct MODULES_OFFSET *modules_offsets;

	if (modules_offsets_get(apsh, &modules_offsets) == APSH_F)
		return APSH_F;

	while (count++ < limit) {
		addr_t tmp_next = 0;
		linux_module_list_t *_ml = malloc(sizeof(linux_module_list_t));

		if (_ml == NULL)
			goto error_exit;

		_ml->entry = current_module;
		_ml->next = NULL;
		_ml->data = NULL;

		apsh_read_addr_va(apsh, current_module, 0, &tmp_next);

		// break if we're at head
		if (head == tmp_next) {
			free(_ml);
			break;
		}

		if (_ml_header == NULL)
			_ml_header = _ml;

		if (_prev)
			_prev->next = _ml;

		_ml->name = linux_modules_get_name(apsh, current_module);
		if (_ml->name == NULL)
			goto error_exit;
		_ml->size = linux_modules_get_size(apsh, current_module, modules_offsets);

		char *name_cpy = g_strndup(_ml->name, strlen(_ml->name) + 1);

		if (name_cpy == NULL)
			goto error_exit;

		_ml->data = g_hash_table_new(NULL, NULL);
		if (_ml->data == NULL) {
			g_free(name_cpy);
			goto error_exit;
		}

		int result;

		result = g_hash_table_insert(_ml->data, GINT_TO_POINTER(DOCA_APSH_MODULES_OFFSET),
					     GINT_TO_POINTER(_ml->entry));
		result &= g_hash_table_insert(_ml->data, GINT_TO_POINTER(DOCA_APSH_MODULES_NAME),
					      name_cpy);
		result &= g_hash_table_insert(_ml->data, GINT_TO_POINTER(DOCA_APSH_MODULES_SIZE),
					      GINT_TO_POINTER(_ml->size));
		if (!result)
			goto error_exit;

get_next_module:
		current_module = tmp_next;
		_prev = _ml;
	}

	retval = APSH_S;
	*modules_list = _ml_header;
	return retval;

error_exit:
	if (_ml_header != NULL)
		linux_free_module_list(_ml_header);

	return retval;
}

void
linux_free_module_list(linux_module_list_t *modules_list)
{
	linux_module_list_t *_ml = modules_list;

	while (_ml) {
		linux_module_list_t *_next = _ml->next;

		if (_ml->name)
			free(_ml->name);
		if (_ml->data) {
			char *name_cpy = g_hash_table_lookup(_ml->data, GINT_TO_POINTER(DOCA_APSH_MODULES_NAME));

			if (name_cpy)
				g_free(name_cpy);
			g_hash_table_destroy(_ml->data);
		}
		free(_ml);

		_ml = _next;
	}
}
