/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "pslist.h"
#include "../../doca_apsh_attr.h"
#include "jsonout.h"
#include "../../../core/APSH/apsh_instance.h"

#include "doca_log.h"

DOCA_LOG_REGISTER(APSH::LINUX::PSLIST);

linux_processes_list_t *
linux_pslist_get_process(linux_processes_list_t *process_list, uint32_t pid)
{
	if (process_list == NULL)
		return NULL;

	linux_processes_list_t *psl = process_list;

	while (psl) {
		if (psl->pid == pid)
			return psl;

		psl = psl->next;
	}

	DOCA_DLOG_DBG("does not find proccess [%d]", pid);
	return NULL;
}

status_t
linux_pslist(apsh_instance_t apsh, linux_processes_list_t *process_list, json_object **json_obj,
	     bool printout)
{

	linux_processes_list_t *psl, *psl_allocation = NULL;

	if (process_list == NULL) {
		if (linux_pslist_processes_list(apsh, &psl_allocation) == APSH_F)
			return APSH_F;
		psl = psl_allocation;
	} else
		psl = process_list;

	char *procname = NULL;
	apsh_pid_t pid = 0;
	uint64_t ppid;
	addr_t current_process = 0;

	addr_t PsActiveProcessHead = psl->address;

	// create new jsonout object
	char *titles[] = {"PID", "PPID", "state", "ImageFileName", "Offset(V)", "Threads",
			  "UID", "GID",	 "\0"};
	json_object *jout = jsonout_new("pslist", titles);

	while (1) {
		json_object *jdata = jsonout_data_element_new(jout);

		jsonout_data_element_value_sprintf(jdata, "%d", psl->pid);
		jsonout_data_element_value_sprintf(jdata, "%d", psl->ppid);
		jsonout_data_element_value_sprintf(jdata, "%d", psl->state);
		jsonout_data_element_value_add(jdata, psl->name);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", psl->address);
		jsonout_data_element_value_sprintf(jdata, "%d", psl->uid);
		jsonout_data_element_value_sprintf(jdata, "%d", psl->gid);

		/* follow the next pointer */
		if (psl->next == NULL)
			break;

		psl = psl->next;

		if (psl->address == PsActiveProcessHead)
			break;
	}

	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);
	if (psl_allocation != NULL)
		linux_free_process_list(psl_allocation);
	return APSH_S;
}

status_t
linux_pslist_processes_list(apsh_instance_t apsh, linux_processes_list_t **pprocess_list)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_PROCESS_LIMIT)));

	status_t retval = APSH_F;

	addr_t tasks_offset;
	addr_t process_name_offset;
	addr_t pid_offset; // PID = Thread ID in linux?!
	addr_t pslist_head;
	addr_t real_parent_offset;
	addr_t cpu_time_offset;
	addr_t parent_offset;
	addr_t tgid_offset; // actual PID
	addr_t cred_offset;
	addr_t uid_offset;
	addr_t gid_offset;
	addr_t mm_offset;
	addr_t arg_start;
	addr_t arg_end;
	addr_t state_offset;

	linux_processes_list_t *psl = NULL;

	/* get offsets in _EPROCESS */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "tasks",
							 &tasks_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "pid", &pid_offset) ==
	    APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "comm",
							 &process_name_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "parent",
							 &parent_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "real_parent",
							 &real_parent_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "sum_exec_runtime",
							 &cpu_time_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "tgid",
							 &tgid_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "cred",
							 &cred_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "state",
							 &state_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "cred", "uid", &uid_offset) ==
	    APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "cred", "gid", &gid_offset) ==
	    APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "mm", &mm_offset) ==
	    APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "arg_start",
							 &arg_start) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "arg_end", &arg_end) ==
	    APSH_F)
		goto error_exit_no_free;

	/* get first ActiveProcessLink head (VA) from global symbol 'PsActiveProcessHead' */
	if (apsh_translate_ksym2v(apsh, "init_task", &pslist_head) == APSH_F)
		goto error_exit_no_free;

	pslist_head += tasks_offset;

	addr_t current_entry = pslist_head;
	addr_t next_entry = 0;
	addr_t current_process = 0;
	uint64_t count = 0;

	// Read current entry for next entry!
	if (apsh_read_addr_va(apsh, current_entry, 0, &next_entry) == APSH_F) {
		DOCA_DLOG_ERR("Failed to read next pointer at %" PRIx64, current_entry);
		goto error_exit_no_free;
	}

	psl = malloc(sizeof(linux_processes_list_t));
	if (psl == NULL) {
		DOCA_DLOG_ERR("Failed to allocate memory for next");
		goto error_exit_free_list;
	}

	*pprocess_list = psl;

	while (count++ < limit) {
		/* calculating current _EPROCESS address */
		current_process = current_entry - tasks_offset;

		psl->address = current_process;
		psl->next = NULL;

		/* get PID from current _EPROCESS */
		apsh_read_32_va(apsh, current_process + tgid_offset, 0,
				(uint32_t *)&psl->pid); // PID
		psl->name = apsh_read_str_va(apsh, current_process + process_name_offset,
					     0); // ImageFileName
		if (psl->name == NULL) {
			DOCA_DLOG_ERR("Failed to read ImageFileName at %" PRIx64,
			       current_process + process_name_offset);
			goto error_exit_free_list;
		}

		apsh_read_64_va(apsh, current_process + state_offset, 0,
				(uint64_t *)&psl->state); // state

		addr_t ppid_struct = 0;
		addr_t cred;

		if (apsh_read_addr_va(apsh, current_process + real_parent_offset, 0, &ppid_struct) == APSH_F || !ppid_struct) {
			if (apsh_read_addr_va(apsh, ppid_struct + parent_offset, 0, &ppid_struct) == APSH_F) {
				DOCA_DLOG_ERR("Failed to find parent at %" PRIx64,
				       current_process + real_parent_offset);
				goto error_exit_free_list;
			}
		}
		apsh_read_32_va(apsh, ppid_struct + tgid_offset, 0, (uint32_t *)&psl->ppid); // PPID

		if (apsh_read_addr_va(apsh, current_process + cred_offset, 0, &cred) == APSH_F) {
			DOCA_DLOG_ERR("Failed to find cred at %" PRIx64,
			       current_process + cred_offset);
			goto error_exit_free_list;
		}
		apsh_read_32_va(apsh, cred + uid_offset, 0, (uint32_t *)&psl->uid); // UID
		apsh_read_32_va(apsh, cred + gid_offset, 0, (uint32_t *)&psl->gid); // GID
		apsh_read_64_va(apsh, current_process + cpu_time_offset, 0, (uint64_t *)&psl->cpu_time); // sum_exec_runtime

		GHashTable *hash = g_hash_table_new(NULL, NULL);

		if (hash == NULL)
			goto error_exit_free_list;

		char *proc_name_copy = g_strndup(psl->name, strlen(psl->name) + 1);

		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_PID),
				    GINT_TO_POINTER(psl->pid));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_LINUX_GID),
				    GINT_TO_POINTER(psl->gid));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_LINUX_UID),
				    GINT_TO_POINTER(psl->uid));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_PPID),
				    GINT_TO_POINTER(psl->ppid));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_CPU_TIME),
				    GINT_TO_POINTER(psl->cpu_time));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_COMM),
				    proc_name_copy);
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_STATE),
				    GINT_TO_POINTER(psl->state));

		psl->hash = hash;

		/* follow the next pointer */
		current_entry = next_entry;
		if (apsh_read_addr_va(apsh, current_entry, 0, &next_entry) == APSH_F) {
			DOCA_DLOG_ERR("Failed to read next pointer at %" PRIx64, current_entry);
			goto error_exit_free_list;
		}

		if (current_entry == pslist_head)
			break;

		psl->next = malloc(sizeof(linux_processes_list_t));
		if (psl->next == NULL) {
			DOCA_DLOG_ERR("Failed to allocate memory for psl->next");
			goto error_exit_free_list;
		}

		psl = psl->next;
	}

	retval = APSH_S;
	return retval;

error_exit_free_list:
	linux_free_process_list(*pprocess_list);
	*pprocess_list = NULL;
error_exit_no_free:
	return retval;
}

void
linux_free_process_list(linux_processes_list_t *process_list)
{
	linux_processes_list_t *p_next;

	while (process_list) {
		p_next = process_list->next;
		free(process_list->name);
		free(g_hash_table_lookup(process_list->hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_COMM)));
		g_hash_table_destroy(process_list->hash);
		free(process_list);
		process_list = p_next;
	}
}
