/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "threadlist.h"
#include "../../doca_apsh_attr.h"
#include "jsonout.h"
#include "doca_log.h"
#include "../../../core/APSH/apsh_instance.h"


DOCA_LOG_REGISTER(APSH::LINUX::STACK);

static void
print_threadlist(linux_thread_list_t *thread_list, bool from_hash)
{

	DOCA_DLOG_DBG("\n\nPRITNING RESULTS:\n");

	DOCA_DLOG_DBG("PID     Process     TID     Offset     Thread Name");

	while (thread_list) {

		if (from_hash) {
			GHashTable *hash_table = thread_list->hash;
			uint32_t pid = g_hash_table_lookup(
				hash_table, GINT_TO_POINTER(DOCA_APSH_THREAD_PID));
			char *process_name = g_hash_table_lookup(
				hash_table, GINT_TO_POINTER(DOCA_APSH_THREAD_LINUX_PROC_NAME));
			uint32_t tid = g_hash_table_lookup(
				hash_table, GINT_TO_POINTER(DOCA_APSH_THREAD_TID));
			char *thread_name = g_hash_table_lookup(
				hash_table, GINT_TO_POINTER(DOCA_APSH_THREAD_LINUX_THREAD_NAME));

			uint64_t state = g_hash_table_lookup(
				hash_table, GINT_TO_POINTER(DOCA_APSH_THREAD_STATE));

			DOCA_DLOG_DBG("[%d]     [%s]     [%d]     [%lx]     [%s]", pid, process_name,
			       tid, thread_list->address, thread_name);
		} else {
			DOCA_DLOG_DBG("[%d]     [%s]     [%d]     [%lx]     [%s]", thread_list->pid,
			       thread_list->proc_name, thread_list->tid, thread_list->address,
			       thread_list->thread_name);
		}
		thread_list = thread_list->next;
	}
	DOCA_DLOG_DBG("\n");
}

status_t
linux_get_threadlist(apsh_instance_t apsh, linux_processes_list_t *process_list, uint32_t pid,
		     linux_thread_list_t **thread_list)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_THREADS_LIMIT)));


	status_t retcode = APSH_F;
	linux_processes_list_t *psl = NULL, *psl_allocation = NULL;
	linux_thread_list_t *thl = NULL, *curr_thead = NULL, *new_thread = NULL;

	if (thread_list == NULL)
		goto error_exit;

	if (process_list == NULL) {
		retcode = linux_pslist_processes_list(apsh, &psl_allocation);
		if (retcode != APSH_S)
			goto error_exit;

		psl = psl_allocation;
	} else
		psl = process_list;

	addr_t thread_offset;
	addr_t thread_group_offset;
	addr_t tasks_offset;
	addr_t process_name_offset;
	addr_t pid_offset; // PID = Thread ID in linux?!
	addr_t prev_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "thread_group",
							 &thread_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "group_leader",
							 &thread_group_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "tasks",
							 &tasks_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "pid", &pid_offset) ==
	    APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "comm",
							 &process_name_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "list_head", "prev", &prev_offset) ==
	    APSH_F)
		goto error_exit;

	addr_t current_process = 0;
	/* walk the process list */
	while (psl) {
		current_process = psl->address;
		if (pid == 0 || pid == psl->pid) {
			addr_t thread_head;

			if (apsh_read_addr_va(apsh, current_process + thread_offset, 0, &thread_head) == APSH_F)
				goto next_process;

			addr_t next = thread_head;
			addr_t current;
			uint64_t count = 0;

			while (count++ < limit) {

				new_thread = malloc(sizeof(linux_thread_list_t));
				if (!new_thread)
					goto error_exit;

				new_thread->next = NULL;
				new_thread->proc_name = strndup(psl->name, strlen(psl->name) + 1);
				new_thread->pid = psl->pid;
				new_thread->state = psl->state;
				if (apsh_read_addr_va(apsh, next + prev_offset, psl->pid, &current) == APSH_F) {
					free(new_thread);
					break;
				}

				addr_t thread_base = current - thread_offset;

				new_thread->address = thread_base;

				/* get PID from current _EPROCESS */

				apsh_read_32_va(apsh, thread_base + pid_offset, psl->pid,
						&(new_thread->tid)); // TID
				new_thread->thread_name =
					apsh_read_str_va(apsh, thread_base + process_name_offset,
							 psl->pid); // thread's name
				if (new_thread->thread_name == NULL) {
					free(new_thread);
					goto error_exit;
				}

				GHashTable *hash = g_hash_table_new(NULL, NULL);

				if (hash == NULL) {
					free(new_thread);
					goto error_exit;
				}

				char *thread_name_copy =
					g_strndup(new_thread->thread_name,
						  strlen(new_thread->thread_name) + 1);

				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_THREAD_PID),
						    GUINT_TO_POINTER(new_thread->pid));
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_THREAD_TID),
						    GUINT_TO_POINTER(new_thread->tid));
				g_hash_table_insert(
					hash, GINT_TO_POINTER(DOCA_APSH_THREAD_LINUX_PROC_NAME),
					new_thread->proc_name);
				g_hash_table_insert(
					hash, GINT_TO_POINTER(DOCA_APSH_THREAD_LINUX_THREAD_NAME),
					thread_name_copy);
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_THREAD_STATE),
						    GUINT_TO_POINTER(new_thread->state));

				new_thread->hash = hash;

				if (thl == NULL) { // 1st node in the list
					thl = curr_thead = new_thread;
				} else {
					curr_thead->next = new_thread;
					curr_thead = curr_thead->next;
				}

				if (APSH_F ==
				    apsh_read_addr_va(apsh, current + prev_offset, psl->pid, &next))
					break;

				if ((next == thread_head) || (next == NULL))
					break;
			}

			if (pid == psl->pid && pid != 0)
				break;
		}

next_process:
		/* follow the next pointer */
		psl = psl->next;
	}
	if (psl_allocation != NULL)
		linux_free_process_list(psl_allocation);
	*thread_list = thl;
	return APSH_S;

error_exit:
	if (psl_allocation != NULL)
		linux_free_process_list(psl_allocation);
	if (thl != NULL)
		linux_free_thread_list(thl);
	return APSH_F;
}

status_t
linux_threadlist(apsh_instance_t apsh, linux_processes_list_t *process_list, uint32_t pid,
		 json_object **json_obj, bool printout)
{

	linux_thread_list_t *thread_list = NULL, *thread_list_head = NULL;

	if (linux_get_threadlist(apsh, process_list, pid, &thread_list) != APSH_S)
		goto error_exit;

	char *titles[] = {"PID", "Process", "TID", "Offset", "Thread Name", "\0"};
	json_object *jout = jsonout_new("threadlist", titles);

	thread_list_head = thread_list;
	while (thread_list) {
		json_object *jdata = jsonout_data_element_new(jout);

		jsonout_data_element_value_sprintf(jdata, "%d", thread_list->pid);
		jsonout_data_element_value_add(jdata, thread_list->proc_name);
		jsonout_data_element_value_sprintf(jdata, "%d", thread_list->tid);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", thread_list->address);
		jsonout_data_element_value_add(jdata, thread_list->thread_name);
		thread_list = thread_list->next;
	}

	thread_list = thread_list_head;

	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	linux_free_thread_list(thread_list);
	return APSH_S;

error_exit:
	return APSH_F;
}

void
linux_free_thread_list(linux_thread_list_t *thread_list)
{

	linux_thread_list_t *p_next = NULL;

	while (thread_list) {
		p_next = thread_list->next;
		free(thread_list->thread_name);
		free(thread_list->proc_name);
		free(g_hash_table_lookup(thread_list->hash, GINT_TO_POINTER(DOCA_APSH_THREAD_LINUX_THREAD_NAME)));
		g_hash_table_destroy(thread_list->hash);
		free(thread_list);
		thread_list = p_next;
	}
}
