/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "vma.h"
#include "doca_apsh_attr.h"
#include <doca_log.h>
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::LINUX::VMC);

struct protection_t {
	char name[32];
	uint32_t mask;
};

static const struct protection_t protect_vma[] = {
	{"r", 0x00000001},
	{"w", 0x00000002},
	{"x", 0x00000004}
};

struct vma_context {
	struct _task_struct {
		addr_t tasks;
		addr_t comm;
		addr_t pid;
		addr_t mm;
	} task_struct;

	struct _mm_struct {
		addr_t mmap;
		addr_t start_brk;
		addr_t brk;
		addr_t start_stack;
		addr_t vdso;
	} mm_struct;

	struct _vm_area_struct {
		addr_t vm_pgoff;
		addr_t vm_start;
		addr_t vm_end;
		addr_t vm_file;
		addr_t vm_page_prot;
		addr_t vm_next;
		addr_t vm_mm;
		addr_t vm_flags;
	} vm_area_struct;

	struct _file_struct {
		addr_t f_inode;
		addr_t f_owner;
		addr_t f_path;
	} file_struct;

	struct _path_struct {
		addr_t dentry;
	} path_struct;

	struct _dentry_struct {
		addr_t d_name;
	} dentry_struct;

	struct _qstr_struct {
		addr_t name;
	} qstr_struct;
};

static status_t
load_offsets_vma(apsh_instance_t apsh, struct vma_context **vma_ctx)
{
	struct vma_context *vma_ctx_ptr = (struct vma_context *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_VMA);

	if (vma_ctx_ptr != NULL) {
		*vma_ctx = vma_ctx_ptr;
		return APSH_S;
	}

	vma_ctx_ptr = malloc(sizeof(*vma_ctx_ptr));

	if (vma_ctx_ptr == NULL) {
		DOCA_DLOG_ERR("Failed to allocate plugin context memory");
		return APSH_F;
	}
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "tasks",
								&vma_ctx_ptr->task_struct.tasks) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "pid",
								&vma_ctx_ptr->task_struct.pid) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "comm",
								&vma_ctx_ptr->task_struct.comm) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "task_struct", "mm",
								&vma_ctx_ptr->task_struct.mm) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "mmap",
								&vma_ctx_ptr->mm_struct.mmap) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "brk",
								&vma_ctx_ptr->mm_struct.brk) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_brk",
								&vma_ctx_ptr->mm_struct.start_brk) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "start_stack",
								&vma_ctx_ptr->mm_struct.start_stack) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_pgoff", &vma_ctx_ptr->vm_area_struct.vm_pgoff) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_start", &vma_ctx_ptr->vm_area_struct.vm_start) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_end",
								&vma_ctx_ptr->vm_area_struct.vm_end) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_file",
								&vma_ctx_ptr->vm_area_struct.vm_file) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_page_prot", &vma_ctx_ptr->vm_area_struct.vm_page_prot) ==
		APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(
			apsh, "vm_area_struct", "vm_flags", &vma_ctx_ptr->vm_area_struct.vm_flags) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_next",
								&vma_ctx_ptr->vm_area_struct.vm_next) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_inode",
								&vma_ctx_ptr->file_struct.f_inode) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_owner",
								&vma_ctx_ptr->file_struct.f_owner) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "file", "f_path",
								&vma_ctx_ptr->file_struct.f_path) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "path", "dentry",
								&vma_ctx_ptr->path_struct.dentry) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "dentry", "d_name",
								&vma_ctx_ptr->dentry_struct.d_name) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "qstr", "name",
								&vma_ctx_ptr->qstr_struct.name) == APSH_F)
		goto error_exit;

	// profile might or might not have this part
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "vm_area_struct", "vm_mm",
								&vma_ctx_ptr->vm_area_struct.vm_mm) == APSH_F)
		vma_ctx_ptr->vm_area_struct.vm_mm = NULL;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "mm_struct", "vdso",
								&vma_ctx_ptr->mm_struct.vdso) == APSH_F)
		vma_ctx_ptr->mm_struct.vdso = NULL;

	if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_VMA, vma_ctx_ptr, free))
		goto error_exit;

	*vma_ctx = vma_ctx_ptr;
	return APSH_S;

error_exit:
	free(vma_ctx_ptr);
	return APSH_F;
}

static status_t
parse_vma(apsh_instance_t apsh, linux_processes_list_t *process_list, struct vma_context *vma_ctx, linux_vma_t **vma_head)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_VADS_LIMIT)));

	addr_t mm = 0;
	addr_t vm_area_struct_head_addr = 0;
	addr_t vm_area_struct_addr = 0;
	addr_t vm_file_addr = 0;
	addr_t dentry_path = 0;
	addr_t vdso_addr = 0;
	char *path = NULL;
	linux_vma_t *vma_prev = NULL;
	uint64_t count = 0;

	if (apsh_read_addr_va(apsh, process_list->address + vma_ctx->task_struct.mm, 0, &mm) == APSH_F)
		return APSH_F;

	if (mm != 0) {
		if (apsh_read_addr_va(apsh, mm + vma_ctx->mm_struct.mmap, process_list->pid, &vm_area_struct_addr) == APSH_S) {
			if (vma_ctx->mm_struct.vdso) {
				apsh_read_addr_va(apsh, mm + vma_ctx->mm_struct.vdso, process_list->pid,
						  &vdso_addr);
			}

			vm_area_struct_head_addr = vm_area_struct_addr;
			*vma_head = NULL; /* prevents memory leak due to uninitialized value */
			while (count++ < limit) {
				addr_t start;
				addr_t end;
				addr_t flag;
				addr_t brk;
				addr_t start_brk;
				addr_t start_stack;

				apsh_read_addr_va(apsh, mm + vma_ctx->mm_struct.brk, process_list->pid, &brk);
				apsh_read_addr_va(apsh, mm + vma_ctx->mm_struct.start_brk, process_list->pid,
						  &start_brk);
				apsh_read_addr_va(apsh, mm + vma_ctx->mm_struct.start_stack,
						  process_list->pid, &start_stack);
				apsh_read_addr_va(apsh,
						  vm_area_struct_addr +vma_ctx-> vm_area_struct.vm_flags,
						  process_list->pid, &flag);
				apsh_read_addr_va(apsh,
						  vm_area_struct_addr + vma_ctx->vm_area_struct.vm_start,
						  process_list->pid, &start);
				apsh_read_addr_va(apsh, vm_area_struct_addr + vma_ctx->vm_area_struct.vm_end,
						  process_list->pid, &end);

				int count = sizeof(protect_vma) / sizeof(struct protection_t);
				char buf[256];
				char *p = buf;
				int i = 0;

				memset(buf, '\0', sizeof(buf));
				for (i = 0; i < count; i++) {
					if ((flag)&protect_vma[i].mask) {
						sprintf(p, "%s", protect_vma[i].name);
						p += strlen(p);
					} else {
						sprintf(p, "-");
						p += strlen(p);
					}
				}

				// allocate
				linux_vma_t *vma = malloc(sizeof(linux_vma_t));

				if(vma == NULL){
					DOCA_DLOG_ERR("failed to malloc");
					return APSH_F;
				}

				if (*vma_head == NULL)
					*vma_head = vma_prev = vma;

				if (vma_prev != NULL)
					vma_prev->next = vma;

				vma->process_name = process_list->name;
				vma->pid = process_list->pid;
				vma->offset = vm_area_struct_addr;
				vma->protection = buf;
				vma->start = start;
				vma->end = end;

				vma->path_alloc = false;

				if (apsh_read_addr_va(apsh, vm_area_struct_addr + vma_ctx->vm_area_struct.vm_file, process_list->pid,
							&vm_file_addr) == APSH_S) {
					if (apsh_read_addr_va(apsh, vm_file_addr + vma_ctx->file_struct.f_path + vma_ctx->path_struct.dentry,
								process_list->pid, &dentry_path) == APSH_S) {
						path = apsh_read_str_va(apsh, dentry_path + vma_ctx->dentry_struct.d_name + vma_ctx->qstr_struct.name + 16,
									process_list->pid);
						if (path) {
							vma->path_alloc = true;
							vma->path = path;
						}
					}
				}

				if (!vma->path_alloc) {
					if (start <= start_brk && end >= brk)
						vma->path = "Heap";
					else if (start <= start_stack && end >= start_stack)
						vma->path = "Stack";
					else if (vdso_addr != NULL && vdso_addr == start)
						vma->path = "Vdso";
					else
						vma->path = "Anonymous Mapping";
				}

				vma->data = g_hash_table_new(NULL, NULL);
				if (vma->data == NULL)
					return APSH_F;
				int result;

				result = g_hash_table_insert(
					vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_PID),
					GINT_TO_POINTER(vma->pid));
				char *name_cpy =
					g_strndup(vma->process_name, strlen(vma->process_name) + 1);
				result &= g_hash_table_insert(
					vma->data,
					GINT_TO_POINTER(DOCA_APSH_VMA_PROCESS_NAME),
					name_cpy);
				result &= g_hash_table_insert(
					vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_OFFSET),
					GINT_TO_POINTER(vma->offset));
				char *protection_cpy =
					g_strndup(vma->protection, strlen(vma->protection) + 1);
				result &= g_hash_table_insert(
					vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_PROTECTION),
					protection_cpy);
				result &= g_hash_table_insert(
					vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_VM_START),
					GINT_TO_POINTER(vma->start));
				result &= g_hash_table_insert(
					vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_VM_END),
					GINT_TO_POINTER(vma->end));
				char *path_cpy = g_strndup(vma->path, strlen(vma->path) + 1);

				result &= g_hash_table_insert(
					vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_FILE_PATH),
					path_cpy);
				if (!result)
					return APSH_F;

				// set current to prev
				vma_prev = vma;

				/* follow the next pointer */
				if (APSH_F ==
				    apsh_read_addr_va(apsh,
						      vm_area_struct_addr + vma_ctx->vm_area_struct.vm_next,
						      process_list->pid, &vm_area_struct_addr)) {
					vma->next = NULL;
					break;
				}

				if (vm_area_struct_addr == vm_area_struct_head_addr ||
				    vm_area_struct_addr == 0) {
					vma->next = NULL;
					break;
				}
			}
		}
	}

	return APSH_S;
}

status_t
linux_get_vma(apsh_instance_t apsh, linux_processes_list_t *process_list, addr_t pid,
	linux_vma_t **vma_list)
{
	linux_processes_list_t *psl = NULL, *psl_allocation = NULL;
	struct vma_context *vma_ctx;
	// load up all the offset once
	if (load_offsets_vma(apsh, &vma_ctx) == APSH_F)
		return APSH_F;

	// get process list
	if (process_list == NULL) {
		if (linux_pslist_processes_list(apsh, &psl_allocation))
			return APSH_F;
		psl = psl_allocation;
	} else
		psl = process_list;

	linux_vma_t *vma_head = NULL;

	/* walk the process list */
	while (psl) {
		if (pid == 0 || pid == psl->pid)
			parse_vma(apsh, psl, vma_ctx, &vma_head);
next_process:
		/* follow the next pointer */
		psl = psl->next;
	};
	*vma_list = vma_head;
	if (psl_allocation != NULL)
		linux_free_process_list(psl_allocation);
	return APSH_S;
}

status_t
linux_vma(apsh_instance_t apsh, linux_processes_list_t *process_list, addr_t pid,
	  json_object **json_obj, bool printout)
{
	status_t retcode = APSH_F;

	// get vma list
	linux_vma_t *vma_list = NULL;
	linux_vma_t *head = NULL;

	if (linux_get_vma(apsh, process_list, pid, &vma_list) == APSH_F)
		goto exit;

	// print out titles
	char *titles[] = {"Process", "PID", "Offset", "Protection", "Start", "End", "Path", "\0"};
	json_object *jout = jsonout_new("vma", titles);

	head = vma_list;
	while (vma_list) {
		json_object *jdata = jsonout_data_element_new(jout);

		jsonout_data_element_value_add(jdata, vma_list->process_name);
		jsonout_data_element_value_sprintf(jdata, "%d", vma_list->pid);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", vma_list->offset);
		jsonout_data_element_value_add(jdata, vma_list->protection);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", vma_list->start);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", vma_list->end);
		jsonout_data_element_value_add(jdata, vma_list->path);

		vma_list = vma_list->next;
	}

	// convert jsonout to string
	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	retcode = APSH_S;
exit:
	linux_free_vma_list(head);
	return retcode;
}

void
linux_free_vma_list(linux_vma_t *vma_list)
{
	linux_vma_t *vma = vma_list;

	while (vma) {
		linux_vma_t *_next = vma->next;

		if (vma->data) {

			char *name = g_hash_table_lookup(vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_PROCESS_NAME));
			char *protection = g_hash_table_lookup(vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_PROTECTION));
			char *path = g_hash_table_lookup(vma->data, GINT_TO_POINTER(DOCA_APSH_VMA_FILE_PATH));

			if (name)
				g_free(name);
			if (protection)
				g_free(protection);
			if (path)
				g_free(path);
			g_hash_table_destroy(vma->data);
		}

		if (vma->path_alloc)
			free(vma->path);
		free(vma);
		vma = _next;
	}
}
