/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dlllist.h"
#include "doca_apsh_attr.h"
#include "jsonout.h"
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::Windows::DLLLIST);

void release_export_info(struct windows_export_info *info)
{
	g_hash_table_destroy(info->exports);
	free(info);
}

status_t
windows_store_dll_functions(apsh_instance_t apsh, addr_t dll_base, uint32_t dll_size, apsh_pid_t pid,
		    char *full_path)
{
	GHashTable *dll_exports = (GHashTable *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_DLL_EXPORTS);

	if (dll_exports == NULL) {
		dll_exports = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, release_export_info);
		if (dll_exports == NULL)
			return APSH_F;

		if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_DLL_EXPORTS, dll_exports, g_hash_table_destroy)) {
			g_hash_table_destroy(dll_exports);
			return APSH_F;
		}
	}

	struct windows_export_info *test =
		(struct windows_export_info *)g_hash_table_lookup(dll_exports, full_path);
	if (test) {
		if (test->dll_base == dll_base)
			return APSH_S;
	}

	struct export_table et;

	ACCESS_CONTEXT(ctx, .translate_mechanism = APSH_TM_PROCESS_PID, .addr = dll_base,
		       .pid = pid);

	if (peparse_get_export_table(apsh, &ctx, &et, NULL, NULL) == APSH_S) {
		struct windows_export_info *info = malloc(sizeof(*info));

		info->exports = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
		info->dll_base = dll_base;
		info->dll_size = dll_size;
		char *name_cpy = g_strndup(full_path, strlen(full_path) + 1);

		/* dump function names into export table for dll */
		addr_t base_addr = dll_base;
		addr_t base1 = base_addr + et.address_of_names;
		addr_t base2 = base_addr + et.address_of_name_ordinals;
		addr_t base3 = base_addr + et.address_of_functions;
		uint32_t i = 0;

		for (; i < et.number_of_names; ++i) {
			uint32_t rva = 0;
			uint16_t ordinal = 0;
			uint32_t loc = 0;
			char *func_name = NULL;

			ctx.addr = base1 + i * sizeof(uint32_t);
			if (apsh_read_32(apsh, &ctx, &rva) == APSH_F)
				continue;

			if (rva) {
				ctx.addr = base_addr + rva;
				func_name = apsh_read_str(apsh, &ctx);
				if (func_name) {
					ctx.addr = base2 + i * sizeof(uint16_t);
					if (apsh_read_16(apsh, &ctx, &ordinal) == APSH_F) {
						free(func_name);
						continue;
					}

					ctx.addr = base3 + ordinal * sizeof(uint32_t);
					if (apsh_read_32(apsh, &ctx, &loc) == APSH_F) {
						free(func_name);
						continue;
					}

					g_hash_table_insert(info->exports, func_name,
							    GINT_TO_POINTER(base_addr + loc));
				}
			}
		}

		g_hash_table_insert(dll_exports, name_cpy, info);
	} else {
		return APSH_F;
	}

	return APSH_S;
}

status_t
windows_save_user_exports(apsh_instance_t apsh, windows_processes_list_t *process_list)
{
	status_t retcode = APSH_F;
	windows_processes_list_t *psl = NULL;

	if (process_list == NULL)
		return APSH_F;
	psl = process_list;

	addr_t ImageFileName_offset;
	addr_t Peb_offset;
	addr_t Ldr_offset;
	addr_t InLoadOrderModuleList_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
							 &ImageFileName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Peb", &Peb_offset) ==
	    APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB", "Ldr", &Ldr_offset) ==
	    APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB_LDR_DATA",
							 "InLoadOrderModuleList",
							 &InLoadOrderModuleList_offset) == APSH_F)
		goto error_exit;

	addr_t InLoadOrderLinks_offset;
	addr_t BaseDllName_offset;
	addr_t FullDllName_offset;
	addr_t DllBase_offset;
	addr_t SizeOfImage_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY",
							 "InLoadOrderLinks",
							 &InLoadOrderLinks_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "BaseDllName", &BaseDllName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "FullDllName", &FullDllName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY", "DllBase",
							 &DllBase_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "SizeOfImage", &SizeOfImage_offset) == APSH_F)
		goto error_exit;

	if (psl) {
		addr_t current_process = psl->address;

		/* get _EPROCESS Peb offset */
		addr_t Peb_va;
		uint64_t Peb = 0;

		Peb_va = current_process + Peb_offset;
		if (apsh_read_64_va(apsh, Peb_va, 0, &Peb) == APSH_F)
			return APSH_F;

		/* get _PEB Ldr */
		addr_t Ldr_va;
		uint64_t Ldr = 0;
		addr_t paddr;

		Ldr_va = Peb + Ldr_offset;

		if (apsh_translate_uv2p(apsh, Peb, psl->pid, &paddr) == APSH_F)
			return APSH_F;

		if (apsh_read_64_va(apsh, Ldr_va, psl->pid, &Ldr) == APSH_F)
			return APSH_F;

		/* get  _PEB_LDR_DATA InLoadOrderModuleList */
		addr_t InLoadOrderModuleList_va;

		struct _LIST_ENTRY {
			addr_t flink;
			addr_t blink;
		} InLoadOrderModuleList;

		InLoadOrderModuleList_va = Ldr + InLoadOrderModuleList_offset;

		if (apsh_read_va(apsh, InLoadOrderModuleList_va, psl->pid,
				 sizeof(struct _LIST_ENTRY), &InLoadOrderModuleList,
				 NULL) == APSH_F)
			return APSH_F;

		/* get  _LDR_DATA_TABLE_ENTRY BaseDllName, FullDllName, DllBase, SizeOfImage */
		addr_t BaseDllName_va;
		addr_t FullDllName_va;
		addr_t DllBase_va, DllBase;
		addr_t SizeOfImage_va;
		uint32_t SizeOfImage;

		// get Flink of InLoadOrderModuleList
		addr_t flink_va, flink;

		flink = InLoadOrderModuleList.flink;

		addr_t link_header = flink;
		uint32_t count = 0;

		unicode_string_t *us = NULL;
		unicode_string_t out = {0};

		unicode_string_t *us1 = NULL;
		unicode_string_t out1 = {0};

		do {
			addr_t ldr_data_va;

			ldr_data_va = flink - InLoadOrderLinks_offset;

			/* get DllBase */
			DllBase_va = ldr_data_va + DllBase_offset;
			apsh_read_64_va(apsh, DllBase_va, psl->pid, &DllBase);

			if (DllBase == 0)
				break;

			/* get SizeOfImage */
			SizeOfImage_va = ldr_data_va + SizeOfImage_offset;
			apsh_read_32_va(apsh, SizeOfImage_va, psl->pid, &SizeOfImage);

			/* get BaseDllName */
			BaseDllName_va = ldr_data_va + BaseDllName_offset;
			us = apsh_read_unicode_str_va(apsh, BaseDllName_va, psl->pid);

			/* get FullDllName */
			FullDllName_va = ldr_data_va + FullDllName_offset;
			us1 = apsh_read_unicode_str_va(apsh, FullDllName_va, psl->pid);

			if (us1 && APSH_S == apsh_convert_str_encoding(us1, &out1, "UTF-8")) {
				if (out1.contents) {
					windows_store_dll_functions(apsh, DllBase, SizeOfImage, psl->pid,
							    out1.contents);
				}
			}

			if (us)
				apsh_free_unicode_str(us);
			if (us1)
				apsh_free_unicode_str(us1);

			/* next Flick */
			flink_va = flink;
			if (apsh_read_64_va(apsh, flink_va, psl->pid, (uint64_t *)&flink) == APSH_F)
				break;

		} while (flink != link_header);
	}
	retcode = APSH_S;

error_exit:
	return retcode;
}

status_t
windows_dlllist(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid, json_object **json_obj,
	bool printout, bool monitor_mode)
{
	status_t retcode = APSH_F;
	windows_processes_list_t *psl = NULL, *psl_allocation = NULL;
	GHashTable *dll_list;
	bool should_init = false;

	if (monitor_mode == true) {
		dll_list = (GHashTable *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_DLL_LIST);

		if (dll_list == NULL) {
			should_init = true;
			dll_list = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
			if (dll_list == NULL)
				return APSH_F;

			if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_DLL_LIST, dll_list, g_hash_table_destroy)) {
				g_hash_table_destroy(dll_list);
				return APSH_F;
			}
		}
	}

	if (process_list == NULL) {
		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F)
			goto error_exit;
		psl = psl_allocation;
	} else {
		psl = process_list;
	}

	addr_t ImageFileName_offset;
	addr_t Peb_offset;
	addr_t Ldr_offset;
	addr_t InLoadOrderModuleList_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
							 &ImageFileName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Peb", &Peb_offset) ==
	    APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB", "Ldr", &Ldr_offset) ==
	    APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB_LDR_DATA",
							 "InLoadOrderModuleList",
							 &InLoadOrderModuleList_offset) == APSH_F)
		goto error_exit;

	addr_t InLoadOrderLinks_offset;
	addr_t BaseDllName_offset;
	addr_t FullDllName_offset;
	addr_t DllBase_offset;
	addr_t SizeOfImage_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY",
							 "InLoadOrderLinks",
							 &InLoadOrderLinks_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "BaseDllName", &BaseDllName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "FullDllName", &FullDllName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY", "DllBase",
							 &DllBase_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "SizeOfImage", &SizeOfImage_offset) == APSH_F)
		goto error_exit;

	/* print out titles */

	/* create new jsonout object */
	char *titles[] = {"PID", "Process", "Base", "Size", "Name", "Path", "\0"};
	json_object *jout = jsonout_new("dlllist", titles);

	/* walk the process list */
	while (psl) {
		addr_t current_process = psl->address;

		/* print out the process name */
		if (pid == 0 || pid == psl->pid) {
			char *procname =
				apsh_read_str_va(apsh, current_process + ImageFileName_offset, 0);

			if (!procname)
				goto next_process;

			/* Process threadlist */

			/* get _EPROCESS Peb offset */
			addr_t Peb_va;
			uint64_t Peb = 0;

			Peb_va = current_process + Peb_offset;
			if (apsh_read_64_va(apsh, Peb_va, 0, &Peb) == APSH_F) {
				free(procname);
				goto next_process;
			}

			/* get _PEB Ldr */
			addr_t Ldr_va;
			uint64_t Ldr = 0;

			Ldr_va = Peb + Ldr_offset;

			addr_t paddr;

			if (apsh_translate_uv2p(apsh, Peb, psl->pid, &paddr) == APSH_F) {
				free(procname);
				goto next_process;
			}

			if (apsh_read_64_va(apsh, Ldr_va, psl->pid, &Ldr) == APSH_F) {
				free(procname);
				goto next_process;
			}

			/* get  _PEB_LDR_DATA InLoadOrderModuleList */
			addr_t InLoadOrderModuleList_va;

			struct _LIST_ENTRY {
				addr_t flink;
				addr_t blink;
			} InLoadOrderModuleList;

			InLoadOrderModuleList_va = Ldr + InLoadOrderModuleList_offset;

			if (apsh_read_va(apsh, InLoadOrderModuleList_va, psl->pid,
					 sizeof(struct _LIST_ENTRY), &InLoadOrderModuleList,
					 NULL) == APSH_F) {
				free(procname);
				goto next_process;
			}

			/* get  _LDR_DATA_TABLE_ENTRY BaseDllName, FullDllName, DllBase, SizeOfImage
			 */
			addr_t BaseDllName_va;
			addr_t FullDllName_va;
			addr_t DllBase_va, DllBase;
			addr_t SizeOfImage_va;
			uint32_t SizeOfImage;

			/* get Flink of InLoadOrderModuleList */
			addr_t flink_va, flink;

			flink = InLoadOrderModuleList.flink;

			addr_t link_header = flink;
			uint32_t count = 0;

			unicode_string_t *us = NULL;
			unicode_string_t out = {0};

			unicode_string_t *us1 = NULL;
			unicode_string_t out1 = {0};

			do {
				addr_t ldr_data_va;

				ldr_data_va = flink - InLoadOrderLinks_offset;

				/* get DllBase */
				DllBase_va = ldr_data_va + DllBase_offset;
				apsh_read_64_va(apsh, DllBase_va, psl->pid, &DllBase);

				if (DllBase == 0)
					break;

				/* get SizeOfImage */
				SizeOfImage_va = ldr_data_va + SizeOfImage_offset;
				apsh_read_32_va(apsh, SizeOfImage_va, psl->pid, &SizeOfImage);

				/* print out thread info */
				json_object *jdata = jsonout_data_element_new(jout);

				jsonout_data_element_value_sprintf(jdata, "%d", psl->pid);
				jsonout_data_element_value_add(jdata, procname);
				jsonout_data_element_value_sprintf(jdata, "0x%lx", DllBase);
				jsonout_data_element_value_sprintf(jdata, "0x%x", SizeOfImage);

				/* get BaseDllName */
				BaseDllName_va = ldr_data_va + BaseDllName_offset;

				us = apsh_read_unicode_str_va(apsh, BaseDllName_va, psl->pid);
				jsonout_data_element_unicode_add(jdata, us);

				/* get FullDllName */
				FullDllName_va = ldr_data_va + FullDllName_offset;

				us1 = apsh_read_unicode_str_va(apsh, FullDllName_va, psl->pid);
				jsonout_data_element_unicode_add(jdata, us1);

				if (monitor_mode) {
					unicode_string_t out = {0};

					if (us1 &&
					    apsh_convert_str_encoding(us1, &out, "UTF-8") ==
						    APSH_S &&
					    us &&
					    apsh_convert_str_encoding(us, &out1, "UTF-8") ==
						    APSH_S) {
						if (out.contents && out1.contents) {
							char *pid_str =
								g_strdup_printf("%d", psl->pid);
							gchar *uniq_dll =
								g_strconcat(pid_str, psl->name,
									    out.contents, NULL);

							if (should_init) {
								char *dll_dup = g_strndup(
									uniq_dll,
									strlen(uniq_dll) + 1);

								g_hash_table_insert(
									dll_list, dll_dup, dll_dup);
							} else if (g_hash_table_lookup(dll_list,
										       uniq_dll) ==
								   NULL) {
								char time_buff[MAX_TIME_STAMP_LEN] = {0};
								if (strcmp(out1.contents,
									   "WS2_32.dll") == 0 &&
								    strcmp(psl->name,
									   "svchost.exe") != 0) {
									DOCA_DLOG_INFO(RED
									       "\n[%s] [Alert ID: 120, Socket Library loaded.] --DLL LOADED-- Process %s (%d) loaded: %s" RESET
									       "\n",
									       get_time_stamp(time_buff),
									       psl->name, psl->pid,
									       out.contents);
								} else {
									DOCA_DLOG_INFO("[%s] " GRN
									       "--DLL LOADED--" RESET
									       " Process " MAG
									       "%s (%d)" RESET
									       " loaded: " YEL
									       "%s" RESET,
									       get_time_stamp(time_buff),
									       psl->name, psl->pid,
									       out.contents);
								}
								char *dll_dup = g_strndup(
									uniq_dll,
									strlen(uniq_dll) + 1);
								g_hash_table_insert(
									dll_list, dll_dup, dll_dup);
							}

							g_free(pid_str);
							g_free(uniq_dll);
						}
						free(out.contents);
					}
				}

				if (us1)
					apsh_free_unicode_str(us1);

				if (us)
					apsh_free_unicode_str(us);

				/* next Flick */
				flink_va = flink;
				if (APSH_F ==
				    apsh_read_64_va(apsh, flink_va, psl->pid, (uint64_t *)&flink)) {
					break;
				}

			} while (flink != link_header);
			if (procname != NULL)
				free(procname);
		}

next_process:
		if (pid == psl->pid)
			break;

		psl = psl->next;
	}

	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	retcode = APSH_S;

error_exit:

	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	return retcode;
}

status_t windows_get_dlls(apsh_instance_t apsh, windows_processes_list_t *process_list, windows_libs_t **libs_list)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_LIBS_LIMIT)));

	status_t retcode = APSH_F;
	windows_processes_list_t *psl = NULL;

	if (process_list == NULL)
		return APSH_F;
	psl = process_list;

	addr_t ImageFileName_offset;
	addr_t Peb_offset;
	addr_t Ldr_offset;
	addr_t InLoadOrderModuleList_offset;

	windows_libs_t *libs_prev = NULL;
	windows_libs_t *libs_head = NULL;

	unicode_string_t *us = NULL;
	unicode_string_t out = {0};
	char *dll_base_name = NULL;

	unicode_string_t *us1 = NULL;
	unicode_string_t out1 = {0};
	char *dll_full_name = NULL;

	char *procname = NULL;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
							 &ImageFileName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Peb", &Peb_offset) ==
	    APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB", "Ldr", &Ldr_offset) ==
	    APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB_LDR_DATA",
							 "InLoadOrderModuleList",
							 &InLoadOrderModuleList_offset) == APSH_F)
		goto error_exit;

	addr_t InLoadOrderLinks_offset;
	addr_t BaseDllName_offset;
	addr_t FullDllName_offset;
	addr_t DllBase_offset;
	addr_t SizeOfImage_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY",
							 "InLoadOrderLinks",
							 &InLoadOrderLinks_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "BaseDllName", &BaseDllName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "FullDllName", &FullDllName_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY", "DllBase",
							 &DllBase_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "SizeOfImage", &SizeOfImage_offset) == APSH_F)
		goto error_exit;

	uint64_t count = 0;

	if (psl) {
		addr_t current_process = psl->address;

		/* get _EPROCESS Peb offset */
		addr_t Peb_va;
		uint64_t Peb = 0;

		Peb_va = current_process + Peb_offset;
		if (apsh_read_64_va(apsh, Peb_va, 0, &Peb) == APSH_F)
			return APSH_F;

		/* get _PEB Ldr */
		addr_t Ldr_va;
		uint64_t Ldr = 0;
		addr_t paddr;

		Ldr_va = Peb + Ldr_offset;

		if (apsh_translate_uv2p(apsh, Peb, psl->pid, &paddr) == APSH_F)
			return APSH_F;

		if (apsh_read_64_va(apsh, Ldr_va, psl->pid, &Ldr) == APSH_F)
			return APSH_F;

		/* get  _PEB_LDR_DATA InLoadOrderModuleList */
		addr_t InLoadOrderModuleList_va;

		struct _LIST_ENTRY {
			addr_t flink;
			addr_t blink;
		} InLoadOrderModuleList;

		InLoadOrderModuleList_va = Ldr + InLoadOrderModuleList_offset;

		if (apsh_read_va(apsh, InLoadOrderModuleList_va, psl->pid,
				 sizeof(struct _LIST_ENTRY), &InLoadOrderModuleList,
				 NULL) == APSH_F)
			return APSH_F;

		/* get  _LDR_DATA_TABLE_ENTRY BaseDllName, FullDllName, DllBase, SizeOfImage */
		addr_t BaseDllName_va;
		addr_t FullDllName_va;
		addr_t DllBase_va, DllBase;
		addr_t SizeOfImage_va;
		uint32_t SizeOfImage;

		// get Flink of InLoadOrderModuleList
		addr_t flink_va, flink;

		flink = InLoadOrderModuleList.flink;

		addr_t link_header = flink;
		uint32_t count = 0;

		do {
			addr_t ldr_data_va;

			ldr_data_va = flink - InLoadOrderLinks_offset;

			/* get DllBase */
			DllBase_va = ldr_data_va + DllBase_offset;
			apsh_read_64_va(apsh, DllBase_va, psl->pid, &DllBase);

			if (DllBase == 0)
				break;

			/* get SizeOfImage */
			SizeOfImage_va = ldr_data_va + SizeOfImage_offset;
			apsh_read_32_va(apsh, SizeOfImage_va, psl->pid, &SizeOfImage);

			/* get BaseDllName */
			dll_base_name = NULL;
			BaseDllName_va = ldr_data_va + BaseDllName_offset;
			us = apsh_read_unicode_str_va(apsh, BaseDllName_va, psl->pid);

			if (us != NULL) {
				if (apsh_convert_str_encoding(us, &out, "UTF-8") == APSH_S && out.contents) {
					dll_base_name = strdup(out.contents);
					free(out.contents);
				}
				apsh_free_unicode_str(us);
			}

			/* get FullDllName */
			dll_full_name = NULL;
			FullDllName_va = ldr_data_va + FullDllName_offset;
			us1 = apsh_read_unicode_str_va(apsh, FullDllName_va, psl->pid);

			if (us1 != NULL) {
				if (apsh_convert_str_encoding(us1, &out1, "UTF-8") == APSH_S && out1.contents) {
					dll_full_name = strdup(out1.contents);
					free(out1.contents);
				}
				apsh_free_unicode_str(us1);
			}

			procname = apsh_read_str_va(apsh, current_process + ImageFileName_offset, 0);

			if (dll_base_name != NULL && dll_full_name != NULL && procname != NULL) {
				// allocate
				windows_libs_t *libs = malloc(sizeof(windows_libs_t));

				libs->data = g_hash_table_new(NULL, NULL);
				if (libs->data == NULL) {
					free(libs);
					goto error_exit;
				}

				libs->pid = psl->pid;
				libs->image_name = procname;
				libs->dll_base_name = dll_base_name;
				libs->dll_full_name = dll_full_name;
				libs->image_size = SizeOfImage;
				libs->next = NULL;

				if (libs_head == NULL)
					libs_head = libs;

				if (libs_prev)
					libs_prev->next = libs;

				// prevent double free in error flow
				dll_base_name = NULL;
				dll_full_name = NULL;
				procname = NULL;

				int result;

				result = g_hash_table_insert(
					libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_PID),
					GINT_TO_POINTER(libs->pid));
				char *image_name_cpy = g_strndup(libs->image_name, strlen(libs->image_name) + 1);

				result &= g_hash_table_insert(
					libs->data,
					GINT_TO_POINTER(DOCA_APSH_LIB_COMM),
					image_name_cpy);

				char *base_dll_copy = g_strndup(libs->dll_base_name, strlen(libs->dll_base_name) + 1);

				result &= g_hash_table_insert(
					libs->data,
					GINT_TO_POINTER(DOCA_APSH_LIB_WINDOWS_FULL_DLL_NAME),
					base_dll_copy);

				char *full_dll_copy = g_strndup(libs->dll_full_name, strlen(libs->dll_full_name) + 1);

				result &= g_hash_table_insert(
					libs->data,
					GINT_TO_POINTER(DOCA_APSH_LIB_LIBRARY_PATH),
					full_dll_copy);

				result &= g_hash_table_insert(
					libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_WINDOWS_SIZE_OFIMAGE),
					GUINT_TO_POINTER(SizeOfImage));

				if (!result)
					goto error_exit;

				// set current to prev
				libs_prev = libs;
			} else {
				if (dll_base_name)
					free(dll_base_name);
				if (dll_full_name)
					free(dll_full_name);
				if (procname)
					free(procname);
			}
			/* next Flick */
			flink_va = flink;
			if (apsh_read_64_va(apsh, flink_va, psl->pid, (uint64_t *)&flink) == APSH_F)
				break;

		} while (flink != link_header && ++count < limit);
	}
	*libs_list = libs_head;
	return APSH_S;

error_exit:
	if (us != NULL)
		apsh_free_unicode_str(us);
	if (us1 != NULL)
		apsh_free_unicode_str(us1);
	if (procname != NULL)
		free(procname);

	if (libs_head != NULL)
		windows_free_libs_list(libs_head);
	*libs_list = NULL;
	return retcode;
}

void
windows_free_libs_list(windows_libs_t *libs_list)
{
	char *image_name;
	char *base_dll_copy;
	char *full_dll_copy;

	windows_libs_t *libs = libs_list;

	while (libs) {
		windows_libs_t *_next = libs->next;

		if (libs->data != NULL) {

			image_name = g_hash_table_lookup(libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_LIBRARY_PATH));
			base_dll_copy = g_hash_table_lookup(libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_COMM));
			full_dll_copy = g_hash_table_lookup(libs->data, GINT_TO_POINTER(DOCA_APSH_LIB_WINDOWS_FULL_DLL_NAME));


			if (image_name != NULL)
				g_free(image_name);

			if (base_dll_copy != NULL)
				g_free(base_dll_copy);

			if (full_dll_copy != NULL)
				g_free(full_dll_copy);

			g_hash_table_destroy(libs->data);
		}

		free(libs->dll_base_name);
		free(libs->dll_full_name);
		free(libs->image_name);
		free(libs);
		libs = _next;
	}
}
