/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "envars.h"
#include "doca_apsh_attr.h"
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::Windows::ENVARS);

/* retrive a pair of variable and value from envars block
 * internally replace separator L'=' with L'\0'
 * return total length of variable and value pair
 */
static uint64_t
get_envar(char *envars, uint64_t envars_length, char **var, char **val)
{
	char *p, *pval;

	p = envars;
	pval = NULL;
	*var = NULL;
	*val = NULL;

	while (envars_length) {
		if (*p == '\0') {
			*var = envars;
			*val = pval;
			p++;
			return (p - envars);
		}

		if (*p == '=') {
			*p = '\0';
			pval = p + 1;
		}

		p++;
		envars_length--;
	}

	return (p - envars);
}

static status_t decode_unicode_utf16(char **pstr, int len)
{
	if (pstr == NULL || (*pstr) == NULL || len <= 0)
		return APSH_F;

	char *str = *pstr;
	char *new_str = calloc(1, len/2);

	if (new_str == NULL)
		return APSH_F;

	for (int i = 0 ; i < (len/2) ; ++i)
		new_str[i] = str[2*i];

	free(str);
	*pstr = new_str;

	return APSH_S;
}

status_t
windows_envars(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid, windows_envars_t **results_list)
{

	status_t retval = APSH_F;
	windows_processes_list_t *psl, *psl_allocation = NULL;
	windows_envars_t *results_list_allocated = NULL, *results_list_allocated_head = NULL;
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_WINDOWS_ENVARS_LIMIT)));

	if (process_list == NULL) {
		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F)
			return APSH_F;
		psl = psl_allocation;
	} else {
		psl = process_list;
	}

	windows_eprocess_struct_t *_eps = NULL;

	if (windows_get_eprocess(apsh, &_eps) == APSH_F)
		return APSH_F;

	char *penvars = calloc(1, sizeof(char) * limit);
	char *penvars_head = penvars;

	if (penvars == NULL)
		goto error_exit;

	/* get _PEB ProcessParameters offset */
	addr_t ProcessParameters_offset;
	addr_t Environment_offset;
	addr_t EnvironmentSize_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB", "ProcessParameters",
							 &ProcessParameters_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_RTL_USER_PROCESS_PARAMETERS",
							 "Environment",
							 &Environment_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_RTL_USER_PROCESS_PARAMETERS",
							 "EnvironmentSize",
							 &EnvironmentSize_offset) == APSH_F)
		goto error_exit;

	/* walk through the process list */
	while (psl) {
		char *procname = NULL;

		if (pid == 0 || pid == psl->pid) {
			apsh_pid_t _pid = psl->pid;
			addr_t process_address = psl->address;

			/* get procname */
			procname = apsh_read_str_va(apsh,
						    process_address + _eps->ImageFileName_offset, 0);

			if (!procname) {
				DOCA_DLOG_ERR("Failed to find procname");
				goto error_exit;
			}

			/* get _EPROCESS.Peb */
			addr_t Peb_va;
			uint64_t Peb = 0;

			Peb_va = process_address + _eps->Peb_offset;
			if (apsh_read_64_va(apsh, Peb_va, _pid, &Peb) == APSH_F)
				goto next_process;

			/* get _PEB.ProcessParameters */
			addr_t ProcessParameters_va;
			uint64_t ProcessParameters = 0;

			ProcessParameters_va = Peb + ProcessParameters_offset;
			if (apsh_read_64_va(apsh, ProcessParameters_va, _pid, &ProcessParameters) == APSH_F)
				goto next_process;

			/* get _RTL_USER_PROCESS_PARAMETERS.Environment */
			addr_t Environment_va;
			uint64_t Environment = 0;

			Environment_va = ProcessParameters + Environment_offset;
			if (apsh_read_64_va(apsh, Environment_va, _pid, &Environment) == APSH_F)
				goto next_process;

			// get _RTL_USER_PROCESS_PARAMETERS.EnvironmentSize
			addr_t EnvironmentSize_va;
			uint64_t EnvironmentSize = 0;

			EnvironmentSize_va = ProcessParameters + EnvironmentSize_offset;
			if (apsh_read_64_va(apsh, EnvironmentSize_va, _pid, &EnvironmentSize) == APSH_F)
				goto next_process;

			if (EnvironmentSize > limit) {
				DOCA_DLOG_ERR("EnvironmentSize too large!");
				goto next_process;
			}

			/* read unicode string */
			if (apsh_read_va(apsh, Environment, _pid, EnvironmentSize, penvars, NULL) == APSH_F) {
				DOCA_DLOG_ERR("failed to read Envars");
				goto next_process;
			}

			if (decode_unicode_utf16(&penvars, EnvironmentSize) == APSH_F) {
				DOCA_DLOG_ERR("failed to allocate memory");
				goto next_process;
			}

			penvars_head = penvars;

			/* handle envars */
			uint64_t length = 0, r_length = EnvironmentSize/2;

			while (r_length > 0) {
				char *var, *val;

				length = get_envar(penvars, r_length, &var, &val);
				if (length <= 1)
					break;

				if (results_list_allocated == NULL) {
					results_list_allocated = calloc(1, sizeof(windows_envars_t));
					results_list_allocated_head = results_list_allocated;
				} else {
					results_list_allocated->next = calloc(1, sizeof(windows_envars_t));
					results_list_allocated = results_list_allocated->next;
				}

				if (results_list_allocated == NULL) {
					free(procname);
					goto error_exit;
				}

				results_list_allocated->next = NULL;
				results_list_allocated->pid = _pid;
				results_list_allocated->procname = g_strdup(procname);
				results_list_allocated->environment = Environment;
				results_list_allocated->variable = g_strdup(var);
				results_list_allocated->value = g_strdup(val);
				results_list_allocated->data = g_hash_table_new(NULL, NULL);

				if (results_list_allocated->variable == NULL || results_list_allocated->value == NULL
				|| results_list_allocated->procname == NULL || results_list_allocated->data == NULL) {
					free(procname);
					goto error_exit;
				}

				char *variable_copy = g_strdup(var);
				char *value_copy = g_strdup(val);
				char *procname_copy = g_strdup(procname);

				if (variable_copy == NULL || value_copy == NULL || procname_copy == NULL) {
					free(procname);
					g_free(variable_copy);
					g_free(value_copy);
					g_free(procname_copy);
					goto error_exit;
				}

				g_hash_table_insert(results_list_allocated->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_PID),
				GINT_TO_POINTER(results_list_allocated->pid));
				g_hash_table_insert(results_list_allocated->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_COMM),
				procname_copy);
				g_hash_table_insert(results_list_allocated->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_WINDOWS_BLOCK),
				GINT_TO_POINTER(results_list_allocated->environment));
				g_hash_table_insert(results_list_allocated->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_VARIABLE),
				variable_copy);
				g_hash_table_insert(results_list_allocated->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_VALUE),
				value_copy);

				penvars += length;
				r_length -= length;
			}
		}

next_process:
		if (procname != NULL)
			free(procname);

		if (pid == psl->pid)
			break;

		/* next process */
		psl = psl->next;
	}

	*results_list = results_list_allocated_head;

	retval = APSH_S;

error_exit:

	if (penvars_head != NULL)
		free(penvars_head);

	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	if (retval != APSH_S)
		windows_free_envars_list(results_list_allocated_head);

	return retval;
}


void windows_free_envars_list(windows_envars_t *envars_list)
{
	while (envars_list) {
		windows_envars_t *_next = envars_list->next;

		if (envars_list->data) {
			char *variable = g_hash_table_lookup(envars_list->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_VARIABLE));
			char *value = g_hash_table_lookup(envars_list->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_VALUE));
			char *procname = g_hash_table_lookup(envars_list->data, GINT_TO_POINTER(DOCA_APSH_ENVARS_COMM));

			if (variable != NULL)
				g_free(variable);

			if (value != NULL)
				g_free(value);

			if (procname != NULL)
				g_free(procname);

			g_hash_table_destroy(envars_list->data);
		}

		g_free(envars_list->variable);
		g_free(envars_list->value);
		g_free(envars_list->procname);
		free(envars_list);
		envars_list = _next;
	}
}
