/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <getopt.h>
#include <gmodule.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <inttypes.h>
#include "openssl/sha.h"
#include "doca_log.h"
#include <zip.h>
#include "private/common_defines.h"

DOCA_LOG_REGISTER(APSH::Windows::HASHBUILD);

#define PAGE_SIZE 0x1000
#define MAX_DIR_LEN 2048
#define ERROR(x) {DOCA_DLOG_ERR("error in %d\n", __LINE__); return x;}
#define ERROR_GO(x) {DOCA_DLOG_ERR("error in %d\n", __LINE__); goto x;}

enum DOCA_APSH_IMAGE_SECTION_HEADER {
	DOCA_APSH_IMAGE_SECTION_HEADER_TYPE_NO_PAD = 0x00000008,
	DOCA_APSH_IMAGE_SECTION_HEADER_CNT_CODE = 0x00000020,
	DOCA_APSH_IMAGE_SECTION_HEADER_CNT_INITIALIZED_DATA = 0x00000040,
	DOCA_APSH_IMAGE_SECTION_HEADER_CNT_UNINITIALIZED_DATA = 0x00000080,
	DOCA_APSH_IMAGE_SECTION_HEADER_GPREL = 0x00008000,
	DOCA_APSH_IMAGE_SECTION_HEADER_MEM_DISCARDABLE = 0x02000000,
	DOCA_APSH_IMAGE_SECTION_HEADER_MEM_NOT_PAGED = 0x08000000,
	DOCA_APSH_IMAGE_SECTION_HEADER_MEM_SHARED = 0x10000000,
	DOCA_APSH_IMAGE_SECTION_HEADER_MEM_EXECUTE = 0x20000000,
	DOCA_APSH_IMAGE_SECTION_HEADER_MEM_READ = 0x40000000,
	DOCA_APSH_IMAGE_SECTION_HEADER_MEM_WRITE = 0x80000000,
};

enum PE_TYPES {
	PE32	  = 0x10b, /** 32bits  */
	PE32_PLUS = 0x20b  /** 64 bits */
};

enum FILE_TYPE {
	SKIP		= 0,
	PE_TYPE		= 1,
	RAW_TYPE	= 2,
};

int string_sort_function(gconstpointer a, gconstpointer b)
{
	const char *str_a = (const char *)a;
	const char *str_b = (const char *)b;

	return strcmp(str_a, str_b);
}

int int_sort_function(gconstpointer a, gconstpointer b)
{
	int int_a = GPOINTER_TO_INT(a);
	int int_b = GPOINTER_TO_INT(b);

	return int_a - int_b;
}

enum FILE_TYPE
is_file_relevant(const char *file_name)
{
	struct file_extentions {
		const char *ext;
		enum FILE_TYPE type;
	};
	static const struct file_extentions file_ext_arr[] = {
		{".dll", PE_TYPE, },
		{".exe", PE_TYPE, },
		{".drv", PE_TYPE, },
		{".cpl", PE_TYPE, },
		{".ocx", PE_TYPE, },
		{".mui", PE_TYPE, },
		{".nls", RAW_TYPE, },
	};
	size_t file_len = strlen(file_name);
	int i;

	for (i = 0; i < ARRAY_SIZE(file_ext_arr); ++i) {
		size_t e_len = strlen(file_ext_arr[i].ext);

		if (file_len > e_len && !strncmp(file_ext_arr[i].ext, file_name + file_len - e_len, e_len))
			return file_ext_arr[i].type;
	}

	return SKIP;
}

int
show_dir_content(const char *path, GHashTable **path_table_p)
{
	char d_path[MAX_DIR_LEN], init_path[MAX_DIR_LEN];
	struct dirent *dir;
	DIR *d = opendir(path);

	if (d == NULL)
		ERROR(-1);
	sprintf(init_path, path[strlen(path)-1] != '/' ? "%s/" : "%s", path);

	if (*path_table_p == NULL)
		*path_table_p = g_hash_table_new_full(g_str_hash, g_str_equal, free, (GDestroyNotify)g_hash_table_destroy);
	GHashTable *path_table = *path_table_p;

	if (path_table == NULL)
		ERROR_GO(close_file);

	while ((dir = readdir(d)) != NULL) {
		sprintf(d_path, "%s%s", init_path, dir->d_name);
		if (dir->d_type == DT_REG) {
			if (is_file_relevant(dir->d_name) == SKIP)
				continue;
			GHashTable *name_table = g_hash_table_lookup(path_table, dir->d_name);

			if (name_table == NULL) {
				name_table = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
				if (name_table == NULL)
					ERROR_GO(free_table);
				char *d_name_dup = strdup(dir->d_name);

				if (d_name_dup == NULL)
					ERROR_GO(free_table);
				g_hash_table_insert(path_table, d_name_dup, name_table);
			}
			if (g_hash_table_lookup(name_table, d_path) == NULL) {
				char *d_path_dup = strdup(d_path);

				if (d_path_dup == NULL)
					ERROR_GO(free_table);
				g_hash_table_insert(name_table, d_path_dup, "full");
			}
		} else if (dir->d_type == DT_DIR && strcmp(dir->d_name, ".") != 0 && strcmp(dir->d_name, "..") != 0) {
			int ret = show_dir_content(d_path, path_table_p);

			if (ret) {
				closedir(d);
				ERROR(ret);
			}
		}
	}
	closedir(d);
	return 0;
free_table:
	g_hash_table_destroy(path_table);
close_file:
	closedir(d);
	ERROR(-1);
}

struct pe_section_t {
	uint8_t name[8];
	uint32_t virtual_size;
	uint32_t virtual_address;
	uint32_t size;
	uint32_t offset;
	uint8_t other[12];
	uint32_t characteristics;
};

struct pe_t {
	uint32_t addr_exeheader;
	uint8_t *nt_header;
	uint32_t magic;
	uint8_t *file_header;
	uint32_t num_sections;
	uint32_t optional_header_size;
	uint32_t header_size;
	uint8_t *directories;
	uint32_t relocs_vaddr;
	uint32_t relocs_size;
	uint32_t iat_vaddr;
	uint32_t iat_size;
	struct pe_section_t *sections;
};

struct file_pe_t {
	char *name;
	uint8_t *bytes;
	GHashTable *alterations;
	GHashTable *pages;
	uint8_t *perms;
	uint8_t *virtual;
	int virtual_size;
	int perms_size;
	struct pe_t pe;
};

struct file_pe_t
create_file_pe()
{
	return (struct file_pe_t){
		NULL, /*path*/
		NULL, /*name*/
		NULL, /*bytes*/
		NULL, /*alterations*/
		NULL, /*pages*/
		NULL, /*perms*/
		NULL, /*virtual*/
		0, /*virtual_size*/
		0, /*perms_size*/
		0 /*pe_binary*/
	};
}

void
free_file_pe(struct file_pe_t *file)
{
	if (file->perms != NULL)
		free(file->perms);
	if (file->bytes != NULL)
		free(file->bytes);
	if (file->virtual != NULL)
		free(file->virtual);
	if (file->alterations != NULL)
		g_hash_table_destroy(file->alterations);
	if (file->pages != NULL)
		g_hash_table_destroy(file->pages);
}

int
pe_parse(struct file_pe_t *file)
{
	if (file->bytes[0] != 'M' || file->bytes[1] != 'Z')
		return -1;
	file->pe.addr_exeheader = *(uint32_t *)(file->bytes + 0x3C);
	file->pe.nt_header = file->bytes + file->pe.addr_exeheader;
	file->pe.magic = *(uint16_t *)(file->pe.nt_header + 0x18);
	file->pe.file_header = file->bytes + file->pe.addr_exeheader + 0x4;
	file->pe.num_sections = *(uint16_t *)(file->pe.file_header + 0x2);
	file->pe.optional_header_size = *(uint16_t *)(file->pe.file_header + 0x10);
	file->pe.header_size = *(uint32_t *)(file->pe.nt_header + 0x18 + 0x3C);
	file->pe.sections = (struct pe_section_t *)(file->bytes + file->pe.addr_exeheader + 0x18 + file->pe.optional_header_size);
	file->pe.directories = file->bytes + file->pe.addr_exeheader + (file->pe.magic == PE32 ? 0x78 : 0x88);
	file->pe.relocs_vaddr = *(uint32_t *)(file->pe.directories + 0x28);
	file->pe.relocs_size = *(uint32_t *)(file->pe.directories + 0x28 + 0x4);
	file->pe.iat_vaddr = *(uint32_t *)(file->pe.directories + 0x60);
	file->pe.iat_size = *(uint32_t *)(file->pe.directories + 0x60 + 0x4);
	return 0;
}

int
section_flag(const struct pe_section_t *section)
{
	int flag = 0;

	if (section->characteristics & DOCA_APSH_IMAGE_SECTION_HEADER_MEM_EXECUTE)
		flag = 2;
	if (section->characteristics & DOCA_APSH_IMAGE_SECTION_HEADER_MEM_WRITE)
		flag = 1;
	if (!strcmp(section->name, ".mrdata"))
		flag = 3;
	return flag;
}

int
expand_section(const struct file_pe_t *file, const struct pe_section_t *section, uint8_t **expanded, int *e_size)
{
	uint32_t size = section->virtual_size;

	if (size % PAGE_SIZE)
		size += (PAGE_SIZE - (size % PAGE_SIZE));
	*expanded = (uint8_t *)calloc(size, sizeof(uint8_t));
	if (*expanded == NULL)
		ERROR(-1);
	uint32_t cp_size = MIN(section->virtual_size, section->size);

	memcpy(*expanded, file->bytes + section->offset, cp_size);
	*e_size = size;
	return 0;
}

int
expand_header(struct file_pe_t *file)
{
	int size = file->pe.header_size;

	file->virtual_size = size;
	if (size % PAGE_SIZE != 0)
		file->virtual_size = size + (PAGE_SIZE - (size % PAGE_SIZE));
	file->virtual = (uint8_t *)calloc(file->virtual_size, sizeof(uint8_t));
	if (file->virtual == NULL)
		ERROR(-1);
	memcpy(file->virtual, file->bytes, size);
	return 0;
}

int
append(struct file_pe_t *file, uint8_t *expanded, int e_size, int vaddr)
{
	uint8_t *virtual = (uint8_t *)calloc(vaddr+e_size, sizeof(uint8_t));

	if (virtual == NULL)
		ERROR(-1);
	int size = vaddr < file->virtual_size ? vaddr : file->virtual_size;

	memcpy(virtual, file->virtual, size);
	free(file->virtual);
	file->virtual = virtual;
	file->virtual_size = vaddr+e_size;
	memcpy(file->virtual + vaddr, expanded, e_size);
	return 0;
}

int
expand(struct file_pe_t *file)
{
	int ret = -1;

	if (expand_header(file))
		ERROR(ret);
	int perms_s = file->virtual_size / PAGE_SIZE;
	uint8_t *prev_perms, *perms = (uint8_t *)calloc(perms_s, sizeof(uint8_t));

	if (perms == NULL)
		ERROR(ret);

	uint8_t prev_perm = 0;
	uint8_t *expanded = NULL;
	size_t i;

	for (i = 0; i < file->pe.num_sections; ++i) {
		struct pe_section_t *section = file->pe.sections + i;
		int vaddr = section->virtual_address;
		int size = (vaddr - file->virtual_size) / PAGE_SIZE;

		perms = realloc(prev_perms = perms, (perms_s + size) * sizeof(uint8_t));
		if (perms == NULL)
			ERROR_GO(error);
		if (memset(perms+perms_s, prev_perm, size) == NULL)
			ERROR_GO(error);
		perms_s += size;
		int e_size;

		if (expand_section(file, section, &expanded, &e_size))
			ERROR_GO(error);
		if (append(file, expanded, e_size, vaddr))
			ERROR_GO(error);
		free(expanded);
		expanded = NULL;
		e_size = e_size / PAGE_SIZE;
		perms = realloc(prev_perms = perms, (perms_s + e_size) * sizeof(uint8_t));
		if (perms == NULL)
			ERROR_GO(error);
		if (memset(perms + perms_s, section_flag(section), e_size) == NULL)
			ERROR_GO(error);
		perms_s += e_size;
		prev_perm = section_flag(section);
	}
	file->perms = perms;
	file->perms_size = perms_s;
	return 0;
error:
	if (perms == NULL)
		free(prev_perms);
	else
		free(perms);
	if (expanded)
		free(expanded);
	return ret;
}

uint8_t *
file_bytes(uint8_t *path)
{
	FILE *fileptr = fopen(path, "rb");

	if (fileptr == NULL)
		ERROR(NULL);
	if (fseek(fileptr, 0, SEEK_END) != 0) {
		fclose(fileptr);
		ERROR(NULL);
	}

	long filelen = ftell(fileptr);

	if (filelen < 0) {
		fclose(fileptr);
		ERROR(NULL);
	}

	rewind(fileptr);
	uint8_t *bytes = (uint8_t *)calloc(filelen, sizeof(uint8_t));

	if (bytes == NULL) {
		fclose(fileptr);
		ERROR(NULL);
	}
	fread(bytes, filelen, 1, fileptr);
	fclose(fileptr);
	return bytes;
}

void
insert_zero(GHashTable *zeroes, uint32_t addr)
{
	g_hash_table_insert(zeroes, GINT_TO_POINTER(addr), "full");
}

int
parse_relocations(GHashTable *zeroes, uint8_t *page_relocs, uint32_t size)
{
	uint32_t offset = 0;
	int64_t last = -1;

	while (offset < size) {
		uint16_t entry = *(uint16_t *)(page_relocs + offset);
		uint32_t reloc_type = (entry & 0xF000) >> 12;
		uint32_t reloc_addr = entry & 0x0FFF;

		if (reloc_type == 3 && reloc_addr > last) {
			insert_zero(zeroes, reloc_addr);
			last = reloc_addr;
		}
		if (reloc_type == 10 && reloc_addr > last && (reloc_addr + 4) > last) {
			insert_zero(zeroes, reloc_addr);
			insert_zero(zeroes, reloc_addr + 4);
			last = reloc_addr + 4;
		}
		offset += 0x2;
	}
	return 0;
}

GHashTable *
look_zeroes(GHashTable *alterations, uint32_t vaddr)
{
	GHashTable *zeroes = g_hash_table_lookup(alterations, GINT_TO_POINTER(vaddr));

	if (zeroes != NULL)
		return zeroes;
	zeroes = g_hash_table_new_full(NULL, NULL, NULL, NULL);
	if (zeroes == NULL)
		ERROR(NULL);

	g_hash_table_insert(alterations, GINT_TO_POINTER(vaddr), zeroes);
	return zeroes;
}

int
relocations(struct file_pe_t *file, GHashTable *alterations)
{
	uint8_t *relocs = file->virtual + file->pe.relocs_vaddr;
	uint32_t offset = 0;

	while (offset < file->pe.relocs_size) {
		uint32_t vaddr = *(uint32_t *)(relocs + offset);
		uint32_t size = *(uint32_t *)(relocs + offset + 0x4);

		if (size == 0)
			break;
		uint8_t *page_relocs = relocs + offset + 0x8;
		GHashTable *zeroes = look_zeroes(alterations, vaddr);

		if (zeroes == NULL)
			ERROR(-1);
		if (parse_relocations(zeroes, page_relocs, size))
			ERROR(-1);
		offset += size;
	}
	return 0;
}

int
parse_iat(GHashTable *alterations, uint32_t base_addr, uint32_t size)
{
	uint32_t offset = base_addr % PAGE_SIZE;
	uint32_t addr = base_addr - offset;

	while (addr + offset < base_addr + size) {
		if (offset == PAGE_SIZE) {
			addr += offset;
			offset = 0;
		}
		GHashTable *zeroes = look_zeroes(alterations, addr);

		if (zeroes == NULL)
			ERROR(-1);
		insert_zero(zeroes, offset);
		offset += 4;
	}
	return 0;
}

int
iat(struct file_pe_t *file, GHashTable *alterations)
{
	if (parse_iat(alterations, file->pe.iat_vaddr, file->pe.iat_size))
		ERROR(-1);
	int i;

	for (i = 0; i < file->pe.num_sections; ++i) {
		struct pe_section_t *section = file->pe.sections + i;

		if (section_flag(section) != 3)
			continue;
		if (parse_iat(alterations, section->virtual_address, section->virtual_size))
			ERROR(-1);
	}
	if (file->pe.magic == PE32) {
		if (parse_iat(alterations, file->pe.addr_exeheader + 0x18 + 28, 4))
			ERROR(-1);
	}
	if (file->pe.magic == PE32_PLUS) {
		if (parse_iat(alterations, file->pe.addr_exeheader + 0x18 + 24, 8))
			ERROR(-1);
	}
	return 0;
}

uint8_t *
look_page(GHashTable *pages, uint32_t vaddr)
{
	uint8_t *page = g_hash_table_lookup(pages, GINT_TO_POINTER(vaddr));

	if (page != NULL)
		return page;
	page = (uint8_t *)calloc(PAGE_SIZE, sizeof(uint8_t));
	if (page == NULL)
		ERROR(NULL);

	g_hash_table_insert(pages, GINT_TO_POINTER(vaddr), page);
	return page;
}

int
parse_pages(struct file_pe_t *file)
{
	file->pages = g_hash_table_new_full(NULL, NULL, NULL, free);
	if (file->pages == NULL)
		ERROR(-1);
	uint32_t vaddr = 0;
	uint32_t unapplied = 0;

	while (vaddr < file->virtual_size) {
		GHashTable *zeroes = g_hash_table_lookup(file->alterations, GINT_TO_POINTER(vaddr));
		uint32_t page_s = 0;
		uint8_t *page = look_page(file->pages, vaddr);

		if (page == NULL)
			ERROR(-1);
		if (zeroes == NULL) {
			memcpy(page, file->virtual + vaddr, PAGE_SIZE);
			vaddr += PAGE_SIZE;
			continue;
		}
		uint32_t offset = 0;

		if (unapplied > 0) {
			offset = unapplied;
			page_s += unapplied;
			insert_zero(zeroes, -(4 - unapplied));
			unapplied = 0;
		}
		GList *zeroes_head = g_hash_table_get_keys(zeroes);
		GList *zeroes_p = g_list_sort(zeroes_head, int_sort_function);
		GList *zeroes_p_const = zeroes_p;

		while (zeroes_p) {
			int32_t zero = GPOINTER_TO_UINT(zeroes_p->data);

			if (zero < 0 || zero < offset) {
				zeroes_p = zeroes_p->next;
				continue;
			}
			memcpy(page+page_s, file->virtual + vaddr + offset, zero - offset);
			page_s += zero - offset;
			if (zero <= PAGE_SIZE - 4) {
				offset = zero + 4;
				page_s += 4;
			} else {
				unapplied = 4 - (PAGE_SIZE - zero);
				offset = PAGE_SIZE;
				page_s += PAGE_SIZE - zero;
			}
			zeroes_p = zeroes_p->next;
		}
		g_list_free(zeroes_p_const);
		if (offset < PAGE_SIZE)
			memcpy(page+page_s, file->virtual + vaddr + offset, PAGE_SIZE - offset);
		vaddr += PAGE_SIZE;
	}
	return 0;
}

int
parse_alterations(struct file_pe_t *file)
{
	file->alterations = g_hash_table_new_full(NULL, NULL, NULL, (GDestroyNotify)g_hash_table_destroy);
	if (file->alterations == NULL)
		ERROR(-1);
	if (relocations(file, file->alterations))
		ERROR(-1);
	if (iat(file, file->alterations))
		ERROR(-1);
	return 0;
}

static void
sha256h_string(uint8_t *string, unsigned int length, unsigned char *hash)
{
	struct timespec ts1, ts2;
	SHA256_CTX sha256;
	unsigned char hash_b[SHA256_DIGEST_LENGTH];

	SHA256_Init(&sha256);
	SHA256_Update(&sha256, string, length);
	SHA256_Final(hash_b, &sha256);
	int i;

	for(i = 0; i < SHA256_DIGEST_LENGTH; i++)
		sprintf(hash + (i * 2), "%02x", hash_b[i]);
}

int
output(struct file_pe_t *file, GHashTable *outputs)
{
	uint32_t offset = 0;
	uint32_t max_offset_digits = floor(log10(file->virtual_size / PAGE_SIZE)) + 1;

	while (offset < file->virtual_size) {
		uint8_t *page = g_hash_table_lookup(file->pages, GINT_TO_POINTER(offset));

		if (page == NULL)
			ERROR(-1);
		unsigned char hash[SHA256_DIGEST_LENGTH*2+1] = {0};

		sha256h_string(page, PAGE_SIZE, hash);
		GHashTable *zeroes = g_hash_table_lookup(file->alterations, GINT_TO_POINTER(offset));
		GList *zeroes_head = zeroes == NULL ? NULL : g_hash_table_get_keys(zeroes);
		GList *zeroes_p = g_list_sort(zeroes_head, int_sort_function);
		GList *zeroes_p_const = zeroes_p;
		uint32_t max_zero_digits = 0;

		if (zeroes_p)
			max_zero_digits = floor(log10(GPOINTER_TO_UINT(g_list_last(zeroes_p)->data))) + 1;
		uint32_t zero_size = g_list_length(zeroes_p) * (max_zero_digits + 1) + 1;
		char *zero_s = (char *)calloc(zero_size, sizeof(char));

		if (zero_s == NULL)
			ERROR(-1);
		uint32_t zero_len = 0;

		while (zeroes_p) {
			uint32_t zero = GPOINTER_TO_UINT(zeroes_p->data);
			int ret = sprintf(zero_s+zero_len, zero_len ? " %d" : "%d", zero);

			if (ret < 0) {
				free(zero_s);
				g_list_free(zeroes_p_const);
				ERROR(ret);
			}
			zero_len += ret;
			if (zero_size <= zero_len) {
				free(zero_s);
				g_list_free(zeroes_p_const);
				ERROR(-1);
			}
			zeroes_p = zeroes_p->next;
		}
		g_list_free(zeroes_p_const);
		uint32_t output_len = strlen(file->name) + 1 + max_offset_digits + 1 + (SHA256_DIGEST_LENGTH*2+1) + 1 + 1 + 1 + zero_len + 1 + 1;
		char *output = (char *)calloc(output_len, sizeof(char));

		if (output == NULL) {
			free(zero_s);
			ERROR(-1);
		}
		int ret = sprintf(output, "%s,%d,%s,%d,%s\n", file->name, offset / PAGE_SIZE, hash, file->perms[offset / PAGE_SIZE], zero_s);

		free(zero_s);
		if (ret < 0) {
			free(output);
			ERROR(ret);
		}
		g_hash_table_insert(outputs, output, "full");
		offset += PAGE_SIZE;
	}
	return 0;
}

int
process_pe(char *content, char *name, GHashTable *outputs)
{
	int ret = -1;
	struct file_pe_t file = create_file_pe();

	file.name = name;
	file.bytes = content;
	if (file.bytes == NULL)
		ERROR_GO(error);
	if (pe_parse(&file))
		ERROR_GO(error);
	if (expand(&file))
		ERROR_GO(error);
	if (parse_alterations(&file))
		ERROR_GO(error);
	if (parse_pages(&file))
		ERROR_GO(error);
	if (output(&file, outputs))
		ERROR_GO(error);
	ret = 0;
error:
	free_file_pe(&file);
	return ret;
}

int
process_raw(const char *content, const char *name, uint64_t size, GHashTable *outputs)
{
	int ret = -1;
	struct file_pe_t file = create_file_pe();
	int perms_s;

	file.name = name;
	file.virtual_size = size + (PAGE_SIZE - (size % PAGE_SIZE));
	file.virtual = (uint8_t *)calloc(file.virtual_size, sizeof(uint8_t));
	if (file.virtual == NULL)
		ERROR_GO(error);
	memcpy(file.virtual, content, size);
	perms_s = file.virtual_size / PAGE_SIZE;

	file.perms = (uint8_t *)calloc(perms_s, sizeof(uint8_t));
	if (file.perms == NULL)
		ERROR_GO(error);
	file.alterations = g_hash_table_new_full(NULL, NULL, NULL, (GDestroyNotify)g_hash_table_destroy);


	if (parse_pages(&file))
		ERROR_GO(error);
	if (output(&file, outputs))
		ERROR_GO(error);
	ret = 0;
error:
	free(content);
	free_file_pe(&file);
	return ret;
}

int output_sort_function(gconstpointer a, gconstpointer b)
{
	const char *str_a = (const char *)a;
	const char *str_b = (const char *)b;
	const char *p_a = strchr(str_a, ',');
	const char *p_b = strchr(str_b, ',');
	int len = p_a - str_a > p_b - str_b ? p_a - str_a : p_b - str_b;
	int ret = strncmp(str_a, str_b, len);

	if (ret)
		return ret;
	return atoi(p_a+1) > atoi(p_b+1);
}

int
write_outputs(GHashTable *outputs, char *irregular_files, char **blob)
{
	int ret = -1;
	*blob = (char *)calloc(0, sizeof(char));

	if (*blob == NULL)
		return -1;
	char *prev_blob = *blob;
	uint32_t blob_s = 0;
	GList *outputs_head = g_hash_table_get_keys(outputs);
	GList *outputs_p = g_list_sort(outputs_head, output_sort_function);
	GList *outputs_p_const = outputs_p;

	while (outputs_p) {
		char *output = outputs_p->data;
		uint32_t len = strlen(output);
		*blob = realloc(prev_blob = (*blob), (blob_s + len) * sizeof(char));
		if ((*blob) == NULL)
			ERROR_GO(error);
		memcpy((*blob)+blob_s, output, len);
		blob_s += len;
		outputs_p = outputs_p->next;
	}
	if (irregular_files != NULL) {
		uint32_t irregular_len = strlen(irregular_files);
		*blob = realloc(prev_blob = (*blob), (blob_s + irregular_len) * sizeof(char));
		if ((*blob) == NULL)
			ERROR_GO(error);
		memcpy((*blob)+blob_s, irregular_files, irregular_len);
		blob_s += irregular_len;
	}
	(*blob)[blob_s-1] = '\0';
	g_list_free(outputs_p_const);
	return 0;
error:
	g_list_free(outputs_p_const);
	free(prev_blob);
	return -1;
}

int
apsh_windows_zip_to_json(char *path_to_zip, char **output)
{
	int ret = -1;
	int err = 0;
	struct zip *zip_file = zip_open(path_to_zip, 0, &err);
	int64_t i, entries = zip_get_num_entries(zip_file, 0);

	if (zip_file == NULL || entries < 0)
		ERROR(ret);

	GHashTable *outputs = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
	char *irregular_files_string = NULL;

	if (outputs == NULL)
		ERROR_GO(close_zip);
	for (i = 0; i < entries; i++) {
		struct zip_stat st;

		zip_stat_init(&st);
		if (zip_stat_index(zip_file, i, 0, &st))
			ERROR_GO(destroy_outputs);
		if (st.size == 0)
			continue;
		enum FILE_TYPE file_type = is_file_relevant(st.name);

		if (file_type == SKIP)
			continue;
		struct zip_file *file = zip_fopen_index(zip_file, i, 0);

		if (file == NULL)
			goto destroy_outputs;
		uint8_t *content = (uint8_t *)calloc(st.size, sizeof(uint8_t));

		if (content == NULL)
			ERROR_GO(destroy_outputs);
		zip_fread(file, content, st.size);
		zip_fclose(file);
		if (strcmp(st.name, "doca_apsh_irregular_files.txt") != 0) {
			if (file_type == PE_TYPE)
				ret = process_pe(content, st.name, outputs);
			else if (file_type == RAW_TYPE)
				ret = process_raw(content, st.name, st.size, outputs);
			if (ret)
				ERROR_GO(destroy_outputs);
		} else {
			irregular_files_string = content;
		}
	}

	ret = write_outputs(outputs, irregular_files_string, output);
destroy_outputs:
	if (irregular_files_string != NULL)
		free(irregular_files_string);
	g_hash_table_destroy(outputs);
close_zip:
	zip_close(zip_file);
	return ret;
}
