/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "hashtest.h"
#include "../../doca_apsh_attr.h"
#include "doca_log.h"
#include "openssl/sha.h"
#include "vadinfo.h"
#include <glib.h>
#include <stdbool.h>
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::Windows::HASHTEST);

static const protection_t winnt_protection[] =
	{
		{"PAGE_NOACCESS", 0x01},
		{"PAGE_READONLY", 0x02},
		{"PAGE_READWRITE", 0x04},
		{"PAGE_WRITECOPY", 0x08},
		{"PAGE_EXECUTE", 0x10},
		{"PAGE_EXECUTE_READ", 0x20},
		{"PAGE_EXECUTE_READWRITE", 0x40},
		{"PAGE_EXECUTE_WRITECOPY", 0x80},
		{"PAGE_GUARD", 0x100},
		{"PAGE_NOCACHE", 0x200},
		{"PAGE_WRITECOMBINE", 0x400},
		{"PAGE_TARGETS_INVALID", 0x40000000}};

struct zero_entry {
	/* offset from the begginning of the page where zero should be applied */
	unsigned int entry_offset;
	struct zero_entry *next;
};

struct page_entry {
	int index;
	unsigned char hash[SHA256_DIGEST_LENGTH];
	int permission;
	struct page_entry *next;
	struct zero_entry *zeroes;
};

struct file_entry {
	unsigned char file_name[MAX_FILENAME];
	bool is_irregular_file;
	struct page_entry *pages;
	struct file_entry *next;
};

struct area_result {
	int pages_present;
	int matching_hashes;
};

static void free_zero_nodes(struct page_entry *page_node);

static bool
is_file_irregular(char *file_name, struct file_entry *file_hashes) {
	struct file_entry *files_ptr = file_hashes;

	while (files_ptr != NULL) {
		if (strcmp(files_ptr->file_name, file_name) == 0)
			return files_ptr->is_irregular_file;
		files_ptr = files_ptr->next;
	}

	return false;
}

static void
set_full_name(char *new_full_name, char *fullpath)
{
	char *full_file_name_p = fullpath + 1;
	int i;

	for (i = 0; i < (MAX_PATH - 1) && full_file_name_p[i] != '\0'; i++) {
		if (full_file_name_p[i] == '\\')
			new_full_name[i] = '_';
		else
			new_full_name[i] = full_file_name_p[i];
	}
	new_full_name[i] = '\0';
}

static bool
are_equal(const char *s1, const char *s2, int length)
{
	for (int i = 0; i < length; i++) {
		if (s1[i] != s2[i])
			return false; /* They are different */
	}
	return true; /* They must be the same */
}

static void
sha256h_string(char *string, unsigned int length, unsigned char *hash)
{

	struct timespec ts1, ts2;
	SHA256_CTX sha256;

	SHA256_Init(&sha256);
	SHA256_Update(&sha256, string, length);
	SHA256_Final(hash, &sha256);
}

static int
print_pages(apsh_instance_t apsh, addr_t start_addr, addr_t end_addr, apsh_pid_t pid)
{

	int committed = 0;
	unsigned long length = end_addr - start_addr;
	unsigned char page_hash[SHA256_DIGEST_LENGTH] = {0};

	if (length < 0xFFFFFFFF && length > 0) {
		while (start_addr < end_addr) {
			char buffer[0x1000] = {0};

			if (apsh_read_va(apsh, start_addr, pid, 0x1000, &buffer, NULL) == APSH_S) {
				char log_buffer[0x2001] = {0};

				++committed;
				DOCA_DLOG_INFO("\n------------------------------------------------------");
				for (int i = 0; i < 0x1000; i++)
					sprintf(log_buffer + 2*i, "%02x", buffer[i]);
				DOCA_DLOG_INFO("%s\n", log_buffer);
			}
			start_addr += 0x1000;
		}
	}
	return committed;
}

static status_t
parse_line(char *line, size_t line_len, struct file_entry **file_hashes_head)
{
	char *zero_line, *ptr;
	char *file_name;
	struct file_entry *file_node, *new_file_node;
	struct page_entry *page_node, *new_page_node;
	struct zero_entry *zero_node, *new_zero_node;
	unsigned char input_hash[SHA256_DIGEST_LENGTH * 2 + 1] = {0};

	if (line_len <= 0 || line == NULL) {
		DOCA_DLOG_ERR("%s: error while prsing the line", __func__);
		return APSH_F;
	}

	if (line[0] == '#') { //irregular file
		new_file_node = calloc(1, sizeof(struct file_entry));
		if (!new_file_node) {
			DOCA_DLOG_ERR("%s: allocation error", __func__);
			goto error_exit;
		}
		new_file_node->next = NULL;
		new_file_node->pages = NULL;
		new_file_node->is_irregular_file = true;
		strncpy(new_file_node->file_name, line + 1, MAX_FILENAME);

		file_node = *file_hashes_head;
		while (file_node->next != NULL)
			file_node = file_node->next;

		if (*file_hashes_head == NULL)
			*file_hashes_head = new_file_node;
		else
			file_node->next = new_file_node;
		return APSH_S;
	}

	file_name = strtok(line, ",");
	file_node = *file_hashes_head;
	bool found = false;

	while (file_node) {
		if (strcmp(file_node->file_name, file_name) == 0) {
			found = true;
			break;
		}
		if (!file_node->next)
			break;
		file_node = file_node->next;
	}

	if (!found) {

		new_file_node = malloc(sizeof(struct file_entry));
		if (!new_file_node) {
			DOCA_DLOG_ERR("%s: allocation error", __func__);
			goto error_exit;
		}
		new_file_node->next = NULL;
		new_file_node->pages = NULL;
		new_file_node->is_irregular_file = false;
		memset(new_file_node->file_name, 0, MAX_FILENAME);
		strncpy(new_file_node->file_name, file_name, MAX_FILENAME);
		if (*file_hashes_head == NULL) {
			file_node = *file_hashes_head = new_file_node;
		} else {
			file_node->next = new_file_node;
			file_node = file_node->next;
		}
	}

	page_node = file_node->pages;
	while (page_node && page_node->next)
		page_node = page_node->next;

	new_page_node = malloc(sizeof(struct page_entry));
	if (!new_page_node) {
		DOCA_DLOG_ERR("%s: allocation error", __func__);
		goto error_exit;
	}
	new_page_node->zeroes = NULL;
	new_page_node->next = NULL;
	new_page_node->index = atoi(strtok(NULL, ","));
	memset(input_hash, 0, SHA256_DIGEST_LENGTH * 2);
	strncpy(input_hash, strtok(NULL, ","), SHA256_DIGEST_LENGTH * 2);

	for (int i = 0; i < SHA256_DIGEST_LENGTH; i++) {
		char base[3] = {0};

		base[0] = input_hash[2 * i];
		base[1] = input_hash[2 * i + 1];
		new_page_node->hash[i] = strtoul(base, NULL, 16);
	}

	new_page_node->permission = atoi(strtok(NULL, ","));
	if (!file_node->pages)
		file_node->pages = new_page_node;
	else
		page_node->next = new_page_node;

	zero_line = strtok(NULL, ",");
	if ((zero_line != NULL) && (strlen(zero_line) > 2)) {
		zero_line = strtok(zero_line, " ");
		while (zero_line != NULL) {
			new_zero_node = malloc(sizeof(struct zero_entry));
			if (!new_zero_node) {
				DOCA_DLOG_ERR("%s: allocation error", __func__);
				goto error_exit;
			}
			new_zero_node->next = NULL;
			new_zero_node->entry_offset = atoi(zero_line);
			if (!new_page_node->zeroes) {
				zero_node = new_page_node->zeroes = new_zero_node;
			} else {
				zero_node->next = new_zero_node;
				zero_node = zero_node->next;
			}
			zero_line = strtok(NULL, " ");
		}
	}
	return APSH_S;

error_exit:
	return APSH_F;
}

static status_t
build_indexes_zip(const char *hash_file_path, struct file_entry **file_hashes_head)
{
	char *output;
	int ret = apsh_windows_zip_to_json(hash_file_path, &output);

	if (ret) {
		DOCA_DLOG_ERR("parse zip error");
		return APSH_F;
	}
	char *cur_line = output;

	while (cur_line) {
		char *next_line = strchr(cur_line, '\n');

		if (next_line)
			*next_line = '\0';
		else
			break;
		char *line = strdup(cur_line);

		if (line == NULL) {
			DOCA_DLOG_ERR("strdup failed");
			free(output);
			return APSH_F;
		}
		if (parse_line(line, next_line-cur_line, file_hashes_head) != APSH_S) {
			DOCA_DLOG_ERR("parse_line error");
			free(line);
			free(output);
			return APSH_F;
		}
		free(line);
		if (next_line)
			*next_line = '\n';
		cur_line = next_line ? (next_line+1) : NULL;
	}
	free(output);
	return APSH_S;
}

static status_t
build_indexes_json(char *hash_file_path, struct file_entry **file_hashes_head)
{
	FILE *fp = NULL;
	char *buffer = NULL;
	size_t bufsize = 1000, line_len = 0;
	status_t ret = APSH_F;

	if (!hash_file_path) {
		DOCA_DLOG_ERR("hashes file path is null");
		return APSH_F;
	}

	buffer = (char *)malloc(bufsize * sizeof(char));
	if (buffer == NULL) {
		DOCA_DLOG_ERR("Unable to allocate buffer");
		return APSH_F;
	}

	/* Just a standard way to open files. */
	fp = fopen(hash_file_path, "r");
	/* If file does not exist, exit. */
	if (!fp) {
		DOCA_DLOG_ERR("error opening hashes file path");
		goto exit;
	}

	/* Read file line by line. */
	while ((line_len = getline(&buffer, &bufsize, fp)) != -1) {
		if (parse_line(buffer, line_len, file_hashes_head) != APSH_S) {
			DOCA_DLOG_ERR("parse_line error");
			goto exit;
		}
	}

	ret = APSH_S;
exit:
	if (fp)
		fclose(fp);
	if (buffer)
		free(buffer);
	return ret;
}

static status_t
build_indexes(char *hash_file_path, struct file_entry **file_hashes_head)
{
	char *zip_ending = ".zip";
	char *file_ending = hash_file_path+strlen(hash_file_path)-strlen(zip_ending);

	if (!strncmp(zip_ending, file_ending, strlen(zip_ending)))
		return build_indexes_zip(hash_file_path, file_hashes_head);
	else
		return build_indexes_json(hash_file_path, file_hashes_head);
}

static void
print_page_node(struct page_entry *page_node)
{
	if (!page_node)
		return;

	char log_buffer[2048] = {0};
	int buffer_offset = 0;

	buffer_offset += sprintf(log_buffer + buffer_offset, "	index: %d, permission: %d, hash: ", page_node->index, page_node->permission);
	for (int i = 0; i < SHA256_DIGEST_LENGTH; i++)
		buffer_offset += sprintf(log_buffer + buffer_offset, "%02x", page_node->hash[i]);

	if (page_node->zeroes) {
		buffer_offset += sprintf(log_buffer + buffer_offset, ", zeroes: ");
		for (struct zero_entry *zero_node = page_node->zeroes; zero_node != NULL;
			zero_node = zero_node->next) {
			buffer_offset += sprintf(log_buffer + buffer_offset, "%u ", zero_node->entry_offset);
		}
	}
	DOCA_DLOG_INFO("%s", log_buffer);
}

static void
print_file_node(struct file_entry *file_node)
{
	if (!file_node)
		return;
	DOCA_DLOG_INFO("file name: %s", file_node->file_name);
	for (struct page_entry *page_node = file_node->pages; page_node != NULL;
		 page_node = page_node->next) {
		print_page_node(page_node);
	}
}

/* Checks if a node exists with the same name
 * if not, it might be a library injection
 */
static bool
file_nodes_exist(const char *file_name, struct file_entry *file_hashes_head)
{

	if (!file_name)
		return false;

	struct file_entry *file_node;

	file_node = file_hashes_head;

	while (file_node) {
		if (strncmp(file_node->file_name, file_name, MAX_PATH) == 0)
			return true;
		file_node = file_node->next;
	}
	return false;
}

static struct page_entry *
find_page_node(const char *file_name, int page_index, struct file_entry *file_hashes_head)
{

	struct file_entry *file_node;
	struct page_entry *page_node;

	file_node = file_hashes_head;

	while (file_node) {
		if (strncmp(file_node->file_name, file_name, MAX_PATH) == 0) {
			page_node = file_node->pages;
			while (page_node) {
				if (page_node->index == page_index)
					return page_node;
				page_node = page_node->next;
			}
			return NULL;
		}
		file_node = file_node->next;
	}
	return NULL;
}

static void
zero_page(char *page_content, int page_size, struct zero_entry *zero_indexes)
{

	unsigned int offset = 0;

	for (struct zero_entry *zero_node = zero_indexes; zero_node != NULL;
		 zero_node = zero_node->next) {

		offset = zero_node->entry_offset;
		if (offset <= page_size - 4) { /* zero 4 bytes from offset inclusive */
			page_content[offset] = 0;
			page_content[offset + 1] = 0;
			page_content[offset + 2] = 0;
			page_content[offset + 3] = 0;
		}
	}
}

static void
print_hash256(const char *hash)
{
	if (!hash)
		return;

	char log_buffer[1024] = {0};
	int buffer_offset = 0;

	buffer_offset += sprintf(log_buffer + buffer_offset, "SHA256:");
	for (int i = 0; i < SHA256_DIGEST_LENGTH; i++)
		buffer_offset += sprintf(log_buffer + buffer_offset, "%02x", hash[i]);
	DOCA_DLOG_INFO("%s", log_buffer);
}

/*
 * returns 1 if page is writable
 * returns 0 if page is NOT writable
 * returns -1 on error (arch is not supported yet)
 */

static int
is_page_writable(apsh_instance_t apsh, page_info_t page_info)
{
	page_mode_t page_mode = apsh_get_page_mode(apsh, 0);
	int ret = 0;

	switch (page_mode) {
	case APSH_PM_LEGACY: /* 32 bit */
	case APSH_PM_PAE:	/* 32 bit extended */
	case APSH_PM_IA32E:  /* 64 bit */
	{
		int bit1 = page_info.x86_ia32e.pte_value & APSH_BIT_MASK(1, 1);
		int bit11 = page_info.x86_ia32e.pte_value & APSH_BIT_MASK(11, 11);

		return bit1 || bit11;
	}
	default: {
		DOCA_DLOG_ERR("%s: arch is not supported yet!", __func__);
		return -1;
	}
	}
}

static status_t
analyze_area_pages(apsh_instance_t apsh, addr_t start_addr, addr_t end_addr, apsh_pid_t pid,
		   const char *path, struct file_entry *file_hashes_head, struct area_result *result)
{

	char page_content[0x1000] = {0};
	char page_content_copy[PAGE_SIZE];
	struct page_entry *page_node = NULL;
	unsigned long length = end_addr - start_addr;
	unsigned char page_hash[SHA256_DIGEST_LENGTH] = {0};

	if (path == NULL || result == NULL) {
		DOCA_DLOG_ERR("null pointers received");
		return APSH_F;
	}

	result->pages_present = 0;
	result->matching_hashes = 0;

	for (addr_t page_addr = start_addr; page_addr < end_addr; page_addr += 0x1000) {
		if (apsh_read_va(apsh, page_addr, pid, 0x1000, &page_content, NULL) == APSH_S) {
			addr_t dtb;

			if (apsh_pid_to_dtb(apsh, pid, &dtb) == APSH_F) {
				DOCA_DLOG_ERR("analzye pages: apsh_pid_to_dtb failed!!");
				return APSH_F;
			}
			page_info_t page_info;

			if (apsh_pagetable_lookup_extended(apsh, dtb, page_addr, &page_info) ==
				APSH_S) {
				int res = is_page_writable(apsh, page_info);

				if (res == -1)
					return APSH_F;

				if (res == 1)
					continue;
			}
			++result->pages_present;
			page_node = find_page_node(path, (page_addr - start_addr) / PAGE_SIZE, file_hashes_head);
			if (page_node == NULL)
				continue;

			if (page_node->zeroes != NULL)
				memcpy(page_content_copy, page_content, PAGE_SIZE);

			zero_page(page_content, PAGE_SIZE, page_node->zeroes);
			sha256h_string(page_content, PAGE_SIZE, page_hash);
			if (are_equal(page_hash, page_node->hash, SHA256_DIGEST_LENGTH)) {
				++result->matching_hashes;
				if (page_node->zeroes != NULL) {
					free_zero_nodes(page_node);
					page_node->zeroes = NULL;
					sha256h_string(page_content_copy, PAGE_SIZE, page_node->hash);
				}
			}
		}
	}
	return APSH_S;
}

static void
free_zero_nodes(struct page_entry *page_node)
{
	struct zero_entry *zero_node = page_node->zeroes;
	struct zero_entry *zero_node_next;

	while (zero_node != NULL) {
		zero_node_next = zero_node->next;
		free(zero_node);
		zero_node = zero_node_next;
	}
}

static void
free_page_nodes(struct file_entry *file_node)
{
	struct page_entry *page_node = file_node->pages;
	struct page_entry *page_node_next;

	while (page_node != NULL) {
		page_node_next = page_node->next;
		free_zero_nodes(page_node);
		free(page_node);
		page_node = page_node_next;
	}
}

static void
free_file_nodes(struct file_entry *file_head)
{
	struct file_entry *file_node = file_head;
	struct file_entry *file_node_next;

	while (file_node != NULL) {
		file_node_next = file_node->next;
		free_page_nodes(file_node);
		free(file_node);
		file_node = file_node_next;
	}
}

static void
free_hashtest_hashtable(GHashTable *hash_table)
{
	char *process_name;
	char *path;
	char *protection;

	process_name = g_hash_table_lookup(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_COMM));
	path = g_hash_table_lookup(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PATH_OF_MEMORY_AREA));
	protection = g_hash_table_lookup(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PROTECTION));

	if (process_name != NULL)
		g_free(process_name);

	if (path != NULL)
		g_free(path);

	if (protection != NULL)
		g_free(protection);

	g_hash_table_destroy(hash_table);
}

void
windows_free_hashtest_list(GSList *results_list)
{
	if (results_list) {
		g_slist_foreach(results_list, free_hashtest_hashtable, NULL);
		g_slist_free(results_list);
	}
}

void
windows_free_hashtest_results(windows_hashtest_t *results_struct)
{
	if (results_struct != NULL) {
		if (results_struct->hash_file != NULL)
			free(results_struct->hash_file);
		if (results_struct->file_hashes != NULL)
			free_file_nodes(results_struct->file_hashes);
		windows_free_hashtest_list(results_struct->results_list);
		free(results_struct);
	}
}

static void
print_hash_table_list(GSList *hash_table_list)
{
	uint32_t pid;
	char *process_name;
	char *path;
	char *protection;
	addr_t start_address;
	addr_t end_address;
	int pages_number;
	int pages_present;
	int matching_hashes;
	bool hash_file_present;
	GHashTable *hash_table;
	int i = 0;

	if (!hash_table_list)
		return;

	while (1) {
		hash_table = g_slist_nth_data(hash_table_list, i);
		if (hash_table == NULL)
			break;

		pid = g_hash_table_lookup(hash_table,
					  GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PID));
		process_name = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_COMM));
		path = g_hash_table_lookup(
			hash_table,
			GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PATH_OF_MEMORY_AREA));
		protection = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PROTECTION));
		start_address = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_START_ADDRESS));
		end_address = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_END_ADDRESS));
		pages_number = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_NUMBER));
		pages_present = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_PRESENT));
		matching_hashes = g_hash_table_lookup(
			hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_MATCHING_HASHES));
		hash_file_present = g_hash_table_lookup(
			hash_table,
			GINT_TO_POINTER(DOCA_APSH_ATTESTATION_HASH_DATA_IS_PRESENT));

		DOCA_DLOG_INFO("[Pid : %d] [Process Name : %s] [Protection : %s] [Start: %lx] [End: %lx] [Path: %s] [Pages Number: %d] [present Pages: %d] [Hash Matches: %d] [Hash Present: %d]\n",
			   pid, process_name, protection, start_address, end_address, path,
			   (end_address - start_address) / 0x1000, pages_present, matching_hashes,
			   hash_file_present);
		i++;
	}
}

static status_t
insert_result(GSList **hash_table_list, uint32_t pid, char *process_name, char *path,
		  char *protection, addr_t start_address, addr_t end_address, int pages_number,
		  int pages_present, int matching_hashes, bool hash_file_present)
{

	gboolean result;

	GHashTable *hash_table = g_hash_table_new(g_direct_hash, g_direct_equal);

	if (hash_table == NULL)
		return APSH_F;

	char *proc_name_copy = g_strndup(process_name, strlen(process_name) + 1);
	char *path_copy = g_strndup(path, strlen(path) + 1);
	char *protection_copy = g_strndup(protection, strlen(protection) + 1);

	result = g_hash_table_insert(hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PID),
					 GINT_TO_POINTER(pid));
	result &= g_hash_table_insert(
		hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_COMM), proc_name_copy);
	result &= g_hash_table_insert(
		hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PATH_OF_MEMORY_AREA),
		path_copy);
	result &= g_hash_table_insert(hash_table,
					  GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PROTECTION),
					  protection_copy);
	result &= g_hash_table_insert(hash_table,
					  GINT_TO_POINTER(DOCA_APSH_ATTESTATION_START_ADDRESS),
					  GUINT_TO_POINTER(start_address));
	result &= g_hash_table_insert(hash_table,
					  GINT_TO_POINTER(DOCA_APSH_ATTESTATION_END_ADDRESS),
					  GUINT_TO_POINTER(end_address));
	result &= g_hash_table_insert(hash_table,
					  GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_NUMBER),
					  GINT_TO_POINTER(pages_number));
	result &= g_hash_table_insert(hash_table,
					  GINT_TO_POINTER(DOCA_APSH_ATTESTATION_PAGES_PRESENT),
					  GINT_TO_POINTER(pages_present));
	result &= g_hash_table_insert(
		hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_MATCHING_HASHES),
		GINT_TO_POINTER(matching_hashes));
	result &= g_hash_table_insert(
		hash_table, GINT_TO_POINTER(DOCA_APSH_ATTESTATION_HASH_DATA_IS_PRESENT),
		GINT_TO_POINTER(hash_file_present));

	if (result) {
		*hash_table_list = g_slist_append(*hash_table_list, hash_table);
		return APSH_S;
	}

	g_hash_table_destroy(hash_table);
	return APSH_F;
}

static status_t
init_plugin(apsh_instance_t apsh, const char *hash_file_path, struct file_entry **file_hashes_head,
		  struct vadinfo_struct **vad_struct)
{

	if (*file_hashes_head == NULL && build_indexes(hash_file_path, file_hashes_head) != APSH_S) {
		DOCA_DLOG_ERR("Failed in building the indexes from hash file");
		return APSH_F;
	}

	struct vadinfo_struct *vad_struct_ptr = (struct vadinfo_struct *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_VMA);

	if (vad_struct_ptr == NULL) {
		vad_struct_ptr = malloc(sizeof(*vad_struct_ptr));

		if (vad_struct_ptr == NULL) {
			DOCA_DLOG_ERR("Failed to allocate plugin context memory");
			return APSH_F;
		}

		if (vadinfo_struct_init(apsh, vad_struct_ptr) == APSH_F) {
			free(vad_struct_ptr);
			return APSH_F;
		}

		if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_VMA, vad_struct_ptr, free)) {
			free(vad_struct_ptr);
			return APSH_F;
		}
	}
	*vad_struct = vad_struct_ptr;
	return APSH_S;
}

static status_t
hashtest_get_tag(apsh_instance_t apsh, addr_t vaddr, char *out)
{
	/* get tag */

	addr_t tag_addr = vaddr - 12; /* for 64bit Windows OS */

	return apsh_read_32_va(apsh, tag_addr, 0, (uint32_t *)out);
}

static status_t
hashtest_mmvad_short(apsh_instance_t apsh, addr_t vaddr, struct vadinfo_struct *vad_struct, vadinfo_t *out)
{
	addr_t node_addr = vaddr;
	mmvad_short_t _mmvad_short;

	/* load mmvad_short data */
	if (apsh_read_va(apsh, node_addr, 0, sizeof(mmvad_short_t), &_mmvad_short, NULL) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get mmvad_short at address 0x%lx", node_addr);
		return APSH_F;
	}

	/* get start vpn */
	out->startvpn = ((uint64_t)_mmvad_short.StartingVpn << 12) |
			((uint64_t)_mmvad_short.StartingVpnHigh << 44);

	/* get end vpn */
	out->endvpn = (((uint64_t)(_mmvad_short.EndingVpn + 1) << 12) |
			   ((uint64_t)_mmvad_short.EndingVpnHigh << 44));

	/* bit 3 ~ 8 */
	uint32_t protection_idx = ((_mmvad_short.u) >> vad_struct->mmvad_flag.protection_start_bit) &
				  BIT_MASK(uint32_t, (vad_struct->mmvad_flag.protection_end_bit -
						      vad_struct->mmvad_flag.protection_start_bit));

	int protection_value = vad_struct->MmProtectToValue[protection_idx];
	int count = sizeof(winnt_protection) / sizeof(protection_t);
	int i = 0;

	char *p = out->protection;

	for (i = 0; i < count; i++) {
		if (protection_value & winnt_protection[i].mask) {
			sprintf(p, "%s", winnt_protection[i].name);
			break;
		}
	}

	uint32_t CommitCharge_value = ((_mmvad_short.u1) >> vad_struct->mmvad_flag.commit_charge_start_bit) &
				      BIT_MASK(uint32_t, (vad_struct->mmvad_flag.commit_charge_end_bit -
							  vad_struct->mmvad_flag.commit_charge_start_bit));
	uint32_t PrivateMemory_value = ((_mmvad_short.u) >> vad_struct->mmvad_flag.private_memory_start_bit) &
				       BIT_MASK(uint32_t, (vad_struct->mmvad_flag.private_memory_end_bit -
							   vad_struct->mmvad_flag.private_memory_start_bit));

	/* get parent */
	uint64_t ParentValue = _mmvad_short.VadNode.ParentValue & ~0x03;

	return APSH_S;
}

static status_t
hashtest_mmvad(apsh_instance_t apsh, addr_t vaddr, struct vadinfo_struct *vad_struct, vadinfo_t *out)
{
	if (hashtest_mmvad_short(apsh, vaddr, vad_struct, out) == APSH_F) {
		DOCA_DLOG_ERR("failed to get mmvad_short");
		return APSH_F;
	}

	addr_t node_addr = vaddr;
	addr_t subsection_va;
	addr_t control_area_va;
	addr_t file_object_va;

	if (apsh_read_addr_va(apsh, node_addr + vad_struct->mmvad.subsection_rva, 0, &subsection_va) == APSH_F) {
		DOCA_DLOG_ERR("Failed to read subsection_va addr.");
		return APSH_F;
	}

	if (apsh_read_addr_va(apsh, subsection_va + vad_struct->subsection.control_area_rva, 0, &control_area_va) == APSH_F) {
		DOCA_DLOG_ERR("Failed to read control_area_va addr.");
		return APSH_F;
	}

	if (apsh_read_addr_va(apsh, control_area_va + vad_struct->control_area.file_pointer_rva, 0, &file_object_va) == APSH_F) {
		DOCA_DLOG_ERR("Failed to read file_pointer_rva addr.");
		return APSH_F;
	}

	/* file_object is a special _EX_FAST_REF pointer, we need to explicitly clear low bits */
	file_object_va &= (~0xFULL);

	addr_t file_name_va = NULL;

	if ((void *)file_object_va != NULL)
		file_name_va = file_object_va + vad_struct->file_object.FileName_rva;

	/* get filename */
	unicode_string_t *us = NULL;

	us = apsh_read_unicode_str_va(apsh, file_name_va, 0);
	if (us == NULL) {
		strncpy(out->filepath, "N/A", MAX_PATH);
	} else {
		unicode_string_t _out = {0};

		if (apsh_convert_str_encoding(us, &_out, "UTF-8") == APSH_S) {
			strncpy(out->filepath, _out.contents, MAX_PATH);
			free(_out.contents);
		} else {
			strncpy(out->filepath, "N/A", MAX_PATH);
		}

		if (us)
			apsh_free_unicode_str(us);
	}

	return APSH_S;
}

static status_t
analyze_memory_map(apsh_instance_t apsh, addr_t vadroot_va, addr_t process_addr, uint32_t pid,
		   char *procname, struct file_entry *file_hashes_head, struct vadinfo_struct *vad_struct,
		   GSList **hash_table_list, uint64_t *count)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_HASHTEST_LIMIT)));

	addr_t node_addr = vadroot_va;
	vadinfo_t _vadinfo;

	_vadinfo.pid = pid;
	_vadinfo.procaddr = process_addr;
	_vadinfo.procname = procname;
	memset(_vadinfo.filepath, 0, MAX_PATH);
	memset(_vadinfo.protection, 0, MAX_PATH);
	memset(_vadinfo.tag, 0, 8);

	if (hashtest_get_tag(apsh, node_addr, &_vadinfo.tag[0]) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get Tag of VAD at 0x%lx", node_addr);
		return APSH_F;
	}

	if ((memcmp(&_vadinfo.tag[0], "VadS", 4) == 0) ||
		(memcmp(&_vadinfo.tag[0], "VadF", 4) == 0)) {
		/* _MMVAD_SHORT */
		hashtest_mmvad_short(apsh, node_addr, vad_struct, &_vadinfo);
	} else if (memcmp(&_vadinfo.tag[0], "Vad", 3) == 0) {
		/* _MMVAD */
		hashtest_mmvad(apsh, node_addr, vad_struct, &_vadinfo);

		char *filename = strrchr(_vadinfo.filepath, '\\');

		if (filename != NULL)
			filename++;

		if (filename && !strcmp(filename, "N/A"))
			filename = NULL;

		struct area_result result = {0};

		if (is_file_irregular(filename, file_hashes_head)) {
			;
		} else if (filename) {
			if (strcmp(filename, "ntdll.dll") &&
				(!strcmp(_vadinfo.protection, "PAGE_READONLY") ||
				!strcmp(_vadinfo.protection, "PAGE_WRITECOPY") ||
				!strcmp(_vadinfo.protection, "PAGE_EXECUTE") ||
				!strcmp(_vadinfo.protection, "PAGE_EXECUTE_READ") ||
				!strcmp(_vadinfo.protection, "PAGE_EXECUTE_WRITECOPY"))) {
				char full_file_name[MAX_PATH + 1];

				set_full_name(full_file_name, _vadinfo.filepath);
				if (analyze_area_pages(apsh, _vadinfo.startvpn, _vadinfo.endvpn, pid,
						full_file_name, file_hashes_head, &result) != APSH_S) {
					if (hash_table_list && *hash_table_list)
						windows_free_hashtest_list(*hash_table_list);
					return APSH_F; /* there was an error allocating structures */
				}
				if (result.pages_present > 0) {
					insert_result(hash_table_list, pid, procname, full_file_name,
						_vadinfo.protection, _vadinfo.startvpn,
						_vadinfo.endvpn,
						(_vadinfo.endvpn - _vadinfo.startvpn) / 0x1000,
						result.pages_present, result.matching_hashes,
						file_nodes_exist(full_file_name, file_hashes_head));
				}
			}
		}
	}

	rtl_balanced_node_t vadnode;

	if (apsh_read_va(apsh, node_addr, 0, sizeof(rtl_balanced_node_t), &vadnode, NULL) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get vadnode at address 0x%lx", node_addr);
		return APSH_F;
	}

	if((*count)++ > limit)
		return APSH_F;
	/* process left tree */
	if (vadnode.Left) {
		if (analyze_memory_map(apsh, vadnode.Left, process_addr, pid, procname, file_hashes_head, vad_struct,
				hash_table_list, count) == APSH_F)
			return APSH_F;
	}

	if(*count > limit)
		return APSH_F;
	/* process right tree */
	if (vadnode.Right) {
		if (analyze_memory_map(apsh, vadnode.Right, process_addr, pid, procname, file_hashes_head, vad_struct,
				hash_table_list, count) == APSH_F)
			return APSH_F;
	}
	return APSH_S;
}

status_t
windows_hashtest_internal(apsh_instance_t apsh, windows_processes_list_t *process, uint32_t pid,
		 const char *hash_file, struct file_entry *file_hashes_head, windows_hashtest_t **results_list)
{
	status_t result;
	windows_processes_list_t *psl = NULL, *psl_allocation = NULL;
	GSList *temp_results_list = NULL;
	struct file_entry *file_hashes = file_hashes_head;
	struct vadinfo_struct *vad_struct;

	if (hash_file == NULL) {
		DOCA_DLOG_ERR("No hash file provided");
		goto error_exit;
	}

	if (init_plugin(apsh, hash_file, &file_hashes, &vad_struct) == APSH_F) {
		DOCA_DLOG_ERR("Failed in plugin init");
		goto error_exit;
	}

	if (process == NULL) {
		/* search for process using pid */
		bool process_found = false;

		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F) {
			DOCA_DLOG_ERR("Failed to retrieve process list");
			goto error_exit;
		}
		psl = psl_allocation;
		while (psl) {
			if (pid == psl->pid) {
				process_found = true;
				process = psl;
				break;
			}
			psl = psl->next;
		}
		if (process_found == false) {
			DOCA_DLOG_ERR("Could not find a process with PID:%d", pid);
			goto error_exit;
		}
	} else {
		if (process->pid != pid) {
			DOCA_DLOG_ERR("process pid doesn't match pid provided");
			goto error_exit;
		}
	}

	addr_t memory_map = 0;

	// find vadroot
	addr_t vadroot_va;
	addr_t rtl_balanced_node;
	uint64_t count = 0;

	vadroot_va = process->address + vad_struct->vadroot_offset;

	result = apsh_read_64_va(apsh, vadroot_va, 0, &rtl_balanced_node);
	if (result == APSH_S && vadroot_va != 0) {

		status_t result =
			analyze_memory_map(apsh, vadroot_va, process->address, process->pid,
					   process->name, file_hashes, vad_struct, &temp_results_list, &count);
		if (result != APSH_S)
			goto error_exit;

		if (results_list == NULL)
			print_hash_table_list(temp_results_list);
		else {
			windows_hashtest_t *result_struct = malloc(sizeof(*result_struct));

			if (result_struct == NULL) {
				DOCA_DLOG_ERR("Failed to allocate memory to hashtest result");
				goto error_exit;
			}

			result_struct->file_hashes = file_hashes;
			result_struct->pid = pid;
			result_struct->hash_file = hash_file;
			result_struct->results_list = temp_results_list;
			*results_list = result_struct;
		}
		if (psl_allocation != NULL)
			windows_free_process_list(psl_allocation);
		return APSH_S;
	} else {
		DOCA_DLOG_ERR("cannot retrieve memory map for process %d", process->pid);
		goto error_exit;
	}

error_exit:
	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);
	return APSH_F;
}

status_t
windows_hashtest(apsh_instance_t apsh, windows_processes_list_t *process, uint32_t pid,
		 const char *hash_file, windows_hashtest_t **results_list)
{
	return windows_hashtest_internal(apsh, process, pid, strndup(hash_file, MAX_FILENAME), NULL, results_list);
}

status_t
windows_hashtest_refresh(apsh_instance_t apsh, windows_processes_list_t *process, windows_hashtest_t **results_list)
{
	status_t ret;
	windows_hashtest_t *new_results_list;

	ret = windows_hashtest_internal(apsh, process, (*results_list)->pid, (*results_list)->hash_file,
							 (*results_list)->file_hashes, &new_results_list);
	if (ret != APSH_S)
		return ret;
	(*results_list)->hash_file = NULL;
	(*results_list)->file_hashes = NULL;
	*results_list = new_results_list;
	return APSH_S;
}
