/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "private/common_defines.h"
#include "ldrmodules.h"
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"
#include "../../doca_apsh_attr.h"


DOCA_LOG_REGISTER(APSH::Windows::LDRMODULES);

static void
remove_last_node(windows_dll_list_t **lm)
{
	if ((*lm)->next == NULL) { //in case there's a single node
		free(*lm);
		*lm = NULL;
	} else {
		windows_dll_list_t *lm_iter = *lm;

		while (lm_iter->next->next != NULL)
			lm_iter = lm_iter->next;
		free(lm_iter->next);
		lm_iter->next = NULL;
	}
}

static windows_ldrmodules_list_t * ldrmodule_list_put(windows_ldrmodules_list_t *lm, addr_t entry)
{
	windows_ldrmodules_list_t *_lm = lm;
	windows_ldrmodules_list_t *_prev = NULL;

	while (_lm) {
		if (_lm->entry == entry)
			return _lm;

		_prev = _lm;
		_lm = _lm->next;
	}

	_lm = malloc(sizeof(windows_ldrmodules_list_t));
	if (_lm == NULL) {
		DOCA_DLOG_ERR("failed to malloc");
		return NULL;
	}
	_lm->entry = entry;
	_lm->inload = false;
	_lm->inmem = false;
	_lm->ininit = false;
	_lm->next = NULL;

	if (_prev)
		_prev->next = _lm;

	return _lm;
}

void windows_free_ldrmodules_list(windows_ldrmodules_list_t *ldrmodules_list)
{
	windows_ldrmodules_list_t *_lm = ldrmodules_list;

	while (_lm) {
		windows_ldrmodules_list_t *_next = _lm->next;
		if (_lm->procname != NULL)
			free(_lm->procname);
		if (_lm->base_dll_name != NULL)
			free(_lm->base_dll_name);
		if (_lm->full_dll_name != NULL)
			free(_lm->full_dll_name);
		g_hash_table_destroy(_lm->hash);
		free(_lm);
		_lm = _next;
	}
}

status_t
windows_ldrmodules(apsh_instance_t apsh, windows_processes_list_t *process_list, apsh_pid_t pid,
		   windows_ldrmodules_list_t **ldrmodules_list)
{
	status_t retval = APSH_F;
	windows_processes_list_t *psl, *psl_allocation = NULL;
	char *procname = NULL;

	if (process_list == NULL) {
		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F)
			return APSH_F;
		psl = psl_allocation;
	} else {
		psl = process_list;
	}

	windows_eprocess_struct_t *_eps = NULL;

	if (windows_get_eprocess(apsh, &_eps) == APSH_F)
		return APSH_F;

	windows_ldrmodules_list_t *ldrmodules_head = NULL;

	/* walk the process list */
	while (psl) {
		if ((pid == 0) || (pid == psl->pid)) {
			windows_dll_list_t *inload = NULL;
			windows_dll_list_t *inmem = NULL;
			windows_dll_list_t *ininit = NULL;

			/* get inload list */
			if (windows_ldrmodules_inloadlist(apsh, psl->pid, psl->address, &inload) == APSH_F)
				goto next_process;

			/* get inmem list */
			if (windows_ldrmodules_inmemlist(apsh, psl->pid, psl->address, &inmem) == APSH_F) {
				DOCA_DLOG_ERR("failed to get inmem");
				goto next_process;
			}

			/* get ininit list */
			if (windows_ldrmodules_ininitlist(apsh, psl->pid, psl->address, &ininit) == APSH_F) {
				DOCA_DLOG_ERR("failed to get ininit");
				goto next_process;
			}

			/* foreach in inload */
			windows_dll_list_t *_dlllst = inload;

			while (_dlllst) {
				windows_ldrmodules_list_t *_lm = ldrmodule_list_put(ldrmodules_head, _dlllst->entry);
				if (ldrmodules_head == NULL)
					ldrmodules_head = _lm;

				_lm->inload = true;
				_lm->pid = psl->pid;
				_lm->address = psl->address;
				_dlllst = _dlllst->next;
			}

			/* foreach in inmem */
			_dlllst = inmem;
			while (_dlllst) {
				windows_ldrmodules_list_t *_lm =
					ldrmodule_list_put(ldrmodules_head, _dlllst->entry);

				if (ldrmodules_head == NULL)
					ldrmodules_head = _lm;

				_lm->inmem = true;
				_lm->pid = psl->pid;
				_lm->address = psl->address;
				_dlllst = _dlllst->next;
			}

			/* foreach in ininit */
			_dlllst = ininit;
			while (_dlllst) {
				windows_ldrmodules_list_t *_lm =
					ldrmodule_list_put(ldrmodules_head, _dlllst->entry);

				if (ldrmodules_head == NULL)
					ldrmodules_head = _lm;

				_lm->ininit = true;
				_lm->pid = psl->pid;
				_lm->address = psl->address;
				_dlllst = _dlllst->next;
			}

		next_process:
			windows_free_dll_list(inload);
			windows_free_dll_list(inmem);
			windows_free_dll_list(ininit);

			if (pid == psl->pid)
				break;
		}

		/* next process */
		psl = psl->next;
	}

	windows_ldrmodules_list_t *_lm = ldrmodules_head;

	addr_t SizeOfImage_offset;
	addr_t DllBase_offset;
	addr_t BaseDllName_offset;
	addr_t FullDllName_offset;

	/* get SizeOfImage_offset, BaseDllName_offset, FullDllName_offset */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "BaseDllName", &BaseDllName_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "FullDllName", &FullDllName_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "SizeOfImage", &SizeOfImage_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY", "DllBase",
							 &DllBase_offset) == APSH_F)
		goto error_exit;

	uint32_t last_pid = 0;

	while (_lm) {
		addr_t ldr_data_va = _lm->entry;
		procname = apsh_read_str_va(apsh, _lm->address + _eps->ImageFileName_offset, 0);
		last_pid = _lm->pid;

		if (procname == NULL)
			goto error_exit;

		_lm->procname = procname;

		/* get SizeOfImage */
		addr_t SizeOfImage_va = ldr_data_va + SizeOfImage_offset;
		uint32_t SizeOfImage;

		apsh_read_32_va(apsh, SizeOfImage_va, _lm->pid, &SizeOfImage);

		/* get DLLBase */
		addr_t DllBase_va = ldr_data_va + DllBase_offset;
		uint64_t DllBase = 0;

		apsh_read_64_va(apsh, DllBase_va, _lm->pid, &DllBase);

		/* get BaseDllName and FullDllName */
		addr_t BaseDllName_va = ldr_data_va + BaseDllName_offset;
		unicode_string_t *usBaseDllName = NULL;
		unicode_string_t sBaseDllName = {0};
		_lm->base_dll_name = NULL;

		usBaseDllName = apsh_read_unicode_str_va(apsh, BaseDllName_va, _lm->pid);
		if (usBaseDllName && APSH_S == apsh_convert_str_encoding(usBaseDllName, &sBaseDllName, "UTF-8")) {
			if (sBaseDllName.contents != NULL)
				_lm->base_dll_name = sBaseDllName.contents;
		}

		addr_t FullDllName_va = ldr_data_va + FullDllName_offset;
		unicode_string_t *usFullDllName = NULL;
		unicode_string_t sFullDllName = {0};
		_lm->full_dll_name = NULL;

		usFullDllName = apsh_read_unicode_str_va(apsh, FullDllName_va, _lm->pid);
		if (usFullDllName && APSH_S == apsh_convert_str_encoding(usFullDllName, &sFullDllName, "UTF-8")) {
			if (sFullDllName.contents != NULL)
				_lm->full_dll_name = sFullDllName.contents;
		}

		unicode_string_t *us1 = NULL;
		unicode_string_t out1 = {0};

		GHashTable *hash = g_hash_table_new(NULL, NULL);

		if (hash == NULL) {
			DOCA_DLOG_ERR("failed to allocate memory for privilege struct");
			goto error_exit;
		}

		_lm->hash = hash;

		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_PID), GINT_TO_POINTER(_lm->pid));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_COMM), _lm->procname);
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_BASE_ADDRESS), GINT_TO_POINTER(DllBase));

		if (_lm->base_dll_name)
			g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_WINDOWS_BASE_DLL_NAME),
					    _lm->base_dll_name);

		if (_lm->full_dll_name)
			g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_LIBRARY_PATH),
					    _lm->full_dll_name);

		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_WINDOWS_SIZE_OF_IMAGE),
				    GINT_TO_POINTER(SizeOfImage));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_WINDOWS_INLOAD),
				    GINT_TO_POINTER(_lm->inload));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_WINDOWS_INMEM),
				    GINT_TO_POINTER(_lm->inmem));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_LDRMODULE_WINDOWS_ININIT),
				    GINT_TO_POINTER(_lm->ininit));

		if (usBaseDllName != NULL)
			apsh_free_unicode_str(usBaseDllName);
		if (usFullDllName != NULL)
			apsh_free_unicode_str(usFullDllName);

		_lm = _lm->next;
	}

	*ldrmodules_list = ldrmodules_head;

	return APSH_S;

error_exit:
	if (ldrmodules_head != NULL)
		windows_free_ldrmodules_list(ldrmodules_head);

	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	return retval;
}

static status_t
getLdr(apsh_instance_t apsh, apsh_pid_t pid, addr_t procaddr, uint64_t *pldr)
{
	/* get _EPROCESS.Peb */
	addr_t Peb_offset, Peb_va;
	uint64_t Peb = 0;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Peb", &Peb_offset) ==
	    APSH_F)
		return APSH_F;

	Peb_va = procaddr + Peb_offset;
	if (apsh_read_64_va(apsh, Peb_va, 0, &Peb) == APSH_F)
		return APSH_F;

	/* get _PEB Ldr */
	addr_t Ldr_offset, Ldr_va;
	uint64_t Ldr = 0;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB", "Ldr", &Ldr_offset) ==
	    APSH_F)
		return APSH_F;

	Ldr_va = Peb + Ldr_offset;

	addr_t paddr;

	if (apsh_translate_uv2p(apsh, Peb, pid, &paddr) == APSH_F) {
		DOCA_DLOG_ERR("failed to translate Pdeb [0x%lx] to physical address", Peb);
		return APSH_F;
	}

	if (apsh_read_64_va(apsh, Ldr_va, pid, &Ldr) == APSH_F) {
		DOCA_DLOG_ERR("failed to get Ldr");
		return APSH_F;
	}

	*pldr = Ldr;

	return APSH_S;
}

static status_t
getLinkedList(apsh_instance_t apsh, apsh_pid_t pid, addr_t linkentry, addr_t linkoffset, windows_dll_list_t **dll_list)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table,
				    GINT_TO_POINTER(DOCA_APSH_LDRMODULES_LIMIT)));
	int count_dlls = 0;

	status_t retval = APSH_F;

	addr_t flink_va, flink;

	flink = linkentry;

	addr_t link_header = flink;
	windows_dll_list_t *dl_header = NULL;
	windows_dll_list_t *_dl;

	/* walk through any of the dll lists (ininit/inload/inmem) and extract the dll_list */
	do {
		if(++count_dlls >= limit)
			break;

		addr_t ldr_data_va;
		ldr_data_va = flink - linkoffset;

		if (dl_header == NULL) {
			_dl = malloc(sizeof(windows_dll_list_t));
			if (_dl == NULL) {
				DOCA_DLOG_ERR("failed to malloc");
				goto error_exit;
			}
			dl_header = _dl;
		} else {
			_dl->next = malloc(sizeof(windows_dll_list_t));
			if (_dl->next == NULL) {
				DOCA_DLOG_ERR("failed to malloc");
				goto error_exit;
			}
			_dl = _dl->next;
		}

		_dl->entry = ldr_data_va;
		_dl->next = NULL;

		/* next Flick */
		flink_va = flink;
		if (apsh_read_64_va(apsh, flink_va, pid, (uint64_t *)&flink) == APSH_F) {
			DOCA_DLOG_ERR("failed to read flink");
			goto error_exit;
		}
	} while (flink != link_header);

	remove_last_node(&dl_header);

	retval = APSH_S;

error_exit:

	*dll_list = dl_header;

	return retval;
}

status_t
windows_ldrmodules_inloadlist(apsh_instance_t apsh, apsh_pid_t pid, addr_t process_address, windows_dll_list_t **dll_list)
{
	/* get _PEB.Ldr */
	uint64_t Ldr = 0;
	if (getLdr(apsh, pid, process_address, &Ldr) == APSH_F)
		return APSH_F;

	/* get _PEB_LDR_DATA.InLoadOrderModuleList header */
	addr_t InLoadOrderModuleList_offset, InLoadOrderModuleList_va;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB_LDR_DATA",
							 "InLoadOrderModuleList",
							 &InLoadOrderModuleList_offset) == APSH_F)
		return APSH_F;

	struct _LIST_ENTRY {
		addr_t flink;
		addr_t blink;
	} InLoadOrderModuleList;

	InLoadOrderModuleList_va = Ldr + InLoadOrderModuleList_offset;
	if (apsh_read_va(apsh, InLoadOrderModuleList_va, pid, sizeof(struct _LIST_ENTRY),
				   &InLoadOrderModuleList, NULL) == APSH_F) {
		DOCA_DLOG_ERR("failed to get InLoadOrderModuleList");
		return APSH_F;
	}

	/* get _LDR_DATA_TABLE_ENTRY.InLoadOrderLinks offset */
	addr_t InLoadOrderLinks_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY",
							 "InLoadOrderLinks",
							 &InLoadOrderLinks_offset) == APSH_F)
		return APSH_F;

	/* get Flink of InLoadOrderModuleList */
	return getLinkedList(apsh, pid, InLoadOrderModuleList.flink, InLoadOrderLinks_offset, dll_list);
}

status_t
windows_ldrmodules_inmemlist(apsh_instance_t apsh, apsh_pid_t pid, addr_t process_address,\
			     windows_dll_list_t **dll_list)
{
	/* get _PEB.Ldr */
	uint64_t Ldr = 0;

	if (getLdr(apsh, pid, process_address, &Ldr) == APSH_F)
		return APSH_F;

	/* get _PEB_LDR_DATA.InMemoryOrderModuleList header */
	addr_t InMemoryOrderModuleList_offset, InMemoryOrderModuleList_va;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_PEB_LDR_DATA",
							 "InMemoryOrderModuleList",
							 &InMemoryOrderModuleList_offset) == APSH_F)
		return APSH_F;

	struct _LIST_ENTRY {
		addr_t flink;
		addr_t blink;
	} InMemoryOrderModuleList;

	InMemoryOrderModuleList_va = Ldr + InMemoryOrderModuleList_offset;
	if (apsh_read_va(apsh, InMemoryOrderModuleList_va, pid,
				   sizeof(struct _LIST_ENTRY), &InMemoryOrderModuleList, NULL) == APSH_F) {
		DOCA_DLOG_ERR("failed to get InLoadOrderModuleList");
		return APSH_F;
	}

	addr_t InMemoryOrderLinks_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_LDR_DATA_TABLE_ENTRY",
							 "InMemoryOrderLinks",
							 &InMemoryOrderLinks_offset) == APSH_F)
		return APSH_F;

	/* get Flink of InLoadOrderModuleList */
	return getLinkedList(apsh, pid, InMemoryOrderModuleList.flink, InMemoryOrderLinks_offset, dll_list);
}

status_t
windows_ldrmodules_ininitlist(apsh_instance_t apsh, apsh_pid_t pid, addr_t process_address, windows_dll_list_t **dll_list)
{
	/* get _PEB.Ldr */
	uint64_t Ldr = 0;

	if (getLdr(apsh, pid, process_address, &Ldr) == APSH_F)
		return APSH_F;

	/* get _PEB_LDR_DATA.InInitializationOrderModuleList header */
	addr_t InInitializationOrderModuleList_offset, InInitializationOrderModuleList_va;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_PEB_LDR_DATA", "InInitializationOrderModuleList",
		    &InInitializationOrderModuleList_offset) == APSH_F)
		return APSH_F;

	struct _LIST_ENTRY {
		addr_t flink;
		addr_t blink;
	} InInitializationOrderModuleList;

	InInitializationOrderModuleList_va = Ldr + InInitializationOrderModuleList_offset;
	if (apsh_read_va(apsh, InInitializationOrderModuleList_va, pid,
				   sizeof(struct _LIST_ENTRY), &InInitializationOrderModuleList,
				   NULL) == APSH_F) {
		DOCA_DLOG_ERR("failed to get InLoadOrderModuleList");
		return APSH_F;
	}

	addr_t InInitializationOrderLinks_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_LDR_DATA_TABLE_ENTRY", "InInitializationOrderLinks",
		    &InInitializationOrderLinks_offset) == APSH_F)
		return APSH_F;

	/* get Flink of InLoadOrderModuleList */
	return getLinkedList(apsh, pid, InInitializationOrderModuleList.flink, InInitializationOrderLinks_offset,
			     dll_list);
}

void
windows_free_dll_list(windows_dll_list_t *dll_list)
{
	windows_dll_list_t *_dl = dll_list;

	while (_dl) {
		windows_dll_list_t *_next = _dl->next;

		free(_dl);
		_dl = _next;
	}
}
