/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "modules.h"
#include "dlllist.h"
#include "doca_log.h"
#include "../../doca_apsh_attr.h"
#include "../../../core/APSH/apsh_instance.h"

DOCA_LOG_REGISTER(APSH::Windows::MODULES);

status_t
windows_save_kernel_exports(apsh_instance_t apsh)
{
	status_t retval = APSH_F;
	windows_module_list_t *module_list;

	if (windows_modules_inloadlist(apsh, &module_list) == APSH_F)
		goto error_exit;

	addr_t SizeOfImage_offset, DllBase_offset, BaseDllName_offset, FullDllName_offset;

	/* _KLDR_DATA_TABLE_ENTRY. SizeOfImage */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "SizeOfImage", &SizeOfImage_offset) == APSH_F)
		goto error_exit;

	/* _KLDR_DATA_TABLE_ENTRY. DllBase */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_KLDR_DATA_TABLE_ENTRY", "DllBase",
							 &DllBase_offset) == APSH_F)
		goto error_exit;

	/* _KLDR_DATA_TABLE_ENTRY. BaseDllName */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "BaseDllName", &BaseDllName_offset) == APSH_F)
		goto error_exit;

	/* _KLDR_DATA_TABLE_ENTRY. FullDllName */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "FullDllName", &FullDllName_offset) == APSH_F)
		goto error_exit;

	/* walk through module list */
	windows_module_list_t *_ml = module_list;

	while (_ml) {
		addr_t kldr_data_va = _ml->entry;

		/* get DllBase */
		addr_t DllBase_va, DllBase = 0;

		DllBase_va = kldr_data_va + DllBase_offset;
		apsh_read_64_va(apsh, DllBase_va, 0, &DllBase);

		/* get SizeOfImage */
		addr_t SizeOfImage_va;
		uint32_t SizeOfImage = 0;

		SizeOfImage_va = kldr_data_va + SizeOfImage_offset;
		apsh_read_32_va(apsh, SizeOfImage_va, 0, &SizeOfImage);

		/* get FullDllName */
		addr_t FullDllName_va = kldr_data_va + FullDllName_offset;
		unicode_string_t *us1 = NULL;
		unicode_string_t out1 = {0};

		us1 = apsh_read_unicode_str_va(apsh, FullDllName_va, 0);

		if (us1 && APSH_S == apsh_convert_str_encoding(us1, &out1, "UTF-8")) {
			if (out1.contents) {
				(void)windows_store_dll_functions(apsh, DllBase, SizeOfImage, 0,
							    out1.contents);
			}
		}

		if (us1)
			apsh_free_unicode_str(us1);

		_ml = _ml->next;
	}

	retval = APSH_S;
error_exit:
	windows_free_module_list(module_list);
	return retval;
}

status_t
windows_modules(apsh_instance_t apsh, json_object **json_obj, bool printout)
{
	status_t retval = APSH_F;
	windows_module_list_t *module_list;

	if (windows_modules_inloadlist(apsh, &module_list) == APSH_F)
		goto error_exit;

	addr_t SizeOfImage_offset, DllBase_offset, BaseDllName_offset, FullDllName_offset;

	/* _KLDR_DATA_TABLE_ENTRY. SizeOfImage */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "SizeOfImage", &SizeOfImage_offset) == APSH_F)
		goto error_exit;

	/* _KLDR_DATA_TABLE_ENTRY. DllBase */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_KLDR_DATA_TABLE_ENTRY", "DllBase",
							 &DllBase_offset) == APSH_F)
		goto error_exit;

	/* _KLDR_DATA_TABLE_ENTRY. BaseDllName */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "BaseDllName", &BaseDllName_offset) == APSH_F)
		goto error_exit;

	/* _KLDR_DATA_TABLE_ENTRY. FullDllName */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "FullDllName", &FullDllName_offset) == APSH_F)
		goto error_exit;

	/* create new jsonout object */
	static const char *titles[] = {"Offset", "Base", "Size", "Name", "Path", "\0"};
	json_object *jout = jsonout_new("modules", titles);

	/* walk through module list */
	windows_module_list_t *_ml = module_list;

	while (_ml) {
		addr_t kldr_data_va = _ml->entry;
		json_object *jdata = jsonout_data_element_new(jout);

		jsonout_data_element_value_sprintf(jdata, "0x%lx", kldr_data_va); /* Offset */

		/* get DllBase */
		addr_t DllBase_va, DllBase = 0;

		DllBase_va = kldr_data_va + DllBase_offset;
		apsh_read_64_va(apsh, DllBase_va, 0, &DllBase);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", DllBase); /* DLLBase */

		/* get SizeOfImage */
		addr_t SizeOfImage_va;
		uint32_t SizeOfImage = 0;

		SizeOfImage_va = kldr_data_va + SizeOfImage_offset;
		apsh_read_32_va(apsh, SizeOfImage_va, 0, &SizeOfImage);
		jsonout_data_element_value_sprintf(jdata, "0x%x", SizeOfImage); /* Size */

		/* get BaseDllNmae */
		addr_t BaseDllName_va = kldr_data_va + BaseDllName_offset;
		unicode_string_t *us = NULL;

		us = apsh_read_unicode_str_va(apsh, BaseDllName_va, 0);
		jsonout_data_element_unicode_add(jdata, us); /* BaseDllName */
		if (us)
			apsh_free_unicode_str(us);

		/* get FullDllName */
		addr_t FullDllName_va = kldr_data_va + FullDllName_offset;
		unicode_string_t *us1 = NULL;

		us1 = apsh_read_unicode_str_va(apsh, FullDllName_va, 0);
		jsonout_data_element_unicode_add(jdata, us1);
		if (us1)
			apsh_free_unicode_str(us1);

		_ml = _ml->next;
	}

	/* convert jsonout to string */
	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	retval = APSH_S;

error_exit:
	windows_free_module_list(module_list);
	return retval;
}

status_t
windows_modules_inloadlist(apsh_instance_t apsh, windows_module_list_t **modules_list)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_MODULES_LIMIT)));

	status_t retval = APSH_F;

	addr_t PsLoadedModuleList;
	addr_t InLoadOrderLinks_offset;
	addr_t SizeOfImage_offset, DllBase_offset, BaseDllName_offset, FullDllName_offset;

	*modules_list = NULL;

	/* get PsLoadedModuleList from global symbol "PsLoadedModuleList" */
	if (apsh_read_addr_ksym(apsh, "PsLoadedModuleList", &PsLoadedModuleList) == APSH_F)
		return APSH_F;

	/* _KLDR_DATA_TABLE_ENTRY. InLoadOrderLinks */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_KLDR_DATA_TABLE_ENTRY",
							 "InLoadOrderLinks",
							 &InLoadOrderLinks_offset) == APSH_F)
		return APSH_F;

	/* _KLDR_DATA_TABLE_ENTRY. SizeOfImage */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "SizeOfImage", &SizeOfImage_offset) == APSH_F)
		return APSH_F;

	/* _KLDR_DATA_TABLE_ENTRY. DllBase */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_KLDR_DATA_TABLE_ENTRY", "DllBase",
							 &DllBase_offset) == APSH_F)
		return APSH_F;

	/* _KLDR_DATA_TABLE_ENTRY. BaseDllName */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "BaseDllName", &BaseDllName_offset) == APSH_F)
		return APSH_F;

	/* _KLDR_DATA_TABLE_ENTRY. FullDllName */
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_KLDR_DATA_TABLE_ENTRY", "FullDllName", &FullDllName_offset) == APSH_F)
		return APSH_F;

	/* walk through list */
	addr_t flink = PsLoadedModuleList;
	windows_module_list_t *_ml_header = NULL;
	windows_module_list_t *_prev = NULL;
	uint64_t count = 0;

	do {
		addr_t current_entry = flink - InLoadOrderLinks_offset;
		windows_module_list_t *_ml = malloc(sizeof(windows_module_list_t));

		if (_ml == NULL) {
			DOCA_DLOG_ERR("failed to malloc");
			goto error_exit;
		}

		_ml->entry = current_entry;
		_ml->next = NULL;
		_ml->hash = NULL;

		if (_ml_header == NULL)
			_ml_header = _ml;

		if (_prev)
			_prev->next = _ml;

		_prev = _ml;

		/* get SizeOfImage */
		addr_t SizeOfImage_va;
		uint32_t SizeOfImage = 0;

		SizeOfImage_va = current_entry + SizeOfImage_offset;
		apsh_read_32_va(apsh, SizeOfImage_va, 0, &SizeOfImage);

		/* get FullDllName */
		addr_t FullDllName_va = current_entry + FullDllName_offset;

		unicode_string_t *name = NULL;
		unicode_string_t out = {0};
		char *name_utf8;
		char *module_name = NULL;

		name = apsh_read_unicode_str_va(apsh, FullDllName_va, 0);

		if (name && apsh_convert_str_encoding(name, &out, "UTF-8") == APSH_S) {
			module_name = g_strndup(out.contents, strlen(out.contents) + 1);
			free(out.contents);
		}

		if (name)
			apsh_free_unicode_str(name);

		if (module_name == NULL)
			goto error_exit;

		_ml->hash = g_hash_table_new(NULL, NULL);

		if (_ml->hash == NULL) {
			free(module_name);
			goto error_exit;
		}

		g_hash_table_insert(_ml->hash,
					GINT_TO_POINTER(DOCA_APSH_MODULES_OFFSET),
					GUINT_TO_POINTER(current_entry));
		g_hash_table_insert(_ml->hash,
					GINT_TO_POINTER(DOCA_APSH_MODULES_NAME),
					module_name);
		g_hash_table_insert(_ml->hash, GINT_TO_POINTER(DOCA_APSH_MODULES_SIZE),
			GUINT_TO_POINTER(SizeOfImage));

		/* next link */
		struct _LIST_ENTRY {
			addr_t flink;
			addr_t blink;
		} InLoadOrderModuleList;

		if (apsh_read_va(apsh, flink, 0, sizeof(struct _LIST_ENTRY),
					   &InLoadOrderModuleList, NULL) == APSH_F) {
			DOCA_DLOG_ERR("failed to get InLoadOrderModuleList");
			goto error_exit;
		}

		flink = InLoadOrderModuleList.flink;

	} while (flink != PsLoadedModuleList && ++count < limit);

	retval = APSH_S;
	*modules_list = _ml_header;

	return retval;

error_exit:
	if (_ml_header != NULL)
		windows_free_module_list(_ml_header);

	*modules_list = NULL;

	return retval;
}

void
windows_free_module_list(windows_module_list_t *modules_list)
{
	windows_module_list_t *_ml = modules_list;

	while (_ml) {
		windows_module_list_t *_next = _ml->next;

		if (_ml->hash) {
			char *name = g_hash_table_lookup(_ml->hash, GINT_TO_POINTER(DOCA_APSH_MODULES_NAME));

			g_free(name);
			g_hash_table_destroy(_ml->hash);
		}
		free(_ml);
		_ml = _next;
	}
}
