/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#include <math.h>
#include "private/common_defines.h"
#include "privs.h"
#include "../../doca_apsh_attr.h"

#include "doca_log.h"

#define NAME_LENGTH 64

DOCA_LOG_REGISTER(APSH::Windows::PRIVS);

struct _Token_Privs {
	uint64_t Present;
	uint64_t Enabled;
	uint64_t EnabledByDefault;
};

struct _privs {
	int index;
	char name[NAME_LENGTH];
};

static const struct _privs windows_privileges[] = {
	{2, "SeCreateTokenPrivilege"},
	{3, "SeAssignPrimaryTokenPrivilege"},
	{4, "SeLockMemoryPrivilege"},
	{5, "SeIncreaseQuotaPrivilege"},
	{6, "SeMachineAccountPrivilege"},
	{7, "SeTcbPrivilege"},
	{8, "SeSecurityPrivilege"},
	{9, "SeTakeOwnershipPrivilege"},
	{10, "SeLoadDriverPrivilege"},
	{11, "SeSystemProfilePrivilege"},
	{12, "SeSystemtimePrivilege"},
	{13, "SeProfileSingleProcessPrivilege"},
	{14, "SeIncreaseBasePriorityPrivilege"},
	{15, "SeCreatePagefilePrivilege"},
	{16, "SeCreatePermanentPrivilege"},
	{17, "SeBackupPrivilege"},
	{18, "SeRestorePrivilege"},
	{19, "SeShutdownPrivilege"},
	{20, "SeDebugPrivilege"},
	{21, "SeAuditPrivilege"},
	{22, "SeSystemEnvironmentPrivilege"},
	{23, "SeChangeNotifyPrivilege"},
	{24, "SeRemoteShutdownPrivilege"},
	{25, "SeUndockPrivilege"},
	{26, "SeSyncAgentPrivilege"},
	{27, "SeEnableDelegationPrivilege"},
	{28, "SeManageVolumePrivilege"},
	{29, "SeImpersonatePrivilege"},
	{30, "SeCreateGlobalPrivilege"},
	{31, "SeTrustedCredManAccessPrivilege"},
	{32, "SeRelabelPrivilege"},
	{33, "SeIncreaseWorkingSetPrivilege"},
	{34, "SeTimeZonePrivilege"},
	{35, "SeCreateSymbolicLinkPrivilege"},
	{36, "SeDelegateSessionUserImpersonatePrivilege"},
};

static char *
get_privs_info(int priv_index)
{
	for (int i = 0; i < ARRAY_SIZE(windows_privileges); i++) {
		if (windows_privileges[i].index == priv_index)
			return windows_privileges[i].name;
	}

	return NULL;
}

status_t
privs(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid,
	windows_privileges_list_t **privileges_list)
{
	status_t retval = APSH_F;
	windows_processes_list_t *psl = NULL, *psl_allocation = NULL;

	windows_privileges_list_t *privs_head = NULL;

	if (process_list == NULL) {
		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F) {
			DOCA_DLOG_ERR("Failed to get process list");
			return retval;
		}
		psl = psl_allocation;
	} else {
		psl = process_list;
	}

	/* get _EPROCSS.Token offset */
	addr_t token_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Token", &token_offset) == APSH_F)
		goto error_exit;

	/* get _TOKEN.Privileges offset */
	addr_t privileges_offset;


	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_TOKEN", "Privileges", &privileges_offset) == APSH_F)
		goto error_exit;

	/* get _EPROCSS.ImageFileName offset */
	addr_t file_name_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName", &file_name_offset) == APSH_F)
		goto error_exit;

	/* walk through process list */
	while (psl) {
		char *procname = NULL;

		if (pid == 0 || pid == psl->pid) {
			uint32_t _pid = psl->pid;
			addr_t process_address = psl->address;

			/* get token pointer from process */
			addr_t fast_ref_va;
			addr_t fast_ref;
			addr_t token_va;
			addr_t file_name_va;
			uint64_t token_pid = 0;

			file_name_va = process_address + file_name_offset;
			procname = apsh_read_str_va(apsh, file_name_va, 0);
			if (procname == NULL)
				goto error_exit;

			fast_ref_va = process_address + token_offset;
			if (apsh_read_64_va(apsh, fast_ref_va, psl->pid, &fast_ref) == APSH_F)
				goto next_process;

			token_va = fast_ref & 0xFFFFFFFFFFFFFFF0L;

			/* get privileges from token */
			addr_t privs_va = token_va + privileges_offset;
			struct _Token_Privs privs;

			if (apsh_read_va(apsh, privs_va, psl->pid, sizeof(struct _Token_Privs), &privs, NULL) == APSH_F)  // &bytes_read
				goto next_process;

			/* parse privileges */
			uint32_t priv_index;
			uint64_t masked = 1;  /* 2**0 */

			windows_privileges_list_t *prev_priv;

			const uint32_t num_of_privs = 64; /* for vista+ */

			for (priv_index = 0; priv_index < num_of_privs; priv_index++) {
				bool present = privs.Present & masked;
				bool enabled = privs.Enabled & masked;
				bool enabled_by_default = privs.EnabledByDefault & masked;

				masked <<= 1;  /* *=2 */

				char *name = get_privs_info(priv_index);

				if (name == NULL)
					continue;

				bool is_on = present && (enabled || enabled_by_default);

				windows_privileges_list_t *priv = calloc(sizeof(*priv), 1);

				if (priv == NULL) {
					DOCA_DLOG_ERR("failed to allocate memory for privilege struct");
					free(procname);
					goto error_exit;
				}

				priv->next = NULL;
				priv->procname = g_strdup(procname);
				if (priv->procname == NULL) {
					DOCA_DLOG_ERR("failed to allocate memory for privilege struct");
					free(procname);
					free(priv);
					goto error_exit;
				}

				priv->name = g_strdup(name);
				if (priv->name == NULL) {
					DOCA_DLOG_ERR("failed to allocate memory for privilege struct");
					free(procname);
					free(priv->procname);
					free(priv);
					goto error_exit;
				}

				priv->present = present;
				priv->enabled = enabled;
				priv->enabled_by_default = enabled_by_default;
				priv->is_on = is_on;
				priv->pid = _pid;

				GHashTable *hash = g_hash_table_new(NULL, NULL);

				if (hash == NULL) {
					DOCA_DLOG_ERR("failed to allocate memory for privilege struct");
					free(procname);
					free(priv->procname);
					g_free(priv->name);
					free(priv);
					goto error_exit;
				}
				priv->hash = hash;

				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_PRIVILEGES_PID),
						    GINT_TO_POINTER(psl->pid));
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_PRIVILEGES_COMM),
						    priv->procname);
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_PRIVILEGES_NAME),
						    priv->name);
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_PRIVILEGES_IS_ON),
						    GINT_TO_POINTER(priv->is_on));
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_PRIVILEGES_WINDOWS_PRESENT),
						    GINT_TO_POINTER(priv->present));
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_PRIVILEGES_WINDOWS_ENABLED),
						    GINT_TO_POINTER(priv->enabled));
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_PRIVILEGES_WINDOWS_DEFAULT),
						    GINT_TO_POINTER(priv->enabled_by_default));

				if (privs_head == NULL)
					privs_head = priv;
				else
					prev_priv->next = priv;

				prev_priv = priv;
			}
		}

next_process:
		if (procname != NULL)
			free(procname);

		if (pid == psl->pid)
			break;

		/* next process */
		psl = psl->next;
	}

	*privileges_list = privs_head;

	return APSH_S;

error_exit:

	if (privs_head != NULL)
		windows_free_privileges_list(privs_head);

	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	return retval;

}

void
windows_free_privileges_list(windows_privileges_list_t *priv_list)
{
	windows_privileges_list_t *p_next = NULL;

	while (priv_list) {
		p_next = priv_list->next;
		if (priv_list->name)
			g_free(priv_list->name);
		if (priv_list->procname)
			g_free(priv_list->procname);
		if (priv_list->hash)
			g_hash_table_destroy(priv_list->hash);
		free(priv_list);
		priv_list = p_next;
	}
}
