/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "pslist.h"
#include "jsonout.h"

#include "../../doca_apsh_attr.h"
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::Windows::PSLIST);

#define BIT_MASK(__TYPE__, __ONE_COUNT__) \
	(((__TYPE__)(-((__ONE_COUNT__) != 0))) & (((__TYPE__)-1) >> ((sizeof(__TYPE__) * CHAR_BIT) - (__ONE_COUNT__))))

status_t
windows_pslist_init_eprocess_struct(apsh_instance_t apsh, windows_eprocess_struct_t *eprocess_struct)
{
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
							 &eprocess_struct->ImageFileName_offset) ==
	    APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ActiveThreads",
							 &eprocess_struct->ActiveThreads_offset) ==
	    APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ObjectTable",
							 &eprocess_struct->ObjectTable_offset) ==
	    APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_EPROCESS", "Session", &eprocess_struct->Session_offset) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_EPROCESS", "InheritedFromUniqueProcessId",
		    &eprocess_struct->InheritedFromUniqueProcessId_offset) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "WoW64Process",
							 &eprocess_struct->WoW64Process_offset) ==
	    APSH_F)
		if (apsh_get_struct_member_offset_from_apsh_json(
			    apsh, "_EPROCESS", "Wow64Process",
			    &eprocess_struct->WoW64Process_offset) == APSH_F)
			return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_EPROCESS", "CreateTime", &eprocess_struct->CreateTime_offset) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_EPROCESS", "ExitTime", &eprocess_struct->ExitTime_offset) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Peb",
							 &eprocess_struct->Peb_offset) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_HANDLE_TABLE", "HandleCount",
							 &eprocess_struct->HandleCount_offset) ==
	    APSH_F) {
		eprocess_struct->has_handlecount = false;
	} else {
		eprocess_struct->has_handlecount = true;
	}
	return APSH_S;
}

status_t
windows_get_eprocess(apsh_instance_t apsh, windows_eprocess_struct_t **eps)
{
	windows_eprocess_struct_t *_eps = (windows_eprocess_struct_t *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_EPROCESS);

	if (_eps == NULL) {
		_eps = malloc(sizeof(*_eps));

		if (_eps == NULL) {
			DOCA_DLOG_ERR("Failed to allocate plugin context memory");
			return APSH_F;
		}
		windows_pslist_init_eprocess_struct(apsh, _eps);

		if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_EPROCESS, _eps, free)) {
			free(_eps);
			return APSH_F;
		}
	}
	*eps = _eps;
	return APSH_S;
}

windows_processes_list_t *
windows_pslist_get_process(windows_processes_list_t *process_list, uint32_t pid)
{
	if (process_list == NULL)
		return NULL;

	windows_processes_list_t *psl = process_list;

	while (psl) {
		if (psl->pid == pid)
			return psl;

		psl = psl->next;
	}
	return NULL;
}

status_t
windows_pslist(apsh_instance_t apsh, windows_processes_list_t *process_list, json_object **json_obj, bool printout)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_PROCESS_LIMIT)));

	status_t retval = APSH_F;
	windows_processes_list_t *psl, *psl_allocation = NULL;

	if (process_list == NULL) {
		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F)
			return APSH_F;
		psl = psl_allocation;
	} else {
		psl = process_list;
	}

	bool has_handlecount = false;

	addr_t ImageFileName_offset;
	addr_t ActiveThreads_offset;
	addr_t ObjectTable_offset;
	addr_t Session_offset;
	addr_t InheritedFromUniqueProcessId_offset;
	addr_t WoW64Process_offset;
	addr_t CreateTime_offset;
	addr_t ExitTime_offset;
	addr_t HandleCount_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
							 &ImageFileName_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ActiveThreads",
							 &ActiveThreads_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ObjectTable",
							 &ObjectTable_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Session",
							 &Session_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_EPROCESS", "InheritedFromUniqueProcessId",
		    &InheritedFromUniqueProcessId_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "WoW64Process",
							 &WoW64Process_offset) == APSH_F)
		if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "Wow64Process",
								 &WoW64Process_offset) == APSH_F)
			goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "CreateTime",
							 &CreateTime_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ExitTime",
							 &ExitTime_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_HANDLE_TABLE", "HandleCount",
							 &HandleCount_offset) == APSH_F) {
		has_handlecount = false;
	} else {
		has_handlecount = true;
	}

	bool has_sessionid = false;
	char *procname = NULL;
	apsh_pid_t pid = 0;
	uint32_t threads;
	uint32_t sessionid;
	uint32_t ppid;
	addr_t current_process = 0;
	addr_t handle_table;
	addr_t session;
	addr_t WoW64Process;
	uint64_t count = 0;

	addr_t PsActiveProcessHead = psl->address;

	/* create new jsonout object */
	char *titles[] = {"PID",       "PPID",	"ImageFileName", "Offset(V)", "Threads", "Handles",
			  "SessionId", "Wow64", "CreateTime",	 "ExitTime",  "\0"};
	json_object *jout = jsonout_new("pslist", titles);

	while (count++ < limit) {
		pid = psl->pid;
		current_process = psl->address;

		procname = apsh_read_str_va(apsh, current_process + ImageFileName_offset,
					    0); /* ImageFileName */
		if (!procname) {
			DOCA_DLOG_ERR("Failed to find ImageFileName");
			goto error_exit;
		}

		if (apsh_read_32_va(apsh, current_process + InheritedFromUniqueProcessId_offset, 0,
				&ppid) == APSH_F) { /* PPID */
			DOCA_DLOG_ERR("Failed to read PPID");
			free(procname);
			goto error_exit;
		}

		if (apsh_read_32_va(apsh, current_process + ActiveThreads_offset, 0,
				&threads) == APSH_F) { /* Threads */
			DOCA_DLOG_ERR("Failed to read ActiveThreads");
			free(procname);
			goto error_exit;
		}

		/* session id */
		apsh_read_addr_va(apsh, current_process + Session_offset, 0, &session); /* session */
		if (session) {
			addr_t sessionid_offset;

			if (apsh_get_struct_member_offset_from_apsh_json(
				    apsh, "_MM_SESSION_SPACE", "SessionId", &sessionid_offset) ==
			    APSH_S) {
				if (apsh_read_32_va(apsh, session + sessionid_offset, 0,
						    &sessionid) == APSH_S)
					has_sessionid = true;
			}
		}

		/* WoW64Process */
		if (apsh_read_addr_va(apsh, current_process + WoW64Process_offset, 0,
						&WoW64Process) == APSH_F) {
			WoW64Process = 0;
		}

		uint64_t create_time, exit_time;

		apsh_read_64_va(apsh, current_process + CreateTime_offset, 0, &create_time);
		apsh_read_64_va(apsh, current_process + ExitTime_offset, 0, &exit_time);

		json_object *jdata = jsonout_data_element_new(jout);

		/* print out the process pid, name and offset */
		jsonout_data_element_value_sprintf(jdata, "%d", pid);
		jsonout_data_element_value_sprintf(jdata, "%d", ppid);
		jsonout_data_element_value_add(jdata, procname);
		jsonout_data_element_value_sprintf(jdata, "0x%lx", current_process);
		jsonout_data_element_value_sprintf(jdata, "%d", threads);

		free(procname);

		if (has_handlecount)
			jsonout_data_element_value_add(jdata, "()");
		else
			jsonout_data_element_value_add(jdata, "-");

		if (has_sessionid)
			jsonout_data_element_value_sprintf(jdata, "%d",
							   sessionid);
		else
			jsonout_data_element_value_add(jdata, "N/A");

		if (WoW64Process)
			jsonout_data_element_value_add(jdata, "True");
		else
			jsonout_data_element_value_add(jdata, "False");

		jsonout_data_element_value_sprintf(jdata, "0x%lx",
						   create_time);
		if (exit_time)
			jsonout_data_element_value_sprintf(
				jdata, "0x%lx", exit_time);
		else
			jsonout_data_element_value_add(jdata, "N/A");


		/* follow the next pointer */
		if (psl->next == NULL)
			break;

		psl = psl->next;

		if (psl->address == PsActiveProcessHead)
			break;
	}

	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	retval = APSH_S;

error_exit:
	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	return retval;
}

status_t
windows_pslist_processes_list(apsh_instance_t apsh, windows_processes_list_t **pprocess_list)
{
	status_t retval = APSH_F;

	addr_t ActiveProcessLinks_offset;
	addr_t ImageFileName_offset;
	addr_t UniqueProcessId_offset;
	addr_t PsActiveProcessHead;
	addr_t InheritedFromUniqueProcessId_offset;
	addr_t ExitTime_offset;
	addr_t ProcessExecutionState_offset;
	size_t ProcessExecutionState_start_bit;
	size_t ProcessExecutionState_end_bit;
	addr_t ProcessExecutionTime_offset;
	addr_t ActiveThreads_offset;

	windows_processes_list_t *psl = NULL;
	windows_processes_list_t *prev_psl = NULL;

	/* get offsets in _EPROCESS */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ActiveProcessLinks",
							 &ActiveProcessLinks_offset) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "UniqueProcessId",
							 &UniqueProcessId_offset) == APSH_F)
		return APSH_F;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
							 &ImageFileName_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(
		    apsh, "_EPROCESS", "InheritedFromUniqueProcessId",
		    &InheritedFromUniqueProcessId_offset) == APSH_F)
		goto error_exit_no_free;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ExitTime",
							 &ExitTime_offset) == APSH_F)
		goto error_exit_no_free;

	if (apsh_get_bitfield_offset_and_size_from_json(
			apsh, apsh_get_kernel_json(apsh), "_EPROCESS", "ProcessExecutionState",
			&ProcessExecutionState_offset, &ProcessExecutionState_start_bit,
			&ProcessExecutionState_end_bit) == APSH_F) {
		goto error_exit_no_free;
	}


	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_KPROCESS", "CycleTime",
							 &ProcessExecutionTime_offset) == APSH_F)
		goto error_exit_no_free;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ActiveThreads",
							 &ActiveThreads_offset) == APSH_F)
		goto error_exit_no_free;

	/* get first ActiveProcessLink head (VA) from global symbol 'PsActiveProcessHead' */
	if (apsh_read_addr_ksym(apsh, "PsActiveProcessHead", &PsActiveProcessHead) == APSH_F)
		return APSH_F;

	addr_t current_entry = PsActiveProcessHead;
	addr_t next_entry = 0;
	addr_t current_process = 0;

	psl = malloc(sizeof(windows_processes_list_t));

	*pprocess_list = psl;

	while (1) {

		/* calculating current _EPROCESS address */
		current_process = current_entry - ActiveProcessLinks_offset;

		/*
		 * Note: the task_struct that we are looking at has a lot of
		 * information.  However, the process name and id are burried
		 * nice and deep.  Instead of doing something sane like mapping
		 * this data to a task_struct, I'm just jumping to the location
		 * with the info that I want.  This helps to make the example
		 * code cleaner, if not more fragile.  In a real app, you'd
		 * want to do this a little more robust :-)  See
		 * include/linux/sched.h for mode details
		 */

		/*
		 * NOTE: _EPROCESS.UniqueProcessId is a really VOID*, but is never > 32 bits,
		 * so this is safe enough for x64 Windows for example purposes
		 */

		psl->address = current_process;
		psl->next = NULL;

		/* get PID from current _EPROCESS */
		if (apsh_read_32_va(apsh, current_process + UniqueProcessId_offset, 0,
				(uint32_t *)&psl->pid) == APSH_F) { /* PID */
			DOCA_DLOG_ERR("Failed to read PID");
			goto error_exit_free_list;
		}

		if (apsh_read_32_va(apsh, current_process + InheritedFromUniqueProcessId_offset, 0,
				&psl->ppid) == APSH_F) { /* PPID */
			DOCA_DLOG_ERR("Failed to read PPID");
			goto error_exit_free_list;
		}

		uint32_t ProcessExecutionState;

		if (apsh_read_32_va(apsh, current_process + ProcessExecutionState_offset, 0,
				&ProcessExecutionState) == APSH_F) { /* ProcessExecutionState */
			DOCA_DLOG_ERR("Failed to read ProcessExecutionState");
			goto error_exit_free_list;
		}

		psl->state = ((ProcessExecutionState) >> ProcessExecutionState_start_bit) &
			     BIT_MASK(uint32_t, (ProcessExecutionState_end_bit - ProcessExecutionState_start_bit));

		psl->name = apsh_read_str_va(apsh, current_process + ImageFileName_offset,
					     0); /* ImageFileName */
		if (psl->name == NULL) {
			DOCA_DLOG_ERR("Failed to read ImageFileName");
			goto error_exit_free_list;
		}

		if (apsh_read_64_va(apsh, current_process + ExitTime_offset, 0, &psl->exit_time) == APSH_F) {
			DOCA_DLOG_ERR("Failed to read ExitTime");
			goto error_exit_free_list;
		}

		if (apsh_read_64_va(apsh, current_process + ProcessExecutionTime_offset, 0, &psl->cpu_time) == APSH_F) {
			DOCA_DLOG_ERR("Failed to read ProcessExecutionTime");
			goto error_exit_free_list;
		}

		if (apsh_read_32_va(apsh, current_process + ActiveThreads_offset, 0, &psl->threads_count) == APSH_F) {
			DOCA_DLOG_ERR("Failed to read ActiveThreads");
			goto error_exit_free_list;
		}

		GHashTable *hash = g_hash_table_new(NULL, NULL);

		if (hash == NULL)
			goto error_exit_free_list;

		char *proc_name_copy = g_strndup(psl->name, strlen(psl->name) + 1);

		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_PID),
				    GINT_TO_POINTER(psl->pid));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_PPID),
				    GINT_TO_POINTER(psl->ppid));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_WINDOWS_OFFSET),
				    GINT_TO_POINTER(psl->address));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_STATE),
				    GINT_TO_POINTER(psl->state));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_COMM),
				    proc_name_copy);
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_WINDOWS_THREADS),
				    GINT_TO_POINTER(psl->threads_count));
		g_hash_table_insert(hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_CPU_TIME),
				    GINT_TO_POINTER(psl->cpu_time));

		psl->hash = hash;

		/* follow the next pointer */
		if (apsh_read_addr_va(apsh, current_entry, 0, &next_entry) == APSH_F) {
			DOCA_DLOG_ERR("Failed to read next pointer at %" PRIx64, current_entry);
			goto error_exit_free_list;
		}

		/*
		 * In Windows, the next pointer points to the head of list, this pointer is actually
		 * the address of PsActiveProcessHead symbol, not the address of an
		 * ActiveProcessLink in EPROCESS struct. It means in Windows, we should stop the
		 * loop at the last element in the list, while
		 */
		if (next_entry == PsActiveProcessHead)
			break;

		current_entry = next_entry;

		psl->next = malloc(sizeof(windows_processes_list_t));
		prev_psl = psl;
		psl = psl->next;
	}

	/* should always enter this */
	if (prev_psl != NULL) {
		prev_psl->next = NULL;
		if (psl->hash != NULL) {
			char *proc_name_cpy = g_hash_table_lookup(psl->hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_COMM));

			if (proc_name_cpy != NULL)
				g_free(proc_name_cpy);
			g_hash_table_destroy(psl->hash);
		}
		if (psl->name != NULL)
			free(psl->name);
		free(psl);
	}

	retval = APSH_S;
	return retval;

error_exit_free_list:
	windows_free_process_list(*pprocess_list);
	*pprocess_list = NULL;
error_exit_no_free:
	return retval;
}

void
windows_free_process_list(windows_processes_list_t *process_list)
{
	if (process_list) {
		windows_processes_list_t *p = process_list;
		windows_processes_list_t *p_next;

		while (p) {
			p_next = p->next;
			char *proc_name_cpy = g_hash_table_lookup(p->hash, GINT_TO_POINTER(DOCA_APSH_PROCESS_COMM));

			if (proc_name_cpy != NULL)
				g_free(proc_name_cpy);
			g_hash_table_destroy(p->hash);
			if (p->name != NULL)
				free(p->name);
			free(p);
			p = p_next;
		}
	}
}
