/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "threadlist.h"
#include "../../doca_apsh_attr.h"
#include "dlllist.h"
#include "jsonout.h"
#include "modules.h"
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::Windows::THREADLIST);

static bool
is_in(addr_t v, GArray *s)
{
	addr_t t;
	int i = 0;

	for (i = 0; i < 1024; i++) {
		t = g_array_index(s, addr_t, i);

		if (t == 0)
			break;

		if (v == t)
			return true;
	}

	return false;
}

static status_t
walk_stalk(apsh_instance_t apsh, addr_t stack_base, addr_t rsp, apsh_pid_t pid)
{
	size_t stack_size = stack_base - rsp;

	if (stack_size < 0x20000) {
		size_t bytes_read = 0;
		uint8_t buf[stack_size + 1];
		/*
		 * read up to stack_size bytes of stack, this may fail returning a partial result
		 * thus, the following for loop analyzes the buffer only up to the `bytes_read`
		 * value
		 */
		(void)apsh_read_va(apsh, rsp, pid, stack_size, buf, &bytes_read);

		size_t stack_width = 8;
		int count = 0;
		size_t i;

		GHashTable *dll_exports = (GHashTable *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_DLL_EXPORTS);

		if (!dll_exports) {
			DOCA_DLOG_DBG("dll_exports not found");
			return APSH_F;
		}

		for (i = 0; i < bytes_read; i += stack_width) {
			uint64_t stack_val = 0;

			memcpy(&stack_val, buf + i, stack_width);

			if (stack_val == 0)
				continue;

			GHashTableIter iter;
			gpointer key, value;

			g_hash_table_iter_init(&iter, dll_exports);
			char *dll = NULL;
			struct windows_export_info *export_functions = NULL;

			/* get the base dll! */
			while (g_hash_table_iter_next(&iter, &key, &value)) {
				char *key_str = (char *)key;
				struct windows_export_info *exp_info = (struct windows_export_info *)value;

				if (stack_val >= exp_info->dll_base) {
					if (stack_val < exp_info->dll_base + exp_info->dll_size) {
						if (!dll || !export_functions) {
							dll = key_str;
							export_functions = exp_info;
						} else if (exp_info->dll_base <
							   export_functions->dll_base) {
							dll = key_str;
							export_functions = exp_info;
						}
					}
				}
			}

			/* crawl through dll for the nearest function */
			if (dll) {
				g_hash_table_iter_init(&iter, export_functions->exports);
				addr_t nearest = 0;
				char *function_name = NULL;

				while (g_hash_table_iter_next(&iter, &key, &value)) {
					char *key_str = (char *)key;
					addr_t value_addr = (addr_t)value;

					if (stack_val >= value_addr) {
						if (nearest == 0) {
							nearest = value_addr;
							function_name = key_str;
						} else if (stack_val - value_addr <=
							   stack_val - nearest) {
							nearest = value_addr;
							function_name = key_str;
						}
					}
				}

				if (nearest) {
					addr_t function_offset = stack_val - nearest;
					DOCA_DLOG_INFO("    Frame %d: 0x%lx     %s!%s+0x%lx", count,
					       stack_val, dll, function_name, function_offset);
					count++;
				}
			}
		}
	}
	return APSH_S;
}

static status_t
threadlist_all(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid,
	       json_object **json_obj, bool printout, bool monitor_mode,
	       windows_thread_list_t **thread_list)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_THREADS_LIMIT)));

	status_t retcode = APSH_F;
	windows_processes_list_t *psl = NULL, *psl_allocation = NULL;
	windows_thread_list_t *thl = NULL, *curr_thead = NULL, *new_thread = NULL;
	addr_t current_process = 0;
	char time_buff[MAX_TIME_STAMP_LEN] = {0};
	char *procname = NULL;
	bool should_init;
	GHashTable *thread_entry;

	if (thread_list == NULL)
		goto error_exit;

	if (monitor_mode == true) {
		thread_entry = (GHashTable *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_THREAD_ENTRY);

		if (!thread_entry) {
			should_init = true;
			thread_entry = g_hash_table_new(g_direct_hash, g_direct_equal);
			if (thread_entry == NULL)
				return APSH_F;

			if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_THREAD_ENTRY, thread_entry, g_hash_table_destroy)) {
				g_hash_table_destroy(thread_entry);
				return APSH_F;
			}
		}
	}

	if (process_list == NULL) {
		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F)
			goto error_exit;
		psl = psl_allocation;
	} else {
		psl = process_list;
	}

	addr_t ImageFileName_offset;
	addr_t ThreadListHead_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
							 &ImageFileName_offset) == APSH_F)
		goto error_exit;

	/* get _EPROCESS ThreadListHead offset */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ThreadListHead",
							 &ThreadListHead_offset) == APSH_F)
		goto error_exit;

	/* get _ETHREAD ThreadListEntry, Cid,  ThreadName, Tcb, CreateTime, ExitTime offsets */
	addr_t ThreadListEntry_offset, Cid_offset, Tcb_offset, ThreadName_offset;
	addr_t CreateTime_offset, ExitTime_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_ETHREAD", "ThreadListEntry",
							 &ThreadListEntry_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_ETHREAD", "Cid", &Cid_offset) ==
	    APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_ETHREAD", "Tcb", &Tcb_offset) ==
	    APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_ETHREAD", "ThreadName",
							 &ThreadName_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_ETHREAD", "CreateTime",
							 &CreateTime_offset) == APSH_F)
		goto error_exit;
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_ETHREAD", "ExitTime",
							 &ExitTime_offset) == APSH_F)
		goto error_exit;

	addr_t state_offset;
	addr_t waitreason_offset;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_KTHREAD", "State",
							 &state_offset) == APSH_F)
		goto error_exit;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_KTHREAD", "WaitReason",
							 &waitreason_offset) == APSH_F)
		goto error_exit;

	/* print out titles */

	char *titles[] = {"TID", "Offset", "State", "WaitReason", "PID", "Process", "\0"};
	json_object *jout = jsonout_new("threadlist", titles);

	/* walk the process list */
	while (psl) {
		current_process = psl->address;

		procname = apsh_read_str_va(apsh, current_process + ImageFileName_offset, 0);

		if (!procname)
			goto next_process;

		if (pid == 0 || pid == psl->pid) {
			/* get ThreadListHead of current process,  Flink = proc.ThreadListHead.Flink */
			addr_t flink_va, flink;

			flink_va = current_process + ThreadListHead_offset;
			if (apsh_read_64_va(apsh, flink_va, 0, (uint64_t *)&flink) == APSH_F)
				goto next_process;

			addr_t link_header = flink_va;

			/* visited tghread list, mixmum = 1024 */
			GArray *visited = g_array_sized_new(false, true, sizeof(addr_t), 1024);
			uint64_t count = 0;

			/* walk through Linked table */
			do {
				if (++count > limit) {
					DOCA_DLOG_ERR("too large in threadlist");
					break;
				}

				visited = g_array_append_val(visited, flink);

				/* calculating ethread address, ethread = (Flink - ThreadListEntry_offset) */
				addr_t thread_va;

				thread_va = flink - ThreadListEntry_offset;

				/* get Tcb */
				addr_t tcb_va;

				tcb_va = thread_va + Tcb_offset;

				/* get Tcb.State */
				addr_t state_va;
				uint8_t state;

				state_va = tcb_va + state_offset;
				if (apsh_read_8_va(apsh, state_va, 0, &state) == APSH_F)
					break;

				/* get Tcb.Wait_reason */
				addr_t waitreason_va;
				uint8_t waitreason;

				waitreason_va = tcb_va + waitreason_offset;
				if (apsh_read_8_va(apsh, waitreason_va, 0, &waitreason) == APSH_F)
					break;

				/* get Cid, UniqueProcess, UniqueThread */
				struct client_id_t {
					addr_t UniqueProcess;
					addr_t UniqueThread;
				};
				addr_t cid_va;
				struct client_id_t cid;

				cid_va = thread_va + Cid_offset;
				if (apsh_read_va(apsh, cid_va, 0, sizeof(struct client_id_t),
							   &cid, NULL) == APSH_F)
					break;

				/* get ThreadName */
				addr_t threadname_va;
				addr_t threadname;

				threadname_va = thread_va + ThreadName_offset;
				apsh_read_64_va(apsh, threadname_va, 0, &threadname);

				/* get CreateTime, ExitTime */
				addr_t create_time_va, exit_time_va;
				uint64_t create_time, exit_time;

				create_time_va = thread_va + CreateTime_offset;
				exit_time_va = thread_va + ExitTime_offset;
				apsh_read_64_va(apsh, create_time_va, 0, &create_time);
				apsh_read_64_va(apsh, exit_time_va, 0, &exit_time);

				new_thread = malloc(sizeof(windows_thread_list_t));
				if (!new_thread)
					goto error_exit;

				new_thread->next = NULL;
				new_thread->pid = cid.UniqueProcess;
				new_thread->tid = cid.UniqueThread;
				new_thread->address = thread_va;

				GHashTable *hash = g_hash_table_new(NULL, NULL);

				if (hash == NULL) {
					free(new_thread);
					goto error_exit;
				}

				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_THREAD_PID),
						    GUINT_TO_POINTER(new_thread->pid));
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_THREAD_TID),
						    GUINT_TO_POINTER(new_thread->tid));
				g_hash_table_insert(
					hash, GINT_TO_POINTER(DOCA_APSH_THREAD_WINDOWS_OFFSET),
					GUINT_TO_POINTER(new_thread->address));
				g_hash_table_insert(hash,
						    GINT_TO_POINTER(DOCA_APSH_THREAD_STATE),
						    GUINT_TO_POINTER(state));
				g_hash_table_insert(
					hash, GINT_TO_POINTER(DOCA_APSH_THREAD_WINDOWS_WAIT_REASON),
					GUINT_TO_POINTER(waitreason));

				new_thread->hash = hash;

				if (thl == NULL) { /* 1st node in the list */
					thl = curr_thead = new_thread;
				} else {
					curr_thead->next = new_thread;
					curr_thead = curr_thead->next;
				}

				/* print out thread info */
				json_object *jdata = jsonout_data_element_new(jout);

				jsonout_data_element_value_sprintf(jdata, "%ld",
								   cid.UniqueThread); /* TID */
				jsonout_data_element_value_sprintf(jdata, "0x%lx",
								   thread_va); /* thread offset */
				jsonout_data_element_value_sprintf(jdata, "%d", state); /* state */
				jsonout_data_element_value_sprintf(jdata, "%d",
								   waitreason); /* wait reason */
				jsonout_data_element_value_sprintf(jdata, "%ld",
								   cid.UniqueProcess); /* PID */

				if (monitor_mode) {
					if (should_init) {
						if (waitreason == 5)
							g_hash_table_insert(
								thread_entry,
								GINT_TO_POINTER(cid.UniqueThread),
								GINT_TO_POINTER(waitreason));
					} else {
						/* insert to hash table if suspended */
						if (waitreason == 5) {
							if (g_hash_table_lookup(
								    thread_entry,
								    GINT_TO_POINTER(
									    cid.UniqueThread)) ==
							    NULL) {
								g_hash_table_insert(
									thread_entry,
									GINT_TO_POINTER(
										cid.UniqueThread),
									GINT_TO_POINTER(
										waitreason));
								DOCA_DLOG_INFO("[%s] " GRN
								       "--Thread Suspended--" RESET
								       " PID %ld, TID %ld, Name: " MAG
								       "%s" RESET "\n",
								       get_time_stamp(time_buff),
								       cid.UniqueProcess,
								       cid.UniqueThread, procname);
							}
						} else if (g_hash_table_lookup(
								   thread_entry,
								   GINT_TO_POINTER(
									   cid.UniqueThread)) !=
							   NULL) {
							/* remove if wait reason is not suspended anymore */
							g_hash_table_remove(
								thread_entry,
								GINT_TO_POINTER(cid.UniqueThread));
						}
					}
				}

				/* next Flick */
				flink_va = flink;
				if (apsh_read_64_va(apsh, flink_va, 0, (uint64_t *)&flink) == APSH_F)
					break;

			} while (flink != link_header);

			g_array_free(visited, true);

			if (pid == psl->pid)
				break;
		}

next_process:
		if (procname != NULL) {
			free(procname);
			procname = NULL;
		}

		/* follow the next pointer */
		psl = psl->next;
	}

	if (json_obj)
		*json_obj = jout;

	*thread_list = thl;

	if (printout)
		jsonout_treegrid_printout(jout);

	retcode = APSH_S;

	if (procname != NULL) {
		free(procname);
		procname = NULL;
	}

	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	return retcode;

error_exit:
	if (procname != NULL) {
		free(procname);
		procname = NULL;
	}

	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	if (thl != NULL)
		windows_free_thread_list(thl);

	return retcode;
}

void
windows_free_thread_list(windows_thread_list_t *thread_list)
{
	windows_thread_list_t *p_next = NULL;

	while (thread_list) {
		p_next = thread_list->next;
		g_hash_table_destroy(thread_list->hash);
		free(thread_list);
		thread_list = p_next;
	}
}

status_t
windows_threadlist(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid,
	   json_object **json_obj, bool printout, bool monitor_mode)
{
	windows_thread_list_t *thread_list;
	status_t result = threadlist_all(apsh, process_list, pid, json_obj, printout, monitor_mode,
					 &thread_list);
	windows_free_thread_list(thread_list);
	return result;
}

status_t
windows_get_threadlist(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid,
	       windows_thread_list_t **thread_list)
{
	return threadlist_all(apsh, process_list, pid, NULL, false, true, thread_list);
}
