/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <bsd/string.h>
#include "vadinfo.h"
#include "../../doca_apsh_attr.h"
#include "jsonout.h"
#include "openssl/sha.h"
#include "private/common_defines.h"
#include "doca_log.h"
#include "../../core/APSH/apsh_instance.h"
#include "../../core/APSH/plugin_context.h"

DOCA_LOG_REGISTER(APSH::Windows::VADINFO);

static const protection_t winnt_protection_vad[] = {
	{"PAGE_NOACCESS", 0x01},
	{"PAGE_READONLY", 0x02},
	{"PAGE_READWRITE", 0x04},
	{"PAGE_WRITECOPY", 0x08},
	{"PAGE_EXECUTE", 0x10},
	{"PAGE_EXECUTE_READ", 0x20},
	{"PAGE_EXECUTE_READWRITE", 0x40},
	{"PAGE_EXECUTE_WRITECOPY", 0x80},
	{"PAGE_GUARD", 0x100},
	{"PAGE_NOCACHE", 0x200},
	{"PAGE_WRITECOMBINE", 0x400},
	{"PAGE_TARGETS_INVALID", 0x40000000}};

static status_t
load_MMProtectToValue(apsh_instance_t apsh, int *MmProtectToValue)
{
	/* refer to volatility3 framework/plugins/windows/vadinfo.py protect_value() */
	addr_t MmProtectToValue_offset;

	if (apsh_read_ksym(apsh, "MmProtectToValue", sizeof(*MmProtectToValue)*MmProtectToValueSIZE, &MmProtectToValue[0],
			   NULL) == APSH_F) {
		DOCA_DLOG_ERR("Failed to find MmProtectToValue");
		return APSH_F;
	}

	return APSH_S;
}

static bool
is_in(addr_t v, GArray *s)
{
	addr_t t;
	int i;

	for (i = 0; i < 1024; i++) {
		t = g_array_index(s, addr_t, i);

		if (t == 0)
			break;

		if (v == t)
			return true;
	}

	return false;
}

static status_t
get_tag(apsh_instance_t apsh, addr_t vaddr, char *out)
{
	/* get tag */

	addr_t tag_addr = vaddr - 12; /* for 64bit Windows OS */

	return apsh_read_32_va(apsh, tag_addr, 0, (uint32_t *)out);
}

static void
sha256_hash_string(unsigned char hash[SHA256_DIGEST_LENGTH], char outputBuffer[65])
{
	int i = 0;

	for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
		sprintf(outputBuffer + (i * 2), "%02x", hash[i]);

	outputBuffer[64] = 0;
}

static void
sha256_string(char *string, char outputBuffer[65])
{
	unsigned char hash[SHA256_DIGEST_LENGTH];
	SHA256_CTX sha256;

	SHA256_Init(&sha256);
	SHA256_Update(&sha256, string, strlen(string));
	SHA256_Final(hash, &sha256);
	int i = 0;

	for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
		sprintf(outputBuffer + (i * 2), "%02x", hash[i]);
	outputBuffer[64] = 0;
}

static status_t
mmvad_short(apsh_instance_t apsh, addr_t vaddr, struct vadinfo_struct *vad_struct, vadinfo_t *out, json_object *jdata)
{
	jsonout_data_element_value_sprintf(jdata, "%d", out->pid);
	jsonout_data_element_value_sprintf(jdata, "%s", out->procname);
	jsonout_data_element_value_sprintf(jdata, "0x%lx", out->procaddr);

	addr_t node_addr = vaddr;
	mmvad_short_t _mmvad_short;

	/* load mmvad_short data */
	if (apsh_read_va(apsh, node_addr, 0, sizeof(mmvad_short_t), &_mmvad_short, NULL) ==
	    APSH_F) {
		DOCA_DLOG_ERR("Failed to get mmvad_shor at address 0x%lx", node_addr);
		return APSH_F;
	}

	/* get start vpn */
	out->startvpn = ((uint64_t)_mmvad_short.StartingVpn << 12) |
			((uint64_t)_mmvad_short.StartingVpnHigh << 44);

	/* get end vpn */
	out->endvpn = (((uint64_t)(_mmvad_short.EndingVpn + 1) << 12) |
		       ((uint64_t)_mmvad_short.EndingVpnHigh << 44));

	jsonout_data_element_value_sprintf(jdata, "0x%lx", out->startvpn);

	jsonout_data_element_value_sprintf(jdata, "0x%lx", out->endvpn);

	jsonout_data_element_value_add(jdata, out->tag);

	/* bit 3 ~ 8 */
	uint32_t protection_idx = ((_mmvad_short.u) >> vad_struct->mmvad_flag.protection_start_bit) &
				  BIT_MASK(uint32_t, (vad_struct->mmvad_flag.protection_end_bit -
						      vad_struct->mmvad_flag.protection_start_bit));

	int protection_value = vad_struct->MmProtectToValue[protection_idx];
	int count = ARRAY_SIZE(winnt_protection_vad);
	int i = 0;
	char buf[256];
	char *p = buf;

	memset(buf, '\0', sizeof(buf));
	for (i = 0; i < count; i++) {
		if (protection_value & winnt_protection_vad[i].mask) {
			sprintf(p, "%s ", winnt_protection_vad[i].name);
			p += strlen(p);
		}
	}
	strlcpy(out->protection, buf, sizeof(out->protection));
	jsonout_data_element_value_add(jdata, buf);

	uint32_t CommitCharge_value = ((_mmvad_short.u1) >> vad_struct->mmvad_flag.commit_charge_start_bit) &
				      BIT_MASK(uint32_t, (vad_struct->mmvad_flag.commit_charge_end_bit -
							  vad_struct->mmvad_flag.commit_charge_start_bit));

	jsonout_data_element_value_sprintf(jdata, "%u", CommitCharge_value);

	uint32_t PrivateMemory_value = ((_mmvad_short.u) >> vad_struct->mmvad_flag.private_memory_start_bit) &
				       BIT_MASK(uint32_t, (vad_struct->mmvad_flag.private_memory_end_bit -
							   vad_struct->mmvad_flag.private_memory_start_bit));
	jsonout_data_element_value_sprintf(jdata, "%u", PrivateMemory_value);

	/* get parent */
	uint64_t ParentValue = _mmvad_short.VadNode.ParentValue & ~0x03;

	jsonout_data_element_value_sprintf(jdata, "0x%lx", ParentValue);

	strlcpy(out->filepath, "N/A", sizeof(out->filepath));

	return APSH_S;
}

static status_t
mmvad(apsh_instance_t apsh, addr_t vaddr, struct vadinfo_struct *vad_struct, vadinfo_t *out, json_object *jdata)
{
	if (mmvad_short(apsh, vaddr, vad_struct, out, jdata) == APSH_F)
		return APSH_F;

	addr_t node_addr = vaddr;

	addr_t subsection_va;

	if (apsh_read_addr_va(apsh, node_addr + vad_struct->mmvad.subsection_rva, 0, &subsection_va) == APSH_F)
		return APSH_F;

	addr_t control_area_va;

	if (apsh_read_addr_va(apsh, subsection_va + vad_struct->subsection.control_area_rva, 0,
			      &control_area_va) == APSH_F)
		return APSH_F;

	addr_t file_object_va;

	if (apsh_read_addr_va(apsh, control_area_va + vad_struct->control_area.file_pointer_rva, 0,
			      &file_object_va) == APSH_F) {
		return APSH_F;
	}

	/* file_object is a special _EX_FAST_REF pointer, we need to explicitly clear low bits */
	file_object_va &= (~0xFULL);

	addr_t file_name_va = NULL;

	if ((void *)file_object_va != NULL)
		file_name_va = file_object_va + vad_struct->file_object.FileName_rva;

	/* get filename */
	unicode_string_t *us = NULL;

	us = apsh_read_unicode_str_va(apsh, file_name_va, 0);
	if (us == NULL) {
		strlcpy(out->filepath, "N/A", sizeof(out->filepath));
		jsonout_data_element_value_add(jdata, "N/A");
	} else {
		unicode_string_t _out = {0};

		if (apsh_convert_str_encoding(us, &_out, "UTF-8") == APSH_S) {
			strlcpy(out->filepath, _out.contents, sizeof(out->filepath));
			jsonout_data_element_value_add(jdata, _out.contents);
			free(_out.contents);
		} else {
			strlcpy(out->filepath, "N/A", sizeof(out->filepath));
			jsonout_data_element_value_add(jdata, "N/A");
		}

		if (us)
			apsh_free_unicode_str(us);
	}

	return APSH_S;
}

static status_t
traverse(apsh_instance_t apsh, addr_t vaddr, GArray *visited, int depth, addr_t process_addr,
	 uint32_t pid, char *procname, struct vadinfo_struct *vad_struct, windows_vad_t **vad_list, uint64_t *count)
{
	int limit = GPOINTER_TO_INT(g_hash_table_lookup(apsh->limits_table, GINT_TO_POINTER(DOCA_APSH_VADS_LIMIT)));

	/* refer to volatility3, framework/symbols/windows/extension/__init__.py, traverse() */

	/* process current node */

	if (vad_list == NULL)
		return APSH_F;

	if (depth > 100) {
		DOCA_DLOG_ERR("Vad tree is too deep.");
		return APSH_F;
	}

	addr_t node_addr = vaddr;

	if (is_in(node_addr, visited) == true)
		return APSH_S;

	visited = g_array_append_val(visited, node_addr);

	vadinfo_t _vadinfo;

	_vadinfo.pid = pid;
	_vadinfo.procaddr = process_addr;
	_vadinfo.procname = procname;

	if (get_tag(apsh, node_addr, &_vadinfo.tag[0]) == APSH_F)
		return APSH_F;

	_vadinfo.tag[4] = '\0';

	json_object *jout = jsonout_last_get(0);

	if ((memcmp(&_vadinfo.tag[0], "VadS", 4) == 0) ||
	    (memcmp(&_vadinfo.tag[0], "VadF", 4) == 0)) {
		json_object *jdata = jsonout_data_element_new(jout);
		/* _MMVAD_SHORT */
		mmvad_short(apsh, node_addr, vad_struct, &_vadinfo, jdata);
	} else if (memcmp(&_vadinfo.tag[0], "Vad", 3) == 0) {
		json_object *jdata = jsonout_data_element_new(jout);
		/* _MMVAD */
		mmvad(apsh, node_addr,vad_struct, &_vadinfo, jdata);
	} else if (depth != 0)
		return APSH_F;

	windows_vad_t *new_vad = malloc(sizeof(windows_vad_t));
	if (!new_vad)
		return APSH_F;
	new_vad->next = NULL;
	new_vad->vadinfo = _vadinfo;
	new_vad->mmvad_flag = vad_struct->mmvad_flag;
	new_vad->vadinfo.procname = strndup(procname, strlen(procname) + 1);
	if(new_vad->vadinfo.procname == NULL){
		free(new_vad);
		return APSH_F;
	}

	GHashTable *hash = g_hash_table_new(NULL, NULL);

	if (hash == NULL) {
		free(new_vad->vadinfo.procname);
		free(new_vad);
		return APSH_F;
	}

	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_PID),
			GUINT_TO_POINTER(new_vad->vadinfo.pid));
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_PROCESS_NAME),
			new_vad->vadinfo.procname);
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_FILE_PATH),
			new_vad->vadinfo.filepath);
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_OFFSET),
			GUINT_TO_POINTER(node_addr));
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_VM_START),
			GUINT_TO_POINTER(new_vad->vadinfo.startvpn));
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_VM_END),
			GUINT_TO_POINTER(new_vad->vadinfo.endvpn));
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_WINDOWS_COMMIT_CHARGE),
			GUINT_TO_POINTER(new_vad->mmvad_flag.commit_charge_offset));
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_WINDOWS_PRIVATE_MEMORY),
			GUINT_TO_POINTER(new_vad->mmvad_flag.private_memory_offset));
	g_hash_table_insert(hash,
			GINT_TO_POINTER(DOCA_APSH_VMA_PROTECTION),
			new_vad->vadinfo.protection);

	new_vad->hash = hash;

	if(*vad_list == NULL){
		*vad_list = new_vad;
		new_vad->tail = new_vad;
	} else {
		(*vad_list)->tail->next = new_vad;
		(*vad_list)->tail = new_vad;
	}

	rtl_balanced_node_t vadnode;

	if (apsh_read_va(apsh, node_addr, 0, sizeof(rtl_balanced_node_t), &vadnode, NULL) ==
	    APSH_F) {
		DOCA_DLOG_ERR("Failed to get vadnode at address 0x%lx", node_addr);
		return APSH_F;
	}

	if((*count)++ > limit)
		return APSH_S;

	/* process left tree */
	if (vadnode.Left) {
		if (traverse(apsh, vadnode.Left, visited, depth + 1, process_addr, pid,
				 procname, vad_struct, vad_list, count) == APSH_F)
			return APSH_F;
	}

	if(*count > limit)
		return APSH_S;
	/* process right tree */
	if (vadnode.Right) {
		if (traverse(apsh, vadnode.Right, visited, depth + 1, process_addr, pid,
			     procname, vad_struct, vad_list, count) == APSH_F)
			return APSH_F;
	}
	return APSH_S;
}

status_t
vadinfo_struct_init(apsh_instance_t apsh, struct vadinfo_struct *vad_struct)
{
	if (vad_struct == NULL)
		return APSH_F;

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "ImageFileName",
						&vad_struct->ImageFileName_offset) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get _EPROCESS ImageFileName");
		return APSH_F;
	}

	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_EPROCESS", "VadRoot",
							&vad_struct->vadroot_offset) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get _EPROCESS VadRoot");
		return APSH_F;
	}

	/* get protection */
	if (apsh_get_bitfield_offset_and_size_from_json(
			apsh, apsh_get_kernel_json(apsh), "_MMVAD_FLAGS", "Protection",
			&vad_struct->mmvad_flag.protection_offset, &vad_struct->mmvad_flag.protection_start_bit,
			&vad_struct->mmvad_flag.protection_end_bit) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get _MMVAD_FLAGS Protection bitfield and offset");

		return APSH_F;
	}
	/* get commit_charge */
	if (apsh_get_bitfield_offset_and_size_from_json(
			apsh, apsh_get_kernel_json(apsh), "_MMVAD_FLAGS1", "CommitCharge",
			&vad_struct->mmvad_flag.commit_charge_offset, &vad_struct->mmvad_flag.commit_charge_start_bit,
			&vad_struct->mmvad_flag.commit_charge_end_bit) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get _MMVAD_FLAGS1 CommitCharge bitfield and offset");
		return APSH_F;
	}

	/* get private memory */
	if (apsh_get_bitfield_offset_and_size_from_json(
			apsh, apsh_get_kernel_json(apsh), "_MMVAD_FLAGS", "PrivateMemory",
			&vad_struct->mmvad_flag.private_memory_offset, &vad_struct->mmvad_flag.private_memory_start_bit,
			&vad_struct->mmvad_flag.private_memory_end_bit) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get _MMVAD_FLAGS Protection bitfield and offset");
		return APSH_F;
	}

	/* get subsection RVA */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_MMVAD", "Subsection",
							&vad_struct->mmvad.subsection_rva) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get subsection offset");
		return APSH_F;
	}

	/* get subsection control area */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_SUBSECTION", "ControlArea",
							&vad_struct->subsection.control_area_rva) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get subsection control area offset");
		return APSH_F;
	}

	/* get control area file pointer */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_CONTROL_AREA", "FilePointer",
							&vad_struct->control_area.file_pointer_rva) ==
		APSH_F) {
		DOCA_DLOG_ERR("Failed to get Control Area FilePointer offset");
		return APSH_F;
	}

	/* get file name */
	if (apsh_get_struct_member_offset_from_apsh_json(apsh, "_FILE_OBJECT", "FileName",
							&vad_struct->file_object.FileName_rva) == APSH_F) {
		DOCA_DLOG_ERR("Failed to get file name offset");
		return APSH_F;
	}

	/* load MMProtectToValue */
	if (load_MMProtectToValue(apsh, vad_struct->MmProtectToValue) == APSH_F)
		return APSH_F;

	return APSH_S;
}

static status_t
vadinfo_all(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid, json_object **json_obj,
	bool printout, windows_vad_t **vad_list)
{
	status_t retcode = APSH_F;
	windows_processes_list_t *psl, *psl_allocation = NULL;

	struct vadinfo_struct *vad_struct = (struct vadinfo_struct *)plugin_ctx_lookup(apsh->plugin_ctx, PLUGIN_VMA);

	if (vad_struct == NULL) {
		vad_struct = malloc(sizeof(*vad_struct));

		if (vad_struct == NULL) {
			DOCA_DLOG_ERR("Failed to allocate plugin context memory");
			return APSH_F;
		}

		if (vadinfo_struct_init(apsh, vad_struct) == APSH_F) {
			free(vad_struct);
			goto error_exit;
		}

		if (!plugin_ctx_insert(apsh->plugin_ctx, PLUGIN_VMA, vad_struct, free)) {
			free(vad_struct);
			return APSH_F;
		}
	}

	if (process_list == NULL) {
		if (windows_pslist_processes_list(apsh, &psl_allocation) == APSH_F)
			goto error_exit;
		psl = psl_allocation;
	} else {
		psl = process_list;
	}

	/* print out titles */
	char *titles[] = {"PID",     "Process", "Offset",     "SHA256",	      "Start VPN",
			  "End VPN", "Tag",	"Protection", "CommitCharge", "PrivateMemory",
			  "Parent",  "File",	"\0"};
	json_object *jout;

	jout = jsonout_new("linux_modules", titles);

	addr_t current_process = 0;
	char *procname = NULL;

	/* walk the process list */
	while (psl) {
		current_process = psl->address;
		procname = apsh_read_str_va(apsh, current_process + vad_struct->ImageFileName_offset,
					    0); /* ImageFileName */

		if (!procname)
			goto next_process;

		if (pid == 0 || pid == psl->pid) {
			addr_t process_base = current_process;

			/* find vadroot */
			addr_t vadroot_va;
			addr_t rtl_balanced_node;
			uint64_t count = 0;

			vadroot_va = process_base + vad_struct->vadroot_offset;
			if (apsh_read_64_va(apsh, vadroot_va, 0, &rtl_balanced_node) == APSH_F)
				goto error_stop;

			GArray *visited = g_array_sized_new(false, true, sizeof(addr_t), 1024);

			status_t result = traverse(apsh, rtl_balanced_node, visited, 0,
						   process_base, psl->pid, procname, vad_struct, vad_list, &count);

			g_array_free(visited, true);

error_stop:
			if (pid == psl->pid) {
				free(procname);
				break;
			}
		}

next_process:
		if (procname != NULL)
			free(procname);
		psl = psl->next;
	}

	if (json_obj)
		*json_obj = jout;

	if (printout)
		jsonout_treegrid_printout(jout);

	retcode = APSH_S;

error_exit:

	if (psl_allocation != NULL)
		windows_free_process_list(psl_allocation);

	return retcode;
}

void
windows_free_vad_list(windows_vad_t *vad_list)
{
	windows_vad_t *p_next = NULL;

	while (vad_list) {
		p_next = vad_list->next;
		g_hash_table_destroy(vad_list->hash);
		free(vad_list->vadinfo.procname);
		free(vad_list);
		vad_list = p_next;
	}
}

status_t
windows_vadinfo(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid, json_object **json_obj, bool printout){
	windows_vad_t *vad_list = NULL;
	status_t result = vadinfo_all(apsh, process_list, pid, json_obj, printout, &vad_list);
	windows_free_vad_list(vad_list);
	return result;
}

status_t windows_get_vadlist(apsh_instance_t apsh, windows_processes_list_t *process_list, uint32_t pid, windows_vad_t **vad_list){
	*vad_list = NULL;
	return vadinfo_all(apsh, process_list, pid, NULL, false, vad_list);
}
