/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <getopt.h>
#include <gmodule.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "Plugins/linux/hashtest.h"
#include "Plugins/linux/libs.h"
#include "Plugins/linux/modules.h"
#include "Plugins/linux/pslist.h"
#include "Plugins/linux/threadlist.h"
#include "Plugins/linux/vma.h"
#include "Plugins/windows/dlllist.h"
#include "Plugins/windows/hashtest.h"
#include "Plugins/windows/modules.h"
#include "Plugins/windows/privs.h"
#include "Plugins/windows/pslist.h"
#include "Plugins/windows/threadlist.h"
#include "Plugins/windows/vadinfo.h"
#include "Plugins/windows/envars.h"
#include "Plugins/windows/ldrmodules.h"
#include "Plugins/windows/handles.h"

#include "doca_apsh.h"
#include "doca_log.h"
#include "priv_doca_dev.h"

#define DEFAULT_LIMIT 10000

DOCA_LOG_REGISTER(APSH);

struct doca_apsh_ctx {
	char *dma_dev_name;
	struct doca_dev *regex_dev;
	int start;
};

struct doca_apsh_system {
	struct doca_apsh_ctx *apsh_ctx;
	char *system_os_symbol_map_path;
	char *system_mem_region_path;
	char *system_kpgd_file_path;
	int vhca_id;
	enum doca_apsh_system_os os_type;
	GHashTable *limits_table;
	struct doca_dev_rep *system_dev;
	int start;
	apsh_instance_t apsh;
};

struct doca_apsh_module {
	addr_t entry;
	void *module_data;
	GHashTable *data;
	struct doca_apsh_system *system;
};

struct doca_apsh_process {
	addr_t entry;
	void *process_data;
	int pid;
	GHashTable *data;
	struct doca_apsh_system *system;
	int count;
};

struct doca_apsh_thread {
	addr_t entry;
	void *thread_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	int count;
};

struct doca_apsh_lib {
	addr_t entry;
	void *lib_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	int count;
};

struct doca_apsh_vad {
	addr_t entry;
	void *vad_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	int count;
};

struct doca_apsh_attestation {
	addr_t entry;
	GSList *attestation_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	char *exec_hash_map_path;
	int count;
};

struct doca_apsh_privilege {
	void *privilege_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	int count;
};

struct doca_apsh_envar {
	void *envar_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	int count;
};

struct doca_apsh_ldrmodule {
	void *ldrmodule_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	int count;
};

struct doca_apsh_handle {
	void *handle_data;
	GHashTable *data;
	struct doca_apsh_process *process;
	int count;
};

struct doca_apsh_ctx *
doca_apsh_create(void)
{
	return (struct doca_apsh_ctx *)calloc(1, sizeof(struct doca_apsh_ctx));
};

doca_error_t
doca_apsh_start(struct doca_apsh_ctx *ctx)
{
	if (ctx == NULL || ctx->dma_dev_name == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	ctx->start = 1;
	return DOCA_SUCCESS;
};

void
doca_apsh_destroy(struct doca_apsh_ctx *ctx)
{
	if (ctx == NULL)
		return;
	if (ctx->dma_dev_name)
		free(ctx->dma_dev_name);
	free(ctx);
};

doca_error_t
doca_apsh_dma_dev_set(struct doca_apsh_ctx *ctx, struct doca_dev *dma_dev)
{
	char dma_dev_name[DOCA_DEVINFO_IBDEV_NAME_SIZE];
	int ret;
	struct doca_devinfo *devinfo;

	if (ctx == NULL || dma_dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	devinfo = doca_dev_as_devinfo(dma_dev);
	ret = doca_devinfo_get_ibdev_name(devinfo, dma_dev_name, sizeof(dma_dev_name));
	if (ret != DOCA_SUCCESS)
		return ret;
	ctx->dma_dev_name = strdup(dma_dev_name);
	if (ctx->dma_dev_name == NULL)
		return DOCA_ERROR_NO_MEMORY;
	return DOCA_SUCCESS;
};

doca_error_t
doca_apsh_regex_dev_set(struct doca_apsh_ctx *ctx, struct doca_dev *regex_dev)
{
	if (ctx == NULL || regex_dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	ctx->regex_dev = regex_dev;
	return DOCA_SUCCESS;
};

struct doca_apsh_system *
doca_apsh_system_create(struct doca_apsh_ctx *ctx)
{
	if (ctx == NULL || !ctx->start)
		return NULL;
	struct doca_apsh_system *system = (struct doca_apsh_system *)calloc(1, sizeof(struct doca_apsh_system));

	if (system != NULL) {
		system->start = 0;
		system->vhca_id = -1;
		system->os_type = -1;
		system->apsh_ctx = ctx;
		system->limits_table = g_hash_table_new(NULL, NULL);
		if (system->limits_table == NULL) {
			free(system);
			return NULL;
		}

		for (int attr = DOCA_APSH_HASHTEST_LIMIT ; attr <= DOCA_APSH_STRING_LIMIT ; ++attr)
			if (g_hash_table_insert(system->limits_table, GINT_TO_POINTER(attr), GINT_TO_POINTER(DEFAULT_LIMIT)) == 0)
				goto error_exit;
	}

	return system;
error_exit:
	g_hash_table_destroy(system->limits_table);
	free(system);
	return NULL;
};

void
doca_apsh_system_destroy(struct doca_apsh_system *system)
{
	if (system == NULL)
		return;
	if (system->system_os_symbol_map_path)
		free(system->system_os_symbol_map_path);
	if (system->system_mem_region_path)
		free(system->system_mem_region_path);
	if (system->system_kpgd_file_path)
		free(system->system_kpgd_file_path);
	g_hash_table_destroy(system->limits_table);
	apsh_destroy(system->apsh);
	free(system);
};

doca_error_t
doca_apsh_system_start(struct doca_apsh_system *system)
{
	uint8_t config = APSH_CONFIG_JSON_PATH;
	void *path = NULL;

	if (system == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (!system->system_os_symbol_map_path)
		return DOCA_ERROR_INVALID_VALUE;
	if (!system->system_mem_region_path)
		return DOCA_ERROR_INVALID_VALUE;
	if (system->vhca_id < 0)
		return DOCA_ERROR_INVALID_VALUE;
	if (system->os_type < 0)
		return DOCA_ERROR_INVALID_VALUE;
	path = system->system_os_symbol_map_path;

	apsh_init_error_t err;

	if (apsh_init_complete(&(system->apsh), "APSH", APSH_INIT_DOMAINNAME, NULL, config, path, 0,
			       system->vhca_id, system->system_mem_region_path, system->system_kpgd_file_path,
			       system->apsh_ctx->dma_dev_name, system->apsh_ctx->regex_dev, system->system_dev, &err)
			        == APSH_F) {
		DOCA_DLOG_ERR("can't init apsh system, err=%d", err);
		return err;
	}
	apsh_set_limits_table(system->apsh, system->limits_table);
	system->start = 1;
	return DOCA_SUCCESS;
};

doca_error_t
doca_apsh_sys_os_symbol_map_set(struct doca_apsh_system *system, const char *system_os_symbol_map_path)
{
	if (system == NULL || system_os_symbol_map_path == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (system->start == 1) {
		DOCA_DLOG_ERR("can't modify system parameters after starting it.");
		return DOCA_ERROR_BAD_STATE;
	}

	system->system_os_symbol_map_path = strdup(system_os_symbol_map_path);
	if (system->system_os_symbol_map_path == NULL)
		return DOCA_ERROR_NO_MEMORY;
	return DOCA_SUCCESS;
};

doca_error_t
doca_apsh_sys_mem_region_set(struct doca_apsh_system *system, const char *system_mem_region_path)
{
	if (system == NULL || system_mem_region_path == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (system->start == 1) {
		DOCA_DLOG_ERR("can't modify system parameters after starting it.");
		return DOCA_ERROR_BAD_STATE;
	}

	system->system_mem_region_path = strdup(system_mem_region_path);
	if (system->system_mem_region_path == NULL)
		return DOCA_ERROR_NO_MEMORY;
	return DOCA_SUCCESS;
};

doca_error_t
doca_apsh_sys_dev_set(struct doca_apsh_system *system, struct doca_dev_rep *dev)
{
	if (system == NULL || dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (system->start == 1) {
		DOCA_DLOG_ERR("can't modify system parameters after starting it.");
		return DOCA_ERROR_BAD_STATE;
	}
	system->system_dev = dev;
	system->vhca_id = priv_doca_dev_vhca_id_get(dev);
	return DOCA_SUCCESS;
};

doca_error_t
doca_apsh_sys_os_type_set(struct doca_apsh_system *ctx, enum doca_apsh_system_os os_type)
{
	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (ctx->start == 1) {
		DOCA_DLOG_ERR("can't modify system parameters after starting it.");
		return DOCA_ERROR_BAD_STATE;
	}

	if (os_type != DOCA_APSH_SYSTEM_WINDOWS && os_type != DOCA_APSH_SYSTEM_LINUX)
		return DOCA_ERROR_NOT_SUPPORTED;
	ctx->os_type = os_type;
	return DOCA_SUCCESS;
};

doca_error_t
doca_apsh_sys_kpgd_file_set(struct doca_apsh_system *system, const char *system_kpgd_file_path)
{
	if (system == NULL || system_kpgd_file_path == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (system->start == 1) {
		DOCA_DLOG_ERR("can't modify system parameters after starting it.");
		return DOCA_ERROR_BAD_STATE;
	}

	system->system_kpgd_file_path = strdup(system_kpgd_file_path);
	if (system->system_kpgd_file_path == NULL)
		return DOCA_ERROR_NO_MEMORY;
	return DOCA_SUCCESS;
};

doca_error_t
__doca_apsh_sys_config(struct doca_apsh_system *system, enum doca_apsh_system_config_attr attr, void *value)
{
	if (system == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (attr < DOCA_APSH_OS_SYMBOL_MAP || attr > DOCA_APSH_STRING_LIMIT)
		return DOCA_ERROR_NOT_SUPPORTED;

	if (system->start)
		return DOCA_ERROR_BAD_STATE;

	switch (attr) {
	case DOCA_APSH_OS_SYMBOL_MAP:
		return doca_apsh_sys_os_symbol_map_set(system, (DOCA_APSH_OS_SYMBOL_MAP_TYPE)value);
	case DOCA_APSH_MEM_REGION:
		return doca_apsh_sys_mem_region_set(system, (DOCA_APSH_MEM_REGION_TYPE)value);
	case DOCA_APSH_KPGD_FILE:
		return doca_apsh_sys_kpgd_file_set(system, (DOCA_APSH_KPGD_FILE_TYPE)value);
	case DOCA_APSH_VHCA_ID:
		return doca_apsh_sys_dev_set(system, (DOCA_APSH_VHCA_ID_TYPE)value);
	case DOCA_APSH_OS_TYPE:
		return doca_apsh_sys_os_type_set(system, (DOCA_APSH_OS_TYPE_TYPE)value);
	default:
		g_hash_table_insert(system->limits_table, GINT_TO_POINTER(attr), GINT_TO_POINTER(value));
		return DOCA_SUCCESS;
	}
}

void *
next_module(void *modules, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_module_list_t *)modules)->next;
	return ((linux_module_list_t *)modules)->next;
}

void
insert_module_data(struct doca_apsh_module *module, void *mmodule)
{
	module->module_data = mmodule;
	if (module->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		module->data = ((windows_module_list_t *)mmodule)->hash;
		module->entry = ((windows_module_list_t *)mmodule)->entry;
	} else {
		module->data = ((linux_module_list_t *)mmodule)->data;
		module->entry = ((linux_module_list_t *)mmodule)->entry;
	}
};

doca_error_t
doca_apsh_modules_get(struct doca_apsh_system *system, struct doca_apsh_module ***modules, int *modules_size)
{
	void *mmodules_list;

	if (system == NULL || !system->start)
		return DOCA_ERROR_INVALID_VALUE;

	if (system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_module_list_t *windows_mmodules_list;

		if (windows_modules_inloadlist(system->apsh, &windows_mmodules_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get modules list");
			return DOCA_ERROR_INITIALIZATION;
		};
		mmodules_list = (void *)windows_mmodules_list;
	} else {
		linux_module_list_t *linux_mmodules_list;

		if (linux_get_modules(system->apsh, &linux_mmodules_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get modules list");
			return DOCA_ERROR_INITIALIZATION;
		};
		mmodules_list = (void *)linux_mmodules_list;
	}
	void *mmodules = mmodules_list;
	int count = 0;

	while (mmodules != NULL) {
		count++;
		mmodules = next_module(mmodules, system->os_type);
	}
	*modules = (struct doca_apsh_module **)malloc(sizeof(struct doca_apsh_module *) * count);
	if (*modules == NULL) {
		if (system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_module_list((windows_module_list_t *)mmodules_list);
		else
			linux_free_module_list((linux_module_list_t *)mmodules_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	*(*modules) = (struct doca_apsh_module *)malloc(sizeof(struct doca_apsh_module) * count);
	if (*(*modules) == NULL) {
		free(*modules);
		if (system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_module_list((windows_module_list_t *)mmodules_list);
		else
			linux_free_module_list((linux_module_list_t *)mmodules_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	mmodules = mmodules_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_module *module = *(*modules) + i;
		(*modules)[i] = module;
		module->system = system;
		insert_module_data(module, mmodules);
		mmodules = next_module(mmodules, system->os_type);
	}
	*modules_size = count;
	return DOCA_SUCCESS;
};

void
doca_apsh_module_free(struct doca_apsh_module **modules)
{
	if (modules == NULL)
		return;
	if (modules[0]->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_module_list((windows_module_list_t *)(modules[0]->module_data));
	else
		linux_free_module_list((linux_module_list_t *)(modules[0]->module_data));
	free(*modules);
	free(modules);
};

const void *
__doca_apsh_module_info_get(struct doca_apsh_module *module, enum doca_apsh_module_attr attr)
{
	return g_hash_table_lookup(module->data, GINT_TO_POINTER(attr));
};

void *
next_process(void *processes, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_processes_list_t *)processes)->next;
	return ((linux_processes_list_t *)processes)->next;
}

void
insert_process_data(struct doca_apsh_process *process, void *pprocess)
{
	process->process_data = pprocess;
	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		process->data = ((windows_processes_list_t *)pprocess)->hash;
		process->entry = ((windows_processes_list_t *)pprocess)->address;
		process->pid = ((windows_processes_list_t *)pprocess)->pid;
	} else {
		process->data = ((linux_processes_list_t *)pprocess)->hash;
		process->entry = ((linux_processes_list_t *)pprocess)->address;
		process->pid = ((linux_processes_list_t *)pprocess)->pid;
	}
};

doca_error_t
doca_apsh_processes_get(struct doca_apsh_system *system, struct doca_apsh_process ***processes, int *processes_size)
{
	void *pprocess_list;

	if (system == NULL || !system->start || processes == NULL || processes_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_processes_list_t *windows_pprocess_list;

		if (windows_pslist_processes_list(system->apsh, &windows_pprocess_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get process list");
			return DOCA_ERROR_INITIALIZATION;
		};
		pprocess_list = (void *)windows_pprocess_list;
	} else {
		linux_processes_list_t *linux_pprocess_list;

		if (linux_pslist_processes_list(system->apsh, &linux_pprocess_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get process list");
			return DOCA_ERROR_INITIALIZATION;
		};
		pprocess_list = (void *)linux_pprocess_list;
	}
	void *pprocess = pprocess_list;
	int count = 0;

	while (pprocess != NULL) {
		count++;
		pprocess = next_process(pprocess, system->os_type);
	}
	*processes = (struct doca_apsh_process **)malloc(sizeof(struct doca_apsh_process *) * count);
	if (*processes == NULL) {
		if (system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_process_list((windows_processes_list_t *)pprocess_list);
		else
			linux_free_process_list((linux_processes_list_t *)pprocess_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	*(*processes) = (struct doca_apsh_process *)malloc(sizeof(struct doca_apsh_process) * count);
	if (*(*processes) == NULL) {
		free(*processes);
		if (system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_process_list((windows_processes_list_t *)pprocess_list);
		else
			linux_free_process_list((linux_processes_list_t *)pprocess_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	pprocess = pprocess_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_process *process = *(*processes) + i;

		(*processes)[i] = process;
		process->system = system;
		process->count = count;
		insert_process_data(process, pprocess);
		pprocess = next_process(pprocess, system->os_type);
	}
	*processes_size = count;
	return DOCA_SUCCESS;
};

void
doca_apsh_processes_free(struct doca_apsh_process **processes)
{
	if (processes == NULL)
		return;
	if (processes[0]->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_process_list((windows_processes_list_t *)(processes[0]->process_data));
	else
		linux_free_process_list((linux_processes_list_t *)(processes[0]->process_data));
	free(*processes);
	free(processes);
};

const void *
__doca_apsh_process_info_get(struct doca_apsh_process *process, enum doca_apsh_process_attr attr)
{
	return g_hash_table_lookup(process->data, GINT_TO_POINTER(attr));
};

void *
next_lib(void *libs, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_libs_t *)libs)->next;
	return ((linux_libs_t *)libs)->next;
}

void
insert_lib_data(struct doca_apsh_lib *lib, void *llib)
{
	lib->lib_data = llib;
	if (lib->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		lib->data = ((windows_libs_t *)llib)->data;
	} else {
		lib->data = ((linux_libs_t *)llib)->data;
		lib->entry = ((linux_libs_t *)llib)->address;
	}
};

doca_error_t
doca_apsh_libs_get(struct doca_apsh_process *process, struct doca_apsh_lib ***libs, int *libs_size)
{
	void *llibs_list;

	if (process == NULL || process->system == NULL || process->system->start == 0 || libs == NULL ||
	    libs_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_libs_t *windows_llibs_list;
		windows_processes_list_t *pprocess = (windows_processes_list_t *)process->process_data;

		if (windows_get_dlls(process->system->apsh, pprocess, &windows_llibs_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get modules list");
			return DOCA_ERROR_INITIALIZATION;
		};
		llibs_list = (void *)windows_llibs_list;
	} else {
		linux_libs_t *llibs_linux_list;
		linux_processes_list_t *pprocess = (linux_processes_list_t *)process->process_data;

		if (linux_get_libs(process->system->apsh, pprocess, pprocess->pid, &llibs_linux_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get modules list");
			return DOCA_ERROR_INITIALIZATION;
		};
		llibs_list = (void *)llibs_linux_list;
	}
	void *llibs = llibs_list;
	int count = 0;

	while (llibs != NULL) {
		count++;
		llibs = next_lib(llibs, process->system->os_type);
	}

	if (count == 0) { /* in case process do not have any libs */
		*libs_size = 0;
		*libs = NULL;
		return DOCA_SUCCESS;
	}

	*libs = (struct doca_apsh_lib **)malloc(sizeof(struct doca_apsh_lib *) * count);

	if (*libs == NULL) {
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_libs_list((windows_libs_t *)llibs_list);
		else
			linux_free_libs_list((linux_libs_t *)llibs_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	*(*libs) = (struct doca_apsh_lib *)malloc(sizeof(struct doca_apsh_lib) * count);
	if (*(*libs) == NULL) {
		free(*libs);
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_libs_list((windows_libs_t *)llibs_list);
		else
			linux_free_libs_list((linux_libs_t *)llibs_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	llibs = llibs_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_lib *lib = *(*libs) + i;
		(*libs)[i] = lib;
		lib->process = process;
		lib->count = count;
		insert_lib_data(lib, llibs);
		llibs = next_lib(llibs, process->system->os_type);
	}
	*libs_size = count;
	return DOCA_SUCCESS;
};

void
doca_apsh_libs_free(struct doca_apsh_lib **libs)
{
	if (libs == NULL || *libs == NULL)
		return;
	if (libs[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_libs_list(libs[0]->lib_data);
	else
		linux_free_libs_list(libs[0]->lib_data);
	free(*libs);
	free(libs);
};

const void *
__doca_apsh_lib_info_get(struct doca_apsh_lib *lib, enum doca_apsh_lib_attr attr)
{
	return g_hash_table_lookup(lib->data, GINT_TO_POINTER(attr));
};

void *
next_thread(void *threads, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_thread_list_t *)threads)->next;
	return ((linux_thread_list_t *)threads)->next;
}

int
insert_thread_data(struct doca_apsh_thread *thread, void *tthreads)
{
	thread->thread_data = tthreads;
	if (thread->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		thread->data = ((windows_thread_list_t *)tthreads)->hash;
		thread->entry = ((windows_thread_list_t *)tthreads)->address;
	} else {
		thread->data = ((linux_thread_list_t *)tthreads)->hash;
		thread->entry = ((linux_thread_list_t *)tthreads)->address;
	}
	return 0;
};

doca_error_t
doca_apsh_threads_get(struct doca_apsh_process *process, struct doca_apsh_thread ***threads, int *threads_size)
{
	void *tthreads_list;

	if (process == NULL || process->system == NULL || !process->system->start || threads == NULL ||
	    threads_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_thread_list_t *windows_tthreads_list;

		if (windows_get_threadlist(process->system->apsh, (windows_processes_list_t *)process->process_data, process->pid,
				   &windows_tthreads_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get thread list");
			return DOCA_ERROR_INITIALIZATION;
		};
		tthreads_list = (void *)windows_tthreads_list;
	} else {
		linux_thread_list_t *linux_tthreads_list;

		if (linux_get_threadlist(process->system->apsh, (linux_processes_list_t *)process->process_data,
					 process->pid, &linux_tthreads_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get thread list");
			return DOCA_ERROR_INITIALIZATION;
		};
		tthreads_list = (void *)linux_tthreads_list;
	}
	void *tthreads = tthreads_list;
	int count = 0;

	while (tthreads != NULL) {
		count++;
		tthreads = next_thread(tthreads, process->system->os_type);
	}

	if (count == 0) { /* in case process do not have any threads */
		*threads_size = 0;
		*threads = NULL;
		return DOCA_SUCCESS;
	}

	*threads = (struct doca_apsh_thread **)malloc(sizeof(struct doca_apsh_thread *) * count);
	if (*threads == NULL) {
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_thread_list(tthreads_list);
		else
			linux_free_thread_list(tthreads_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	*(*threads) = (struct doca_apsh_thread *)malloc(sizeof(struct doca_apsh_thread) * count);
	if (*(*threads) == NULL) {
		free(*threads);
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_thread_list(tthreads_list);
		else
			linux_free_thread_list(tthreads_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	tthreads = tthreads_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_thread *thread = *(*threads) + i;

		(*threads)[i] = thread;
		thread->process = process;
		thread->count = count;
		insert_thread_data(thread, tthreads);
		tthreads = next_thread(tthreads, process->system->os_type);
	}
	*threads_size = count;
	return DOCA_SUCCESS;
};

void
doca_apsh_threads_free(struct doca_apsh_thread **threads)
{
	if (threads == NULL)
		return;
	if (threads[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_thread_list((windows_thread_list_t *)(threads[0]->thread_data));
	else
		linux_free_thread_list((linux_thread_list_t *)(threads[0]->thread_data));
	free(*threads);
	free(threads);
};

const void *
__doca_apsh_thread_info_get(struct doca_apsh_thread *thread, enum doca_apsh_thread_attr attr)
{
	return g_hash_table_lookup(thread->data, GINT_TO_POINTER(attr));
};

void *
next_vad(void *vads, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_vad_t *)vads)->next;
	return ((linux_vma_t *)vads)->next;
}

int
insert_vad_data(struct doca_apsh_vad *vad, void *vvads)
{
	vad->vad_data = vvads;
	if (vad->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		vad->data = ((windows_vad_t *)vvads)->hash;
	} else {
		vad->data = ((linux_vma_t *)vvads)->data;
	}
	return 0;
};

doca_error_t
doca_apsh_vads_get(struct doca_apsh_process *process, struct doca_apsh_vad ***vads, int *vads_size)
{
	void *vvads_list;

	if (process == NULL || process->system == NULL || !process->system->start || vads == NULL ||
	    vads_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_vad_t *windows_vvads_list;

		if (windows_get_vadlist(process->system->apsh,
					     (windows_processes_list_t *)process->process_data,
					     process->pid, &windows_vvads_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get vad list");
			return DOCA_ERROR_INITIALIZATION;
		};
		vvads_list = (void *)windows_vvads_list;
	} else {
		linux_vma_t *linux_vvads_list;

		if (linux_get_vma(process->system->apsh,
					(linux_processes_list_t *)process->process_data,
					process->pid, &linux_vvads_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get vad list");
			return DOCA_ERROR_INITIALIZATION;
		};
		vvads_list = (void *)linux_vvads_list;
	}
	void *vvads = vvads_list;
	int count = 0;

	while (vvads != NULL) {
		count++;
		vvads = next_vad(vvads, process->system->os_type);
	}

	if (count == 0) { /* in case process do not have any vads */
		*vads_size = 0;
		*vads = NULL;
		return DOCA_SUCCESS;
	}

	*vads = (struct doca_apsh_vad **)malloc(sizeof(struct doca_apsh_vad *) * count);
	if (*vads == NULL) {
		linux_free_vma_list(vvads_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	*(*vads) = (struct doca_apsh_vad *)malloc(sizeof(struct doca_apsh_vad) * count);
	if (*(*vads) == NULL) {
		free(*vads);
		linux_free_vma_list(vvads_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	vvads = vvads_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_vad *vad = *(*vads) + i;

		(*vads)[i] = vad;
		vad->process = process;
		vad->count = count;
		insert_vad_data(vad, vvads);
		vvads = next_vad(vvads, process->system->os_type);
	}
	*vads_size = count;
	return DOCA_SUCCESS;
};

void
doca_apsh_vads_free(struct doca_apsh_vad **vads)
{
	if (vads == NULL)
		return;
	if (vads[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_vad_list((windows_vad_t *)(vads[0]->vad_data));
	else
		linux_free_vma_list((linux_vma_t *)(vads[0]->vad_data));
	free(*vads);
	free(vads);
};

const void *
__doca_apsh_vad_info_get(struct doca_apsh_vad *vad, enum doca_apsh_vad_attr attr)
{
	return g_hash_table_lookup(vad->data, GINT_TO_POINTER(attr));
};

doca_error_t
attestation_create(struct doca_apsh_process *process, const char *exec_hash_map_path,
						 GSList *attestation_list, void *attestation_result,
						 struct doca_apsh_attestation ***attestation, int *attestation_size)
{
	int count = 0;
	char *path;
	GSList *attestation_list_p = attestation_list;

	while (attestation_list_p != NULL) {
		count++;
		attestation_list_p = attestation_list_p->next;
	}

	if (count == 0)
		return DOCA_ERROR_NOT_FOUND;

	*attestation = (struct doca_apsh_attestation **)malloc(sizeof(struct doca_apsh_attestation *) * count);
	if (*attestation == NULL)
		return DOCA_ERROR_NO_MEMORY;
	*(*attestation) = (struct doca_apsh_attestation *)malloc(sizeof(struct doca_apsh_attestation) * count);
	if (*(*attestation) == NULL) {
		free(*attestation);
		return DOCA_ERROR_NO_MEMORY;
	}
	attestation_list_p = attestation_list;
	path = strdup(exec_hash_map_path);
	if (path == NULL) {
		free(*(*attestation));
		free(*attestation);
		return DOCA_ERROR_NO_MEMORY;
	}

	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_attestation *att = *(*attestation) + i;

		(*attestation)[i] = att;
		att->attestation_data = attestation_result;
		att->data = attestation_list_p->data;
		att->process = process;
		att->exec_hash_map_path = path;
		attestation_list_p = attestation_list_p->next;
	}
	*attestation_size = count;
	return DOCA_SUCCESS;
}

doca_error_t
doca_apsh_attestation_get(struct doca_apsh_process *process, const char *exec_hash_map_path,
			  struct doca_apsh_attestation ***attestation, int *attestation_size)
{
	int count = 0;
	GSList *attestation_list = NULL;
	void *attestation_result;
	char *path;

	if (process == NULL || exec_hash_map_path == NULL || attestation == NULL || attestation_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_hashtest_t *windows_attestation_result;

		if (windows_hashtest(process->system->apsh, (windows_processes_list_t *)(process->process_data), process->pid,
				exec_hash_map_path, &windows_attestation_result) == APSH_F) {
			DOCA_DLOG_ERR("can't get module list");
			return DOCA_ERROR_INITIALIZATION;
		}
		attestation_result = (void *)windows_attestation_result;
		attestation_list = windows_attestation_result->results_list;
	} else {
		linux_hashtest_t *linux_attestation_result;

		if (linux_hashtest(process->system->apsh, (linux_processes_list_t *)(process->process_data), process->pid,
				exec_hash_map_path, &linux_attestation_result) == APSH_F) {
			DOCA_DLOG_ERR("can't get module list");
			return DOCA_ERROR_INITIALIZATION;
		}
		attestation_result = (void *)linux_attestation_result;
		attestation_list = linux_attestation_result->results_list;
	}

	return attestation_create(process, exec_hash_map_path, attestation_list, attestation_result,
				  attestation, attestation_size);
};

void
doca_apsh_attestation_free(struct doca_apsh_attestation **attestation)
{
	if (attestation == NULL || (attestation[0]->process) == NULL ||
			(attestation[0]->process->system) == NULL)
		return;
	if (attestation[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_hashtest_results(attestation[0]->attestation_data);
	else
		linux_free_hashtest_results(attestation[0]->attestation_data);
	free(attestation[0]->exec_hash_map_path);
	free(*attestation);
	free(attestation);
};

const void *
__doca_apsh_attst_info_get(struct doca_apsh_attestation *attestation, enum doca_apsh_attestation_attr attr)
{
	return g_hash_table_lookup(attestation->data, GINT_TO_POINTER(attr));
};

doca_error_t
doca_apsh_attst_refresh(struct doca_apsh_attestation ***attestation, int *attestation_size)
{
	struct doca_apsh_attestation **newAttestation;
	int new_attestation_size;
	GSList *attestation_list = NULL;
	void *attestation_result;
	char *path;
	doca_error_t ret;

	if (attestation == NULL || attestation_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if ((*attestation)[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_hashtest_t *windows_attestation_result = (*attestation)[0]->attestation_data;

		if (windows_hashtest_refresh((*attestation)[0]->process->system->apsh,
				(windows_processes_list_t *)((*attestation)[0]->process->process_data),
				&windows_attestation_result) == APSH_F) {
			DOCA_DLOG_ERR("can't get module list");
			return DOCA_ERROR_INITIALIZATION;
		}
		attestation_result = (void *)windows_attestation_result;
		attestation_list = windows_attestation_result->results_list;
	} else {
		linux_hashtest_t *linux_attestation_result = (*attestation)[0]->attestation_data;

		if (linux_hashtest_refresh((*attestation)[0]->process->system->apsh,
				(linux_processes_list_t *)((*attestation)[0]->process->process_data),
				&linux_attestation_result) == APSH_F) {
			DOCA_DLOG_ERR("can't get module list");
			return DOCA_ERROR_INITIALIZATION;
		}
		attestation_result = (void *)linux_attestation_result;
		attestation_list = linux_attestation_result->results_list;
	}

	ret = attestation_create((*attestation)[0]->process, (*attestation)[0]->exec_hash_map_path,
			attestation_list, attestation_result, &newAttestation, &new_attestation_size);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("can't get module list");
		return ret;
	}
	doca_apsh_attestation_free(*attestation);
	*attestation = newAttestation;
	*attestation_size = new_attestation_size;
	return DOCA_SUCCESS;
};

void *
next_privilege(void *privileges, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_privileges_list_t *)privileges)->next;
	return NULL;
}

int
insert_privilege_data(struct doca_apsh_privilege *privilege, void *pprivileges)
{
	privilege->privilege_data = pprivileges;
	if (privilege->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		privilege->data = ((windows_privileges_list_t *)pprivileges)->hash;
	return 0;
};

doca_error_t
doca_apsh_privileges_get(struct doca_apsh_process *process, struct doca_apsh_privilege ***privileges, int *privileges_size)
{
	void *pprivileges_list;

	if (process == NULL || process->system == NULL || !process->system->start || privileges == NULL ||
	    privileges_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_privileges_list_t *windows_pprivileges_list;

		if (privs(process->system->apsh, (windows_processes_list_t *)process->process_data, process->pid,
				   &windows_pprivileges_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get privilege list");
			return DOCA_ERROR_INITIALIZATION;
		};
		pprivileges_list = (void *)windows_pprivileges_list;
	} else {
		return DOCA_ERROR_NOT_SUPPORTED;
	}
	void *pprivileges = pprivileges_list;
	int count = 0;

	while (pprivileges != NULL) {
		count++;
		pprivileges = next_privilege(pprivileges, process->system->os_type);
	}

	if (count == 0) { /* in case process do not have any privileges */
		*privileges_size = 0;
		*privileges = NULL;
		return DOCA_SUCCESS;
	}

	*privileges = (struct doca_apsh_privilege **)malloc(sizeof(struct doca_apsh_privilege *) * count);
	if (*privileges == NULL) {
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_privileges_list(pprivileges_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	*(*privileges) = (struct doca_apsh_privilege *)malloc(sizeof(struct doca_apsh_privilege) * count);
	if (*(*privileges) == NULL) {
		free(*privileges);
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_privileges_list(pprivileges_list);
		return DOCA_ERROR_NO_MEMORY;
	}
	pprivileges = pprivileges_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_privilege *privilege = *(*privileges) + i;

		(*privileges)[i] = privilege;
		privilege->process = process;
		privilege->count = count;
		insert_privilege_data(privilege, pprivileges);
		pprivileges = next_privilege(pprivileges, process->system->os_type);
	}
	*privileges_size = count;
	return DOCA_SUCCESS;
};

void
doca_apsh_privileges_free(struct doca_apsh_privilege **privileges)
{
	if (privileges == NULL)
		return;
	if (privileges[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_privileges_list((windows_privileges_list_t *)(privileges[0]->privilege_data));
	free(*privileges);
	free(privileges);
};

const void *
__doca_apsh_privilege_info_get(struct doca_apsh_privilege *privilege, enum doca_apsh_privilege_attr attr)
{
	return g_hash_table_lookup(privilege->data, GINT_TO_POINTER(attr));
};

void *
next_envar(void *envar, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_envars_t *)envar)->next;
	return NULL;
}

int
insert_envar_data(struct doca_apsh_envar *envar, void *eenvar)
{
	envar->envar_data = eenvar;
	if (envar->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		envar->data = ((windows_envars_t *)eenvar)->data;
	return 0;
};

doca_error_t doca_apsh_envars_get(struct doca_apsh_process *process, struct doca_apsh_envar ***envars, int *envars_size)
{
	void *eenvars_list;

	if (process == NULL || envars == NULL || envars_size == NULL || process->system == NULL || !process->system->start) {
		DOCA_DLOG_ERR("one of the inputs is null");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_envars_t *windows_eenvars_list;

		if (windows_envars(process->system->apsh, (windows_processes_list_t *)process->process_data, process->pid,
				   &windows_eenvars_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get envars list");
			return DOCA_ERROR_INITIALIZATION;
		};
		eenvars_list = (void *)windows_eenvars_list;
	} else {
		DOCA_DLOG_ERR("envars supports only windows");
		return DOCA_ERROR_NOT_SUPPORTED;
	}
	void *eenvars = eenvars_list;
	int count = 0;

	while (eenvars != NULL) {
		count++;
		eenvars = next_envar(eenvars, process->system->os_type);
	}

	if (count == 0) { /* in case process do not have any envars */
		*envars_size = 0;
		*envars = NULL;
		return DOCA_SUCCESS;
	}

	*envars = (struct doca_apsh_envar **)calloc(count, sizeof(struct doca_apsh_envar *));
	if (*envars == NULL) {
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_envars_list(eenvars_list);
		DOCA_DLOG_ERR("failed to allocate memory for envars array");
		return DOCA_ERROR_NO_MEMORY;
	}
	*(*envars) = (struct doca_apsh_envar *)calloc(count, sizeof(struct doca_apsh_envar));
	if (*(*envars) == NULL) {
		free(*envars);
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_envars_list(eenvars_list);
		DOCA_DLOG_ERR("failed to allocate memory for envars array");
		return DOCA_ERROR_NO_MEMORY;
	}
	eenvars = eenvars_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_envar *envar = *(*envars) + i;

		(*envars)[i] = envar;
		envar->process = process;
		envar->count = count;
		insert_envar_data(envar, eenvars);
		eenvars = next_envar(eenvars, process->system->os_type);
	}
	*envars_size = count;
	return DOCA_SUCCESS;
};

void
doca_apsh_envars_free(struct doca_apsh_envar **envars)
{
	if (envars == NULL)
		return;
	if (envars[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_envars_list((windows_envars_t *)(envars[0]->envar_data));
	free(*envars);
	free(envars);
};

const void *
__doca_apsh_envar_info_get(struct doca_apsh_envar *envar, enum doca_apsh_envar_attr attr)
{
	return g_hash_table_lookup(envar->data, GINT_TO_POINTER(attr));
};

int
insert_ldrmodule_data(struct doca_apsh_ldrmodule *ldrmodule, void *pldrmodules)
{
	ldrmodule->ldrmodule_data = pldrmodules;
	if (ldrmodule->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		ldrmodule->data = ((windows_ldrmodules_list_t *)pldrmodules)->hash;
	return 0;
};

void *
next_ldrmodule(void *ldrmodules, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_ldrmodules_list_t *)ldrmodules)->next;
	return NULL;
}

doca_error_t
doca_apsh_ldrmodules_get(struct doca_apsh_process *process, struct doca_apsh_ldrmodule ***ldrmodules, int *ldrmodules_size)
{
	void *pldrmodules_list;

	if (process == NULL || process->system == NULL || !process->system->start || ldrmodules == NULL ||
		ldrmodules_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_ldrmodules_list_t *windows_pldrmodules_list;

		if (windows_ldrmodules(process->system->apsh, (windows_processes_list_t *)process->process_data, process->pid,
								&windows_pldrmodules_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get ldrmodules list");
			return DOCA_ERROR_INITIALIZATION;
		}
		pldrmodules_list = (void *)windows_pldrmodules_list;
	} else
		return DOCA_ERROR_NOT_SUPPORTED;

	void *pldrmodules = pldrmodules_list;
	int count = 0;

	while (pldrmodules != NULL) {
		count++;
		pldrmodules = next_ldrmodule(pldrmodules, process->system->os_type);
	}

	if (count == 0) { /* in case process do not have any privileges */
		*ldrmodules_size = 0;
		*ldrmodules = NULL;
		return DOCA_SUCCESS;
	}

	*ldrmodules = (struct doca_apsh_ldrmodule **)malloc(sizeof(struct doca_apsh_ldrmodule *) * count);
	if (*ldrmodules == NULL) {
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_ldrmodules_list(pldrmodules_list);
		return DOCA_ERROR_NO_MEMORY;
	}

	*(*ldrmodules) = (struct doca_apsh_ldrmodule *)malloc(sizeof(struct doca_apsh_ldrmodule) * count);
	if (*(*ldrmodules) == NULL) {
		free(*ldrmodules);
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_ldrmodules_list(pldrmodules_list);
		return DOCA_ERROR_NO_MEMORY;
	}

	pldrmodules = pldrmodules_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_ldrmodule *ldrmodule = *(*ldrmodules) + i;

		(*ldrmodules)[i] = ldrmodule;
		ldrmodule->process = process;
		ldrmodule->count = count;
		insert_ldrmodule_data(ldrmodule, pldrmodules);
		pldrmodules = next_ldrmodule(pldrmodules, process->system->os_type);
	}
	*ldrmodules_size = count;

	return DOCA_SUCCESS;
}

void
doca_apsh_ldrmodules_free(struct doca_apsh_ldrmodule **ldrmodules)
{
	if (ldrmodules == NULL)
		return;
	if (ldrmodules[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_ldrmodules_list((windows_ldrmodules_list_t *)(ldrmodules[0]->ldrmodule_data));
	free(*ldrmodules);
	free(ldrmodules);
}

const void *
__doca_apsh_ldrmodule_info_get(struct doca_apsh_ldrmodule *ldrmodule, enum doca_apsh_ldrmodule_attr attr)
{
	return g_hash_table_lookup(ldrmodule->data, GINT_TO_POINTER(attr));
}

int
insert_handle_data(struct doca_apsh_handle *handle, void *handles_ptr)
{
	handle->handle_data = handles_ptr;
	if (handle->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		handle->data = ((windows_handles_t *)handles_ptr)->hash;
	return 0;
};

void *
next_handle(void *handle, int type)
{
	if (type == DOCA_APSH_SYSTEM_WINDOWS)
		return ((windows_handles_t *)handle)->next;
	return NULL; // linux not implemented
}

doca_error_t
doca_apsh_handles_get(struct doca_apsh_process *process, struct doca_apsh_handle ***handles, int *handles_size)
{
	*handles_size = 0;
	void *handles_list;

	if (process == NULL || handles == NULL || handles_size == NULL || process->system == NULL ||
		!process->system->start) {
		DOCA_DLOG_ERR("one of the inputs is null");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS) {
		windows_handles_t *windows_handles_list;

		if (windows_handles(process->system->apsh, (windows_processes_list_t *)process->process_data,
				    process->pid, &windows_handles_list) == APSH_F) {
			DOCA_DLOG_ERR("can't get envars list");
			return DOCA_ERROR_INITIALIZATION;
		};
		handles_list = (void *) windows_handles_list;
	} else {
		DOCA_DLOG_ERR("handles supports only windows");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	void *handles_ptr = handles_list;
	int count = 0;

	while (handles_ptr != NULL) {
		count++;
		handles_ptr = next_handle(handles_ptr, process->system->os_type);
	}

	if (count == 0) { /* in case process do not have any handles */
		*handles_size = 0;
		*handles = NULL;
		return DOCA_SUCCESS;
	}

	*handles = (struct doca_apsh_handle **)malloc(sizeof(struct doca_apsh_handle *) * count);
	if (*handles == NULL) {
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_handles(handles_list);
		DOCA_DLOG_ERR("failed to allocate memory for handles array");
		return DOCA_ERROR_NO_MEMORY;
	}

	*(*handles) = (struct doca_apsh_handle *)malloc(sizeof(struct doca_apsh_handle) * count);
	if (*(*handles) == NULL) {
		free(*handles);
		if (process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
			windows_free_handles(handles_list);
		DOCA_DLOG_ERR("failed to allocate memory for handles array");
		return DOCA_ERROR_NO_MEMORY;
	}

	handles_ptr = handles_list;
	for (size_t i = 0; i < count; i++) {
		struct doca_apsh_handle *handle = *(*handles) + i;

		(*handles)[i] = handle;
		handle->process = process;
		handle->count = count;
		insert_handle_data(handle, handles_ptr);
		handles_ptr = next_handle(handles_ptr, process->system->os_type);
	}

	*handles_size = count;
	return DOCA_SUCCESS;
}

void
doca_apsh_handles_free(struct doca_apsh_handle **handles)
{
	if (handles == NULL)
		return;
	if (handles[0]->process->system->os_type == DOCA_APSH_SYSTEM_WINDOWS)
		windows_free_handles((windows_handles_t *)(handles[0]->handle_data));
	free(*handles);
	free(handles);
}

const void *
__doca_apsh_handle_info_get(struct doca_apsh_handle *handle, enum doca_apsh_handle_attr attr)
{
	return g_hash_table_lookup(handle->data, GINT_TO_POINTER(attr));
}
