/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "jsonout.h"

static const char *count_key = "count";
static const char *pluginname_key = "plugin";
static const char *titles_key = "titles";
static const char *data_key = "data";

static json_object *pool[JSONOUT_POOL_SIZE];
static uint16_t curpos;

void
jsonout_init(void)
{
	int i = 0;

	curpos = 0;
	for (i = 0; i < JSONOUT_POOL_SIZE; i++)
		pool[i] = NULL;
}

json_object *
jsonout_new(char *pluginname, char *titles[])
{
	if (pool[curpos] == NULL) {
	} else {
		/* free jsonout object */
		while (json_object_put(pool[curpos]) == 0) {
			/* keep freeing */
		}

		pool[curpos] = NULL;
	}

	pool[curpos] = json_object_new_object();
	json_object *jpluginname = json_object_new_string(pluginname);
	json_object *jtitleslist = json_object_new_array();
	int i = 0;

	while (1) {
		char *title = titles[i++];

		if (!(*title))
			break;

		json_object *jtitle = json_object_new_string(title);

		json_object_array_add(jtitleslist, jtitle);
	}

	json_object_object_add(pool[curpos], pluginname_key, jpluginname);	 // add plugin name
	json_object_object_add(pool[curpos], titles_key, jtitleslist);		 // add titles list
	json_object_object_add(pool[curpos], data_key, json_object_new_array()); // add data
	json_object_object_add(pool[curpos], count_key, json_object_new_int(0)); // add count

	return pool[curpos];
}

json_object *
jsonout_last_get(int offset)
{
	int index = curpos;
	return pool[index];
}

json_object *
jsonout_data_element_new(json_object *jsonout)
{
	json_object *jdata;

	json_object_object_get_ex(jsonout, data_key, &jdata);
	json_object *jdataele = json_object_new_array();

	json_object_array_add(jdata, jdataele);

	/* update count */
	int count = json_object_array_length(jdata);

	json_object_object_add(pool[curpos], count_key, json_object_new_int(count));

	return json_object_array_get_idx(jdata, count - 1); /* return the last element */
}

void
jsonout_data_element_value_add(json_object *dataelement, char *value)
{
	json_object_array_add(dataelement, json_object_new_string(value));
}

void
jsonout_data_element_unicode_add(json_object *dataelement, unicode_string_t *us)
{
	unicode_string_t out = {0};

	if (us && APSH_S == apsh_convert_str_encoding(us, &out, "UTF-8")) {
		jsonout_data_element_value_add(dataelement, (char *)out.contents);
		free(out.contents);
	}
}

void
jsonout_data_element_unicode_string_add(json_object *dataelement, char *wstr)
{
	char buf[1024];
	char *pbuf = buf;
	char *p = wstr;
	uint32_t len = sizeof(buf) - 2;

	while (len) {
		if (*p == 0 && *(p + 1) == 0)
			break;
		sprintf(pbuf, "%c", *p);
		p += 2;
		pbuf++;
		len--;
	}

	*pbuf = '\0';

	jsonout_data_element_value_add(dataelement, buf);
}

void
jsonout_data_element_value_sprintf(json_object *dataelement, const char *format, ...)
{
	va_list args;
	char *p = NULL;

	va_start(args, format);
	if (vasprintf(&p, format, args) < 0)
		p = NULL;
	va_end(args);
	if (p) {
		json_object_array_add(dataelement, json_object_new_string(p));
		free(p);
	}
}

#if 0
uint32_t jsonout_add_data(json_object *jsonout, json_object *data)
{
	json_object *dataarray;

	json_object_object_get_ex(jsonout, data_key, &dataarray);
	json_object_array_add(dataarray, data);
    /* update count */
	int count = json_object_array_length(dataarray);

	json_object_object_add(pool[curpos], count_key, json_object_new_int(count));
	return count;
}

json_object *jsonout_get_by_name(char *pluginname)
{
}

json_object *jsonout_get_last(int offset)
{
}

#endif

const char *
jsonout_to_string(json_object *jsonout)
{
	return json_object_to_json_string(jsonout);
}

void
jsonout_treegrid_printout(json_object *jsonout)
{
	/* print titles */
	int i = 0, j = 0, k = 0;
	json_object *jtitles;

	if (false == json_object_object_get_ex(jsonout, "titles", &jtitles))
		return;

	int colnum = json_object_array_length(jtitles);

	for (i = 0; i < colnum; i++) {
		json_object *title = json_object_array_get_idx(jtitles, i);

		printf(" %s\t", json_object_get_string(title));
	}
	printf("\n");

	/* print data */
	json_object *jdata;

	if (false == json_object_object_get_ex(jsonout, "data", &jdata))
		return;

	int count = json_object_array_length(jdata);

	for (j = 0; j < count; j++) {
		/* print one line */
		json_object *data_element = json_object_array_get_idx(jdata, j);

		colnum = json_object_array_length(data_element);
		for (k = 0; k < colnum; k++) {
			json_object *value = json_object_array_get_idx(data_element, k);

			printf(" %s\t", json_object_get_string(value));
		}
		printf("\n");
	}
}

void
jsonout_deinit(void)
{
	int i = 0;

	for (i = 0; i < JSONOUT_POOL_SIZE; i++) {
		if (pool[i]) {
			while (json_object_put(pool[i]) != 1) {
				// printf("keep freeing\n");
			}

			pool[i] = NULL;
		}
	}
}
