/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef __JSONOUT_H__
#define __JSONOUT_H__

#include <stdbool.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdio.h>
#include <json-c/json.h>
#include "vmm.h"

#define JSONOUT_POOL_SIZE  10


void jsonout_init(void);

/**
 * jsonout_new
 * initialize an jsonout object from pool
 * e.g {  "plugin":<plugin name>, 
 *         "tiltes": [<tiltes>],
 *         "data": [], 
 *         "count": 0
 *     }
 * @param[in] plugin name
 * @param[in] string array holding titles 
 * @return json object 
 */
json_object * jsonout_new(char * pluginname, char * titles[]);

/**
 * jsonout_last_get
 * get last jsonout object
 * @param[in] offset, 0 - current/last,  -1 the second last, as so on
 * @return last jsonout object 
 */
json_object * jsonout_last_get(int offset);

/**
 * jsonout_data_element_new
 * add empty element in jsonout object data array
 * e.g {  "plugin":<plugin name>, 
 *         "tiltes": [<tiltes>],
 *         "data": [ [], ],  // empty element
 *         "count": 0
 *     }
 * @param[in] json out object
 * @return data object
 */
json_object * jsonout_data_element_new(json_object * jsonout);

/**
 * jsonout_data_element_value_add
 * add value in data element
 * e.g {  "plugin":<plugin name>, 
 *         "tiltes": [<tiltes>],
 *         "data": [ [<value>, ], ],  // added value
 *         "count": 0
 *     } 
 * @param[in] json data element
 * @return data object
 */
void jsonout_data_element_value_add(json_object * dataelement, char * value);

void jsonout_data_element_unicode_add(json_object * dataelement, unicode_string_t * us);

void jsonout_data_element_unicode_string_add(json_object * dataelement, char * wstr);

void jsonout_data_element_value_sprintf(json_object * dataelement, const char * format, ... ) __attribute__ ((format (printf, 2, 3)));

/**
 * jsonout_to_string
 * copnvewrt jsonout object to json string 
 * @return json object 
 */
const char * jsonout_to_string(json_object * jsonout);

/**
 * jsonout_treegrid_printout
 * print jsonout object in volatility treegrid style
 * @param[in] json out object
 */
void jsonout_treegrid_printout(json_object * jsonout);

/*free jsonout objects in the pool */
void jsonout_deinit(void);

#endif /* __JSONOUT_H__ */
