/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef QP_CHANNEL_H_
#define QP_CHANNEL_H_

#include <stddef.h>
#include <stdbool.h>
#include <stdint.h>
#include <limits.h>

#include <doca_comm_channel.h>
#include <doca_dev.h>

#define MAX_DATA_SIZE 4080 /** calculated from the MTU supported by the channel, minus 16B for header */
#define MIN_DATA_SIZE 256
#define MAX_QUEUE_SIZE 8192
#define MIN_QUEUE_SIZE 16


enum {
	HELLO_MSG,
	REJECT_MSG,
	ARRIVE_MSG,
	SEND_MSG,
	ERR_SEND_MSG,
	ERR_RECV_MSG,
	UNKNOWN_MSG,
	NO_MSG_FOUND,
	WRONG_QPN,
	DESTROYED_QP,
	QUEUE_FULL,
};

/**
 * @brief Header types for the messages.
 */
enum msg_header_type {
	DOCA_CC_HANDSHAKE = 1,	/**< Configuration message, part of the first handshake */
	DOCA_CC_SEND_MSG,	/**< Data message */
	DOCA_CC_REJECT,
};

/**
 * @brief Error types to handle.
 */
enum cqe_error_type {
	DOCA_CC_FLUSH_ERR = 1,
	DOCA_CC_PEER_ERR,
};


/**
 * @brief 32 bit header structure for every handshake message sent.
 */
#pragma pack(push, 1)
struct handshake_header {
	uint32_t version : 4;
	uint32_t type : 4;
	uint32_t mtu : 14;
	uint32_t reserved : 10;
};
#pragma pack(pop)
static_assert((CHAR_BIT * sizeof(struct handshake_header)) == 32, "Invalid size for handshake_header");

/**
 * @brief 32 bit immediate message structure.
 */
union imm_msg {
	struct handshake_header handshake;
	uint32_t imm_msg_uint;
};
static_assert((CHAR_BIT * sizeof(union imm_msg)) == 32, "Invalid size for imm_msg");

struct doca_cc_properties {
	struct doca_dev *hw_dev;
	struct doca_dev_rep *dev_rep;
	uint16_t log_msg_size;
	uint16_t log_rcv_queue_size;
	uint16_t log_snd_queue_size;
	uint16_t max_clients;
};

struct cc_qp;

struct cc_cq_obj;

struct cc_qp_common;

struct ibv_context *get_ibv_ctx(struct cc_qp_common *common);

struct doca_dev *get_doca_dev(struct cc_qp_common *common);

struct priv_doca_comp_events_channel *cc_qp_event_ch_create(struct cc_qp_common *common_rs, int flag);

int cc_event_channel_subscribe(struct cc_cq_obj *cq, struct priv_doca_comp_events_channel *event_channel);

struct cc_qp_common *cc_qp_common_init(struct doca_cc_properties *cc_properties, bool *is_service);

void cc_qp_common_destroy(struct cc_qp_common *common_resource);

struct cc_av {
	int dct;
};

struct destroyed_qp;

struct destroyed_qps_list {
	struct destroyed_qp *destroyed_qp;
};
/*
* return true if link layer is IB, false if it is ETH
*/
bool cc_get_common_ib_link_layer(const struct cc_qp_common *common_rs);


/*
 * Create CQ for the use of all endpoint's QPs.
 */
struct cc_cq_obj *cq_create(struct cc_qp_common *common_rs, int log_cq_size);

void cq_destroy(struct cc_cq_obj *cq);

/*
 * Create 2 QPs, one for rcv and one for snd
 */
struct cc_qp *cc_qpp_create(struct cc_qp_common *common, const struct doca_cc_properties *cc_attr,

			    struct cc_cq_obj *snd_cq, struct cc_cq_obj *rcv_cq);

void cc_qpp_destroy(struct cc_qp *qpp, struct destroyed_qps_list *destroyed_list);

/*
 * qp - qp to connect
 * service - service to register to
 * Is called only by the client (host), will pratically create 2 sevices: snd_$service, rcv_$service
 * This called only by the client, need to work also on windows.
 */
int cc_qp_register_client(struct cc_qp *qpp, struct cc_qp_common *common,
			  const char *service);


/*
 * qp - qp to connect
 * service - service to register to
 * max_clients - the maximal amount of client connections the doca_dev on the DPU can maintain.
 * Is called only by the server (ARM), will pratically create 2 sevices: snd_$service, rcv_$service
 * This is called only by the server, no need for windows.
 */
int cc_qp_register_server(struct cc_qp *qpp, struct cc_qp_common *common, const char *service, uint16_t max_clients);

/*
 * To send a (non handshake) msg:
 * void *msg = cc_qp_msg_alloc(qp);
 * ... prepare/get the msg into "msg" (for UDS - read(uds_fd, msg))
 * cc_qp_msg_send(qp, msg, len);
 * Will update the wqe and send doorbell.
 * Currently - Order of alloc must be the same as order of send
 * Returns -1 if no credits
 */
int cc_qp_msg_send(struct cc_qp *qpp, struct cc_cq_obj *cq, const void *data, size_t len,
		   struct cc_av *dest_av, const union imm_msg *send_header);


/*
 * To send a handshake msg:
 * void *msg = cc_qp_msg_alloc(qp);
 * cc_qp_msg_send_handshake(qp, msg, len);
 * Will update the wqe and send doorbell.
 * Currently - Order of alloc must be the same as order of send
 * Returns -1 if no credits
 */
int
cc_qp_msg_send_handshake(struct cc_qp *qpp, struct cc_cq_obj *cq,
			 struct cc_av *dest_av, const union imm_msg *send_header);

/*
 *
 * output message code enum.
 *
 *
 */
int cc_qp_msg_recv(struct cc_qp *qpp, struct cc_cq_obj *cq, void *data, size_t *len,
		   struct destroyed_qps_list *destroyed_qps_list, union imm_msg *recv_header);

int cc_qp_check_send_cqe(struct cc_qp *qpp, struct cc_cq_obj *cq, struct destroyed_qps_list *destroyed_qps_list, struct cc_av *av);

void cc_channel_destroy(struct cc_qp *qpp);


int vhca_id_get(struct cc_qp *qpp);

void vhca_id_set(int vhca_id, struct cc_qp *qpp);

struct cc_av *av_get(struct cc_qp *qpp);

void destroy_destroyed_qps_list(struct destroyed_qps_list *list);

bool is_armed(const struct cc_cq_obj *cq);

int recover_dci(struct cc_qp *qpp);

uint32_t get_cqe_error_syndrom(const struct cc_qp *qpp);

int get_cqe_error_dest_dct(const struct cc_qp *qpp);

void add_wqe_to_repost_buf(const struct cc_qp *qpp, uint8_t *temp_buf, int bad_dctn);

void repost_flushed_wqes(const struct cc_qp *qpp, uint8_t *temp_repost_buf);

void cc_force_arm(struct cc_cq_obj *cq);
#endif /* QP_CHANNEL_H_ */
