/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <malloc.h>

#include <doca_log.h>
#include <priv_doca_dev.h>
#include <priv_doca_os_abstraction.hpp>
#include <doca_types.h>

#include "cc_devx.h"

#define MLX5_GENERAL_OBJ_TYPE_CHANNEL_SERVICE 0x003B
#define MLX5_GENERAL_OBJ_TYPE_CHANNEL_CONNECTION 0x003C
#define RQ_TYPE_SRQ 1
#define RECV_WQE_SIZE 16
#define LOG_MLX5_SEND_WQE_BB 6
#define UD_ADDRESS_VECTOR_LONG 48
#define SEND_WQE_CTRL_INDEX_MASK 0xffff00

DOCA_LOG_REGISTER(COMM_CH::DEVX);

void
arm_cq(struct cc_cq_obj *cq)
{
	priv_doca_comp_events_channel_arm(cq->events_channel);
	cq->is_armed = true;
}

void
force_arm(struct cc_cq_obj *cq)
{
#ifdef __linux__
	if (!priv_doca_cq_empty(cq->cq_q))
		return;
#endif // __linux__

	priv_doca_comp_events_channel_force_arm(cq->events_channel);
	cq->is_armed = true;
}

int
devx_reg_mr(struct devx_context *devx_ctx, void *buff, const struct qp_params_reg_mr *params,
	    struct doca_devx_object **mr_obj, struct doca_umem_object **umem)
{
	uint32_t in[MLX5_ST_SZ_DW(create_mkey_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(create_mkey_out)] = {0};
	struct doca_umem_object *mem = NULL;
	struct doca_devx_object *mr;
	uint32_t mem_umem_id = 0;
	int mem_access_flags = 0;
	doca_error_t status = DOCA_SUCCESS;

	if (params->lw)
		mem_access_flags |= DOCA_ACCESS_LOCAL_WRITE;

	status = priv_doca_umem_create(devx_ctx->doca_device, buff, params->size, mem_access_flags, &mem);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_umem_create failed in %s()", __func__);
		return 0;
	}
	priv_doca_umem_id_get(mem, &mem_umem_id);

	MLX5_SET(create_mkey_in, in, opcode, MLX5_CMD_OP_CREATE_MKEY);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.access_mode_1_0,params->access_mode_1_0);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.lw, params->lw);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.lr, params->lr);
	MLX5_SET64(create_mkey_in, in, memory_key_mkey_entry.start_addr, (uintptr_t)buff);
	MLX5_SET64(create_mkey_in, in, memory_key_mkey_entry.len, params->size);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.pd, params->pd);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.qpn, params->qpn);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.mkey_7_0, params->mkey_7_0);
	MLX5_SET(create_mkey_in, in, pg_access, params->pg_access);
	MLX5_SET(create_mkey_in, in, mkey_umem_id, mem_umem_id);
	MLX5_SET(create_mkey_in, in, mkey_umem_valid, 1);

	status = priv_doca_dev_devx_object_create(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out), &mr);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_devx_object_create failed in %s()", __func__);
		priv_doca_umem_destroy(mem);
		return 0;
	}
	*mr_obj = mr;
	*umem = mem;

	return MLX5_GET(create_mkey_out, out, mkey_index);
}

int
cc_q_obj_destroy(struct cc_q_obj *q_obj)
{
	if (q_obj->q_object != NULL) {
		auto status = priv_doca_devx_object_destroy(q_obj->q_object);
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR(
				"failed to destroy q_object in cc_q_obj - priv_doca_devx_object_destroy returned %s",
				doca_get_error_name(status));
			return -1;
		}
	}
	if (q_obj->wq_mem != NULL) {
		auto status = priv_doca_umem_destroy(q_obj->wq_mem);
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("failed to deregister wq_mem in cc_q_obj - priv_doca_umem_destroy returned %s",
				       doca_get_error_name(status));
			return -1;
		}
	}
	if (q_obj->dbrm != NULL) {
		auto status = priv_doca_umem_destroy(q_obj->dbrm);
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("failed to deregister dbrm in cc_q_obj - priv_doca_umem_destroy returned %s",
				      doca_get_error_name(status));
			return -1;
		}
	}
	delete q_obj->wq_buff;
	delete q_obj->dbr;
	return 0;
}

int
create_dci(struct devx_context *devx_ctx, struct priv_doca_uar *uar,
	   int cqn, int pd, size_t log_sq_size, struct cc_q_obj *q)
{
	u8 in[MLX5_ST_SZ_BYTES(create_qp_in)] = {0};
	u8 out[MLX5_ST_SZ_BYTES(create_qp_out)] = {0};
	struct doca_umem_object *pas, *dbrm;
	uint32_t pas_umem_id = 0, dbrm_umem_id = 0;
	void *qpc;
	os_abstraction::aligned_buffer<uint8_t> *buff = nullptr;
	os_abstraction::aligned_buffer<uint32_t> *dbr = nullptr;
	int wq_buff_size;
	doca_error_t status = DOCA_SUCCESS;
	uint32_t uar_page_id = 0;

	if (priv_doca_uar_id_get(uar, &uar_page_id) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_uar_id_get failed in %s()", __func__);
		goto error;
	}

	wq_buff_size = 1 << (log_sq_size + LOG_MLX5_SEND_WQE_BB);
	try {
		buff = new os_abstraction::aligned_buffer<uint8_t>(0x1000, wq_buff_size);
	} catch (...) {
		DOCA_DLOG_ERR("WQ buff buff_sz allocation failed in %s()", __func__);
		goto error;
	}
	memset(buff->data(), 0, wq_buff_size);
	status = priv_doca_umem_create(devx_ctx->doca_device, buff->data(), wq_buff_size, 0, &pas);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_umem_create for pas failed in %s()", __func__);
		goto error_free_wq_buff;
	}
	priv_doca_umem_id_get(pas, &pas_umem_id);

	try {
		dbr = new os_abstraction::aligned_buffer<uint32_t>(0x40, 2);
	} catch (...) {
		DOCA_DLOG_ERR("dbr allocation failed in %s()", __func__);
		goto error_free_buff_umem;
	}

	status = priv_doca_umem_create(devx_ctx->doca_device, dbr->data(), 0x8, 0, &dbrm);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_umem_create for dbrm fail in %s()", __func__);
		goto error_free_dbr;
	}
	priv_doca_umem_id_get(dbrm, &dbrm_umem_id);

	MLX5_SET(create_qp_in, in, opcode, MLX5_CMD_OP_CREATE_QP);

	qpc = MLX5_ADDR_OF(create_qp_in, in, qpc);
	MLX5_SET(qpc, qpc, st, MLX5_QPC_ST_DCI);
	MLX5_SET(qpc, qpc, pd, pd);
	MLX5_SET(qpc, qpc, uar_page, uar_page_id);
	MLX5_SET(qpc, qpc, cqn_snd, cqn);
	MLX5_SET(qpc, qpc, cqn_rcv, cqn);
	MLX5_SET(qpc, qpc, log_sq_size, static_cast<uint32_t>(log_sq_size));
	MLX5_SET(qpc, qpc, rq_type, RQ_TYPE_SRQ);
	MLX5_SET(qpc, qpc, log_num_dci_stream_channels, 1);
	MLX5_SET(create_qp_in, in, wq_umem_id, pas_umem_id);
	MLX5_SET(qpc, qpc, dbr_umem_id, dbrm_umem_id);
	MLX5_SET64(qpc, qpc, dbr_addr, 0);
	MLX5_SET(qpc, qpc, ts_format, MLX5_QPC_TIMESTAMP_FORMAT_DEFAULT);

	MLX5_SET(qpc, qpc, primary_address_path.fl, 1);

	status = priv_doca_dev_devx_object_create(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out),
						  &q->q_object);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_devx_object_create for q_object failed in %s()", __func__);
		goto error_free_dbr_umem;
	}

	q->dbr = dbr;
	q->wq_buff = buff;
	q->wq_mem = pas;
	q->dbrm = dbrm;

	return MLX5_GET(create_qp_out, out, qpn);

error_free_dbr_umem:
	priv_doca_umem_destroy(dbrm);
error_free_dbr:
	delete dbr;
error_free_buff_umem:
	priv_doca_umem_destroy(pas);
error_free_wq_buff:
	delete buff;
error:
	return 0;
}

int
create_rmp(struct devx_context *devx_ctx, int pd, int log_rq_size, struct cc_q_obj *rmp_obj)
{
	struct doca_umem_object *wq_mem = nullptr, *dbrm = nullptr;
	os_abstraction::aligned_buffer<uint8_t> *wq_buff = nullptr;
	os_abstraction::aligned_buffer<uint32_t> *dbr = nullptr;
	u8 in[MLX5_ST_SZ_BYTES(create_rmp_in)] = {0};
	u8 out[MLX5_ST_SZ_BYTES(create_rmp_out)] = {0};
	void *rmpc, *wqc;
	int wq_buff_size;
	int mem_access_flags = 0; /* allow only local read */
	doca_error_t status = DOCA_SUCCESS;
	uint32_t wq_mem_umem_id = 0, dbrm_umem_id = 0;

	try {
		dbr = new os_abstraction::aligned_buffer<uint32_t>(0x40, 2);
	} catch (...) {
		DOCA_DLOG_ERR("dbr allocation failed in %s()", __func__);
		goto error;
	}
	status = priv_doca_umem_create(devx_ctx->doca_device, dbr->data(), 0x8, mem_access_flags, &dbrm);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_umem_create for dbrm failed in %s()", __func__);
		goto error_free_dbr;
	}
	priv_doca_umem_id_get(dbrm, &dbrm_umem_id);

	wq_buff_size = (1 << log_rq_size) * RECV_WQE_SIZE;
	try {
		wq_buff = new os_abstraction::aligned_buffer<uint8_t>(0x40, wq_buff_size);
	} catch (...) {
		DOCA_DLOG_ERR("WQ buff allocation failed in %s()", __func__);
		goto error_free_dbr_umem;
	}
	status = priv_doca_umem_create(devx_ctx->doca_device, wq_buff->data(), wq_buff_size, mem_access_flags, &wq_mem);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_umem_create for wq_mem failed in %s()", __func__);
		goto error_free_wq_buff;
	}
	priv_doca_umem_id_get(wq_mem, &wq_mem_umem_id);

	MLX5_SET(create_rmp_in, in, opcode, MLX5_CMD_OP_CREATE_RMP);
	rmpc = MLX5_ADDR_OF(create_rmp_in, in, ctx);
	MLX5_SET(rmpc, rmpc, state, 1);
	MLX5_SET(rmpc, rmpc, basic_cyclic_rcv_wqe, 1);

	wqc = MLX5_ADDR_OF(rmpc, rmpc, wq);
	MLX5_SET(wq, wqc, pd, pd);
	MLX5_SET(wq, wqc, dbr_umem_id, dbrm_umem_id);
	MLX5_SET(wq, wqc, wq_umem_id, wq_mem_umem_id);
	MLX5_SET(wq, wqc, log_wq_stride, 4);
	MLX5_SET(wq, wqc, log_wq_sz, log_rq_size);
	MLX5_SET(wq, wqc, wq_type, MLX5_WQ_TYPE_CYCLIC);

	status = priv_doca_dev_devx_object_create(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out),
						  &rmp_obj->q_object);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_devx_object_create failed for q_object in %s()", __func__);
		goto error_free_wq_buff_umem;
	}
	rmp_obj->wq_buff = wq_buff;
	rmp_obj->dbr = dbr;
	rmp_obj->wq_mem = wq_mem;
	rmp_obj->dbrm = dbrm;

	return MLX5_GET(create_rmp_out, out, rmpn);
error_free_wq_buff_umem:
	priv_doca_umem_destroy(wq_mem);
error_free_wq_buff:
	delete wq_buff;
error_free_dbr_umem:
	priv_doca_umem_destroy(dbrm);
error_free_dbr:
	delete dbr;
error:
	return 0;
}

int
create_dct(struct devx_context *devx_ctx, int cqn, int pd, uint16_t port_num, int srqn, struct cc_q_obj *dct_obj)
{
	u8 in[MLX5_ST_SZ_BYTES(create_dct_in)] = {0};
	u8 out[MLX5_ST_SZ_BYTES(create_dct_out)] = {0};
	void *dctc;
	doca_error_t status;

	MLX5_SET(create_dct_in, in, opcode, MLX5_CMD_OP_CREATE_DCT);
	dctc = MLX5_ADDR_OF(create_dct_in, in, dct_context_entry);
	MLX5_SET(dctc, dctc, cqn, cqn);
	MLX5_SET(dctc, dctc, port, port_num);
	MLX5_SET(dctc, dctc, mtu, MLX5_QPC_MTU_4K_BYTES);
	MLX5_SET(dctc, dctc, srqn_xrqn, srqn);
	MLX5_SET(dctc, dctc, pd, pd);
	MLX5_SET(dctc, dctc, my_addr_index, 0xff);
	MLX5_SET(dctc, dctc, min_rnr_nak, 1);
	MLX5_SET64(dctc, dctc, dc_access_key, 0);

	status = priv_doca_dev_devx_object_create(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out),
						  &dct_obj->q_object);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_devx_object_create failed for q_object in %s()", __func__);
		return 0;
	}
	dct_obj->wq_mem = NULL;
	dct_obj->dbrm = NULL;
	dct_obj->wq_buff = NULL;
	dct_obj->dbr = NULL;

	return MLX5_GET(create_dct_out, out, dctn);
}

int
get_issi(struct devx_context *devx_ctx)
{
	int issi;

	u8 in[MLX5_ST_SZ_BYTES(query_issi_in)] = {0};
	u8 out[MLX5_ST_SZ_BYTES(query_issi_out)] = {0};

	MLX5_SET(query_issi_in, in, opcode, MLX5_CMD_OP_QUERY_ISSI);

	auto status = priv_doca_dev_prm_cmd_send(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_prm_cmd_send failed for QUERY_ISSI");
		return -1;
	}

	issi = MLX5_GET(query_issi_out, out, current_issi);
	return issi;
}

int
devx_to_init(struct doca_devx_object *obj, const struct qp_params_init *params)
{
	uint32_t in[MLX5_ST_SZ_DW(rst2init_qp_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(rst2init_qp_out)] = {0};
	void *qpc = MLX5_ADDR_OF(rst2init_qp_in, in, qpc);

	MLX5_SET(rst2init_qp_in, in, opcode, MLX5_CMD_OP_RST2INIT_QP);
	MLX5_SET(rst2init_qp_in, in, qpn, params->qp);

	MLX5_SET(qpc, qpc, primary_address_path.vhca_port_num, params->vhca_port_num);
	MLX5_SET(qpc, qpc, pm_state, MLX5_QPC_PM_STATE_MIGRATED);

	return priv_doca_dev_devx_object_modify(obj, in, sizeof(in), out, sizeof(out));
}

int
devx_to_rtr(struct doca_devx_object *obj, const struct qp_params_rtr *params)
{
	uint32_t in[MLX5_ST_SZ_DW(init2rtr_qp_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(init2rtr_qp_out)] = {0};
	void *qpc = MLX5_ADDR_OF(init2rtr_qp_in, in, qpc);

	MLX5_SET(init2rtr_qp_in, in, opcode, MLX5_CMD_OP_INIT2RTR_QP);
	MLX5_SET(init2rtr_qp_in, in, qpn, params->qp);

	MLX5_SET(qpc, qpc, log_msg_max, params->log_msg_max);
	MLX5_SET(qpc, qpc, mtu, params->mtu);
	MLX5_SET(qpc, qpc, deth_sqpn_or_initiator_src_dct, params->dctn);
	MLX5_SET(qpc, qpc, primary_address_path.hop_limit, params->hop_limit);

	return priv_doca_dev_devx_object_modify(obj, in, sizeof(in), out, sizeof(out));
}

int
devx_to_rts(struct doca_devx_object *obj, const struct qp_params_rts *params)
{
	uint32_t in[MLX5_ST_SZ_DW(rtr2rts_qp_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(rtr2rts_qp_out)] = {0};
	void *qpc = MLX5_ADDR_OF(rtr2rts_qp_in, in, qpc);

	MLX5_SET(rtr2rts_qp_in, in, opcode, MLX5_CMD_OP_RTR2RTS_QP);
	MLX5_SET(rtr2rts_qp_in, in, qpn, params->qp);

	MLX5_SET(qpc, qpc, log_ack_req_freq, params->log_ack_req_freq);
	MLX5_SET(qpc, qpc, retry_count, params->retry_count);
	MLX5_SET(qpc, qpc, rnr_retry, params->rnr_retry);
	MLX5_SET(qpc, qpc, primary_address_path.ack_timeout, params->ack_timeout);

	return priv_doca_dev_devx_object_modify(obj, in, sizeof(in), out, sizeof(out));
}

int
to_rst(struct doca_devx_object *obj, int qpn)
{
	uint32_t in[MLX5_ST_SZ_DW(qp_2rst_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(qp_2rst_out)] = {0};

	MLX5_SET(qp_2rst_in, in, opcode, MLX5_CMD_OP_2RST_QP);
	MLX5_SET(qp_2rst_in, in, qpn, qpn);

	return priv_doca_dev_devx_object_modify(obj, in, sizeof(in), out, sizeof(out));
}

void
devx_set_ud_addr(uint8_t *ud_addr, const struct qp_params_set_ud_addr *params)
{
	memset(ud_addr, 0, UD_ADDRESS_VECTOR_LONG);

	MLX5_SET(ud_adrs_vector, ud_addr, ext, params->ext);
	MLX5_SET(ud_adrs_vector, ud_addr, destination_qp_dct, params->dest_qpn);
	MLX5_SET(ud_adrs_vector, ud_addr, rlid_udp_sport, params->rlid);
	MLX5_SET(ud_adrs_vector, ud_addr, src_addr_index, params->src_addr_index);
}

int
devx_query_capabilities(const struct devx_context *devx_ctx, struct cc_qp_cap *cap)
{
	uint32_t in[MLX5_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};

	MLX5_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	MLX5_SET(query_hca_cap_in, in, op_mod, MLX5_HCA_CAP_OPMOD_GET_CUR | MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE);
	auto status = priv_doca_dev_prm_cmd_send(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS)
		return -1;

	cap->log_max_msg = MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.log_max_msg);
	cap->log_max_queue_sz = MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.log_max_qp_sz);
	cap->num_ports = MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.num_ports);
	cap->num_vhca_ports = MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.num_vhca_ports);
	cap->port_type = MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.port_type);

	if (MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.hca_cap_2) == 1) {
		memset(in, 0, sizeof(in));
		memset(out, 0, sizeof(out));

		MLX5_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
		MLX5_SET(query_hca_cap_in, in, op_mod,
			 MLX5_HCA_CAP_OPMOD_GET_CUR | MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE_CAP_2);
		status = priv_doca_dev_prm_cmd_send(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out));
		if (status != DOCA_SUCCESS)
			return -1;

		cap->log_max_connections = MLX5_GET(query_hca_cap_out, out,
					   	    capability.cmd_hca_cap_2.log_max_channel_service_connection);
	}

	return 0;
}

struct doca_devx_object *
register_channel_connection(struct devx_context *devx_ctx, const struct channel_params *params)
{
	struct doca_devx_object *connection_obj;
	doca_error_t status;
	uint32_t in[MLX5_ST_SZ_DW(create_channel_connection_obj_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(create_channel_connection_obj_out)] = {0};

	void *hdr = MLX5_ADDR_OF(create_channel_connection_obj_in, in, hdr);
	void *ch_conn_obj = MLX5_ADDR_OF(create_channel_connection_obj_in, in, ch_conn_obj);
	void *service_name = MLX5_ADDR_OF(channel_connection_obj, ch_conn_obj, service_name);

	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJ_TYPE_CHANNEL_CONNECTION);
	MLX5_SET(channel_connection_obj, ch_conn_obj, connection_type, params->flag);
	MLX5_SET(channel_connection_obj, ch_conn_obj, dst_vhca_id, params->vhca_id);
	if (params->service_name != NULL)
		os_abstraction::priv_doca_strlcpy((char *)service_name, params->service_name,
			MLX5_FLD_SZ_BYTES(channel_connection_obj, service_name));
	MLX5_SET(channel_connection_obj, ch_conn_obj, my_qpn_or_dci, params->dci_dct_num);
	MLX5_SET(channel_connection_obj, ch_conn_obj, transport_type, params->transport_type);

	status = priv_doca_dev_devx_object_create(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out),
						  &connection_obj);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create channel connection object with error %s",
			      doca_get_error_name(status));
		return NULL;
	}

	return connection_obj;
}

struct doca_devx_object *
register_channel_service(struct devx_context *devx_ctx, const struct channel_params *params)
{
	struct doca_devx_object *service_obj;
	doca_error_t status;
	uint32_t in[MLX5_ST_SZ_DW(create_channel_service_obj_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(create_channel_service_obj_out)] = {0};

	void *hdr = MLX5_ADDR_OF(create_channel_service_obj_in, in, hdr);
	void *ch_svc_obj = MLX5_ADDR_OF(create_channel_service_obj_in, in, ch_svc_obj);
	void *service_name = MLX5_ADDR_OF(channel_service_obj, ch_svc_obj, service_name);

	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJ_TYPE_CHANNEL_SERVICE);

	MLX5_SET(channel_service_obj, ch_svc_obj, allowed_vhca_id, params->vhca_id);
	MLX5_SET(channel_service_obj, ch_svc_obj, service_qpn_or_dct, params->dci_dct_num);
	MLX5_SET(channel_service_obj, ch_svc_obj, service_qkey, params->qkey);
	MLX5_SET(channel_service_obj, ch_svc_obj, transport_type, params->transport_type);
	MLX5_SET(channel_service_obj, ch_svc_obj, max_num_connected_client, params->max_num_connected_client);

	os_abstraction::priv_doca_strlcpy((char *)service_name, params->service_name,
					  MLX5_FLD_SZ_BYTES(channel_connection_obj, service_name));

	status = priv_doca_dev_devx_object_create(devx_ctx->doca_device, in, sizeof(in), out, sizeof(out),
						  &service_obj);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create channel service object with error %s",
			      doca_get_error_name(status));
		return NULL;
	}

	return service_obj;
}

void
update_wqe_ctrl_index(uint8_t *wqe_ptr, uint32_t value)
{
	((mlx5_wqe_ctrl_seg *)wqe_ptr)->opmod_idx_opcode = htobe32(((0xffff & (value)) << 8) |
			   ((~SEND_WQE_CTRL_INDEX_MASK) & (be32toh(((mlx5_wqe_ctrl_seg *)wqe_ptr)->opmod_idx_opcode))));
}
