/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <private/common_defines.h>
#include <doca_log.h>
#include <doca_dev.h>
#include <priv_doca_dev.h>
#include <priv_doca_os_abstraction.hpp>
#include <priv_doca_cq.h>
#include <doca_types.h>
#include <cc_devx.h>
#include <qp_channel.h>
#include <doca_comm_channel.h>

#ifndef __linux__
#include <windows.h>
#include <stdexcept>
#endif /* __linux__ */

#define CURRENT_COMM_CHANNEL_VERSION 0
#define DEFAULT_MAX_MSG_SIZE 1500
#define DEFAULT_SND_QUEUE_SIZE 128
#define	DEFAULT_RCV_QUEUE_SIZE 128

DOCA_LOG_REGISTER(COMM_CH::CC);

/**
 * @brief Service max name size includeing the terminating null byte ('\0').
 */
#define MAX_SERVICE_NAME_LEN 120

struct doca_comm_channel_ep_t {
	struct cc_qp_common *common_rs; /**< Common resources handle for device. */
	struct cc_cq_obj *snd_cq_obj;
	struct cc_cq_obj *rcv_cq_obj;
	struct doca_cc_properties ep_properties;
	struct cc_qp **qpp;
	struct priv_doca_comp_events_channel *snd_event_ch;
	struct priv_doca_comp_events_channel *rcv_event_ch;
	struct peer_addr_list *peer_addr_list;
	size_t num_ports;
	uint32_t flags;
	bool is_service;
	bool is_active; /**< for service - was listen called successfully*/
	bool connection_in_progress; /**< for client - marks if a connection is in progress*/
	struct destroyed_qps_list destroyed_qps_list;
	size_t mtu;
};

struct doca_comm_channel_addr_stat_t {
	uint64_t rx_bytes;
	uint64_t tx_bytes;
	uint64_t rx_messages;
	uint64_t tx_messages;
	uint64_t tx_in_flight_messages;
};

struct doca_comm_channel_addr_t {
	struct doca_comm_channel_ep_t *ep;
	int vhca_id;
	uint32_t version;
	uint64_t user_data;
	struct cc_av av;
	bool error;
	bool user_notified;
	struct doca_comm_channel_addr_t *next;
	struct doca_comm_channel_addr_stat_t current_stat;
	struct doca_comm_channel_addr_stat_t snapshot_stat; /**< info updated through the update_info function */
};

struct peer_addr_list {
	int len;
	struct doca_comm_channel_addr_t *peer_addr;
};

static inline int
get_peers_vhca_id_index(const struct doca_comm_channel_ep_t *local_ep, int peers_vhca_id)
{
	int i;

	for (i = 0; i < local_ep->num_ports; ++i) {
		if (local_ep->qpp[i] == NULL)
			continue;

		if (vhca_id_get(local_ep->qpp[i]) == peers_vhca_id)
			return i;
	}

	return -1;
}

doca_error_t
doca_comm_channel_get_max_service_name_len(uint32_t *max_service_name_len)
{
	if (max_service_name_len == NULL) {
		DOCA_DLOG_ERR("Unable to get max service name size: Received invalid input - max_service_name_len=%p",
			      max_service_name_len);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*max_service_name_len = MAX_SERVICE_NAME_LEN;

	return DOCA_SUCCESS;
}

doca_error_t
doca_comm_channel_get_max_message_size(struct doca_devinfo *devinfo, uint32_t *max_message_size)
{
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t dev_log_max_msg = 0;
	uint32_t dev_max_msg = 0;
	doca_error_t status;

	if (devinfo == NULL) {
		DOCA_DLOG_ERR("Unable to get max message size: Received invalid input - devinfo=%p",
			      devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (max_message_size == NULL) {
		DOCA_DLOG_ERR("Unable to get max message size: Received invalid input - max_message_size=%p",
			      max_message_size);
		return DOCA_ERROR_INVALID_VALUE;
	}

	status = priv_doca_devinfo_hca_caps_by_opmod_query(devinfo,
						(MLX5_HCA_CAP_OPMOD_GET_CUR | MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE),
						out, sizeof(out));

	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to get max message size: Querying capabilities failed");
		return DOCA_ERROR_UNEXPECTED;
	}
	dev_log_max_msg = MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.log_max_msg);
	dev_max_msg = (dev_log_max_msg == 0 ? 0 : 1 << dev_log_max_msg);

	*max_message_size = (dev_max_msg < MAX_DATA_SIZE ? dev_max_msg : MAX_DATA_SIZE);

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_comm_channel_get_max_queue_size(struct doca_devinfo *devinfo, uint32_t *max_queue_size)
{
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t dev_log_max_queue_size = 0;
	uint32_t dev_max_queue_size = 0;
	doca_error_t status;

	if (devinfo == NULL) {
		DOCA_DLOG_ERR("Unable to get max queue size: Received invalid input - devinfo=%p",
			      devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	status = priv_doca_devinfo_hca_caps_by_opmod_query(devinfo,
						(MLX5_HCA_CAP_OPMOD_GET_CUR | MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE),
						out, sizeof(out));

	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to get max queue size: Querying capabilities failed");
		return DOCA_ERROR_UNEXPECTED;
	}
	dev_log_max_queue_size = MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.log_max_msg);
	dev_max_queue_size  = (dev_log_max_queue_size == 0 ? 0 : 1 << dev_log_max_queue_size);

	*max_queue_size = (dev_max_queue_size < MAX_QUEUE_SIZE ? dev_max_queue_size : MAX_QUEUE_SIZE);

	return DOCA_SUCCESS;
}

doca_error_t
doca_comm_channel_get_max_send_queue_size(struct doca_devinfo *devinfo, uint32_t *max_send_queue_size)
{
	if (max_send_queue_size == NULL) {
		DOCA_DLOG_ERR(
			"Unable to get max send queue size: Received invalid input - max_send_queue_size=%p",
			max_send_queue_size);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return priv_doca_comm_channel_get_max_queue_size(devinfo, max_send_queue_size);
}

doca_error_t
doca_comm_channel_get_max_recv_queue_size(struct doca_devinfo *devinfo, uint32_t *max_receive_queue_size)
{
	if (max_receive_queue_size == NULL) {
		DOCA_DLOG_ERR(
			"Unable to get max receive queue size: Received invalid input - max_receive_queue_size=%p",
			max_receive_queue_size);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return priv_doca_comm_channel_get_max_queue_size(devinfo, max_receive_queue_size);
}

doca_error_t
doca_comm_channel_get_service_max_num_connections(struct doca_devinfo *devinfo, uint32_t *max_num_connections)
{
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t dev_log_max_num_connections = 0;
	doca_error_t status;

	if (devinfo == NULL) {
		DOCA_DLOG_ERR(
			"Unable to get service max number of connections size: Received invalid input - devinfo=%p",
			devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (max_num_connections == NULL) {
		DOCA_DLOG_ERR(
			"Unable to get max number of connections: Received invalid input - max_num_connections=%p",
			max_num_connections);
		return DOCA_ERROR_INVALID_VALUE;
	}

	status = priv_doca_devinfo_hca_caps_by_opmod_query(devinfo,
						(MLX5_HCA_CAP_OPMOD_GET_CUR | MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE),
						out, sizeof(out));

	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to get max number of connections: Querying capabilities failed");
		return DOCA_ERROR_UNEXPECTED;
	}

	if (MLX5_GET(query_hca_cap_out, out, capability.cmd_hca_cap.hca_cap_2) == 1) {
		memset(out, 0, sizeof(out));

		status = priv_doca_devinfo_hca_caps_by_opmod_query(devinfo,
						(MLX5_HCA_CAP_OPMOD_GET_CUR | MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE_CAP_2),
						out, sizeof(out));

		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Unable to get max number of connections: Querying capabilities failed");
			return DOCA_ERROR_UNEXPECTED;
		}
		dev_log_max_num_connections = MLX5_GET(query_hca_cap_out, out,
								capability.cmd_hca_cap_2.log_max_channel_service_connection);
	} else {
		DOCA_DLOG_ERR("Unable to get max number of connections: HCA CAP 2 is not supported on this device");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	*max_num_connections = (dev_log_max_num_connections == 0 ? 0 : 1 << dev_log_max_num_connections);

	return DOCA_SUCCESS;
}

// Create local End Point handle
doca_error_t
doca_comm_channel_ep_create(struct doca_comm_channel_ep_t **ep)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("no valid ep pointer provided in ep_create\n");
		return DOCA_ERROR_INVALID_VALUE;
	}

	// open ibv_context, ibv_comp_channel (fd), ibv_pd,
	struct doca_comm_channel_ep_t *new_ep =
	    (struct doca_comm_channel_ep_t *)calloc(1, sizeof(struct doca_comm_channel_ep_t));
	if (new_ep == NULL) {
		DOCA_DLOG_ERR("endpoint mem allocation failed\n");
		return DOCA_ERROR_NO_MEMORY;
	}

	new_ep->qpp = NULL;
	new_ep->peer_addr_list = NULL;
	new_ep->flags = 0;
	new_ep->mtu = DEFAULT_MAX_MSG_SIZE;
	new_ep->common_rs = NULL;
	new_ep->snd_cq_obj = NULL;
	new_ep->rcv_cq_obj = NULL;
	new_ep->snd_event_ch = NULL;
	new_ep->rcv_event_ch = NULL;
	new_ep->peer_addr_list = NULL;
	new_ep->is_active = false;
	new_ep->connection_in_progress = false;
	new_ep->ep_properties.hw_dev = NULL;
	new_ep->ep_properties.dev_rep = NULL;
	new_ep->ep_properties.log_msg_size = log2_by_bits(DEFAULT_MAX_MSG_SIZE);
	new_ep->ep_properties.log_rcv_queue_size = log2_by_bits(DEFAULT_RCV_QUEUE_SIZE);
	new_ep->ep_properties.log_snd_queue_size = log2_by_bits(DEFAULT_SND_QUEUE_SIZE);
	new_ep->ep_properties.max_clients = 0;

	new_ep->destroyed_qps_list.destroyed_qp = NULL;

	*ep = new_ep;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_get_device(struct doca_comm_channel_ep_t *ep, struct doca_dev **device)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (device == NULL) {
		DOCA_DLOG_ERR("NULL output parameter received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*device = ep->ep_properties.hw_dev;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_get_device_rep(struct doca_comm_channel_ep_t *ep, struct doca_dev_rep **device_rep)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (device_rep == NULL) {
		DOCA_DLOG_ERR("NULL output parameter received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*device_rep = ep->ep_properties.dev_rep;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_get_max_msg_size(struct doca_comm_channel_ep_t *ep, uint16_t *max_ms_sz)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (max_ms_sz == NULL) {
		DOCA_DLOG_ERR("NULL output parameter received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*max_ms_sz = static_cast<uint16_t>(ep->mtu);

	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_get_send_queue_size(struct doca_comm_channel_ep_t *ep, uint16_t *snd_queue_size)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (snd_queue_size == NULL) {
		DOCA_DLOG_ERR("NULL output parameter received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*snd_queue_size = 1 << ep->ep_properties.log_snd_queue_size;

	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_get_recv_queue_size(struct doca_comm_channel_ep_t *ep, uint16_t *rcv_queue_size)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rcv_queue_size == NULL) {
		DOCA_DLOG_ERR("NULL output parameter received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*rcv_queue_size = 1 << ep->ep_properties.log_rcv_queue_size;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_set_device(struct doca_comm_channel_ep_t *ep, struct doca_dev *device)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (ep->is_active) {
		DOCA_DLOG_ERR("value can't be set when endpoint is active in %s", __func__);
		return DOCA_ERROR_BAD_STATE;
	}

	if (device == NULL) {
		DOCA_DLOG_ERR("invalid device received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	ep->ep_properties.hw_dev = device;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_set_device_rep(struct doca_comm_channel_ep_t *ep, struct doca_dev_rep *device_rep)
{
	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (ep->is_active) {
		DOCA_DLOG_ERR("value can't be set when endpoint is active in %s", __func__);
		return DOCA_ERROR_BAD_STATE;
	}

	if (device_rep == NULL) {
		DOCA_DLOG_ERR("invalid device received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	ep->ep_properties.dev_rep = device_rep;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_set_max_msg_size(struct doca_comm_channel_ep_t *ep, uint16_t max_ms_sz)
{
	uint32_t actual_msg_size;

	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (ep->is_active) {
		DOCA_DLOG_ERR("value can't be set when endpoint is active in %s", __func__);
		return DOCA_ERROR_BAD_STATE;
	}

	if (max_ms_sz > MAX_DATA_SIZE || max_ms_sz == 0) {
		DOCA_DLOG_ERR("Requested max messages is 0 or is larger than max messages %d\n", MAX_DATA_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (max_ms_sz < MIN_DATA_SIZE) {
		DOCA_DLOG_DBG("INFO: requested msg size smaller than minimal msg size, %d is used instead\n", MIN_DATA_SIZE);
		actual_msg_size = MIN_DATA_SIZE;
	} else
		actual_msg_size = max_ms_sz;


	ep->ep_properties.log_msg_size = log2_by_bits(actual_msg_size);
	ep->mtu = actual_msg_size;
	return DOCA_SUCCESS;
}


doca_error_t doca_comm_channel_ep_set_send_queue_size(struct doca_comm_channel_ep_t *ep, uint16_t snd_queue_size)
{
	uint32_t actual_queue_size;

	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (ep->is_active) {
		DOCA_DLOG_ERR("value can't be set when endpoint is active in %s", __func__);
		return DOCA_ERROR_BAD_STATE;
	}

	if (snd_queue_size > MAX_QUEUE_SIZE || snd_queue_size == 0) {
		DOCA_DLOG_ERR("Requested max messages is 0 or is larger than max messages %d\n", MAX_QUEUE_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (snd_queue_size < MIN_QUEUE_SIZE) {
		DOCA_DLOG_DBG("requested msg size smaller than minimal msg size, %d is used instead\n", MIN_QUEUE_SIZE);
		actual_queue_size = MIN_QUEUE_SIZE;
	} else
		actual_queue_size = snd_queue_size;


	ep->ep_properties.log_snd_queue_size = log2_by_bits(actual_queue_size);
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_set_recv_queue_size(struct doca_comm_channel_ep_t *ep, uint16_t rcv_queue_size)
{
	uint32_t actual_queue_size;

	if (ep == NULL) {
		DOCA_DLOG_ERR("invalid ep received in %s", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (ep->is_active) {
		DOCA_DLOG_ERR("%s is called for an active endpoint", __func__);
		return DOCA_ERROR_BAD_STATE;
	}

	if (rcv_queue_size > MAX_QUEUE_SIZE || rcv_queue_size == 0) {
		DOCA_DLOG_ERR("Requested max messages is 0 or is larger than max messages %d\n", MAX_QUEUE_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rcv_queue_size < MIN_QUEUE_SIZE) {
		DOCA_DLOG_DBG("requested msg size smaller than minimal msg size, %d is used instead\n", MIN_QUEUE_SIZE);
		actual_queue_size = MIN_QUEUE_SIZE;
	} else
		actual_queue_size = rcv_queue_size;


	ep->ep_properties.log_rcv_queue_size = log2_by_bits(actual_queue_size);
	return DOCA_SUCCESS;
}

doca_error_t init_event_channel(struct doca_comm_channel_ep_t *local_ep)
{
	doca_error_t err;

	local_ep->snd_event_ch = cc_qp_event_ch_create(local_ep->common_rs, 1);
	if (local_ep->snd_event_ch == NULL) {
		DOCA_DLOG_ERR("cc_qp_event_ch_create() failed for new endpoint\n");
		err = DOCA_ERROR_INITIALIZATION;
		goto error;
	}

	local_ep->rcv_event_ch = cc_qp_event_ch_create(local_ep->common_rs, 1);
	if (local_ep->rcv_event_ch == NULL) {
		DOCA_DLOG_ERR("cc_qp_event_ch_create() failed for new endpoint\n");
		err = DOCA_ERROR_INITIALIZATION;
		goto error_free_snd_event_ch;
	}

	if (cc_event_channel_subscribe(local_ep->snd_cq_obj, local_ep->snd_event_ch)) {
		DOCA_DLOG_ERR("channel event subscribe failed for snd_cq");
		err = DOCA_ERROR_INITIALIZATION;
		goto error_free_rcv_event_ch;
	}

	if (cc_event_channel_subscribe(local_ep->rcv_cq_obj, local_ep->rcv_event_ch)) {
		DOCA_DLOG_ERR("channel event subscribe failed for rcv_cq");
		err = DOCA_ERROR_INITIALIZATION;
		goto error_free_rcv_event_ch;
	}

	return DOCA_SUCCESS;


error_free_rcv_event_ch:
	priv_doca_dev_comp_events_channel_destroy(local_ep->rcv_event_ch);
error_free_snd_event_ch:
	priv_doca_dev_comp_events_channel_destroy(local_ep->snd_event_ch);
error:
	return err;
}

doca_error_t init_cqs(struct doca_comm_channel_ep_t *local_ep)
{
	doca_error_t err;
	int log_snd_cq_size;
	int log_rcv_cq_size;

	local_ep->num_ports = 1;

	log_snd_cq_size = log2_by_bits(static_cast<uint32_t>(local_ep->num_ports)) + local_ep->ep_properties.log_snd_queue_size + 1;
	log_rcv_cq_size = log2_by_bits(static_cast<uint32_t>(local_ep->num_ports)) + local_ep->ep_properties.log_rcv_queue_size + 1;

	local_ep->snd_cq_obj = cq_create(local_ep->common_rs, log_snd_cq_size);
	if (!local_ep->snd_cq_obj) {
		DOCA_DLOG_ERR("snd cq creation failed in create_ep\n");
		err = DOCA_ERROR_INITIALIZATION;
		goto error;
	}

	local_ep->rcv_cq_obj = cq_create(local_ep->common_rs, log_rcv_cq_size);
	if (!local_ep->rcv_cq_obj) {
		DOCA_DLOG_ERR("rcv cq creation failed in create_ep\n");
		err = DOCA_ERROR_INITIALIZATION;
		goto error_free_snd_cq;
	}

	return DOCA_SUCCESS;

error_free_snd_cq:
	cq_destroy(local_ep->snd_cq_obj);
	local_ep->snd_cq_obj = NULL;
error:
	return err;
}

void free_cqs(struct doca_comm_channel_ep_t *local_ep)
{
	if (local_ep->snd_cq_obj != NULL) {
		cq_destroy(local_ep->snd_cq_obj);
		local_ep->snd_cq_obj = NULL;
	}
	if (local_ep->rcv_cq_obj != NULL) {
		cq_destroy(local_ep->rcv_cq_obj);
		local_ep->rcv_cq_obj = NULL;
	}
}

void free_event_channel(struct doca_comm_channel_ep_t *local_ep)
{
	priv_doca_dev_comp_events_channel_destroy(local_ep->rcv_event_ch);
	local_ep->rcv_event_ch = NULL;
	priv_doca_dev_comp_events_channel_destroy(local_ep->snd_event_ch);
	local_ep->snd_event_ch = NULL;
}

// End Point notification file descriptor
// for blocking with epoll for recv ready event
doca_error_t doca_comm_channel_ep_get_event_channel(struct doca_comm_channel_ep_t *local_ep, doca_event_channel_t *snd_event_channel, doca_event_channel_t *rcv_event_channel)
{
	if (local_ep == NULL) {
		DOCA_DLOG_ERR("no local_ep received in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (local_ep->is_active) {
		doca_event_handle_t snd_event_handle;
		doca_event_handle_t rcv_event_handle;

		if (snd_event_channel == NULL && rcv_event_channel == NULL) {
			DOCA_DLOG_ERR("neither snd_event_channel or rcv_event_channel received in %s()", __func__);
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (snd_event_channel != NULL){
			auto snd_status = priv_doca_comp_events_channel_handle_get(local_ep->snd_event_ch, &snd_event_handle);
			if (snd_status != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("priv_doca_comp_events_channel_handle_get failed for send event channel in %s()", __func__);
				return DOCA_ERROR_NOT_FOUND;
			}
			*snd_event_channel = static_cast<doca_event_channel_t>(snd_event_handle);
		}

		if (rcv_event_channel != NULL) {
			auto rcv_status = priv_doca_comp_events_channel_handle_get(local_ep->rcv_event_ch, &rcv_event_handle);
			if (rcv_status != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("priv_doca_comp_events_channel_handle_get failed for receive event channel in %s()", __func__);
				return DOCA_ERROR_NOT_FOUND;
			}
			*rcv_event_channel = static_cast<doca_event_channel_t>(rcv_event_handle);
		}
		return DOCA_SUCCESS;
	} else {
		DOCA_DLOG_ERR("endpoint not yet initialized in %s()", __func__);
		return DOCA_ERROR_BAD_STATE;
	}
}

doca_error_t doca_comm_channel_ep_event_handle_arm_send(struct doca_comm_channel_ep_t *local_ep)
{
	if (local_ep == NULL) {
		DOCA_DLOG_ERR("local_ep is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	cc_force_arm(local_ep->snd_cq_obj);

	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_ep_event_handle_arm_recv(struct doca_comm_channel_ep_t *local_ep)
{
	if (local_ep == NULL) {
		DOCA_DLOG_ERR("local_ep is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	cc_force_arm(local_ep->rcv_cq_obj);

	return DOCA_SUCCESS;
}

// Service side listen on all interafces
// 'name' identifies the service, as a MAX_SERVICE_NAME_LEN bytes size null-terminated string.
doca_error_t
doca_comm_channel_ep_listen(struct doca_comm_channel_ep_t *local_ep, const char *name)
{
	uint8_t num_ports = 1;
	int i;
	doca_error_t ret;
	struct cc_qp **qpp_list;
	int cur_qpp, num_of_qpps = 0;
	size_t name_len;
	struct cc_qp_common *common_rs;

	if (local_ep == NULL) {
		DOCA_DLOG_ERR("no ep provided for ep_listen\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (name == NULL) {
		DOCA_DLOG_ERR("no name provided for ep_listen\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	name_len = strnlen(name, MAX_SERVICE_NAME_LEN);
	if (name_len == 0 || name_len == MAX_SERVICE_NAME_LEN) {
		DOCA_DLOG_ERR("name with illegal len provided in ep_listen\n");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (local_ep->is_active == true) {
		DOCA_DLOG_ERR("listen already called for local_ep\n");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	if (local_ep->ep_properties.hw_dev == NULL) {
		DOCA_DLOG_ERR("no doca_dev was set\n");
		return DOCA_ERROR_BAD_STATE;
	}

	if (local_ep->ep_properties.dev_rep == NULL) {
		DOCA_DLOG_ERR("no doca_dev_rep was set\n");
		return DOCA_ERROR_BAD_STATE;
	}

	common_rs = cc_qp_common_init(&local_ep->ep_properties, &local_ep->is_service);
	if (common_rs == NULL) {
		DOCA_DLOG_ERR("common resources initiation failed in %s()", __func__);
		return DOCA_ERROR_INITIALIZATION;
	}
	local_ep->common_rs = common_rs;

	if (local_ep->is_service == false) {
		DOCA_DLOG_ERR("listen function is allowed only for service side in Comm Channel\n");
		ret = DOCA_ERROR_NOT_PERMITTED;
		goto error_free_common_rs;
	}

	ret = init_cqs(local_ep);
	if (ret != DOCA_SUCCESS)
		goto error_free_common_rs;

	ret = init_event_channel(local_ep);
	if (ret != DOCA_SUCCESS)
		goto error_free_cqs;

	qpp_list = (struct cc_qp **)calloc(1, sizeof(struct cc_qp *) * num_ports);
	if (qpp_list == NULL) {
		DOCA_DLOG_ERR("temp qpp_list alloc failed\n");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error_free_event_channel;
	}

	qpp_list[num_of_qpps] = cc_qpp_create(local_ep->common_rs, &local_ep->ep_properties,
							local_ep->snd_cq_obj, local_ep->rcv_cq_obj);
	if (qpp_list[num_of_qpps] == NULL) {
			DOCA_DLOG_ERR("qpp number %d creation failed on IB link layer\n", num_of_qpps);
			ret = DOCA_ERROR_INITIALIZATION;
			goto error_free_qpp_list;
		}
	vhca_id_set(priv_doca_dev_vhca_id_get(local_ep->ep_properties.dev_rep), qpp_list[num_of_qpps]);
	if (cc_qp_register_server(qpp_list[num_of_qpps], local_ep->common_rs, name,
				  local_ep->ep_properties.max_clients)) {
			DOCA_DLOG_ERR("server registration failed on IB link layer\n");
			ret = DOCA_ERROR_CONNECTION_ABORTED;
			goto error_free_qp_register;
		}
	num_of_qpps++;

	/* create the accurate number of QPPs and peer_addresses needed for the endpont */
	local_ep->num_ports = num_of_qpps;
	local_ep->qpp = (struct cc_qp **)calloc(1, sizeof(struct cc_qp *) * num_of_qpps);
	if (local_ep->qpp == NULL) {
		DOCA_DLOG_ERR("qpp array allocation failed\n");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error_free_qp_register;
	}
	local_ep->peer_addr_list = (struct peer_addr_list *)calloc(
						1, sizeof(struct peer_addr_list) * num_of_qpps);
	if (local_ep->peer_addr_list == NULL) {
		DOCA_DLOG_ERR("peer_addr_list allocation failed\n");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error_free_qpp;
	}
	for (i = 0 ; i < num_of_qpps ; i++) {
		local_ep->qpp[i] = qpp_list[i];
		local_ep->peer_addr_list[i].len = 0;
		local_ep->peer_addr_list[i].peer_addr = NULL;
	}
	free(qpp_list);
	local_ep->is_active = true;
	return DOCA_SUCCESS;

error_free_qpp:
	free(local_ep->qpp);
error_free_qp_register:
	for (cur_qpp = num_of_qpps - 1; cur_qpp >= 0; cur_qpp--)
		cc_qpp_destroy(qpp_list[cur_qpp], &(local_ep->destroyed_qps_list));
error_free_qpp_list:
	free(qpp_list);
error_free_event_channel:
	free_event_channel(local_ep);
error_free_cqs:
	free_cqs(local_ep);
error_free_common_rs:
	cc_qp_common_destroy(local_ep->common_rs);
	local_ep->common_rs = NULL;
	return ret;
}

/** cc_preform_handshake is called only by the client's side */
static doca_error_t
cc_preform_handshake(struct doca_comm_channel_ep_t *local_ep, uint32_t *common_version)
{
	union imm_msg hello_header = {0};
	struct cc_qp *qpp;
	int rc;

	hello_header.handshake.type = DOCA_CC_HANDSHAKE;
	hello_header.handshake.mtu = local_ep->mtu;
	hello_header.handshake.version = CURRENT_COMM_CHANNEL_VERSION;
	qpp = local_ep->qpp[0];

	rc = cc_qp_msg_send_handshake(qpp, local_ep->snd_cq_obj, NULL, &hello_header);
	if (unlikely(rc)) {
		DOCA_DLOG_ERR("Error sending handshake message");
		return DOCA_ERROR_UNKNOWN;
	}

	local_ep->connection_in_progress = true;
	return DOCA_SUCCESS;
}

// Client side Connect
// 'name' identifies the service to connect to, as a MAX_SERVICE_NAME_LEN bytes size null-terminated
// string. Errors: DOCA_CC_EFAULT, ECONNREFUSED,
doca_error_t
doca_comm_channel_ep_connect(struct doca_comm_channel_ep_t *local_ep, const char *name,
			     struct doca_comm_channel_addr_t **peer_addr)
{
	doca_error_t ret, handshake;
	struct doca_comm_channel_addr_t *new_peer_addr;
	size_t name_len;
	struct cc_qp_common *common_rs;
	uint32_t common_version = CURRENT_COMM_CHANNEL_VERSION;

	if (local_ep == NULL) {
		DOCA_DLOG_ERR("local_ep is null in ep_connect\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (name == NULL) {
		DOCA_DLOG_ERR("no name provided in ep_connect\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	name_len = strnlen(name, MAX_SERVICE_NAME_LEN);
	if (name_len == 0 || name_len == MAX_SERVICE_NAME_LEN) {
		DOCA_DLOG_ERR("name with illegal len provided in ep_connect\n");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("peer_addr is null in ep_connect\n");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (local_ep->is_active) {
		DOCA_DLOG_ERR("Client already connected to service\n");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	if (local_ep->ep_properties.hw_dev == NULL) {
		DOCA_DLOG_ERR("no doca_dev was set\n");
		return DOCA_ERROR_BAD_STATE;
	}

	common_rs = cc_qp_common_init(&local_ep->ep_properties, &local_ep->is_service);
	if (common_rs == NULL) {
		DOCA_DLOG_ERR("common resources initiation failed in %s()", __func__);
		return DOCA_ERROR_INITIALIZATION;
	}
	local_ep->common_rs = common_rs;
	if (local_ep->is_service) {
		DOCA_DLOG_ERR("connect function is allowed only for client side in Comm Channel\n");
		ret = DOCA_ERROR_NOT_PERMITTED;
		goto error_free_common_rs;
	}

	ret = init_cqs(local_ep);
	if (ret != DOCA_SUCCESS)
		goto error_free_common_rs;

	ret = init_event_channel(local_ep);
	if (ret != DOCA_SUCCESS)
		goto error_free_cqs;

	local_ep->num_ports = 1;

	local_ep->peer_addr_list = (struct peer_addr_list *)calloc(
								1, sizeof(struct peer_addr_list));
	if (local_ep->peer_addr_list == NULL) {
		DOCA_DLOG_ERR("peer_addr_list allocation failed\n");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error_free_event_channel;
	}

	new_peer_addr =
	    (struct doca_comm_channel_addr_t *)calloc(1, sizeof(struct doca_comm_channel_addr_t));
	if (new_peer_addr == NULL) {
		DOCA_DLOG_ERR("peer_addr malloc failed\n");
		ret = DOCA_ERROR_NO_MEMORY;
		goto peer_addr_err;
	}
	new_peer_addr->ep = local_ep;
	new_peer_addr->next = NULL;
	new_peer_addr->user_data = 0;
	new_peer_addr->error = false;
	new_peer_addr->user_notified = false;
	memset(&new_peer_addr->current_stat, 0, sizeof(new_peer_addr->current_stat));
	memset(&new_peer_addr->snapshot_stat, 0, sizeof(new_peer_addr->snapshot_stat));

	local_ep->peer_addr_list->peer_addr = new_peer_addr;
	local_ep->peer_addr_list->len = 1;

	local_ep->qpp = (struct cc_qp **)calloc(1, sizeof(struct cc_qp *));
	if (local_ep->qpp == NULL) {
		DOCA_DLOG_ERR("qpp alloc failed\n");
		ret = DOCA_ERROR_NO_MEMORY;
		goto qpp_array_err;
	}

	*local_ep->qpp = cc_qpp_create(local_ep->common_rs, &local_ep->ep_properties,
				       local_ep->snd_cq_obj, local_ep->rcv_cq_obj);
	if (*local_ep->qpp == NULL) {
		DOCA_DLOG_ERR("qpp create failed for %s()\n", __func__);
		ret = DOCA_ERROR_INITIALIZATION;
		goto qpp_create_err;
	}
	if (cc_qp_register_client(*local_ep->qpp, local_ep->common_rs, name)) {
		DOCA_DLOG_ERR("client registration failed for %s()\n", __func__);
		ret = DOCA_ERROR_CONNECTION_ABORTED;
		goto registration_err;
	}

	handshake = cc_preform_handshake(local_ep, &common_version);
	if (handshake != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Connection failed due to failed handshake.\n");
		ret = handshake;
		goto handshake_err;
	}
	/*sent hello message*/
	new_peer_addr->current_stat.tx_in_flight_messages++;
	new_peer_addr->version = common_version;
	local_ep->is_active = true;
	*peer_addr = new_peer_addr;
	return DOCA_SUCCESS;
handshake_err:
	cc_channel_destroy(local_ep->qpp[0]);
registration_err:
	cc_qpp_destroy(*local_ep->qpp, &(local_ep->destroyed_qps_list));
	*local_ep->qpp = NULL;
qpp_create_err:
	free(local_ep->qpp);
	local_ep->qpp = NULL;
qpp_array_err:
	free(new_peer_addr);
peer_addr_err:
	free(local_ep->peer_addr_list);
	local_ep->peer_addr_list = NULL;
error_free_event_channel:
	free_event_channel(local_ep);
error_free_cqs:
	free_cqs(local_ep);
error_free_common_rs:
	cc_qp_common_destroy(local_ep->common_rs);
	local_ep->common_rs = NULL;
	return ret;
}

// Disconnect the End Point from the remote peer
// block until all resources related to peer address are freed
// new connection could be created on the End Point
doca_error_t
doca_comm_channel_ep_disconnect(struct doca_comm_channel_ep_t *local_ep,
				struct doca_comm_channel_addr_t *peer_addr)
{
	if (local_ep == NULL) {
		DOCA_DLOG_ERR("local_ep is null in ep_disconnect\n");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (local_ep->qpp == NULL) {
		DOCA_DLOG_ERR("no qpp available on the endpoint in ep_disconnect\n");
		return DOCA_ERROR_NOT_CONNECTED;
	}
	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("no peer address to disconnect from in ep_disconnect\n");
		return DOCA_ERROR_NOT_CONNECTED;
	}

	if (local_ep->is_service) {
                struct peer_addr_list *addr_list;
                struct doca_comm_channel_addr_t *ep_peer_addr_cur = NULL;
                struct doca_comm_channel_addr_t *ep_peer_addr_prev = NULL;
		int i = 0;

		i = get_peers_vhca_id_index(local_ep, peer_addr->vhca_id);
		if (i < 0) {
			DOCA_DLOG_ERR("qpp for peer_addr with vhca_id = %d could not be found\n",
			       peer_addr->vhca_id);
			return DOCA_ERROR_NOT_CONNECTED;
		}

		addr_list = local_ep->peer_addr_list + i;
		ep_peer_addr_cur = addr_list->peer_addr;

		for (i = 0 ; i < addr_list->len ; i++) {
			if (ep_peer_addr_cur == NULL)
				break;

			if (ep_peer_addr_cur->av.dct == peer_addr->av.dct) {
				peer_addr = ep_peer_addr_cur;
				break;
			}
			ep_peer_addr_prev = ep_peer_addr_cur;
			ep_peer_addr_cur = ep_peer_addr_prev->next;
		}
		if (ep_peer_addr_cur == NULL) {
			DOCA_DLOG_DBG("INFO: the peer_addr couldn't be found in this endpoint\n");
			return DOCA_ERROR_NOT_CONNECTED;
		}
		if (ep_peer_addr_prev != NULL)
			ep_peer_addr_prev->next = peer_addr->next;
		else
			addr_list->peer_addr = peer_addr->next;
		addr_list->len--;
	} else {
		/* client has only one peer */
		cc_channel_destroy(local_ep->qpp[0]);
		cc_qpp_destroy(local_ep->qpp[0], &(local_ep->destroyed_qps_list));
		local_ep->qpp[0] = NULL;
		free(local_ep->qpp);
		local_ep->qpp = NULL;
		free_event_channel(local_ep);
		free_cqs(local_ep);
		cc_qp_common_destroy(local_ep->common_rs);
		local_ep->common_rs = NULL;
		free(local_ep->peer_addr_list);
		local_ep->peer_addr_list = NULL;
		local_ep->is_active = false;
	}
	free(peer_addr);

	return DOCA_SUCCESS;
}

static void
clear_peer_addr_list(struct doca_comm_channel_addr_t *peer_addr)
{
	struct doca_comm_channel_addr_t *peer_addr_last;

	while (peer_addr != NULL) {
		peer_addr_last = peer_addr;
		peer_addr = peer_addr_last->next;
		free(peer_addr_last);
	}
}

// Release End Point handles (remote and local)
// block until all queued messages are sent
// close fd, and release all internal resources
doca_error_t
doca_comm_channel_ep_destroy(struct doca_comm_channel_ep_t *ep)
{
	int i;

	if (ep == NULL)
		return DOCA_ERROR_NOT_CONNECTED;
	if (ep->qpp == NULL)
		goto finish;

	for (i = 0; i < ep->num_ports ; i++) {
		if (ep->qpp[i] == NULL)
			continue;

		if (ep->peer_addr_list[i].len > 0)
			clear_peer_addr_list(ep->peer_addr_list[i].peer_addr);
		cc_qpp_destroy(ep->qpp[i], &(ep->destroyed_qps_list));
	}

finish:
	if (ep->qpp != NULL)
		free(ep->qpp);
	if (ep->peer_addr_list != NULL)
		free(ep->peer_addr_list);
	free_event_channel(ep);
	free_cqs(ep);
	if (ep->common_rs != NULL)
		cc_qp_common_destroy(ep->common_rs);
	destroy_destroyed_qps_list(&(ep->destroyed_qps_list));

	free(ep);
	return DOCA_SUCCESS;
}

static struct doca_comm_channel_addr_t *
get_peer_addr(struct cc_av *av, struct doca_comm_channel_addr_t *peer_addr)
{
	while (peer_addr != NULL) {
		if (av->dct == peer_addr->av.dct)
			return peer_addr;
		peer_addr = peer_addr->next;
	}
	return NULL;
}


static doca_error_t
empty_send_cq(struct doca_comm_channel_ep_t *local_ep)
{
	int i = 0;
	int ret;
	doca_error_t err_code;
	struct cc_av av = {};
	struct doca_comm_channel_addr_t *peer_addr;
	uint8_t *temp_repost_buf = NULL;
	uint32_t err_type;

	while (1) {
		for (i = 0; i < local_ep->num_ports; i++) {
			ret = cc_qp_check_send_cqe(local_ep->qpp[i], local_ep->snd_cq_obj, &(local_ep->destroyed_qps_list), &av);
			switch (ret) {
			case SEND_MSG:
				if (local_ep->is_service) {
					peer_addr = get_peer_addr(&av, local_ep->peer_addr_list[i].peer_addr);
					if (unlikely(peer_addr == NULL)) {
						DOCA_DLOG_ERR("got send cqe for unknown peer_address\n");
						break;
					}
				} else
					peer_addr = local_ep->peer_addr_list[i].peer_addr;
				peer_addr->current_stat.tx_in_flight_messages--;
				break;
			case WRONG_QPN:
				break;
			case NO_MSG_FOUND:
				if (temp_repost_buf != NULL) {
					repost_flushed_wqes(local_ep->qpp[i], temp_repost_buf);
					DOCA_DLOG_DBG("re-post all flushed wqes\n");
				}
				err_code = DOCA_SUCCESS;
				goto exit;
			case ERR_SEND_MSG:
				err_type = get_cqe_error_syndrom(local_ep->qpp[i]);
				DOCA_DLOG_ERR("A message failed to send for qpp on vhca_id %d with error enum 0x%x",
						vhca_id_get(local_ep->qpp[i]), err_type);
				if (err_type == DOCA_CC_PEER_ERR) {
					temp_repost_buf = (uint8_t *)calloc((1 << (local_ep->ep_properties.log_snd_queue_size + 1))
									    * MLX5_SEND_WQE_BB, sizeof(uint8_t));
					if (temp_repost_buf == NULL) {
						DOCA_DLOG_ERR("memory allocation for temporary buffer for reposting flushed wqes failed");
						err_code = DOCA_ERROR_NO_MEMORY;
						goto exit;
					}

					av.dct = recover_dci(local_ep->qpp[i]);
					if (av.dct < 0) {
						DOCA_DLOG_ERR("DCI recovery failed in %s()", __func__);
						err_code = DOCA_ERROR_INITIALIZATION;
						goto exit;
					}
					if (local_ep->is_service) {
						peer_addr = get_peer_addr(&av, local_ep->peer_addr_list[i].peer_addr);
						if (unlikely(peer_addr == NULL)) {
							DOCA_DLOG_ERR("got send error cqe for unknown peer_address\n");
							break;
						}
					} else
						peer_addr = local_ep->peer_addr_list[i].peer_addr;
					peer_addr->error = true;
				} else if (err_type == DOCA_CC_FLUSH_ERR) {
					if (av.dct == 0) {
						DOCA_DLOG_ERR("can't get flush_error when not come after real error");
						err_code = DOCA_ERROR_UNKNOWN;
						goto exit;
					}
					add_wqe_to_repost_buf(local_ep->qpp[i], temp_repost_buf, av.dct);
				}
				break;
			default:
				DOCA_DLOG_ERR("return code %d is unexpected when checking for empty send queue on vhca_id %d\n",
						ret, vhca_id_get(local_ep->qpp[i]));
				err_code = DOCA_ERROR_UNKNOWN;
				goto exit;
			}
		}
	}

exit:
	if (temp_repost_buf != NULL)
		free(temp_repost_buf);
	return err_code;
}

// Data Path send
// @msg, len: pointer to the message to be sent, of length bytes
// @flags: MSG_DONTWAIT, MSG_MORE
// @peer_addr: destination address handle of the send operation
// Errors: EAGAIN, ENOTCONN, DOCA_CC_EFAULT
doca_error_t
doca_comm_channel_ep_sendto(struct doca_comm_channel_ep_t *local_ep, const void *msg, size_t len,
			    int flags, struct doca_comm_channel_addr_t *peer_addr)
{
	struct cc_qp *qpp;
	struct cc_av *av = NULL;
	int i = 0, ret;
	doca_error_t empty_cq_ret;
	int max_msg_size;

	if (unlikely(local_ep == NULL)) {
		DOCA_DLOG_ERR("local_ep is null in ep_sendto\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(local_ep->is_active == false)) {
		DOCA_DLOG_ERR("local_ep is not connected in %s\n", __func__);
		return DOCA_ERROR_NOT_CONNECTED;
	}
	if (unlikely(msg == NULL)) {
		DOCA_DLOG_ERR("msg is null in ep_sendto\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(len == 0)) {
		DOCA_DLOG_ERR("len is 0 in ep_sendto\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(flags != DOCA_CC_MSG_FLAG_NONE)) {
		DOCA_DLOG_ERR("non-supported flag received in %s\n", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(peer_addr == NULL)) {
		DOCA_DLOG_ERR("must insert valid peer_addr in ep_sendto\n");
		return DOCA_ERROR_NOT_CONNECTED;
	}
	if (unlikely(local_ep->connection_in_progress == true)) {
		DOCA_DLOG_ERR("peer address not yet connected in %s\n", __func__);
		return DOCA_ERROR_NOT_CONNECTED;
	}

	max_msg_size = 1 << local_ep->ep_properties.log_msg_size;
	if (unlikely(len > max_msg_size)) {
		DOCA_DLOG_ERR("Msg size larger than maximal msg size which is %d.\n", max_msg_size);
		return DOCA_ERROR_INVALID_VALUE;
	}

	empty_cq_ret = empty_send_cq(local_ep);
	if (unlikely(empty_cq_ret != DOCA_SUCCESS)) {
		DOCA_DLOG_ERR("empty send queue failed.");
		return empty_cq_ret;
	}

	if (local_ep->is_service) {
		i = get_peers_vhca_id_index(local_ep, peer_addr->vhca_id);
		if (unlikely(i < 0)) {
			DOCA_DLOG_ERR("qpp for vhca_id %d could not be found\n", peer_addr->vhca_id);
			return DOCA_ERROR_NOT_CONNECTED;
		}
		av = &(peer_addr->av);
	}
	qpp = local_ep->qpp[i];

	if (unlikely(peer_addr->error == true)) {
		DOCA_DLOG_ERR("the provided peer_addr is in error state and it needs to be disconnected\n");
		peer_addr->user_notified = true;
		return DOCA_ERROR_CONNECTION_RESET;
	}
	ret = cc_qp_msg_send(qpp, local_ep->snd_cq_obj, msg, len, av, NULL);
	if (ret == -QUEUE_FULL) {
		DOCA_DLOG_DBG("msg_send failed due to full send queue\n");
		return DOCA_ERROR_AGAIN;
	}

	if (unlikely(ret < 0)) {
		DOCA_DLOG_ERR("msg_send failed on vhca_id %d due to transmit error", vhca_id_get(qpp));
		return DOCA_ERROR_UNKNOWN;
	}

	peer_addr->current_stat.tx_messages++;
	peer_addr->current_stat.tx_bytes += len;
	peer_addr->current_stat.tx_in_flight_messages++;

	return DOCA_SUCCESS;
}


doca_error_t
doca_comm_channel_ep_recvfrom(struct doca_comm_channel_ep_t *local_ep, void *msg, size_t *len,
			      int flags, struct doca_comm_channel_addr_t **peer_addr)
{
	int i, ret, send_ret;
	struct doca_comm_channel_addr_t *new_peer_addr;
	union imm_msg recv_header = {0}, send_header = {0};
	struct cc_av *av = NULL;
	struct cc_av fail_av;
	uint32_t handshake_dct;
	doca_error_t empty_cq_ret;

	if (unlikely(local_ep == NULL)) {
		DOCA_DLOG_ERR("local_ep is null in ep_recvfrom\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(msg == NULL)) {
		DOCA_DLOG_ERR("msg is null in ep_recvfrom\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(len == NULL || *len == 0)) {
		DOCA_DLOG_ERR("len is null or 0 in ep_recvfrom\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(peer_addr == NULL)) {
		DOCA_DLOG_ERR("peer_addr is null in ep_recvfrom\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(flags != DOCA_CC_MSG_FLAG_NONE)) {
		DOCA_DLOG_ERR("non-supported flag received in %s\n", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(local_ep->is_active == false)) {
		DOCA_DLOG_ERR("ep_recvfrom called before listen/connect\n");
		return DOCA_ERROR_NOT_CONNECTED;
	}

	if (unlikely(*len > MAX_DATA_SIZE)) {
		*len = MAX_DATA_SIZE;
		DOCA_DLOG_DBG("Warning: msg buffer supplied is bigger then max received msg size\n"
			"INFO: max message size is %d\n", MAX_DATA_SIZE);
	}


start_receive:
	if (local_ep->is_service) {
		for (i = 0; i < local_ep->num_ports; i++) {
			ret = cc_qp_msg_recv(local_ep->qpp[i], local_ep->rcv_cq_obj, msg, len,
					     &(local_ep->destroyed_qps_list), &recv_header);
			switch (ret) {
			case HELLO_MSG:
				send_header.handshake.type = DOCA_CC_HANDSHAKE;
				send_header.handshake.mtu = local_ep->mtu;
				send_header.handshake.version = CURRENT_COMM_CHANNEL_VERSION;

				handshake_dct = av_get(local_ep->qpp[i])->dct;
				*peer_addr = get_peer_addr(av_get(local_ep->qpp[i]),
							   local_ep->peer_addr_list[i].peer_addr);
				/** empty CQ and send queue*/
				empty_cq_ret = empty_send_cq(local_ep);
				if (unlikely(empty_cq_ret != DOCA_SUCCESS)) {
					DOCA_DLOG_ERR("got unexpected message in send CQ: %s\n", doca_get_error_name(empty_cq_ret));
					return DOCA_ERROR_UNKNOWN;
				}

				if (recv_header.handshake.mtu == local_ep->mtu && *peer_addr == NULL) {
					new_peer_addr = (struct doca_comm_channel_addr_t *)calloc(
						1, sizeof(struct doca_comm_channel_addr_t));
					if (new_peer_addr == NULL) {
						DOCA_DLOG_ERR("peer_addr malloc failed for new peer from handshake");
						return DOCA_ERROR_NO_MEMORY;
					}
					new_peer_addr->ep = local_ep;
					new_peer_addr->vhca_id = vhca_id_get(local_ep->qpp[i]);

					if (send_header.handshake.version < recv_header.handshake.version) {
						DOCA_DLOG_INFO( "Current comm channel version is %d, peer has version %d of comm channel",
								send_header.handshake.version,
								recv_header.handshake.version);
						DOCA_DLOG_INFO("A newer version for comm channel is available. Updating to the newest version is recommended");
						new_peer_addr->version = send_header.handshake.version;
					} else
						new_peer_addr->version = recv_header.handshake.version;

					new_peer_addr->av.dct = handshake_dct;
					new_peer_addr->next = local_ep->peer_addr_list[i].peer_addr;
					new_peer_addr->user_data = 0;
					new_peer_addr->error = false;
					new_peer_addr->user_notified = false;
					memset(&new_peer_addr->current_stat, 0, sizeof(new_peer_addr->current_stat));
					av = &(new_peer_addr->av);
					send_ret = cc_qp_msg_send_handshake(local_ep->qpp[i], local_ep->snd_cq_obj, av, &send_header);
					if (unlikely(send_ret)) {
						DOCA_DLOG_ERR("Error sending handshake message in vhca_id %d",
								vhca_id_get(local_ep->qpp[i]));
						free(new_peer_addr);
						return DOCA_ERROR_UNKNOWN;
					}
					new_peer_addr->current_stat.tx_in_flight_messages++;
					local_ep->peer_addr_list[i].peer_addr = new_peer_addr;
					local_ep->peer_addr_list[i].len++;
				} else {
					if (unlikely(*peer_addr != NULL)) {
						DOCA_DLOG_ERR("received HELLO_MSG from existing peer_addr");
					} else {
						DOCA_DLOG_DBG("received HELLO_MSG with mtu %d, while local ep mtu is %zu, connection dropped\n",
							recv_header.handshake.mtu, local_ep->mtu);
					}
					send_header.handshake.type = DOCA_CC_REJECT;
					fail_av.dct = av_get(local_ep->qpp[i])->dct;
					send_ret = cc_qp_msg_send(local_ep->qpp[i], local_ep->snd_cq_obj, NULL, 0, &fail_av, &send_header);
					if (unlikely(send_ret)) {
						DOCA_DLOG_ERR("Error sending handshake message in vhca_id %d",
								vhca_id_get(local_ep->qpp[i]));
						return DOCA_ERROR_UNKNOWN;
					}
				}
				goto start_receive;
			case ARRIVE_MSG:
				*peer_addr = get_peer_addr(av_get(local_ep->qpp[i]),
							local_ep->peer_addr_list[i].peer_addr);
				if (unlikely(*peer_addr == NULL)) {
					DOCA_DLOG_ERR("Got message from an unknown peer address in vhca_id %d",
							vhca_id_get(local_ep->qpp[i]));
					return DOCA_ERROR_UNKNOWN;
				}
				(*(peer_addr))->current_stat.rx_messages++;
				(*(peer_addr))->current_stat.rx_bytes += *len;
				if (unlikely((*peer_addr)->error == true)) {
					if ((*peer_addr)->user_notified)
						goto start_receive;
					DOCA_DLOG_ERR("got message from peer_addr that has error and is need to be disconnected\n");
					(*peer_addr)->user_notified = true;
					return DOCA_ERROR_CONNECTION_RESET;
				}
				return DOCA_SUCCESS;
			case NO_MSG_FOUND:
				goto no_msg;
			case WRONG_QPN:
				break;
			default:
				DOCA_DLOG_ERR("ret value %d in vhca_id %d from msg_recv is illegal\n",
						ret, vhca_id_get(local_ep->qpp[i]));
				return DOCA_ERROR_UNKNOWN;
			}
		}
	} else {
		if (unlikely(local_ep->peer_addr_list[0].peer_addr == NULL)) {
			DOCA_DLOG_ERR("can't use %s() on client side without an active peer_addr", __func__);
			return DOCA_ERROR_UNKNOWN;
		}
		ret = cc_qp_msg_recv(local_ep->qpp[0], local_ep->rcv_cq_obj, msg, len,
				     &(local_ep->destroyed_qps_list), NULL);
		switch (ret) {
		case ARRIVE_MSG:
			*peer_addr = local_ep->peer_addr_list[0].peer_addr;
			if (unlikely((*peer_addr)->error == true)) {
				if ((*peer_addr)->user_notified)
					goto start_receive;
				DOCA_DLOG_ERR("got message from a peer_addr which is in error state and it needs to be disconnected\n");
				(*peer_addr)->user_notified = true;
				return DOCA_ERROR_CONNECTION_RESET;
			}
			return DOCA_SUCCESS;
		case NO_MSG_FOUND:
			goto no_msg;
		case WRONG_QPN:
			break;
		default:
			DOCA_DLOG_ERR("ret value %d from msg_recv is illegal on client side\n", ret);
			return DOCA_ERROR_UNKNOWN;
		}
	}
no_msg:
	return DOCA_ERROR_AGAIN;
}

// Extract 'user_data' from peer_addr handle
doca_error_t
doca_comm_channel_peer_addr_get_user_data(struct doca_comm_channel_addr_t *peer_addr, uint64_t *user_data)
{
	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("no peer_addr supplied for %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (user_data == NULL) {
		DOCA_DLOG_ERR("user_data provided is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}
	*user_data = peer_addr->user_data;
	return DOCA_SUCCESS;
}

// Save 'user_data' in peer_addr handle
doca_error_t
doca_comm_channel_peer_addr_set_user_data(struct doca_comm_channel_addr_t *peer_addr,
					  uint64_t user_data)
{
	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("no peer_addr supplied for %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}
	peer_addr->user_data = user_data;
	return DOCA_SUCCESS;
}

doca_error_t
static inline check_handshake_response(struct doca_comm_channel_ep_t *local_ep,
				       struct doca_comm_channel_addr_t *peer_addr, struct cc_qp *qpp)
{
	union imm_msg recv_header = {0};
	size_t rc;
	size_t msg_size = DEFAULT_MAX_MSG_SIZE; /*we don't expect any data, but must provide a buffer to receive */
	uint8_t buff[DEFAULT_MAX_MSG_SIZE];

	rc = cc_qp_msg_recv(qpp, local_ep->rcv_cq_obj, buff, &msg_size,
						&(local_ep->destroyed_qps_list), &recv_header);

	if (likely(rc == HELLO_MSG)) {
		if (CURRENT_COMM_CHANNEL_VERSION < recv_header.handshake.version) {
			DOCA_DLOG_INFO("Current comm channel version is %d, peer has version %d of comm channel",
				       CURRENT_COMM_CHANNEL_VERSION, recv_header.handshake.version);
			DOCA_DLOG_INFO(
			"A newer version for comm channel is available. Updating to the newest version is recommended");
			peer_addr->version = CURRENT_COMM_CHANNEL_VERSION;
		} else
			peer_addr->version = recv_header.handshake.version;

		local_ep->connection_in_progress = false;
		return DOCA_SUCCESS;
	} else if (rc == NO_MSG_FOUND) {
		return DOCA_ERROR_CONNECTION_INPROGRESS;
	} else {
		peer_addr->error = true;
		switch (rc) {
		case REJECT_MSG:
			if (recv_header.handshake.mtu != local_ep->mtu)
				DOCA_DLOG_ERR("Connection failed due to MTU mismatch");
			else
				DOCA_DLOG_ERR("Connection failed due to connection already exists\n");

			return DOCA_ERROR_CONNECTION_ABORTED;
		case WRONG_QPN:
			DOCA_DLOG_ERR("Received hello message to an unknown qpn.\n");
			return DOCA_ERROR_CONNECTION_ABORTED;
		default:
			DOCA_DLOG_ERR("Received an unexpected message while waiting for hello message.\n");
			return DOCA_ERROR_CONNECTION_ABORTED;
		}
	}
}

doca_error_t doca_comm_channel_peer_addr_update_info(struct doca_comm_channel_addr_t *peer_addr)
{
	struct doca_comm_channel_ep_t *local_ep;
	doca_error_t ret;
	struct cc_qp *qpp;

	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("no peer_addr supplied for %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	local_ep = peer_addr->ep;

	/* if we already saw the connection failed or the peer address is in error */
	if (peer_addr->error == true) {
		DOCA_DLOG_ERR("peer address is in error state");
		return DOCA_ERROR_CONNECTION_ABORTED;
	}

	/*if connection was not yet established, we need to check the recv queue */
	if (local_ep->connection_in_progress == true) {
		qpp = local_ep->qpp[0];
		ret = check_handshake_response(local_ep, peer_addr, qpp);
		if (ret != DOCA_SUCCESS) {
			if (ret == DOCA_ERROR_CONNECTION_INPROGRESS)
				DOCA_DLOG_DBG("no handshake response received");
			else
				DOCA_DLOG_ERR("Handshake failed");
			return ret;
		}
	}

	/* empty send cq to update the in flight messages*/
	ret = empty_send_cq(local_ep);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("failed to empty send cq in %s", __func__);
		return ret;
	}

	memcpy(&peer_addr->snapshot_stat, &peer_addr->current_stat, sizeof(peer_addr->current_stat));

	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_peer_addr_get_send_messages(const struct doca_comm_channel_addr_t *peer_addr, uint64_t *tx_messages)
{
	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("peer_addr is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (tx_messages == NULL) {
		DOCA_DLOG_ERR("tx_messages is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*tx_messages = peer_addr->snapshot_stat.tx_messages;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_peer_addr_get_send_bytes(const struct doca_comm_channel_addr_t *peer_addr, uint64_t *tx_bytes)
{
	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("peer_addr is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (tx_bytes == NULL) {
		DOCA_DLOG_ERR("tx_bytes is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*tx_bytes = peer_addr->snapshot_stat.tx_bytes;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_peer_addr_get_recv_messages(const struct doca_comm_channel_addr_t *peer_addr, uint64_t *rx_messages)
{
	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("peer_addr is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rx_messages == NULL) {
		DOCA_DLOG_ERR("rx_messages is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*rx_messages = peer_addr->snapshot_stat.rx_messages;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_peer_addr_get_recv_bytes(const struct doca_comm_channel_addr_t *peer_addr, uint64_t *rx_bytes)
{
	if (rx_bytes == NULL) {
		DOCA_DLOG_ERR("peer_addr is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rx_bytes == NULL) {
		DOCA_DLOG_ERR("tx_messages is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*rx_bytes = peer_addr->snapshot_stat.rx_bytes;
	return DOCA_SUCCESS;
}

doca_error_t doca_comm_channel_peer_addr_get_send_in_flight_messages(const struct doca_comm_channel_addr_t *peer_addr, uint64_t *tx_in_flight_messages)
{
	if (peer_addr == NULL) {
		DOCA_DLOG_ERR("peer_addr is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (tx_in_flight_messages == NULL) {
		DOCA_DLOG_ERR("tx_in_flight_messages is null in %s()", __func__);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*tx_in_flight_messages = peer_addr->snapshot_stat.tx_in_flight_messages;
	return DOCA_SUCCESS;
}
