/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <errno.h>
#include <doca_log.h>
#include <private/common_defines.h>
#include <doca_dev.h>
#include <priv_doca_dev.h>
#include <priv_doca_os_abstraction.hpp>
#include <priv_doca_os_abstraction_endian.h>
#include <doca_types.h>
#include <cc_devx.h>
#include <qp_channel.h>
#include <doca_comm_channel.h>

DOCA_LOG_REGISTER(COMM_CH::QP);


#define DEST_QP_SIZE (0xffffff)

#define CONNECT_BY_SERVICE_NAME 0
#define CONNECT_BY_VHCA_ID 1
#define CREDIT_LOG_QUEUE_SIZE 13
#define MIN_BUFFER_SIZE 0x1000
#define SERVICE_NAME_FULL 128
#define EVENT_SZ 1
#define UD_ADDRESS_VECTOR_SHORT 16
#define UD_ADDRESS_VECTOR_LONG 48
#define UD_ADDRESS_LEN_IN_16_BYTE 1
#define DC_ADDRESS_LEN_IN_16_BYTE 3
#define DC_KEY 0x0
#define MIN_LOG_SZ_FOR_QP_BUFF 13
#define NO_SEND_CQE 0
#define BF_DCT_PORT 1
#define RECV_WQE_SIZE 16
#define UD_TRANSPORT_TYPE 0
#define DC_TRANSPORT_TYPE 1

enum {
	MLX5_CQE_LOCAL_LENGTH_ERR = 0x1,
	MLX5_CQE_LOCAL_QP_OPERATION_ERR = 0x2,
	MLX5_CQE_LOCAL_PROTECTION_ERR = 0x4,
	MLX5_CQE_WORK_REQUEST_FLUSHED_ERR = 0x5,
	MLX5_CQE_MEMORY_WINDOW_BIND_ERR = 0x6,
	MLX5_CQE_BAD_RESPONSE_ERR = 0x10,
	MLX5_CQE_LOCAL_ADDESS_ERR = 0x11,
	MLX5_CQE_REMOTE_INVALID_REQUEST_ERR = 0x12,
	MLX5_CQE_REMOTE_ACCESS_ERR = 0x13,
	MLX5_CQE_REMOTE_OPERATION_ERR = 0x14,
	MLX5_CQE_TRANSPORT_RETRY_COUNTER_EXCEEDED = 0x15,
	MLX5_CQE_RNR_RETRY_COUNTER_EXCEEDED = 0x16,
	MLX5_CQE_ABORTED_ERR = 0x22,
};


struct error_handling_info {
	uint32_t syndrom;
	uint32_t destination_dct;
	int wqe_counter;
};

/* channel_obj holding the essential information to recognize and connect to a QP.
 */
struct cc_qp_obj {
	uint32_t dci_num;
	uint32_t dct_num;
	uint32_t srq_num;
	uint32_t dest_dct;
	char *name;
	void *mkey_buff;
	int mkey;
	uint32_t rqi_out;
	uint32_t rqi_in;
	uint32_t sqi;
	uint32_t sqi_sent;
	size_t log_sq_size;
	size_t log_rq_size;
	size_t rcv_msg_sz;
	size_t snd_msg_sz;
	uint8_t *uar_base_addr;
	struct doca_devx_object *mr_obj;
	struct doca_umem_object *umem;
	struct cc_q_obj *dci_obj;
	struct cc_q_obj *dct_obj;
	struct cc_q_obj *srq_obj;
	struct error_handling_info err_info;
	uint32_t err_counter;
	size_t flush_offset;
};

struct cc_qp {
	struct cc_qp_obj *qp;
	struct cc_cq_obj cq_obj;
	struct cc_av ret_av;
	struct doca_devx_object *connection_obj;
	struct doca_devx_object *service_obj;
	bool is_service;
	int vhca_id;
	bool ib_link;
};

struct cc_qp_common {
	struct devx_context *devx_ctx;
	struct priv_doca_uar *uar;
	uint32_t pd;
	int num_supported_vhca_id;
	bool ib_link;
};

struct destroyed_qp {
	uint32_t dctn;
	uint32_t dcin;
	struct destroyed_qp *next;
};

bool cc_get_common_ib_link_layer(const struct cc_qp_common *common_rs)
{
	return common_rs->ib_link;
}

static inline
const void *
get_recv_msg_address_by_index(const struct cc_qp_obj *qp, uint32_t index)
{
	return (const uint8_t *)qp->mkey_buff +
		((index % (1 << qp->log_rq_size)) * qp->rcv_msg_sz);
}

static inline
void *
get_send_msg_address_by_index(struct cc_qp_obj *qp, uint32_t index)
{
	index += (uint32_t)qp->flush_offset;
	return (uint8_t *)qp->mkey_buff + ((1ull << qp->log_rq_size) * qp->rcv_msg_sz) +
		((index % (1 << qp->log_sq_size)) * qp->snd_msg_sz);
}

static int
reg_mr(struct cc_qp_common *common_resource, struct cc_qp_obj *qp_ch, int pd, void *buff,
	size_t size)
{
	struct qp_params_reg_mr params = {0};

	params.pd = pd;
	params.size = size;
	params.access_mode_1_0 = MLX5_MKC_ACCESS_MODE_MTT;
	params.lw = 1;
	params.lr = 1;
	params.qpn = 0xffffff;
	params.mkey_7_0 = 0x42;
	params.pg_access = 1;

	return devx_reg_mr(common_resource->devx_ctx, buff, &params, &qp_ch->mr_obj,
			   &qp_ch->umem) << 8 | params.mkey_7_0;
}

static void
cq_event_handle_callback(union doca_data user_data)
{
	struct cc_cq_obj *cq = reinterpret_cast<struct cc_cq_obj *>(user_data.ptr);
	cq->is_armed = false;
}

struct cc_cq_obj *
cq_create(struct cc_qp_common *common_rs, int log_cq_size)
{
	doca_error_t status = DOCA_SUCCESS;
	int cq_size = 1 << log_cq_size;
	struct cc_cq_obj *cq_obj = (struct cc_cq_obj *)calloc(1, sizeof(struct cc_cq_obj));
	auto cq_attributes = priv_doca_cq_attr_create(cq_size);

	if (cq_obj == NULL) {
		DOCA_DLOG_ERR("malloc failed for cq_obj in create_cq");
		return NULL;
	}

	if (!common_rs->uar) {
		status = priv_doca_dev_uar_create(common_rs->devx_ctx->doca_device,
			PRIV_DOCA_UAR_ALLOCATION_TYPE_BLUEFLAME, &common_rs->uar);
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("priv_doca_dev_uar_create failed in create_cq");
			goto error_free_cq_obj;
		}
	}

	cq_attributes.initiator_src_dct = true;
	cq_attributes.handler = cq_event_handle_callback;
	cq_attributes.user_data.ptr = cq_obj;
	status = priv_doca_dev_cq_create(common_rs->devx_ctx->doca_device, common_rs->uar, &cq_attributes,
					 &cq_obj->cq_q);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_cq_create failed in create_cq");
		goto error_free_cq_obj;
	}

	priv_doca_cq_number_get(cq_obj->cq_q, &cq_obj->cqn);

	return cq_obj;

error_free_cq_obj:
	free(cq_obj);

	return NULL;
}

void
cq_destroy(struct cc_cq_obj *cq)
{
	auto status = priv_doca_dev_cq_destroy(cq->cq_q);
	if (status != DOCA_SUCCESS)
		DOCA_DLOG_ERR("priv_doca_dev_cq_destroy failed");

	free(cq);
}

static int
to_init(struct doca_devx_object *obj, int qp)
{
	struct qp_params_init params = {0};

	params.qp = qp;
	params.vhca_port_num = 1;

	return devx_to_init(obj, &params);
}

static int
to_rtr(struct doca_devx_object *obj, int qp, int dctn, int ib_link)
{
	struct qp_params_rtr params = {0};

	params.qp = qp;
	params.log_msg_max = log2_by_bits(MAX_DATA_SIZE);
	params.mtu = MLX5_QPC_MTU_4K_BYTES;
	params.dctn = dctn;

	if (!ib_link)
		params.hop_limit = 1;

	return devx_to_rtr(obj, &params);
}

static int
to_rts(struct doca_devx_object *obj, int qp)
{
	struct qp_params_rts params = {0};

	params.qp = qp;
	params.log_ack_req_freq = 8;
	params.retry_count = 7;
	params.rnr_retry = 7;
	params.ack_timeout = 14;

	return devx_to_rts(obj, &params);
}

static void
set_ud_addr(uint8_t *ud_addr, struct cc_av *av, bool is_service, bool ib_link)
{
	struct qp_params_set_ud_addr params = {0};

	params.ext = 1;

	if (ib_link == false)
		params.rlid = 0xc000;
	else
		params.rlid = 0x0000;

	if (is_service && av != NULL)
		params.dest_qpn = av->dct;
	else
		params.dest_qpn = 0xffffff;

	params.src_addr_index = 0xff;

	devx_set_ud_addr(ud_addr, &params);
}

static int
initiate_qp_state_to_work(struct cc_qp_obj *qp, bool ib_link)
{
	if (to_init(qp->dci_obj->q_object, qp->dci_num)) {
		DOCA_DLOG_ERR("to_init failed\n");
		return -1;
	}
	if (to_rtr(qp->dci_obj->q_object, qp->dci_num, qp->dct_num, ib_link)) {
		DOCA_DLOG_ERR("to_rtr failed\n");
		return -1;
	}
	if (to_rts(qp->dci_obj->q_object, qp->dci_num)) {
		DOCA_DLOG_ERR("to_rts failed\n");
		return -1;
	}
	return 0;
}

int
recover_dci(struct cc_qp *qpp)
{
	struct cc_qp_obj *qp = qpp->qp;
	int tmp_dct = qp->err_info.destination_dct;

	if (to_rst(qp->dci_obj->q_object, qp->dci_num)) {
		DOCA_DLOG_ERR("to_rst failed in %s()\n", __func__);
		return -1;
	}
	if (initiate_qp_state_to_work(qp, qpp->ib_link))
		return -1;

	qp->flush_offset = qp->sqi_sent;
	qp->rqi_in = 0;
	qp->rqi_out = 0;
	qp->sqi = 0;
	qp->sqi_sent = 0;
	qp->err_info.syndrom = 0;
	qp->err_info.destination_dct = 0;
	return tmp_dct;
}

static uint32_t
get_dest_dct_from_cqe(const struct mlx5_64b_cqe *cqe, const struct cc_qp_obj *qp)
{
	int wqe_counter = be16toh(cqe->wqe_counter);
	uint32_t true_sqi = wqe_counter % (1 << qp->log_sq_size);
	uint8_t *target_wqe, *ud_addr_struct;

	target_wqe = qp->dci_obj->wq_buff->data() + (true_sqi * MLX5_SEND_WQE_BB);

	ud_addr_struct = (uint8_t *)((struct mlx5_wqe_ctrl_seg *)target_wqe + 1);

	return MLX5_GET(ud_adrs_vector, ud_addr_struct, destination_qp_dct);
}

static int
extract_cqe_error_code(const struct mlx5_64b_cqe *cqe, struct cc_qp_obj *qp)
{
	uint8_t syndrom;

	syndrom = MLX5_GET(cqe_error_syndrom, MLX5_ADDR_OF(cqe_error_layout, cqe, cqe_error_syndrom), syndrom);

	switch (syndrom) {
	case MLX5_CQE_LOCAL_LENGTH_ERR:
	case MLX5_CQE_LOCAL_QP_OPERATION_ERR:
	case MLX5_CQE_LOCAL_PROTECTION_ERR:
	case MLX5_CQE_LOCAL_ADDESS_ERR:
	case MLX5_CQE_TRANSPORT_RETRY_COUNTER_EXCEEDED:
	case MLX5_CQE_RNR_RETRY_COUNTER_EXCEEDED:
	case MLX5_CQE_MEMORY_WINDOW_BIND_ERR:
	case MLX5_CQE_BAD_RESPONSE_ERR:
	case MLX5_CQE_REMOTE_INVALID_REQUEST_ERR:
	case MLX5_CQE_REMOTE_ACCESS_ERR:
	case MLX5_CQE_REMOTE_OPERATION_ERR:
	case MLX5_CQE_ABORTED_ERR:
		DOCA_DLOG_ERR("cqe with error syndrom 0x%x", syndrom);
		qp->err_info.syndrom = DOCA_CC_PEER_ERR;
		qp->err_info.destination_dct = get_dest_dct_from_cqe(cqe, qp);
		qp->err_info.wqe_counter = -1;
		qp->err_counter++;
		return ERR_SEND_MSG;
	case MLX5_CQE_WORK_REQUEST_FLUSHED_ERR: // in October need to re-post, for now just skip this cqe
		qp->err_info.syndrom = DOCA_CC_FLUSH_ERR;
		qp->err_info.destination_dct = get_dest_dct_from_cqe(cqe, qp);
		qp->err_info.wqe_counter = be16toh(cqe->wqe_counter);
		return ERR_SEND_MSG;
	default:
		DOCA_DLOG_ERR("cqe with error syndrom 0x%x is unknown and unsupported", syndrom);
		return UNKNOWN_MSG;
	}
}

static int
poll_cq(struct cc_cq_obj *cq, struct cc_qp_obj *qp, void *data, bool is_service, struct cc_av *av, size_t *len,
	struct destroyed_qps_list *destroyed_qps_list, union imm_msg *return_header)
{
	int ret = ARRIVE_MSG;
	int rq_in;
	union imm_msg msg_hdr = {0};
	const void *msg_data;
	int data_len;
	int cqe_qpn;
	struct priv_doca_64b_cqe cqe;

	if (priv_doca_cq_poll_one(cq->cq_q, &cqe) == DOCA_ERROR_AGAIN)
		return NO_MSG_FOUND;

	cqe_qpn = MLX5_GET(cqe, &cqe, qpn);
	/**< check if the cqe belongs to the right qp */
	if (cqe_qpn != qp->dct_num && cqe_qpn != qp->dci_num) {
		if (destroyed_qps_list) {
			const struct destroyed_qp *destroyed_qp = destroyed_qps_list->destroyed_qp;

			while (destroyed_qp) {
				if (cqe_qpn == destroyed_qp->dctn || cqe_qpn == destroyed_qp->dcin)
					return DESTROYED_QP;
				destroyed_qp = destroyed_qp->next;
			}
		}
		return WRONG_QPN;
	}

	auto polled_cqe = reinterpret_cast<const mlx5_64b_cqe *>(&cqe);
	auto cqe_code = doca_get_cqe_opcode(polled_cqe);
	switch (cqe_code) {
	case MLX5_CQE_RESP_SEND_IMM:
		msg_hdr.imm_msg_uint = doca_get_cqe_imm(polled_cqe);
		if (msg_hdr.handshake.type == DOCA_CC_HANDSHAKE)
			ret = HELLO_MSG;
		else if (msg_hdr.handshake.type == DOCA_CC_REJECT)
			ret = REJECT_MSG;
		else {
			DOCA_DLOG_ERR("message type %d is unrecognized for the comm_channel",
				       msg_hdr.handshake.type);
			return UNKNOWN_MSG;
		}

		if (return_header != NULL)
			memcpy(return_header, &msg_hdr, sizeof(union imm_msg));

		// fall through
	case MLX5_CQE_RESP_SEND:
		rq_in = be16toh(polled_cqe->wqe_counter);
		msg_data = get_recv_msg_address_by_index(qp, rq_in);
		if (is_service)
			av->dct = be32toh(polled_cqe->flags_rqpn) & 0xffffff;
		if ((ret == ARRIVE_MSG) && (*len > 0)) {
			data_len = be32toh(polled_cqe->byte_cnt);

			if (*len > data_len)
				*len = data_len;
			memcpy(data, msg_data, *len);
		}
		qp->rqi_out++;
		break;
	case MLX5_CQE_REQ:
		av->dct = get_dest_dct_from_cqe(polled_cqe, qp);
		qp->sqi_sent += 2;
		ret = SEND_MSG;
		break;
	case MLX5_CQE_REQ_ERR:
		qp->sqi_sent += 2;
		ret = extract_cqe_error_code(polled_cqe, qp);
		DOCA_DLOG_DBG("got error CQE for send WQE on DCI number 0x%x", qp->dci_num);
		break;
	case MLX5_CQE_RESP_ERR:
		qp->rqi_out++;
		ret = ERR_RECV_MSG;
		DOCA_DLOG_DBG("got error CQE for receive WQE on DCT number 0x%x", qp->dct_num);
		break;
	default:
		DOCA_DLOG_ERR("opcode %d is unrecognized for the comm_channel", cqe_code);
		ret = UNKNOWN_MSG;
		break;
	}

	return ret;
}


static void
cc_xmit(struct cc_qp_obj *qp, void *addr, size_t size, struct cc_av *av, bool is_service, bool ib_link,
	const union imm_msg *imm)
{
	uint8_t opcode = MLX5_OPCODE_SEND;
	uint32_t msg_header = 0;
	uint32_t true_sqi = qp->sqi % (1 << qp->log_sq_size);

	struct mlx5_wqe_ctrl_seg *ctrl =
		(struct mlx5_wqe_ctrl_seg *)(qp->dci_obj->wq_buff->data() + (true_sqi * MLX5_SEND_WQE_BB));

	int data_seg_sz = !!size;

	if (imm != NULL) {
		opcode = MLX5_OPCODE_SEND_IMM;
		msg_header = htobe32(imm->imm_msg_uint);
	}

	mlx5dv_set_ctrl_seg(ctrl, qp->sqi, opcode, 0, qp->dci_num, MLX5_WQE_CTRL_CQ_UPDATE,
			    1 + DC_ADDRESS_LEN_IN_16_BYTE + data_seg_sz, 0, msg_header);

	uint8_t *ud_addr = (uint8_t *)(ctrl + 1);

	set_ud_addr(ud_addr, av, is_service, ib_link);

	if (data_seg_sz) {
		struct mlx5_wqe_data_seg *dseg = (struct mlx5_wqe_data_seg *)(ud_addr + UD_ADDRESS_VECTOR_LONG);

		mlx5dv_set_data_seg(dseg, static_cast<uint32_t>(size), qp->mkey, (uintptr_t)addr);
	}

	qp->sqi += 2;
	priv_doca_memory_barrier();
	(*qp->dci_obj->dbr)[MLX5_SND_DBR] = htobe32(qp->sqi & 0xffff);
	priv_doca_memory_barrier();
	*(uint64_t *)(qp->uar_base_addr + 0x800) = *(uint64_t *)ctrl;
	priv_doca_memory_barrier();
}

static void
cc_recv(struct cc_qp_obj *qp, void const *addr, size_t size)
{
	uint32_t true_rqi = qp->rqi_in % (1 << qp->log_rq_size);

	struct mlx5_wqe_data_seg *dseg = (struct mlx5_wqe_data_seg *)((uint8_t *)qp->srq_obj->wq_buff->data() +
									true_rqi * RECV_WQE_SIZE);
	mlx5dv_set_data_seg(dseg, static_cast<uint32_t>(size), qp->mkey, (uintptr_t)addr);
	qp->rqi_in++;
	priv_doca_memory_barrier();
	(*qp->srq_obj->dbr)[MLX5_RCV_DBR] = htobe32(qp->rqi_in & 0xffff);
}

struct ibv_context *
get_ibv_ctx(struct cc_qp_common *common)
{
	return common->devx_ctx->ctx;
}

struct doca_dev *
get_doca_dev(struct cc_qp_common *common)
{
	return common->devx_ctx->doca_device;
}

struct priv_doca_comp_events_channel *
cc_qp_event_ch_create(struct cc_qp_common *common_rs, int flag)
{
	bool is_blocking = false; /*DOCA CC only supports non-block */
	priv_doca_comp_events_channel *evt_channel = nullptr;
	auto status = priv_doca_dev_comp_events_channel_create(common_rs->devx_ctx->doca_device, is_blocking,
							       &evt_channel);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_comp_events_channel_create failed with status %s",
			      doca_get_error_name(status));
		return nullptr;
	}

	return evt_channel;
}

int
cc_event_channel_subscribe(struct cc_cq_obj *cq, struct priv_doca_comp_events_channel *event_channel)
{
	auto status = priv_doca_comp_events_channel_register(event_channel, cq->cq_q);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_comp_events_channel_register failed with status %s",
			      doca_get_error_name(status));
		return -1;
	}

	cq->events_channel = event_channel;

	arm_cq(cq);

	return 0;
}

static int
check_user_attr(struct doca_cc_properties *cc_properties, struct cc_qp_common *common_rs)
{
	int ret;
	struct cc_qp_cap cap = {0};

	ret = devx_query_capabilities(common_rs->devx_ctx, &cap);
	if (ret)
		return ret;

	if (cap.log_max_connections != 0) {
		cc_properties->max_clients = 1 << cap.log_max_connections;
		DOCA_DLOG_INFO("LOG_MAX_CONNECTIONS is %d, MAX_CLIENTS set to: %d", cap.log_max_connections,
				cc_properties->max_clients);
	}

	if (cc_properties->log_rcv_queue_size > cap.log_max_queue_sz) {
		DOCA_DLOG_ERR("max number of rcv messages must be <= %d, and is wanted to be %d",
			1 << cap.log_max_queue_sz, 1 << cc_properties->log_rcv_queue_size);
		return -1;
	}
	if (cc_properties->log_snd_queue_size > cap.log_max_queue_sz) {
		DOCA_DLOG_ERR("max number of snd messages must be <= %d, and is wanted to be %d",
			1 << cap.log_max_queue_sz, 1 << cc_properties->log_snd_queue_size);
		return -1;
	}
	if (cc_properties->log_msg_size > cap.log_max_msg) {
		DOCA_DLOG_ERR("max of message size must be <= %d, and is wanted to be %d",
			1 << cap.log_max_msg, 1 << cc_properties->log_msg_size);
		return -1;
	}

	if (cap.num_ports == 1) {
		common_rs->num_supported_vhca_id = cap.num_vhca_ports;
	}

	return 0;
}

static int
cc_is_ib(struct devx_context *devx_ctx)
{
	int ret;
	struct cc_qp_cap cap = {0};

	ret = devx_query_capabilities(devx_ctx, &cap);
	if (ret)
		return -1;

	switch (cap.port_type) {
	case MLX5_CAP_PORT_TYPE_IB:
		return 1;
	case MLX5_CAP_PORT_TYPE_ETH:
		return 0;
	default:
		DOCA_DLOG_ERR("unexpected port link layer %d", cap.port_type);
		return -1;
	}
}

struct cc_qp_common *
cc_qp_common_init(struct doca_cc_properties *cc_properties, bool *is_service)
{
	int rc;
	struct devx_context *devx_ctx;
	struct cc_qp_common *common_resource;

	devx_ctx = (struct devx_context *)calloc(1, sizeof(struct devx_context));
	if (devx_ctx == NULL) {
		DOCA_DLOG_ERR("devx_ctx memory allocation failed in %s()", __func__);
		return NULL;
	}

	devx_ctx->doca_device = cc_properties->hw_dev;

	common_resource = (struct cc_qp_common *)calloc(1, sizeof(struct cc_qp_common));
	if (common_resource == NULL) {
		DOCA_DLOG_ERR("malloc for cc_qp_common failed in %s()", __func__);
		goto error_free_ctx;
	}

	common_resource->devx_ctx = devx_ctx;

	rc = cc_is_ib(devx_ctx);
	if (rc < 0) {
		DOCA_DLOG_ERR("query for link type failed in %s", __func__);
		goto error_free_common;
	}
	common_resource->ib_link = rc;

	if (priv_doca_dev_uar_create(devx_ctx->doca_device, PRIV_DOCA_UAR_ALLOCATION_TYPE_BLUEFLAME,
				     &common_resource->uar) == DOCA_SUCCESS)
		*is_service = false;
	else {
		*is_service = true;
		if (priv_doca_dev_uar_create(devx_ctx->doca_device, PRIV_DOCA_UAR_ALLOCATION_TYPE_NONCACHE,
					     &common_resource->uar) != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("priv_doca_dev_uar_create failed in %s()", __func__);
			goto error_free_common;
		}
	}

	if (priv_doca_dev_pd_id_get(common_resource->devx_ctx->doca_device, &common_resource->pd) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_dev_pd_id_get failed in %s()", __func__);
		goto error_free_uar;
	}

	if (check_user_attr(cc_properties, common_resource))
		goto error_free_uar;

	return common_resource;

error_free_uar:
	priv_doca_dev_uar_destroy(common_resource->uar);
error_free_common:
	free(common_resource);
error_free_ctx:
	free(devx_ctx);
	return NULL;
}

void
cc_qp_common_destroy(struct cc_qp_common *common_resource)
{
	if (common_resource == NULL)
		return;

	if (common_resource->uar)
		priv_doca_dev_uar_destroy(common_resource->uar);

	if (common_resource->devx_ctx) {
		free(common_resource->devx_ctx);
	}

	free(common_resource);
}

struct cc_qp_obj *cc_one_qp_create(bool is_snd, const struct doca_cc_properties *cc_attr,
				   struct cc_qp_common *common_resource, struct cc_cq_obj *snd_cq, struct cc_cq_obj *rcv_cq);

static void cc_one_qp_destroy(struct cc_qp_obj *qp, struct destroyed_qps_list *destroyed_list);

struct cc_qp *
cc_qpp_create(struct cc_qp_common *common, const struct doca_cc_properties *cc_attr, struct cc_cq_obj *snd_cq,
		struct cc_cq_obj *rcv_cq)
{
	struct cc_qp *qpp;

	qpp = (struct cc_qp *)calloc(1, sizeof(struct cc_qp));
	if (qpp == NULL) {
		DOCA_DLOG_ERR("qpp mem alocation failed");
		return NULL;
	}

	DOCA_DLOG_DBG("Creating send side QP...");
	qpp->qp = cc_one_qp_create(1, cc_attr, common, snd_cq, rcv_cq);
	if (qpp->qp == NULL) {
		DOCA_DLOG_ERR("qp_create failed\n");
		goto error_free_qp;
	}

	qpp->connection_obj = NULL;
	qpp->service_obj = NULL;
	qpp->is_service = false;
	qpp->ib_link = common->ib_link;
	qpp->vhca_id = 0;

	return qpp;


error_free_qp:
	free(qpp);
	return NULL;
}

struct cc_qp_obj *
cc_one_qp_create(bool is_snd, const struct doca_cc_properties *cc_attr,
		 struct cc_qp_common *common_resource, struct cc_cq_obj *snd_cq, struct cc_cq_obj *rcv_cq)
{
	int i, ret;
	int srqn, dctn, dcin;
	size_t log_rq_size = cc_attr->log_rcv_queue_size;
	size_t log_sq_size = cc_attr->log_snd_queue_size + 1; /**< DC snd takes double space */
	struct cc_qp_obj *qp_channel;
	size_t buff_sz;
	int mkey;
	struct cc_q_obj *srq_obj;
	struct cc_q_obj *dct_obj;
	struct cc_q_obj *dci_obj;
	void *uar_base_addr;
	doca_error_t status = DOCA_SUCCESS;

	/* buff_sz represent the data buffer needed for the QP. count only for receive or transmit
	 * because every QP is for one side communication (send_qp or recv_qp)
	 */

	buff_sz = (1 << (cc_attr->log_msg_size + log_rq_size)) + (1 << (cc_attr->log_msg_size + log_sq_size));
	if (buff_sz < MIN_BUFFER_SIZE) {
		DOCA_DLOG_DBG("Info: mkey buffer size must be >= %d, updating buffer size", MIN_BUFFER_SIZE);
		buff_sz = MIN_BUFFER_SIZE;
	}

	qp_channel = (struct cc_qp_obj *)calloc(1, sizeof(struct cc_qp_obj));
	if (!qp_channel) {
		DOCA_DLOG_ERR("malloc failed for qp_channel in %s()", __func__);
		return NULL;
	}

	// void *buff = (void *)memalign(0x800, buff_sz*2);
	char *buff = (char *)malloc(buff_sz);

	if (buff == NULL) {
		DOCA_DLOG_ERR("malloc failed for data buffer in %s()", __func__);
		goto error_free_qp;
	}

	srq_obj = (struct cc_q_obj *)calloc(1, sizeof(struct cc_q_obj));
	if (srq_obj == NULL) {
		DOCA_DLOG_ERR("malloc failed for srq_obj in %s()", __func__);
		goto error_free_buff;
	}

	dci_obj = (struct cc_q_obj *)calloc(1, sizeof(struct cc_q_obj));
	if (dci_obj == NULL) {
		DOCA_DLOG_ERR("malloc failed for dci_obj in %s()", __func__);
		goto error_free_srq_obj;
	}

	dct_obj = (struct cc_q_obj *)calloc(1, sizeof(struct cc_q_obj));
	if (dct_obj == NULL) {
		DOCA_DLOG_ERR("malloc failed for dct_obj in %s()", __func__);
		goto error_free_dci_obj;
	}

	mkey = reg_mr(common_resource, qp_channel, common_resource->pd, buff, buff_sz);
	if (!mkey)
		goto error_free_dct_obj;

	ret = get_issi(common_resource->devx_ctx);
	if (ret != 1) {
		DOCA_DLOG_ERR("Unsupported ISSI - %d", ret);
		goto error_dereg_mem;
	}
	srqn = create_rmp(common_resource->devx_ctx, common_resource->pd, static_cast<int>(log_rq_size), srq_obj);
	if (srqn == 0)
		goto error_dereg_mem;

	dctn = create_dct(common_resource->devx_ctx, rcv_cq->cqn, common_resource->pd, BF_DCT_PORT, srqn, dct_obj);
	if (dctn == 0)
		goto error_free_srq;

	dcin = create_dci(common_resource->devx_ctx, common_resource->uar, snd_cq->cqn,
			   common_resource->pd, log_sq_size, dci_obj);
	if (dcin == 0)
		goto error_free_dct;

	status = priv_doca_uar_page_get(common_resource->uar, &uar_base_addr);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("priv_doca_uar_page_get failed in %s()", __func__);
		goto error_free_dci;
	}

	qp_channel->mkey = mkey;
	qp_channel->mkey_buff = buff;
	qp_channel->uar_base_addr = reinterpret_cast<uint8_t *>(uar_base_addr);
	qp_channel->dci_num = dcin;
	qp_channel->dct_num = dctn;
	qp_channel->srq_num = srqn;
	qp_channel->dest_dct = 0;
	qp_channel->rqi_in = 0;
	qp_channel->rqi_out = 0;
	qp_channel->sqi = 0;
	qp_channel->sqi_sent = 0;
	qp_channel->log_sq_size = log_sq_size;
	qp_channel->log_rq_size = log_rq_size;
	qp_channel->dci_obj = dci_obj;
	qp_channel->dct_obj = dct_obj;
	qp_channel->srq_obj = srq_obj;
	qp_channel->snd_msg_sz = (1ull << cc_attr->log_msg_size);
	qp_channel->rcv_msg_sz = (1ull << cc_attr->log_msg_size);
	qp_channel->err_info.syndrom = 0;
	qp_channel->err_info.destination_dct = 0;
	qp_channel->err_counter = 0;
	qp_channel->flush_offset = 0;
	DOCA_DLOG_DBG("DCIN = 0x%x, DCTN = 0x%x, SRQN = 0x%x",qp_channel->dci_num, qp_channel->dct_num, qp_channel->srq_num);

	if (initiate_qp_state_to_work(qp_channel, common_resource->ib_link))
		goto error_free_dci;

	for (i = 0; i < (1 << log_rq_size); i++)
		cc_recv(qp_channel, get_recv_msg_address_by_index(qp_channel, i),
			qp_channel->rcv_msg_sz);

	DOCA_DLOG_DBG("DCI number = 0x%x, SRQ number = 0x%x, DCT number = 0x%x", dcin, srqn, dctn);
	return qp_channel;

error_free_dci:
	cc_q_obj_destroy(dci_obj);
error_free_dct:
	cc_q_obj_destroy(dct_obj);
error_free_srq:
	cc_q_obj_destroy(srq_obj);
error_dereg_mem:
	priv_doca_devx_object_destroy(qp_channel->mr_obj);
	priv_doca_umem_destroy(qp_channel->umem);
error_free_dct_obj:
	free(dct_obj);
error_free_dci_obj:
	free(dci_obj);
error_free_srq_obj:
	free(srq_obj);
error_free_buff:
	free(buff);
error_free_qp:
	free(qp_channel);
	return NULL;
}

void
cc_channel_destroy(struct cc_qp *qpp)
{
	if (qpp == NULL)
		return;
	if (qpp->connection_obj != NULL) {
		auto status = priv_doca_devx_object_destroy(qpp->connection_obj);
		if (status != DOCA_SUCCESS)
			DOCA_DLOG_ERR("priv_doca_devx_object_destroy failed for connection_obj with error %s",
				      doca_get_error_name(status));
		else
			qpp->connection_obj = NULL;
	}
	if (qpp->service_obj != NULL) {
		auto status = priv_doca_devx_object_destroy(qpp->service_obj);
		if (status != DOCA_SUCCESS)
			DOCA_DLOG_ERR("priv_doca_devx_object_destroy failed for service_obj with error %s",
				      doca_get_error_name(status));
		else
			qpp->service_obj = NULL;
	}
}

static void
cc_one_qp_destroy(struct cc_qp_obj *qp, struct destroyed_qps_list *destroyed_list)
{
	struct destroyed_qp *new_destroyed;
	doca_error_t status;
	int ret;

	if (qp == NULL)
		return;

	ret = cc_q_obj_destroy(qp->dci_obj);
	if (ret)
		DOCA_DLOG_ERR("devx_obj_destroy failed for dci_obj with errno = %d", ret);
	free(qp->dci_obj);

	ret = cc_q_obj_destroy(qp->dct_obj);
	if (ret)
		DOCA_DLOG_ERR("devx_obj_destroy failed for dct_obj with errno = %d", ret);
	free(qp->dct_obj);

	ret = cc_q_obj_destroy(qp->srq_obj);
	if (ret)
		DOCA_DLOG_ERR("devx_obj_destroy failed for srq_obj with errno = %d", ret);
	free(qp->srq_obj);

	status = priv_doca_devx_object_destroy(qp->mr_obj);
	if (status != DOCA_SUCCESS)
		DOCA_DLOG_ERR("priv_doca_devx_object_destroy failed for mr_obj with error %s",
			      doca_get_error_name(status));

	status = priv_doca_umem_destroy(qp->umem);
	if (status != DOCA_SUCCESS)
		DOCA_DLOG_ERR("priv_doca_umem_destroy failed with status %s", doca_get_error_name(status));

	if (qp->mkey_buff)
		free(qp->mkey_buff);

	if (destroyed_list) {
		new_destroyed = (struct destroyed_qp *)calloc(1, sizeof(struct destroyed_qp));
		if (new_destroyed == NULL) {
			DOCA_DLOG_ERR("error allocating destroyed QP object");
		} else {
			new_destroyed->dctn = qp->dct_num;
			new_destroyed->dcin = qp->dci_num;
			new_destroyed->next = destroyed_list->destroyed_qp;
			destroyed_list->destroyed_qp = new_destroyed;
		}
	}

	free(qp);
}

void
cc_qpp_destroy(struct cc_qp *qpp, struct destroyed_qps_list *destroyed_list)
{
	if (qpp == NULL)
		return;

	cc_channel_destroy(qpp);
	if (qpp->qp != NULL)
		cc_one_qp_destroy(qpp->qp, destroyed_list);
	free(qpp);
}

int
cc_qp_msg_send(struct cc_qp *qpp, struct cc_cq_obj *cq, const void *data, size_t len, struct cc_av *dest_av,
	       const union imm_msg *send_header)
{
	uint8_t *mkey_buff_send_addr;
	struct cc_qp_obj *qp = qpp->qp;
	uint32_t true_sqi = qp->sqi % (1 << qp->log_sq_size);
	uint32_t true_sqo = qp->sqi_sent % (1 << qp->log_sq_size);

	if (unlikely((true_sqi == true_sqo) && ((qp->sqi >> qp->log_sq_size) != (qp->sqi_sent >> qp->log_sq_size)))) {
		return -QUEUE_FULL;
	}

	mkey_buff_send_addr = (uint8_t *)get_send_msg_address_by_index(qpp->qp, qpp->qp->sqi);

	memcpy(mkey_buff_send_addr, data, len);

	cc_xmit(qpp->qp, mkey_buff_send_addr, len, dest_av, qpp->is_service, qpp->ib_link, send_header);

	return 0;
}

int
cc_qp_msg_send_handshake(struct cc_qp *qpp, struct cc_cq_obj *cq, struct cc_av *dest_av,
			 const union imm_msg *send_header)
{
	int data = 0;

	return cc_qp_msg_send(qpp, cq, &data, sizeof(data),
			      dest_av, send_header);
}

int cc_qp_msg_recv(struct cc_qp *qpp, struct cc_cq_obj *cq, void *data, size_t *len,
		   struct destroyed_qps_list *destroyed_qps_list, union imm_msg *recv_header)
{
	int ret;
	const void *mkey_buff_recv_addr = get_recv_msg_address_by_index(qpp->qp, qpp->qp->rqi_in);

	do {
		ret = poll_cq(cq, qpp->qp, data, qpp->is_service, &(qpp->ret_av), len, destroyed_qps_list, recv_header);
		switch (ret) {
		case NO_MSG_FOUND:
			return ret;
		case ERR_RECV_MSG:
			DOCA_DLOG_ERR("receive wqe failed");
		/* FALLTHROUGH */
		case UNKNOWN_MSG: /* FALLTHROUGH */
		case WRONG_QPN:
			return ret;
		default:
			continue;
		}

	} while (ret == SEND_MSG || ret == DESTROYED_QP);

	cc_recv(qpp->qp, mkey_buff_recv_addr, qpp->qp->rcv_msg_sz);
	return ret;
}

int cc_qp_check_send_cqe(struct cc_qp *qpp, struct cc_cq_obj *cq, struct destroyed_qps_list *destroyed_qps_list, struct cc_av *av)
{
	int ret;
	size_t len = 0;

	do {
		ret = poll_cq(cq, qpp->qp, NULL, qpp->is_service, &(qpp->ret_av), &len, destroyed_qps_list, NULL);
	} while (ret == DESTROYED_QP);
	if (ret == SEND_MSG)
		av->dct = qpp->ret_av.dct; /* if it is send cqe, we want to know to where it was sent*/
	return ret;

}

int
cc_qp_register_client(struct cc_qp *qpp, struct cc_qp_common *common_rs,
			const char *service)
{
	struct channel_params ch_params = {0};
	char name[SERVICE_NAME_FULL];

	os_abstraction::priv_doca_strlcpy(name, service, SERVICE_NAME_FULL);
	ch_params.service_name = name;
	ch_params.dci_dct_num = qpp->qp->dci_num;
	ch_params.transport_type = DC_TRANSPORT_TYPE;
	ch_params.vhca_id = 0; /* vhca_id not allowed in client registration */
	ch_params.flag = CONNECT_BY_SERVICE_NAME; /* connection type */
	ch_params.qkey = 0;
	qpp->connection_obj = register_channel_connection(common_rs->devx_ctx, &ch_params);
	if (qpp->connection_obj == NULL) {
		DOCA_DLOG_ERR("client registration failed for send side\n");
		return -1;
	}

	return 0;
}

static int
cc_qp_register_service_connection(struct cc_qp *qpp, struct cc_qp_common *common_rs, int type)
{
	struct channel_params ch_params = {0};

	ch_params.dci_dct_num = qpp->qp->dci_num;
	ch_params.transport_type = type;
	ch_params.vhca_id = qpp->vhca_id;
	ch_params.flag = CONNECT_BY_VHCA_ID; /* connection type */
	ch_params.qkey = 0;
	qpp->connection_obj = register_channel_connection(common_rs->devx_ctx, &ch_params);
	if (qpp->connection_obj == NULL) {
		DOCA_DLOG_ERR("client registration failed for recv side\n");
		return -1;
	}
	return 0;
}

int
cc_qp_register_server(struct cc_qp *qpp, struct cc_qp_common *common, const char *service, uint16_t max_clients)
{
	struct channel_params ch_params = {0};
	char name[SERVICE_NAME_FULL];
	int type = DC_TRANSPORT_TYPE;

	os_abstraction::priv_doca_strlcpy(name, service, SERVICE_NAME_FULL);
	ch_params.service_name = name;
	ch_params.dci_dct_num = qpp->qp->dct_num;
	ch_params.vhca_id = qpp->vhca_id;
	ch_params.transport_type = type;
	ch_params.max_num_connected_client = max_clients;

	qpp->service_obj = register_channel_service(common->devx_ctx, &ch_params);
	if (qpp->service_obj == NULL) {
		DOCA_DLOG_ERR("server registration failed for recv side\n");
		return -1;
	}

	/* register service connection */
	if (cc_qp_register_service_connection(qpp, common, type)) {
		DOCA_DLOG_ERR("service channel connection failed");
		return -EFAULT;
	}

	qpp->is_service = true;
	return 0;
}


int
vhca_id_get(struct cc_qp *qpp)
{
	return qpp->vhca_id;
}

void
vhca_id_set(int vhca_id, struct cc_qp *qpp)
{
	qpp->vhca_id = vhca_id;
}

struct cc_av *
av_get(struct cc_qp *qpp)
{
	return &(qpp->ret_av);
}

void destroy_destroyed_qps_list(struct destroyed_qps_list *list)
{
	struct destroyed_qp *curr_destroyed_qp = list->destroyed_qp;
	struct destroyed_qp *next_destroyed_qp = NULL;

	while (curr_destroyed_qp) {
		next_destroyed_qp = curr_destroyed_qp->next;
		free(curr_destroyed_qp);
		curr_destroyed_qp = next_destroyed_qp;
	}
	list->destroyed_qp = NULL;
}

bool
is_armed(const struct cc_cq_obj *cq)
{
	return cq->is_armed;
}

uint32_t
get_cqe_error_syndrom(const struct cc_qp *qpp)
{
	return qpp->qp->err_info.syndrom;
}

int
get_cqe_error_dest_dct(const struct cc_qp *qpp)
{
	return qpp->qp->err_info.destination_dct;
}

void
add_wqe_to_repost_buf(const struct cc_qp *qpp, uint8_t *temp_buf, int bad_dctn)
{
	uint32_t recovered_sqi = qpp->qp->sqi % (1 << qpp->qp->log_sq_size);
	uint32_t flushed_sqi = qpp->qp->err_info.wqe_counter % (1 << qpp->qp->log_sq_size);
	uint8_t *temp_wqe = temp_buf + (recovered_sqi * MLX5_SEND_WQE_BB);
	uint8_t *target_wqe = (uint8_t *)qpp->qp->dci_obj->wq_buff->data() + (flushed_sqi * MLX5_SEND_WQE_BB);

	if (qpp->qp->err_info.destination_dct != bad_dctn) {
		memcpy(temp_wqe, target_wqe, MLX5_SEND_WQE_BB * 2);
		update_wqe_ctrl_index(temp_wqe, recovered_sqi);
		qpp->qp->sqi += 2;
	}
}

void
repost_flushed_wqes(const struct cc_qp *qpp, uint8_t *temp_repost_buf)
{
	void *last_wqe_posted;
	uint32_t true_sqi = qpp->qp->sqi % (1 << qpp->qp->log_sq_size);

	if (qpp->qp->sqi == 0) /* true only if add_wqe_to_repost_buf() been called successfully before */
		return;

	qpp->qp->sqi_sent = 0;

	memcpy(qpp->qp->dci_obj->wq_buff->data(), temp_repost_buf, true_sqi * MLX5_SEND_WQE_BB);

	memset(qpp->qp->dci_obj->wq_buff->data() + true_sqi * MLX5_SEND_WQE_BB, 0, 2 * MLX5_SEND_WQE_BB);

	last_wqe_posted = qpp->qp->dci_obj->wq_buff->data() + (((qpp->qp->sqi - 2) % (1 << qpp->qp->log_sq_size)) * MLX5_SEND_WQE_BB);

	priv_doca_memory_barrier();
	(*qpp->qp->dci_obj->dbr)[MLX5_SND_DBR] = htobe32(qpp->qp->sqi & 0xffff);
	priv_doca_memory_barrier();
	*(uint64_t *)((uint8_t *)qpp->qp->uar_base_addr + 0x800) = *(uint64_t *)last_wqe_posted;
	priv_doca_memory_barrier();
}

void
cc_force_arm(struct cc_cq_obj *cq)
{
	force_arm(cq);
}
