/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include "doca_compress.h"

#include <errno.h>
#include <pthread.h>
#include <stdint.h>
#include <stdlib.h>

#include <infiniband/mlx5dv.h>
#include <infiniband/verbs.h>

#include <doca_log.h>
#include <doca_rdma.h>
#include <priv_doca_buf.h>
#include <priv_doca_ctx.h>
#include <priv_doca_dev.h>
#include <priv_doca_malloc.h>
#include <priv_doca_umr.h>
#include <private/common_defines.h>

#include "impl/compress_mlx5_common.h"
#include "impl/compress_mlx5_prm.h"

DOCA_LOG_REGISTER(DOCA_COMPRESS);

#define LOG_QP_DEPTH 10
#define WORKQ_CTX_ALIGNMENT 32
#define MAX_COMPRESS_BUF_LIST_SIZE 128

struct workq_context_base {
	// FUTURE - try to order these so that usage patterns in enqueue and dequeue use them sequentially, cache line
	// at a time
	struct doca_dev *dev;
	struct ibv_context *ibv_ctx;
	uint8_t *uar_base_addr;
	struct ibv_pd *pd;
	struct ibv_mr *metadata_mr;
	struct priv_doca_cq *compress_cq;
	struct compress_qp *compress_qp;
	struct priv_doca_uar *uar;
	uint8_t registered_with_compress_ctx;
	uint32_t device_max_compress_job_len;
	uint32_t device_max_decompress_job_len;
	bool umr_support;
};

struct polled_workq_context {
	struct workq_context_base base;
} __attribute__((aligned(WORKQ_CTX_ALIGNMENT)));

struct event_driven_workq_context {
	struct workq_context_base base;
	struct priv_doca_comp_events_channel *channel;
	doca_event_handle_t channel_handle;
	struct doca_workq *workq;
	struct priv_doca_workq_progress_entry *poller;
} __attribute__((aligned(WORKQ_CTX_ALIGNMENT)));

struct compress_caps {
	bool compress_qp_enabled;
	bool compress_sq_enabled;
	bool decompress_qp_enabled;
	bool decompress_sq_enabled;
	uint32_t device_qp_ts_format;
	uint32_t device_max_compress_job_len;
	uint32_t device_max_decompress_job_len;
	bool umr_enable;
};

struct doca_compress {
	struct doca_ctx base;
	struct doca_dev *dev;
	pthread_mutex_t *mutex;	     // FUTURE - put this into the base?
	size_t attached_workq_count; // FUTURE - put this into the base?
	uint8_t started;	     // FUTURE - put this into the base?
	uint32_t device_max_compress_job_len;
	uint32_t device_max_decompress_job_len;
};

/**********************************************************************************************************************
 * Local private function declarations.
 */

/**
 * Initialise the generic context vtable with appropriate compress function calls.
 *
 * @param [in] ops
 * The vtable to populate.
 */
static void init_ops(struct priv_doca_ctx_ops *ops);

/**
 * Verify the doca_ctx is not null, cast it to the concrete doca_compress type and lock the mutex to prevent
 * concurrency issues mutating the context.
 *
 * @param [in] ctx
 * Pointer to the context as a pointer to the base ctx type.
 *
 * @param [out] compress_ctx
 * Pointer to be populated with the correct casting to restore the concrete type from the base ctx type.
 *
 * @Return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t validate_object_and_lock(struct doca_ctx *ctx, struct doca_compress **compress_ctx);

/**
 * Query COMPRESS relevant capabilities of given device
 *
 * @param [in] info
 * The DOCA device information.
 * @param [out] caps
 * The available capabilities.
 *
 * @return
 * On success returns DOCA_SUCCESS. Otherwise:
 *
 */
static doca_error_t compress_caps_get(const struct doca_devinfo *info, struct compress_caps *caps);

/**
 * Check if given COMPRESS caps are enough for compress
 *
 * @param [in] caps
 * The COMPRESS capabilities, output of compress_caps_get()
 *
 * @return
 * true if given caps are enough for compress, false otherwise.
 */
static bool compress_caps_supported(struct compress_caps *caps);

/**
 * Check if given COMPRESS caps are enough for decompress
 *
 * @param [in] caps
 * The COMPRESS capabilities, output of compress_caps_get()
 *
 * @return
 * true if given caps are enough for decompress, false otherwise.
 */
static bool decompress_caps_supported(struct compress_caps *caps);

/**
 * Create a polling work queue context. Each workq context can be used with a single work queue,
 * and is expected to never be called from multiple threads so no concurrency primitives are provided.
 *
 * @param [in] dev
 * The doca dev this context will use - cached from the compress_context to avoid access to the compress_context.
 *
 * @return
 * Non NULL upon success, NULL otherwise.
 */
static struct polled_workq_context *create_polled_workq_ctx(struct doca_dev *dev);

/**
 * Create an event driven work queue context. Each workq context can be used with a single work queue,
 * and is expected to never be called from multiple threads so no concurrency primitives are provided.
 * This context provides support for blocking events.
 *
 * @param [in] dev
 * The doca dev this context will use - cached from the compress_context to avoid access to the compress_context.
 *
 * @return
 * Non NULL upon success, NULL otherwise.
 */
static struct event_driven_workq_context *create_event_driven_workq_ctx(struct doca_dev *dev);

/**
 * Common creation logic of the work queue context. Unified creation logic for both event and polling contexts.
 * Used by create_polled_workq_ctx & create_event_driven_workq_ctx.
 *
 * @param [in] dev
 * The doca dev this context will use - cached from the compress_context to avoid access to the compress_context.
 * @param [in] workq_ctx
 * A newly malloc'd work queue context.
 *
 * @return
 * DOCA_SUCESS upon success, appropriate doca_error_t otherwise.
 */
static doca_error_t init_workq_ctx_base(struct doca_dev *dev, struct workq_context_base *workq_ctx);

/**
 * Destroy work queue context and all resources owned by it.
 *
 * @param [in] workq_ctx
 * Pointer to workq to be destroyed.
 */
static void destroy_polled_workq_ctx(struct polled_workq_context *workq_ctx);

/**
 * Destroy event driven work queue context and all resources owned by it.
 *
 * @param [in] workq_ctx
 * Pointer to workq to be destroyed.
 */
static void destroy_event_driven_workq_ctx(struct event_driven_workq_context *workq_ctx);

/**
 * Destroy resources that are common for both polling & event mode work queue contexts.
 * Used by destroy_polled_workq_ctx & destroy_event_driven_workq_ctx.
 *
 * @param [in] workq_ctx
 * Pointer to workq to be destroyed.
 */
static void destroy_workq_ctx_base(struct workq_context_base *workq_ctx);

/**
 * Validate settings and start the COMPRESS context.
 *
 * @note
 * A COMPRESS context must be started before any work queues may be attached as we cannot safely update the work queue
 * contexts which may be used by another thread and placing locking mechanisms in the hot path is not acceptable so we
 * must ensure everything is valid in this context first so that when creating work queues it can copy / use / cache any
 * values from the context with confidence that they are correct and essentially const while the context is running.
 *
 * @param [in] ctx
 * The COMPRESS context.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_ctx_start(struct doca_ctx *ctx);

/**
 * Stop the COMPRESS context.
 *
 * @note
 * Any attached work queue must be detached before the context will allow itself to return to the stopped state.
 *
 * @note
 * After stopping the context can be reconfigured or destroyed.
 *
 * @param [in] ctx
 * The COMPRESS context.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_ctx_stop(struct doca_ctx *ctx);

/**
 * Attach a device to the COMPRESS context.
 *
 * @note
 * The COMPRESS context only supports a single device being attached. Previous device should be detached before
 * attaching a new one.
 *
 * @param [in] ctx
 * The COMPRESS context.
 *
 * @param [in] dev
 * The device to attach to COMPRESS.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_ctx_dev_add(struct doca_ctx *ctx, struct doca_dev *dev);

/**
 * Detach a device from the COMPRESS context.
 *
 * @param [in] ctx
 * The COMPRESS context.
 *
 * @param [in] dev
 * The device to detach from COMPRESS.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_ctx_dev_rm(struct doca_ctx *ctx, struct doca_dev *dev);

/**
 * Attach a work queue to the COMPRESS instance.
 *
 * @param [in] ctx
 * The COMPRESS context.
 *
 * @param [in] workq
 * The work queue to attach.
 *
 * @param [out] data_priv
 * Opaque pointer which the engine must populate with a priv_doca_workq_progress_entry which is allocated by the
 * COMPRESS context. This will hold the polled_workq_context, and progress callbacks.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_polled_workq_ctx_add(struct doca_ctx *ctx, struct doca_workq *workq,
						  union doca_data *data_priv);

/**
 * Attach an event driven work queue to the COMPRESS instance.
 *
 * @details In case context was added to work queue in event driven mode, this callback will be invoked instead of
 * compress_polled_workq_ctx_add().
 *
 * @param [in] ctx
 * The COMPRESS context.
 *
 * @param [in] workq
 * The work queue to attach.
 *
 * @param [out] data_priv
 * Opaque pointer which the engine must populate with a priv_doca_workq_progress_entry which is allocated by the
 * COMPRESS context. This will hold the event_driven_workq_context base, and progress callbacks.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_event_driven_workq_ctx_add(struct doca_ctx *ctx, struct doca_workq *workq,
							union doca_data *data_priv);

/**
 * Logic that is common for attaching a polling or event mode work queue context.
 *
 * @param [in] compress_ctx
 * The COMPRESS context.
 *
 * @param [in] workq
 * The work queue to attach.
 *
 * @param [in] workq_ctx
 * The work queue context base structure.
 *
 * @param [out] poller_out
 * The newly created progress entry, that was added to workq.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_workq_ctx_base_add(struct doca_compress *compress_ctx, struct doca_workq *workq,
						struct workq_context_base *workq_ctx,
						struct priv_doca_workq_progress_entry **poller_out);

/**
 * Detach a work queue from the COMPRESS context.
 *
 * @param [in] ctx
 *
 * @param [in] workq
 * The work queue to detach.
 *
 * @param [out] data_priv
 * Opaque pointer which will refer to the value output during compress_polled_workq_ctx_add in its data_priv out
 * param. The COMPRESS instance MUST release all resources held by this object.
 * (Undo everything done during compress_polled_workq_ctx_add).
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_polled_workq_ctx_rm(struct doca_ctx *ctx, struct doca_workq *workq,
						 union doca_data data_priv);

/**
 * Detach an event driven work queue from the COMPRESS context.
 *
 * @param [in] ctx
 * The COMPRESS context.
 *
 * @param [in] workq
 * The work queue to detach.
 *
 * @param [out] data_priv
 * Opaque pointer which will refer to the value output during compress_event_driven_workq_ctx_add in its data_priv out
 * param. The COMPRESS instance MUST release all resources held by this object.
 * (Undo everything done during compress_event_driven_workq_ctx_add).
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_event_driven_workq_ctx_rm(struct doca_ctx *ctx, struct doca_workq *workq,
						       union doca_data data_priv);

/**
 * Logic that is common for detaching polling or event mode work queue from the COMPRESS context.
 *
 * @param [in] compress_ctx
 * The COMPRESS context.
 *
 * @param [in] workq
 * The work queue to detach.
 *
 * @param [out] poller
 * The progress entry that was previously added to workq in compress_workq_ctx_base_add().
 *
 * @param [out] workq_ctx_out
 * Will hold a pointer to the work queue context stored within data_priv.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_workq_ctx_base_rm(struct doca_compress *compress_ctx, struct doca_workq *workq,
					       struct priv_doca_workq_progress_entry *poller,
					       struct workq_context_base **workq_ctx_out);

/**
 * Submit a COMPRESS job
 *
 * @param [in] data_priv
 * Opaque pointer to data created by the engine during compress_workq_ctx_base_add and emitted through its data_priv out
 * param. The COMPRESS instance can reinterpret this as the appropriate object which will be a
 * priv_doca_workq_progress_entry which in turn holds a workq_context_base which can be used to create WQEs and
 * submit to the doca device.
 *
 * @param [in] job
 * Job as passed to: doca_workq_submit. This should have the doca_job::type set to an appropriate value to suggest this
 * is a COMPRESS JOB.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_ctx_job_submit(union doca_data data_priv, struct doca_job const *job);

/**
 * Retrieve a COMPRESS event (response)
 *
 * @param [in] data_priv
 * Opaque pointer to data created by the engine during compress_workq_ctx_base_add and emitted through its data_priv out
 * param. The COMPRESS instance can reinterpret this is the appropriate object which will be a
 * priv_doca_workq_progress_entry which in turn holds a workq_context_base which can be used to receive CQEs to get
 * responses from the device.
 *
 * @param [in] ev
 * Event structure to be populated as appropriate when this function returns a success code.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_workq_progress(union doca_data data_priv, struct doca_event *ev);

/**
 * Arm work queue context for an event. Requesting an event on next CQE received.
 *
 * @details Callback to be used once application requests an event from the work queue while working in event driven
 * mode. Callback will be trigerred once application calls doca_workq_event_handle_arm().
 *
 * @param [in] event_data
 * Opaque pointer to a priv_doca_comp_events_channel belonging to an event workq context. The channel can be used for
 * arming of events. The pointer was previously provided upon registering a doca_event_handle_t to the work queue
 * using priv_doca_workq_event_handle_register().
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_ctx_arm_event(union doca_data event_data);

/**
 * Clear events retrieved for the work queue context
 *
 * @details Callback to be used once application requests clearing of events received on the work queue,
 * using doca_workq_event_handle_clear().
 * The method's goal is to change state of the event handle from 'ready to read' to 'pending'.
 *
 * @param [in] event_data
 * Opaque pointer to a priv_doca_comp_events_channel belonging to an event workq context. The channel can be used for
 * clearing of events. The pointer was previously provided upon registering a doca_event_handle_t to the work queue
 * using priv_doca_workq_event_handle_register().
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t compress_ctx_clear_event(union doca_data event_data);

/**********************************************************************************************************************
 * Public API implementation
 */
doca_error_t
doca_compress_create(struct doca_compress **compress)
{
	struct doca_compress *inst;
	int ret;
	doca_error_t err;

	if (compress == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	inst = priv_doca_malloc(1, sizeof(struct doca_compress));
	if (inst == NULL)
		return DOCA_ERROR_NO_MEMORY;

	inst->mutex = priv_doca_malloc(1, sizeof(pthread_mutex_t));
	if (inst->mutex == NULL) {
		priv_doca_free(inst);
		return DOCA_ERROR_NO_MEMORY;
	}

	ret = pthread_mutex_init(inst->mutex, NULL);
	if (ret != 0) {
		priv_doca_free(inst->mutex);
		priv_doca_free(inst);
		return DOCA_ERROR_INITIALIZATION;
	}

	err = priv_doca_ctx_init(&(inst->base), true);
	if (err != DOCA_SUCCESS) {
		pthread_mutex_destroy(inst->mutex);
		priv_doca_free(inst->mutex);
		priv_doca_free(inst);
		return err;
	}

	init_ops(&(inst->base.ops));

	inst->dev = NULL;
	inst->attached_workq_count = 0;
	inst->started = false;

	*compress = inst;

	return DOCA_SUCCESS;
}

doca_error_t
doca_compress_destroy(struct doca_compress *compress)
{
	pthread_mutex_t *mutex;

	if (compress == NULL)
		return DOCA_SUCCESS;

	if (pthread_mutex_trylock(compress->mutex) != 0)
		return DOCA_ERROR_IN_USE;

	if (compress->attached_workq_count != 0) {
		pthread_mutex_unlock(compress->mutex);
		return DOCA_ERROR_IN_USE;
	}

	/* point of no return - extend life of mutex as long as possible to prevent any thread form doing anything for
	 * as long as possible.
	 */
	mutex = compress->mutex;
	compress->mutex = NULL;

	priv_doca_ctx_destroy(&(compress->base));

	priv_doca_free(compress);

	pthread_mutex_unlock(mutex);
	pthread_mutex_destroy(mutex);
	priv_doca_free(mutex);

	return DOCA_SUCCESS;
}

doca_error_t
doca_compress_job_get_supported(struct doca_devinfo *devinfo, enum doca_compress_job_types job_type)
{
	struct compress_caps caps_compress = {0};

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	const doca_error_t err = compress_caps_get(devinfo, &caps_compress);

	if (err != DOCA_SUCCESS)
		return err;

	switch (job_type) {
	case DOCA_COMPRESS_DEFLATE_JOB:
		if (compress_caps_supported(&caps_compress))
			return DOCA_SUCCESS;
		break;
	case DOCA_DECOMPRESS_DEFLATE_JOB:
		if (decompress_caps_supported(&caps_compress))
			return DOCA_SUCCESS;
		break;
	default:
		return DOCA_ERROR_INVALID_VALUE;
	}
	return DOCA_ERROR_NOT_SUPPORTED;
}

doca_error_t
doca_compress_get_max_buffer_size(const struct doca_devinfo *devinfo, enum doca_compress_job_types job_type,
				  uint32_t *max_buffer_size)
{
	struct compress_caps caps_compress = {0};

	if (devinfo == NULL || max_buffer_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	const doca_error_t err = compress_caps_get(devinfo, &caps_compress);

	if (err != DOCA_SUCCESS)
		return err;

	switch (job_type) {
	case DOCA_COMPRESS_DEFLATE_JOB:
		if (compress_caps_supported(&caps_compress)) {
			*max_buffer_size = caps_compress.device_max_compress_job_len;
			return DOCA_SUCCESS;
		}
		break;
	case DOCA_DECOMPRESS_DEFLATE_JOB:
		if (decompress_caps_supported(&caps_compress)) {
			*max_buffer_size = caps_compress.device_max_decompress_job_len;
			return DOCA_SUCCESS;
		}
		break;
	default:
		DOCA_DLOG_ERR("Unknown doca_compress job type: %d is not supported", job_type);
		return DOCA_ERROR_INVALID_VALUE;
	}
	return DOCA_ERROR_NOT_SUPPORTED;
}

doca_error_t
doca_compress_get_max_list_buf_num_elem(const struct doca_devinfo *devinfo, uint32_t *max_list_num_elem)
{
	struct compress_caps caps_compress = {0};
	doca_error_t ret;

	if (devinfo == NULL || max_list_num_elem == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = compress_caps_get(devinfo, &caps_compress);
	if (ret != DOCA_SUCCESS)
		return ret;

	*max_list_num_elem = caps_compress.umr_enable ? MAX_COMPRESS_BUF_LIST_SIZE : 1;
	return DOCA_SUCCESS;
}

struct doca_ctx *
doca_compress_as_ctx(struct doca_compress *compress_ctx)
{
	if (compress_ctx == NULL)
		return NULL;

	return &compress_ctx->base;
}

/**********************************************************************************************************************
 * Local private function definitions.
 */
static void
init_ops(struct priv_doca_ctx_ops *ops)
{
	ops->start = compress_ctx_start;
	ops->stop = compress_ctx_stop;
	ops->dev_add = compress_ctx_dev_add;
	ops->dev_rm = compress_ctx_dev_rm;
	ops->workq_add = compress_polled_workq_ctx_add;
	ops->event_workq_add = compress_event_driven_workq_ctx_add;
	ops->workq_rm = compress_polled_workq_ctx_rm;
	ops->event_workq_rm = compress_event_driven_workq_ctx_rm;
	ops->job_submit = compress_ctx_job_submit;
}

static doca_error_t
validate_object_and_lock(struct doca_ctx *ctx, struct doca_compress **compress_ctx)
{
	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*compress_ctx = container_of(ctx, struct doca_compress, base);

	if (pthread_mutex_lock((*compress_ctx)->mutex) != 0)
		return DOCA_ERROR_NO_LOCK;

	return DOCA_SUCCESS;
}

static doca_error_t
compress_caps_get(const struct doca_devinfo *info, struct compress_caps *caps)
{
	uint32_t out[DEVX_ST_SZ_DW(query_hca_cap_out)] = {0};
	const void *atts = DEVX_ADDR_OF(query_hca_cap_out, out, capability);
	doca_error_t ret;

	ret = priv_doca_devinfo_hca_caps_by_opmod_query(info,
							MLX5_GET_HCA_CAP_OP_MOD_GENERAL_DEVICE | MLX5_HCA_CAP_OPMOD_GET_CUR,
							out, sizeof(out));
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to query device capabilities. Maybe old FW?");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	/* Pull out compress specific info from return. */
	caps->device_max_compress_job_len = 1U << DEVX_GET(cmd_hca_cap, atts, log_compress_mmo_size);
	caps->device_max_decompress_job_len = 1U << DEVX_GET(cmd_hca_cap, atts, log_decompress_mmo_size);
	caps->compress_qp_enabled = DEVX_GET(cmd_hca_cap, atts, compress_mmo_qp);
	caps->compress_sq_enabled = DEVX_GET(cmd_hca_cap, atts, compress_mmo_sq);
	caps->decompress_qp_enabled = DEVX_GET(cmd_hca_cap, atts, decompress_mmo_qp);
	caps->decompress_sq_enabled = DEVX_GET(cmd_hca_cap, atts, decompress_mmo_sq);
	caps->umr_enable = !(DEVX_GET(cmd_hca_cap, atts, umr_indirect_mkey_disabled));

	/* Confirm RDMA over ethernet. */
	if (DEVX_GET(cmd_hca_cap, atts, roce)) {
		memset(out, 0, sizeof(out));

		ret = priv_doca_devinfo_hca_caps_by_opmod_query(info,
								MLX5_GET_HCA_CAP_OP_MOD_ROCE | MLX5_HCA_CAP_OPMOD_GET_CUR,
								out, sizeof(out));
		if (ret != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to query ROCE device capabilities. Maybe old FW?");
			return DOCA_ERROR_NOT_SUPPORTED;
		}

		caps->device_qp_ts_format = DEVX_GET(roce_caps, atts, qp_ts_format);
	} else
		caps->device_qp_ts_format = 0;

	return DOCA_SUCCESS;
}

static bool
compress_caps_supported(struct compress_caps *caps)
{
	if ((!caps->compress_qp_enabled && !caps->compress_sq_enabled)) {
		DOCA_LOG_ERR("COMPRESS not supported on the device.");
		return false;
	}
	return true;
}

static bool
decompress_caps_supported(struct compress_caps *caps)
{
	if ((!caps->decompress_qp_enabled && !caps->decompress_sq_enabled)) {
		DOCA_LOG_ERR("DECOMPRESS not supported on the device.");
		return false;
	}
	return true;
}

static doca_error_t
create_uar(struct workq_context_base *workq_ctx, struct priv_doca_uar **uar, uint8_t **uar_base_addr)
{
	/* First attempt to allocate to cached memory (will require write barrier after writes) */
	doca_error_t ret = priv_doca_dev_uar_create(workq_ctx->dev, PRIV_DOCA_UAR_ALLOCATION_TYPE_BLUEFLAME, uar);

	if (ret != DOCA_SUCCESS)
		/* If fails, try to allocate to non-cached memory (no barriers required for write). */
		ret = priv_doca_dev_uar_create(workq_ctx->dev, PRIV_DOCA_UAR_ALLOCATION_TYPE_NONCACHE, uar);

	if (ret == DOCA_SUCCESS)
		ret = priv_doca_uar_page_get(*uar, (void **)uar_base_addr);

	return ret;
}

static doca_error_t
create_metadata_mr(struct workq_context_base *workq_ctx, struct ibv_mr **mr)
{
	void *ptr;
	const int metadata_len = MLX5_COMPRESS_METADATA_BUF_LEN;
	const uint16_t qp_size = 1U << LOG_QP_DEPTH;
	const size_t buf_size = metadata_len * qp_size;

	ptr = aligned_alloc(metadata_len, buf_size);
	if (ptr == NULL) {
		DOCA_DLOG_ERR("Failed to allocate metadata buffer");
		return DOCA_ERROR_NO_MEMORY;
	}

	memset(ptr, 0, buf_size);

	*mr = ibv_reg_mr(workq_ctx->pd, ptr, buf_size, IBV_ACCESS_LOCAL_WRITE);
	if (*mr == NULL) {
		DOCA_DLOG_ERR("Failed to register metadata mr. err=%d", errno);
		free(ptr);
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
destroy_metadata_mr(struct workq_context_base *workq_ctx)
{
	void *ptr = workq_ctx->metadata_mr->addr;
	int rc;

	rc = ibv_dereg_mr(workq_ctx->metadata_mr);
	if (rc != 0) {
		DOCA_DLOG_ERR("Failed to deregister metadata mr. err=%d", errno);
		return DOCA_ERROR_INVALID_VALUE;
	}

	free(ptr);

	return DOCA_SUCCESS;
}

static doca_error_t
create_cq(struct workq_context_base *workq_ctx, uint16_t log_cq_size, struct priv_doca_cq **cq)
{
	int ret;
	const struct priv_doca_cq_attr cq_attr = priv_doca_cq_attr_create(1UL << log_cq_size);

	ret = priv_doca_dev_cq_create(workq_ctx->dev, workq_ctx->uar, &cq_attr, cq);
	if (ret != DOCA_SUCCESS)
		DOCA_LOG_ERR("COMPRESS work queue context, failed to create cq. err=%s", doca_get_error_name(ret));

	return ret;
}

static doca_error_t
destroy_cq(struct priv_doca_cq *cq)
{
	doca_error_t ret;

	ret = priv_doca_dev_cq_destroy(cq);
	if (ret != DOCA_SUCCESS)
		DOCA_LOG_ERR("COMPRESS work queue context, failed to destroy cq. err=%s", doca_get_error_name(ret));

	return ret;
}

static doca_error_t
create_qp(struct workq_context_base *workq_ctx, uint16_t log_qp_size, struct compress_qp **qp)
{
	uint32_t pd_num = 0;
	uint32_t cq_num = 0;
	uint32_t uar_id = 0;
	doca_error_t ret;
	uint16_t qp_size = 1U << log_qp_size;

	ret = compress_mlx5_devx_get_pdn(workq_ctx->pd, &pd_num);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = priv_doca_cq_number_get(workq_ctx->compress_cq, &cq_num);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = priv_doca_uar_id_get(workq_ctx->uar, &uar_id);
	if (ret != DOCA_SUCCESS)
		return ret;

	struct compress_mlx5_devx_qp_attr attr = {
		.cqn = cq_num,
		.uar_index = uar_id,
		.ts_format = MLX5_QPC_TIMESTAMP_FORMAT_DEFAULT,
		.pd = pd_num,
		.rq_size = 0,
		.sq_size = qp_size,
		.mmo = 1,
	};

	*qp = malloc(sizeof(**qp));
	if (*qp == NULL) {
		DOCA_DLOG_ERR("Failed to allocate memory for qp. err=%d", errno);
		return DOCA_ERROR_NO_MEMORY;
	}

	if (workq_ctx->umr_support) {
		ret = priv_doca_dev_umr_create(workq_ctx->dev, qp_size >> 2, &(*qp)->umr);
		if (ret != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to create UMR for num_jobs=%u", qp_size >> 2);
			free(*qp);
			*qp = NULL;
			return ret;
		}
	} else
		(*qp)->umr = NULL;

	ret = compress_mlx5_devx_qp_init(workq_ctx->ibv_ctx, *qp, log_qp_size, &attr);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create qp. err=%s", doca_get_error_name(ret));
		if (workq_ctx->umr_support)
			priv_doca_dev_umr_destroy((*qp)->umr);
		free(*qp);
		*qp = NULL;
		return ret;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
destroy_qp(struct compress_qp *qp)
{
	doca_error_t ret;

	if (qp->umr != NULL) {
		ret = priv_doca_dev_umr_destroy(qp->umr);
		if (ret != DOCA_SUCCESS)
			return ret;
	}

	ret = compress_mlx5_devx_qp_destroy(qp);
	if (ret != DOCA_SUCCESS)
		return ret;

	free(qp);

	return DOCA_SUCCESS;
}

static doca_error_t
prepare_qp_for_use(struct compress_qp *qp)
{
	return compress_mlx5_devx_qp2rts(qp, qp->devx_obj->id);
}

static struct polled_workq_context *
create_polled_workq_ctx(struct doca_dev *dev)
{
	struct polled_workq_context *workq_ctx;

	workq_ctx = priv_doca_malloc(1, sizeof(*workq_ctx));
	if (workq_ctx == NULL)
		return NULL;

	if (init_workq_ctx_base(dev, &workq_ctx->base) != DOCA_SUCCESS) {
		priv_doca_free(workq_ctx);
		return NULL;
	}

	return workq_ctx;
}

static struct event_driven_workq_context *
create_event_driven_workq_ctx(struct doca_dev *dev)
{
	struct event_driven_workq_context *workq_ctx;

	workq_ctx = priv_doca_malloc(1, sizeof(*workq_ctx));
	if (workq_ctx == NULL)
		return NULL;

	if (priv_doca_dev_comp_events_channel_create(dev, true, &workq_ctx->channel) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to create completion events channel.");
		priv_doca_free(workq_ctx);
		return NULL;
	}

	if (priv_doca_comp_events_channel_handle_get(workq_ctx->channel, &workq_ctx->channel_handle) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to get completion events channel handle.");
		priv_doca_dev_comp_events_channel_destroy(workq_ctx->channel);
		priv_doca_free(workq_ctx);
		return NULL;
	}

	if (init_workq_ctx_base(dev, &workq_ctx->base) != DOCA_SUCCESS) {
		priv_doca_dev_comp_events_channel_destroy(workq_ctx->channel);
		priv_doca_free(workq_ctx);
		return NULL;
	}

	if (priv_doca_comp_events_channel_register(workq_ctx->channel, workq_ctx->base.compress_cq) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to register CQ with completion events channel.");
		destroy_workq_ctx_base(&workq_ctx->base);
		priv_doca_dev_comp_events_channel_destroy(workq_ctx->channel);
		priv_doca_free(workq_ctx);
		return NULL;
	}

	return workq_ctx;
}

static doca_error_t
init_workq_ctx_base(struct doca_dev *dev, struct workq_context_base *workq_ctx)
{
	doca_error_t ret;
	struct compress_caps caps_compress = {0};

	workq_ctx->dev = dev;
	ret = compress_caps_get(doca_dev_as_devinfo(dev), &caps_compress);
	if (ret != DOCA_SUCCESS)
		return ret;

	workq_ctx->ibv_ctx = priv_doca_dev_to_ibv_ctx(dev);
	if (workq_ctx->ibv_ctx == NULL) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to get device context");
		destroy_workq_ctx_base(workq_ctx);
		return DOCA_ERROR_INITIALIZATION;
	}

	ret = doca_dev_get_pd(dev, &workq_ctx->pd);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to get PD");
		destroy_workq_ctx_base(workq_ctx);
		return ret;
	}

	ret = create_uar(workq_ctx, &workq_ctx->uar, &workq_ctx->uar_base_addr);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to allocate UAR. err=%s", doca_get_error_name(ret));
		destroy_workq_ctx_base(workq_ctx);
		return ret;
	}

	ret = create_metadata_mr(workq_ctx, &workq_ctx->metadata_mr);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to register metadata MR. err=%s",
			     doca_get_error_name(ret));
		destroy_workq_ctx_base(workq_ctx);
		return ret;
	}

	ret = create_cq(workq_ctx, LOG_QP_DEPTH, &workq_ctx->compress_cq);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to create CQ. err=%s", doca_get_error_name(ret));
		destroy_workq_ctx_base(workq_ctx);
		return ret;
	}

	workq_ctx->umr_support = caps_compress.umr_enable;

	ret = create_qp(workq_ctx, LOG_QP_DEPTH, &workq_ctx->compress_qp);
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("COMPRESS work queue context unable to create QP. err=%s", doca_get_error_name(ret));
		destroy_workq_ctx_base(workq_ctx);
		return ret;
	}

	ret = prepare_qp_for_use(workq_ctx->compress_qp);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("COMPRESS work queue context unable move QP to RTS state. err=%s",
			      doca_get_error_name(ret));
		destroy_workq_ctx_base(workq_ctx);
		return ret;
	}

	return DOCA_SUCCESS;
}

static void
destroy_polled_workq_ctx(struct polled_workq_context *workq_ctx)
{
	destroy_workq_ctx_base(&workq_ctx->base);

	priv_doca_free(workq_ctx);
}

static void
destroy_event_driven_workq_ctx(struct event_driven_workq_context *workq_ctx)
{
	doca_error_t ret;

	if (workq_ctx->base.compress_cq != NULL && workq_ctx->channel != NULL) {
		ret = priv_doca_comp_events_channel_unregister(workq_ctx->channel, workq_ctx->base.compress_cq);
		/* In case create did not reach registration phase, not found may be returned.*/
		if (ret != DOCA_SUCCESS && ret != DOCA_ERROR_NOT_FOUND)
			DOCA_LOG_ERR("Failed to unregister COMPRESS workq CQ from completion events channel. err=%s",
				     doca_get_error_name(ret));
	}

	destroy_workq_ctx_base(&workq_ctx->base);

	if (workq_ctx->channel != NULL)
		priv_doca_dev_comp_events_channel_destroy(workq_ctx->channel);

	priv_doca_free(workq_ctx);
}

static void
destroy_workq_ctx_base(struct workq_context_base *workq_ctx)
{
	doca_error_t ret;

	if (workq_ctx->compress_qp != NULL) {
		ret = destroy_qp(workq_ctx->compress_qp);
		if (ret != DOCA_SUCCESS)
			DOCA_LOG_ERR("Failed to destroy COMPRESS workq QP. err=%s", doca_get_error_name(ret));
	}

	if (workq_ctx->compress_cq != NULL) {
		ret = destroy_cq(workq_ctx->compress_cq);
		if (ret != DOCA_SUCCESS)
			DOCA_LOG_ERR("Failed to destroy COMPRESS workq CQ. err=%s", doca_get_error_name(ret));
	}

	if (workq_ctx->metadata_mr != NULL) {
		ret = destroy_metadata_mr(workq_ctx);
		if (ret != DOCA_SUCCESS)
			DOCA_LOG_ERR("Failed to destroy COMPRESS workq metadata MR. err=%s", doca_get_error_name(ret));
	}

	if (workq_ctx->uar != NULL)
		priv_doca_dev_uar_destroy(workq_ctx->uar);
}

static doca_error_t
compress_ctx_start(struct doca_ctx *ctx)
{
	struct doca_compress *compress_ctx;
	doca_error_t result;

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	if (compress_ctx->started)
		result = DOCA_ERROR_BAD_STATE;
	else if (compress_ctx->dev != NULL && compress_ctx->attached_workq_count == 0) {
		compress_ctx->started = true;
		result = DOCA_SUCCESS;
	} else
		result = DOCA_ERROR_INVALID_VALUE;

	pthread_mutex_unlock(compress_ctx->mutex);

	return result;
}

static doca_error_t
compress_ctx_stop(struct doca_ctx *ctx)
{
	struct doca_compress *compress_ctx;
	doca_error_t result;

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	if (!compress_ctx->started)
		result = DOCA_ERROR_BAD_STATE;
	else if (compress_ctx->attached_workq_count == 0) {
		result = DOCA_SUCCESS;
		compress_ctx->started = false;
	} else
		result = DOCA_ERROR_IN_USE;

	pthread_mutex_unlock(compress_ctx->mutex);

	return result;
}

static doca_error_t
compress_ctx_dev_add(struct doca_ctx *ctx, struct doca_dev *dev)
{
	struct doca_compress *compress_ctx;
	struct compress_caps caps_compress = {0};
	doca_error_t result;

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	/* Extract max compress and decompress job len */
	result = compress_caps_get(doca_dev_as_devinfo(dev), &caps_compress);
	if (result != DOCA_SUCCESS)
		return result;
	compress_ctx->device_max_compress_job_len = caps_compress.device_max_compress_job_len;
	compress_ctx->device_max_decompress_job_len = caps_compress.device_max_decompress_job_len;

	// FUTURE: fail if running
	if (compress_ctx->dev != dev) {
		compress_ctx->dev = dev;
		result = DOCA_SUCCESS;
	} else
		result = DOCA_ERROR_INVALID_VALUE;

	pthread_mutex_unlock(compress_ctx->mutex);

	return result;
}

static doca_error_t
compress_ctx_dev_rm(struct doca_ctx *ctx, struct doca_dev *dev)
{
	struct doca_compress *compress_ctx;
	doca_error_t result;

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	// FUTURE: fail if running
	if (compress_ctx->dev == dev) {
		compress_ctx->dev = NULL;
		result = DOCA_SUCCESS;
	} else
		result = DOCA_ERROR_INVALID_VALUE;

	pthread_mutex_unlock(compress_ctx->mutex);

	return result;
}

static doca_error_t
compress_workq_ctx_base_add(struct doca_compress *compress_ctx, struct doca_workq *workq,
			    struct workq_context_base *workq_ctx, struct priv_doca_workq_progress_entry **poller_out)
{
	struct priv_doca_workq_progress_entry *poller;
	doca_error_t result;

	poller = priv_doca_malloc(1, sizeof(struct priv_doca_workq_progress_entry));
	if (poller == NULL)
		return DOCA_ERROR_NO_MEMORY;

	poller->progress_cb = compress_workq_progress;

	result = priv_doca_workq_progress_entry_add(workq, poller);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to register progress entry");
		priv_doca_free(poller);
		return DOCA_ERROR_INITIALIZATION;
	}

	poller->data_priv.ptr = (void *)workq_ctx;
	++(compress_ctx->attached_workq_count);
	workq_ctx->registered_with_compress_ctx = true;
	workq_ctx->device_max_compress_job_len = compress_ctx->device_max_compress_job_len;
	workq_ctx->device_max_decompress_job_len = compress_ctx->device_max_decompress_job_len;

	*poller_out = poller;

	return result;
}

static doca_error_t
compress_ctx_arm_event(union doca_data event_data)
{
	doca_error_t ret = DOCA_SUCCESS;
	struct event_driven_workq_context *workq_ctx = (struct event_driven_workq_context *)event_data.ptr;

	priv_doca_comp_events_channel_arm(workq_ctx->channel);
	if (priv_doca_cq_empty(workq_ctx->base.compress_cq))
		/* All CQEs that generated an event, have been drained. */
		ret = priv_doca_workq_progress_entry_state_update(workq_ctx->workq, workq_ctx->poller,
								  PRIV_DOCA_WORKQ_PROGRESS_SUSPENDED);

	return ret;
}

static doca_error_t
compress_ctx_clear_event(union doca_data event_data)
{
	doca_error_t ret;
	struct event_driven_workq_context *workq_ctx = (struct event_driven_workq_context *)event_data.ptr;

	ret = priv_doca_workq_progress_entry_state_update(workq_ctx->workq, workq_ctx->poller,
							  PRIV_DOCA_WORKQ_PROGRESS_RUNNING);
	if (unlikely(ret != DOCA_SUCCESS))
		return ret;

	return priv_doca_comp_events_channel_handle_events(workq_ctx->channel, 0 /*ignored*/);
}

static doca_error_t
compress_event_driven_workq_ctx_add(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data *data_priv)
{
	struct event_driven_workq_context *workq_ctx;
	struct doca_compress *compress_ctx;
	struct priv_doca_workq_progress_entry *poller = NULL;
	union doca_data event_data;
	doca_error_t result;

	static const struct priv_doca_event_handle_ops event_ops = {
		.arm_cb = compress_ctx_arm_event,
		.clear_cb = compress_ctx_clear_event,
	};

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	workq_ctx = create_event_driven_workq_ctx(compress_ctx->dev);
	if (workq_ctx == NULL) {
		pthread_mutex_unlock(compress_ctx->mutex);
		DOCA_LOG_ERR("Unable to create COMPRESS work queue context");
		return DOCA_ERROR_NO_MEMORY;
	}

	event_data.ptr = workq_ctx;
	result = priv_doca_workq_event_handle_register(workq, workq_ctx->channel_handle, &event_ops, event_data);
	if (result != DOCA_SUCCESS) {
		destroy_event_driven_workq_ctx(workq_ctx);
		pthread_mutex_unlock(compress_ctx->mutex);
		DOCA_LOG_ERR("Unable to register COMPRESS event handle to work queue.");
		return result;
	}

	result = compress_workq_ctx_base_add(compress_ctx, workq, &workq_ctx->base, &poller);
	if (result != DOCA_SUCCESS) {
		if (priv_doca_workq_event_handle_unregister(workq, workq_ctx->channel_handle) != DOCA_SUCCESS)
			DOCA_DLOG_ERR("Unable to unregister COMPRESS workq ctx event handle from workq.");
		destroy_event_driven_workq_ctx(workq_ctx);
		pthread_mutex_unlock(compress_ctx->mutex);
		return result;
	}

	data_priv->ptr = (void *)poller;
	/* Store workq & poller to be able to change state according to received event. */
	workq_ctx->poller = poller;
	workq_ctx->workq = workq;

	pthread_mutex_unlock(compress_ctx->mutex);

	return DOCA_SUCCESS;
}

static doca_error_t
compress_polled_workq_ctx_add(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data *data_priv)
{
	struct polled_workq_context *workq_ctx;
	struct doca_compress *compress_ctx;
	struct priv_doca_workq_progress_entry *poller = NULL;
	doca_error_t result;

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	workq_ctx = create_polled_workq_ctx(compress_ctx->dev);
	if (workq_ctx == NULL) {
		pthread_mutex_unlock(compress_ctx->mutex);
		DOCA_LOG_ERR("Unable to create COMPRESS work queue context");
		return DOCA_ERROR_NO_MEMORY;
	}

	result = compress_workq_ctx_base_add(compress_ctx, workq, &workq_ctx->base, &poller);
	if (result != DOCA_SUCCESS)
		destroy_polled_workq_ctx(workq_ctx);

	data_priv->ptr = (void *)poller;

	pthread_mutex_unlock(compress_ctx->mutex);

	return result;
}

static doca_error_t
compress_workq_ctx_base_rm(struct doca_compress *compress_ctx, struct doca_workq *workq,
			   struct priv_doca_workq_progress_entry *poller, struct workq_context_base **workq_ctx_out)
{
	struct workq_context_base *workq_ctx;
	doca_error_t result;

	workq_ctx = poller->data_priv.ptr;
	if (workq_ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	/* Unregister with the compress ctx */
	if (workq_ctx->registered_with_compress_ctx) {
		--(compress_ctx->attached_workq_count);
		workq_ctx->registered_with_compress_ctx = false;
	}

	/* Remove the progress entry - this will unregister the progress_cb.*/
	result = priv_doca_workq_progress_entry_rm(workq, poller);
	if (result != DOCA_SUCCESS)
		return result;

	/* Destroy the poller */
	priv_doca_free(poller);
	*workq_ctx_out = workq_ctx;

	return DOCA_SUCCESS;
}

static doca_error_t
compress_polled_workq_ctx_rm(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data data_priv)
{
	struct doca_compress *compress_ctx;
	struct workq_context_base *workq_ctx_base;
	struct polled_workq_context *workq_ctx;
	struct priv_doca_workq_progress_entry *poller;
	doca_error_t result;

	if (data_priv.ptr == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	/* Grab objects from data_priv */
	poller = (struct priv_doca_workq_progress_entry *)data_priv.ptr;

	result = compress_workq_ctx_base_rm(compress_ctx, workq, poller, &workq_ctx_base);
	if (result != DOCA_SUCCESS) {
		pthread_mutex_unlock(compress_ctx->mutex);
		return result;
	}
	workq_ctx = container_of(workq_ctx_base, struct polled_workq_context, base);

	destroy_polled_workq_ctx(workq_ctx);
	pthread_mutex_unlock(compress_ctx->mutex);

	return DOCA_SUCCESS;
}

static doca_error_t
compress_event_driven_workq_ctx_rm(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data data_priv)
{
	struct doca_compress *compress_ctx;
	struct workq_context_base *workq_ctx_base;
	struct event_driven_workq_context *workq_ctx;
	struct priv_doca_workq_progress_entry *poller;
	doca_error_t result;

	if (data_priv.ptr == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	result = validate_object_and_lock(ctx, &compress_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	poller = (struct priv_doca_workq_progress_entry *)data_priv.ptr;

	result = compress_workq_ctx_base_rm(compress_ctx, workq, poller, &workq_ctx_base);
	if (result != DOCA_SUCCESS) {
		pthread_mutex_unlock(compress_ctx->mutex);
		return result;
	}
	workq_ctx = container_of(workq_ctx_base, struct event_driven_workq_context, base);

	result = priv_doca_workq_event_handle_unregister(workq, workq_ctx->channel_handle);
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to unregister COMPRESS workq ctx event handle from workq.");
		return result;
	}

	destroy_event_driven_workq_ctx(workq_ctx);
	pthread_mutex_unlock(compress_ctx->mutex);

	return DOCA_SUCCESS;
}

static doca_error_t
compress_ctx_job_submit(union doca_data data_priv, struct doca_job const *job)
{
	struct doca_compress_job const *compress_job;
	struct workq_context_base *workq_ctx;
	struct priv_doca_workq_progress_entry *poller;
	uint32_t dest_lkey;
	uint32_t src_lkey;
	int true_pi;
	uint8_t opmod;
	uint32_t max_job_len;
	size_t dst_buf_len;
	uint32_t num_elements_src;
	uint32_t num_elements_dest;
	unsigned int wqe_to_post;

	compress_job = container_of(job, struct doca_compress_job const, base);
	poller = (struct priv_doca_workq_progress_entry *)data_priv.ptr;
	workq_ctx = (struct workq_context_base *)poller->data_priv.ptr;
	wqe_to_post = workq_ctx->umr_support ? 4 : 1; /* if UMR is supported then we add 3 wqes to every compress job */

	doca_buf_list_num_elements(compress_job->src_buff, &num_elements_src);
	doca_buf_list_num_elements(compress_job->dst_buff, &num_elements_dest);

	if (unlikely(num_elements_dest > 1)) {
		DOCA_LOG_ERR("linked list is prohibited for destination buffer");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	switch (job->type) {
	case DOCA_COMPRESS_DEFLATE_JOB:
		opmod = MLX5_OP_MOD_MMO_COMP;
		max_job_len = workq_ctx->device_max_compress_job_len;
		break;
	case DOCA_DECOMPRESS_DEFLATE_JOB:
		opmod = MLX5_OP_MOD_MMO_DECOMP;
		max_job_len = workq_ctx->device_max_decompress_job_len;
		break;
	default:
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (unlikely(priv_doca_buf_get_data_len(compress_job->src_buff) > max_job_len))
		return DOCA_ERROR_INVALID_VALUE;

	if (unlikely(compress_mlx5_devx_is_qp_overflow(workq_ctx->compress_qp, wqe_to_post - 1))) {
		DOCA_DLOG_DBG("Info: QP is full, can't add another wqe, pi = %d and ci = %d",
			      workq_ctx->compress_qp->pi, workq_ctx->compress_qp->ci);
		return DOCA_ERROR_NO_MEMORY;
	}

	if (unlikely(doca_buf_get_lkey(compress_job->dst_buff, workq_ctx->dev,
				       DOCA_ACCESS_LOCAL_READ | DOCA_ACCESS_LOCAL_WRITE, &dest_lkey) != DOCA_SUCCESS))
		return DOCA_ERROR_INVALID_VALUE;
	if (unlikely(doca_buf_get_lkey(compress_job->src_buff, workq_ctx->dev,
				       DOCA_ACCESS_LOCAL_READ, &src_lkey) != DOCA_SUCCESS))
		return DOCA_ERROR_INVALID_VALUE;

	dst_buf_len = priv_doca_buf_get_len(compress_job->dst_buff) -
		      (priv_doca_buf_get_data(compress_job->dst_buff) - priv_doca_buf_get_addr(compress_job->dst_buff));

	if (workq_ctx->umr_support) {
		true_pi = compress_mlx5_devx_post_umr_wqe(workq_ctx->compress_qp, workq_ctx->metadata_mr,
							  workq_ctx->uar_base_addr, dest_lkey,
							  (uint64_t)priv_doca_buf_get_data(compress_job->dst_buff),
							  (struct doca_buf *)compress_job->src_buff,
							  dst_buf_len, opmod);
		if (unlikely(true_pi < 0))
			return DOCA_ERROR_IO_FAILED;
	} else {
		if (unlikely(num_elements_src + num_elements_dest > 2)) {
			DOCA_LOG_ERR("Device added to compress context does not support linked list, consider adding a different device.");
			return DOCA_ERROR_NOT_SUPPORTED;
		}
		true_pi = compress_mlx5_devx_post_one(workq_ctx->compress_qp, workq_ctx->metadata_mr,
						      workq_ctx->uar_base_addr, dest_lkey,
						      (uint64_t)priv_doca_buf_get_data(compress_job->dst_buff),
						      src_lkey, (uint64_t)priv_doca_buf_get_data(compress_job->src_buff),
						      priv_doca_buf_get_data_len(compress_job->src_buff),
						      dst_buf_len, job->user_data.u64, opmod);
	}


	/* Save the wr-id (so CQE will use it later) */
	workq_ctx->compress_qp->wr_id_arr[true_pi].user_data = job->user_data.u64;
	workq_ctx->compress_qp->wr_id_arr[true_pi].dst_buf = compress_job->dst_buff;
	workq_ctx->compress_qp->wr_id_arr[true_pi].type = job->type;
	workq_ctx->compress_qp->wr_id_arr[true_pi].output_chksum = compress_job->output_chksum;

	return DOCA_SUCCESS;
}

static doca_error_t
compress_workq_progress(union doca_data data_priv, struct doca_event *ev)
{
	struct workq_context_base *workq_ctx;
	bool is_err;
	struct compress_wc compress_wc;
	struct priv_doca_64b_cqe cqe;
	void *data;

	workq_ctx = (struct workq_context_base *)data_priv.ptr;

	/* Result not ready yet. Ask user to poll again.*/
	if (likely(priv_doca_cq_poll_one(workq_ctx->compress_cq, &cqe) == DOCA_ERROR_AGAIN))
		return DOCA_ERROR_AGAIN;

	is_err = !compress_mlx5_devx_process_cqe((struct mlx5_cqe64 *)&cqe, workq_ctx->compress_qp, &compress_wc);

	ev->type = compress_wc.wr_id->type;

	/* num_polled == 1 */
	ev->user_data.u64 = compress_wc.wr_id->user_data;

	/* return umr key */
	if (workq_ctx->compress_qp->umr != NULL)
		priv_doca_umr_mkey_free(workq_ctx->compress_qp->umr, compress_wc.wr_id->umr_handle);

	/* Completion with error.*/
	if (unlikely(is_err)) {
		DOCA_LOG_ERR("CQ received for failed job: status=%d, vendor_error=%d",
			     compress_wc.status, compress_wc.vendor_err);
		ev->result.u64 = DOCA_ERROR_UNKNOWN;
		return DOCA_ERROR_IO_FAILED;
	}

	/* Set dst_buf data */
	doca_buf_get_data(compress_wc.wr_id->dst_buf, &data);
	doca_buf_set_data(compress_wc.wr_id->dst_buf, data, compress_wc.byte_cnt);

	/* Set output checksum */
	if (compress_wc.wr_id->output_chksum) {
		struct compress_metadata *metadata = ((struct compress_metadata *)(workq_ctx->metadata_mr->addr +
						      (compress_wc.wr_id_idx * MLX5_COMPRESS_METADATA_BUF_LEN)));

		*(compress_wc.wr_id->output_chksum) = ((uint64_t)be32toh(metadata->crc_32) |
						       ((uint64_t)be32toh(metadata->adler_32) << 32));
	}

	ev->result.u64 = DOCA_SUCCESS;

	return DOCA_SUCCESS;
}
