/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <infiniband/mlx5dv.h>
#include <infiniband/verbs.h>

#include <doca_log.h>

#include "impl/compress_mlx5_common.h"
#include "impl/compress_mlx5_prm.h"

DOCA_LOG_REGISTER(COMPRESS_MLX5_COMMON);

static inline size_t
get_mem_page_size(void)
{
	static size_t page_size;

	if (page_size == 0)
		page_size = sysconf(_SC_PAGESIZE);

	return page_size;
}

static inline uint32_t
get_umem_id(const void *umem)
{
	if (umem == NULL)
		return 0;

	return ((const struct mlx5dv_devx_umem *)umem)->umem_id;
}

static inline uint32_t
log2_u32(uint32_t x)
{
	if (x == 0)
		/* log(0) is undefined */
		return 0;

	return 31u - __builtin_clz(x);
}

static struct compress_devx_obj *
compress_mlx5_devx_cmd_create_qp(void *ctx, const struct compress_mlx5_devx_qp_attr *attr)
{
	uint32_t in[DEVX_ST_SZ_DW(create_qp_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(create_qp_out)] = {0};
	void *qpc = DEVX_ADDR_OF(create_qp_in, in, qpc);
	struct compress_devx_obj *qp_obj = calloc(1, sizeof(*qp_obj));

	if (qp_obj == NULL) {
		DOCA_DLOG_ERR("Failed to allocate QP data.");
		return NULL;
	}
	DEVX_SET(create_qp_in, in, opcode, MLX5_CMD_OP_CREATE_QP);
	DEVX_SET(qpc, qpc, st, MLX5_QP_ST_RC);
	DEVX_SET(qpc, qpc, pd, attr->pd);
	DEVX_SET(qpc, qpc, ts_format, attr->ts_format);
	DEVX_SET(qpc, qpc, user_index, attr->user_index);
	if (attr->uar_index) {
		if (attr->mmo) {
			void *qpc_ext_and_pas_list =
				DEVX_ADDR_OF(create_qp_in, in, qpc_extension_and_pas_list);
			void *qpc_ext = DEVX_ADDR_OF(qpc_extension_and_pas_list,
				qpc_ext_and_pas_list, qpc_data_extension);

			DEVX_SET(create_qp_in, in, qpc_ext, 1);
			DEVX_SET(qpc_extension, qpc_ext, mmo, 1);
		}
		DEVX_SET(qpc, qpc, pm_state, MLX5_QP_PM_MIGRATED);
		DEVX_SET(qpc, qpc, uar_page, attr->uar_index);
		if (attr->log_page_size > MLX5_ADAPTER_PAGE_SHIFT)
			DEVX_SET(qpc, qpc, log_page_size,
				 attr->log_page_size - MLX5_ADAPTER_PAGE_SHIFT);
		if (attr->sq_size) {
			DEVX_SET(qpc, qpc, cqn_snd, attr->cqn);
			DEVX_SET(qpc, qpc, log_sq_size, log2_u32(attr->sq_size));
		} else
			DEVX_SET(qpc, qpc, no_sq, 1);

		if (attr->rq_size) {
			DEVX_SET(qpc, qpc, cqn_rcv, attr->cqn);
			DEVX_SET(qpc, qpc, log_rq_stride,
				 attr->log_rq_stride - MLX5_LOG_RQ_STRIDE_SHIFT);
			DEVX_SET(qpc, qpc, log_rq_size, log2_u32(attr->rq_size));
			DEVX_SET(qpc, qpc, rq_type, MLX5_NON_ZERO_RQ);
		} else
			DEVX_SET(qpc, qpc, rq_type, MLX5_ZERO_LEN_RQ);

		if (attr->dbr_umem_valid) {
			DEVX_SET(qpc, qpc, dbr_umem_valid, attr->dbr_umem_valid);
			DEVX_SET(qpc, qpc, dbr_umem_id, attr->dbr_umem_id);
		}
		DEVX_SET64(qpc, qpc, dbr_addr, attr->dbr_address);
		DEVX_SET64(create_qp_in, in, wq_umem_offset, attr->wq_umem_offset);
		DEVX_SET(create_qp_in, in, wq_umem_id, attr->wq_umem_id);
		DEVX_SET(create_qp_in, in, wq_umem_valid, 1);
	} else {
		/* Special QP to be managed by FW - no SQ\RQ\CQ\UAR\DB rec. */
		DEVX_SET(qpc, qpc, rq_type, MLX5_ZERO_LEN_RQ);
		DEVX_SET(qpc, qpc, no_sq, 1);
	}

	qp_obj->obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (qp_obj->obj == NULL) {
		DOCA_DLOG_ERR("Failed to create QP Obj using DevX.");
		free(qp_obj);
		return NULL;
	}

	qp_obj->id = DEVX_GET(create_qp_out, out, qpn);

	DOCA_DLOG_INFO("QP number 0x%x was created", qp_obj->id);

	return qp_obj;
}

doca_error_t
compress_mlx5_devx_qp_init(void *ctx, struct compress_qp *new_qp, uint16_t log_wqbb_n,
		      struct compress_mlx5_devx_qp_attr *attr)
{
	struct compress_devx_obj *qp = NULL;
	struct mlx5dv_devx_umem *umem_obj = NULL;
	void *umem_buf = NULL;
	void *wr_id_arr = NULL;
	size_t page_size = get_mem_page_size();
	size_t alignment = page_size;
	uint32_t umem_size, umem_dbrec;
	uint16_t qp_size = 1 << log_wqbb_n;
	doca_error_t ret;

	if (alignment == (size_t)-1) {
		DOCA_DLOG_ERR("Failed to get WQE buf alignment.");
		return DOCA_ERROR_NO_MEMORY;
	}

	wr_id_arr = malloc(sizeof(struct compress_wr_id) * qp_size);
	if (wr_id_arr == NULL) {
		DOCA_DLOG_ERR("Failed to allocate wr-id memory for QP.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}

	/* Allocate memory buffer for WQEs and doorbell record. */
	umem_size = sizeof(struct mlx5_mmo_wqe) * qp_size;
	umem_dbrec = COMPRESS_MLX5_ALIGN(umem_size, COMPRESS_MLX5_DBR_SIZE);
	umem_size += COMPRESS_MLX5_DBR_SIZE;
	umem_size += alignment - (umem_size % alignment);

	umem_buf = aligned_alloc(alignment, umem_size);
	if (umem_buf == NULL) {
		DOCA_DLOG_ERR("Failed to allocate memory for QP work queues.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}
	memset(umem_buf, 0, umem_size);

	/* Register allocated buffer in user space with DevX. */
	umem_obj = mlx5dv_devx_umem_reg(ctx, (void *)(uintptr_t)umem_buf, umem_size, 1);
	if (umem_obj == NULL) {
		DOCA_DLOG_ERR("Failed to register umem for QP.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}

	/* Fill attributes for SQ object creation. */
	attr->wq_umem_id = get_umem_id(umem_obj);
	attr->wq_umem_offset = 0;
	attr->dbr_umem_valid = 1;
	attr->dbr_umem_id = attr->wq_umem_id;
	attr->dbr_address = umem_dbrec;
	attr->log_page_size = log2_u32(page_size);
	/* Create send queue object with DevX. */
	qp = compress_mlx5_devx_cmd_create_qp(ctx, attr);
	if (qp == NULL) {
		DOCA_DLOG_ERR("Can't create DevX QP object.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}

	new_qp->qp_size = qp_size;
	new_qp->ci = 0;
	new_qp->pi = 0;
	new_qp->wr_id_arr = wr_id_arr;
	new_qp->umem_buf = umem_buf;
	new_qp->umem_obj = umem_obj;
	new_qp->devx_obj = qp;
	new_qp->db_rec = DOCA_PTR_ADD(new_qp->umem_buf, umem_dbrec);

	return DOCA_SUCCESS;
error:
	if (umem_obj != NULL)
		mlx5dv_devx_umem_dereg(umem_obj);
	if (umem_buf != NULL)
		free(umem_buf);
	if (wr_id_arr != NULL)
		free(wr_id_arr);

	return ret;
}

doca_error_t
compress_mlx5_devx_qp_destroy(struct compress_qp *qp)
{
	int ret;

	if (qp->devx_obj != NULL) {
		ret = mlx5dv_devx_obj_destroy(qp->devx_obj->obj);
		if (ret != 0) {
			DOCA_DLOG_ERR("Failed to destroy QP DevX object. ret=%d", ret);
			return DOCA_ERROR_INVALID_VALUE;
		}
		free(qp->devx_obj);
	}
	if (qp->umem_obj != NULL) {
		ret = mlx5dv_devx_umem_dereg(qp->umem_obj);
		if (ret != 0) {
			DOCA_DLOG_ERR("Failed to destroy QP umem object. ret=%d", ret);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}
	if (qp->umem_buf != NULL)
		free(qp->umem_buf);
	if (qp->wr_id_arr != NULL)
		free(qp->wr_id_arr);

	return DOCA_SUCCESS;
}

static doca_error_t
compress_mlx5_devx_cmd_modify_qp_rst2init(struct compress_devx_obj *qp)
{
	uint32_t in[DEVX_ST_SZ_DW(rst2init_qp_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(rst2init_qp_out)] = {0};
	void *qpc = DEVX_ADDR_OF(rst2init_qp_in, &in, qpc);
	int ret;

	DEVX_SET(rst2init_qp_in, &in, opcode, MLX5_CMD_OP_RST2INIT_QP);
	DEVX_SET(rst2init_qp_in, &in, qpn, qp->id);

	DEVX_SET(qpc, qpc, primary_address_path.vhca_port_num, 1);
	DEVX_SET(qpc, qpc, rre, 0);
	DEVX_SET(qpc, qpc, rwe, 0);
	DEVX_SET(qpc, qpc, pm_state, MLX5_QP_PM_MIGRATED);

	ret = mlx5dv_devx_obj_modify(qp->obj, &in, sizeof(in), &out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to modify QP rst2init using DevX. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
compress_mlx5_devx_cmd_modify_qp_init2rtr(struct compress_devx_obj *qp, uint32_t remote_qp_id)
{
	uint32_t in[DEVX_ST_SZ_DW(init2rtr_qp_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(init2rtr_qp_out)] = {0};
	void *qpc = DEVX_ADDR_OF(init2rtr_qp_in, &in, qpc);
	int ret;

	DEVX_SET(init2rtr_qp_in, &in, opcode, MLX5_CMD_OP_INIT2RTR_QP);
	DEVX_SET(init2rtr_qp_in, &in, qpn, qp->id);

	DEVX_SET(qpc, qpc, primary_address_path.fl, 1);
	DEVX_SET(qpc, qpc, primary_address_path.vhca_port_num, 1);
	DEVX_SET(qpc, qpc, mtu, 3);
	DEVX_SET(qpc, qpc, log_msg_max, 30);
	DEVX_SET(qpc, qpc, remote_qpn, remote_qp_id);
	DEVX_SET(qpc, qpc, min_rnr_nak, 0xc);
	DEVX_SET(qpc, qpc, next_rcv_psn, 0);

	ret = mlx5dv_devx_obj_modify(qp->obj, &in, sizeof(in), &out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to modify QP init2rtr using DevX. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
compress_mlx5_devx_cmd_modify_qp_rtr2rts(struct compress_devx_obj *qp)
{
	uint32_t in[DEVX_ST_SZ_DW(rtr2rts_qp_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(rtr2rts_qp_out)] = {0};
	void *qpc = DEVX_ADDR_OF(rtr2rts_qp_in, &in, qpc);
	int ret;

	DEVX_SET(rtr2rts_qp_in, &in, opcode, MLX5_CMD_OP_RTR2RTS_QP);
	DEVX_SET(rtr2rts_qp_in, &in, qpn, qp->id);

	DEVX_SET(qpc, qpc, next_send_psn, 0);
	DEVX_SET(qpc, qpc, primary_address_path.ack_timeout, 14);
	DEVX_SET(qpc, qpc, log_ack_req_freq, 0);
	DEVX_SET(qpc, qpc, retry_count, 0);
	DEVX_SET(qpc, qpc, rnr_retry, 3);

	ret = mlx5dv_devx_obj_modify(qp->obj, &in, sizeof(in), &out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to modify QP rtr2rts using DevX. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

doca_error_t
compress_mlx5_devx_qp2rts(struct compress_qp *qp, uint32_t remote_qp_id)
{
	doca_error_t ret;

	ret = compress_mlx5_devx_cmd_modify_qp_rst2init(qp->devx_obj);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = compress_mlx5_devx_cmd_modify_qp_init2rtr(qp->devx_obj, remote_qp_id);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = compress_mlx5_devx_cmd_modify_qp_rtr2rts(qp->devx_obj);
	if (ret != DOCA_SUCCESS)
		return ret;

	return DOCA_SUCCESS;
}

doca_error_t
compress_mlx5_devx_get_pdn(void *pd, uint32_t *pdn)
{
	struct mlx5dv_obj obj = {0};
	struct mlx5dv_pd pd_info = {0};
	int ret;

	obj.pd.in = pd;
	obj.pd.out = &pd_info;

	ret = mlx5dv_init_obj(&obj, MLX5DV_OBJ_PD);
	if (ret != 0) {
		DOCA_DLOG_ERR("Failed to get PD number. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	*pdn = pd_info.pdn;

	return DOCA_SUCCESS;
}
