/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef COMPRESS_MLX5_COMMON_H_
#define COMPRESS_MLX5_COMMON_H_

#include <doca_build_config.h>

#include <stddef.h>
#include <stdint.h>


#ifdef DOCA_ARCH_X86
#include <immintrin.h>
#endif

#include <infiniband/verbs.h>

#include <private/common_defines.h>
#include <doca_buf.h>
#include <priv_doca_umr.h>
#include <priv_doca_os_abstraction_memory.h>

#define DOCA_PTR_ADD(ptr, x) ((void *)((uintptr_t)(ptr) + (x)))

#define COMPRESS_MLX5_ALIGN(v, mul) \
        ((((v) + (typeof(v))(mul) - 1) / ((typeof(v))(mul))) * (typeof(v))(mul))

/* NOTE: hardcoded for now - 64 is fine for BF ARM and x86. */
#define CACHELINE_SIZE 64

#define COMPRESS_MLX5_DBR_SIZE CACHELINE_SIZE

/* Just keep enough space for the metadata of a single WQE.
 * Since we don't need the info that the device provides in this region,
 * all the WQEs will point to the same metadata region.
 */
#define MLX5_COMPRESS_METADATA_BUF_LEN 64

#define COMPRESS_DEFAULT_WINDOW_SIZE 6u
#define COMPRESS_DEFAULT_BLOCK_SIZE 15u
#define COMPRESS_DEFAULT_DYNAMIC_SIZE 0u

#define COMPRESS_WINDOW_SIZE_OFFSET 12u
#define COMPRESS_BLOCK_SIZE_OFFSET 16u
#define COMPRESS_DYNAMIC_SIZE_OFFSET 20u

enum {
	MLX5_CQ_SET_CI = 0,
	MLX5_CQ_ARM_DB = 1,
};

enum {
	MLX5_OP_MOD_MMO_COMP = 0x2,
	MLX5_OP_MOD_MMO_DECOMP = 0x3,
};

struct compress_devx_obj {
	void *obj;	/* The DevX object. */
	int id;		/* The object ID. */
};

/* QP attributes structure, used by QP operations. */
struct compress_mlx5_devx_qp_attr {
	uint32_t pd : 24;
	uint32_t uar_index : 24;
	uint32_t cqn : 24;
	uint32_t log_page_size : 5;
	uint32_t rq_size : 17; /* Must be power of 2. */
	uint32_t log_rq_stride : 3;
	uint32_t sq_size : 17; /* Must be power of 2. */
	uint32_t ts_format : 2;
	uint32_t dbr_umem_valid : 1;
	uint32_t dbr_umem_id;
	uint64_t dbr_address;
	uint32_t wq_umem_id;
	uint64_t wq_umem_offset;
	uint32_t user_index : 24;
	uint32_t mmo : 1;
};

/**
 * COMPRESS WorkQ IDs.
 */
struct compress_wr_id {
	uint64_t user_data;
	struct doca_buf *dst_buf;
	int type;
	uint64_t *output_chksum;
	struct priv_doca_umr_handle *umr_handle;
};

/**
 * COMPRESS engine queue pair.
 */
struct compress_qp {
	struct compress_wr_id *wr_id_arr;	/* Array for holding wr_ids */
	struct compress_devx_obj *devx_obj;	/* The QP DevX object. */
	void *umem_obj;				/* The QP umem object. */
	union {
		void *umem_buf;
		struct mlx5_mmo_wqe *wqes;	/* The QP ring buffer. */
	};
	uint32_t pi;				/* Producer index. */
	uint32_t ci;				/* Consumer index. */
	uint16_t qp_size;			/* Number of entries in this QP (value is power of two). */
	uint32_t *db_rec;			/* The QP doorbell record. */
	struct priv_doca_umr *umr;
};

struct mlx5_mmo_metadata_seg {
	uint32_t mmo_control_31_0;
	uint32_t local_key;
	uint64_t local_address;
};

struct mlx5_mmo_wqe {
	struct mlx5_wqe_ctrl_seg ctrl;
	struct mlx5_mmo_metadata_seg mmo_meta;
	struct mlx5_wqe_data_seg src;
	struct mlx5_wqe_data_seg dest;
};

struct compress_mlx5_wqebb {
	uint8_t reserved[64];
};

struct compress_mlx5_umr_wqe {
	/* 1st WQEBB */
	struct compress_mlx5_wqebb umr_segment[3];
	/* 4th WQEBB */
	struct mlx5_mmo_wqe mmo_seg;
};

/**
 * COMPRESS engine completion: only the needed values of the CQE.
 */
struct compress_wc {
	struct compress_wr_id *wr_id;
	uint16_t wr_id_idx;
	uint32_t status;
	uint32_t vendor_err;
	uint32_t byte_cnt;
};

struct compress_metadata {
	uint32_t reserved_at_0[8];
	uint32_t crc_32;
	uint32_t adler_32;
	uint32_t reserved_at_28[6];
};

/* return true - if good CQE was found,
 *        false - if CQE with error was found
 */
static inline bool
compress_mlx5_devx_process_cqe(struct mlx5_cqe64 *cqe, struct compress_qp *qp, struct compress_wc *compress_wc)
{
	const uint8_t op_code = mlx5dv_get_cqe_opcode(cqe);
	bool ret;
	uint16_t wqe_counter;
	const bool has_umr = qp->umr != NULL;

	/* CQE with error */
	if (unlikely(op_code == MLX5_CQE_REQ_ERR || op_code == MLX5_CQE_RESP_ERR)) {
		const struct mlx5_err_cqe *err_cqe = (const struct mlx5_err_cqe *)cqe;

		compress_wc->status = err_cqe->syndrome;
		compress_wc->vendor_err = err_cqe->vendor_err_synd;
		ret = false;
	} else { /* Good CQE */
		compress_wc->status = 0;
		ret = true;
	}

	/* Get the wr_id of the job */
	wqe_counter = be16toh(cqe->wqe_counter);

	/* Divide wqe_counter by 4 to get the right job ID */
	if (has_umr)
		compress_wc->wr_id_idx = ((wqe_counter) & (qp->qp_size - 1)) >> 2;
	else
		compress_wc->wr_id_idx = wqe_counter & (qp->qp_size - 1);

	compress_wc->wr_id = &(qp->wr_id_arr[compress_wc->wr_id_idx]);
	compress_wc->byte_cnt = be32toh(cqe->byte_cnt);

	/* Update the CI according to the wqe-index (to consider unsignaled WQEs) */
	qp->ci += (uint16_t)(wqe_counter - (uint16_t)qp->ci + 1);

	return ret;
}

static inline int
compress_mlx5_devx_is_qp_overflow(const struct compress_qp *qp, int nreq)
{
	unsigned cur = qp->pi - qp->ci;

	if (likely(cur + nreq < qp->qp_size))
		return 0;

	return cur + nreq >= qp->qp_size;
}

static inline uint32_t
compress_mmo_control_fields_arrange()
{
	return COMPRESS_DEFAULT_WINDOW_SIZE << COMPRESS_WINDOW_SIZE_OFFSET |
	       COMPRESS_DEFAULT_BLOCK_SIZE << COMPRESS_BLOCK_SIZE_OFFSET |
	       COMPRESS_DEFAULT_DYNAMIC_SIZE << COMPRESS_DYNAMIC_SIZE_OFFSET;
}

static inline void
compress_mlx5_devx_fill_compress_wqe(struct mlx5_mmo_wqe *compress_wqe, uint16_t pi, uint32_t qp_num,
				     struct ibv_mr *metadata_mr, uint32_t dest_lkey, uint64_t dest_addr,
				     uint32_t src_lkey, uint64_t src_addr, uint64_t src_len, uint64_t dest_len,
				     uint8_t opmod, bool has_umr, uint32_t true_pi)
{
	uint32_t imm = compress_mmo_control_fields_arrange();

	mlx5dv_set_ctrl_seg(&compress_wqe->ctrl,	/* struct mlx5_wqe_ctrl_seg *seg */
			    pi,				/* uint16_t pi */
			    MLX5_OPCODE_MMO,		/* uint8_t opcode 0x2f */
			    opmod,			/* uint8_t opmod */
			    qp_num,			/* uint32_t qp_num */
			    MLX5_WQE_CTRL_CQ_UPDATE,	/* uint8_t fm_ce_se */
			    sizeof(*compress_wqe) / 16,	/* ds: space for ctrl_segement, metadata + 2 * mlx5_wqe_data_seg */
			    0,				/* uint8_t signature */
			    htobe32(imm));		/* uint32_t imm */

	/* All WQEs point to the same metadata region (since we don't really need it) */
	compress_wqe->mmo_meta.mmo_control_31_0 = 0;
	compress_wqe->mmo_meta.local_key = htobe32(metadata_mr->lkey);
	compress_wqe->mmo_meta.local_address = htobe64((uint64_t)((uintptr_t)(metadata_mr->addr) +
						       (true_pi * MLX5_COMPRESS_METADATA_BUF_LEN)));

	if (!has_umr)
		mlx5dv_set_data_seg(&compress_wqe->src, src_len, src_lkey, src_addr);
	mlx5dv_set_data_seg(&compress_wqe->dest, dest_len, dest_lkey, dest_addr);
}

static inline uint32_t
compress_mlx5_devx_post_one(struct compress_qp *qp, struct ibv_mr *metadata_mr, uint8_t *uar_base_addr,
			    uint32_t dest_lkey, uint64_t dest_addr,
			    uint32_t src_lkey, uint64_t src_addr,
			    uint64_t src_len, uint64_t dst_len, uint64_t wr_id, uint8_t opmod)
{
	/* We want to post only one WQE */
	const unsigned int wqe_to_post = 1;
	uint32_t true_pi = qp->pi & (qp->qp_size - 1);

	struct mlx5_mmo_wqe *compress_wqe =
				(struct mlx5_mmo_wqe *)((uint8_t *)qp->umem_buf + (true_pi * MLX5_SEND_WQE_BB));

	compress_mlx5_devx_fill_compress_wqe(compress_wqe, qp->pi, qp->devx_obj->id, metadata_mr, dest_lkey, dest_addr,
					     src_lkey, src_addr, src_len, dst_len, opmod, false, true_pi);

	/* Every WQE needs 1 WQEBB */
	qp->pi += 1 * wqe_to_post;

	priv_doca_udma_to_device_barrier();
	qp->db_rec[MLX5_SND_DBR] = htobe32(qp->pi & 0xffff);
	priv_doca_mmio_wc_start();
	*(uint64_t *)(uar_base_addr + 0x800) = *(uint64_t *)compress_wqe;
	priv_doca_mmio_flush_writes();

	qp->wr_id_arr[true_pi].umr_handle = NULL;

	return true_pi;
}

static inline int
compress_mlx5_devx_post_umr_wqe(struct compress_qp *qp, struct ibv_mr *metadata_mr,
			   uint8_t *uar_base_addr, uint32_t dest_lkey, uint64_t dest_addr, struct doca_buf *src_buf,
			   uint64_t dest_len, uint8_t opmod)
{
	struct priv_doca_umr_handle *handle;
	doca_error_t ret;
	bool need_fence = false;
	uint32_t true_pi = qp->pi & (qp->qp_size - 1);

	/* qp->pi is the actual pi which represents (job_idx * 4). So to find job_idx we divide pi by 4.*/
	true_pi >>= 2;


	struct compress_mlx5_umr_wqe *umr_compress_wqe = ((struct compress_mlx5_umr_wqe *)qp->umem_buf + true_pi);
	ret = priv_doca_umr_buf_to_wqe(qp->umr, src_buf, umr_compress_wqe, qp->pi, qp->devx_obj->id,
				       (struct priv_doca_data_segment *)&umr_compress_wqe->mmo_seg.src,
				       &need_fence, &handle);
	if (unlikely(ret != DOCA_SUCCESS))
		return -1;
	/* UMR WQE needs 3 WQEBB */
	qp->pi += 3;

	compress_mlx5_devx_fill_compress_wqe(&umr_compress_wqe->mmo_seg, qp->pi, qp->devx_obj->id, metadata_mr,
					     dest_lkey, dest_addr, 0, 0, 0, dest_len, opmod, true, true_pi);
	if (need_fence)
		/* Make sure UMR processing ends before COMPRESS operation starts. */
		umr_compress_wqe->mmo_seg.ctrl.fm_ce_se |= MLX5_WQE_CTRL_INITIATOR_SMALL_FENCE;

	/* COMPRESS MMO WQE needs 1 WQEBB */
	qp->pi++;

	priv_doca_udma_to_device_barrier();
	qp->db_rec[MLX5_SND_DBR] = htobe32(qp->pi & 0xffff);
	priv_doca_mmio_wc_start();
	*(uint64_t *)(uar_base_addr + 0x800) = *(uint64_t *)&umr_compress_wqe->mmo_seg.ctrl;
	priv_doca_mmio_flush_writes();

	qp->wr_id_arr[true_pi].umr_handle = handle;

	return true_pi;
}

doca_error_t
compress_mlx5_devx_get_pdn(void *pd, uint32_t *pdn);

doca_error_t
compress_mlx5_devx_qp_init(void *ctx, struct compress_qp *new_qp, uint16_t log_wqbb_n,
		      struct compress_mlx5_devx_qp_attr *attr);
doca_error_t
compress_mlx5_devx_qp_destroy(struct compress_qp *qp);

doca_error_t
compress_mlx5_devx_qp2rts(struct compress_qp *qp, uint32_t remote_qp_id);

#endif /* COMPRESS_MLX5_COMMON_H_ */
