/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <unistd.h>

#include <doca_log.h>

#include <multi_os_prm.h>

#include "impl/dma_mlx5_common.h"

DOCA_LOG_REGISTER(DMA_MLX5_COMMON);

static inline size_t
get_mem_page_size(void)
{
	static size_t page_size;

	if (page_size == 0)
		page_size = sysconf(_SC_PAGESIZE);

	return page_size;
}

static inline uint32_t
get_umem_id(const void *umem)
{
	if (umem == NULL)
		return 0;

	return ((const struct mlx5dv_devx_umem *)umem)->umem_id;
}

static inline uint32_t
log2_u32(uint32_t x)
{
	if (x == 0)
		/* log(0) is undefined */
		return 0;

	return 31u - __builtin_clz(x);
}

static struct dma_devx_obj *
dma_mlx5_devx_cmd_create_cq(void *ctx, const struct dma_mlx5_devx_cq_attr *attr)
{
	uint32_t in[MLX5_ST_SZ_DW(create_cq_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(create_cq_out)] = {0};
	struct dma_devx_obj *cq_obj;
	void *cq_ctx = MLX5_ADDR_OF(create_cq_in, in, cq_context);

	cq_obj = calloc(1, sizeof(*cq_obj));
	if (cq_obj == NULL) {
		DOCA_DLOG_ERR("Failed to allocate CQ object memory");
		return NULL;
	}

	MLX5_SET(create_cq_in, in, opcode, MLX5_CMD_OP_CREATE_CQ);
	if (attr->db_umem_valid) {
		MLX5_SET(cqc, cq_ctx, dbr_umem_valid, attr->db_umem_valid);
		MLX5_SET(cqc, cq_ctx, dbr_umem_id, attr->db_umem_id);
		MLX5_SET64(cqc, cq_ctx, dbr_addr, attr->db_umem_offset);
	} else
		MLX5_SET64(cqc, cq_ctx, dbr_addr, attr->db_addr);

	MLX5_SET(cqc, cq_ctx, cqe_sz, (CACHELINE_SIZE == 128) ? MLX5_CQE_SIZE_128B : MLX5_CQE_SIZE_64B);

	MLX5_SET(cqc, cq_ctx, cc, attr->use_first_only);
	MLX5_SET(cqc, cq_ctx, oi, attr->overrun_ignore);
	MLX5_SET(cqc, cq_ctx, log_cq_size, attr->log_cq_size);
	MLX5_SET(cqc, cq_ctx, log_page_size, attr->log_page_size - MLX5_ADAPTER_PAGE_SHIFT);
	MLX5_SET(cqc, cq_ctx, c_eqn, attr->eqn);
	MLX5_SET(cqc, cq_ctx, uar_page, attr->uar_page_id);
	MLX5_SET(cqc, cq_ctx, cqe_comp_en, !!attr->cqe_comp_en);
	MLX5_SET(cqc, cq_ctx, mini_cqe_res_format, attr->mini_cqe_res_format);
	MLX5_SET(cqc, cq_ctx, mini_cqe_res_format_ext, attr->mini_cqe_res_format_ext);
	if (attr->q_umem_valid) {
		MLX5_SET(create_cq_in, in, cq_umem_valid, attr->q_umem_valid);
		MLX5_SET(create_cq_in, in, cq_umem_id, attr->q_umem_id);
		MLX5_SET64(create_cq_in, in, cq_umem_offset, attr->q_umem_offset);
	}

	cq_obj->obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (cq_obj->obj == NULL) {
		DOCA_DLOG_ERR("Failed to create CQ using DevX");
		free(cq_obj);
		return NULL;
	}
	cq_obj->id = MLX5_GET(create_cq_out, out, cqn);

	DOCA_DLOG_INFO("CQ number 0x%x was created", cq_obj->id);

	return cq_obj;
}

static void
cq_init(struct mlx5_cqe64 *cqes, uint16_t cq_size)
{
	uint16_t i;

	for (i = 0; i < cq_size; i++)
		cqes[i].op_own = (MLX5_CQE_INVALID << 4) | MLX5_CQE_OWNER_MASK;
}

doca_error_t
dma_mlx5_devx_cq_init(void *ctx, struct dma_cq *new_cq, uint16_t log_cq_size, struct dma_mlx5_devx_cq_attr *attr)
{
	struct dma_devx_obj *cq = NULL;
	struct mlx5dv_devx_umem *umem_obj = NULL;
	void *umem_buf = NULL;
	size_t page_size = get_mem_page_size();
	size_t alignment = page_size;
	uint32_t umem_size, umem_dbrec;
	uint32_t eqn;
	uint16_t cq_size = 1U << log_cq_size;
	int ret;

	if (page_size == (size_t)-1 || alignment == (size_t)-1)
		return DOCA_ERROR_NO_MEMORY;

	/* Query first EQN. */
	ret = mlx5dv_devx_query_eqn(ctx, 0, &eqn);
	if (ret) {
		DOCA_DLOG_ERR("Failed to query event queue number on CQ. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	/* Allocate memory buffer for CQEs and doorbell record. */
	umem_size = sizeof(struct mlx5_cqe64) * cq_size;
	umem_dbrec = DMA_MLX5_ALIGN(umem_size, DMA_MLX5_DBR_SIZE);
	umem_size += DMA_MLX5_DBR_SIZE;
	umem_size += alignment - (umem_size % alignment);
	umem_buf = aligned_alloc(alignment, umem_size);
	if (umem_buf == NULL) {
		DOCA_DLOG_ERR("CQ memory failure.");
		return DOCA_ERROR_NO_MEMORY;
	}
	memset(umem_buf, 0, umem_size);

	/* Register allocated buffer in user space with DevX. */
	umem_obj = mlx5dv_devx_umem_reg(ctx, (void *)(uintptr_t)umem_buf, umem_size, IBV_ACCESS_LOCAL_WRITE);
	if (umem_obj == NULL) {
		DOCA_DLOG_ERR("Failed to register umem for CQ");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}

	/* Fill attributes for CQ object creation. */
	attr->q_umem_valid = 1;
	attr->q_umem_id = get_umem_id(umem_obj);
	attr->q_umem_offset = 0;
	attr->db_umem_valid = 1;
	attr->db_umem_id = attr->q_umem_id;
	attr->db_umem_offset = umem_dbrec;
	attr->eqn = eqn;
	attr->log_cq_size = log_cq_size;
	attr->log_page_size = log2_u32(page_size);

	/* Create completion queue object with DevX. */
	cq = dma_mlx5_devx_cmd_create_cq(ctx, attr);
	if (cq == NULL) {
		DOCA_DLOG_ERR("DevX CQ object failure");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}
	new_cq->cq_size = cq_size;
	new_cq->ci = 0;
	new_cq->umem_buf = umem_buf;
	new_cq->umem_obj = umem_obj;
	new_cq->devx_obj = cq;
	new_cq->db_rec = DOCA_PTR_ADD(new_cq->umem_buf, umem_dbrec);

	/* Mark all CQEs initially as invalid. */
	cq_init(new_cq->cqes, cq_size);

	return DOCA_SUCCESS;
error:
	if (umem_obj != NULL)
		mlx5dv_devx_umem_dereg(umem_obj);
	if (umem_buf != NULL)
		free(umem_buf);

	return ret;
}

doca_error_t
dma_mlx5_devx_cq_destroy(struct dma_cq *cq)
{
	int ret;

	if (cq->devx_obj != NULL) {
		ret = mlx5dv_devx_obj_destroy(cq->devx_obj->obj);
		if (ret != 0) {
			DOCA_DLOG_ERR("Failed to destroy CQ DevX object. ret=%d", ret);
			return DOCA_ERROR_INVALID_VALUE;
		}
		free(cq->devx_obj);
	}
	if (cq->umem_obj != NULL) {
		ret = mlx5dv_devx_umem_dereg(cq->umem_obj);
		if (ret != 0) {
			DOCA_DLOG_ERR("Failed to destroy CQ umem object. ret=%d", ret);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}
	if (cq->umem_buf != NULL)
		free(cq->umem_buf);

	return DOCA_SUCCESS;
}

static struct dma_devx_obj *
dma_mlx5_devx_cmd_create_qp(void *ctx, const struct dma_mlx5_devx_qp_attr *attr)
{
	uint32_t in[MLX5_ST_SZ_DW(create_qp_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(create_qp_out)] = {0};
	void *qpc = MLX5_ADDR_OF(create_qp_in, in, qpc);
	struct dma_devx_obj *qp_obj = calloc(1, sizeof(*qp_obj));

	if (qp_obj == NULL) {
		DOCA_DLOG_ERR("Failed to allocate QP data.");
		return NULL;
	}
	MLX5_SET(create_qp_in, in, opcode, MLX5_CMD_OP_CREATE_QP);
	MLX5_SET(qpc, qpc, st, MLX5_QPC_ST_RC);
	MLX5_SET(qpc, qpc, pd, attr->pd);
	MLX5_SET(qpc, qpc, ts_format, attr->ts_format);
	MLX5_SET(qpc, qpc, user_index, attr->user_index);
	if (attr->uar_index) {
		if (attr->mmo) {
			void *qpc_ext_and_pas_list = MLX5_ADDR_OF(create_qp_in, in, qpc_extension_and_pas_list);
			void *qpc_ext =
				MLX5_ADDR_OF(qpc_extension_and_pas_list, qpc_ext_and_pas_list, qpc_data_extension);

			MLX5_SET(create_qp_in, in, qpc_ext, 1);
			MLX5_SET(qpc_extension, qpc_ext, mmo, 1);
		}
		MLX5_SET(qpc, qpc, pm_state, MLX5_QPC_PM_STATE_MIGRATED);
		MLX5_SET(qpc, qpc, uar_page, attr->uar_index);
		if (attr->log_page_size > MLX5_ADAPTER_PAGE_SHIFT)
			MLX5_SET(qpc, qpc, log_page_size, attr->log_page_size - MLX5_ADAPTER_PAGE_SHIFT);
		if (attr->sq_size) {
			MLX5_SET(qpc, qpc, cqn_snd, attr->cqn);
			MLX5_SET(qpc, qpc, log_sq_size, log2_u32(attr->sq_size));
		} else
			MLX5_SET(qpc, qpc, no_sq, 1);

		if (attr->rq_size) {
			MLX5_SET(qpc, qpc, cqn_rcv, attr->cqn);
			MLX5_SET(qpc, qpc, log_rq_stride, attr->log_rq_stride - MLX5_LOG_RQ_STRIDE_SHIFT);
			MLX5_SET(qpc, qpc, log_rq_size, log2_u32(attr->rq_size));
			MLX5_SET(qpc, qpc, rq_type, MLX5_QPC_RQ_TYPE_REGULAR);
		} else
			MLX5_SET(qpc, qpc, rq_type, MLX5_QPC_RQ_TYPE_ZERO_SIZE_RQ);

		if (attr->dbr_umem_valid) {
			MLX5_SET(qpc, qpc, dbr_umem_valid, attr->dbr_umem_valid);
			MLX5_SET(qpc, qpc, dbr_umem_id, attr->dbr_umem_id);
		}
		MLX5_SET64(qpc, qpc, dbr_addr, attr->dbr_address);
		MLX5_SET64(create_qp_in, in, wq_umem_offset, attr->wq_umem_offset);
		MLX5_SET(create_qp_in, in, wq_umem_id, attr->wq_umem_id);
		MLX5_SET(create_qp_in, in, wq_umem_valid, 1);
	} else {
		/* Special QP to be managed by FW - no SQ\RQ\CQ\UAR\DB rec. */
		MLX5_SET(qpc, qpc, rq_type, MLX5_QPC_RQ_TYPE_ZERO_SIZE_RQ);
		MLX5_SET(qpc, qpc, no_sq, 1);
	}

	qp_obj->obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (qp_obj->obj == NULL) {
		DOCA_DLOG_ERR("Failed to create QP Obj using DevX.");
		free(qp_obj);
		return NULL;
	}

	qp_obj->id = MLX5_GET(create_qp_out, out, qpn);

	DOCA_DLOG_INFO("QP number 0x%x was created", qp_obj->id);

	return qp_obj;
}

doca_error_t
dma_mlx5_devx_qp_init(void *ctx, struct dma_qp *new_qp, uint16_t log_wqbb_n, struct dma_mlx5_devx_qp_attr *attr)
{
	struct dma_devx_obj *qp = NULL;
	struct mlx5dv_devx_umem *umem_obj = NULL;
	void *umem_buf = NULL;
	void *wr_id_arr = NULL;
	size_t page_size = get_mem_page_size();
	size_t alignment = page_size;
	uint32_t umem_size, umem_dbrec;
	uint16_t qp_size = 1 << log_wqbb_n;
	doca_error_t ret;

	if (alignment == (size_t)-1) {
		DOCA_DLOG_ERR("Failed to get WQE buf alignment.");
		return DOCA_ERROR_NO_MEMORY;
	}

	wr_id_arr = malloc(sizeof(struct dma_wr_id) * qp_size);
	if (wr_id_arr == NULL) {
		DOCA_DLOG_ERR("Failed to allocate wr-id memory for QP.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}

	/* Allocate memory buffer for WQEs and doorbell record. */
	umem_size = sizeof(struct mlx5_mmo_wqe) * qp_size;
	umem_dbrec = DMA_MLX5_ALIGN(umem_size, DMA_MLX5_DBR_SIZE);
	umem_size += DMA_MLX5_DBR_SIZE;
	umem_size += alignment - (umem_size % alignment);

	umem_buf = aligned_alloc(alignment, umem_size);
	if (umem_buf == NULL) {
		DOCA_DLOG_ERR("Failed to allocate memory for QP work queues.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}
	memset(umem_buf, 0, umem_size);

	/* Register allocated buffer in user space with DevX. */
	umem_obj = mlx5dv_devx_umem_reg(ctx, (void *)(uintptr_t)umem_buf, umem_size, 1);
	if (umem_obj == NULL) {
		DOCA_DLOG_ERR("Failed to register umem for QP.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}

	/* Fill attributes for SQ object creation. */
	attr->wq_umem_id = get_umem_id(umem_obj);
	attr->wq_umem_offset = 0;
	attr->dbr_umem_valid = 1;
	attr->dbr_umem_id = attr->wq_umem_id;
	attr->dbr_address = umem_dbrec;
	attr->log_page_size = log2_u32(page_size);
	/* Create send queue object with DevX. */
	qp = dma_mlx5_devx_cmd_create_qp(ctx, attr);
	if (qp == NULL) {
		DOCA_DLOG_ERR("Can't create DevX QP object.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto error;
	}

	new_qp->qp_size = qp_size;
	new_qp->ci = 0;
	new_qp->pi = 0;
	new_qp->wr_id_arr = wr_id_arr;
	new_qp->wr_id_front = 0;
	new_qp->wr_id_back = 0;
	new_qp->umem_buf = umem_buf;
	new_qp->umem_obj = umem_obj;
	new_qp->devx_obj = qp;
	new_qp->db_rec = DOCA_PTR_ADD(new_qp->umem_buf, umem_dbrec);

	return DOCA_SUCCESS;
error:
	if (umem_obj != NULL)
		mlx5dv_devx_umem_dereg(umem_obj);
	if (umem_buf != NULL)
		free(umem_buf);
	if (wr_id_arr != NULL)
		free(wr_id_arr);

	return ret;
}

doca_error_t
dma_mlx5_devx_qp_destroy(struct dma_qp *qp)
{
	int ret;

	if (qp->devx_obj != NULL) {
		ret = mlx5dv_devx_obj_destroy(qp->devx_obj->obj);
		if (ret != 0) {
			DOCA_DLOG_ERR("Failed to destroy QP DevX object. ret=%d", ret);
			return DOCA_ERROR_INVALID_VALUE;
		}
		free(qp->devx_obj);
	}
	if (qp->umem_obj != NULL) {
		ret = mlx5dv_devx_umem_dereg(qp->umem_obj);
		if (ret != 0) {
			DOCA_DLOG_ERR("Failed to destroy QP umem object. ret=%d", ret);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}
	if (qp->umem_buf != NULL)
		free(qp->umem_buf);
	if (qp->wr_id_arr != NULL)
		free(qp->wr_id_arr);

	return DOCA_SUCCESS;
}

static doca_error_t
dma_mlx5_devx_cmd_modify_qp_rst2init(struct dma_devx_obj *qp)
{
	uint32_t in[MLX5_ST_SZ_DW(rst2init_qp_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(rst2init_qp_out)] = {0};
	void *qpc = MLX5_ADDR_OF(rst2init_qp_in, &in, qpc);
	int ret;

	MLX5_SET(rst2init_qp_in, &in, opcode, MLX5_CMD_OP_RST2INIT_QP);
	MLX5_SET(rst2init_qp_in, &in, qpn, qp->id);

	MLX5_SET(qpc, qpc, primary_address_path.vhca_port_num, 1);
	MLX5_SET(qpc, qpc, rre, 0);
	MLX5_SET(qpc, qpc, rwe, 0);
	MLX5_SET(qpc, qpc, pm_state, MLX5_QPC_PM_STATE_MIGRATED);

	ret = mlx5dv_devx_obj_modify(qp->obj, &in, sizeof(in), &out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to modify QP rst2init using DevX. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
dma_mlx5_devx_cmd_modify_qp_init2rtr(struct dma_devx_obj *qp, uint32_t remote_qp_id)
{
	uint32_t in[MLX5_ST_SZ_DW(init2rtr_qp_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(init2rtr_qp_out)] = {0};
	void *qpc = MLX5_ADDR_OF(init2rtr_qp_in, &in, qpc);
	int ret;

	MLX5_SET(init2rtr_qp_in, &in, opcode, MLX5_CMD_OP_INIT2RTR_QP);
	MLX5_SET(init2rtr_qp_in, &in, qpn, qp->id);

	MLX5_SET(qpc, qpc, primary_address_path.fl, 1);
	MLX5_SET(qpc, qpc, primary_address_path.vhca_port_num, 1);
	MLX5_SET(qpc, qpc, mtu, 3);
	MLX5_SET(qpc, qpc, log_msg_max, 30);
	MLX5_SET(qpc, qpc, remote_qpn, remote_qp_id);
	MLX5_SET(qpc, qpc, min_rnr_nak, 0xc);
	MLX5_SET(qpc, qpc, next_rcv_psn, 0);

	ret = mlx5dv_devx_obj_modify(qp->obj, &in, sizeof(in), &out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to modify QP init2rtr using DevX. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
dma_mlx5_devx_cmd_modify_qp_rtr2rts(struct dma_devx_obj *qp)
{
	uint32_t in[MLX5_ST_SZ_DW(rtr2rts_qp_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(rtr2rts_qp_out)] = {0};
	void *qpc = MLX5_ADDR_OF(rtr2rts_qp_in, &in, qpc);
	int ret;

	MLX5_SET(rtr2rts_qp_in, &in, opcode, MLX5_CMD_OP_RTR2RTS_QP);
	MLX5_SET(rtr2rts_qp_in, &in, qpn, qp->id);

	MLX5_SET(qpc, qpc, next_send_psn, 0);
	MLX5_SET(qpc, qpc, primary_address_path.ack_timeout, 14);
	MLX5_SET(qpc, qpc, log_ack_req_freq, 0);
	MLX5_SET(qpc, qpc, retry_count, 0);
	MLX5_SET(qpc, qpc, rnr_retry, 3);

	ret = mlx5dv_devx_obj_modify(qp->obj, &in, sizeof(in), &out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to modify QP rtr2rts using DevX. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

doca_error_t
dma_mlx5_devx_qp2rts(struct dma_qp *qp, uint32_t remote_qp_id)
{
	doca_error_t ret;

	ret = dma_mlx5_devx_cmd_modify_qp_rst2init(qp->devx_obj);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = dma_mlx5_devx_cmd_modify_qp_init2rtr(qp->devx_obj, remote_qp_id);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = dma_mlx5_devx_cmd_modify_qp_rtr2rts(qp->devx_obj);
	if (ret != DOCA_SUCCESS)
		return ret;

	return DOCA_SUCCESS;
}

doca_error_t
dma_mlx5_devx_get_pdn(void *pd, uint32_t *pdn)
{
	struct mlx5dv_obj obj = {0};
	struct mlx5dv_pd pd_info = {0};
	int ret;

	obj.pd.in = pd;
	obj.pd.out = &pd_info;

	ret = mlx5dv_init_obj(&obj, MLX5DV_OBJ_PD);
	if (ret != 0) {
		DOCA_DLOG_ERR("Failed to get PD number. ret=%d", ret);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	*pdn = pd_info.pdn;

	return DOCA_SUCCESS;
}
