/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DMA_MLX5_COMMON_H_
#define DMA_MLX5_COMMON_H_

#include <doca_build_config.h>

#include <stddef.h>
#include <stdint.h>


#ifdef DOCA_ARCH_X86
#include <immintrin.h>
#endif

#include <infiniband/verbs.h>
#include <private/common_defines.h>
#include <priv_doca_os_abstraction_memory.h>

#define DOCA_PTR_ADD(ptr, x) ((void *)((uintptr_t)(ptr) + (x)))

#define DMA_MLX5_ALIGN(v, mul) ((((v) + (typeof(v))(mul)-1) / ((typeof(v))(mul))) * (typeof(v))(mul))

/* NOTE: hardcoded for now - 64 is fine for BF ARM and x86. */
#define CACHELINE_SIZE 64

#define DMA_MLX5_DBR_SIZE CACHELINE_SIZE

#define MLX5_LOG_RQ_STRIDE_SHIFT 4

/* Just keep enough space for the metadata of a single WQE.
 * Since we don't need the info that the device provides in this region,
 * all the WQEs will point to the same metadata region.
 */
#define MLX5_DMA_METADATA_BUF_LEN 64

enum { MLX5_OP_MOD_MMO_DMA = 0x1,
};

struct dma_devx_obj {
	void *obj; /* The DevX object. */
	int id;	   /* The object ID. */
};

/* CQ attributes structure, used by CQ operations. */
struct dma_mlx5_devx_cq_attr {
	uint32_t q_umem_valid : 1;
	uint32_t db_umem_valid : 1;
	uint32_t use_first_only : 1;
	uint32_t overrun_ignore : 1;
	uint32_t cqe_comp_en : 1;
	uint32_t mini_cqe_res_format : 2;
	uint32_t mini_cqe_res_format_ext : 2;
	uint32_t log_cq_size : 5;
	uint32_t log_page_size : 5;
	uint32_t uar_page_id;
	uint32_t q_umem_id;
	uint64_t q_umem_offset;
	uint32_t db_umem_id;
	uint64_t db_umem_offset;
	uint32_t eqn;
	uint64_t db_addr;
};

/* QP attributes structure, used by QP operations. */
struct dma_mlx5_devx_qp_attr {
	uint32_t pd : 24;
	uint32_t uar_index : 24;
	uint32_t cqn : 24;
	uint32_t log_page_size : 5;
	uint32_t rq_size : 17; /* Must be power of 2. */
	uint32_t log_rq_stride : 3;
	uint32_t sq_size : 17; /* Must be power of 2. */
	uint32_t ts_format : 2;
	uint32_t dbr_umem_valid : 1;
	uint32_t dbr_umem_id;
	uint64_t dbr_address;
	uint32_t wq_umem_id;
	uint64_t wq_umem_offset;
	uint32_t user_index : 24;
	uint32_t mmo : 1;
};

/**
 * DMA engine completion queue.
 */
struct dma_cq {
	uint32_t cq_size;	       /* Number of entries in this CQ (value is power of two). */
	struct dma_devx_obj *devx_obj; /* The CQ DevX object. */
	void *umem_obj;		       /* The CQ umem object. */
	union {
		void *umem_buf;
		struct mlx5_cqe64 *cqes; /* The CQ ring buffer. */
	};
	uint32_t *db_rec; /* The CQ doorbell record. */
	uint32_t ci;	  /* Consumer index. */
};

/**
 * DMA WorkQ ID.
 */
struct dma_wr_id {
	uint64_t user_data;
	struct doca_buf *dst_buf;
};

/**
 * DMA engine queue pair.
 */
struct dma_qp {
	struct dma_wr_id *wr_id_arr;   /* Array for holding wr_ids */
	struct dma_devx_obj *devx_obj; /* The QP DevX object. */
	void *umem_obj;		       /* The QP umem object. */
	union {
		void *umem_buf;
		struct mlx5_mmo_wqe *wqes; /* The QP ring buffer. */
	};
	uint32_t *db_rec; /* The QP doorbell record. */
	uint32_t pi;	  /* Producer index. */
	uint32_t ci;	  /* Consumer index. */
	uint16_t qp_size; /* Number of entries in this QP (value is power of two). */
	uint16_t wr_id_front;
	uint16_t wr_id_back;
};

struct mlx5_mmo_metadata_seg {
	uint32_t mmo_control_31_0;
	uint32_t local_key;
	uint64_t local_address;
};

struct mlx5_mmo_wqe {
	struct mlx5_wqe_ctrl_seg ctrl;
	struct mlx5_mmo_metadata_seg mmo_meta;
	struct mlx5_wqe_data_seg src;
	struct mlx5_wqe_data_seg dest;
};

/**
 * DMA engine completion: only the needed values of the CQE.
 */
struct dma_wc {
	struct dma_wr_id *wr_id;
	uint32_t status;
	uint32_t vendor_err;
	uint32_t byte_count;
};

/* return true - if good CQE was found,
 *        false - if CQE with error was found
 */
static inline bool
dma_mlx5_devx_process_cqe(struct mlx5_cqe64 *cqe, struct dma_qp *qp, struct dma_wc *dma_wc)
{
	const uint8_t op_code = mlx5dv_get_cqe_opcode(cqe);
	bool ret;

	/* CQE with error */
	if (unlikely(op_code == MLX5_CQE_REQ_ERR || op_code == MLX5_CQE_RESP_ERR)) {
		const struct mlx5_err_cqe *err_cqe = (const struct mlx5_err_cqe *)cqe;

		dma_wc->status = err_cqe->syndrome;
		dma_wc->vendor_err = err_cqe->vendor_err_synd;
		ret = false;
	} else { /* Good CQE */
		dma_wc->status = 0;
		ret = true;
	}

	/* Get the wr_id of the job */
	dma_wc->wr_id = &qp->wr_id_arr[qp->wr_id_back];
	qp->wr_id_back = (qp->wr_id_back + 1) & (qp->qp_size - 1);

	dma_wc->byte_count = be32toh(cqe->byte_cnt);

	/* Update the CI according to the wqe-index (to consider unsignaled WQEs) */
	qp->ci += (uint16_t)(be16toh(cqe->wqe_counter) - (uint16_t)qp->ci + 1);

	return ret;
}

static inline int
dma_mlx5_devx_is_qp_overflow(const struct dma_qp *qp, int nreq)
{
	unsigned cur = qp->pi - qp->ci;

	if (likely(cur + nreq < qp->qp_size))
		return 0;

	return cur + nreq >= qp->qp_size;
}

static inline void
dma_mlx5_devx_post_n(struct dma_qp *qp, struct mlx5_mmo_wqe *dma_wqe, uint8_t *uar_base_addr, uint16_t nb_post)
{
	qp->pi = qp->pi + nb_post;

	priv_doca_udma_to_device_barrier();
	qp->db_rec[MLX5_SND_DBR] = htobe32(qp->pi & 0xffff);
	priv_doca_mmio_wc_start();
	*(uint64_t *)(uar_base_addr + 0x800) = *(uint64_t *)dma_wqe;
	priv_doca_mmio_flush_writes();
}

static inline void
dma_mlx5_devx_fill_one(struct dma_qp *qp, struct mlx5_mmo_wqe *dma_wqe, uint16_t pi, struct ibv_mr *metadata_mr,
		       uint8_t fm_ce_se, uint32_t dest_lkey, uint64_t dest_addr, uint32_t src_lkey,
		       uint64_t src_addr, uint64_t num_bytes)
{
	mlx5dv_set_ctrl_seg(&dma_wqe->ctrl,	     /* struct mlx5_wqe_ctrl_seg *seg */
			    pi,			     /* uint16_t pi */
			    MLX5_OPCODE_MMO,	     /* uint8_t opcode */
			    MLX5_OP_MOD_MMO_DMA,     /* uint8_t opmod */
			    qp->devx_obj->id,	     /* uint32_t qp_num */
			    fm_ce_se,		     /* uint8_t fm_ce_se */
			    sizeof(*dma_wqe) / 16,   /* ds: space for ctrl_segement, metadata + 2 * mlx5_wqe_data_seg */
			    0,			     // uint8_t signature
			    0);			     // uint32_t imm

	/* All WQEs point to the same metadata region (since we don't really need it) */
	dma_wqe->mmo_meta.mmo_control_31_0 = 0;
	dma_wqe->mmo_meta.local_key = htobe32(metadata_mr->lkey);
	dma_wqe->mmo_meta.local_address = htobe64((uint64_t)(uintptr_t)metadata_mr->addr);

	mlx5dv_set_data_seg(&dma_wqe->src, num_bytes, src_lkey, src_addr);
	mlx5dv_set_data_seg(&dma_wqe->dest, num_bytes, dest_lkey, dest_addr);
}

doca_error_t dma_mlx5_devx_get_pdn(void *pd, uint32_t *pdn);

doca_error_t dma_mlx5_devx_cq_init(void *ctx, struct dma_cq *new_cq, uint16_t log_cq_size,
				   struct dma_mlx5_devx_cq_attr *attr);

doca_error_t dma_mlx5_devx_cq_destroy(struct dma_cq *cq);

doca_error_t dma_mlx5_devx_qp_init(void *ctx, struct dma_qp *new_qp, uint16_t log_wqbb_n,
				   struct dma_mlx5_devx_qp_attr *attr);
doca_error_t dma_mlx5_devx_qp_destroy(struct dma_qp *qp);

doca_error_t dma_mlx5_devx_qp2rts(struct dma_qp *qp, uint32_t remote_qp_id);

#endif /* DMA_MLX5_COMMON_H_ */
