/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


/**
 * Every logical DPI queue is composed of 2 queues:
 * 1. SW queue.
 * 2. HW regex queue.
 * When enqueue is called a new job is posted on the SW queue. Each SW queue job
 * may create multiple regex jobs.
 * Each mbuf is parsed with the DPI parser. The DPI parser is producing regex jobs for the
 * regex queue.
 * SW queue entry is considered done when all regex queue entries generated by this
 * entry are done.
 *
 * │        │                                      │         │
 * │        │                                      │         │
 * ├────────┤           ┌────────────────┐         ├─────────┤
 * │SW Q    │           │                ├────────►│ Regex   │
 * │Entry   ├──────────►│  DPI Parser    │         │ Entry   │
 * ├────────┤           │                │         ├─────────┤
 * │        │           │                ├────────►│         │
 * │        │           └───────────────┬┘         │         │
 * ├────────┤                           │          ├─────────┤
 * │        │                           └─────────►│         │
 * │        │                                      │         │
 * ├────────┤                                      ├─────────┤
 *
 * Each enqueued mbuf is creating an indirect mbuf. The indirect mbuf will exist for as long as
 * The parser requires it.
 *
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <bsd/string.h>

#include <rte_mbuf.h>
#include <rte_malloc.h>
#include <rte_regexdev.h>
#include <rte_hash.h>
#include <rte_spinlock.h>
#include <rte_ring.h>

#include <doca_log.h>

#include "doca_dpi.h"
#include "dpi_regex.h"
#include "cdo_decoder.h"
#include "parser/dpi_parser.h"
#include "parser/ssl_parser.h"
#include "parser/dns_cache.h"
#include "parser/ftp_cache.h"

#ifdef DOCA_DPI_USE_HS
	#define LOG_SW_QUEUE 4
#else
	#define LOG_SW_QUEUE 10
#endif
#define DPI_MAX_SW_QUEUE_SIZE (1 << LOG_SW_QUEUE)
#define DPI_MAX_HW_QUEUE_SIZE (DPI_MAX_SW_QUEUE_SIZE * 2)
#define REGEX_DEV_ID 0
#define REGEX_MAX_BUF (1 << 14)
#define DPI_MAX_SIG_MATCH_LEN 5000
#define DNS_PORT 53
#define FTP_PORT 21
#define FIELD_MAX_LEN 17
#define PROTOCOL_MAX_LEN 6
#define META_MAX_LEN 200

#define IPV6_SIZE 128
#define BYTE_SIZE 8

#define DPI_IPV4 0x0800
#define DPI_IPV6 0x86DD

struct enum_to_field {
	int protocol_num;
	int field_num;
	const char *name;
};

struct enum_to_protocol {
	int protocol_num;
	const char *name;
};

static struct enum_to_field field_to_str[] = {
	{DPI_PARSER_HTTP_1, HTTP_CONTENT, "content"},
	{DPI_PARSER_HTTP_1, HTTP_URI, "uri"},
	{DPI_PARSER_HTTP_1, HTTP_USER_AGENT, "user_agent"},
	{DPI_PARSER_HTTP_1, HTTP_LOCATION, "location"},
	{DPI_PARSER_HTTP_1, HTTP_SERVER_AGENT, "server"},
	{DPI_PARSER_HTTP_1, HTTP_SERVER, "host"},
	{DPI_PARSER_HTTP_1, HTTP_MIME_TYPE, "content_type"},
	{DPI_PARSER_HTTP_1, HTTP_COOKIE, "cookie"},
	{DPI_PARSER_HTTP_1, HTTP_REFERER, "referer"},
	{DPI_PARSER_SSL, SSL_COMMON_NAME, "common_name"},
	{DPI_PARSER_SSL, SSL_SERVER_NAME, "server_name"},
	{DPI_PARSER_SSL, SSL_SUBJECT_ALT_NAME, "subject_alt_name"},
	{DPI_PARSER_RTMP, RTMP_PAGE_URL, "page_url"},
	{DPI_PARSER_DNS, DNS_QUERY, "query"},
	{DPI_PARSER_DNS, DNS_RESOLVE_NAME, "resolv_name"},
	{DPI_PARSER_STUN, STUN_REALM, "realm"},
	{DPI_PARSER_STUN, STUN_SOFTWARE, "software"}
};

static struct enum_to_protocol protocol_to_str[] = {
	{DPI_PARSER_HTTP_1, "http"},
	{DPI_PARSER_HTTP_2, "http2"},
	{DPI_PARSER_SSL, "ssl"},
	{DPI_PARSER_SIP, "sip"},
	{DPI_PARSER_RTMP, "rtmp"},
	{DPI_PARSER_QUIC, "quic"},
	{DPI_PARSER_STUN, "stun"},
	{DPI_PARSER_DNS, "dns"},
};

enum dpi_counter {
	DPI_CNT_SCANNED,
	DPI_CNT_ERRORS,
	DPI_CNT_MATCHES,
	DPI_CNT_HTTP,
	DPI_CNT_SSL,
	DPI_CNT_TCP,
	DPI_CNT_UDP,
	DPI_CNT_OTHER_L4,
	DPI_CNT_OTHER_L7,
	DPI_MAX_COUNTERS,
};

struct flow_dir;

enum flow_direction_t {
	DPI_FLOW_SERVER,
	DPI_FLOW_CLIENT,
};

struct cond_match {
	LIST_ENTRY(cond_match) matches;
	uint32_t rule_id;
};

struct doca_dpi_flow_ctx;

struct flow_dir {
	struct dpi_parser_ctx *p_ctx;
	struct rte_mbuf *flow_chain;
	uint32_t total_payload_parsed;
	uint32_t first_mbuf_offset;
	uint16_t dpi_q;
	struct doca_dpi_flow_ctx *flow_ctx;
	bool dirty;
};

struct dpi_sig_info {
	struct pp_table_entry pp_entry;
};

struct doca_dpi_flow_ctx {
	bool matched;
	int bypass_state;
	LIST_HEAD(cond_matches, cond_match) cond_match_list;
	struct flow_dir dir_obj[2];
	struct doca_dpi_parsing_info parsing_info;
	bool counted;
	struct dpi_sig_info sig_info;
	bool destroyed;
	uint32_t currently_in_queue;
	struct dpi_queue_context *dpi_q_ctx;
};

struct dpi_regex_entry_meta {
	struct dpi_sw_queue_entry *sw_entry;
	int protocol;
	int field;
	uint16_t len;
	bool stream_start;
};

/*
 * Regex queue holds the hw facing queue.
 */
struct dpi_regex_queue {
	uint32_t hw_queue_id;
	struct dpi_regex_entry_meta entries[DPI_MAX_HW_QUEUE_SIZE];
	struct rte_ring *free_list;
};

enum dpi_sw_queue_entry_state {
	DPI_SW_QUEUE_ENTRY_DONE,
	DPI_SW_QUEUE_ENTRY_PROCESSSING,
};

struct dpi_sw_queue_entry {
	struct flow_dir *dir_ctx;
	struct dpi_regex_queue *regex_queue;
	struct rte_mbuf *mbuf;
	uint16_t outstanding_regex_jobs;
	void *user_data;
	int initiator;
	uint32_t payload_offset;
	struct rte_mbuf *mbufs_to_free;
};

/*
 * SW queue holds the user facing queue.
 * An entry may be consumed only partially by the
 * regex.
 */
struct dpi_sw_queue {
	uint32_t nb_entries;
	struct dpi_sw_queue_entry entries[DPI_MAX_SW_QUEUE_SIZE];
	struct rte_ring *free_list;
	struct rte_ring *done_list;
};

struct doca_dpi_ctx;

struct dpi_queue_context {
	uint32_t outstanding;
	/*
	 * Because regex currently does not support
	 * gather list, need to copy packet to a
	 * single contiguous memory.
	 */

	struct rte_mempool *ind_pkts;
	struct dpi_cdo_data *cdo_data;
	struct dpi_sw_queue sw_queue;
	struct dpi_regex_queue regex_queue;
	rte_spinlock_t db_lock;
	uint32_t counters[DPI_MAX_COUNTERS];
	uint16_t dpi_q;
	struct doca_dpi_ctx *dpi_ctx;
	struct dpi_parser_main_ctx *parser_ctx;
	struct dpi_regex_queue_ctx *regex_queue_ctx;
};

struct doca_dpi_ctx {
	struct doca_dpi_config_t cfg;
	struct dpi_queue_context *dpi_q;
	struct dns_cache *dns_cache;
	struct ftp_cache *ftp_cache;
	struct dpi_regex_dev_ctx *regex_dev;
};

DOCA_LOG_REGISTER(DPI::CORE);

#define PRINT_MISMATCH(TYPE, EXPECTED, ACTUAL) \
	DOCA_DLOG_DBG(TYPE":\t"TYPE" mismatch: expected=%d, actual=%d", \
	EXPECTED, ACTUAL)

static bool
is_initiator(struct flow_dir *dir_obj)
{
	return &dir_obj->flow_ctx->dir_obj[DPI_FLOW_CLIENT] == dir_obj;
}

static int
update_and_check_conditions(const struct dpi_cdo_data *data,
			    struct doca_dpi_flow_ctx *flow_ctx,
			    const struct pp_table_entry *pp_entry);

static bool
entry_valid(const struct pp_table_entry *pp_entry, int type)
{
	return pp_entry->bitmask & type;
}

/*
 * Remove as much mbufs as possible from chain when parser is done with them.
 */
static void
dpi_mbuf_unchain(struct flow_dir *dir_obj, struct dpi_sw_queue_entry *sw_entry)
{
	while (rte_pktmbuf_data_len(dir_obj->flow_chain) <
	       dir_obj->total_payload_parsed - dir_obj->first_mbuf_offset) {
		struct rte_mbuf *mbuf;
		struct rte_mbuf *old_head = dir_obj->flow_chain;

		if (old_head->next == NULL)
			break;
		mbuf = old_head->next;
		mbuf->nb_segs = old_head->nb_segs - 1;
		mbuf->pkt_len = rte_pktmbuf_pkt_len(old_head) - rte_pktmbuf_data_len(old_head);
		old_head->nb_segs = 1;
		old_head->pkt_len = old_head->data_len;
		old_head->next = NULL;
		dir_obj->first_mbuf_offset += old_head->data_len;
		dir_obj->flow_chain = mbuf;
		/*
		 * We need to hold the mbuf until dequeue, only then we can be sure
		 * HW is done with it.
		 */
		if (sw_entry->mbufs_to_free)
			rte_pktmbuf_chain(sw_entry->mbufs_to_free, old_head);
		else
			sw_entry->mbufs_to_free = old_head;
	}
}

static int
match_one(const struct dpi_regex_entry_meta *regex_entry,
	  const struct rte_regexdev_match *regex_match,
	  const struct pp_table_entry *pp_entry, const struct doca_dpi_flow_ctx *flow_ctx)
{
	if (regex_entry && regex_match) {
		if (pp_entry->anchored && (regex_match->start_offset != 0 || !regex_entry->stream_start)) {
			PRINT_MISMATCH("Anchored", 1, 0);
			return 0;
		}
		if (entry_valid(pp_entry, DPI_PP_L4_PROTOCOL)) {
			if (entry_valid(pp_entry, DPI_PP_OFFSET)) {
				if (!regex_entry->stream_start)
					return 0;
				if (regex_match->start_offset < pp_entry->offset)
					return 0;
			}
		}
	}
	return 1;
}

static bool
has_higher_prio(const struct pp_table_entry *pp_entry_curr, const struct pp_table_entry *pp_entry_last)
{
	if (pp_entry_curr->order1 < pp_entry_last->order1)
		return true;
	else if (pp_entry_curr->order1 == pp_entry_last->order1 &&
			pp_entry_curr->order2 < pp_entry_last->order2)
		return true;
	return false;
}

static void match_single_pp(const struct pp_table_entry *pp_entry,
			    struct dpi_regex_entry_meta *regex_entry,
			    const struct rte_regexdev_match *regex_match,
			    struct doca_dpi_flow_ctx *flow_ctx, const struct dpi_cdo_data *cdo_data)
{
	int matched = match_one(regex_entry, regex_match, pp_entry, flow_ctx);

	if (matched) {
		int cond_match;

		if (flow_ctx->matched && !has_higher_prio(pp_entry, &flow_ctx->sig_info.pp_entry)) {
			/* Did not find higher prio sig, not point to further check */
			return;
		}
		cond_match = update_and_check_conditions(cdo_data, flow_ctx, pp_entry);
		if (cond_match == 1) {
			flow_ctx->matched = true;
			flow_ctx->sig_info.pp_entry = *pp_entry;
		}
	}
}
/*
 * Returns the sw_entry if the dequeued job
 */
static struct dpi_sw_queue_entry *
regex_queue_dequeue_locked(struct dpi_queue_context *dpi_queue)
{
	struct dpi_regex_queue *regex_queue = &dpi_queue->regex_queue;
	struct dpi_regex_entry_meta *regex_entry = NULL;
	struct dpi_sw_queue_entry *sw_entry = NULL;
	const struct rte_regex_ops *ops;
	struct flow_dir *dir_ctx;
	int i, ret;

	struct dpi_regex_result result;

	ret = dpi_regex_dequeue(dpi_queue->regex_queue_ctx, &result);
	if (ret == 0)
		return NULL;

	regex_entry = (struct dpi_regex_entry_meta *)result.user_ptr;
	sw_entry = regex_entry->sw_entry;
	ops = result.op;
	dir_ctx = sw_entry->dir_ctx;

	for (i = 0; i < ops->nb_matches; i++) {
		const struct pp_table_entry *pp_entry =
			cdo_pp_entry_get(dpi_queue->cdo_data, ops->matches[i].rule_id);

		for (; pp_entry != NULL; pp_entry = pp_entry->next)
			match_single_pp(pp_entry, regex_entry, &ops->matches[i], dir_ctx->flow_ctx,
					dpi_queue->cdo_data);
	}
	rte_ring_enqueue(regex_queue->free_list, regex_entry);
	return sw_entry;
}

static struct dpi_sw_queue_entry *
regex_queue_dequeue(struct dpi_queue_context *dpi_queue)
{
	struct dpi_sw_queue_entry *sw_entry;

	rte_spinlock_lock(&dpi_queue->db_lock);
	sw_entry = regex_queue_dequeue_locked(dpi_queue);
	rte_spinlock_unlock(&dpi_queue->db_lock);
	return sw_entry;
}

/*
 * Return sw_entry if all regex entries dequeued
 */
static struct dpi_sw_queue_entry *
sw_queue_dequeue(struct dpi_queue_context *dpi_queue)
{
	struct dpi_sw_queue_entry *sw_entry;

	while ((sw_entry = regex_queue_dequeue(dpi_queue))) {
		sw_entry->outstanding_regex_jobs--;
		if (sw_entry->outstanding_regex_jobs == 0) {
			if (sw_entry->mbufs_to_free) {
				rte_pktmbuf_free(sw_entry->mbufs_to_free);
				sw_entry->mbufs_to_free = NULL;
			}
			return sw_entry;
		}
	}
	return NULL;

}

/*
 * Convert ipv4 string format to ipv6 format
 */
static void
ipv4_to_ipv6(const char *ipv4str, char *ipv6buf)
{
	size_t pos = 0, last_pos = 0, dst_pos = 0;
	int ip_num;
	int len = strlen(ipv4str);
	int block_num = 0;
	char ip_hex[3] = {'\0'};
	char tmp_num[4] = {'\0'};

	while (pos < len) {
		while (ipv4str[pos] != '.') {
			pos++;
			if (pos == len)
				break;
		}
		/* assuming that the ipv4 string is legal */
		strlcpy(tmp_num, ipv4str + last_pos, pos - last_pos + 1);
		ip_num = atoi(tmp_num);
		sprintf(ip_hex, "%x", ip_num);
		if (ip_num < 16) {
			ipv6buf[dst_pos] = '0';
			dst_pos++;
		}
		strcpy(ipv6buf + dst_pos, ip_hex);
		dst_pos += ip_num < 16 ? 1 : 2;
		if (block_num == 1) {
			ipv6buf[dst_pos] = ':';
			dst_pos++;
		}
		block_num++;
		pos++;
		last_pos = pos;
	}
	ipv6buf[dst_pos] = '\0';
}

static void
ipv6_expander(char *str, const struct in6_addr *addr)
{
	sprintf(str, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
		(int)addr->s6_addr[0], (int)addr->s6_addr[1],
		(int)addr->s6_addr[2], (int)addr->s6_addr[3],
		(int)addr->s6_addr[4], (int)addr->s6_addr[5],
		(int)addr->s6_addr[6], (int)addr->s6_addr[7],
		(int)addr->s6_addr[8], (int)addr->s6_addr[9],
		(int)addr->s6_addr[10], (int)addr->s6_addr[11],
		(int)addr->s6_addr[12], (int)addr->s6_addr[13],
		(int)addr->s6_addr[14], (int)addr->s6_addr[15]);
}

/*
 * Update the meta data according to the information from the parsers
 */
static void
update_meta(char *meta, uint16_t protocol, uint16_t field, struct dpi_sw_queue_entry *sw_entry)
{
	int i;
	char ipv4_src_addr[INET_ADDRSTRLEN] = {'\0'};
	char ipv4_dst_addr[INET_ADDRSTRLEN] = {'\0'};
	char ipv6_src_addr[INET6_ADDRSTRLEN] = {'\0'};
	char ipv6_dst_addr[INET6_ADDRSTRLEN] = {'\0'};
	char field_str[FIELD_MAX_LEN] = {'\0'};
	char protocol_str[PROTOCOL_MAX_LEN] = {'\0'};
	char initiator[4] = {'\0'};
	struct doca_dpi_parsing_info *parsing_info = &sw_entry->dir_ctx->flow_ctx->parsing_info;

	if (sw_entry->initiator)
		strcpy(initiator, "cts");
	else
		strcpy(initiator, "stc");

	for (i = 0 ; i < ARRAY_SIZE(field_to_str) ; i++) {
		if (protocol == field_to_str[i].protocol_num &&
			field == field_to_str[i].field_num) {
			strlcpy(field_str, field_to_str[i].name, sizeof(field_str));
		}
	}
	if (!strlen(field_str))
		strcpy(field_str, "stream");
	for (i = 0 ; i < ARRAY_SIZE(protocol_to_str) ; i++) {
		if (protocol == protocol_to_str[i].protocol_num) {
			assert(strlen(protocol_to_str[i].name) < PROTOCOL_MAX_LEN);
			strlcpy(protocol_str, protocol_to_str[i].name, sizeof(protocol_str));
		}
	}
	if (!strlen(protocol_str))
		strcpy(protocol_str, "l4");
	if (rte_be_to_cpu_16(parsing_info->ethertype) == DPI_IPV4) {
		ipv4_to_ipv6(inet_ntoa(parsing_info->src_ip.ipv4), ipv4_src_addr);
		ipv4_to_ipv6(inet_ntoa(parsing_info->dst_ip.ipv4), ipv4_dst_addr);
		sprintf(meta, "&&&&&%s&%s&%d&%s&%d&%s&%d&%s&%d",
			field_str, protocol_str, parsing_info->l4_protocol, initiator,
			rte_be_to_cpu_16(parsing_info->l4_dport), ipv4_dst_addr,
			rte_be_to_cpu_16(parsing_info->l4_sport), ipv4_src_addr,
			rte_be_to_cpu_16(parsing_info->ethertype));
	} else {
		ipv6_expander(ipv6_src_addr, &parsing_info->src_ip.ipv6);
		ipv6_expander(ipv6_dst_addr, &parsing_info->dst_ip.ipv6);
		sprintf(meta, "&&&&&%s&%s&%d&%s&%d&%s&%d&%s&%d",
			field_str, protocol_str, parsing_info->l4_protocol, initiator,
			rte_be_to_cpu_16(parsing_info->l4_dport), ipv6_dst_addr,
			rte_be_to_cpu_16(parsing_info->l4_sport), ipv6_src_addr,
			rte_be_to_cpu_16(parsing_info->ethertype));
	}
}

/*
 * Post a job to the regex HW using chained mbufs, returns 1 on success
 */
static int
regex_queue_submit_mbuf_job(struct dpi_sw_queue_entry *sw_entry, uint16_t protocol,
		       uint16_t field, struct rte_mbuf *mbuf, bool stream_start)
{
	struct dpi_regex_entry_meta *regex_entry;
	int groups[4] = {1, 0, 0, 0};
	int ret;
	char meta[META_MAX_LEN];

	if (rte_ring_dequeue(sw_entry->regex_queue->free_list, (void **)&regex_entry))
		return 0;

	regex_entry->sw_entry = sw_entry;
	regex_entry->protocol = protocol;
	regex_entry->len = rte_pktmbuf_pkt_len(mbuf);
	regex_entry->field = field;
	regex_entry->stream_start = stream_start;

	update_meta(meta, protocol, field, sw_entry);
	ret = dpi_regex_enqueue_mbuf(sw_entry->dir_ctx->flow_ctx->dpi_q_ctx->regex_queue_ctx, mbuf,
				     groups, regex_entry, meta);
	if (ret)
		sw_entry->outstanding_regex_jobs++;
	else
		rte_ring_enqueue(sw_entry->regex_queue->free_list, (void *)regex_entry);
	return ret;
}

/*
 * Post a job to the regex HW using only one mbuf, returns 1 on success
 */
static int
regex_queue_submit_job(struct dpi_sw_queue_entry *sw_entry, uint16_t protocol,
		       uint16_t field, void *buf, uint16_t len, bool stream_start)
{
	struct dpi_regex_entry_meta *regex_entry;
	int groups[4] = {1, 0, 0, 0};
	int ret;
	char meta[META_MAX_LEN];

	if (rte_ring_dequeue(sw_entry->regex_queue->free_list, (void **)&regex_entry))
		return 0;

	regex_entry->sw_entry = sw_entry;
	regex_entry->protocol = protocol;
	regex_entry->len = len;
	regex_entry->field = field;
	regex_entry->stream_start = stream_start;
	update_meta(meta, protocol, field, sw_entry);
	ret = dpi_regex_enqueue_buf(sw_entry->dir_ctx->flow_ctx->dpi_q_ctx->regex_queue_ctx, buf,
				    len, groups, regex_entry, meta);
	if (ret)
		sw_entry->outstanding_regex_jobs++;
	else
		rte_ring_enqueue(sw_entry->regex_queue->free_list, (void *)regex_entry);
	return ret;
}

static void
parser_field_cb(void *user_data, void *start_or_mbuf,
		uint16_t len, const struct dpi_parser_cb_metadata *meta)
{
	struct dpi_sw_queue_entry *sw_entry = (typeof(sw_entry))user_data;
	struct dpi_sw_queue_entry *dequeued_sw_entry;

	DOCA_DLOG_DBG("PARSER_CB:	Found protocol=%d, field=%d len=%d", meta->prot, meta->field, len);
	if (meta->mbuf)
		while (!regex_queue_submit_mbuf_job(sw_entry, meta->prot, meta->field,
						    (struct rte_mbuf *)start_or_mbuf, meta->stream_start)) {
			dequeued_sw_entry = sw_queue_dequeue(sw_entry->dir_ctx->flow_ctx->dpi_q_ctx);
			if (dequeued_sw_entry && (dequeued_sw_entry != sw_entry))
				rte_ring_enqueue(sw_entry->dir_ctx->flow_ctx->dpi_q_ctx->sw_queue.done_list,
						 (void *)dequeued_sw_entry);
		}
	else
		while (!regex_queue_submit_job(sw_entry, meta->prot, meta->field, start_or_mbuf, len,
					       meta->stream_start)) {
			dequeued_sw_entry = sw_queue_dequeue(sw_entry->dir_ctx->flow_ctx->dpi_q_ctx);
			if (dequeued_sw_entry && (dequeued_sw_entry != sw_entry))
				rte_ring_enqueue(sw_entry->dir_ctx->flow_ctx->dpi_q_ctx->sw_queue.done_list,
						 (void *)dequeued_sw_entry);
		}
}

/*
 * Returns 1 if job submitted
 */
static void
sw_queue_enqueue(struct dpi_queue_context *dpi_queue,
		 struct flow_dir *dir_ctx, struct rte_mbuf *mbuf, void *user_data,
		 uint32_t payload_offset)
{
	struct dpi_sw_queue_entry *sw_entry = NULL;
#ifdef NDEBUG
	int ret __attribute__((unused));
#else
	int ret;
#endif

	ret = rte_ring_dequeue(dpi_queue->sw_queue.free_list, (void **)&sw_entry);
	assert(!ret && sw_entry != NULL);

	sw_entry->mbuf = mbuf;
	sw_entry->dir_ctx = dir_ctx;
	sw_entry->user_data = user_data;
	sw_entry->initiator = is_initiator(dir_ctx);
	sw_entry->payload_offset = payload_offset;
	dir_ctx->total_payload_parsed += dpi_parser_parse(dpi_queue->dpi_q, sw_entry->dir_ctx->p_ctx,
							  dir_ctx->flow_chain, dir_ctx->first_mbuf_offset,
							  parser_field_cb, sw_entry);
	dpi_mbuf_unchain(dir_ctx, sw_entry);
}

static int
setup_dpi_queue(struct dpi_queue_context *dpi_q_ctx)
{
	char name[100];
	int res, j;

	sprintf(name, "queue_free_list%d", dpi_q_ctx->dpi_q);
	dpi_q_ctx->sw_queue.free_list =
		rte_ring_create(name, DPI_MAX_SW_QUEUE_SIZE, rte_socket_id(),
				RING_F_SP_ENQ | RING_F_SC_DEQ);
	if (dpi_q_ctx->sw_queue.free_list == NULL)
		return -ENOMEM;
	sprintf(name, "sw_queue_done_list%d", dpi_q_ctx->dpi_q);
	dpi_q_ctx->sw_queue.done_list =
		rte_ring_create(name, DPI_MAX_SW_QUEUE_SIZE, rte_socket_id(),
				RING_F_SP_ENQ | RING_F_SC_DEQ);
	if (dpi_q_ctx->sw_queue.done_list == NULL) {
		res = -ENOMEM;
		goto sw_free_q_err;
	}
	sprintf(name, "regex_free_list%d", dpi_q_ctx->dpi_q);
	dpi_q_ctx->regex_queue.free_list =
		rte_ring_create(name, DPI_MAX_HW_QUEUE_SIZE, rte_socket_id(),
				RING_F_SP_ENQ | RING_F_SC_DEQ);
	if (dpi_q_ctx->regex_queue.free_list == NULL) {
		res = -ENOMEM;
		goto regex_free_q_err;
	}
	for (j = 0; j < DPI_MAX_HW_QUEUE_SIZE; j++)
		rte_ring_enqueue(dpi_q_ctx->regex_queue.free_list, &dpi_q_ctx->regex_queue.entries[j]);
	for (j = 0; j < DPI_MAX_SW_QUEUE_SIZE; j++)
		rte_ring_enqueue(dpi_q_ctx->sw_queue.free_list, &dpi_q_ctx->sw_queue.entries[j]);

	sprintf(name, "queue_ind_mbuf%d", dpi_q_ctx->dpi_q);
	dpi_q_ctx->ind_pkts =
		rte_pktmbuf_pool_create(name, dpi_q_ctx->dpi_ctx->cfg.max_packets_per_queue,
					0, RTE_MBUF_PRIV_ALIGN,
					0, rte_socket_id());
	if (!dpi_q_ctx->ind_pkts) {
		res = -ENOMEM;
		DOCA_DLOG_ERR("Failed to setup ind buffers pool, please adjust max_flows_per_queue or add more huge pages.");
		goto ind_pkt_err;
	}
	dpi_q_ctx->parser_ctx = dpi_parser_init(dpi_q_ctx->dpi_ctx->cfg.max_sig_match_len);
	if (!dpi_q_ctx->parser_ctx) {
		res = -ENOMEM;
		goto parser_init_err;
	}
	dpi_q_ctx->regex_queue.hw_queue_id = dpi_q_ctx->dpi_q;
	dpi_q_ctx->sw_queue.nb_entries = DPI_MAX_SW_QUEUE_SIZE;
	for (j = 0; j < DPI_MAX_SW_QUEUE_SIZE; j++)
		dpi_q_ctx->sw_queue.entries[j].regex_queue = &dpi_q_ctx->regex_queue;
	rte_spinlock_init(&dpi_q_ctx->db_lock);
	dpi_q_ctx->regex_queue_ctx =
		dpi_regex_queue_alloc(dpi_q_ctx->dpi_ctx->regex_dev,
				      dpi_q_ctx->dpi_q, DPI_MAX_HW_QUEUE_SIZE);
	if (!dpi_q_ctx->regex_queue_ctx) {
		res = -ENOMEM;
		DOCA_DLOG_ERR("Failed to setup regex queue, please add more huge pages.");
		goto parser_init_err;
	}
	return 0;

parser_init_err:
	rte_mempool_free(dpi_q_ctx->ind_pkts);
ind_pkt_err:
	rte_ring_free(dpi_q_ctx->regex_queue.free_list);
regex_free_q_err:
	rte_ring_free(dpi_q_ctx->sw_queue.done_list);
sw_free_q_err:
	rte_ring_free(dpi_q_ctx->sw_queue.free_list);
	return res;
}

static void
close_dpi_queue(struct dpi_queue_context *dpi_q_ctx)
{
	dpi_regex_queue_dealloc(dpi_q_ctx->regex_queue_ctx);
	if (dpi_q_ctx->ind_pkts)
		rte_mempool_free(dpi_q_ctx->ind_pkts);

	if (dpi_q_ctx->regex_queue.free_list)
		rte_ring_free(dpi_q_ctx->regex_queue.free_list);
	if (dpi_q_ctx->sw_queue.done_list)
		rte_ring_free(dpi_q_ctx->sw_queue.done_list);
	if (dpi_q_ctx->sw_queue.free_list)
		rte_ring_free(dpi_q_ctx->sw_queue.free_list);
	if (dpi_q_ctx->parser_ctx)
		dpi_parser_destroy(dpi_q_ctx->parser_ctx);
}

static void
update_counters(struct dpi_queue_context *dpi_q_ctx, int type)
{
	dpi_q_ctx->counters[type]++;
}

struct doca_dpi_ctx *
doca_dpi_init(const struct doca_dpi_config_t *config, int *error)
{
	int i, j, res;
	struct doca_dpi_ctx *dpi_ctx;

	if (RTE_VER_YEAR < 20 ||
	    (RTE_VER_YEAR == 20 && RTE_VER_MONTH < 8)) {
		DOCA_DLOG_ERR("DPDK version must be >= 20.08");
		*error = -ENOTSUP;
		return NULL;
	}
	if (config->max_sig_match_len > DPI_MAX_SIG_MATCH_LEN) {
		DOCA_DLOG_ERR("max_sig_match_len must be <= 5000");
		*error = -EINVAL;
		return NULL;
	}
	if (config->max_packets_per_queue == 0) {
		DOCA_DLOG_ERR("max_packets_per_queue must be > 0");
		*error = -EINVAL;
		return NULL;
	}
	if (config->nb_queues == 0) {
		DOCA_DLOG_ERR("nb_queues must be > 0");
		*error = -EINVAL;
		return NULL;
	}

	dpi_ctx = rte_zmalloc(NULL, sizeof(*dpi_ctx), 0);
	if (!dpi_ctx) {
		*error = -ENOMEM;
		return NULL;
	}

	dpi_ctx->regex_dev = dpi_regex_dev_alloc(REGEX_DEV_ID, config->nb_queues, error);
	if (dpi_ctx->regex_dev == NULL)
		goto regex_cfg_err;
	dpi_ctx->cfg = *config;
	dpi_ctx->dpi_q = rte_calloc(NULL, config->nb_queues,
				    sizeof(struct dpi_queue_context), 0);
	if (!dpi_ctx->dpi_q) {
		*error = -ENOMEM;
		goto dpi_q_err;
	}
	for (i = 0; i < config->nb_queues; i++) {
		dpi_ctx->dpi_q[i].dpi_q = i;
		dpi_ctx->dpi_q[i].dpi_ctx = dpi_ctx;
		res = setup_dpi_queue(&dpi_ctx->dpi_q[i]);
		if (res) {
			DOCA_DLOG_ERR("Failed to alloc DPI queues");
			*error = res;
			goto setup_q_err;
		}
	}
	dpi_ctx->dns_cache = dns_cache_create(64);
	if (!dpi_ctx->dns_cache) {
		DOCA_DLOG_ERR("Failed to create DNS cache");
		*error = -ENOMEM;
		goto dns_cache_err;
	}
	dpi_ctx->ftp_cache = ftp_cache_create(64);
	if (!dpi_ctx->ftp_cache) {
		DOCA_DLOG_ERR("Failed to create FTP cache");
		*error = -ENOMEM;
		goto ftp_cache_err;
	}
	return dpi_ctx;

ftp_cache_err:
	dns_cache_free(dpi_ctx->dns_cache);

dns_cache_err:

setup_q_err:
	for (j = 0; j < i; j++) {
		dpi_regex_queue_dealloc(dpi_ctx->dpi_q[j].regex_queue_ctx);
		rte_mempool_free(dpi_ctx->dpi_q[j].ind_pkts);
		rte_ring_free(dpi_ctx->dpi_q[j].regex_queue.free_list);
		rte_ring_free(dpi_ctx->dpi_q[j].sw_queue.done_list);
		rte_ring_free(dpi_ctx->dpi_q[j].sw_queue.free_list);
	}
	rte_free(dpi_ctx->dpi_q);
dpi_q_err:
	dpi_regex_dev_dealloc(dpi_ctx->regex_dev);
regex_cfg_err:
	rte_free(dpi_ctx);
	return NULL;
}

int
doca_dpi_load_signatures(struct doca_dpi_ctx *dpi_ctx, const char *cdofile)
{
	uint32_t rxp_len = 0, hs_len = 0;
	char *rxp_buf = NULL;
	char *hs_buf = NULL;
	int i, ret = 0;
	struct dpi_cdo_data *cdo_data;
	struct dpi_cdo_data *old_data = dpi_ctx->dpi_q[0].cdo_data;

	ret = cdo_decode(cdofile, &cdo_data, &rxp_buf, &rxp_len, &hs_buf, &hs_len);
	if (ret)
		return ret;
	DOCA_DLOG_DBG("INIT:	Loading signatures to HW...");
	for (i = 0; i < dpi_ctx->cfg.nb_queues; i++) {
		rte_spinlock_lock(&dpi_ctx->dpi_q[i].db_lock);
		dpi_ctx->dpi_q[i].cdo_data = cdo_data;
		rte_spinlock_unlock(&dpi_ctx->dpi_q[i].db_lock);
	}
	if (rxp_len) {
		ret = dpi_regex_load_sig(dpi_ctx->regex_dev, DPI_REGEX_TYPE_RXP, rxp_buf, rxp_len);
		if (ret < 0)
			goto load_out;
	}
	if (hs_len) {
		ret = dpi_regex_load_sig(dpi_ctx->regex_dev, DPI_REGEX_TYPE_HS, hs_buf, hs_len);
		if (ret < 0)
			goto load_out;
	}
load_out:
	if (hs_buf)
		free(hs_buf);
	if (rxp_buf)
		free(rxp_buf);
	if (old_data)
		cdo_dealloc(old_data);
	return ret;
}

static int dpi_process_packet_enqueue(
	struct rte_mbuf *dpkt,
	struct doca_dpi_flow_ctx *flow_ctx, bool initiator,
	uint32_t payload_offset, void *user_data)
{
	struct flow_dir *dir_obj;
	struct rte_mbuf *pkt;

	dir_obj = &flow_ctx->dir_obj[initiator];
	pkt = rte_pktmbuf_clone(dpkt, flow_ctx->dpi_q_ctx->ind_pkts);
	if (unlikely(!pkt))
		return -ENOMEM;
	rte_pktmbuf_adj(pkt, payload_offset);
	if (unlikely(dir_obj->flow_chain == NULL))
		dir_obj->flow_chain = pkt;
	else
		rte_pktmbuf_chain(dir_obj->flow_chain, pkt);

	sw_queue_enqueue(flow_ctx->dpi_q_ctx, dir_obj, dpkt, user_data, payload_offset);
	update_counters(flow_ctx->dpi_q_ctx, DPI_CNT_SCANNED);
	flow_ctx->currently_in_queue++;
	return DOCA_DPI_ENQ_PROCESSING;
}

/*
 * Return 1 if found all sig_ids in condition list.
 */
static int
update_and_check_conditions(const struct dpi_cdo_data *data,
			    struct doca_dpi_flow_ctx *flow_ctx,
			    const struct pp_table_entry *pp_entry)
{
	const struct condition_table_entry *cond_table_entry;
	struct cond_match *cond = NULL;
	struct cond_match *new_cond;
	int ret;
	int i;

	if (!entry_valid(pp_entry, DPI_PP_COND))
		return 1;
	cond_table_entry = cdo_cond_entry_get(data, pp_entry->cond_id);
	LIST_FOREACH(cond, &flow_ctx->cond_match_list, matches) {
		if (cond->rule_id == pp_entry->rule_id) {
			ret = 0;
			goto out;
		}
		if (cond->rule_id > pp_entry->rule_id)
			break;
	}
	new_cond = rte_zmalloc(NULL, sizeof(*new_cond), 0);
	if (!new_cond) {
		DOCA_DLOG_ERR("Failed to allocate memory for new rule");
		ret = -ENOMEM;
		goto out;
	}
	new_cond->rule_id = pp_entry->rule_id;
	DOCA_DLOG_DBG("POST:	Add new cond rule %d", pp_entry->rule_id);
	if (cond != NULL)
		LIST_INSERT_BEFORE(cond, new_cond, matches);
	else
		LIST_INSERT_HEAD(&flow_ctx->cond_match_list, new_cond, matches);

	for (i = 0; i < cond_table_entry->nb_cond; i++) {
		int matched = 0;

		LIST_FOREACH(cond, &flow_ctx->cond_match_list, matches) {
			if (cond_table_entry->conds[i] == cond->rule_id) {
				matched = 1;
				break;
			}
		}
		if (!matched) {
			ret = 0;
			goto out;
		}
	}
	ret = 1;
out:
	return ret;
}

static int
dpi_init_directions(uint16_t dpi_q, struct flow_dir *dir_obj1, struct flow_dir *dir_obj2,
		   const struct doca_dpi_parsing_info *info)
{
	int ret;

	ret = dpi_parser_ctx_alloc(dir_obj1->flow_ctx->dpi_q_ctx->parser_ctx, &dir_obj1->p_ctx,
				    &dir_obj2->p_ctx, rte_be_to_cpu_16(info->l4_dport),
				    info->l4_protocol);
	if (ret)
		return ret;

	dir_obj1->dpi_q = dpi_q;
	dir_obj2->dpi_q = dpi_q;
	return 0;
}

static void
dpi_terminate_direction(struct flow_dir *dir_obj)
{
	dpi_parser_ctx_dealloc(dir_obj->p_ctx);
	if (dir_obj->flow_chain) {
		rte_pktmbuf_free(dir_obj->flow_chain);
		dir_obj->flow_chain = NULL;
	}
}

struct doca_dpi_flow_ctx *
doca_dpi_flow_create(struct doca_dpi_ctx *dpi_ctx,
		     uint16_t dpi_q,
		     const struct doca_dpi_parsing_info *parsing_info,
		     int *error,
		     struct doca_dpi_result *result)
{
	struct doca_dpi_flow_ctx *flow_ctx;
	int ret;

	*error = 0;
	if (unlikely(dpi_ctx->cfg.nb_queues <= dpi_q)) {
		*error = -EINVAL;
		DOCA_DLOG_ERR("Trying to create flow for invalid queue");
		return NULL;
	}

	flow_ctx = rte_zmalloc(__func__, sizeof(*flow_ctx), 0);
	if (unlikely(!flow_ctx)) {
		*error = -ENOMEM;
		return NULL;
	}
	flow_ctx->dpi_q_ctx = &dpi_ctx->dpi_q[dpi_q];
	flow_ctx->dir_obj[DPI_FLOW_SERVER].flow_ctx = flow_ctx;
	flow_ctx->dir_obj[DPI_FLOW_CLIENT].flow_ctx = flow_ctx;
	ret = dpi_init_directions(dpi_q, &flow_ctx->dir_obj[DPI_FLOW_CLIENT],
				  &flow_ctx->dir_obj[DPI_FLOW_SERVER], parsing_info);
	if (ret)
		goto parser_err;

	flow_ctx->parsing_info = *parsing_info;
	LIST_INIT(&flow_ctx->cond_match_list);

	if (flow_ctx->dpi_q_ctx->cdo_data) {

		ret = dns_cache_ip_lookup(dpi_ctx->dns_cache, &(parsing_info->dst_ip.ipv4));
		if (ret >= 0) {
			flow_ctx->matched = true;
			flow_ctx->sig_info.pp_entry.sig_id = ret;
			DOCA_DLOG_DBG("DNS:	DNS cache matched created flow");
		} else if (ret == -EINVAL) {
			goto dns_ftp_err;
		}

		ret = ftp_cache_ip_lookup(dpi_ctx->ftp_cache, &(parsing_info->dst_ip.ipv4),
					  rte_be_to_cpu_16(parsing_info->l4_dport));
		if (ret >= 0) {
			flow_ctx->matched = true;
			flow_ctx->sig_info.pp_entry.sig_id = ret;
			DOCA_DLOG_DBG("FTP:	FTP cache matched created flow");
		} else if (ret == -EINVAL) {
			goto dns_ftp_err;
		}
	}
	result->matched = flow_ctx->matched;
	result->info.sig_id = flow_ctx->sig_info.pp_entry.sig_id;
	result->info.action = result->matched == false ? DOCA_DPI_SIG_ACTION_NA
						       : flow_ctx->sig_info.pp_entry.action;
	return flow_ctx;

dns_ftp_err:
	dpi_terminate_direction(&(flow_ctx)->dir_obj[DPI_FLOW_SERVER]);
	dpi_terminate_direction(&(flow_ctx)->dir_obj[DPI_FLOW_CLIENT]);
parser_err:
	rte_free(flow_ctx);
	*error = ret;
	return NULL;
}

static void
dpi_flow_ctx_free(struct doca_dpi_flow_ctx *flow_ctx)
{
	struct cond_match *cond;

	dpi_terminate_direction(&flow_ctx->dir_obj[DPI_FLOW_CLIENT]);
	dpi_terminate_direction(&flow_ctx->dir_obj[DPI_FLOW_SERVER]);
	while (!LIST_EMPTY(&flow_ctx->cond_match_list)) {
		cond = LIST_FIRST(&flow_ctx->cond_match_list);
		LIST_REMOVE(cond, matches);
		rte_free(cond);
	}
	rte_free(flow_ctx);
}

void
doca_dpi_flow_destroy(struct doca_dpi_flow_ctx *flow_ctx)
{
	if (flow_ctx->currently_in_queue == 0)
		dpi_flow_ctx_free(flow_ctx);
	else
		flow_ctx->destroyed = true;
}

int
doca_dpi_enqueue(struct doca_dpi_flow_ctx *flow_ctx, struct rte_mbuf *pkt,
		 bool initiator, uint32_t payload_offset, void *user_data)
{
	if (!flow_ctx->dpi_q_ctx->cdo_data)
		return DOCA_DPI_ENQ_INVALID_DB;
	if (flow_ctx->bypass_state)
		return flow_ctx->bypass_state;
	if (rte_pktmbuf_pkt_len(pkt) <= payload_offset)
		return DOCA_DPI_ENQ_PACKET_EMPTY;
	if (rte_ring_empty(flow_ctx->dpi_q_ctx->sw_queue.free_list))
		return DOCA_DPI_ENQ_BUSY;
	return dpi_process_packet_enqueue(pkt, flow_ctx, initiator,
					  payload_offset, user_data);
}

int
doca_dpi_flow_match_get(const struct doca_dpi_flow_ctx *flow_ctx,
			struct doca_dpi_result *result)
{
	result->matched = false;
	if (flow_ctx->matched) {
		result->matched = true;
		result->info.sig_id = flow_ctx->sig_info.pp_entry.sig_id;
		result->info.action = flow_ctx->sig_info.pp_entry.action;
	}
	return 0;
}

static void
collect_stats(struct dpi_queue_context *dpi_q_ctx, struct pp_table_entry *pp_entry)
{
	int protocol;

	if (entry_valid(pp_entry, DPI_PP_L7_PROTOCOL)) {
		protocol = pp_entry->l7_protocol;
		switch (protocol) {
		case DPI_PARSER_HTTP_1:
			update_counters(dpi_q_ctx, DPI_CNT_HTTP);
			break;
		case DPI_PARSER_SSL:
			update_counters(dpi_q_ctx, DPI_CNT_SSL);
			break;
		default:
			update_counters(dpi_q_ctx, DPI_CNT_OTHER_L7);
		}
	} else if (entry_valid(pp_entry, DPI_PP_L4_PROTOCOL)) {
		protocol = pp_entry->l4_protocol;
		switch (protocol) {
		case DPI_PARSER_UDP:
			update_counters(dpi_q_ctx, DPI_CNT_UDP);
			break;
		case DPI_PARSER_TCP:
			update_counters(dpi_q_ctx, DPI_CNT_TCP);
			break;
		default:
			update_counters(dpi_q_ctx, DPI_CNT_OTHER_L4);
		}
	}
	update_counters(dpi_q_ctx, DPI_CNT_MATCHES);
}

int
doca_dpi_dequeue(struct doca_dpi_ctx *dpi_ctx, uint16_t dpi_q, struct doca_dpi_result *result)
{
	struct dpi_sw_queue_entry *sw_entry;
	struct doca_dpi_flow_ctx *flow_ctx;
	struct flow_dir *dir_ctx;
	int ret;

	if (unlikely(dpi_ctx->cfg.nb_queues <= dpi_q)) {
		DOCA_DLOG_ERR("Trying to dequeue for invalid queue");
		return -EINVAL;
	}

	ret = rte_ring_dequeue(dpi_ctx->dpi_q[dpi_q].sw_queue.done_list, (void **)&sw_entry);
	if (ret)
		sw_entry = sw_queue_dequeue(&dpi_ctx->dpi_q[dpi_q]);
	if (!sw_entry)
		return DOCA_DPI_DEQ_NA;

	rte_ring_enqueue(dpi_ctx->dpi_q[dpi_q].sw_queue.free_list, sw_entry);
	result->status_flags = 0;
	result->pkt = sw_entry->mbuf;
	dir_ctx = sw_entry->dir_ctx;
	flow_ctx = dir_ctx->flow_ctx;
	flow_ctx->currently_in_queue--;
	result->user_data = sw_entry->user_data;
	result->matched = dir_ctx->flow_ctx->matched;
	result->info.sig_id = dir_ctx->flow_ctx->sig_info.pp_entry.sig_id;
	result->info.action = dir_ctx->flow_ctx->sig_info.pp_entry.action;
	if (flow_ctx->currently_in_queue == 0)
		result->status_flags |= DOCA_DPI_STATUS_LAST_PACKET;

	if (result->matched && !dir_ctx->flow_ctx->counted) {
		collect_stats(dir_ctx->flow_ctx->dpi_q_ctx, &(dir_ctx->flow_ctx->sig_info.pp_entry));
		dir_ctx->flow_ctx->counted = true;
		result->status_flags |= DOCA_DPI_STATUS_NEW_MATCH;
		if (flow_ctx->parsing_info.l4_dport == rte_cpu_to_be_16(DNS_PORT)) {
			ret = dpi_parser_update_dns_cache(sw_entry->mbuf, sw_entry->payload_offset,
							  dpi_ctx->dns_cache,
							  flow_ctx->sig_info.pp_entry.sig_id);
			if (ret == 0)
				DOCA_DLOG_DBG("DNS:	DNS cache updated on match");
		}

		if (flow_ctx->parsing_info.l4_dport == rte_cpu_to_be_16(FTP_PORT)) {
			ret = dpi_parser_update_ftp_cache(sw_entry->mbuf, sw_entry->payload_offset,
							  dpi_ctx->ftp_cache,
							  flow_ctx->sig_info.pp_entry.sig_id);
			if (ret == 0)
				DOCA_DLOG_DBG("FTP:	FTP cache updated on match");
		}
	}

	result->matched = dir_ctx->flow_ctx->matched;
	result->info.sig_id = dir_ctx->flow_ctx->sig_info.pp_entry.sig_id;
	result->info.action = dir_ctx->flow_ctx->sig_info.pp_entry.action;
	if (flow_ctx->destroyed) {
		result->status_flags |= DOCA_DPI_STATUS_DESTROYED;
		if (flow_ctx->currently_in_queue == 0)
			dpi_flow_ctx_free(flow_ctx);
	}
	return DOCA_DPI_DEQ_READY;
}

void
doca_dpi_destroy(struct doca_dpi_ctx *dpi_ctx)
{
	int i;

	if (dpi_ctx->cfg.nb_queues && dpi_ctx->dpi_q[0].cdo_data)
		cdo_dealloc(dpi_ctx->dpi_q[0].cdo_data);
	for (i = 0; i < dpi_ctx->cfg.nb_queues; i++)
		close_dpi_queue(&dpi_ctx->dpi_q[i]);
	if (dpi_ctx->dpi_q)
		rte_free(dpi_ctx->dpi_q);
	if (dpi_ctx->regex_dev)
		dpi_regex_dev_dealloc(dpi_ctx->regex_dev);
	if (dpi_ctx->dns_cache)
		dns_cache_free(dpi_ctx->dns_cache);
	if (dpi_ctx->ftp_cache)
		ftp_cache_free(dpi_ctx->ftp_cache);
	rte_free(dpi_ctx);
}

int
doca_dpi_signature_get(const struct doca_dpi_ctx *dpi_ctx, uint32_t sig_id, struct doca_dpi_sig_data *sig_data)
{
	return cdo_signature_get(dpi_ctx->dpi_q[0].cdo_data, sig_id, sig_data);
}

int
doca_dpi_signatures_get(const struct doca_dpi_ctx *dpi_ctx, struct doca_dpi_sig_data **sig_data)
{
	return cdo_signatures_get(dpi_ctx->dpi_q[0].cdo_data, sig_data);
}

void
doca_dpi_stat_get(const struct doca_dpi_ctx *dpi_ctx, bool clear, struct doca_dpi_stat_info *stats)
{
	int i;

	memset(stats, 0, sizeof(*stats));
	for (i = 0; i < dpi_ctx->cfg.nb_queues; i++) {
		stats->nb_matches += dpi_ctx->dpi_q[i].counters[DPI_CNT_MATCHES];
		stats->nb_other_l4 += dpi_ctx->dpi_q[i].counters[DPI_CNT_OTHER_L4];
		stats->nb_other_l7 += dpi_ctx->dpi_q[i].counters[DPI_CNT_OTHER_L7];
		stats->nb_ssl_parser_based += dpi_ctx->dpi_q[i].counters[DPI_CNT_SSL];
		stats->nb_tcp_based += dpi_ctx->dpi_q[i].counters[DPI_CNT_TCP];
		stats->nb_udp_based += dpi_ctx->dpi_q[i].counters[DPI_CNT_UDP];
		stats->nb_http_parser_based += dpi_ctx->dpi_q[i].counters[DPI_CNT_HTTP];
		stats->nb_scanned_pkts += dpi_ctx->dpi_q[i].counters[DPI_CNT_SCANNED];
		if (clear) {
			int j;

			for (j = 0; j < DPI_MAX_COUNTERS; j++)
				dpi_ctx->dpi_q[i].counters[j] = 0;
		}
	}

}

#ifdef DOCA_DPI_USE_HS
#include <hs_compat.c>
#endif
