/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <fstream>
#include <string.h>
#include <stdlib.h>
#include <bsd/string.h>
#include <endian.h>

#include <grpcpp/grpcpp.h>

#include <doca_log.h>

#include "doca_dpi_grpc_client.h"
#include "common.pb.h"
#include "common.grpc.pb.h"
#include "doca_dpi.pb.h"
#include "doca_dpi.grpc.pb.h"

DOCA_LOG_REGISTER(DPI::GRPC::CLIENT);

#define LIB_NAME "DOCA DPI"
#define MOCK_DPI_CTX_PTR 0xffff69d8a900
#define ETHERTYPE_IPV4 0x0800

using grpc::Status;

struct doca_dpi_ctx {};
struct doca_dpi_flow_ctx {};

static uint32_t nb_microservices;
static uint32_t nb_dpi_queues;
static DocaDpi::Stub **stubs;
static bool doca_dpi_init_was_invoked_and_active = false;

#define get_error_code_and_log_errors(a,b) _get_error_code_and_log_errors(a,b,__func__)

/* Checks the grpc status to make sure grpc call was invoked properly, then check no internal DPI error occurred.
 * Returns a negative value on error and err.error_code otherwise. */
static int
_get_error_code_and_log_errors(Status& status, const DocaDpiErrorInfo& err, const char *func_name)
{
	if (!status.ok()) {
		DOCA_DLOG_ERR("%s rpc failed: %s", func_name, status.error_message().c_str());
		return -status.error_code();  /* Grpc statuses are positive */
	}

	return err.error_code();
}

static inline bool
is_server_online(std::shared_ptr<grpc::Channel> channel)
{
	grpc::ClientContext context;
	HealthCheckReq request;
	HealthCheckResp response;
	Status status;
	/* Create stub to ping server */
	DocaOrchestration::Stub orchestration_stub(channel);

	/* Ping the server */
	status = orchestration_stub.HealthCheck(&context, request, &response);

	/* Return true only if request arrived and returned "OK" */
	return status.ok();
}

static inline int
_init_remote_dpi_context(DocaDpi::Stub * stub, const struct doca_dpi_config_t *config)
{
	DOCA_DLOG_DBG(LIB_NAME " Init request");
	grpc::ClientContext context;
	DocaDpiInitParams request;
	DocaDpiInitResponse response;
	Status status;
	int ret_val;

	/* Set the request parameters */
	request.set_uint16_nb_queues(config->nb_queues);
	request.set_max_packets_per_queue(config->max_packets_per_queue);
	request.set_max_sig_match_len(config->max_sig_match_len);

	/* Submitting gRPC call */
	status = stub->DocaDpiInit(&context, request, &response);

	/* Check response is valid and take the number of remote active services */
	ret_val = get_error_code_and_log_errors(status, response.erreno());
	if (ret_val == 0) {
		nb_microservices = response.nb_microservices();
		/* If less cores than microservices are asked to be used, ignore extra microservices */
		if (nb_microservices > config->nb_queues)
			nb_microservices = config->nb_queues;
		nb_dpi_queues = config->nb_queues;
	}

	/* Return if gRPC and DOCA DPI succeeded */
	return ret_val;
}

static inline DocaDpi::Stub*
_construct_stub(std::string& base_ip, int port)
{
	DocaDpi::Stub* ret_val;

	/* Construct a channel connection to the server */
	std::string grpc_full_address = base_ip + ":" + std::to_string(port);
	std::shared_ptr<grpc::Channel> ch = grpc::CreateChannel(grpc_full_address, grpc::InsecureChannelCredentials());

	/* Check if service is active */
	if (!is_server_online(ch)) {
		DOCA_DLOG_ERR("Failed to connect to the gRPC server on the DPU");
		return nullptr;
	}

	/* Create a connection to the remote service */
	ret_val = new(std::nothrow) DocaDpi::Stub(ch);
	if (ret_val == nullptr)
		DOCA_DLOG_ERR("Failed to create client layer above the gRPC connection");

	return ret_val;
}

struct doca_dpi_ctx*
doca_dpi_grpc_init(const struct doca_dpi_config_t *config, int *error)
{
	DOCA_DLOG_INFO("Connecting to the " LIB_NAME " gRPC server on the DPU: %s", config->server_address);
	int i;
	std::string base_ip = std::string(config->server_address);
	size_t dots_pos = base_ip.find(':');
	int base_port = eNetworkPort::k_DocaDPI;
	DocaDpi::Stub *main_stub;
	uint32_t temp;

	/* Check valid state */
	if (doca_dpi_init_was_invoked_and_active) {
		/* Using log since DOCA DPI lib won't warn the user, but it will generate undefined behaviour.
		 * This is an init call so performace can be a bit slower. */
		DOCA_LOG_ERR("doca_dpi_init should be invoked once per process.");
	}
	doca_dpi_init_was_invoked_and_active = true;

	/* Check if we got a non-default port with the IP */
	if (dots_pos != std::string::npos) {
		base_port = std::stoi(base_ip.substr(dots_pos + 1));
		base_ip = base_ip.substr(0, dots_pos);
	}

	/* Call the remote init method - this also sets the global nb_microservices */
	main_stub = _construct_stub(base_ip, base_port);
	if (main_stub == nullptr) {
		*error = GENERAL_ERRORCODE;
		doca_dpi_init_was_invoked_and_active = false;
		return nullptr;
	}
	*error = _init_remote_dpi_context(main_stub, config);
	if (*error != 0)
	{
		delete main_stub;
		doca_dpi_init_was_invoked_and_active = false;
		return nullptr;
	}

	/* Create the Stubs array */
	stubs = new(std::nothrow) DocaDpi::Stub*[config->nb_queues];
	if (stubs == nullptr)
	{
		delete main_stub;
		doca_dpi_init_was_invoked_and_active = false;
		return nullptr;
	}

	/* Create a stub to each remote microservice (lcore-server) */
	stubs[0] = main_stub;
	for (i = 1; i < nb_microservices; ++i) {
		/* Create a connection to the remote service */
		stubs[i] = _construct_stub(base_ip, base_port + i);
		if (stubs[i] == nullptr) {
			*error = GENERAL_ERRORCODE;
			temp = nb_microservices;
			nb_microservices = i;
			doca_dpi_grpc_destroy(reinterpret_cast<struct doca_dpi_ctx *>(MOCK_DPI_CTX_PTR));
			nb_microservices = temp;
			return nullptr;
		}
	}

	/* Spread the use of the stubs equally between all client threads */
	for (; i < config->nb_queues; ++i) {
		stubs[i] = stubs[i % nb_microservices];
	}

	DOCA_DLOG_INFO("Connected to server.");

	return reinterpret_cast<struct doca_dpi_ctx *>(MOCK_DPI_CTX_PTR);
}

static void
assert_is_valid_doca_dpi_ptr(const struct doca_dpi_ctx *ctx)
{
	uintptr_t ctx_as_int = reinterpret_cast<uintptr_t>(ctx);

	/* If bad pointer, generate an error similar to an error user will get on the DPU.
	 * On the DPU it will crash, using "exit" as it is not possible to crash on purpose. */
	if (doca_dpi_init_was_invoked_and_active && ctx_as_int != MOCK_DPI_CTX_PTR) {
		DOCA_LOG_ERR("Bad doca_dpi_ctx, unable to read fields properly - might be corrupted.");
		exit(GENERAL_ERRORCODE);
	}
}

void
doca_dpi_grpc_destroy(struct doca_dpi_ctx* ctx)
{
	assert_is_valid_doca_dpi_ptr(ctx);
	int i;
	Status status;
	grpc::ClientContext context;
	DocaDpiDestroyParams request;
	DocaDpiDestroyResponse response;

	/* Invoke remote destroy */
	status = stubs[0]->DocaDpiDestroy(&context, request, &response);

	/* This deletes all stubs and their inner channel */
	for (i = 0; i < nb_microservices; ++i) {
		delete ((DocaDpi::Stub *) stubs[i]);
	}
	delete[] stubs;

	doca_dpi_init_was_invoked_and_active = false;
}

int
doca_dpi_grpc_load_signatures(struct doca_dpi_ctx *ctx, const char *path_to_cdo)
{
	DOCA_DLOG_DBG(LIB_NAME " Load signatures request");
	(void)ctx;
	grpc::ClientContext context;
	DocaDpiLoadSignaturesParams request;
	DocaDpiErrorInfo response;
	Status status;

	assert_is_valid_doca_dpi_ptr(ctx);

	/* Set the request parameters */
	request.set_cdo_filename(path_to_cdo);

	/* Submitting gRPC call */
	status = stubs[pthread_self() % nb_microservices]->DocaDpiLoadSignatures(&context, request, &response);

	/* Return if gRPC and DOCA DPI succeeded */
	return get_error_code_and_log_errors(status, response);
}

int
doca_dpi_grpc_enqueue(struct doca_dpi_flow_ctx *flow_ctx, struct doca_dpi_grpc_generic_packet *pkt, bool initiator,
		      uint32_t payload_offset, void *user_data, size_t user_data_len, uint16_t dpi_q)
{
	DOCA_DLOG_DBG(LIB_NAME " Enqueue request");
	grpc::ClientContext context;
	DocaDpiEnqueueParams request;
	DocaDpiErrorInfo response;
	Status status;
	uintptr_t ctx_as_int = reinterpret_cast<uintptr_t>(flow_ctx);

	if (GPR_UNLIKELY(dpi_q >= nb_dpi_queues)) {
		DOCA_DLOG_ERR("Got a non-existing DPI queue number: %u, ABORT action.", dpi_q);
		return GENERAL_ERRORCODE;
	}

	request.mutable_context()->set_unique_id(ctx_as_int);
	request.mutable_packet()->set_segment(pkt->segment, pkt->seg_len);
	request.set_initiator(initiator);
	request.set_payload_offset(payload_offset);

	if (user_data != nullptr)
		request.set_user_data(user_data, user_data_len);

	/* Submitting gRPC call */
	status = stubs[dpi_q]->DocaDpiEnqueue(&context, request, &response);

	return get_error_code_and_log_errors(status, response);
}

static int
convert_result(struct doca_dpi_grpc_result *dst, const DocaDpiResult& src)
{
	size_t data_size;

	dst->user_data = nullptr;
	dst->pkt = nullptr;

	dst->matched = src.matched();
	if (src.has_user_data()) {
		data_size = src.user_data().size();
		dst->user_data = malloc(data_size);
		if (dst->user_data == nullptr)
			goto malloc_failed;
		memcpy(dst->user_data, src.user_data().c_str(), data_size);
	}

	data_size = src.pkt().segment().size();
	if (data_size > 0) {
		dst->pkt = (struct doca_dpi_grpc_generic_packet*) malloc(sizeof(struct doca_dpi_grpc_generic_packet));
		if (dst->pkt == nullptr)
			goto malloc_failed;
		dst->pkt->seg_len = data_size;
		dst->pkt->segment = (uint8_t *) malloc(dst->pkt->seg_len);
		if (dst->pkt->segment == nullptr)
			goto malloc_failed;
		memcpy(dst->pkt->segment, src.pkt().segment().c_str(), dst->pkt->seg_len);
	}

	dst->info.sig_id = src.info().sig_id();
	dst->info.action = src.info().action();
	dst->status_flags = src.status_flags();
	return 0;

malloc_failed:
	if (dst->user_data != nullptr)
		free(dst->user_data);
	if (dst->pkt != nullptr)
		free(dst->pkt);
	DOCA_DLOG_ERR("Malloc failed, not enough memory.");
	return -ENOMEM;
}

int
doca_dpi_grpc_dequeue(struct doca_dpi_ctx *ctx, uint16_t dpi_queue, struct doca_dpi_grpc_result *result)
{
	DOCA_DLOG_DBG(LIB_NAME " Dequeue request");
	grpc::ClientContext context;
	DocaDpiDequeueParams request;
	DocaDpiActionResponse response;
	Status status;
	int ret_val;
	int convert_result_status;

	assert_is_valid_doca_dpi_ptr(ctx);

	if (GPR_UNLIKELY(dpi_queue >= nb_dpi_queues)) {
		DOCA_DLOG_ERR("Got a non-existing DPI queue number: %u, ABORT action.", dpi_queue);
		return GENERAL_ERRORCODE;
	}

	request.set_uint16_dpi_q(dpi_queue);

	/* Submitting gRPC call */
	status = stubs[dpi_queue]->DocaDpiDequeue(&context, request, &response);

	ret_val = get_error_code_and_log_errors(status, response.erreno());
	if (ret_val < 0 || ret_val == DOCA_DPI_DEQ_NA)
		return ret_val;

	convert_result_status = convert_result(result, response.result());
	if (convert_result_status < 0)
		return convert_result_status;

	return ret_val;
}

struct doca_dpi_flow_ctx *
doca_dpi_grpc_flow_create(struct doca_dpi_ctx *ctx, uint16_t dpi_q, const struct doca_dpi_parsing_info *parsing_info,
			  int *error, struct doca_dpi_grpc_result *result)
{
	DOCA_DLOG_DBG(LIB_NAME " Flow create request");
	grpc::ClientContext context;
	DocaDpiFlowCreateParams request;
	DocaDpiFlowCreateResponse response;
	Status status;
	DocaDpiParsingInfo *connection_info = request.mutable_parsing_info();

	assert_is_valid_doca_dpi_ptr(ctx);

	if (GPR_UNLIKELY(dpi_q >= nb_dpi_queues)) {
		DOCA_DLOG_ERR("Got a non-existing DPI queue number: %u, ABORT action.", dpi_q);
		*error = GENERAL_ERRORCODE;
		return nullptr;
	}

	/* Prepare request parameters */
	connection_info->set_ethertype(parsing_info->ethertype);
	connection_info->set_l4_protocol(parsing_info->l4_protocol);
	connection_info->set_l4_dst_port(parsing_info->l4_dport);
	connection_info->set_l4_src_port(parsing_info->l4_sport);
	if (be16toh(parsing_info->ethertype) == ETHERTYPE_IPV4) {
		connection_info->set_dst_ipv4(parsing_info->dst_ip.ipv4.s_addr);
		connection_info->set_src_ipv4(parsing_info->src_ip.ipv4.s_addr);
	}
	else {
		connection_info->mutable_dst_ipv6()->
			set_ipv6_addr(parsing_info->dst_ip.ipv6.__in6_u.__u6_addr8, IPV6_ADDER_LEN);
		connection_info->mutable_src_ipv6()->
			set_ipv6_addr(parsing_info->src_ip.ipv6.__in6_u.__u6_addr8, IPV6_ADDER_LEN);
	}
	request.set_uint16_dpi_q(dpi_q);

	/* Invoke gRPC call */
	status = stubs[dpi_q]->DocaDpiFlowCreate(&context, request, &response);

	*error = get_error_code_and_log_errors(status, response.erreno());
	if (*error < 0)
		return nullptr;

	/* Return results */
	/* This won't fail, since result of flow_create only includes result->matched and result->parsing_info .
	 * If you still think this should be checked and handled, know that you can't return 0 if this
	 * fails, it will cause flow context leaks on service.
	 */
	convert_result(result, response.result());
	return reinterpret_cast<struct doca_dpi_flow_ctx *>(response.flow_context().unique_id());
}

void
doca_dpi_grpc_flow_destroy(struct doca_dpi_flow_ctx *ctx, uint16_t dpi_q)
{
	DOCA_DLOG_DBG(LIB_NAME " Destroy flow context request");
	grpc::ClientContext context;
	DocaDpiFlowDestroyParams request;
	DocaDpiFlowDestroyResponse response;
	Status status;
	uintptr_t ctx_as_int = reinterpret_cast<uintptr_t>(ctx);

	if (GPR_UNLIKELY(dpi_q >= nb_dpi_queues)) {
		DOCA_DLOG_ERR("Got a non-existing DPI queue number: %u, ABORT action.", dpi_q);
		return;
	}

	request.mutable_context()->set_unique_id(ctx_as_int);

	/* Invoke gRPC call */
	status = stubs[dpi_q]->DocaDpiFlowDestroy(&context, request, &response);

	if (!status.ok())
		DOCA_DLOG_ERR("%s rpc failed: %s", __func__, status.error_message().c_str());
}

int
doca_dpi_grpc_flow_match_get(const struct doca_dpi_flow_ctx *flow_ctx, struct doca_dpi_grpc_result *result, uint16_t dpi_q)
{
	DOCA_DLOG_DBG(LIB_NAME " Flow match request");
	grpc::ClientContext context;
	DocaDpiFlowMatchGetParams request;
	DocaDpiActionResponse response;
	Status status;
	int err_code;
	uintptr_t ctx_as_int = reinterpret_cast<uintptr_t>(flow_ctx);

	if (GPR_UNLIKELY(dpi_q >= nb_dpi_queues)) {
		DOCA_DLOG_ERR("Got a non-existing DPI queue number: %u, ABORT action.", dpi_q);
		return GENERAL_ERRORCODE;
	}

	request.mutable_context()->set_unique_id(ctx_as_int);

	/* Submitting gRPC call */
	status = stubs[dpi_q]->DocaDpiFlowMatchGet(&context, request, &response);

	err_code = get_error_code_and_log_errors(status, response.erreno());
	if (err_code != 0)
		return err_code;

	return convert_result(result, response.result());
}

static inline void
convert_sig_data(struct doca_dpi_sig_data *dst, const DocaDpiSigData& src)
{
	dst->sig_id = src.sig_id();
	strlcpy(dst->name, src.name().c_str(), sizeof(dst->name));
}

int
doca_dpi_grpc_signature_get(const struct doca_dpi_ctx *ctx, uint32_t sig_id, struct doca_dpi_sig_data *sig_data)
{
	DOCA_DLOG_DBG(LIB_NAME " SignatureGet request");
	grpc::ClientContext context;
	DocaDpiSignatureGetParams request;
	DocaDpiSignatureGetResponse response;
	Status status;
	int err_code;

	assert_is_valid_doca_dpi_ptr(ctx);

	request.set_sign_id(sig_id);

	/* Submitting gRPC call */
	status = stubs[pthread_self() % nb_microservices]->DocaDpiSignatureGet(&context, request, &response);

	err_code = get_error_code_and_log_errors(status, response.erreno());
	if (err_code != 0)
		return err_code;

	convert_sig_data(sig_data, response.sign_data());
	return 0;
}

int
doca_dpi_grpc_signatures_get(const struct doca_dpi_ctx *ctx, struct doca_dpi_sig_data **sig_data)
{
	DOCA_DLOG_DBG(LIB_NAME " SignaturesGet request");
	grpc::ClientContext context;
	DocaDpiSignaturesGetParams request;
	DocaDpiSignaturesGetResponse response;
	Status status;
	int err_code;
	struct doca_dpi_sig_data *sig_arr;
	size_t num_signs;

	assert_is_valid_doca_dpi_ptr(ctx);

	/* Submitting gRPC call */
	status = stubs[pthread_self() % nb_microservices]->DocaDpiSignaturesGet(&context, request, &response);

	err_code = get_error_code_and_log_errors(status, response.erreno());
	if (err_code != 0) {
		*sig_data = nullptr;
		return err_code;
	}

	num_signs = response.signs_size();  /* Actual array size, not set by sender. */
	/* No int overflow can happen here since num_signs is the real size of the signs array sent from the server. */
	sig_arr = (struct doca_dpi_sig_data*) malloc(num_signs * sizeof(struct doca_dpi_sig_data));
	if (sig_arr == nullptr)
		return -ENOMEM;
	*sig_data = sig_arr;
	for (const DocaDpiSigData& sign : response.signs()) {
		convert_sig_data(sig_arr, sign);
		++sig_arr;
	}

	return num_signs;
}

void
doca_dpi_grpc_stat_get(const struct doca_dpi_ctx *ctx, bool clear, struct doca_dpi_stat_info *stats)
{
	DOCA_DLOG_DBG(LIB_NAME " StatGet request");
	grpc::ClientContext context;
	DocaDpiStatGetParams request;
	DocaDpiStatInfo response;
	Status status;

	assert_is_valid_doca_dpi_ptr(ctx);

	request.set_clear(clear);

	/* Submitting gRPC call */
	status = stubs[pthread_self() % nb_microservices]->DocaDpiStatGet(&context, request, &response);

	if (!status.ok()) {
		DOCA_DLOG_ERR("%s rpc failed: %s", __func__, status.error_message().c_str());
		return;
	}

	stats->nb_scanned_pkts = response.nb_scanned_pkts();
	stats->nb_matches = response.nb_matches();
	stats->nb_http_parser_based = response.nb_http_parser_based();
	stats->nb_ssl_parser_based = response.nb_ssl_parser_based();
	stats->nb_tcp_based = response.nb_tcp_based();
	stats->nb_udp_based = response.nb_udp_based();
	stats->nb_other_l4 = response.nb_other_l4();
	stats->nb_other_l7 = response.nb_other_l7();
}
