/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <thread>

#include <rte_byteorder.h>
#include <rte_eal.h>

#include <doca_argp.h>
#include <doca_log.h>

#include <dpdk_utils.h>

#include "common.pb.h"
#include "doca_dpi_grpc_handler.hpp"
#include "doca_dpi_grpc_async_server.hpp"
#include "orchestration.hpp"

DOCA_LOG_REGISTER(DPI::GRPC::ASYNC);

struct server_per_core_params {
	int service_idx;
	bool used;
};

using namespace std;
using grpc::Server;
using grpc::ServerBuilder;
using grpc::ServerCompletionQueue;
using grpc::ServerContext;
using grpc::Status;
using grpc::Channel;

static ServerImpl *server;
struct server_per_core_params core_params_arr[RTE_MAX_LCORE];

Status
DocaOrchestrationForAsyncImpl::HealthCheck(ServerContext *context, const HealthCheckReq *request,
					   HealthCheckResp *response)
{
	(void)context;
	(void)request;
	(void)response;

	/* Show the service that we are responsive */
	return Status::OK;
}

Status
DocaOrchestrationForAsyncImpl::Destroy(ServerContext *context, const DestroyReq *request, DestroyResp *response)
{
	(void)context;
	(void)request;
	(void)response;
	pthread_t son_tid;

	/* Shutdown the server and all its threads */

	/* The sync_grpc_thread that executes this method shares grpc resources with one of the async microservices.
	 * Shutting down all the microservices will also release this sync_thread grpc-resources and cause it to stop
	 * immediately. Performing the Shutdown in a different thread will let this grpc call to finish gracefully
	 * (and won't hang the calling client). */
	pthread_create(
		&son_tid, nullptr,
		[](void *arg) -> void * {
			server->Shutdown();
			return nullptr;
		},
		nullptr);

	return Status::OK;
}

static void
doca_dpi_server_map_queues(uint16_t nb_lcores_in_use, uint16_t nb_services)
{
	size_t i, service_idx = 0;

	memset(core_params_arr, 0, sizeof(core_params_arr));
	for (i = 0; i < RTE_MAX_LCORE; i++) {
		if (rte_lcore_is_enabled(i))
		{
			core_params_arr[i].service_idx = service_idx % nb_services;
			core_params_arr[i].used = true;
			++service_idx;
			if (service_idx >= nb_lcores_in_use)
				break;
		}
	}
}

MicroService::MicroService(std::string server_address, unsigned int nb_microservices) :
	dpi_adapter_(nb_microservices)
{
	/* Start the gRPC server */
	ServerBuilder builder;

	/* Listen on the given address without any authentication mechanism */
	builder.AddListeningPort(server_address, grpc::InsecureServerCredentials());

	/* Register "service" as the instance through which we'll communicate with clients.
	* The synchronous/asynchronous of the service depends on its class type. */
	builder.RegisterService(&service_);
	builder.RegisterService(&orchestration_service);

	/* Finally assemble the server and completion queue */
	cq_ = builder.AddCompletionQueue();
	server_ = builder.BuildAndStart();
}

void
MicroService::RequestStartReceiving()
{
	DocaDpiRpcHandler base(&service_, cq_.get(), &dpi_adapter_);
	new InitRpcHandler(base);
	new DestroyRpcHandler(base);
	new LoadSignaturesRpcHandler(base);
	new EnqueueRpcHandler(base);
	new DequeueRpcHandler(base);
	new FlowCreateRpcHandler(base);
	new FlowDestroyRpcHandler(base);
	new FlowMatchGetRpcHandler(base);
	new SignatureGetRpcHandler(base);
	new SignaturesGetRpcHandler(base);
	new StatGetRpcHandler(base);
}

void
MicroService::HandleRpcs()
{
	RequestStartReceiving();

	void* tag;
	bool ok;
	DocaDpiRpcHandler *call;

	while (cq_->Next(&tag, &ok))
	{
		/* Keep calling Next if "ok" is false - need to drain the cq_ to perform a successful Shutdown of it.
		 * If we'll not check "ok" and it is false, we'll get an undefined behavior (all API calls will be
		 * invoked once with uninitialized arguments). */
		if (ok) {
			call = static_cast<DocaDpiRpcHandler*>(tag);
			call->Proceed();
		}
	}
}

static int
microservice_routine(void *service)
{
	uint32_t core_id = rte_lcore_id();
	MicroService **curr_service = (MicroService **)service;

	if (core_params_arr[core_id].used)
		curr_service[core_params_arr[core_id].service_idx]->HandleRpcs();

	return 0;
}

void
ServerImpl::Run()
{
	/* Start the micro-services */
	rte_eal_mp_remote_launch(microservice_routine, microservices, CALL_MAIN);
	rte_eal_mp_wait_lcore();
}

ServerImpl::ServerImpl(const char *raw_server_address, unsigned int nb_lcores_to_use)
{
	/* Set inner members */
	this->nb_microservices = nb_lcores_to_use;
	this->microservices = new MicroService*[nb_microservices];

	/* Created objects for the micro-services */
	std::string base_ip = std::string(raw_server_address);
	size_t dots_pos = base_ip.find(':');
	int base_port = eNetworkPort::k_DocaDPI;
	int i;

	/* Check if we got a non-default port with the IP. */
	if (dots_pos != std::string::npos) {
		base_port = std::stoi(base_ip.substr(dots_pos + 1));
		base_ip = base_ip.substr(0, dots_pos);
	}

	/* Allocate resources for each micro-service */
	for (i = 0; i < nb_microservices; ++i)
	{
		std::string server_address(base_ip + ":" + std::to_string(base_port + i));
		microservices[i] = new MicroService(server_address, this->nb_microservices);
	}

	/* Setting mapping between the lcores and the microservices */
	doca_dpi_server_map_queues(nb_microservices, nb_microservices);

	/* Notify users only NOW they can start communicating with the server */
	DOCA_LOG_CRIT("gRPC server has started and listening on %s:%d-%d", base_ip.c_str(), base_port,
		      base_port + nb_microservices - 1);
}

int
main(int argc, char *argv[])
{
	struct application_dpdk_config dpdk_config = {};
	uint32_t nb_used_lcores;
	const char *grpc_address;
	doca_error_t result;
	int exit_status = EXIT_SUCCESS;

	/* Init the DPDK configuration struct */
	dpdk_config.port_config.nb_ports = 0;
	dpdk_config.port_config.nb_queues = 1; /* minimum 1 */
	dpdk_config.port_config.nb_hairpin_q = 0;
	dpdk_config.reserve_main_thread = false;

	/* Parse cmdline/json arguments */
	result = doca_argp_init("doca_dpi_grpc_async_server", nullptr);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init ARGP resources: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}
	doca_argp_set_dpdk_program(dpdk_init);
	doca_argp_set_grpc_program();

	/* No program-specific parameters */
	result = doca_argp_start(argc, argv);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse server input: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	/* Initialize the DPDK settings */
	result = dpdk_queues_and_ports_init(&dpdk_config);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to update application ports and queues: %s", doca_get_error_string(result));
		exit_status = EXIT_FAILURE;
		goto dpdk_destroy;
	}

	/* Best performance was using half the lcores. */
	nb_used_lcores = rte_lcore_count();  /* Best value is 4 (when there's at least 8 clients, otherwise it
					      * should be half the clients).
					      * For BF3 need to test what is fastest -
					      * half the lcores or 4 lcores or just leaving out 4 lcores unused. */

	/* Init Server static env */
	result = DocaDpiImpl::InitServerEnv();
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init server: %s", doca_get_error_string(result));
		exit_status = EXIT_FAILURE;
		goto dpdk_cleanup;
	}

	if (doca_argp_get_grpc_addr(&grpc_address) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to get grpc address");
		exit_status = EXIT_FAILURE;
		goto dpdk_cleanup;
	}

	/* Start the server */
	server = new ServerImpl(grpc_address, nb_used_lcores);
	server->Run();

	/* Releasing the memory */
	delete server;
dpdk_cleanup:
	dpdk_queues_and_ports_fini(&dpdk_config);
dpdk_destroy:
	dpdk_fini();
	doca_argp_destroy();
	return exit_status;
}
