/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DOCA_DPI_GRPC_ASYNC_SERVER_H_
#define DOCA_DPI_GRPC_ASYNC_SERVER_H_

#include <stdint.h>

#include <grpcpp/grpcpp.h>

#include <doca_dpi.h>

#include "doca_dpi.grpc.pb.h"
#include "doca_dpi.pb.h"
#include "orchestration.hpp"
#include "server_core.hpp"

/** This class holds on the variables needed by a microservice to operate as a DOCA gRPC DPI service. */
class MicroService {

	protected:
		/** The means of communication with the gRPC runtime for asynchronous server */
		DocaDpi::AsyncService service_;
		/** The instance that provides this microservice with the DocaOrchestration gRPC API support */
		DocaOrchestrationForAsyncImpl orchestration_service;
		/** The adapter class between DOCA DPI and gRPC */
		DocaDpiImpl dpi_adapter_;
		/** A running server which is ready for processing calls. GRPC engine uses it, */
		std::unique_ptr<grpc::Server> server_;
		/** The producer-consumer queue where for asynchronous server notifications.*/
		std::unique_ptr<grpc::ServerCompletionQueue> cq_;

	protected:
		/** Allocates handlers for the API. This determines what async API calls the microservice supports. */
		void RequestStartReceiving();

	public:
		/** Constructor
		 *
		 * param server_address
		 * 	IPv4 address for this microservice, format - "<IP>:<port>"
		 * param nb_microservices
		 * 	The number of total microservices that are running in the server.
		 */
		MicroService(std::string server_address, unsigned int nb_microservices);

		/** This stops the microservice. */
		void Shutdown()
		{
			/** Always shutdown the completion queue after the server. */
			server_->Shutdown();
			cq_->Shutdown();
		}

		/** This runs the microservice and does not return until shutdown. */
		void HandleRpcs();
};

/** This class holds all microservices that construct the whole server. */
class ServerImpl final {

	/** Number of microservices in the Server.
	 * This number also indicates how many lcores are is use and on how many ports the server is active. */
	unsigned int nb_microservices;
	/** Dynamically allocates array of microservices resources. */
	MicroService **microservices;

	public:
		/** Constructor
		 *
		 * param raw_server_address
		 * 	IP address (and optionally base port) for this server, format - "<IP>:<port>" or "<IP>".
		 * 	If a port number wasn't given, the default one will be used.
		 * 	The microservices IPs are the same, but their port is: base-port + lcore_id().
		 * param nb_microservices
		 * 	The number of total microservices to run in the server.
		 */
		ServerImpl(const char *raw_server_address, unsigned int nb_lcores_to_use);

		/** Deletes the server. Always call "Shutdown" before deleteing the server. */
		~ServerImpl()
		{
			int i;
			for (i = 0; i < nb_microservices; ++i) {
				delete microservices[i];
			}
			delete[] microservices;
		}

		/** This starts the server. This will not return until server is shutdown. */
		void Run();

		/** Gracefully stops the server and all its microservices.
		 * This cannot be invoked by a gRPC thread or the microservice thread that is being shutdown */
		void Shutdown()
		{
			int i;
			for (i = 0; i < nb_microservices; i++)
				microservices[i]->Shutdown();
		}
};

#endif /* DOCA_DPI_GRPC_ASYNC_SERVER_H_ */
