/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#ifndef DOCA_DPI_GRPC_HANDLER_H_
#define DOCA_DPI_GRPC_HANDLER_H_

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <errno.h>
#include <string>
#include <cstring>
#include <queue>

#include <grpcpp/grpcpp.h>
#include <grpcpp/health_check_service_interface.h>

#include <doca_dpi.h>

#include "doca_dpi.grpc.pb.h"
#include "server_core.hpp"

/** Class encompassing the state and logic needed to serve an async request. */
class DocaDpiRpcHandler {

        protected:
		/** The means of communication with the gRPC runtime for asynchronous server */
		DocaDpi::AsyncService* service_;
		/** The producer-consumer queue where for asynchronous server notifications.*/
		grpc::ServerCompletionQueue* cq_;
		/** The adapter class between DOCA DPI and gRPC */
		DocaDpiImpl* dpi_adapter_;

		/** Context for the rpc, allowing to tweak aspects of it such as the use
		* of compression, authentication, as well as to send metadata back to theclient.
		*/
		grpc::ServerContext ctx_;

		/** The following states represents the possible states of our state machine for RPCs handling. */
		enum DocaDpiRpcStatus { PROCESS, FINISH };
        	/** The current serving state. */
		DocaDpiRpcStatus status_;

        public:
                /** Constructor.
		 * Take in the "service" instance (in this case representing an asynchronous
                 * server) and the completion queue "cq" used for asynchronous communication
                 * with the gRPC runtime.
                 */
                DocaDpiRpcHandler(DocaDpi::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaDpiImpl* adapter)
                        : service_(service), cq_(cq), dpi_adapter_(adapter), status_(PROCESS)
			{ };

		/** Copy Constructor.
		 * Creates a new handler using the pointers to the resources in the given handler.
                 */
                DocaDpiRpcHandler(const DocaDpiRpcHandler& other)
                        : service_(other.service_), cq_(other.cq_), dpi_adapter_(other.dpi_adapter_), status_(PROCESS)
			{ };

		/** Progress the execution of an async gRPC call. */
		virtual void Proceed(void);

		virtual ~DocaDpiRpcHandler() {};
};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class InitRpcHandler : public DocaDpiRpcHandler {

	/** Request object to be filled by gRPC upon incoming RPC. */
        DocaDpiInitParams request_;
	/** Replay object to be filled by "Proceed" method and being passed to gRPC as replay to the client. */
        DocaDpiInitResponse reply_;
	/** This object is used to pass the response to the client using gRPC. */
        grpc::ServerAsyncResponseWriter<DocaDpiInitResponse> responder_;

	public:
		/** Copy Constructor. */
		InitRpcHandler(DocaDpiRpcHandler &other)
			: DocaDpiRpcHandler(other), responder_(&ctx_) {
			service_->RequestDocaDpiInit(&ctx_, &request_, &responder_, cq_, cq_, this);
		};

		virtual void Proceed(void) override;
		virtual ~InitRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class DestroyRpcHandler : public DocaDpiRpcHandler {

        DocaDpiDestroyParams request_;
        DocaDpiDestroyResponse reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiDestroyResponse> responder_;

	public:
		/** Copy Constructor. */
		DestroyRpcHandler(DocaDpiRpcHandler &other)
			: DocaDpiRpcHandler(other), responder_(&ctx_) {
			service_->RequestDocaDpiDestroy(&ctx_, &request_, &responder_, cq_, cq_, this);
		};

		virtual void Proceed(void) override;
		virtual ~DestroyRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class LoadSignaturesRpcHandler : public DocaDpiRpcHandler {

        DocaDpiLoadSignaturesParams request_;
        DocaDpiErrorInfo reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiErrorInfo> responder_;

        public:
		/** Copy Constructor. */
                LoadSignaturesRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiLoadSignatures(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;

                virtual ~LoadSignaturesRpcHandler() override {};
};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class EnqueueRpcHandler : public DocaDpiRpcHandler {

        DocaDpiEnqueueParams request_;
        DocaDpiErrorInfo reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiErrorInfo> responder_;

        public:
		/** Copy Constructor. */
                EnqueueRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiEnqueue(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~EnqueueRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class DequeueRpcHandler : public DocaDpiRpcHandler {

        DocaDpiDequeueParams request_;
        DocaDpiActionResponse reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiActionResponse> responder_;

        public:
		/** Copy Constructor. */
                DequeueRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiDequeue(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~DequeueRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class FlowCreateRpcHandler : public DocaDpiRpcHandler {

        DocaDpiFlowCreateParams request_;
        DocaDpiFlowCreateResponse reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiFlowCreateResponse> responder_;

        public:
		/** Copy Constructor. */
                FlowCreateRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiFlowCreate(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

		virtual void Proceed(void) override;
                virtual ~FlowCreateRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class FlowDestroyRpcHandler : public DocaDpiRpcHandler {

        DocaDpiFlowDestroyParams request_;
        DocaDpiFlowDestroyResponse reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiFlowDestroyResponse> responder_;

        public:
		/** Copy Constructor. */
                FlowDestroyRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiFlowDestroy(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~FlowDestroyRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class FlowMatchGetRpcHandler : public DocaDpiRpcHandler {

        DocaDpiFlowMatchGetParams request_;
        DocaDpiActionResponse reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiActionResponse> responder_;

        public:
		/** Copy Constructor. */
                FlowMatchGetRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiFlowMatchGet(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~FlowMatchGetRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class SignatureGetRpcHandler : public DocaDpiRpcHandler {

        DocaDpiSignatureGetParams request_;
        DocaDpiSignatureGetResponse reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiSignatureGetResponse> responder_;

        public:
		/** Copy Constructor. */
                SignatureGetRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiSignatureGet(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~SignatureGetRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class SignaturesGetRpcHandler : public DocaDpiRpcHandler {

        DocaDpiSignaturesGetParams request_;
        DocaDpiSignaturesGetResponse reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiSignaturesGetResponse> responder_;

        public:
		/** Copy Constructor. */
                SignaturesGetRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiSignaturesGet(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~SignaturesGetRpcHandler() override {};

};

/** This class hold the resources and logic to serve a specific DPI gRPC request. */
class StatGetRpcHandler : public DocaDpiRpcHandler {

        DocaDpiStatGetParams request_;
        DocaDpiStatInfo reply_;
        grpc::ServerAsyncResponseWriter<DocaDpiStatInfo> responder_;

        public:
		/** Copy Constructor. */
                StatGetRpcHandler(const DocaDpiRpcHandler& other)
                        : DocaDpiRpcHandler(other), responder_(&ctx_) {
                         service_->RequestDocaDpiStatGet(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~StatGetRpcHandler() override {};

};

#endif /* DOCA_DPI_GRPC_HANDLER_H_ */
