/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdint.h>
#include <string.h>

#include <rte_malloc.h>

#include <doca_log.h>

#include "grpc_type_converter.hpp"

DOCA_LOG_REGISTER(DPI::GRPC::CONVERTER);

struct postfix_len_user_data {
	void *buf;
	uint32_t len;
};

static struct rte_mempool *mem_pool;
static uint16_t mbuf_segment_buffer_size;

void
converter_init(struct rte_mempool *mp, uint16_t data_buffer_size_in_mbuf)
{
	mem_pool = mp;
	mbuf_segment_buffer_size = data_buffer_size_in_mbuf;
}

DocaDpiGenericPacket *
convert_rte_mbuf(const struct rte_mbuf *src)
{
	DocaDpiGenericPacket *pkt = new DocaDpiGenericPacket();
	std::string *segment = pkt->mutable_segment();
	while (src != NULL) {
		segment->append(rte_pktmbuf_mtod(src, char *), src->data_len);
		src = src->next;
	}
	return pkt;
}

static DocaDpiSigInfo *
convert_sig_info(const struct doca_dpi_sig_info *src)
{
	DocaDpiSigInfo *dst = new DocaDpiSigInfo();
	dst->set_sig_id(src->sig_id);
	dst->set_action(src->action);
	return dst;
}

void *
convert_grpc_user_data(const std::string &user_data)
{
	struct postfix_len_user_data *ret =
	    (struct postfix_len_user_data *)rte_malloc(NULL, sizeof(struct postfix_len_user_data), 0);
	if (ret == NULL) {
		DOCA_DLOG_ERR("Could not allocate memory - not enough memory");
		return NULL;
	}
	ret->len = user_data.length();
	ret->buf = rte_malloc(NULL, ret->len, 0);
	if (ret->buf == NULL) {
		DOCA_DLOG_ERR("Could not allocate user_data buffer memory - not enough memory");
		rte_free(ret);
		return NULL;
	}
	memcpy(ret->buf, user_data.c_str(), ret->len);
	return ret;
}

void
free_user_data(void *converted_user_data)
{
	struct postfix_len_user_data *user_data = (struct postfix_len_user_data *)converted_user_data;

	if (user_data != NULL) {
		rte_free(user_data->buf);
		rte_free(user_data);
	}
}

DocaDpiResult *
convert_result(const struct doca_dpi_result *src)
{
	struct postfix_len_user_data *user_data = (struct postfix_len_user_data *)src->user_data;
	DocaDpiResult *result = new DocaDpiResult();
	result->set_matched(src->matched);
	if (user_data != NULL)
		result->set_user_data(user_data->buf, user_data->len);
	if (src->pkt != NULL)
		result->set_allocated_pkt(convert_rte_mbuf(src->pkt));
	result->set_allocated_info(convert_sig_info(&src->info));
	result->set_status_flags(src->status_flags);
	return result;
}

static struct rte_mbuf *
create_rte_mbuf_node(const void *payload, uint16_t data_len)
{
	struct rte_mbuf *pkt = rte_pktmbuf_alloc(mem_pool);
	if (pkt == NULL) {
		DOCA_DLOG_ERR("Could not allocate mbuf memory - not enough memory");
		return NULL;
	}
	pkt->data_len = data_len;
	memcpy(rte_pktmbuf_mtod(pkt, void *), payload, data_len);

	return pkt;
}

struct rte_mbuf *
convert_grpc_generic_packet(const DocaDpiGenericPacket *src)
{
	uint16_t pkt_len = src->segment().length();
	uint16_t nb_segs =
	    pkt_len / mbuf_segment_buffer_size + (pkt_len % mbuf_segment_buffer_size != 0);
	const char *curr_segment_start =
	    src->segment().c_str() + pkt_len - pkt_len % mbuf_segment_buffer_size;
	struct rte_mbuf *new_head;
	/* Allocate the last packet of the chain with the "residue" data */
	struct rte_mbuf *curr_head =
	    create_rte_mbuf_node(curr_segment_start, pkt_len % mbuf_segment_buffer_size);
	if (curr_head == NULL) {
		return NULL;
	}
	/* Chain the rest of the data in reverse order */
	for (int i = 1; i < nb_segs; ++i) {
		curr_segment_start -= mbuf_segment_buffer_size;
		new_head = create_rte_mbuf_node(curr_segment_start, mbuf_segment_buffer_size);
		if (new_head != NULL) {
			new_head->next = curr_head;
			curr_head = new_head;
		} else {
			curr_head->nb_segs = i;
			rte_pktmbuf_free(curr_head);
			return NULL;
		}
	}
	/* Assgin pkt_len and nb_segs to only first packet */
	curr_head->pkt_len = pkt_len;
	curr_head->nb_segs = nb_segs;

	return curr_head;
}

DocaDpiSigData *
convert_sig_data(const struct doca_dpi_sig_data *src, DocaDpiSigData *buff)
{
	DocaDpiSigData *ret = buff;
	if (buff == NULL)
		ret = new DocaDpiSigData();
	ret->set_sig_id(src->sig_id);
	ret->set_name(src->name);
	return ret;
}
