/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <condition_variable>
#include <functional>
#include <fstream>
#include <mutex>
#include <rte_byteorder.h>
#include <rte_eal.h>
#include <rte_jhash.h>
#include <rte_malloc.h>
#include <rte_mbuf.h>
#include <rte_mbuf_core.h>
#include <rte_random.h>
#include <rte_errno.h>
#include <stdint.h>
#include <string>
#include <time.h>

#include <doca_dpi.h>
#include <doca_log.h>

#include <grpc/log_forwarder.hpp>

#include "server_core.hpp"
#include "grpc_type_converter.hpp"

DOCA_LOG_REGISTER(DPI::GRPC::Core);

#define LCORE_CACHE_SIZE 512
#define HASH_PTR(x) rte_jhash(&(x), sizeof(void *), seed)
#define TMP_CDO_FILE_PREFIX "tmp_cdo_request_file_"

using grpc::Server;
using grpc::ServerBuilder;
using grpc::ServerContext;
using grpc::ServerWriter;
using grpc::Status;
using grpc::StatusCode;

/* Generating random seed for HASH macro. The default seed of rte_random is the time. */
static uint32_t seed = static_cast<uint32_t>(rte_rand());

static struct synchronized_queue log_records_queue;
static std::string latest_log_records[RTE_MAX_LCORE];
static char log_buffer[1024] = {};
static struct rte_mempool *mem_pool;
static struct doca_dpi_ctx *dpi_ctx;

/* Function used for inserting the log messages to the messages queue */
static void
flush_buffer(char *buffer)
{
	static thread_local unsigned int lcore_id = rte_lcore_id();

	/* If in sync server mode */
	if (lcore_id == LCORE_ID_ANY) {
		/* Insert log message to queue, this should happen only on error or critical */
		synchronized_queue_enqueue(&log_records_queue, std::string(buffer));
	} else {
		/* Insert log message to queue */
		latest_log_records[lcore_id] = std::string(buffer);
	}
}

doca_error_t
DocaDpiImpl::InitServerEnv()
{
	doca_error_t result;
	struct doca_logger_backend *logger;

	/* Allocating a log that adds error messages to queue */
	result = doca_log_create_buffer_backend(log_buffer, sizeof(log_buffer), flush_buffer, &logger);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to allocate logger");
		return result;
	}
	doca_log_backend_level_set(logger, DOCA_LOG_LEVEL_ERROR);
	return DOCA_SUCCESS;
}

DocaDpiImpl::~DocaDpiImpl()
{
	struct doca_dpi_flow_ctx *flow_curr;
	for (auto kv : this->doca_dpi_flow_ctx_pool.hash_table) {
		flow_curr = (struct doca_dpi_flow_ctx *)kv.second;
		doca_dpi_flow_destroy(flow_curr);
	}
}

DocaDpiErrorInfo *
DocaDpiImpl::generate_error_info(int error_code, DocaDpiErrorInfo *inplace_buf)
{
	static thread_local unsigned int lcore_id = rte_lcore_id();
	DocaDpiErrorInfo *ret = inplace_buf;
	std::string err_msg;

	if (!ret)
		ret = new DocaDpiErrorInfo();

	ret->set_error_code(error_code);
	/* If in sync server mode */
	if (lcore_id == LCORE_ID_ANY)
		err_msg = synchronized_queue_dequeue(&log_records_queue);
	else {  /* If in async server mode */
		err_msg = latest_log_records[lcore_id];
		latest_log_records[lcore_id] = "";
	}

	if (err_msg.size() > 0) {
		ret->set_err_msg(err_msg);
	}
	return ret;
}

Status
DocaDpiImpl::DocaDpiInit(ServerContext *context, const DocaDpiInitParams *request,
			    DocaDpiInitResponse *response)
{
	(void)context;

	int error_code;

	/* Preparing to call doca_dpi */
	struct doca_dpi_config_t conf = {.nb_queues = static_cast<uint16_t>(request->uint16_nb_queues()),
					 .max_packets_per_queue = request->max_packets_per_queue(),
					 .max_sig_match_len = request->max_sig_match_len()};

	/* If DocaDpiInit was called, cancel */
	if (dpi_ctx != nullptr)
		return Status(StatusCode::ALREADY_EXISTS,
			      "DPI context already exists! Perhaps another process is using the server.");

	/* Allocating mempool for rte_mbuf creation. */
	mem_pool = rte_pktmbuf_pool_create(
	    "dpi_grpc_mbuf_mempool", conf.nb_queues * conf.max_packets_per_queue - 1, LCORE_CACHE_SIZE,
	    RTE_MBUF_PRIV_ALIGN, RTE_MBUF_DEFAULT_BUF_SIZE, SOCKET_ID_ANY);

	if (unlikely(mem_pool == nullptr)) {
		DOCA_DLOG_CRIT("Cannot allocate mbuf pool, DPDK error code: %d : %s", rte_errno,
			       rte_strerror(rte_errno));
		return Status(StatusCode::INTERNAL,
			      "Failed to allocate mempool, it might exist already or requests size in too large");
	}

	/* Init the converter that is shared by all microservices.  */
	converter_init(mem_pool, RTE_MBUF_DEFAULT_DATAROOM);

	/* Create a new DPI context to be shared by all micro services. */
	dpi_ctx = doca_dpi_init(&conf, &error_code);

	/* Preparing response */
	generate_error_info(error_code, response->mutable_erreno());
	response->set_nb_microservices(this->nb_microservices);

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiDestroy(ServerContext *context, const DocaDpiDestroyParams *request,
			    DocaDpiDestroyResponse *response)
{
	(void)context;
	(void)request;
	(void)response;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Destroying DPI without stopping the server. Server is reusable. */
	doca_dpi_destroy(dpi_ctx);
	dpi_ctx = nullptr;

	/* Destroying the memory pool */
	converter_init(nullptr, 0);
	rte_mempool_free(mem_pool);

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiLoadSignatures(ServerContext *context,
				   const DocaDpiLoadSignaturesParams *request, DocaDpiErrorInfo *response)
{
	(void)context;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Reading arguments */
	int error_code;
	std::string sig_path;
	std::fstream tmp_cdo_file;

	if (request->cdo_case() == DocaDpiLoadSignaturesParams::kCdoFilename)
		sig_path = request->cdo_filename();
	else {
		sig_path = TMP_CDO_FILE_PREFIX + std::to_string(rte_rand());
		tmp_cdo_file.open(sig_path, std::fstream::out | std::fstream::trunc);
		tmp_cdo_file << request->cdo_data();
		tmp_cdo_file.close();
	}

	/* Directing to doca dpi */
	error_code = doca_dpi_load_signatures(dpi_ctx, sig_path.c_str());

	/* Deleting tmp file */
	if (request->cdo_case() != DocaDpiLoadSignaturesParams::kCdoFilename) {
		if (std::remove(sig_path.c_str()) != 0)
			DOCA_DLOG_WARN("Could not delete temp cdo file");
	}

	/* Preparing response */
	generate_error_info(error_code, response);

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiEnqueue(ServerContext *context, const DocaDpiEnqueueParams *request,
			    DocaDpiErrorInfo *response)
{
	(void)context;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	if (request->packet().segment().length() >= 0x10000) {
		DOCA_DLOG_WARN("Max packet length is 65535 (0xffff).");
		return Status(StatusCode::INVALID_ARGUMENT, "DocaDpiGenericPacket segment is too big (max: 65535).");
	}

	/* Preparing to call doca_dpi */
	int error_code;
	struct doca_dpi_flow_ctx *dpi_flow =
		(struct doca_dpi_flow_ctx *)sync_map_get(&this->doca_dpi_flow_ctx_pool, request->context().unique_id());

	if (dpi_flow == nullptr)
		return Status(StatusCode::NOT_FOUND, "Unknown dpi flow context");

	const DocaDpiGenericPacket &in_pkt = request->packet();
	struct rte_mbuf *pkt = convert_grpc_generic_packet(&in_pkt);
	bool initiator = request->initiator();
	uint32_t payload_offset = request->payload_offset();
	void *user_data = nullptr;

	if (request->has_user_data()) {
		user_data = convert_grpc_user_data(request->user_data());
		if (unlikely(user_data == nullptr))
			return Status(StatusCode::RESOURCE_EXHAUSTED, "Not enough memory to store user_data");
	}
	if (unlikely(pkt == nullptr))
		return Status(StatusCode::RESOURCE_EXHAUSTED, "Not enough memory to run");

	/* Directing call to doca_dpi */
	error_code = doca_dpi_enqueue(dpi_flow, pkt, initiator, payload_offset, user_data);
	if (error_code != DOCA_DPI_ENQ_PROCESSING)
		rte_pktmbuf_free(pkt);

	/* Preparing response */
	generate_error_info(error_code, response);

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiDequeue(ServerContext *context, const DocaDpiDequeueParams *request,
			    DocaDpiActionResponse *response)
{
	(void)context;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Preparing to call doca_dpi */
	int error_code;
	struct doca_dpi_result result = {};
	uint16_t dpi_q = static_cast<uint16_t>(request->uint16_dpi_q());

	/* Directing call to doca_dpi */
	error_code = doca_dpi_dequeue(dpi_ctx, dpi_q, &result);

	/* Preparing response */
	response->set_allocated_erreno(generate_error_info(error_code));
	response->set_allocated_result(convert_result(&result));

	/* Freeing the dequeued packet memory */
	free_user_data(result.user_data);
	rte_pktmbuf_free(result.pkt);

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiFlowCreate(ServerContext *context, const DocaDpiFlowCreateParams *request,
			       DocaDpiFlowCreateResponse *response)
{
	(void)context;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Preparing to call doca_dpi */
	struct doca_dpi_flow_ctx *new_dpi_flow;
	DocaDpiFlowCtx *flow_ctx;
	uint64_t hash_val;
	const DocaDpiParsingInfo &in_parse_info = request->parsing_info();
	int dpi_q = request->uint16_dpi_q();
	int error_code;
	struct doca_dpi_result result = {};
	struct doca_dpi_parsing_info parsing_info = {
	    .ethertype = rte_cpu_to_be_16(in_parse_info.ethertype()),
	    .l4_protocol = static_cast<uint8_t>(in_parse_info.l4_protocol()),
	    .l4_dport = rte_cpu_to_be_16(in_parse_info.l4_dst_port()),
	    .l4_sport = rte_cpu_to_be_16(in_parse_info.l4_src_port())};

	if (in_parse_info.dst_ip_case() == DocaDpiParsingInfo::kDstIpv4)
		parsing_info.dst_ip.ipv4.s_addr = in_parse_info.dst_ipv4();
	else
		memcpy(&parsing_info.dst_ip.ipv6, in_parse_info.dst_ipv6().ipv6_addr().c_str(),
		       in_parse_info.dst_ipv6().ipv6_addr().length());

	if (in_parse_info.src_ip_case() == DocaDpiParsingInfo::kSrcIpv4)
		parsing_info.src_ip.ipv4.s_addr = in_parse_info.src_ipv4();
	else
		memcpy(&parsing_info.src_ip.ipv6, in_parse_info.src_ipv6().ipv6_addr().c_str(),
		       in_parse_info.src_ipv6().ipv6_addr().length());

	/* Directing call to doca_dpi */
	new_dpi_flow = doca_dpi_flow_create(dpi_ctx, dpi_q, &parsing_info, &error_code, &result);

	/* Preparing response */
	response->set_allocated_erreno(generate_error_info(error_code)); /* "response" takes ownership */
	if (new_dpi_flow == nullptr)
		return Status::OK;

	flow_ctx = new DocaDpiFlowCtx();
	hash_val = HASH_PTR(new_dpi_flow);
	flow_ctx->set_unique_id(hash_val);
	response->set_allocated_flow_context(flow_ctx); /* "response" takes ownership on memory */
	response->set_allocated_result(convert_result(&result));

	/* Adding new dpi flow context to the allocated pool */
	sync_map_add(&this->doca_dpi_flow_ctx_pool, hash_val, new_dpi_flow);

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiFlowDestroy(ServerContext *context, const DocaDpiFlowDestroyParams *request,
				DocaDpiFlowDestroyResponse *response)
{
	(void)context;
	(void)response;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Preparing to call doca_dpi */
	struct doca_dpi_flow_ctx *dpi_flow = (struct doca_dpi_flow_ctx *)sync_map_remove(
	    &this->doca_dpi_flow_ctx_pool, request->context().unique_id());

	if (dpi_flow == nullptr)
		return Status(StatusCode::NOT_FOUND, "Unknown dpi flow context");

	/* Directing call to doca_dpi */
	doca_dpi_flow_destroy(dpi_flow);

	/* Preparing response */
	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiFlowMatchGet(ServerContext *context, const DocaDpiFlowMatchGetParams *request,
				 DocaDpiActionResponse *response)
{
	(void)context;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Preparing to call doca_dpi */
	int error_code;
	struct doca_dpi_result result = {};
	struct doca_dpi_flow_ctx *dpi_flow = (struct doca_dpi_flow_ctx *)sync_map_get(
	    &this->doca_dpi_flow_ctx_pool, request->context().unique_id());
	if (dpi_flow == nullptr)
		return Status(StatusCode::NOT_FOUND, "Unknown dpi flow context");

	/* Directing call to doca_dpi */
	error_code = doca_dpi_flow_match_get(dpi_flow, &result);

	/* Preparing response */
	response->set_allocated_erreno(generate_error_info(error_code));
	response->set_allocated_result(convert_result(&result));

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiSignatureGet(ServerContext *context, const DocaDpiSignatureGetParams *request,
				 DocaDpiSignatureGetResponse *response)
{
	(void)context;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Preparing to call doca_dpi */
	int error_code;
	struct doca_dpi_sig_data sig_data = {};
	DocaDpiSigData *response_sig_data = new DocaDpiSigData();

	/* Directing call to doca_dpi */
	error_code = doca_dpi_signature_get(dpi_ctx, request->sign_id(), &sig_data);

	/* Preparing response */
	response_sig_data->set_sig_id(sig_data.sig_id);
	response_sig_data->set_name(sig_data.name);
	response->set_allocated_erreno(generate_error_info(error_code));
	response->set_allocated_sign_data(response_sig_data); /* "response" takes ownership */

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiSignaturesGet(ServerContext *context,
				  const DocaDpiSignaturesGetParams *request,
				  DocaDpiSignaturesGetResponse *response)
{
	(void)context;
	(void)request;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Preparing to call doca_dpi */
	int error_code;
	int num_signs;
	struct doca_dpi_sig_data *sig_data_arr;
	DocaDpiSigData *added_sig_data;
	google::protobuf::RepeatedPtrField<DocaDpiSigData> *response_signs = response->mutable_signs();

	/* Directing call to doca_dpi */
	num_signs = doca_dpi_signatures_get(dpi_ctx, &sig_data_arr);
	error_code = (num_signs >= 0) ? 0 : num_signs;

	/* Preparing response */
	response->set_allocated_erreno(generate_error_info(error_code));
	for (int i = 0; i < num_signs; ++i) {
		added_sig_data = response_signs->Add();
		convert_sig_data(sig_data_arr + i, added_sig_data);
	}

	/* free copied data */
	rte_free(sig_data_arr);

	return Status::OK;
}

Status
DocaDpiImpl::DocaDpiStatGet(ServerContext *context, const DocaDpiStatGetParams *request,
			    DocaDpiStatInfo *response)
{
	(void)context;

	/* If DocaDpiInit wasn't called, cancel */
	if (dpi_ctx == nullptr)
		return Status(StatusCode::FAILED_PRECONDITION, "No existing DPI context, please call DocaDpiInit first.");

	/* Preparing to call doca_dpi */
	bool clear = request->clear();
	struct doca_dpi_stat_info stats;

	/* Directing call to doca_dpi */
	doca_dpi_stat_get(dpi_ctx, clear, &stats);

	/* Preparing response */
	response->set_nb_scanned_pkts(stats.nb_scanned_pkts);
	response->set_nb_matches(stats.nb_matches);
	response->set_nb_http_parser_based(stats.nb_http_parser_based);
	response->set_nb_ssl_parser_based(stats.nb_ssl_parser_based);
	response->set_nb_tcp_based(stats.nb_tcp_based);
	response->set_nb_udp_based(stats.nb_udp_based);
	response->set_nb_other_l4(stats.nb_other_l4);
	response->set_nb_other_l7(stats.nb_other_l7);

	return Status::OK;
}
