/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <bsd/string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <rte_malloc.h>
#include <json-c/json.h>

#include <doca_version.h>
#include <doca_log.h>

#include "cdo_decoder.h"
#include "parser/dpi_parser_types.h"

#define MAX_ADDRESS_LEN 43
#define IPV4_LEN 15

struct dpi_cdo_data {
	rte_atomic32_t ref_cnt;
	struct post_processing_table pp_table;
	struct condition_table cond_table;
	struct dpi_sig_info_table info_table;
	struct category_table cat_table;
	struct family_table fam_table;
};

DOCA_LOG_REGISTER(DPI::DECODER);

static long
read_rules_db(const char *file, char **buf)
{
	FILE *fp;
	long buf_len = 0;
	size_t read_len;
	int res = 0;
	*buf = NULL;

	fp = fopen(file, "r");
	if (!fp) {
		DOCA_DLOG_ERR("can't open file %s, err = %d", file, errno);
		return -EIO;
	}
	if (fseek(fp, 0L, SEEK_END) == 0) {
		buf_len = ftell(fp);
		if (buf_len == -1) {
			DOCA_DLOG_ERR("ftell function fail");
			res = EIO;
			goto error;
		}
		*buf = (char *)calloc((buf_len + 1), sizeof(char));
		if (!*buf) {
			DOCA_DLOG_ERR("malloc function fail");
			res = ENOMEM;
			goto error;
		}
		if (fseek(fp, 0L, SEEK_SET) != 0) {
			DOCA_DLOG_ERR("fseek function fail");
			res = EIO;
			goto error;
		}
		read_len = fread(*buf, sizeof(char), buf_len, fp);
		if (read_len != (unsigned long)buf_len) {
			DOCA_DLOG_ERR("fread function fail");
			res = EIO;
			goto error;
		}
	}
	fclose(fp);
	return buf_len;
error:
	if (fp)
		fclose(fp);
	if (*buf)
		free(*buf);
	return -res;
}

#ifdef DOCA_LOGGING_ALLOW_DLOG
static const char*
print_json_object(json_object *json_elem)
{
	switch (json_object_get_type(json_elem)) {
	case json_type_int:
		return "json_type_int";
	case json_type_boolean:
		return "json_type_boolean";
	case json_type_double:
		return "json_type_double";
	case json_type_string:
		return "json_type_string";
	case json_type_array:
		return "json_type_array";
	case json_type_object:
		return "json_type_object";
	case json_type_null:
		return "json_type_null";
	default:
		return "\nwrong json_object\n";
	}
}

#endif

static int
action_str2num(const char *sig_action)
{
	if (!strcmp(sig_action, "drop"))
		return DOCA_DPI_SIG_ACTION_DROP;
	else if (!strcmp(sig_action, "alert"))
		return DOCA_DPI_SIG_ACTION_ALERT;
	else if (!strcmp(sig_action, "pass"))
		return DOCA_DPI_SIG_ACTION_PASS;
	else if (!strcmp(sig_action, "reject"))
		return DOCA_DPI_SIG_ACTION_REJECT;
	else if (!strcmp(sig_action, "rejectsrc"))
		return DOCA_DPI_SIG_ACTION_REJECTSRC;
	else if (!strcmp(sig_action, "rejectdst"))
		return DOCA_DPI_SIG_ACTION_REJECTDST;
	else if (!strcmp(sig_action, "rejectboth"))
		return DOCA_DPI_SIG_ACTION_REJECTBOTH;
	else
		return DOCA_DPI_SIG_ACTION_NA;
}

static int
fill_prot_translate_map(json_object *prot_table, int **prot_translt)
{
	int prot_num, *tmp_prot_trn;

	if (json_object_get_type(prot_table) != json_type_object) {
		DOCA_DLOG_ERR("wrong object type for protocol_to_number, json type is: %s",
			print_json_object(prot_table));
		return -1;
	}

	*prot_translt = (int *)rte_zmalloc(NULL,
					sizeof(int)*DPI_PARSER_MAX_PROTOCOL, 0);
	if (*prot_translt == NULL) {
		DOCA_DLOG_ERR("memory allocation error");
		return -1;
	}
	tmp_prot_trn = *prot_translt;

	json_object_object_foreach(prot_table, key, val) {
		switch (json_object_get_type(val)) {
		case json_type_int:
			prot_num = json_object_get_int(val);
			if (prot_num >= DPI_PARSER_MAX_PROTOCOL) {
				DOCA_DLOG_ERR("unsupported protocol received, in translation map");
				return -1;
			}
			if (!strcasecmp(key, "http"))
				tmp_prot_trn[prot_num] = DPI_PARSER_HTTP_1;
			else if (!strcasecmp(key, "http2"))
				tmp_prot_trn[prot_num] = DPI_PARSER_HTTP_2;
			else if (!strcasecmp(key, "tcp"))
				tmp_prot_trn[prot_num] = DPI_PARSER_TCP;
			else if (!strcasecmp(key, "udp"))
				tmp_prot_trn[prot_num] = DPI_PARSER_UDP;
			else if (!strcasecmp(key, "ssl"))
				tmp_prot_trn[prot_num] = DPI_PARSER_SSL;
			else if (!strcasecmp(key, "tls"))
				tmp_prot_trn[prot_num] = DPI_PARSER_SSL;
			else if (!strcasecmp(key, "sip"))
				tmp_prot_trn[prot_num] = DPI_PARSER_SIP;
			else if (!strcasecmp(key, "rtmp"))
				tmp_prot_trn[prot_num] = DPI_PARSER_RTMP;
			else if (!strcasecmp(key, "icmp"))
				tmp_prot_trn[prot_num] = DPI_PARSER_ICMP;
			else if (!strcasecmp(key, "ip"))
				tmp_prot_trn[prot_num] = DPI_PARSER_IP;
			else if (!strcasecmp(key, "ip6"))
				tmp_prot_trn[prot_num] = DPI_PARSER_IP6;
			else if (!strcasecmp(key, "quic"))
				tmp_prot_trn[prot_num] = DPI_PARSER_QUIC;
			else if (!strcasecmp(key, "dns"))
				tmp_prot_trn[prot_num] = DPI_PARSER_DNS;
			else if (!strcasecmp(key, "stun"))
				tmp_prot_trn[prot_num] = DPI_PARSER_STUN;
			else if (!strcasecmp(key, "unknown"))
				tmp_prot_trn[prot_num] =
						  DPI_PARSER_PROTOCOL_NOT_READY;
			else {
				DOCA_DLOG_ERR("protocol %s is not supported",
					key);
				return -1;
			}
			continue;
		default:
			DOCA_DLOG_ERR("wrong type in protocol translation table");
			return -1;
		}
	}
	return 0;
}

static int
parse_pp_entry(struct json_object *sig_obj, struct pp_table_entry *ppt_entry, int *prot_translt)
{
	int int_val;

	json_object_object_foreach(sig_obj, key, val) {
		switch (json_object_get_type(val)) {
		case json_type_int:
			int_val = json_object_get_int(val);
			if (!strcmp(key, "offset")) {
				if (int_val > 0) {
					ppt_entry->offset = int_val;
					ppt_entry->bitmask = ppt_entry->bitmask | DPI_PP_OFFSET;
				}
			} else if (!strcmp(key, "l7_protocol")) {
				if (int_val > 0) {
					ppt_entry->l7_protocol = prot_translt[int_val];
					ppt_entry->bitmask = ppt_entry->bitmask |
							     DPI_PP_L7_PROTOCOL;
				}
			} else if (!strcmp(key, "l4_protocol")) {
				if (int_val > 0) {
					ppt_entry->l4_protocol = int_val;
					ppt_entry->bitmask = ppt_entry->bitmask |
							     DPI_PP_L4_PROTOCOL;
				}
			} else if (!strcmp(key, "regex_id")) {
				continue;
			} else {
				DOCA_DLOG_ERR("unsupported field of type int in signature. key = %s",
					key);
				return -1;
			}
			continue;
		case json_type_string:
			if (!strcmp(key, "regex"))
				continue;
			else {
				DOCA_DLOG_ERR("unsupported field of type string in signature. key = %s",
					key);
				return -1;
			}
		case json_type_boolean:
			if (!strcmp(key, "anchor")) {
				ppt_entry->anchored = json_object_get_boolean(val);
				continue;
			} else {
				DOCA_DLOG_ERR("unsupported field of type string in signature. key = %s",
					key);
				return -1;
			}
		default:
			DOCA_DLOG_ERR("wrong type - string isn't expected");
			return -1;
		}
	}
	return 0;
}

static int
get_sig_action(struct dpi_sig_info_table *info_table, int sig_id)
{
	int i;

	for (i = 0; i < info_table->nb_entries; i++) {
		if (info_table->entry[i].sig_id == sig_id)
			return info_table->entry[i].action;
	}
	DOCA_DLOG_ERR("action parsing failed");
	return -1;
}

static int
parse_pp_table(struct json_object *ppt_obj, struct dpi_cdo_data *data, int *prot_translt,
	       int regex_num, int cond_num)
{
	struct json_object *ppt_entry;
	struct json_object *regex_id;
	struct post_processing_table *ppt_struct = &(data->pp_table);
	struct condition_table *cond_table = &data->cond_table;
	struct condition_table_entry *cond_entry = NULL;
	struct json_object *rules;
	struct json_object *rule;
	struct json_object *sig_id_obj;
	struct json_object *group_obj;
	struct json_object *priority_obj;
	struct json_object *json_obj;
	uint32_t num_rules, sig_id, num_entries;
	int order1 = -1, order2 = -1;
	int reg_id, ret, j = 0, i = 0, k, cond_id = 0;
	bool conditions_exist = false;

	if (json_object_get_type(ppt_obj) != json_type_array) {
		DOCA_DLOG_ERR("wrong object type for pp_table, json type is: %s",
			print_json_object(ppt_obj));
		return -EINVAL;
	}

	ppt_struct->nb_entries = regex_num;
	if (regex_num == 0)
		return 0;

	ppt_struct->entries = (struct pp_table_entry *)rte_zmalloc(NULL,
				   sizeof(struct pp_table_entry)*regex_num, 0);
	if (ppt_struct->entries == NULL) {
		DOCA_DLOG_ERR("mem allocation failed for pp_table\n");
		return -EINVAL;
	}
	num_entries = json_object_array_length(ppt_obj);
	if (num_entries == 0) {
		DOCA_DLOG_ERR("zero entries in pp_table is illegal\n");
		return -EINVAL;
	}

	cond_table->nb_entries = cond_num;
	if (cond_num > 0) {
		conditions_exist = true;

		cond_table->entries = (struct condition_table_entry *)rte_zmalloc(NULL,
					sizeof(struct condition_table_entry)*cond_num, 0);
		if (cond_table->entries == NULL) {
			DOCA_DLOG_ERR("mem allocation failed for condition table\n");
			ret = -EINVAL;
			goto cond_table_err;
		}
	}

	for (j = 0 ; j < num_entries ; j++) {
		order1 = -1, order2 = -1;

		ppt_entry = json_object_array_get_idx(ppt_obj, j);
		if (!json_object_object_get_ex(ppt_entry, "group", &group_obj)) {
			DOCA_DLOG_ERR("error parsing group_obj!!");
			ret = -EINVAL;
			goto entries_error;
		}

		if (json_object_object_get_ex(group_obj, "priority", &priority_obj)) {
			json_obj = json_object_array_get_idx(priority_obj, 0);
			order1 = json_object_get_int(json_obj);
			json_obj = json_object_array_get_idx(priority_obj, 1);
			order2 = json_object_get_int(json_obj);
		}

		if (!json_object_object_get_ex(ppt_entry, "sig_id", &sig_id_obj)) {
			DOCA_DLOG_ERR("error parsing sig_id!!");
			ret = -EINVAL;
			goto entries_error;
		}
		sig_id = json_object_get_int(sig_id_obj);

		if (!json_object_object_get_ex(group_obj, "rules", &rules)) {
			DOCA_DLOG_ERR("error parsing rules array!!");
			ret = -EINVAL;
			goto entries_error;
		}

		num_rules = json_object_array_length(rules);
		if (num_rules > 1) {
			cond_entry = cond_table->entries + cond_id;
			cond_entry->conds = (uint32_t *)rte_malloc(NULL,
							sizeof(uint32_t)*num_rules, 0);
			if (cond_entry->conds == NULL) {
				DOCA_DLOG_ERR("mem allocation failed for cond_entry\n");
				ret = -EINVAL;
				goto parse_error;
			}
			cond_entry->nb_cond = num_rules;
		}
		for (i = 0 ; i < num_rules ; i++) {
			rule = json_object_array_get_idx(rules, i);
			if (!json_object_object_get_ex(rule, "regex_id", &regex_id)) {
				DOCA_DLOG_ERR("error parsing regex_id!!");
				ret = -EINVAL;
				goto parse_error;
			}

			reg_id = json_object_get_int(regex_id);
			ppt_struct->entries[reg_id-1].sig_id = sig_id;
			ppt_struct->entries[reg_id-1].order1 = order1;
			ppt_struct->entries[reg_id-1].order2 = order2;
			ppt_struct->entries[reg_id-1].bitmask =
					ppt_struct->entries[reg_id-1].bitmask | DPI_PP_ORDER;
			ppt_struct->entries[reg_id-1].rule_id = reg_id;
			if (num_rules > 1) {
				ppt_struct->entries[reg_id-1].cond_id = cond_id;
				ppt_struct->entries[reg_id-1].bitmask =
						ppt_struct->entries[reg_id-1].bitmask | DPI_PP_COND;
				cond_entry->conds[i] = reg_id;
			}
			ret = parse_pp_entry(rule, &(ppt_struct->entries[reg_id-1]), prot_translt);
			if (ret) {
				ret = -EINVAL;
				goto parse_error;
			}

			ret = get_sig_action(&data->info_table,
						ppt_struct->entries[reg_id-1].sig_id);
			if (ret < 0)
				goto parse_error;
			ppt_struct->entries[reg_id-1].action = ret;
		}
		if (num_rules > 1)
			cond_id++;
	}
	return 0;

parse_error:
	DOCA_DLOG_ERR("Error parsing signature in regex_id %d, sig num = %d", j+1, i);
	if (conditions_exist) {
		for (k = 0 ; k <= cond_id ; k++)
			rte_free(cond_table->entries[k].conds);
	}
entries_error:
	if (conditions_exist)
		rte_free(cond_table->entries);
cond_table_err:
	rte_free(ppt_struct->entries);
	return ret;
}

static int
parse_cat_table(struct json_object *cat_t_arr, struct category_table *cat_table)
{
	struct json_object *cat_obj, *json_elem;
	uint32_t nb_entries;
	int i, cat_id;
	const char *cat_name;

	if (json_object_get_type(cat_t_arr) != json_type_array) {
		DOCA_DLOG_ERR("wrong object type for category_metadata, json type is: %s",
			print_json_object(cat_t_arr));
		return -1;
	}

	nb_entries = json_object_array_length(cat_t_arr);
	cat_table->nb_entries = nb_entries;
	if (nb_entries == 0)
		return 0;
	cat_table->entry = (char **)rte_malloc(NULL,
						  sizeof(char *)*nb_entries, 0);
	if (cat_table->entry == NULL)
		return -1;

	for (i = 0 ; i < nb_entries ; i++) {
		cat_obj = json_object_array_get_idx(cat_t_arr, i);

		if (!json_object_object_get_ex(cat_obj, "cat_id", &json_elem)) {
			DOCA_DLOG_ERR("couldn't parse category id");
			goto error;
		}
		if (json_object_get_type(json_elem) != json_type_int) {
			DOCA_DLOG_ERR("cat_id json object isn't int");
			goto error;
		}
		cat_id = json_object_get_int(json_elem);
		if (cat_id >= nb_entries) {
			DOCA_DLOG_ERR("cat_id is bigger than %d",
				nb_entries);
			goto error;
		}
		if (!json_object_object_get_ex(cat_obj, "cat_name",
					       &json_elem)) {
			DOCA_DLOG_ERR("couldn't parse category name");
			goto error;
		}
		if (json_object_get_type(json_elem) != json_type_string) {
			DOCA_DLOG_ERR("cat_name json object isn't string");
			goto error;
		}
		cat_name = json_object_get_string(json_elem);
		cat_table->entry[cat_id] = (char *)rte_malloc(NULL,
					sizeof(char)*(strlen(cat_name) + 1), 0);
		if (cat_table->entry[cat_id] == NULL) {
			DOCA_DLOG_ERR("malloc failed at %s", __func__);
			goto error;
		}
		strcpy(cat_table->entry[cat_id], cat_name);
	}
	return 0;
error:
	for (i = 0 ; i < nb_entries ; i++)
		rte_free(cat_table->entry[i]);
	rte_free(cat_table->entry);
	return -1;
}

static int
parse_fam_table(struct json_object *fam_t_arr, struct family_table *fam_table)
{
	struct json_object *fam_obj, *json_elem;
	uint32_t nb_entries;
	int i, fam_id;
	const char *fam_name;

	if (json_object_get_type(fam_t_arr) != json_type_array) {
		DOCA_DLOG_ERR("wrong object type for family_metadata, json type is: %s",
			print_json_object(fam_t_arr));
		return -1;
	}

	nb_entries = json_object_array_length(fam_t_arr);
	fam_table->nb_entries = nb_entries;
	if (nb_entries == 0)
		return 0;
	fam_table->entry = (char **)rte_malloc(NULL,
						  sizeof(char *)*nb_entries, 0);
	if (fam_table->entry == NULL)
		return -1;

	for (i = 0 ; i < nb_entries ; i++) {
		fam_obj = json_object_array_get_idx(fam_t_arr, i);

		if (!json_object_object_get_ex(fam_obj, "fam_id", &json_elem)) {
			DOCA_DLOG_ERR("couldn't parse family id");
			goto error;
		}
		if (json_object_get_type(json_elem) != json_type_int) {
			DOCA_DLOG_ERR("fam_id json object isn't int");
			goto error;
		}
		fam_id = json_object_get_int(json_elem);
		if (fam_id >= nb_entries) {
			DOCA_DLOG_ERR("fam_id is bigger than %d",
				nb_entries);
			goto error;
		}
		if (!json_object_object_get_ex(fam_obj, "fam_name",
					       &json_elem)) {
			DOCA_DLOG_ERR("couldn't parse family name");
			goto error;
		}
		if (json_object_get_type(json_elem) != json_type_string) {
			DOCA_DLOG_ERR("fam_name json object isn't string");
			goto error;
		}
		fam_name = json_object_get_string(json_elem);
		fam_table->entry[fam_id] = (char *)rte_malloc(NULL,
					sizeof(char)*(strlen(fam_name) + 1), 0);
		if (fam_table->entry[fam_id] == NULL) {
			DOCA_DLOG_ERR("malloc failed at %s", __func__);
			goto error;
		}
		strcpy(fam_table->entry[fam_id], fam_name);
	}
	return 0;
error:
	for (i = 0 ; i < nb_entries ; i++)
		rte_free(fam_table->entry[i]);
	rte_free(fam_table->entry);
	return -1;
}

static int
parse_sig_info_table(struct json_object *metadata_arr,
		     struct dpi_cdo_data *data)
{
	struct json_object *app_obj, *json_elem, *cat_arr;
	struct cdo_sig_info *app_info;
	uint32_t nb_entries;
	int i, j, tmp_int, nb_cat;
	const char *sig_name, *sig_action;

	if (json_object_get_type(metadata_arr) != json_type_array) {
		DOCA_DLOG_ERR("wrong object type for app_metadata, json type is: %s",
			print_json_object(metadata_arr));
		return -1;
	}

	nb_entries = json_object_array_length(metadata_arr);
	data->info_table.nb_entries = nb_entries;
	if (nb_entries == 0)
		return 0;
	data->info_table.entry = (struct cdo_sig_info *)rte_malloc(NULL,
				     sizeof(struct cdo_sig_info)*nb_entries, 0);
	if (data->info_table.entry == NULL) {
		DOCA_DLOG_ERR("memory allocation failed");
		return -1;
	}

	for (i = 0 ; i < nb_entries ; i++) {
		app_info = data->info_table.entry + i;
		app_obj = json_object_array_get_idx(metadata_arr, i);

		if (!json_object_object_get_ex(app_obj, "sig_id", &json_elem)) {
			DOCA_DLOG_ERR("couldn't parse sig_id");
			return -1;
		}
		if (json_object_get_type(json_elem) != json_type_int) {
			DOCA_DLOG_ERR("wrong object type for sig_id");
			return -1;
		}
		app_info->sig_id = json_object_get_int(json_elem);

		if (!json_object_object_get_ex(app_obj, "name", &json_elem)) {
			DOCA_DLOG_ERR("couldn't parse sig_name");
			return -1;
		}
		if (json_object_get_type(json_elem) != json_type_string) {
			DOCA_DLOG_ERR("wrong object type for sig_name");
			return -1;
		}
		sig_name = json_object_get_string(json_elem);
		app_info->name = (char *)rte_malloc(NULL,
					sizeof(char)*(strlen(sig_name) + 1), 0);
		if (app_info->name == NULL) {
			DOCA_DLOG_ERR("malloc failed at %s", __func__);
			return -1;
		}
		strcpy(app_info->name, sig_name);

		if (json_object_object_get_ex(app_obj, "action", &json_elem)) {
			if (json_object_get_type(json_elem) != json_type_string) {
				DOCA_DLOG_ERR("wrong object type for sig action");
				return -1;
			}
			sig_action = json_object_get_string(json_elem);
			tmp_int = action_str2num(sig_action);
			app_info->action = tmp_int;
		}

		if (json_object_object_get_ex(app_obj, "aging", &json_elem)) {
			if (json_object_get_type(json_elem) != json_type_int) {
				DOCA_DLOG_ERR("wrong object type for sig aging");
				return -1;
			}
			app_info->aging_ms = json_object_get_int(json_elem);
		} else
			app_info->aging_ms = 0;

		if (json_object_object_get_ex(app_obj, "family", &json_elem)) {
			if (json_object_get_type(json_elem) != json_type_int) {
				DOCA_DLOG_ERR("wrong object type for sig family");
				return -1;
			}
			tmp_int = json_object_get_int(json_elem);
			if (tmp_int >= data->fam_table.nb_entries) {
				DOCA_DLOG_ERR("app metadata has wrong family id");
				return -1;
			}
			app_info->family = data->fam_table.entry[tmp_int];
		} else
			app_info->family = NULL;

		if (!json_object_object_get_ex(app_obj, "categories", &cat_arr)) {
			app_info->categories = NULL;
			continue;
		}
		if (json_object_get_type(cat_arr) != json_type_array) {
			DOCA_DLOG_ERR("wrong object type for sig categories");
			return -1;
		}
		nb_cat = json_object_array_length(cat_arr);
		app_info->nb_categories = nb_cat;
		if (nb_cat == 0)
			continue;
		app_info->categories = (char **)rte_malloc(NULL,
						      sizeof(char *)*nb_cat, 0);
		if (app_info->categories == NULL) {
			DOCA_DLOG_ERR("memory allocation failed");
			return -1;
		}
		for (j = 0 ; j < nb_cat ; j++) {
			json_elem = json_object_array_get_idx(cat_arr, j);
			if (json_object_get_type(json_elem) != json_type_int) {
				DOCA_DLOG_ERR("wrong object type for category id");
				return -1;
			}
			tmp_int = json_object_get_int(json_elem);
			if (tmp_int >= data->cat_table.nb_entries) {
				DOCA_DLOG_ERR("app metadata has wrong category id");
				return -1;
			}
			app_info->categories[j] = data->cat_table.entry[tmp_int];
		}
	}
	return 0;
}

static bool __rte_unused
version_check(struct json_object *maj_obj, struct json_object *min_obj, struct json_object *patch_obj)
{
	bool failed = false;

	if (json_object_get_type(maj_obj) != json_type_int ||
	    json_object_get_type(min_obj) != json_type_int) {
		DOCA_DLOG_ERR("all version objects should be of type int.\n"
			      "major_ver type = %s, minor_ver type = %s",
				print_json_object(maj_obj), print_json_object(min_obj));
		return -1;
	}
	/* checking that the CDO version is not higher than the DPI */
	if (!DOCA_VERSION_LTE_CURRENT(json_object_get_int(maj_obj), json_object_get_int(min_obj),
				      0)) {
		DOCA_LOG_ERR("CDO file version is higher then current DPI version");
		failed = true;
	}
	/* checking that the CDO version is at least 1.5.1000 */
	if (DOCA_VERSION_NUM(1, 5, 1000) > DOCA_VERSION_NUM(json_object_get_int(maj_obj), json_object_get_int(min_obj),
								json_object_get_int(patch_obj))) {
		DOCA_LOG_ERR("DPI version must be at least 1.5.1000");
		failed = true;
	}
	return failed;
}

int cdo_decode(const char *filename, struct dpi_cdo_data **data,
		   char **rof2_buf, uint32_t *rof2_buf_size,
		   char **hs_buf, uint32_t *hs_buf_size)
{
	char *buffer = NULL;
	struct json_object *parsed_json;
	struct json_object *prot_tran;
	struct json_object *minor_ver;
	struct json_object *patch_ver;
	struct json_object *major_ver;
	struct json_object *pp_table;
	struct json_object *app_metadata;
	struct json_object *category_metadata;
	struct json_object *family_metadata;
	struct json_object *regex_rof2;
	struct json_object *regex_rof2_bin;
	struct json_object *regex_hs;
	struct json_object *json_obj;
	long res;
	int regex_num, conds_num;
	const char *regex_rof;
	int *prot_translt = NULL;
	int i = 0, nb_entries = 0;
	struct pp_table_entry *pp_t_e, *pp_t_e_tmp;
	bool hs = true;
	bool rof_bin = true;

	res = read_rules_db(filename, &buffer);
	if (res < 0)
		return res;

	*data = (struct dpi_cdo_data *)rte_malloc(NULL, sizeof(struct dpi_cdo_data), 0);
	if (*data == NULL) {
		DOCA_DLOG_ERR("malloc failed for dpi_cdo_data struct");
		res = -ENOMEM;
		goto error_data;
	}
	parsed_json = json_tokener_parse(buffer);
	if (parsed_json == NULL) {
		DOCA_DLOG_ERR("couldn't parse the input file as json tokens");
		res = -1;
		goto error_json_parse;
	}

	if (!json_object_object_get_ex(parsed_json, "protocol_to_number",
				       &prot_tran)) {
		DOCA_DLOG_ERR("couldn't parse protocol_to_number object");
		res = -1;
		goto error_protocol;
	}
	if (fill_prot_translate_map(prot_tran, &prot_translt)) {
		res = -1;
		goto error_protocol;
	}
	if (!json_object_object_get_ex(parsed_json, "patch_version",
				       &patch_ver)) {
		DOCA_DLOG_ERR("couldn't parse patch_version object");
		res = -1;
		goto error_version;
	}
	if (!json_object_object_get_ex(parsed_json, "minor_version",
				       &minor_ver)) {
		DOCA_DLOG_ERR("couldn't parse minor_version object");
		res = -1;
		goto error_version;
	}
	if (!json_object_object_get_ex(parsed_json, "major_version",
				       &major_ver)) {
		DOCA_DLOG_ERR("couldn't parse major_version object");
		res = -1;
		goto error_version;
	}

#ifndef DOCA_DEBUG
	if (version_check(major_ver, minor_ver, patch_ver)) {
		res = -ENOTSUP;
		goto error_version;
	}
#endif /* DOCA_DEBUG */

	if (!json_object_object_get_ex(parsed_json, "number_of_regex_rules",
				       &json_obj)) {
		DOCA_DLOG_ERR("couldn't parse number_of_regex_rules object");
		res = -1;
		goto error_version;
	}
	regex_num = json_object_get_int(json_obj);

	if (!json_object_object_get_ex(parsed_json, "number_of_conditions",
				       &json_obj)) {
		DOCA_DLOG_ERR("couldn't parse number_of_conditions object");
		res = -1;
		goto error_version;
	}
	conds_num = json_object_get_int(json_obj);

	if (!json_object_object_get_ex(parsed_json, "category_metadata",
				       &category_metadata)) {
		DOCA_DLOG_ERR("couldn't parse category_metadata object");
		res = -1;
		goto error_cat_table;
	}
	if (parse_cat_table(category_metadata, &(*data)->cat_table)) {
		DOCA_DLOG_ERR("parse_category function failed");
		res = -1;
		goto error_cat_table;
	}

	if (!json_object_object_get_ex(parsed_json, "family_metadata",
				       &family_metadata)) {
		DOCA_DLOG_ERR("couldn't parse family_metadata object");
		res = -1;
		goto error_fam_table;
	}
	if (parse_fam_table(family_metadata, &(*data)->fam_table)) {
		DOCA_DLOG_ERR("parse_family function failed");
		res = -1;
		goto error_fam_table;
	}

	if (!json_object_object_get_ex(parsed_json, "app_metadata",
				       &app_metadata)) {
		DOCA_DLOG_ERR("couldn't parse app_metadata object");
		res = -1;
		goto error_sig_info;
	}
	if (parse_sig_info_table(app_metadata, (*data))) {
		DOCA_DLOG_ERR("parse_sig_info function failed");
		res = -1;
		goto error_sig_info;
	}

	if (!json_object_object_get_ex(parsed_json, "pp_table", &pp_table)) {
		DOCA_DLOG_ERR("couldn't parse pp_table object");
		res = -1;
		goto error_pp_table;
	}
	if (parse_pp_table(pp_table, (*data), prot_translt, regex_num, conds_num)) {
		res = -1;
		goto error_pp_table;
	}

	if (!json_object_object_get_ex(parsed_json, "regex_bin", &regex_rof2_bin)) {
		DOCA_DLOG_DBG("couldn't parse regex_bin object");
		rof_bin = false;

		/* if there is no binary rof file then check for textual rof */
		if (!json_object_object_get_ex(parsed_json, "regex", &regex_rof2)) {
			DOCA_DLOG_ERR("couldn't parse regex object");
			res = -EINVAL;
			goto error_rof2;
		}
		if (json_object_get_type(regex_rof2) != json_type_string) {
			DOCA_DLOG_ERR("wrong object type for regex, json type is: %s",
				print_json_object(regex_rof2));
			res = -EINVAL;
			goto error_rof2;
		}
		regex_rof = json_object_get_string(regex_rof2);
		*rof2_buf = strdup(regex_rof);
		if (*rof2_buf == NULL) {
			DOCA_DLOG_ERR("strdup failed for rof2_buf");
			res = -ENOMEM;
			goto error_rof2;
		}
		*rof2_buf_size = json_object_get_string_len(regex_rof2);
	}

	if (rof_bin) {
		int i;

		*rof2_buf_size = json_object_array_length(regex_rof2_bin);
		*rof2_buf = (char *)malloc(*rof2_buf_size * sizeof(char));
		if (*rof2_buf == NULL) {
			DOCA_DLOG_ERR("malloc failed for rof2_buf_bin");
			res = -ENOMEM;
			goto error_rof2;
		}
		for (i = 0; i < *rof2_buf_size; i++) {
			struct json_object *c = json_object_array_get_idx(regex_rof2_bin, i);

			(*rof2_buf)[i] = json_object_get_int(c);
		}
	}


	if (!json_object_object_get_ex(parsed_json, "hs_regex", &regex_hs)) {
		DOCA_DLOG_ERR("couldn't parse hs regex object");
		hs = false;
	}
	if (hs) {
		int i;

		*hs_buf_size = json_object_array_length(regex_hs);
		if (*hs_buf_size) {
			*hs_buf = (char *)malloc(*hs_buf_size * sizeof(char));
			if (*hs_buf == NULL) {
				DOCA_DLOG_ERR("malloc failed for rof2_buf_bin");
				res = -ENOMEM;
				goto error_hs;
			}
			for (i = 0; i < *hs_buf_size; i++) {
				struct json_object *c = json_object_array_get_idx(regex_hs, i);

				(*hs_buf)[i] = json_object_get_int(c);
			}
		}
	}
	json_object_put(parsed_json);
	free(buffer);
	return 0;

error_hs:
	free(*rof2_buf);
error_rof2:
	nb_entries = (*data)->pp_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++) {
		pp_t_e = &((*data)->pp_table.entries[i]);
		pp_t_e_tmp = pp_t_e->next;
		while (pp_t_e_tmp != NULL) {
			pp_t_e = pp_t_e_tmp->next;
			rte_free(pp_t_e_tmp);
			pp_t_e_tmp = pp_t_e;
		};
	}
	rte_free((*data)->pp_table.entries);

error_pp_table:
	nb_entries = (*data)->info_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++) {
		rte_free((*data)->info_table.entry[i].categories);
		rte_free((*data)->info_table.entry[i].name);
	}
	rte_free((*data)->info_table.entry);

error_sig_info:
	nb_entries = (*data)->fam_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++)
		rte_free((*data)->fam_table.entry[i]);
	rte_free((*data)->fam_table.entry);

error_fam_table:
	nb_entries = (*data)->cat_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++)
		rte_free((*data)->cat_table.entry[i]);
	rte_free((*data)->cat_table.entry);

error_cat_table:
	nb_entries = (*data)->cond_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++)
		rte_free((*data)->cond_table.entries[i].conds);
	rte_free((*data)->cond_table.entries);

error_version:
error_protocol:
	json_object_put(parsed_json);
	rte_free(prot_translt);
error_json_parse:
	rte_free(*data);
error_data:
	free(buffer);
	DOCA_DLOG_DBG("dpi_dcode_cdo function failed");
	return res;
}

void cdo_dealloc(struct dpi_cdo_data *data)
{
	if (data == NULL)
		return;

	int nb_entries, i;
	struct pp_table_entry *pp_t_e, *pp_t_e_tmp;

/* remove pp_table */
	nb_entries = data->pp_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++) {
		pp_t_e = &(data->pp_table.entries[i]);
		pp_t_e_tmp = pp_t_e->next;
		while (pp_t_e_tmp != NULL) {
			pp_t_e = pp_t_e_tmp->next;
			rte_free(pp_t_e_tmp);
			pp_t_e_tmp = pp_t_e;
		};
	}
	if (nb_entries)
		rte_free(data->pp_table.entries);
	DOCA_DLOG_DBG("pp_table dealloc succeeded");

/* remove info_table */
	nb_entries = data->info_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++) {
		rte_free(data->info_table.entry[i].categories);
		rte_free(data->info_table.entry[i].name);
	}
	if (nb_entries)
		rte_free(data->info_table.entry);
	DOCA_DLOG_DBG("info_table dealloc succeeded");

/* remove fam_table */
	nb_entries = data->fam_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++)
		rte_free(data->fam_table.entry[i]);
	if (nb_entries)
		rte_free(data->fam_table.entry);
	DOCA_DLOG_DBG("fam_table dealloc succeeded");

/* remove cat_table */
	nb_entries = data->cat_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++)
		rte_free(data->cat_table.entry[i]);
	if (nb_entries)
		rte_free(data->cat_table.entry);
	DOCA_DLOG_DBG("cat_table dealloc succeeded");

/* remove cond_table */
	nb_entries = data->cond_table.nb_entries;
	for (i = 0 ; i < nb_entries ; i++)
		rte_free(data->cond_table.entries[i].conds);
	if (nb_entries)
		rte_free(data->cond_table.entries);
	rte_free(data);
	DOCA_DLOG_DBG("cond_table dealloc succeeded");
}

/* get the first entry from the pp_table for the given regex_id */
const struct pp_table_entry *cdo_pp_entry_get(const struct dpi_cdo_data *data,
					      uint32_t regex_id)
{
	assert(regex_id > 0);
	if (data->pp_table.nb_entries < regex_id) {
		DOCA_DLOG_ERR("there is no regex_id %d in the post processing table",
			regex_id);
		return NULL;
	}
	return &data->pp_table.entries[regex_id - 1];
}

/* get the condition_table's entry for given cond_id */
const struct condition_table_entry *
cdo_cond_entry_get(const struct dpi_cdo_data *data, uint32_t cond_id)
{
	if (data->cond_table.nb_entries <= cond_id) {
		DOCA_DLOG_ERR("there is no entry number %d in the condition table",
			cond_id);
		return NULL;
	}
	return &data->cond_table.entries[cond_id];
}

/* get name and id for the given sig_id */
int
cdo_signature_get(const struct dpi_cdo_data *data, uint32_t sig_id,
		  struct doca_dpi_sig_data *sig_data)
{
	int i;

	for (i = 0; i < data->info_table.nb_entries; i++)
		if (data->info_table.entry[i].sig_id == sig_id) {
			strlcpy(sig_data->name,
				data->info_table.entry[i].name,
				sizeof(sig_data->name));
			sig_data->sig_id = sig_id;
			return 0;
		}

	DOCA_DLOG_ERR("sig_id %d couldn't be found", sig_id);
	return -EINVAL;
}

/* get name and id for all sig_ids */
int
cdo_signatures_get(const struct dpi_cdo_data *data,
		   struct doca_dpi_sig_data **sig_data)
{
	int i;

	if (data->info_table.nb_entries == 0) {
		DOCA_DLOG_DBG("no signatures metadata is available");
		return 0;
	}
	*sig_data = (struct doca_dpi_sig_data *)rte_malloc(NULL,
					sizeof(struct doca_dpi_sig_data)
					* data->info_table.nb_entries, 0);
	if (!*sig_data)
		return -ENOMEM;
	for (i = 0; i < data->info_table.nb_entries; i++) {
		strlcpy((*sig_data)[i].name, data->info_table.entry[i].name,
			sizeof((*sig_data)[i].name));
		(*sig_data)[i].sig_id = data->info_table.entry[i].sig_id;
	}
	return data->info_table.nb_entries;
}
