/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file
 *
 * CDO decoder
 *
 * CDO decoder to a json CDO file produced by the DPI compiler.
 *
 * CDO file contains:
 * 1. Compiled Regex rules (rof2).
 * 2. Post processing table.
 * 3. Condition table.
 * 4. Signatures metadata.
 * 5. Category metadata.
 * 6. L3 next app mapping.
 * 7. TCP/UDP port mapping*.
 *
 * TCP/UDP port mapping maps l4 port to a signatures id. This applies
 * for signatures that have no regex.
 *
 * The Json format is expected to be a s follows:
 * Invalid fields will not exist in the Jason.
 * {
 *         minor_version: <number>
 *         major_version: <number>
 *         protocol_table : [
 *                 {[<protocol>,<id>]}
 *         ]
 *         field_table : [
 *                 {[<field>, <id>]}
 *         ]
 *         num_signatures : <number>
 *         pp_table : [
 *		{
 *		"sig_id":1,
 *		"group":{
 *			"priority":[
 *				order <number>
 *				order <number>
 *			],
 *			"rules":[
 *				{
 *					"regex_id": <number>
 *					"regex": <string>
 *					"field": <number>
 *					"direction": <number>
 *				},
 *				...
 *				]
 *			}
 *		},
 *              ...
 *         ]
 *         app_metadata: [
 *                 {
 *			sig_id: <number>
 *			name: <app name>
 *			action: <action>
 *			categories: [<cat ID>, <cat ID>, ...]
 *			family: <number>
 *			agining: <number>
 *                 }
 *                 ...
 *         ]
 *         category_metadata: [
 *              {
 *                           cat_id: <number>
 *                           cat_name: <name>
 *                 }
 *                 ...
 *         ]
 *         family_metadata: [
 *                 {
 *                           cat_id: <number>
 *                           cat_name: <name>
 *                 }
 *                 ...
 *         ]
 *	  regex: "string of rof2 file"
 *	  regex_bin: [ array of rof2 binary file ]
 * }
 *
 * The decoder will map the CDO file into a dpi_cdo_data struct.
 */

#ifndef CDO_DECODER_H_
#define CDO_DECODER_H_

#include <stdint.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "parser/dpi_parser_types.h"
#include "doca_dpi.h"

#ifdef __cplusplus
extern "C" {
#endif

enum dpi_pp_entry_p {
	DPI_PP_L7_PROTOCOL = 1 << 0,
	DPI_PP_FIELD       = 1 << 1,
	DPI_PP_PORT        = 1 << 2,
	DPI_PP_COND        = 1 << 3,
	DPI_PP_ORDER       = 1 << 4,
	DPI_PP_DIRECTION   = 1 << 5,
	DPI_PP_MINDATA     = 1 << 6,
	DPI_PP_L4_PROTOCOL = 1 << 7,
	DPI_PP_L3_PROTOCOL = 1 << 8,
	DPI_PP_OFFSET	   = 1 << 9,
	DPI_PP_ADDRESS     = 1 << 10,
	DPI_PP_ADDR_IPV4   = 1 << 11,
	DPI_PP_ADDR_IPV6   = 1 << 12,
};


enum dpi_direction {
	DPI_DIR_CTS,
	DPI_DIR_STC,
	DPI_DIR_BOTH,
};

union addr {
		struct in_addr ipv4;
		struct in6_addr ipv6;
};

struct cdo_sig_info {
	uint32_t aging_ms; /* Non-zero means agining timeout needs to
			    * be updated for this flow
			    */
	uint32_t sig_id;
	char *name;
	int action;
	uint32_t nb_categories;
	char **categories;
	char *family;
};

struct pp_table_entry {
	uint32_t bitmask;
	uint32_t rule_id;
	uint32_t sig_id;
	uint16_t port_from;
	uint16_t port_to;
	uint8_t l7_protocol;
	uint8_t l4_protocol;
	uint8_t l3_protocol;
	uint8_t field;
	uint8_t anchored;
	uint8_t dollar;
	uint32_t cond_id;
	uint32_t order1;
	uint32_t order2;
	uint32_t offset;
	uint32_t direction; /* CTS, STC, BOTH */
	uint8_t subnet_mask;
	union addr addr;
	struct pp_table_entry *next;
	int action;
};

struct post_processing_table {
	uint32_t nb_entries;
	struct pp_table_entry *entries;
};

struct condition_table_entry {
	uint32_t nb_cond;
	uint32_t *conds;
};

struct condition_table {
	uint32_t nb_entries;
	struct condition_table_entry *entries;
};

struct dpi_sig_info_table {
	uint32_t nb_entries;
	struct cdo_sig_info *entry;
};

struct category_table {
	uint32_t nb_entries;
	char **entry;
};

struct family_table {
	uint32_t nb_entries;
	char **entry;
};

struct dpi_cdo_data;

int cdo_decode(const char *filename, struct dpi_cdo_data **data,
	       char **rof2_buf, uint32_t *rof2_buf_size,
	       char **hs_rof2_buf, uint32_t *hs_rof2_buf_size);
void cdo_dealloc(struct dpi_cdo_data *data);
const struct pp_table_entry *cdo_pp_entry_get(const struct dpi_cdo_data *data,
					      uint32_t regex_id);
const struct condition_table_entry *cdo_cond_entry_get(
	const struct dpi_cdo_data *data, uint32_t cond_id);

int cdo_signature_get(const struct dpi_cdo_data *data, uint32_t sig_id,
			  struct doca_dpi_sig_data *sig_data);
int cdo_signatures_get(const struct dpi_cdo_data *data,
			   struct doca_dpi_sig_data **sig_data);

#ifdef __cplusplus
}
#endif

#endif /* CDO_DECODER_H_ */
