/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <json-c/json.h>
#include <rte_malloc.h>
#include <rte_regexdev.h>
#include <rte_ring.h>

#ifdef DOCA_DPI_USE_HS
#include <hs/hs_compile.h>
#include <hs/hs_runtime.h>
#endif

#include <doca_log.h>

#include "dpi_regex.h"

DOCA_LOG_REGISTER(DPI::REGEX);

#define META_MAX_BUF (1 << 10)
#define REGEX_MAX_BUF (1 << 14)
#define REGEX_MAX_NB_SEGS 32
#define MAX_REGEX_HS_MATCH 254

struct dpi_regex_queue_entry {
	uint16_t nb_matches;
	struct rte_regex_ops *op;
	struct rte_mbuf *regex_buf;
	struct rte_mbuf *meta_buf;
	void *user_ptr;
	bool hs;
};

struct dpi_regex_queue_ctx {
	struct dpi_regex_dev_ctx *dev_ctx;
	uint32_t hw_queue_id;
	struct dpi_regex_queue_entry *entries;
	uint16_t nb_entries;
	struct rte_ring *free_list;
	struct rte_ring *enqueued_list;
	struct rte_mempool *regex_buffers;
	struct rte_mempool *meta_buffers;
	struct rte_regex_ops *op;

#ifdef DOCA_DPI_USE_HS
	hs_scratch_t *hs_scratch;
#endif
};

struct dpi_regex_dev_ctx {
	bool hs;
	bool rxp;
	int dev_id;
	struct rte_regexdev_info info;
	int num_queues;
	struct dpi_regex_queue_ctx *regex_queue;
#ifdef DOCA_DPI_USE_HS
	hs_database_t *hs_db;
#endif
};

static int
dpdk_dev_setup(struct dpi_regex_dev_ctx *dev_ctx)
{
	int res;
	struct rte_regexdev_config dev_conf = {
		.nb_queue_pairs = dev_ctx->num_queues,
		.nb_groups = 1,
		.nb_max_matches = 254,
		.nb_rules_per_group = 1,
	};

	res = rte_regexdev_info_get(dev_ctx->dev_id, &dev_ctx->info);
	if (res < 0) {
		DOCA_DLOG_ERR("Failed to query regex device %d info.", dev_ctx->dev_id);
		return res;
	}
	res = rte_regexdev_configure(dev_ctx->dev_id, &dev_conf);
	if (res < 0) {
		DOCA_DLOG_ERR("Failed to configure regex device %d.", dev_ctx->dev_id);
		return res;
	}
	dev_ctx->rxp = true;
	return 0;
}

struct dpi_regex_dev_ctx *
dpi_regex_dev_alloc(int dpdk_dev_id, int num_queues, int *error)
{
	struct dpi_regex_dev_ctx *dev_ctx;


	dev_ctx = rte_zmalloc(NULL, sizeof(*dev_ctx), 0);
	if (!dev_ctx) {
		*error = -ENOMEM;
		return NULL;
	}

	dev_ctx->regex_queue = rte_calloc(NULL, num_queues, sizeof(*dev_ctx->regex_queue), 0);
	if (!dev_ctx->regex_queue) {
		*error = -ENOMEM;
		goto regex_alloc_q_err;
	}
	dev_ctx->dev_id = dpdk_dev_id;
	dev_ctx->num_queues = num_queues;
	if (rte_regexdev_count() == 0) {
#ifndef DOCA_DPI_USE_HS
		DOCA_DLOG_ERR("No regex device found");
		*error = -EEXIST;
		goto regex_cfg_err;
#else
		dev_ctx->info.max_matches = MAX_REGEX_HS_MATCH;
#endif
	} else {
		*error = dpdk_dev_setup(dev_ctx);
		if (*error < 0)
			goto regex_cfg_err;
	}
	return dev_ctx;

regex_cfg_err:
	rte_free(dev_ctx->regex_queue);
regex_alloc_q_err:
	rte_free(dev_ctx);
	return NULL;
}

void
dpi_regex_dev_dealloc(struct dpi_regex_dev_ctx *dev_ctx)
{
#ifdef DOCA_DPI_USE_HS
	if (dev_ctx->hs) {
		hs_free_database(dev_ctx->hs_db);
		dev_ctx->hs = false;
	}
#else
	if (dev_ctx->rxp) {
		rte_regexdev_stop(dev_ctx->dev_id);
		dev_ctx->rxp = false;
	}
#endif
	if (dev_ctx->regex_queue)
		rte_free(dev_ctx->regex_queue);
	rte_free(dev_ctx);
}

int
dpi_regex_load_sig(struct dpi_regex_dev_ctx *dev_ctx, int db_type, void *db, size_t len)
{
	int ret;

	if (db_type == DPI_REGEX_TYPE_RXP && dev_ctx->rxp) {
		ret = rte_regexdev_rule_db_import(dev_ctx->dev_id, db, len);
		if (ret)
			return ret;
	}
	if (db_type == DPI_REGEX_TYPE_HS) {
#ifdef DOCA_DPI_USE_HS
		int i, j;

		if (dev_ctx->hs)
			hs_free_database(dev_ctx->hs_db);
		ret = hs_deserialize_database(db, len, &dev_ctx->hs_db);
		if (ret)
			return ret;
		for (i = 0; i < dev_ctx->num_queues; i++) {
			ret = hs_alloc_scratch(dev_ctx->hs_db, &dev_ctx->regex_queue[i].hs_scratch);
			if (ret < 0) {
				for (j = 0; j < i; j++)
					hs_free_scratch(dev_ctx->regex_queue[j].hs_scratch);
				hs_free_database(dev_ctx->hs_db);
				return ret;
			}
		}
		dev_ctx->hs = true;
#else
		return -ENOTSUP;
#endif
	}
	return 0;
}

static int
regex_entry_init(struct dpi_regex_queue_ctx *queue_ctx, struct dpi_regex_queue_entry *entry)
{
	entry->op = rte_zmalloc(NULL, sizeof(*entry->op) +
				sizeof(struct rte_regexdev_match) *
				queue_ctx->dev_ctx->info.max_matches, 0);
	if (!entry->op)
		return -ENOMEM;

	entry->regex_buf = rte_pktmbuf_alloc(queue_ctx->regex_buffers);
	if (!entry->regex_buf) {
		rte_free(entry->op);
		return -ENOMEM;
	}
	entry->meta_buf = rte_pktmbuf_alloc(queue_ctx->meta_buffers);
	if (!entry->meta_buf) {
		rte_pktmbuf_free(entry->regex_buf);
		rte_free(entry->op);
		return -ENOMEM;
	}
	return 0;
}

static void
regex_entry_destroy(struct dpi_regex_queue_entry *entry)
{
	if (entry->op)
		rte_free(entry->op);
	if (entry->regex_buf)
		rte_pktmbuf_free(entry->regex_buf);
	if (entry->meta_buf)
		rte_pktmbuf_free(entry->meta_buf);
}

static void
dpdk_regex_queue_destroy(struct dpi_regex_queue_ctx *ctx)
{
	int j;

	for (j = 0; j < ctx->nb_entries; j++)
		regex_entry_destroy(&ctx->entries[j]);
	rte_mempool_free(ctx->regex_buffers);
	rte_mempool_free(ctx->meta_buffers);
}

static int
dpdk_regex_queue_setup(struct dpi_regex_queue_ctx *ctx)
{
	char regex_name[256];
	char meta_name[256];
	int res;
	struct rte_regexdev_qp_conf qp_conf = {
		.nb_desc = ctx->nb_entries,
		.qp_conf_flags = 0, /* in order */
	};

	if (ctx->dev_ctx->rxp) {
		res = rte_regexdev_queue_pair_setup(ctx->dev_ctx->dev_id, ctx->hw_queue_id, &qp_conf);
		if (res < 0)
			return res;
	}

	sprintf(regex_name, "queue_regex_buf%d", ctx->hw_queue_id);
	sprintf(meta_name, "queue_meta_buf%d", ctx->hw_queue_id);

	ctx->regex_buffers = rte_pktmbuf_pool_create(regex_name,
				ctx->nb_entries, 0, RTE_MBUF_PRIV_ALIGN,
				RTE_PKTMBUF_HEADROOM + REGEX_MAX_BUF, rte_socket_id());
	if (!ctx->regex_buffers) {
		DOCA_DLOG_ERR("Failed to setup regex buffers pool.");
		return -ENOMEM;
	}
	ctx->meta_buffers = rte_pktmbuf_pool_create(meta_name,
				ctx->nb_entries, 0, RTE_MBUF_PRIV_ALIGN,
				RTE_PKTMBUF_HEADROOM + META_MAX_BUF, rte_socket_id());
	if (!ctx->meta_buffers) {
		DOCA_DLOG_ERR("Failed to setup meta buffers pool.");
		rte_mempool_free(ctx->regex_buffers);
		return -ENOMEM;
	}
	return 0;
}

void
dpi_regex_queue_dealloc(struct dpi_regex_queue_ctx *ctx)
{
	dpdk_regex_queue_destroy(ctx);
	rte_free(ctx->entries);
	rte_ring_free(ctx->enqueued_list);
	rte_ring_free(ctx->free_list);
#ifdef DOCA_DPI_USE_HS
	if (ctx->hs_scratch)
		hs_free_scratch(ctx->hs_scratch);
#endif
}

struct dpi_regex_queue_ctx *
dpi_regex_queue_alloc(struct dpi_regex_dev_ctx *dev_ctx, int hw_queue_id, uint16_t queue_sz)
{
	struct dpi_regex_queue_ctx *ctx;
	char name[256];
	int j, i, ret;

	if (hw_queue_id >= dev_ctx->num_queues)
		return NULL;
	ctx = &dev_ctx->regex_queue[hw_queue_id];
	ctx->nb_entries = queue_sz;
	ctx->hw_queue_id = hw_queue_id;
	ctx->dev_ctx = dev_ctx;

	sprintf(name, "regex_free__list%d", hw_queue_id);
	ctx->free_list = rte_ring_create(name, queue_sz * 2, rte_socket_id(),
					 RING_F_SP_ENQ | RING_F_SC_DEQ);
	if (ctx->free_list == NULL)
		return NULL;
	sprintf(name, "regex_enq__list%d", hw_queue_id);
	ctx->enqueued_list = rte_ring_create(name, queue_sz * 2, rte_socket_id(),
					     RING_F_SP_ENQ | RING_F_SC_DEQ);
	if (ctx->enqueued_list == NULL)
		goto regex_en_q_err;

	ctx->entries = rte_calloc(NULL, queue_sz, sizeof(*ctx->entries), 0);
	if (ctx->entries == NULL)
		goto entries_alloc_err;

	ret = dpdk_regex_queue_setup(ctx);
	if (ret < 0)
		goto q_setup_err;

	ctx->op = rte_zmalloc(NULL, sizeof(*ctx->op) +
		sizeof(struct rte_regexdev_match) *
		ctx->dev_ctx->info.max_matches, 0);
	if (!ctx->op)
		goto regex_op_err;


	for (j = 0; j < queue_sz; j++) {
		ret = rte_ring_enqueue(ctx->free_list, &ctx->entries[j]);
		if (ret < 0)
			goto entry_init_err;
		ret = regex_entry_init(ctx, &ctx->entries[j]);
		if (ret < 0)
			goto entry_init_err;
	}

	return ctx;
entry_init_err:
	for (i = 0; i < j; i++)
		regex_entry_destroy(&ctx->entries[j]);
	rte_free(ctx->op);
regex_op_err:
	rte_mempool_free(ctx->regex_buffers);
	rte_mempool_free(ctx->meta_buffers);
q_setup_err:
	rte_free(ctx->entries);
entries_alloc_err:
	rte_ring_free(ctx->enqueued_list);
regex_en_q_err:
	rte_ring_free(ctx->free_list);
	return NULL;
}

#ifdef DOCA_DPI_USE_HS
static int HS_CDECL
scan_match_handler(unsigned int id, unsigned long long from, unsigned long long to,
		   unsigned int flags, void *context)
{
	const struct dpi_regex_queue_entry *regex_entry = (struct dpi_regex_queue_entry *)context;
	struct rte_regex_ops **ops = (struct rte_regex_ops **)&regex_entry->op;

	ops[0]->matches[ops[0]->nb_matches].start_offset = from;
	ops[0]->matches[ops[0]->nb_matches].end_offset = to;
	ops[0]->matches[ops[0]->nb_matches].len = from - to;
	ops[0]->matches[ops[0]->nb_matches].rule_id = id;
	ops[0]->nb_matches++;
	assert(id > 0);

	return ops[0]->nb_matches >= MAX_REGEX_HS_MATCH;
}
#endif

static int
__dpi_regex_enqueue_mbuf(struct dpi_regex_queue_ctx *ctx, struct rte_mbuf *mbuf,
		  struct dpi_regex_queue_entry *regex_entry,
		  int groups[4], void *user_ptr)
{
	struct rte_regex_ops *ops[1];
	struct rte_regex_ops op = {};
#ifdef NDEBUG
	int ret __attribute__((unused));
#else
	int ret;
#endif

	op.mbuf = mbuf;
	op.group_id0 = groups[0];
	op.group_id1 = groups[1];
	op.group_id2 = groups[2];
	op.group_id3 = groups[3];
	op.user_ptr = regex_entry;
	assert(mbuf->pkt_len);
	assert(mbuf->pkt_len <= (REGEX_MAX_BUF));
	assert(mbuf->nb_segs <= REGEX_MAX_NB_SEGS);
	regex_entry->user_ptr = user_ptr;
	regex_entry->op->nb_matches = 0;
	ops[0] = &op;
	if (ctx->dev_ctx->rxp) {
		ret = rte_regexdev_enqueue_burst(ctx->dev_ctx->dev_id, ctx->hw_queue_id,
						 (struct rte_regex_ops **)ops, 1);
		assert(ret == 1);
		regex_entry->hs = false;
	} else {
		if (ctx->dev_ctx->hs) {
#ifdef DOCA_DPI_USE_HS
			char buf[rte_pktmbuf_pkt_len(mbuf)];
			uint16_t len = rte_pktmbuf_pkt_len(mbuf);

			memcpy(buf, rte_pktmbuf_read(mbuf, 0, len, buf), len);
			ret = hs_scan(ctx->dev_ctx->hs_db, &buf[0], len, 0, ctx->hs_scratch,
				scan_match_handler, regex_entry);
			assert(ret == 0 || ret == HS_SCAN_TERMINATED);
			regex_entry->hs = true;
#endif
		} else {
			return 0;
		}
	}

	return 1;
}

static struct rte_mbuf *
tail_get(struct rte_mbuf *head)
{
	if (head->next)
		return tail_get(head->next);
	return head;
}

int
dpi_regex_enqueue_mbuf(struct dpi_regex_queue_ctx *ctx, struct rte_mbuf *mbuf,
		  int groups[4], void *user_ptr, char *meta)
{
	struct dpi_regex_queue_entry *regex_entry;
	struct rte_mbuf *tail = tail_get(mbuf);
	int meta_len;
#ifdef NDEBUG
	int ret __attribute__((unused));
#else
	int ret;
#endif

	if (rte_ring_dequeue(ctx->free_list, (void **)&regex_entry))
		return 0;

	meta_len = strlen(meta);
	regex_entry->meta_buf->data_len = meta_len;
	regex_entry->meta_buf->pkt_len = meta_len;
	regex_entry->meta_buf->nb_segs = 1;
	regex_entry->meta_buf->next = NULL;
	memcpy(rte_pktmbuf_mtod(regex_entry->meta_buf, void *), meta, meta_len);
	rte_pktmbuf_chain(mbuf, regex_entry->meta_buf);
	__dpi_regex_enqueue_mbuf(ctx, mbuf, regex_entry, groups, user_ptr);
	tail->next = NULL;
	mbuf->pkt_len -= meta_len;
	mbuf->nb_segs--;
	ret = rte_ring_enqueue(ctx->enqueued_list, (void *)regex_entry);
	assert(ret == 0);

	return 1;
}

int
dpi_regex_enqueue_buf(struct dpi_regex_queue_ctx *ctx, const void *buf, uint16_t len,
		  int groups[4], void *user_ptr, char *meta)
{
	struct dpi_regex_queue_entry *regex_entry;
	int meta_len;
#ifdef NDEBUG
	int ret __attribute__((unused));
#else
	int ret;
#endif

	if (rte_ring_dequeue(ctx->free_list, (void **)&regex_entry))
		return 0;
	meta_len = strlen(meta);
	memcpy(rte_pktmbuf_mtod(regex_entry->meta_buf, void *), meta, meta_len);
	regex_entry->meta_buf->pkt_len = meta_len;
	regex_entry->meta_buf->data_len = meta_len;
	regex_entry->meta_buf->nb_segs = 1;
	regex_entry->meta_buf->next = NULL;
	if (len) {
		memcpy(rte_pktmbuf_mtod(regex_entry->regex_buf, void *), buf, len);
		regex_entry->regex_buf->pkt_len = len + meta_len;
		regex_entry->regex_buf->data_len = len;
		regex_entry->regex_buf->nb_segs = 2;
		regex_entry->regex_buf->next = regex_entry->meta_buf;
		__dpi_regex_enqueue_mbuf(ctx, regex_entry->regex_buf, regex_entry, groups, user_ptr);
	} else {
		__dpi_regex_enqueue_mbuf(ctx, regex_entry->meta_buf, regex_entry, groups, user_ptr);
	}
	ret = rte_ring_enqueue(ctx->enqueued_list, (void *)regex_entry);
	assert(ret == 0);
	return 1;
}

int
dpi_regex_dequeue(struct dpi_regex_queue_ctx *ctx, struct dpi_regex_result *result)
{
	struct dpi_regex_queue_entry *regex_entry;
#ifdef NDEBUG
	int ret __attribute__((unused));
#else
	int ret;
#endif

	if (ctx->dev_ctx->rxp) {
		ret = rte_regexdev_dequeue_burst(ctx->dev_ctx->dev_id, ctx->hw_queue_id, &ctx->op, 1);
		if (ret == 0)
			return 0;
	}

	ret = rte_ring_dequeue(ctx->enqueued_list, (void **)&regex_entry);
	if (ret < 0)
		return 0;

	ret = rte_ring_enqueue(ctx->free_list, (void *)regex_entry);
	assert(ret == 0);

	if (regex_entry->hs)
		result->op = regex_entry->op;
	else
		result->op = ctx->op;
	if (result->op->nb_matches)
		assert(result->op->matches[0].rule_id);
	result->user_ptr = regex_entry->user_ptr;
	return 1;
}
