/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_malloc.h>
#include <rte_errno.h>
#include <rte_jhash.h>
#include <rte_hash.h>

#include <doca_log.h>

#include "dns_cache.h"

DOCA_LOG_REGISTER(DPI::DNS_CACHE);

struct dns_cache {
	struct rte_hash *cache;
};

struct cache_data {
	int sig_id;
	time_t ttl;
};

struct dns_cache *
dns_cache_create(int size)
{
	struct rte_hash_parameters dns_hash_parameters = {
		.name = "dns",
		.entries = size,
		.reserved = 0,
		.key_len = sizeof(uint32_t),
		.hash_func = rte_jhash,
		.hash_func_init_val = 0,
	};
	struct dns_cache *dns_cache = (struct dns_cache *)rte_zmalloc(NULL,
								      sizeof(struct dns_cache), 0);

	if (!dns_cache)
		return NULL;
	dns_cache->cache = rte_hash_create(&dns_hash_parameters);
	if (!dns_cache->cache) {
		DOCA_DLOG_ERR("create cache failed - %s", rte_strerror(rte_errno));
		rte_free(dns_cache);
		return NULL;
	}
	return dns_cache;
}

int
dns_cache_sig_store(const struct dns_cache *cache, const struct in_addr *ipv4,
		    int sig_id, uint32_t ttl)
{
	int ret = 0;
	struct cache_data *data = NULL;
	int lookup_res = rte_hash_lookup_data(cache->cache, &(ipv4->s_addr), (void **)(&data));

	if (lookup_res < 0) {
		data = (struct cache_data *)rte_zmalloc(NULL, sizeof(struct cache_data), 0);
		if (!data)
			return -ENOMEM;
		data->sig_id = sig_id;
		data->ttl = time(NULL) + ttl;
		ret = rte_hash_add_key_data(cache->cache, (const void *)(&ipv4->s_addr), (void *)data);
		if (ret < 0)
			DOCA_DLOG_ERR("add to cache failed - %s",
					  rte_strerror(rte_errno));
	} else {
		data->sig_id = sig_id;
		data->ttl = time(NULL) + ttl;
	}
	return ret;
}

int
dns_cache_ip_lookup(const struct dns_cache *cache, const struct in_addr *ipv4)
{
	struct cache_data *data = NULL;
	int ret = rte_hash_lookup_data(cache->cache, &(ipv4->s_addr), (void **)(&data));

	if (ret < 0) {
		DOCA_DLOG_DBG("find in cache failed");
		return ret;
	}
	if (data->ttl < time(NULL)) {
		int32_t index = rte_hash_del_key(cache->cache, &(ipv4->s_addr));

		if (index < 0)
			DOCA_DLOG_ERR("erase from cache failed - %s",
					  rte_strerror(rte_errno));
		rte_free(data);
		return -ENOENT;
	}
	return data->sig_id;
}

void
dns_cache_free(struct dns_cache *cache)
{
	uint32_t *key = NULL;
	struct cache_data *data = NULL;
	uint32_t next = 0;

	while ((rte_hash_iterate(cache->cache, (const void **)&key, (void **)&data, &next)) >= 0)
		rte_free(data);
	rte_hash_free(cache->cache);
	rte_free(cache);
}

int
dns_cache_count(const struct dns_cache *cache)
{
	return rte_hash_count(cache->cache);
}
