/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DNS_CACHE_H_
#define DNS_CACHE_H_

#include <stdint.h>
#include <netinet/in.h>

#ifdef __cplusplus
extern "C" {
#endif

struct dns_cache;

/**
 * Create a new DNS cache.
 *
 * @param size
 *   The wanted size of the new DNS cache.
 * @return
 *   Pointer to DNS cache structure that is used in future DNS cache
 *   operations, or NULL on error
 */
struct dns_cache *dns_cache_create(int size);

/**
 * De-allocate all memory used by DNS cache.
 *
 * @param cache
 *   DNS cache to free
 */
void dns_cache_free(struct dns_cache *cache);

/**
 * Add Key(ipv4) + Data(sig_id+TTL) to DNS cache
 *
 * @param cache
 *   DNS cache to add the information to.
 * @param ipv4
 *   An IP address
 * @param sig_id
 * @param ttl
 *   The time to live of the ip address.
 * @return
 *   - 0 if added successfully
 *   - -EINVAL if the parameters are invalid.
 *   - -ENOSPC if there is no space in the cache for this key.
 *   - -ENOMEM if out of memory
 */
int dns_cache_sig_store(const struct dns_cache *cache, const struct in_addr *ipv4,
			int sig_id, uint32_t ttl);

/**
 * Find Data(sig_id) by Key(ipv4) in DNS cache
 *
 * @param cache
 *   DNS cache to lookup.
 * @param ipv4
 *   An IP address
 * @return
 *   - sig_id if the key exists in the cache.
 *   - -EINVAL if the parameters are invalid.
 *   - -ENOENT if the key is not found.
 */
int dns_cache_ip_lookup(const struct dns_cache *cache, const struct in_addr *ipv4);

/**
 * Return the number of elements in the DNS cache
 *
 * @param cache
 *  DNS cache to query from
 * @return
 *   - -EINVAL if parameters are invalid
 *   - A value indicating how many elements were inserted in the cache.
 */
int dns_cache_count(const struct dns_cache *cache);

#ifdef __cplusplus
}
#endif

#endif /* DNS_CACHE_H_ */
