/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include <rte_malloc.h>
#include <rte_mbuf.h>
#include <rte_errno.h>

#include <doca_log.h>

#include "dns_parser.h"
#include "dns_cache.h"

DOCA_LOG_REGISTER(DPI::DNS_PARSER);

struct dns_parser_ctx {
	enum dpi_parser_protocol protocol;
	enum dpi_parser_state state;
	bool first;
};

int
dns_parser_cache_update(struct rte_mbuf *mbuf, uint32_t payload_offset,
			const struct dns_cache *cache, int sig_id)
{
	const unsigned char *data = rte_pktmbuf_mtod_offset(mbuf, unsigned char *, payload_offset);
	int len = rte_pktmbuf_data_len(mbuf) - payload_offset;
	ns_rr rr;
	int rrnum;
	ns_sect section = ns_s_an;
	ns_msg handle;

	if (!rte_pktmbuf_is_contiguous(mbuf)) {
		DOCA_DLOG_ERR("scattered mbuf is not supported - aborting DNS parser");
		return 0;
	}

	if (ns_initparse(data, len, &handle) < 0) {
		DOCA_DLOG_DBG("DNS parser failed on update cache");
		return -EINVAL;
	}

	if (ns_msg_getflag(handle, ns_f_qr) == 0)
		return -EINVAL;

	for (rrnum = 0; rrnum < (ns_msg_count(handle, section)); rrnum++) {
		if (ns_parserr(&(handle), ns_s_an, rrnum, &rr) < 0) {
			DOCA_DLOG_ERR("response record parser failed, %s", strerror(errno));
			return -EINVAL;
		}
		if (ns_rr_type(rr) == ns_t_a) {
			struct in_addr ipv4;

			memcpy(&ipv4.s_addr, ns_rr_rdata(rr), sizeof(struct in_addr));
			dns_cache_sig_store(cache, &ipv4, sig_id, rr.ttl);
		}
	}
	return 0;
}

/*
 * receive contex, mbuf and its data offset.
 * extract the data from the mbuf and send it to dns parser function
 * return the parser state.
 */
static int
dns_parser_parse(void *pctx, struct rte_mbuf *mbuf,
		 field_cb_t field_cb, void *user_data, uint32_t *consumed)
{
	struct dns_parser_ctx *ctx = (struct dns_parser_ctx *)pctx;
	unsigned char *data = rte_pktmbuf_mtod(mbuf, unsigned char *);
	int len = rte_pktmbuf_data_len(mbuf);
	ns_msg handle;
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_DNS,
		.mbuf = false,
		.stream_start = true,
	};

	if (ns_initparse(data, len, &handle) < 0) {
		ctx->protocol = DPI_PARSER_PROTOCOL_MISS;
		ctx->state = PARSER_STATE_ERROR;
		*consumed = 0;
		DOCA_DLOG_DBG("DNS parser failed");
		return PARSER_STATE_ERROR;
	}

	if (ctx->first) {
		meta.stream_start = false;
		meta.field = EMPTY_FIELD;
		field_cb(user_data, (void *)data, 0, &meta);
		ctx->first = false;
		meta.stream_start = true;
	}

	const unsigned char *eom = data + len;
	const unsigned char *query_start = handle._sections[ns_s_qd];

	uint32_t offset = (query_start - data) + 1; /* add 1 to skip opening tag */
	int name_len_res = dn_skipname(query_start, eom);

	if (name_len_res < 0) {
		ctx->protocol = DPI_PARSER_DNS;
		ctx->state = PARSER_STATE_PROT_VERIFIED;
		*consumed = len;
		DOCA_DLOG_ERR("Parser failed to get Dns domain name length");
		return PARSER_STATE_ERROR;
	}
	uint32_t domain_name_len = (uint32_t)name_len_res - 2; /* subtract 2 to reduce opening tag and closing tag */

	ctx->protocol = DPI_PARSER_DNS;
	ctx->state = PARSER_STATE_DONE;
	*consumed = len;

	if (ns_msg_getflag(handle, ns_f_qr) == 0) {
		meta.field = DNS_QUERY;
		field_cb(user_data, (void *)(data + offset),
			 domain_name_len, &meta);
	} else {
		meta.field = DNS_RESOLVE_NAME;
		field_cb(user_data, (void *)(data + offset),
			 domain_name_len, &meta);
	}
	return PARSER_STATE_DONE;
}

static int
ctx_alloc(void **pctx)
{
	int ret = 0;
	struct dns_parser_ctx **ctx = (struct dns_parser_ctx **)pctx;

	*ctx = (struct dns_parser_ctx *)rte_zmalloc(NULL, sizeof(struct dns_parser_ctx), 0);
	if (*ctx == NULL) {
		ret = -ENOMEM;
		goto err_ctx;
	}
	(*ctx)->state = PARSER_STATE_PART;
	(*ctx)->protocol = DPI_PARSER_PROTOCOL_NOT_READY;
	(*ctx)->first = true;
	return ret;
err_ctx:
	return ret;
}

static void
ctx_dealloc(void *pctx)
{
	struct dns_parser_ctx *ctx = (struct dns_parser_ctx *)pctx;

	if (ctx != NULL)
		rte_free(ctx);
}

static struct dpi_parser dns_parser = {
	.ctx_alloc = ctx_alloc,
	.ctx_dealloc = ctx_dealloc,
	.parse = dns_parser_parse,
};

struct dpi_parser *dns_parser_get(void)
{
	return &dns_parser;
}
