/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/*
 * Each protocol behaves differently, some protocols cannot detect mal format and
 * report parsing not done until they find a specific character.
 * For example, HTTP can have very long URI. FTP can have a long command, etc.
 * We assume a protocol is valid until proven otherwise by the parser, or until
 * MAX_PARSER_BUF cannot be consumed by the parser.
 *
 *                    offset_in_mbuf
 *                        ┌────┐
 *                        │    │
 * ┌─────────┐ ┌────────┐ ┌────┬────────┐  ┌──────┐  ┌─────┐   ┌─────┐
 * │   60    ├─┤   40   ├─┤    │ 80     ├──┤      ├──┤     ├───┤     │
 * └─────────┘ └────────┘ └────┴────────┘  └──────┘  └─────┘   └─────┘
 *                        ▲    ▲
 *                        │    │
 *                        │    │
 *                        │ min_consumed
 *                        │    │
 *                        │    │
 *                       100  120                 210
 *  │                     │    │                   │                 │
 *  ├─────────────────────┼────┼───────────────────┼─────────────────┤
 *  │                     ▲    ▲                   ▲  .  .  .        │
 *                        │    │                   │
 *  first_mbuf_offset─────┘    │                   │
 *                       http-consumed         stream-consumed
 *
 *  <─────────────────────><──────────────────────────────────────────>
 *      unchained mbufs                    mbuf_chain
 *
 */
#include <rte_malloc.h>

#include <doca_log.h>

#include "dpi_parser.h"
#include "base_parser.h"
#include "ssl_parser.h"
#include "http_parser.h"
#include "http2_parser.h"
#include "dns_parser.h"
#include "stream_parser.h"
#include "ftp_parser.h"

#define DNS_PORT 53
#define FTP_PORT 21
#define SSL_PORT 443
#define META_MAX_LEN 200

DOCA_LOG_REGISTER(DPI::PARSER);

#define MAX_PARSER_BUF (1 << 14)
#define MAX_FIELD_LEN 5000
#define MAX_NB_SEGS 32

static uint16_t max_sig_length = 5000;

struct parser_entry {
	LIST_ENTRY(parser_entry) parsers;
	struct dpi_parser *parser;
	void *ctx;
	uint32_t offset;
	int state;
};

struct dpi_parser_main_ctx {
	uint8_t q_buf[MAX_PARSER_BUF];
};

struct dpi_parser_ctx {
	LIST_HEAD(parsers, parser_entry) parser_list;
	uint32_t offset;
	struct dpi_parser_main_ctx *main_ctx;
};

struct dpi_parser_main_ctx *
dpi_parser_init(uint32_t max_sig_len)
{
	struct dpi_parser_main_ctx *ctx;

	ctx = rte_zmalloc(NULL, sizeof(*ctx), 0);
	if (!ctx)
		return NULL;
	max_sig_length = max_sig_len;
	stream_parser_set_max_sig_len(max_sig_len);
	return ctx;
}

void
dpi_parser_destroy(struct dpi_parser_main_ctx *ctx)
{
	if (ctx)
		rte_free(ctx);
}

static int
add_parser(struct dpi_parser_ctx *ctx1, struct dpi_parser_ctx *ctx2, struct dpi_parser *parser)
{
	struct parser_entry *entry1 = rte_zmalloc(NULL, sizeof(struct parser_entry), 0);

	if (!entry1)
		return -ENOMEM;
	struct parser_entry *entry2 = rte_zmalloc(NULL, sizeof(struct parser_entry), 0);

	if (!entry2) {
		rte_free(entry1);
		return -ENOMEM;
	}
	entry1->parser = parser;
	entry2->parser = parser;
	entry1->parser->ctx_alloc(&entry1->ctx);
	entry2->parser->ctx_alloc(&entry2->ctx);
	LIST_INSERT_HEAD(&ctx1->parser_list, entry1, parsers);
	LIST_INSERT_HEAD(&ctx2->parser_list, entry2, parsers);
	if (entry1->parser->peer_set != NULL && entry2->parser->peer_set != NULL) {
		entry1->parser->peer_set(entry1->ctx, entry2->ctx);
		entry2->parser->peer_set(entry2->ctx, entry1->ctx);
	}
	return 0;
}

int dpi_parser_ctx_alloc(struct dpi_parser_main_ctx *main_ctx, struct dpi_parser_ctx **pctx1,
			 struct dpi_parser_ctx **pctx2, uint16_t port, int l4_type)
{
	struct dpi_parser_ctx *ctx1, *ctx2;

	ctx1 = rte_zmalloc(NULL, sizeof(struct dpi_parser_ctx), 0);
	if (!ctx1)
		return -ENOMEM;
	ctx2 = rte_zmalloc(NULL, sizeof(struct dpi_parser_ctx), 0);
	if (!ctx2) {
		rte_free(ctx1);
		return -ENOMEM;
	}
	*pctx1 = ctx1;
	*pctx2 = ctx2;
	LIST_INIT(&ctx1->parser_list);
	LIST_INIT(&ctx2->parser_list);
	add_parser(ctx1, ctx2, http_parser_get());
	add_parser(ctx1, ctx2, http2_parser_get());
	if (port == SSL_PORT)
		add_parser(ctx1, ctx2, ssl_parser_get());
	if (port == DNS_PORT)
		add_parser(ctx1, ctx2, dns_parser_get());
	if (l4_type == DPI_PARSER_TCP)
		add_parser(ctx1, ctx2, tcp_parser_get());
	else if (l4_type == DPI_PARSER_UDP)
		add_parser(ctx1, ctx2, udp_parser_get());
	else
		add_parser(ctx1, ctx2, other_l4_protocol_parser_get());
	ctx1->main_ctx = main_ctx;
	ctx2->main_ctx = main_ctx;
	return 0;
}

void dpi_parser_ctx_dealloc(struct dpi_parser_ctx *ctx)
{
	struct parser_entry *entry;

	while (!LIST_EMPTY(&ctx->parser_list)) {
		entry = LIST_FIRST(&ctx->parser_list);
		LIST_REMOVE(entry, parsers);
		entry->parser->ctx_dealloc(entry->ctx);
		rte_free(entry);
	}
	rte_free(ctx);
}

static void __rte_unused
dpi_debug_status(int ret, struct dpi_parser_status *s)
{
	DOCA_DLOG_DBG("offset=%d, Start=%d,End=%d, field=%d, protocol=%d ret=%d",
	       s->payload_offset, s->start, s->end, s->field, s->protocol, ret);
}

#define DPI_DEBUG_STAT(RET, STAT) dpi_debug_status(RET, STAT)

static int
parse_data(struct parser_entry *entry, uint8_t *data, uint32_t len,
		     field_cb_t field_cb,
		     void *user_data,
		     uint32_t *consumed)
{
	struct rte_mbuf dummy_mbuf = {};

	dummy_mbuf.buf_addr = data;
	dummy_mbuf.nb_segs = 1;
	dummy_mbuf.pkt_len = len;
	dummy_mbuf.data_len = len;
	return entry->parser->parse(entry->ctx, &dummy_mbuf, field_cb, user_data, consumed);
}

static struct rte_mbuf *
update_fake_head(uint32_t offset, struct rte_mbuf *fake_head)
{
	uint32_t pkt_len = fake_head->pkt_len;
	uint16_t nb_segs = fake_head->nb_segs;

	while (fake_head->next && (offset > fake_head->data_len) && (pkt_len > fake_head->data_len) && (nb_segs > 0)) {
		offset -= fake_head->data_len;
		pkt_len -= fake_head->data_len;
		nb_segs--;
		fake_head = fake_head->next;
	}
	assert(pkt_len <= (MAX_PARSER_BUF - META_MAX_LEN));
	fake_head->pkt_len = pkt_len;
	fake_head->nb_segs = nb_segs;
	return fake_head;
}

static void
update_stream_consumed(uint32_t pkt_len, struct rte_mbuf *fake_head, uint32_t *consumed)
{
	uint32_t overlap_with_next =
		pkt_len > max_sig_length ? max_sig_length : pkt_len;
	uint32_t stream_consumed = pkt_len - overlap_with_next;

	while (fake_head && (stream_consumed > 0) && (stream_consumed >= fake_head->data_len)) {
		stream_consumed -= fake_head->data_len;
		*consumed += fake_head->data_len;
		fake_head = fake_head->next;
	}
}

/**
 * Wraps rte_pktmbuf_read but uses a mutable mbuf so that the returned pointer can be non const.
 *
 * @NOTE
 * This function may be promoted so some utility header if its needed in more than one place.
 *
 * @WARNING
 * This function exists to codify our assumption that both the input mbuf and buffer are both mutable.
 * Without this the user could have changed the input mbuf to be a const mbuf and we have then broken our justification
 * for accepting this occurrence of casting away const as valid. so we cannot place the pragma at the call site as
 * it cannot be enforced over time, this way can and will be verified by the compiler.
 */
static inline void *
mutable_rte_pktmbuf_read(struct rte_mbuf *m, uint32_t off, uint32_t len, void *buf)
{
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual"
	/* Justification for casting away const:
	 * rte_pktmbuf_read operates upon a const mbuf so it must return a const pointer to maintain that contract.
	 * However we know for sure the input mbuf is mutable and the user defined buffer is also mutable so the result
	 * from rte_pktmbuf_read can only be a pointer to a region which we know is mutable, so we can safely remove
	 * the const qualifier from the output which rte_pktmbuf_read had to assume was necessary.
	 */
	return (void *)rte_pktmbuf_read(m, off, len, buf);
#pragma GCC diagnostic pop
}

/**
 * Parse new flow, or continue parsing prev flow.
 * Parser should be able to store the parsing state on the ctx.
 * returns the amount of processed bytes.
 */
uint32_t
dpi_parser_parse(int dpi_q, struct dpi_parser_ctx *ctx, struct rte_mbuf *mbuf_chain,
		 uint32_t first_mbuf_offset, field_cb_t field_cb, void *user_data)
{
	struct parser_entry *entry;
	uint8_t *data;
	uint32_t offset_in_mbuf; /* local offset inside chain */
	uint32_t len;
	uint32_t min_offset = ctx->offset + rte_pktmbuf_pkt_len(mbuf_chain);
	uint32_t min_consumed = 0;

	/* first copy mbuf into the parser buf */
	offset_in_mbuf = ctx->offset - first_mbuf_offset;
	len = rte_pktmbuf_pkt_len(mbuf_chain) - offset_in_mbuf;
	if (len > (MAX_PARSER_BUF - META_MAX_LEN))
		len = MAX_PARSER_BUF - META_MAX_LEN;
	data = (uint8_t *)mutable_rte_pktmbuf_read(mbuf_chain, offset_in_mbuf, len, ctx->main_ctx->q_buf);

	LIST_FOREACH(entry, &ctx->parser_list, parsers) {
		uint32_t consumed = 0;

		if (entry->state == PARSER_STATE_ERROR)
			continue;
		/* Each parser is in a different location in the stream */
		if (entry->parser->support_mbuf && entry->parser->support_mbuf(entry->ctx)) {
		/**
		 * For stream parser we send the mbuf_chain without copying it
		 * unless the mbuf_chain contains more than 32 nb_segs
		 * because the regex can't handle more than 32 nb_segs
		 */
			uint32_t offset = entry->offset - first_mbuf_offset;
			struct rte_mbuf *fake_head = mbuf_chain;

			/* promote the mbuf_chain according to the offset of the parser */
			fake_head = update_fake_head(offset, fake_head);
			if (fake_head->nb_segs >= MAX_NB_SEGS) {
				entry->state = parse_data(entry,
					data + (entry->offset - ctx->offset),
					len - (entry->offset - ctx->offset),
					field_cb, user_data, &consumed);
				update_stream_consumed(len - (entry->offset - ctx->offset),
							fake_head, &consumed);
			} else {
				entry->state = entry->parser->parse(entry->ctx, fake_head,
							    field_cb, user_data,
							    &consumed);
			}
		} else {
			/**
			 * For all the rest of the parsers we copy the mbuf_chain to a single mbuf
			 * and parse the data from the place that we stoped the previous time
			 */
			entry->state = parse_data(entry,
				data + (entry->offset - ctx->offset),
				len - (entry->offset - ctx->offset),
				field_cb, user_data, &consumed);
			if (len - (entry->offset - ctx->offset) >= MAX_FIELD_LEN)
				entry->state = PARSER_STATE_ERROR;
		}
		if (entry->state == PARSER_STATE_ERROR)
			continue;
		entry->offset += consumed;
		if (entry->offset < min_offset)
			min_offset = entry->offset;
	}
	/* Report only the minimun consumed parser */
	min_consumed = min_offset - ctx->offset;
	ctx->offset = min_offset;
	return min_consumed;
}

int
dpi_parser_update_dns_cache(struct rte_mbuf *mbuf, uint32_t payload_offset,
			    struct dns_cache *dns_cache, int sig_id)
{
	return dns_parser_cache_update(mbuf, payload_offset, dns_cache, sig_id);
}

int
dpi_parser_update_ftp_cache(struct rte_mbuf *mbuf, uint32_t payload_offset,
			    struct ftp_cache *ftp_cache, int sig_id)
{
	return ftp_parser_cache_update(mbuf, payload_offset, ftp_cache, sig_id);
}
