/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_malloc.h>
#include <rte_errno.h>
#include <rte_jhash.h>
#include <rte_hash.h>

#include <doca_log.h>

#include "ftp_cache.h"

DOCA_LOG_REGISTER(DPI::FTP_CACHE);

struct ftp_cache {
	struct rte_hash *cache;
};

struct cache_data {
	int sig_id;
};

struct cache_key {
	uint32_t ip_addr;
	uint16_t port;
};

struct ftp_cache *
ftp_cache_create(int size)
{
	struct rte_hash_parameters ftp_hash_parameters = {
		.name = "ftp",
		.entries = size,
		.reserved = 0,
		.key_len = sizeof(struct cache_key),
		.hash_func = rte_jhash,
		.hash_func_init_val = 0,
	};
	struct ftp_cache *ftp_cache = (struct ftp_cache *)rte_zmalloc(NULL,
								      sizeof(struct ftp_cache), 0);

	if (!ftp_cache)
		return NULL;
	ftp_cache->cache = rte_hash_create(&ftp_hash_parameters);
	if (!ftp_cache->cache) {
		DOCA_DLOG_ERR("create cache failed - %s", rte_strerror(rte_errno));
		rte_free(ftp_cache);
		return NULL;
	}
	return ftp_cache;
}

int
ftp_cache_sig_store(const struct ftp_cache *cache, const struct in_addr *ipv4, uint16_t port,
		    int sig_id)
{
	int ret = 0;
	struct cache_data *data = NULL;
	struct cache_key key = {
		.ip_addr = ipv4->s_addr,
		.port = port,
	};

	int lookup_res = rte_hash_lookup_data(cache->cache, &key, (void **)(&data));

	if (lookup_res < 0) {
		data = (struct cache_data *)rte_zmalloc(NULL, sizeof(struct cache_data), 0);
		if (!data)
			return -ENOMEM;
		data->sig_id = sig_id;
		ret = rte_hash_add_key_data(cache->cache, (void *)(&key), (void *)data);
		if (ret < 0)
			DOCA_DLOG_ERR("add to cache failed - %s", rte_strerror(rte_errno));
	} else {
		data->sig_id = sig_id;
	}
	return ret;
}

int
ftp_cache_ip_lookup(const struct ftp_cache *cache, const struct in_addr *ipv4, uint16_t port)
{
	struct cache_data *data = NULL;
	struct cache_key key = {
		.ip_addr = ipv4->s_addr,
		.port = port,
	};


	int ret = rte_hash_lookup_data(cache->cache, &key, (void **)(&data));

	if (ret < 0) {
		DOCA_DLOG_DBG("find in cache failed");
		return ret;
	}
	return data->sig_id;
}

void
ftp_cache_free(struct ftp_cache *cache)
{
	uint32_t *key = NULL;
	struct cache_data *data = NULL;
	uint32_t next = 0;

	while ((rte_hash_iterate(cache->cache, (const void **)&key, (void **)&data, &next)) >= 0)
		rte_free(data);
	rte_hash_free(cache->cache);
	rte_free(cache);
}

int
ftp_cache_count(const struct ftp_cache *cache)
{
	return rte_hash_count(cache->cache);
}
