/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef FTP_CACHE_H_
#define FTP_CACHE_H_

#include <stdint.h>
#include <netinet/in.h>

#ifdef __cplusplus
extern "C" {
#endif

struct ftp_cache;

/**
 * Create a new FTP cache.
 *
 * @param size
 *   The wanted size of the new FTP cache.
 * @return
 *   Pointer to FTP cache structure that is used in future FTP cache
 *   operations, or NULL on error
 */
struct ftp_cache *ftp_cache_create(int size);

/**
 * De-allocate all memory used by FTP cache.
 *
 * @param cache
 *   FTP cache to free
 */
void ftp_cache_free(struct ftp_cache *cache);

/**
 * Add Key(ipv4) + Data(sig_id+TTL) to FTP cache
 *
 * @param cache
 *   FTP cache to add the information to.
 * @param ipv4
 *   An IP address
 * @param sig_id
 * @param ttl
 *   The time to live of the ip address.
 * @return
 *   - 0 if added successfully
 *   - -EINVAL if the parameters are invalid.
 *   - -ENOSPC if there is no space in the cache for this key.
 *   - -ENOMEM if out of memory
 */
int ftp_cache_sig_store(const struct ftp_cache *cache, const struct in_addr *ipv4, uint16_t port,
			int sig_id);

/**
 * Find Data(sig_id) by Key(ipv4) in FTP cache
 *
 * @param cache
 *   FTP cache to lookup.
 * @param ipv4
 *   An IP address
 * @return
 *   - sig_id if the key exists in the cache.
 *   - -EINVAL if the parameters are invalid.
 *   - -ENOENT if the key is not found.
 */
int ftp_cache_ip_lookup(const struct ftp_cache *cache, const struct in_addr *ipv4, uint16_t port);

/**
 * Return the number of elements in the FTP cache
 *
 * @param cache
 *  FTP cache to query from
 * @return
 *   - -EINVAL if parameters are invalid
 *   - A value indicating how many elements were inserted in the cache.
 */
int ftp_cache_count(const struct ftp_cache *cache);

#ifdef __cplusplus
}
#endif

#endif /* FTP_CACHE_H_ */
