/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>
#include <arpa/inet.h>
#include <errno.h>

#include <rte_malloc.h>
#include <rte_mbuf.h>
#include <rte_errno.h>

#include <doca_log.h>

#include "ftp_parser.h"
#include "ftp_cache.h"

DOCA_LOG_REGISTER(DPI::FTP_PARSER);

#define IP_FIELD_SIZE 4

/*
 * The FTP parser recives FTP messages stream and split it to single messages according to the
 * '\r\n' sequense.
 * The consumed value is set to the stream length till the last '\r\n' sequense that was found,
 * or zero if not found.
 */
static int
ftp_parser_parse(struct rte_mbuf *mbuf, uint32_t payload_offset,
		 const struct ftp_cache *cache, int sig_id)
{
	char *data = rte_pktmbuf_mtod_offset(mbuf, char *, payload_offset);
	int len = rte_pktmbuf_data_len(mbuf) - payload_offset;
	char *msg = data;
	int i, flag = 0, ftp_packet_len = 0;
	char ip_msg[16] = "";
	int ip_field_size = IP_FIELD_SIZE;
	int consumed = 0;
	char *passive_mode_msg = "227 Entering Passive Mode";
	struct in_addr ipv4;
	uint16_t port;

	if (!rte_pktmbuf_is_contiguous(mbuf)) {
		DOCA_DLOG_ERR("scattered mbuf is not supported - aborting FTP parser");
		return 0;
	}

	while (consumed < len) {
		flag = 0;
		ftp_packet_len = 0;
		for (i = consumed; i < len - 1; i++) {
			if (data[i] == '\r' && data[i+1] == '\n') {
				ftp_packet_len += 2;
				flag = 1;
				break;
			}
			ftp_packet_len++;
		}
		if (!flag) {
			DOCA_DLOG_DBG("FTP message did not reach the end");
			break;
		}
		msg = data + consumed;
		if (ftp_packet_len < strlen(passive_mode_msg) ||
		strncmp(msg, passive_mode_msg, strlen(passive_mode_msg)) != 0) {
			DOCA_DLOG_DBG("Not FTP entering passive mode message");
			consumed += ftp_packet_len;
			continue;
		}

		msg = msg + strlen("227 Entering Passive Mode (");
		for (i = 0; ip_field_size > 0; i++, msg++) {
			if ((*msg) == ',') {
				ip_field_size--;
				if (ip_field_size)
					ip_msg[i] = '.';
			} else
				ip_msg[i] = *msg;
		}
		inet_pton(AF_INET, ip_msg, &(ipv4.s_addr));
		int port_num1 = strtol(strtok(msg, ","), (char **)NULL, 10);
		int port_num2 = strtol(strtok(NULL, ")"), (char **)NULL, 10);
		int port_num = port_num1*256 + port_num2;

		port = (uint16_t)port_num;
		consumed += ftp_packet_len;
		ftp_cache_sig_store(cache, &(ipv4), port, sig_id);
	}
	return 0;
}

int
ftp_parser_cache_update(struct rte_mbuf *mbuf, uint32_t payload_offset,
			const struct ftp_cache *cache, int sig_id)
{
	return ftp_parser_parse(mbuf, payload_offset, cache, sig_id);
}
