/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>

#include <rte_mbuf.h>
#include <rte_malloc.h>
#include <rte_errno.h>
#include <nghttp2/nghttp2.h>

#include <doca_log.h>

#include "http2_parser.h"

DOCA_LOG_REGISTER(DPI::HTTP2_PARSER);

#define FIELD_TO_ENUM_MAX_ENTRIES 9

/**
 * HTTP2 frame types
 */
enum http2_handshake_type {
	/** Data frame type */
	HTTP2_DATA	    = 0,
	/** Headers frame type */
	HTTP2_HEADERS	    = 1,
	/** Priority frame type */
	HTTP2_PRIORITY	    = 2,
	/** Rst_stream frame type */
	HTTP2_RST_STREAM    = 3,
	/** Settings frame type */
	HTTP2_SETTINGS	    = 4,
	/** Push_Promise frame type (TLS 1.3) */
	HTTP2_PUSH_PROMISE  = 5,
	/** Ping frame type (TLS 1.3) */
	HTTP2_PING	    = 6,
	/** Goaway frame type (TLS 1.3) */
	HTTP2_GOAWAY	    = 7,
	/** Window update frame type (TLS 1.3) */
	HTTP2_WINDOW_UPDATE = 8,
	/** Continuation frame type */
	HTTP2_CONTINUATION  = 9,
	/** Altsvc frame type */
	HTTP2_ALTSVC	    = 10,
	/** Origin frame type */
	HTTP2_ORIGIN	    = 12,
	HTTP2_TYPE_MAX
};


struct field2enum {
	char *name;
	int num;
};

static struct field2enum FIELD_TO_ENUM[] = {
	{"host", HTTP_SERVER},
	{"server", HTTP_SERVER_AGENT},
	{"user-agent", HTTP_USER_AGENT},
	{"referer", HTTP_REFERER},
	{"location", HTTP_LOCATION},
	{"cookie", HTTP_COOKIE},
	{"content", HTTP_CONTENT},
	{"content-type", HTTP_MIME_TYPE},
	{"content-length", HTTP_CONTENT_LENGTH},
	{":path", HTTP_URI}
};

struct http2_parser_ctx {
	enum dpi_parser_protocol protocol;
	enum dpi_parser_state state;
	uint32_t content_len;
	nghttp2_hd_inflater *inflater_ptr;
	struct http2_parser_ctx *peer_ctx;
	bool first;
};

struct http2_payload_len {
	/** Length is 3-Byte long, This is the MSB */
	uint8_t length1;
	/** Length is 3-Byte long, This is the 2 LSB */
	uint16_t length2;
} __rte_packed;

struct http2_frame_header {
	/** Length of the message payload. */
	struct http2_payload_len len;
	/** Type of the frame (one of ::http2_handshake_type) */
	uint8_t type;
	/** Flags for frame type */
	uint8_t flags;
	/** Stream identifier */
	uint32_t stream_id;
} __rte_packed;


static uint32_t get_http2_len(struct http2_payload_len *http2_len_hdr)
{
	uint32_t len = rte_be_to_cpu_16(http2_len_hdr->length2) |
		       (http2_len_hdr->length1 << 16);
	return len;
}

static uint32_t get_frame_payload_len(struct http2_frame_header *msg_hdr)
{
	uint32_t len = get_http2_len(&msg_hdr->len);

	return sizeof(struct http2_frame_header) + len;
}

static int
header_field2num(const char *name, int name_len)
{
	int i;

	for (i = 0 ; i < ARRAY_SIZE(FIELD_TO_ENUM); i++) {
		if (strlen(FIELD_TO_ENUM[i].name) == name_len)
			if (!strncasecmp(name, FIELD_TO_ENUM[i].name, name_len))
				return FIELD_TO_ENUM[i].num;
	}

	DOCA_DLOG_DBG("field %s is not supported", name);
	return -1;
}

static void
parse_header(struct http2_parser_ctx *ctx, void *user_data, nghttp2_nv *nv_out, field_cb_t field_cb)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_HTTP_1,
		.mbuf = false,
		.stream_start = true,
	};
	int field_enum = -1;

	field_enum = header_field2num((const char *)nv_out->name, nv_out->namelen);
	if (field_enum < 0)
		return;
	if (field_enum == HTTP_CONTENT_LENGTH)
		ctx->content_len = atoi((const char *)nv_out->value);
	meta.field = field_enum;
	field_cb(user_data, nv_out->value, nv_out->valuelen, &meta);
}

static int
parse_frame(struct http2_parser_ctx *ctx, void *user_data, uint8_t *data, uint32_t len,
	    field_cb_t field_cb)
{
	struct http2_frame_header *frame_header = (struct http2_frame_header *)data;
	nghttp2_nv nv_out;
	ssize_t ret;
	size_t proc_len;
	int inflate_flags = 0;
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_HTTP_1,
		.mbuf = false,
		.stream_start = true,
	};

	switch (frame_header->type) {
	/* Parse SETTINGS packet, must arrive before any HEADERS packet */
	case HTTP2_SETTINGS:
		if (ctx->first) {
			meta.stream_start = false;
			meta.field = EMPTY_FIELD;
			field_cb(user_data, data, 0, &meta);
			ctx->first = false;
			meta.stream_start = true;
		}
		data += sizeof(struct http2_frame_header);
		len -= sizeof(struct http2_frame_header);
		ctx->state = PARSER_STATE_PROT_VERIFIED;
		ctx->peer_ctx->state = PARSER_STATE_PROT_VERIFIED;
		break;
	/* Parse HEADERS packet, all headers in the packet are deflated and arrive as binary data */
	case HTTP2_HEADERS:
		data += sizeof(struct http2_frame_header);
		len -= sizeof(struct http2_frame_header);
		if (ctx->state != PARSER_STATE_PROT_VERIFIED) {
			DOCA_DLOG_DBG("http2 preface error. HEADERS packet must come after SETTINGS packet\n");
			ctx->protocol = DPI_PARSER_PROTOCOL_MISS;
			return -1;
		}
		for (;;) {
			/* Inflate headers */
			ret = nghttp2_hd_inflate_hd2(ctx->inflater_ptr, &nv_out, &inflate_flags,
				data, len, 1);
			if (ret < 0) {
				DOCA_DLOG_DBG("inflate failed with error code %zd", ret);
				return -1;
			}
			proc_len = ret;
			data += proc_len;
			len -= proc_len;

			/* Header deflate success */
			if (inflate_flags & NGHTTP2_HD_INFLATE_EMIT)
				parse_header(ctx, user_data, &nv_out, field_cb);
			/* Data deflate finish */
			if (inflate_flags & NGHTTP2_HD_INFLATE_FINAL) {
				nghttp2_hd_inflate_end_headers(ctx->inflater_ptr);
				break;
			}
			if ((inflate_flags & NGHTTP2_HD_INFLATE_EMIT) == 0 && len == 0)
				break;
		}
		break;
	/* Parse DATA packet, contain the payload (content) of the transaction */
	case HTTP2_DATA:
		data += sizeof(struct http2_frame_header);
		len -= sizeof(struct http2_frame_header);
		if (ctx->state != PARSER_STATE_PROT_VERIFIED) {
			DOCA_DLOG_DBG("http2 preface error. DATA packet must come after protocol is verified\n");
			ctx->protocol = DPI_PARSER_PROTOCOL_MISS;
			return -1;
		}
		if (len < ctx->content_len) {
			DOCA_DLOG_DBG("http2 DATA packet length is shorter then content-length field");
			return -1;
		}
		if (len == 0)
			break;
		len -= ctx->content_len;
		meta.field = HTTP_CONTENT;
		field_cb(user_data, data,
			 ctx->content_len, &meta);
		break;
	/* All the known types we accept as frame type */
	case HTTP2_PRIORITY:
	case HTTP2_RST_STREAM:
	case HTTP2_PUSH_PROMISE:
	case HTTP2_PING:
	case HTTP2_GOAWAY:
	case HTTP2_WINDOW_UPDATE:
	case HTTP2_CONTINUATION:
	case HTTP2_ALTSVC:
	case HTTP2_ORIGIN:
		break;
	default:
		DOCA_DLOG_DBG("http2 packet type wasn't recognize");
		return -1;
	}
	return 0;
}

static bool
is_http2(struct http2_frame_header *frame_header, uint32_t data_len)
{
	if (data_len < sizeof(struct http2_frame_header))
		return false;
	if (frame_header->type > HTTP2_TYPE_MAX || frame_header->type == 11)
		return false;
	return true;
}

static int
http2_parser_parse(void *pctx, struct rte_mbuf *mbuf,
		   field_cb_t field_cb, void *user_data, uint32_t *consumed)
{
	struct http2_parser_ctx *ctx = (struct http2_parser_ctx *)pctx;
	uint8_t *data = rte_pktmbuf_mtod(mbuf, uint8_t *);
	uint32_t left_to_scan = rte_pktmbuf_pkt_len(mbuf);
	uint32_t msg_len;
	struct http2_frame_header *frame_header;
	uint8_t *frame, *ret, *tmp_data;
	const char *preface_packet = "PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n";

	*consumed = 0;
	/*
	 * http2 can be recognized by the 24 byte pattern "PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n"
	 * this preface packet must arrive from the client side in order to recognize the
	 * transaction as an http2
	 */
	if (ctx->protocol != DPI_PARSER_HTTP_2) {
		ret = (uint8_t *)strstr((const char *)data, preface_packet);
		if (ret == NULL) {
			DOCA_DLOG_DBG("http2 was not recognized");
			ctx->state = left_to_scan < 24 ? PARSER_STATE_PART : PARSER_STATE_ERROR;
			*consumed += left_to_scan;
			return ctx->state;
		}
		ctx->protocol = DPI_PARSER_HTTP_2;
		ctx->peer_ctx->protocol = DPI_PARSER_HTTP_2;
		tmp_data = ret + 24;
		left_to_scan -= tmp_data - data;
		*consumed += tmp_data - data;
		data = tmp_data;
	}

	if (left_to_scan < sizeof(struct http2_frame_header)) {
		ctx->state = PARSER_STATE_PART;
		return PARSER_STATE_PART;
	}

	while (left_to_scan) {
		frame = data;
		frame_header = (struct http2_frame_header *)frame;
		if (!is_http2(frame_header, left_to_scan)) {
			DOCA_DLOG_DBG("http2 packet received has error in frame header");
			ctx->state = PARSER_STATE_ERROR;
			return -1;
		}
		msg_len = get_frame_payload_len(frame_header);
		if (msg_len > left_to_scan)
			break;
		if (parse_frame(ctx, user_data, frame, msg_len, field_cb)) {
			DOCA_DLOG_DBG("parse_frame() failed for http2 parser");
			ctx->state = PARSER_STATE_ERROR;
			return -1;
		}
		data += msg_len;
		left_to_scan -= msg_len;
		*consumed += msg_len;
	}

	return ctx->state;
}

static int
http2_parser_ctx_alloc(void **pctx)
{
	int ret = 0;
	struct http2_parser_ctx **ctx = (struct http2_parser_ctx **)pctx;

	*ctx = (struct http2_parser_ctx *)rte_zmalloc(NULL, sizeof(struct http2_parser_ctx), 0);
	if (*ctx == NULL) {
		ret = -ENOMEM;
		DOCA_DLOG_ERR("http2 parser ctx alloc failed");
		goto err_ctx;
	}
	(*ctx)->state = PARSER_STATE_PART;
	(*ctx)->protocol = DPI_PARSER_PROTOCOL_NOT_READY;
	(*ctx)->first = true;
	if (nghttp2_hd_inflate_new(&((*ctx)->inflater_ptr))) {
		ret = NGHTTP2_ERR_NOMEM;
		DOCA_DLOG_ERR("http2 parser inflator alloc failed");
		goto err_parser;
	}
	return ret;
err_parser:
	rte_free(*ctx);
err_ctx:
	return ret;
}

static void
http2_parser_ctx_dealloc(void *pctx)
{
	if (pctx == NULL)
		return;
	struct http2_parser_ctx *ctx = (struct http2_parser_ctx *)pctx;

	nghttp2_hd_inflate_del(ctx->inflater_ptr);
	if (ctx != NULL)
		rte_free(ctx);
}

static void
http2_peer_set(void *pctx, void *peer_pctx)
{
	struct http2_parser_ctx *ctx1 = (struct http2_parser_ctx *)pctx;
	struct http2_parser_ctx *ctx2 = (struct http2_parser_ctx *)peer_pctx;

	ctx1->peer_ctx = ctx2;
}

static struct dpi_parser http2_parser = {
	.ctx_alloc = http2_parser_ctx_alloc,
	.ctx_dealloc = http2_parser_ctx_dealloc,
	.parse = http2_parser_parse,
	.peer_set = http2_peer_set,
};

struct dpi_parser *http2_parser_get(void)
{
	return &http2_parser;
}
