/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <rte_mbuf.h>
#include <rte_malloc.h>
#include <rte_errno.h>
#include <uriparser/Uri.h>
#include <uriparser/UriIp4.h>

#include <doca_log.h>

#include <picohttpparser.h>
#include "http_parser.h"


#define MAX_HEADERS_NUM 32
#define MIN_FIELD_LEN 4

DOCA_LOG_REGISTER(DPI::HTTP_PARSER);

#define CHECK_VERSION(ver, ret)								\
	do {										\
		if (ver < 0) {								\
			DOCA_DLOG_DBG("http version has not been recognize");	\
			return ret == -2 ? -EOP : -UNKNOWN_ERROR;			\
		}									\
	} while (0)

enum packet_parse_return {
	PARSE_ERROR,
	EOP, /* End Of Packet */
	UNKNOWN_ERROR
};

struct field2enum {
	const char *name;
	int num;
};

static struct field2enum FIELD_TO_ENUM[] = {
	{"host", HTTP_SERVER},
	{"server-agent", HTTP_SERVER_AGENT},
	{"content-type", HTTP_MIME_TYPE},
	{"user-agent", HTTP_USER_AGENT},
	{"referer", HTTP_REFERER},
	{"server", HTTP_SERVER_AGENT},
	{"location", HTTP_LOCATION},
	{"cookie", HTTP_COOKIE},
	{"content", HTTP_CONTENT},
	{"content-length", HTTP_CONTENT_LENGTH}
};

struct http_parser_ctx {
	enum dpi_parser_protocol protocol;
	enum dpi_parser_state state;
	uint32_t content_len;
	bool content_left;
	bool first;
};

static int
header_field2num(const char *name, int name_len)
{
	int i;

	for (i = 0 ; i < ARRAY_SIZE(FIELD_TO_ENUM) ; i++) {
		if (strlen(FIELD_TO_ENUM[i].name) == name_len)
			if (!strncasecmp(name, FIELD_TO_ENUM[i].name, name_len))
				return FIELD_TO_ENUM[i].num;
	}
	return -1;
}

/*
 * Go through all headers found and report it with callback
 */
static void
report_headers_list(struct http_parser_ctx *ctx, struct phr_header *headers, void *user_data,
		  size_t num_headers, uint32_t *payload_offset,
		  field_cb_t field_cb, char *init_data)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_HTTP_1,
		.mbuf = false,
		.stream_start = true,
	};
	int field_enum = -1, i;

	for (i = 0 ; i < num_headers ; i++) {
		field_enum = header_field2num(headers[i].name,
					      headers[i].name_len);
		if (field_enum < 0)
			continue;
		if (field_enum == HTTP_CONTENT_LENGTH)
			ctx->content_len = atoi(init_data + headers[i].value_offset +
						*payload_offset);

		meta.field = field_enum;
		field_cb(user_data,
			 init_data + headers[i].value_offset + *payload_offset,
			 headers[i].value_len, &meta);
	}
	*payload_offset += headers[i-1].value_offset + headers[i-1].value_len + 2;
}

/*
 * Check if the data length is enough to consume all of the content value.
 * if it is, call field_cb and return 1, else return 0.
 * if data length is shorter then content length update ctx->content_left
 */
static int
consume_rest_of_content(struct rte_mbuf *mbuf, void *user_data, field_cb_t field_cb,
			struct http_parser_ctx *ctx, int *payload_offset)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_HTTP_1,
		.mbuf = false,
		.stream_start = true,
	};
	int data_len = rte_pktmbuf_pkt_len(mbuf) - *payload_offset;
	char *init_data = rte_pktmbuf_mtod_offset(mbuf, char *, *payload_offset);

	/*
	 * content.len is 0 if the value of content-length wasn't valid for atoi() or didn't exist.
	 */
	if (ctx->content_len == 0)
		return data_len == 0 ? 0 : 1;
	if (ctx->content_len <= data_len) {
		meta.field = HTTP_CONTENT;
		field_cb(user_data, init_data, ctx->content_len, &meta);
		*payload_offset += ctx->content_len;
		uint32_t content_len_val = ctx->content_len;

		ctx->content_len = 0;
		return content_len_val == data_len ? 0 : 1;
	}
	ctx->content_left = true;
	return 0;
}

/*
 * Normalize uri and send it to callback function.
 */
static int
normalize_uri(void *user_data, field_cb_t field_cb, int prot,
			 const char *start, size_t len)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_HTTP_1,
		.mbuf = false,
		.stream_start = true,
	};
	UriParserStateA state;
	UriUriA uri;
	char uri_text[len + 1];
	char *normalized_uri;
	int chars_required;
	int ret = 0;

	memcpy(uri_text, start, len);
	uri_text[len] = '\0';
	state.uri = &uri;
	if (uriParseUriA(&state, uri_text) != URI_SUCCESS) {
		DOCA_DLOG_ERR("failed to parse uri");
		ret = -1;
		goto err;
	}
	/* Normalizes all components of a URI. */
	if (uriNormalizeSyntaxA(&uri) != URI_SUCCESS) {
		DOCA_DLOG_ERR("failed to normalize uri");
		ret = -1;
		goto err;
	}
	/*
	 * Calculates the number of characters needed to store
	 * the string representation of the given URI excluding the terminator.
	 */
	if (uriToStringCharsRequiredA(&uri, &chars_required) != URI_SUCCESS) {
		DOCA_DLOG_ERR("failed at uri chars required");
		ret = -1;
		goto err;
	}
	/* For the terminator */
	chars_required++;
	normalized_uri = rte_malloc(NULL, chars_required * sizeof(char), 0);
	if (normalized_uri == NULL) {
		DOCA_DLOG_ERR("failed to allocate uriString");
		ret = -1;
		goto err;
	}
	/* Converting a URI structure back to text. */
	if (uriToStringA(normalized_uri, &uri, chars_required, NULL) != URI_SUCCESS) {
		DOCA_DLOG_ERR("failed at uri to string");
		ret = -1;
		goto free_uri;
	}
	meta.field = HTTP_URI_NORM;
	field_cb(user_data, normalized_uri, chars_required, &meta);
free_uri:
	rte_free(normalized_uri);
err:
	uriFreeUriMembersA(&uri);
	return ret;
}

/*
 * Receive mbuf, its data offset and flow context.
 * go through all the data in the mbuf and parse uri path, http version and headers.
 * return:
 *	-2: unexpected error
 *	-1: end of packet
 *	 0: parsing error
 *	>0: http packet ended successfully
 */
static int
http_parser_packet(struct rte_mbuf *mbuf, uint32_t *payload_offset, void *user_data,
		   struct http_parser_ctx *ctx, field_cb_t field_cb)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_HTTP_1,
		.mbuf = false,
		.stream_start = true,
	};
	uint32_t offset = *payload_offset;
	int ret = 0;
	const char *method;
	size_t method_len;
	const char *msg;
	size_t msg_len;
	int status;
	int path = 0;
	size_t path_len = 0;
	int minor_version;
	struct phr_header headers[MAX_HEADERS_NUM];
	size_t num_headers;
	char *init_data = rte_pktmbuf_mtod(mbuf, char *);
	int init_len = rte_pktmbuf_data_len(mbuf) - offset;
	int len = rte_pktmbuf_data_len(mbuf) - offset;
	char *data = rte_pktmbuf_mtod_offset(mbuf, char *, offset);

	while (true) {
		len = init_len - ret;
		data += ret;
		num_headers = MAX_HEADERS_NUM;

		/* check for request packet */
		ret = phr_parse_request(data, len, &method, &method_len, &path,
					&path_len, &minor_version, headers,
					&num_headers, 0);
		if (ret > 0 || ret == -2) {
			if (path == 0 || path_len == 0) {
				DOCA_DLOG_DBG("uri path has not been recognize");
				return ret == -2 ? -EOP : -UNKNOWN_ERROR;
			}
			CHECK_VERSION(minor_version, ret);
			ctx->state = PARSER_STATE_PROT_VERIFIED;
			ctx->protocol = DPI_PARSER_HTTP_1;
			if (ctx->first) {
				meta.stream_start = false;
				meta.field = EMPTY_FIELD;
				field_cb(user_data, init_data, 0, &meta);
				ctx->first = false;
				meta.stream_start = true;
			}
			meta.field = HTTP_URI;
			field_cb(user_data, init_data + path + *payload_offset, path_len, &meta);
			if (normalize_uri(user_data, field_cb, ctx->protocol, (char *)
					(init_data + path + *payload_offset), path_len) != 0)
				return -UNKNOWN_ERROR;
			if (num_headers > 0)
				report_headers_list(ctx, headers, user_data, num_headers,
						  payload_offset, field_cb, init_data);
			if (ret == -2)
				return -EOP;
			ret += offset;
			offset = ret;
			*payload_offset = ret;
			if (consume_rest_of_content(mbuf, user_data, field_cb, ctx, &ret))
				continue;
			else
				return ret;
		}

		/* check for response packet */
		num_headers = MAX_HEADERS_NUM;
		ret = phr_parse_response(data, len, &minor_version, &status,
					 &msg, &msg_len, headers, &num_headers,
					 0);
		if (ret > 0 || ret == -2) {
			CHECK_VERSION(minor_version, ret);
			ctx->protocol = DPI_PARSER_HTTP_1;
			ctx->state = PARSER_STATE_PROT_VERIFIED;
			if (ctx->first) {
				meta.stream_start = false;
				meta.field = EMPTY_FIELD;
				field_cb(user_data, init_data, 0, &meta);
				ctx->first = false;
				meta.stream_start = true;
			}
			if (num_headers > 0)
				report_headers_list(ctx, headers, user_data, num_headers,
						payload_offset, field_cb, init_data);
			if (ret == -2)
				return -EOP;
			ret += offset;
			offset = ret;
			*payload_offset = ret;
			if (consume_rest_of_content(mbuf, user_data, field_cb, ctx, &ret))
				continue;
			else
				return ret;
		}

		/* check for headers packet */
		if (ctx->protocol == DPI_PARSER_HTTP_1 &&
		    ctx->state == PARSER_STATE_PROT_VERIFIED) {
			num_headers = MAX_HEADERS_NUM;
			ret = phr_parse_headers(data, len, headers, &num_headers, 0);
			if (ret > 0 || ret == -2) {
				if (num_headers > 0)
					report_headers_list(ctx, headers, user_data, num_headers,
							payload_offset, field_cb, init_data);
				if (ret == -2)
					return -EOP;
				ret += offset;
				offset = ret;
				*payload_offset = ret;
				if (consume_rest_of_content(mbuf, user_data, field_cb, ctx, &ret))
					continue;
				else
					return ret;
			}
		}

		DOCA_DLOG_DBG("HttpPicoParse failed\nret = %d, version = %d, num_headers = %ld",
			       ret, minor_version, num_headers);
		return -PARSE_ERROR;
	}
}

static int
http_parser_ctx_alloc(void **pctx)
{
	int ret = 0;
	struct http_parser_ctx **ctx = (struct http_parser_ctx **)pctx;

	*ctx = (struct http_parser_ctx *)rte_zmalloc(NULL,
					sizeof(struct http_parser_ctx), 0);
	if (*ctx == NULL) {
		ret = -ENOMEM;
		goto err_ctx;
	}
	(*ctx)->content_len = 0;
	(*ctx)->state = PARSER_STATE_PART;
	(*ctx)->protocol = DPI_PARSER_PROTOCOL_NOT_READY;
	(*ctx)->first = true;
	return ret;
err_ctx:
	return ret;
}

static void
http_parser_ctx_dealloc(void *pctx)
{
	struct http_parser_ctx *ctx = (struct http_parser_ctx *)pctx;

	if (ctx != NULL)
		rte_free(ctx);
}

/*
 * Parse new flow, or continue parsing prev flow.
 * The parser walk through the mbuf's data and parse it as if its http packet.
 * For not an http packet the parser will stop and update the flow context accordingly.
 * For http packet, the parser will call back with the uri path and headers it finds.
 * return:
 *	3    : parsing error - not an http
 *	else : done / still parsing the packet.
 */
static int
http_parser_parse(void *pctx, struct rte_mbuf *mbuf,
		  field_cb_t field_cb,
		  void *user_data, uint32_t *consumed)
{
	struct http_parser_ctx *ctx = (struct http_parser_ctx *)pctx;
	uint32_t payload_offset = 0;

	if (ctx->content_left) {
		if (!consume_rest_of_content(mbuf, user_data, field_cb, ctx,
					     (int *)&payload_offset)) {
			*consumed = payload_offset;
			return ctx->state;
		}
		ctx->content_left = false;
	}

	int ret = http_parser_packet(mbuf, &payload_offset, user_data, ctx, field_cb);

	if (ret > 0) {
		*consumed = ret;
		ctx->state = PARSER_STATE_DONE;
		return PARSER_STATE_DONE;
	}
	switch (ret) {
	case -PARSE_ERROR:
		DOCA_DLOG_DBG("transaction did not recognized as http");
		ctx->protocol = DPI_PARSER_PROTOCOL_MISS;
		*consumed = -1;
		ctx->state = PARSER_STATE_ERROR;
		return PARSER_STATE_ERROR;
	case -EOP:
		*consumed = payload_offset;
		return ctx->state;
	case -UNKNOWN_ERROR:
		DOCA_DLOG_ERR("payload_offset returned positive while http request/response did not fully parsed");
		*consumed = -1;
		return PARSER_STATE_ERROR;
	default:
		DOCA_DLOG_ERR("packet_parse function failed");
		*consumed = -1;
		ctx->protocol = -1;
		ctx->state = PARSER_STATE_ERROR;
		return PARSER_STATE_ERROR;
	}
}

static struct dpi_parser http_parser = {
	.ctx_alloc = http_parser_ctx_alloc,
	.ctx_dealloc = http_parser_ctx_dealloc,
	.parse = http_parser_parse,
};

struct dpi_parser *http_parser_get(void)
{
	return &http_parser;
}
