/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef HTTP_PARSER_H_
#define HTTP_PARSER_H_

/*
 * The DPI parser is designed to parse the application layer in a flow.
 * Because it can spread across packets, the parser needs to store the parsing
 * state of the flow, so it can continue parsing in the next packet.
 *
 * The user can query the parsing results at any state of the flow parsing,
 * But should be aware that if parsing is not done the parsing results are
 * incomplete.
 *
 * For performance reasons, The parser will not check the correctness of
 * protocol and may produce false positives.
 *
 * The input to the parser is mbuf, which represents a packet.
 * The user should provide parsing offset of each mbuf provided,
 * i.e. the beginning of the L7 payload.
 *
 * The parsing interim result is reported with the status struct.
 * The status contains data about the following:
 * 1. The flow's protocol being suspected to be.
 * 2. whether a field was found, and if so then also an offset to its
 *    begginning and end.
 * 3. An offset from which the flow should continue the next time its being
 *    called by the parser.
 *
 * When the parser is done, the user can get access to the offset and
 * length of every header and value found in the transaction.
 * For example, URI value which is spread across
 * multiple packets, will be represented by offset to the point in the mbuf
 * where it starts, and a length of the URI value.
 *
 * Because the parser does not copy the packet but holds a data to the field's
 * location, user should not delete or reuse the mbuf while the flow context
 * exists.
 *
 * A typical flow would be:
 *
 * if (new flow)
 *		dpi_parser_ctx_alloc(&(client_obj->p_ctx))
 * else
 *		client_obj = sft_client_obj_get(flow_id)
 * state = dpi_parser_parse(client_obj->p_ctx, mbuf, offset, &status)
 * if (state == PARSER_STATE_DONE)
 *		...
 *		protocol = dpi_parser_protocol_get(client_obj->p_ctx)
 *		uri = dpi_parser_field_get(HTTP_URI, ctx)
 *		...
 *
 * if (state == PARSER_STATE_PART)
 *		...
 *		protocol = dpi_parser_protocol_get(client_obj->p_ctx)
 *		uri_offset = dpi_parser_field_offset_get(ctx, HTTP_URI)
 *		uri_len = dpi_parser_field_len_get(ctx, HTTP_URI)
 *		...
 *
 * When done regex/post processing
 * dpi_parser_ctx_dealloc(client_obj->p_ctx)
 */

#include "base_parser.h"

#ifdef __cplusplus
extern "C" {
#endif

struct dpi_parser *http_parser_get(void);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* HTTP_PARSER_H_ */
