/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_malloc.h>
#include <openssl/x509v3.h>
#include <openssl/x509.h>
#include <openssl/bio.h>
#include <openssl/err.h>

#include <doca_log.h>

#include "ssl_parser.h"

DOCA_LOG_REGISTER(DPI::SSL_PARSER);

#define CN_SIZE 256

/**
 * SSL/TLS handshake message types
 */
enum ssl_handshake_type {
	/** Hello-request message type */
	DOCA_SSL_HELLO_REQUEST        = 0,
	/** Client-hello message type */
	DOCA_SSL_CLIENT_HELLO         = 1,
	/** Server-hello message type */
	DOCA_SSL_SERVER_HELLO         = 2,
	/** New-session-ticket message type */
	DOCA_SSL_NEW_SESSION_TICKET   = 4,
	/** End-of-early-data message type (TLS 1.3) */
	DOCA_SSL_END_OF_EARLY_DATE    = 5,
	/** Encrypted-extensions message type (TLS 1.3) */
	DOCA_SSL_ENCRYPTED_EXTENSIONS = 8,
	/** Certificate message type */
	DOCA_SSL_CERTIFICATE          = 11,
	/** Server-key-exchange message type */
	DOCA_SSL_SERVER_KEY_EXCHANGE  = 12,
	/** Certificate-request message type */
	DOCA_SSL_CERTIFICATE_REQUEST  = 13,
	/** Server-hello-done message type */
	DOCA_SSL_SERVER_DONE          = 14,
	/** Certificate-verify message type */
	DOCA_SSL_CERTIFICATE_VERIFY   = 15,
	/** Client-key-exchange message type */
	DOCA_SSL_CLIENT_KEY_EXCHANGE  = 16,
	/** Finish message type */
	DOCA_SSL_FINISHED             = 20,
	/** Key-update message type (TLS 1.3) */
	DOCA_SSL_KEY_UPDATE           = 24,
	/** Unknown SSL handshake message */
	DOCA_SSL_HANDSHAKE_UNKNOWN    = 255
};

enum ssl_ext_type {
	SSL_EXT_SERVER_NAME = 0,
};

struct ssl_tls_len {
	/** Length is 3-Byte long, This is the MSB */
	uint8_t length1;
	/** Length is 3-Byte long, This is the 2 LSB */
	uint16_t length2;
} __rte_packed;

struct ssl_tls_handshake_layer {
	/** Type of the handshake message (one of ::SSLHandshakeType) */
	uint8_t handshake_type;
	/** Length of the message. */
	struct ssl_tls_len len;
} __rte_packed;

struct ssl_tls_client_server_hello {
	/** SSL/TLS handshake version */
	uint16_t handshakeVersion;
	/** 32-bytes random number */
	uint8_t random[32];
} __rte_packed;

struct ssl_tls_record_layer {
	/** Message (record) type (one of ::SSLRecordType) */
	uint8_t record_type;
	/** Message (record) version (one of SSLVersion::SSLVersionEnum) */
	uint16_t record_version;
	/** Message (record) length in bytes */
	uint16_t length;
} __rte_packed;

struct session_id_hdr {
	uint8_t len;
	uint32_t offset;
};

struct ssl_parser_ctx {
	int last_handshake_msg;
	int state;
	struct session_id_hdr session_id;
	bool first;
};

static bool
is_ssl(uint8_t *data, size_t data_len)
{
	struct ssl_tls_record_layer *record_layer;

	if (data_len < sizeof(struct ssl_tls_record_layer))
		return false;
	record_layer = (struct ssl_tls_record_layer *)data;
	/* there is no SSL message with length 0 */
	if (record_layer->length == 0)
		return false;
	if (record_layer->record_type < 20 || record_layer->record_type > 23)
		return false;

	return true;
}

static
uint16_t get_cipher_suit_count(uint8_t *data)
{
	uint16_t cipher_suit_len;

	memcpy(&cipher_suit_len, data, sizeof(uint16_t));
	return rte_be_to_cpu_16(cipher_suit_len) / 2;
}

static uint32_t
get_ssl_len(const struct ssl_tls_len *cert_len_hdr)
{
	uint32_t len = rte_be_to_cpu_16(cert_len_hdr->length2) |
		       (cert_len_hdr->length1 << 16);
	return len;
}

static int
get_common_name(X509 * const cert, void *user_data, field_cb_t field_cb)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_SSL,
		.mbuf = false,
		.stream_start = true,
	};
	X509_NAME *subjectName;
	char commonName[CN_SIZE];
	int len = 0;

	subjectName = X509_get_subject_name(cert);
	if (subjectName == NULL) {
		DOCA_DLOG_DBG("no X509_NAME struct was found in the tls certificate\n");
		return 0;
	}

	/* Get subject extension from the certificate */
	char *certSubject = X509_NAME_oneline(subjectName, 0, 0);

	len = strlen(certSubject);
	meta.field = SSL_CERT_SUBJECT;
	field_cb(user_data, certSubject, len, &meta);

	/* Get only common_name */
	len = X509_NAME_get_text_by_NID(subjectName, NID_commonName, commonName, CN_SIZE);
	if (len <= 0) {
		DOCA_DLOG_ERR("no Common Name found in the sll certificate");
		OPENSSL_free(certSubject);
		return -1;
	}
	meta.field = SSL_COMMON_NAME;
	field_cb(user_data, commonName, len, &meta);
	OPENSSL_free(certSubject);
	return 0;
}

static int
get_sub_alt_name(X509 * const cert, void *user_data, field_cb_t field_cb)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_SSL,
		.mbuf = false,
		.stream_start = true,
	};
	int i = 0, success = 0;
	GENERAL_NAMES *names = NULL;
	GENERAL_NAME *entry = NULL;
	unsigned char *utf8 = NULL;
	int count = 0, ret;

	assert(cert);

	names = X509_get_ext_d2i(cert, NID_subject_alt_name, 0, 0);
	if (!names) {
		DOCA_DLOG_DBG("X509_get_ext_d2i function failed");
		ret = 0;
		goto no_name;
	};
	count = sk_GENERAL_NAME_num(names);

	if (!count) {
		DOCA_DLOG_DBG("sk_GENERAL_NAME_num function failed");
		ret = 0;
		goto zero_names;
	};

	for (i = 0; i < count; ++i) {
		entry = sk_GENERAL_NAME_value(names, i);
		if (!entry)
			continue;

		if (entry->type == GEN_DNS) {
			int len1 = 0, len2 = -1;

			len1 = ASN1_STRING_to_UTF8(&utf8, entry->d.dNSName);
			if (len1 < 0) {
				ret = -1;
				DOCA_DLOG_ERR("function ASN1_STRING_to_UTF8 failed. error type is %lu",
					    ERR_get_error());
				goto error;
			}
			if (utf8)
				len2 = (int)strlen((const char *)utf8);

			if (len1 != len2) {
				DOCA_DLOG_ERR("alt_name len does not match its buffer len from ASN1_STRING: %d vs %d", len2, len1);
				ret = -1;
				goto error;
			}

			if (utf8 && len1 && len2) {
				meta.field = SSL_SUBJECT_ALT_NAME;
				field_cb(user_data, utf8, len1, &meta);
				success = 1;
			}

			if (utf8)
				OPENSSL_free(utf8), utf8 = NULL;
		} else {
			DOCA_DLOG_DBG("Unknown GENERAL_NAME type: %d", entry->type);
			ret = 0;
			goto out;
		}
	}
	ret = 0;
	goto out;

out:
	if (!success)
		DOCA_DLOG_DBG("Alternative Names are not available");
error:
	if (utf8)
		OPENSSL_free(utf8);
zero_names:
	GENERAL_NAMES_free(names);
no_name:
	return ret;
}

static uint32_t
parse_server_name_cert(uint8_t *data, uint32_t len, void *user_data, field_cb_t field_cb)
{
	const unsigned char *cert_data = (unsigned char *)data;
	const struct ssl_tls_len *cert_len_hdr;
	uint32_t cert_len;
	X509 *x509_cert;

	cert_data += sizeof(struct ssl_tls_handshake_layer);
	cert_len_hdr = (const struct ssl_tls_len *)cert_data;
	cert_data += sizeof(struct ssl_tls_len);
	cert_len_hdr = (const struct ssl_tls_len *)cert_data;
	cert_len = get_ssl_len(cert_len_hdr);
	cert_data += sizeof(const struct ssl_tls_len);

	/* Parse data from ssl certificate */
	x509_cert = d2i_X509(NULL, &cert_data, cert_len);
	if (!x509_cert) {
		DOCA_DLOG_ERR("unable to parse certificate from data");
		X509_free(x509_cert);
		return -1;
	}

	/* Get common_name from the certificate */
	if (get_common_name(x509_cert, user_data, field_cb))
		return -1;

	if (get_sub_alt_name(x509_cert, user_data, field_cb))
		return -1;

	X509_free(x509_cert);
	return 0;
}

static uint32_t
parse_server_name_ch(uint8_t *data, uint32_t len, void *user_data, field_cb_t field_cb)
{
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_SSL,
		.mbuf = false,
		.stream_start = true,
	};
	uint32_t min_ssl_ext_len = 2*sizeof(uint16_t) + sizeof(uint8_t);
	uint16_t ext_len;
	uint8_t session_id_len;
	uint16_t cipher_suit_count;
	uint8_t compression_len;
	uint8_t *cur_pos, *ext_pos;

	cur_pos = data + sizeof(struct ssl_tls_handshake_layer);
	cur_pos += sizeof(struct ssl_tls_client_server_hello);
	if (cur_pos - data > len)
		return 0;
	session_id_len = *cur_pos;
	if (session_id_len > 0) {
		cur_pos += sizeof(uint8_t);
		cur_pos += session_id_len;
	} else
		cur_pos += session_id_len + sizeof(uint8_t);
	if (cur_pos - data > len)
		return 0;
	cipher_suit_count = get_cipher_suit_count(cur_pos);
	cur_pos += sizeof(uint16_t)*cipher_suit_count + sizeof(uint16_t);
	if (cur_pos - data > len)
		return 0;
	compression_len = *cur_pos;
	cur_pos += compression_len + sizeof(uint8_t);
	if (cur_pos - data > len)
		return 0;
	memcpy(&ext_len, cur_pos, sizeof(uint16_t));
	ext_len = rte_be_to_cpu_16(ext_len);

	cur_pos += sizeof(uint16_t);
	ext_pos = cur_pos;

	while ((cur_pos - ext_pos) < (int)ext_len
		&& (cur_pos - data) < (int)len
		&& (int)len - (cur_pos - data) >= (int)min_ssl_ext_len) {
		uint16_t ssl_ext_type, ext_len;
		uint8_t *ext_len_pos;

		memcpy(&ssl_ext_type, cur_pos, sizeof(uint16_t));
		ssl_ext_type = rte_be_to_cpu_16(ssl_ext_type);
		ext_len_pos = cur_pos + sizeof(uint16_t);
		memcpy(&ext_len, ext_len_pos, sizeof(uint16_t));
		ext_len = rte_be_to_cpu_16(ext_len);

		switch (ssl_ext_type) {
		case SSL_EXT_SERVER_NAME: {
			uint8_t *sni_pos = ext_len_pos + sizeof(uint16_t);
			uint16_t *sni_val_len_pos = (uint16_t *)(sni_pos +
				sizeof(uint16_t)  + sizeof(uint8_t));
			uint8_t *sni_val_pos = sni_pos + sizeof(uint16_t) +
				sizeof(uint8_t) + sizeof(uint16_t);
			uint16_t sni_len;

			memcpy(&sni_len, sni_val_len_pos, sizeof(uint16_t));
			sni_len = rte_be_to_cpu_16(sni_len);

			/* Remove zeros at the end */
			while (sni_len) {
				if (sni_val_pos[sni_len - 1])
					break;
				sni_len--;
			}
			meta.field = SSL_COMMON_NAME;
			field_cb(user_data, sni_val_pos, sni_len, &meta);
			return 1;
		}
		}

		cur_pos += ext_len + sizeof(uint16_t) + sizeof(uint16_t);
	}
	return 0;
}

static uint32_t get_handshake_msg_len(struct ssl_tls_handshake_layer *msg_hdr)
{
	uint32_t len = get_ssl_len(&msg_hdr->len);

	return sizeof(struct ssl_tls_handshake_layer) + len;
}

static int
parse_handshake(struct ssl_parser_ctx *ctx, void *user_data, uint8_t *data, uint32_t len, field_cb_t field_cb)
{
	struct ssl_tls_handshake_layer *msg_hdr =
		(struct ssl_tls_handshake_layer *)data;

	switch (msg_hdr->handshake_type) {
	case DOCA_SSL_CLIENT_HELLO:
		parse_server_name_ch(data, len, user_data, field_cb);
		break;
	case DOCA_SSL_CERTIFICATE:
		parse_server_name_cert(data, len, user_data, field_cb);
	case DOCA_SSL_SERVER_HELLO:
		break;
	}

	ctx->last_handshake_msg = msg_hdr->handshake_type;
	return 0;
}

static int
ssl_parser_parse(void *pctx, struct rte_mbuf *mbuf,
		 field_cb_t field_cb, void *user_data, uint32_t *consumed)
{
	struct ssl_parser_ctx *ctx = (struct ssl_parser_ctx *)pctx;
	uint8_t *data = rte_pktmbuf_mtod(mbuf, uint8_t *);
	uint8_t *init_data = data;
	uint32_t left_to_scan = rte_pktmbuf_pkt_len(mbuf);
	struct ssl_tls_record_layer *record;
	uint8_t *handshake_layer;
	struct ssl_tls_handshake_layer *msg_hdr;
	uint32_t msg_len;
	uint32_t next_record = 0;
	uint32_t payload_offset = 0;

	if (left_to_scan < sizeof(struct ssl_tls_record_layer)) {
		ctx->state = PARSER_STATE_PART;
		return PARSER_STATE_PART;
	}
	struct dpi_parser_cb_metadata meta = {
		.prot = DPI_PARSER_SSL,
		.mbuf = false,
		.stream_start = true,
	};

	while (left_to_scan) {
		if (payload_offset == next_record) {
			record = (struct ssl_tls_record_layer *)data;
			next_record += rte_be_to_cpu_16(record->length) +
				sizeof(struct ssl_tls_record_layer);
			if (!is_ssl(data, left_to_scan)) {
				ctx->state = PARSER_STATE_ERROR;
				return PARSER_STATE_ERROR;
			}
			if (ctx->first) {
				meta.stream_start = false;
				meta.field = EMPTY_FIELD;
				field_cb(user_data, init_data, 0, &meta);
				ctx->first = false;
				meta.stream_start = true;
			}
			if (left_to_scan < (rte_be_to_cpu_16(record->length) +
					    sizeof(struct ssl_tls_record_layer))) {
				goto finish;
			}
			data += sizeof(struct ssl_tls_record_layer);
			left_to_scan -= sizeof(struct ssl_tls_record_layer);
			payload_offset += sizeof(struct ssl_tls_record_layer);
		}
		handshake_layer = data;
		msg_hdr = (struct ssl_tls_handshake_layer *)handshake_layer;
		msg_len = get_handshake_msg_len(msg_hdr);
		if (msg_len > left_to_scan)
			break;
		parse_handshake(ctx, user_data, handshake_layer, msg_len, field_cb);
		data += msg_len;
		left_to_scan -= msg_len;
		payload_offset += msg_len;
	}

finish:
	ctx->state = PARSER_STATE_PROT_VERIFIED;
	*consumed = payload_offset;
	return ctx->state;
}

static int
ctx_alloc(void **pctx)
{
	struct ssl_parser_ctx *ctx =
		rte_zmalloc(NULL, sizeof(struct ssl_parser_ctx), 0);

	if (!ctx)
		return -ENOMEM;
	ctx->state = PARSER_STATE_PART;
	ctx->first = true;
	*pctx = ctx;
	return 0;
}

static void
ctx_dealloc(void *ctx)
{
	rte_free(ctx);
}

static struct dpi_parser ssl_parser = {
	.ctx_alloc = ctx_alloc,
	.ctx_dealloc = ctx_dealloc,
	.parse = ssl_parser_parse,
};

struct dpi_parser *
ssl_parser_get(void)
{
	return &ssl_parser;
}
