/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>

#include <rte_malloc.h>
#include <rte_mbuf.h>
#include <rte_errno.h>

#include <doca_log.h>

#include "dns_parser.h"
#include "dns_cache.h"
#include "stream_parser.h"

DOCA_LOG_REGISTER(DPI::STREAM_PARSER);

struct field {
	uint32_t offset;
	uint32_t len;
};

struct stream_parser_ctx {
	uint32_t last_offset;
	int protocol;
	int field;
	bool first;
};

static uint16_t max_sig_length = 5000;

void stream_parser_set_max_sig_len(uint16_t len)
{
	max_sig_length = len;
}

static int
stream_parser_parse(void *pctx, struct rte_mbuf *mbuf,
		 field_cb_t field_cb, void *user_data, uint32_t *consumed)
{
	struct stream_parser_ctx *ctx = (struct stream_parser_ctx *)pctx;
	struct dpi_parser_cb_metadata meta = {
		.prot = ctx->protocol,
		.mbuf = true,
		.stream_start = ctx->last_offset == 0,
		.field = ctx->field,
	};

	if (ctx->first) {
		meta.field = EMPTY_FIELD;
		meta.stream_start = false;
		meta.mbuf = false;
		field_cb(user_data, "", 0, &meta);
		ctx->first = false;
		meta.mbuf = true;
		meta.stream_start = ctx->last_offset == 0;
		meta.field = ctx->field;
	}
	if (mbuf->pool == NULL) {
		meta.mbuf = false;
		field_cb(user_data, rte_pktmbuf_mtod(mbuf, void *), rte_pktmbuf_pkt_len(mbuf), &meta);
	} else {
		field_cb(user_data, mbuf, rte_pktmbuf_pkt_len(mbuf), &meta);
		uint32_t pkt_len = rte_pktmbuf_pkt_len(mbuf);
		uint32_t data_len = rte_pktmbuf_data_len(mbuf);

		while ((pkt_len >= max_sig_length) && ((pkt_len - data_len) >= max_sig_length)) {
			*consumed += data_len;
			mbuf = mbuf->next;
			if (!mbuf || !mbuf->next)
				break;
			data_len = rte_pktmbuf_data_len(mbuf);
			pkt_len -= rte_pktmbuf_data_len(mbuf);

		}
	}
	ctx->last_offset += *consumed;
	return PARSER_STATE_PROT_VERIFIED;
}

static int
ctx_alloc(void **pctx)
{
	struct stream_parser_ctx **ctx = (struct stream_parser_ctx **)pctx;

	*ctx = (struct stream_parser_ctx *)rte_zmalloc(NULL, sizeof(struct stream_parser_ctx), 0);
	if (*ctx == NULL)
		return -ENOMEM;
	(*ctx)->first = true;
	return 0;
}

static int
tcp_ctx_alloc(void **pctx)
{
	int ret;
	struct stream_parser_ctx **ctx = (struct stream_parser_ctx **)pctx;

	ret = ctx_alloc(pctx);
	if (ret < 0)
		return ret;

	(*ctx)->protocol = DPI_PARSER_TCP;
	(*ctx)->field = TCP_STREAM;
	return 0;
}

static int
udp_ctx_alloc(void **pctx)
{
	int ret;
	struct stream_parser_ctx **ctx = (struct stream_parser_ctx **)pctx;

	ret = ctx_alloc(pctx);
	if (ret < 0)
		return ret;

	(*ctx)->protocol = DPI_PARSER_UDP;
	(*ctx)->field = UDP_STREAM;
	return 0;
}

static int
other_l4_protocol_ctx_alloc(void **pctx)
{
	int ret;
	struct stream_parser_ctx **ctx = (struct stream_parser_ctx **)pctx;

	ret = ctx_alloc(pctx);
	if (ret < 0)
		return ret;

	(*ctx)->protocol = DPI_PARSER_OTHER_L4;
	(*ctx)->field = OTHER_L4_STREAM;
	return 0;
}


static void
ctx_dealloc(void *pctx)
{
	struct stream_parser_ctx *ctx = (struct stream_parser_ctx *)pctx;

	if (ctx != NULL)
		rte_free(ctx);
}

static bool
support_mbuf(void *ctx)
{
	return true;
}

static struct dpi_parser tcp_parser = {
	.ctx_alloc = tcp_ctx_alloc,
	.ctx_dealloc = ctx_dealloc,
	.parse = stream_parser_parse,
	.support_mbuf = support_mbuf,
};

static struct dpi_parser udp_parser = {
	.ctx_alloc = udp_ctx_alloc,
	.ctx_dealloc = ctx_dealloc,
	.parse = stream_parser_parse,
	.support_mbuf = support_mbuf,
};

static struct dpi_parser other_l4_protocol_parser = {
	.ctx_alloc = other_l4_protocol_ctx_alloc,
	.ctx_dealloc = ctx_dealloc,
	.parse = stream_parser_parse,
	.support_mbuf = support_mbuf,
};

struct dpi_parser *tcp_parser_get(void)
{
	return &tcp_parser;
}

struct dpi_parser *udp_parser_get(void)
{
	return &udp_parser;
}

struct dpi_parser *other_l4_protocol_parser_get(void)
{
	return &other_l4_protocol_parser;
}
