/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <bsd/string.h>

#include <rte_ethdev.h>
#include <grpcpp/client_context.h>
#include <grpcpp/create_channel.h>

#include <doca_log.h>

#include "doca_flow_grpc_client.h"
#include "common.pb.h"
#include "doca_flow.grpc.pb.h"

DOCA_LOG_REGISTER(DOCA_FLOW_GRPC_CLIENT);

using grpc::Channel;
using grpc::ClientAsyncResponseReader;
using grpc::ClientContext;
using grpc::CompletionQueue;
using grpc::Status;

#define MAIN_CLIENT_IDX (0)	/*< main client index in clients list. */
#define DOCA_ETHER_ADDR_LEN (6) /*< length of ether address. */
#define DOCA_IPV6_ADDR_LEN (4)	/*< length of ipv6 address. */
#define MAX_BUFF_SIZE (256)
#define USER_META_MAX_SIZE (4)	/*< maximum size of user meta-data in 32-bit resolution. */

static int nb_queues;

struct AsyncClientCall {
	DocaFlowResponse reply;
	ClientContext context;
	Status status;
	std::unique_ptr<ClientAsyncResponseReader<DocaFlowResponse>> response_reader;
};

struct DocaFlowClient {
	std::unique_ptr<DocaFlow::Stub> stub_;
	CompletionQueue cq_;
	int queue;
};

struct DocaFlowClient *client_list;
std::unique_ptr<DocaFlow::Stub> main_stub;
std::string server_address;
int base_port;

static char tmp_error_buffer[MAX_BUFF_SIZE];

static void
createMacAddress(const uint8_t *mac_address, uint64_t *mac_address_req)
{
	int byte_idx;
	uint8_t *mac_address_byte_ptr = (uint8_t *)mac_address_req;

	/* need only the first 6 bytes for the mac address */
	for (byte_idx = 0; byte_idx < DOCA_ETHER_ADDR_LEN; byte_idx++)
		mac_address_byte_ptr[byte_idx] = mac_address[byte_idx];
}

static void
createIpAddress(const struct doca_flow_ip_addr *ip_add, DocaFlowIPAddress *ip_add_req)
{
	int iter;

	ip_add_req->set_type((DocaFlowIPtype)ip_add->type);
	if (ip_add->type == DOCA_FLOW_IP4_ADDR)
		ip_add_req->set_ipv4_addr(ip_add->ipv4_addr);

	if (ip_add->type == DOCA_FLOW_IP6_ADDR) {
		for (iter = 0; iter < DOCA_IPV6_ADDR_LEN; iter++)
			ip_add_req->add_ipv6_addr(ip_add->ipv6_addr[iter]);
	}
}

static void
createTun(const struct doca_flow_tun *tun, DocaFlowTun *tun_req)
{
	tun_req->set_type((DocaFlowTunType)tun->type);
	switch (tun_req->type()) {
	case TUN_VXLAN:
		tun_req->set_vxlan_tun_id(tun->vxlan_tun_id);
		break;
	case TUN_GRE:
		tun_req->set_gre_key(tun->gre_key);
		break;
	case TUN_GTPU:
		tun_req->set_gtp_teid(tun->gtp_teid);
		break;
	default:
		DOCA_LOG_DBG("doca_flow_tun configuration - tun type not defined");
	}
}

static void
createMeta(const struct doca_flow_meta *meta, DocaFlowMeta *meta_req)
{
	meta_req->set_pkt_meta(meta->pkt_meta);
	meta_req->set_mark(meta->mark);
	for(uint32_t i = 0; i < USER_META_MAX_SIZE; i++)
		meta_req->add_u32(meta->u32[i]);
}

static void
createMatch(const struct doca_flow_match *match, DocaFlowMatch *match_req)
{
	uint64_t out_src_mac, out_dst_mac, in_src_mac, in_dst_mac;
	DocaFlowIPAddress *out_src_ip;
	DocaFlowIPAddress *out_dst_ip;
	DocaFlowTun *tun_req;
	DocaFlowIPAddress *in_src_ip;
	DocaFlowIPAddress *in_dst_ip;

	match_req->set_flags(match->flags);
	createMeta(&(match->meta), match_req->mutable_meta());
	createMacAddress(match->out_src_mac, &out_src_mac);
	createMacAddress(match->out_dst_mac, &out_dst_mac);
	match_req->set_out_src_mac(out_src_mac);
	match_req->set_out_dst_mac(out_dst_mac);
	match_req->set_out_eth_type(match->out_eth_type);
	match_req->set_out_vlan_tci(match->out_vlan_tci);
	out_src_ip = match_req->mutable_out_src_ip();
	out_dst_ip = match_req->mutable_out_dst_ip();
	createIpAddress(&(match->out_src_ip), out_src_ip);
	createIpAddress(&(match->out_dst_ip), out_dst_ip);
	match_req->set_out_l4_type(match->out_l4_type);
	match_req->set_out_tcp_flags(match->out_tcp_flags);
	match_req->set_out_src_port(match->out_src_port);
	match_req->set_out_dst_port(match->out_dst_port);

	tun_req = match_req->mutable_tun();
	createTun(&(match->tun), tun_req);
	createMacAddress(match->in_src_mac, &in_src_mac);
	createMacAddress(match->in_dst_mac, &in_dst_mac);
	match_req->set_in_src_mac(in_src_mac);
	match_req->set_in_dst_mac(in_dst_mac);
	match_req->set_in_eth_type(match->in_eth_type);
	match_req->set_in_vlan_tci(match->in_vlan_tci);
	in_src_ip = match_req->mutable_in_src_ip();
	in_dst_ip = match_req->mutable_in_dst_ip();
	createIpAddress(&(match->in_src_ip), in_src_ip);
	createIpAddress(&(match->in_dst_ip), in_dst_ip);
	match_req->set_in_l4_type(match->in_l4_type);
	match_req->set_in_tcp_flags(match->in_tcp_flags);
	match_req->set_in_src_port(match->in_src_port);
	match_req->set_in_dst_port(match->in_dst_port);
}

static void
createEncapAction(const struct doca_flow_encap_action *encap_action, DocaFlowEncapAction *encap_action_req)
{
	uint64_t src_mac;
	uint64_t dst_mac;
	DocaFlowIPAddress *src_ip;
	DocaFlowIPAddress *dst_ip;
	DocaFlowTun *tun_req;

	createMacAddress(encap_action->src_mac, &src_mac);
	createMacAddress(encap_action->dst_mac, &dst_mac);
	encap_action_req->set_src_mac(src_mac);
	encap_action_req->set_dst_mac(dst_mac);

	encap_action_req->set_vlan_tci(encap_action->vlan_tci);

	src_ip = encap_action_req->mutable_src_ip();
	dst_ip = encap_action_req->mutable_dst_ip();
	createIpAddress(&(encap_action->src_ip), src_ip);
	createIpAddress(&(encap_action->dst_ip), dst_ip);

	tun_req = encap_action_req->mutable_tun();
	createTun(&(encap_action->tun), tun_req);
}

static void
createActions(const struct doca_flow_actions *actions, DocaFlowActions *actions_req)
{
	uint64_t mod_src_mac;
	uint64_t mod_dst_mac;
	DocaFlowIPAddress *mod_src_ip;
	DocaFlowIPAddress *mod_dst_ip;
	DocaFlowEncapAction *encap_action_req;

	actions_req->set_action_idx(actions->action_idx);
	actions_req->set_flags(actions->flags);
	actions_req->set_decap(actions->decap);

	createMeta(&actions->meta, actions_req->mutable_meta());

	createMacAddress(actions->mod_src_mac, &mod_src_mac);
	createMacAddress(actions->mod_dst_mac, &mod_dst_mac);
	actions_req->set_mod_src_mac(mod_src_mac);
	actions_req->set_mod_dst_mac(mod_dst_mac);
	actions_req->set_mod_vlan_id(actions->mod_vlan_id);
	mod_src_ip = actions_req->mutable_mod_src_ip();
	mod_dst_ip = actions_req->mutable_mod_dst_ip();
	createIpAddress(&(actions->mod_src_ip), mod_src_ip);
	createIpAddress(&(actions->mod_dst_ip), mod_dst_ip);

	actions_req->set_mod_src_port(actions->mod_src_port);
	actions_req->set_mod_dst_port(actions->mod_dst_port);
	actions_req->set_ttl(actions->ttl);
	actions_req->set_encap_exist(actions->has_encap);

	encap_action_req = actions_req->mutable_encap();
	createEncapAction(&(actions->encap), encap_action_req);
}

static void
createActionField(const struct doca_flow_action_field *action_field, DocaFlowActionField *action_field_req, void *match_address)
{
	if(action_field->address)
	{
		uint64_t address_offset = (uint8_t *)(action_field->address) - (uint8_t *)match_address;
		action_field_req->set_address_offset(address_offset);
	}

	action_field_req->set_offset(action_field->offset);
}


static void
createActionDesc(const struct doca_flow_action_desc *action_desc, DocaFlowActionDesc *action_desc_req, void *match_address)
{
	DocaFlowActionCopy *copy;
	DocaFlowActionMask *mask;
	struct doca_flow_action_field zeroed_action_field;

	memset(&zeroed_action_field, 0, sizeof(zeroed_action_field));

	action_desc_req->set_type(DocaFlowActionType(action_desc->type));
	switch (action_desc->type)
	{
		case DOCA_FLOW_ACTION_COPY:
			copy = action_desc_req->mutable_copy();
			if(memcmp(&(action_desc->copy.src), &zeroed_action_field, sizeof(zeroed_action_field)))
				createActionField(&(action_desc->copy.src), copy->mutable_src(), match_address);
			else
				createActionField(&(action_desc->copy.dst), copy->mutable_dst(), match_address);
			copy->set_width(action_desc->copy.width);
			break;
		default:
			mask = action_desc_req->mutable_mask();
			for(int u8_iter = 0; u8_iter < 16; u8_iter++)
                		mask->add_u8(action_desc->mask.u8[u8_iter]);
			break;
	}
}

static void
createActionDescsMeta(struct doca_flow_action_descs_meta *action_descs_meta, DocaFlowActionDescsMeta *action_descs_meta_req, void *match_address)
{
	struct doca_flow_action_desc zeroed_action_desc;
	memset(&zeroed_action_desc, 0, sizeof(zeroed_action_desc));


	if(memcmp(&(action_descs_meta->pkt_meta), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs_meta->pkt_meta, action_descs_meta_req->mutable_pkt_meta(), match_address);


        for (int u32_iter = 0; u32_iter < USER_META_MAX_SIZE; u32_iter++)
               		createActionDesc(&action_descs_meta->u32[u32_iter], action_descs_meta_req->add_u32(), match_address);
}

static void
createActionDescs(struct doca_flow_action_descs *action_descs, DocaFlowActionDescs *action_descs_req, void *match_address)
{
	struct doca_flow_action_desc zeroed_action_desc;
	struct doca_flow_action_descs_meta zeroed_action_descs_meta;

	memset(&zeroed_action_desc, 0, sizeof(zeroed_action_desc));
	memset(&zeroed_action_descs_meta, 0, sizeof(zeroed_action_descs_meta));

	if(memcmp(&(action_descs->src_mac), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->src_mac, action_descs_req->mutable_src_mac(), match_address);

	if(memcmp(&(action_descs->dst_mac), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->dst_mac, action_descs_req->mutable_dst_mac(), match_address);

	if(memcmp(&(action_descs->eth_type), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->eth_type, action_descs_req->mutable_eth_type(), match_address);

	if(memcmp(&(action_descs->vlan_id), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->vlan_id, action_descs_req->mutable_vlan(), match_address);

	if(memcmp(&(action_descs->src_ip), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->src_ip, action_descs_req->mutable_src_ip(), match_address);

	if(memcmp(&(action_descs->dst_ip), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->dst_ip, action_descs_req->mutable_dst_ip(), match_address);

	if(memcmp(&(action_descs->ttl), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->ttl, action_descs_req->mutable_ttl(), match_address);

	if(memcmp(&(action_descs->src_port), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->src_port, action_descs_req->mutable_src_port(), match_address);

	if(memcmp(&(action_descs->dst_port), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&action_descs->dst_port, action_descs_req->mutable_dst_port(), match_address);

	if(memcmp(&(action_descs->meta), &zeroed_action_descs_meta, sizeof(zeroed_action_descs_meta)))
		createActionDescsMeta(&(action_descs->meta), action_descs_req->mutable_meta(), match_address);

	if(memcmp(&(action_descs->tunnel), &zeroed_action_desc, sizeof(zeroed_action_desc)))
		createActionDesc(&(action_descs->tunnel), action_descs_req->mutable_tunnel(), match_address);
}

static void
createPipeAttr(struct doca_flow_pipe_attr *attr, DocaFlowPipeAttr *attr_req)
{
	if (attr->name)
		attr_req->set_name(attr->name);
	attr_req->set_type(DocaFlowPipeType(attr->type));
	attr_req->set_is_root(attr->is_root);
	attr_req->set_nb_flows(attr->nb_flows);
	attr_req->set_nb_actions(attr->nb_actions);
}

static void
createMonitor(const struct doca_flow_monitor *monitor, DocaFlowMonitor *monitor_req)
{
	monitor_req->set_flags(monitor->flags);

	monitor_req->set_cir(monitor->cir);
	monitor_req->set_cbs(monitor->cbs);

        monitor_req->set_shared_meter_id(monitor->shared_meter_id);
        monitor_req->set_shared_counter_id(monitor->shared_counter_id);

	monitor_req->set_aging(monitor->aging);
	monitor_req->set_user_data(monitor->user_data);
}

static void
createPipeCfg(const struct doca_flow_grpc_pipe_cfg *client_cfg, DocaFlowPipeCfg *cfg_req)
{
	int iter = 0;
	struct doca_flow_pipe_cfg *cfg;
	DocaFlowActions actions_req;
	DocaFlowActionDescs action_descs_req;
	if (!client_cfg->cfg) {
		DOCA_LOG_DBG("missing doca flow pipe configuration");
		return;
	}

	cfg = client_cfg->cfg;

	cfg_req->set_port_id(client_cfg->port_id);

	createPipeAttr(&cfg->attr, cfg_req->mutable_attr());

	if (cfg->match) {
		DocaFlowMatch *match = cfg_req->mutable_match();
		createMatch(cfg->match, match);
	}

	if (cfg->match_mask) {
		DocaFlowMatch *match_mask = cfg_req->mutable_match_mask();
		createMatch(cfg->match_mask, match_mask);
	}

	if (cfg->actions) {
		for(iter = 0; iter < cfg->attr.nb_actions; iter++)
			createActions(cfg->actions[iter], cfg_req->add_actions());
	}

	if(cfg->action_descs) {
		for(iter = 0; iter < cfg->attr.nb_actions; iter++)
			createActionDescs(cfg->action_descs[iter],cfg_req->add_action_descs(), cfg->match);
	}

	if (cfg->monitor) {
		DocaFlowMonitor *monitor = cfg_req->mutable_monitor();
		createMonitor(cfg->monitor, monitor);
	}
}

static void
createFwd(const struct doca_flow_grpc_fwd *client_fwd, DocaFlowFwd *fwd_req)
{
	struct doca_flow_fwd *fwd;

	if (!client_fwd->fwd) {
		DOCA_LOG_DBG("missing doca flow fwd");
		return;
	}

	fwd = client_fwd->fwd;
	fwd_req->set_type((DocaFlowFwdType)fwd->type);

	switch (fwd->type) {
	case DOCA_FLOW_FWD_RSS:
		fwd_req->set_rss_flags(fwd->rss_flags);
		fwd_req->set_num_of_queues(fwd->num_of_queues);
		for (int i = 0; i < fwd->num_of_queues; i++)
			fwd_req->add_rss_queues(fwd->rss_queues[i]);
		break;
	case DOCA_FLOW_FWD_PORT:
		fwd_req->set_port_id(fwd->port_id);
		break;
	case DOCA_FLOW_FWD_PIPE:
		fwd_req->set_next_pipe_id(client_fwd->next_pipe_id);
		break;
	case DOCA_FLOW_FWD_DROP:
		break;
	default:
		DOCA_LOG_DBG("doca_flow_fwd configuration - fwd type not found");
	}
}

static void
createResponse(DocaFlowResponse &response_msg, Status status, struct doca_flow_grpc_response *response)
{
	DocaFlowError error = response_msg.error();
	DocaFlowAgingHandleRes aging_res = response_msg.handle_aging_res();
	if (!status.ok()) {
		response->success = false;
		strlcpy(tmp_error_buffer, status.error_message().c_str(), MAX_BUFF_SIZE);
	}
	else {
		response->success = response_msg.success();
		response->error.type = (doca_flow_error_type)error.type();
		strlcpy(tmp_error_buffer, error.err_message().c_str(), MAX_BUFF_SIZE);
	}

	response->error.message = tmp_error_buffer;
	response->pipe_id = response_msg.pipe_id();
	response->entry_id = response_msg.entry_id();
	response->switch_port_id = response_msg.switch_port_id();
	response->aging_res = aging_res.res();
	response->nb_entries_processed = response_msg.nb_entries_processed();
	response->entry_status = (doca_flow_entry_status)response_msg.status();
}

static void
create_rpc_call(AsyncClientCall *call, CompletionQueue *cq_, struct doca_flow_grpc_response *response)
{
	void *got_tag;
	bool ok = false;

	call->response_reader->StartCall();
	call->response_reader->Finish(&call->reply, &call->status, (void *)call);
	(*cq_).Next(&got_tag, &ok);
	call = static_cast<AsyncClientCall *>(got_tag);
	createResponse(call->reply, call->status, response);
}

void
doca_flow_grpc_client_create(const char *grpc_address)
{
	DOCA_LOG_DBG("Doca Flow grpc client create request");
	std::string grpc_full_address(grpc_address);
	size_t port_idx = grpc_full_address.find(':');

	if (port_idx == std::string::npos) {
		grpc_full_address += ":" + std::to_string(eNetworkPort::k_DocaFlow);
		server_address = grpc_address;
		base_port = eNetworkPort::k_DocaFlow;
	} else {
		server_address = std::string(grpc_address).substr(0, port_idx);
		base_port = stoi(std::string(grpc_address).substr(port_idx + 1));
	}

	std::shared_ptr<grpc::ChannelInterface> channel =
		grpc::CreateChannel(grpc_full_address, grpc::InsecureChannelCredentials());
	main_stub = DocaFlow::NewStub(channel);
}

struct doca_flow_grpc_response
doca_flow_grpc_init(const struct doca_flow_cfg *cfg)
{
	DOCA_LOG_DBG("Doca Flow initialize request");
	DocaFlowCfg request;
	DocaFlowResources *resource_req = request.mutable_resource();
	struct doca_flow_grpc_response response;
	CompletionQueue cq_;

	memset(&response, 0, sizeof(response));

	request.set_queues(cfg->queues);
	resource_req->set_nb_counters(cfg->resource.nb_counters);
	resource_req->set_nb_meters(cfg->resource.nb_meters);
	request.set_mode_args(std::string(cfg->mode_args));
	for (int i = 0; i < DOCA_FLOW_SHARED_RESOURCE_MAX; i++)
		request.add_nr_shared_resources(cfg->nr_shared_resources[i]);

	request.set_queue_depth(cfg->queue_depth);

	AsyncClientCall *init_call = new AsyncClientCall;
	init_call->response_reader = main_stub->PrepareAsyncDocaFlowInit(&init_call->context, request, &cq_);
	create_rpc_call(init_call, &cq_, &response);

	if (response.success) {
		nb_queues = cfg->queues;
		client_list = new DocaFlowClient[cfg->queues];
		client_list[0].stub_ = std::move(main_stub);
		for (int i = 1; i < cfg->queues; i++) {
			int port = base_port + i;
			std::string channel_address(server_address + ":" + std::to_string(port));
			std::shared_ptr<grpc::ChannelInterface> channel =
				grpc::CreateChannel(channel_address, grpc::InsecureChannelCredentials());
			client_list[i].stub_ = DocaFlow::NewStub(channel);
		}
	}
	delete init_call;

	return response;
}

void
doca_flow_grpc_destroy()
{
	DOCA_LOG_DBG("Doca Flow destroy request");
	DocaFlowEmptyRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	AsyncClientCall *port_start_call = new AsyncClientCall;
	port_start_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowDestroy(
		&port_start_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(port_start_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete port_start_call;
}

struct doca_flow_grpc_response
doca_flow_grpc_port_start(const struct doca_flow_port_cfg *cfg)
{
	DOCA_LOG_DBG("Doca Flow port start request");
	DocaFlowPortCfg request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_port_id(cfg->port_id);
	request.set_type((DocaFlowPortType)cfg->type);
	request.set_devargs(cfg->devargs);
	request.set_priv_data_size(cfg->priv_data_size);

	AsyncClientCall *port_start_call = new AsyncClientCall;
	port_start_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPortStart(
		&port_start_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(port_start_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete port_start_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_port_stop(uint16_t port_id)
{
	DOCA_LOG_DBG("Doca Flow port stop request");
	DocaFlowPort request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_port_id(port_id);

	AsyncClientCall *port_stop_call = new AsyncClientCall;
	port_stop_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPortStop(
		&port_stop_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(port_stop_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete port_stop_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_port_pair(uint16_t port_id, uint16_t pair_port_id)
{
	DOCA_LOG_DBG("Doca Flow port pair request");
	DocaFlowPortPairRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_port_id(port_id);
	request.set_pair_port_id(pair_port_id);

	AsyncClientCall *port_pair_call = new AsyncClientCall;
	port_pair_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPortPair(
		&port_pair_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(port_pair_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete port_pair_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_shared_resource_cfg(enum doca_flow_shared_resource_type type, uint32_t id,
		struct doca_flow_shared_resource_cfg *cfg)
{
	DOCA_LOG_DBG("Doca Flow shared resources cfg request");
	DocaFlowSharedResourceCfgRequest request;
	SharedResourceCfg *cfg_req = request.mutable_cfg();
	DocaFlowResourceMeterCfg *meter_cfg_req = cfg_req->mutable_meter_cfg();
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_type((DocaFlowSharedResourceType)type);
	request.set_id(id);
	meter_cfg_req->set_cir(cfg->meter_cfg.cir);
	meter_cfg_req->set_cbs(cfg->meter_cfg.cbs);

	AsyncClientCall *shared_resource_cfg_call = new AsyncClientCall;
	shared_resource_cfg_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowSharedResourceCfg(
		&shared_resource_cfg_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(shared_resource_cfg_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete shared_resource_cfg_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_shared_resources_bind(enum doca_flow_shared_resource_type type, uint32_t *res_array,
		uint32_t res_array_len,	struct doca_flow_grpc_bindable_obj *bindable_obj_id)
{
	DOCA_LOG_DBG("Doca Flow shared resources bind request");
	DocaFlowSharedResourcesBindRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_type((DocaFlowSharedResourceType)type);
	for (int i = 0; i < res_array_len; i++)
		request.add_resource_arr(res_array[i]);
	if (bindable_obj_id->type == DOCA_FLOW_GRPC_BIND_TYPE_PORT)
		request.set_port_id(bindable_obj_id->pipe_id);
	if (bindable_obj_id->type == DOCA_FLOW_GRPC_BIND_TYPE_PIPE)
		request.set_pipe_id(bindable_obj_id->port_id);

	AsyncClientCall *shared_resources_bind_call = new AsyncClientCall;
	shared_resources_bind_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowSharedResourcesBind(
		&shared_resources_bind_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(shared_resources_bind_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete shared_resources_bind_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_shared_resources_query(enum doca_flow_shared_resource_type type, uint32_t *res_array,
		struct doca_flow_shared_resource_result *query_results_array, uint32_t array_len)
{
	DOCA_LOG_DBG("Doca Flow shared resources query request");
	DocaFlowSharedResourcesQueryRequest request;
	struct doca_flow_grpc_response response;
	int i;

	request.set_type((DocaFlowSharedResourceType)type);
	for (i = 0; i < array_len; i++) {
		request.add_res_array(res_array[i]);
	}

	AsyncClientCall *shared_resources_query_call = new AsyncClientCall;
	shared_resources_query_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowSharedResourcesQuery(
		&shared_resources_query_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(shared_resources_query_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);

	for (i = 0; i < array_len; i++) {
		query_results_array[i].counter.total_bytes = shared_resources_query_call->reply.query_results(i).total_bytes();
		query_results_array[i].counter.total_pkts = shared_resources_query_call->reply.query_results(i).total_pkts();
	}

	delete shared_resources_query_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_pipe_create(const struct doca_flow_grpc_pipe_cfg *client_cfg, const struct doca_flow_grpc_fwd *client_fwd,
			   const struct doca_flow_grpc_fwd *client_fwd_miss)
{
	DOCA_LOG_DBG("Doca Flow create pipe request");
	DocaFlowPipeCreateRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	if (client_cfg) {
		DocaFlowPipeCfg *cfg_req = request.mutable_cfg();
		createPipeCfg(client_cfg, cfg_req);
	}

	if (client_fwd) {
		DocaFlowFwd *fwd_req = request.mutable_fwd();
		createFwd(client_fwd, fwd_req);
	}

	if (client_fwd_miss) {
		DocaFlowFwd *fwd_miss_req = request.mutable_fwd_miss();
		createFwd(client_fwd_miss, fwd_miss_req);
	}

	AsyncClientCall *create_pipe_call = new AsyncClientCall;
	create_pipe_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPipeCreate(
		&create_pipe_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(create_pipe_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete create_pipe_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_pipe_add_entry(uint16_t pipe_queue, uint64_t pipe_id, const struct doca_flow_match *match,
			      const struct doca_flow_actions *actions, const struct doca_flow_monitor *monitor,
			      const struct doca_flow_grpc_fwd *client_fwd, uint32_t flags)
{
	DOCA_LOG_DBG("Doca Flow pipe add entry request");
	DocaFlowPipeAddEntryRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	if (pipe_queue >= nb_queues) {
		DOCA_LOG_ERR("Invalid pipe_queue, must be in range 0-%d", nb_queues - 1);
		response.success = false;
		response.error.message = "Invalid pipe_queue";
		return response;
	}

	request.set_pipe_queue(pipe_queue);
	request.set_pipe_id(pipe_id);
	request.set_flags(flags);

	if (match) {
		DocaFlowMatch *match_req = request.mutable_match();
		createMatch(match, match_req);
	}

	if (actions) {
		DocaFlowActions *actions_req = request.mutable_actions();
		createActions(actions, actions_req);
	}

	if (monitor) {
		DocaFlowMonitor *monitor_req = request.mutable_monitor();
		createMonitor(monitor, monitor_req);
	}

	if (client_fwd) {
		DocaFlowFwd *fwd_req = request.mutable_fwd();
		createFwd(client_fwd, fwd_req);
	}

	AsyncClientCall *pipe_add_entry_call = new AsyncClientCall;
	pipe_add_entry_call->response_reader = client_list[pipe_queue].stub_->PrepareAsyncDocaFlowPipeAddEntry(
		&pipe_add_entry_call->context, request, &client_list[pipe_queue].cq_);
	create_rpc_call(pipe_add_entry_call, &client_list[pipe_queue].cq_, &response);
	delete pipe_add_entry_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_pipe_control_add_entry(uint16_t pipe_queue, uint8_t priority, uint64_t pipe_id,
				      const struct doca_flow_match *match, const struct doca_flow_match *match_mask,
				      const struct doca_flow_grpc_fwd *client_fwd)
{
	DOCA_LOG_DBG("Doca Flow control pipe add entry request");
	DocaFlowPipeControlAddEntryRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	if (pipe_queue >= nb_queues) {
		DOCA_LOG_ERR("Invalid pipe_queue, must be in range 0-%d", nb_queues - 1);
		response.success = false;
		response.error.message = "Invalid pipe_queue";
		return response;
	}

	request.set_pipe_queue(pipe_queue);
	request.set_priority(priority);
	request.set_pipe_id(pipe_id);

	if (match) {
		DocaFlowMatch *match_req = request.mutable_match();
		createMatch(match, match_req);
	}

	if (match_mask) {
		DocaFlowMatch *match_mask_req = request.mutable_match_mask();
		createMatch(match_mask, match_mask_req);
	}

	if (client_fwd) {
		DocaFlowFwd *fwd_req = request.mutable_fwd();
		createFwd(client_fwd, fwd_req);
	}

	AsyncClientCall *control_pipe_add_entry_call = new AsyncClientCall;
	control_pipe_add_entry_call->response_reader =
		client_list[pipe_queue].stub_->PrepareAsyncDocaFlowPipeControlAddEntry(
			&control_pipe_add_entry_call->context, request, &client_list[pipe_queue].cq_);
	create_rpc_call(control_pipe_add_entry_call, &client_list[pipe_queue].cq_, &response);
	delete control_pipe_add_entry_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_pipe_lpm_add_entry(uint16_t pipe_queue, uint64_t pipe_id, const struct doca_flow_match *match,
		const struct doca_flow_match *match_mask, const struct doca_flow_actions *actions,
		const struct doca_flow_monitor *monitor, const struct doca_flow_grpc_fwd *client_fwd,
		const enum doca_flow_flags_type flag)
{
	DOCA_LOG_DBG("Doca Flow LPM pipe add entry request");
	DocaFlowPipeLpmAddEntryRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	if (pipe_queue >= nb_queues) {
		DOCA_LOG_ERR("Invalid pipe_queue, must be in range 0-%d", nb_queues - 1);
		response.success = false;
		response.error.message = "Invalid pipe_queue";
		return response;
	}

	request.set_pipe_queue(pipe_queue);
	request.set_pipe_id(pipe_id);
	request.set_flag(flag);

	if (match) {
		DocaFlowMatch *match_req = request.mutable_match();
		createMatch(match, match_req);
	}

	if (match_mask) {
		DocaFlowMatch *match_mask_req = request.mutable_match_mask();
		createMatch(match_mask, match_mask_req);
	}

	if (actions) {
		DocaFlowActions *actions_req = request.mutable_actions();
		createActions(actions, actions_req);
	}

	if (monitor) {
		DocaFlowMonitor *monitor_req = request.mutable_monitor();
		createMonitor(monitor, monitor_req);
	}

	if (client_fwd) {
		DocaFlowFwd *fwd_req = request.mutable_fwd();
		createFwd(client_fwd, fwd_req);
	}

	AsyncClientCall *lpm_pipe_add_entry_call = new AsyncClientCall;
	lpm_pipe_add_entry_call->response_reader = client_list[pipe_queue].stub_->PrepareAsyncDocaFlowPipeLpmAddEntry(
		&lpm_pipe_add_entry_call->context, request, &client_list[pipe_queue].cq_);
	create_rpc_call(lpm_pipe_add_entry_call, &client_list[pipe_queue].cq_, &response);
	delete lpm_pipe_add_entry_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_pipe_rm_entry(uint16_t pipe_queue, uint64_t entry_id)
{
	DOCA_LOG_DBG("Doca Flow remove entry request");
	DocaFlowPipeRmEntryRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	if (pipe_queue >= nb_queues) {
		DOCA_LOG_ERR("Invalid pipe_queue, must be in range 0-%d", nb_queues - 1);
		response.success = false;
		response.error.message = "Invalid pipe_queue";
		return response;
	}

	request.set_pipe_queue(pipe_queue);
	request.set_entry_id(entry_id);

	AsyncClientCall *pipe_rm_entry_call = new AsyncClientCall;
	pipe_rm_entry_call->response_reader = client_list[pipe_queue].stub_->PrepareAsyncDocaFlowPipeRmEntry(
		&pipe_rm_entry_call->context, request, &client_list[pipe_queue].cq_);
	create_rpc_call(pipe_rm_entry_call, &client_list[pipe_queue].cq_, &response);
	delete pipe_rm_entry_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_pipe_destroy(uint64_t pipe_id)
{
	DOCA_LOG_DBG("Doca Flow destroy pipe request");
	DocaFlowPipeDestroyRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_pipe_id(pipe_id);

	AsyncClientCall *destroy_pipe_call = new AsyncClientCall;
	destroy_pipe_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPipeDestroy(
		&destroy_pipe_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(destroy_pipe_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete destroy_pipe_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_port_pipes_flush(uint16_t port_id)
{
	DOCA_LOG_DBG("Doca Flow port pipes flush request");
	DocaFlowPort request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_port_id(port_id);

	AsyncClientCall *port_pipes_flush_call = new AsyncClientCall;
	port_pipes_flush_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPortPipesFlush(
		&port_pipes_flush_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(port_pipes_flush_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete port_pipes_flush_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_port_destroy(uint16_t port_id)
{
	DOCA_LOG_DBG("Doca Flow destroy port request");
	DocaFlowPort request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_port_id(port_id);

	AsyncClientCall *destroy_port_call = new AsyncClientCall;
	destroy_port_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPortDestroy(
		&destroy_port_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(destroy_port_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete destroy_port_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_port_pipes_dump(uint16_t port_id, FILE *f)
{
	DOCA_LOG_DBG("Doca Flow port pipes dump request");
	DocaFlowPort request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_port_id(port_id);

	AsyncClientCall *port_pipes_dump_call = new AsyncClientCall;
	port_pipes_dump_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPortPipesDump(
		&port_pipes_dump_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(port_pipes_dump_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);

	fputs(port_pipes_dump_call->reply.port_pipes_dump().c_str(), f);
	delete port_pipes_dump_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_query(uint64_t entry_id, struct doca_flow_query *query_stats)
{
	DOCA_LOG_DBG("Doca Flow destroy port request");
	DocaFlowQueryRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_entry_id(entry_id);

	AsyncClientCall *query_call = new AsyncClientCall;
	query_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowQuery(
		&query_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(query_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);

	query_stats->total_bytes = query_call->reply.query_stats().total_bytes();
	query_stats->total_pkts = query_call->reply.query_stats().total_pkts();
	delete query_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_aging_handle(uint16_t port_id, uint16_t queue, uint64_t quota, uint64_t *entries_id, int len)
{
	DOCA_LOG_DBG("Doca Flow handle aging request");
	DocaFlowAgingHandleRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_port_id(port_id);
	request.set_queue(queue);
	request.set_quota(quota);
        request.set_len(len);

	AsyncClientCall *handle_aging_call = new AsyncClientCall;
	handle_aging_call->response_reader = client_list[queue].stub_->PrepareAsyncDocaFlowAgingHandle(
		&handle_aging_call->context, request, &client_list[queue].cq_);
	create_rpc_call(handle_aging_call, &client_list[queue].cq_, &response);

	if (response.success) {
                for (int entry_idx = 0; entry_idx < response.aging_res; entry_idx++)
                        entries_id[entry_idx] = handle_aging_call->reply.handle_aging_res().entries(entry_idx);
        }
	delete handle_aging_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_entries_process(uint16_t port_id, uint16_t pipe_queue, uint64_t timeout, uint32_t max_processed_entries)
{
	DOCA_LOG_DBG("Doca Flow entries process request");
	DocaFlowEntriesProcessRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	if (pipe_queue >= nb_queues) {
		DOCA_LOG_ERR("Invalid pipe_queue, must be in range 0-%d", nb_queues - 1);
		response.success = false;
		response.error.message = "Invalid pipe_queue";
		return response;
	}

	request.set_port_id(port_id);
	request.set_pipe_queue(pipe_queue);
	request.set_timeout(timeout);
	request.set_max_processed_entries(max_processed_entries);

	AsyncClientCall *entries_process_call = new AsyncClientCall;
	entries_process_call->response_reader = client_list[pipe_queue].stub_->PrepareAsyncDocaFlowEntriesProcess(
		&entries_process_call->context, request, &client_list[pipe_queue].cq_);
	create_rpc_call(entries_process_call, &client_list[pipe_queue].cq_, &response);
	delete entries_process_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_pipe_entry_get_status(uint64_t entry_id)
{
	DOCA_LOG_DBG("Doca Flow entry get status request");
	DocaFlowPipeEntryGetStatusRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	request.set_entry_id(entry_id);

	AsyncClientCall *entry_get_status_call = new AsyncClientCall;
	entry_get_status_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPipeEntryGetStatus(
		&entry_get_status_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(entry_get_status_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete entry_get_status_call;

	return response;
}

struct doca_flow_grpc_response
doca_flow_grpc_port_switch_get()
{
	DOCA_LOG_DBG("Doca Flow port switch get request");
	DocaFlowEmptyRequest request;
	struct doca_flow_grpc_response response;

	memset(&response, 0, sizeof(response));

	AsyncClientCall *port_switch_get_call = new AsyncClientCall;
	port_switch_get_call->response_reader = client_list[MAIN_CLIENT_IDX].stub_->PrepareAsyncDocaFlowPortSwitchGet(
		&port_switch_get_call->context, request, &client_list[MAIN_CLIENT_IDX].cq_);
	create_rpc_call(port_switch_get_call, &client_list[MAIN_CLIENT_IDX].cq_, &response);
	delete port_switch_get_call;

	return response;
}
