/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_log.h>

#include "flow_grpc_handler.hpp"

#define USER_META_MAX_SIZE 4    /* User meta-data size in 32-bit resolution. */

#ifdef PACKET_BUFFERING_SUPPORT
#define NB_PACKET_PROCESSING_LCORES 1
#endif

DOCA_LOG_REGISTER(doca_flow::GRPC::Handler);

doca_error_t DocaFlowServerParams::InitParams(unsigned int nb_ports, unsigned int  nb_queues)
{
	int lcore, microservice = 0, queue = 0;
        doca_error_t result;
	this->nb_ports = nb_ports;
	this->nb_queues = nb_queues;
	this->ports = (struct doca_flow_port **)malloc(sizeof(struct doca_flow_port *) * nb_ports);
	if (this->ports == NULL)
		return DOCA_ERROR_NO_MEMORY;
	this->ports_pipes_manager = new PipesManager();
	if (this->ports_pipes_manager == NULL) {
		free(this->ports);
		return DOCA_ERROR_NO_MEMORY;
	}
	/* Init the pipes manager */
	result = this->ports_pipes_manager->init();
	if (result != DOCA_SUCCESS) {
		free(this->ports);
		delete this->ports_pipes_manager;
		return result;
	}

        /* init per lcore params */
        memset(lcore_params_arr, 0, sizeof(lcore_params_arr));

	RTE_LCORE_FOREACH(lcore) {
		lcore_params_arr[lcore].used = true;
		lcore_params_arr[lcore].lcore = lcore;
#ifdef PACKET_BUFFERING_SUPPORT
		if (unlikely(lcore == rte_get_main_lcore())) {
			lcore_params_arr[lcore].type = CONTROL;
			continue;
		}
		if(lcore <= NB_PACKET_PROCESSING_LCORES) {
			lcore_params_arr[lcore].type = PACKET_PROCESSING;
			continue;
		}
#endif
		/* At this point, in both modes (w/ or w/o Packet Buffering) the lcoreshould be responsible for RPC handling */
		lcore_params_arr[lcore].type = RPC_HANDLER;
		lcore_params_arr[lcore].microservice = microservice++;
		lcore_params_arr[lcore].queue = queue++;

        }

	return DOCA_SUCCESS;
}

void DocaFlowServerParams::SetPort(struct doca_flow_port *port)
{
		if(next_port_id >= nb_queues)
                        return;
                this->ports[next_port_id] = port;
                next_port_id++;
}

PipesManager * DocaFlowServerParams::GetPipesManager()
{
        return ports_pipes_manager;
}

unsigned int DocaFlowServerParams::GetNumberOfPorts() const
{
        return nb_ports;
}

unsigned int DocaFlowServerParams::GetNumberOfQueues() const
{
        return nb_queues;
}

struct doca_flow_port *  DocaFlowServerParams::GetPort(uint64_t port_id) const
{
        return ports[port_id];
}

struct server_per_lcore_params * DocaFlowServerParams::GetServerPerLcoreParams()
{
        return lcore_params_arr;
}

void DocaFlowServerParams::Destroy()
{
        if (ports != NULL)
		free(ports);

	ports = NULL;
        delete ports_pipes_manager;
}

void DocaFlowRpcHandler::fillMacAddress(uint64_t mac_address_req, uint8_t *mac_address)
{

        int byte_idx;
        uint8_t *byte_ptr = (uint8_t*) &mac_address_req;

        /* we need only the first 6 bytes for the mac address */
        for (byte_idx = 0; byte_idx < DOCA_ETHER_ADDR_LEN; byte_idx++)
                mac_address[byte_idx] = byte_ptr[byte_idx];
}

void DocaFlowRpcHandler::fillIpAddress(DocaFlowIPAddress *ip_add_req, struct doca_flow_ip_addr *ip_add)
{
        int iter;
        memset(ip_add, 0,sizeof(*ip_add));
        ip_add->type =(enum doca_flow_ip_type) ip_add_req->type();

        if(ip_add->type == DOCA_FLOW_IP4_ADDR)
                ip_add->ipv4_addr = ip_add_req->ipv4_addr();
        if(ip_add->type == DOCA_FLOW_IP6_ADDR){
                for(iter = 0; iter< ip_add_req->ipv6_addr_size(); iter++)
                        ip_add->ipv6_addr[iter] = ip_add_req->ipv6_addr(iter);
        }
}

void DocaFlowRpcHandler::fillTun(DocaFlowTun *tun_req, struct doca_flow_tun *tun)
{
        memset(tun, 0,sizeof(*tun));
        tun->type =  (enum doca_flow_tun_type) tun_req->type();
        switch (tun_req->tun_oneof_case())
        {
        case DocaFlowTun::TunOneofCase::kVxlanTunId:
                tun->vxlan_tun_id = tun_req->vxlan_tun_id();
                break;
        case DocaFlowTun::TunOneofCase::kGreKey:
                tun->gre_key = tun_req->gre_key();
                break;
        case DocaFlowTun::TunOneofCase::kGtpTeid:
                tun->gtp_teid = tun_req->gtp_teid();
                break;
        default:
                break;
        }
}

void DocaFlowRpcHandler::fillMeta(DocaFlowMeta *meta_req, struct doca_flow_meta *meta)
{
        int u32_iter;
        meta->pkt_meta = meta_req->pkt_meta();

        for (u32_iter = 0; u32_iter < USER_META_MAX_SIZE; u32_iter++)
        {
               meta->u32[u32_iter] = meta_req->u32(u32_iter);
        }

        meta->port_meta = meta_req->port_meta();
        meta->mark = meta_req->mark();
}

void DocaFlowRpcHandler::fillMatch(DocaFlowMatch* match_req, struct doca_flow_match *match)
{
        memset(match, 0, sizeof(*match));

        match->flags = match_req->flags();

	if (match_req->has_meta())
            fillMeta(match_req->mutable_meta(), &(match->meta));

        fillMacAddress(match_req->out_src_mac(), match->out_src_mac);
        fillMacAddress(match_req->out_dst_mac(), match->out_dst_mac);

        match->out_eth_type = match_req->out_eth_type();

        match->out_vlan_tci = match_req->out_vlan_tci();

        fillIpAddress(match_req->mutable_out_src_ip(), &(match->out_src_ip));
        fillIpAddress(match_req->mutable_out_dst_ip(), &(match->out_dst_ip));

        match->out_l4_type = match_req->out_l4_type();
        match->out_tcp_flags = match_req->out_tcp_flags();
        match->out_src_port = match_req->out_src_port();
        match->out_dst_port = match_req->out_dst_port();

        fillTun(match_req->mutable_tun(), &(match->tun));

        fillMacAddress(match_req->in_src_mac(), match->in_src_mac);
        fillMacAddress(match_req->in_dst_mac(), match->in_dst_mac);

        match->in_eth_type = match_req->in_eth_type();
        match->in_vlan_tci = match_req->in_vlan_tci();

        fillIpAddress(match_req->mutable_in_src_ip(), &(match->in_src_ip));
        fillIpAddress(match_req->mutable_in_dst_ip(), &(match->in_dst_ip));

        match->in_l4_type = match_req->in_l4_type();
        match->in_tcp_flags = match_req->in_tcp_flags();
        match->in_src_port = match_req->in_src_port();
        match->in_dst_port = match_req->in_dst_port();
}

void DocaFlowRpcHandler::fillActionField(DocaFlowActionField *action_field_req, doca_flow_action_field *action_field, void *match_address)
{
        uint8_t address_offset = action_field_req->address_offset();
        action_field->address = (uint8_t *)match_address + address_offset;
        action_field->offset  = action_field_req->offset();
}

void DocaFlowRpcHandler::fillEncapAction(DocaFlowEncapAction *encap_action_req, struct doca_flow_encap_action *encap_action)
{
        fillMacAddress(encap_action_req->src_mac(), encap_action->src_mac);
        fillMacAddress(encap_action_req->dst_mac(), encap_action->dst_mac);

        encap_action->vlan_tci = encap_action_req->vlan_tci();

        fillIpAddress(encap_action_req->mutable_src_ip(), &(encap_action->src_ip));
        fillIpAddress(encap_action_req->mutable_dst_ip(), &(encap_action->dst_ip));

        fillTun(encap_action_req->mutable_tun(), &(encap_action->tun));
}

void DocaFlowRpcHandler::fillActions(DocaFlowActions *actions_req, struct doca_flow_actions *actions)
{
        memset(actions, 0, sizeof(*actions));

        uint8_t action_idx = actions_req->action_idx();

        actions->action_idx = action_idx;
        actions->flags = actions_req->flags();
        actions->decap = actions_req->decap();

        fillMacAddress(actions_req->mod_src_mac(), actions->mod_src_mac);
        fillMacAddress(actions_req->mod_dst_mac(), actions->mod_dst_mac);

        fillIpAddress(actions_req->mutable_mod_src_ip(), &(actions->mod_src_ip));
        fillIpAddress(actions_req->mutable_mod_dst_ip(), &(actions->mod_dst_ip));

        actions->mod_src_port = actions_req->mod_src_port();
        actions->mod_dst_port = actions_req->mod_dst_port();

        actions->ttl = actions_req->ttl();

        actions->has_encap = actions_req->encap_exist();
        fillEncapAction(actions_req->mutable_encap(), &(actions->encap));
        fillMeta(actions_req->mutable_meta(), &(actions->meta));
}

void DocaFlowRpcHandler::fillActionCopy(DocaFlowActionCopy *copy, struct doca_flow_action_desc *action_desc, void *match_address)
{

        if(copy->has_src())
                fillActionField(copy->mutable_src(), &(action_desc->copy.src), match_address);

        if(copy->has_dst())
                fillActionField(copy->mutable_dst(), &(action_desc->copy.dst), match_address);

        action_desc->copy.width = copy->width();
}

void DocaFlowRpcHandler::fillActionMask(DocaFlowActionMask *mask, struct doca_flow_action_desc *action_desc, void *match_address)
{
        for(int u8_iter = 0; u8_iter < 16; u8_iter++)
                action_desc->mask.u8[u8_iter] = mask->u8(u8_iter);
}

void DocaFlowRpcHandler::fillActionDesc(DocaFlowActionDesc *action_desc_req, struct doca_flow_action_desc *action_desc, void *match_address)
{
        action_desc->type = (doca_flow_action_type) action_desc_req->type();

        switch(action_desc_req->action_types_case())
        {
        case DocaFlowActionDesc::ActionTypesCase::kMask:
                fillActionMask(action_desc_req->mutable_mask(), action_desc, match_address);
                break;
        case DocaFlowActionDesc::ActionTypesCase::kCopy:
                fillActionCopy(action_desc_req->mutable_copy(), action_desc, match_address);
                break;
        default:
                break;
        }
}

void DocaFlowRpcHandler::fillActionDescsMeta(DocaFlowActionDescsMeta *action_descs_meta_req, struct doca_flow_action_descs_meta *action_descs_meta, void *match_address)
{

        memset(action_descs_meta, 0, sizeof(*action_descs_meta));

        if(action_descs_meta_req->has_pkt_meta())
                fillActionDesc(action_descs_meta_req->mutable_pkt_meta(), &action_descs_meta->pkt_meta, match_address);

        for(int u32_iter = 0; u32_iter < USER_META_MAX_SIZE; u32_iter++)
                fillActionDesc(action_descs_meta_req->mutable_u32(u32_iter), &action_descs_meta->u32[u32_iter], match_address);

}

void DocaFlowRpcHandler::fillActionDescs(DocaFlowActionDescs *action_descs_req, struct doca_flow_action_descs *action_descs, void *match_address)
{

        memset(action_descs, 0, sizeof(*action_descs));

        if(action_descs_req->has_src_mac())
                fillActionDesc(action_descs_req->mutable_src_mac(), &action_descs->src_mac, match_address);

        if(action_descs_req->has_dst_mac())
                fillActionDesc(action_descs_req->mutable_dst_mac(), &action_descs->dst_mac, match_address);

        if(action_descs_req->has_eth_type())
                fillActionDesc(action_descs_req->mutable_eth_type(), &action_descs->eth_type, match_address);

        if(action_descs_req->has_vlan())
                fillActionDesc(action_descs_req->mutable_vlan(), &action_descs->vlan_id, match_address);

        if(action_descs_req->has_src_ip())
                fillActionDesc(action_descs_req->mutable_src_ip(), &action_descs->src_ip, match_address);

        if(action_descs_req->has_dst_ip())
                fillActionDesc(action_descs_req->mutable_dst_ip(), &action_descs->dst_ip, match_address);

        if(action_descs_req->has_ttl())
                fillActionDesc(action_descs_req->mutable_ttl(), &action_descs->ttl, match_address);

        if(action_descs_req->has_src_port())
                fillActionDesc(action_descs_req->mutable_src_port(), &action_descs->src_port, match_address);

        if(action_descs_req->has_dst_port())
                fillActionDesc(action_descs_req->mutable_dst_port(), &action_descs->dst_port, match_address);

        if(action_descs_req->has_meta())
                fillActionDescsMeta(action_descs_req->mutable_meta(), &action_descs->meta, match_address);

        if(action_descs_req->has_tunnel())
                fillActionDesc(action_descs_req->mutable_tunnel(), &action_descs->tunnel, match_address);
}

void DocaFlowRpcHandler::fillMonitor(const DocaFlowMonitor *monitor_req, struct doca_flow_monitor *monitor)
{
        memset(monitor, 0, sizeof(*monitor));
        monitor->flags = monitor_req->flags();
        monitor->cir = monitor_req->cir();
        monitor->cbs = monitor_req->cbs();
        monitor->shared_meter_id = monitor_req->shared_meter_id();
        monitor->shared_counter_id = monitor_req->shared_counter_id();
        monitor->aging = monitor_req->aging();
        monitor->user_data = monitor_req->user_data();
}

void DocaFlowRpcHandler::fillFwd(const DocaFlowFwd *fwd_req, struct doca_flow_fwd *fwd, uint16_t **rss_queues_ptr)
{
        int q_iter;
        uint16_t *rss_queues = *rss_queues_ptr;
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        memset(fwd, 0, sizeof(*fwd));
        /* filling fwd component */
        fwd->type = (doca_flow_fwd_type)fwd_req->type();
        switch(fwd->type)
        {
        case DOCA_FLOW_FWD_RSS:
                fwd->rss_flags = fwd_req->rss_flags();
                fwd->num_of_queues = fwd_req->num_of_queues();
                for (q_iter = 0; q_iter<fwd_req->rss_queues_size(); q_iter++)
                        rss_queues[q_iter]=fwd_req->rss_queues(q_iter);
                fwd->rss_queues = rss_queues;
                break;
        case DOCA_FLOW_FWD_PORT:
                fwd->port_id = fwd_req->port_id();
                break;
        case DOCA_FLOW_FWD_PIPE:
                fwd->next_pipe = ports_pipes_manager->getPipe(fwd_req->next_pipe_id());
                break;
        case DOCA_FLOW_FWD_DROP:
                break;
        default:
                DOCA_DLOG_INFO("Wrong doca_flow_fwd configuration - next pipe is not found");
        }
}

void DocaFlowRpcHandler::fillErrorResponse(DocaFlowResponse *response, struct doca_flow_error* error)
{
        DocaFlowError *err_response = response->mutable_error();
        err_response->set_type((DocaFlowErrorType)error->type);
        if(error->message)
                err_response->set_err_message(std::string((char *)error->message));
        response->set_success(false);
}

void DocaFlowRpcHandler::fillPipeAttr(const DocaFlowPipeAttr *attr_req, struct doca_flow_pipe_attr *attr)
{
        /* filling pipe's attributes */
        attr->name = attr_req->name().c_str();
        attr->type = (enum doca_flow_pipe_type)attr_req->type();
        attr->is_root = attr_req->is_root();
        attr->nb_flows = attr_req->nb_flows();
        attr->nb_actions = attr_req->nb_actions();
}

int InitRpcHandler::DocaFlowInit(const DocaFlowCfg * request, DocaFlowResponse *response)
{
        struct doca_flow_cfg cfg = {0};
        struct doca_flow_error error;
        std::string mode_str = request->mode_args();
        char mode[mode_str.length() + 1];
        int res;

        /* initialize doca framework */
        cfg.queues = request->queues();
        cfg.resource.nb_counters = (request->resource()).nb_counters();
        cfg.resource.nb_meters = (request->resource()).nb_meters();
        std::copy(mode_str.begin(), mode_str.end(), mode);
        cfg.mode_args = mode;
        cfg.queue_depth = request->queue_depth();
        cfg.mode_args = (request->mode_args()).c_str();

        for(int i = 0; i < request->nr_shared_resources_size(); i++)
        {
                cfg.nr_shared_resources[i] = request->nr_shared_resources(i);
        }

        /* should be called after decoding the request appropriately */
        res = doca_flow_init(&cfg, &error);
        if (res < 0) {
                /* doca flow framework initialization failed */
                fillErrorResponse(response,&error);
                return res;
        }

        /* doca flow framework initialization succeeded */
        response->set_success(true);
        return 0;
}

int DestroyRpcHandler::DocaFlowDestroy(const DocaFlowEmptyRequest * request, DocaFlowResponse *response)
{
        server_params_->Destroy();
        doca_flow_destroy();
        response->set_success(true);
        return 0;
}

int PortStartRpcHandler::DocaFlowPortStart(const DocaFlowPortCfg * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        struct doca_flow_port_cfg cfg;
        struct doca_flow_error error;
        struct doca_flow_port* port;
        uint64_t port_id = request->port_id();

        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        cfg.port_id = port_id;
        cfg.type = (enum doca_flow_port_type) request->type();
        cfg.devargs = (request->devargs()).c_str();
        cfg.priv_data_size = request->priv_data_size();

        port = doca_flow_port_start(&cfg, &error);
        if (port == NULL){
                fillErrorResponse(response,&error);
                return -1;
        }

        server_params_->SetPort(port);

        /* doca flow starting ports succeeded */
        response->set_success(true);
        response->set_port_id(port_id);
        return 0;
}

int PortStopRpcHandler::DocaFlowPortStop(const DocaFlowPort * request, DocaFlowResponse *response)
{
        int res;
        DocaFlowError *err_response = response->mutable_error();
        uint64_t port_id = request->port_id();

        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        res = doca_flow_port_stop(server_params_->GetPort(port_id));
        if (res < 0){
                err_response->set_err_message(std::string("doca flow port stop failed"));
                response->set_success(false);
                return -1;
        }
        response->set_success(true);
        return 0;
}

int PortPairRpcHandler::DocaFlowPortPair(const DocaFlowPortPairRequest *request, DocaFlowResponse *response)
{
        int res;
        DocaFlowError *err_response = response->mutable_error();
        uint64_t port_id = request->port_id();
        uint pair_port_id = request->pair_port_id();

        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        /* check if pair port id is legal */
        if (pair_port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("pair port id is illegal"));
                response->set_success(false);
                return -1;
        }

        res = doca_flow_port_pair(server_params_->GetPort(port_id), server_params_->GetPort(pair_port_id));
        if (res < 0){
                err_response->set_err_message(std::string("doca flow pair port failed"));
                response->set_success(false);
                return res;
        }
        response->set_success(true);
        return 0;
}

int PortPrivDataRpcHandler::DocaFlowPortPrivData( const DocaFlowPort * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        uint64_t port_id = request->port_id();
        void* priv_data;
        /* check if port id is legal */
        if (port_id >=server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        priv_data = (void*)doca_flow_port_priv_data(server_params_->GetPort(port_id));
        std::string priv_data_str = std::string((char *)priv_data);
        response->set_priv_data(priv_data_str);
        response->set_success(true);
        return 0;
}

int SharedResourceCfgRpcHandler::DocaFlowSharedResourceCfg( const DocaFlowSharedResourceCfgRequest * request, DocaFlowResponse *response)
{
        enum doca_flow_shared_resource_type type = (enum doca_flow_shared_resource_type)request->type();
        uint32_t id = request->id();
        SharedResourceCfg req_cfg = request->cfg();
        DocaFlowResourceMeterCfg meter_cfg = req_cfg.meter_cfg();
        struct doca_flow_shared_resource_cfg cfg = {0};
        struct doca_flow_error error;
        int res = 0;

        cfg.meter_cfg.cir =  meter_cfg.cir();
        cfg.meter_cfg.cbs =  meter_cfg.cbs();

        res = doca_flow_shared_resource_cfg(type, id, &cfg, &error);
        if (res < 0){
                fillErrorResponse(response,&error);
                return -1;
        }

        response->set_success(true);
        return 0;
}

int SharedResourcesBindRpcHandler::DocaFlowSharedResourcesBind( const DocaFlowSharedResourcesBindRequest * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        enum doca_flow_shared_resource_type type = (enum doca_flow_shared_resource_type)request->type();
        uint32_t res_array_len = request->resource_arr_size();
        uint32_t res_array[res_array_len];
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();
        int res = 0;
        uint64_t port_id;
        uint64_t pipe_id;
        struct doca_flow_error error;
        void *bindable_obj = NULL;

        for(uint32_t i = 0; i < res_array_len; i++)
                res_array[i] = request->resource_arr(i);

        switch (request->bindable_obj_id_case())
        {
        case DocaFlowSharedResourcesBindRequest::BindableObjIdCase::kPortId:
                port_id = request->port_id();
                if (port_id >= server_params_->GetNumberOfPorts()){
                        err_response->set_err_message(std::string("port id is illegal"));
                        response->set_success(false);
                        return -1;
                }
                bindable_obj = server_params_->GetPort(port_id);
                break;
        case DocaFlowSharedResourcesBindRequest::BindableObjIdCase::kPipeId:
                pipe_id = request->pipe_id();
                bindable_obj = ports_pipes_manager->getPipe(pipe_id);
                break;
        default:
                break;
        }

        res = doca_flow_shared_resources_bind(type, res_array, res_array_len, bindable_obj, &error);
        if (res < 0){
                fillErrorResponse(response,&error);
                return -1;
        }

        response->set_success(true);
        return 0;
}

int SharedResourcesQueryRpcHandler::DocaFlowSharedResourcesQuery( const DocaFlowSharedResourcesQueryRequest * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        enum doca_flow_shared_resource_type type = (enum doca_flow_shared_resource_type)request->type();
        uint32_t res_array_len = request->res_array_size();
        uint32_t res_array[res_array_len];
        struct doca_flow_shared_resource_result query_results_array[res_array_len];
        int res = 0;
        struct doca_flow_error error;
        uint32_t i = 0;
        DocaFlowQueryStats *curr_query_res;

        if(res_array_len == 0) {
                err_response->set_err_message(std::string("Array length is zero"));
                response->set_success(false);
                return -1;
        }

        for(i = 0; i < res_array_len; i++)
                res_array[i] = request->res_array(i);

        res = doca_flow_shared_resources_query(type, res_array, query_results_array, res_array_len, &error);
        if (res < 0){
                fillErrorResponse(response, &error);
                return -1;
        }

        for(i = 0; i < res_array_len; i++)
        {
                curr_query_res = response->mutable_query_results(i);
                curr_query_res->set_total_bytes(query_results_array[i].counter.total_bytes);
                curr_query_res->set_total_pkts(query_results_array[i].counter.total_pkts);
        }

        response->set_success(true);
        return 0;
}


int PipeCreateRpcHandler::DocaFlowPipeCreate(const DocaFlowPipeCreateRequest * request, DocaFlowResponse *response)
{
        int iter = 0;
        int action_descs_size = 0;
        int actions_size = 0;
        struct doca_flow_pipe_cfg cfg = {0};
        struct doca_flow_match match;
	struct doca_flow_match match_mask;
        struct doca_flow_actions **actions = NULL;
	struct doca_flow_action_descs **action_descs = NULL;
        struct doca_flow_monitor monitor;
        struct doca_flow_fwd fwd;
        struct doca_flow_fwd fwd_miss;
        struct doca_flow_fwd *fwd_ptr = NULL;
        struct doca_flow_fwd *fwd_miss_ptr = NULL;
        struct doca_flow_error error;
        struct doca_flow_pipe *pipe;
        DocaFlowError *err_response = response->mutable_error();
        DocaFlowPipeCfg cfg_req = request->cfg();
        DocaFlowPipeAttr *attr_req;
        DocaFlowMatch match_req;
        DocaFlowMatch match_mask_req;
        DocaFlowActions actions_req;
        DocaFlowActionDescs actions_descs_req;
        DocaFlowMonitor monitor_req;
        DocaFlowFwd fwd_req = request->fwd();
        uint16_t rss_queues[fwd_req.rss_queues_size()];
	uint64_t pipe_id_res;
        uint16_t *rss_queues_aux = rss_queues;
        uint16_t **rss_queues_ptr = &rss_queues_aux;
        DocaFlowFwd fwd_miss_req = request->fwd_miss();
        memset(&cfg, 0, sizeof(cfg));
        uint  port_id = cfg_req.port_id();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

         /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        if(cfg_req.has_attr()){
                attr_req = cfg_req.mutable_attr();
                fillPipeAttr(attr_req, &cfg.attr);
        }

        cfg.port = server_params_->GetPort(port_id);

         if (cfg_req.has_match()){
                match_req = cfg_req.match();
                fillMatch(&match_req, &match);
                cfg.match = &match;
        }

        if (cfg_req.has_match_mask()){
                match_mask_req =cfg_req.match_mask();
                fillMatch(&match_mask_req, &match_mask);
                cfg.match_mask = &match_mask;
        }

        if (cfg.attr.nb_actions != 0) {
                actions = ( struct doca_flow_actions **) malloc(sizeof(*actions)*cfg.attr.nb_actions);
                if(actions == NULL) {
                        err_response->set_err_message(std::string("Out of memory"));
                        response->set_success(false);
                        return -1;
                }
                actions_size =  cfg_req.actions_size();
                for(iter = 0; iter < actions_size; iter++)
                {
                        actions[iter] = (struct doca_flow_actions *) malloc(sizeof(*actions[iter]));
                        fillActions(cfg_req.mutable_actions(iter), actions[iter]);
                }
                cfg.actions = actions;

                action_descs_size = cfg_req.action_descs_size();
                if(action_descs_size != 0) {
                        action_descs = ( struct doca_flow_action_descs **) malloc(sizeof(*action_descs)*cfg.attr.nb_actions);
                        if(action_descs == NULL) {
                                err_response->set_err_message(std::string("Out of memory"));
                                response->set_success(false);
                                return -1;
                        }

                        for(iter = 0; iter < action_descs_size; iter++)
                        {               action_descs[iter] = (struct doca_flow_action_descs *) malloc(sizeof(*action_descs[iter]));
                                        fillActionDescs(cfg_req.mutable_action_descs(iter), action_descs[iter], cfg.match);
                        }
                        cfg.action_descs = action_descs;
                }
        }

        if (cfg_req.has_monitor()){
                monitor_req = cfg_req.monitor();
                fillMonitor(&monitor_req, &monitor);
                cfg.monitor = &monitor;
        }

        /* filling fwd component of the pipe */
        if (request->has_fwd()){
                fillFwd(&fwd_req, &fwd, rss_queues_ptr);
                fwd_ptr = &fwd;
        }

        /* filling fwd miss component of the pipe */
        if (request->has_fwd_miss()){
                fillFwd(&fwd_miss_req, &fwd_miss, rss_queues_ptr);
                fwd_miss_ptr = &fwd_miss;
        }

        pipe = doca_flow_pipe_create(&cfg, fwd_ptr, fwd_miss_ptr, &error);
        if (pipe == NULL){
                fillErrorResponse(response,&error);
                return -1;
        }

        /* check if pipe id is not 0 which indicates a failure  */
	pipe_id_res = ports_pipes_manager->addPipe(pipe, port_id);
	if (pipe_id_res == 0) {
		doca_flow_pipe_destroy(pipe);
		response->set_success(false);
		err_response->set_err_message(std::string("Pipe id was not found"));
		return -1;

	}

        if(actions)
        {
                for(iter = 0; iter < actions_size; iter++)
                {
                        if(actions[iter])
                                free(actions[iter]);
                }
                free(actions);
        }

        if(action_descs)
        {
                for(iter = 0; iter < action_descs_size; iter++)
                {
                        if(action_descs[iter])
                                free(action_descs[iter]);
                }
                        free(action_descs);
        }

        response->set_success(true);
	response->set_port_id(port_id);
	response->set_pipe_id(pipe_id_res);

        return 0;
}

int PipeAddEntryRpcHandler::DocaFlowPipeAddEntry(const DocaFlowPipeAddEntryRequest * request, DocaFlowResponse *response)
{
        struct doca_flow_match match;
        struct doca_flow_match *match_ptr = NULL;
        struct doca_flow_actions actions;
        struct doca_flow_actions *actions_ptr = NULL;
        struct doca_flow_monitor monitor;
        struct doca_flow_monitor *monitor_ptr = NULL;
        struct doca_flow_fwd fwd;
        struct doca_flow_fwd *fwd_ptr = NULL;
        struct doca_flow_error error;
        struct doca_flow_pipe *pipe;
        struct doca_flow_pipe_entry *entry;

        uint16_t pipe_queue = request->pipe_queue();
        uint64_t pipe_id = request->pipe_id();
	uint64_t entry_id_res;
        DocaFlowMatch match_req;
        DocaFlowActions actions_req;
        DocaFlowMonitor monitor_req;
        DocaFlowFwd fwd_req;
        DocaFlowError *err_response = response->mutable_error();
        uint16_t rss_queues[fwd_req.rss_queues_size()];
        uint16_t *rss_queues_aux = rss_queues;
        uint16_t **rss_queues_ptr = &rss_queues_aux;
        uint32_t flags = request->flags();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        pipe = ports_pipes_manager->getPipe(pipe_id);
        if (pipe == NULL){
                err_response->set_err_message(std::string("no pipe found matching the given pipe id"));
                response->set_success(false);
                return -1;
        }

        if (request->has_match()){
                match_req = request->match();
                fillMatch(&match_req, &match);
                match_ptr = &match;
        }

        if (request->has_actions()){
                actions_req = request->actions();
                fillActions(&actions_req, &actions);
                actions_ptr = &actions;
        }

        if (request->has_monitor()){
                monitor_req = request->monitor();
                fillMonitor(&monitor_req, &monitor);
                monitor_ptr = &monitor;
        }

        if (request->has_fwd()){
                fwd_req = request->fwd();
                fillFwd(&fwd_req, &fwd, rss_queues_ptr);
                fwd_ptr = &fwd;
        }

        entry = doca_flow_pipe_add_entry(pipe_queue, pipe, match_ptr, actions_ptr,
                        monitor_ptr, fwd_ptr, flags, NULL, &error);
        if (entry == NULL){
                DOCA_LOG_DBG("Entry Insertion Failed.");
                fillErrorResponse(response,&error);
                return -1;
        }

        /* check if entry id is not 0 which indicates a failure */
	entry_id_res = ports_pipes_manager->addPipeEntry(pipe, entry);
	if (entry_id_res == 0) {
		doca_flow_pipe_rm_entry(pipe_queue, NULL, entry);
		response->set_success(false);
		err_response->set_err_message(std::string("Entry id was not found"));
		return -1;
	}

	response->set_success(true);
	response->set_entry_id(entry_id_res);

        return 0;
}

int PipeControlAddEntryRpcHandler::DocaFlowPipeControlAddEntry(const DocaFlowPipeControlAddEntryRequest * request, DocaFlowResponse *response)
{
        struct doca_flow_match match;
        struct doca_flow_match *match_ptr = NULL;
        struct doca_flow_match match_mask;
        struct doca_flow_match *match_mask_ptr = NULL;
        struct doca_flow_fwd fwd;
        struct doca_flow_fwd *fwd_ptr = NULL;
        struct doca_flow_error error;
        struct doca_flow_pipe *pipe;
        struct doca_flow_pipe_entry *entry;
        DocaFlowMatch match_req;
        DocaFlowMatch match_mask_req;
        DocaFlowMonitor monitor_req;
        DocaFlowFwd fwd_req;
        DocaFlowError *err_response = response->mutable_error();
        uint8_t priority = request->priority();
        uint16_t pipe_queue = request->pipe_queue();
        uint64_t pipe_id = request->pipe_id();
	uint64_t entry_id_res;
        uint16_t rss_queues[fwd_req.rss_queues_size()];
        uint16_t *rss_queues_aux = rss_queues;
        uint16_t **rss_queues_ptr = &rss_queues_aux;
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        pipe = ports_pipes_manager->getPipe(pipe_id);
        if (pipe == NULL){
                err_response->set_err_message(std::string("no filter pipe found matching the given pipe id"));
                response->set_success(false);
                return -1;
        }

        if (request->has_match()){
                match_req = request->match();
                fillMatch(&match_req, &match);
                match_ptr = &match;
        }

        if (request->has_match_mask()){
                match_mask_req = request->match_mask();
                fillMatch(&match_mask_req, &match_mask);
                match_mask_ptr = &match_mask;
        }
        if (request->has_fwd()){
                fwd_req = request->fwd();
                fillFwd(&fwd_req, &fwd, rss_queues_ptr);
                fwd_ptr = &fwd;
        }

        entry = doca_flow_pipe_control_add_entry(pipe_queue, priority, pipe, match_ptr,
                match_mask_ptr, NULL, NULL, NULL, fwd_ptr, &error);
        if (entry == NULL){
                fillErrorResponse(response,&error);
                return -1;
        }

	/* check if entry id is not 0 which indicates a failure */
	entry_id_res = ports_pipes_manager->addPipeEntry(pipe, entry);
	if (entry_id_res == 0) {
		doca_flow_pipe_rm_entry(pipe_queue, NULL, entry);
		response->set_success(false);
		err_response->set_err_message(std::string("Control pipe entry id was not found"));
		return -1;

	}

	response->set_success(true);
	response->set_entry_id(entry_id_res);

        return 0;
}

int PipeLpmAddEntryRpcHandler::DocaFlowPipeLpmAddEntry(const DocaFlowPipeLpmAddEntryRequest * request, DocaFlowResponse *response)
{
        struct doca_flow_match match;
        struct doca_flow_match *match_ptr = NULL;
        struct doca_flow_match match_mask;
        struct doca_flow_match *match_mask_ptr = NULL;
        struct doca_flow_actions actions;
        struct doca_flow_actions *actions_ptr = NULL;
        struct doca_flow_monitor monitor;
        struct doca_flow_monitor *monitor_ptr = NULL;
        struct doca_flow_fwd fwd;
        struct doca_flow_fwd *fwd_ptr = NULL;
        struct doca_flow_error error;
        struct doca_flow_pipe *pipe;
        struct doca_flow_pipe_entry *entry;

        uint16_t pipe_queue = request->pipe_queue();
        uint64_t pipe_id = request->pipe_id();
	uint64_t entry_id_res;
        DocaFlowMatch match_req;
        DocaFlowMatch match_mask_req;
        DocaFlowActions actions_req;
        DocaFlowMonitor monitor_req;
        DocaFlowFwd fwd_req;
        DocaFlowError *err_response = response->mutable_error();
        uint16_t rss_queues[fwd_req.rss_queues_size()];
        uint16_t *rss_queues_aux = rss_queues;
        uint16_t **rss_queues_ptr = &rss_queues_aux;
        enum doca_flow_flags_type flag = (enum doca_flow_flags_type) request->flag();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        pipe = ports_pipes_manager->getPipe(pipe_id);
        if (pipe == NULL){
                err_response->set_err_message(std::string("no pipe found matching the given pipe id"));
                response->set_success(false);
                return -1;
        }

        if (request->has_match()){
                match_req = request->match();
                fillMatch(&match_req, &match);
                match_ptr = &match;
        }

        if (request->has_match_mask()){
                match_mask_req = request->match_mask();
                fillMatch(&match_mask_req, &match_mask);
                match_mask_ptr = &match_mask;
        }

        if (request->has_actions()){
                actions_req = request->actions();
                fillActions(&actions_req, &actions);
                actions_ptr = &actions;
        }

        if (request->has_monitor()){
                monitor_req = request->monitor();
                fillMonitor(&monitor_req, &monitor);
                monitor_ptr = &monitor;
        }

        if (request->has_fwd()){
                fwd_req = request->fwd();
                fillFwd(&fwd_req, &fwd, rss_queues_ptr);
                fwd_ptr = &fwd;
        }

        entry = doca_flow_pipe_lpm_add_entry(pipe_queue, pipe, match_ptr, match_mask_ptr, actions_ptr,
                        monitor_ptr, fwd_ptr, flag, NULL, &error);
        if (entry == NULL){
                DOCA_LOG_DBG("Entry Insertion Failed.");
                fillErrorResponse(response,&error);
                return -1;
        }

	/* check if entry id is not 0 which indicates a failure */
	entry_id_res = ports_pipes_manager->addPipeEntry(pipe, entry);
	if (entry_id_res == 0) {
		doca_flow_pipe_rm_entry(pipe_queue, NULL, entry);
		response->set_success(false);
		err_response->set_err_message(std::string("Control pipe entry id was not found"));
		return -1;

	}

	response->set_success(true);
	response->set_entry_id(entry_id_res);

        return 0;
}

int PipeRmEntryRpcHandler::DocaFlowPipeRmEntry(const DocaFlowPipeRmEntryRequest * request, DocaFlowResponse *response)
{
        int res;
        struct doca_flow_pipe_entry *entry;
        DocaFlowError *err_response = response->mutable_error();

        /* get remove entry request info */
        uint16_t pipe_queue = request->pipe_queue();
        uint64_t entry_id = request->entry_id();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        /* check if the entry is legal */
        entry = ports_pipes_manager->getPipeEntry(entry_id);
        if (entry == NULL){
                err_response->set_err_message(std::string("no entry found matching the given entry id"));
                response->set_success(false);
                return -1;
        }

	if (!ports_pipes_manager->removePipeEntry(entry_id)) {
		err_response->set_err_message(std::string("no pipe found matching the given entry id"));
		response->set_success(false);
		return -1;
	}

        /* if the returned value is negative then the given pipe queue of the entry is wrong */
        res = doca_flow_pipe_rm_entry(pipe_queue, NULL, entry);
        if (res < 0){
                err_response->set_err_message(std::string("no entry found matching the given pipe queue"));
                response->set_success(false);
                return -1;
        }

        response->set_success(true);
        return 0;
}

int PipeDestroyRpcHandler::DocaFlowPipeDestroy(const DocaFlowPipeDestroyRequest * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        uint64_t pipe_id = request->pipe_id();
        struct doca_flow_pipe * pipe;
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        /* check if the pipe id is legal */
        pipe = ports_pipes_manager->getPipe(pipe_id);
        if (pipe == NULL){
                err_response->set_err_message(std::string("no pipe found matching the given pipe id"));
                response->set_success(false);
                return -1;
        }

        /* remove pipe and its all entries from port's pipes manager */
        if (!ports_pipes_manager->removePipe(pipe_id)) {
		err_response->set_err_message(std::string("failed to destroy pipe"));
		response->set_success(false);
		return -1;
	}
        /* destroying the pipe */
        doca_flow_pipe_destroy(pipe);
        response->set_success(true);
        return 0;
}

int PortPipesFlushRpcHandler::DocaFlowPortPipesFlush(const DocaFlowPort * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        uint64_t port_id = request->port_id();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();
        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        /* removing all stored pipes from port's pipes manager */
        if (!ports_pipes_manager->portPipesFlush(port_id)) {
		err_response->set_err_message(std::string("failed to flush pipe for"));
		response->set_success(false);
		return -1;
	}

        /* flushing the actual pipe */
        doca_flow_port_pipes_flush(server_params_->GetPort(port_id));
        response->set_success(true);
        return 0;
}

int PortDestroyRpcHandler::DocaFlowPortDestroy(const DocaFlowPort * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        uint64_t port_id = request->port_id();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        /* removing all stored pipes from port's pipes manager */
        ports_pipes_manager->portPipesFlush(port_id);
        /* destroying port */
        doca_flow_port_destroy(server_params_->GetPort(port_id));
        response->set_success(true);
        return 0;
}

int PortPipesDumpRpcHandler::DocaFlowPortPipesDump(const DocaFlowPort * request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        std::string port_pipes_dump;
        uint64_t port_id = request->port_id();
        FILE* port_pipes_dump_file;
        const char* port_pipes_dump_file_name = "port_pipes_dump_log.txt";
        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        port_pipes_dump_file = fopen(port_pipes_dump_file_name,"w");
        if (port_pipes_dump_file == NULL){
                err_response->set_err_message(std::string("doca flow server: cannot create and open log file"));
                response->set_success(false);
                return -1;
        }

        /* The actual call of doca flow API for dumping all pipes information for the given port */
        doca_flow_port_pipes_dump(server_params_->GetPort(port_id) ,port_pipes_dump_file);
        fclose(port_pipes_dump_file);
        std::ifstream ifs(port_pipes_dump_file_name);
        port_pipes_dump.assign( (std::istreambuf_iterator<char>(ifs) ), (std::istreambuf_iterator<char>()) );
        response->set_port_pipes_dump(port_pipes_dump);
        response->set_success(true);
        return 0;
}

int  QueryRpcHandler::DocaFlowQuery(const DocaFlowQueryRequest * request, DocaFlowResponse *response)
{
        int res;
        struct doca_flow_pipe_entry *entry;
        struct doca_flow_query query_stats;
        DocaFlowError *err_response = response->mutable_error();
        DocaFlowQueryStats *query_stats_res = response->mutable_query_stats();
        uint64_t entry_id = request->entry_id();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();

        /* check if the entry is legal */
        entry = ports_pipes_manager->getPipeEntry(entry_id);
        if (entry == NULL){
                err_response->set_err_message(std::string("no entry found matching the given entry id"));
                response->set_success(false);
                return -1;
        }

        /* if the returned value is negative then the given pipe queue of the entry is wrong */
        res = doca_flow_query(entry, &query_stats);
        if (res < 0){
                err_response->set_err_message(std::string("doca flow query error"));
                response->set_success(false);
                return -1;
        }

        query_stats_res->set_total_bytes(query_stats.total_bytes);
        query_stats_res->set_total_pkts(query_stats.total_pkts);
        response->set_success(true);
        return 0;
}

int AgingHandleRpcHandler::DocaFlowAgingHandle(const DocaFlowAgingHandleRequest *request, DocaFlowResponse *response)
{
        int res;
        uint64_t port_id = request->port_id();
        uint queue = request->queue();
        uint64_t quota = request->quota();
        int len = request->len();
        DocaFlowError *err_response = response->mutable_error();
        DocaFlowAgingHandleRes *handle_aging_res;
        struct doca_flow_pipe_entry *entry;
	struct doca_flow_aged_query *aged_entries;

        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        if(len <= 0){
                err_response->set_err_message(std::string("Number of required aged entries should be positive!"));
                response->set_success(false);
                return -1;
        }

        aged_entries = (struct doca_flow_aged_query *)malloc(sizeof(*aged_entries) * len);
        if(aged_entries == NULL){
                err_response->set_err_message(std::string("Memory allocation error!"));
                response->set_success(false);
                return -1;
        }

        handle_aging_res = response->mutable_handle_aging_res();

        /* the actual call to doca flow aging handler */
        res = doca_flow_aging_handle(this->server_params_->GetPort(port_id), queue, quota, aged_entries, len);

        /* add all ids of aged entries */
        for (int entry_idx = 0; entry_idx < res; entry_idx++){
                entry = (struct doca_flow_pipe_entry *)((void *)aged_entries[entry_idx].user_data);
                handle_aging_res->add_entries((server_params_->GetPipesManager())->getEntryId(entry));
        }

        if (aged_entries)
                free(aged_entries);

        /* set the returned result of handle aging API */
        handle_aging_res->set_res(res);
        response->set_success(true);
        return 0;
}

int EntriesProcessRpcHandler::DocaFlowEntriesProcess(const DocaFlowEntriesProcessRequest *request, DocaFlowResponse *response)
{
        int res;
        uint64_t port_id = request->port_id();
        uint32_t pipe_queue = request->pipe_queue();
        uint64_t timeout = request->timeout();
        uint32_t max_processed_entries = request->max_processed_entries();
        DocaFlowError *err_response = response->mutable_error();

        /* check if port id is legal */
        if (port_id >= server_params_->GetNumberOfPorts()){
                err_response->set_err_message(std::string("port id is illegal"));
                response->set_success(false);
                return -1;
        }

        res = doca_flow_entries_process(server_params_->GetPort(port_id), pipe_queue, timeout, max_processed_entries);
        if(res < 0)
        {
                err_response->set_err_message(std::string("Doca Flow Process entries failure!"));
                response->set_success(false);
                return -1;
        }
        /* set the returned result of handle aging API */
        response->set_nb_entries_processed(res);
        response->set_success(true);
        return 0;
}

int PipeEntryGetStatusRpcHandler::DocaFlowPipeEntryGetStatus(const DocaFlowPipeEntryGetStatusRequest *request, DocaFlowResponse *response)
{
        enum doca_flow_entry_status res;
        struct doca_flow_pipe_entry *entry;
        uint64_t entry_id = request->entry_id();
        DocaFlowError *err_response = response->mutable_error();
        PipesManager *ports_pipes_manager = server_params_->GetPipesManager();
        uint nb_ports = server_params_->GetNumberOfPorts();

        for(uint64_t port_id = 0; port_id < nb_ports; port_id++)
        {
                /* check if the entry is found in the current port id */
                entry = ports_pipes_manager->getPipeEntry(entry_id);
                if (entry)
                {
                        res = doca_flow_pipe_entry_get_status(entry);
                        response->set_status(DocaFlowEntryStatus(res));
                        response->set_success(true);
                        return 0;
                }
        }
                /* Otherwise, the entry is not found */
        err_response->set_err_message(std::string("no entry found matching the given entry identifier!"));
        response->set_success(false);
        return -1;
}

int PortSwitchGetRpcHandler::DocaFlowPortSwitchGet(const DocaFlowEmptyRequest *request, DocaFlowResponse *response)
{
        DocaFlowError *err_response = response->mutable_error();
        uint nb_ports = server_params_->GetNumberOfPorts();
        struct doca_flow_port *switch_port = doca_flow_port_switch_get();
        for(uint64_t port_id = 0; port_id < nb_ports; port_id++)
        {
                if(server_params_->GetPort(port_id) != switch_port)
                        continue;
                response->set_switch_port_id(port_id);
                response->set_success(true);
                return 0;
        }

        /* Otherwise, the entry is not found */
        err_response->set_err_message(std::string("no switch port found!"));
        response->set_success(false);
        return -1;
}

void InitRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("DOCA FLOW INIT request received");
                DocaFlowInit(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                GPR_ASSERT(status_ == FINISH);
                break;
        case FINISH:
                DOCA_DLOG_DBG("DOCA FLOW INIT request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void DestroyRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("DESTROY request received");
                DocaFlowDestroy(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                GPR_ASSERT(status_ == FINISH);
                break;
        case FINISH:
                DOCA_DLOG_DBG("DESTROY request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PortStartRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PORT START request received, port_id: %ld", request_.port_id());
                new PortStartRpcHandler(service_, cq_, server_params_);
                DocaFlowPortStart(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PORT START request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PortStopRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PORT STOP request received, port_id: %ld", request_.port_id());
                new PortStopRpcHandler(service_, cq_, server_params_);
                DocaFlowPortStop(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PORT STOP request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PortPairRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PORT PAIR request received, port_id: %ld", request_.port_id());
                new PortPairRpcHandler(service_, cq_, server_params_);
                DocaFlowPortPair(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PORT PAIR request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PortPrivDataRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PORT PRIV DATA request received, port_id: %ld", request_.port_id());
                new PortPrivDataRpcHandler(service_, cq_, server_params_);
                DocaFlowPortPrivData(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PORT PRIV DATA request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void SharedResourceCfgRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_LOG_DBG("DOCA FLOW SHARED RESOURCE CONFIG request received");
                new SharedResourceCfgRpcHandler(service_, cq_, server_params_);
                DocaFlowSharedResourceCfg(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                GPR_ASSERT(status_ == FINISH);
                break;
        case FINISH:
                DOCA_LOG_DBG("DOCA FLOW SHARED RESOURCE CONFIG request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void SharedResourcesBindRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_LOG_DBG("DOCA FLOW SHARED RESOURCE BIND request received");
                new SharedResourcesBindRpcHandler(service_, cq_, server_params_);
                DocaFlowSharedResourcesBind(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                GPR_ASSERT(status_ == FINISH);
                break;
        case FINISH:
                DOCA_LOG_DBG("DOCA FLOW SHARED RESOURCE BIND request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void SharedResourcesQueryRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_LOG_DBG("DOCA FLOW SHARED RESOURCE QUERY request received");
                new SharedResourcesQueryRpcHandler(service_, cq_, server_params_);
                DocaFlowSharedResourcesQuery(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                GPR_ASSERT(status_ == FINISH);
                break;
        case FINISH:
                DOCA_LOG_DBG("DOCA FLOW SHARED RESOURCE QUERY request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PipeCreateRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("CREATE PIPE request received");
                new PipeCreateRpcHandler(service_, cq_, server_params_);
                DocaFlowPipeCreate(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("CREATE PIPE request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PipeAddEntryRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PIPE ADD ENTRY request received");
                new PipeAddEntryRpcHandler(service_, cq_, server_params_);
                DocaFlowPipeAddEntry(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PIPE ADD ENTRY request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PipeControlAddEntryRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("CONTROL PIPE ADD ENTRY request received");
                new PipeControlAddEntryRpcHandler(service_, cq_, server_params_);
                DocaFlowPipeControlAddEntry(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("CONTROL PIPE ADD ENTRY request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PipeLpmAddEntryRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("LPM PIPE ADD ENTRY request received");
                new PipeLpmAddEntryRpcHandler(service_, cq_, server_params_);
                DocaFlowPipeLpmAddEntry(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("LPM PIPE ADD ENTRY request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PipeRmEntryRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("REMOVE ENTRY request received");
                new PipeRmEntryRpcHandler(service_, cq_, server_params_);
                DocaFlowPipeRmEntry(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("REMOVE ENTRY request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PipeDestroyRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("DESTROY PIPE request received");
                new PipeDestroyRpcHandler(service_, cq_, server_params_);
                DocaFlowPipeDestroy(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("DESTROY PIPE request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PortPipesFlushRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PORT PIPES FLUSH PIPES request received");
                new PortPipesFlushRpcHandler(service_, cq_, server_params_);
                DocaFlowPortPipesFlush(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PORT PIPES FLUSH  request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PortDestroyRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("DESTROY PORT request received");
                new PortDestroyRpcHandler(service_, cq_, server_params_);
                DocaFlowPortDestroy(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("DESTROY PORT request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void PortPipesDumpRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PORT DUMP PIPES request received");
                new PortPipesDumpRpcHandler(service_, cq_, server_params_);
                DocaFlowPortPipesDump(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PORT DUMP PIPES request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void QueryRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("ENTRY QUERY request received");
                new QueryRpcHandler(service_, cq_, server_params_);
                DocaFlowQuery(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("ENTRY QUERY request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void AgingHandleRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("HANDLE AGING request received");
                new AgingHandleRpcHandler(service_, cq_, server_params_);
                DocaFlowAgingHandle(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("HANDLE AGING request processed & finished");
                delete this;
                /* should not get here the destructor is called. */
                break;
        default:
                break;
        }
}

void EntriesProcessRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("ENTRIES PROCESS request received");
                new EntriesProcessRpcHandler(service_, cq_, server_params_);
                DocaFlowEntriesProcess(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("ENTRIES PROCESS request processed & finished");
                delete this;
        default:
                break;
        }
}

void PipeEntryGetStatusRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("ENTRY GET STATUS request received");
                new PipeEntryGetStatusRpcHandler(service_, cq_, server_params_);
                DocaFlowPipeEntryGetStatus(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("ENTRY GET STATUS request processed & finished");
                delete this;
        default:
                break;
        }
}

void PortSwitchGetRpcHandler::Proceed()
{
        switch(status_)
        {
        case PROCESS:
                DOCA_DLOG_DBG("PORT SWITCH GET request received");
                new PortSwitchGetRpcHandler(service_, cq_, server_params_);
                DocaFlowPortSwitchGet(&request_, &reply_);
                status_ = FINISH;
                responder_.Finish(reply_, grpc::Status::OK, this);
                break;
        case FINISH:
                DOCA_DLOG_DBG("PORT SWITCH GET request processed & finished");
                delete this;
        default:
                break;
        }
}
