/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#ifndef FLOW_GRPC_HANDLER_H_
#define FLOW_GRPC_HANDLER_H_

#include <iostream>
#include <fstream>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <string>
#include <cstring>
#include <queue>

#include <doca_flow.h>
#include <doca_error.h>

#include <dpdk_utils.h>

#include <grpcpp/grpcpp.h>
#include <grpcpp/health_check_service_interface.h>

#include "doca_flow.grpc.pb.h"
#include "flow_server_pipes_manager.hpp"

#define MAX_CORES_NUM 8
#define DEFAULT_PORTS_NUM 2

enum lcore_type { CONTROL ,  RPC_HANDLER, PACKET_PROCESSING };

struct server_per_lcore_params {
	int queue;
	bool used;
        lcore_type type;
        int microservice;
        int lcore;
};

class DocaFlowServerParams {

        /* Array of PipesManager, for every port we have pipes manager,
	 * to manage all pipes and entries attached to it.
	 */
	PipesManager *ports_pipes_manager;

        /* Number of ports and queues per port used by client application. */
        unsigned int nb_ports, nb_queues;

        /* Array of pointers for holding initialized doca flow ports. */
        struct doca_flow_port **ports;

        /* Next port_id to use for ports intialization. */
        uint64_t next_port_id;

        struct server_per_lcore_params lcore_params_arr[RTE_MAX_LCORE];

        public:
                DocaFlowServerParams() {
                        /* default values */
                        nb_ports = DEFAULT_PORTS_NUM;
                        nb_queues = MAX_CORES_NUM;
                        ports_pipes_manager = NULL;
                        ports = NULL;
                        next_port_id = 0;
                        memset(lcore_params_arr, 0, sizeof(struct server_per_lcore_params) * RTE_MAX_LCORE);
                };

                /* Initializes all needed data structures. */
                doca_error_t InitParams(unsigned int nb_ports, unsigned int  nb_queues);

                /* Stores the created doca flow port. */
                void SetPort(struct doca_flow_port *port);

                /* Returns the array of pointer of PipesManager. */
                PipesManager * GetPipesManager();

                /* Returns the configured number of ports. */
                unsigned int GetNumberOfPorts() const;

                /* Returns the configured number of queues per port. */
                unsigned int GetNumberOfQueues() const;

                /* Returns a pointer of doca flow port of the given port id */
                struct doca_flow_port * GetPort(uint64_t port_id) const;

                /* Get Server per lcore parameters */
                struct server_per_lcore_params *GetServerPerLcoreParams();

                /* Deallocates all held data structures */
                void Destroy();
};

/* Class encompasing the state and logic needed to serve a DOCA Flow request */
class DocaFlowRpcHandler {

        protected:

        /* The means of communication with the gRPC runtime for asynchronous server */
        DocaFlow::AsyncService* service_;
        /* The producer-consumer queue where for asynchronous server notifications.*/
        grpc::ServerCompletionQueue* cq_;

        /* Context for the rpc, allowing to tweak aspects of it such as the use
         * of compression, authentication, as well as to send metadata back to theclient.
         */
        grpc::ServerContext ctx_;

        /* what the server send back to the client */
        DocaFlowResponse reply_;

        /* The means to get back to the client. */
        grpc::ServerAsyncResponseWriter<DocaFlowResponse> responder_;

        unsigned int port_start_ops = 0;
        /* The following states represents the possibile states of our state machine for RPCs handling. */
        enum DocaFlowRpcStatus { CREATE, QUEUE , PROCESS, FINISH };

        /* The current serving state. */
        DocaFlowRpcStatus status_;

        DocaFlowServerParams *server_params_;
        /* method for filling the mac address */
	void fillMacAddress(uint64_t mac_address_req, uint8_t *mac_address);

	/* method for filling the ip address struct */
	void fillIpAddress(DocaFlowIPAddress *ip_add_req,
		struct doca_flow_ip_addr *ip_add);

	/* method for filling the tunnel */
	void fillTun(DocaFlowTun *tun_req, struct doca_flow_tun *tun);

        /* method for filling the metadata */
        void fillMeta(DocaFlowMeta *meta_req, struct doca_flow_meta *meta);

	/* method for filling the match */
	void fillMatch(DocaFlowMatch *match_req, struct doca_flow_match *match);

	/* method for filling the encap action*/
	void fillEncapAction(DocaFlowEncapAction *encap_action_req,
		struct doca_flow_encap_action *encap_action);

	/* method for filling the actions */
	void fillActions(DocaFlowActions *actions_req, struct doca_flow_actions *actions);

        /* method for filling action field */
        void fillActionField(DocaFlowActionField *action_field_req, doca_flow_action_field *action_field, void *match_address);

        void fillActionMask(DocaFlowActionMask *mask, struct doca_flow_action_desc *action_desc, void *match_address);

        /* method for filling copy action */
        void fillActionCopy(DocaFlowActionCopy *copy, struct doca_flow_action_desc *action_desc, void *match_address);

        /* method for filling action descriptor */
        void fillActionDesc(DocaFlowActionDesc *action_desc_req, struct doca_flow_action_desc *action_desc, void *match_address);

        void fillActionDescsMeta(DocaFlowActionDescsMeta *action_descs_meta_req, struct doca_flow_action_descs_meta *action_descs_meta, void *match_address);

        /* method for filling action descriptors */
        void fillActionDescs(DocaFlowActionDescs *action_descs_req, struct doca_flow_action_descs *action_descs, void *match_address);

 	/* method for filling  monitor struct */
	void fillMonitor(const DocaFlowMonitor *monitor_req, struct doca_flow_monitor *monitor);

	/* method for filling the FORWARDING component */
	void fillFwd(const DocaFlowFwd *fwd_req, struct doca_flow_fwd *fwd, uint16_t **rss_queues_ptr);

         /* method for filling pipe's attributes */
        void fillPipeAttr(const DocaFlowPipeAttr *attr_req, struct doca_flow_pipe_attr *attr);

	/* method for filling error response */
	void fillErrorResponse(DocaFlowResponse *response, struct doca_flow_error *error);

        public:
                /* Take in the "Doca Flow service" instance (in this case representing an asynchronous
                 * server) and the completion queue "cq" used for asynchronous communication
                 * with the gRPC runtime.
                 */
                DocaFlowRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : service_(service), cq_(cq), responder_(&ctx_), status_(PROCESS) {
                                server_params_ = server_params;
                        };

                virtual void Proceed() = 0;
                virtual ~DocaFlowRpcHandler() {};
};

class InitRpcHandler : public DocaFlowRpcHandler {

        DocaFlowCfg request_;

        int DocaFlowInit(const DocaFlowCfg *request, DocaFlowResponse *response);

        public:

                InitRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                       service_->RequestDocaFlowInit(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~InitRpcHandler() override {};
};

class DestroyRpcHandler : public DocaFlowRpcHandler {

        DocaFlowEmptyRequest request_;

        int DocaFlowDestroy(const DocaFlowEmptyRequest *request, DocaFlowResponse *response);

        public:

                DestroyRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                         service_->RequestDocaFlowDestroy(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                virtual void Proceed(void) override;
                virtual ~DestroyRpcHandler() override {};

};

class PortStartRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPortCfg request_;

        int DocaFlowPortStart(const DocaFlowPortCfg *request, DocaFlowResponse *response);

        public:

                PortStartRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                         service_->RequestDocaFlowPortStart(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~PortStartRpcHandler() override {};
};

class PortStopRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPort request_;

        int DocaFlowPortStop(const DocaFlowPort *request, DocaFlowResponse *response);

        public:

                PortStopRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                         service_->RequestDocaFlowPortStop(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~PortStopRpcHandler() override {};
};

class PortPairRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPortPairRequest request_;

        int DocaFlowPortPair(const DocaFlowPortPairRequest *request, DocaFlowResponse *response);

        public:

                PortPairRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                         service_->RequestDocaFlowPortPair(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~PortPairRpcHandler() override {};
};

class PortPrivDataRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPort request_;

        int DocaFlowPortPrivData(const DocaFlowPort *request, DocaFlowResponse *response);

        public:

                PortPrivDataRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                         service_->RequestDocaFlowPortPrivData(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~PortPrivDataRpcHandler() override {};
};

class SharedResourceCfgRpcHandler : public DocaFlowRpcHandler {

        DocaFlowSharedResourceCfgRequest request_;

        int DocaFlowSharedResourceCfg(const DocaFlowSharedResourceCfgRequest *request, DocaFlowResponse *response);

        public:

                SharedResourceCfgRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowSharedResourceCfg(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~SharedResourceCfgRpcHandler() override {};
};

class SharedResourcesBindRpcHandler : public DocaFlowRpcHandler {

        DocaFlowSharedResourcesBindRequest request_;

        int DocaFlowSharedResourcesBind(const DocaFlowSharedResourcesBindRequest *request, DocaFlowResponse *response);

        public:

                SharedResourcesBindRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowSharedResourcesBind(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~SharedResourcesBindRpcHandler() override {};
};

class SharedResourcesQueryRpcHandler : public DocaFlowRpcHandler {

        DocaFlowSharedResourcesQueryRequest request_;

        int DocaFlowSharedResourcesQuery(const DocaFlowSharedResourcesQueryRequest *request, DocaFlowResponse *response);

        public:

                SharedResourcesQueryRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowSharedResourcesQuery(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~SharedResourcesQueryRpcHandler() override {};
};

class PipeCreateRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPipeCreateRequest request_;

        int DocaFlowPipeCreate(const DocaFlowPipeCreateRequest *request, DocaFlowResponse *response);

        public:

                PipeCreateRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPipeCreate(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~PipeCreateRpcHandler() override {};
};

class PipeAddEntryRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPipeAddEntryRequest request_;

        int DocaFlowPipeAddEntry(const DocaFlowPipeAddEntryRequest *request, DocaFlowResponse *response);

        public:

               PipeAddEntryRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPipeAddEntry(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~PipeAddEntryRpcHandler() override {};
};

class PipeControlAddEntryRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPipeControlAddEntryRequest request_;

        int DocaFlowPipeControlAddEntry(const DocaFlowPipeControlAddEntryRequest *request, DocaFlowResponse *response);

        public:

               PipeControlAddEntryRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPipeControlAddEntry(&ctx_, &request_, &responder_, cq_, cq_, this);
                }

                void Proceed(void) override;
                virtual ~PipeControlAddEntryRpcHandler() override {};
};

class PipeLpmAddEntryRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPipeLpmAddEntryRequest request_;

        int DocaFlowPipeLpmAddEntry(const DocaFlowPipeLpmAddEntryRequest *request, DocaFlowResponse *response);

        public:

               PipeLpmAddEntryRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPipeLpmAddEntry(&ctx_, &request_, &responder_, cq_, cq_, this);
                }

                void Proceed(void) override;
                virtual ~PipeLpmAddEntryRpcHandler() override {};
};

class PipeRmEntryRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPipeRmEntryRequest request_;

        int DocaFlowPipeRmEntry(const DocaFlowPipeRmEntryRequest *request, DocaFlowResponse *response);

        public:

                PipeRmEntryRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPipeRmEntry(&ctx_, &request_, &responder_, cq_, cq_, this);
                }

                void Proceed(void) override;
                virtual ~PipeRmEntryRpcHandler() override {};
};

class PipeDestroyRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPipeDestroyRequest request_;

        int DocaFlowPipeDestroy(const DocaFlowPipeDestroyRequest *request, DocaFlowResponse *response);

        public:

                PipeDestroyRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPipeDestroy(&ctx_, &request_, &responder_, cq_, cq_, this);
                }

                void Proceed(void) override;
                virtual ~PipeDestroyRpcHandler() override {};

};

class PortPipesFlushRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPort request_;

        int DocaFlowPortPipesFlush(const DocaFlowPort *request, DocaFlowResponse *response);

        public:

                PortPipesFlushRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPortPipesFlush(&ctx_, &request_, &responder_, cq_, cq_, this);
                }

                void Proceed(void) override;
                virtual ~PortPipesFlushRpcHandler() override {};
};

class PortDestroyRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPort request_;

        int DocaFlowPortDestroy(const DocaFlowPort *request, DocaFlowResponse *response);

        public:

                PortDestroyRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPortDestroy(&ctx_, &request_, &responder_, cq_, cq_, this);
                }

                void Proceed(void) override;
                virtual ~PortDestroyRpcHandler() override {};

};

class QueryRpcHandler : public DocaFlowRpcHandler {

        DocaFlowQueryRequest request_;

        int DocaFlowQuery(const DocaFlowQueryRequest *request, DocaFlowResponse *response);

        public:

               QueryRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowQuery(&ctx_, &request_, &responder_, cq_, cq_, this);
                };

                void Proceed(void) override;
                virtual ~QueryRpcHandler() override {};
};

class PortPipesDumpRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPort request_;

        int DocaFlowPortPipesDump(const DocaFlowPort *request, DocaFlowResponse *response);

        public:

                PortPipesDumpRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPortPipesDump(&ctx_, &request_, &responder_, cq_, cq_, this);
                }
                void Proceed(void) override;
                virtual ~PortPipesDumpRpcHandler() override {};
};

class AgingHandleRpcHandler : public DocaFlowRpcHandler {

        DocaFlowAgingHandleRequest request_;

        int DocaFlowAgingHandle(const DocaFlowAgingHandleRequest *request, DocaFlowResponse *response);

        public:

                AgingHandleRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowAgingHandle(&ctx_, &request_, &responder_, cq_, cq_, this);
                }
                void Proceed(void) override;
                virtual ~AgingHandleRpcHandler() override {};
};

class EntriesProcessRpcHandler : public DocaFlowRpcHandler {

        DocaFlowEntriesProcessRequest request_;

        int DocaFlowEntriesProcess(const DocaFlowEntriesProcessRequest *request, DocaFlowResponse *response);

        public:

                EntriesProcessRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowEntriesProcess(&ctx_, &request_, &responder_, cq_, cq_, this);
                }
                void Proceed(void) override;
                virtual ~EntriesProcessRpcHandler() override {};
};

class PipeEntryGetStatusRpcHandler : public DocaFlowRpcHandler {

        DocaFlowPipeEntryGetStatusRequest request_;

        int DocaFlowPipeEntryGetStatus(const DocaFlowPipeEntryGetStatusRequest *request, DocaFlowResponse *response);

        public:

                PipeEntryGetStatusRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPipeEntryGetStatus(&ctx_, &request_, &responder_, cq_, cq_, this);
                }
                void Proceed(void) override;
                virtual ~PipeEntryGetStatusRpcHandler() override {};
};

class PortSwitchGetRpcHandler : public DocaFlowRpcHandler {

        DocaFlowEmptyRequest request_;

        int DocaFlowPortSwitchGet(const DocaFlowEmptyRequest *request, DocaFlowResponse *response);

        public:

                PortSwitchGetRpcHandler(DocaFlow::AsyncService *service, grpc::ServerCompletionQueue *cq, DocaFlowServerParams *server_params)
                        : DocaFlowRpcHandler(service, cq, server_params) {
                        service_->RequestDocaFlowPortSwitchGet(&ctx_, &request_, &responder_, cq_, cq_, this);
                }
                void Proceed(void) override;
                virtual ~PortSwitchGetRpcHandler() override {};
};

#endif /* DOCA_FLOW_GRPC_HANDLER_H_ */
