/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_argp.h>
#include <doca_log.h>

#ifdef PACKET_BUFFERING_SUPPORT
#include "../packet_buffering/packet_buffering_core.hpp"
#else
#include "flow_server_core.hpp"
#endif

#define NB_FLOW_HANDLERS_W_PB 3         /* Number of lcores that are responsible for Flow RPC Handling */

DOCA_LOG_REGISTER(doca_flow::GRPC);

using namespace std;
using grpc::Server;
using grpc::ServerBuilder;
using grpc::ServerCompletionQueue;
using grpc::ServerContext;
using grpc::Status;
using grpc::Channel;

DocaFlowServerParams *server_params;

Status DocaOrchestrationImpl::HealthCheck(ServerContext *context, const HealthCheckReq *request,
				   HealthCheckResp *response)
{
	(void)context;
	(void)request;
	(void)response;

	/* Show the service that we are responsive */
	return Status::OK;
}

Status DocaOrchestrationImpl::Destroy(ServerContext *context, const DestroyReq *request,
			       DestroyResp *response)
{
	(void)context;
	(void)request;
	(void)response;

	return Status::OK;
}

int handle_rpcs(void *microservices)
{
	uint32_t lcore = rte_lcore_id();
	struct server_per_lcore_params *params = server_params->GetServerPerLcoreParams();
	MicroService **microservices_ = (MicroService **)microservices;

	if(params[lcore].used && microservices_[params[lcore].microservice] != NULL)
		microservices_[params[lcore].microservice]->HandleRpcs();
	else {
		DOCA_LOG_ERR("Core is initialized by EAL but not used! ");
		for (int i = 0; i < MAX_CORES_NUM; i++)
			if (microservices_[i])
				microservices_[i]->Shutdown();
		return -1;
	}

	return 0;
}

class ServerImpl final {
	MicroService **microservices = NULL;

#ifdef PACKET_BUFFERING_SUPPORT
	unsigned int nb_microservices = MAX_CORES_NUM - 2; /* Main lcore is reserved fdor control and one lcore for packet processing */
	unsigned int nb_flow_microservices = NB_FLOW_HANDLERS_W_PB;
#else
	unsigned int nb_microservices = MAX_CORES_NUM;
	unsigned int nb_flow_microservices = MAX_CORES_NUM;
#endif
	std::unique_ptr<Server> server_;
	std::unique_ptr<ServerCompletionQueue> cq_;

	public:

		~ServerImpl()
		{
			if (microservices == NULL)
				return;

			for(unsigned int i = 0; i < nb_microservices ; i++) {
				if (microservices[i] == NULL)
					continue;
				microservices[i]->Shutdown();
				delete microservices[i];
			}
			delete[] microservices;
		}

		doca_error_t LaunchLcores(struct application_dpdk_config *dpdk_config)
		{
#ifndef PACKET_BUFFERING_SUPPORT
			rte_eal_mp_remote_launch(handle_rpcs, microservices, CALL_MAIN);
#else
			int lcore = 0;
			struct server_per_lcore_params *lcore_params_arr = server_params->GetServerPerLcoreParams();
			/* marco to browse all running lcores except the main lcore */
			RTE_LCORE_FOREACH(lcore)
			{
				switch(lcore_params_arr[lcore].type)
				{
					case CONTROL:
						break;
					case PACKET_PROCESSING:
						rte_eal_remote_launch(process_packets, dpdk_config, lcore);
						break;
					case RPC_HANDLER:
						rte_eal_remote_launch(handle_rpcs, microservices, lcore);
						break;
				}
			}
			rte_eal_mp_wait_lcore();
#endif
			return DOCA_SUCCESS;
		}

		doca_error_t Run(const char *grpc_address, struct application_dpdk_config *dpdk_config)
		{
			doca_error_t result;
			unsigned int base_port = 0;
			std::string  ip_address, check_server_address, server_address;

			/* Check if we got a port or if we are using the default one */
			check_server_address = std::string(grpc_address);
			if (check_server_address.find(':') == std::string::npos) {
				ip_address = check_server_address;
				base_port = eNetworkPort::k_DocaFlow;
			}
			else {
				ip_address = std::string(check_server_address.substr(0,  check_server_address.find(':')));
				base_port = stoi(check_server_address.substr(check_server_address.find(':') + 1,  check_server_address.length()));
			}

			server_address = std::string(std::string(ip_address) + ":" + std::to_string(base_port));

			server_params = new DocaFlowServerParams();
			if (server_params == NULL)
				return DOCA_ERROR_NO_MEMORY;
			result = server_params->InitParams(dpdk_config->port_config.nb_ports, dpdk_config->port_config.nb_queues);
			DOCA_LOG_INFO("Server Parameters initialization succeeded! ");
			if (result != DOCA_SUCCESS) {
				delete server_params;
				return result;
			}

			microservices = new MicroService*[nb_microservices];
			for(unsigned int i = 0; i < nb_microservices; i++)
			{
#ifdef PACKET_BUFFERING_SUPPORT
				if(i < nb_flow_microservices)
					microservices[i] = new DocaFlowMicroService(i, server_params);
				else
					microservices[i] = new PacketBufferingMicroService(i, server_params);
#else
				microservices[i] = new DocaFlowMicroService(i, server_params);
#endif
				server_address = std::string(std::string(ip_address) + ":" + std::to_string(base_port + i));
				microservices[i]->InitMicroService(server_address);
			}

			LaunchLcores(dpdk_config);

			return DOCA_SUCCESS;
		}
};

static doca_error_t
flow_grpc_hairpinq_callback(void *param, void *config)
{
	int* hairpin_q_cfg = (int *)config;
	int *nb_hairpin_qs = (int *)param;

	if (*nb_hairpin_qs >= 0)
		*hairpin_q_cfg = *nb_hairpin_qs;
	else {
		DOCA_LOG_ERR("Hairpin queues number should be non-negative = [%d]", *nb_hairpin_qs);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return DOCA_SUCCESS;
}

doca_error_t
register_flow_server_params()
{
	doca_error_t result;
	struct doca_argp_param *nb_hairpin_qs;

	/* Create and register flow grpc server number of haripin queues. */
	result = doca_argp_param_create(&nb_hairpin_qs);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(nb_hairpin_qs, "hqs");
	doca_argp_param_set_long_name(nb_hairpin_qs, "nb_hairpin_qs");
	doca_argp_param_set_description(nb_hairpin_qs, "Set number of hairpin queues");
	doca_argp_param_set_callback(nb_hairpin_qs, flow_grpc_hairpinq_callback);
	doca_argp_param_set_type(nb_hairpin_qs, DOCA_ARGP_TYPE_INT);
	doca_argp_param_set_mandatory(nb_hairpin_qs);
	result = doca_argp_register_param(nb_hairpin_qs);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}
	return DOCA_SUCCESS;
}

int
main(int argc, char *argv[])
{
	doca_error_t result;
	int exit_status = EXIT_SUCCESS;
	ServerImpl server;
	const char *grpc_address;
	struct application_dpdk_config dpdk_config = {0};

	dpdk_config.port_config.nb_ports = DEFAULT_PORTS_NUM;
	dpdk_config.port_config.nb_queues = MAX_CORES_NUM;
	dpdk_config.reserve_main_thread = false;

	/* Parse cmdline/json arguments, no program-specific parameters */
	result = doca_argp_init("flow_grpc", &dpdk_config.port_config.nb_hairpin_q);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init ARGP resources: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}
	doca_argp_set_dpdk_program(dpdk_init);
	doca_argp_set_grpc_program();
	result = register_flow_server_params();
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register ARGP params: %s",  doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}
	result = doca_argp_start(argc, argv);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse server input: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}
	/* Update queues and ports */
	result = dpdk_queues_and_ports_init(&dpdk_config);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to update application ports and queues: %s", doca_get_error_string(result));
		exit_status = EXIT_FAILURE;
		goto dpdk_destroy;
	}

	if (doca_argp_get_grpc_addr(&grpc_address) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to get grpc address");
		exit_status = EXIT_FAILURE;
		goto dpdk_cleanup;
	}

	/* Trigger DOCA Flow server */
        result = server.Run(grpc_address, &dpdk_config);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to run the gRPC server");
		exit_status = EXIT_FAILURE;
	}

dpdk_cleanup:
	/* DPDK cleanup */
	dpdk_queues_and_ports_fini(&dpdk_config);
dpdk_destroy:
	dpdk_fini();

	doca_argp_destroy();

	return exit_status;
}
