/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <cstdlib>
#include <unistd.h>
#include <fcntl.h>

#include <doca_log.h>

#include "flow_server_pipes_manager.hpp"

DOCA_LOG_REGISTER(doca_flow::GRPC::PipesManager);

static const int ELEMENT_ERASED_SUCCESS = 1;
static const int ENTRIES_TABLE_IDX = 2;

PipesManager::PipesManager() : private_key(0) {}

doca_error_t
PipesManager::init()
{
	int fd;
	ssize_t nread;

	fd = open("/dev/urandom", O_RDONLY);
	if (fd < 0) {
		DOCA_LOG_ERR("Unable to open /dev/urandom");
		return DOCA_ERROR_IO_FAILED;
	}

	nread = read(fd, &this->private_key, sizeof(this->private_key));
	if (nread < 0) {
		DOCA_LOG_ERR("Unable to init the secret key");
		close(fd);
		return DOCA_ERROR_IO_FAILED;
	}

	close(fd);
	return DOCA_SUCCESS;
}

uint64_t
PipesManager::generateID(void *ptr) const
{
	return (uint64_t) ptr ^ private_key;
}

uint64_t
PipesManager::addPipe(struct doca_flow_pipe *pipe, uint16_t port_id)
{
	uint64_t pipe_id;

	pipe_id = generateID(pipe);

	if (pipe_id_to_tuple_map.find(pipe_id) != pipe_id_to_tuple_map.end())
		return 0;

	auto port_to_pipes_itr = port_id_to_pipes_id_map.find(port_id);

	if (port_to_pipes_itr == port_id_to_pipes_id_map.end()) {
		/* this port id is new, create a new pipes table for it and insert the newly created pipe */
		auto res = port_id_to_pipes_id_map.insert({port_id, unordered_map<uint64_t, uint64_t>()});
		if (res.second == false)
			return 0;

		port_to_pipes_itr = res.first;
	}

	/* insert the pipe id into pipes table of current port id and check result */
	if (port_to_pipes_itr->second.insert({pipe_id, pipe_id}).second == false) {
		port_id_to_pipes_id_map.erase(port_id);
		return 0;
	}

	/* save a reference to pipes table for this port id for future deletion */
	unordered_map<uint64_t, uint64_t> &pipes_map = port_to_pipes_itr->second;

	/* create a pipe tuple */
	tuple<struct doca_flow_pipe *, unordered_map<uint64_t, uint64_t> &,
	      unordered_map<uint64_t, struct doca_flow_pipe_entry *>>
		pipe_tuple(pipe, pipes_map, unordered_map<uint64_t, doca_flow_pipe_entry *>());

	/* insert the tuple into the tuples table for this pipe id and check result */
	if (pipe_id_to_tuple_map.insert({pipe_id, pipe_tuple}).second == false) {
		port_id_to_pipes_id_map.erase(port_id);
		port_to_pipes_itr->second.erase(pipe_id);
		return 0;
	}

	return pipe_id;
}

uint64_t
PipesManager::addPipeEntry(struct doca_flow_pipe *pipe, struct doca_flow_pipe_entry *entry)
{
	uint64_t pipe_id, entry_id;

	pipe_id = generateID(pipe);

	auto pipe_to_tuple_itr = pipe_id_to_tuple_map.find(pipe_id);
	if (pipe_to_tuple_itr == pipe_id_to_tuple_map.end())
		return 0;

	entry_id = generateID(entry);

	std::unordered_map<unsigned long, doca_flow_pipe_entry*>& entries_map = get<ENTRIES_TABLE_IDX>(pipe_to_tuple_itr->second);

	if (entries_map.find(entry_id) != entries_map.end())
		return 0;

	if (entries_map.insert({entry_id, entry}).second == false)
		return 0;

	if (entry_id_to_pipe_id_map.insert({entry_id, pipe_id}).second == false) {
		entries_map.erase(entry_id);
		return 0;
	}

	return entry_id;
}

struct doca_flow_pipe *
PipesManager::getPipe(uint64_t pipe_id)
{
	auto itr = pipe_id_to_tuple_map.find(pipe_id);
        if (itr == pipe_id_to_tuple_map.end())
                return nullptr;

        return get<0>(itr->second);
}

struct doca_flow_pipe_entry* PipesManager::getPipeEntry(uint64_t entry_id) {

        auto entry_to_pipe_itr = entry_id_to_pipe_id_map.find(entry_id);
        if (entry_to_pipe_itr == entry_id_to_pipe_id_map.end())
                return nullptr;

        uint64_t pipe_id = entry_to_pipe_itr->second;

        auto pipe_to_tuple_itr = pipe_id_to_tuple_map.find(pipe_id);
        if (pipe_to_tuple_itr == pipe_id_to_tuple_map.end())
                return nullptr;

	std::unordered_map<unsigned long, doca_flow_pipe_entry*>& entries_map = get<ENTRIES_TABLE_IDX>(pipe_to_tuple_itr->second);

        if(entries_map.find(entry_id) == entries_map.end())
                return nullptr;

        return entries_map.find(entry_id)->second;
}

bool PipesManager::removePipe(uint64_t pipe_id) {

	bool is_success = false;
        auto pipe_to_tuple_itr = pipe_id_to_tuple_map.find(pipe_id);
        if (pipe_to_tuple_itr == pipe_id_to_tuple_map.end())
                return is_success;

	/* remove pipe id directly from relevant port id to pipes table*/
	if (get<1>(pipe_to_tuple_itr->second).erase(pipe_id) != ELEMENT_ERASED_SUCCESS)
		return is_success;

	std::unordered_map<unsigned long, doca_flow_pipe_entry*>& entries_map = get<ENTRIES_TABLE_IDX>(pipe_to_tuple_itr->second);
        for (auto &pair : entries_map)
		entry_id_to_pipe_id_map.erase(pair.first);

	is_success = pipe_id_to_tuple_map.erase(pipe_id);

	return is_success;
}

bool PipesManager::removePipeEntry(uint64_t entry_id)
{
	bool is_success = false;
        auto entry_to_pipe_itr = entry_id_to_pipe_id_map.find(entry_id);
        if (entry_to_pipe_itr == entry_id_to_pipe_id_map.end())
                return is_success;

        auto pipe_to_tuple_itr = pipe_id_to_tuple_map.find(entry_to_pipe_itr->second);
        if (pipe_to_tuple_itr == pipe_id_to_tuple_map.end())
                return is_success;

	if (entry_id_to_pipe_id_map.erase(entry_id) != ELEMENT_ERASED_SUCCESS)
		return is_success;

	std::unordered_map<unsigned long, doca_flow_pipe_entry*>& entries_map = get<ENTRIES_TABLE_IDX>(pipe_to_tuple_itr->second);

	is_success = entries_map.erase(entry_id);
        if (is_success != ELEMENT_ERASED_SUCCESS)
		entry_id_to_pipe_id_map.insert({entry_id, entry_to_pipe_itr->second});

	return is_success;
}

bool PipesManager::portPipesFlush(uint16_t port_id) {
	auto port_to_pipes_itr = port_id_to_pipes_id_map.find(port_id);
	if (port_to_pipes_itr == port_id_to_pipes_id_map.end())
		return false;

	for (auto & pipe_itr : port_to_pipes_itr->second) {
		auto pipe_to_tuple_itr = pipe_id_to_tuple_map.find(pipe_itr.first);
		if (pipe_to_tuple_itr == pipe_id_to_tuple_map.end())
			continue;

		for (auto & entry_itr : get<ENTRIES_TABLE_IDX>(pipe_to_tuple_itr->second))
			entry_id_to_pipe_id_map.erase(entry_itr.first);

		pipe_id_to_tuple_map.erase(pipe_itr.first);
	}
	port_id_to_pipes_id_map.erase(port_id);
	return true;
}

uint64_t
PipesManager::getEntryId(struct doca_flow_pipe_entry *entry)
{
	return generateID(entry);
}
