/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "doca_regex.h"

#include <pthread.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <doca_log.h>

#include <impl/drivers/regex_driver.h>
#include <impl/regex_defines.h>
#include <impl/regex_device_ops.h>
#include <priv_doca_buf.h>
#include <priv_doca_cq.h>
#include <priv_doca_ctx.h>
#include <private/common_defines.h>

/**********************************************************************************************************************
 *
 * #defines and macros
 *
 *********************************************************************************************************************/

DOCA_LOG_REGISTER(DOCA_REGEX);

#define PER_Q_TEMP_JOBS_SIZE REGEX_MAX_PER_QP_NUM_DESCS
#define JOB_ID_CACHE_SIZE REGEX_MAX_PER_QP_NUM_DESCS
#define JOB_ID_BITMASK_ARR_LEN (JOB_ID_CACHE_SIZE / 64)
#define ALIGN_CACHE_LINE 64
#define DO_NOT_AGGREGATE false
#define ALLOW_AGGREGATION true
#define DEFAULT_MEMPOOL_SIZE 1024

/**********************************************************************************************************************
 *
 * Type definitions
 *
 *********************************************************************************************************************/
enum ping_pong {
	/** The driver that will be polled next dequeue */
	PING_PONG_NEXT = 0,

	/** The driver that was most recently read from */
	PING_PONG_LAST_USED = 1
};

/**
 * Context to maintain the relationship between a driver and a queue
 */
struct driver_queue_ctx {
	/** Shared non owning reference to a driver instance */
	struct regex_driver *driver;

	/** Queue that is unique to this context */
	struct regex_driver_queue *queue;
};

/**
 * Internal data required to push the parts of a job through the RXP until they are all done.
 */
struct sending_job {
	/**
	 * For the unavoidable case where we have to fall-back from hw to sw, but an active job already exists, it can
	 * be chained on here. No new jobs will be accepted until this backlog is cleared.
	 */
	struct sending_job *backlog;

	/** The driver queue this job is using. */
	struct driver_queue_ctx *driver_queue_ctx;

	/**
	 * Copy of request data to be resent (after being manipulated) with each sub job.
	 */
	struct doca_regex_job_search request;

	/** Bytes to send in next chunk */
	uint32_t send_chunk_size;

	/** Send offset */
	uint32_t send_offset;

	/** Bytes remaining to be sent. */
	uint32_t outstanding_bytes;

	/**
	 * Number of chunks of data yet to be sent to the associated device.
	 */
	uint32_t outstanding_chunks;

	/**
	 * Total number of chunks of data to be sent to the associated device.
	 */
	uint32_t total_chunks;

	/**
	 * Number of bytes to step forward each iteration.
	 */
	uint32_t step_size;
};

/**
 * Internal data required to recollect the fragments of a user job to the RXP. Keeps tracking of work done so far and
 * which FD if any to signal upon completion. Only one of these structures will exist per job submitted to the engine.
 * All responses from SW and or HW devices will be aggregated together.
 */
struct receive_job {

	/** User supplied value to identify this job and relate the response to a result. */
	uint64_t correlation_id;

	/**
	 * Store rule group id's for re-use in case of fall-back from HW -> SW
	 */
	uint16_t orig_rule_group_ids[4];

	/**
	 * Pointer to next receive_job to support keeping a linked list of jobs.
	 */
	struct receive_job *next_job;

	/**
	 * Pointer to application provided job buffer, needed for fall-back.
	 */
	struct doca_buf const *orig_buffer;

	/**
	 * Performance gain - store the tail match for fast insertion at the end.
	 */
	struct doca_regex_match *last_match;

	/**
	 * Number of outstanding fragment responses to complete this job.
	 */
	uint32_t outstanding_responses;

	/**
	 * Flag to indicate if the job has performed a fall-back from HW -> SW. Ensures we only do this once and let us
	 * know upon receipt of the last fragment to not emit this response. We keep this job in use until all of its
	 * fragments are completed to avoid a new job re-using this slot to get responses from the driver that are not
	 * related to it.
	 */
	uint8_t has_done_fallback;
};

/**
 * Holds some additional meta-data we need for each fragment to ease processing / recombination of results and to
 * correlate fragments with the users original request.
 */
struct fragment_metadata {
	/** dual purpose id, for huge jobs this refers to the index of the receive_job in the mempool, for a
	 * small job this is the input job id
	 */
	uint64_t ref_id;

	/**
	 * Store rule group id's for re-use in case of fall-back from HW -> SW
	 */
	uint16_t rule_group_ids[4];

	/**
	 * Store pointer to doca_buffer for re-use in case of fall-back from HW -> SW
	 */
	struct doca_buf const *buffer;

	/** control bit fields to determine:
	 *  - if a job is a regular (single fragment) or HJE (multiple fragment) job
	 *  - the index of this fragment within the overall job
	 *  - some free bits for later use
	 */
	struct {
		uint32_t is_hje_job : 1;
		uint32_t reserved : 7;
		uint32_t fragment_idx : 24;
	} ctrl;

	/**
	 * Step size for HJE fragments to calculate the true offset of a received match For example if fragment_idx >
	 * 0 then match.match_start += hje_step_size * fragment_idx.
	 */
	uint32_t hje_step_size;

	/**
	 * User supplied result storage object.
	 */
	struct doca_regex_search_result *result;
};

/**
 * Holds a collection of per fragment meta-data along with a large bitmask array to track which entries are currently in
 * use
 */
struct fragment_metadata_cache {
	/** Array of metadata */
	struct fragment_metadata data[JOB_ID_CACHE_SIZE];
	/** array of 64bit values for use with ffsll to find a free slot in the data array for each bit 1 = free, 0 = in
	 * use
	 */
	uint64_t inuse_mask[JOB_ID_BITMASK_ARR_LEN];
};

/**
 * Per work queue state.
 *
 * @note Everything in this structure may assume it's used by only a single thread.
 */
struct __attribute__((aligned(ALIGN_CACHE_LINE))) regex_workq_context {
	/** meta data required per job to correctly process responses */
	struct fragment_metadata_cache *metadata_cache;

	/**
	 * A memory pool which can be used to get match objects without the need to allocate and reallocate memory as we
	 * aggregate responses from the devices into a job_internal.
	 */
	struct doca_regex_mempool *matches_mp;

	/**
	 * A self owned pool from which to allocate internal send and receive job objects.
	 */
	struct doca_regex_mempool *jobs_mp;

	/** Queue for jobs to be sent to hardware driver */
	struct driver_queue_ctx hardware_ctx;

	/** Queue for jobs to be sent to software driver */
	struct driver_queue_ctx software_ctx;

	/**
	 * All jobs start here. Once the last block of data has been sent to the next job can be sent. A corresponding
	 * receive_job will be placed on the waiting_jobs queue so that any responses received before all the data has
	 * been sent to the engine can be aggregated.
	 *
	 * @note If a sending_job exists and we need to enqueue another job due to a fall-back or similar internal
	 * source then extra sending_job can be linked onto the tail of this one.
	 *
	 * @note All enqueue operations will be rejected until this sending_job and its backlog have been dispatched.
	 */
	struct sending_job *sending_job;

	/**
	 * Ping-pong buffer. Each time we fill the cache we switch which driver we use, this provides multiple things:
	 *  - We know if the cache entries are from hardware or software.
	 *  - We can simplify the hot path, both pointers here are always valid so no NULL checks are needed.
	 *  - By switching which driver we read from each dequeue we avoid starvation of any driver.
	 */
	struct driver_queue_ctx *driver_ping_pong[2];

	/**
	 * Cache for a batch of responses read from a regex driver queue.
	 */
	struct regex_driver_queue_result_batch response_cache;

	/**
	 * Non owning pointer to event channel
	 */
	struct priv_doca_comp_events_channel *channel;

	/** Reference to parent workq. */
	struct doca_workq *workq;

	/** Reference to poller. */
	struct priv_doca_workq_progress_entry *poller;

	/**
	 * Event channel handle
	 */
	doca_event_handle_t channel_handle;

	/**
	 * How many of the items in temp_responses are processed, we fetch another batch from the driver once we have
	 * processed all the responses in the batch we fetched.
	 */
	uint32_t response_cache_read_idx;

	/** Value set by: DOCA_REGEX_PROPERTY_HUGE_JOB_EMULATION_OVERLAP */
	uint16_t hje_overlap_size;

	/** Value set by: DOCA_REGEX_PROPERTY_SMALL_JOB_OFFLOAD_THRESHOLD */
	uint16_t small_job_offload_threshold;

	/** Value set by: DOCA_REGEX_PROPERTY_FAILED_JOB_FALLBACK */
	uint8_t sw_fallback_enabled;
};

struct binary_blob {
	/** binary blob of bytes */
	uint8_t *data;

	/** size of blob */
	size_t size;
};

/**
 * An instance of doca_regex.
 */
struct doca_regex {
	/** doca_ctx */
	struct doca_ctx base;

	/** mutex */
	pthread_mutex_t *mutex; // FUTURE - put this into the base?

	/** Hardware accelerated driver. */
	struct regex_driver *hardware_driver;

	/** Software based driver. */
	struct regex_driver *software_driver;

	/** Value set by: DOCA_REGEX_PROPERTY_HARDWARE_BINARY_RULES */
	struct binary_blob hw_binary_rules;

	/** Value set by: DOCA_REGEX_PROPERTY_SOFTWARE_BINARY_RULES */
	struct binary_blob sw_binary_rules;

	/** Value set by: DOCA_REGEX_PROPERTY_HARDWARE_UNCOMPILED_RULES */
	struct binary_blob hw_uncompiled_rules;

	/** Value set by: DOCA_REGEX_PROPERTY_SOFTWARE_UNCOMPILED_RULES */
	struct binary_blob sw_uncompiled_rules;

	/** Value set by: DOCA_REGEX_PROPERTY_MATCHES_MEMORY_POOL_SIZE */
	uint32_t workq_mempool_size;

	/** Number of attached workq objects. */
	uint32_t workq_count; // FUTURE - put this into the base?

	/** Value set by: DOCA_REGEX_PROPERTY_HUGE_JOB_EMULATION_OVERLAP */
	uint16_t hje_overlap_size;

	/** Value set by: DOCA_REGEX_PROPERTY_SMALL_JOB_OFFLOAD_THRESHOLD */
	uint16_t small_job_offload_threshold;

	/** Value set by: DOCA_REGEX_PROPERTY_FAILED_JOB_FALLBACK */
	uint8_t sw_fallback_enabled;
};

/**********************************************************************************************************************
 *
 * Forward declarations
 *
 *********************************************************************************************************************/
static doca_error_t regex_ctx_arm_event(union doca_data event_data);
static doca_error_t regex_ctx_clear_event(union doca_data event_data);
static doca_error_t regex_workq_progress(union doca_data data_priv, struct doca_event *event);

typedef doca_error_t (*regex_ctx_validator_fn)(struct doca_regex const *const regex);

/**********************************************************************************************************************
 *
 * Private internal functions
 *
 *********************************************************************************************************************/
static doca_error_t
clone_blob(struct binary_blob const *blob, void **out_data, size_t *out_len)
{
	if (blob->data == NULL) {
		*out_data = NULL;
		*out_len = 0;
		return DOCA_SUCCESS;
	}

	void *const copy = malloc(blob->size);

	if (copy == NULL)
		return DOCA_ERROR_NO_MEMORY;

	memcpy(copy, blob->data, blob->size);
	*out_data = copy;
	*out_len = blob->size;

	return DOCA_SUCCESS;
}

/**
 * Determine if a doca_dev is a software device
 */
static bool
is_sw_device(struct doca_dev const *dev)
{
	/* HACK ATTACK: This is a hack for unit testing to allow us to "detect" a software device. This will be replaced
	 * with some doca_core routine once doca software devices are supported.
	 *
	 * Using a random small address that should never ever be used in the real world.
	 */
	return dev == (struct doca_dev const *)0x7;
}

/**
 * Determine if a doca_dev_info represents a software device
 */
static bool
is_sw_devinfo(struct doca_devinfo const *devinfo)
{
	/* HACK ATTACK: This is a hack for unit testing to allow us to "detect" a software device. This will be replaced
	 * with some doca_core routine once doca software devices are supported.
	 *
	 * Using a random small address that should never ever be used in the real world.
	 */
	return devinfo == (struct doca_devinfo const *)0x7;
}

doca_error_t
get_devinfo_caps(struct doca_devinfo const *devinfo, struct regex_driver_info *regex_driver_info)
{
	if (devinfo == NULL || regex_driver_info == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	enum regex_device_kind const driver_kind =
		is_sw_devinfo(devinfo) ? REGEX_DEVICE_KIND_HS : REGEX_DEVICE_KIND_BF2;

	if (!is_device_kind_supported(driver_kind))
		return DOCA_ERROR_NOT_SUPPORTED;

	doca_error_t const result = query_regex_device_kind_info(driver_kind, devinfo, regex_driver_info);

	return result;
}

/**
 * Procedure for adding a driver with all required compatibility checks
 */
static doca_error_t
install_driver(struct doca_regex *regex, struct doca_dev *dev)
{
	struct regex_driver *driver = NULL;
	enum regex_device_kind const driver_kind = is_sw_device(dev) ? REGEX_DEVICE_KIND_HS : REGEX_DEVICE_KIND_BF2;

	if (!is_device_kind_supported(driver_kind))
		return DOCA_ERROR_NOT_SUPPORTED;

	driver = create_regex_device_driver(driver_kind, dev);
	if (driver == NULL)
		return DOCA_ERROR_NO_MEMORY;

	if (regex_driver_get_max_single_job_size(driver) == 0) {
		regex_driver_destroy(driver);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (driver_kind == REGEX_DEVICE_KIND_HS) {
		if (regex->software_driver != NULL)
			regex_driver_destroy(regex->software_driver);
		regex->software_driver = driver;
	} else {
		if (regex->hardware_driver != NULL)
			regex_driver_destroy(regex->hardware_driver);
		regex->hardware_driver = driver;
	}

	return DOCA_SUCCESS;
}

/**
 * Procedure to remove a driver
 */
static doca_error_t
uninstall_driver(struct doca_regex *regex, struct doca_dev *dev)
{
	if (regex->hardware_driver != NULL && regex->hardware_driver->dev == dev) {
		regex_driver_destroy(regex->hardware_driver);
		regex->hardware_driver = NULL;
		return DOCA_SUCCESS;
	}

	if (regex->software_driver != NULL && regex->software_driver->dev == dev) {
		regex_driver_destroy(regex->software_driver);
		regex->software_driver = NULL;
		return DOCA_SUCCESS;
	}

	return DOCA_ERROR_INVALID_VALUE;
}

/**
 * For values that default to non zero values they should be initialised here.
 */
static void
set_default_configuration_values(struct doca_regex *regex)
{
	regex->workq_mempool_size = DEFAULT_MEMPOOL_SIZE;
}

/**
 * Store rules binary blob value
 */
static doca_error_t
store_rules(uint8_t const *value, uint32_t length, struct binary_blob *blob)
{
	uint8_t *copy;

	if (blob->data != NULL) {
		free(blob->data);
		blob->data = NULL;
		blob->size = 0;
	}

	if (length == 0 && value == NULL)
		return DOCA_SUCCESS;

	if (length == 0 || value == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	copy = malloc(length);
	if (copy == NULL)
		return DOCA_ERROR_NO_MEMORY;

	memcpy(copy, value, length);
	blob->data = copy;
	blob->size = length;

	return DOCA_SUCCESS;
}

/**
 * Validate configuration
 *
 * @param [in] regex
 * Regex instance
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
validate_configuration(struct doca_regex *regex)
{
	uint32_t const hw_max_job_size = regex_driver_get_max_single_job_size(regex->hardware_driver);
	uint32_t const sw_max_job_size = regex_driver_get_max_single_job_size(regex->software_driver);

	/* validate at least one driver provided */
	if (regex->hardware_driver == NULL && regex->software_driver == NULL) {
		DOCA_LOG_ERR("No RegEx drivers have been configured");
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* validate hardware driver and rules */
	if (regex->hardware_driver != NULL) {
		if (hw_max_job_size == 0)
			return DOCA_ERROR_INVALID_VALUE;

		if ((regex->hw_binary_rules.data == NULL || regex->hw_binary_rules.size == 0) &&
		    (regex->hw_uncompiled_rules.data == NULL || regex->hw_uncompiled_rules.size == 0)) {
			DOCA_LOG_ERR(
				"Neither binary rules nor uncompiled rules have been configured for RegEx hardware device");
			return DOCA_ERROR_INVALID_VALUE;
		}
		if (regex->hw_binary_rules.data != NULL && regex->hw_uncompiled_rules.data != NULL) {
			DOCA_LOG_ERR(
				"Only one type of rules should be configured for Regex hardware device either binary or uncompiled");
			return DOCA_ERROR_INVALID_VALUE;
		}
	} else if ((regex->hw_binary_rules.data != NULL || regex->hw_uncompiled_rules.data != NULL) &&
		   regex->hardware_driver == NULL) {
		DOCA_LOG_ERR("No hardware device configured to utilise provided hardware rules");
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* validate software driver and rules */
	if (regex->software_driver != NULL) {
		if (sw_max_job_size == 0)
			return DOCA_ERROR_INVALID_VALUE;

		if ((regex->sw_binary_rules.data == NULL || regex->sw_binary_rules.size == 0) &&
		    (regex->sw_uncompiled_rules.data == NULL || regex->sw_uncompiled_rules.size == 0)) {
			DOCA_LOG_ERR(
				"Neither binary rules nor uncompiled rules have been configured for RegEx software device");
			return DOCA_ERROR_INVALID_VALUE;
		}
		if (regex->sw_binary_rules.data != NULL && regex->sw_uncompiled_rules.data != NULL) {
			DOCA_LOG_ERR(
				"Only one type of rules should be configured for Regex software device, either binary or uncompiled");
			return DOCA_ERROR_INVALID_VALUE;
		}
	} else if ((regex->sw_binary_rules.data != NULL || regex->sw_uncompiled_rules.data != NULL) &&
		   regex->software_driver == NULL) {
		DOCA_LOG_ERR("No software device configured to utilise provided software rules");
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* validate hje overlap size is smaller than drivers job size */
	if (regex->hje_overlap_size != 0) {
		if (regex->hardware_driver != NULL && regex->hje_overlap_size >= hw_max_job_size) {
			DOCA_LOG_ERR(
				"Huge job emulation overlap(%u) must be less than hardware device max job size(%u)",
				regex->hje_overlap_size, hw_max_job_size);
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (regex->software_driver != NULL && regex->hje_overlap_size >= sw_max_job_size) {
			DOCA_LOG_ERR(
				"Huge job emulation overlap(%u) must be less than software device max job size(%u)",
				regex->hje_overlap_size, sw_max_job_size);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}

	/* validate small job offload */
	if (regex->small_job_offload_threshold != 0) {
		if (regex->software_driver == NULL) {
			DOCA_LOG_ERR("Small job software off-load enabled but no software device was configured");
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (regex->hardware_driver == NULL) {
			DOCA_LOG_ERR("Small job software off-load enabled but no hardware device was configured");
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (regex->small_job_offload_threshold >= hw_max_job_size) {
			DOCA_LOG_ERR("Small job off-load size(%u) must be less than hardware device max job size(%u)",
				     regex->small_job_offload_threshold, hw_max_job_size);
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (regex->small_job_offload_threshold >= sw_max_job_size) {
			DOCA_LOG_ERR("Small job off-load size(%u) must be less than software device max job size(%u)",
				     regex->small_job_offload_threshold, sw_max_job_size);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}

	/* validate software offload */
	if (regex->sw_fallback_enabled) {
		if (regex->software_driver == NULL) {
			DOCA_LOG_ERR("Software fall-back enabled but no software device was configured");
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (regex->hardware_driver == NULL) {
			DOCA_LOG_ERR("Software fall-back enabled but no hardware device was configured");
			return DOCA_ERROR_INVALID_VALUE;
		}
	}

	return DOCA_SUCCESS;
}

/**
 * Fill response cache
 *
 * @param [in] ctx
 * Thread context in use.
 *
 * @return
 * DOCA_SUCCESS - cache has been filled with 0 or more items
 * DOCA_ERROR_INVALID_VALUE - Invalid driver / driver state
 */
static doca_error_t
fill_response_cache(struct regex_workq_context *const ctx)
{
	struct driver_queue_ctx *const active_driver = ctx->driver_ping_pong[PING_PONG_NEXT];

	doca_error_t const result =
		regex_driver_queue_dequeue(active_driver->queue, ctx->matches_mp, &ctx->response_cache);

	if (result != DOCA_ERROR_AGAIN && result != DOCA_SUCCESS)
		return result;

	/* swap */
	ctx->driver_ping_pong[PING_PONG_NEXT] = ctx->driver_ping_pong[PING_PONG_LAST_USED];
	ctx->driver_ping_pong[PING_PONG_LAST_USED] = active_driver;

	return DOCA_SUCCESS;
}

/**
 * Program installed drivers
 *
 * @param [in] regex
 * Regex instance
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - invalid driver, empty rules
 * DOCA_ERROR_NO_MEMORY
 * DOCA_ERROR_NOT_SUPPORTED - failed to get PDN / Device rejected rules data
 */
static doca_error_t
program_drivers(struct doca_regex *regex)
{
	doca_error_t result;

	if (regex->hardware_driver != NULL) {
		if (regex->hw_binary_rules.data != NULL && regex->hw_binary_rules.size != 0) {
			result = regex_driver_program_compiled_rules(
				regex->hardware_driver, regex->hw_binary_rules.data, regex->hw_binary_rules.size);
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to program hardware device binary rules: %s",
					     doca_get_error_string(result));
				return result;
			}
		} else if (regex->hw_uncompiled_rules.data != NULL && regex->hw_uncompiled_rules.size != 0) {
			struct binary_blob hw_compiled_rules = {0};

			result = regex_driver_compile_rules(regex->hardware_driver, regex->hw_uncompiled_rules.data,
							    regex->hw_uncompiled_rules.size,
							    (void **)&(hw_compiled_rules.data),
							    (uint32_t *)&(hw_compiled_rules.size));
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to compile hardware device uncompiled rules: %s",
					     doca_get_error_string(result));
				return result;
			}
			result = regex_driver_program_compiled_rules(regex->hardware_driver, hw_compiled_rules.data,
								     hw_compiled_rules.size);
			free(hw_compiled_rules.data);
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to program hardware device compiled rules: %s",
					     doca_get_error_string(result));
				return result;
			}
		} else {
			DOCA_LOG_ERR("No rules provided for hardware device");
			return DOCA_ERROR_INVALID_VALUE;
		}
	}

	if (regex->software_driver != NULL) {
		if (regex->sw_binary_rules.data != NULL && regex->sw_binary_rules.size != 0) {
			result = regex_driver_program_compiled_rules(
				regex->software_driver, regex->sw_binary_rules.data, regex->sw_binary_rules.size);
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to program software device binary rules: %s",
					     doca_get_error_string(result));
				return result;
			}
		} else if (regex->sw_uncompiled_rules.data != NULL && regex->sw_uncompiled_rules.size != 0) {
			struct binary_blob sw_compiled_rules = {0};

			result = regex_driver_compile_rules(regex->software_driver, regex->sw_uncompiled_rules.data,
							    regex->sw_uncompiled_rules.size,
							    (void **)&(sw_compiled_rules.data),
							    (uint32_t *)&(sw_compiled_rules.size));
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to compile software device uncompiled rules: %s",
					     doca_get_error_string(result));
				return result;
			}
			result = regex_driver_program_compiled_rules(regex->software_driver, sw_compiled_rules.data,
								     sw_compiled_rules.size);
			free(sw_compiled_rules.data);
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to program software device compiled rules: %s",
					     doca_get_error_string(result));
				return result;
			}
		} else {
			DOCA_LOG_ERR("No rules provided for software device");
			return DOCA_ERROR_INVALID_VALUE;
		}
	}

	return DOCA_SUCCESS;
}

/**
 * Copy a group id array
 *
 * @param [in,out] dst
 * Destination of copied value
 *
 * @param [in] b
 * Value to be copied
 */
static inline void
copy_group_ids(uint16_t dst[4], uint16_t const src[4])
{
	dst[0] = src[0];
	dst[1] = src[1];
	dst[2] = src[2];
	dst[3] = src[3];
}

/**
 * Zip a matches linked list to find the tail element
 *
 * @param [in] match
 * Starting position
 *
 * @return
 * The last match in the supplied linked list, Will be NULL only when match is NULL.
 */
static inline struct doca_regex_match *
match_tail(struct doca_regex_match *match)
{
	if (match == NULL)
		return NULL;

	for (; match->next != NULL; match = match->next)
		;

	return match;
}

/**
 * Zip a linked list of matches releasing each match back into the given memory pool.
 *
 * @param [in] matches_mp
 * Memory pool owning the matches
 *
 * @param [in,out] match_ptr
 * Pointer to a pointer to the first match. MUST NOT BE NULL, the pointer pointed to by this pointer may be NULL.
 */
static inline void
release_matches(struct doca_regex_mempool *matches_mp, struct doca_regex_match **match_ptr)
{
	struct doca_regex_match *match = *match_ptr;

	while (match != NULL) {
		struct doca_regex_match *next = match->next;

		doca_regex_mempool_put_obj(matches_mp, match);
		match = next;
	}

	*match_ptr = NULL;
}

/**
 * Determine which registered regex driver and associated queue should be used to process a given job.
 *
 * @param [in] wq_ctx
 * Work queue context.
 *
 * @param [in] job
 * The job to consider.
 *
 * @return
 * A pointer to a driver and queue to use.
 */
static struct driver_queue_ctx *
select_driver_and_queue(struct regex_workq_context *wq_ctx, struct doca_regex_job_search const *job)
{
	struct driver_queue_ctx *result;

	if (wq_ctx->hardware_ctx.driver == NULL) {
		/* If we only have software we must use it */
		result = &wq_ctx->software_ctx;
	} else if (wq_ctx->software_ctx.driver == NULL) {
		/* If we only have hardware we must use it */
		result = &wq_ctx->hardware_ctx;
	} else if (priv_doca_buf_get_data_len(job->buffer) < wq_ctx->small_job_offload_threshold) {
		/* prefer SW for small jobs */
		result = &wq_ctx->software_ctx;
	} else {
		/* nothing made us choose otherwise so pick HW */
		result = &wq_ctx->hardware_ctx;
	}

	return result;
}

/**
 * Release any resources held by a given thread domain.
 *
 * @param [in] ctx
 * The domain to cleanup.
 */
static void
destroy_workq_ctx(struct regex_workq_context *ctx)
{
	if (ctx == NULL)
		return;

	if (ctx->hardware_ctx.queue != NULL) {
		regex_driver_queue_destroy(ctx->hardware_ctx.queue);
		ctx->hardware_ctx.queue = NULL;
	}

	if (ctx->software_ctx.queue != NULL) {
		regex_driver_queue_destroy(ctx->software_ctx.queue);
		ctx->software_ctx.queue = NULL;
	}

	/* release any sending jobs */
	while (ctx->sending_job != NULL) {
		struct sending_job *next = ctx->sending_job->backlog;

		doca_regex_mempool_put_obj(ctx->jobs_mp, ctx->sending_job);
		ctx->sending_job = next;
	}

	doca_regex_mempool_destroy(ctx->matches_mp);
	doca_regex_mempool_destroy(ctx->jobs_mp);
	free(ctx->response_cache.responses);
	free(ctx->metadata_cache);
	free(ctx);
}

/**
 * Create a workq context
 *
 * @param [in] regex
 * The regex instance to use
 *
 * @param [out] out
 * Output pointer to created workq_context upon success. Otherwise this param is not altered.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INV
 * DOCA_ERROR_NO_MEMORY
 */
static doca_error_t
create_workq_ctx(struct doca_regex *regex, struct regex_workq_context **workq_ctx)
{
	size_t ii;
	struct regex_workq_context *new_ctx;

	new_ctx = aligned_calloc(ALIGN_CACHE_LINE, sizeof(struct regex_workq_context));
	if (new_ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	new_ctx->hje_overlap_size = regex->hje_overlap_size;
	new_ctx->small_job_offload_threshold = regex->small_job_offload_threshold;
	new_ctx->sw_fallback_enabled = regex->sw_fallback_enabled;

	new_ctx->jobs_mp = doca_regex_mempool_create(MAX(sizeof(struct sending_job), sizeof(struct receive_job)),
						     REGEX_MAX_PER_QP_NUM_DESCS);
	if (new_ctx->jobs_mp == NULL) {
		destroy_workq_ctx(new_ctx);
		return DOCA_ERROR_NO_MEMORY;
	}

	new_ctx->response_cache.responses =
		aligned_calloc(ALIGN_CACHE_LINE, PER_Q_TEMP_JOBS_SIZE * sizeof(struct regex_driver_response));
	if (new_ctx->response_cache.responses == NULL) {
		destroy_workq_ctx(new_ctx);
		return DOCA_ERROR_NO_MEMORY;
	}
	new_ctx->response_cache.responses_size = PER_Q_TEMP_JOBS_SIZE;

	if (regex->workq_mempool_size != 0) {
		new_ctx->matches_mp =
			doca_regex_mempool_create(sizeof(struct doca_regex_match), regex->workq_mempool_size);
		if (new_ctx->matches_mp == NULL) {
			destroy_workq_ctx(new_ctx);
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	new_ctx->metadata_cache = aligned_calloc(ALIGN_CACHE_LINE, sizeof(struct fragment_metadata_cache));
	if (new_ctx->metadata_cache == NULL) {
		destroy_workq_ctx(new_ctx);
		return DOCA_ERROR_NO_MEMORY;
	}

	for (ii = 0; ii != JOB_ID_BITMASK_ARR_LEN; ++ii)
		new_ctx->metadata_cache->inuse_mask[ii] = 0xFFFFFFFFFFFFFFFFULL;

	if (regex->hardware_driver != NULL) {
		new_ctx->hardware_ctx.driver = regex->hardware_driver;
		new_ctx->hardware_ctx.queue = regex_driver_create_queue(regex->hardware_driver);
		if (new_ctx->hardware_ctx.queue == NULL) {
			destroy_workq_ctx(new_ctx);
			return DOCA_ERROR_NO_MEMORY;
		}

		new_ctx->driver_ping_pong[0] = &new_ctx->hardware_ctx;
	}

	if (regex->software_driver != NULL) {
		new_ctx->software_ctx.driver = regex->software_driver;
		new_ctx->software_ctx.queue = regex_driver_create_queue(regex->software_driver);
		if (new_ctx->software_ctx.queue == NULL) {
			destroy_workq_ctx(new_ctx);
			return DOCA_ERROR_NO_MEMORY;
		}

		new_ctx->driver_ping_pong[1] = &new_ctx->software_ctx;
	}

	if (new_ctx->driver_ping_pong[0] == NULL)
		new_ctx->driver_ping_pong[0] = new_ctx->driver_ping_pong[1];

	if (new_ctx->driver_ping_pong[1] == NULL)
		new_ctx->driver_ping_pong[1] = new_ctx->driver_ping_pong[0];

	*workq_ctx = new_ctx;

	return DOCA_SUCCESS;
}

/**
 * Find an empty job_metadata slot in the meta data array
 *
 * @param [in] ctx
 * Thread context to user
 *
 * @param [out] found_entry
 * Upon successful completion of this function *found_entry will point to a valid job_metadata.
 *
 * @param [out] found_entry_idx
 * Upon successful completion of this function *found_entry_idx be the index of the value pointed to by *found_entry.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_NO_MEMORY - cache full
 */
static doca_error_t
find_free_metadata_slot(struct regex_workq_context *ctx, struct fragment_metadata **found_entry,
			size_t *found_entry_idx)
{
	size_t entry_idx = 0;
	size_t ii;

	for (ii = 0; ii != JOB_ID_BITMASK_ARR_LEN && entry_idx == 0; ++ii)
		entry_idx = ffsll(ctx->metadata_cache->inuse_mask[ii]);

	if (entry_idx == 0)
		return DOCA_ERROR_NO_MEMORY;

	/* go back to the row where the slot was found, ii is incremented before the condition of entry_idx == 0 breaks
	 * the loop
	 */
	--ii;

	/* Take one off to switch from the 1-64 range of ffsl to 0-63 range for array index */
	--entry_idx;
	ctx->metadata_cache->inuse_mask[ii] &= ~(1ULL << entry_idx);
	entry_idx += ii * 64;
	*found_entry_idx = entry_idx;
	*found_entry = ctx->metadata_cache->data + entry_idx;

	return DOCA_SUCCESS;
}

/**
 * return a job_metadata slot back to the pool.
 *
 * @param [in] ctx
 * Thread context to user
 *
 * @param [in] idx
 * Index of the pool to release.
 */
static void
release_metadata_slot(struct regex_workq_context *ctx, size_t idx)
{
	uint64_t const mask_idx = idx / 64;
	uint64_t const offset = idx % 64;

	ctx->metadata_cache->inuse_mask[mask_idx] |= (1ULL << offset);
}

/**
 * Send a non huge job avoiding the overhead of configuring a hje meta-data storage
 *
 * @param [in] workq_ctx
 * The work queue context to use.
 *
 * @param [in] queue_ctx
 * The driver queue context to use.
 *
 * @param [in] job
 * The job to send.
 *
 * @param [in] allow_aggregation
 * Allow the device to aggregate jobs prior to sending.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - Invalid argument(s) provided.
 * DOCA_ERROR_NO_MEMORY - cache / queue is full
 */
static doca_error_t
send_single_buffer_job(struct regex_workq_context *workq_ctx, struct driver_queue_ctx *queue_ctx,
		       struct doca_regex_job_search const *job, bool allow_aggregation)
{
	size_t metadata_slot_idx;
	struct fragment_metadata *metadata_slot;
	doca_error_t result;

	result = find_free_metadata_slot(workq_ctx, &metadata_slot, &metadata_slot_idx);
	if (result != DOCA_SUCCESS)
		return result;

	metadata_slot->ref_id = job->base.user_data.u64;
	metadata_slot->ctrl.is_hje_job = 0;
	metadata_slot->buffer = job->buffer;
	metadata_slot->result = job->result;
	copy_group_ids(metadata_slot->rule_group_ids, job->rule_group_ids);

	struct regex_driver_job const driver_job = {
		.job_id = metadata_slot_idx,
		.rule_group_ids = {job->rule_group_ids[0], job->rule_group_ids[1], job->rule_group_ids[2],
				   job->rule_group_ids[3]},
		.buffer = job->buffer,
		.buffer_offset = 0,
		.buffer_bytes_to_send = priv_doca_buf_get_data_len(job->buffer),
		.allow_batching = allow_aggregation,
		.flags = job->base.flags,
	};

	result = regex_driver_queue_enqueue(queue_ctx->queue, &driver_job);

	if (result != DOCA_SUCCESS)
		release_metadata_slot(workq_ctx, metadata_slot_idx);

	return result;
}

/**
 * Set up the internal state to send and track progress of a job.
 *
 * @note DOES NOT send any data to the regex devices.
 *
 * @param [in] workq_ctx
 * The work queue context to use.
 *
 * @param [in] queue_ctx
 * The driver queue context to use.
 *
 * @param [in] job
 * The job to send.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE
 * DOCA_ERROR_NO_MEMORY
 */
static doca_error_t
prepare_hje_job(struct regex_workq_context *workq_ctx, struct driver_queue_ctx *queue_ctx,
		struct doca_regex_job_search const *job)
{
	struct sending_job *sj;
	struct receive_job *rj = doca_regex_mempool_get_obj(workq_ctx->jobs_mp);
	uint32_t num_hje_steps;
	uint32_t const engine_max_job_len = regex_driver_get_max_single_job_size(queue_ctx->driver);
	uint32_t const hje_window_size = priv_doca_buf_get_data_len(job->buffer) - engine_max_job_len;

	if (rj == NULL)
		return DOCA_ERROR_NO_MEMORY;

	sj = doca_regex_mempool_get_obj(workq_ctx->jobs_mp);
	if (sj == NULL) {
		doca_regex_mempool_put_obj(workq_ctx->jobs_mp, rj);
		return DOCA_ERROR_NO_MEMORY;
	}

	/* prepare sending job */
	sj->backlog = NULL;
	sj->driver_queue_ctx = queue_ctx;
	memcpy(&sj->request, job, sizeof(struct doca_regex_job_search));
	sj->outstanding_bytes = priv_doca_buf_get_data_len(sj->request.buffer);
	sj->request.base.user_data.u64 = doca_regex_mempool_index_of(workq_ctx->jobs_mp, rj);
	sj->send_chunk_size = engine_max_job_len;
	sj->send_offset = 0;
	sj->step_size = engine_max_job_len - workq_ctx->hje_overlap_size;
	num_hje_steps = hje_window_size / sj->step_size;
	if ((num_hje_steps * sj->step_size + engine_max_job_len) < priv_doca_buf_get_data_len(job->buffer))
		num_hje_steps += 1;

	sj->total_chunks = 1 + num_hje_steps;

	sj->outstanding_chunks = sj->total_chunks;

	/* configure receive response status / state */
	rj->correlation_id = job->base.user_data.u64;
	rj->orig_buffer = job->buffer;
	copy_group_ids(rj->orig_rule_group_ids, job->rule_group_ids);
	rj->outstanding_responses = sj->total_chunks;
	rj->next_job = NULL;
	rj->last_match = NULL;
	rj->has_done_fallback = false;

	/* Add sending job to the send queue_ctx */
	if (workq_ctx->sending_job == NULL)
		workq_ctx->sending_job = sj;
	else {
		struct sending_job *tail = workq_ctx->sending_job;

		while (tail->backlog != NULL)
			tail = tail->backlog;

		tail->backlog = sj;
	}

	return DOCA_SUCCESS;
}

/**
 * Send as much more of the prepared job(s) as possible.
 *
 * Once the current job is fully sent then this thread domain can accept another job.
 *
 * @param [in] ctx
 * Thread context.
 *
 * @param [in] allow_aggregation
 * Allow the device to aggregate jobs prior to sending.
 *
 * @return number of chunks sent this try
 */
static uint32_t
progress_sending_job(struct regex_workq_context *ctx, bool allow_aggregation)
{
	struct driver_queue_ctx *const queue_ctx = ctx->sending_job->driver_queue_ctx;
	struct sending_job *const sj = ctx->sending_job;
	uint32_t sent_count = 0;

	while (sj->outstanding_chunks != 0) {
		size_t metadata_slot_idx;
		struct fragment_metadata *metadata_slot;
		struct doca_regex_job_search const *job_request = &sj->request;

		if (find_free_metadata_slot(ctx, &metadata_slot, &metadata_slot_idx) != DOCA_SUCCESS)
			break;

		metadata_slot->ref_id = sj->request.base.user_data.u64;
		metadata_slot->ctrl.is_hje_job = 1;
		metadata_slot->ctrl.fragment_idx = sj->total_chunks - sj->outstanding_chunks;
		metadata_slot->hje_step_size = sj->step_size;
		metadata_slot->result = job_request->result;
		metadata_slot->result->matches_mempool = ctx->matches_mp;

		struct regex_driver_job const driver_job = {
			.job_id = metadata_slot_idx,
			.rule_group_ids = {job_request->rule_group_ids[0], job_request->rule_group_ids[1],
					   job_request->rule_group_ids[2], job_request->rule_group_ids[3]},
			.buffer = job_request->buffer,
			.buffer_offset = sj->send_offset,
			.buffer_bytes_to_send = sj->send_chunk_size,
			.allow_batching = allow_aggregation && sj->outstanding_chunks != 1,
		};

		if (regex_driver_queue_enqueue(queue_ctx->queue, &driver_job) != DOCA_SUCCESS) {
			release_metadata_slot(ctx, metadata_slot_idx);
			break;
		}

		++sent_count;
		sj->send_offset += sj->step_size;
		sj->outstanding_bytes -= sj->step_size;
		--sj->outstanding_chunks;
		if (sj->outstanding_chunks == 1 && sj->total_chunks > 1)
			sj->send_chunk_size = sj->outstanding_bytes;
	}

	if (sj->outstanding_chunks == 0 || sj->total_chunks == sj->outstanding_chunks) {
		/* Job is finished or nothing was enqueued, move to the next back-logged job if it exists or clear the
		 * send queue_ctx.
		 */
		if (sj->total_chunks == sj->outstanding_chunks) {
			/* nothing was enqueued, also free the rj */
			struct receive_job *const rj = (struct receive_job *)doca_regex_mempool_get_nth_element(
				ctx->jobs_mp, sj->request.base.user_data.u64);

			doca_regex_mempool_put_obj(ctx->jobs_mp, rj);
		}

		if (sj->backlog != NULL)
			ctx->sending_job = sj->backlog;
		else
			ctx->sending_job = NULL;

		doca_regex_mempool_put_obj(ctx->jobs_mp, sj);
	}

	return sent_count;
}

/**
 * Attempt to perform fall-back for a failed job from a hardware device to a software device.
 *
 * There a number of physical limits in the hardware, if these are exceeded then we can fall-back the job to software
 * which will be slow, but will not be artificially limited or constrained like the hardware is.
 *
 * @param [in] ctx
 * Thread context.
 *
 * @param [in] meta_idx
 * Index into the meta index for the job / fragment to fall-back. This meta data will identify the response as a
 * stand-alone response or as a fragment of a huge search. The appropriate fall-back steps will be taken based
 * on the type of job the response belongs to.
 *
 * @return
 * Was the fall-back successful? If not it should be processed as if fall-back was disabled to avoid losing a response.
 *
 */
static bool
try_fallback(struct regex_workq_context *ctx, size_t meta_idx)
{
	struct fragment_metadata *const meta = ctx->metadata_cache->data + meta_idx;

	if (meta->ctrl.is_hje_job) {
		struct receive_job *const rj =
			(struct receive_job *)doca_regex_mempool_get_nth_element(ctx->jobs_mp, meta->ref_id);

		if (rj == NULL)
			return false; /* should never happen */

		rj->outstanding_responses -= 1;

		if (rj->has_done_fallback) {
			release_metadata_slot(ctx, meta_idx);
			return true;
		}

		struct doca_regex_job_search const job = {
			.base = {.user_data = {.u64 = rj->correlation_id}},
			.rule_group_ids = {rj->orig_rule_group_ids[0], rj->orig_rule_group_ids[1],
					   rj->orig_rule_group_ids[2], rj->orig_rule_group_ids[3]},
			.buffer = rj->orig_buffer,
			.result = meta->result,
			.allow_batching = false,
		};

		doca_error_t const prep_fallback = prepare_hje_job(ctx, &ctx->software_ctx, &job);

		if (prep_fallback == DOCA_SUCCESS) {
			release_matches(ctx->matches_mp, &meta->result->matches);
			meta->result->detected_matches = 0;
			meta->result->num_matches = 0;
			rj->has_done_fallback = true;

			release_metadata_slot(ctx, meta_idx);
			return true;
		}

		if (rj->outstanding_responses == 0) {
			DOCA_LOG_ERR("Fallback of hje job %lu failed: %s", rj->correlation_id,
				     doca_get_error_string(prep_fallback));
		}

	} else {
		/* If doing fallback then do not attempt aggregation. */
		struct regex_driver_job const driver_job = {
			.job_id = meta_idx,
			.rule_group_ids = {meta->rule_group_ids[0], meta->rule_group_ids[1], meta->rule_group_ids[2],
					   meta->rule_group_ids[3]},
			.buffer = meta->buffer,
			.buffer_offset = 0,
			.buffer_bytes_to_send = priv_doca_buf_get_data_len(meta->buffer),
			.allow_batching = DO_NOT_AGGREGATE,
		};

		doca_error_t const enq_result = regex_driver_queue_enqueue(ctx->software_ctx.queue, &driver_job);

		if (enq_result == DOCA_SUCCESS)
			return true;

		DOCA_LOG_ERR("Fallback of job %lu failed. SW device rejected job (%s)", meta->ref_id,
			     doca_get_error_string(enq_result));
	}

	/* NOTE: returning false will drop us into normal response processing which will use the metadata so
	 * don't release it if we will return this way
	 */
	return false;
}

/**
 * Remove any duplicate matches from a response
 *
 * @note A duplicate match is one that occurs entirely within the overlap area of a buffer that is part of a huge job
 *
 * @param [in,out] response
 * Response to prune.
 *
 * @param [in] overlap_size
 * Number of bytes of overlap in the response.
 *
 * @param [in] matches_mp
 * Pool to return pruned matches to.
 *
 * @param [in] chunk_offset
 * Amount by which to adjust the start index of non duplicate jobs to shift them into the correct position in the range
 * of job instead of the range of the scanned fragment.
 */
static void
prune_duplicate_matches(struct regex_driver_response *response, uint32_t overlap_size,
			struct doca_regex_mempool *matches_mp, uint32_t const chunk_offset)
{
	struct doca_regex_match *prev = NULL;
	struct doca_regex_match *match = response->matches;
	/*
	 * Clear the matches list pointer this will be populated with the first non duplicate match and any subsequent
	 * non duplicate matches will be chained off it. We cannot assume the matches come back with any kind of
	 * ordering so we have to check them all.
	 */
	response->matches = NULL;
	while (match != NULL) {
		struct doca_regex_match *next = match->next;

		if (match->match_start + match->length > overlap_size) {
			/* Match value exceeds the overlap - keep it */
			if (prev != NULL)
				prev->next = match;
			else
				response->matches = match;

			/* adjust the start pointer to account for the fragments that came before this job */
			match->match_start += chunk_offset;

			match->next = NULL; /* Unlink the next one, there might not be another acceptable one */
			prev = match;
		} else {
			/* Match is contained within the overlap region - belongs to the previous huge job result
			 * set. Drop it
			 */
			response->num_matches -= 1;
			response->detected_matches -= 1;
			doca_regex_mempool_put_obj(matches_mp, match);
		}
		match = next;
	}
}

/**
 * Process a response that belongs to a huge job search
 *
 * @param [in] workq_ctx
 * Work queue context.
 *
 * @param [in] response
 * The response to process
 *
 * @param [in] overlap_size
 * The size of overlap that was configured for the device which executed this job.
 *
 * @param [in] meta_idx
 * Index into the meta data array.
 *
 * @param [out] event
 * The event to populate if this fragment is the final fragment of the huge job.
 *
 * @return
 * true if event was populated, false otherwise.
 */
static bool
process_hje_response(struct regex_workq_context *workq_ctx, struct regex_driver_response *response,
		     uint32_t overlap_size, size_t meta_idx, struct doca_event *event)
{
	struct fragment_metadata *metadata = workq_ctx->metadata_cache->data + meta_idx;
	struct receive_job *const rj =
		(struct receive_job *)doca_regex_mempool_get_nth_element(workq_ctx->jobs_mp, metadata->ref_id);

	if (rj == NULL) {
		release_matches(workq_ctx->matches_mp, &response->matches);
		return false;
	}

	if (rj->has_done_fallback)
		release_matches(workq_ctx->matches_mp, &response->matches);

	rj->outstanding_responses -= 1;

	if (response->matches != NULL) {
		struct doca_regex_match *tail;

		if (metadata->ctrl.fragment_idx != 0)
			prune_duplicate_matches(response, overlap_size, workq_ctx->matches_mp,
						metadata->ctrl.fragment_idx * metadata->hje_step_size);

		if (response->matches != NULL) {
			tail = match_tail(response->matches);

			if (rj->last_match == NULL)
				metadata->result->matches = response->matches;
			else
				rj->last_match->next = response->matches;

			rj->last_match = tail;
			response->matches = NULL;

			metadata->result->num_matches += response->num_matches;
		}
	}

	metadata->result->detected_matches += response->detected_matches;

	if (rj->outstanding_responses == 0 && rj->has_done_fallback == false) {
		event->type = DOCA_REGEX_JOB_SEARCH;
		event->user_data.u64 = rj->correlation_id;
		event->result.ptr = metadata->result;

		doca_regex_mempool_put_obj(workq_ctx->jobs_mp, rj);

		return true;
	}

	return false;
}

/**
 * Process a response received from a regex driver.
 *
 * Checks if this response is a fragment of a larger job, or a single fragment job and handles it appropriately. If the
 * response should be provided to the caller / user then an item is added to the responses output buffer.
 *
 * @param [in] workq_ctx
 * The work queue context.
 *
 * @param [in] response
 * Response to process
 *
 * @param [in] fallback_allowed
 * Flag to indicate if fall-back is permitted for failure responses
 *
 * @param [out] event
 * Output area for completed job responses.
 *
 * @return
 * true if event was populated, false otherwise.
 */
static bool
process_response(struct regex_workq_context *workq_ctx, struct regex_driver_response *response, bool fallback_allowed,
		 struct doca_event *event)

{
	size_t const meta_idx = response->id;
	bool populated_response = false;

	if ((response->status_flags & REGEX_DEVICE_FALLBACK_FLAGS) != 0) {
		if (fallback_allowed && try_fallback(workq_ctx, meta_idx)) {
			release_matches(workq_ctx->matches_mp, &response->matches);
			memset(response, 0, sizeof(*response));

			return false;
		}

		response->status_flags = DOCA_REGEX_STATUS_SEARCH_FAILED;

	} else {
		response->status_flags = 0;
	}

	if (workq_ctx->metadata_cache->data[meta_idx].ctrl.is_hje_job) {
		populated_response =
			process_hje_response(workq_ctx, response, workq_ctx->hje_overlap_size, meta_idx, event);
	} else {
		workq_ctx->metadata_cache->data[meta_idx].result->status_flags = response->status_flags;
		workq_ctx->metadata_cache->data[meta_idx].result->detected_matches = response->detected_matches;
		workq_ctx->metadata_cache->data[meta_idx].result->num_matches = response->num_matches;
		workq_ctx->metadata_cache->data[meta_idx].result->matches = response->matches;
		workq_ctx->metadata_cache->data[meta_idx].result->matches_mempool = response->matches_mempool;
		event->type = DOCA_REGEX_JOB_SEARCH;
		event->user_data.u64 = workq_ctx->metadata_cache->data[meta_idx].ref_id;
		event->result.ptr = workq_ctx->metadata_cache->data[meta_idx].result;

		populated_response = true;
	}

	release_metadata_slot(workq_ctx, meta_idx);

	return populated_response;
}

/**
 * Validator for regex not started yet
 */
static doca_error_t
validate_regex_is_stopped(struct doca_regex const *const regex)
{
	return regex->base.is_started ? DOCA_ERROR_IN_USE : DOCA_SUCCESS;
}

/**
 * Validator for regex is running (started)
 */
static doca_error_t
validate_regex_is_running(struct doca_regex const *const regex)
{
	return regex->base.is_started ? DOCA_SUCCESS : DOCA_ERROR_BAD_STATE;
}

/**
 * Validator for regex is running (started) and there is no installed software devices
 */
static doca_error_t
validate_regex_is_running_and_software_is_not_enabled(struct doca_regex const *const regex)
{
	if (regex->base.is_started && regex->software_driver == NULL)
		return DOCA_SUCCESS;

	if (!regex->base.is_started)
		return DOCA_ERROR_BAD_STATE;
	else
		return DOCA_ERROR_NOT_SUPPORTED;
}

/**
 * Check the regex instance is not NULL then lock the mutex to prevent concurrency issues
 * mutating the context.
 *
 * @param [in] regex
 * Regex instance
 *
 * @param [in] validator
 * Predicate which when not NULL is executed before returning to the caller. This allows us to do more complex
 * precondition checks while under the control of the mutex but while keeping cumbersome duplicate checks wrapped away
 * in one place.
 *
 * @Return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
null_check_lock_and_verify(struct doca_regex *const regex, regex_ctx_validator_fn validator)
{
	doca_error_t res;

	if (regex == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (pthread_mutex_lock(regex->mutex) != 0)
		return DOCA_ERROR_NO_LOCK;

	if (validator != NULL) {
		res = validator(regex);
		if (res != DOCA_SUCCESS) {
			pthread_mutex_unlock(regex->mutex);
			return res;
		}
	} else {
		res = DOCA_SUCCESS;
	}

	return res;
}

/**
 * Reinterpret the doca_ctx as a doca_regex, check its not null then lock the mutex to prevent concurrency issues
 * mutating the context.
 *
 * @param [in] ctx
 * Pointer to the context as a pointer to the base ctx type.
 *
 * @param [in] validator
 * Predicate which when not NULL is executed before returning to the caller. This allows us to do more complex
 * precondition checks while under the control of the mutex but while keeping cumbersome duplicate checks wrapped away
 * in one place.
 *
 * @param [out] regex_ctx
 * Reinterpreted pointer
 *
 * @Return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
reinterpret_cast_lock_and_verify(struct doca_ctx *ctx, regex_ctx_validator_fn validator, struct doca_regex **regex_ctx)
{
	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*regex_ctx = container_of(ctx, struct doca_regex, base);

	return null_check_lock_and_verify(*regex_ctx, validator);
}

/**
 * Common routine to create a workq context
 */
static doca_error_t
create_and_register_workq(struct doca_regex *regex, struct doca_workq *workq, bool const event_driven_support,
			  struct priv_doca_workq_progress_entry **poller_out)
{
	/* This struct looks identical for everyone so we just store a single static version and avoid mallocing one */
	static const struct priv_doca_event_handle_ops event_ops = {
		.arm_cb = regex_ctx_arm_event,
		.clear_cb = regex_ctx_clear_event,
	};

	struct regex_workq_context *workq_ctx;
	struct priv_doca_workq_progress_entry *poller;
	doca_error_t res;

	/* We can only support event driven mode with a hardware device so if this is not present or a software device
	 * is present then reject this as unsupported.
	 */
	if (event_driven_support && (regex->hardware_driver == NULL || regex->software_driver != NULL)) {
		res = DOCA_ERROR_NOT_SUPPORTED;
		goto err_exit;
	}

	res = create_workq_ctx(regex, &workq_ctx);
	if (res != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create regex work queue context");
		goto err_exit;
	}

	poller = aligned_calloc(ALIGN_CACHE_LINE, sizeof(struct priv_doca_workq_progress_entry));
	if (poller == NULL) {
		DOCA_LOG_ERR("Unable to create progress entry");
		res = DOCA_ERROR_NO_MEMORY;
		goto release_driver_ctx;
	}

	poller->progress_cb = regex_workq_progress;
	poller->data_priv.ptr = workq_ctx;
	res = priv_doca_workq_progress_entry_add(workq, poller);
	if (res != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to register progress entry");
		goto release_poller;
	}

	if (event_driven_support) {
		union doca_data const event_data = {.ptr = workq_ctx};

		res = regex_driver_queue_get_event_channel(workq_ctx->hardware_ctx.queue, &workq_ctx->channel,
							   &workq_ctx->channel_handle);
		if (res != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to create event channel");
			goto deregister_poller;
		}

		res = priv_doca_workq_event_handle_register(workq, workq_ctx->channel_handle, &event_ops, event_data);
		if (res != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Unable to register event handle to work queue.");
			goto deregister_poller;
		}
	}

	++regex->workq_count;

	workq_ctx->poller = poller;
	workq_ctx->workq = workq;
	*poller_out = poller;

	return DOCA_SUCCESS;

deregister_poller:
	priv_doca_workq_progress_entry_rm(workq, poller);
release_poller:
	free(poller);
release_driver_ctx:
	destroy_workq_ctx(workq_ctx);
err_exit:
	return res;
}

/**
 * Common routine to destroy a workq context
 */
static doca_error_t
destroy_and_deregister_workq(struct doca_regex *regex, struct doca_workq *workq,
			     struct priv_doca_workq_progress_entry *progress_entry)
{
	struct regex_workq_context *workq_ctx;
	doca_error_t result;

	if (workq == NULL || progress_entry == NULL || progress_entry->data_priv.ptr == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	workq_ctx = progress_entry->data_priv.ptr;

	if (workq_ctx->channel != NULL) {
		result = priv_doca_workq_event_handle_unregister(workq, workq_ctx->channel_handle);
		if (result != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Unable de-register event handle from workq.");
			return result;
		}

		workq_ctx->channel = NULL;
	}

	/* Remove the progress entry - this will unregister the progress_cb.*/
	result = priv_doca_workq_progress_entry_rm(workq, progress_entry);
	if (result != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable de-register progress entry from workq.");
		return result;
	}

	/* Destroy workq ctx and poller */
	free(progress_entry);
	destroy_workq_ctx(workq_ctx);

	/* Unregister with the regex ctx */
	--(regex->workq_count);

	return DOCA_SUCCESS;
}

/**********************************************************************************************************************
 *
 * Virtual functions
 *
 *********************************************************************************************************************/

/**
 * Attach a device to the regex context.
 *
 * @note
 * Any existing device driver of the same category (software or hardware device) will be released and replaced by the
 * new one.
 *
 * @param [in] ctx
 * The regex context.
 *
 * @param [in] dev
 * The device to attach to regex.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - Invalid context
 * DOCA_ERROR_NO_LOCK - Failed to lock mutex
 * DOCA_ERROR_NOT_SUPPORTED - Device not supported
 * DOCA_ERROR_NO_MEMORY - Unable to allocate necessary resource(s)
 */
static doca_error_t
regex_ctx_dev_add(struct doca_ctx *ctx, struct doca_dev *dev)
{
	struct doca_regex *regex_ctx;
	doca_error_t result;

	result = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_stopped, &regex_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	result = install_driver(regex_ctx, dev);

	pthread_mutex_unlock(regex_ctx->mutex);

	return result;
}

/**
 * Detach a device from the regex context.
 *
 * @param [in] ctx
 * The regex context.
 *
 * @param [in] dev
 * The device to detach from regex.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - Invalid context, device or device not installed
 * DOCA_ERROR_NO_LOCK - Failed to lock mutex
 * DOCA_ERROR_NOT_SUPPORTED - Device not supported
 */
static doca_error_t
regex_ctx_dev_rm(struct doca_ctx *ctx, struct doca_dev *dev)
{
	struct doca_regex *regex_ctx;
	doca_error_t result;

	result = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_stopped, &regex_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	result = uninstall_driver(regex_ctx, dev);

	pthread_mutex_unlock(regex_ctx->mutex);

	return result;
}

/**
 * Validate settings and start the regex context.
 *
 * @note
 * A regex context must be started before any work queues may be attached as we cannot safely update the work
 * queue contexts which may be used by another thread and placing locking mechanisms in the hot path is not
 * acceptable so we must ensure everything is valid in this context first so that when creating work queues it
 * can copy / use / cache any values from the context with confidence that they are correct and essentially
 * const while the context is running.
 *
 * @param [in] ctx
 * The regex context.
 *
 * @return
 * DOCA_SUCCES
 * DOCA_ERROR_INVALID_VALUE - Invalid context / invalid configuration
 * DOCA_ERROR_NO_LOCK - Failed to lock mutex
 * DOCA_ERROR_NO_MEMORY - Unable to allocate necessary resource(s)
 * DOCA_ERROR_NOT_SUPPORTED - failed to get PDN / Device rejected rules data
 *
 *
 */
static doca_error_t
regex_ctx_start(struct doca_ctx *ctx)
{
	struct doca_regex *regex_ctx;
	doca_error_t result;

	result = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_stopped, &regex_ctx);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to start doca regex. Unable to acquire lock: %s", doca_get_error_string(result));
		return result;
	}

	result = validate_configuration(regex_ctx);
	if (result != DOCA_SUCCESS) {
		pthread_mutex_unlock(regex_ctx->mutex);
		return result;
	}

	result = program_drivers(regex_ctx);
	if (result != DOCA_SUCCESS) {
		pthread_mutex_unlock(regex_ctx->mutex);
		return result;
	}

	pthread_mutex_unlock(regex_ctx->mutex);

	return DOCA_SUCCESS;
}

/**
 * Stop the regex context.
 *
 * @note
 * Any attached work queue must be detached before the context will allow itself to return to the stopped state.
 *
 * @note
 * After stopping the context can be reconfigured or destroyed.
 *
 * @param [in] ctx
 * The regex context.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - Invalid context
 * DOCA_ERROR_NO_LOCK - Failed to lock mutex
 * DOCA_ERROR_IN_USE - One or more workq objects are still attached.
 */
static doca_error_t
regex_ctx_stop(struct doca_ctx *ctx)
{
	struct doca_regex *regex_ctx;
	doca_error_t result;

	result = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_running, &regex_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	if (regex_ctx->workq_count == 0)
		result = DOCA_SUCCESS;
	else
		result = DOCA_ERROR_IN_USE;

	pthread_mutex_unlock(regex_ctx->mutex);

	return result;
}

/**
 * Submit a regex job
 *
 * @param [in] data_priv
 * Opaque pointer to data created by the engine during regex_ctx_workq_add and emitted through its data_priv out
 * param. The regex instance can reinterpret this as the appropriate object which will be a
 * priv_doca_workq_progress_entry which in turn holds a regex_workq_context which can be used to create WQEs and
 * submit to the doca device.
 *
 * @param [in] job
 * Job as passed to: doca_workq_submit. This should have the doca_job::type set to an appropriate value to
 * suggest this is a regex JOB. Beyond that there is not much we can actually validate and we have to blindly
 * typecast this object to a doca_regex_job and hope thats what the user actually provided to us.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - Invalid job, job data or result
 * DOCA_ERROR_NO_MEMORY
 */
static doca_error_t
regex_ctx_job_submit(union doca_data data_priv, struct doca_job const *job)
{
	struct doca_regex_job_search const *regex_job;
	struct regex_workq_context *workq_ctx;
	struct priv_doca_workq_progress_entry *poller;

	regex_job = container_of(job, struct doca_regex_job_search const, base);
	poller = (struct priv_doca_workq_progress_entry *)data_priv.ptr;
	workq_ctx = (struct regex_workq_context *)poller->data_priv.ptr;

	if (regex_job == NULL || regex_job->buffer == NULL || regex_job->result == NULL ||
	    job->type != DOCA_REGEX_JOB_SEARCH || priv_doca_buf_get_data_len(regex_job->buffer) == 0)
		return DOCA_ERROR_INVALID_VALUE;

	if (workq_ctx->sending_job != NULL) {
		progress_sending_job(workq_ctx, true);
		if (workq_ctx->sending_job != NULL)
			return DOCA_ERROR_NO_MEMORY;
	}

	struct driver_queue_ctx *const queue_ctx = select_driver_and_queue(workq_ctx, regex_job);

	/* ensure result is empty */
	memset(regex_job->result, 0, sizeof(struct doca_regex_search_result));

	if (priv_doca_buf_get_data_len(regex_job->buffer) <= regex_driver_get_max_single_job_size(queue_ctx->driver))
		return send_single_buffer_job(workq_ctx, queue_ctx, regex_job, regex_job->allow_batching);

	/* Highest priority match not supported for huge jobs */
	if (regex_job->base.flags & DOCA_REGEX_SEARCH_JOB_FLAG_HIGHEST_PRIORITY_MATCH)
		return DOCA_ERROR_NOT_SUPPORTED;

	doca_error_t const prepare_result = prepare_hje_job(workq_ctx, queue_ctx, regex_job);

	if (prepare_result != DOCA_SUCCESS)
		return prepare_result;

	return progress_sending_job(workq_ctx, regex_job->allow_batching) == 0 ? DOCA_ERROR_NO_MEMORY : DOCA_SUCCESS;
}

/**
 * Retrieve a regex event (response)
 *
 * @param [in] data_priv
 * Opaque pointer to data created by the engine during regex_ctx_workq_add and emitted through its data_priv out
 * param. The regex instance can reinterpret this is the appropriate object which will be a
 * priv_doca_workq_progress_entry which in turn holds a regex_workq_context which can be used to receive CQEs to get
 * responses from the device.
 *
 * @param [in] event
 * Event structure to be populated as appropriate when this function returns a success code.
 *
 * @return
 * DOCA_SUCCESS - A result has been retrieved
 * DOCA_ERROR_INVALID_VALUE - Invalid event
 * DOCA_ERROR_AGAIN - No result could be returned at this time
 */
static doca_error_t
regex_workq_progress(union doca_data data_priv, struct doca_event *event)
{
	struct regex_workq_context *workq_ctx;
	bool populated_response = false;
	size_t ii;

	workq_ctx = (struct regex_workq_context *)data_priv.ptr;

	if (event == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (workq_ctx->response_cache_read_idx == workq_ctx->response_cache.nb_populated_responses) {
		workq_ctx->response_cache.nb_populated_responses = 0;
		workq_ctx->response_cache_read_idx = 0;

		doca_error_t const result = fill_response_cache(workq_ctx);

		if (result != DOCA_SUCCESS)
			return result;
	}

	bool const fallback_allowed = (workq_ctx->sw_fallback_enabled) && (workq_ctx->software_ctx.driver != NULL) &&
				      (workq_ctx->driver_ping_pong[PING_PONG_LAST_USED] == &workq_ctx->hardware_ctx);

	for (ii = workq_ctx->response_cache_read_idx; ii != workq_ctx->response_cache.nb_populated_responses; ++ii) {
		populated_response =
			process_response(workq_ctx, workq_ctx->response_cache.responses + ii, fallback_allowed, event);
		++workq_ctx->response_cache_read_idx;

		if (populated_response)
			break;
	}

	/* To ensure the user can enqueue one job then poll dequeue until they get a response we need to try to progress
	 * the sending job each time after we lift any available responses
	 */
	if (workq_ctx->sending_job != NULL)
		progress_sending_job(workq_ctx, DO_NOT_AGGREGATE);

	return populated_response ? DOCA_SUCCESS : DOCA_ERROR_AGAIN;
}

/**
 * Attach a work queue to the regex instance.
 *
 * @param [in] ctx
 * The regex context.
 *
 * @param [in] workq
 * The work queue to attach.
 *
 * @param [out] data_priv
 * Opaque pointer which the engine must populate with a priv_doca_workq_progress_entry which is allocated by the
 * regex context. This will hold the regex_workq_context, and progress callbacks.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
regex_ctx_polled_workq_add(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data *data_priv)
{
	struct priv_doca_workq_progress_entry *poller = NULL;
	struct doca_regex *regex;
	doca_error_t res;

	res = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_running, &regex);
	if (res != DOCA_SUCCESS)
		return res;

	res = create_and_register_workq(regex, workq, /* with event driven support */ false, &poller);
	data_priv->ptr = poller;

	pthread_mutex_unlock(regex->mutex);

	return res;
}

/**
 * Arm work queue context for an event. Requesting an event on next CQE received.
 *
 * @details Callback to be used once application requests an event from the work queue while working in event driven
 * mode. Callback will be trigerred once application calls doca_workq_event_handle_arm().
 *
 * @param [in] event_data
 * Opaque pointer to a priv_doca_comp_events_channel belonging to an event workq context. The channel can be used for
 * arming of events. The pointer was previously provided upon registering a doca_event_handle_t to the work queue
 * using priv_doca_workq_event_handle_register().
 *
 * @return
 * DOCA_SUCCESS
 */
static doca_error_t
regex_ctx_arm_event(union doca_data const event_data)
{
	struct regex_workq_context *workq_ctx = (struct regex_workq_context *)event_data.ptr;

	priv_doca_comp_events_channel_arm(workq_ctx->channel);

	/* Right now only consider the hardware queue as software has no event channel support. */
	if (regex_driver_queue_is_empty(workq_ctx->hardware_ctx.queue)) {
		/* No CQEs left to be processed, remove ourselves from the set of contexts to be polled as part of
		 * progress retrieve. regex_ctx_clear_event will be called when the event channel sees an event for this
		 * queue
		 */
		return priv_doca_workq_progress_entry_state_update(workq_ctx->workq, workq_ctx->poller,
								   PRIV_DOCA_WORKQ_PROGRESS_SUSPENDED);
	}

	return DOCA_SUCCESS;
}

/**
 * Clear events retrieved for the work queue context
 *
 * @details Callback to be used once application requests clearing of events received on the work queue,
 * using doca_workq_event_handle_clear().
 * The method's goal is to change state of the event handle from 'ready to read' to 'pending'.
 *
 * @param [in] event_data
 * Opaque pointer to a priv_doca_comp_events_channel belonging to an event workq context. The channel can be used for
 * clearing of events. The pointer was previously provided upon registering a doca_event_handle_t to the work queue
 * using priv_doca_workq_event_handle_register().
 *
 * DOCA_SUCCESS - successfully handled completion events.
 * DOCA_ERROR_INVALID_INPUT - channel is NULL.
 * DOCA_ERROR_OPERATING_SYSTEM - call to device driver failed.
 */
static doca_error_t
regex_ctx_clear_event(union doca_data const event_data)
{
	struct regex_workq_context *workq_ctx = (struct regex_workq_context *)event_data.ptr;

	doca_error_t const ret = priv_doca_workq_progress_entry_state_update(workq_ctx->workq, workq_ctx->poller,
									     PRIV_DOCA_WORKQ_PROGRESS_RUNNING);
	if (ret != DOCA_SUCCESS)
		return ret;

	priv_doca_comp_events_channel_arm(workq_ctx->channel);

	return DOCA_SUCCESS;
}

/**
 * Detach a work queue from the regex context.
 *
 * @param [in] ctx
 *
 * @param [in] workq
 * The work queue to detach.
 *
 * @param [out] data_priv
 * Opaque pointer which will refer to the value output during regex_ctx_workq_add in its data_priv out param.
 * The regex instance MUST release all resources held by this object. (Undo everything done during
 * regex_ctx_workq_add).
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
regex_ctx_polled_workq_rm(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data data_priv)
{
	struct doca_regex *regex;
	doca_error_t res;

	res = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_running, &regex);
	if (res != DOCA_SUCCESS)
		return res;

	res = destroy_and_deregister_workq(regex, workq, (struct priv_doca_workq_progress_entry *)data_priv.ptr);

	pthread_mutex_unlock(regex->mutex);

	return res;
}

/**
 * Attach a work queue to the regex instance.
 *
 * @param [in] ctx
 * The regex context.
 *
 * @param [in] workq
 * The work queue to attach.
 *
 * @param [out] data_priv
 * Opaque pointer which the engine must populate with a priv_doca_workq_progress_entry which is allocated by the
 * regex context. This will hold the regex_workq_context, and progress callbacks.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
regex_ctx_event_driven_workq_add(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data *data_priv)
{

	struct priv_doca_workq_progress_entry *poller = NULL;
	struct doca_regex *regex;
	doca_error_t res;

	res = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_running_and_software_is_not_enabled, &regex);
	if (res != DOCA_SUCCESS)
		return res;

	res = create_and_register_workq(regex, workq, /* with event driven support */ true, &poller);
	data_priv->ptr = poller;

	pthread_mutex_unlock(regex->mutex);

	return res;
}

/**
 * Detach a work queue from the regex context.
 *
 * @param [in] ctx
 *
 * @param [in] workq
 * The work queue to detach.
 *
 * @param [out] data_priv
 * Opaque pointer which will refer to the value output during regex_ctx_workq_add in its data_priv out param.
 * The regex instance MUST release all resources held by this object. (Undo everything done during
 * regex_ctx_workq_add).
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
regex_ctx_event_driven_workq_rm(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data data_priv)
{
	struct doca_regex *regex;
	doca_error_t res;

	res = reinterpret_cast_lock_and_verify(ctx, validate_regex_is_running, &regex);
	if (res != DOCA_SUCCESS)
		return res;

	res = destroy_and_deregister_workq(regex, workq, (struct priv_doca_workq_progress_entry *)data_priv.ptr);

	pthread_mutex_unlock(regex->mutex);

	return res;
}

/**********************************************************************************************************************
 *
 * Public API
 *
 *********************************************************************************************************************/

doca_error_t
doca_regex_create(struct doca_regex **regex)
{
	struct doca_regex *inst;

	if (regex == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	inst = aligned_calloc(ALIGN_CACHE_LINE, sizeof(struct doca_regex));
	if (inst == NULL)
		return DOCA_ERROR_NO_MEMORY;

	if (priv_doca_ctx_init(&(inst->base), true) != DOCA_SUCCESS) {
		free(inst);
		return DOCA_ERROR_NO_MEMORY;
	}

	inst->mutex = aligned_calloc(ALIGN_CACHE_LINE, sizeof(pthread_mutex_t));
	if (inst->mutex == NULL) {
		priv_doca_ctx_destroy(&(inst->base));
		free(inst);
		return DOCA_ERROR_NO_MEMORY;
	}

	int const ret = pthread_mutex_init(inst->mutex, NULL);

	if (ret != 0) {
		free(inst->mutex);
		priv_doca_ctx_destroy(&(inst->base));
		free(inst);
		return DOCA_ERROR_NO_MEMORY;
	}

	inst->base.ops.start = regex_ctx_start;
	inst->base.ops.stop = regex_ctx_stop;
	inst->base.ops.dev_add = regex_ctx_dev_add;
	inst->base.ops.dev_rm = regex_ctx_dev_rm;
	inst->base.ops.workq_add = regex_ctx_polled_workq_add;
	inst->base.ops.workq_rm = regex_ctx_polled_workq_rm;
	inst->base.ops.event_workq_add = regex_ctx_event_driven_workq_add;
	inst->base.ops.event_workq_rm = regex_ctx_event_driven_workq_rm;
	inst->base.ops.job_submit = regex_ctx_job_submit;

	set_default_configuration_values(inst);

	*regex = inst;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_destroy(struct doca_regex *regex)
{
	if (regex == NULL)
		return DOCA_SUCCESS;

	if (pthread_mutex_lock(regex->mutex) != 0)
		return DOCA_ERROR_NO_LOCK;

	if (regex->workq_count != 0) {
		pthread_mutex_unlock(regex->mutex);
		return DOCA_ERROR_IN_USE;
	}

	/* Point of no return. If any other thread is waiting on the mutex they are going to have a bad time.
	 * Not much we can do about that. The user must ensure no thread can be attempting to access or will
	 * access this object in the future.
	 */
	priv_doca_ctx_destroy(&(regex->base));

	/* users are lazy, if they still have drivers attached, destroy those */
	if (regex->hardware_driver != NULL)
		regex_driver_destroy(regex->hardware_driver);

	if (regex->software_driver != NULL)
		regex_driver_destroy(regex->software_driver);

	/* Release stored rules */
	store_rules(NULL, 0, &(regex->hw_binary_rules));
	store_rules(NULL, 0, &(regex->sw_binary_rules));
	store_rules(NULL, 0, &(regex->hw_uncompiled_rules));
	store_rules(NULL, 0, &(regex->sw_uncompiled_rules));

	/* Release, then cleanup mutex. */
	pthread_mutex_unlock(regex->mutex);
	pthread_mutex_destroy(regex->mutex);
	free(regex->mutex);

	/* Release the object */
	free(regex);

	return DOCA_SUCCESS;
}

struct doca_ctx *
doca_regex_as_ctx(struct doca_regex *regex)
{
	return &(regex->base);
}

doca_error_t
doca_regex_set_hardware_compiled_rules(struct doca_regex *regex, void const *rules_data, size_t rules_data_size)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	res = store_rules(rules_data, rules_data_size, &regex->hw_binary_rules);

	pthread_mutex_unlock(regex->mutex);

	return res;
}

doca_error_t
doca_regex_set_hardware_uncompiled_rules(struct doca_regex *regex, void const *rules_data, size_t rules_data_size)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	res = store_rules(rules_data, rules_data_size, &regex->hw_uncompiled_rules);

	pthread_mutex_unlock(regex->mutex);

	return res;
}

doca_error_t
doca_regex_set_software_compiled_rules(struct doca_regex *regex, void const *rules_data, size_t rules_data_size)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	res = store_rules(rules_data, rules_data_size, &regex->sw_binary_rules);

	pthread_mutex_unlock(regex->mutex);

	return res;
}

doca_error_t
doca_regex_set_software_uncompiled_rules(struct doca_regex *regex, void const *rules_data, size_t rules_data_size)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	res = store_rules(rules_data, rules_data_size, &regex->sw_uncompiled_rules);

	pthread_mutex_unlock(regex->mutex);

	return res;
}

doca_error_t
doca_regex_set_workq_matches_memory_pool_size(struct doca_regex *regex, uint32_t pool_size)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	regex->workq_mempool_size = pool_size;

	pthread_mutex_unlock(regex->mutex);

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_set_huge_job_emulation_overlap_size(struct doca_regex *regex, uint16_t nb_overlap_bytes)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	regex->hje_overlap_size = nb_overlap_bytes;

	pthread_mutex_unlock(regex->mutex);

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_set_small_job_offload_threshold(struct doca_regex *regex, uint16_t threshold)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	regex->small_job_offload_threshold = threshold;

	pthread_mutex_unlock(regex->mutex);

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_set_failed_job_fallback_enabled(struct doca_regex *regex, bool enabled)
{
	doca_error_t res = null_check_lock_and_verify(regex, validate_regex_is_stopped);

	if (res != DOCA_SUCCESS)
		return res;

	regex->sw_fallback_enabled = enabled;

	pthread_mutex_unlock(regex->mutex);

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_get_hardware_compiled_rules(struct doca_regex const *regex, void **rules_data, size_t *rules_data_size)
{
	if (regex == NULL || rules_data == NULL || rules_data_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	return clone_blob(&regex->hw_binary_rules, rules_data, rules_data_size);
}

doca_error_t
doca_regex_get_hardware_uncompiled_rules(struct doca_regex const *regex, void **rules_data, size_t *rules_data_size)
{
	if (regex == NULL || rules_data == NULL || rules_data_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	return clone_blob(&regex->hw_uncompiled_rules, rules_data, rules_data_size);
}

doca_error_t
doca_regex_get_software_compiled_rules(struct doca_regex const *regex, void **rules_data, size_t *rules_data_size)
{
	if (regex == NULL || rules_data == NULL || rules_data_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	return clone_blob(&regex->sw_binary_rules, rules_data, rules_data_size);
}

doca_error_t
doca_regex_get_software_uncompiled_rules(struct doca_regex const *regex, void **rules_data, size_t *rules_data_size)
{
	if (regex == NULL || rules_data == NULL || rules_data_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	return clone_blob(&regex->sw_uncompiled_rules, rules_data, rules_data_size);
}

doca_error_t
doca_regex_get_workq_matches_memory_pool_size(struct doca_regex const *regex, uint32_t *pool_size)
{
	if (regex == NULL || pool_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*pool_size = regex->workq_mempool_size;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_get_huge_job_emulation_overlap_size(struct doca_regex const *regex, uint16_t *nb_overlap_bytes)
{
	if (regex == NULL || nb_overlap_bytes == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*nb_overlap_bytes = regex->hje_overlap_size;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_get_small_job_offload_threshold(struct doca_regex const *regex, uint16_t *threshold)
{
	if (regex == NULL || threshold == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*threshold = regex->small_job_offload_threshold;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_get_failed_job_fallback_enabled(struct doca_regex const *regex, bool *enabled)
{
	if (regex == NULL || enabled == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*enabled = regex->sw_fallback_enabled != 0;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_is_supported(struct doca_devinfo const *devinfo)
{
	struct regex_driver_info info;
	doca_error_t res;

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	res = get_devinfo_caps(devinfo, &info);
	if (res != DOCA_SUCCESS)
		return res;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_get_hardware_supported(struct doca_devinfo const *devinfo)
{
	struct regex_driver_info info;
	doca_error_t res;

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	res = get_devinfo_caps(devinfo, &info);
	if (res != DOCA_SUCCESS)
		return res;

	return info.supports_hardware_acceleration ? DOCA_SUCCESS : DOCA_ERROR_NOT_SUPPORTED;
}

doca_error_t
doca_regex_get_software_supported(struct doca_devinfo const *devinfo)
{
	struct regex_driver_info info;
	doca_error_t res;

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	res = get_devinfo_caps(devinfo, &info);
	if (res != DOCA_SUCCESS)
		return res;

	return info.supports_hardware_acceleration ? DOCA_ERROR_NOT_SUPPORTED : DOCA_SUCCESS;
}

doca_error_t
doca_regex_get_maximum_job_size(struct doca_devinfo const *devinfo, uint64_t *max_job_len)
{
	struct regex_driver_info info;
	doca_error_t res;

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	res = get_devinfo_caps(devinfo, &info);
	if (res != DOCA_SUCCESS)
		return res;

	*max_job_len = info.max_hje_job_length;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_get_maximum_non_huge_job_size(struct doca_devinfo const *devinfo, uint64_t *max_job_len)
{
	struct regex_driver_info info;
	doca_error_t res;

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	res = get_devinfo_caps(devinfo, &info);
	if (res != DOCA_SUCCESS)
		return res;

	*max_job_len = info.max_single_job_length;

	return DOCA_SUCCESS;
}

doca_error_t
doca_regex_job_get_supported(struct doca_devinfo const *devinfo, enum doca_regex_job_types job_type)
{
	struct regex_driver_info info;
	doca_error_t res;

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	res = get_devinfo_caps(devinfo, &info);
	if (res != DOCA_SUCCESS)
		return res;

	res = DOCA_ERROR_NOT_SUPPORTED;

	switch (job_type) {
	case DOCA_REGEX_JOB_SEARCH:
		res = DOCA_SUCCESS;
		break;
	}

	return res;
}

doca_error_t
doca_regex_search_job_flag_get_highest_priority_match_supported(struct doca_devinfo const *devinfo)
{
	struct regex_driver_info info;
	doca_error_t res;

	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	res = get_devinfo_caps(devinfo, &info);
	if (res != DOCA_SUCCESS)
		return res;

	if (info.supports_hardware_acceleration)
		return DOCA_SUCCESS;

	return DOCA_ERROR_NOT_SUPPORTED;
}

doca_error_t
doca_regex_search_job_flag_get_stop_on_any_match_supported(struct doca_devinfo const *devinfo)
{
	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	return DOCA_ERROR_NOT_SUPPORTED;
}
