/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "doca_regex_mempool.h"

#include <errno.h>
#include <stdint.h>
#include <stdlib.h>

#define MEMPOOL_NAME_LEN 32

struct elem {
	struct elem *prev;
	struct elem *next;
};

struct doca_regex_mempool {
	struct elem *cont_begin;
	struct elem *cont_end;
	struct elem *free_list;
	uint32_t num_elems;
	uint32_t elem_size;
};

struct doca_regex_mempool *
doca_regex_mempool_create(size_t elem_size, size_t nb_elems)
{
	if (elem_size == 0 || nb_elems == 0)
		return NULL;

	struct doca_regex_mempool *obj = malloc(sizeof(struct doca_regex_mempool));

	if (obj == NULL)
		return NULL;

	size_t aligned_obj_size = sizeof(struct elem) + elem_size;

	if (aligned_obj_size % 8 != 0)
		aligned_obj_size = (1 + (aligned_obj_size / 8)) * 8;

	obj->elem_size = aligned_obj_size;
	obj->cont_begin = malloc(aligned_obj_size * nb_elems);
	if (obj->cont_begin == NULL) {
		free(obj);
		return NULL;
	}

	uint8_t *iter;
	struct elem *prev;

	/* init pool fields */
	obj->num_elems = nb_elems;
	obj->cont_end = (struct elem *)(((uint8_t *)obj->cont_begin) + (obj->num_elems * obj->elem_size));

	/* build initial free list */
	iter = (uint8_t *)obj->cont_begin;
	prev = (struct elem *)iter;
	prev->next = NULL;
	prev->prev = NULL;
	obj->free_list = prev;
	iter += obj->elem_size;
	for (; (struct elem *)iter != obj->cont_end; iter += obj->elem_size) {
		struct elem *const elem = (struct elem *)iter;

		elem->next = NULL;
		prev->next = elem;
		elem->prev = prev;
		prev = elem;
	}

	return obj;
}

void
doca_regex_mempool_destroy(struct doca_regex_mempool *pool)
{
	if (pool == NULL)
		return;

	free(pool->cont_begin);
#if DEBUG_DOCA_REGEX_MEMPOOL
	/* mark these so its clear this is a deleted object if we had a stale pointer */
	pool->cont_begin = 0xDEADBEEF;
	pool->cont_end = 0xDEADBEEF;
	pool->free_list = 0xDEADBEEF;
#endif /* DEBUG_DOCA_REGEX_MEMPOOL */
	free(pool);
}

void *
doca_regex_mempool_get_obj(struct doca_regex_mempool *pool)
{
	if (pool == NULL)
		return NULL;

	struct elem *elem = pool->free_list;

	if (elem != NULL) {
		pool->free_list = elem->next;
		if (pool->free_list != NULL)
			pool->free_list->prev = NULL;

		elem->prev = NULL;

		return (void *)(((uint8_t *)elem) + sizeof(struct elem));
	}

	return NULL;
}

void
doca_regex_mempool_put_obj(struct doca_regex_mempool *pool, void *obj)
{
	if (pool == NULL || obj <= (void *)pool->cont_begin || obj >= (void *)pool->cont_end) {
		/* invalid access */
		return;
	}

	struct elem *const elem = (struct elem *)(((uint8_t *)obj) - sizeof(struct elem));

	/* return to free list */
	elem->prev = NULL;
	elem->next = pool->free_list;
	if (elem->next != NULL)
		elem->next->prev = elem;

	pool->free_list = elem;
}

int
doca_regex_mempool_index_of(struct doca_regex_mempool const *pool, void const *obj)
{
	if (pool == NULL || obj <= (void *)pool->cont_begin || obj >= (void *)pool->cont_end) {
		/* invalid access */
		return -EINVAL;
	}

	uint8_t const *elem_begin = (((uint8_t const *)obj) - sizeof(struct elem));
	ptrdiff_t distance = (elem_begin - (uint8_t const *)pool->cont_begin);

	return distance / pool->elem_size;
}

void *
doca_regex_mempool_get_nth_element(struct doca_regex_mempool *pool, size_t n)
{
	if (pool == NULL || n >= pool->num_elems) {
		/* invalid access */
		return NULL;
	}

	return (((uint8_t *)pool->cont_begin) + (pool->elem_size * n) + sizeof(struct elem));
}
