/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "bf2_driver.h"

#include <endian.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <regex-compiler.h>

#include <doca_log.h>
#include <doca_rdma.h>
#include <doca_regex.h>
#include <doca_types.h>
#include <priv_doca_buf.h>
#include <priv_doca_cq.h>
#include <priv_doca_dev.h>
#include <priv_doca_os_abstraction_memory.h>
#include <priv_doca_uar.h>
#include <private/common_defines.h>

#include <impl/drivers/regex_driver.h>
#include <impl/regex_defines.h>

#include "bf2_regex_headers.h"
#include "mlx5_common.h"
#include "mlx5_prm.h"

DOCA_LOG_REGISTER(BF2_DRIVER);

/**********************************************************************************************************************
 *
 * #defines and macros
 *
 *********************************************************************************************************************/

#define ALIGN_4K 4096
#define ALIGN_64 64

#define PRIV_MLX5_REGEX_NUM_WQE_PER_PAGE (4096 / 64)
/* UMR mode requires 4 times the descriptors so convert to log value + 2. */
#define PRIV_MLX5_REGEX_WQE_LOG_NUM(log_desc) ((log_desc) + 2)

/* Access flags from the verb.h. */
#define PRIV_UMEM_ACCESS_FLAGS (IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ)

#define PRIV_MLX5_BIT32(nr) (UINT32_C(1) << (nr))
#define PRIV_MLX5_REGEX_METADATA_SIZE UINT32_C(64)
#define PRIV_MLX5_REGEX_MAX_OUTPUT PRIV_MLX5_BIT32(11)
#define PRIV_MLX5_REGEX_WQE_METADATA_OFFSET 16

/* The maximum KLMs can be added to one UMR indirect mkey. */
#define PRIV_MLX5_REGEX_MAX_KLM_NUM 128
/* The KLM array size for one job. */
#define PRIV_MLX5_REGEX_KLMS_SIZE ((PRIV_MLX5_REGEX_MAX_KLM_NUM) * sizeof(struct priv_mlx5_klm))

#define PRIV_MLX5_REGEX_WQE_SCATTER_OFFSET 48
#define PRIV_MLX5_REGEX_METADATA_OFF 32
#define PRIV_MLX5_REGEX_UMR_WQE_SIZE 192

#define PRIV_MLX5_OPCODE_MMO 0x2f
#define PRIV_MLX5_OPC_MOD_MMO_REGEX 0x4
#define PRIV_MLX5_REGEX_WQE_GATHER_OFFSET 32

#define PRIV_MLX5_REGEX_MAX_WQE_INDEX 0xffff
#define PRIV_MLX5_REGEX_UMR_QP_PI_IDX(pi, ops) (((pi) + (ops)) & (PRIV_MLX5_REGEX_MAX_WQE_INDEX >> 2))

#define PRIV_MLX5_REGEX_RESP_SZ 8

#define BF2_CQE_OWNER(op_own) ((op_own)&MLX5_CQE_OWNER_MASK)
#define BF2_CQE_OPCODE(op_own) (((op_own)&0xf0) >> 4)

#define COMBINED_ROF_HEADER_SIZE 16
#define COMBINED_ROF_V1_NUM_BLOCKS_OFFSET 12
#define COMBINED_ROF_V1_ROF_BLOB_HEADER_SIZE 8

/*
 * The flag is currently hard-coded but may change later. This signals out of order (OOO) jobs mode so sets many
 * devx_qps as opposed to one. Results are processed in order per queue so multiple queues means can processed
 * OOO.
 */
#define OOO_EXECUTION_SUPPORTED 1

/*
 * MACRO related to compile uncompiled rules.
 */
#define COMPILATION_OUTPUT_DIR "compiled_output"
#define COMPILATION_OUTPUT_DIR_BF2 "bf2"
#define FORMATTED_RULES_FILE_NAME COMPILATION_OUTPUT_DIR "/" COMPILATION_OUTPUT_DIR_BF2 "/rules.txt"
#define COMPILED_RULES_FILE_NAME COMPILATION_OUTPUT_DIR "/" COMPILATION_OUTPUT_DIR_BF2 "/rules.rof2"

/**********************************************************************************************************************
 *
 * Type definitions
 *
 *********************************************************************************************************************/

struct regex_mkey {
	struct mlx5dv_devx_umem *umem;
	struct priv_mlx5_devx_obj *mkey;
	uint64_t offset;
};

struct devx_qp_ctx {
	uint16_t log_nb_desc;		  /* Log 2 number of desc for this object. */
	struct priv_mlx5_devx_qp devx_qp; /* The QP DevX object. */
	size_t pi, db_pi;
	size_t ci;
	uint32_t qpn;
};

struct regex_qp_ctx {
	uint32_t nb_desc;	      /* Total number of desc for this qp. */
	struct devx_qp_ctx *devx_qps; /* Pointer to qp array. */
	uint16_t nb_obj;	      /* Number of qp objects. */
	struct priv_doca_cq *cq;      /* CQ struct. */
	uint64_t free_qps;
	struct job_metadata *jobs_metadata;
	struct ibv_mr *metadata;
	struct ibv_mr *outputs;
	struct ibv_mr *imkey_addr; /* Indirect mkey array region. */
	size_t ci, pi;
};

struct batch_pending_qp {
	uint64_t hw_qpid;
	struct devx_qp_ctx *devx_qp;
};

struct driver_ctx {
	struct doca_dev *dev;
	struct priv_doca_uar *uar; /* UAR object. */
	uint32_t nb_engines;	   /* Number of RegEx engines. */
	uint32_t eqn;		   /* EQ number required for DEVX commands. */
	uint32_t qp_ts_format;	   /* Whether QP supports timestamp formats. */
	uint32_t regexp_version;   /* Regex hardware version */
	struct regex_driver_info properties;
};

struct driver_queue_ctx {
	struct doca_dev *driver_dev;
	struct priv_doca_uar *uar;		       /* UAR object. */
	struct regex_qp_ctx *qp_ctx;		       /* Pointer to the QP array. */
	struct batch_pending_qp active_qp;	       /* QP objs with jobs pending. */
	uint8_t *uar_base_addr;			       /* Base address of UAR */
	struct priv_doca_comp_events_channel *channel; /* Events channel (owned by the queue) */
	doca_event_handle_t channel_handle;	       /* Handle needed to de-register the events channel */
	uint16_t per_qp_depth;			       /* Number of items that can be placed on a single qp */
	uint32_t regexp_version;		       /* Regex hardware version */
};

struct job_metadata {
	uint64_t user_id;
	volatile uint8_t *output;
	volatile uint8_t *metadata;
	struct priv_mlx5_klm *imkey_array; /* Indirect mkey's KLM array. */
	struct priv_mlx5_devx_obj *imkey;  /* UMR WQE's indirect meky. */
};

enum cqe_status {
	CQE_STATUS_SW_OWN = -1,
	CQE_STATUS_HW_OWN = -2,
	CQE_STATUS_ERR = -3,
};

static const uint64_t combined_rof_tag = 0xff52544424a52475;

/**********************************************************************************************************************
 *
 * Internal functions
 *
 *********************************************************************************************************************/

static int make_dir(char const *path);
static int write_to_file(char const *buffer, size_t buffer_len, char const *file_name);
static void release_regex_ruleset(struct regex_ruleset *ruleset);
static int load_regex_raw_rules_file(char const *file_name, struct regex_ruleset *ruleset);
static int load_regex_binary_rof_file(char const *path, char **out_buffer, size_t *out_buffer_len);

static int
load_regex_binary_rof_file(char const *path, char **out_buffer, size_t *out_buffer_len)
{
	FILE *file;
	char *bytes;

	file = fopen(path, "rb");
	if (!file)
		return -ENOENT;

	if (fseek(file, 0, SEEK_END) != 0) {
		fclose(file);
		return -ENOENT;
	}

	long const nb_file_bytes = ftell(file);

	if (nb_file_bytes == -1) {
		fclose(file);
		return -ENOENT;
	}
	/* should we enforce some size limit here so we don't try to allocate something stupidly big? */
	bytes = malloc(nb_file_bytes);
	if (!bytes) {
		fclose(file);
		return -ENOMEM;
	}

	if (fseek(file, 0, SEEK_SET) != 0) {
		free(bytes);
		fclose(file);
		return -ENOENT;
	}

	size_t const read_byte_count = fread(bytes, 1, nb_file_bytes, file);

	fclose(file);

	if (read_byte_count != nb_file_bytes) {
		free(bytes);
		return -EIO;
	}

	*out_buffer = bytes;
	*out_buffer_len = read_byte_count;

	return 0;
}

static int
make_dir(char const *path)
{
	int const result = mkdir(path, 0755);

	if (result != 0 && errno != EEXIST)
		return -errno;

	return 0;
}

static int
write_to_file(char const *buffer, size_t buffer_len, char const *file_name)
{
	FILE *of = fopen(file_name, "w");

	if (of == NULL)
		return -errno;

	int const nb_written = fwrite(buffer, 1, buffer_len, of);

	fclose(of);

	if (nb_written != buffer_len)
		return -EIO;

	return 0;
}

static void
release_regex_ruleset(struct regex_ruleset *ruleset)
{
	uint32_t i, j;

	if (ruleset->number_of_entries) {
		for (i = 0; i < ruleset->number_of_entries; i++) {
			for (j = 0; j < ruleset->rules[i].number_of_prefix_entries; j++)
				free(ruleset->rules[i].prefix[j]);
			free(ruleset->rules[i].prefix);
			free(ruleset->rules[i].rule);
		}

		free(ruleset->rules);
	}
}

static int
load_regex_raw_rules_file(char const *file_name, struct regex_ruleset *ruleset)
{
	char *ruleset_err = NULL;
	char *file_name_mutable;
	int ret;

	if (ruleset == NULL || file_name == NULL)
		return -EINVAL;

	file_name_mutable = strdup(file_name);
	if (file_name_mutable == NULL)
		return -ENOMEM;

	ruleset->number_of_entries = 0;
	ret = regex_read_ruleset_file(file_name_mutable, ruleset, &ruleset_err);
	if (ret != REGEX_STATUS_OK) {
		DOCA_LOG_ERR("Failed reading ruleset from raw file (%s) Error: %s.\n", file_name,
			     ruleset_err != NULL ? ruleset_err : "Unexpected error");
		free(ruleset_err);
		free(file_name_mutable);
		release_regex_ruleset(ruleset);
		return -EINVAL;
	}

	free(file_name_mutable);
	free(ruleset_err);
	return 0;
}

static int
compile_bf2_rules(char const *buffer, uint32_t buffer_len, void **out_compiled_rules_data,
		  uint32_t *out_compiled_rules_len, enum regex_version compiled_regex_version)
{
	struct regex_compilation_statistics *comp_stats = NULL;
	struct regex_uncompiled_rules_log *uncompiled_rules_log;
	uint32_t global_regex_opts = 0;
	struct regex_rof *rof = NULL;
	struct regex_ruleset ruleset;
	uint32_t compiler_output;
	uint32_t compiler_opts;
	char *compiled_rules_data;
	size_t compiled_rules_data_size;
	int ret = 0;

	if (buffer == NULL || buffer_len == 0)
		return -EINVAL;

	ret = make_dir(COMPILATION_OUTPUT_DIR);
	if (ret < 0) {
		DOCA_LOG_ERR("Unable to create compilation output directory: %s (%s)", COMPILATION_OUTPUT_DIR,
			     strerror(abs(ret)));
		return ret;
	}

	ret = make_dir(COMPILATION_OUTPUT_DIR "/" COMPILATION_OUTPUT_DIR_BF2);
	if (ret < 0) {
		DOCA_LOG_ERR("Unable to create compilation output directory: %s (%s)",
			     COMPILATION_OUTPUT_DIR "/" COMPILATION_OUTPUT_DIR_BF2, strerror(abs(ret)));
		return ret;
	}

	ret = write_to_file(buffer, buffer_len, FORMATTED_RULES_FILE_NAME);
	if (ret < 0)
		return ret;

	ret = load_regex_raw_rules_file(FORMATTED_RULES_FILE_NAME, &ruleset);
	if (ret < 0)
		return ret;

	compiler_output = REGEX_COMPILER_OUTPUT_COMPILATION_STATISTICS;
	compiler_opts = REGEX_COMPILER_OPTIONS_DISABLE_BIDIRECTIONAL | REGEX_COMPILER_OPTIONS_FORCE |
			REGEX_GLOBAL_REGEX_OPTIONS_NO_SINGLE;

	ret = regex_compile(&ruleset, NULL, compiler_opts, global_regex_opts, 1, compiled_regex_version,
			    REGEX_VERBOSE_LEVEL_1, compiler_output, &comp_stats, NULL, NULL, &uncompiled_rules_log,
			    &rof, NULL);
	if (ret != REGEX_STATUS_OK) {
		DOCA_LOG_ERR("Failed to compile regex rules.");
		release_regex_ruleset(&ruleset);
		regex_free_structs(&comp_stats, NULL, NULL, &uncompiled_rules_log, &rof, NULL, NULL, NULL, 0);
		return -EINVAL;
	}

	char *rof_file_name = strdup(COMPILED_RULES_FILE_NAME);

	if (rof_file_name == NULL) {
		release_regex_ruleset(&ruleset);
		regex_free_structs(&comp_stats, NULL, NULL, &uncompiled_rules_log, &rof, NULL, NULL, NULL, 0);
		return -ENOMEM;
	}

	ret = regex_write_rof_file(rof_file_name, rof);
	if (ret != REGEX_STATUS_OK) {
		DOCA_LOG_ERR("Unable to write BF2 rof file");
		free(rof_file_name);
		release_regex_ruleset(&ruleset);
		regex_free_structs(&comp_stats, NULL, NULL, &uncompiled_rules_log, &rof, NULL, NULL, NULL, 0);
		return -ENOMEM;
	}

	free(rof_file_name);

	release_regex_ruleset(&ruleset);
	memset(&ruleset, 0, sizeof(struct regex_ruleset));
	regex_free_structs(&comp_stats, NULL, NULL, &uncompiled_rules_log, &rof, NULL, NULL, NULL, 0);

	ret = load_regex_binary_rof_file(COMPILED_RULES_FILE_NAME ".binary", &compiled_rules_data,
					 &compiled_rules_data_size);
	if (ret != 0)
		return -ENOMEM;

	*out_compiled_rules_data = compiled_rules_data;
	*out_compiled_rules_len = compiled_rules_data_size;

	return 0;
}

static inline uint32_t
get_supported_timestamp_format(uint32_t ts_format)
{
	return ts_format == PRIV_MLX5_HCA_CAP_TIMESTAMP_FORMAT_FR ? PRIV_MLX5_QPC_TIMESTAMP_FORMAT_FREE_RUNNING
								  : PRIV_MLX5_QPC_TIMESTAMP_FORMAT_DEFAULT;
}

static uint16_t
regex_ctrl_get_nb_obj(uint16_t nb_desc)
{
	return ((nb_desc / PRIV_MLX5_REGEX_NUM_WQE_PER_PAGE) + !!(nb_desc % PRIV_MLX5_REGEX_NUM_WQE_PER_PAGE));
}

static doca_error_t
device_has_regex_support(struct ibv_context *ibv_ctx, struct driver_ctx *driver)
{
	uint32_t in[PRIV_MLX5_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint32_t out[PRIV_MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t umr_modify_entity_size_disabled;
	uint32_t umr_indirect_mkey_disabled;
	uint32_t regexp_version;
	uint32_t regexp_params;
	uint32_t qp_ts_format = 0;
	uint32_t nb_engines;
	uint8_t regexp_qp_en;
	uint8_t regexp_sq_en;
	void *atts;

	PRIV_MLX5_SET(query_hca_cap_in, in, opcode, PRIV_MLX5_CMD_OP_QUERY_HCA_CAP);
	PRIV_MLX5_SET(query_hca_cap_in, in, op_mod,
		      PRIV_MLX5_GET_HCA_CAP_OP_MOD_GENERAL_DEVICE | PRIV_MLX5_HCA_CAP_OPMOD_GET_CUR);

	if (mlx5dv_devx_general_cmd(ibv_ctx, &in, sizeof(in), out, sizeof(out)) != 0) {
		DOCA_LOG_ERR("HCA query failed.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	atts = PRIV_MLX5_ADDR_OF(query_hca_cap_out, out, capability);

	/* Pull out regex specific info from return. */
	nb_engines = PRIV_MLX5_GET(cmd_hca_cap, atts, regexp_num_of_engines);
	regexp_params = PRIV_MLX5_GET(cmd_hca_cap, atts, regexp_params);
	regexp_qp_en = PRIV_MLX5_GET(cmd_hca_cap, atts, regexp_mmo_qp);
	regexp_sq_en = PRIV_MLX5_GET(cmd_hca_cap, atts, regexp_mmo_sq);
	regexp_version = PRIV_MLX5_GET(cmd_hca_cap, atts, regexp_version);

	if (!regexp_params || nb_engines == 0 || (!regexp_qp_en && !regexp_sq_en)) {
		DOCA_LOG_ERR("Regex not supported on the device.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	if (regexp_version != RXP_IDENTIFIER_BF2) {
		DOCA_LOG_ERR("Card not detected as being BF2.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	umr_indirect_mkey_disabled = PRIV_MLX5_GET(cmd_hca_cap, atts, umr_indirect_mkey_disabled);
	umr_modify_entity_size_disabled = PRIV_MLX5_GET(cmd_hca_cap, atts, umr_modify_entity_size_disabled);

	/* Verify if device supports User-Mode Memory Registration. */
	if (umr_indirect_mkey_disabled || umr_modify_entity_size_disabled) {
		DOCA_LOG_ERR("User-Mode Memory Registration not supported.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	/* If supported, query RDMA over ethernet for QP timestamp format. */
	if (PRIV_MLX5_GET(cmd_hca_cap, atts, roce)) {
		memset(in, 0, sizeof(in));
		memset(out, 0, sizeof(out));
		PRIV_MLX5_SET(query_hca_cap_in, in, opcode, PRIV_MLX5_CMD_OP_QUERY_HCA_CAP);
		PRIV_MLX5_SET(query_hca_cap_in, in, op_mod,
			      PRIV_MLX5_GET_HCA_CAP_OP_MOD_ROCE | PRIV_MLX5_HCA_CAP_OPMOD_GET_CUR);
		if (mlx5dv_devx_general_cmd(ibv_ctx, in, sizeof(in), out, sizeof(out)) != 0) {
			DOCA_LOG_ERR("HCA ROCE query failed.");
			return DOCA_ERROR_NOT_SUPPORTED;
		}

		qp_ts_format = PRIV_MLX5_GET(roce_caps, atts, qp_ts_format);
	}

	if (driver != NULL) {
		driver->nb_engines = nb_engines;
		driver->qp_ts_format = qp_ts_format;
		driver->regexp_version = regexp_version;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
init_driver_ctx(struct driver_ctx *driver, struct doca_dev *dev)
{
	struct ibv_context *ctx = priv_doca_dev_to_ibv_ctx(dev);

	driver->dev = dev;
	doca_error_t const result = device_has_regex_support(ctx, driver);

	if (result != DOCA_SUCCESS)
		return result;

	/* Get device eq number for event WQ doorbells. */
	if (mlx5dv_devx_query_eqn(ctx, 0, &driver->eqn) != 0) {
		DOCA_LOG_ERR("Cannot query event queue number.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

static void
teardown_devx_qp_ctx(struct devx_qp_ctx *devx_qp)
{
	priv_mlx5_devx_qp_destroy(&devx_qp->devx_qp);
	memset(devx_qp, 0, sizeof(*devx_qp));
}

static doca_error_t
init_devx_qp_ctx(struct regex_qp_ctx *qp_ctx, uint16_t q_ind, uint16_t log_nb_desc, struct doca_dev *dev,
		 struct priv_doca_uar *uar, uint32_t qp_ts_format)
{
	struct devx_qp_ctx *devx_qp = &qp_ctx->devx_qps[q_ind];
	uint32_t pd_num = 0;
	struct ibv_pd *pd = NULL;
	uint32_t cq_num = 0;
	uint32_t uar_page_id;
	doca_error_t result;

	result = priv_doca_cq_number_get(qp_ctx->cq, &cq_num);
	if (result != DOCA_SUCCESS)
		return result;

	result = priv_doca_uar_id_get(uar, &uar_page_id);
	if (result != DOCA_SUCCESS)
		return result;

	devx_qp->log_nb_desc = log_nb_desc;
	devx_qp->qpn = q_ind;
	devx_qp->ci = 0;
	devx_qp->pi = 0;

	result = doca_dev_get_pd(dev, &pd);
	if (result != DOCA_SUCCESS)
		return result;

	if (bf2_priv_mlx5_get_pdn(pd, &pd_num) != 0)
		return DOCA_ERROR_INITIALIZATION;

	struct priv_mlx5_devx_qp_attr attr = {
		.cqn = cq_num,
		.uar_index = uar_page_id,
		.ts_format = get_supported_timestamp_format(qp_ts_format),
		.user_index = q_ind,
		.pd = pd_num,
		.rq_size = 0,
		.sq_size = PRIV_MLX5_BIT32(PRIV_MLX5_REGEX_WQE_LOG_NUM(log_nb_desc)),
		.mmo = 1,
	};

	if (priv_mlx5_devx_qp_create(priv_doca_dev_to_ibv_ctx(dev), &devx_qp->devx_qp,
				     PRIV_MLX5_REGEX_WQE_LOG_NUM(log_nb_desc), &attr) != 0) {
		DOCA_LOG_ERR("Can't create QP object.");
		return DOCA_ERROR_NO_MEMORY;
	}

	if (priv_mlx5_devx_qp2rts(&devx_qp->devx_qp, 0) != 0) {
		DOCA_LOG_ERR("Can't change QP state to RTS.");
		teardown_devx_qp_ctx(devx_qp);
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
create_regex_qp_ctx_umr_buffers(struct driver_ctx *driver, struct regex_qp_ctx *regex_qp)
{
	struct ibv_pd *pd = NULL;
	uint32_t i;
	void *ptr;
	doca_error_t result;

	const size_t meta_size = regex_qp->nb_desc * PRIV_MLX5_REGEX_METADATA_SIZE;
	const size_t max_out_size = regex_qp->nb_desc * PRIV_MLX5_REGEX_MAX_OUTPUT;
	const size_t klms_size = regex_qp->nb_desc * PRIV_MLX5_REGEX_KLMS_SIZE;

	result = doca_dev_get_pd(driver->dev, &pd);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to get pd.");
		return result;
	}

	ptr = aligned_calloc(PRIV_MLX5_REGEX_METADATA_SIZE, meta_size);
	if (ptr == NULL)
		return DOCA_ERROR_NOT_SUPPORTED;

	regex_qp->metadata = ibv_reg_mr(pd, ptr, meta_size, IBV_ACCESS_LOCAL_WRITE);
	if (regex_qp->metadata == NULL) {
		DOCA_LOG_ERR("Failed to register metadata.");
		free(ptr);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	ptr = aligned_calloc(PRIV_MLX5_REGEX_MAX_OUTPUT, max_out_size);
	if (ptr == NULL) {
		DOCA_LOG_ERR("Failed to allocate output buffers");
		result = DOCA_ERROR_NO_MEMORY;
		goto err_regex_output;
	}

	regex_qp->outputs = ibv_reg_mr(pd, ptr, max_out_size, IBV_ACCESS_LOCAL_WRITE);
	if (regex_qp->outputs == NULL) {
		free(ptr);
		DOCA_LOG_ERR("Failed to register output.");
		result = DOCA_ERROR_NO_MEMORY;
		goto err_regex_output;
	}

	ptr = aligned_calloc(PRIV_MLX5_REGEX_KLMS_SIZE, klms_size);
	if (!ptr) {
		result = DOCA_ERROR_NO_MEMORY;
		goto err_imkey;
	}

	regex_qp->imkey_addr = ibv_reg_mr(pd, ptr, klms_size, IBV_ACCESS_LOCAL_WRITE);
	if (!regex_qp->imkey_addr) {
		free(ptr);
		DOCA_LOG_ERR("Failed to register output.");
		result = DOCA_ERROR_NOT_SUPPORTED;
		goto err_imkey;
	}

	/* Distribute buffers to metadata. */
	for (i = 0; i < regex_qp->nb_desc; i++) {
		regex_qp->jobs_metadata[i].output =
			(uint8_t *)regex_qp->outputs->addr + (i % regex_qp->nb_desc) * PRIV_MLX5_REGEX_MAX_OUTPUT;
		regex_qp->jobs_metadata[i].metadata =
			(uint8_t *)regex_qp->metadata->addr + (i % regex_qp->nb_desc) * PRIV_MLX5_REGEX_METADATA_SIZE;
		if (regex_qp->imkey_addr)
			regex_qp->jobs_metadata[i].imkey_array = (struct priv_mlx5_klm *)regex_qp->imkey_addr->addr +
								 (i % regex_qp->nb_desc) * PRIV_MLX5_REGEX_MAX_KLM_NUM;
	}

	return DOCA_SUCCESS;

err_imkey:
	free(regex_qp->outputs->addr);
	ibv_dereg_mr(regex_qp->outputs);
	regex_qp->outputs = NULL;
err_regex_output:
	free(regex_qp->metadata->addr);
	ibv_dereg_mr(regex_qp->metadata);
	regex_qp->metadata = NULL;

	return result;
}

static void
destroy_regex_qp_ctx_umr_buffers(struct regex_qp_ctx *regex_qp)
{
	if (regex_qp->imkey_addr) {
		free(regex_qp->imkey_addr->addr);
		ibv_dereg_mr(regex_qp->imkey_addr);
		regex_qp->imkey_addr = NULL;
	}
	if (regex_qp->outputs) {
		free(regex_qp->outputs->addr);
		ibv_dereg_mr(regex_qp->outputs);
		regex_qp->outputs = NULL;
	}
	if (regex_qp->metadata) {
		free(regex_qp->metadata->addr);
		ibv_dereg_mr(regex_qp->metadata);
		regex_qp->metadata = NULL;
	}
}

static inline uint32_t
get_qp_size(struct devx_qp_ctx *qp)
{
	return (1U << qp->log_nb_desc);
}

static void
init_regex_qp_ctx_wqes(struct regex_qp_ctx *queue)
{
	size_t hw_qpid, entry;
	uint32_t job_id;

	for (hw_qpid = 0; hw_qpid < queue->nb_obj; hw_qpid++) {
		struct devx_qp_ctx *devx_qp = &queue->devx_qps[hw_qpid];
		uint8_t *wqe = (uint8_t *)devx_qp->devx_qp.wqes;

		for (entry = 0; entry < get_qp_size(devx_qp); entry++) {
			job_id = hw_qpid * get_qp_size(devx_qp) + entry;
			struct job_metadata *job = &queue->jobs_metadata[job_id];

			/* Fill UMR WQE with NOP in advance. */
			set_wqe_ctrl_seg((struct mlx5_wqe_ctrl_seg *)wqe, entry * 2, MLX5_OPCODE_NOP, 0,
					 devx_qp->devx_qp.qp->id, 0, 12, 0, 0);
			wqe += PRIV_MLX5_REGEX_UMR_WQE_SIZE;
			set_metadata_seg(
				(struct priv_mlx5_wqe_metadata_seg *)(wqe + PRIV_MLX5_REGEX_WQE_METADATA_OFFSET), 0,
				queue->metadata->lkey, (uintptr_t)job->metadata);
			set_data_seg((struct mlx5_wqe_data_seg *)(wqe + PRIV_MLX5_REGEX_WQE_SCATTER_OFFSET),
				     PRIV_MLX5_REGEX_MAX_OUTPUT, queue->outputs->lkey, (uintptr_t)job->output);
			wqe += 64;
		}
		queue->free_qps |= 1ULL << hw_qpid;
	}
}

static void
destroy_regex_qp_ctx_jobs(struct regex_qp_ctx *regex_qp)
{
	uint32_t i;

	if (regex_qp == NULL)
		return;

	if (regex_qp->jobs_metadata) {
		for (i = 0; i < regex_qp->nb_desc; i++) {
			if (regex_qp->jobs_metadata[i].imkey) {
				mlx5dv_devx_obj_destroy(regex_qp->jobs_metadata[i].imkey->obj);
				free(regex_qp->jobs_metadata[i].imkey);
			}
		}
		destroy_regex_qp_ctx_umr_buffers(regex_qp);
		free(regex_qp->jobs_metadata);
		regex_qp->jobs_metadata = NULL;
	}
}

static doca_error_t
create_regex_qp_ctx_jobs(struct driver_ctx *driver, struct regex_qp_ctx *regex_qp)
{
	int i;
	uint32_t pdn;
	struct ibv_pd *pd = NULL;
	doca_error_t result;

	result = doca_dev_get_pd(driver->dev, &pd);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to get pd.");
		return result;
	}

	if (bf2_priv_mlx5_get_pdn(pd, &pdn) != 0) {
		DOCA_LOG_ERR("Failed to get pdn.");
		return DOCA_ERROR_INITIALIZATION;
	}

	regex_qp->jobs_metadata = aligned_calloc(ALIGN_64, regex_qp->nb_desc * sizeof(*regex_qp->jobs_metadata));
	if (!regex_qp->jobs_metadata) {
		DOCA_LOG_ERR("Failed to allocate qp job memory.");
		return DOCA_ERROR_NO_MEMORY;
	}

	if (create_regex_qp_ctx_umr_buffers(driver, regex_qp) != DOCA_SUCCESS) {
		free(regex_qp->jobs_metadata);
		return DOCA_ERROR_NO_MEMORY;
	}

	init_regex_qp_ctx_wqes(regex_qp);

	for (i = 0; i < regex_qp->nb_desc; i++) {
		struct priv_mlx5_devx_mkey_attr attr = {
			.umr_en = 1,
			.pd = pdn,
			.klm_num = PRIV_MLX5_REGEX_MAX_KLM_NUM,
			.klm_array = regex_qp->jobs_metadata[i].imkey_array,
		};

		regex_qp->jobs_metadata[i].imkey =
			priv_mlx5_devx_cmd_mkey_create(priv_doca_dev_to_ibv_ctx(driver->dev), &attr);
		if (regex_qp->jobs_metadata[i].imkey == NULL) {
			DOCA_LOG_ERR("Failed to allocate imkey.");
			destroy_regex_qp_ctx_jobs(regex_qp);
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	return DOCA_SUCCESS;
}

static doca_error_t
create_cq(struct doca_dev const *const dev, struct priv_doca_uar const *const uar, uint16_t log_cq_size,
	  struct priv_doca_cq **cq)
{
	struct priv_doca_cq_attr const cq_attr = priv_doca_cq_attr_create(1UL << log_cq_size);
	doca_error_t const ret = priv_doca_dev_cq_create(dev, uar, &cq_attr, cq);

	if (ret != DOCA_SUCCESS)
		DOCA_LOG_ERR("DMA work queue context, failed to create cq. err=%s", doca_get_error_string(ret));

	return ret;
}

static doca_error_t
destroy_cq(struct priv_doca_cq *cq)
{
	doca_error_t const ret = priv_doca_dev_cq_destroy(cq);

	if (ret != DOCA_SUCCESS)

		DOCA_LOG_ERR("DMA work queue context, failed to destroy cq. err=%s", doca_get_error_name(ret));

	return ret;
}

static struct regex_qp_ctx *
create_regex_qp_ctx(struct driver_queue_ctx *queue, struct driver_ctx *driver)
{
	uint16_t const total_qp_log_desc = bf2_priv_mlx5_log2_u32(REGEX_MAX_PER_QP_NUM_DESCS);
	uint16_t const cq_log_desc = total_qp_log_desc + 1; /* UMR mode so doubling the number of descriptors. */
	uint16_t const nb_qp_objects =
		OOO_EXECUTION_SUPPORTED ? regex_ctrl_get_nb_obj(1 << PRIV_MLX5_REGEX_WQE_LOG_NUM(total_qp_log_desc))
					: 1;
	uint16_t const per_qp_nb_desc = bf2_priv_mlx5_log2_u32((1 << total_qp_log_desc) / nb_qp_objects);
	struct regex_qp_ctx *regex_qp;
	int i;

	regex_qp = calloc(1, sizeof(struct regex_qp_ctx));
	if (regex_qp == NULL)
		return NULL;

	regex_qp->nb_desc = 1 << total_qp_log_desc;

	regex_qp->devx_qps = aligned_calloc(ALIGN_64, nb_qp_objects * sizeof(struct devx_qp_ctx));
	if (!regex_qp->devx_qps) {
		DOCA_LOG_ERR("Can't allocate qp array memory.");
		goto release_obj;
	}

	if (create_cq(queue->driver_dev, queue->uar, cq_log_desc, &regex_qp->cq) != DOCA_SUCCESS)
		goto release_qps;

	for (i = 0; i < nb_qp_objects; i++) {
		if (init_devx_qp_ctx(regex_qp, i, per_qp_nb_desc, queue->driver_dev, queue->uar,
				     driver->qp_ts_format) != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Can't create qp object.");
			goto destroy_qps;
		}
		++(regex_qp->nb_obj);
	}

	if (create_regex_qp_ctx_jobs(driver, regex_qp) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Cannot set up fast path.");
		goto destroy_qps;
	}

	queue->per_qp_depth = 1U << per_qp_nb_desc;

	return regex_qp;

destroy_qps:
	for (i = 0; i < regex_qp->nb_obj; i++)
		teardown_devx_qp_ctx(queue->qp_ctx->devx_qps + i);
	destroy_cq(regex_qp->cq);
release_qps:
	free(regex_qp->devx_qps);
release_obj:
	free(regex_qp);

	return NULL;
}

static void
destroy_regex_qp_ctx(struct regex_qp_ctx *qp_ctx)
{
	size_t ii;

	for (ii = 0; ii != qp_ctx->nb_obj; ++ii)
		teardown_devx_qp_ctx(qp_ctx->devx_qps + ii);

	destroy_regex_qp_ctx_jobs(qp_ctx);
	destroy_cq(qp_ctx->cq);
	free(qp_ctx->devx_qps);
	qp_ctx->devx_qps = NULL;
	free(qp_ctx);
}

static struct regex_driver_queue_impl *
create_driver_queue_ctx(struct driver_ctx *driver)
{
	struct driver_queue_ctx *queue;
	doca_error_t ret;

	queue = aligned_calloc(ALIGN_64, sizeof(struct driver_queue_ctx));
	if (queue == NULL) {
		DOCA_LOG_ERR("Failed to allocate driver queue context");
		return NULL;
	}

	queue->driver_dev = driver->dev;
	queue->regexp_version = driver->regexp_version;

	/*
	 * Allocate the User Access Region. First attempt to allocate to cached memory (will require write barrier
	 * after writes). If fails, try to allocate to non-cached memory (no barriers required for write).
	 */
	ret = priv_doca_dev_uar_create(driver->dev, PRIV_DOCA_UAR_ALLOCATION_TYPE_BLUEFLAME, &(queue->uar));
	if (ret != DOCA_SUCCESS) {
		/* If fails, try to allocate to non-cached memory (no barriers required for write). */
		ret = priv_doca_dev_uar_create(driver->dev, PRIV_DOCA_UAR_ALLOCATION_TYPE_NONCACHE, &(queue->uar));
	}

	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("UAR alloc failed.");
		goto release_queue;
	}

	ret = priv_doca_uar_page_get(queue->uar, (void **)(&queue->uar_base_addr));
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("UAR init failed.");
		goto release_uar;
	}

	queue->qp_ctx = create_regex_qp_ctx(queue, driver);
	if (queue->qp_ctx == NULL)
		goto release_uar;

	return (struct regex_driver_queue_impl *)queue;

release_uar:
	priv_doca_dev_uar_destroy(queue->uar);
release_queue:
	free(queue);
	return NULL;
}

/**
 * Return the expected ROF version number for the requested RXP hardware type.
 *
 * @param [in] regexp_version
 * The version of Bluefield hardware being run on
 *
 * @param [out] target_rxp_vers
 * On successful completion of this function target_rxp_vers will contain the rof version identifier for the specified
 * Bluefield hardware version
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - unsupported regexp_version
 */
static doca_error_t
rxp_identifier_to_rof_identifier(uint32_t regexp_version, uint32_t *target_rxp_vers)
{
	switch (regexp_version) {
	case RXP_IDENTIFIER_BF2:
		*target_rxp_vers = ROF_IDENTIFIER_BF2;
		return DOCA_SUCCESS;
	case RXP_IDENTIFIER_BF3:
		*target_rxp_vers = ROF_IDENTIFIER_BF3;
		return DOCA_SUCCESS;
	}
	DOCA_LOG_ERR("Unsupported rxp version: %u", regexp_version);
	return DOCA_ERROR_INVALID_VALUE;
}

/**
 * Parse a version 1 combined rof format to locate the rof entry relevant to the RXP
 * hardware being run on.
 *
 * @param [in] rules_db
 * Pointer to the combined rof binary file
 *
 * @param [in] rules_db_len
 * Length of the combined rof binary file in bytes
 *
 * @param [in] target_rxp_vers
 * The version of the Bluefield that a binary rof is needed for
 *
 * @param [out] found_rof_binary_data
 * On successful completion of this function *found_rof_binary_data will point to the binary rof data to be programmed
 *
 * @param [out] found_rof_binary_size
 * On successful completion of this function found_rof_binary_size will the size of the binary rof data to be
 * programmed.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - Invalid rules database length or uncompatible combined rof file
 */
static doca_error_t
extract_rof_from_v1_format_combined_rof(const char *rules_db, uint32_t const rules_db_len,
					uint32_t const target_rxp_vers, const char **found_rof_binary_data,
					uint32_t *found_rof_binary_size)
{
	uint32_t i = 0;
	const char *rof_blob_ptr = rules_db;
	uint32_t rxpc_vers = 0;
	uint32_t rof_blob_len = 0;
	bool rof_binary_found = false;
	uint32_t ptr_diff = 0;
	const uint32_t num_rof_blocks = *((uint32_t *)(rules_db + COMBINED_ROF_V1_NUM_BLOCKS_OFFSET));

	/* Need at least a combined rof header and one rof blob header to process */
	if (rules_db_len < (COMBINED_ROF_HEADER_SIZE + COMBINED_ROF_V1_ROF_BLOB_HEADER_SIZE))
		return DOCA_ERROR_INVALID_VALUE;

	if (num_rof_blocks == 0)
		return DOCA_ERROR_INVALID_VALUE;

	/* Try to find the rof binary blob for this version of rxp */
	rof_blob_ptr += COMBINED_ROF_HEADER_SIZE;
	for (i = 0; i != num_rof_blocks; ++i) {
		rxpc_vers = *((uint32_t *)rof_blob_ptr);
		rof_blob_len = *((uint32_t *)(rof_blob_ptr + sizeof(rxpc_vers)));
		rof_blob_ptr += COMBINED_ROF_V1_ROF_BLOB_HEADER_SIZE;
		ptr_diff = rof_blob_ptr - rules_db;
		if (rxpc_vers == target_rxp_vers) {
			/* Found corresponding binary rof entry */
			if (ptr_diff + rof_blob_len <= (rules_db_len))
				rof_binary_found = true;
			else
				DOCA_LOG_ERR("Compatible rof file found - invalid length!");
			break;
		}
		/* Move on to next rof blob header */
		if (ptr_diff + rof_blob_len + COMBINED_ROF_V1_ROF_BLOB_HEADER_SIZE < (rules_db_len)) {
			rof_blob_ptr += rof_blob_len;
		} else {
			/* Cannot parse any more of combined rof file */
			break;
		}
	}

	if (rof_binary_found == true) {
		*found_rof_binary_data = rof_blob_ptr;
		*found_rof_binary_size = rof_blob_len;
	} else {
		DOCA_LOG_ERR("Compatible rof file not found!");
		return DOCA_ERROR_INVALID_VALUE;
	}

	return DOCA_SUCCESS;
}

/**
 * Check for and parse a combined rof format to extract the relevant rof for the RXP hardware
 * being run on.
 *
 * @param [in] rules_db
 * Pointer to the combined rof binary file
 *
 * @param [in] rules_db_len
 * Length of the combined rof binary file in bytes
 *
 * @param [in] target_rxp_vers
 * The version of the Bluefield that a binary rof is needed for
 *
 * @param [out] found_rof_binary_data
 * On successful completion of this function *found_rof_binary_data will point to the binary rof data to be programmed
 *
 * @param [out] found_rof_binary_size
 * On successful completion of this function found_rof_binary_size will the size of the binary rof data to be
 * programmed.
 *
 * @return
 * DOCA_SUCCESS
 * DOCA_ERROR_INVALID_VALUE - Invalid rules database length or unsupported combined rof version
 */
static doca_error_t
extract_appropriate_rof_db_from_combined_rof(const char *rules_db, uint32_t rules_db_len, uint32_t target_rxp_vers,
					     const char **found_rof_binary_data, uint32_t *found_rof_binary_size)
{
	const char *rof_ptr = rules_db;
	uint32_t combined_rof_vers = 0;
	uint64_t combined_rof_magic = *((uint64_t *)rof_ptr);
	uint32_t combined_rof_fmt = *((uint32_t *)(rof_ptr + 8));

	/* Need minimum of 8 bytes to process single or combined rof */
	if (rules_db_len < 8)
		return DOCA_ERROR_INVALID_VALUE;

	if (combined_rof_magic == combined_rof_tag) {
		switch (combined_rof_fmt) {
		case 1:
			return extract_rof_from_v1_format_combined_rof(rules_db, rules_db_len, target_rxp_vers,
								       found_rof_binary_data, found_rof_binary_size);
		}

		DOCA_LOG_ERR("Unsupported combined rof version: %u", combined_rof_vers);
		return DOCA_ERROR_INVALID_VALUE;
	}
	/* Nothing to do, it's a plain rof file. Set out parameters to in parameters */
	*found_rof_binary_data = rules_db;
	*found_rof_binary_size = rules_db_len;

	return DOCA_SUCCESS;
}

static doca_error_t
program_regex_rules_memory(struct driver_ctx *driver, void const *rules, uint32_t rules_len)
{
	struct priv_mlx5_devx_mkey_attr mkey_attr;
	struct regex_mkey regex_mkey;
	char *aligned_rules;
	uint32_t id;
	struct ibv_context *ctx;
	struct ibv_pd *pd = NULL;
	doca_error_t result;
	uint32_t target_rxp_vers = 0;
	uint32_t found_rof_binary_size = rules_len;
	const char *found_rof_binary_data = (const char *)rules;

	if (rules == NULL || rules_len == 0) {
		DOCA_LOG_ERR("Invalid rule input.");
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* Get the version of rxp we need the rof for */
	result = rxp_identifier_to_rof_identifier(driver->regexp_version, &target_rxp_vers);
	if (result != DOCA_SUCCESS)
		return result;

	result = extract_appropriate_rof_db_from_combined_rof(rules, rules_len, target_rxp_vers, &found_rof_binary_data,
							      &found_rof_binary_size);
	if (result != DOCA_SUCCESS)
		return result;

	/* Data is required to be 4k aligned. Ensure this is the case by making a copy. */
	aligned_rules = aligned_calloc(ALIGN_4K, sizeof(char) * found_rof_binary_size);
	if (aligned_rules == NULL) {
		DOCA_LOG_ERR("Failed to copy rules file.");
		return DOCA_ERROR_NO_MEMORY;
	}

	memcpy(aligned_rules, found_rof_binary_data, found_rof_binary_size);
	ctx = priv_doca_dev_to_ibv_ctx(driver->dev);

	/*
	 * Get umem address for rules. Access is 'enum ibv_access_flags' bitmask. 7 is local and remote read/write
	 */
	regex_mkey.umem = mlx5dv_devx_umem_reg(ctx, aligned_rules, found_rof_binary_size, PRIV_UMEM_ACCESS_FLAGS);
	if (!regex_mkey.umem) {
		DOCA_LOG_ERR("Failed to allocate umem mkey for rules write.");
		return DOCA_ERROR_NO_MEMORY;
	}

	mkey_attr = (struct priv_mlx5_devx_mkey_attr){
		.addr = (uintptr_t)aligned_rules,
		.size = (uint32_t)found_rof_binary_size,
		.umem_id = regex_mkey.umem->umem_id,
		.pg_access = 1,
		.umr_en = 0,
	};

	result = doca_dev_get_pd(driver->dev, &pd);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to get pd.");
		goto free_umem;
	}

	if (bf2_priv_mlx5_get_pdn(pd, &mkey_attr.pd) < 0) {
		DOCA_LOG_ERR("Failed to get pdn.");
		result = DOCA_ERROR_NOT_SUPPORTED;
		goto free_umem;
	}

	regex_mkey.mkey = priv_mlx5_devx_cmd_mkey_create(ctx, &mkey_attr);
	if (!regex_mkey.mkey) {
		DOCA_LOG_ERR("Failed to create direct mkey for rules write.");
		result = DOCA_ERROR_NO_MEMORY;
		goto free_umem;
	}

	result = DOCA_SUCCESS;

	for (id = 0; id < driver->nb_engines; id++) {
		int const ret = priv_mlx5_devx_regex_rules_program(ctx, id, regex_mkey.mkey->id, found_rof_binary_size,
								   (uintptr_t)aligned_rules);
		if (ret != 0) {
			DOCA_LOG_ERR("Failed on rules programming.");
			result = DOCA_ERROR_NOT_SUPPORTED;
			break;
		}
	}

	/* Free mkey for rules - no longer required. */
	mlx5dv_devx_obj_destroy(regex_mkey.mkey->obj);
	free(regex_mkey.mkey);
free_umem:
	mlx5dv_devx_umem_dereg(regex_mkey.umem);
	free(aligned_rules);

	return result;
}

static inline void
set_regex_ctrl_seg(void *seg, uint8_t le, uint16_t subset_id0, uint16_t subset_id1, uint16_t subset_id2,
		   uint16_t subset_id3, uint8_t ctrl)
{
	PRIV_MLX5_SET(regexp_mmo_control, seg, le, le);
	PRIV_MLX5_SET(regexp_mmo_control, seg, ctrl, ctrl);
	PRIV_MLX5_SET(regexp_mmo_control, seg, subset_id_0, subset_id0);
	PRIV_MLX5_SET(regexp_mmo_control, seg, subset_id_1, subset_id1);
	PRIV_MLX5_SET(regexp_mmo_control, seg, subset_id_2, subset_id2);
	PRIV_MLX5_SET(regexp_mmo_control, seg, subset_id_3, subset_id3);
}

static inline uint32_t
job_id_get(uint32_t qid, size_t qp_size, size_t index)
{
	return qid * qp_size + (index & (qp_size - 1));
}

static inline int
get_free(struct devx_qp_ctx *devx_qp)
{
	return (get_qp_size(devx_qp) - ((devx_qp->pi - devx_qp->ci) & (PRIV_MLX5_REGEX_MAX_WQE_INDEX >> 2)));
}

static inline void
prep_one(struct devx_qp_ctx *devx_qp, struct regex_driver_job const *job, size_t pi, struct priv_mlx5_klm *klm)
{
	size_t wqe_offset = (pi & (get_qp_size(devx_qp) - 1)) * (MLX5_SEND_WQE_BB << 2) + PRIV_MLX5_REGEX_UMR_WQE_SIZE;
	uint16_t group0 = job->rule_group_ids[0];
	uint16_t group1 = job->rule_group_ids[1];
	uint16_t group2 = job->rule_group_ids[2];
	uint16_t group3 = job->rule_group_ids[3];

	/* If none are set then default to group 0 only. */
	if (!group0 && !group1 && !group2 && !group3)
		group0 = 1;

	/* Can set control to 1 for Highest priority match. */
	uint8_t control = 0;

	if (job->flags & DOCA_REGEX_SEARCH_JOB_FLAG_HIGHEST_PRIORITY_MATCH)
		control |= RXP_JOB_CONTROL_HIGHEST_PRIORITY_MATCH;

	uint8_t *wqe = (uint8_t *)(uintptr_t)devx_qp->devx_qp.wqes + wqe_offset;
	int ds = 4; /*  ctrl + meta + input + output */

	set_wqe_ctrl_seg((struct mlx5_wqe_ctrl_seg *)wqe, (pi * 4 + 3), PRIV_MLX5_OPCODE_MMO,
			 PRIV_MLX5_OPC_MOD_MMO_REGEX, devx_qp->devx_qp.qp->id, 0, ds, 0, 0);
	set_regex_ctrl_seg(wqe + 12, 0, group0, group1, group2, group3, control);
	struct mlx5_wqe_data_seg *input_seg = (struct mlx5_wqe_data_seg *)(wqe + PRIV_MLX5_REGEX_WQE_GATHER_OFFSET);

	input_seg->byte_count = htobe32(klm->byte_count);
	input_seg->addr = htobe64(klm->address);
	input_seg->lkey = htobe32(klm->mkey);
}

static inline void
prep_nop_regex_wqe_set(struct devx_qp_ctx *devx_qp, struct regex_driver_job const *job, size_t pi,
		       struct priv_mlx5_klm *klm)
{
	size_t wqe_offset = (pi & (get_qp_size(devx_qp) - 1)) * (MLX5_SEND_WQE_BB << 2);
	struct mlx5_wqe_ctrl_seg *wqe =
		(struct mlx5_wqe_ctrl_seg *)((uint8_t *)(uintptr_t)devx_qp->devx_qp.wqes + wqe_offset);

	/* Clear the WQE memory used as UMR WQE previously. */
	if ((be32toh(wqe->opmod_idx_opcode) & 0xff) != MLX5_OPCODE_NOP)
		memset(wqe, 0, PRIV_MLX5_REGEX_UMR_WQE_SIZE);
	/* UMR WQE size is 9 DS, align nop WQE to 3 WQEBBS(12 DS). */
	set_wqe_ctrl_seg(wqe, pi * 4, MLX5_OPCODE_NOP, 0, devx_qp->devx_qp.qp->id, 0, 12, 0, 0);
	prep_one(devx_qp, job, pi, klm);
}

static inline void
prep_regex_umr_wqe_set(struct driver_queue_ctx *queue, struct regex_qp_ctx *regex_qp, struct devx_qp_ctx *devx_qp,
		       struct regex_driver_job const *jobs, size_t no_jobs)
{
	size_t job_idx = no_jobs;
	struct priv_mlx5_klm klm;
	uint32_t buf_klm_mkey;

	while (job_idx--)
		priv_prefetch0(jobs + job_idx);
	job_idx = no_jobs;

	/*
	 * Build the WQE set by reverse. In case the burst may consume multiple mkeys, build the WQE set as normal will
	 * hard to address the last mkey index, since we will only know the last RegEx WQE's index when finishes
	 * building.
	 */
	while (job_idx--) {
		size_t pi = PRIV_MLX5_REGEX_UMR_QP_PI_IDX(devx_qp->pi, job_idx);

		/* FUTURE: Scatter stuff needs to go in here. */

		/* Get doca buf mkey */
		if (doca_buf_get_lkey(jobs[job_idx].buffer, queue->driver_dev, DOCA_ACCESS_LOCAL_READ, &buf_klm_mkey) !=
		    DOCA_SUCCESS) {
			DOCA_LOG_ERR("Unable to get mkey for buffer.");
			return;
		}

		/* The single mubf case. Build the KLM directly. */
		klm.mkey = buf_klm_mkey;
		klm.address = (uint64_t)(priv_doca_buf_get_data(jobs[job_idx].buffer) + jobs[job_idx].buffer_offset);
		klm.byte_count = jobs[job_idx].buffer_bytes_to_send;

		regex_qp->jobs_metadata[job_id_get(devx_qp->qpn, get_qp_size(devx_qp), pi)].user_id =
			jobs[job_idx].job_id;
		/*
		 * Build the nop + RegEx WQE set by default. The fist nop WQE will be updated later as UMR WQE if
		 * scattered mubf exist.
		 */
		prep_nop_regex_wqe_set(devx_qp, jobs + job_idx, pi, &klm);
	}

	/*
	 * Scattered mbuf have been added to the KLM array. Complete the build of UMR WQE, update the first nop WQE as
	 *UMR WQE.
	 *
	 *	if (mkey_klm)
	 *		complete_umr_wqe(qp, devx_qp, &qp->jobs[mkey_job_id], devx_qp->pi,
	 *				 klm_num, len);
	 */

	devx_qp->db_pi = PRIV_MLX5_REGEX_UMR_QP_PI_IDX(devx_qp->pi, no_jobs - 1);
	devx_qp->pi = PRIV_MLX5_REGEX_UMR_QP_PI_IDX(devx_qp->pi, no_jobs);
}

static inline void
send_doorbell(struct driver_queue_ctx *queue, struct devx_qp_ctx *devx_qp)
{
	size_t wqe_offset;
	uint8_t *wqe;

	wqe_offset =
		(devx_qp->db_pi & (get_qp_size(devx_qp) - 1)) * (MLX5_SEND_WQE_BB << 2) + PRIV_MLX5_REGEX_UMR_WQE_SIZE;
	wqe = (uint8_t *)(uintptr_t)devx_qp->devx_qp.wqes + wqe_offset;
	((struct mlx5_wqe_ctrl_seg *)wqe)->fm_ce_se |= MLX5_WQE_CTRL_CQ_UPDATE;

	priv_doca_udma_to_device_barrier();
	devx_qp->devx_qp.db_rec[MLX5_SND_DBR] = htobe32(((1 + devx_qp->db_pi) * 4) & PRIV_MLX5_REGEX_MAX_WQE_INDEX);
	priv_doca_mmio_wc_start();
	*(uint64_t *)(queue->uar_base_addr + 0x800) = *(uint64_t *)wqe;
	priv_doca_mmio_flush_writes();
}

static inline void
extract_result(struct job_metadata *job, struct regex_driver_response *resp, struct doca_regex_mempool *matches_mp)
{
	struct doca_regex_match *last_match;
	size_t i;
	size_t offset;
	uint32_t detected_matches;
	uint32_t reported_matches;
	uint16_t status;

	resp->id = job->user_id;

	detected_matches = PRIV_MLX5_GET_VOLATILE(regexp_metadata, job->metadata + PRIV_MLX5_REGEX_METADATA_OFF,
						  detected_match_count);
	reported_matches =
		PRIV_MLX5_GET_VOLATILE(regexp_metadata, job->metadata + PRIV_MLX5_REGEX_METADATA_OFF, match_count);

	resp->detected_matches = detected_matches;
	resp->num_matches = 0;
	resp->matches = NULL;
	resp->matches_mempool = matches_mp;

	status = PRIV_MLX5_GET_VOLATILE(regexp_metadata, job->metadata + PRIV_MLX5_REGEX_METADATA_OFF, status);
	resp->status_flags = status;

	if (matches_mp == NULL)
		return;

	last_match = NULL;
	for (i = 0; i < reported_matches; i++) {
		struct doca_regex_match *match = doca_regex_mempool_get_obj(matches_mp);

		if (!match)
			break;

		offset = PRIV_MLX5_REGEX_RESP_SZ * i;
		match->next = NULL;
		match->rule_id = PRIV_MLX5_GET_VOLATILE(regexp_match_tuple, (job->output + offset), rule_id);
		match->match_start = PRIV_MLX5_GET_VOLATILE(regexp_match_tuple, (job->output + offset), start_ptr);
		match->length = PRIV_MLX5_GET_VOLATILE(regexp_match_tuple, (job->output + offset), length);
		resp->num_matches++;

		if (last_match == NULL)
			resp->matches = match;
		else
			last_match->next = match;

		last_match = match;
	}
}

bool
is_valid_regex_response(struct priv_mlx5_cqe const *const cqe)
{
	uint8_t const op_code = BF2_CQE_OPCODE(cqe->op_own);

	if (op_code == MLX5_CQE_INVALID)
		return false;

	if ((op_code == MLX5_CQE_RESP_ERR || op_code == MLX5_CQE_REQ_ERR)) {
		struct mlx5_err_cqe const *const err_cqe = (struct mlx5_err_cqe const *)cqe;

		DOCA_LOG_ERR("Completion with error! syndrome: %u err: %u", err_cqe->syndrome,
			     err_cqe->vendor_err_synd);
		return false;
	}

	return true;
}

/**********************************************************************************************************************
 *
 * Virtual function declarations
 *
 *********************************************************************************************************************/

static void bf2_driver_destroy(struct regex_driver *driver);
struct regex_driver_info const *bf2_driver_get_properties(struct regex_driver const *driver);
static doca_error_t bf2_driver_compile_rules(struct regex_driver *driver, void const *rules_data, uint32_t rules_len,
					     void **out_compiled_rules_data, uint32_t *out_compiled_rules_len);
static doca_error_t bf2_driver_program_compiled_rules(struct regex_driver *driver, void const *data, uint32_t data_len);
static struct regex_driver_queue *bf2_driver_create_queue(struct regex_driver *driver);

static void bf2_queue_destroy(struct regex_driver_queue *queue);
static doca_error_t bf2_get_event_channel(struct regex_driver_queue *queue,
					  struct priv_doca_comp_events_channel **channel, doca_event_handle_t *handle);
static bool bf2_queue_is_empty(struct regex_driver_queue const *queue);
static doca_error_t bf2_queue_enqueue_job(struct regex_driver_queue *queue, struct regex_driver_job const *job);
static doca_error_t bf2_queue_dequeue_jobs(struct regex_driver_queue *queue, struct doca_regex_mempool *matches_mp,
					   struct regex_driver_queue_result_batch *result_batch);

/**********************************************************************************************************************
 *
 * Virtual interface template
 *
 *********************************************************************************************************************/
static const struct regex_driver driver_template = {
	NULL,
	NULL,
	bf2_driver_destroy,
	bf2_driver_get_properties,
	bf2_driver_compile_rules,
	bf2_driver_program_compiled_rules,
	bf2_driver_create_queue,
};

static const struct regex_driver_queue queue_template = {
	NULL,
	bf2_queue_destroy,
	bf2_get_event_channel,
	bf2_queue_is_empty,
	bf2_queue_enqueue_job,
	bf2_queue_dequeue_jobs,
};

/**********************************************************************************************************************
 *
 * Virtual function definitions
 *
 *********************************************************************************************************************/

static void
bf2_driver_destroy(struct regex_driver *driver)
{
	if (driver == NULL)
		return;

	if (driver->inst == NULL) {
		free(driver);
		return;
	}

	free(driver->inst);
	free(driver);
}

struct regex_driver_info const *
bf2_driver_get_properties(struct regex_driver const *driver)
{
	struct driver_ctx const *ctx;

	if (driver == NULL || driver->inst == NULL)
		return NULL;

	ctx = (struct driver_ctx *)driver->inst;

	return &ctx->properties;
}

static doca_error_t
bf2_driver_compile_rules(struct regex_driver *driver, void const *rules_data, uint32_t rules_len,
			 void **out_compiled_rules_data, uint32_t *out_compiled_rules_len)
{
	enum regex_version compiled_regex_version;
	struct driver_ctx *const ctx = (struct driver_ctx *)driver->inst;

	switch (ctx->regexp_version) {
	case RXP_IDENTIFIER_BF2:
		compiled_regex_version = REGEX_VERSION_V5_7;
		break;
	case RXP_IDENTIFIER_BF3:
		compiled_regex_version = REGEX_VERSION_V6_0;
		break;
	default:
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (compile_bf2_rules(rules_data, rules_len, out_compiled_rules_data, out_compiled_rules_len,
			      compiled_regex_version) != 0)
		return DOCA_ERROR_NOT_SUPPORTED;
	return DOCA_SUCCESS;
}

static doca_error_t
bf2_driver_program_compiled_rules(struct regex_driver *driver, void const *data, uint32_t data_len)
{
	struct driver_ctx *ctx = (struct driver_ctx *)driver->inst;

	return program_regex_rules_memory(ctx, data, data_len);
}

static struct regex_driver_queue *
bf2_driver_create_queue(struct regex_driver *driver)
{
	struct regex_driver_queue *queue;

	if (driver == NULL || driver->inst == NULL)
		return NULL;

	queue = malloc(sizeof(struct regex_driver_queue));
	if (queue == NULL)
		return NULL;

	memcpy(queue, &queue_template, sizeof(struct regex_driver_queue));
	queue->inst = create_driver_queue_ctx((struct driver_ctx *)driver->inst);
	if (queue->inst == NULL) {
		free(queue);
		return NULL;
	}

	return queue;
}

static void
bf2_queue_destroy(struct regex_driver_queue *queue)
{
	if (queue == NULL)
		return;

	if (queue->inst == NULL) {
		free(queue);
		return;
	}

	struct driver_queue_ctx *const ctx = (struct driver_queue_ctx *)queue->inst;

	if (ctx->channel != NULL) {
		doca_error_t const res = priv_doca_comp_events_channel_unregister(ctx->channel, ctx->qp_ctx->cq);
		/* In case create did not reach registration phase, not found may be returned.*/
		if (res != DOCA_SUCCESS && res != DOCA_ERROR_NOT_FOUND)
			DOCA_LOG_ERR("Failed to unregister DMA workq CQ from completion events channel. Reason: %s",
				     doca_get_error_string(res));

		priv_doca_dev_comp_events_channel_destroy(ctx->channel);
	}

	destroy_regex_qp_ctx(ctx->qp_ctx);

	if (ctx->uar != NULL)
		priv_doca_dev_uar_destroy(ctx->uar);

	free(queue->inst);
	free(queue);
}

static doca_error_t
bf2_get_event_channel(struct regex_driver_queue *queue, struct priv_doca_comp_events_channel **channel,
		      doca_event_handle_t *handle)
{
	doca_error_t res;

	if (queue == NULL || queue->inst == NULL) {
		DOCA_LOG_ERR("Invalid queue for create event channel.");
		return DOCA_ERROR_INVALID_VALUE;
	}

	struct driver_queue_ctx *const ctx = (struct driver_queue_ctx *)queue->inst;

	if (ctx->channel != NULL)
		goto populate_result;

	res = priv_doca_dev_comp_events_channel_create(ctx->driver_dev, /* blocking */ true, &ctx->channel);
	if (res != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to create completion events channel. Reason: %s", doca_get_error_string(res));
		return res;
	}

	res = priv_doca_comp_events_channel_handle_get(ctx->channel, &ctx->channel_handle);
	if (res != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to get completion events channel handle. Reason: %s", doca_get_error_string(res));
		priv_doca_dev_comp_events_channel_destroy(ctx->channel);
		return res;
	}

	res = priv_doca_comp_events_channel_register(ctx->channel, ctx->qp_ctx->cq);
	if (res != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Unable to register CQ with completion events channel. Reason: %s",
			     doca_get_error_string(res));
		priv_doca_dev_comp_events_channel_destroy(ctx->channel);
		return res;
	}

populate_result:
	*channel = ctx->channel;
	*handle = ctx->channel_handle;

	return DOCA_SUCCESS;
}

static bool
bf2_queue_is_empty(struct regex_driver_queue const *queue)
{

	if (queue == NULL || queue->inst == NULL) {
		DOCA_LOG_ERR("Invalid queue for is_empty check.");
		return DOCA_ERROR_INVALID_VALUE;
	}

	return priv_doca_cq_empty(((struct driver_queue_ctx *)queue->inst)->qp_ctx->cq);
}

static doca_error_t
bf2_queue_enqueue_job(struct regex_driver_queue *queue, struct regex_driver_job const *job)
{
	bool should_ring_doorbell;

	if (queue == NULL || queue->inst == NULL) {
		DOCA_LOG_ERR("Invalid queue for enqueue.");
		return DOCA_ERROR_INVALID_VALUE;
	}
	struct driver_queue_ctx *const ctx = (struct driver_queue_ctx *)queue->inst;
	struct regex_qp_ctx *const regex_qp = ctx->qp_ctx;

	if ((job->flags & DOCA_REGEX_SEARCH_JOB_FLAG_STOP_ON_ANY_MATCH) && (ctx->regexp_version == RXP_IDENTIFIER_BF2))
		return DOCA_ERROR_NOT_SUPPORTED;

	if (ctx->active_qp.devx_qp == NULL) {
		ctx->active_qp.hw_qpid = ffsll(regex_qp->free_qps);
		if (ctx->active_qp.hw_qpid == 0)
			return DOCA_ERROR_NO_MEMORY;

		--ctx->active_qp.hw_qpid; /* ffsll returns 1 for bit 0 */
		ctx->active_qp.devx_qp = &regex_qp->devx_qps[ctx->active_qp.hw_qpid];
	}

	uint64_t const nb_desc = get_free(ctx->active_qp.devx_qp);

	should_ring_doorbell = !job->allow_batching;
	if (nb_desc == 1) {
		/* Taking the last slot, mark qp exhausted for now */
		regex_qp->free_qps &= ~(1ULL << ctx->active_qp.hw_qpid);
		should_ring_doorbell = true;
	}

	prep_regex_umr_wqe_set(ctx, regex_qp, ctx->active_qp.devx_qp, job, 1);
	if (should_ring_doorbell) {
		send_doorbell(ctx, ctx->active_qp.devx_qp);
		/* Data sent so no pending data on qp. */
		ctx->active_qp.devx_qp = NULL;
	}

	++regex_qp->pi;

	return DOCA_SUCCESS;
}

static doca_error_t
bf2_queue_dequeue_jobs(struct regex_driver_queue *queue, struct doca_regex_mempool *matches_mp,
		       struct regex_driver_queue_result_batch *const result_batch)
{
	struct driver_queue_ctx *ctx;
	struct priv_mlx5_cqe cqe;
	struct regex_qp_ctx *regex_qp;
	uint16_t nb_dequeued = 0;
	uint16_t batch_slots_remaining = result_batch->responses_size;

	if (queue == NULL || queue->inst == NULL) {
		DOCA_LOG_ERR("Invalid queue for dequeue.");
		return DOCA_ERROR_INVALID_VALUE;
	}

	ctx = (struct driver_queue_ctx *)queue->inst;
	regex_qp = ctx->qp_ctx;

	while (batch_slots_remaining >= ctx->per_qp_depth &&
	       priv_doca_cq_poll_one(regex_qp->cq, (struct priv_doca_64b_cqe *)&cqe) != DOCA_ERROR_AGAIN) {
		if (!is_valid_regex_response(&cqe))
			continue;

		uint16_t wq_counter = (be16toh(cqe.wqe_counter) + 1) & PRIV_MLX5_REGEX_MAX_WQE_INDEX;
		size_t hw_qpid = cqe.user_index_bytes[2];
		struct devx_qp_ctx *devx_qp = &regex_qp->devx_qps[hw_qpid];

		/* UMR mode WQE counter - move as WQE set(4 WQEBBS).*/
		wq_counter >>= 2;

		while (devx_qp->ci != wq_counter) {
			uint32_t job_id;

			job_id = job_id_get(hw_qpid, get_qp_size(devx_qp), devx_qp->ci);
			extract_result(&regex_qp->jobs_metadata[job_id], result_batch->responses + nb_dequeued,
				       matches_mp);
			devx_qp->ci = (devx_qp->ci + 1) & (PRIV_MLX5_REGEX_MAX_WQE_INDEX >> 2);
			++nb_dequeued;
			--batch_slots_remaining;
		}
		regex_qp->free_qps |= (1ULL << hw_qpid);
	}

	regex_qp->ci += nb_dequeued;
	result_batch->nb_populated_responses = nb_dequeued;

	return (nb_dequeued == 0) ? DOCA_ERROR_AGAIN : DOCA_SUCCESS;
}

/**********************************************************************************************************************
 *
 * Public API
 *
 *********************************************************************************************************************/

struct regex_driver *
create_bf2_driver(struct regex_device_ops_impl *inst, struct doca_dev *dev)
{
	struct regex_driver *obj;
	struct driver_ctx *ctx;
	doca_error_t res;

	if (dev == NULL) {
		DOCA_LOG_ERR("Device must not be NULL.");
		return NULL;
	}

	obj = malloc(sizeof(struct regex_driver));
	if (obj == NULL)
		return NULL;

	memcpy(obj, &driver_template, sizeof(*obj));

	ctx = calloc(1, sizeof(struct driver_ctx));
	if (ctx == NULL) {
		free(obj);
		return NULL;
	}

	res = init_driver_ctx(ctx, dev);
	if (res != DOCA_SUCCESS) {
		free(ctx);
		free(obj);
		DOCA_LOG_ERR("Failed to open regex device: %s", doca_get_error_string(res));
		return NULL;
	}

	res = get_bf2_driver_info(inst, doca_dev_as_devinfo(dev), &ctx->properties);
	if (res != DOCA_SUCCESS) {
		free(ctx);
		free(obj);
		DOCA_LOG_ERR("Failed to determine device caps %s", doca_get_error_string(res));
		return NULL;
	}

	obj->inst = (struct regex_driver_impl *)ctx;
	obj->dev = dev;

	return obj;
}

doca_error_t
get_bf2_driver_info(struct regex_device_ops_impl *inst, struct doca_devinfo const *devinfo,
		    struct regex_driver_info *regex_driver_info)
{
	doca_error_t ret;

	if (devinfo == NULL || regex_driver_info == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = device_has_regex_support(priv_doca_devinfo_to_ibv_ctx(devinfo), NULL);
	if (ret == DOCA_SUCCESS) {
		regex_driver_info->max_hje_job_length = INT32_MAX;
		regex_driver_info->max_single_job_length = 16384;
		regex_driver_info->max_matches_per_job = 254;
		regex_driver_info->supports_hardware_acceleration = true;
	}

	return ret;
}
