/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * Definition of an abstract implementation to be used by doca regex, Provides a separation between implementation of
 * regex routines for the higher level functionality of the doca regex api.
 */

#ifndef LIBS_DOCA_REGEX_IMPL_DRIVERS_REGEX_DRIVER_H_
#define LIBS_DOCA_REGEX_IMPL_DRIVERS_REGEX_DRIVER_H_

#include <errno.h>
#include <stdbool.h>
#include <stdint.h>

#include <doca_buf.h>
#include <doca_dev.h>
#include <doca_error.h>
#include <doca_regex_mempool.h>
#include <doca_types.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declarations */
struct regex_driver;
struct regex_driver_impl;
struct regex_driver_queue;
struct regex_driver_queue_impl;
struct priv_doca_comp_events_channel;

/**
 * Properties that vary between drivers / implementations
 */
struct regex_driver_info {
	/** Maximum job size in bytes .*/
	uint64_t max_hje_job_length;
	/** Maximum single job length */
	uint64_t max_single_job_length;
	/** Maximum number of matches per job. */
	uint64_t max_matches_per_job;
	/** Is the device hardware accelerated */
	bool supports_hardware_acceleration;
};

/**
 * Data required by a driver to handle a job
 */
struct regex_driver_job {
	/** ID to use to correlate this request with a response. */
	uint64_t job_id;
	/** Group id set used to select a subset of rules to match against. */
	uint16_t rule_group_ids[4];
	/** Flags to be used when scanning the job */
	uint32_t flags;
	/** data to be scanned. */
	struct doca_buf const *buffer;
	/** Offset into buffer to start scanning. */
	uint32_t buffer_offset;
	/** Number of bytes following the given offset to scan. */
	uint32_t buffer_bytes_to_send;
	/** Is a driver allowed to defer the execution of this job into a batch to improve performance. */
	bool allow_batching;
};

/**
 * Generic form of response which can hold all the data returned from the regex device. This type is purposely unique
 * from doca_regex_search_result so they can vary independently.
 */
struct regex_driver_response {
	uint64_t id;

	/** Response flags. A bit masked field for zero or more status flags. See doca_regex_status_flag. */
	uint64_t status_flags;

	/** Total number of detected matches. */
	uint32_t detected_matches;

	/** Total number of returned matches. */
	uint32_t num_matches;

	/** Returned matches. Contains num_matches elements as a linked list. */
	struct doca_regex_match *matches;

	/** Memory pool owning the matches. */
	struct doca_regex_mempool *matches_mempool;
};

/**
 * A batch of results from a driver queue
 */
struct regex_driver_queue_result_batch {
	/** Pointer to array of responses which contains at least responses_size elements */
	struct regex_driver_response *responses;
	/** Size of responses array */
	uint32_t responses_size;
	/** Number of elements in responses that have been populated */
	uint32_t nb_populated_responses;
};

/**
 * An abstract queue provided by a driver to facilitate enqueue and dequeue of jobs
 */
struct regex_driver_queue {
	/** Opaque queue handle */
	struct regex_driver_queue_impl *inst;

	/**
	 * @brief
	 * Destructor
	 *
	 * @param [in] queue
	 * Queue to destroy.
	 */
	void (*destroy_fn)(struct regex_driver_queue *queue);

	/**
	 * @brief
	 * Add ability to be used with event driven mode
	 *
	 * @param [in] queue
	 * Instance of the queue.
	 *
	 * @param [out] channel
	 * Upon success this will contain a NON-OWNING pointer to the channel. The caller must NOT attempt to destroy
	 * this object. The channel is owned and managed by the driver queue.
	 *
	 * @param [out] handle
	 * Upon success this will contain a handle for the channel, used to register and de-register with the channel.
	 *
	 * @return
	 * DOCA_SUCCESS - Success.
	 * DOCA_ERROR_NO_MEMORY
	 * DOCA_ERROR_NOT_SUPPORTED
	 */
	doca_error_t (*create_event_channel_fn)(struct regex_driver_queue *queue,
						struct priv_doca_comp_events_channel **channel,
						doca_event_handle_t *handle);

	/**
	 * Query the queue to determine if queue holds any results or not.
	 *
	 * @param [in] queue
	 * Instance of the queue.
	 *
	 * @return true if the queue is empty, false if the queue has results that could be read by calling
	 * dequeue_jobs_fn
	 */
	bool (*is_empty_fn)(struct regex_driver_queue const *queue);
	/**
	 * @brief
	 * Send a job to the regex engine to be processed.
	 *
	 * @param [in] queue
	 * Instance of the queue.
	 *
	 * @param [in] job
	 * Details of the job. Caller retains ownership of these objects;
	 *
	 * @return
	 * DOCA_SUCCESS - Success.
	 * DOCA_ERROR_INVALID_VALUE - Invalid argument(s) provided.
	 * DOCA_ERROR_NO_MEMORY - Queue is full.
	 */
	doca_error_t (*enqueue_job_fn)(struct regex_driver_queue *queue, struct regex_driver_job const *job);

	/**
	 * @brief
	 * Fetch a batch of job responses from the queue.
	 *
	 * @note
	 * It is expected that the batch is always empty (nb_populated_responses == ) and that it is responses_size
	 * large enough to receive every response that can be notified by a single CQE (so it may need to be as large
	 * as the maximum QP depth)
	 *
	 * @param [in] queue
	 * Instance of the queue.
	 *
	 * @param [in] matches_mp
	 * To get matches back from the regex search provide a non null memory pool which will return items which are
	 * least as big as a doca_regex_match. Provide A NULL pool to just get a summary of the number of matches
	 * detected.
	 *
	 * @param [out] result_batch
	 * Pointer to a batch of results.
	 *
	 * @return
	 * DOCA_SUCCESS - Success, result_batch::nb_populated_responses specifies how many elements are populated.
	 * DOCA_ERROR_INVALID_VALUE - Invalid argument(s) provided.
	 * DOCA_ERROR_AGAIN - Nothing to dequeue.
	 */
	doca_error_t (*dequeue_jobs_fn)(struct regex_driver_queue *queue, struct doca_regex_mempool *matches_mp,
					struct regex_driver_queue_result_batch *result_batch);
};

/**
 * Structure containing all the functionality required of a regex device by the DOCA regex.
 *
 * @details
 * Allows each DOCA regex instance to be supported by any regex_impl. Such as Hyperscan, DPDK regex, or any other regex
 * library that can be adapted to this interface.
 */
struct regex_driver {
	/** A place to bind any required extra data into the impl. */
	struct regex_driver_impl *inst;

	/** The doca_dev used by this driver */
	struct doca_dev *dev;

	/**
	 * @brief
	 * Function suitable to destroy this doca_regex_device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @details
	 * Set this to NULL to indicate that the engine should not attempt to destroy this object when the engine is
	 * destroyed.
	 */
	void (*destroy_fn)(struct regex_driver *inst);

	/**
	 * @brief
	 * Query information about the regex device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @return
	 * Populated properties structure
	 *
	 * @note This function MUST NOT return NULL when driver points to a valid driver
	 */
	struct regex_driver_info const *(*get_properties_fn)(struct regex_driver const *inst);

	/**
	 * @brief
	 * Function to transform the uncompiled rules into compiled/programmable rules.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] rules_data
	 * Input uncompiled rules.
	 *
	 * @param [in] rules_len
	 * Input uncompiled rules data length.
	 *
	 * @param [out] out_compiled_rules_data
	 * The compiled rules data.
	 *
	 * @param [out] out_compiled_rules_len
	 * The compiled rules data length.
	 *
	 * @return
	 * DOCA_SUCCESS or DOCA_ERROR_XXXX.
	 */
	doca_error_t (*compile_rules_fn)(struct regex_driver *inst, void const *rules_data, uint32_t rules_len,
					 void **out_compiled_rules_data, uint32_t *out_compiled_rules_len);

	/**
	 * @brief
	 * Program regex device with the supplied compiled rules data.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] data
	 * A blob of memory in the same format that would be produced by compile_rules_fn. However it may have been
	 * read from a pre-compiled input file and not necessarily compiled by this regex_impl. It will be the
	 * applications responsibility to ensure that when providing a path to a pre-compiled data within
	 * doca_regex_rules_cfg that the file contains data in a format suitable for the selected regex_impl.
	 *
	 * @param [in] data_len
	 * Number of bytes in the blob pointed to by data.
	 *
	 * @param [in] incremental
	 * Set to true when providing additional rules to a regex_impl, false
	 * to start from scratch.
	 *
	 * @return
	 * DOCA_SUCCESS - Success.
	 * DOCA_ERROR_INVALID_VALUE - Invalid argument(s) provided.
	 * DOCA_ERROR_NO_MEMORY - Unable to allocate the resources needed to perform rules programming.
	 * DOCA_ERROR_NOT_SUPPORTED - Device rejected this attempt to configure it's rules database.
	 */
	doca_error_t (*program_compiled_rules_fn)(struct regex_driver *inst, void const *data, uint32_t data_len);

	/**
	 * Create a new queue for use with this driver.
	 *
	 * @details
	 * Queues are designed to be used by a single thread, a single thread can have multiple queues if it wishes.
	 *
	 * @param [in] driver
	 * Driver instance.
	 *
	 * @return
	 * Non NULL upon success, NULL otherwise
	 */
	struct regex_driver_queue *(*create_queue_fn)(struct regex_driver *driver);
};

static inline void
regex_driver_destroy(struct regex_driver *driver)
{
	if (driver != NULL)
		driver->destroy_fn(driver);
}

static inline uint32_t
regex_driver_get_max_single_job_size(struct regex_driver const *driver)
{
	struct regex_driver_info const *properties = NULL;

	if (driver != NULL && driver->inst != NULL)
		properties = driver->get_properties_fn(driver);

	return properties == NULL ? 0 : properties->max_single_job_length;
}

static inline doca_error_t
regex_driver_compile_rules(struct regex_driver *driver, void const *rules_data, uint32_t rules_len,
			   void **out_compiled_rules_data, uint32_t *out_compiled_rules_len)
{
	if (driver != NULL && driver->inst != NULL)
		return driver->compile_rules_fn(driver, rules_data, rules_len, out_compiled_rules_data,
						out_compiled_rules_len);

	return DOCA_ERROR_INVALID_VALUE;
}

static inline doca_error_t
regex_driver_program_compiled_rules(struct regex_driver *driver, void const *data, uint32_t data_len)
{
	if (driver != NULL && driver->inst != NULL)
		return driver->program_compiled_rules_fn(driver, data, data_len);

	return DOCA_ERROR_INVALID_VALUE;
}

static inline struct regex_driver_queue *
regex_driver_create_queue(struct regex_driver *driver)
{
	if (driver != NULL && driver->inst != NULL)
		return driver->create_queue_fn(driver);

	return NULL;
}

static inline void
regex_driver_queue_destroy(struct regex_driver_queue *queue)
{
	if (queue != NULL && queue->inst != NULL)
		queue->destroy_fn(queue);
}

static inline doca_error_t
regex_driver_queue_get_event_channel(struct regex_driver_queue *queue, struct priv_doca_comp_events_channel **channel,
				     doca_event_handle_t *handle)
{
	return queue->create_event_channel_fn(queue, channel, handle);
}

static inline bool
regex_driver_queue_is_empty(struct regex_driver_queue const *queue)
{
	return queue->is_empty_fn(queue);
}

static inline doca_error_t
regex_driver_queue_enqueue(struct regex_driver_queue *queue, struct regex_driver_job const *job)
{
	return queue->enqueue_job_fn(queue, job);
}

static inline doca_error_t
regex_driver_queue_dequeue(struct regex_driver_queue *queue, struct doca_regex_mempool *matches_mp,
			   struct regex_driver_queue_result_batch *result_batch)
{
	return queue->dequeue_jobs_fn(queue, matches_mp, result_batch);
}

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* LIBS_DOCA_REGEX_IMPL_DRIVERS_REGEX_DRIVER_H_ */
