/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "regex_device_ops.h"

#include <string.h>

#include <doca_log.h>

#ifdef DOCA_REGEX_SUPPORTS_BF2
#include <impl/drivers/bf/bf2_driver.h>
#endif
#ifdef DOCA_REGEX_SUPPORTS_HS
#include <impl/drivers/hs/hyperscan_driver.h>
#endif

DOCA_LOG_REGISTER(REGEX_DEVICE_OPS);

#define OPS_TABLE_SIZE 2

/**********************************************************************************************************************
 *
 * Default ops
 *
 *********************************************************************************************************************/
/**
 * Immutable default ops
 */
static const struct regex_device_ops default_bf2_ops = {
#ifdef DOCA_REGEX_SUPPORTS_BF2
	.inst = NULL,
	.get_driver_info_fn = get_bf2_driver_info,
	.create_driver_fn = create_bf2_driver,
#else  /* !DOCA_REGEX_SUPPORTS_BF2 */
	NULL, NULL, NULL
#endif /* DOCA_REGEX_SUPPORTS_BF2 */
};

static const struct regex_device_ops default_hs_ops = {
#ifdef DOCA_REGEX_SUPPORTS_HS
	.inst = NULL,
	.get_driver_info_fn = get_hyperscan_driver_info,
	.create_driver_fn = create_hyperscan_driver,

#else  /* !DOCA_REGEX_SUPPORTS_HS */
	NULL, NULL, NULL
#endif /* DOCA_REGEX_SUPPORTS_HS */
};

/**
 * Runtime mutable ops table. Unit tests will modify this to allow for injecting mocks.
 */
static const struct regex_device_ops *ops_table[OPS_TABLE_SIZE] = {
	&default_hs_ops,
	&default_bf2_ops,
};

/**********************************************************************************************************************
 *
 * Public (internal) API
 *
 *********************************************************************************************************************/

bool
is_device_kind_supported(enum regex_device_kind kind)
{
	return kind >= REGEX_DEVICE_KIND_HS && kind < OPS_TABLE_SIZE && ops_table[kind]->create_driver_fn != NULL;
}

struct regex_device_ops const *
get_regex_ops_for_device_kind(enum regex_device_kind kind)
{
	if (kind >= REGEX_DEVICE_KIND_HS && kind < OPS_TABLE_SIZE)
		return ops_table[kind];

	return NULL;
}

void
set_regex_ops_for_device_kind(enum regex_device_kind kind, struct regex_device_ops const *ops)
{
	if (kind >= REGEX_DEVICE_KIND_HS && kind < OPS_TABLE_SIZE)
		ops_table[kind] = ops;
}

void
reset_all_regex_device_ops()
{
	ops_table[REGEX_DEVICE_KIND_HS] = &default_hs_ops;
	ops_table[REGEX_DEVICE_KIND_BF2] = &default_bf2_ops;
}
