/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef LIBS_DOCA_REGEX_IMPL_REGEX_DEVICE_OPS_H_
#define LIBS_DOCA_REGEX_IMPL_REGEX_DEVICE_OPS_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include <doca_dev.h>
#include <doca_error.h>

#include <priv_doca_compat.h>

#include <impl/drivers/regex_driver.h>

/**********************************************************************************************************************
 *
 * Public types
 *
 *********************************************************************************************************************/
enum regex_device_kind {
	REGEX_DEVICE_KIND_HS,
	REGEX_DEVICE_KIND_BF2,
};

struct regex_device_ops_impl;

/**
 * Operations that can be performed per device kind.
 */
struct regex_device_ops {
	struct regex_device_ops_impl *inst;

	/**
	 * @brief
	 * Get some details about the driver that could be created by this devinfo.
	 *
	 * @param [in] inst
	 * Implementation specific data.
	 *
	 * @param [in] devinfo
	 * Devinfo to use to execute caps query.
	 *
	 * @param [out] reger_device_info
	 * Driver information and capabilities
	 *
	 * @return
	 * DOCA_SUCCESS on success or an appropriate DOCA_ERROR_XXX.
	 */
	doca_error_t (*get_driver_info_fn)(struct regex_device_ops_impl *inst, struct doca_devinfo const *devinfo,
					   struct regex_driver_info *regex_driver_info);

	/**
	 * @brief
	 * Create a new driver from a given device.
	 *
	 * @param [in] inst
	 * Implementation specific data.
	 *
	 * @param [in] dev
	 * Device to create a driver from.
	 *
	 * @return
	 * Non NULL on success, otherwise NULL.
	 */
	struct regex_driver *(*create_driver_fn)(struct regex_device_ops_impl *inst, struct doca_dev *dev);
};

/**********************************************************************************************************************
 *
 * Public (internal) API
 *
 *********************************************************************************************************************/

/**
 * Is a given device kind supported.
 */
DOCA_PRIVATE
bool is_device_kind_supported(enum regex_device_kind kind);

/**
 * Get the ops supported for a given device kind.
 */
DOCA_PRIVATE
struct regex_device_ops const *get_regex_ops_for_device_kind(enum regex_device_kind kind);

/**
 * Set the ops for a given device kind.
 */
DOCA_PRIVATE
void set_regex_ops_for_device_kind(enum regex_device_kind kind, struct regex_device_ops const *ops);

/**
 * Reset All ops to their default values.
 */
DOCA_PRIVATE
void reset_all_regex_device_ops();

/**
 * inline helper
 */
static inline doca_error_t
query_regex_device_kind_info(enum regex_device_kind kind, struct doca_devinfo const *devinfo,
			     struct regex_driver_info *regex_driver_info)
{
	struct regex_device_ops const *ops = get_regex_ops_for_device_kind(kind);

	if (ops == NULL || regex_driver_info == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (ops->get_driver_info_fn == NULL)
		return DOCA_ERROR_NOT_SUPPORTED;

	return ops->get_driver_info_fn(ops->inst, devinfo, regex_driver_info);
}

/**
 * inline helper
 */
static inline struct regex_driver *
create_regex_device_driver(enum regex_device_kind kind, struct doca_dev *dev)
{
	struct regex_device_ops const *ops = get_regex_ops_for_device_kind(kind);

	if (ops == NULL || ops->create_driver_fn == NULL)
		return NULL;

	return ops->create_driver_fn(ops->inst, dev);
}

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* LIBS_DOCA_REGEX_IMPL_REGEX_DEVICE_OPS_H_ */
