/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "doca_mlx5_common.h"

#include <stdlib.h>

#include <doca_log.h>

#include "doca_mlx5_prm.h"

DOCA_LOG_REGISTER(BF2_DOCA_MLX5_COMMON);

struct doca_mlx5_devx_obj *
doca_mlx5_devx_cmd_mkey_create(void *ctx, const struct doca_mlx5_devx_mkey_attr *attr)
{
	struct doca_mlx5_klm *klm_array = attr->klm_array;
	int klm_num = attr->klm_num;
	int in_size_dw = DOCA_MLX5_ST_SZ_DW(create_mkey_in) +
			 (klm_num ? DOCA_MLX5_ALIGN(klm_num, 4) : 0) * DOCA_MLX5_ST_SZ_DW(klm);
	uint32_t in[in_size_dw];
	uint32_t out[DOCA_MLX5_ST_SZ_DW(create_mkey_out)] = {0};
	struct doca_mlx5_devx_obj *mkey;
	uint32_t translation_size;
	size_t pgsize;
	void *mkc;

	mkey = calloc(1, sizeof(*mkey));
	if (!mkey) {
		DOCA_DLOG_ERR("no memory");
		return NULL;
	}
	memset(in, 0, in_size_dw * 4);

	pgsize = get_mem_page_size();
	if (pgsize == (size_t)-1) {
		free(mkey);
		DOCA_DLOG_ERR("Failed to get page size");
		return NULL;
	}
	DOCA_MLX5_SET(create_mkey_in, in, opcode, DOCA_MLX5_CMD_OP_CREATE_MKEY);
	mkc = DOCA_MLX5_ADDR_OF(create_mkey_in, in, memory_key_mkey_entry);
	/* NOTE: I don't think we'll ever need > 0 here. */
	if (klm_num > 0) {
		int i;
		uint8_t *klm = (uint8_t *)DOCA_MLX5_ADDR_OF(create_mkey_in, in, klm_pas_mtt);

		translation_size = DOCA_MLX5_ALIGN(klm_num, 4);
		for (i = 0; i < klm_num; i++) {
			DOCA_MLX5_SET(klm, klm, byte_count, klm_array[i].byte_count);
			DOCA_MLX5_SET(klm, klm, mkey, klm_array[i].mkey);
			DOCA_MLX5_SET64(klm, klm, address, klm_array[i].address);
			klm += DOCA_MLX5_ST_SZ_BYTES(klm);
		}
		for (; i < (int)translation_size; i++) {
			DOCA_MLX5_SET(klm, klm, mkey, 0x0);
			DOCA_MLX5_SET64(klm, klm, address, 0x0);
			klm += DOCA_MLX5_ST_SZ_BYTES(klm);
		}
		DOCA_MLX5_SET(mkc, mkc, access_mode_1_0,
			      attr->log_entity_size ? DOCA_MLX5_MKC_ACCESS_MODE_KLM_FBS
						    : DOCA_MLX5_MKC_ACCESS_MODE_KLM);
		DOCA_MLX5_SET(mkc, mkc, log_page_size, attr->log_entity_size);
	} else {
		translation_size = (DOCA_MLX5_ALIGN(attr->size, pgsize) * 8) / 16;
		DOCA_MLX5_SET(mkc, mkc, access_mode_1_0, DOCA_MLX5_MKC_ACCESS_MODE_MTT);
		DOCA_MLX5_SET(mkc, mkc, log_page_size, bf2_doca_mlx5_log2_u32(pgsize));
	}
	DOCA_MLX5_SET(create_mkey_in, in, translations_octword_actual_size, translation_size);
	DOCA_MLX5_SET(create_mkey_in, in, mkey_umem_id, attr->umem_id);
	DOCA_MLX5_SET(create_mkey_in, in, pg_access, attr->pg_access);
	DOCA_MLX5_SET(mkc, mkc, lw, 0x1);
	DOCA_MLX5_SET(mkc, mkc, lr, 0x1);
	DOCA_MLX5_SET(mkc, mkc, qpn, 0xffffff);
	DOCA_MLX5_SET(mkc, mkc, pd, attr->pd);
	DOCA_MLX5_SET(mkc, mkc, mkey_7_0, attr->umem_id & 0xFF);
	DOCA_MLX5_SET(mkc, mkc, umr_en, attr->umr_en);
	DOCA_MLX5_SET(mkc, mkc, translations_octword_size, translation_size);
	DOCA_MLX5_SET(mkc, mkc, relaxed_ordering_write, attr->relaxed_ordering_write);
	DOCA_MLX5_SET(mkc, mkc, relaxed_ordering_read, attr->relaxed_ordering_read);
	DOCA_MLX5_SET64(mkc, mkc, start_addr, attr->addr);
	DOCA_MLX5_SET64(mkc, mkc, len, attr->size);
	mkey->obj = mlx5dv_devx_obj_create(ctx, in, in_size_dw * 4, out, sizeof(out));
	if (mkey->obj == NULL) {
		DOCA_DLOG_ERR("Can't create %sdirect mkey - error %d\n", klm_num ? "an in" : "a ",
			     errno);
		free(mkey);
		return NULL;
	}
	mkey->id = DOCA_MLX5_GET(create_mkey_out, out, mkey_index);
	mkey->id = (mkey->id << 8) | (attr->umem_id & 0xFF);

	return mkey;
}

struct doca_mlx5_devx_dbr_page *
doca_mlx5_alloc_dbr_page(void *ctx)
{
	struct doca_mlx5_devx_dbr_page *page;

	/* Allocate space for door-bell page and management data. */
	page = aligned_alloc(CACHELINE_SIZE, sizeof(struct doca_mlx5_devx_dbr_page));
	if (!page) {
		DOCA_DLOG_ERR("cannot allocate dbr page");
		return NULL;
	}
	/* Did align malloc so set to 0 now - MAYBE A NEW FUNCTION FOR THIS?*/
	memset(page, 0, sizeof(struct doca_mlx5_devx_dbr_page));

	/* Register allocated memory. */
	page->umem = mlx5dv_devx_umem_reg(ctx, page->dbrs, DOCA_MLX5_DBR_PAGE_SIZE, 0);

	if (!page->umem) {
		DOCA_DLOG_ERR("cannot umem reg dbr page");
		free(page);
		return NULL;
	}

	return page;
}

int64_t
doca_mlx5_get_dbr(void *ctx, struct doca_mlx5_dbr_page_list *head,
		  struct doca_mlx5_devx_dbr_page **dbr_page)
{
	struct doca_mlx5_devx_dbr_page *page = NULL;
	uint32_t i, j;

	LIST_FOREACH(page, head, next)
	if (page->dbr_count < DOCA_MLX5_DBR_PER_PAGE)
		break;

	if (!page) { /* No page with free door-bell exists. */
		page = doca_mlx5_alloc_dbr_page(ctx);
		if (!page) /* Failed to allocate new page. */
			return (-1);
		LIST_INSERT_HEAD(head, page, next);
	}
	/* Loop to find bitmap part with clear bit. */
	for (i = 0; i < DOCA_MLX5_DBR_BITMAP_SIZE && page->dbr_bitmap[i] == UINT64_MAX; i++)
		; /* Empty. */
	/* Find the first clear bit. */
	DOCA_MLX5_ASSERT(i < DOCA_MLX5_DBR_BITMAP_SIZE);
	j = bf2_doca_mlx5_bsf64(~page->dbr_bitmap[i]);
	page->dbr_bitmap[i] |= (UINT64_C(1) << j);
	page->dbr_count++;
	*dbr_page = page;

	return (i * CHAR_BIT * sizeof(uint64_t) + j) * DOCA_MLX5_DBR_SIZE;
}

int32_t
doca_mlx5_release_dbr(struct doca_mlx5_dbr_page_list *head, uint32_t umem_id, uint64_t offset)
{
	struct doca_mlx5_devx_dbr_page *page = NULL;
	int ret = 0;

	LIST_FOREACH(page, head, next)
	/* Find the page this address belongs to. */
	if (doca_mlx5_os_get_umem_id(page->umem) == umem_id)
		break;
	if (!page)
		return -EINVAL;
	page->dbr_count--;
	if (!page->dbr_count) {
		/* Page not used, free it and remove from list. */
		LIST_REMOVE(page, next);
		if (page->umem)
			ret = -mlx5dv_devx_umem_dereg(page->umem);
		free(page);
	} else {
		/* Mark in bitmap that this door-bell is not in use. */
		offset /= DOCA_MLX5_DBR_SIZE;
		int i = offset / 64;
		int j = offset % 64;

		page->dbr_bitmap[i] &= ~(UINT64_C(1) << j);
	}
	return ret;
}

struct doca_mlx5_devx_obj *
doca_mlx5_devx_cmd_create_cq(void *ctx, const struct doca_mlx5_devx_cq_attr *attr)
{
	uint32_t in[DOCA_MLX5_ST_SZ_DW(create_cq_in)] = {0};
	uint32_t out[DOCA_MLX5_ST_SZ_DW(create_cq_out)] = {0};
	struct doca_mlx5_devx_obj *cq_obj;
	void *cqctx;

	cq_obj = calloc(1, sizeof(*cq_obj));
	if (!cq_obj) {
		DOCA_DLOG_ERR("Failed to allocate CQ object memory.");
		return NULL;
	}

	cqctx = DOCA_MLX5_ADDR_OF(create_cq_in, in, cq_context);
	DOCA_MLX5_SET(create_cq_in, in, opcode, DOCA_MLX5_CMD_OP_CREATE_CQ);
	if (attr->db_umem_valid) {
		DOCA_MLX5_SET(cqc, cqctx, dbr_umem_valid, attr->db_umem_valid);
		DOCA_MLX5_SET(cqc, cqctx, dbr_umem_id, attr->db_umem_id);
		DOCA_MLX5_SET64(cqc, cqctx, dbr_addr, attr->db_umem_offset);
	} else {
		DOCA_MLX5_SET64(cqc, cqctx, dbr_addr, attr->db_addr);
	}
	DOCA_MLX5_SET(cqc, cqctx, cqe_sz,
		      (CACHELINE_SIZE == 128) ? DOCA_MLX5_CQE_SIZE_128B : DOCA_MLX5_CQE_SIZE_64B);

	DOCA_MLX5_SET(cqc, cqctx, cc, attr->use_first_only);
	DOCA_MLX5_SET(cqc, cqctx, oi, attr->overrun_ignore);
	DOCA_MLX5_SET(cqc, cqctx, log_cq_size, attr->log_cq_size);
	DOCA_MLX5_SET(cqc, cqctx, log_page_size,
		      attr->log_page_size - DOCA_MLX5_ADAPTER_PAGE_SHIFT);
	DOCA_MLX5_SET(cqc, cqctx, c_eqn, attr->eqn);
	DOCA_MLX5_SET(cqc, cqctx, uar_page, attr->uar_page_id);
	DOCA_MLX5_SET(cqc, cqctx, cqe_comp_en, !!attr->cqe_comp_en);
	DOCA_MLX5_SET(cqc, cqctx, mini_cqe_res_format, attr->mini_cqe_res_format);
	DOCA_MLX5_SET(cqc, cqctx, mini_cqe_res_format_ext, attr->mini_cqe_res_format_ext);
	if (attr->q_umem_valid) {
		DOCA_MLX5_SET(create_cq_in, in, cq_umem_valid, attr->q_umem_valid);
		DOCA_MLX5_SET(create_cq_in, in, cq_umem_id, attr->q_umem_id);
		DOCA_MLX5_SET64(create_cq_in, in, cq_umem_offset, attr->q_umem_offset);
	}

	cq_obj->obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!cq_obj->obj) {
		DOCA_DLOG_ERR("Failed to create CQ using DevX\n");
		free(cq_obj);
		return NULL;
	}
	cq_obj->id = DOCA_MLX5_GET(create_cq_out, out, cqn);

	return cq_obj;
}

void
doca_mlx5_devx_cq_destroy(struct doca_mlx5_devx_cq *cq)
{
	if (cq->cq) {
		mlx5dv_devx_obj_destroy(cq->cq->obj);
		free(cq->cq);
	}
	if (cq->umem_obj)
		mlx5dv_devx_umem_dereg(cq->umem_obj);
	if (cq->umem_buf)
		free((void *)(uintptr_t)cq->umem_buf);
}

static struct doca_mlx5_devx_obj *
doca_mlx5_devx_cmd_create_qp(void *ctx, struct doca_mlx5_devx_qp_attr *attr)
{
	uint32_t in[DOCA_MLX5_ST_SZ_DW(create_qp_in)] = {0};
	uint32_t out[DOCA_MLX5_ST_SZ_DW(create_qp_out)] = {0};
	struct doca_mlx5_devx_obj *qp_obj = calloc(1, sizeof(*qp_obj));
	void *qpc = DOCA_MLX5_ADDR_OF(create_qp_in, in, qpc);

	if (!qp_obj) {
		DOCA_DLOG_ERR("Failed to allocate QP data.");
		return NULL;
	}
	DOCA_MLX5_SET(create_qp_in, in, opcode, DOCA_MLX5_CMD_OP_CREATE_QP);
	DOCA_MLX5_SET(qpc, qpc, st, DOCA_MLX5_QP_ST_RC);
	DOCA_MLX5_SET(qpc, qpc, pd, attr->pd);
	DOCA_MLX5_SET(qpc, qpc, ts_format, attr->ts_format);
	DOCA_MLX5_SET(qpc, qpc, user_index, attr->user_index);
	if (attr->uar_index) {
		if (attr->mmo) {
			void *qpc_ext_and_pas_list =
				DOCA_MLX5_ADDR_OF(create_qp_in, in, qpc_extension_and_pas_list);
			void *qpc_ext = DOCA_MLX5_ADDR_OF(qpc_extension_and_pas_list,
							  qpc_ext_and_pas_list, qpc_data_extension);

			DOCA_MLX5_SET(create_qp_in, in, qpc_ext, 1);
			DOCA_MLX5_SET(qpc_extension, qpc_ext, mmo, 1);
		}
		DOCA_MLX5_SET(qpc, qpc, pm_state, DOCA_MLX5_QP_PM_MIGRATED);
		DOCA_MLX5_SET(qpc, qpc, uar_page, attr->uar_index);
		if (attr->log_page_size > DOCA_MLX5_ADAPTER_PAGE_SHIFT)
			DOCA_MLX5_SET(qpc, qpc, log_page_size,
				      attr->log_page_size - DOCA_MLX5_ADAPTER_PAGE_SHIFT);
		if (attr->sq_size) {
			DOCA_MLX5_SET(qpc, qpc, cqn_snd, attr->cqn);
			DOCA_MLX5_SET(qpc, qpc, log_sq_size, bf2_doca_mlx5_log2_u32(attr->sq_size));
		} else {
			DOCA_MLX5_SET(qpc, qpc, no_sq, 1);
		}
		if (attr->rq_size) {
			DOCA_MLX5_SET(qpc, qpc, cqn_rcv, attr->cqn);
			DOCA_MLX5_SET(qpc, qpc, log_rq_stride,
				      attr->log_rq_stride - DOCA_MLX5_LOG_RQ_STRIDE_SHIFT);
			DOCA_MLX5_SET(qpc, qpc, log_rq_size, bf2_doca_mlx5_log2_u32(attr->rq_size));
			DOCA_MLX5_SET(qpc, qpc, rq_type, DOCA_MLX5_NON_ZERO_RQ);
		} else {
			DOCA_MLX5_SET(qpc, qpc, rq_type, DOCA_MLX5_ZERO_LEN_RQ);
		}
		if (attr->dbr_umem_valid) {
			DOCA_MLX5_SET(qpc, qpc, dbr_umem_valid, attr->dbr_umem_valid);
			DOCA_MLX5_SET(qpc, qpc, dbr_umem_id, attr->dbr_umem_id);
		}
		DOCA_MLX5_SET64(qpc, qpc, dbr_addr, attr->dbr_address);
		DOCA_MLX5_SET64(create_qp_in, in, wq_umem_offset, attr->wq_umem_offset);
		DOCA_MLX5_SET(create_qp_in, in, wq_umem_id, attr->wq_umem_id);
		DOCA_MLX5_SET(create_qp_in, in, wq_umem_valid, 1);
	} else {
		/* Special QP to be managed by FW - no SQ\RQ\CQ\UAR\DB rec. */
		DOCA_MLX5_SET(qpc, qpc, rq_type, DOCA_MLX5_ZERO_LEN_RQ);
		DOCA_MLX5_SET(qpc, qpc, no_sq, 1);
	}
	qp_obj->obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!qp_obj->obj) {
		DOCA_DLOG_ERR("Failed to create QP Obj using DevX.");
		free(qp_obj);
		return NULL;
	}
	qp_obj->id = DOCA_MLX5_GET(create_qp_out, out, qpn);

	return qp_obj;
}

int
doca_mlx5_devx_qp_create(void *ctx, struct doca_mlx5_devx_qp *qp_obj, uint16_t log_wqbb_n,
			 struct doca_mlx5_devx_qp_attr *attr)
{
	struct doca_mlx5_devx_obj *qp = NULL;
	struct mlx5dv_devx_umem *umem_obj = NULL;
	void *umem_buf = NULL;
	size_t alignment = get_mem_page_size();
	uint32_t umem_size, umem_dbrec;
	uint16_t qp_size = 1 << log_wqbb_n;
	int ret;

	if (alignment == (size_t)-1) {
		DOCA_DLOG_ERR("Failed to get WQE buf alignment.");
		return -ENOMEM;
	}

	/* Allocate memory buffer for WQEs and doorbell record. */
	umem_size = sizeof(struct doca_mlx5_wqe) * qp_size;
	umem_dbrec = DOCA_MLX5_ALIGN(umem_size, DOCA_MLX5_DBR_SIZE);
	umem_size += DOCA_MLX5_DBR_SIZE;
	umem_size += alignment - (umem_size % alignment);
	umem_buf = aligned_alloc(alignment, umem_size);
	if (!umem_buf) {
		DOCA_DLOG_ERR("Failed to allocate memory for QP.");
		return -ENOMEM;
	}
	memset(umem_buf, 0, umem_size);

	/* Register allocated buffer in user space with DevX. */
	umem_obj = mlx5dv_devx_umem_reg(ctx, (void *)(uintptr_t)umem_buf, umem_size, 1);
	if (!umem_obj) {
		DOCA_DLOG_ERR("Failed to register umem for QP.");
		ret = errno;
		goto error;
	}

	/* Fill attributes for SQ object creation. */
	attr->wq_umem_id = doca_mlx5_os_get_umem_id(umem_obj);
	attr->wq_umem_offset = 0;
	attr->dbr_umem_valid = 1;
	attr->dbr_umem_id = attr->wq_umem_id;
	attr->dbr_address = umem_dbrec;
	attr->log_page_size = DOCA_MLX5_LOG_PAGE_SIZE;
	/* Create send queue object with DevX. */
	qp = doca_mlx5_devx_cmd_create_qp(ctx, attr);
	if (!qp) {
		DOCA_DLOG_ERR("Can't create DevX QP object.");
		ret = -ENOMEM;
		goto error;
	}

	qp_obj->umem_buf = umem_buf;
	qp_obj->umem_obj = umem_obj;
	qp_obj->qp = qp;
	qp_obj->db_rec = DOCA_PTR_ADD(qp_obj->umem_buf, umem_dbrec);

	return 0;
error:
	if (umem_obj)
		mlx5dv_devx_umem_dereg(umem_obj);
	if (umem_buf)
		free((void *)(uintptr_t)umem_buf);

	return ret;
}

void
doca_mlx5_devx_qp_destroy(struct doca_mlx5_devx_qp *qp)
{
	if (qp->qp) {
		mlx5dv_devx_obj_destroy(qp->qp->obj);
		free(qp->qp);
	}
	if (qp->umem_obj)
		mlx5dv_devx_umem_dereg(qp->umem_obj);
	if (qp->umem_buf)
		free((void *)(uintptr_t)qp->umem_buf);
}

static int
doca_mlx5_devx_cmd_modify_qp_state(struct doca_mlx5_devx_obj *qp, uint32_t qp_st_mod_op,
				   uint32_t remote_qp_id)
{
	union {
		uint32_t rst2init[DOCA_MLX5_ST_SZ_DW(rst2init_qp_in)];
		uint32_t init2rtr[DOCA_MLX5_ST_SZ_DW(init2rtr_qp_in)];
		uint32_t rtr2rts[DOCA_MLX5_ST_SZ_DW(rtr2rts_qp_in)];
	} in;
	union {
		uint32_t rst2init[DOCA_MLX5_ST_SZ_DW(rst2init_qp_out)];
		uint32_t init2rtr[DOCA_MLX5_ST_SZ_DW(init2rtr_qp_out)];
		uint32_t rtr2rts[DOCA_MLX5_ST_SZ_DW(rtr2rts_qp_out)];
	} out;
	void *qpc;
	int ret;
	unsigned int inlen;
	unsigned int outlen;

	memset(&in, 0, sizeof(in));
	memset(&out, 0, sizeof(out));
	DOCA_MLX5_SET(rst2init_qp_in, &in, opcode, qp_st_mod_op);
	switch (qp_st_mod_op) {
	case DOCA_MLX5_CMD_OP_RST2INIT_QP:
		DOCA_MLX5_SET(rst2init_qp_in, &in, qpn, qp->id);
		qpc = DOCA_MLX5_ADDR_OF(rst2init_qp_in, &in, qpc);
		DOCA_MLX5_SET(qpc, qpc, primary_address_path.vhca_port_num, 1);
		DOCA_MLX5_SET(qpc, qpc, rre, 1);
		DOCA_MLX5_SET(qpc, qpc, rwe, 1);
		DOCA_MLX5_SET(qpc, qpc, pm_state, DOCA_MLX5_QP_PM_MIGRATED);
		inlen = sizeof(in.rst2init);
		outlen = sizeof(out.rst2init);
		break;
	case DOCA_MLX5_CMD_OP_INIT2RTR_QP:
		DOCA_MLX5_SET(init2rtr_qp_in, &in, qpn, qp->id);
		qpc = DOCA_MLX5_ADDR_OF(init2rtr_qp_in, &in, qpc);
		DOCA_MLX5_SET(qpc, qpc, primary_address_path.fl, 1);
		DOCA_MLX5_SET(qpc, qpc, primary_address_path.vhca_port_num, 1);
		DOCA_MLX5_SET(qpc, qpc, mtu, 1);
		DOCA_MLX5_SET(qpc, qpc, log_msg_max, 30);
		DOCA_MLX5_SET(qpc, qpc, remote_qpn, remote_qp_id);
		DOCA_MLX5_SET(qpc, qpc, min_rnr_nak, 0);
		inlen = sizeof(in.init2rtr);
		outlen = sizeof(out.init2rtr);
		break;
	case DOCA_MLX5_CMD_OP_RTR2RTS_QP:
		qpc = DOCA_MLX5_ADDR_OF(rtr2rts_qp_in, &in, qpc);
		DOCA_MLX5_SET(rtr2rts_qp_in, &in, qpn, qp->id);
		DOCA_MLX5_SET(qpc, qpc, primary_address_path.ack_timeout, 14);
		DOCA_MLX5_SET(qpc, qpc, log_ack_req_freq, 0);
		DOCA_MLX5_SET(qpc, qpc, retry_count, 7);
		DOCA_MLX5_SET(qpc, qpc, rnr_retry, 7);
		inlen = sizeof(in.rtr2rts);
		outlen = sizeof(out.rtr2rts);
		break;
	default:
		DOCA_DLOG_ERR("Invalid or unsupported QP modify op.");
		return -EINVAL;
	}
	ret = mlx5dv_devx_obj_modify(qp->obj, &in, inlen, &out, outlen);
	if (ret) {
		DOCA_DLOG_ERR("Failed to modify QP using DevX.");
		return -errno;
	}

	return ret;
}

int
doca_mlx5_devx_qp2rts(struct doca_mlx5_devx_qp *qp, uint32_t remote_qp_id)
{
	if (doca_mlx5_devx_cmd_modify_qp_state(qp->qp, DOCA_MLX5_CMD_OP_RST2INIT_QP,
					       remote_qp_id)) {
		DOCA_DLOG_ERR("Failed to modify QP to INIT state.");
		return -ENOTSUP;
	}
	if (doca_mlx5_devx_cmd_modify_qp_state(qp->qp, DOCA_MLX5_CMD_OP_INIT2RTR_QP,
					       remote_qp_id)) {
		DOCA_DLOG_ERR("Failed to modify QP to RTR state.");
		return -ENOTSUP;
	}
	if (doca_mlx5_devx_cmd_modify_qp_state(qp->qp, DOCA_MLX5_CMD_OP_RTR2RTS_QP, remote_qp_id)) {
		DOCA_DLOG_ERR("Failed to modify QP to RTS state.");
		return -ENOTSUP;
	}
	return 0;
}

