/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef LIB_DOCA_SHA_BF2_DOCA_MLX5_COMMON_H_
#define LIB_DOCA_SHA_BF2_DOCA_MLX5_COMMON_H_

#include <limits.h>
#include <sys/queue.h>
#include <unistd.h>
#include <endian.h>
#include <infiniband/mlx5dv.h>
#include <infiniband/verbs.h>

/* NOTE: hardcoded for now - 64 is fine for BF ARM and x86. */
#define CACHELINE_SIZE 64

#define DOCA_MLX5_LOG_PAGE_SIZE 12
#define DOCA_PTR_ADD(ptr, x) ((void *)((uintptr_t)(ptr) + (x)))

struct doca_mlx5_devx_obj {
	void *obj; /* The DV object. */
	int id;	   /* The object ID. */
};

struct doca_mlx5_devx_qp {
	struct doca_mlx5_devx_obj *qp; /* The QP DevX object. */
	void *umem_obj;		       /* The QP umem object. */
	union {
		void *umem_buf;
		struct doca_mlx5_wqe *wqes; /* The QP ring buffer. */
		struct doca_mlx5_aso_wqe *aso_wqes;
	};
	volatile uint32_t *db_rec; /* The QP doorbell record. */
};

struct doca_mlx5_devx_cq {
	struct doca_mlx5_devx_obj *cq; /* The CQ DevX object. */
	void *umem_obj;		       /* The CQ umem object. */
	union {
		volatile void *umem_buf;
		volatile struct doca_mlx5_cqe *cqes; /* The CQ ring buffer. */
	};
	volatile uint32_t *db_rec; /* The CQ doorbell record. */
};

/* UMR memory buffer used to define 1 entry in indirect mkey. */
struct doca_mlx5_klm {
	uint32_t byte_count;
	uint32_t mkey;
	uint64_t address;
};

static inline size_t
get_mem_page_size(void)
{
	static size_t page_size;

	if (!page_size)
		page_size = sysconf(_SC_PAGESIZE);

	return page_size;
}

static inline uint32_t
bf2_doca_mlx5_align32pow2(uint32_t x)
{
	x--;
	x |= x >> 1;
	x |= x >> 2;
	x |= x >> 4;
	x |= x >> 8;
	x |= x >> 16;

	return x + 1;
}

static inline uint32_t
bf2_doca_mlx5_bsf32(uint32_t v)
{
	return __builtin_ctz(v);
}

static inline uint32_t
bf2_doca_mlx5_bsf64(uint64_t v)
{
	return (uint32_t)__builtin_ctzll(v);
}

static inline uint32_t
bf2_doca_mlx5_log2_u32(uint32_t v)
{
	if (v == 0)
		return 0;

	v = bf2_doca_mlx5_align32pow2(v);

	return bf2_doca_mlx5_bsf32(v);
}

static inline int
bf2_doca_mlx5_get_pdn(void *pd, uint32_t *pdn)
{
	struct mlx5dv_obj obj;
	struct mlx5dv_pd pd_info;
	int ret = 0;

	obj.pd.in = pd;
	obj.pd.out = &pd_info;
	ret = mlx5dv_init_obj(&obj, MLX5DV_OBJ_PD);
	if (ret != 0)
		return ret;

	*pdn = pd_info.pdn;

	return 0;
}

static inline uint32_t
doca_mlx5_os_get_umem_id(void *umem)
{
	if (!umem)
		return 0;

	return ((struct mlx5dv_devx_umem *)umem)->umem_id;
}

static inline void
set_data_seg(struct mlx5_wqe_data_seg *seg, uint32_t length, uint32_t lkey, uintptr_t address)
{
	seg->byte_count = htobe32(length);
	seg->lkey = htobe32(lkey);
	seg->addr = htobe64(address);
}

struct doca_mlx5_wqe_metadata_seg {
	uint32_t mmo_control_31_0; /* mmo_control_63_32 is in ctrl_seg.imm */
	uint32_t lkey;
	uint64_t addr;
};

static inline void
set_metadata_seg(struct doca_mlx5_wqe_metadata_seg *seg, uint32_t mmo_control_31_0, uint32_t lkey,
		 uintptr_t address)
{
	seg->mmo_control_31_0 = htobe32(mmo_control_31_0);
	seg->lkey = htobe32(lkey);
	seg->addr = htobe64(address);
}

#define DOCA_MLX5_DBR_SIZE CACHELINE_SIZE
#define DOCA_MLX5_DBR_PER_PAGE 64
/* Must be >= CHAR_BIT * sizeof(uint64_t) */
#define DOCA_MLX5_DBR_PAGE_SIZE (DOCA_MLX5_DBR_PER_PAGE * DOCA_MLX5_DBR_SIZE)
/* Page size must be >= 512. */
#define DOCA_MLX5_DBR_BITMAP_SIZE (DOCA_MLX5_DBR_PER_PAGE / (CHAR_BIT * sizeof(uint64_t)))

struct doca_mlx5_devx_dbr_page {
	/* Door-bell records, must be first member in structure. */
	uint8_t dbrs[DOCA_MLX5_DBR_PAGE_SIZE];
	LIST_ENTRY(doca_mlx5_devx_dbr_page) next; /* Pointer to the next element. */
	void *umem;
	uint32_t dbr_count; /* Number of door-bell records in use. */
	/* 1 bit marks matching door-bell is in use. */
	uint64_t dbr_bitmap[DOCA_MLX5_DBR_BITMAP_SIZE];
};

LIST_HEAD(doca_mlx5_dbr_page_list, doca_mlx5_devx_dbr_page);

struct doca_mlx5_devx_mkey_attr {
	uint64_t addr;
	uint64_t size;
	uint32_t umem_id;
	uint32_t pd;
	uint32_t log_entity_size;
	uint32_t pg_access : 1;
	uint32_t relaxed_ordering_write : 1;
	uint32_t relaxed_ordering_read : 1;
	uint32_t umr_en : 1;
	struct doca_mlx5_klm *klm_array;
	int klm_num;
};

/* CQ attributes structure, used by CQ operations. */
struct doca_mlx5_devx_cq_attr {
	uint32_t q_umem_valid : 1;
	uint32_t db_umem_valid : 1;
	uint32_t use_first_only : 1;
	uint32_t overrun_ignore : 1;
	uint32_t cqe_comp_en : 1;
	uint32_t mini_cqe_res_format : 2;
	uint32_t mini_cqe_res_format_ext : 2;
	uint32_t log_cq_size : 5;
	uint32_t log_page_size : 5;
	uint32_t uar_page_id;
	uint32_t q_umem_id;
	uint64_t q_umem_offset;
	uint32_t db_umem_id;
	uint64_t db_umem_offset;
	uint32_t eqn;
	uint64_t db_addr;
};

struct doca_mlx5_devx_wq_attr {
	uint32_t wq_type : 4;
	uint32_t wq_signature : 1;
	uint32_t end_padding_mode : 2;
	uint32_t cd_slave : 1;
	uint32_t hds_skip_first_sge : 1;
	uint32_t log2_hds_buf_size : 3;
	uint32_t page_offset : 5;
	uint32_t lwm : 16;
	uint32_t pd : 24;
	uint32_t uar_page : 24;
	uint64_t dbr_addr;
	uint32_t hw_counter;
	uint32_t sw_counter;
	uint32_t log_wq_stride : 4;
	uint32_t log_wq_pg_sz : 5;
	uint32_t log_wq_sz : 5;
	uint32_t dbr_umem_valid : 1;
	uint32_t wq_umem_valid : 1;
	uint32_t log_hairpin_num_packets : 5;
	uint32_t log_hairpin_data_sz : 5;
	uint32_t single_wqe_log_num_of_strides : 4;
	uint32_t two_byte_shift_en : 1;
	uint32_t single_stride_log_num_of_bytes : 3;
	uint32_t dbr_umem_id;
	uint32_t wq_umem_id;
	uint64_t wq_umem_offset;
};

struct doca_mlx5_devx_qp_attr {
	uint32_t pd : 24;
	uint32_t uar_index : 24;
	uint32_t cqn : 24;
	uint32_t log_page_size : 5;
	uint32_t rq_size : 17; /* Must be power of 2. */
	uint32_t log_rq_stride : 3;
	uint32_t sq_size : 17; /* Must be power of 2. */
	uint32_t ts_format : 2;
	uint32_t dbr_umem_valid : 1;
	uint32_t dbr_umem_id;
	uint64_t dbr_address;
	uint32_t wq_umem_id;
	uint64_t wq_umem_offset;
	uint32_t user_index : 24;
	uint32_t mmo : 1;
};

struct doca_mlx5_devx_obj *
doca_mlx5_devx_cmd_mkey_create(void *ctx, const struct doca_mlx5_devx_mkey_attr *attr);
struct doca_mlx5_devx_dbr_page *doca_mlx5_alloc_dbr_page(void *ctx);
int64_t doca_mlx5_get_dbr(void *ctx, struct doca_mlx5_dbr_page_list *head,
			  struct doca_mlx5_devx_dbr_page **dbr_page);
int32_t doca_mlx5_release_dbr(struct doca_mlx5_dbr_page_list *head, uint32_t umem_id,
			      uint64_t offset);
struct doca_mlx5_devx_obj *doca_mlx5_devx_cmd_create_cq(void *ctx,
							const struct doca_mlx5_devx_cq_attr *attr);
void doca_mlx5_devx_cq_destroy(struct doca_mlx5_devx_cq *cq);
int doca_mlx5_devx_qp_create(void *ctx, struct doca_mlx5_devx_qp *qp_obj, uint16_t log_wqbb_n,
			     struct doca_mlx5_devx_qp_attr *attr);
void doca_mlx5_devx_qp_destroy(struct doca_mlx5_devx_qp *qp);
int doca_mlx5_devx_qp2rts(struct doca_mlx5_devx_qp *qp, uint32_t remote_qp_id);

#endif /* LIB_DOCA_SHA_BF2_DOCA_MLX5_COMMON_H_ */
