/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <doca_log.h>
#include <private/common_defines.h>
#include <priv_doca_ctx.h>
#include <priv_doca_dev.h>
#include <priv_doca_malloc.h>
#include <doca_types.h>
#include <pthread.h>

#include "doca_sha.h"
#include "doca_sha_workq_ctx.h"
#include "doca_sha_impl.h"

DOCA_LOG_REGISTER(DOCA_SHA);

#define SHA_MAX_WORKQ_NUM 16 /**< Place holder for the max number of workq for this device. */

/**
 * The doca sha context definition
 */
struct doca_sha {
	struct doca_ctx base;				/**< Opaque object. */
	struct doca_dev *dev;				/**< The attached device. */
	uint8_t nb_workq_attached;			/**< Number of datapath instantiated. */
	bool use_sw_fallback;				/**< Indicate whether use openssl software SHA. */
	pthread_mutex_t *lock;				/**< Mutex for muti-thread sync. */
};

static doca_error_t
validate_object_and_lock(struct doca_ctx *ctx, struct doca_sha **sha_ctx)
{
	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*sha_ctx = container_of(ctx, struct doca_sha, base);

	if (pthread_mutex_lock((*sha_ctx)->lock) != 0)
		return DOCA_ERROR_NO_LOCK;

	return DOCA_SUCCESS;
}

/**
 * Validate settings and start the sha context.
 *
 * @param [in] ctx
 * The sha context.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
sha_start(struct doca_ctx *ctx)
{
	struct doca_sha *sha_ctx = NULL;
	doca_error_t result;

	result = validate_object_and_lock(ctx, &sha_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	if (ctx->is_started)
		result = DOCA_ERROR_IN_USE;
	else if (sha_ctx->dev != NULL && sha_ctx->nb_workq_attached == 0)
		result = DOCA_SUCCESS;
	else
		result = DOCA_ERROR_INVALID_VALUE;

	pthread_mutex_unlock(sha_ctx->lock);
	return result;
}

/**
 * Stop the sha context.
 *
 * @note
 * After stopping the context can be reconfigured or destroyed.
 *
 * @param [in] ctx
 * The sha context.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
sha_stop(struct doca_ctx *ctx)
{
	struct doca_sha *sha_ctx = NULL;
	doca_error_t result;

	result = validate_object_and_lock(ctx, &sha_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	if (!ctx->is_started)
		result = DOCA_ERROR_BAD_STATE;
	else if (sha_ctx->nb_workq_attached == 0)
		result = DOCA_SUCCESS;
	else
		result = DOCA_ERROR_IN_USE;

	pthread_mutex_unlock(sha_ctx->lock);
	return result;
}

/**
 * Attach a device to the sha context.
 *
 * @note
 * The sha context only supports a single device being attached.
 * Previous device should be detached before attaching a new one.
 *
 * @param [in] ctx
 * The sha context.
 *
 * @param [in] dev
 * The device to attach to sha.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
sha_dev_add(struct doca_ctx *ctx, struct doca_dev *dev)
{
	doca_error_t ret = DOCA_SUCCESS;
	struct doca_sha *sha_ctx = NULL;

	if (dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = validate_object_and_lock(ctx, &sha_ctx);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (ctx->is_started) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_IN_USE;
	}

	/*
	 * Assume the hardware is always supported,
	 * so set use_sw_fallback to false.
	 */
	sha_ctx->use_sw_fallback = false;
	/*
	 * After find the doca_sha_device is no occupied
	 * we need continue to check whether this device has the required sha-engine ability
	 */
	ret = device_has_sha_support(dev);
	if (ret == DOCA_SUCCESS)
		sha_ctx->dev = dev;
	else {
#ifdef SUPPORT_OPENSSL_CRYPTO
		/*
		 * If openssl library is available, we use openssl SHA as fallback.
		 */
		sha_ctx->use_sw_fallback = true;
		DOCA_DLOG_WARN("The specified device has not SHA-engine enabled, so use openssl SHA instead!");
		sha_ctx->dev = dev;
		ret = DOCA_SUCCESS;
#else /* SUPPORT_OPENSSL_CRYPTO */
		/*
		 * If the specified device has no SHA-enigne enabled, and openssl library is unavailable,
		 * then, return unsupported.
		 */
		sha_ctx->dev = NULL;
		ret = DOCA_ERROR_NOT_SUPPORTED;
#endif /* SUPPORT_OPENSSL_CRYPTO */
	}

	pthread_mutex_unlock(sha_ctx->lock);
	return ret;
}

/**
 * Detach a device from the sha context.
 *
 * @param [in] ctx
 * The sha context.
 *
 * @param [in] dev
 * The device to detach from sha context.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
sha_dev_rm(struct doca_ctx *ctx, struct doca_dev *dev)
{
	struct doca_sha *sha_ctx = NULL;
	doca_error_t result;

	result = validate_object_and_lock(ctx, &sha_ctx);
	if (result != DOCA_SUCCESS)
		return result;

	if (ctx->is_started) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_BAD_STATE;
	}

	if ((sha_ctx->dev != NULL) && (sha_ctx->dev == dev)) {
		/* free the device attached */
		sha_ctx->dev = NULL;
		result = DOCA_SUCCESS;
	} else {
		/*
		 * No same device is found
		 */
		result = DOCA_ERROR_INVALID_VALUE;
	}

	pthread_mutex_unlock(sha_ctx->lock);
	return result;
}

/**
 * Retrieve a sha event (response)
 *
 * @param [in] data_priv
 * Opaque pointer to data created by the engine during sha_ctx_workq_add and emitted through its data_priv out param.
 *
 * @param [in] ev
 * Event structure to be populated as appropriate when this function returns a success code.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
sha_progress(union doca_data data_priv, struct doca_event *ev)
{
	struct sha_workq_impl *workq_impl = (struct sha_workq_impl *)(data_priv.ptr);
	/*
	 * Use the actual workq to do response retrieve
	 */
	return workq_impl->progress_retrieve_fn(workq_impl->workq_ctx, ev);
}

/**
 * Logic that is common for attaching a polling or event mode work queue context.
 *
 * @param [in] sha_ctx
 * The SHA context.
 *
 * @param [in] workq
 * The work queue to attach.
 *
 * @param [in] workq_impl
 * The work queue context base structure.
 *
 * @param [out] poller_out
 * The newly created progress entry, that was added to workq.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
workq_ctx_base_add(struct doca_sha *sha_ctx, struct doca_workq *workq, struct sha_workq_impl *workq_impl,
		   struct priv_doca_workq_progress_entry **poller_out)
{
	struct priv_doca_workq_progress_entry *poller = NULL;
	doca_error_t ret = DOCA_SUCCESS;

	poller = priv_doca_malloc(1, sizeof(*poller));
	if (poller == NULL)
		return DOCA_ERROR_NO_MEMORY;

	poller->progress_cb = sha_progress;
	poller->data_priv.ptr = (void *)(workq_impl);

	ret = priv_doca_workq_progress_entry_add(workq, poller);
	if (ret != DOCA_SUCCESS)
		goto err_clean_poller;

	*poller_out = poller;
	sha_ctx->nb_workq_attached += 1;

	return DOCA_SUCCESS;
err_clean_poller:
	priv_doca_free(poller);
	return ret;
}

/**
 * Logic that is common for detaching a polling or event mode work queue context.
 *
 * @param [in] sha_ctx
 * The SHA context.
 *
 * @param [in] workq
 * The work queue to attach.
 *
 * @param [out] poller
 * The progress entry that was previously added to workq in dma_workq_ctx_base_add().
 *
 * @param [out] workq_impl_out
 * The work queue context base structure.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
workq_ctx_base_rm(struct doca_sha *sha_ctx, struct doca_workq *workq,
		  struct priv_doca_workq_progress_entry *poller,
		  struct sha_workq_impl **workq_impl_out)
{
	struct sha_workq_impl *workq_impl = NULL;
	doca_error_t ret = DOCA_SUCCESS;

	ret = priv_doca_workq_progress_entry_rm(workq, poller);
	if (ret != DOCA_SUCCESS)
		return ret;

	workq_impl = (struct sha_workq_impl *)(poller->data_priv.ptr);
	sha_ctx->nb_workq_attached--;
	*workq_impl_out = workq_impl;

	priv_doca_free(poller);
	poller = NULL;
	return DOCA_SUCCESS;
}

/**
 * Initialise the sha workq.
 *
 * @param [in] sha_ctx
 * The doca_sha instance.
 *
 * @param [out] workq_ctx
 * The workq to be initialized.
 *
 * @param [in] dev
 * The device used by the workq.
 */
static doca_error_t
sha_polled_workq_ctx_create(struct doca_sha *sha_ctx, struct sha_polled_workq_ctx **workq_ctx, struct doca_dev *dev)
{
	int ret;
	struct sha_polled_workq_ctx *ctx;
	struct sha_workq_impl *workq_impl = NULL;

	if (workq_ctx == NULL || dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	ctx = priv_doca_malloc(1, sizeof(struct sha_polled_workq_ctx));
	if (ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	workq_impl = &(ctx->base);
	/*
	 * Accoridng to the value of use_sw_fallback, to select the suitable workq instance type.
	 */
	ret = sha_workq_factory_partial_init_workq(sha_ctx->use_sw_fallback, workq_impl);
	if (ret != DOCA_SUCCESS) {
		priv_doca_free(ctx);
		DOCA_DLOG_ERR("factory_partial_init_workq failure");
		return ret;
	}

	ret = workq_impl->workq_create_fn((void **)&(workq_impl->workq_ctx),
			(void **)&(workq_impl->cq), sha_ctx->dev);
	if (ret != DOCA_SUCCESS) {
		priv_doca_free(ctx);
		DOCA_DLOG_ERR("polled workq ctx base create failure");
		return ret;
	}
	workq_impl->sha_set_event_driven_mode_fn(workq_impl->workq_ctx, false);
	*workq_ctx = ctx;
	return ret;
}

/**
 * Destroy work queue context and all resources owned by it.
 *
 * @param [in] ctx
 * Pointer to workq to be destroyed.
 */
static void
sha_polled_workq_ctx_destroy(struct sha_polled_workq_ctx *ctx)
{
	struct sha_workq_impl *workq_impl = &(ctx->base);

	workq_impl->workq_destroy_fn(workq_impl->workq_ctx);
	priv_doca_free(ctx);
}

static doca_error_t
sha_polled_workq_add(struct doca_ctx *ctx, struct doca_workq *workq,
	      union doca_data *data_priv)
{
	struct doca_sha *sha_ctx = NULL;
	struct sha_polled_workq_ctx *workq_ctx = NULL;
	struct priv_doca_workq_progress_entry *poller = NULL;
	doca_error_t ret = DOCA_SUCCESS;

	ret = validate_object_and_lock(ctx, &sha_ctx);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (!ctx->is_started) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_BAD_STATE;
	}

	/* if no dev attached, return failure */
	if (sha_ctx->dev == NULL) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_NOT_PERMITTED;
	}
	/* if no space, return failure */
	if ((sha_ctx->nb_workq_attached + 1) > SHA_MAX_WORKQ_NUM) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_NOT_PERMITTED;
	}

	ret = sha_polled_workq_ctx_create(sha_ctx, &workq_ctx, sha_ctx->dev);
	if (ret != DOCA_SUCCESS) {
		pthread_mutex_unlock(sha_ctx->lock);
		return ret;
	}

	ret = workq_ctx_base_add(sha_ctx, workq, &(workq_ctx->base), &poller);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("polled work ctx base add failure.");
		sha_polled_workq_ctx_destroy(workq_ctx);
	}

	data_priv->ptr = (void *)poller;

	pthread_mutex_unlock(sha_ctx->lock);
	return DOCA_SUCCESS;
}

static doca_error_t
sha_polled_workq_rm(struct doca_ctx *ctx, struct doca_workq *workq,
		union doca_data data_priv)
{
	struct doca_sha *sha_ctx = NULL;
	struct sha_workq_impl *workq_impl = NULL;
	struct sha_polled_workq_ctx *workq_ctx = NULL;
	struct priv_doca_workq_progress_entry *poller;
	doca_error_t ret = DOCA_SUCCESS;

	if (data_priv.ptr == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = validate_object_and_lock(ctx, &sha_ctx);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (!ctx->is_started) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_BAD_STATE;
	}

	/* Grab objects from data_priv */
	poller = (struct priv_doca_workq_progress_entry *)data_priv.ptr;

	ret = workq_ctx_base_rm(sha_ctx, workq, poller, &workq_impl);
	if (ret != DOCA_SUCCESS) {
		pthread_mutex_unlock(sha_ctx->lock);
		DOCA_DLOG_ERR("polled workq ctx base rm failure");
		return ret;
	}

	workq_ctx = container_of(workq_impl, struct sha_polled_workq_ctx, base);
	sha_polled_workq_ctx_destroy(workq_ctx);

	pthread_mutex_unlock(sha_ctx->lock);
	return DOCA_SUCCESS;
}

/**
 * Arm work queue context for an event. Requesting an event on next CQE received.
 *
 * @details Callback to be used once application requests an event from the work queue while working in event driven
 * mode. Callback will be trigerred once application calls doca_workq_event_handle_arm().
 *
 * @param [in] event_data
 * Opaque pointer to a priv_doca_comp_events_channel belonging to an event workq context. The channel can be used for
 * arming of events. The pointer was previously provided upon registering a doca_event_handle_t to the work queue
 * using priv_doca_workq_event_handle_register().
 */
static doca_error_t
sha_ctx_arm_event(union doca_data event_data)
{
	doca_error_t ret = DOCA_SUCCESS;
	struct sha_event_driven_workq_ctx *workq_ctx = (struct sha_event_driven_workq_ctx *)event_data.ptr;

	priv_doca_comp_events_channel_arm(workq_ctx->channel);
	/*
	 * Only suspend hardware engine state based on whether there is cqe.
	 * Because openssl software fallback does not have cq, so do not suspends it in fallback mode.
	 */
	if ((workq_ctx->base.cq != NULL) && priv_doca_cq_empty(workq_ctx->base.cq))
		/* All CQEs that generated an event, have been drained. */
		ret = priv_doca_workq_progress_entry_state_update(workq_ctx->workq, workq_ctx->poller,
								  PRIV_DOCA_WORKQ_PROGRESS_SUSPENDED);

	return ret;
}

/**
 * Clear events retrieved for the work queue context
 *
 * @details Callback to be used once application requests clearing of events received on the work queue,
 * using doca_workq_event_handle_clear().
 * The method's goal is to change state of the event handle from 'ready to read' to 'pending'.
 *
 * @param [in] event_data
 * Opaque pointer to a priv_doca_comp_events_channel belonging to an event workq context. The channel can be used for
 * clearing of events. The pointer was previously provided upon registering a doca_event_handle_t to the work queue
 * using priv_doca_workq_event_handle_register().
 */
static doca_error_t
sha_ctx_clear_event(union doca_data event_data)
{
	doca_error_t ret;
	struct sha_event_driven_workq_ctx *workq_ctx = (struct sha_event_driven_workq_ctx *)event_data.ptr;

	ret = priv_doca_workq_progress_entry_state_update(workq_ctx->workq, workq_ctx->poller,
							  PRIV_DOCA_WORKQ_PROGRESS_RUNNING);
	if (unlikely(ret != DOCA_SUCCESS))
		return ret;

	return priv_doca_comp_events_channel_handle_events(workq_ctx->channel, 0 /*ignored*/);
}

/**
 * Initialise the sha workq.
 *
 * @param [in] sha_ctx
 * The doca_sha instance.
 *
 * @param [out] workq_ctx
 * The workq to be initialized.
 *
 * @param [in] dev
 * The device used by the workq.
 */
static doca_error_t
sha_event_driven_workq_ctx_create(struct doca_sha *sha_ctx,
		struct sha_event_driven_workq_ctx **workq_ctx, struct doca_dev *dev)
{
	doca_error_t ret = DOCA_SUCCESS;
	struct sha_event_driven_workq_ctx *ctx;
	struct sha_workq_impl *workq_impl;

	if (workq_ctx == NULL || dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ctx = priv_doca_malloc(1, sizeof(struct sha_event_driven_workq_ctx));
	if (ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	ret = priv_doca_dev_comp_events_channel_create(dev, true, &(ctx->channel));
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("SHA work queue context unable to create completion events channel.");
		goto err_event_channle_create;
	}

	ret = priv_doca_comp_events_channel_handle_get(ctx->channel, &(ctx->channel_handle));
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("SHA work queue context unable to get completion events channel handle.");
		goto err_event_channel_handle_get;
	}

	workq_impl = &(ctx->base);
	/*
	 * Accoridng to the value of use_sw_fallback, to select the suitable workq instance type.
	 */
	ret = sha_workq_factory_partial_init_workq(sha_ctx->use_sw_fallback, workq_impl);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("event driven workq ctx workq factory partial init failure");
		goto err_workq_factory_partial_init;
	}

	ret = workq_impl->workq_create_fn((void **)&(workq_impl->workq_ctx),
			(void **)&(workq_impl->cq), sha_ctx->dev);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("event driven workq ctx base create failure");
		goto err_workq_ctx_base_create;
	}

	workq_impl->sha_set_event_driven_mode_fn(workq_impl->workq_ctx, true);

	/*
	 * Register event channel only in hardware mode.
	 * In software fallback mode, it is not required.
	 */
	if (sha_ctx->use_sw_fallback == false) {
		if (workq_impl->cq != NULL) {
			ret = priv_doca_comp_events_channel_register(ctx->channel, workq_impl->cq);
			if (ret != DOCA_SUCCESS) {
				DOCA_LOG_ERR("SHA work queue context unable to register CQ with completion events channel.");
				goto err_event_channel_register;
			}
		} else {
			DOCA_LOG_ERR("Unexpected situation occurs.");
			ret = DOCA_ERROR_UNEXPECTED;
			goto err_event_channel_register;
		}
	}

	*workq_ctx = ctx;
	return DOCA_SUCCESS;

err_event_channel_register:
	workq_impl->workq_destroy_fn(workq_impl->workq_ctx);
err_workq_ctx_base_create:
err_workq_factory_partial_init:
err_event_channel_handle_get:
	priv_doca_dev_comp_events_channel_destroy(ctx->channel);
err_event_channle_create:
	priv_doca_free(ctx);
	return ret;
}

/**
 * Destroy work queue context and all resources owned by it.
 *
 * @param [in] ctx
 * Pointer to workq to be destroyed.
 */
static void
sha_event_driven_workq_ctx_destroy(struct sha_event_driven_workq_ctx *ctx)
{
	doca_error_t ret;
	struct sha_workq_impl *workq_impl = &(ctx->base);

	if ((workq_impl->cq != NULL) && (ctx->channel != NULL)) {
		ret = priv_doca_comp_events_channel_unregister(ctx->channel, workq_impl->cq);
		/* In case create did not reach registration phase, not found may be returned.*/
		if (ret != DOCA_SUCCESS && ret != DOCA_ERROR_NOT_FOUND)
			DOCA_LOG_ERR("Failed to unregister SHA workq CQ from completion events channel. err = %s",
				     doca_get_error_name(ret));
	}

	workq_impl->workq_destroy_fn(workq_impl->workq_ctx);

	if (ctx->channel != NULL)
		priv_doca_dev_comp_events_channel_destroy(ctx->channel);

	priv_doca_free(ctx);
}

static doca_error_t
sha_event_driven_workq_add(struct doca_ctx *ctx, struct doca_workq *workq,
	      union doca_data *data_priv)
{
	struct doca_sha *sha_ctx = NULL;
	struct sha_event_driven_workq_ctx *workq_ctx = NULL;
	struct priv_doca_workq_progress_entry *poller = NULL;
	union doca_data event_data;
	doca_error_t ret = DOCA_SUCCESS;

	ret = validate_object_and_lock(ctx, &sha_ctx);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (!ctx->is_started) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_BAD_STATE;
	}

	/* if no dev attached, return failure */
	if (sha_ctx->dev == NULL) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_NOT_PERMITTED;
	}
	/* if no space, return failure */
	if ((sha_ctx->nb_workq_attached + 1) > SHA_MAX_WORKQ_NUM) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_NOT_PERMITTED;
	}

	ret = sha_event_driven_workq_ctx_create(sha_ctx, &workq_ctx, sha_ctx->dev);
	if (ret != DOCA_SUCCESS) {
		pthread_mutex_unlock(sha_ctx->lock);
		return ret;
	}

	static const struct priv_doca_event_handle_ops event_ops = {
		.arm_cb = sha_ctx_arm_event,
		.clear_cb = sha_ctx_clear_event,
	};

	event_data.ptr = workq_ctx;
	ret = priv_doca_workq_event_handle_register(workq, workq_ctx->channel_handle, &event_ops, event_data);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to register SHA event handle to work queue.");
		goto err_event_handle_register;
	}

	ret = workq_ctx_base_add(sha_ctx, workq, &(workq_ctx->base), &poller);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("event_driven work ctx base add failure.");
		goto err_event_driven_ctx_base_add;
	}

	data_priv->ptr = (void *)poller;
	/* Store workq & poller to be able to change state according to received event. */
	workq_ctx->poller = poller;
	workq_ctx->workq = workq;

	pthread_mutex_unlock(sha_ctx->lock);
	return DOCA_SUCCESS;
err_event_driven_ctx_base_add:
	if (priv_doca_workq_event_handle_unregister(workq, workq_ctx->channel_handle) != DOCA_SUCCESS)
		DOCA_DLOG_ERR("Unable to unregister DMA workq ctx event handle from workq.");
err_event_handle_register:
	sha_event_driven_workq_ctx_destroy(workq_ctx);
	pthread_mutex_unlock(sha_ctx->lock);
	return ret;
}

static doca_error_t
sha_event_driven_workq_rm(struct doca_ctx *ctx, struct doca_workq *workq,
		union doca_data data_priv)
{
	struct doca_sha *sha_ctx = NULL;
	struct sha_workq_impl *workq_impl = NULL;
	struct sha_event_driven_workq_ctx *workq_ctx = NULL;
	struct priv_doca_workq_progress_entry *poller;
	doca_error_t ret = DOCA_SUCCESS;

	if (data_priv.ptr == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = validate_object_and_lock(ctx, &sha_ctx);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (!ctx->is_started) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_BAD_STATE;
	}

	poller = (struct priv_doca_workq_progress_entry *)data_priv.ptr;

	ret = workq_ctx_base_rm(sha_ctx, workq, poller, &workq_impl);
	if (ret != DOCA_SUCCESS) {
		pthread_mutex_unlock(sha_ctx->lock);
		DOCA_DLOG_ERR("polled workq ctx base rm failure");
		return ret;
	}

	workq_ctx = container_of(workq_impl, struct sha_event_driven_workq_ctx, base);
	ret = priv_doca_workq_event_handle_unregister(workq, workq_ctx->channel_handle);
	if (ret != DOCA_SUCCESS) {
		pthread_mutex_unlock(sha_ctx->lock);
		DOCA_DLOG_ERR("Unable to unregister DMA workq ctx event handle from workq.");
		return ret;
	}
	sha_event_driven_workq_ctx_destroy(workq_ctx);

	pthread_mutex_unlock(sha_ctx->lock);
	return DOCA_SUCCESS;
}

/**
 * Submit a sha job
 *
 * @param [in] data_priv
 * Opaque pointer to data created by the engine during sha_ctx_workq_add and emitted through its data_priv out param.
 *
 * @param [in] job
 * Job as passed to: doca_workq_submit. This should have the doca_job::type set to an appropriate value to suggest this
 * is a sha job.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
static doca_error_t
sha_job_submit(union doca_data data_priv, struct doca_job const *job)
{
	struct sha_workq_impl *workq_impl = NULL;
	struct priv_doca_workq_progress_entry *poller = NULL;
	struct doca_sha_partial_job const *sha_job = NULL;

	poller = (struct priv_doca_workq_progress_entry *)(data_priv.ptr);
	workq_impl = (struct sha_workq_impl *)(poller->data_priv.ptr);

	/*
	 * When the user submits a job, the sha algorithm for this job is determined,
	 * so we can use job->type to decide the actual sha engine.
	 * Then, we use the actual workq to do job submit
	 */
	switch (job->type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA512:
	case DOCA_SHA_JOB_SHA1_PARTIAL:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		sha_job = (struct doca_sha_partial_job const *)
			  container_of(job, struct doca_sha_job const, base);
		return workq_impl->job_submit_fn(workq_impl->workq_ctx, sha_job);
	default:
		DOCA_DLOG_ERR("Unsupported sha job type [%d] is found", job->type);
		return DOCA_ERROR_INVALID_VALUE;
	}
}

/**
 * Initialise the generic context vtable with appropriate sha function calls.
 */
static const struct priv_doca_ctx_ops sha_ops = {
	.start = sha_start,
	.stop = sha_stop,
	.dev_add = sha_dev_add,
	.dev_rm = sha_dev_rm,
	.workq_add = sha_polled_workq_add,
	.event_workq_add = sha_event_driven_workq_add,
	.workq_rm = sha_polled_workq_rm,
	.event_workq_rm = sha_event_driven_workq_rm,
	.job_submit = sha_job_submit
};

doca_error_t
doca_sha_create(struct doca_sha **sha_ctx)
{
	struct doca_sha *ctx = NULL;
	doca_error_t ret = DOCA_SUCCESS;

	if (sha_ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ctx = priv_doca_malloc(1, sizeof(struct doca_sha));
	if (ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	memset(ctx, 0, sizeof(struct doca_sha));

	ctx->lock = priv_doca_malloc(1, sizeof(pthread_mutex_t));
	if (ctx->lock == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_mutex_create;
	}

	if (pthread_mutex_init(ctx->lock, NULL) != 0) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_mutex_init;
	}

	ret = priv_doca_ctx_init(&(ctx->base), true);
	if (ret != DOCA_SUCCESS)
		goto err_priv_doca_ctx_init;

	ctx->base.ops = sha_ops;

	*sha_ctx = ctx;

	return DOCA_SUCCESS;

err_priv_doca_ctx_init:
	pthread_mutex_destroy(ctx->lock);
err_mutex_init:
	priv_doca_free(ctx->lock);
err_mutex_create:
	priv_doca_free(ctx);
	return ret;
}

doca_error_t
doca_sha_destroy(struct doca_sha *sha_ctx)
{
	pthread_mutex_t *lock;

	if (sha_ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (pthread_mutex_trylock(sha_ctx->lock) != 0)
		return DOCA_ERROR_IN_USE;

	if (sha_ctx->nb_workq_attached > 0) {
		pthread_mutex_unlock(sha_ctx->lock);
		return DOCA_ERROR_IN_USE;
	}

	lock = sha_ctx->lock;
	sha_ctx->lock = NULL;

	sha_ctx->nb_workq_attached = 0;

	/* free the device attached */
	sha_ctx->dev = NULL;

	priv_doca_ctx_destroy(&(sha_ctx->base));

	priv_doca_free(sha_ctx);

	pthread_mutex_unlock(lock);
	pthread_mutex_destroy(lock);
	priv_doca_free(lock);

	return DOCA_SUCCESS;
}

doca_error_t
doca_sha_job_get_supported(const struct doca_devinfo *devinfo, enum doca_sha_job_type job_type)
{
	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	switch (job_type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA512:
	case DOCA_SHA_JOB_SHA1_PARTIAL:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
#ifdef SUPPORT_OPENSSL_CRYPTO
		/* If openssl is found, do not need to check hardware capability anymore. */
		return DOCA_SUCCESS;
#else /* SUPPORT_OPENSSL_CRYPTO */
		return devinfo_has_sha_support(devinfo);
#endif /* SUPPORT_OPENSSL_CRYPTO */
	default:
		DOCA_DLOG_ERR("Unknown doca_sha job type: %d is not supported", job_type);
		return DOCA_ERROR_INVALID_VALUE;
	}
}

doca_error_t
doca_sha_get_max_list_buf_num_elem(const struct doca_devinfo *devinfo, uint32_t *max_list_num_elem)
{
	doca_error_t ret;

	if ((devinfo == NULL) || (max_list_num_elem == NULL))
		return DOCA_ERROR_INVALID_VALUE;

	/*
	 * First check whether the devinfo supports sha.
	 */
	ret = doca_sha_job_get_supported(devinfo, DOCA_SHA_JOB_SHA1);
	if (ret != DOCA_SUCCESS)
		return ret;

	/*
	 * The doca_sha does internal-split, so in theory, it can use any list element number.
	 * Because the maximum job size is DOCA_SHA_MAX_JOB_BYTE_COUNT, and each doca_buf contains at least 1byte,
	 * so use the value.
	 */
	*max_list_num_elem = DOCA_SHA_MAX_JOB_BYTE_COUNT;
	return DOCA_SUCCESS;
}

doca_error_t
doca_sha_get_max_src_buffer_size(const struct doca_devinfo *devinfo, uint64_t *max_buffer_size)
{
	doca_error_t ret;

	if (devinfo == NULL || max_buffer_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	/*
	 * First check whether the devinfo supports sha.
	 */
	ret = doca_sha_job_get_supported(devinfo, DOCA_SHA_JOB_SHA1);
	if (ret != DOCA_SUCCESS)
		return ret;

	/*
	 * The doca_sha does internal-split, its maximum job size for all job types is the same.
	 */
	*max_buffer_size = DOCA_SHA_MAX_JOB_BYTE_COUNT;
	return DOCA_SUCCESS;
}

doca_error_t doca_sha_get_min_dst_buffer_size(const struct doca_devinfo *devinfo,
					       enum doca_sha_job_type job_type, uint32_t *min_buffer_size)
{
	doca_error_t ret;

	if (devinfo == NULL || min_buffer_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	/*
	 * First check whether the devinfo supports sha.
	 */
	ret = doca_sha_job_get_supported(devinfo, job_type);
	if (ret != DOCA_SUCCESS)
		return ret;

	/*
	 * Then output the size.
	 */
	switch (job_type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA1_PARTIAL:
		*min_buffer_size = DOCA_SHA1_BYTE_COUNT;
		break;
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
		*min_buffer_size = DOCA_SHA256_BYTE_COUNT;
		break;
	case DOCA_SHA_JOB_SHA512:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		*min_buffer_size = DOCA_SHA512_BYTE_COUNT;
		break;
	default:
		DOCA_DLOG_ERR("Unexpected error");
		return DOCA_ERROR_NOT_SUPPORTED;
	}
	return DOCA_SUCCESS;
}

doca_error_t doca_sha_get_hardware_supported(const struct doca_devinfo *devinfo)
{
	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	return devinfo_has_sha_support(devinfo);
}

struct doca_ctx *
doca_sha_as_ctx(struct doca_sha *sha_ctx)
{
	if (sha_ctx == NULL)
		return NULL;
	return &(sha_ctx->base);
}

/**
 * A helper to get the priv workq data.
 *
 * The pre-allocated sessions are bind to specific workq: a software thread.
 * Here, in order to get/put a session object belonging to a specific workq,
 * so need to get the workq_data first.
 */
static inline doca_error_t
get_sha_workq_data(struct doca_sha *ctx, struct doca_workq *workq,
		struct sha_workq_impl **workq_impl)
{
	union doca_data data_priv;
	struct priv_doca_workq_progress_entry *poller = NULL;
	doca_error_t ret;

	if ((ctx == NULL) || (workq == NULL) || (workq_impl == NULL))
		return DOCA_ERROR_INVALID_VALUE;

	ret = priv_doca_workq_data_priv_get(workq, doca_sha_as_ctx(ctx)->id, &data_priv);
	if (ret == DOCA_SUCCESS) {
		poller = (struct priv_doca_workq_progress_entry *)(data_priv.ptr);
		*workq_impl = (struct sha_workq_impl *)(poller->data_priv.ptr);
	}
	return ret;
}

/**
 * Get a new session object.
 */
doca_error_t
doca_sha_partial_session_create(struct doca_sha *ctx, struct doca_workq *workq,
				struct doca_sha_partial_session **session)
{
	doca_error_t ret;
	struct sha_workq_impl *workq_impl = NULL;

	ret = get_sha_workq_data(ctx, workq, &workq_impl);
	if (ret == DOCA_SUCCESS)
		ret = workq_impl->sha_partial_session_create_fn(workq_impl->workq_ctx, session);
	return ret;
}

/**
 * Put back the session object acquired before.
 */
doca_error_t
doca_sha_partial_session_destroy(struct doca_sha *ctx, struct doca_workq *workq,
				 struct doca_sha_partial_session *session)
{
	doca_error_t ret;
	struct sha_workq_impl *workq_impl = NULL;

	ret = get_sha_workq_data(ctx, workq, &workq_impl);
	if (ret == DOCA_SUCCESS)
		ret = workq_impl->sha_partial_session_destroy_fn(workq_impl->workq_ctx, session);
	return ret;
}

doca_error_t
doca_sha_partial_session_copy(struct doca_sha *ctx,
			      struct doca_workq *workq,
			      const struct doca_sha_partial_session *from,
			      struct doca_sha_partial_session *to)
{
	doca_error_t ret;
	struct sha_workq_impl *workq_impl = NULL;

	ret = get_sha_workq_data(ctx, workq, &workq_impl);
	if (ret == DOCA_SUCCESS)
		ret = workq_impl->sha_partial_session_copy_fn(workq_impl->workq_ctx, from, to);
	return ret;
}

