/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <endian.h>
#include <doca_rdma.h>
#include <doca_log.h>
#include <priv_doca_malloc.h>
#include <priv_doca_buf.h>
#include "doca_sha_impl.h"
#include "doca_sha_bf2_impl.h"
#include "doca_mlx5_common.h"
#include "doca_mlx5_prm.h"
#include "doca_sha_workq_ctx.h"
#include <priv_doca_dev.h>
#include <priv_doca_os_abstraction_memory.h>

DOCA_LOG_REGISTER(BF2_IMPL);

///////////////////////////////////
/* Global config setting section */
//////////////////////////////////
//#define USE_SHA_DEBUG_PRINT
//#define DISABLE_UMR

////////////////////////////////////////////////////////////
/* macro setting section, mainly used for qp, wqe setting */
////////////////////////////////////////////////////////////
/**
 * Macro to align a value to a given power-of-two. The resultant value
 * will be of the same type as the first parameter, and will be no
 * bigger than the first parameter. Second parameter must be a
 * power-of-two value.
 */
#define SHA_UMR_ALIGN_FLOOR(val, align) \
	((typeof(val))((val) & (~((typeof(val))((align) - 1)))))

/**
 * Macro to align a value to a given power-of-two. The resultant value
 * will be of the same type as the first parameter, and will be no lower
 * than the first parameter. Second parameter must be a power-of-two
 * value.
 */
#define SHA_UMR_ALIGN_CEIL(val, align) \
	SHA_UMR_ALIGN_FLOOR(((val) + ((typeof(val)) (align) - 1)), align)
/**
 * Macro to align a value to a given power-of-two. The resultant
 * value will be of the same type as the first parameter, and
 * will be no lower than the first parameter. Second parameter
 * must be a power-of-two value.
 * This function is the same as SHA_UMR_ALIGN_CEIL
 */
#define SHA_UMR_ALIGN(val, align) SHA_UMR_ALIGN_CEIL(val, align)

#define ALIGN_4K 4096
#define ALIGN_64 64

/* define the max entries per QP - for now until we have a better way - should be a multiple of 2 */
#define MAX_PER_QP_NUM_DESCS (PER_Q_TEMP_JOBS_SIZE)

#define DOCA_MLX5_MMO_NUM_WQE_PER_PAGE (4096 / 64)
/* UMR mode requires 4 times the descriptors so convert to log value + 2. */
#define DOCA_MLX5_MMO_WQE_LOG_NUM(has_umr, log_desc) ((has_umr) ? ((log_desc) + 2) : (log_desc))
/* Access flags from the verb.h. */
#define DOCA_UMEM_ACCESS_FLAGS (IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ)

#define DOCA_MLX5_BIT32(nr)                (UINT32_C(1) << (nr))
#define DOCA_MLX5_MMO_SHA_METADATA_SIZE    (UINT32_C(256))
#define DOCA_MLX5_MMO_SHA_MAX_OUTPUT       (DOCA_MLX5_BIT32(6))
#define DOCA_MLX5_MMO_WQE_METADATA_OFFSET  16
#define DOCA_MLX5_MMO_WQE_GATHER_OFFSET    32
#define DOCA_MLX5_MMO_WQE_SCATTER_OFFSET   48
#define DOCA_MLX5_MMO_SHA_METADATA_GATHERED_LENGTH_OFFSET  12
#define DOCA_MLX5_MMO_SHA_METADATA_PARTIAL_SHA_OFFSET      64
/* The maximum KLMs can be added to one UMR indirect mkey. */
#define DOCA_MLX5_MMO_MAX_KLM_NUM          (MAX_LL_DOCA_BUF_NUM)
/* The KLM array size for one job. */
#define DOCA_MLX5_MMO_KLMS_SIZE            ((DOCA_MLX5_MMO_MAX_KLM_NUM) * sizeof(struct doca_mlx5_klm))
#define DOCA_MLX5_MMO_UMR_WQE_SIZE         192
#define DOCA_MLX5_OPCODE_MMO               0x2f
#define DOCA_MLX5_OPC_MOD_MMO_SHA          0x0

#define DOCA_MLX5_MMO_MAX_WQE_INDEX        0xffff
#define DOCA_MLX5_MMO_UMR_QP_PI_IDX(pi, ops, has_umr) \
	(((pi) + (ops)) & (DOCA_MLX5_MMO_MAX_WQE_INDEX >> ((has_umr) ? 2 : 0)))

#define BF2_CQE_OWNER(op_own)  ((op_own)&MLX5_CQE_OWNER_MASK)
#define BF2_CQE_OPCODE(op_own) (((op_own)&0xf0) >> 4)

enum bf2_cqe_status {
	BF2_CQE_STATUS_SW_OWN = -1,
	BF2_CQE_STATUS_HW_OWN = -2,
	BF2_CQE_STATUS_ERR = -3,
};

enum bf2_sha_func_type {
	BF2_SHA1 = 0,
	BF2_SHA256 = 2,
	BF2_SHA512 = 4,
};

#ifdef USE_SHA_DEBUG_PRINT
uint64_t global_enqueue_count;
uint64_t global_dequeue_count;
uint64_t global_enqueue_count_prev;
#endif /* USE_SHA_DEBUG_PRINT */

#ifdef USE_SHA_DEBUG_PRINT
static void
print_value_as_bits(uint32_t value)
{
	char arr[32] = {0};

	for (uint64_t i = value, j = 0; i > 0; i = i / 2)
		arr[j++] = (i % 2) ? 1 : 0;
	for (int i = 0; i < 8; i++) {
		for (int j = 0; j < 4; j++)
			DOCA_DLOG_DBG("%d", arr[31 - (i * 4 + j)]);
		DOCA_DLOG_DBG("  ");
	}
	DOCA_DLOG_DBG("\n");
}

static void
debug_print_wqe(struct doca_mlx5_wqe *wqe, bool has_umr)
{
	int j = 0;
	struct doca_mlx5_wqe_dseg *metaseg = (struct doca_mlx5_wqe_dseg *)&(wqe->eseg);

	DOCA_DLOG_DBG("wqe address is: %p\n", (void *)wqe);
	DOCA_DLOG_DBG("wqe control seg:\n");
	DOCA_DLOG_DBG("-- opcode       = 0x%08x\n", wqe->cseg.opcode);
	DOCA_DLOG_DBG("-- sq_ds        = 0x%08x\n", wqe->cseg.sq_ds);
	DOCA_DLOG_DBG("-- flags        = 0x%08x\n", wqe->cseg.flags);
	DOCA_DLOG_DBG("-- misc         = 0x%08x\n", wqe->cseg.misc);

	DOCA_DLOG_DBG("-- metadata dseg:\n");
	DOCA_DLOG_DBG("   -- bcout = 0x%08x\n", metaseg->bcount);
	DOCA_DLOG_DBG("   -- lkey  = 0x%08x\n", metaseg->lkey);
	DOCA_DLOG_DBG("   -- pbuf  = 0x%016lx\n", metaseg->pbuf);
	DOCA_DLOG_DBG("   -- htobe64(pbuf)  = 0x%016lx\n", htobe64(metaseg->pbuf));
	DOCA_DLOG_DBG("   -- the pbuf content length is: 0x%08x\n", htobe32(metaseg->bcount));
	DOCA_DLOG_DBG("   -- the pbuf content hex format is: ");
	for (j = 0; (j < DOCA_SHA_OUTPUT_BYTE_COUNT * 2) && (j < htobe32(metaseg->bcount)); j++) {
		char ch = ((char *)(htobe64(metaseg->pbuf)))[j];

		DOCA_DLOG_DBG("%02x ", ch & 0xff);
	}
	DOCA_DLOG_DBG("\n");
	DOCA_DLOG_DBG("-- gather dseg:\n");
	DOCA_DLOG_DBG("   -- bcout = 0x%08x\n", wqe->dseg[0].bcount);
	DOCA_DLOG_DBG("   -- lkey  = 0x%08x\n", wqe->dseg[0].lkey);
	DOCA_DLOG_DBG("   -- pbuf  = 0x%016lx\n", wqe->dseg[0].pbuf);
	DOCA_DLOG_DBG("   -- htobe64(pbuf)  = 0x%016lx\n", htobe64(wqe->dseg[0].pbuf));
	if (!has_umr) {
		DOCA_DLOG_DBG("   -- the pbuf content length is: 0x%08x\n", htobe32(wqe->dseg[0].bcount));
		DOCA_DLOG_DBG("   -- the pbuf content hex format is: ");
		for (j = 0; j < htobe32(wqe->dseg[0].bcount); j++)
			DOCA_DLOG_DBG("%02x ", ((char *)(htobe64(wqe->dseg[0].pbuf)))[j] & 0xff);
		DOCA_DLOG_DBG("\n");
	} else {
		DOCA_DLOG_DBG("   -- indirect mkey is used, so no data printed,");
		DOCA_DLOG_DBG(" but the above byte-count == total-byte-count in its corresponding umr-wqe\n");
	}
	DOCA_DLOG_DBG("-- scatter dseg:\n");
	DOCA_DLOG_DBG("   -- bcout = 0x%08x\n", wqe->dseg[1].bcount);
	DOCA_DLOG_DBG("   -- lkey  = 0x%08x\n", wqe->dseg[1].lkey);
	DOCA_DLOG_DBG("   -- pbuf  = 0x%016lx\n", wqe->dseg[1].pbuf);
	DOCA_DLOG_DBG("   -- htobe64(pbuf)  = 0x%016lx\n", htobe64(wqe->dseg[1].pbuf));
	DOCA_DLOG_DBG("   -- the pbuf content length is: 0x%08x\n", htobe32(wqe->dseg[1].bcount));
	DOCA_DLOG_DBG("   -- the pbuf content hex format is: ");
	for (j = 0; (j < DOCA_SHA_OUTPUT_BYTE_COUNT) && (j < htobe32(wqe->dseg[1].bcount)); j++) {
		char ch = ((char *)(htobe64(wqe->dseg[1].pbuf)))[j];

		DOCA_DLOG_DBG("%02x ", ch & 0xff);
	}
	DOCA_DLOG_DBG("\n");
	DOCA_DLOG_DBG("The sha-engine ctrl segment bits are: ");
	print_value_as_bits(htobe32(wqe->cseg.misc));
}
#endif /* USE_SHA_DEBUG_PRINT */

////////////////////////////////////////////////////////////
/**
 * Global device init seting section
 * only contain non-perf-critical control functions
 * for a device start/close, only for 1 time enough
 */
////////////////////////////////////////////////////////////
static doca_error_t
bf2_sha_hca_caps_get(const struct doca_devinfo *info, struct bf2_sha_hca_caps *caps)
{
	uint32_t out[DOCA_MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	const void *atts = DOCA_MLX5_ADDR_OF(query_hca_cap_out, out, capability);
	int ret;

	ret = priv_doca_devinfo_hca_caps_by_opmod_query(info,
							DOCA_MLX5_GET_HCA_CAP_OP_MOD_GENERAL_DEVICE | DOCA_MLX5_HCA_CAP_OPMOD_GET_CUR,
							out, sizeof(out));
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("HCA query failed.");
		return DOCA_ERROR_INITIALIZATION;
	}

	/* Pull out sha specific info from return. */
	caps->sha_mmo_qp = DOCA_MLX5_GET(cmd_hca_cap, atts, sha_mmo_qp);
	caps->log_sha_mmo_max_size = DOCA_MLX5_GET(cmd_hca_cap, atts, log_sha_mmo_max_size);
	caps->umr_indirect_mkey_disabled = DOCA_MLX5_GET(cmd_hca_cap, atts, umr_indirect_mkey_disabled);
	caps->umr_modify_entity_size_disabled = DOCA_MLX5_GET(cmd_hca_cap, atts, umr_modify_entity_size_disabled);

	/* Confirm RDMA over ethernet. */
	if (DOCA_MLX5_GET(cmd_hca_cap, atts, roce)) {
		memset(out, 0, sizeof(out));

		ret = priv_doca_devinfo_hca_caps_by_opmod_query(info,
								DOCA_MLX5_GET_HCA_CAP_OP_MOD_ROCE | DOCA_MLX5_HCA_CAP_OPMOD_GET_CUR,
								out, sizeof(out));
		if (ret != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("HCA ROCE query failed.");
			return DOCA_ERROR_INITIALIZATION;
		}

		caps->qp_ts_format = DOCA_MLX5_GET(roce_caps, atts, qp_ts_format);
	} else
		caps->qp_ts_format = 0;

	return DOCA_SUCCESS;
}

static doca_error_t
bf2_sha_supported(const struct bf2_sha_hca_caps *caps)
{
	if (caps == NULL)
		return DOCA_ERROR_NOT_SUPPORTED;
	if (caps->sha_mmo_qp == 0) {
		DOCA_DLOG_ERR("SHA QP unsupported");
		return DOCA_ERROR_NOT_SUPPORTED;
	}
	return DOCA_SUCCESS;
}

doca_error_t
devinfo_has_sha_support(const struct doca_devinfo *devinfo)
{
	struct bf2_sha_hca_caps caps = {0};

	doca_error_t caps_err = bf2_sha_hca_caps_get(devinfo, &caps);

	if (caps_err != DOCA_SUCCESS)
		return caps_err;

	return bf2_sha_supported(&caps);
}

doca_error_t
device_has_sha_support(struct doca_dev *dev)
{
	return devinfo_has_sha_support(doca_dev_as_devinfo(dev));
}

////////////////////////////////////////////////////////////
/**
 * Global qps init setting section
 * only contain medium-perf-critical control functions
 * every thread represents 1 qp
 * when a new thread created,
 * these functions need to be called once
 */
////////////////////////////////////////////////////////////

static inline uint32_t
doca_mlx5_ts_format_conv(uint32_t ts_format)
{
	return ts_format == DOCA_MLX5_HCA_CAP_TIMESTAMP_FORMAT_FR
			? DOCA_MLX5_QPC_TIMESTAMP_FORMAT_FREE_RUNNING
			: DOCA_MLX5_QPC_TIMESTAMP_FORMAT_DEFAULT;
}

static inline uint16_t
sha_ctrl_get_nb_obj(uint16_t nb_desc)
{
	return ((nb_desc / DOCA_MLX5_MMO_NUM_WQE_PER_PAGE) +
		!!(nb_desc % DOCA_MLX5_MMO_NUM_WQE_PER_PAGE));
}

static doca_error_t
create_cq(struct bf2_sha_priv *priv, uint16_t log_cq_size, struct priv_doca_cq **cq)
{
	doca_error_t ret;
	const struct priv_doca_cq_attr cq_attr = priv_doca_cq_attr_create(1UL << log_cq_size);

	ret = priv_doca_dev_cq_create(priv->dev, priv->uar, &cq_attr, cq);
	if (ret != DOCA_SUCCESS)
		DOCA_LOG_ERR("SHA work queue failed to create cq. err = %s", doca_get_error_name(ret));

	return ret;
}

static doca_error_t
destroy_cq(struct priv_doca_cq *cq)
{
	doca_error_t ret;

	ret = priv_doca_dev_cq_destroy(cq);
	if (ret != DOCA_SUCCESS)
		DOCA_LOG_ERR("SHA work queue failed to destroy cq. err = %s", doca_get_error_name(ret));

	return ret;
}

static doca_error_t
destroy_hw_qp(struct bf2_sha_qp *qp, uint16_t q_in)
{
	struct bf2_sha_hw_qp *qp_obj = &qp->qps[q_in];

	doca_mlx5_devx_qp_destroy(&qp_obj->qp_obj);
	memset(qp_obj, 0, sizeof(*qp_obj));

	return DOCA_SUCCESS;
}

static doca_error_t
create_hw_qp(struct bf2_sha_priv *bf2_priv, struct bf2_sha_qp *qp, uint16_t q_ind,
	     uint16_t log_nb_desc)
{
	struct bf2_sha_hw_qp *qp_obj = &qp->qps[q_ind];
	uint32_t pd_num = 0;
	uint32_t uar_id = 0;
	uint32_t cq_num = 0;
	doca_error_t ret;

	ret = priv_doca_cq_number_get(qp->cq, &cq_num);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = priv_doca_uar_id_get(bf2_priv->uar, &uar_id);
	if (ret != DOCA_SUCCESS)
		return ret;

	qp_obj->log_nb_desc = log_nb_desc;
	qp_obj->qpn = q_ind;
	qp_obj->ci = 0;
	qp_obj->pi = 0;
	ret = bf2_doca_mlx5_get_pdn(bf2_priv->pd, &pd_num);
	if (ret)
		return DOCA_ERROR_INITIALIZATION;

	struct doca_mlx5_devx_qp_attr attr = {
		.ts_format = doca_mlx5_ts_format_conv(bf2_priv->qp_ts_format),
		.user_index = q_ind,
		.pd = pd_num,
		.rq_size = 0,
		.sq_size = DOCA_MLX5_BIT32(DOCA_MLX5_MMO_WQE_LOG_NUM(bf2_priv->has_umr, log_nb_desc)),
		.mmo = 1,
	};
	attr.cqn = cq_num;
	attr.uar_index = uar_id;

	ret = doca_mlx5_devx_qp_create(bf2_priv->ctx, &qp_obj->qp_obj,
				       DOCA_MLX5_MMO_WQE_LOG_NUM(bf2_priv->has_umr, log_nb_desc), &attr);
	if (ret) {
		DOCA_DLOG_ERR("Can't create QP object.");
		return DOCA_ERROR_NO_MEMORY;
	}

	ret = doca_mlx5_devx_qp2rts(&qp_obj->qp_obj, 0);
	if (ret) {
		DOCA_DLOG_ERR("Can't change QP state to RTS.");
		destroy_hw_qp(qp, q_ind);
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
setup_buffers(struct bf2_sha_priv *bf2_priv, struct bf2_sha_qp *qp)
{
	struct ibv_pd *pd = bf2_priv->pd;
	uint32_t i;
	void *ptr;
	int ret;

	const size_t meta_size = qp->nb_desc * DOCA_MLX5_MMO_SHA_METADATA_SIZE;

	ptr = aligned_alloc(DOCA_MLX5_MMO_SHA_METADATA_SIZE, meta_size);
	if (ptr == NULL)
		return DOCA_ERROR_NO_MEMORY;

	memset(ptr, 0, meta_size);
	qp->metadata = ibv_reg_mr(pd, ptr, meta_size, IBV_ACCESS_LOCAL_WRITE);
	if (qp->metadata == NULL) {
		DOCA_DLOG_ERR("Failed to register metadata.");
		free(ptr);
		return DOCA_ERROR_NO_MEMORY;
	}

	const size_t max_out_size = qp->nb_desc * DOCA_MLX5_MMO_SHA_MAX_OUTPUT;

	ptr = aligned_alloc(DOCA_MLX5_MMO_SHA_MAX_OUTPUT, max_out_size);
	if (ptr == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_sha_output;
	}

	memset(ptr, 0, max_out_size);
	qp->outputs = ibv_reg_mr(pd, ptr, max_out_size, IBV_ACCESS_LOCAL_WRITE);
	if (qp->outputs == NULL) {
		DOCA_DLOG_ERR("Failed to register output.");
		free(ptr);
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_sha_output;
	}

	if (bf2_priv->has_umr) {
		const size_t klms_size = qp->nb_desc * DOCA_MLX5_MMO_KLMS_SIZE;

		ptr = aligned_alloc(DOCA_MLX5_MMO_KLMS_SIZE, klms_size);
		if (!ptr) {
			ret = DOCA_ERROR_NO_MEMORY;
			goto err_imkey;
		}

		memset(ptr, 0, klms_size);
		qp->imkey_addr = ibv_reg_mr(pd, ptr, klms_size, IBV_ACCESS_LOCAL_WRITE);
		if (!qp->imkey_addr) {
			free(ptr);
			DOCA_DLOG_ERR("Failed to register imkey.");
			ret = DOCA_ERROR_NO_MEMORY;
			goto err_imkey;
		}
	}

	/* Distribute buffers to jobs. */
	for (i = 0; i < qp->nb_desc; i++) {
		qp->jobs[i].output = (uint8_t *)qp->outputs->addr +
					(i % qp->nb_desc) * DOCA_MLX5_MMO_SHA_MAX_OUTPUT;
		qp->jobs[i].output_lkey = qp->outputs->lkey;
		qp->jobs[i].metadata = (uint8_t *)qp->metadata->addr +
					(i % qp->nb_desc) * DOCA_MLX5_MMO_SHA_METADATA_SIZE;
		qp->jobs[i].metadata_lkey = qp->metadata->lkey;
		if (bf2_priv->has_umr && qp->imkey_addr) {
			qp->jobs[i].imkey_array = (struct doca_mlx5_klm *)qp->imkey_addr->addr +
					(i % qp->nb_desc) * DOCA_MLX5_MMO_MAX_KLM_NUM;
		}
	}

	return DOCA_SUCCESS;

err_imkey:
	ptr = qp->outputs->addr;
	ibv_dereg_mr(qp->outputs);
	free(ptr);
err_sha_output:
	ptr = qp->metadata->addr;
	ibv_dereg_mr(qp->metadata);
	free(ptr);

	return ret;
}

static doca_error_t
create_metadata_buffer_pool(struct bf2_sha_priv *bf2_priv, struct bf2_sha_qp *qp)
{
	struct ibv_pd *pd = bf2_priv->pd;
	const size_t meta_size = qp->nb_desc * DOCA_MLX5_MMO_SHA_METADATA_SIZE;

	if (sha_buffer_pool_create(&(bf2_priv->meta_pool.pool), DOCA_MLX5_MMO_SHA_METADATA_SIZE,
				qp->nb_desc, ALIGN_64) != DOCA_SUCCESS)
		return DOCA_ERROR_NO_MEMORY;

	bf2_priv->meta_pool.mr = ibv_reg_mr(pd, bf2_priv->meta_pool.pool->addr, meta_size, IBV_ACCESS_LOCAL_WRITE);
	if (bf2_priv->meta_pool.mr == NULL) {
		DOCA_DLOG_ERR("Failed to register metadata pool.");
		sha_buffer_pool_destroy(bf2_priv->meta_pool.pool);
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

static void
destroy_metadata_buffer_pool(struct bf2_sha_priv *bf2_priv)
{
	void *ptr = bf2_priv->meta_pool.pool->addr;

	ibv_dereg_mr(bf2_priv->meta_pool.mr);
	bf2_priv->meta_pool.pool->addr = ptr;
	sha_buffer_pool_destroy(bf2_priv->meta_pool.pool);
}

inline void *
get_meta_buf(struct sha_impl *impl)
{
	struct bf2_sha_priv *bf2_priv = NULL;

	if (!impl)
		return NULL;
	if (!impl->cookie)
		return NULL;
	bf2_priv = (struct bf2_sha_priv *)(impl->cookie);

	return get_buf_from_pool(bf2_priv->meta_pool.pool);
}

inline bool
is_meta_buf_valid(const struct sha_impl *impl, const void *addr)
{
	const struct bf2_sha_priv *bf2_priv = NULL;
	uint32_t id;

	if (!impl)
		return false;
	if (!impl->cookie)
		return false;
	bf2_priv = (const struct bf2_sha_priv *)(impl->cookie);
	return is_buf_valid(bf2_priv->meta_pool.pool, addr, &id);
}

inline doca_error_t
put_meta_buf(struct sha_impl *impl, void *addr)
{
	struct bf2_sha_priv *bf2_priv = NULL;

	if (!impl)
		return DOCA_ERROR_INVALID_VALUE;
	if (!impl->cookie)
		return DOCA_ERROR_INVALID_VALUE;
	bf2_priv = (struct bf2_sha_priv *)(impl->cookie);
	return put_buf_to_pool(bf2_priv->meta_pool.pool, addr);
}

static inline uint32_t
qp_size_get(struct bf2_sha_hw_qp *qp)
{
	return (1U << qp->log_nb_desc);
}

static void
setup_qps(struct bf2_sha_priv *bf2_priv, struct bf2_sha_qp *queue)
{
	size_t hw_qpid, entry;
	uint32_t job_id;
	struct bf2_sha_hw_qp *qp_obj = NULL;
	uint8_t *wqe = NULL;
	struct bf2_sha_job *job = NULL;

	for (hw_qpid = 0; hw_qpid < queue->nb_obj; hw_qpid++) {
		qp_obj = &queue->qps[hw_qpid];
		wqe = (uint8_t *)qp_obj->qp_obj.wqes;

		for (entry = 0; entry < qp_size_get(qp_obj); entry++) {
			job_id = hw_qpid * qp_size_get(qp_obj) + entry;
			job = &queue->jobs[job_id];

			if (bf2_priv->has_umr) {
				/* Fill UMR WQE with NOP in advance. */
				mlx5dv_set_ctrl_seg((struct mlx5_wqe_ctrl_seg *)wqe, entry * 2,
						MLX5_OPCODE_NOP, 0, qp_obj->qp_obj.qp->id, 0, 12, 0, 0);
				wqe += DOCA_MLX5_MMO_UMR_WQE_SIZE;
			}
			set_metadata_seg((struct doca_mlx5_wqe_metadata_seg *)
					(wqe + DOCA_MLX5_MMO_WQE_METADATA_OFFSET),
					0, queue->metadata->lkey, (uintptr_t)job->metadata);
			set_data_seg((struct mlx5_wqe_data_seg *)(wqe + DOCA_MLX5_MMO_WQE_SCATTER_OFFSET),
				DOCA_MLX5_MMO_SHA_MAX_OUTPUT, queue->outputs->lkey, (uintptr_t)job->output);
			wqe += 64;
		}
		queue->free_qps |= ((uint64_t)1) << hw_qpid;
	}
}

static void
free_buffers(struct bf2_sha_qp *qp)
{
	void *addr;

	if (qp->imkey_addr) {
		addr = qp->imkey_addr->addr;
		ibv_dereg_mr(qp->imkey_addr);
		free(addr);
	}
	if (qp->metadata) {
		addr = qp->metadata->addr;
		ibv_dereg_mr(qp->metadata);
		free(addr);
	}
	if (qp->outputs) {
		addr = qp->outputs->addr;
		ibv_dereg_mr(qp->outputs);
		free(addr);
	}
}

static void
doca_mlx5_mmodev_teardown_fastpath(struct bf2_sha_priv *bf2_priv, uint32_t qp_id)
{
	struct bf2_sha_qp *qp;
	uint32_t i;

	if (bf2_priv->qps == NULL)
		return;

	qp = &bf2_priv->qps[qp_id];

	if (qp->jobs) {
		for (i = 0; i < qp->nb_desc; i++) {
			if (qp->jobs[i].imkey) {
				mlx5dv_devx_obj_destroy(qp->jobs[i].imkey->obj);
				free(qp->jobs[i].imkey);
			}
		}
		free(qp->jobs);
		qp->jobs = NULL;
	}
	free_buffers(qp);
}

static doca_error_t
doca_mlx5_mmodev_setup_fastpath(struct bf2_sha_priv *bf2_priv, uint32_t qp_id)
{
	struct bf2_sha_qp *qp = &bf2_priv->qps[qp_id];
	struct doca_mlx5_klm klm = {0};
	int ret, i;

	struct doca_mlx5_devx_mkey_attr attr = {
		.klm_array = &klm,
		.klm_num = 1,
		.umr_en = 1,
	};

	qp->jobs = aligned_alloc(ALIGN_64, qp->nb_desc * sizeof(*qp->jobs));
	if (qp->jobs == NULL) {
		DOCA_DLOG_ERR("Failed to allocate qp job memory.");
		return DOCA_ERROR_NO_MEMORY;
	}
	memset(qp->jobs, 0, qp->nb_desc * sizeof(*qp->jobs));

	ret = setup_buffers(bf2_priv, qp);
	if (ret != 0) {
		free(qp->jobs);
		qp->jobs = NULL;
		return DOCA_ERROR_INITIALIZATION;
	}

	setup_qps(bf2_priv, qp);

	ret = bf2_doca_mlx5_get_pdn(bf2_priv->pd, &attr.pd);
	if (ret != 0) {
		DOCA_DLOG_ERR("Failed to get pdn.");
		doca_mlx5_mmodev_teardown_fastpath(bf2_priv, qp_id);
		return DOCA_ERROR_INITIALIZATION;
	}

	if (bf2_priv->has_umr) {
		for (i = 0; i < qp->nb_desc; i++) {
			attr.klm_num = DOCA_MLX5_MMO_MAX_KLM_NUM;
			attr.klm_array = qp->jobs[i].imkey_array;
			qp->jobs[i].imkey = doca_mlx5_devx_cmd_mkey_create(bf2_priv->ctx, &attr);
			if (qp->jobs[i].imkey == NULL) {
				DOCA_DLOG_ERR("Failed to allocate imkey.");
				doca_mlx5_mmodev_teardown_fastpath(bf2_priv, qp_id);
				return DOCA_ERROR_NO_MEMORY;
			}
		}
	}

	return DOCA_SUCCESS;
}

static doca_error_t
create_sha_queue_pair(struct bf2_sha_priv *bf2_priv, uint16_t qp_id)
{
	struct bf2_sha_qp *qp = &bf2_priv->qps[qp_id];
	uint32_t qp_flags = 1;
	uint16_t log_desc;
	uint32_t nb_desc;
	int qps_objs;
	int ret;
	int i;

	qp->flags = qp_flags;

	nb_desc = MAX_PER_QP_NUM_DESCS;

	log_desc = bf2_doca_mlx5_log2_u32(nb_desc);
	qp->nb_desc = 1 << log_desc;

	/* UMR mode so add 1 to the log doubling the descriptors. */
	qp->log_cq_size = (bf2_priv->has_umr) ? (log_desc + 1) : log_desc;
	/*
	 * The flag is currently hardcoded by may change later.
	 * This signals out of order (OOO) jobs mode so sets many qp_objs as opposed to one.
	 * Results are processed in order per queue so multiple queues means can processed OOO.
	 */
	if (qp->flags & 1)
		qp->nb_obj = sha_ctrl_get_nb_obj(1 << DOCA_MLX5_MMO_WQE_LOG_NUM(bf2_priv->has_umr, log_desc));
	else
		qp->nb_obj = 1;
#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("qp->nb_obj = %d\n", qp->nb_obj);
#endif

	qp->qps = aligned_alloc(ALIGN_64, qp->nb_obj * sizeof(struct bf2_sha_hw_qp));
	if (!qp->qps) {
		DOCA_DLOG_ERR("Can't allocate qp array memory.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_cq;
	}
	log_desc = bf2_doca_mlx5_log2_u32((1 << log_desc) / qp->nb_obj);
#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("nb_desc per qp-obj = %d\n", log_desc);
#endif
	ret = create_cq(bf2_priv, qp->log_cq_size, &qp->cq);
	if (ret != 0) {
		DOCA_DLOG_ERR("Failed to create completion queue.");
		ret = DOCA_ERROR_INITIALIZATION;
		goto err_cq;
	}

	qps_objs = 0;
	for (i = 0; i < qp->nb_obj; i++) {
		ret = create_hw_qp(bf2_priv, qp, i, log_desc);
		if (ret) {
			DOCA_DLOG_ERR("Can't create qp object.");
			ret = DOCA_ERROR_INITIALIZATION;
			goto err_hw_qp;
		}
		++qps_objs;
	}

	ret = doca_mlx5_mmodev_setup_fastpath(bf2_priv, qp_id);
	if (ret) {
		DOCA_DLOG_ERR("Cannot set up fast path.");
		ret = DOCA_ERROR_INITIALIZATION;
		goto err_hw_qp;
	}

	return DOCA_SUCCESS;

err_hw_qp:
	for (i = 0; i < qps_objs; i++)
		destroy_hw_qp(qp, i);
	destroy_cq(qp->cq);
err_cq:
	free(qp->qps);
	qp->qps = NULL;
	qp->nb_obj = 0;

	return ret;
}

static void
destroy_sha_queue_pairs(struct bf2_sha_priv *bf2_priv)
{
	struct bf2_sha_qp *qp;
	int qp_ind;
	int i;

	if (!bf2_priv->qps)
		return;

	for (qp_ind = 0; qp_ind < bf2_priv->nb_queues; qp_ind++) {
		qp = &bf2_priv->qps[qp_ind];
		doca_mlx5_mmodev_teardown_fastpath(bf2_priv, qp_ind);
		for (i = 0; i < qp->nb_obj; i++)
			destroy_hw_qp(qp, i);
		destroy_cq(qp->cq);
		free(qp->qps);
		qp->qps = NULL;
	}
}

////////////////////////////////////////////////////////////
/**
 * wqe, doorbell setting section
 * contain perf-critical enqueue, dequeue functions
 */
////////////////////////////////////////////////////////////
static inline void
set_sha_ctrl_seg(void *seg, uint8_t re, uint8_t le, uint8_t sha_func,
		uint8_t pe)
{
	DOCA_MLX5_SET(sha_mmo_control, seg, re, re);
	DOCA_MLX5_SET(sha_mmo_control, seg, le, le);
	DOCA_MLX5_SET(sha_mmo_control, seg, sha_func, sha_func);
	DOCA_MLX5_SET(sha_mmo_control, seg, pe, pe);
}

static inline uint32_t
job_id_get(uint32_t qid, size_t qp_size, size_t index)
{
	return qid * qp_size + (index & (qp_size - 1));
}

static inline int
get_free(struct bf2_sha_hw_qp *qp_obj, uint8_t has_umr)
{
	return (qp_size_get(qp_obj) -
		((qp_obj->pi - qp_obj->ci) & (DOCA_MLX5_MMO_MAX_WQE_INDEX >> (has_umr ? 2 : 0))));
}

static inline uint8_t
get_sha_func(enum doca_sha_job_type job_type)
{
	switch (job_type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA1_PARTIAL:
		return BF2_SHA1;
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
		return BF2_SHA256;
	case DOCA_SHA_JOB_SHA512:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		return BF2_SHA512;
	default:
		/*
		 * This case impossible to occurs.
		 * Because the doca_sha_job validity check will found it.
		 */
		DOCA_DLOG_ERR("receive wrong SHA job type [%d]", job_type);
		return BF2_SHA1;
	}
}

static inline void
prep_one(struct bf2_sha_priv *bf2_priv, struct wqe_job *req,
	 struct bf2_sha_hw_qp *qp_obj, struct bf2_sha_job *job, size_t pi, struct doca_mlx5_klm *klm,
	 struct doca_mlx5_klm *meta_klm)
{
	struct mlx5_wqe_data_seg *input_seg = NULL;
	struct mlx5_wqe_data_seg *output_seg = NULL;
	struct mlx5_wqe_data_seg *meta_seg = NULL;
	size_t wqe_offset = (pi & (qp_size_get(qp_obj) - 1)) * (MLX5_SEND_WQE_BB << (bf2_priv->has_umr ? 2 : 0)) +
			(bf2_priv->has_umr ? DOCA_MLX5_MMO_UMR_WQE_SIZE : 0);

#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("in prep_one, wqe_offset = %lu\n", wqe_offset);
	DOCA_DLOG_DBG("-- pi = %lu\n", pi);
#endif
	uint8_t *wqe = (uint8_t *)(uintptr_t)qp_obj->qp_obj.wqes + wqe_offset;
	const int ds = 4; /*  ctrl + meta + input + output */
	uint8_t pe;
	uint8_t re;
	uint8_t sha_func = 0;
	struct partial_job *partial;

	mlx5dv_set_ctrl_seg((struct mlx5_wqe_ctrl_seg *)wqe, (bf2_priv->has_umr ? (pi * 4 + 3) : pi),
			DOCA_MLX5_OPCODE_MMO, DOCA_MLX5_OPC_MOD_MMO_SHA,
			qp_obj->qp_obj.qp->id, 0, ds, 0, 0);

	input_seg = (struct mlx5_wqe_data_seg *)(wqe + DOCA_MLX5_MMO_WQE_GATHER_OFFSET);
	input_seg->byte_count = htobe32(klm->byte_count);
	input_seg->addr = htobe64(klm->address);
	input_seg->lkey = klm->mkey;
	output_seg = (struct mlx5_wqe_data_seg *)(wqe + DOCA_MLX5_MMO_WQE_SCATTER_OFFSET);
	meta_seg = (struct mlx5_wqe_data_seg *)(wqe + DOCA_MLX5_MMO_WQE_METADATA_OFFSET);

	if (req->is_single_wqe_job == true) {
		sha_func = get_sha_func(req->data.whole.state.base.type);
#ifdef USE_SHA_DEBUG_PRINT
		DOCA_DLOG_DBG("in whole-job, sha_func = %d\n", sha_func);
#endif /* USE_SHA_DEBUG_PRINT */
		set_sha_ctrl_seg(wqe + 12, 0, 0, sha_func, 1);
		output_seg->byte_count = htobe32(64);
		output_seg->addr = htobe64((uint64_t)(job->output));
		output_seg->lkey = htobe32(job->output_lkey);
		meta_seg->byte_count = htobe32(DOCA_MLX5_MMO_SHA_METADATA_SIZE);
		meta_seg->addr = htobe64((uint64_t)(job->metadata));
		meta_seg->lkey = htobe32(job->metadata_lkey);
	} else {
		partial = &(req->data.partial);
		pe = !!(partial->is_last_wqe);
		re = !(partial->is_first_wqe);
		sha_func = get_sha_func(partial->state->sha_job.base.type);
#ifdef USE_SHA_DEBUG_PRINT
		DOCA_DLOG_DBG("in partial-job, re = %d, pe = %d, sha_func = %d\n", re, pe, sha_func);
#endif /* USE_SHA_DEBUG_PRINT */
		set_sha_ctrl_seg(wqe + 12, re, 0, sha_func, pe);
		if (pe == 1) {
			/*
			 * ((pe == 1) && (re == 0))
			 * This is most usual case, no parital-sha case.
			 * Do not need to set the meta_seg.
			 */
			output_seg->byte_count = htobe32(DOCA_SHA512_BYTE_COUNT);
			output_seg->addr = htobe64((uint64_t)(job->output));
			output_seg->lkey = htobe32(job->output_lkey);
			if (re == 1) {
				/*
				 * ((pe == 1) && (re == 1))
				 * This is the partial-sha last segment case.
				 */
				meta_seg->byte_count = htobe32(meta_klm->byte_count);
				meta_seg->addr = htobe64(meta_klm->address);
				meta_seg->lkey = meta_klm->mkey;
			} else {
				meta_seg->byte_count = htobe32(DOCA_MLX5_MMO_SHA_METADATA_SIZE);
				meta_seg->addr = htobe64((uint64_t)(job->metadata));
				meta_seg->lkey = htobe32(job->metadata_lkey);
			}
		} else {
			/*
			 * This the partial-sha non-last segment case.
			 * output_address == meta_seg_address + 64
			 */
			output_seg->byte_count = htobe32(meta_klm->byte_count - 64);
			output_seg->addr = htobe64(meta_klm->address + 64);
			output_seg->lkey = meta_klm->mkey;
			meta_seg->byte_count = htobe32(meta_klm->byte_count);
			meta_seg->addr = htobe64(meta_klm->address);
			meta_seg->lkey = meta_klm->mkey;
		}
		if (partial->fence_mode == true)
			((struct mlx5_wqe_ctrl_seg *)wqe)->fm_ce_se |= MLX5_WQE_CTRL_FENCE;
	}
#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("The enqueue wqe is:\n");
	debug_print_wqe((struct doca_mlx5_wqe *) wqe, bf2_priv->has_umr);
#endif /* USE_SHA_DEBUG_PRINT */
#ifdef USE_SHA_DEBUG_PRINT
	++global_enqueue_count;
#endif /* USE_SHA_DEBUG_PRINT */
}

static inline void
prep_nop_sha_wqe_set(struct bf2_sha_priv *bf2_priv, struct wqe_job *req,
		     struct bf2_sha_hw_qp *qp_obj, struct bf2_sha_job *job, size_t pi,
		     struct doca_mlx5_klm *klm, struct doca_mlx5_klm *meta_klm)
{
	size_t wqe_offset = (pi & (qp_size_get(qp_obj) - 1)) * (MLX5_SEND_WQE_BB << ((bf2_priv->has_umr) ? 2 : 0));
	struct mlx5_wqe_ctrl_seg *wqe =
		(struct mlx5_wqe_ctrl_seg *)((uint8_t *)(uintptr_t)qp_obj->qp_obj.wqes +
					     wqe_offset);
#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("in prep_nop_sha_wqe_set, wqe_offset = %lu\n", wqe_offset);
#endif
	if (bf2_priv->has_umr) {
		/* Clear the WQE memory used as UMR WQE previously. */
		if ((be32toh(wqe->opmod_idx_opcode) & 0xff) != MLX5_OPCODE_NOP)
			memset(wqe, 0, DOCA_MLX5_MMO_UMR_WQE_SIZE);
		/* UMR WQE size is 9 DS, align nop WQE to 3 WQEBBS(12 DS). */
		mlx5dv_set_ctrl_seg(wqe, pi * 4, MLX5_OPCODE_NOP, 0, qp_obj->qp_obj.qp->id, 0, 12, 0, 0);
	}
	prep_one(bf2_priv, req, qp_obj, job, pi, klm, meta_klm);
}

static inline void
complete_umr_wqe(struct bf2_sha_qp *qp, struct bf2_sha_hw_qp *qp_obj,
		 struct bf2_sha_job *mkey_job,
		 size_t umr_index, uint32_t klm_size, uint32_t total_len)
{
#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("in complete_umr_wqe, wqe_offset = %lu\n",
		(umr_index & (qp_size_get(qp_obj) - 1)) * (MLX5_SEND_WQE_BB * 4));
#endif
	struct mlx5_wqe_ctrl_seg *wqe = (struct mlx5_wqe_ctrl_seg *)((uint8_t *)
				   (uintptr_t)qp_obj->qp_obj.wqes +
				   ((umr_index & (qp_size_get(qp_obj) - 1)) * (MLX5_SEND_WQE_BB * 4)));
	struct mlx5_wqe_umr_ctrl_seg *ucseg =
				(struct mlx5_wqe_umr_ctrl_seg *)(wqe + 1);
	struct mlx5_wqe_mkey_context_seg *mkc =
				(struct mlx5_wqe_mkey_context_seg *)(ucseg + 1);
	struct doca_mlx5_klm *iklm = (struct doca_mlx5_klm *)(mkc + 1);
	const uint16_t klm_align = SHA_UMR_ALIGN(klm_size, 4);

	memset(wqe, 0, DOCA_MLX5_MMO_UMR_WQE_SIZE);
	/* Set WQE control seg. Non-inline KLM UMR WQE size must be 9 WQE_DS. */
	mlx5dv_set_ctrl_seg(wqe, (umr_index * 4), MLX5_OPCODE_UMR,
			 0, qp_obj->qp_obj.qp->id, 0, 9, 0,
			 htobe32(mkey_job->imkey->id));
	/* Set UMR WQE control seg. */
	ucseg->mkey_mask |= htobe64(MLX5_WQE_UMR_CTRL_MKEY_MASK_LEN |
				MLX5_WQE_UMR_CTRL_FLAG_TRNSLATION_OFFSET |
				MLX5_WQE_UMR_CTRL_MKEY_MASK_ACCESS_LOCAL_WRITE);
	ucseg->klm_octowords = htobe16(klm_align);
	/* Set mkey context seg. */
	mkc->len = htobe64(total_len);
	mkc->qpn_mkey = htobe32(0xffffff00 | (mkey_job->imkey->id & 0xff));
	/* Set UMR pointer to data seg. */
	iklm->address = htobe64((uintptr_t)((uint8_t *)mkey_job->imkey_array));
	iklm->mkey = htobe32(qp->imkey_addr->lkey);
	iklm->byte_count = htobe32(klm_align);
	/* Clear the padding memory. */
	memset((uint8_t *)&mkey_job->imkey_array[klm_size], 0,
	       sizeof(struct doca_mlx5_klm) * (klm_align - klm_size));

	/* Add the following RegEx WQE with fence. */
	wqe = (struct mlx5_wqe_ctrl_seg *)
				(((uint8_t *)wqe) + DOCA_MLX5_MMO_UMR_WQE_SIZE);
	wqe->fm_ce_se |= MLX5_WQE_CTRL_INITIATOR_SMALL_FENCE;
}

static inline int
mkey_klm_available(struct doca_mlx5_klm *klm, uint32_t pos, uint32_t new)
{
	return (klm && ((pos + new) <= DOCA_MLX5_MMO_MAX_KLM_NUM));
}

static inline doca_error_t
prep_sha_umr_wqe_set(struct bf2_sha_priv *bf2_priv, struct bf2_sha_qp *qp,
		     struct bf2_sha_hw_qp *qp_obj, struct wqe_job **reqs,
		     size_t no_reqs, uint16_t front)
{
	struct doca_buf *buf;
	uint8_t *meta_buf = NULL;
	struct bf2_sha_job *job = NULL;
	size_t hw_qpid = qp_obj->qpn;
	size_t left_reqs = no_reqs;
	struct doca_mlx5_klm klm;
	struct doca_mlx5_klm meta_klm;
	size_t pi;
	uint32_t buf_klm_mkey;
	uint16_t req_index = 0;
	struct partial_job *partial = NULL;
	struct whole_job *whole = NULL;
	struct doca_mlx5_klm *mkey_klm = NULL;
	size_t klm_num = 0;
	size_t len = 0;
	size_t mkey_job_id = 0;
	uint32_t i = 0;
	struct doca_buf *next_buf;

	while (left_reqs--)
		doca_prefetch0(reqs[(left_reqs + front) % PER_Q_TEMP_JOBS_SIZE]);
	left_reqs = no_reqs;

	/*
	 * Build the WQE set by reverse.
	 * In case the burst may consume multiple mkeys, build the WQE set as normal will hard to address the last mkey
	 * index, since we will only know the last SHA WQE's index when finishes building.
	 */
	while (left_reqs--) {
		pi = DOCA_MLX5_MMO_UMR_QP_PI_IDX(qp_obj->pi, left_reqs, bf2_priv->has_umr);

		req_index = (left_reqs + front) % PER_Q_TEMP_JOBS_SIZE;
		if (reqs[req_index]->is_single_wqe_job == true) {
			whole = &(reqs[req_index]->data.whole);
			buf = whole->state.req_buf;
			meta_buf = NULL;
			klm.address = (uint64_t)priv_doca_buf_get_data(buf);
			klm.byte_count = priv_doca_buf_get_data_len(buf);
		} else {
			partial = &(reqs[req_index]->data.partial);
			buf = partial->start_buf;
			meta_buf = partial->state->session->meta_buf;
			klm.address = (uint64_t)priv_doca_buf_get_data(buf) + partial->start_buf_offset;
			if (partial->start_buf == partial->last_buf)
				klm.byte_count = partial->last_buf_offset - partial->start_buf_offset;
			else
				klm.byte_count = priv_doca_buf_get_data_len(buf) - partial->start_buf_offset;
			meta_klm.mkey = htobe32(bf2_priv->meta_pool.mr->lkey);
			meta_klm.address = (uint64_t)(meta_buf);
			meta_klm.byte_count = DOCA_MLX5_MMO_SHA_METADATA_SIZE;
		}
		/* Build the KLM directly. */
		if (doca_buf_get_lkey(buf, bf2_priv->dev, DOCA_ACCESS_LOCAL_READ, &buf_klm_mkey) != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("get req_buf lkey error");
			return DOCA_ERROR_INVALID_VALUE;
		}
		klm.mkey = htobe32(buf_klm_mkey);

		const uint32_t new_num = (reqs[req_index]->is_single_wqe_job == true) ? whole->nb_ll_buf :
											partial->nb_ll_buf;

		if (new_num > 1) {
			if (!mkey_klm_available(mkey_klm, klm_num, new_num)) {
				/*
				 * The mkey's KLM is not big enough, create the UMR WQE in the next WQE set.
				 * The previous wqe is a umr-wqe, so finish it first, and move to the next umr-wqe.
				 */
				if (mkey_klm)
					complete_umr_wqe(qp, qp_obj, &qp->jobs[mkey_job_id],
						 DOCA_MLX5_MMO_UMR_QP_PI_IDX(pi, 1, bf2_priv->has_umr),
						 klm_num, len);
				/*
				 * Get the indirect mkey and klm array index from the next wqe set.
				 */
				mkey_job_id = job_id_get(hw_qpid, qp_size_get(qp_obj), pi);
				mkey_klm = qp->jobs[mkey_job_id].imkey_array;
				klm_num = 0;
				len = 0;
			}

			/*
			 * This is the first umr wqe klm element.
			 * Because the sha-wqe is already set, so do an exhange here:
			 * put the sha-wqe's content to first wqe-wqe, and set the sha-wqe points to the first umr-wqe.
			 */
			mkey_klm->address = htobe64(klm.address);
			klm.address = len;
			mkey_klm->byte_count = htobe32(klm.byte_count);
			len += klm.byte_count;
			klm.byte_count = (reqs[req_index]->is_single_wqe_job == true) ? whole->src_len :
											partial->src_len;
			mkey_klm->mkey = klm.mkey;
			klm.mkey = htobe32(qp->jobs[mkey_job_id].imkey->id);
			++mkey_klm;
			++klm_num;

			for (i = 1; i < new_num; i++) {
				if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
					DOCA_DLOG_ERR("cannot get next linked-list buf");
					return DOCA_ERROR_INVALID_VALUE;
				}
				mkey_klm->address = htobe64((uint64_t)priv_doca_buf_get_data(next_buf));
				if (reqs[req_index]->is_single_wqe_job == true)
					mkey_klm->byte_count = htobe32(priv_doca_buf_get_data_len(next_buf));
				else if (i == new_num - 1) {
					if (next_buf != partial->last_buf)
						mkey_klm->byte_count = htobe32(priv_doca_buf_get_data_len(next_buf));
					else
						mkey_klm->byte_count = htobe32(partial->last_buf_offset);
				} else
					mkey_klm->byte_count = htobe32(priv_doca_buf_get_data_len(next_buf));
				if (doca_buf_get_lkey(next_buf, bf2_priv->dev,
					DOCA_ACCESS_LOCAL_READ, &buf_klm_mkey) != DOCA_SUCCESS) {
					DOCA_DLOG_ERR("get req_buf lkey error");
					return DOCA_ERROR_INVALID_VALUE;
				}
				mkey_klm->mkey = htobe32(buf_klm_mkey);
#ifdef USE_SHA_DEBUG_PRINT
				DOCA_DLOG_DBG("[%d]-th umr wqe address = %p, byte_count = 0x%x, mkey = %u\n", i,
					(void *)(be64toh(mkey_klm->address)),
					be32toh(mkey_klm->byte_count),
					buf_klm_mkey);
#endif //USE_SHA_DEBUG_PRINT
				len += be32toh(mkey_klm->byte_count);
				++mkey_klm;
				++klm_num;
				buf = next_buf;
			}
		}

		job = &qp->jobs[job_id_get(hw_qpid, qp_size_get(qp_obj), pi)];
		job->user_job = (void *)reqs[req_index];
		/*
		 * Build the nop + sha WQE set by default. The fist nop WQE
		 * will be updated later as UMR WQE if scattered mbuf exist.
		 */
		prep_nop_sha_wqe_set(bf2_priv, reqs[req_index], qp_obj, job, pi,
				     &klm, ((meta_buf != NULL) ? &meta_klm : NULL));
	}

	/*
	 * Scattered mbuf have been added to the KLM array. Complete the build
	 * of UMR WQE, update the first nop WQE as UMR WQE.
	 */
	if (mkey_klm)
		complete_umr_wqe(qp, qp_obj, &qp->jobs[mkey_job_id], qp_obj->pi, klm_num, len);

	qp_obj->db_pi = DOCA_MLX5_MMO_UMR_QP_PI_IDX(qp_obj->pi, no_reqs - 1, bf2_priv->has_umr);
	qp_obj->pi = DOCA_MLX5_MMO_UMR_QP_PI_IDX(qp_obj->pi, no_reqs, bf2_priv->has_umr);
	return DOCA_SUCCESS;
}

static inline void
send_doorbell(struct bf2_sha_priv *bf2_priv, struct bf2_sha_hw_qp *qp_obj)
{
	uint64_t *doorbell_addr;
	size_t wqe_offset;
	uint8_t *wqe;

	wqe_offset = (qp_obj->db_pi & (qp_size_get(qp_obj) - 1)) * (MLX5_SEND_WQE_BB << (bf2_priv->has_umr ? 2 : 0)) +
		     (bf2_priv->has_umr ? DOCA_MLX5_MMO_UMR_WQE_SIZE : 0);
	wqe = (uint8_t *)(uintptr_t)qp_obj->qp_obj.wqes + wqe_offset;
	((struct mlx5_wqe_ctrl_seg *)wqe)->fm_ce_se |= MLX5_WQE_CTRL_CQ_UPDATE;
#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("qp_obj->db_pi = %lu, wqe_offset = %lu, wqe = %p\n",
		qp_obj->db_pi, wqe_offset, (void *)wqe);
	DOCA_DLOG_DBG("after doorbell The enqueue wqe is:\n");
	debug_print_wqe((struct doca_mlx5_wqe *) wqe, bf2_priv->has_umr);
#endif
	doorbell_addr = (uint64_t *)((uint8_t *)bf2_priv->uar_base_addr + 0x800);
	priv_doca_udma_to_device_barrier();
	qp_obj->qp_obj.db_rec[MLX5_SND_DBR] =
		htobe32((bf2_priv->has_umr ? (qp_obj->db_pi * 4 + 3) : qp_obj->db_pi) & DOCA_MLX5_MMO_MAX_WQE_INDEX);
	priv_doca_mmio_wc_start();
	*doorbell_addr = *(volatile uint64_t *)wqe;
	priv_doca_mmio_flush_writes();
}

static int
bf2_impl_enqueue_jobs(struct sha_impl *impl, uint16_t qpid, struct wqe_job **jobs,
		      uint16_t jobs_count, uint16_t job_front, bool use_burst)
{
	struct bf2_sha_priv *bf2_priv = NULL;
	struct bf2_sha_hw_qp *qp_obj;
	struct bf2_sha_qp *qp;
	size_t hw_qpid, nb_desc;
	size_t try_sent = 0;
	size_t nb_sent = 0;
	size_t i = 0;
	size_t j = 0;
	uint16_t i_index = 0;
	uint16_t j_index = 0;

	if (!impl)
		return -1;
	if (!impl->cookie)
		return -1;
	bf2_priv = (struct bf2_sha_priv *)(impl->cookie);
	if (qpid >= bf2_priv->nb_queues) {
		DOCA_DLOG_ERR("Invalid qid for enqueue.");
		return -1;
	}

	if (use_burst) {
		if (jobs_count < DOCA_SHA_MAX_BURST_COUNT)
			return 0;
	}

	qp = &bf2_priv->qps[qpid];

	while ((hw_qpid = ffsll(qp->free_qps)) && jobs_count) {
		hw_qpid--; /* ffs returns 1 for bit 0 */
		qp_obj = &qp->qps[hw_qpid];

		try_sent = jobs_count;
		nb_desc = get_free(qp_obj, bf2_priv->has_umr);
		if (try_sent > nb_desc)
			try_sent = nb_desc;
		if (try_sent) {
			if (nb_desc == try_sent) {
				/* Taking the last slot, mark qp exhausted for now */
				qp->free_qps &= ~(((uint64_t)1) << hw_qpid);
			}
			/*
			 * set the fence_mode
			 */
			for (i = 0; i < try_sent; i++) {
				i_index = (job_front + nb_sent + i) % PER_Q_TEMP_JOBS_SIZE;
				if (jobs[i_index]->is_single_wqe_job == true)
					continue;
				if (i == 0) {
					jobs[i_index]->data.partial.fence_mode = false;
					continue;
				}
				j = i - 1;
				j_index = (job_front + nb_sent + j) % PER_Q_TEMP_JOBS_SIZE;
				jobs[i_index]->data.partial.fence_mode =
				    ((jobs[j_index]->is_single_wqe_job == false) &&
				     (jobs[i_index]->data.partial.state == jobs[j_index]->data.partial.state));
			}
			if (prep_sha_umr_wqe_set(bf2_priv, qp, qp_obj, jobs, try_sent,
						 job_front + nb_sent) != DOCA_SUCCESS)
				return -1;
			send_doorbell(bf2_priv, qp_obj);

			nb_sent += try_sent;
			jobs_count -= try_sent;
			/*
			 * this break is very important
			 * it guarantees the wqes of the same job is not put into the different qp_obj
			 * because the fence-mode only valid for the same qp_obj.
			 * so that the sequential execution can be guaranteed.
			 */
			break;
		}
	}

	qp->pi += nb_sent;
	return nb_sent;
}

static inline int
check_cqe(struct mlx5_cqe64 *cqe, struct bf2_sha_cqerr *cqerr)
{
	const uint8_t op_code = mlx5dv_get_cqe_opcode(cqe);
	int ret;

	/* CQE with error */
	if (op_code == MLX5_CQE_REQ_ERR || op_code == MLX5_CQE_RESP_ERR) {
		const struct mlx5_err_cqe *err_cqe = (const struct mlx5_err_cqe *)cqe;

		cqerr->status = err_cqe->syndrome;
		cqerr->vendor_err = err_cqe->vendor_err_synd;
		ret = -1;
	} else { /* Good CQE */
		cqerr->status = 0;
		ret = 1;
	}
	return ret;
}

static int
bf2_impl_dequeue_jobs(struct sha_impl *impl, uint16_t qpid,
		      struct wqe_job **finished_jobs,
		      uint16_t finished_job_front,
		      uint16_t max_results)
{
	struct bf2_sha_priv *bf2_priv = NULL;
	struct doca_mlx5_cqe *cqe;
	struct bf2_sha_cqerr *cqerr;
	struct bf2_sha_qp *qp;
	struct priv_doca_cq *cq;
	uint16_t i = 0;
	uint32_t job_id;
	uint16_t wq_counter;
	size_t hw_qpid;
	struct bf2_sha_hw_qp *qp_obj = NULL;
	struct wqe_job *user_job;
	uint8_t resp_byte_count = 0;

	if (!impl)
		return -1;
	if (!impl->cookie)
		return -1;
	bf2_priv = (struct bf2_sha_priv *)(impl->cookie);
	if (qpid >= bf2_priv->nb_queues) {
		DOCA_DLOG_ERR("Invalid queue pair id for dequeue.");
		return -1;
	}

	qp = &bf2_priv->qps[qpid];
	cq = qp->cq;
	cqe = &(qp->cqe);
	cqerr = &(qp->cqerr);
	while (1) {
		if (priv_doca_cq_poll_one(cq, (struct priv_doca_64b_cqe *)cqe) == DOCA_ERROR_AGAIN)
			break;
		if (check_cqe((struct mlx5_cqe64 *)cqe, cqerr) == -1) {
			DOCA_DLOG_ERR("cq received for failed job: status = %d, vendor_error = %d",
				cqerr->status, cqerr->vendor_err);
			return -1;
		}
		wq_counter =
			(be16toh(cqe->wqe_counter) + 1) & DOCA_MLX5_MMO_MAX_WQE_INDEX;
#ifdef USE_SHA_DEBUG_PRINT
		DOCA_DLOG_DBG("wqe_counter = %u\n", wq_counter);
#endif /* USE_SHA_DEBUG_PRINT */
		hw_qpid = cqe->user_index_bytes[2];
		qp_obj = &qp->qps[hw_qpid];
		/* UMR mode WQE counter - move as WQE set(4 WQEBBS).*/
		if (bf2_priv->has_umr)
			wq_counter >>= 2;

#ifdef USE_SHA_DEBUG_PRINT
		DOCA_DLOG_DBG("The dequeue received data byte count: %u\n", be32toh(cqe->byte_cnt));
		DOCA_DLOG_DBG("The completed wqe count: %d, qp_obj->ci = %lu\n",
			be16toh(cqe->wqe_counter) + 1, qp_obj->ci);
		DOCA_DLOG_DBG("The completed send wqe count: %d\n",
			(be16toh(cqe->wqe_counter) + 1) / (bf2_priv->has_umr ? 4 : 1));
#endif /* USE_SHA_DEBUG_PRINT */

		while (qp_obj->ci != wq_counter) {
			if (i == max_results) {
				DOCA_DLOG_DBG("Please increase resp buf count to aovid duplicate resps\n");
				DOCA_DLOG_DBG("qp_obj->ci = %lu, wq_counter = %d\n", qp_obj->ci, wq_counter);
				/* Return without updating cq->ci */
				goto out;
			}

			job_id = job_id_get(hw_qpid, qp_size_get(qp_obj), qp_obj->ci);
#ifdef USE_SHA_DEBUG_PRINT
			DOCA_DLOG_DBG("dequeue's job_id = %08x\n", job_id);
			size_t wqe_offset = (qp_obj->ci & (qp_size_get(qp_obj) - 1)) *
				(MLX5_SEND_WQE_BB << (bf2_priv->has_umr ? 2 : 0)) +
				(bf2_priv->has_umr ? DOCA_MLX5_MMO_UMR_WQE_SIZE : 0);
			uint8_t *wqe = (uint8_t *)(uintptr_t)qp_obj->qp_obj.wqes + wqe_offset;

			DOCA_DLOG_DBG("The dequeue wqe is:\n");
			DOCA_DLOG_DBG("-- db_pi = %lu\n", qp_obj->db_pi);
			DOCA_DLOG_DBG("-- ci = %lu\n", qp_obj->ci);
			debug_print_wqe((struct doca_mlx5_wqe *) wqe, bf2_priv->has_umr);
#endif /* USE_SHA_DEBUG_PRINT */

			user_job = (struct wqe_job *)(qp->jobs[job_id].user_job);
			if (user_job->is_single_wqe_job == true) {
				resp_byte_count = get_response_byte_count(user_job->data.whole.state.base.type);
				/*
				 * Copy data from pre-allocated buf to user-buf
				 */
				memcpy(priv_doca_buf_get_data(user_job->data.whole.state.resp_buf),
					(void *)(qp->jobs[job_id].output),
					resp_byte_count);
				if (doca_buf_set_data(user_job->data.whole.state.resp_buf,
					priv_doca_buf_get_data(user_job->data.whole.state.resp_buf),
					get_response_byte_count(user_job->data.whole.state.base.type))
					!= DOCA_SUCCESS) {
					DOCA_DLOG_ERR("doca_buf_set_data() for response doca_buf failure");
					goto out;
				}
			} else if (user_job->data.partial.is_last_wqe == true) {
				resp_byte_count = get_response_byte_count(
					user_job->data.partial.state->sha_job.base.type);
				/*
				 * Copy data from pre-allocated buf to user-buf
				 */
				memcpy(priv_doca_buf_get_data(user_job->data.partial.state->sha_job.resp_buf),
					(void *)(qp->jobs[job_id].output),
					resp_byte_count);
				if (doca_buf_set_data(user_job->data.partial.state->sha_job.resp_buf,
					priv_doca_buf_get_data(user_job->data.partial.state->sha_job.resp_buf),
					get_response_byte_count(user_job->data.partial.state->sha_job.base.type))
					!= DOCA_SUCCESS) {
					DOCA_DLOG_ERR("doca_buf_set_data() for response doca_buf failure");
					goto out;
				}
			}
			finished_jobs[finished_job_front] = user_job;
			finished_job_front = (finished_job_front + 1) % PER_Q_TEMP_FINISHED_JOBS_SIZE;

			qp_obj->ci = (qp_obj->ci + 1) & (DOCA_MLX5_MMO_MAX_WQE_INDEX >> (bf2_priv->has_umr ? 2 : 0));
			++i;
#ifdef USE_SHA_DEBUG_PRINT
			global_dequeue_count += 1;
#endif /* USE_SHA_DEBUG_PRINT */
		}
		qp->free_qps |= (((uint64_t)1) << hw_qpid);
	}

out:
	qp->ci += i;
	return i;
}

static doca_error_t
bf2_impl_qps_cleanup(struct sha_impl *impl)
{
	struct bf2_sha_priv *bf2_priv = NULL;

	if (!impl)
		return DOCA_ERROR_INVALID_VALUE;
	if (!impl->cookie)
		return DOCA_ERROR_INVALID_VALUE;

	bf2_priv = impl->cookie;

	destroy_metadata_buffer_pool(bf2_priv);

	destroy_sha_queue_pairs(bf2_priv);
	if (bf2_priv->uar != NULL) {
		priv_doca_dev_uar_destroy(bf2_priv->uar);
		bf2_priv->uar = NULL;
	}
	if (bf2_priv->qps) {
		priv_doca_free(bf2_priv->qps);
		bf2_priv->qps = NULL;
	}
	priv_doca_free(bf2_priv);
	impl->cookie = NULL;

#ifdef USE_SHA_DEBUG_PRINT
	DOCA_DLOG_DBG("at [%s], global_enqueue_count = [%lu], global_dequeue_count = [%lu]\n",
		__func__, global_enqueue_count, global_dequeue_count);
#endif /* USE_SHA_DEBUG_PRINT */

	return DOCA_SUCCESS;
}

static void
bf2_impl_qps_destroy(struct sha_impl *impl)
{
	bf2_impl_qps_cleanup(impl);
	if (impl) {
		priv_doca_free(impl);
		impl = NULL;
	}
}

static doca_error_t
bf2_impl_qps_init(struct sha_impl *impl, struct doca_dev *dev)
{
	struct bf2_sha_priv *bf2_priv = NULL;
	int ret;
	int qid = 0;
	struct ibv_context *ibv_ctx = NULL;
	struct ibv_pd *pd_obj = NULL;
	struct bf2_sha_hca_caps caps = {0};

	if (!impl || !dev)
		return DOCA_ERROR_INVALID_VALUE;

	ibv_ctx = priv_doca_dev_to_ibv_ctx(dev);
	ret = doca_dev_get_pd(dev, &pd_obj);
	if (!(ibv_ctx) || (ret != DOCA_SUCCESS))
		return DOCA_ERROR_INVALID_VALUE;

	if (!impl->cookie) {
		impl->cookie = (void *)priv_doca_malloc(1, sizeof(struct bf2_sha_priv));
		if (!impl->cookie)
			return DOCA_ERROR_NO_MEMORY;
	}

	bf2_priv = (struct bf2_sha_priv *)(impl->cookie);
	bf2_priv->dev = dev;
	bf2_priv->ctx = ibv_ctx;
	bf2_priv->pd = pd_obj;
	bf2_priv->nb_queues = 1;

	if (bf2_sha_hca_caps_get(doca_dev_as_devinfo(dev), &caps) != DOCA_SUCCESS)
		return DOCA_ERROR_NOT_SUPPORTED;

	if (bf2_sha_supported(&caps) != DOCA_SUCCESS)
		return DOCA_ERROR_NOT_SUPPORTED;

	if ((caps.umr_indirect_mkey_disabled == 0) && (caps.umr_modify_entity_size_disabled == 0))
		bf2_priv->has_umr = 1;
	else
		bf2_priv->has_umr = 0;
#ifdef DISABLE_UMR
	bf2_priv->has_umr = 0;
#endif /* DISABLE_UMR */

	bf2_priv->qp_ts_format = caps.qp_ts_format;

	/* Get device eq number for event WQ doorbells. */
	ret = mlx5dv_devx_query_eqn(bf2_priv->ctx, 0, &bf2_priv->eqn);
	if (ret) {
		DOCA_DLOG_ERR("Cannot query event queue number.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	/*
	 * Allocate the User Access Region.
	 * First attempt to allocate to cached memory (will require write barrier after writes).
	 * If fails, try to allocate to non-cached memory (no barriers required for write).
	 */
	ret = priv_doca_dev_uar_create(bf2_priv->dev, PRIV_DOCA_UAR_ALLOCATION_TYPE_BLUEFLAME, &(bf2_priv->uar));
	if (ret != DOCA_SUCCESS)
		ret = priv_doca_dev_uar_create(bf2_priv->dev, PRIV_DOCA_UAR_ALLOCATION_TYPE_NONCACHE, &(bf2_priv->uar));
	if (ret == DOCA_SUCCESS)
		ret = priv_doca_uar_page_get(bf2_priv->uar, (void **)&(bf2_priv->uar_base_addr));
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create user memory region.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_free_uar;
	}

	bf2_priv->qps = priv_doca_malloc(bf2_priv->nb_queues, sizeof(struct bf2_sha_qp));
	if (bf2_priv->qps == NULL) {
		DOCA_DLOG_ERR("Failed to allocate qp memory.");
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_free_uar;
	}

	for (qid = 0; qid < bf2_priv->nb_queues; qid++) {
		ret = create_sha_queue_pair(bf2_priv, qid);
		if (ret != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to create queue pair on id [%d].", qid);
			goto free_qp_priv;
		}
	}

	ret = create_metadata_buffer_pool(bf2_priv, bf2_priv->qps);
	if (ret != DOCA_SUCCESS)
		goto free_queue_pair;

#ifdef USE_SHA_DEBUG_PRINT
	global_enqueue_count = 0;
	global_dequeue_count = 0;
	global_enqueue_count_prev = global_enqueue_count;
#endif /* USE_SHA_DEBUG_PRINT */

	return DOCA_SUCCESS;
free_queue_pair:
free_qp_priv:
	destroy_sha_queue_pairs(bf2_priv);
	priv_doca_free(bf2_priv->qps);
err_free_uar:
	priv_doca_dev_uar_destroy(bf2_priv->uar);
	return ret;
}

////////////////////////////////////////////////////////////
/**
 * Export bf2-impl to oursider
 */
////////////////////////////////////////////////////////////
static const struct sha_impl impl_tmp = {
	.cookie      = NULL,
	.has_sha_support_fn = device_has_sha_support,
	.init_fn       = bf2_impl_qps_init,
	.cleanup_fn    = bf2_impl_qps_cleanup,
	.destroy_fn    = bf2_impl_qps_destroy,
	.enqueue_fn    = bf2_impl_enqueue_jobs,
	.dequeue_fn    = bf2_impl_dequeue_jobs
};

doca_error_t
create_sha_impl(struct sha_impl **sha_impl)
{
	struct sha_impl *obj = NULL;

	if (sha_impl == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	obj = priv_doca_malloc(1, sizeof(struct sha_impl));
	if (obj == NULL)
		return DOCA_ERROR_NO_MEMORY;
	memcpy(obj, &impl_tmp, sizeof(*obj));
	obj->cookie = priv_doca_malloc(1, sizeof(struct bf2_sha_priv));
	if (obj->cookie == NULL) {
		priv_doca_free(obj);
		return DOCA_ERROR_NO_MEMORY;
	}

	*sha_impl = obj;

	return DOCA_SUCCESS;
}

