/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef LIB_DOCA_SHA_IMPL_BF2_IMPL_HEADERS_H_
#define LIB_DOCA_SHA_IMPL_BF2_IMPL_HEADERS_H_

#include <limits.h>

#include <priv_doca_os_abstraction_memory.h>

#include "doca_mlx5_common.h"
#include "doca_mlx5_prm.h"
#include "sha_buffer_pool.h"

#if DOCA_SHA_ARM_OS
#if DOCA_SHA_AARCH64
static inline void
doca_prefetch0(const volatile void *p)
{
	asm volatile("PRFM PLDL1KEEP, [%0]" : : "r"(p));
}
#else /* !DOCA_SHA_AARCH64 */
static inline void
doca_prefetch0(const volatile void *p)
{
	asm volatile("pld [%0]" : : "r"(p));
}
#endif
#else /* !DOCA_SHA_ARM_OS */
static inline void
doca_prefetch0(const volatile void *p)
{
	asm volatile("prefetcht0 %[p]" : : [ p ] "m"(*(const volatile char *)p));
}
#endif

/**
 * SHA engine job initialization data.
 */
struct bf2_sha_job {
	void *user_job;			   /**< user job ptr. */
	uint64_t user_id;		   /**< user job id. */
	volatile uint8_t *output;	   /**< engine output address. */
	uint32_t output_lkey;		   /**< engine output address mkey. */
	volatile uint8_t *metadata;	   /**< metadata address. */
	uint32_t metadata_lkey;		   /**< metadata address mkey. */
	struct doca_mlx5_klm *imkey_array; /**< Indirect mkey's KLM array. */
	struct doca_mlx5_devx_obj *imkey;  /**< UMR WQE's indirect meky. */
};

/**
 * SHA engine mkey.
 */
struct bf2_sha_mkey {
	struct mlx5dv_devx_umem *umem;   /**< memory pointed by the mkey. */
	struct doca_mlx5_devx_obj *mkey; /**< mkey object. */
	uint64_t offset;		 /**< offset in a mkey. */
};

/**
 * SHA engine completion queue.
 */
struct bf2_sha_cq {
	uint32_t log_nb_desc;		 /**< Log 2 number of desc for this object. */
	struct doca_mlx5_devx_cq cq_obj; /**< The CQ DevX object. */
	size_t ci;			 /**< cq index. */
};

/**
 * SHA engine queue pair hardware object.
 */
struct bf2_sha_hw_qp {
	uint16_t log_nb_desc;		 /**< Log 2 number of desc for this object. */
	struct doca_mlx5_devx_qp qp_obj; /**< The QP DevX object. */
	size_t pi;			 /**< Producer index. */
	size_t db_pi;			 /**< doorbell index. */
	size_t ci;			 /**< Global completion index. */
	uint32_t qpn;			 /**< Queue pair object index. */
};

/**
 * Complete error report.
 */
struct bf2_sha_cqerr {
	uint64_t wr_id;		/**< cq wr_id. */
	uint32_t status;	/**< cq status. */
	uint32_t vendor_err;	/**< cq vendor error. */
};

/**
 * SHA engine queue pair definition.
 */
struct bf2_sha_qp {
	uint32_t flags;		     /**< QP user flags. */
	uint32_t nb_desc;	     /**< Total number of desc for this qp. */
	struct bf2_sha_hw_qp *qps;   /**< Pointer to qp array. */
	uint16_t nb_obj;	     /**< Number of qp objects. */
	uint32_t log_cq_size;	     /**< Log 2 number of desc for cq object. */
	struct priv_doca_cq *cq;     /**< complete queue. */
	struct doca_mlx5_cqe cqe;    /**< complete queue event. */
	struct bf2_sha_cqerr cqerr;  /**< complete queue error report. */
	uint64_t free_qps;	     /**< bit fields to indicate free qp. */
	struct bf2_sha_job *jobs;    /**< job array. */
	struct ibv_mr *metadata;     /**< metadata region. */
	struct ibv_mr *outputs;      /**< engine output region. */
	struct ibv_mr *imkey_addr;   /**< Indirect mkey array region. */
	size_t ci;		     /**< Global completion index. */
	size_t pi;		     /**< Global producer index. */
};

/**
 * Meta_buf pool used for sha_partial session.
 */
struct meta_buf_pool {
	struct ibv_mr *mr;		/**< metadata region. */
	struct sha_buffer_pool *pool;	/**< buffer pool. */
};

/**
 * SHA engine workq hardware definition.
 */
struct bf2_sha_priv {
	struct doca_dev *dev;		       /**< Doca device. */
	struct ibv_context *ctx;	       /**< Device context. */
	uint8_t *uar_base_addr;		       /**< UAR base address. */
	struct priv_doca_uar *uar;	       /**< UAR object. */
	struct ibv_pd *pd;		       /**< Protection domain for device. */
	struct doca_mlx5_dbr_page_list dbrpgs; /**< Door-bell pages. */
	struct bf2_sha_qp *qps;	               /**< Pointer to the QP array. */
	uint16_t nb_queues;		       /**< Number of queues. Usually everytime workq creation, set it to 1. */
	uint32_t eqn;			       /**< EQ number required for DEVX commands. */
	uint8_t qp_ts_format;		       /**< Whether QP supports timestamp formats. */
	uint8_t has_umr;		       /**< Indicate whether to use UMR wqe */
	struct meta_buf_pool meta_pool;        /**< metadata buffer pool for user job. */
};

/**
 * SHA engine HCA capability bits.
 */
struct bf2_sha_hca_caps {
	uint32_t log_sha_mmo_max_size;		/**< The maximum single wqe job size for sha engine. */
	uint8_t sha_mmo_qp;			/**< Whether sha engine is enabled. */
	uint8_t umr_indirect_mkey_disabled;	/**< Whether umr wqe supported. */
	uint8_t umr_modify_entity_size_disabled;/**< Whether umr wqe supported. */
	uint8_t qp_ts_format;			/**< The queue pair timestamp formart supported. */
};

#endif /* LIB_DOCA_SHA_IMPL_BF2_IMPL_HEADERS_H_ */
