/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * Definition of an abstract implementation to be used by doca sha, Provides a separation between
 * implementation of sha routines for the higher level functionality of the doca sha api.
 */

#ifndef DOCA_SHA_IMPL_H_
#define DOCA_SHA_IMPL_H_

#include <stdint.h>
#include <doca_error.h>

#ifdef __cplusplus
extern "C" {
#endif

struct wqe_job;

/**
 * Structure containing all the functionality required of a sha device by the DOCA sha.
 *
 * @details
 * Allows each DOCA sha instance to be supported by any sha_impl.
 * or any other sha library that can be adapted to this interface.
 */
struct sha_impl {
	/** A place to bind any required extra data into the impl. */
	void *cookie;

	/**
	 * @brief
	 * Check the given doca device supports sha.
	 *
	 * @param [in] dev
	 * The doca device.
	 *
	 * @return
	 * DOCA_SUCCESS or DOCA_ERROR_XXXX.
	 */
	doca_error_t (*has_sha_support_fn)(struct doca_dev *dev);

	/**
	 * @brief
	 * Do any one time setup work that may be required before performing sha operations.
	 *
	 * @param [in] impl
	 * Instance of the impl.
	 *
	 * @param [in] dev
	 * The doca device.
	 *
	 * @return
	 * DOCA_SUCCESS or DOCA_ERROR_XXXX.
	 */
	doca_error_t (*init_fn)(struct sha_impl *impl, struct doca_dev *dev);

	/**
	 * @brief
	 * Cleanup all resources acquired by the cryto_impl from either one time setup or during
	 * runtime operation.
	 *
	 * @param [in] impl
	 * Instance of the impl.
	 *
	 * @return
	 * DOCA_SUCCESS or DOCA_ERROR_XXXX.
	 */
	doca_error_t (*cleanup_fn)(struct sha_impl *impl);

	/**
	 * @brief
	 * Function suitable to destroy this doca_sha_impl.
	 *
	 * @param [in] impl
	 * Instance of the impl.
	 *
	 * @details
	 * Set this to NULL to indicate that the engine should not attempt to destroy this object
	 * when the engine is destroyed.
	 */
	void (*destroy_fn)(struct sha_impl *impl);

	/**
	 * @brief
	 * Send a job to the sha engine to be processed.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] qpid
	 * ID of qp to use
	 *
	 * @param [in] jobs
	 * Details of the jobs. Caller retains ownership of these objects;
	 *
	 * @param [in] jobs_count
	 * Number of the jobs.
	 *
	 * @param [in] tail.
	 * The tail pointer of the jobs circular buffer.
	 *
	 * @param [in] use_burst
	 * Indicate wheither use burst for this reqeust.
	 *
	 * @return
	 * [0..count] Number of jobs enqueued or a negative posix status code.
	 */
	int (*enqueue_fn)(struct sha_impl *inst, uint16_t qpid,
			  struct wqe_job **jobs, uint16_t jobs_count,
			  uint16_t tail, bool use_burst);
	/**
	 * @brief
	 * Fetch job responses from the sha_impl.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in, out] finished_jobs
	 * A pointer to an array of jobs that had received respnses.
	 * this array until either the maximum number of responses are returned or all available
	 * responses have been returned.
	 *
	 * @param [in] front
	 * The front pointer of the finished jobs circular buffer.
	 *
	 * @param [in] max_results
	 * The maximum number of responses the engine is able to accept from the sha_impl.
	 *
	 * @return
	 * [0..max_results] Number of resps dequeued or a negative posix status code.
	 */
	int (*dequeue_fn)(struct sha_impl *inst, uint16_t qpid,
			  struct wqe_job **finished_jobs, uint16_t front,
			  uint16_t max_results);
};

/**
 * @brief
 * Create a doca sha actual implementation.
 *
 * @param [out] impl
 * Instance of the impl.
 */
doca_error_t create_sha_impl(struct sha_impl **impl);

/**
 * @brief
 * Check the given doca device supports sha.
 *
 * @param [in] dev
 * The doca device.
 *
 * @return
 * DOCA_SUCCESS or DOCA_ERROR_XXXX.
 */
doca_error_t device_has_sha_support(struct doca_dev *dev);

/**
 * @brief
 * Check the given doca devinfo supports sha.
 *
 * @param [in] devinfo
 * The doca device.
 *
 * @return
 * DOCA_SUCCESS or DOCA_ERROR_XXXX.
 */
doca_error_t devinfo_has_sha_support(const struct doca_devinfo *devinfo);

/**
 * @brief
 * Acquire a metadata buffer if available.
 *
 * @param [in] impl
 * Instance of the impl.
 *
 * @return
 * Metadata buffer address or NULL.
 */
void *
get_meta_buf(struct sha_impl *impl);

/**
 * @brief
 * Release a metadata buffer.
 *
 * @param [in] impl
 * Instance of the impl.
 * @param [in] addr
 * The buffer address to be released.
 *
 * @return
 * DOCA_SUCCESS or DOCA_ERROR_XXXX.
 */
doca_error_t
put_meta_buf(struct sha_impl *impl, void *addr);

/**
 * @brief
 * check validity of a metadata buffer.
 *
 * @param [in] impl
 * Instance of the impl.
 * @param [in] addr
 * The buffer address to be checked.
 *
 * @return
 * true or false.
 */
bool
is_meta_buf_valid(const struct sha_impl *impl, const void *addr);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_SHA_IMPL_H_ */
