/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <private/common_defines.h>
#include <doca_log.h>
#include <priv_doca_buf.h>
#include <doca_buf.h>
#include <priv_doca_malloc.h>
#include "doca_sha.h"
#include "doca_sha_workq_ctx.h"
#include "doca_sha_impl.h"
#include "doca_sha_bf2_impl.h"
#include "sha_buffer_pool.h"

DOCA_LOG_REGISTER(DOCA_SHA_WORKQ);

static inline bool
is_sha_partial_session_valid(const struct bf2_sha_workq_ctx *ctx, const struct doca_sha_partial_session *session)
{
	uint32_t id;

	if ((ctx == NULL) || (session == NULL)) {
		DOCA_DLOG_ERR("sha_partial_session cannot be NULL for DOCA_SHA_JOB_SHA_PARTIAL");
		return false;
	}
	if (is_buf_valid(ctx->session_pool, session, &id) == false)
		return false;
	return is_meta_buf_valid(ctx->sha_impl, session->meta_buf);
}

static inline bool
is_non_last_segment_valid(const int job_type, const size_t src_len)
{
	switch (job_type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA1_PARTIAL:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
		if ((src_len % 64) != 0) {
			DOCA_DLOG_ERR("src_len = [%lu], but for [%s], non-last-seg length must be multiple of 64",
				src_len, get_job_type_string(job_type));
			return false;
		}
		return true;
	case DOCA_SHA_JOB_SHA512:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		if ((src_len % 128) != 0) {
			DOCA_DLOG_ERR("src_len = [%lu], but for [%s], non-last-seg length must be multiple of 128",
				src_len, get_job_type_string(job_type));
			return false;
		}
		return true;
	default:
		DOCA_DLOG_ERR("unsupported sha job type [%d] found", job_type);
		return false;
	}
}

static inline bool
is_job_data_valid(const struct bf2_sha_workq_ctx *ctx, const struct doca_sha_partial_job *job,
		size_t *req_len_ret, uint32_t *nb_ll_buf_ret)
{
	size_t req_len = 0;
	size_t resp_len = 0;
	uint32_t gathered_bytes = 0;
	uint32_t i = 0;
	uint32_t nb_ll_buf = 1;
	struct doca_buf *next_buf = NULL;
	struct doca_buf *buf = NULL;
	size_t len = 0;

	if ((ctx == NULL) || (job == NULL)) {
		DOCA_DLOG_ERR("ctx or job pointer is NULL");
		return false;
	}
	if (doca_buf_list_num_elements(job->sha_job.req_buf, &nb_ll_buf) != DOCA_SUCCESS)
		nb_ll_buf = 1;
	*nb_ll_buf_ret = nb_ll_buf;

	if (doca_buf_get_data_len(job->sha_job.req_buf, &req_len) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("job request buffer get length error");
		return false;
	}

	if (nb_ll_buf > 1) {
		buf = job->sha_job.req_buf;
		for (i = 1; i < nb_ll_buf; i++) {
			if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("job request buffer linked_list next_buf not set");
				return false;
			}
			if (doca_buf_get_data_len(next_buf, &len) != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("job request buffer linked list next_buf get length error");
				return false;
			}
			req_len += len;
			buf = next_buf;
		}
#ifdef USE_SHA_DEBUG_PRINT
		DOCA_DLOG_DBG("linked_list doca_buf total req_len is: %ld\n", req_len);
#endif /* USE_SHA_DEBUG_PRINT */
	}
	*req_len_ret = req_len;

	if (req_len == 0) {
		DOCA_DLOG_ERR("job request buffer length cannot be 0");
		return false;
	}
	if (req_len > DOCA_SHA_MAX_JOB_BYTE_COUNT) {
		DOCA_DLOG_ERR("job request byte count must <= 0x%lx", DOCA_SHA_MAX_JOB_BYTE_COUNT);
		return false;
	}
	/*
	 * after doca-core is updated, i can use priv_doca_buf_get_data_len() to replace these 2 lines.
	 */
	resp_len = priv_doca_buf_get_len(job->sha_job.resp_buf);
	resp_len -= (size_t)priv_doca_buf_get_data(job->sha_job.resp_buf) -
		    (size_t)priv_doca_buf_get_addr(job->sha_job.resp_buf);

	if ((resp_len == 0) || (resp_len < get_response_byte_count(job->sha_job.base.type))) {
		DOCA_DLOG_ERR("For [%s], job response buffer length need >= %d",
			get_job_type_string(job->sha_job.base.type),
			get_response_byte_count(job->sha_job.base.type));
		return false;
	}
	switch (job->sha_job.base.type) {
	case DOCA_SHA_JOB_SHA1_PARTIAL:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		/*
		 * Do a job->session validity check.
		 * Just make sure illegal address does not stall the engine.
		 * If sha_partial is used, user must provide meta_buf for partial sha result.
		 * and the sha_partial buf must >= 128byte.
		 * If user calls doca_sha_partial_session_create() provided by us, this requirement can be met.
		 * If user just pass a random value, the check should fail.
		 */
		if (is_sha_partial_session_valid(ctx, job->session) != true) {
			DOCA_DLOG_ERR("invalid sha_partial_session found");
			return false;
		}

		/*
		 * sha_partial job length check
		 * only the last segment allows seg-byte-count != multiple-of-64 for sha1 and sha256
		 * only the last segment allows seg-byte-count != multiple-of-128 for sha512
		 */
		if (job->sha_job.flags == DOCA_SHA_JOB_FLAGS_NONE) {
			if (is_non_last_segment_valid(job->sha_job.base.type, req_len) == false)
				return false;
		}
		if (job->session->meta_buf == NULL) {
			DOCA_DLOG_ERR("meta_buf cannot be NULL for DOCA_SHA_JOB_SHA_PARTIAL");
			return false;
		}
		/*
		 * check accumulated job size <= max-allowed-job-size
		 */
		gathered_bytes = be32toh(((uint32_t *)(job->session->meta_buf))[METADATA_BUF_GATHERED_LENGTH_UINT32_OFFSET]);
#ifdef USE_SHA_DEBUG_PRINT
		DOCA_DLOG_DBG("The sha-engine gathered_bytes = 0x%012x\n", gathered_bytes);
		DOCA_DLOG_DBG("The accumulated processed bytes = 0x%012x\n", job->session->processed_bytes);
#endif /* USE_SHA_DEBUG_PRINT */
		if (job->session->processed_bytes + req_len > DOCA_SHA_MAX_JOB_BYTE_COUNT) {
			DOCA_DLOG_ERR("the request job byte count must <= 0x%lx", DOCA_SHA_MAX_JOB_BYTE_COUNT);
			return false;
		}
		if (gathered_bytes != job->session->processed_bytes) {
			DOCA_DLOG_ERR("internal state for sha_partial session corrupted");
			return false;
		}
		break;
	default:
		break;
	}
	return true;
}

static inline int
sha_dispatch_job(struct bf2_sha_workq_ctx *ctx, bool use_burst)
{
	int enq_ret = 0;

	if (ctx->nb_job_in_batch > 0) {
		if (ctx->sha_impl == NULL)
			return -1;

		enq_ret = ctx->sha_impl->enqueue_fn(ctx->sha_impl, 0,
				ctx->jobs,
				MIN(PER_Q_TEMP_JOBS_SIZE, ctx->nb_job_in_batch),
				ctx->job_in_batch_tail,
				use_burst);

		if ((enq_ret > 0) && (enq_ret <= ctx->nb_job_in_batch)) {
			ctx->nb_job_sent_in_batch += enq_ret;
			ctx->nb_job_in_batch -= enq_ret;
			ctx->job_in_batch_tail = (ctx->job_in_batch_tail + enq_ret) % PER_Q_TEMP_JOBS_SIZE;
		} else if (enq_ret < 0) {
			DOCA_DLOG_ERR("sha-enq return wrong value [%d]", enq_ret);
			enq_ret = -1;
		}
	}

	return enq_ret;
}

/**
 * This function extracts the next wqe_job candidate to be put into the queue for enqueue.
 *
 * The bf2 SHA engine has data_length limitation for a single wqe: 128Mbyte (2^27).
 * In the queue pair setting, I pre_config 128 UMR_mkey elements for a wqe: 1 linked_list_element is for 1 UMR_mkey.
 *
 * The pre_config wqe buffer for user job can be seen as a 2-dimension array:
 * each row represnets a wqe_job:
 *     sha_wqe_0/(umr_wqe_0), umr_mkey_0_0, umr_mkey_0_1, ..., umr_mkey_0_127
 *     sha_wqe_1/(umr_wqe_1), umr_mkey_1_0, umr_mkey_1_1, ..., umr_mkey_1_127
 *     sha_wqe_2/(umr_wqe_2), umr_mkey_2_0, umr_mkey_2_1, ..., umr_mkey_2_127
 *     ...
 *     sha_wqe_N/(umr_wqe_N), umr_mkey_N_0, umr_mkey_N_1, ..., umr_mkey_N_127
 * The first elment in each row is configured into: sha_wqe or umr_wqe, depending on whether the user_job is
 * a linked_list or a non_linked_list job.
 *
 * So for a non-linked-list user job, when doing the wqe_job extraction:
 *     the first element in each row is configured into a: sha_wqe. and all the subsequent umr_mkeys are useless.
 *     only 1 restrction (128M) is considered, and only data-split is needed.
 *     and the purpose of this data-split is to make sure: wqe_data_len <= 128M.
 *
 * but for a linked-list user job, when doing the wqe_job extraction:
 *     the first element in each row is configured into a: umr_wqe, and the subsequent umr_mkeys also need suitable
 *     values depending on the linked_list doca_buf elements.
 *     we need consider 2 restrctions (128M and 128UMR_mkey), thus, both data-split and data-concatenation is needed.
 *     and the purpose of this data-split is to make sure: data_len <= 128M, and linked_list_elements <= 128.
 *     and the purpose of the data-concatenation is to make sure: the extracted wqe_job count as small as possible.
 *
 * @param [in]
 * remain_len: the data byte count still need to be processed.
 * @param [in]
 * remain_nb_ll_buf: the linked_list element count still need to be processed.
 * @param [in]
 * len_threshold: 128Mbyte.
 * @param [in]
 * nb_ll_buf_threshold: 128.
 * @param [in]
 * partial->start_buf: the first doca_buf to be processed.
 * @param [in]
 * partial->start_buf_offset: the first byte in the first doca_buf to be processed.
 *                            so its value range is >=0, and <= start_buf_len-1.
 *
 * @param [out]
 * partial->last_buf: the last doca_buf to be processed if partial-data of the last doca_buf will be used.
 *               the next doca_buf of the last doca_buf to be processed if all-data of the last doca_buf will be used.
 * @param [out]
 * partial->last_buf_offset: the byte next to the last byte in the last doca_buf to be processed.
 *                      so its value range is:
 *                          ==0, when the data in the last_buf will not be used.
 *                          >0 and <=last_buf_len-1, whe partial-data in the last_buf will be used.
 *                          ==lastt_buf_len, when whole-data in the last_buf will be used.
 * @param [out]
 * partial->nb_ll_buf: the linked_list element count will be used including both the start_buf and last_buf.
 * @param [out]
 * partial->src_len: the total source data extracted.
 *              For the start_buf, its tail_part is included, and data_len == start_buf_len - start_buf_offset.
 *              For the middle linked_list bufs, all of them data is included.
 *              For the last_buf, its head_part in included, and data_len == last_buf_offset.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t
extract_ll_doca_buf_range_within_threshold(const uint32_t remain_len, const uint32_t remain_nb_ll_buf,
					const uint32_t len_threshold, const uint32_t nb_ll_buf_threshold,
					struct partial_job *partial)
{
	struct doca_buf *buf;
	uint32_t start_buf_offset;
	struct doca_buf *next_buf;
	uint32_t new_src_len;
	size_t len;
	struct doca_buf *last_buf;
	uint32_t last_buf_offset;
	uint32_t nb_ll_buf;
	uint32_t src_len;

	if (remain_len == 0 || remain_nb_ll_buf == 0)
		return DOCA_ERROR_INVALID_VALUE;
	if (partial == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	buf = partial->start_buf;
	start_buf_offset = partial->start_buf_offset;
	if (doca_buf_get_data_len(buf, &len) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Get doca_buf len error");
		return DOCA_ERROR_INVALID_VALUE;
	}
	src_len = len - start_buf_offset;

	/*
	 * If the current doca_buf data length >= 128MB, return immediately.
	 * This single doca_buf is used as a single_wqe job when its data length == 128MB.
	 * This single doca_buf will be split into several single_wqe jobs when its data length > 128MB.
	 * If we only have 1 linked_list elment left, we also need return.
	 */
	if (src_len > len_threshold) {
		/*
		 * split the current buf.
		 */
		partial->nb_ll_buf = 1;
		partial->last_buf = buf;
		partial->last_buf_offset = start_buf_offset + len_threshold;
		partial->src_len = len_threshold;
		return DOCA_SUCCESS;
	} else if (src_len == len_threshold) {
		nb_ll_buf = 1;
		if (remain_nb_ll_buf == 1) {
			/*
			 * if remain_nb_ll_buf == 1
			 * this is the only left doca_buf.
			 */
			last_buf = buf;
			last_buf_offset = len;
		} else if (remain_nb_ll_buf > 1) {
			/*
			 * if remain_nb_ll_buf > 1, we move point to next_buf.
			 */
			if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("get next linked-list doa_buf error");
				return DOCA_ERROR_INVALID_VALUE;
			}
			last_buf = next_buf;
			last_buf_offset = 0;
		} else {
			DOCA_DLOG_ERR("Unexpected remain_nb_ll_buf < 1");
			return DOCA_ERROR_INVALID_VALUE;
		}
		partial->nb_ll_buf = 1;
		partial->last_buf = last_buf;
		partial->last_buf_offset = last_buf_offset;
		partial->src_len = src_len;
		return DOCA_SUCCESS;
	}
	/*
	 * src_len < len_threshold:
	 * If the current doca_buf data length < 128MB, we continue to search
	 */
	nb_ll_buf = 1;
	last_buf = buf;
	last_buf_offset = len;
	while ((src_len < remain_len) && (src_len < len_threshold) &&
	       (nb_ll_buf < remain_nb_ll_buf) && (nb_ll_buf < nb_ll_buf_threshold)) {
		if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("get next linked-list doa_buf error");
			return DOCA_ERROR_INVALID_VALUE;
		}
		if (doca_buf_get_data_len(next_buf, &len) != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Get doca_buf len error");
			return DOCA_ERROR_INVALID_VALUE;
		}
		new_src_len = src_len + len;
		if (new_src_len == len_threshold) {
			nb_ll_buf += 1;
			src_len = new_src_len;
			if (remain_nb_ll_buf == nb_ll_buf) {
				/*
				 * if remain_nb_ll_buf == nb_ll_buf
				 * this is the only left doca_buf.
				 */
				last_buf = next_buf;
				last_buf_offset = len;
			} else if (remain_nb_ll_buf > nb_ll_buf) {
				/*
				 * if remain_nb_ll_buf > nb_ll_buf, we move point to next_buf.
				 */
				buf = next_buf;
				if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
					DOCA_DLOG_ERR("get next linked-list doa_buf error");
					return DOCA_ERROR_INVALID_VALUE;
				}
				last_buf = next_buf;
				last_buf_offset = 0;
			} else {
				DOCA_DLOG_ERR("Unexpected remain_nb_ll_buf < nb_ll_buf");
				return DOCA_ERROR_INVALID_VALUE;
			}
			partial->nb_ll_buf = nb_ll_buf;
			partial->last_buf = last_buf;
			partial->last_buf_offset = last_buf_offset;
			partial->src_len = src_len;
			return DOCA_SUCCESS;
		} else if (new_src_len > len_threshold) {
			/*
			 * If the new_src_len > 128MB, then, we do split.
			 */
			nb_ll_buf += 1;
			last_buf = next_buf;
			last_buf_offset = len_threshold - src_len;
			src_len = len_threshold;
			partial->nb_ll_buf = nb_ll_buf;
			partial->last_buf = last_buf;
			partial->last_buf_offset = last_buf_offset;
			partial->src_len = src_len;
			return DOCA_SUCCESS;
		}
		/* if (new_src_len < len_threshold) */
		nb_ll_buf += 1;
		last_buf = next_buf;
		last_buf_offset = len;
		buf = next_buf;
		src_len = new_src_len;
	}
	/*
	 * We reach here, it mean: src_len < len_threshold
	 * and the last_buf above is completely included, and we need to check whether we can move to next_buf.
	 */
	if (remain_nb_ll_buf > nb_ll_buf) {
		if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("get next linked-list doa_buf error");
			return DOCA_ERROR_INVALID_VALUE;
		}
		last_buf = next_buf;
		last_buf_offset = 0;
	}
	partial->nb_ll_buf = nb_ll_buf;
	partial->last_buf = last_buf;
	partial->last_buf_offset = last_buf_offset;
	partial->src_len = src_len;
	return DOCA_SUCCESS;
}

static inline doca_error_t
sha_job_state_submit(struct bf2_sha_workq_ctx *ctx)
{
	struct doca_sha_partial_job *state = NULL;
	struct wqe_job *job = NULL;
	struct partial_job *partial = NULL;
	uint16_t i = 0;
	uint16_t cur_job_state_index;
	uint32_t remain_len = 0;
	uint16_t nb_state_in_batch_backup;
	bool no_free_job = false;
	uint32_t remain_nb_ll_buf = 0;

	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	/*
	 * Iterate the job_states buffer, spawn job, insert job into the job buffer
	 */
	nb_state_in_batch_backup = ctx->nb_state_in_batch;
	for (i = 0; (i < nb_state_in_batch_backup) && (ctx->nb_job_in_batch < PER_Q_TEMP_JOBS_SIZE); i++) {
		cur_job_state_index = ctx->state_in_batch_tail;
		state = ctx->states[cur_job_state_index];
		if (state == NULL) {
			DOCA_DLOG_ERR("NULL job-state ptr found");
			return DOCA_ERROR_BAD_STATE;
		}

		/*
		 * A wqe_job == a user_job.
		 */
		if (state->session->wait_resp == true) {
			/*
			 * mv this state to the last position.
			 * this is a fair-play to give all job-states equal chance to be checked.
			 * make sure the not-null positions are continous.
			 */
			if (ctx->nb_state_in_batch < PER_Q_TEMP_STATES_SIZE) {
				ctx->state_in_batch_tail = (ctx->state_in_batch_tail + 1) % PER_Q_TEMP_STATES_SIZE;
				ctx->states[ctx->state_in_batch_front] = state;
				ctx->state_in_batch_front = (ctx->state_in_batch_front + 1) % PER_Q_TEMP_STATES_SIZE;
			}
			continue;
		}

		while ((remain_len = state->session->req_len - state->session->sent_bytes) > 0) {
			if (ctx->nb_job_in_batch >= PER_Q_TEMP_JOBS_SIZE) {
				no_free_job = true;
				break;
			}
			remain_nb_ll_buf = state->session->nb_ll_buf - state->session->sent_nb_ll_buf;

			job = (struct wqe_job *)get_buf_from_pool(ctx->job_pool);
			if (job == NULL) {
				no_free_job = true;
				break;
			}
			job->is_single_wqe_job = false;
			partial = &(job->data.partial);
			partial->state = state;
			partial->start_buf = state->session->start_buf;
			partial->start_buf_offset = state->session->start_buf_offset;
			if (extract_ll_doca_buf_range_within_threshold(remain_len, remain_nb_ll_buf,
					MAX_BYTE_COUNT_PER_WQE, MAX_LL_DOCA_BUF_NUM,
					partial) != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("get a valid doca buf range failure");
				return DOCA_ERROR_INVALID_VALUE;
			}
			if (state->sha_job.base.type >= DOCA_SHA_JOB_SHA1_PARTIAL) {
				partial->is_first_wqe =
					((state->session->is_first_segment == true) &&
					 (partial->start_buf == state->sha_job.req_buf) &&
					 (partial->start_buf_offset == 0));
				partial->is_last_wqe =
					((state->sha_job.flags == DOCA_SHA_JOB_FLAGS_SHA_PARTIAL_FINAL) &&
					 (state->session->req_len ==
					  (size_t)(state->session->sent_bytes + partial->src_len)));
			} else {
				partial->is_first_wqe =
					((partial->start_buf == state->sha_job.req_buf) &&
					 (partial->start_buf_offset == 0));
				partial->is_last_wqe =
					(state->session->req_len ==
					 (size_t)(state->session->sent_bytes + partial->src_len));
			}

			if ((partial->is_last_wqe == false) && (partial->nb_ll_buf > 1)) {
				/*
				 * If the extracted data is a linked_list wqe job, although it is within the range
				 * of nb_ll_buf_threshold, but it cannot guarantee the output meets the requirement of
				 * partial_sha non_last_segment length requirement.
				 * For example, the whole segment length == 256, but its nb_ll_buf == 129, and with
				 * the first 128 ll_bufs length_sum == 255.
				 * So we need this length check.
				 */
				if (is_non_last_segment_valid(partial->state->sha_job.base.type, partial->src_len) == false)
					return DOCA_ERROR_BAD_STATE;
			}

			state->session->start_buf = partial->last_buf;
			state->session->start_buf_offset = partial->last_buf_offset;
			state->session->sent_bytes += partial->src_len;
			state->session->sent_nb_ll_buf += partial->nb_ll_buf;
			if ((partial->last_buf_offset > 0) &&
			    (partial->last_buf_offset < priv_doca_buf_get_data_len(partial->last_buf))) {
				/*
				 * When partial-data of the last_buf is need, this last_buf will be the start_buf of
				 * the next wqe_job, so we need to reduce 1 for the total sent linked_list elment count.
				 */
				state->session->sent_nb_ll_buf -= 1;
			}
			state->session->wait_resp = true;

			ctx->jobs[ctx->job_in_batch_front] = job;
			ctx->job_in_batch_front = (ctx->job_in_batch_front + 1) % PER_Q_TEMP_JOBS_SIZE;
			ctx->nb_job_in_batch++;

			ctx->nb_successive_job_total++;

			/*
			 * The test shows:
			 * (1) if a non_linked_list job is split into several single wqe jobs, the fence_mode is useful
			 *     to make sure their sequential processing.
			 * (2) if a linked_list job is split into several umr wqe jobs, the fence_mode does not work
			 *     anymore, so when the state is a linked-list user job, only a single extracted umr-wqe
			 *     job is allowed for a single enqueue, to make sure the extracted umr wqe jobs of the same
			 *     user job can be sequentially processed.
			 */
			if (state->session->nb_ll_buf > 1)
				break;
		}

		if (state->session->sent_bytes >= state->session->req_len) {
			if (state->session->sent_nb_ll_buf < state->session->nb_ll_buf) {
				DOCA_DLOG_ERR("req_len and nb_ll_buf should be consumed at the same time");
				return DOCA_ERROR_INVALID_VALUE;
			}
			ctx->states[cur_job_state_index] = NULL;
			/*
			 * make sure the not-null positions are continous
			 */
			ctx->state_in_batch_tail = (ctx->state_in_batch_tail + 1) % PER_Q_TEMP_STATES_SIZE;
			ctx->nb_state_in_batch--;
		} else {
			/*
			 * mv the checked state to the last position.
			 * this is a fair-play to give all job-states equal chance to be checked.
			 * make sure the not-null positions are continous.
			 */
			if (ctx->nb_state_in_batch < PER_Q_TEMP_STATES_SIZE) {
				ctx->state_in_batch_tail = (ctx->state_in_batch_tail + 1) % PER_Q_TEMP_STATES_SIZE;
				ctx->states[ctx->state_in_batch_front] = state;
				ctx->state_in_batch_front = (ctx->state_in_batch_front + 1) % PER_Q_TEMP_STATES_SIZE;
			}
		}
		if (no_free_job)
			break;
	}

	return DOCA_SUCCESS;
}

doca_error_t
bf2_sha_partial_session_copy(void *ctx,
			const struct doca_sha_partial_session *from,
			struct doca_sha_partial_session *to)
{
	if ((ctx == NULL) || (from == NULL) || (to == NULL))
		return DOCA_ERROR_INVALID_VALUE;

	if ((from->meta_buf == NULL) || (to->meta_buf == NULL))
		return DOCA_ERROR_UNEXPECTED;

	((uint32_t *)(to->meta_buf))[METADATA_BUF_GATHERED_LENGTH_UINT32_OFFSET] =
			((uint32_t *)(from->meta_buf))[METADATA_BUF_GATHERED_LENGTH_UINT32_OFFSET];
	memcpy(to->meta_buf + METADATA_BUF_SHA_PARTIAL_RESULT_UINT8_OFFSET,
			from->meta_buf + METADATA_BUF_SHA_PARTIAL_RESULT_UINT8_OFFSET,
			METADATA_BUF_SHA_PARTIAL_RESULT_BYTE_COUNT);
	to->processed_bytes = from->processed_bytes;
	return DOCA_SUCCESS;
}

static inline struct doca_sha_partial_session *
get_sha_partial_session(struct bf2_sha_workq_ctx *ctx, int from)
{
	struct doca_sha_partial_session *session = get_buf_from_pool(ctx->session_pool);

	if (session == NULL)
		return NULL;
	session->meta_buf = get_meta_buf(ctx->sha_impl);
	if (session->meta_buf == NULL) {
		put_buf_to_pool(ctx->session_pool, session);
		return NULL;
	}
	((uint32_t *)(session->meta_buf))[METADATA_BUF_GATHERED_LENGTH_UINT32_OFFSET] = 0;
	session->processed_bytes = 0;
	session->sent_bytes = 0;
	session->sent_nb_ll_buf = 0;
	session->recv_bytes = 0;
	session->is_first_segment = true;
	session->wait_resp = false;
	session->alloc_status = from;
	return session;
}

doca_error_t
bf2_sha_partial_session_create(void *workq_ctx,
				struct doca_sha_partial_session **session)
{
	struct bf2_sha_workq_ctx *ctx = (struct bf2_sha_workq_ctx *)workq_ctx;

	*session = get_sha_partial_session(ctx, SESSION_ALLOC_STATUS_EXTERNAL);
	if (*session == NULL)
		return DOCA_ERROR_NO_MEMORY;
	return DOCA_SUCCESS;
}

static inline doca_error_t
put_sha_partial_session(struct bf2_sha_workq_ctx *ctx, struct doca_sha_partial_session *session, int from)
{
	doca_error_t ret;

	if (session == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	if ((session->alloc_status == SESSION_ALLOC_STATUS_FREE) ||
	     (from == SESSION_ALLOC_STATUS_FREE))
		return DOCA_SUCCESS;
	if (session->alloc_status != from) {
		switch (session->alloc_status) {
		case SESSION_ALLOC_STATUS_INTERNAL:
			/*
			 * The internally created session cannot be freed by external.
			 * This situation is: the session once used by external,
			 * and auto freed by internal at the end of job before user calls session_destroy().
			 * And now it is used by another internally created session.
			 */
			return DOCA_SUCCESS;
		case SESSION_ALLOC_STATUS_EXTERNAL:
			/*
			 * This situation is: the externally created session is auto
			 * freed by internal at the end of job before user calls session_destroy().
			 */
			break;
		default:
			DOCA_DLOG_ERR("Unsupported session alloc status [%d]", session->alloc_status);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}
	ret = put_meta_buf(ctx->sha_impl, session->meta_buf);
	if (ret == DOCA_SUCCESS) {
		ret = put_buf_to_pool(ctx->session_pool, session);
		if (ret == DOCA_SUCCESS)
			session->alloc_status = SESSION_ALLOC_STATUS_FREE;
	}
	return ret;
}

doca_error_t
bf2_sha_partial_session_destroy(void *workq_ctx,
				struct doca_sha_partial_session *session)
{
	struct bf2_sha_workq_ctx *ctx = (struct bf2_sha_workq_ctx *)workq_ctx;

	return put_sha_partial_session(ctx, session, SESSION_ALLOC_STATUS_EXTERNAL);
}

doca_error_t
bf2_sha_workq_job_submit(void *workq_ctx, struct doca_sha_partial_job const *user_job)
{
	struct bf2_sha_workq_ctx *ctx = (struct bf2_sha_workq_ctx *)workq_ctx;
	struct doca_sha_partial_session *session = NULL;
	struct doca_sha_partial_job *state = NULL;
	bool is_single_wqe_job;
	bool is_large_size_job; /* indicate a non-partial user job with length > threshold */
	struct wqe_job *job;
	doca_error_t ret;
	size_t req_len = 0;
	bool is_large_ll_job; /* indicate a non-partial user job with linked_list element count > threshold */
	uint32_t nb_ll_buf = 0;

	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	if (ctx->jobs == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	/*
	 * when the job batch is full, send first.
	 */
	if (ctx->nb_job_in_batch >= PER_Q_TEMP_JOBS_SIZE) {
		if (sha_dispatch_job(ctx, false) < 0)
			return DOCA_ERROR_BAD_STATE;
		return DOCA_ERROR_NO_MEMORY;
	}

	if (ctx->nb_state_in_batch >= PER_Q_TEMP_STATES_SIZE)
		return DOCA_ERROR_NO_MEMORY;

	/*
	 * Check job data validity
	 * only pass the data integrity check, we can continue.
	 * And to avoid redundant function call, during the check processing,
	 * we also extract this job's total_request_len and total_linked_list_element_count.
	 */
	if (!is_job_data_valid(ctx, user_job, &req_len, &nb_ll_buf))
		return DOCA_ERROR_INVALID_VALUE;
	is_single_wqe_job = ((user_job->sha_job.base.type < DOCA_SHA_JOB_SHA1_PARTIAL) &&
			     (req_len <= MAX_BYTE_COUNT_PER_WQE) &&
			     (nb_ll_buf <= MAX_LL_DOCA_BUF_NUM));

	if (is_single_wqe_job == true) {
		job = (struct wqe_job *)get_buf_from_pool(ctx->job_pool);
		if (job == NULL)
			return DOCA_ERROR_NO_MEMORY;
		job->is_single_wqe_job = true;
		/*
		 * Copy user_job into the job_pool
		 */
		memcpy(&(job->data.whole), user_job, sizeof(struct doca_sha_job));
		job->data.whole.nb_ll_buf = nb_ll_buf;
		job->data.whole.src_len = req_len;

		ctx->jobs[ctx->job_in_batch_front] = job;
		ctx->job_in_batch_front = (ctx->job_in_batch_front + 1) % PER_Q_TEMP_JOBS_SIZE;
		ctx->nb_job_in_batch++;

		ctx->nb_successive_job_total++;
	} else {
		is_large_size_job = ((user_job->sha_job.base.type < DOCA_SHA_JOB_SHA1_PARTIAL) &&
				     (req_len > MAX_BYTE_COUNT_PER_WQE));
		is_large_ll_job = ((user_job->sha_job.base.type < DOCA_SHA_JOB_SHA1_PARTIAL) &&
				   (nb_ll_buf > MAX_LL_DOCA_BUF_NUM));
		if ((is_large_size_job == true) || (is_large_ll_job == true)) {
			/*
			 * If sha_partial is not enabled,
			 * but the request byte count > 128Mbyte, or the linked_list buf element count > 128,
			 * we also need a session.
			 */
			session = get_sha_partial_session(ctx, SESSION_ALLOC_STATUS_INTERNAL);
			if (session == NULL)
				return DOCA_ERROR_NO_MEMORY;
		}

		/*
		 * Try to get a job_state from state_pool
		 */
		state = (struct doca_sha_partial_job *)get_buf_from_pool(ctx->state_pool);
		if (state == NULL) {
			/*
			 * If the session is grabbed internally for a large-size job or a large-linked-list job,
			 * we need release it here.
			 */
			if ((session != NULL) &&
			    ((is_large_size_job == true) || (is_large_ll_job == true)))
				put_sha_partial_session(ctx, session, SESSION_ALLOC_STATUS_INTERNAL);
			return DOCA_ERROR_NO_MEMORY;
		}
		state->session = NULL;

		if (user_job->sha_job.base.type >= DOCA_SHA_JOB_SHA1_PARTIAL) {
			/*
			 * Copy user_job into the job_state
			 */
			memcpy(state, user_job, sizeof(struct doca_sha_partial_job));
			if (((uint32_t *)(user_job->session->meta_buf))[METADATA_BUF_GATHERED_LENGTH_UINT32_OFFSET] != 0)
				user_job->session->is_first_segment = false;
			user_job->session->sent_bytes = 0;
			user_job->session->sent_nb_ll_buf = 0;
			user_job->session->recv_bytes = 0;
		} else {
			/*
			 * Copy user_job into the job_state
			 */
			memcpy(state, user_job, sizeof(struct doca_sha_job));
		}
		if ((is_large_size_job == true) || (is_large_ll_job == true))
			state->session = session;

		state->session->req_len = req_len;
		state->session->nb_ll_buf = nb_ll_buf;
		state->session->start_buf = state->sha_job.req_buf;
		state->session->start_buf_offset = 0;

		/*
		 * Put this job into sha_job_state buffer
		 */
		ctx->states[ctx->state_in_batch_front] = state;
		ctx->state_in_batch_front = (ctx->state_in_batch_front + 1) % PER_Q_TEMP_STATES_SIZE;
		ctx->nb_state_in_batch++;

		/*
		 * Until, user_job submission completed.
		 * Start the process of converting user_job into wqe_jobs.
		 */
		ret = sha_job_state_submit(ctx);
		if (ret != DOCA_SUCCESS)
			return ret;
	}

	/*
	 * This is the wqe_job enqueue.
	 */
	const bool use_burst = !(ctx->event_driven_mode);

	if (sha_dispatch_job(ctx, use_burst) < 0)
		return DOCA_ERROR_BAD_STATE;

	return DOCA_SUCCESS;
}

doca_error_t
bf2_sha_workq_progress(void *workq_ctx, struct doca_event *ev)
{
	struct bf2_sha_workq_ctx *ctx = (struct bf2_sha_workq_ctx *)workq_ctx;
	doca_error_t ret = DOCA_ERROR_AGAIN;
	int nb_resp = 0;
	struct wqe_job *job = NULL;
	struct partial_job *partial = NULL;
	struct doca_sha_job *whole = NULL;
	struct doca_sha_partial_session *session = NULL;

	/*
	 * always try fetch some responses.
	 */
	if (ctx->nb_job_sent_in_batch > 0) {
		if (ctx->sha_impl == NULL)
			return DOCA_ERROR_INVALID_VALUE;
		if (ctx->finished_jobs == NULL)
			return DOCA_ERROR_INVALID_VALUE;
		nb_resp = ctx->sha_impl->dequeue_fn(ctx->sha_impl, 0, ctx->finished_jobs,
						    ctx->finished_job_front,
						    PER_Q_TEMP_FINISHED_JOBS_SIZE - ctx->nb_job_finished);
		if (nb_resp < 0)
			return DOCA_ERROR_IO_FAILED;
		else if (nb_resp > 0) {
			ctx->nb_job_finished += nb_resp;
			ctx->nb_job_sent_in_batch -= nb_resp;
			ctx->finished_job_front = (ctx->finished_job_front + nb_resp) % PER_Q_TEMP_FINISHED_JOBS_SIZE;
			ret = DOCA_SUCCESS;
		}
	}

	/*
	 * ASAP return the first finished job
	 */
	ev->type = DOCA_ACTION_NONE;
	while (ctx->nb_job_finished > 0) {
		job = ctx->finished_jobs[ctx->finished_job_tail];

		if (job == NULL) {
			DOCA_DLOG_ERR("dequeue return a NULL wqe_job");
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (job->is_single_wqe_job == true) {
			whole = &(job->data.whole.state);
			ev->type = whole->base.type;
			ev->user_data = whole->base.user_data;
			ev->result.u64 = DOCA_SUCCESS;
			ret = DOCA_SUCCESS;
		} else {
			partial = &(job->data.partial);
			session = partial->state->session;
			session->recv_bytes += partial->src_len;
			session->processed_bytes += partial->src_len;
			if (session->recv_bytes == session->sent_bytes)
				session->wait_resp = false;
			/*
			 * session->recv_bytes represents the accumulated processed byte of a segment.
			 * session->processed_bytes represents the accumulated processed byte of a whole user job.
			 */
			if (((session->alloc_status == SESSION_ALLOC_STATUS_EXTERNAL) &&
			     (session->recv_bytes == session->req_len)) ||
			    ((session->alloc_status == SESSION_ALLOC_STATUS_INTERNAL) &&
			     (session->processed_bytes == session->req_len))) {
				ev->type = partial->state->sha_job.base.type;
				ev->user_data = partial->state->sha_job.base.user_data;
				ev->result.u64 = DOCA_SUCCESS;

				if ((session->alloc_status == SESSION_ALLOC_STATUS_INTERNAL) &&
				    (partial->is_last_wqe == true))
					put_sha_partial_session(ctx, partial->state->session,
						SESSION_ALLOC_STATUS_INTERNAL);

				put_buf_to_pool(ctx->state_pool, (void *)(partial->state));
				ret = DOCA_SUCCESS;
			} else
				ret = DOCA_ERROR_AGAIN;
		}
		put_buf_to_pool(ctx->job_pool, (void *)(job));
		ctx->nb_job_finished--;
		ctx->finished_job_tail = (ctx->finished_job_tail + 1) % PER_Q_TEMP_FINISHED_JOBS_SIZE;

		if (ret == DOCA_SUCCESS)
			break;
	}

	/*
	 * No response, No finished job, try to send jobs
	 */
	if (ctx->nb_successive_job_total_prev == ctx->nb_successive_job_total) {
		if (sha_job_state_submit(ctx) != DOCA_SUCCESS)
			return DOCA_ERROR_IO_FAILED;
		if (sha_dispatch_job(ctx, false) < 0)
			return DOCA_ERROR_IO_FAILED;
		ctx->nb_successive_job_total = 0;
	}
	ctx->nb_successive_job_total_prev = ctx->nb_successive_job_total;

	return ret;
}

doca_error_t
bf2_sha_workq_create(void **workq_ctx, void **cq, struct doca_dev *dev)
{
	int ret;
	struct bf2_sha_workq_ctx *ctx;

	if (workq_ctx == NULL || cq == NULL || dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	ctx = priv_doca_malloc(1, sizeof(struct bf2_sha_workq_ctx));
	if (ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	if (sha_buffer_pool_create(&(ctx->session_pool), sizeof(struct doca_sha_partial_session),
				   PER_Q_TEMP_SESSIONS_SIZE, 0) != DOCA_SUCCESS) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_session_pool;
	}

	if (sha_buffer_pool_create(&(ctx->state_pool), sizeof(struct doca_sha_partial_job),
				   PER_Q_TEMP_STATES_SIZE, 0) != DOCA_SUCCESS) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_state_pool;
	}

	if (sha_buffer_pool_create(&(ctx->job_pool), sizeof(struct wqe_job),
				   PER_Q_TEMP_JOBS_SIZE, 0) != DOCA_SUCCESS) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_job_pool;
	}

	ctx->states = priv_doca_malloc(PER_Q_TEMP_STATES_SIZE, sizeof(struct doca_sha_partial_job *));
	if (ctx->states == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_states;
	}

	ctx->jobs = priv_doca_malloc(PER_Q_TEMP_JOBS_SIZE, sizeof(struct wqe_job *));
	if (ctx->jobs == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_jobs;
	}

	ctx->finished_jobs = priv_doca_malloc(PER_Q_TEMP_FINISHED_JOBS_SIZE, sizeof(struct wqe_job *));
	if (ctx->finished_jobs == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_finished_jobs;
	}

	ret = create_sha_impl(&(ctx->sha_impl));
	if (ret != DOCA_SUCCESS)
		goto err_create_impl;

	if (ctx->sha_impl != NULL) {
		if (ctx->sha_impl->cookie != NULL) {
			ret = ctx->sha_impl->init_fn(ctx->sha_impl, dev);
			if (ret != DOCA_SUCCESS) {
				ctx->sha_impl->cleanup_fn(ctx->sha_impl);
				ctx->sha_impl->destroy_fn(ctx->sha_impl);
				goto err_create_impl;
			}
		}
	}

	*workq_ctx = ctx;

	*cq = ((struct bf2_sha_priv *)(ctx->sha_impl->cookie))->qps[0].cq;

	return DOCA_SUCCESS;

err_create_impl:
	if (ctx->sha_impl != NULL) {
		if (ctx->sha_impl->cookie != NULL)
			priv_doca_free(ctx->sha_impl->cookie);
		priv_doca_free(ctx->sha_impl);
	}
	priv_doca_free(ctx->finished_jobs);
err_create_finished_jobs:
	priv_doca_free(ctx->jobs);
err_create_jobs:
	priv_doca_free(ctx->states);
err_create_states:
	sha_buffer_pool_destroy(ctx->job_pool);
err_create_job_pool:
	sha_buffer_pool_destroy(ctx->state_pool);
err_create_state_pool:
	sha_buffer_pool_destroy(ctx->session_pool);
err_create_session_pool:
	priv_doca_free(ctx);
	*workq_ctx = NULL;
	return ret;
}

void
bf2_sha_workq_ctx_destroy(void *workq_ctx)
{
	struct bf2_sha_workq_ctx *ctx = (struct bf2_sha_workq_ctx *)workq_ctx;

	if (ctx == NULL)
		return;
	if (ctx->jobs != NULL)
		priv_doca_free(ctx->jobs);
	if (ctx->finished_jobs != NULL)
		priv_doca_free(ctx->finished_jobs);
	if (ctx->sha_impl != NULL) {
		if (ctx->sha_impl->cookie != NULL) {
			ctx->sha_impl->cleanup_fn(ctx->sha_impl);
			ctx->sha_impl->destroy_fn(ctx->sha_impl);
		}
	}
	if (ctx->states != NULL)
		priv_doca_free(ctx->states);
	if (ctx->job_pool != NULL)
		sha_buffer_pool_destroy(ctx->job_pool);
	if (ctx->state_pool != NULL)
		sha_buffer_pool_destroy(ctx->state_pool);
	if (ctx->session_pool != NULL)
		sha_buffer_pool_destroy(ctx->session_pool);
	priv_doca_free(ctx);
	ctx = NULL;
}

doca_error_t
bf2_sha_set_event_driven_mode(void *workq_ctx, bool value)
{
	((struct bf2_sha_workq_ctx *)workq_ctx)->event_driven_mode = value;
	return DOCA_SUCCESS;
}

/**
 * The workq instance for bf2 SHA engine.
 */
static struct sha_workq_impl bf2_workq_impl = {
	.workq_ctx = NULL,
	.job_submit_fn = bf2_sha_workq_job_submit,
	.progress_retrieve_fn = bf2_sha_workq_progress,
	.workq_create_fn = bf2_sha_workq_create,
	.workq_destroy_fn = bf2_sha_workq_ctx_destroy,
	.sha_partial_session_create_fn = bf2_sha_partial_session_create,
	.sha_partial_session_destroy_fn = bf2_sha_partial_session_destroy,
	.sha_partial_session_copy_fn = bf2_sha_partial_session_copy,
	.sha_set_event_driven_mode_fn = bf2_sha_set_event_driven_mode
};

#ifdef SUPPORT_OPENSSL_CRYPTO
/**
 * The workq instance for openssl software SHA.
 */
static struct sha_workq_impl ossl_workq_impl = {
	.workq_ctx = NULL,
	.job_submit_fn = ossl_sha_workq_job_submit,
	.progress_retrieve_fn = ossl_sha_workq_progress,
	.workq_create_fn = ossl_sha_workq_create,
	.workq_destroy_fn = ossl_sha_workq_ctx_destroy,
	.sha_partial_session_create_fn = ossl_sha_partial_session_create,
	.sha_partial_session_destroy_fn = ossl_sha_partial_session_destroy,
	.sha_partial_session_copy_fn = ossl_sha_partial_session_copy,
	.sha_set_event_driven_mode_fn = ossl_sha_set_event_driven_mode
};
#endif /* SUPPORT_OPENSSL_CRYPTO */

doca_error_t
sha_workq_factory_partial_init_workq(bool use_sw_fallback, struct sha_workq_impl *workq_impl)
{
	if (workq_impl == NULL)
		return DOCA_ERROR_NO_MEMORY;

	/*
	 * Accoridng to the value of use_sw_fallback, to select the suitable workq instance type.
	 */
	if (use_sw_fallback == false)
		memcpy(workq_impl, &bf2_workq_impl, sizeof(*workq_impl));
	else {
#ifdef SUPPORT_OPENSSL_CRYPTO
		memcpy(workq_impl, &ossl_workq_impl, sizeof(*workq_impl));
#else /* SUPPORT_OPENSSL_CRYPTO */
		DOCA_DLOG_ERROR("There is no openssl library found, not possible to use software fallback");
		return DOCA_ERROR_INVALID_VALUE;
#endif /* SUPPORT_OPENSSL_CRYPTO */
	}

	return DOCA_SUCCESS;
}

