/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * Definition of an abstract implementation to be used by doca sha, Provides a separation between
 * implementation of sha routines for the higher level functionality of the doca sha api.
 */

#ifndef DOCA_SHA_WORKQ_CTX_H_
#define DOCA_SHA_WORKQ_CTX_H_

#include <stdint.h>
#include <doca_dev.h>
#include <doca_ctx.h>
#include "doca_sha.h"
#include "sha_buffer_pool.h"
#ifdef SUPPORT_OPENSSL_CRYPTO
#include <openssl/evp.h>
#include <openssl/sha.h>
#include <pthread.h>
#endif /* SUPPORT_OPENSSL_CRYPTO */

#ifdef __cplusplus
extern "C" {
#endif

#define DOCA_SHA_MAX_BURST_COUNT (16)

#define PER_Q_TEMP_JOBS_SIZE (1024)

#define PER_Q_TEMP_STATES_SIZE (PER_Q_TEMP_JOBS_SIZE)

#define PER_Q_TEMP_SESSIONS_SIZE (PER_Q_TEMP_JOBS_SIZE)

#define PER_Q_TEMP_FINISHED_JOBS_SIZE (PER_Q_TEMP_JOBS_SIZE)

#define DOCA_SHA_MAX_JOB_BYTE_COUNT (2UL * 1024 * 1024 * 1024) /**< MAX request byte count per SHA job, 2Gbyte. */

#define MAX_BYTE_COUNT_PER_WQE (128 * 1024 * 1024) /**< MAX request byte count per SHA wqe, 128Mbyte. */

#define METADATA_BUF_GATHERED_LENGTH_UINT32_OFFSET 3

#define METADATA_BUF_SHA_PARTIAL_RESULT_UINT8_OFFSET 64

#define METADATA_BUF_SHA_PARTIAL_RESULT_BYTE_COUNT 64

#define MAX_LL_DOCA_BUF_NUM 64

/**
 * sha_partial session allocated status.
 * If a session is acquired by user call session_create(),
 * it can be freed by user call session_destroy(),
 * and it can also be automatically freed interanlly at the end of a job.
 * But if a session is interally created, it must be freed internally.
 */
enum sha_partial_session_alloc_status {
	SESSION_ALLOC_STATUS_FREE = 0,	/**< This session is free to use for both external or internal. */
	SESSION_ALLOC_STATUS_INTERNAL,	/**< This session is internally allocated. */
	SESSION_ALLOC_STATUS_EXTERNAL	/**< This session is allocated by user call session_create(). */
};

/**
 * Sha_partial session defintion.
 *
 * The doca_sha_partial_job represents a segment of a whole job for the stateful sha calculation.
 *
 * The session is used to for keeping state for stateful sha.
 * It is also used for keeping information for large-size (>128Mbyte) job/segment.
 * Because it is very possible that a single enqueue cannot enqueue all needed wqes.
 * In order to get correct result, the wqes of the same job must be sequentially processed
 *     1, if the wqes are enqueued together, the fence-mode need to be set properly
 *        1.1, the wqes enqueued together must be put in the same hardware qp-obj, and set the fence-mode.
 *     2, if a wqe is not enqueued, it needs to wait until all previous enqueued wqes finished
 */
struct doca_sha_partial_session {
	uint8_t *meta_buf;			/**< state address for sha_partial.
						  *  For sha_partial job, its value is acquired by user calling
						  *  doca_sha_partial_session_create().
						  *  And assigned by user in user_job construction.
						  *  But for a large-size (>128Mbyte) NORMAL SHA job, we cannot get its
						  *  value form user_job, so we find its value internally to help us
						  *  to do an internally job-splitting.
						  */
	struct doca_buf *start_buf;		/**< The first doca_buf to be processed.
						  *   For non-linked-list job, we can easily get this value.
						  *   For linked-list job, we need record it.
						  */
	uint32_t req_len;			/**< The total bytes of the request job starting from start_buf.
						  *  For non-linked-list job, we can easily get this value.
						  *  For linked-list job, we need iterate all linked-list elements to
						  *  get this value, so we need record it here.
						  */
	uint32_t nb_ll_buf;			/**< The number of linked_list elements starting from start_buf.
						  *  For non-linked-list job, we can easily get this value.
						  *  For linked-list job, we need record it.
						  */
	uint32_t start_buf_offset;		/**< The byte address to be processed in the start_buf.
						  *  Because we need to do the job data split and concatenation,
						  *  so it is necessary to record it.
						  */
	uint32_t sent_nb_ll_buf;		/**< Accumulated processed linked-list element count. */
	uint32_t processed_bytes;		/**< Accumulated processed byte count for this session.
						  *  Usd to verify gathered_length in meta_buf is correct.
						  *  So we know whether this meta_buf is corrupted.
						  */
	uint32_t sent_bytes;			/**< Total sent data bytes to now for this segment. */
	uint32_t recv_bytes;			/**< Total recvd data bytes to now for this segment. */
	uint32_t alloc_status;			/**< session allocation status.
						  *  It is used to avoid mis-free session by user.
						  */
	bool is_first_segment;			/**< whether it is the first segment of a whole job. */
	bool wait_resp;				/**< Indicate whether it is waiting response. */
};

/**
 * Workq implementation interface: context and operations.
 * workq_ctx: the real implementation context, bf2 or openssl.
 * function_ptrs: all the virtual function needed.
 */
struct sha_workq_impl {
	/**
	 * workq context depending on implememntation.
	 */
	void *workq_ctx;

	/**
	 * complete queue for event_driven mode.
	 */
	void *cq;

	/**
	 * @brief
	 * Submit a sha job
	 *
	 * @param [in] workq_ctx
	 * Pointer to sha workq.
	 *
	 * @param [in] job
	 * Job as passed to: doca_workq_submit.
	 *
	 * @return
	 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
	 */
	doca_error_t (*job_submit_fn)(void *workq_ctx, struct doca_sha_partial_job const *job);

	/**
	 * @brief
	 * Retrieve a sha event (response)
	 *
	 * @param [in] workq_ctx
	 * Pointer to sha workq.
	 *
	 * @param [in] ev
	 * The doca_event used to retrieve response.
	 *
	 * @return
	 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
	 */
	doca_error_t (*progress_retrieve_fn)(void *workq_ctx, struct doca_event *ev);

	/**
	 * @brief
	 * Initialise the sha workq.
	 *
	 * @param [in and out] workq_ctx
	 * The workq to be initialized.
	 *
	 * @param [in and out ] cq
	 * Pointer to complete queue to be used in event_driven mode.
	 *
	 * @param [in] dev
	 * The device used by the workq.
	 *
	 * @return
	 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
	 */
	doca_error_t (*workq_create_fn)(void **workq_ctx, void **cq, struct doca_dev *dev);

	/**
	 * @brief
	 * Destroy work queue context and all resources owned by it.
	 *
	 * @param [in] workq_ctx
	 * Pointer to workq to be destroyed.
	 */
	void (*workq_destroy_fn)(void *workq_ctx);

	/**
	 * @brief
	 * Get a new sha_partial session object.
	 *
	 * @param [in] workq_ctx
	 * Workq instance.
	 *
	 * @param [out] session
	 * The created sha_partial_session object.
	 *
	 * @return
	 * DOCA_SUCCESS - in case of success.
	 * Error code - in case of failure:
	 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
	 */
	doca_error_t (*sha_partial_session_create_fn)(void *workq_ctx,
		struct doca_sha_partial_session **session);

	/**
	 * @brief
	 * Put back a sha_partial session object.
	 *
	 * @param [in] workq_ctx
	 * Workq instance.
	 * @param [in] session
	 * The sha_partial_session object to be freed.
	 *
	 * @return
	 * DOCA_SUCCESS - in case of success.
	 * Error code - in case of failure:
	 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
	 */
	doca_error_t (*sha_partial_session_destroy_fn)(void *workq_ctx,
		struct doca_sha_partial_session *session);

	/**
	 * @brief
	 * Copy the sha_partial session state from "from_session" to "to_session".
	 * we do not copy all the data, only the useful elements are copied.
	 *
	 * @param [in] workq_ctx
	 * Workq instance.
	 * @param [in] from_session
	 * The source sha_partial_session object to be copied.
	 *
	 * @param [out] to_session
	 * The dest sha_partial_session object.
	 *
	 * @return
	 * DOCA_SUCCESS - in case of success.
	 * Error code - in case of failure:
	 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
	 */
	doca_error_t (*sha_partial_session_copy_fn)(void *workq_ctx,
		const struct doca_sha_partial_session *from_session,
		struct doca_sha_partial_session *to_session);

	/**
	 * @brief
	 * Set the workq event_driven_mode.
	 *
	 * @param [in] workq_ctx
	 * Workq instance.
	 * @param [in] value
	 * The value of event_driven_mode.
	 *
	 * @return
	 * DOCA_SUCCESS - in case of success.
	 * Error code - in case of failure:
	 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
	 */
	doca_error_t (*sha_set_event_driven_mode_fn)(void *workq_ctx, bool value);
};

/**
 * Polled workq ctx definition.
 */
struct sha_polled_workq_ctx {
	struct sha_workq_impl base;	/**< Actually sha workq implementation. */
};

/**
 * Event_driven workq ctx definition.
 */
struct sha_event_driven_workq_ctx {
	struct sha_workq_impl base;			/**< Actually sha workq implementation. */
	struct priv_doca_comp_events_channel *channel;  /**< Event Channel. */
	doca_event_handle_t channel_handle;		/**< Event Channel handle. */
	struct doca_workq *workq;			/**< Workq. */
	struct priv_doca_workq_progress_entry *poller;	/**< Workq poller. */
};

/**
 * The is the actual partial job struct (of a multi-segment job) used to do the enqueue.
 *
 * When it is non-linked-list job:
 * for all the wqes of the same job/segment, the only difference is the input data info:
 * {src_addr, src_len, is_first_wqe, is_last_wqe, fence_mode}.
 *
 * When it is linked-list job:
 * for all the wqes of the same job/segment, the difference of the input data info:
 * besides the above difference, still need to make account of
 * {start_doca_buf of the src_addr, last_doca_buf, last_doca_buf_offset, nb_linked_list_element}.
 *
 * The output {resp_buf, meta_buf} address always keeps the same.
 *
 * Here, I call user_job as a state.
 */
struct partial_job {
	struct doca_sha_partial_job *state;	     /**< The user_job used to generate this job. */
	struct doca_buf *start_buf;		     /**< The first doca_buf of this wqe job. */
	struct doca_buf *last_buf;		     /**< The last doca_buf of this wqe job. */
	uint32_t start_buf_offset;		     /**< The address offset of the start_buf,
						       *  so start_buf_src_addr = start_buf_addr + start_buf_offset.
						       *  so start_buf_src_len = start_buf_len - start_buf_offset.
						       */
	uint32_t last_buf_offset;		     /**< The address offset of the last_buf,
						       *  so last_buf_src_addr = last_buf_addr.
						       *  so last_buf_src_len = last_buf_offset.
						       */
	uint32_t nb_ll_buf;			     /**< The total linked-list element count for this wqe job.
						       *  including the start_buf and the last_buf.
						       */
	uint32_t src_len;			     /**< The total source data length for this wqe job. */
	bool is_first_wqe;			     /**< Whether it is the first wqe of a whole job.
						       *  For the first wqe, the meta_buf is not needed.
						       *  For non-first wqes, the meta_buf is needed.
						       */
	bool is_last_wqe;			     /**< Whether it is the last wqe of a whole job.
						       *  For the last wqe, padding must be enabled.
						       *  For non-last wqe, padding should be disabled.
						       */
	bool fence_mode;			     /**< Whether we should enable fence-mode for this wqe.
						       *  Fence mode greately degrades the performance. But we need it.
						       *  Fence-mode used to keep the wqes sequentially executed.
						       *  For wqes of the same job, if they are enqueued together,
						       *  it is needed; if enqueued not together, not needed.
						       *  For wqes of different jobs, of course, not needed.
						       */
};

/**
 * The is the actual single wqe job struct (of a linked_list job) used to do the enqueue.
 */
struct whole_job {
	struct doca_sha_job state;		     /**< The user_job used to generate this job. */
	uint32_t nb_ll_buf;			     /**< The total linked-list element count for this wqe job. */
	uint32_t src_len;			     /**< The total source data length for this wqe job. */
};

/**
 * wqe_job data union.
 */
union wqe_job_data {
	struct partial_job partial;	  /**< Partial job data. */
	struct whole_job whole;		  /**< Whole job data. */
};

/**
 * The is the actual job struct used to do the enqueue.
 * A wqe_job represents a wqe.
 * A wqe_job has two options: (1) single_wqe job; (2) a partial job of a multi-segment job.
 */
struct wqe_job {
	bool is_single_wqe_job;  /**< Indicate whether this wqe represents a job with the two conditions:
				   *   data_len <= 128Mbyte, and linked_list_element_count <= 128.
				   */
	union wqe_job_data data; /**< The true wqe job data. */
};

/**
 * Structure containing all the elements required by the DOCA bf2 sha workq.
 * Here, I call user_job as a state.
 */
struct bf2_sha_workq_ctx {
	struct sha_buffer_pool *session_pool;		 /**< The pre-allocated sha_partial session pool. */
	struct sha_buffer_pool *state_pool;		 /**< The pre-allocated user_job pool. */
	struct sha_buffer_pool *job_pool;		 /**< The pre-allocated wqe_job pool. */

	struct doca_sha_partial_job **states;		 /**< The state batch used as state pointers storage. */
	uint16_t nb_state_in_batch;			 /**< The number of states in the state_batch to be sent. */
	uint16_t state_in_batch_front;			 /**< The state_batch front index. */
	uint16_t state_in_batch_tail;			 /**< The state_batch tail index. */

	struct wqe_job **jobs;				 /**< The user job_batch used as job pointers storage. */
	uint16_t nb_job_in_batch;			 /**< The number of jobs in the job_batch to be sent. */
	uint16_t job_in_batch_front;			 /**< The job_batch front index. */
	uint16_t job_in_batch_tail;			 /**< The job_batch tail index. */
	uint16_t nb_job_sent_in_batch;			 /**< The number of job already sent, but no response yet. */

	struct wqe_job **finished_jobs;			 /**< The user finished job pointers storage. */
	uint16_t nb_job_finished;			 /**< The number of jobs finished. */
	uint16_t finished_job_front;			 /**< The finished_jobs front index. */
	uint16_t finished_job_tail;			 /**< The finished_jobs tail index. */

	uint64_t nb_successive_job_total;		 /**< Special flag to indicate whether to finish burst. */
	uint64_t nb_successive_job_total_prev;		 /**< Special flag to indicate whether to finish burst. */

	struct sha_impl *sha_impl;			 /**< Pointers of SHA engine hardware driver. */

	bool event_driven_mode;				 /**< Indicate the working mode of this workq. */
};

/**
 * Submit a sha job
 *
 * @param [in] workq_ctx
 * Pointer to sha workq.
 *
 * @param [in] job
 * Job as passed to: doca_workq_submit.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
doca_error_t bf2_sha_workq_job_submit(void *ctx, struct doca_sha_partial_job const *job);

/**
 * Retrieve a sha event (response)
 *
 * @param [in] workq_ctx
 * Pointer to sha workq.
 *
 * @param [in] ev
 * The doca_event used to retrieve response.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
doca_error_t bf2_sha_workq_progress(void *ctx, struct doca_event *ev);

/**
 * Initialise the sha workq.
 *
 * @param [in and out] workq_ctx
 * The workq to be initialized.
 *
 * @param [in] dev
 * The device used by the workq.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
doca_error_t bf2_sha_workq_create(void **ctx, void **cq, struct doca_dev *dev);

/**
 * Destroy work queue context and all resources owned by it.
 *
 * @param [in] workq_ctx
 * Pointer to workq to be destroyed.
 */
void bf2_sha_workq_ctx_destroy(void *ctx);

/**
 * Get a new sha_partial session object.
 *
 * @param [in] workq_ctx
 * Workq instance.
 *
 * @param [out] session
 * The created sha_partial_session object.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t bf2_sha_partial_session_create(void *ctx, struct doca_sha_partial_session **session);

/**
 * Put back a sha_partial session object.
 *
 * @param [in] workq_ctx
 * Workq instance.
 * @param [in] session
 * The sha_partial_session object to be freed.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t bf2_sha_partial_session_destroy(void *ctx, struct doca_sha_partial_session *session);

/**
 * Copy the sha_partial session state from "from_session" to "to_session".
 * we do not copy all the data, only the useful elements are copied.
 *
 * @param [in] workq_ctx
 * Workq instance.
 * @param [in] from_session
 * The source sha_partial_session object to be copied.
 *
 * @param [out] to_session
 * The dest sha_partial_session object.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t bf2_sha_partial_session_copy(void *ctx,
	const struct doca_sha_partial_session *from_session,
	struct doca_sha_partial_session *to_session);

/**
 * @brief
 * Set the workq event_driven_mode.
 *
 * @param [in] workq_ctx
 * Workq instance.
 * @param [in] value
 * The value of event_driven_mode.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t bf2_sha_set_event_driven_mode(void *workq_ctx, bool value);

#ifdef SUPPORT_OPENSSL_CRYPTO
/**
 * Structure containing all the elements required by the DOCA sha openssl workq.
 *
 * Although the openssl sha is synchronous, the input job still need be backup-ed, because job input and output
 * functions for doca_sha is different: job_submit() and progress_retrieve().
 *
 * Because openssl sha is synchronous, there are two choices to call openssl_sha(), either in job_submit() or in
 * progress_retrieve().
 * In this implementation, in order not to save addition variables, the openssl_sha calculation is done immediately
 * in the job_submit() after the job is received.
 *
 * The response event is reported in progress_retrieve() by accessing the elements in finished_jobs.
 *
 * Because the finished_jobs are shared by two different functions, a mutex/lock is needed to make sure thread-safe.
 */
struct ossl_sha_workq_ctx {
	struct sha_buffer_pool *job_pool;		 /**< The pre-allocated user_job pool. */

	pthread_mutex_t lock;				 /**< The lock for accessing finished_job array. */

	struct doca_sha_partial_job **finished_jobs;	 /**< The user finished job pointers storage. */
	uint16_t nb_job_finished;			 /**< The number of jobs finished. */
	uint16_t finished_job_front;			 /**< The finished_jobs front index. */
	uint16_t finished_job_tail;			 /**< The finished_jobs tail index. */
};

/**
 * Submit a sha job
 *
 * @param [in] workq_ctx
 * Pointer to sha workq.
 *
 * @param [in] job
 * Job as passed to: doca_workq_submit.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
doca_error_t ossl_sha_workq_job_submit(void *ctx, struct doca_sha_partial_job const *job);

/**
 * Retrieve a sha event (response)
 *
 * @param [in] workq_ctx
 * Pointer to sha workq.
 *
 * @param [in] ev
 * The doca_event used to retrieve response.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
doca_error_t ossl_sha_workq_progress(void *ctx, struct doca_event *ev);

/**
 * Initialise the sha workq.
 *
 * @param [in and out] workq_ctx
 * The workq to be initialized.
 *
 * @param [in] dev
 * The device used by the workq.
 *
 * @return
 * DOCA_SUCCESS or an appropriate DOCA_ERROR_XXX value.
 */
doca_error_t ossl_sha_workq_create(void **ctx, void **cq, struct doca_dev *dev);

/**
 * Destroy work queue context and all resources owned by it.
 *
 * @param [in] workq_ctx
 * Pointer to workq to be destroyed.
 */
void ossl_sha_workq_ctx_destroy(void *ctx);

/**
 * Get a new sha_partial session object.
 *
 * @param [in] workq_ctx
 * Workq instance.
 *
 * @param [out] session
 * The created sha_partial_session object.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t ossl_sha_partial_session_create(void *ctx, struct doca_sha_partial_session **session);

/**
 * Put back a sha_partial session object.
 *
 * @param [in] workq_ctx
 * Workq instance.
 * @param [in] session
 * The sha_partial_session object to be freed.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t ossl_sha_partial_session_destroy(void *ctx, struct doca_sha_partial_session *session);

/**
 * Copy the sha_partial session state from "from_session" to "to_session".
 * we do not copy all the data, only the useful elements are copied.
 *
 * @param [in] workq_ctx
 * Workq instance.
 * @param [in] from_session
 * The source sha_partial_session object to be copied.
 *
 * @param [out] to_session
 * The dest sha_partial_session object.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t ossl_sha_partial_session_copy(void *ctx,
	const struct doca_sha_partial_session *from_session,
	struct doca_sha_partial_session *to_session);

/**
 * @brief
 * Set the workq event_driven_mode.
 *
 * @param [in] workq_ctx
 * Workq instance.
 * @param [in] value
 * The value of event_driven_mode.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
doca_error_t ossl_sha_set_event_driven_mode(void *workq_ctx, bool value);

#endif /* SUPPORT_OPENSSL_CRYPTO */

/**
 * Create the wokrq_implementation according to whether to use openssl software fallback.
 *
 * @param [in] workq_impl
 * Workq instance pointer.
 * @param [in] use_sw_fallback
 * Indicate whether to use software fallback.
 *
 * @param [out] workq_impl
 * Wokrq instance.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - malloc failure.
 */
doca_error_t sha_workq_factory_partial_init_workq(bool use_sw_fallback, struct sha_workq_impl *workq_impl);

/**
 * Utility function.
 * To get the SHA job type string description.
 */
static inline const char *
get_job_type_string(enum doca_sha_job_type job_type)
{
	switch (job_type) {
	case DOCA_SHA_JOB_SHA1:
		return "DOCA_SHA_JOB_SHA1";
	case DOCA_SHA_JOB_SHA1_PARTIAL:
		return "DOCA_SHA_JOB_SHA1_PARTIAL";
	case DOCA_SHA_JOB_SHA256:
		return "DOCA_SHA_JOB_SHA256";
	case DOCA_SHA_JOB_SHA256_PARTIAL:
		return "DOCA_SHA_JOB_SHA256_PARTIAL";
	case DOCA_SHA_JOB_SHA512:
		return "DOCA_SHA_JOB_SHA512";
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		return "DOCA_SHA_JOB_SHA512_PARTIAL";
	default:
		return "Unsupported_job_type";
	}
}

/**
 * Utility function.
 * To get the standard/minimum response byte count of a SHA job type.
 */
static inline uint8_t
get_response_byte_count(enum doca_sha_job_type job_type)
{
	switch (job_type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA1_PARTIAL:
		return DOCA_SHA1_BYTE_COUNT;
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
		return DOCA_SHA256_BYTE_COUNT;
	case DOCA_SHA_JOB_SHA512:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		return DOCA_SHA512_BYTE_COUNT;
	default:
		return 0;
	}
}

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_SHA_WORKQ_CTX_H_ */
