/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <private/common_defines.h>
#include <doca_log.h>
#include <priv_doca_buf.h>
#include <doca_buf.h>
#include <priv_doca_malloc.h>
#include <doca_sha.h>
#include "doca_sha_workq_ctx.h"
#include "sha_buffer_pool.h"

DOCA_LOG_REGISTER(OSSL_SHA_WORKQ);

/**
 * For openssl, there is no input restrictions, such as:
 *     job data size,
 *     linked_list element count,
 *     partial sha segment data size.
 * so do not check them for openssl.
 * but still need to check output size.
 */
static inline bool
is_job_data_valid(const struct ossl_sha_workq_ctx *ctx, const struct doca_sha_partial_job *job,
		size_t *req_len_ret, uint32_t *nb_ll_buf_ret)
{
	size_t req_len = 0;
	size_t resp_len = 0;
	uint32_t i = 0;
	uint32_t nb_ll_buf = 1;
	struct doca_buf *next_buf = NULL;
	struct doca_buf *buf = NULL;
	size_t len = 0;

	if ((ctx == NULL) || (job == NULL)) {
		DOCA_DLOG_ERR("ctx or job pointer is NULL");
		return false;
	}
	if (doca_buf_list_num_elements(job->sha_job.req_buf, &nb_ll_buf) != DOCA_SUCCESS)
		nb_ll_buf = 1;
	*nb_ll_buf_ret = nb_ll_buf;

	if (doca_buf_get_data_len(job->sha_job.req_buf, &req_len) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("job request buffer get length error");
		return false;
	}

	if (nb_ll_buf > 1) {
		buf = job->sha_job.req_buf;
		for (i = 1; i < nb_ll_buf; i++) {
			if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("job request buffer linked_list next_buf not set");
				return false;
			}
			if (doca_buf_get_data_len(next_buf, &len) != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("job request buffer linked list next_buf get length error");
				return false;
			}
			req_len += len;
			buf = next_buf;
		}
#ifdef USE_SHA_DEBUG_PRINT
		DOCA_DLOG_DBG("linked_list doca_buf total req_len is: %ld\n", req_len);
#endif /* USE_SHA_DEBUG_PRINT */
	}
	*req_len_ret = req_len;

	/*
	 * after doca-core is updated, i can use priv_doca_buf_get_data_len() to replace these 2 lines.
	 */
	resp_len = priv_doca_buf_get_len(job->sha_job.resp_buf);
	resp_len -= (size_t)priv_doca_buf_get_data(job->sha_job.resp_buf) -
		    (size_t)priv_doca_buf_get_addr(job->sha_job.resp_buf);
	if ((resp_len == 0) || (resp_len < get_response_byte_count(job->sha_job.base.type))) {
		DOCA_DLOG_ERR("For [%s], job response buffer length need >= %d",
			get_job_type_string(job->sha_job.base.type),
			get_response_byte_count(job->sha_job.base.type));
		return false;
	}
	switch (job->sha_job.base.type) {
	case DOCA_SHA_JOB_SHA1_PARTIAL:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		/*
		 * Do a job->session validity check.
		 */
		if (job->session == NULL) {
			DOCA_DLOG_ERR("sha_partial_session cannot be NULL for DOCA_SHA_JOB_SHA_PARTIAL");
			return false;
		}
	default:
		break;
	}
	return true;
}

doca_error_t
ossl_sha_partial_session_copy(void *ctx,
			const struct doca_sha_partial_session *from,
			struct doca_sha_partial_session *to)
{
	if ((ctx == NULL) || (from == NULL) || (to == NULL))
		return DOCA_ERROR_INVALID_VALUE;
	if (EVP_MD_CTX_copy_ex((EVP_MD_CTX *)to, (EVP_MD_CTX *)from) == 1)
		return DOCA_SUCCESS;
	return DOCA_ERROR_INVALID_VALUE;
}

doca_error_t
ossl_sha_partial_session_create(void *workq_ctx,
				struct doca_sha_partial_session **session)
{
	EVP_MD_CTX *mdctx = EVP_MD_CTX_create();
	*session = (struct doca_sha_partial_session *)mdctx;
	if (*session == NULL) {
		DOCA_DLOG_ERR("mdctx create failure");
		return DOCA_ERROR_NO_MEMORY;
	}
	return DOCA_SUCCESS;
}

doca_error_t
ossl_sha_partial_session_destroy(void *workq_ctx,
				struct doca_sha_partial_session *session)
{
	if (session == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	EVP_MD_CTX_destroy((EVP_MD_CTX *)session);
	return DOCA_SUCCESS;
}

static inline const EVP_MD *
get_md(enum doca_sha_job_type job_type)
{
	switch (job_type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA1_PARTIAL:
		return EVP_sha1();
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
		return EVP_sha256();
	case DOCA_SHA_JOB_SHA512:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		return EVP_sha512();
	default:
		DOCA_DLOG_ERR("unsupported sha job_type [%d]", job_type);
		return NULL;
	}
}

static inline doca_error_t
one_shot_sha(struct doca_buf *buf, EVP_MD_CTX *mdctx)
{
	unsigned char *msg = NULL;
	size_t msg_len = 0;

	if (doca_buf_get_data(buf, (void **)&msg) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("job request buffer get addr error");
		return DOCA_ERROR_BAD_STATE;
	}
	if (doca_buf_get_data_len(buf, &msg_len) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("job request buffer get length error");
		return DOCA_ERROR_BAD_STATE;
	}
	if (EVP_DigestUpdate(mdctx, msg, msg_len) == 0) {
		DOCA_DLOG_ERR("evp_digestupdate failure\n");
		return DOCA_ERROR_BAD_STATE;
	}
	return DOCA_SUCCESS;
}

static inline doca_error_t
sha_final(struct doca_buf *buf, EVP_MD_CTX *mdctx)
{
	unsigned char *digest = NULL;
	unsigned int digest_len = 0;

	if (doca_buf_get_data(buf, (void **)&digest) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("job response buffer get addr error");
		return DOCA_ERROR_BAD_STATE;
	}
	if (EVP_DigestFinal_ex(mdctx, digest, &digest_len) == 0) {
		DOCA_DLOG_ERR("evp_digestfinal failure\n");
		return DOCA_ERROR_BAD_STATE;
	}
	if (doca_buf_set_data(buf,
		priv_doca_buf_get_data(buf),
		digest_len) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("doca_buf_set_data() for response doca_buf failure");
		return DOCA_ERROR_BAD_STATE;
	}
	return DOCA_SUCCESS;
}

/**
 * Openssl sha calculation.
 *
 * @param [in] job->req_buf
 * the first request doca_buf.
 *
 * @param [in] req_len
 * the total request len whatever whether it is linked_list_buf.
 *
 * @param [in] nb_ll_buf
 * the total number of request doca_buf.
 *
 * @param [in] job->base.job_type
 * the 6 input job_type.
 *
 * @param [in] job->sha_partial_session
 * it is the EVP_MD_CTX, if the job_type is sha_partial, mdctx is useful; otherwise, we create a new one.
 *
 * @param [out] job->resp_buf
 * the resopnse doca_buf.
 *
 * @return
 * DOCA_SUCCESS or DOCA_ERROR_XXX
 */
static inline doca_error_t
openssl_sha(struct doca_sha_partial_job *job, const size_t req_len,
	    const uint32_t nb_ll_buf)
{
	struct doca_buf *buf;
	struct doca_buf *next_buf;
	const EVP_MD *evp_md;
	EVP_MD_CTX *mdctx = NULL;
	doca_error_t ret;
	uint32_t i = 0;

	if (job == NULL) {
		DOCA_DLOG_ERR("job is NULL\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (nb_ll_buf < 1) {
		DOCA_DLOG_ERR("request doca_buf number should >= 1\n");
		return DOCA_ERROR_INVALID_VALUE;
	}

	evp_md = get_md(job->sha_job.base.type);
	if (evp_md == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	switch (job->sha_job.base.type) {
	case DOCA_SHA_JOB_SHA1:
	case DOCA_SHA_JOB_SHA256:
	case DOCA_SHA_JOB_SHA512:
		goto openssl_sha_create_ctx;
	case DOCA_SHA_JOB_SHA1_PARTIAL:
	case DOCA_SHA_JOB_SHA256_PARTIAL:
	case DOCA_SHA_JOB_SHA512_PARTIAL:
		goto openssl_sha_check_ctx;
	default:
		DOCA_DLOG_ERR("unsupported sha job_type [%d]", job->sha_job.base.type);
		return DOCA_ERROR_INVALID_VALUE;
	}

openssl_sha_check_ctx:
	mdctx = (EVP_MD_CTX *)(job->session);
	if (mdctx == NULL) {
		DOCA_DLOG_ERR("session is NULL for sha_partial job");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (EVP_MD_CTX_md(mdctx) == NULL) {
		/* first time to use this mdxtx. */
		goto openssl_sha_init_ctx;
	} else {
		/* Not the first time, so use it. */
		goto openssl_sha_use_ctx;
	}
openssl_sha_create_ctx:
	/*
	 * This section is only useful for non_partial_sha job.
	 */
	mdctx = EVP_MD_CTX_create();
	if (mdctx == NULL) {
		DOCA_DLOG_ERR("mdctx create failure\n");
		return DOCA_ERROR_NO_MEMORY;
	}
openssl_sha_init_ctx:
	if (EVP_DigestInit_ex(mdctx, evp_md, NULL) == 0) {
		DOCA_DLOG_ERR("evp_digestinit_ex failure\n");
		ret = DOCA_ERROR_NO_MEMORY;
		goto evp_digestinit_error;
	}

openssl_sha_use_ctx:
	/*
	 * This section is useful for both non_partial and partial job.
	 */
	buf = job->sha_job.req_buf;
	ret = one_shot_sha(buf, mdctx);
	if (ret != DOCA_SUCCESS)
		goto evp_digestupdate_error;
	for (i = 1; i < nb_ll_buf; i++) {
		if (doca_buf_list_next(buf, &next_buf) != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("job request buffer linked_list next_buf not set");
			ret = DOCA_ERROR_BAD_STATE;
			goto evp_digestupdate_error;
		}
		ret = one_shot_sha(next_buf, mdctx);
		if (ret != DOCA_SUCCESS)
			goto evp_digestupdate_error;
		buf = next_buf;
	}
	/*
	 * If this is not the final segment job of a whole partial_sha job, return immediately.
	 */
	if ((job->sha_job.base.type >= DOCA_SHA_JOB_SHA1_PARTIAL) &&
	    (job->sha_job.flags != DOCA_SHA_JOB_FLAGS_SHA_PARTIAL_FINAL))
		return DOCA_SUCCESS;

	/*
	 * This section is useful for non_partial job.
	 * And useful for partial_sha's final segment job.
	 */
	ret = sha_final(job->sha_job.resp_buf, mdctx);

evp_digestupdate_error:
evp_digestinit_error:
	if (job->sha_job.base.type < DOCA_SHA_JOB_SHA1_PARTIAL)
		EVP_MD_CTX_destroy(mdctx);
	return ret;
}


doca_error_t
ossl_sha_workq_job_submit(void *workq_ctx, struct doca_sha_partial_job const *user_job)
{
	struct ossl_sha_workq_ctx *ctx = (struct ossl_sha_workq_ctx *)workq_ctx;
	struct doca_sha_partial_job *job;
	doca_error_t ret;
	size_t req_len = 0;
	uint32_t nb_ll_buf = 0;

	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	if (ctx->finished_jobs == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	/*
	 * First to get the lock.
	 * Aftr lock is get, we can do process the shared resource.
	 */
	pthread_mutex_lock(&(ctx->lock));

	/*
	 * when the job batch is full, send first.
	 */
	if (ctx->nb_job_finished >= PER_Q_TEMP_FINISHED_JOBS_SIZE) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto job_submit_ret_after_unlock;
	}

	/*
	 * Check job data validity
	 * only pass the data integrity check, we can continue.
	 * And to avoid redundancry function call, during the check processing,
	 * we also extract this job's total_request_len and total_linked_list_element_count.
	 */
	if (!is_job_data_valid(ctx, user_job, &req_len, &nb_ll_buf)) {
		ret = DOCA_ERROR_INVALID_VALUE;
		goto job_submit_ret_after_unlock;
	}

	job = (struct doca_sha_partial_job *)get_buf_from_pool(ctx->job_pool);
	if (job == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto job_submit_ret_after_unlock;
	}
	/*
	 * Copy user_job into the job_pool.
	 */
	if (user_job->sha_job.base.type < DOCA_SHA_JOB_SHA1_PARTIAL)
		memcpy(job, user_job, sizeof(struct doca_sha_job));
	else
		memcpy(job, user_job, sizeof(struct doca_sha_partial_job));

	/*
	 * Do openssl sha calculation immediately.
	 */
	ret = openssl_sha(job, req_len, nb_ll_buf);
	if (ret != DOCA_SUCCESS) {
		ret = DOCA_ERROR_BAD_STATE;
		goto job_submit_ret_after_unlock;
	}

	ctx->finished_jobs[ctx->finished_job_front] = job;
	ctx->finished_job_front = (ctx->finished_job_front + 1) % PER_Q_TEMP_FINISHED_JOBS_SIZE;
	ctx->nb_job_finished++;

job_submit_ret_after_unlock:
	/*
	 * Release the lock.
	 */
	pthread_mutex_unlock(&(ctx->lock));

	return ret;
}

doca_error_t
ossl_sha_workq_progress(void *workq_ctx, struct doca_event *ev)
{
	struct ossl_sha_workq_ctx *ctx = (struct ossl_sha_workq_ctx *)workq_ctx;
	doca_error_t ret = DOCA_ERROR_AGAIN;
	struct doca_sha_partial_job *job = NULL;

	/*
	 * First to get the lock.
	 */
	pthread_mutex_lock(&(ctx->lock));

	/*
	 * Aftr lock is get, we can do process.
	 */
	if (ctx->nb_job_finished == 0) {
		ret = DOCA_ERROR_AGAIN;
		goto progress_ret_after_unlock;
	}

	/*
	 * ASAP return the first finished job
	 */
	ev->type = DOCA_ACTION_NONE;
	while (ctx->nb_job_finished > 0) {
		job = ctx->finished_jobs[ctx->finished_job_tail];

		if (job == NULL) {
			DOCA_DLOG_ERR("openssl sha enqueue return a NULL job");
			ret = DOCA_ERROR_INVALID_VALUE;
			goto progress_ret_after_unlock;
		}

		ev->type = job->sha_job.base.type;
		ev->user_data = job->sha_job.base.user_data;
		ev->result.u64 = DOCA_SUCCESS;
		ret = DOCA_SUCCESS;

		put_buf_to_pool(ctx->job_pool, (void *)(job));
		ctx->nb_job_finished--;
		ctx->finished_job_tail = (ctx->finished_job_tail + 1) % PER_Q_TEMP_FINISHED_JOBS_SIZE;

		break;
	}

progress_ret_after_unlock:
	/*
	 * Release the lock.
	 */
	pthread_mutex_unlock(&(ctx->lock));

	return ret;
}

doca_error_t
ossl_sha_workq_create(void **workq_ctx, void **cq, struct doca_dev *dev)
{
	int ret;
	struct ossl_sha_workq_ctx *ctx;

	(void)dev;

	if (workq_ctx == NULL || cq == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	ctx = priv_doca_malloc(1, sizeof(struct ossl_sha_workq_ctx));
	if (ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	if (sha_buffer_pool_create(&(ctx->job_pool), sizeof(struct doca_sha_partial_job),
				   PER_Q_TEMP_FINISHED_JOBS_SIZE, 0) != DOCA_SUCCESS) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_job_pool;
	}

	ctx->finished_jobs = priv_doca_malloc(PER_Q_TEMP_FINISHED_JOBS_SIZE, sizeof(struct doca_sha_partial_job *));
	if (ctx->finished_jobs == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_finished_jobs;
	}

	if (pthread_mutex_init(&(ctx->lock), NULL) != 0) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_create_lock;
	}

	*workq_ctx = ctx;
	*cq = NULL;

	return DOCA_SUCCESS;
err_create_lock:
	priv_doca_free(ctx->finished_jobs);
err_create_finished_jobs:
	sha_buffer_pool_destroy(ctx->job_pool);
err_create_job_pool:
	priv_doca_free(ctx);
	*workq_ctx = NULL;
	return ret;
}

void
ossl_sha_workq_ctx_destroy(void *workq_ctx)
{
	struct ossl_sha_workq_ctx *ctx = (struct ossl_sha_workq_ctx *)workq_ctx;

	if (ctx == NULL)
		return;
	pthread_mutex_destroy(&(ctx->lock));
	if (ctx->finished_jobs != NULL)
		priv_doca_free(ctx->finished_jobs);
	if (ctx->job_pool != NULL)
		sha_buffer_pool_destroy(ctx->job_pool);
	priv_doca_free(ctx);
	ctx = NULL;
}

doca_error_t
ossl_sha_set_event_driven_mode(void *workq_ctx, bool value)
{
	return DOCA_SUCCESS;
}

