/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "sha_buffer_pool.h"

DOCA_LOG_REGISTER(SHA_BUFFER_POOL);

doca_error_t
sha_buffer_pool_create(struct sha_buffer_pool **buf_pool, const uint32_t unit_buf_size, const uint32_t nb_buf,
			uint32_t align_size)
{
	struct sha_buffer_pool *pool = NULL;

	if (buf_pool == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	pool = calloc(1, sizeof(struct sha_buffer_pool));
	if (pool == NULL)
		return DOCA_ERROR_NO_MEMORY;

	pool->unit_buf_size = unit_buf_size;
	pool->nb_buf = nb_buf;

	const size_t pool_size = pool->unit_buf_size * pool->nb_buf;

	if (align_size > 0)
		pool->addr = aligned_alloc(align_size, pool_size);
	else
		pool->addr = calloc(pool->nb_buf, pool->unit_buf_size);
	if (pool->addr == NULL) {
		free(pool);
		return DOCA_ERROR_NO_MEMORY;
	}
	pool->id_alloc = id_allocator_create(pool->nb_buf);
	if (pool->id_alloc == NULL) {
		DOCA_DLOG_ERR("Failed to allocate id-alloc for buffer pool.");
		free(pool->addr);
		free(pool);
		return DOCA_ERROR_NO_MEMORY;
	}

	*buf_pool = pool;
	return DOCA_SUCCESS;
}

inline void*
get_buf_from_pool(struct sha_buffer_pool *pool)
{
	uint32_t id = 0;

	if (pool == NULL)
		return NULL;
	if (pool->id_alloc == NULL)
		return NULL;

	id = id_allocator_get(pool->id_alloc);

	if (id == ID_ALLOCATOR_INVALID_ID)
		return NULL;
	id--;
	return pool->addr + id * pool->unit_buf_size;
}

inline bool
is_buf_valid(const struct sha_buffer_pool *pool, const void *addr, uint32_t *id)
{
	if ((pool == NULL) || (addr == NULL) || (id == NULL))
		return false;
	if (pool->id_alloc == NULL)
		return false;

	if (addr < pool->addr) {
		DOCA_DLOG_ERR("buf address [%p] is less than pool address [%p]", addr, pool);
		return false;
	}

	if (((addr - pool->addr) % pool->unit_buf_size) != 0) {
		DOCA_DLOG_ERR("buf address [%p] is not aligned to pool unit_buf_size [%d]", addr, pool->unit_buf_size);
		return false;
	}

	*id = (addr - pool->addr) / pool->unit_buf_size;
	if (*id > pool->nb_buf) {
		DOCA_DLOG_ERR("buf address [%p] is greater than maximum pool buf address", addr);
		return false;
	}

	return true;
}

inline doca_error_t
put_buf_to_pool(struct sha_buffer_pool *pool, const void *addr)
{
	uint32_t id = 0;

	if (is_buf_valid(pool, addr, &id) == true) {
		id++;
		id_allocator_put(pool->id_alloc, id);
		return DOCA_SUCCESS;
	}
	return DOCA_ERROR_INVALID_VALUE;
}

void
sha_buffer_pool_destroy(struct sha_buffer_pool *pool)
{
	if (pool == NULL)
		return;

	if (pool->addr != NULL)
		free(pool->addr);

	if (pool->id_alloc != NULL)
		id_allocator_destroy(pool->id_alloc);

	free(pool);
}
