/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <doca_log.h>
#include <doca_version.h>

#include "doca_telemetry.h"
#include "telemetry_impl.h"


/* Default buffer size for the DOCA sources */
#define DOCA_TELEMETRY_DEFAULT_BUFFER_SIZE 60000
/* Default maximal file age */
#define DOCA_TELEMETRY_DEFAULT_FILE_AGE (60 * 60 * 1000000ULL) /* 1 Hour */
/* Default maximal file size */
#define DOCA_TELEMETRY_DEFAULT_FILE_SIZE (1 * 1024 * 1024) /* 1 MB */
/* Default data root folder */
#define DOCA_TELEMETRY_DEFAULT_DATA_ROOT "/opt/mellanox/doca/services/telemetry/data/"
/* Default config root folder */
#define DOCA_TELEMETRY_DEFAULT_CONFIG_ROOT "/opt/mellanox/doca/services/telemetry/config"
/* Default IPC reconnection time */
#define DOCA_TELEMETRY_DEFAULT_IPC_RECONNECT_TIME 100
/* Default number of IPC reconnection attempts */
#define DOCA_TELEMETRY_DEFAULT_IPC_RECONNECT_RETRIES 3
/* Default IPC socket timeout */
#define DOCA_TELEMETRY_DEFAULT_IPC_SOCKET_TIMEOUT 3000
/* Default IPC socket directory */
#define DOCA_TELEMETRY_DEFAULT_IPC_SOCKET_DIR "/opt/mellanox/doca/services/telemetry/ipc_sockets"

#define SCHEMA_DIR_POSTFIX ("/schema")
#define TYPE_INIT_CAP 64
#define MAX_TYPE_CAP (1024 * 1024) /* 1M fields */
#define OWNERSHIP_DOCA_TELEMETRY 1

DOCA_LOG_REGISTER(TELEMETRY);

/* ============================== Telemetry Source name ============================== */
static inline void _schema_destroy_attributes(struct doca_telemetry_schema *ds)
{
	if (ds->base_parameters != NULL) {
		free(ds->base_parameters->data_root);
		ds->base_parameters->data_root = NULL;

		free(ds->base_parameters->schema_path);
		ds->base_parameters->schema_path = NULL;

		free(ds->base_parameters->ipc_sockets_dir);
		ds->base_parameters->ipc_sockets_dir = NULL;
	}
}

void
doca_telemetry_source_set_id(struct doca_telemetry_source *doca_source, const char *source_id)
{
	if (doca_source == NULL)
		return;

	doca_source->attributes.source_id = source_id;
}

void
doca_telemetry_source_set_tag(struct doca_telemetry_source *doca_source, const char *source_tag)
{
	if (doca_source == NULL)
		return;

	doca_source->attributes.source_tag = source_tag;
}

/* ================================ Telemetry Buffer ================================ */

void
doca_telemetry_schema_set_buffer_size(struct doca_telemetry_schema *ds, uint64_t size)
{
	if (ds == NULL)
		return;

	ds->base_parameters->buffer_size = size;
}

doca_error_t
doca_telemetry_schema_get_buffer_size(struct doca_telemetry_schema *ds, uint64_t *size)
{
	if (ds == NULL || size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*size = ds->base_parameters->buffer_size;
	return DOCA_SUCCESS;
}

void
doca_telemetry_schema_set_buffer_data_root(struct doca_telemetry_schema *ds, const char *path)
{
	if (ds == NULL)
		return;
	ds->attributes.buffer_data_root = path;
}

doca_error_t
doca_telemetry_schema_get_buffer_data_root(struct doca_telemetry_schema *doca_schema, const char **path)
{
	if (doca_schema == NULL || path == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*path = strdup(doca_schema->attributes.buffer_data_root);
	if (*path == NULL)
		return DOCA_ERROR_NO_MEMORY;
	return DOCA_SUCCESS;
}

static inline doca_error_t
_schema_buffer_attr_set(struct doca_telemetry_schema *ds)
{
	int schema_path_len = 0;
	char *schema_path;

	_STRDUP(ds->base_parameters->data_root, ds->attributes.buffer_data_root);

	/* Schema path is $DATA_ROOT/schema */
	schema_path_len = strlen(ds->base_parameters->data_root) + strlen(SCHEMA_DIR_POSTFIX) + 1;
	schema_path     = calloc(schema_path_len, sizeof(char));
	if (schema_path == NULL) {
		DOCA_DLOG_ERR("Cannot alloc data root for doca_schema");
		free(ds->base_parameters->data_root);
		ds->base_parameters->data_root = NULL;
		return DOCA_ERROR_NO_MEMORY;
	}

	snprintf(schema_path, schema_path_len, "%s/schema", ds->base_parameters->data_root);

	free(ds->base_parameters->schema_path);
	ds->base_parameters->schema_path = schema_path;

	return DOCA_SUCCESS;
}

/* ================================ Telemetry file write ================================ */

void
doca_telemetry_schema_set_file_write_enabled(struct doca_telemetry_schema *ds)
{
	if (ds == NULL)
		return;

	ds->base_parameters->file_write_enabled = true;
}

void
doca_telemetry_schema_set_file_write_max_size(struct doca_telemetry_schema *ds, size_t size)
{
	if (ds == NULL)
		return;

	ds->base_parameters->max_file_size = size;
}

doca_error_t
doca_telemetry_schema_get_file_write_max_size(struct doca_telemetry_schema *doca_schema, size_t *size)
{
	if (doca_schema == NULL || size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*size = doca_schema->base_parameters->max_file_size;
	return DOCA_SUCCESS;
}

void
doca_telemetry_schema_set_file_write_max_age(struct doca_telemetry_schema *ds, doca_telemetry_timestamp_t max_age)
{
	if (ds == NULL)
		return;

	ds->base_parameters->max_file_age = max_age;
}

doca_error_t
doca_telemetry_schema_get_file_write_max_age(struct doca_telemetry_schema *doca_schema,
						doca_telemetry_timestamp_t *max_age)
{
	if (doca_schema == NULL || max_age == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*max_age = doca_schema->base_parameters->max_file_age;
	return DOCA_SUCCESS;
}

/* ================================ Telemetry IPC ================================ */

void
doca_telemetry_schema_set_ipc_enabled(struct doca_telemetry_schema *ds)
{
	if (ds == NULL)
		return;

	ds->base_parameters->ipc_enabled = true;
}

void
doca_telemetry_schema_set_ipc_sockets_dir(struct doca_telemetry_schema *ds, const char *sockets_dir)
{
	if (ds == NULL)
		return;

	ds->attributes.ipc_sockets_dir = sockets_dir;
}

doca_error_t
doca_telemetry_schema_get_ipc_sockets_dir(struct doca_telemetry_schema *doca_schema, const char **sockets_dir)
{
	if (doca_schema == NULL || sockets_dir == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*sockets_dir = strdup(doca_schema->attributes.ipc_sockets_dir);
	if (*sockets_dir == NULL)
		return DOCA_ERROR_NO_MEMORY;
	return DOCA_SUCCESS;
}

void
doca_telemetry_schema_set_ipc_reconnect_time(struct doca_telemetry_schema *ds, uint32_t max_time)
{
	if (ds == NULL)
		return;

	ds->base_parameters->ipc_max_reattach_time_msec = max_time;
}


doca_error_t
doca_telemetry_schema_get_ipc_reconnect_time(struct doca_telemetry_schema *doca_schema, uint32_t *max_time)
{
	if (doca_schema == NULL || max_time == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*max_time = doca_schema->base_parameters->ipc_max_reattach_time_msec;
	return DOCA_SUCCESS;
}


void
doca_telemetry_schema_set_ipc_reconnect_tries(struct doca_telemetry_schema *ds, uint8_t tries)
{
	if (ds == NULL)
		return;

	ds->base_parameters->ipc_max_reattach_tries = tries;
}

doca_error_t
doca_telemetry_schema_get_ipc_reconnect_tries(struct doca_telemetry_schema *doca_schema, uint8_t *tries)
{
	if (doca_schema == NULL || tries == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*tries = doca_schema->base_parameters->ipc_max_reattach_tries;
	return DOCA_SUCCESS;
}

void
doca_telemetry_schema_set_ipc_socket_timeout(struct doca_telemetry_schema *ds, uint32_t max_time)
{
	if (ds == NULL)
		return;

	ds->base_parameters->ipc_socket_timeout_msec = max_time;
}

doca_error_t
doca_telemetry_schema_get_ipc_socket_timeout(struct doca_telemetry_schema *doca_schema, uint32_t *timeout)
{
	if (doca_schema == NULL || timeout == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*timeout = doca_schema->base_parameters->ipc_socket_timeout_msec;
	return DOCA_SUCCESS;
}

static inline doca_error_t
_schema_ipc_attr_set(struct doca_telemetry_schema *ds)
{
	_STRDUP(ds->base_parameters->ipc_sockets_dir, ds->attributes.ipc_sockets_dir);
	return DOCA_SUCCESS;
}

/* ================================ Telemetry opaque events ================================ */

void doca_telemetry_schema_set_opaque_events_enabled(struct doca_telemetry_schema *ds)
{
	if (ds == NULL)
		return;

	ds->base_parameters->enable_opaque_events = true;
}

/* ================================ Telemetry field ================================ */

doca_error_t
doca_telemetry_field_create(struct doca_telemetry_field **field)
{
	if (field == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*field = (struct doca_telemetry_field *) calloc(1, sizeof(**field)); /* All fields are initalized to 0. (default) */
	if (*field == NULL)
		return DOCA_ERROR_NO_MEMORY;

	(*field)->is_netflow_field = false;
	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_field_destroy(struct doca_telemetry_field *field)
{
	if (field == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	if (field->field_owner == FIELD_OWNER_DOCA && !field->is_netflow_field) {
		free((char *)field->field.field_name);
		free((char *)field->field.description);
		free((char *)field->field.type_name);
	}

	free(field);

	return DOCA_SUCCESS;
}

void
doca_telemetry_field_set_name(struct doca_telemetry_field *field_info, const char *name)
{
	if (field_info == NULL)
		return;
	field_info->field.field_name = name;
}

void
doca_telemetry_field_set_description(struct doca_telemetry_field *field_info, const char *desc)
{
	if (field_info == NULL)
		return;
	field_info->field.description = desc;
}

void
doca_telemetry_field_set_type_name(struct doca_telemetry_field *field_info, const char *type)
{
	if (field_info == NULL)
		return;
	field_info->field.type_name = type;
}

void
doca_telemetry_field_set_array_length(struct doca_telemetry_field *field_info, uint16_t len)
{
	if (field_info == NULL)
		return;
	field_info->field.array_length = len;
}

static inline doca_error_t
_field_set_attributes(struct doca_telemetry_field *field)
{
	if (field->is_netflow_field)
		return DOCA_SUCCESS;

	field->field.field_name = strdup(field->field.field_name);
	if (field->field.field_name == NULL)
		return DOCA_ERROR_NO_MEMORY;

	field->field.description = strdup(field->field.description);
	if (field->field.description == NULL) {
		free((char *) field->field.field_name);
		field->field.field_name = (const char *) NULL;
		return DOCA_ERROR_NO_MEMORY;
	}
	field->field.type_name = strdup(field->field.type_name);
	if (field->field.type_name == NULL) {
		free((char *) field->field.field_name);
		free((char *) field->field.description);
		field->field.field_name = (const char *) NULL;
		field->field.description = (const char *) NULL;
		return DOCA_ERROR_NO_MEMORY;
	}
	return DOCA_SUCCESS;
}


/* ================================ Telemetry type ================================ */

doca_error_t
doca_telemetry_type_create(struct doca_telemetry_type **type)
{
	struct doca_telemetry_field **field_array;

	if (type == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*type = (struct doca_telemetry_type *) calloc(1, sizeof(**type));
	if (*type == NULL)
		return DOCA_ERROR_NO_MEMORY;

	field_array = (struct doca_telemetry_field **) calloc(TYPE_INIT_CAP, sizeof(*field_array));
	if (field_array == NULL) {
		free(*type);
		return DOCA_ERROR_NO_MEMORY;
	}

	(*type)->field_array = field_array;
	(*type)->array_capacity = TYPE_INIT_CAP;
	(*type)->nb_fields = 0;

	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_type_add_field(struct doca_telemetry_type *type, struct doca_telemetry_field *field)
{
	struct doca_telemetry_field **new_array;
	uint32_t new_array_cap;

	if (type == NULL || field == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	/* Same field should not be passed twice */
	if (field->field_owner == FIELD_OWNER_DOCA)
		return DOCA_ERROR_BAD_STATE;

	if (type->nb_fields >= type->array_capacity) {
		/* Limit number of fields per type */
		if (type->array_capacity > MAX_TYPE_CAP)
			return DOCA_ERROR_NO_MEMORY;
		new_array_cap = type->array_capacity * 2;
		/* Field array is full, try to increase capacity by two times the current capacity. */
		new_array = realloc(type->field_array, new_array_cap * sizeof(struct doca_telemetry_field *));
		if (new_array == NULL)
			return DOCA_ERROR_NO_MEMORY;
		type->array_capacity = new_array_cap;
		type->field_array = new_array;
	}
	if (_field_set_attributes(field) != DOCA_SUCCESS)
		return DOCA_ERROR_NO_MEMORY;

	type->field_array[type->nb_fields] = field;
	type->nb_fields++;
	field->field_owner = FIELD_OWNER_DOCA;

	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_type_destroy(struct doca_telemetry_type *type)
{
	int idx;
	doca_error_t ret;

	if (type == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	for (idx = 0; idx < type->nb_fields; idx++) {
		ret = doca_telemetry_field_destroy(type->field_array[idx]);
		if (ret != DOCA_SUCCESS)
			return ret;
		type->field_array[idx] = NULL;
	}
	free(type->field_array);
	free(type);

	return DOCA_SUCCESS;
}


doca_error_t
doca_telemetry_schema_add_type(struct doca_telemetry_schema *ds,
			       const char *new_type_name,
			       struct doca_telemetry_type *type,
			       doca_telemetry_type_index_t *type_index)
{
	int i = 0;
	int ret = 0;
	struct doca_telemetry_field **fields = type->field_array;
	clx_api_event_field_info_t tmp_fields[type->nb_fields];

	for (i = 0; i < type->nb_fields; i++) {
		tmp_fields[i].field_name    = fields[i]->field.field_name;
		tmp_fields[i].description   = fields[i]->field.description;
		tmp_fields[i].type_name     = fields[i]->field.type_name;
		tmp_fields[i].counting_mode = CLX_FIELD_VALUE_ABSOLUTE;
		tmp_fields[i].array_length  = fields[i]->field.array_length;

	}

	ret = clx_api_add_event_type(ds->base_ctx, (char *)new_type_name, tmp_fields, type->nb_fields,
								  type_index);

	switch (ret) {
	case CLX_API_NOMEMORY:
		return DOCA_ERROR_NO_MEMORY;
	case CLX_API_DUPLICATE:
	case CLX_API_UNDEFINED:
		return DOCA_ERROR_INVALID_VALUE;
	}
	return doca_telemetry_type_destroy(type);
}


/* ============================================================================= */


doca_error_t
doca_telemetry_source_create(struct doca_telemetry_schema *doca_schema, struct doca_telemetry_source **doca_source)
{
	struct doca_telemetry_source *new_source = calloc(1, sizeof(*new_source));

	if (new_source == NULL) {
		DOCA_DLOG_ERR("Failed to create doca source: allocation failed");
		return DOCA_ERROR_NO_MEMORY;
	}
	new_source->doca_schema = doca_schema;
	new_source->clone_parameters = calloc(1, sizeof(clx_api_params_t));
	if (!new_source->clone_parameters) {
		DOCA_DLOG_ERR("Failed to create doca source: cannot allocate parameters");
		free(new_source);
		return DOCA_ERROR_NO_MEMORY;
	}
	memcpy(new_source->clone_parameters, doca_schema->base_parameters, sizeof(clx_api_params_t));
	*doca_source = new_source;
	return DOCA_SUCCESS;
}

static doca_error_t _source_apply_attributes(struct doca_telemetry_source *doca_source)
{
	if (doca_source->attributes.source_id == NULL || doca_source->attributes.source_tag == NULL) {
		DOCA_DLOG_ERR("Cannot start source, because source attributes were not set");
		return DOCA_ERROR_INVALID_VALUE;
	}

	_STRDUP(doca_source->clone_parameters->source_id, doca_source->attributes.source_id);
	_STRDUP(doca_source->clone_parameters->source_tag, doca_source->attributes.source_tag);
	return DOCA_SUCCESS;

}

doca_error_t
doca_telemetry_source_start(struct doca_telemetry_source *ds)
{
	if (ds == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	/* Apply source attributes */
	if (_source_apply_attributes(ds) != DOCA_SUCCESS)
		return DOCA_ERROR_INVALID_VALUE;

	ds->clone_ctx = clx_api_clone_context(ds->doca_schema->base_ctx, ds->clone_parameters);
	if (ds->clone_ctx == NULL) {
		DOCA_DLOG_ERR("Cannot start source with id='%s' and tag='%s'",
				ds->clone_parameters->source_id,
				ds->clone_parameters->source_tag);
		return DOCA_ERROR_NO_MEMORY;
	}
	return DOCA_SUCCESS;
}


doca_error_t
doca_telemetry_source_destroy(struct doca_telemetry_source *ds)
{
	if (ds == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	clx_api_destroy_context(ds->clone_ctx);

	free(ds->clone_parameters->source_id);
	ds->clone_parameters->source_id = NULL;

	free(ds->clone_parameters->source_tag);
	ds->clone_parameters->source_tag = NULL;

	free(ds->clone_parameters);
	ds->clone_parameters = NULL;

	free(ds);
	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_source_report(struct doca_telemetry_source *ds,
			     doca_telemetry_type_index_t index,
			     void *data, int count)
{
	clx_api_ipc_connect_if_needed(ds->clone_ctx);

	if (!clx_api_write_event(ds->clone_ctx, data, index, count)) {
		DOCA_DLOG_ERR("DOCA source report failed. Internal buffer error - buffer size too small?");
		return DOCA_ERROR_NO_MEMORY;
	}
	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_source_opaque_report(struct doca_telemetry_source *ds,
			const doca_guid_t app_id, uint64_t user_defined1, uint64_t user_defined2,
			const void *data, uint32_t data_size)
{
	/* write / report opaque data */
	return clx_api_opaque_event_write(ds->clone_ctx, app_id, user_defined1, user_defined2, data, data_size)
			   ? DOCA_SUCCESS : DOCA_ERROR_NO_MEMORY;
}

doca_error_t
doca_telemetry_source_get_opaque_report_max_data_size(struct doca_telemetry_source *ds, uint32_t *max_data_size)
{
	if (ds == NULL || max_data_size == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*max_data_size = clx_api_opaque_event_max_data_size(ds->clone_ctx);

	return DOCA_SUCCESS;
}


doca_error_t
doca_telemetry_get_timestamp(doca_telemetry_timestamp_t *timestamp)
{
	if (timestamp == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*timestamp = clx_api_get_timestamp();

	return DOCA_SUCCESS;
}


doca_error_t
doca_telemetry_source_flush(struct doca_telemetry_source *ds)
{
	if (ds == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	clx_api_force_write(ds->clone_ctx);

	return DOCA_SUCCESS;
}


doca_error_t
doca_telemetry_check_ipc_status(struct doca_telemetry_source *ds, doca_telemetry_ipc_status_t *status)
{
	int ipc_status = 0;

	if (ds == NULL || status == NULL) {
		DOCA_DLOG_ERR("Cannot get IPC status: doca telemetry source is NULL.");
		return DOCA_ERROR_INVALID_VALUE;
	}

	ipc_status = clx_api_get_ipc_status(ds->clone_ctx);

	switch (ipc_status) {
	case CLX_API_IPC_DISABLED:
		*status = DOCA_TELEMETRY_IPC_STATUS_DISABLED;
		break;
	case CLX_API_IPC_SUCCESS:
		*status = DOCA_TELEMETRY_IPC_STATUS_CONNECTED;
		break;
	case CLX_API_IPC_FAILED:
	default:
		*status = DOCA_TELEMETRY_IPC_STATUS_FAILED;
		break;
	}

	return DOCA_SUCCESS;
}


/* ================================ DOCA SCHEMA ============================= */

static doca_error_t
_schema_apply_attributes(struct doca_telemetry_schema *ds)
{
	doca_error_t ret;

	ret = _schema_buffer_attr_set(ds);
	if (ret != DOCA_SUCCESS)
		return ret;
	ret = _schema_ipc_attr_set(ds);
	if (ret != DOCA_SUCCESS)
		return ret;
	return ret;
}

doca_error_t
doca_telemetry_schema_init(const char *schema_name, struct doca_telemetry_schema **doca_schema_output)
{
	clx_api_version_t ver = { { DOCA_VER_MAJOR, DOCA_VER_MINOR, DOCA_VER_PATCH } };
	struct doca_telemetry_schema *doca_schema;
	doca_error_t ret = DOCA_SUCCESS;

	if (doca_schema_output == NULL || schema_name == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	doca_schema = calloc(1, sizeof(*doca_schema));

	if (doca_schema == NULL) {
		DOCA_DLOG_ERR("Cannot allocate DOCA schema");
		return DOCA_ERROR_NO_MEMORY;
	}

	doca_schema->base_parameters = calloc(1, sizeof(clx_api_params_t));
	if (doca_schema->base_parameters == NULL) {
		DOCA_DLOG_ERR("Cannot allocate DOCA schema base parameters");
		doca_telemetry_schema_destroy(doca_schema);
		return DOCA_ERROR_NO_MEMORY;
	}

	doca_schema->base_ctx = clx_api_init_context_with_schema(schema_name, ver);
	if (doca_schema->base_ctx == NULL) {
		DOCA_DLOG_ERR("Cannot init base_ctx for DOCA schema");
		doca_telemetry_schema_destroy(doca_schema);
		return DOCA_ERROR_INITIALIZATION;
	}


	/* Buffer attributes */
	doca_schema->attributes.buffer_data_root = DOCA_TELEMETRY_DEFAULT_DATA_ROOT;
	doca_schema->attributes.ipc_sockets_dir = DOCA_TELEMETRY_DEFAULT_IPC_SOCKET_DIR;

	doca_schema->base_parameters->buffer_size = DOCA_TELEMETRY_DEFAULT_BUFFER_SIZE;
	doca_schema->base_parameters->schema_path = NULL;
	doca_schema->base_parameters->data_root = NULL;

	/* Storage attributes */
	doca_schema->base_parameters->file_write_enabled	= false;
	doca_schema->base_parameters->max_file_age		= DOCA_TELEMETRY_DEFAULT_FILE_AGE;
	doca_schema->base_parameters->max_file_size		= DOCA_TELEMETRY_DEFAULT_FILE_SIZE;

	doca_schema->base_parameters->data_path_template	= CLX_API_DATA_PATH_TEMPLATE;

	doca_schema->base_parameters->enable_opaque_events	= 0;

	doca_schema->base_parameters->source_id			= NULL;
	doca_schema->base_parameters->source_tag		= NULL;

	doca_schema->base_parameters->prometheus_enpoint	= "";
	doca_schema->base_parameters->cset_dir_path		= "";

	doca_schema->base_parameters->ipc_enabled		= false;
	doca_schema->base_parameters->ipc_sockets_dir		= NULL;

	doca_schema->base_parameters->ipc_max_reattach_time_msec = DOCA_TELEMETRY_DEFAULT_IPC_RECONNECT_TIME;
	doca_schema->base_parameters->ipc_max_reattach_tries     = DOCA_TELEMETRY_DEFAULT_IPC_RECONNECT_RETRIES;
	doca_schema->base_parameters->ipc_socket_timeout_msec    = DOCA_TELEMETRY_DEFAULT_IPC_SOCKET_TIMEOUT;

	/* Apply DEFAULT attributes. */
	ret = _schema_apply_attributes(doca_schema);
	if (ret != DOCA_SUCCESS) {
		doca_telemetry_schema_destroy(doca_schema);
		return ret;
	}

	*doca_schema_output = doca_schema;
	return DOCA_SUCCESS;
}


doca_error_t
doca_telemetry_schema_start(struct doca_telemetry_schema *ds)
{
	doca_error_t ret = DOCA_SUCCESS;

	if (ds == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	/* Apply attributes */
	ret = _schema_apply_attributes(ds);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (!clx_api_context_apply_schema(ds->base_ctx, ds->base_parameters)) {
		DOCA_DLOG_ERR("doca_telemetry_schema_apply failed");
		return DOCA_ERROR_INITIALIZATION;
	}

	/* Will only be a real error when the first report operation will fail */
	if (ds->base_parameters->ipc_enabled && clx_api_get_ipc_status(ds->base_ctx) != CLX_API_IPC_SUCCESS)
		DOCA_DLOG_WARN("IPC transport failed to connect at start");

	return DOCA_SUCCESS;
}


doca_error_t
doca_telemetry_schema_destroy(struct doca_telemetry_schema *ds)
{
	if (ds == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (ds->base_ctx != NULL)
		clx_api_destroy_context(ds->base_ctx);
	ds->base_ctx = NULL;

	_schema_destroy_attributes(ds);

	if (ds->base_parameters != NULL)
		free(ds->base_parameters);
	ds->base_parameters = NULL;

	free(ds);
	return DOCA_SUCCESS;
}
