/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/sysinfo.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#include <doca_log.h>

#include "doca_telemetry_netflow.h"
#include "telemetry_impl.h"

DOCA_LOG_REGISTER(TELEMETRY_NETFLOW);

#define NETFLOW_PROTOCOL_V9 0x0009
#define NETFLOW_FLOWSET_PADDING 4
#define MSEC_IN_SEC 1000
#define MAKE_U64(low, hi) ((uint64_t)(((uint32_t)(low)) | (((uint64_t)((uint32_t)(hi))) << 32)))
#define MAKE_U32(low, hi) ((uint32_t)(((uint16_t)(low)) | (((uint32_t)((uint16_t)(hi))) << 16)))

static const doca_guid_t doca_telemetry_netflow_app_id = DOCA_NETFLOW_APP_ID;

struct doca_telemetry_netflow_ctx {
	struct doca_telemetry_schema	*doca_schema;
	struct doca_telemetry_source	*doca_source;
	uint32_t			source_id;
	uint32_t			flow_sequence;
	uint32_t			opaque_report_max_data_size;
	const char			*collector_addr;
	uint16_t			collector_port;
	uint16_t			max_packet_size;
};

struct doca_telemetry_netflow_flowset_field_v9 {
	uint16_t type;
	uint16_t length;
} __attribute__((packed));

struct doca_telemetry_netflow_template_v9 {
	uint16_t template_id;
	uint16_t field_count;
} __attribute__((packed));

struct doca_telemetry_netflow_section_header_v9 {
	uint16_t id;
	uint16_t length;
} __attribute__((packed));

struct doca_telemetry_netflow_flowSet_header_v9 {
	struct doca_telemetry_netflow_section_header_v9 section_header;
	struct doca_telemetry_netflow_template_v9       flowSet_template_header;
} __attribute__((packed));

struct doca_telemetry_netflow_header_v9 {
	uint16_t                                        version;
	uint16_t                                        count;
	uint32_t                                        sys_up_time;
	uint32_t                                        unix_secs;
	uint32_t                                        flow_sequence;
	uint32_t                                        source_id;
	struct doca_telemetry_netflow_flowSet_header_v9 flowSet_header;
} __attribute__((packed));

static struct doca_telemetry_netflow_ctx ctx = {0};

struct doca_telemetry_netflow_flowset_field {
	struct doca_telemetry_field *field;
};

struct doca_telemetry_netflow_template {
	struct doca_telemetry_type *fields;
};

doca_error_t
doca_telemetry_netflow_init(uint16_t source_id)
{
	doca_error_t ret;

	if (ctx.doca_schema != NULL)
		return DOCA_ERROR_BAD_STATE;

	ret = doca_telemetry_schema_init("DOCA NETFLOW API", &ctx.doca_schema);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = doca_telemetry_source_create(ctx.doca_schema, &ctx.doca_source);
	if (ret != DOCA_SUCCESS) {
		doca_telemetry_schema_destroy(ctx.doca_schema);
		return ret;
	}

	doca_telemetry_schema_set_opaque_events_enabled(ctx.doca_schema);

	ctx.doca_schema->base_parameters->netflow_collector_addr = NULL;
	ctx.collector_port = DOCA_NETFLOW_DEFAULT_PORT;
	ctx.max_packet_size = 0;
	ctx.source_id = source_id;
	ctx.flow_sequence = 0;
	return DOCA_SUCCESS;
}

void
doca_telemetry_netflow_set_collector_addr(const char *collector_addr)
{
	if (ctx.doca_schema == NULL)
		return;

	ctx.collector_addr = collector_addr;
}

void
doca_telemetry_netflow_set_collector_port(uint16_t collector_port)
{
	if (ctx.doca_schema == NULL)
		return;

	ctx.collector_port = collector_port;
}

void
doca_telemetry_netflow_set_max_packet_size(uint16_t max_packet_size)
{
	if (ctx.doca_schema == NULL)
		return;

	ctx.max_packet_size = max_packet_size;
}

static doca_error_t
_netflow_send_attr_set()
{
	struct doca_telemetry_schema *ds = ctx.doca_schema;

	_STRDUP(ds->base_parameters->netflow_collector_addr, ctx.collector_addr);
	ds->base_parameters->netflow_collector_port = ctx.collector_port;
	ds->base_parameters->netflow_max_packet_size = ctx.max_packet_size;
	return DOCA_SUCCESS;
}



doca_error_t
doca_telemetry_netflow_field_create(struct doca_telemetry_netflow_flowset_field **field)
{
	doca_error_t ret;

	if (field == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*field = (struct doca_telemetry_netflow_flowset_field *)calloc(1, sizeof(**field));
	if (*field == NULL)
		return DOCA_ERROR_NO_MEMORY;

	ret = doca_telemetry_field_create(&(*field)->field);
	if (ret != DOCA_SUCCESS) {
		free(field);
		return ret;
	}
	(*field)->field->is_netflow_field = true;

	return DOCA_SUCCESS;
}

void
doca_telemetry_netflow_field_set_type(struct doca_telemetry_netflow_flowset_field *field, uint16_t type)
{
	if (field == NULL)
		return;

	field->field->netflow_field.netflow_type = type;
}

void
doca_telemetry_netflow_field_set_length(struct doca_telemetry_netflow_flowset_field *field, uint16_t length)
{
	if (field == NULL)
		return;

	field->field->netflow_field.netflow_length = length;
}

doca_error_t
doca_telemetry_netflow_field_destroy(struct doca_telemetry_netflow_flowset_field *field)
{
	doca_error_t ret;

	if (field == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = doca_telemetry_field_destroy(field->field);
	if (ret != DOCA_SUCCESS)
		return ret;

	free(field);

	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_netflow_template_create(struct doca_telemetry_netflow_template **netflow_template)
{
	doca_error_t ret;

	if (netflow_template == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*netflow_template = (struct doca_telemetry_netflow_template *)calloc(1, sizeof(**netflow_template));
	if (*netflow_template == NULL)
		return DOCA_ERROR_NO_MEMORY;

	ret = doca_telemetry_type_create(&(*netflow_template)->fields);
	if (ret != DOCA_SUCCESS) {
		free(netflow_template);
		return ret;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_netflow_template_add_field(struct doca_telemetry_netflow_template *netflow_template,
						struct doca_telemetry_netflow_flowset_field *field)
{
	doca_error_t ret;

	if (netflow_template == NULL || field == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = doca_telemetry_type_add_field(netflow_template->fields, field->field);
	if (ret != DOCA_SUCCESS)
		return ret;

	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_netflow_template_destroy(struct doca_telemetry_netflow_template *netflow_template)
{
	doca_error_t ret;

	if (netflow_template == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ret = doca_telemetry_type_destroy(netflow_template->fields);
	if (ret != DOCA_SUCCESS)
		return ret;
	free(netflow_template);

	return DOCA_SUCCESS;
}

void
doca_telemetry_netflow_set_ipc_enabled(void)
{
	doca_telemetry_schema_set_ipc_enabled(ctx.doca_schema);
}

void
doca_telemetry_netflow_set_ipc_sockets_dir(const char *path)
{
	doca_telemetry_schema_set_ipc_sockets_dir(ctx.doca_schema, path);
}

doca_error_t
doca_telemetry_netflow_get_ipc_sockets_dir(const char **path)
{
	return doca_telemetry_schema_get_ipc_sockets_dir(ctx.doca_schema, path);
}

void
doca_telemetry_netflow_set_file_write_enabled(void)
{
	doca_telemetry_schema_set_file_write_enabled(ctx.doca_schema);
}

void
doca_telemetry_netflow_set_file_write_max_size(size_t size)
{
	doca_telemetry_schema_set_file_write_max_size(ctx.doca_schema, size);
}

doca_error_t
doca_telemetry_netflow_get_file_write_max_size(size_t *size)
{
	return doca_telemetry_schema_get_file_write_max_size(ctx.doca_schema, size);
}

void
doca_telemetry_netflow_set_file_write_max_age(doca_telemetry_timestamp_t max_age)
{
	doca_telemetry_schema_set_file_write_max_age(ctx.doca_schema, max_age);
}

doca_error_t
doca_telemetry_netflow_get_file_write_max_age(doca_telemetry_timestamp_t *max_age)
{
	return doca_telemetry_schema_get_file_write_max_age(ctx.doca_schema, max_age);
}

void
doca_telemetry_netflow_set_buffer_size(uint64_t size)
{
	doca_telemetry_schema_set_buffer_size(ctx.doca_schema, size);
}

doca_error_t doca_telemetry_netflow_get_buffer_size(uint64_t *size)
{
	return doca_telemetry_schema_get_buffer_size(ctx.doca_schema, size);
}

void
doca_telemetry_netflow_set_buffer_data_root(const char *root)
{
	doca_telemetry_schema_set_buffer_data_root(ctx.doca_schema, root);
}

doca_error_t
doca_telemetry_netflow_get_buffer_data_root(const char **path)
{
	return doca_telemetry_schema_get_buffer_data_root(ctx.doca_schema, path);
}

void
doca_telemetry_netflow_source_set_id(const char *source_id)
{
	doca_telemetry_source_set_id(ctx.doca_source, source_id);
}

void
doca_telemetry_netflow_source_set_tag(const char *source_tag)
{
	doca_telemetry_source_set_tag(ctx.doca_source, source_tag);
}

doca_error_t
doca_telemetry_netflow_start(void)
{
	doca_error_t ret;

	if (ctx.doca_schema == NULL || ctx.doca_source == NULL)
		return DOCA_ERROR_BAD_STATE;

	/* Set collector configuration */
	ret = _netflow_send_attr_set();
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = doca_telemetry_schema_start(ctx.doca_schema);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = doca_telemetry_source_start(ctx.doca_source);
	if (ret != DOCA_SUCCESS)
		return ret;

	ret = doca_telemetry_source_get_opaque_report_max_data_size(ctx.doca_source, &ctx.opaque_report_max_data_size);
	if (ret != DOCA_SUCCESS)
		return ret;

	return ret;
}

static doca_error_t
_doca_telemetry_netflow_send_records_v9(struct doca_telemetry_netflow_header_v9 *header,
		const struct doca_telemetry_netflow_template *netflow_template,
		const void **records, uint32_t number_of_records,
		uint32_t record_len, uint32_t padding, uint8_t *buffer, size_t buffer_size)
{
	struct doca_telemetry_netflow_header_v9         *buf_header;
	struct doca_telemetry_netflow_flowset_field_v9  *template_v9;
	struct doca_telemetry_netflow_section_header_v9 *flows_header_v9;
	uint8_t                                         *buffer_pos = buffer;
	uint16_t                                         i;

	buf_header = (struct doca_telemetry_netflow_header_v9 *)buffer_pos;
	buffer_pos += sizeof(struct doca_telemetry_netflow_header_v9);
	template_v9 = (struct doca_telemetry_netflow_flowset_field_v9 *)buffer_pos;
	buffer_pos += sizeof(struct doca_telemetry_netflow_flowset_field_v9)
			* netflow_template->fields->nb_fields;
	flows_header_v9 = (struct doca_telemetry_netflow_section_header_v9 *)buffer_pos;
	buffer_pos += sizeof(struct doca_telemetry_netflow_section_header_v9);

	buf_header->version                              = htobe16(header->version);
	buf_header->count                                = htobe16(header->count);
	buf_header->sys_up_time                          = htobe32(header->sys_up_time);
	buf_header->unix_secs                            = htobe32(header->unix_secs);
	buf_header->flow_sequence                        = htobe32(header->flow_sequence);
	buf_header->source_id                            = htobe32(header->source_id);
	buf_header->flowSet_header.section_header.id     = htobe16(0);
	buf_header->flowSet_header.section_header.length =
		htobe16(sizeof(struct doca_telemetry_netflow_flowSet_header_v9) +
			sizeof(struct doca_telemetry_netflow_flowset_field_v9) *
				netflow_template->fields->nb_fields);
	buf_header->flowSet_header.flowSet_template_header.template_id = htobe16(258);
	buf_header->flowSet_header.flowSet_template_header.field_count =
		htobe16(netflow_template->fields->nb_fields);
	for (i = 0; i < netflow_template->fields->nb_fields; i++) {
		template_v9[i].type   = htobe16(netflow_template->fields->field_array[i]->netflow_field.netflow_type);
		template_v9[i].length = htobe16(netflow_template->fields->field_array[i]->netflow_field.netflow_length);
	}
	flows_header_v9->id     = htobe16(258);
	flows_header_v9->length = htobe16(number_of_records * record_len + padding +
					sizeof(struct doca_telemetry_netflow_section_header_v9));
	/* In this loops logic, header count stands only for records and not the template */
	for (i = 0; i < number_of_records; i++) {
		uint8_t *record = (uint8_t *)(records[i]);

		memcpy(buffer_pos, record, record_len);
		buffer_pos += record_len;
	}
	memset(buffer_pos, 0, padding);
	buffer_pos += padding;

	return doca_telemetry_source_opaque_report(ctx.doca_source, doca_telemetry_netflow_app_id,
					MAKE_U64(MAKE_U32(record_len, number_of_records), 0), 0, buffer, buffer_size);
}

doca_error_t
doca_telemetry_netflow_send(const struct doca_telemetry_netflow_template *netflow_template,
			const void **records, size_t nof_records, size_t *nof_records_sent)
{
	struct doca_telemetry_netflow_header_v9 header = {
		.version   = NETFLOW_PROTOCOL_V9,
		.source_id = ctx.source_id,
	};
	struct sysinfo s_info;
	time_t         t;

	if (!ctx.doca_schema || !ctx.doca_source)
		return DOCA_ERROR_BAD_STATE;

	sysinfo(&s_info);
	time(&t);
	header.unix_secs   = t;
	header.sys_up_time = s_info.uptime * MSEC_IN_SEC;

	if (nof_records_sent)
		*nof_records_sent = 0;

	uint16_t i           = 0;
	uint32_t record_size = 0;
	uint32_t padding = 0;

	/* calc record size */
	for (i = 0; i < netflow_template->fields->nb_fields; i++)
		record_size += netflow_template->fields->field_array[i]->netflow_field.netflow_length;

	if (record_size == 0)
		return DOCA_ERROR_INVALID_VALUE;

	/* calc headers size */
	uint32_t headers_size = sizeof(struct doca_telemetry_netflow_header_v9) +
		sizeof(struct doca_telemetry_netflow_section_header_v9) +
		sizeof(struct doca_telemetry_netflow_flowset_field_v9) * netflow_template->fields->nb_fields;
		/*
		 * Netflow header + Netflow section header +
		 * Template size [(field = one record) * number of fields]
		 */

	/* calc max records which can be sent via opaque events mechanism */
	size_t max_records;

	max_records = (ctx.opaque_report_max_data_size - headers_size) / record_size;

	if (max_records >= nof_records)
		max_records = nof_records;

	/* make sure we have a place for padding */
	if (max_records * record_size == (ctx.opaque_report_max_data_size - headers_size))
		--max_records;

	/* calc FlowSet padding */
	padding = (headers_size + max_records * record_size) % NETFLOW_FLOWSET_PADDING;
	if (padding)
		padding = NETFLOW_FLOWSET_PADDING - padding;

	/* calc required buffer size */
	size_t   buffer_size = headers_size + max_records * record_size + padding;
	uint8_t *buffer      = (uint8_t *)malloc(buffer_size);

	if (!buffer)
		return DOCA_ERROR_NO_MEMORY;

	/* send the records. if necessary, split the batch */
	doca_error_t res = DOCA_SUCCESS;

	while (nof_records) {
		size_t nof_records_to_send =
			(nof_records >= max_records) ? max_records : nof_records;

		header.flow_sequence = ctx.flow_sequence;
		/* Number of FlowSet records (both template FlowSet and data FlowSet) within this packet */
		header.count         = nof_records_to_send + 1;

		res = _doca_telemetry_netflow_send_records_v9(&header, netflow_template, records, nof_records_to_send,
					record_size, padding, buffer, buffer_size);
		if (res != DOCA_SUCCESS)
			break;

		++ctx.flow_sequence;
		if (nof_records_sent)
			*nof_records_sent += nof_records_to_send;
		nof_records -= nof_records_to_send;
		records += nof_records_to_send;
	}

	free(buffer);
	return res;
}

doca_error_t
doca_telemetry_netflow_flush(void)
{
	if (ctx.doca_source == NULL)
		return DOCA_ERROR_BAD_STATE;

	doca_telemetry_source_flush(ctx.doca_source);
	return DOCA_SUCCESS;
}

doca_error_t
doca_telemetry_netflow_destroy(void)
{
	if (ctx.doca_source) {
		doca_telemetry_source_destroy(ctx.doca_source);
		ctx.doca_source = NULL;
	}

	if (ctx.doca_schema) {
		doca_telemetry_schema_destroy(ctx.doca_schema);
		ctx.doca_schema = NULL;
	}
	return DOCA_SUCCESS;
}
