/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#ifndef DOCA_TELEMETRY_IMPL_H_
#define DOCA_TELEMETRY_IMPL_H_

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <clx_api.h>

#define DOCA_TELEMETRY_FIELD_VALUE_ABSOLUTE     CLX_FIELD_VALUE_ABSOLUTE
#define _STRDUP(dst, src)				\
	do {						\
		free(dst);				\
		dst = strdup(src);			\
		if (dst == NULL)			\
			return DOCA_ERROR_NO_MEMORY;	\
	} while (0)					\


/** @brief DOCA Telemetry schema internal type.
 *
 * Schema is used to prepare user types and create Doca sources.
 *
 * It contains default parameters for CLX_API context and
 * base CLX_API context that is shared with the rest of the contexts.
 * Base context is hidden from user and does not collect/report data.
 */
struct doca_telemetry_schema {
	clx_api_params_t	*base_parameters;	/* Default parameters for base context. */
	clx_api_context_t	*base_ctx;		/* API context that is used as a base for DOCA contexts. */
	struct {
		const char 	*buffer_data_root;	/* Root path for the data. */
		const char 	*ipc_sockets_dir;	/* Directory for the IPC sockets. */
	} attributes;
};


/** @brief DOCA Telemetry source internal type.
 *
 * Created from DOCA Telemetry schema. Sources created from the same DOCA
 * schema have the same types, only source_[id|tag] are different.
 *
 * It contains a clone of CLX_API base_context which is used to collect/report data.
 * clone_parameters is a copy of base_parameters with overwritten source_[id|tag].
 * doca_schema points to the parent schema.
 */
struct doca_telemetry_source {
	struct doca_telemetry_schema *doca_schema;	/* Pointer to DOCA schema with parent API context. */
	clx_api_params_t        *clone_parameters;	/* Parameters for cloning of CLX_API context. */
	clx_api_context_t       *clone_ctx;		/* Cloned CLX_API context that is used to collect/report data. */
	struct {
		const char 	*source_id;		/* Source ID. */
		const char 	*source_tag;		/* Source tag. */
	} attributes;
};

enum {
	FIELD_OWNER_USER = 0,
	FIELD_OWNER_DOCA,
};

struct doca_telemetry_field {
	union {
		struct {
			const char *field_name;		/* Name of field */
			const char *description;	/* Field description */
			const char *type_name;		/* Name of type that is already in schema (including built-in types). */
			uint16_t    array_length;	/* Array length */
		} field;
		struct {
			uint16_t netflow_type;		/* Netflow field type */
			uint16_t netflow_length;	/* Netflow field length */
		} netflow_field;
	};
	int field_owner;			/* Field ownership, user = 0 , doca_telemetry = OWNERSHIP_DOCA_TELEMETRY */
	bool is_netflow_field;		/* Whether this field is a netflow field */
};

struct doca_telemetry_type {
	struct doca_telemetry_field **field_array;
	uint32_t array_capacity;	/* Indicates the capacity of the array */
	uint32_t nb_fields;		/* Indicates number of fields in the array */
};


#endif /* DOCA_TELEMETRY_IMPL_H_ */
