/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_version.h>
#include <json-c/json.h>
#include <netinet/if_ether.h>
#include <set>
#include <map>
#include <list>
#include <vector>
#include <string.h>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <doca_log.h>
#include <regex-compiler.h>
#ifdef DOCA_DPI_USE_HS
#include <hs/hs_compile.h>
#endif
#include "dpi_compiler.hpp"

using namespace std;

#define RULES_FILE_NAME "signatures.rules"
#define ROF_FILE_NAME "signatures_compiled.rof2"
#define IPV4_LEN 15
#define HS_RULES_FILE_NAME "hs_signatures.rules"

DOCA_LOG_REGISTER(COMPILER);

#define ADD_INT_OBJ(src, name, dest) json_object_object_add(dest, name, json_object_new_int(src));

#define ADD_STRING_OBJ(src, name, dest)								\
	json_object_object_add(dest, name, json_object_new_string(src));

#define ADD_BOOL_OBJ(src, name, dest) json_object_object_add(dest, name, json_object_new_boolean(src));

enum group_property_enum {
	DOCA_DPI_PRIOR_MASK		= 1 << 0,
};

enum metadata_patch_order_enum {
	ETH_TYPE,
	SRC_PORT,
	SRC_ADDR,
	DEST_PORT,
	DEST_ADDR,
	DIRECTION,
	FIELD,
	L4_PROT,
	L7_PROT,
	MAX_METADATA_ENUM
};

int metadata_order[MAX_METADATA_ENUM] = {
	FIELD,
	L7_PROT,
	L4_PROT,
	DIRECTION,
	DEST_PORT,
	DEST_ADDR,
	SRC_PORT,
	SRC_ADDR,
	ETH_TYPE,
};

const map < int , string > field_to_str = {{DOCA_DPI_HTTP_CONTENT, "content"},
					   {DOCA_DPI_HTTP_URI, "uri"},
					   {DOCA_DPI_HTTP_USER_AGENT, "user_agent"},
					   {DOCA_DPI_HTTP_LOCATION, "location"},
					   {DOCA_DPI_HTTP_SERVER, "server"},
					   {DOCA_DPI_HTTP_HOST, "host"},
					   {DOCA_DPI_HTTP_CONTENT_TYPE, "content_type"},
					   {DOCA_DPI_HTTP_COOKIE, "cookie"},
					   {DOCA_DPI_HTTP_REFERER, "referer"},
					   {DOCA_DPI_SSL_COMMON_NAME, "common_name"},
					   {DOCA_DPI_SSL_SERVER_NAME, "server_name"},
					   {DOCA_DPI_SSL_SUBJECT_ALT_NAME, "subject_alt_name"},
					   {DOCA_DPI_SSL_RECORD_VER, "record_version"},
					   {DOCA_DPI_SSL_HANDSHAKE_VER, "handshake_version"},
					   {DOCA_DPI_SSL_CERT_SUBJECT, "ssl_cert_subject"},
					   {DOCA_DPI_RTMP_PAGE_URL, "page_url"},
					   {DOCA_DPI_DNS_QUERY, "query"},
					   {DOCA_DPI_STUN_REALM, "realm"},
					   {DOCA_DPI_STUN_SOFTWARE, "software"},
					   {DOCA_DPI_IP_RESOLV_NAME, "resolv_name"}
};

const map < int , string > l7_prot_to_str = {{DOCA_DPI_HTTP, "http"},
					     {DOCA_DPI_HTTP2, "http2"},
					     {DOCA_DPI_TLS, "ssl"},
					     {DOCA_DPI_SSL, "ssl"},
					     {DOCA_DPI_SIP, "sip"},
					     {DOCA_DPI_RTMP, "rtmp"},
					     {DOCA_DPI_QUIC, "quic"},
					     {DOCA_DPI_STUN, "stun"},
					     {DOCA_DPI_DNS, "dns"}
};
/* poc */

struct group_property {
	int order1;
	int order2;
	int bitmask;
	int l7_prot;
	int l4_prot;
	int eth_type;
	int field;
	int direction;
	string src_addr;
	int src_mask;
	string dest_addr;
	int dest_mask;
	bool src_is_ipv6;
	bool dest_is_ipv6;
	int src_port_start;
	int src_port_end;
	int dest_port_start;
	int dest_port_end;
	bool update_done;
};

enum rules_types {
	REGEX_TYPE,
	ADDRESS_TYPE,
	PORT_TYPE,
	SIMPLE_TYPE
};

struct rule_obj {
	struct json_object *json;
	int type;
};

struct regex_obj_fields {
	string regex;
	int l7_prot;
	int regex_id;
	int field;
	int direction;
	int offset;
	bool dollar;
	bool anchor;
};

struct doca_dpi_compiler {
	/* map of < subgroup id, < regex id , regex string > > to arrange regex ids to subgroups*/
	map < int, map < int, string > > subgrp2reg_map;
	/* number of regex strings inserted */
	int num_regex;
	/* number of groups with more than one rule */
	int num_conds;
	/* number of ports inserted */
	int num_ports;
	/* number of addresses inserted */
	int num_addresses;
	/* map of < sig id, < group id, list < rule's json and type > > > for  the CDO pp_table */
	map < int, map < int, list < struct rule_obj > > > grp2rules_map;
	/* map of < sig_id < group_id, list < regex fields > > > to hold regex until all
	 * fields arrived
	 */
	map < int, map < int, list < struct regex_obj_fields > > > group2regex_list_map;
	/* map of < sig_id < group_id, group_property > > to hold group's property */
	map < int, map < int, struct group_property > > group_prop_map;
	/* store signatures according to sig_id */
	map < int, struct doca_dpi_sig_info > sig_map;
};

/* allocating the necessary memory for dpi compiler.
 * return pointer to "struct doca_dpi_compiler" for success and NULL for failure.
 */
struct doca_dpi_compiler *doca_dpi_compiler_alloc()
{
	struct doca_dpi_compiler *ctx = new(nothrow) struct doca_dpi_compiler;
	if (ctx == NULL) {
		DOCA_DLOG_ERR("Error: compiler allocation failed\n");
		return NULL;
	}
	ctx->num_regex = 0;
	ctx->num_addresses = 0;
	ctx->num_ports = 0;
	ctx->num_conds = 0;
	return ctx;
}

void doca_dpi_compiler_dealloc(struct doca_dpi_compiler *ctx)
{
	if (ctx == NULL)
		return;

	/* clearing subgroup map */
	for (auto i : ctx->subgrp2reg_map) {
		i.second.clear();
	}
	ctx->subgrp2reg_map.clear();

	/* clearing groups map */
	for (auto i : ctx->grp2rules_map) {
		for (auto j : i.second) {
			j.second.clear();
		}
		i.second.clear();
	}
	ctx->grp2rules_map.clear();

	/* clearing regex rules properties */
	ctx->group2regex_list_map.clear();

	/* clearing groups properties */
	ctx->group_prop_map.clear();

	/* clearing signaturs map */
	for (auto i : ctx->sig_map) {
		free(i.second.name);
		if (i.second.action != NULL)
			free(i.second.action);
	}
	ctx->sig_map.clear();

	delete ctx;
	return;
}

int doca_dpi_compiler_set_sig(struct doca_dpi_compiler *ctx, int sig_id,
			      const struct doca_dpi_sig_info* sig_info)
{
	struct doca_dpi_sig_info sig_info_new;

	if (sig_id < 1) {
		DOCA_DLOG_ERR("Sig id must be greater than 0\n");
		return -1;
	}

	sig_info_new.name = strdup(sig_info->name);
	if (sig_info_new.name == NULL) {
		DOCA_DLOG_ERR("Error: strdup failed for sig name\n");
		return -1;
	}
	sig_info_new.action = strdup(sig_info->action);
	if (sig_info_new.action == NULL) {
		DOCA_DLOG_ERR("Error: strdup failed for sig action\n");
		free(sig_info_new.name);
		return -1;
	}
	if (ctx->sig_map.find(sig_id) != ctx->sig_map.end()) {
		if (ctx->sig_map[sig_id].name != NULL)
			free(ctx->sig_map[sig_id].name);
		if (ctx->sig_map[sig_id].action != NULL)
			free(ctx->sig_map[sig_id].action);
		ctx->sig_map.erase(sig_id);
	}

	ctx->sig_map[sig_id].name = sig_info_new.name;
	ctx->sig_map[sig_id].action = sig_info_new.action;
	return 0;
}

/* add trivial ranges */
void trivial_finish_range(int div, string &range_str)
{
	while (div > 0) {
		range_str += "[0-9]";
		div = div / 10;
	}
	range_str += ")|";
}

/* handle inner level digits in port range */
int range_inner(int start, int end, int div, bool is_upper, string &range_str)
{
	bool non_found = true;
	int from, to;

	while (div > 0) {
		from = start / div;
		to = end / div;
		start = start - (div * from);
		end = end - (div * to);
		div = div / 10;

		/* write digit range */
		if (to == from) {
			range_str += to_string(from);
			continue;
		}
		if (is_upper) {
			/* top digit */
			range_str += "(" + to_string(to) + "[0-" + to_string(end / div) + "]";
			trivial_finish_range(div / 10, range_str);

			/* low part */
			if (to > 0) {
				range_str += "([0-" + to_string(to - 1) + "]";
				trivial_finish_range(div, range_str);
			}
		} else {
			/* top part */
			if (from < 9) {
				range_str += "([" + to_string(from + 1) + "-9]";
				trivial_finish_range(div, range_str);
			}

			/* low digit */
			range_str += "(" + to_string(from) + "[" + to_string(start / div) + "-9]";
			trivial_finish_range(div / 10, range_str);
		}

		non_found = false;
		break;
	}
	if (non_found)
		range_str += ")|";
	return div;
}

/* support only up to 9999 */
static char *port_rang_to_string(int start, int end)
{
	if (start == end)
		return strdup(to_string(start).c_str());
	if (start > end) {
		DOCA_DLOG_ERR("port %d:%d is written in the wrong direction.\n", end, start);
		DOCA_DLOG_ERR("port range should be written in the form of <start_num> : <end_num>\n");
		return NULL;
	}

	int from, to, div = 1000;
	string range_str = "(";

	while (div > 0) {
		from = start / div;
		to = end / div;

		/* find the highest digit */
		if (to == 0) {
			div = div / 10;
			continue;
		}

		start = start - (div * from);
		end = end - (div * to);
		div = div / 10;
		/* write digit range */
		if (to == from) {
			range_str += to_string(from);
			continue;
		} else if (to > from) {
			/* top part */
			range_str += "(" + to_string(to);
			if (range_inner(start, end, div, true, range_str) < 0)
				return NULL;
			range_str.replace(range_str.size() - 1, 1, ")|");

			/* middle part */
			if (to - from > 1) {
				range_str += "([" + to_string(from + 1) + "-" + to_string(to + 1) +
					     "]";
				trivial_finish_range(div, range_str);
			}

			/* low part */
			range_str += "(";
			if (from > 0)
				range_str += to_string(from);
			if (range_inner(start, end, div, false, range_str) < 0)
				return NULL;
			range_str.replace(range_str.size() - 1, 1, ")|");
			break;
		} else {
			DOCA_DLOG_ERR("port range %d:%d failed in translation\n",start, end);
			return NULL;
		}
	}
	range_str.erase(range_str.size() - 1);
	range_str += ")";
	return strdup(range_str.c_str());
}

static void reformat_regex(string &regex, bool has_dollar)
{
	bool caseinsensitive = false;
	/* chage '/i' case insensitive to '(?i)...(?-i)' */
	if (regex.rfind('/') != string::npos && regex.rfind('i') != string::npos
						&& regex.rfind('/') < regex.rfind('i'))
		caseinsensitive = true;

	/* remove $ anchor */
	if (has_dollar)
		regex.erase(regex.rfind('$'), 1);

	/* clear '/' from beginning and end of string */
	if (regex.at(0) == '/') {
		if (caseinsensitive) {
			regex.replace(0, 1, "(?i)");
			regex.erase(regex.rfind('/'));
			regex.append("(?-i)");
		} else {
			regex.erase(0, 1);
			regex.erase(regex.rfind('/'), 1);
		}
	}
}

static char *char_to_range(int first_bit, const char c)
{
	int hex_val;

	if (first_bit == 3)
		return (strdup("[0-9a-fA-F]"));

	if ((hex_val = c - 'a') >= 0)
		hex_val += 10;
	else if ((hex_val = c - 'A') >= 0)
		hex_val += 10;
	else
		hex_val = c - '0';

	if (first_bit == 2) {
		if (hex_val > 7)
			return (strdup("[8-9a-fA-F]"));
		else
			return (strdup("[0-7]"));
	} else if (first_bit == 1) {
		if (hex_val > 11)
			return (strdup("[c-fC-F]"));
		else if (hex_val > 7)
			return (strdup("[8-9a-bA-B]"));
		else if (hex_val > 3)
			return (strdup("[4-7]"));
		else
			return (strdup("[0-3]"));
	} else {
		if (hex_val > 13)
			return (strdup("[e-fE-F]"));
		else if (hex_val > 11)
			return (strdup("[c-dC-D]"));
		else if (hex_val > 9)
			return (strdup("[a-bA-B]"));
		else if (hex_val > 7)
			return (strdup("[8-9]"));
		else if (hex_val > 5)
			return (strdup("[6-7]"));
		else if (hex_val > 3)
			return (strdup("[4-5]"));
		else if (hex_val > 1)
			return (strdup("[2-3]"));
		else
			return (strdup("[0-1]"));
	}
}

void reformat_address_to_ipv6(string &ip_addr)
{
	size_t pos = 0, last_pos = 0;
	int ip_num;
	bool even_block = false;
	char ip_hex[3] = {'\0'};
	string new_addr;

	while ((pos = ip_addr.find('.', pos+1)) < ip_addr.length()) {
		ip_num = stoi(ip_addr.substr(last_pos, pos - last_pos));
		sprintf(ip_hex, "%x", ip_num);
		if (ip_num < 16)
			new_addr += "0";
		new_addr += ip_hex;
		if (even_block) {
			new_addr += ":";
			even_block = false;
		} else
			even_block = true;

		last_pos = pos + 1;
	}
	ip_num = stoi(ip_addr.substr(last_pos, ip_addr.length() - last_pos));
	sprintf(ip_hex, "%x", ip_num);
	if (ip_num < 16)
		new_addr += "0";
	new_addr += ip_hex;
	ip_addr.clear();
	ip_addr = new_addr;
	return;
}

void expand_ipv6_address(string &ip_addr)
{
	struct in6_addr tmp;
	char ipv6_addr[INET6_ADDRSTRLEN] = {'\0'};
	const char* char_array = ip_addr.c_str();

	if (inet_pton(AF_INET6, char_array, &tmp) == 1) {
		sprintf(ipv6_addr,
			"%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
		(int)tmp.s6_addr[0], (int)tmp.s6_addr[1],
		(int)tmp.s6_addr[2], (int)tmp.s6_addr[3],
		(int)tmp.s6_addr[4], (int)tmp.s6_addr[5],
		(int)tmp.s6_addr[6], (int)tmp.s6_addr[7],
		(int)tmp.s6_addr[8], (int)tmp.s6_addr[9],
		(int)tmp.s6_addr[10], (int)tmp.s6_addr[11],
		(int)tmp.s6_addr[12], (int)tmp.s6_addr[13],
		(int)tmp.s6_addr[14], (int)tmp.s6_addr[15]);
	}
	ip_addr = ipv6_addr;
}

/* translate ipv4 address struct to that of ipv6 (XXXX:XXXX:XXXX...) and write the regex rule
 * according to the mask provided.
 */
int fix_dot_and_mask(bool is_ipv6, int mask, string &ip_addr)
{
	if (mask > (32 + (is_ipv6 * 96)) || mask < 0) {
		DOCA_DLOG_ERR("netmask %d is not legal for address %s", mask, ip_addr.c_str());
		return -1;
	}
	if (!is_ipv6)
		reformat_address_to_ipv6(ip_addr);
	else
		expand_ipv6_address(ip_addr);
	if (mask == 0)
		return 0;

	size_t last_pos = 0, pos = 0;
	char delim = ':';
	/* calculate which blocks in the ip address are masked */
	int first_masked_blocks = 0, first_digit = 0, first_bit = 0, block_num = 0;
	int max_blocks = is_ipv6 ? 8 : 2;
	bool passed_first_block = false;
	char *tmp_str;
	string new_addr_str = "", rest_of_addr;

	/* find the first masked block in the address in form X.X.X.X.X.X.X.X */
	first_masked_blocks = mask / 16;
	/* find the first digit in the first block being masked, in blocks of FFFF */
	first_digit = (mask % 16) / 4;
	first_bit = 3 - (mask % 4);

	while ((pos = ip_addr.find(delim, pos + 1)) < ip_addr.length()) {
		if (block_num == first_masked_blocks) {
			/* every block has 4 chars to represent 16 bits */
			for (int i = 0 ; i < 4 ; i++) {
				if (i < first_digit) {
					new_addr_str += ip_addr.at(last_pos + i);
					continue;
				} else if (i == first_digit) {
					tmp_str = char_to_range(first_bit, ip_addr.at(last_pos + i));
					new_addr_str += tmp_str;
					free(tmp_str);
					passed_first_block = true;
				} else {
					new_addr_str += "[0-9a-fA-F]{" + to_string(4-i) + "}";
					break;
				}
			}
			ip_addr.replace(last_pos, pos - last_pos, new_addr_str);
			pos = ip_addr.find(delim, pos);
		}
		else if (passed_first_block) {
			if (max_blocks > block_num + 1)
				rest_of_addr = "([0-9a-fA-F]{4}:){" +
							to_string(max_blocks - block_num - 1) + "}";
			rest_of_addr += "[0-9a-fA-F]{4}";
			ip_addr.replace(last_pos, ip_addr.length() - last_pos, rest_of_addr);
			break;
		}
		block_num++;
		last_pos = pos + 1;
	}
	if (block_num == first_masked_blocks) {
		/* every block has 4 chars to represent 16 bits */
		for (int i = 0 ; i < 4 ; i++) {
			if (i < first_digit) {
				new_addr_str += ip_addr.at(last_pos + i);
				continue;
			} else if (i == first_digit) {
				tmp_str = char_to_range(first_bit, ip_addr.at(last_pos + i));
				new_addr_str += tmp_str;
				free(tmp_str);
				passed_first_block = true;
			} else {
				new_addr_str += "[0-9a-fA-F]{" + to_string(4-i) + "}";
				break;
			}
		}
		ip_addr.replace(last_pos, pos - last_pos, new_addr_str);
	}
	return 0;
}

/*
 * concatenate metadata to the end of the regex signature.
 * order is: eth_type:src_port:src_ip:dst_pot:dst_ip:field:l4_prot:l7_prot.
 */
static int add_regex_metadata_patch(struct group_property *metadata, string &regex,
					int field, int l7_prot, bool dollar,
					int direction)
{
	int i;
	string any_addr = "(([0-9a-fA-F]{1,4}:[0-9a-fA-F]{1,4})|(([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}))";
	char *tmp_str;
	bool eth_been = metadata->eth_type > 0;
	bool src_adr_been = eth_been | (!metadata->src_addr.empty());
	bool src_port_been = src_adr_been | (metadata->src_port_start > 0);
	bool dst_adr_been = src_port_been | (!metadata->dest_addr.empty());
	bool dst_port_been = dst_adr_been | (metadata->dest_port_start > 0);
	bool dir_been = dst_port_been | (direction != DOCA_DPI_BOTH);
	bool l4_prot_been = dir_been | (metadata->l4_prot > 0);

	if (!dollar)
		regex.append(".*");
	regex.append("(&&&&");

	for (i = 0 ; i < MAX_METADATA_ENUM ; i++) {
		switch (metadata_order[i]) {
		case SRC_PORT:
			if (!src_port_been) {
				regex += ")(&[^&]+){" + to_string(MAX_METADATA_ENUM - i) + "}$";
				goto end;
			}
			if (metadata->src_port_start > 0) {
				regex.append("&(");
				tmp_str = port_rang_to_string(metadata->src_port_start,
							      metadata->src_port_end);
				if (tmp_str == NULL)
					return -1;
				regex.append(tmp_str);
				free(tmp_str);
			} else
				regex.append("&([0-9]{1,6}");
			regex.append(")");
			break;
		case DEST_PORT:
			if (!dst_port_been) {
				regex += ")(&[^&]+){" + to_string(MAX_METADATA_ENUM - i) + "}$";
				goto end;
			}
			if (metadata->dest_port_start > 0) {
				regex.append("&(");
				tmp_str = port_rang_to_string(metadata->dest_port_start,
							      metadata->dest_port_end);
				if (tmp_str == NULL)
					return -1;
				regex.append(tmp_str);
				free(tmp_str);
			} else
				regex.append("&([0-9]{1,6}");
			regex.append(")");
			break;
		case SRC_ADDR:
			if (!src_adr_been) {
				regex += ")(&[^&]+){" + to_string(MAX_METADATA_ENUM - i) + "}$";
				goto end;
			}
			if (!metadata->src_addr.empty()) {
				regex.append("&(");
				if (!metadata->update_done)
					fix_dot_and_mask(metadata->src_is_ipv6, metadata->src_mask,
							 metadata->src_addr);
				regex.append(metadata->src_addr);
			} else {
				regex.append("&(");
				regex.append(any_addr);
			}
			regex.append(")");
			break;
		case DEST_ADDR:
			if (!dst_adr_been) {
				regex += ")(&[^&]+){" + to_string(MAX_METADATA_ENUM - i) + "}$";
				goto end;
			}
			if (!metadata->dest_addr.empty()) {
				regex.append("&(");
				if (!metadata->update_done)
					fix_dot_and_mask(metadata->dest_is_ipv6,
							 metadata->dest_mask, metadata->dest_addr);
				regex.append(metadata->dest_addr);
			} else {
				regex.append("&(");
				regex.append(any_addr);
			}
			regex.append(")");
			break;
		case L7_PROT:
			regex.append("&(");
			if (l7_prot_to_str.count(l7_prot) > 0) {
				regex.append(l7_prot_to_str.at(l7_prot));
			} else {
				regex.append("l4"); /* for second regex run only on l3_l4 stream */
			}
			regex.append(")");
			break;
		case FIELD:
			regex.append("&(");
			if (field > 0 && field_to_str.count(field) > 0) {
				regex.append(field_to_str.at(field));
			} else {
				regex.append("[a-z_]{1,20}");
			}
			regex.append(")");
			break;
		case ETH_TYPE:
			if (!eth_been) {
				regex += ")(&[^&]+){" + to_string(MAX_METADATA_ENUM - i) + "}$";
				goto end;
			}
			regex.append("&(");
			if (metadata->eth_type > 0)
				regex.append(to_string(metadata->eth_type));
			else
				regex.append("[0-9]{1,5}");
			regex.append(")");
			break;
		case L4_PROT:
			if (!l4_prot_been) {
				regex += ")(&[^&]+){" + to_string(MAX_METADATA_ENUM - i) + "}$";
				goto end;
			}
			regex.append("&(");
			if (metadata->l4_prot > 0)
				regex.append(to_string(metadata->l4_prot));
			else
				regex.append("[0-9]{1,5}");
			regex.append(")");
			break;
		case DIRECTION:
			if (!dir_been) {
				regex += ")(&[^&]+){" + to_string(MAX_METADATA_ENUM - i) + "}$";
				goto end;
			}
			regex.append("&(");
			if (direction == DOCA_DPI_CTS)
				regex.append("cts");
			else if (direction == DOCA_DPI_STC)
				regex.append("stc");
			else
				regex.append("[a-z]{3}");
			regex.append(")");
			break;
		}
	}
	regex.append(")$");
end:
	metadata->update_done = true;
	return 0;
}

/* extract layer 7 protocol from a given field */
static int
extract_l7prot(int field)
{
	if (field < DOCA_DPI_MAX_HTTP_FIELD)
		return DOCA_DPI_HTTP;
	else if (field < DOCA_DPI_MAX_SSL_FIELD)
		return DOCA_DPI_SSL;
	else if (field < DOCA_DPI_MAX_RTMP_FIELD)
		return DOCA_DPI_RTMP;
	else if (field < DOCA_DPI_MAX_STREAM_FIELD)
		return 0;
	else if (field < DOCA_DPI_MAX_DNS_FIELD)
		return DOCA_DPI_DNS;
	else if (field < DOCA_DPI_MAX_STUN_FIELD)
		return DOCA_DPI_STUN;
	else if (field < DOCA_DPI_MAX_IP_FIELD)
		return DOCA_DPI_DNS;
	return 0;
}

/* insert new rule field and update regex sub-groups
 */
int doca_dpi_compiler_add_rule(struct doca_dpi_compiler *ctx, int sig_id, int group_id,
			       int direction, int rule_field, union rule_value *val)
{
	if (!(ctx->sig_map.count(sig_id))) {
		DOCA_DLOG_ERR("Error: sig id provided for rule is not exist in the compiler's context\n");
		return -1;
	}

	struct regex_obj_fields reg_fields = {};
	string regex;
	bool is_ipv6 = false;
	struct in6_addr tmp_buf;

	switch(rule_field) {
	case DOCA_DPI_REGEX:
		ctx->num_regex++;
		/* insert regex to sub-group map in compiler ctx */
		regex = val->reg_field.regex;

		if (val->reg_field.offset > 0) {
			if (regex.find('^') < 2) { /* look for "^" in start index or after "/" */
				regex.erase(regex.find('^'), 1);
				reg_fields.anchor = true;
			} else
				reg_fields.anchor = false;
			reg_fields.offset = val->reg_field.offset;
		} else
			reg_fields.offset = 0;

		if (regex.back() == '$')
			reg_fields.dollar = true;
		else if (regex.at(0) == '/')
			if (regex.at(regex.rfind('/') - 1) == '$')
				reg_fields.dollar = true;

		/* fix the regex string to RXP format */
		reformat_regex(regex, reg_fields.dollar);
		reg_fields.regex = regex;
		reg_fields.regex_id = ctx->num_regex;
		reg_fields.field = val->reg_field.field;
		reg_fields.l7_prot = extract_l7prot(val->reg_field.field);
		reg_fields.direction = direction;
		(ctx->group2regex_list_map[sig_id])[group_id].push_back(reg_fields);

		break;
	case DOCA_DPI_PORT:
		ctx->num_ports++;

		if (ctx->group_prop_map[sig_id].count(group_id) == 0)
			(ctx->group_prop_map[sig_id])[group_id] = {0};
		(ctx->group_prop_map[sig_id])[group_id].direction = direction;
		if (strcmp(val->port_field.side, "client")) {
			(ctx->group_prop_map[sig_id])[group_id].dest_port_start = val->port_field.start_port;
			(ctx->group_prop_map[sig_id])[group_id].dest_port_end = val->port_field.end_port;
		} else {
			(ctx->group_prop_map[sig_id])[group_id].src_port_start = val->port_field.start_port;
			(ctx->group_prop_map[sig_id])[group_id].src_port_end = val->port_field.end_port;
		}
		break;
	case DOCA_DPI_ADDRESS:
		ctx->num_addresses++;
		if (inet_pton(AF_INET6, val->addr_field.ip_address, &tmp_buf))
			is_ipv6 = true;

		if (ctx->group_prop_map[sig_id].count(group_id) == 0)
			(ctx->group_prop_map[sig_id])[group_id] = {0};
		(ctx->group_prop_map[sig_id])[group_id].direction = direction;
		if (strcmp(val->addr_field.side, "client")) {
			(ctx->group_prop_map[sig_id])[group_id].dest_addr = val->addr_field.ip_address;
			(ctx->group_prop_map[sig_id])[group_id].dest_mask =
					val->addr_field.ip_mask < 0 ? 0 : val->addr_field.ip_mask;
			(ctx->group_prop_map[sig_id])[group_id].dest_is_ipv6 = is_ipv6;
		} else {
			(ctx->group_prop_map[sig_id])[group_id].src_addr = val->addr_field.ip_address;
			(ctx->group_prop_map[sig_id])[group_id].src_mask =
					val->addr_field.ip_mask < 0 ? 0 : val->addr_field.ip_mask;
			(ctx->group_prop_map[sig_id])[group_id].src_is_ipv6 = is_ipv6;
		}
		/* poc */
		break;
	case DOCA_DPI_PRIORITY:
		if (ctx->group_prop_map[sig_id].count(group_id) == 0)
			(ctx->group_prop_map[sig_id])[group_id] = {0};
		(ctx->group_prop_map[sig_id])[group_id].order1 = val->priority.prior1;
		(ctx->group_prop_map[sig_id])[group_id].order2 = val->priority.prior2;
		(ctx->group_prop_map[sig_id])[group_id].bitmask =
			      (ctx->group_prop_map[sig_id])[group_id].bitmask | DOCA_DPI_PRIOR_MASK;
		break;
	case DOCA_DPI_L7_PROT:
		/* for use of layer 7 stream in future use */
		if (ctx->group_prop_map[sig_id].count(group_id) == 0)
			(ctx->group_prop_map[sig_id])[group_id] = {0};
		(ctx->group_prop_map[sig_id])[group_id].l7_prot = val->primitive;
		break;
	case DOCA_DPI_IP_PROT:
		/* poc */
		if (ctx->group_prop_map[sig_id].count(group_id) == 0)
			(ctx->group_prop_map[sig_id])[group_id] = {0};
		(ctx->group_prop_map[sig_id])[group_id].l4_prot = val->primitive;
		/* poc */
		break;
	case DOCA_DPI_ETH_PROT:
		/* poc */
		if (ctx->group_prop_map[sig_id].count(group_id) == 0)
			(ctx->group_prop_map[sig_id])[group_id] = {0};
		(ctx->group_prop_map[sig_id])[group_id].eth_type = val->primitive;
		/* poc */
		break;
	}
	return 0;
}

static int create_dir();
static int erase_dir();
static void create_rules_file(map < int, map < int, string > > *subgrp2reg_map);
static int create_rof2_file();
static int build_cdo(struct doca_dpi_compiler *ctx, struct doca_dpi_compiler_in_args *args);
static int add_field_suffix_to_regex(struct doca_dpi_compiler *ctx);

int doca_dpi_compiler_save(struct doca_dpi_compiler *ctx, struct doca_dpi_compiler_in_args *args)
{
	if (add_field_suffix_to_regex(ctx))
		return -1;
	if (create_dir())
		return -1;
	create_rules_file(&(ctx->subgrp2reg_map));
	if (create_rof2_file())
		return -1;
	if (build_cdo(ctx, args)) {
		return -1;
	}
	if (erase_dir())
		return -1;

	return 0;
}

static int add_field_suffix_to_regex(struct doca_dpi_compiler *ctx)
{
	int sig_id, group_id;
	string no_regex;
	struct rule_obj val_obj;

	for (auto sig_lvl : (ctx->group_prop_map)) {
		sig_id = sig_lvl.first;
		for (auto grp_lvl : sig_lvl.second) {
			group_id = grp_lvl.first;
			if ((ctx->group2regex_list_map[sig_id])[group_id].empty()) {
				no_regex = "";
				if (add_regex_metadata_patch(&grp_lvl.second, no_regex,
							 DOCA_DPI_NO_REGEX, 0,
							 true, grp_lvl.second.direction))
					return -1;
				ctx->num_regex++;
				val_obj.json = json_object_new_object();
				ADD_INT_OBJ(ctx->num_regex, "regex_id", val_obj.json);
				ADD_STRING_OBJ(no_regex.c_str(), "regex", val_obj.json);
				ADD_INT_OBJ(0, "offset", val_obj.json);
				ADD_BOOL_OBJ(false, "anchor", val_obj.json);
				ADD_INT_OBJ(grp_lvl.second.l7_prot, "l7_protocol", val_obj.json);
				ADD_INT_OBJ(grp_lvl.second.l4_prot, "l4_protocol", val_obj.json);
				val_obj.type = REGEX_TYPE;
				(ctx->grp2rules_map[sig_id])[group_id].push_back(val_obj);
				(ctx->subgrp2reg_map[DOCA_DPI_NO_REGEX])[ctx->num_regex] = no_regex;
			} else {
				for (auto regex_rule :
						(ctx->group2regex_list_map[sig_id])[group_id]) {
					if (add_regex_metadata_patch(&grp_lvl.second,
									regex_rule.regex,
									regex_rule.field,
									regex_rule.l7_prot,
									regex_rule.dollar,
									regex_rule.direction))
						return -1;
					val_obj.json = json_object_new_object();
					ADD_INT_OBJ(regex_rule.regex_id, "regex_id", val_obj.json);
					ADD_STRING_OBJ(regex_rule.regex.c_str(), "regex",
										val_obj.json);
					ADD_INT_OBJ(regex_rule.offset, "offset", val_obj.json);
					ADD_BOOL_OBJ(regex_rule.anchor, "anchor", val_obj.json);
					ADD_INT_OBJ(grp_lvl.second.l7_prot, "l7_protocol", val_obj.json);
					ADD_INT_OBJ(grp_lvl.second.l4_prot, "l4_protocol", val_obj.json);
					val_obj.type = REGEX_TYPE;
					(ctx->grp2rules_map[sig_id])[group_id].push_back(val_obj);
					(ctx->subgrp2reg_map[regex_rule.field])[regex_rule.regex_id]
										= regex_rule.regex;
				}
			}
		}
	}
	return 0;
}

static long read_rules_db(const char *file, char **buf);

static bool version_check(struct json_object *maj_obj, struct json_object *min_obj,
				       struct json_object *p_obj);
static int parse_app_metadata(struct doca_dpi_compiler *ctx, struct json_object *app_metadata);
static int parse_pp_table(struct doca_dpi_compiler *ctx, struct json_object *pp_table);

int doca_dpi_compiler_load(struct doca_dpi_compiler *ctx, const char *cdo)
{
	char *buffer = NULL;
	long ret;
	struct json_object *parsed_json;
	struct json_object *minor_ver;
	struct json_object *major_ver;
	struct json_object *patch_ver;
	struct json_object *pp_table;
	struct json_object *app_metadata;

	ret = read_rules_db(cdo, &buffer);
	if (ret < 0)
		return ret;

	parsed_json = json_tokener_parse(buffer);
	if (parsed_json == NULL) {
		DOCA_DLOG_ERR("Error: couldn't parse the input file as json tokens\n");
		ret = -1;
		goto out;
	}

	if (!json_object_object_get_ex(parsed_json, "minor_version",
				       &minor_ver)) {
		DOCA_DLOG_ERR("Error: couldn't parse minor_version object");
		ret = -1;
		goto out;
	}
	if (!json_object_object_get_ex(parsed_json, "major_version",
				       &major_ver)) {
		DOCA_DLOG_ERR("Error: couldn't parse major_version object");
		ret = -1;
		goto out;
	}
	if (!json_object_object_get_ex(parsed_json, "patch_version",
				       &patch_ver)) {
		DOCA_DLOG_ERR("Error: couldn't parse patch_version object");
		ret = -ENOTSUP;
		goto out;
	}
	if (version_check(major_ver, minor_ver, patch_ver)) {
		ret = -ENOTSUP;
		goto out;
	}

	if (!json_object_object_get_ex(parsed_json, "app_metadata",
				       &app_metadata)) {
		DOCA_DLOG_ERR("Error: couldn't parse app_metadata object. signatures must be provided\n");
		goto out;
	}
	if (parse_app_metadata(ctx, app_metadata)) {
		ret = -1;
		goto out;
	}

	/* check for regex rules */
	if (!json_object_object_get_ex(parsed_json, "pp_table", &pp_table)) {
		DOCA_DLOG_ERR("Error: couldn't parse pp_table object");
		ret = -1;
		goto out;
	}
	if (parse_pp_table(ctx, pp_table)) {
		ret = -1;
		goto out;
	}

out:
	free(buffer);
	return ret;
}

#ifdef DOCA_LOGGING_ALLOW_DLOG
static const char*
print_json_object(json_object *json_elem)
{
	switch (json_object_get_type(json_elem)) {
	case json_type_int:
		return "json_type_int";
	case json_type_boolean:
		return "json_type_boolean";
	case json_type_double:
		return "json_type_double";
	case json_type_string:
		return "json_type_string";
	case json_type_array:
		return "json_type_array";
	case json_type_object:
		return "json_type_object";
	case json_type_null:
		return "json_type_null";
	default:
		return "\nError: wrong json_object\n";
	}
}
#endif

static int
create_dir()
{
	string pid_dir = "/tmp/" + to_string(getpid()) + "/";
	string mkdir = "mkdir " + pid_dir;
	if (system(mkdir.c_str())) {
		DOCA_DLOG_ERR("Error: mkdir command \"%s\" failed\n", mkdir.c_str());
		return 1;
	}
	string rof_dir = pid_dir + "rof/";
	mkdir = "mkdir " + rof_dir;
	if (system(mkdir.c_str())) {
		DOCA_DLOG_ERR("Error: mkdir command \"%s\" failed\n", mkdir.c_str());
		return 1;
	}
	return 0;
}

static int
erase_dir()
{
	string pid_dir = "/tmp/" + to_string(getpid()) + "/";
	string rm = "rm -rf " + pid_dir;
	if (system(rm.c_str())) {
		DOCA_DLOG_ERR("Error: remove of directory %s failed\n", pid_dir.c_str());
		return 1;
	}
	return 0;
}

static void
compile_hs_db(vector<const char *> & regex, vector<unsigned> & id)
{
#ifdef DOCA_DPI_USE_HS
	string rof_dir = "/tmp/" + to_string(getpid()) + "/";
	ofstream rules_file(rof_dir + HS_RULES_FILE_NAME);
	if (!rules_file.is_open()) {
		DOCA_DLOG_ERR("rules file %s didn't open\n", (rof_dir + RULES_FILE_NAME).c_str());
		cerr << strerror(errno) << endl;
		return;
	}

	string converted, tmp_rule;
	string converter_err;
	hs_database *db;
	hs_compile_error_t *err;
	char *serialized_db;
	size_t size;
	vector<unsigned int> flags;

	cout << rof_dir + HS_RULES_FILE_NAME << endl;
	printf("================************=================\n");
	for (auto &it : regex) {
		flags.push_back(HS_FLAG_DOTALL);
		hs_compile(it, HS_FLAG_DOTALL, HS_MODE_BLOCK, NULL,  &db, &err);
		if (err != HS_SUCCESS) {
			DOCA_DLOG_ERR("Compile id failed: %s\n", err->message);
			continue;
		}
		hs_free_database(db);
	}
	cout << "compiling all rules..." << endl;
	hs_compile_multi(regex.data(), flags.data(), id.data(), regex.size(), HS_MODE_BLOCK, NULL, &db, &err);
	if (err != HS_SUCCESS) {
		if (err && err->message) {
			DOCA_DLOG_ERR("Compile failed: %s\n", err->message);
		} else {
			DOCA_DLOG_ERR("hs_compile_multi failed on expression number %d\n",
					err->expression);
		}
		hs_free_compile_error(err);
		goto compile_multi_err;
	}
	hs_serialize_database(db, &serialized_db, &size);
	printf("%ld ================************=================\n", size);

	rules_file.write(serialized_db, size);
	free(serialized_db);
compile_multi_err:
	hs_free_database(db);
	rules_file.close();
	return;
#endif
}

/* create ruleset file for the rxpc
 */
static void
create_rules_file(map < int, map < int, string > > *subgrp2reg_map)
{
	vector<const char *> regex_vec;
	vector<unsigned> id_vec;
	set <string> seen_regex;
	string rof_dir = "/tmp/" + to_string(getpid()) + "/";
	ofstream rules_file(rof_dir + RULES_FILE_NAME);
	if (!rules_file.is_open()) {
		DOCA_DLOG_ERR("Error rules file %s didn't open\n",
				      (rof_dir + RULES_FILE_NAME).c_str());
		cerr << strerror(errno) << endl;
		return;
	}
	if (subgrp2reg_map == NULL) {
		rules_file.close();
		return;
	}

	cout << rof_dir + RULES_FILE_NAME << endl;
	for (auto &subgrp : (*subgrp2reg_map)) {
		// rules_file << "subset_id=" << subgrp.first << endl;
		for (auto &regex : subgrp.second) {
			if (seen_regex.count(regex.second))
				continue;
			seen_regex.insert(regex.second);
			rules_file << regex.first << ", " << regex.second << endl;
			regex_vec.push_back(regex.second.c_str());
			id_vec.push_back(regex.first);
		}

	}
	compile_hs_db(regex_vec, id_vec);
	seen_regex.clear();
	rules_file.close();
	return;
}

void
print_regex_status_error(int ret, const char *func_name)
{
	if (ret == REGEX_STATUS_NO_RULES_FOUND)
		DOCA_DLOG_ERR("%s failed, no rules have been found\n", func_name);
	else if (ret == REGEX_STATUS_SYNTAX_ERROR)
		DOCA_DLOG_ERR("%s failed, syntax error was found\n", func_name);
	else if (ret == REGEX_STATUS_FATAL_ERROR)
		DOCA_DLOG_ERR("%s failed, compilation was aborted due to a fatal error\n",
			func_name);
	else
		DOCA_DLOG_ERR("%s failed, unknown error\n", func_name);
}

/* sending command to the rxpc to create therof file
 */
static int
create_rof2_file()
{
	string pid_dir = "/tmp/" + to_string(getpid()) + "/";
	string rof_dir = pid_dir + "rof/";
	string input_file_str = pid_dir + RULES_FILE_NAME;
	string rof_out_str = rof_dir + ROF_FILE_NAME;

	struct regex_ruleset *regex_ruleset = NULL;
	struct regex_prefix_selection_control_list *regex_prefix_selection_control_list = NULL;
	uint32_t compiler_options = REGEX_COMPILER_OPTIONS_DISABLE_BIDIRECTIONAL; /* similar to -W */
	//uint32_t global_regex_options = REGEX_GLOBAL_REGEX_OPTIONS_NO_SINGLE; /* similar to -S */
	uint32_t compiler_output_flags = 0;
	struct regex_compilation_statistics *compilation_statistics = NULL;
	struct regex_compiled_rules_log *compiled_rules_log = NULL;
	struct regex_critical_rules_rank *critical_rules_rank = NULL;
	struct regex_uncompiled_rules_log *uncompiled_rules_log = NULL;
	struct regex_rof *rof = NULL;
	struct regex_roff *roff = NULL;
	enum regex_version version = REGEX_VERSION_V5_7; /* similar to -V */
	int threads = 1;
	char *input_file, *output_file, *ruleset_errors = NULL;
	int ret = 0, i = 0, j = 0;

	/* read rules file */
	input_file = strdup(input_file_str.c_str());
	if (input_file == NULL) {
		DOCA_DLOG_ERR("Error: strdup failed for input file name\n");
		return -1;
	}
	regex_ruleset = (struct regex_ruleset*) calloc (1, sizeof(struct regex_ruleset));
	if (regex_ruleset == NULL) {
		DOCA_DLOG_ERR("Error: calloc failed for regex ruleset\n");
		goto alloc_err;
	}
	printf("rules file is %s\n",input_file);
	ret = regex_read_ruleset_file(input_file, regex_ruleset, &ruleset_errors);
	if (ret != REGEX_STATUS_OK) {
		print_regex_status_error(ret, "regex_read_ruleset_file");
		ret = -1;
		goto compile_err;
	}

	/* create rof struct */
	// ret = regex_compile(&regex_ruleset, regex_prefix_selection_control_list, /* INPUT */
	// 		    compiler_options, global_regex_options, threads, version, /* OPTIONS */
	// 		    compiler_output_flags, /* OPTIONS */
	// 		    &compilation_statistics, &compiled_rules_log, /* OUTPUT */
	// 		    &critical_rules_rank, &uncompiled_rules_log, &rof, &roff /* OUTPUT */);

	ret = regex_compile_advanced(regex_ruleset, regex_prefix_selection_control_list, NULL, NULL, NULL, NULL, NULL, /* INPUT */
			    compiler_options, REGEX_PREFIX_CAPACITY_32K, 0, 0, 0, REGEX_COMPILER_OBJECTIVE_5, 10, 65535,
			    threads, version, /* OPTIONS */
			    REGEX_VERBOSE_LEVEL_3, REGEX_TPE_DATA_WIDTH_4, REGEX_VIRTUAL_PREFIX_MODE_0, compiler_output_flags,
			    1, /* OPTIONS */
			    &compilation_statistics, &compiled_rules_log, /* OUTPUT */
			    &critical_rules_rank, &uncompiled_rules_log, &rof, &roff, /* OUTPUT */
			    NULL,NULL );

	if (ret != REGEX_STATUS_OK) {
		print_regex_status_error(ret, "regex_compile");
		ret = -1;
		goto compile_err;
	}

	/* write rof file */
	output_file = strdup(rof_out_str.c_str());
	if (output_file == NULL) {
		DOCA_DLOG_ERR("Error: strdup failed for output file name\n");
		ret = -1;
		goto err_rof;
	}

	ret = regex_write_rof_file(output_file, rof);
	if (ret != REGEX_STATUS_OK) {
		print_regex_status_error(ret, "regex_write_rof_file");
		ret = -1;
		goto err_out;
	}

	ret = 0;
err_out:
	free(output_file);
err_rof:
	if (regex_free_structs(NULL /* regex_compilation_statistics */,
				  NULL /* regex_compiled_rules_log */,
				  NULL /* regex_critical_rules_rank */,
				  NULL /* regex_uncompiled_rules_log */,
				  &rof,
				  NULL /* regex_roff */,
				  NULL /* regex_rofi */,
				  NULL /* regex_prefix_selection_control_list */,
				  1 /* divide_ruleset */) != REGEX_STATUS_OK) {
		DOCA_DLOG_ERR("regex_free_structs failed\n");
		ret = -1;
	}
compile_err:
	if (ruleset_errors != NULL)
		free(ruleset_errors);
	if (regex_ruleset != NULL) {
		for (i = 0; i < regex_ruleset->number_of_entries; i++) {
			free(regex_ruleset->rules[i].rule);
			for (j = 0; j < regex_ruleset->rules[i].number_of_prefix_entries; j++)
				free(regex_ruleset->rules[i].prefix[j]);
			free(regex_ruleset->rules[i].prefix);
		}
		if (regex_ruleset->number_of_entries > 0)
			free(regex_ruleset->rules);
		free(regex_ruleset);
	}
alloc_err:
	free(input_file);
	return ret;
}

static void
create_app_metadata_array(struct json_object *app_metadata,
			  map < int, struct doca_dpi_sig_info > *sig_map)
{
	struct json_object *mdata_obj;

	for (auto &it : (*sig_map)) {
		mdata_obj = json_object_new_object();
		ADD_INT_OBJ(it.first, "sig_id", mdata_obj);
		ADD_STRING_OBJ(it.second.name, "name", mdata_obj);
		ADD_STRING_OBJ(it.second.action, "action", mdata_obj);
		json_object_array_add(app_metadata, mdata_obj);
	}
}

/* add signature and regex_id object to the post processing table json array.
 */
static int
create_sig_id_rules(int sig_id, struct json_object * pp_table,
		    map < int, list < struct rule_obj > > *group2rule_map,
		    map < int, struct group_property > *property_map, int & conds_num)
{
	struct json_object *pp_entry;
	struct json_object *group_entry;
	struct json_object *priority_obj;
	struct json_object *regex_arr;
	string addr;
	string port;
	int regex_per_group;

	for (auto grp : *group2rule_map) {
		pp_entry = json_object_new_object();
		group_entry = json_object_new_object();
		regex_arr = json_object_new_array();
		regex_per_group = 0;

		if ((*property_map)[grp.first].bitmask & DOCA_DPI_PRIOR_MASK) {
			priority_obj = json_object_new_array();
			json_object_array_add(priority_obj,
					    json_object_new_int((*property_map)[grp.first].order1));
			json_object_array_add(priority_obj,
					    json_object_new_int((*property_map)[grp.first].order2));
			json_object_object_add(group_entry, "priority", priority_obj);
		}

		for (auto rule_obj : grp.second) {
			switch (rule_obj.type) {
			case REGEX_TYPE:
				json_object_array_add(regex_arr, rule_obj.json);
				regex_per_group++;
				break;
			default:
				DOCA_DLOG_ERR("Error: rule object with unrecognized type: %d\n",
						      rule_obj.type);
				goto create_sig_id_rules_err;
			}
		}

		json_object_object_add(group_entry, "rules", regex_arr);
		if (regex_per_group > 1)
			conds_num++;

		/* add pp_entry object to the pp_table */
		ADD_INT_OBJ(sig_id, "sig_id", pp_entry);
		json_object_object_add(pp_entry, "group", group_entry);
		json_object_array_add(pp_table, pp_entry);
	}
	return 0;
create_sig_id_rules_err:
	json_object_put(pp_entry);
	json_object_put(group_entry);
	json_object_put(regex_arr);
	return -1;
}

static void build_patch_order(struct json_object *metadata_patch_order)
{
	int i;

	for (i = 0 ; i < MAX_METADATA_ENUM ; i++) {
		switch (metadata_order[i]) {
		case SRC_PORT:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("src_port"));
			break;
		case DEST_PORT:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("dest_port"));
			break;
		case SRC_ADDR:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("src_addr"));
			break;
		case DEST_ADDR:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("dest_addr"));
			break;
		case L7_PROT:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("l7_protocol"));
			break;
		case FIELD:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("field"));
			break;
		case L4_PROT:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("l4_protocol"));
			break;
		case ETH_TYPE:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("ethernet_type"));
			break;
		case DIRECTION:
			json_object_array_add(metadata_patch_order,
						json_object_new_string("direction"));
			break;
		}
	}
}

void
build_prot2num_table(struct json_object *prot2num_obj)
{
	ADD_INT_OBJ(DOCA_DPI_HTTP, "http", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_HTTP2, "http2", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_TLS, "tls", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_SSL, "ssl", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_SIP, "sip", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_RTMP, "rtmp", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_QUIC, "quic", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_STUN, "stun", prot2num_obj);
	ADD_INT_OBJ(DOCA_DPI_DNS, "dns", prot2num_obj);
}

/* create CDO_output file from all data gathered from rules file
 */
static int
build_cdo(struct doca_dpi_compiler *ctx, struct doca_dpi_compiler_in_args *args)
{
	FILE *fp;
	struct json_object *decoded_json = json_object_new_object();
	struct json_object *minor_ver = json_object_new_int(DOCA_VER_MINOR);
	struct json_object *major_ver = json_object_new_int(DOCA_VER_MAJOR);
	struct json_object *patch_ver = json_object_new_int(DOCA_VER_PATCH);
	struct json_object *pp_table = json_object_new_array();
	struct json_object *app_metadata = json_object_new_array();
	struct json_object *category_metadata = json_object_new_array();
	struct json_object *family_metadata = json_object_new_array();
	struct json_object *metadata_patch_order = json_object_new_array();
	struct json_object *prot2num_obj = json_object_new_object();
	struct json_object *regex_bin_arr = NULL;
	struct json_object *hs_regex_obj = json_object_new_array();
	int ret = 0;
	bool rxp_bin = false;
	char c;

	/* handle rof2 file */
	ifstream rof2_file("/tmp/" + to_string(getpid()) + "/rof/" + ROF_FILE_NAME);
	string rof2_content((istreambuf_iterator<char>(rof2_file)), (istreambuf_iterator<char>()));
	struct json_object *regex_obj = json_object_new_string(rof2_content.c_str());
	rof2_file.close();

	/* handle rof2 binary file */
	ifstream rof2_file_bin("/tmp/" + to_string(getpid()) + "/rof/" + ROF_FILE_NAME +
				".binary", ios::binary);
	if (!rof2_file_bin) {
		DOCA_DLOG_DBG("rof.binary file is not available\n");
	} else if (!(rof2_file_bin.peek() == std::ifstream::traits_type::eof())) {
		regex_bin_arr = json_object_new_array();
		while (rof2_file_bin.get(c))
			json_object_array_add(regex_bin_arr, json_object_new_int(c));
		rof2_file_bin.close();
		rxp_bin = true;
	}

	/* handle hyperscan file */
	ifstream hs_file("/tmp/" + to_string(getpid()) + "/" + HS_RULES_FILE_NAME);
	while (hs_file.get(c)) {
		struct json_object *hs_regex_c = json_object_new_int(c);

		json_object_array_add(hs_regex_obj, hs_regex_c);
	}
	hs_file.close();

	fp = fopen(args->cdo_file_name, "w");
	if (!fp) {
		DOCA_DLOG_ERR("Error: can't open file %s, err = %d\n", args->cdo_file_name, EIO);
		ret = EIO;
		goto err_out;
	}

	create_app_metadata_array(app_metadata, &ctx->sig_map);

	for (auto i : (ctx->grp2rules_map)) {
		if (create_sig_id_rules(i.first, pp_table, &(i.second),
					&ctx->group_prop_map[i.first], ctx->num_conds) < 0) {
			DOCA_DLOG_ERR("Error: create_group_obj failed\n");
			ret = 1;
			goto err_out;
		}
	}

	DOCA_DLOG_DBG("All signatures been transfer to json_object\n");

	/* poc */
	build_patch_order(metadata_patch_order);
	build_prot2num_table(prot2num_obj);

	json_object_object_add(decoded_json, "metadata_patch_order", metadata_patch_order);
	json_object_object_add(decoded_json, "protocol_to_number", prot2num_obj);
	json_object_object_add(decoded_json, "minor_version", minor_ver);
	json_object_object_add(decoded_json, "major_version", major_ver);
	json_object_object_add(decoded_json, "patch_version", patch_ver);
	ADD_INT_OBJ(ctx->num_regex, "number_of_regex_rules", decoded_json);
	ADD_INT_OBJ(ctx->num_conds, "number_of_conditions", decoded_json);
	ADD_INT_OBJ(ctx->num_addresses, "number_of_addresses", decoded_json);
	ADD_INT_OBJ(ctx->num_ports, "number_of_ports", decoded_json);
	json_object_object_add(decoded_json, "pp_table", pp_table);
	json_object_object_add(decoded_json, "app_metadata", app_metadata);
	json_object_object_add(decoded_json, "category_metadata", category_metadata);
	json_object_object_add(decoded_json, "family_metadata", family_metadata);
	json_object_object_add(decoded_json, "regex", regex_obj);
	if (rxp_bin)
		json_object_object_add(decoded_json, "regex_bin", regex_bin_arr);
	json_object_object_add(decoded_json, "hs_regex", hs_regex_obj);

	if (fprintf(fp, "%s",
		    json_object_to_json_string_ext(decoded_json, JSON_C_TO_STRING_PRETTY)) < 0) {
		ret = EIO;
		DOCA_DLOG_ERR("Error: couldn't write to file %s\n, err = %d", args->cdo_file_name, EIO);
		goto out;
	}
	ret = 0;
	goto out;
err_out:
	json_object_put(minor_ver);
	json_object_put(major_ver);
	json_object_put(patch_ver);
	json_object_put(pp_table);
	json_object_put(app_metadata);
	json_object_put(category_metadata);
	json_object_put(family_metadata);
	json_object_put(metadata_patch_order);
	json_object_put(prot2num_obj);
	json_object_put(regex_obj);
	if (rxp_bin)
		json_object_put(regex_bin_arr);
	json_object_put(hs_regex_obj);
out:
	json_object_put(decoded_json);
	if (fp != NULL)
		fclose(fp);
	return -ret;
}

static long
read_rules_db(const char *file, char **buf)
{
	FILE *fp;
	long buf_len = 0;
	size_t read_len;
	int res = 0;
	*buf = NULL;

	fp = fopen(file, "r");
	if (!fp) {
		DOCA_DLOG_ERR("Error, can't open file %s, err = %d\n", file, errno);
		return -EIO;
	}
	if (fseek(fp, 0L, SEEK_END) == 0) {
		buf_len = ftell(fp);
		if (buf_len == -1) {
			DOCA_DLOG_ERR("Error: ftell function fail\n");
			res = EIO;
			goto error;
		}
		*buf = (char *)malloc(sizeof(char)*(buf_len + 1));
		if (!*buf) {
			DOCA_DLOG_ERR("Error: malloc function fail\n");
			res = ENOMEM;
			goto error;
		}
		if (fseek(fp, 0L, SEEK_SET) != 0) {
			DOCA_DLOG_ERR("Error: fseek function fail\n");
			res = EIO;
			goto error;
		}
		read_len = fread(*buf, sizeof(char), buf_len, fp);
		if (read_len != (unsigned long)buf_len) {
			DOCA_DLOG_ERR("Error: fread function fail\n");
			res = EIO;
			goto error;
		}
	}
	fclose(fp);
	return buf_len;
error:
	if (fp)
		fclose(fp);
	if (*buf)
		free(*buf);
	return -res;
}

static bool
version_check(struct json_object *maj_obj, struct json_object *min_obj, struct json_object *p_obj)
{
	bool failed = false;

	if (json_object_get_type(maj_obj) != json_type_int ||
	    json_object_get_type(min_obj) != json_type_int ||
	    json_object_get_type(p_obj) != json_type_int) {
		DOCA_DLOG_ERR("Error: wrong object type for version objects\n"
				"major_ver type = %s, minor_ver type = %s, patch_version type = %s",
			print_json_object(maj_obj), print_json_object(min_obj),
			print_json_object(p_obj));
		return -1;
	}
	if (!DOCA_VERSION_LTE_CURRENT(json_object_get_int(maj_obj), json_object_get_int(min_obj),
				     json_object_get_int(p_obj))) {
		DOCA_DLOG_ERR("Error: CDO file version is higher then current DPI version\n");
		failed = true;
		}
	return failed;
}

static int parse_app_metadata(struct doca_dpi_compiler *ctx, struct json_object *app_metadata)
{
	struct json_object *sig_entry;
	struct doca_dpi_sig_info sig_info;
	uint32_t entry_num, nb_entries;
	int sig_id;
	const char *sig_name, *action;

	if (json_object_get_type(app_metadata) != json_type_array) {
		DOCA_DLOG_ERR("Error: wrong object type for app_metadata, json type is: %s\n",
			print_json_object(app_metadata));
		return -1;
	}

	nb_entries = json_object_array_length(app_metadata);
	if (nb_entries == 0) {
		DOCA_DLOG_ERR("Error: app_metadata object is empty\n");
		return -1;
	}

	for (entry_num = 0 ; entry_num < nb_entries ; entry_num++) {
		sig_entry = json_object_array_get_idx(app_metadata, entry_num);
		sig_id = -1;
		sig_name = "";
		action = "";
		json_object_object_foreach(sig_entry, key, val) {
			switch (json_object_get_type(val)) {
			case json_type_int:
				/* extracting sig_id */
				sig_id = json_object_get_int(val);

				if (strcasecmp(key, "sig_id")) {
					DOCA_DLOG_ERR("Error: int type in app_metadata is allowed only for sig_id\n");
					return -1;
				}
				continue;
			case json_type_string:
				/* extracting signature name and action */
				if (!strcasecmp(key, "name"))
					sig_name = json_object_get_string(val);
				else if (!strcasecmp(key, "action"))
					action = json_object_get_string(val);
				else {
					DOCA_DLOG_ERR("Error: string type in app_metadata is allowed only for name and action\n");
					return -1;
				}
				continue;
			default:
				DOCA_DLOG_ERR("Error: wrong type in signature object\n");
				return -1;
			}
		}
		if (strlen(sig_name) == 0 || sig_id < 0) {
			DOCA_DLOG_ERR("Error: name and id must be provided for every sig_info\n");
			return -1;
		}
		if ((sig_info.name = strdup(sig_name)) == NULL) {
			DOCA_DLOG_ERR("Error: memory allocation failed for sig_name\n");
			return -1;
		}
		if (strlen(action) > 0) {
			if ((sig_info.action = strdup(action)) == NULL) {
				DOCA_DLOG_ERR("Error: memory allocation failed for sig_action\n");
				free(sig_info.name);
				return -1;
			}
		} else
			sig_info.action = NULL;

		ctx->sig_map[sig_id].name = sig_info.name;
		ctx->sig_map[sig_id].action = sig_info.action;
	}
	return 0;
}

// static int parse_rule_obj_fields(struct json_object *sig_field, struct rule_obj *rule_obj)
// {
// 	struct json_object *json_elem;
// 	int tmp_int;

// 	json_object_object_foreach(sig_field, key, val) {
// 		switch (json_object_get_type(val)) {
// 		case json_type_int:
// 			tmp_int = json_object_get_int(val);

// 			if (!strcasecmp(key, "sig_id"))
// 				rule_obj->sig_id = tmp_int;
// 			else if (!strcasecmp(key, "rule_id"))
// 				rule_obj->rule_id = tmp_int;
// 			else if (!strcasecmp(key, "cond_id"))
// 				rule_obj->cond_id = tmp_int;
// 			else if (!strcasecmp(key, "l7_protocol")) {
// 				rule_obj->rule.l7_protocol = tmp_int;
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask
// 								   | DOCA_DPI_L7_PROTOCOL;
// 			} else if (!strcasecmp(key, "l4_protocol")) {
// 				rule_obj->rule.l4_protocol = tmp_int;
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask
// 								   | DOCA_DPI_L4_PROTOCOL;
// 			} else if (!strcasecmp(key, "l3_protocol")) {
// 				rule_obj->rule.l3_protocol = tmp_int;
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask
// 								   | DOCA_DPI_L3_PROTOCOL;
// 			} else if (!strcasecmp(key, "field")) {
// 				rule_obj->rule.field = tmp_int;
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask | DOCA_DPI_FIELD;
// 			} else if (!strcasecmp(key, "direction")) {
// 				rule_obj->rule.direction = tmp_int;
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask
// 								   | DOCA_DPI_DIRECTION;
// 			} else if (!strcasecmp(key, "offset")) {
// 				rule_obj->rule.offset = tmp_int;
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask | DOCA_DPI_OFFSET;
// 			} else if (!strcasecmp(key, "ip_mask")) {
// 				rule_obj->rule.ip_mask = tmp_int;
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask | DOCA_DPI_IP_MASK;
// 			}
// 			continue;
// 		case json_type_boolean:
// 			if (!strcasecmp(key, "anchored")) {
// 				rule_obj->rule.anchor = json_object_get_boolean(val);
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask | DOCA_DPI_ANCHOR;
// 			} else if (!strcasecmp(key, "dollar")) {
// 				rule_obj->rule.dollar = json_object_get_boolean(val);
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask | DOCA_DPI_DOLLAR;
// 			}
// 			continue;
// 		case json_type_array:
// 			if (!strcasecmp(key, "port")) {
// 				json_elem = json_object_array_get_idx(val, 0);
// 				rule_obj->rule.start_port = json_object_get_int(json_elem);
// 				json_elem = json_object_array_get_idx(val, 1);
// 				rule_obj->rule.end_port = json_object_get_int(json_elem);
// 				rule_obj->rule.bitmask = rule_obj->rule.bitmask | DOCA_DPI_PORT;
// 			}

// 		default:
// 			DOCA_DLOG_ERR("Error: wrong type in signature field object\n");
// 			return -1;
// 		}
// 	}
// 	return 0;
// }

/* extract fields from signature field and add it to ctx as rule_obj
 */
// static int parse_signatures_fields(struct doca_dpi_compiler *ctx,
// 				   struct json_object *signatures_obj,
// 				   int regex_id, string regex_rule)
// {
// 	struct json_object *sig_field;
// 	struct rule_obj *rule_obj;
// 	int nb_entries, entry_num;

// 	nb_entries = json_object_array_length(signatures_obj);
// 	if (nb_entries == 0) {
// 		DOCA_DLOG_ERR("Error: signatures_obj object is empty\n");
// 		return -1;
// 	}

// 	for (entry_num = 0 ; entry_num < nb_entries ; entry_num++) {
// 		sig_field = json_object_array_get_idx(signatures_obj, entry_num);
// 		rule_obj = new(nothrow) struct rule_obj;
// 		if (rule_obj == NULL) {
// 			DOCA_DLOG_ERR("Error: memory allocation failed for rule_obj in parse_signatures_fields()\n");
// 			return -1;
// 		}
// 		if (parse_rule_obj_fields(sig_field, rule_obj)) {
// 			delete rule_obj;
// 			return -1;
// 		}

// 		rule_obj->rule.nocase = check_nocase(regex_rule);
// 		rule_obj->rule.regex = strdup(regex_rule.c_str());
// 		ctx->regex_rules_map[regex_id].push_back(rule_obj);
// 	}
// 	return 0;
// }

static int parse_pp_table(struct doca_dpi_compiler *ctx, struct json_object *pp_table)
{
// 	struct json_object *sig_obj, *tmp_obj, *rule_obj, *regex_id_obj, *signatures_obj;
// 	map < int, string > id2regex_map;
// 	char *regex_str, *regex_str_tmp, *rules_string;
// 	int regex_id, nb_entries, entry_num, sig_id;

// 	if (json_object_get_type(pp_table) != json_type_array) {
// 		DOCA_DLOG_ERR("Error: wrong object type for pp_table, json type is: %s\n",
// 			print_json_object(pp_table));
// 		return -1;
// 	}

// 	/* extract rules fields */
// 	nb_entries = json_object_array_length(pp_table);
// 	if (nb_entries == 0) {
// 		DOCA_DLOG_ERR("Error: pp_table object is empty\n");
// 		return -1;
// 	}

// 	for (entry_num = 0 ; entry_num < nb_entries ; entry_num++) {
// 		sig_id = -1;
// 		sig_obj = json_object_array_get_idx(pp_table, entry_num);

// 		if (json_object_object_get_ex(sig_obj, "sig_id", &tmp_obj)) {
// 			if (json_object_get_type(tmp_obj) == json_type_int)
// 				sig_id = json_object_get_int(tmp_obj);
// 		}
// 		if (sig_id < 1) {
// 			DOCA_DLOG_ERR("Error: no/wrong regex_id provided\n");
// 			return -1;
// 		}

// 		if (!json_object_object_get_ex(sig_obj, "group", &tmp_obj)) {
// 			DOCA_DLOG_ERR("Error: no groups objects provided\n");
// 			return -1;
// 		}


// /* map of < subgroup id, < regex id , regex string > > to arrange regex ids to subgroups*/
// map < int, map < int, string > > subgrp2reg_map;
// /* number of regex strings inserted */
// int num_regex;
// /* number of ports inserted */
// int num_ports;
// /* number of addresses inserted */
// int num_addresses;
// /* number of simple inserted */
// int num_simple;
// /* map of < sig id, < group id, list < rule's json and type > > > for  the CDO pp_table */
// map < int, map < int, list < struct rule_obj > > > grp2rules_map;
// /* group's priority */
// map < int, map < int, struct group_property > > group_prop_map;

// 			return -1;
// 	}
	return 0;
}
