/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef __DOCA_DPI_COMPILER_HPP__
#define __DOCA_DPI_COMPILER_HPP__

#include <netinet/in.h>
#include <stdbool.h>

enum doca_dpi_l7_protocol {
	DOCA_DPI_HTTP = 1,
	DOCA_DPI_HTTP2,
	DOCA_DPI_TLS,
	DOCA_DPI_SSL,
	DOCA_DPI_SIP,
	DOCA_DPI_RTMP,
	DOCA_DPI_QUIC,
	DOCA_DPI_STUN,
	DOCA_DPI_DNS,
	DOCA_DPI_MAX_PROTOCOL
};

enum doca_dpi_fields {
	/* http fields */
	DOCA_DPI_HTTP_URI = 1,
	DOCA_DPI_HTTP_HOST,
	DOCA_DPI_HTTP_SERVER,
	DOCA_DPI_HTTP_CONTENT_TYPE,
	DOCA_DPI_HTTP_USER_AGENT,
	DOCA_DPI_HTTP_REFERER,
	DOCA_DPI_HTTP_LOCATION,
	DOCA_DPI_HTTP_COOKIE,
	DOCA_DPI_HTTP_CONTENT,
	DOCA_DPI_MAX_HTTP_FIELD,
	/* ssl fields */
	DOCA_DPI_SSL_COMMON_NAME,
	DOCA_DPI_SSL_SERVER_NAME,
	DOCA_DPI_SSL_SUBJECT_ALT_NAME,
	DOCA_DPI_SSL_RECORD_VER,
	DOCA_DPI_SSL_HANDSHAKE_VER,
	DOCA_DPI_SSL_CERT_SUBJECT,
	DOCA_DPI_MAX_SSL_FIELD,
	/* rtmp fields */
	DOCA_DPI_RTMP_PAGE_URL,
	DOCA_DPI_MAX_RTMP_FIELD,
	/* l4 fields */
	DOCA_DPI_L4_STREAM,
	DOCA_DPI_MAX_STREAM_FIELD,
	/* dns fields */
	DOCA_DPI_DNS_QUERY,
	DOCA_DPI_MAX_DNS_FIELD,
	/* stun fields */
	DOCA_DPI_STUN_REALM,
	DOCA_DPI_STUN_SOFTWARE,
	DOCA_DPI_MAX_STUN_FIELD,
	/* ip fields */
	DOCA_DPI_IP_RESOLV_NAME,
	DOCA_DPI_MAX_IP_FIELD,
	/* no regex */
	DOCA_DPI_NO_REGEX
};

enum doca_dpi_direction {
	DOCA_DPI_BOTH,
	DOCA_DPI_CTS,
	DOCA_DPI_STC
};

enum rule_field {
	DOCA_DPI_REGEX,
	DOCA_DPI_PORT,
	DOCA_DPI_ADDRESS,
	DOCA_DPI_PRIORITY,
	DOCA_DPI_L7_PROT,
	DOCA_DPI_IP_PROT,
	DOCA_DPI_ETH_PROT,
	DOCA_DPI_ALLOW_BIDIRECTIONAL,
};

struct doca_dpi_compiler_in_args {
	const char *cdo_file_name; /* name for the CDO file */
};

struct doca_dpi_compiler;

struct doca_dpi_sig_info {
	char *name;
	char *action;
};

struct address_field {
	const char *ip_address;
	int ip_mask;
	const char *side; /* indicate client or server side */
};

struct port_field {
	uint16_t start_port;
	uint16_t end_port;
	const char *side; /* indicate client or server side */
};

struct regex_field {
	char *regex;
	int field;
	int offset; /* if not in use set to 0 */
};

struct priority {
	int prior1;
	int prior2;
};

union rule_value {
	/* Ip address field */
	struct address_field addr_field;
	/* Port number */
	struct port_field port_field;
	/* Regex field */
	struct regex_field reg_field;
	/* Priority field */
	struct priority priority;
	/* Group's primitive field */
	int primitive;
};

/* allocating all memory and data structs needed for the compiler to work */
struct doca_dpi_compiler *doca_dpi_compiler_alloc();

/* free all data allocated to the compiler */
void doca_dpi_compiler_dealloc(struct doca_dpi_compiler *ctx);

/* add signature data to the compiler's list */
int doca_dpi_compiler_set_sig(struct doca_dpi_compiler *ctx, int sig_id,
			      const struct doca_dpi_sig_info *sig_info);

/* add rule data to the compiler's list */
int doca_dpi_compiler_add_rule(struct doca_dpi_compiler *ctx, int sig_id, int group_id,
			       int direction, int rule_field, union rule_value *val);

int doca_dpi_compiler_commit(struct doca_dpi_compiler *ctx);

/* save all data in the compiler's context to a CDO file */
int doca_dpi_compiler_save(struct doca_dpi_compiler *ctx, struct doca_dpi_compiler_in_args *args);

/* load data from CDO file to an allocated compiler's context */
int doca_dpi_compiler_load(struct doca_dpi_compiler *ctx, const char *cdo);

/* get signature information */
struct doca_dpi_sig_info *doca_dpi_compiler_get_signuture(int sig_id);

#endif /* __DOCA_DPI_COMPILER_HPP__ */
