/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>
#include <unistd.h>
#include <bsd/string.h>

#include "dpi_compiler_common.hpp"

static doca_error_t
args_validation_callback(void *cfg)
{
	struct in_args *dpi_compiler_cfg = (struct in_args *)cfg;

	if (dpi_compiler_cfg->rules_format == INVALID) {
		printf("Error: format entered is not supported\n");
		return DOCA_ERROR_INVALID_VALUE;
	}
	return DOCA_SUCCESS;
}

static doca_error_t
input_callback(void *param, void *config)
{
	char *input_file = (char *)param;
	struct in_args *dpi_compiler_cfg = (struct in_args *)config;

	if(strnlen(input_file, MAX_FILE_NAME) == MAX_FILE_NAME) {
		printf("Error: input file name is too long - MAX=%d\n", MAX_FILE_NAME - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (access(input_file, F_OK) == -1) {
		printf("Error: input file not found %s\n", input_file);
		return DOCA_ERROR_NOT_FOUND;
	}
	strlcpy(dpi_compiler_cfg->input_file, input_file, MAX_FILE_NAME);
	return DOCA_SUCCESS;
}

static doca_error_t
output_callback(void *param, void *config)
{
	char *output_file = (char *)param;
	struct in_args *dpi_compiler_cfg = (struct in_args *)config;

	if(strnlen(output_file, MAX_FILE_NAME) == MAX_FILE_NAME) {
		printf("Error: CDO output file name is too long - MAX=%d\n", MAX_FILE_NAME - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strlcpy(dpi_compiler_cfg->cdo_file_name, output_file, MAX_FILE_NAME);
	return DOCA_SUCCESS;
}

static doca_error_t
format_callback(void *param, void *config)
{
	struct in_args *dpi_compiler_cfg = (struct in_args *)config;
	char *format = (char *)param;

	if (!strcasecmp(format, "suricata"))
		dpi_compiler_cfg->rules_format = SURICATA;
	else if (!strcasecmp(format, "qosmos"))
		dpi_compiler_cfg->rules_format = QOSMOS;
	else {
		printf("The rules format \"%s\" is not supported.\n", format);
		return DOCA_ERROR_NOT_SUPPORTED;
	}
	return DOCA_SUCCESS;
}

static doca_error_t
classtype_callback(void *param, void *config)
{
	char *class_type_path = (char *)param;
	struct in_args *dpi_compiler_cfg = (struct in_args *)config;

	if(strnlen(class_type_path, MAX_FILE_NAME) == MAX_FILE_NAME) {
		printf("Error: path to class type file is too long - MAX=%d\n", MAX_FILE_NAME - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strlcpy(dpi_compiler_cfg->classtype_file, class_type_path, MAX_FILE_NAME);
	return DOCA_SUCCESS;
}

doca_error_t
register_dpi_compiler_params()
{
	doca_error_t ret;
	struct doca_argp_param *input_param, *format_param, *output_param, *classtype_param;

	/* Create and register input_param param */
	ret = doca_argp_param_create(&input_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to create ARGP param: %s\n", doca_get_error_string(ret));
		return ret;
	}
	doca_argp_param_set_short_name(input_param, "i");
	doca_argp_param_set_long_name(input_param, "input");
	doca_argp_param_set_arguments(input_param, "<path>");
	doca_argp_param_set_description(input_param, "Full path to signature file");
	doca_argp_param_set_callback(input_param, input_callback);
	doca_argp_param_set_type(input_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(input_param);
	ret = doca_argp_register_param(input_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to register program param: %s\n", doca_get_error_string(ret));
		return ret;
	}
	/* Create and register format_param param */
	ret = doca_argp_param_create(&format_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to create ARGP param: %s\n", doca_get_error_string(ret));
		return ret;
	}
	doca_argp_param_set_short_name(format_param, "f");
	doca_argp_param_set_long_name(format_param, "format");
	doca_argp_param_set_description(format_param, "Signature file format");
	doca_argp_param_set_callback(format_param, format_callback);
	doca_argp_param_set_type(format_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(format_param);
	ret = doca_argp_register_param(format_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to register program param: %s\n", doca_get_error_string(ret));
		return ret;
	}

	/* Create and register output_param param */
	ret = doca_argp_param_create(&output_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to create ARGP param: %s\n", doca_get_error_string(ret));
		return ret;
	}
	doca_argp_param_set_short_name(output_param, "o");
	doca_argp_param_set_long_name(output_param, "output");
	doca_argp_param_set_arguments(output_param, "<path>");
	doca_argp_param_set_description(output_param, "Full path to CDO file");
	doca_argp_param_set_callback(output_param, output_callback);
	doca_argp_param_set_type(output_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(output_param);
	ret = doca_argp_register_param(output_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to register program param: %s\n", doca_get_error_string(ret));
		return ret;
	}

	/* Create and register classtype_param param */
	ret = doca_argp_param_create(&classtype_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to create ARGP param: %s\n", doca_get_error_string(ret));
		return ret;
	}
	doca_argp_param_set_short_name(classtype_param, "c");
	doca_argp_param_set_long_name(classtype_param, "classtype");
	doca_argp_param_set_arguments(classtype_param, "<path>");
	doca_argp_param_set_description(classtype_param, "Path to the classtype file for priority in suricata file");
	doca_argp_param_set_callback(classtype_param, classtype_callback);
	doca_argp_param_set_type(classtype_param, DOCA_ARGP_TYPE_STRING);
	ret = doca_argp_register_param(classtype_param);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to register program param: %s\n", doca_get_error_string(ret));
		return ret;
	}

	/* Register tool validation callback */
	ret = doca_argp_register_validation_callback(args_validation_callback);
	if (ret != DOCA_SUCCESS) {
		printf("Failed to register validation callback: %s\n", doca_get_error_string(ret));
		return ret;
	}
	return DOCA_SUCCESS;
}
