/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <getopt.h>
#include <unistd.h>
#include <stdio.h>
#include <json-c/json.h>
#include <string.h>
#ifdef DOCA_DPI_USE_HS
#include <hs/hs_compile.h>
#endif
#include <iostream>
#include <fstream>
#include <map>
#include <set>
#include <iterator>
#include <list>
#include <doca_version.h>
#include <doca_log.h>

#include "rxp_qosmos/src/rxp_qosmos_class.hpp"
#include "dpi_compiler.hpp"
#include "dpi_compiler_common.hpp"

using namespace std;
DOCA_LOG_REGISTER(COMPILER:QOSMOS);

#define RXP_VERSION "5.7"
#define RULES_FILE_NAME "signatures.rules"
#define ROF_FILE_NAME "signatures_compiled"
#define HS_RULES_FILE_NAME "hs_signatures.rules"
#define MAX_RANGE_SIZE 32
#define REGEX_MAX_SIZE 255
#define MAX_FILE_NAME 255

#define EXPECT_CHAR_NO_CHECK(ch)										\
	if (*buf++ != ch) {											\
        	*ret = -1;											\
		return NULL;											\
	}

#define GET_JSON_OBJ_SOFT(src_obj, name, dest_obj, ret)	       							\
	if (!json_object_object_get_ex(src_obj, name, dest_obj)) {	       					\
		DOCA_DLOG_DBG("field \"%s\" does not exist in the parsed pdd\n",name);				\
		ret = 0;											\
	} else													\
		ret = 1;											\

#define GET_JSON_OBJ_HARD(src_obj, name, dest_obj, ret, src_name)						\
	if (!json_object_object_get_ex(src_obj, name, dest_obj)){	       					\
		DOCA_DLOG_ERR("field \"%s\" does not exist in the %s\n",name, src_name);\
		return ret;											\
	}

#define ADD_INT_OBJ(src, name, dest)					       					\
	json_object_object_add(dest, name, json_object_new_int(src));	       					\

#define ADD_STRING_OBJ(src, name, dest)					       					\
	json_object_object_add(dest, name, json_object_new_string(src));	       				\

#define CHECK_OBJ_INT(val)						       					\
	json_object_get_type(val) == json_type_int			       					\

#define CHECK_OBJ_STRING(val)						       					\
	json_object_get_type(val) == json_type_string			       					\

#define CHECK_OBJ_ARRAY(val)						       					\
	json_object_get_type(val) == json_type_array			       					\

#define CHECK_OBJ_OBJECT(val)						       					\
	json_object_get_type(val) == json_type_object			       					\


enum l3_l4_protocols {
	DOCA_DPI_UDP = 17,
	DOCA_DPI_TCP = 6,
	DOCA_DPI_ICMP = 1,
	DOCA_DPI_IP = 0x0800,
	DOCA_DPI_IP6 = 0x86DD
};

static long
read_rules_db(const char *file, char **buf)
{
        FILE *fp;
        long buf_len = 0;
        size_t read_len = 0;
        int res = 0;

        fp = fopen(file, "r");
        if (!fp) {
		DOCA_DLOG_ERR("can't open file %s\n, err = %d", file, EIO);
		return -EIO;
	}
        if (fseek(fp, 0L, SEEK_END) == 0) {
                *buf = NULL;

                buf_len = ftell(fp);
                if (buf_len == -1) {
                        res = EIO;
                        goto error;
                }
                *buf = (char *)malloc(sizeof(char) * (buf_len + 1));
                if (!*buf) {
                        res = ENOMEM;
                        goto error;
                }
                if (fseek(fp, 0L, SEEK_SET) != 0) {
                        res = EIO;
                        goto error;
                }
                read_len = fread(*buf, sizeof(char), buf_len, fp);
                if (read_len != (unsigned long)buf_len) {
                        res = EIO;
                        goto error;
                }

                (*buf)[read_len] = 0;
        }
        fclose(fp);
        return buf_len;
error:
        DOCA_DLOG_ERR("can't open file %s\n, err = %d", file, res);
        if (fp)
                fclose(fp);
        if (*buf)
                free (*buf);
        return -res;
}

const map<string, int> prot2num = {
	{"http", DOCA_DPI_HTTP},
	{"tcp", DOCA_DPI_TCP},
	{"udp", DOCA_DPI_UDP},
	{"ssl", DOCA_DPI_SSL},
	{"sip", DOCA_DPI_SIP},
	{"rtmp", DOCA_DPI_RTMP},
	{"icmp", DOCA_DPI_ICMP},
	{"ip", DOCA_DPI_IP},
	{"ip6", DOCA_DPI_IP6},
	{"quic", DOCA_DPI_QUIC},
	{"stun", DOCA_DPI_STUN},
	{"dns", DOCA_DPI_DNS},
	{"tls", DOCA_DPI_TLS}
};

const map<string, int> field2num = {
	{"uri", DOCA_DPI_HTTP_URI},
	{"server", DOCA_DPI_HTTP_HOST},
	{"user_agent", DOCA_DPI_HTTP_USER_AGENT},
	{"referer", DOCA_DPI_HTTP_REFERER},
	{"server_agent", DOCA_DPI_HTTP_SERVER},
	{"location", DOCA_DPI_HTTP_LOCATION},
	{"cookie", DOCA_DPI_HTTP_COOKIE},
	{"mime_type", DOCA_DPI_HTTP_CONTENT_TYPE},
	{"content", DOCA_DPI_HTTP_CONTENT},
	{"common_name", DOCA_DPI_SSL_COMMON_NAME},
	{"subject_alt_name", DOCA_DPI_SSL_SUBJECT_ALT_NAME},
	{"page_url", DOCA_DPI_RTMP_PAGE_URL},
	{"stream", DOCA_DPI_L4_STREAM},
	{"software", DOCA_DPI_STUN_SOFTWARE},
	{"realm", DOCA_DPI_STUN_REALM},
	{"query", DOCA_DPI_DNS_QUERY},
	{"server_name", DOCA_DPI_SSL_SERVER_NAME},
	{"resolv_name", DOCA_DPI_IP_RESOLV_NAME},
	{"stream_offset", DOCA_DPI_L4_STREAM}
};

static int
check_direction(const char *field)
{
	if (strcmp(field,"")) {
		if (!strncmp(field, "cts", 3))
			return DOCA_DPI_CTS;
		else if (!strncmp(field, "stc", 3))
			return DOCA_DPI_STC;
		return DOCA_DPI_BOTH;
	}
	return DOCA_DPI_BOTH;
}

static void
fill_l3_protocols(set<string> *l3_prot)
{
	l3_prot->insert("ip");
	l3_prot->insert("ip6");
	return;
}

static void
fill_l4_protocols(set<string> *l4_prot)
{
	l4_prot->insert("tcp");
	l4_prot->insert("udp");
	l4_prot->insert("icmp");
	l4_prot->insert("spx");
	return;
}

static void
fill_l7_protocols(set<string> *l7_prot)
{
	l7_prot->insert("http");
	l7_prot->insert("ssl");
	l7_prot->insert("sip");
	l7_prot->insert("rtmp");
	l7_prot->insert("smtp");
	l7_prot->insert("telnet");
	l7_prot->insert("dns");
	l7_prot->insert("quic");
	l7_prot->insert("stun");
	return;
}

static int
handle_stream_offset(struct json_object *stream_offset, const char* key, union rule_value *tmp_rule)
{
	struct json_object *json_elem;
	const char *field;
	string regex_str, converted_str, converter_err;
	int direction = -1;
	rxp_qosmos_class converter;

	if (strstr(key, "stream_offset") == NULL) {
		DOCA_DLOG_ERR("pdd format error: json_object_object allowed only as stream_offset\n");
		return -1;
	}
	direction = check_direction(key);
	field = direction == DOCA_DPI_BOTH ? key : key+4;
	if (field2num.count(field) == 0) {
		DOCA_DLOG_ERR("field %s is not supported\n",
			field);
	}
	tmp_rule->reg_field.field = field2num.at(field);

	GET_JSON_OBJ_HARD(stream_offset, "offset", &json_elem, 0, "stream_offset field");
	if (!(CHECK_OBJ_INT(json_elem))) {
		DOCA_DLOG_ERR("pdd format error: field \"offset\" stream_offset isn't of type int\n");
		return -1;
	}
	tmp_rule->reg_field.offset = json_object_get_int(json_elem);

	GET_JSON_OBJ_HARD(stream_offset, "pattern", &json_elem, 0, "stream_offset field");
	if (!(CHECK_OBJ_STRING(json_elem))) {
		DOCA_DLOG_ERR("pdd format error: field \"offset\" stream_offset isn't of type string\n");
		return -1;
	}
	regex_str = json_object_get_string(json_elem);
	/* converting to RXP format */
	if (regex_str.at(0) == '/') {
		regex_str.replace(0, 1, "//");
		regex_str.append("/");
	}
	if (!converter.convert_qosmos_rule_to_rxp(regex_str, converted_str, converter_err)) {
		cout << "Error: regex rule to rxp converter failed" << endl;
		cout << converter_err << endl;
		cout << "regex is: " << regex_str << endl;
		return -1;
	}
	tmp_rule->reg_field.regex = strdup(converted_str.c_str());
	return 0;
}

/* receive pdd object of one app and parse it into rule_structs
 */
static int
parse_app_pdd(struct doca_dpi_compiler *ctx, struct json_object *parsed_json)
{
	struct json_object *id;
	struct json_object *rules_list;
	struct json_object *rule;
	struct json_object *inner_rule;
	struct json_object *fields;
	struct json_object *name;
	struct json_object *protocol;
	struct json_object *order_obj;
	struct json_object *tags;
	struct doca_dpi_sig_info sig_info;
	size_t n_rules, n_inner_rules, i_out, i_in, idx;
	uint32_t pdd_id;
	int int_data, direction, parse_check;
	const char *field, *pdd_name, *r_prot;
	string field_str, netmask, address, converter_err, converted_str;
	union rule_value tmp_rule;
	set<string> l7_prot;
	set<string> l4_prot;
	set<string> l3_prot;
	rxp_qosmos_class converter;

	fill_l3_protocols(&l3_prot);
	fill_l4_protocols(&l4_prot);
	fill_l7_protocols(&l7_prot);

	GET_JSON_OBJ_HARD(parsed_json, "name", &name, -1, "parsed file");
	GET_JSON_OBJ_HARD(parsed_json, "id", &id, -1, "parsed file");
	GET_JSON_OBJ_HARD(parsed_json, "tags", &tags, -1, "parsed file");
	GET_JSON_OBJ_SOFT(parsed_json, "rules", &rules_list, parse_check);

	pdd_name = json_object_get_string(name);
	pdd_id = json_object_get_int(id);

	/* handle signatures data */
	sig_info.name = strdup(pdd_name);
	if (sig_info.name == NULL) {
		DOCA_DLOG_ERR("strdup failed for sig name\n");
		return -1;
	}
	sig_info.action = strdup("none");
	if (sig_info.action == NULL) {
		DOCA_DLOG_ERR("strdup failed for sig action\n");
		free(sig_info.name);
		return -1;
	}
	if (doca_dpi_compiler_set_sig(ctx, pdd_id, &sig_info)) {
		free(sig_info.name);
		free(sig_info.action);
		return -1;
	}

	free(sig_info.name);
	free(sig_info.action);

	n_rules = (parse_check && CHECK_OBJ_ARRAY(rules_list)) ?
				       json_object_array_length(rules_list) : 0;
	for (i_out = 0 ; i_out < n_rules ; i_out++) {
		rule = json_object_array_get_idx(rules_list, i_out);
		if (!json_object_object_get_ex(rule, "rule", &inner_rule)) {
			DOCA_DLOG_DBG("rule object without rule field.\n");
			continue;
		}
		n_inner_rules = CHECK_OBJ_ARRAY(inner_rule) ?
				       json_object_array_length(inner_rule) : 0;
		if (json_object_object_get_ex(rule, "order", &order_obj)) {
			if (CHECK_OBJ_STRING(order_obj)) {
				field_str = json_object_get_string(order_obj);
				try {
					idx = field_str.find(':');
					if (idx == string::npos) {
						DOCA_DLOG_ERR("order is not in format <NUM> : <NUM>\n");
						return -1;
					}
					tmp_rule.priority.prior1 =
							stoi(field_str.substr(0, idx).c_str());
					tmp_rule.priority.prior2 =
							stoi(field_str.substr(idx + 1).c_str());
				}
				catch(...) {
					DOCA_DLOG_ERR("order string is not of pattern <NUM>:<NUM>\n");
					return -1;
				}
				if (doca_dpi_compiler_add_rule(ctx, pdd_id, i_out,
							       0, DOCA_DPI_PRIORITY, &tmp_rule))
					return -1;
			}
		}


		for (i_in = 0 ; i_in < n_inner_rules ; i_in++) {
			fields = json_object_array_get_idx(inner_rule, i_in);
			r_prot = "";
			tmp_rule.primitive = -1;
			if (json_object_object_get_ex(fields, "protocol", &protocol)) {
				if (CHECK_OBJ_STRING(protocol)) {
					r_prot = json_object_get_string(protocol);
					if (l7_prot.count(r_prot)) {
						tmp_rule.primitive = prot2num.at(r_prot);
						doca_dpi_compiler_add_rule(ctx, pdd_id, i_out,
								0, DOCA_DPI_L7_PROT, &tmp_rule);
					} else if (l4_prot.count(r_prot)) {
						tmp_rule.primitive = prot2num.at(r_prot);
						doca_dpi_compiler_add_rule(ctx, pdd_id, i_out,
								0, DOCA_DPI_IP_PROT, &tmp_rule);
					} else if (l3_prot.count(r_prot)) {
						tmp_rule.primitive = prot2num.at(r_prot);
						doca_dpi_compiler_add_rule(ctx, pdd_id, i_out,
								0, DOCA_DPI_ETH_PROT, &tmp_rule);
					}
				}
			} else
				DOCA_DLOG_DBG("no protocol found on pdd_id %d.\n", pdd_id);

			json_object_object_foreach(fields,key,val) {
				field = "";
				field_str.clear();
				int_data = -1;

				if (CHECK_OBJ_STRING(val)) {
					/* adding rule with string object */
					if (strcasecmp(key,"protocol") == 0)
						continue;
					for (size_t i=0; i<strlen(key); i++)
						key[i] = tolower(key[i]);
					field_str = json_object_get_string(val);

					if (!strcmp(key, "address")) {
						/* handle address rule */
						char *duplicated_addr;

						idx = field_str.find('/');
						if (idx == string::npos) {
							tmp_rule.addr_field.ip_mask = 0;
							duplicated_addr = strdup(field_str.c_str());
						} else {
							tmp_rule.addr_field.ip_mask =
								atoi(field_str.substr(idx + 1).c_str());

							duplicated_addr = strdup(field_str.substr(0, idx).c_str());
						}
						if (duplicated_addr == NULL) {
							DOCA_DLOG_ERR("failed to allocate memory for ip address\n");
							return -1;
						}
						tmp_rule.addr_field.ip_address = duplicated_addr;
						tmp_rule.addr_field.side = "server";
						doca_dpi_compiler_add_rule(ctx, pdd_id,
							i_out, DOCA_DPI_BOTH, DOCA_DPI_ADDRESS,
							&tmp_rule);
						free(duplicated_addr);

					} else if (!strcmp(key, "protoid")) {
						/* handle proto_id rule */
						try {
							tmp_rule.primitive = stoi(field_str);
							doca_dpi_compiler_add_rule(ctx,
									pdd_id, i_out, 0,
									DOCA_DPI_IP_PROT,
									&tmp_rule);
						}
						catch(...) {
							DOCA_DLOG_ERR("failed to convert protoid to integer\n");
							return -1;
						}

					} else if (!strcmp(key, "port")) {
						/* handle port rule */
						idx = field_str.find('-');
						if (idx == string::npos) {
							int_data = atoi(field_str.c_str());
							tmp_rule.port_field.start_port = int_data;
							tmp_rule.port_field.end_port = int_data;
							tmp_rule.port_field.side = "server";
							doca_dpi_compiler_add_rule(ctx, pdd_id,
									i_out, DOCA_DPI_BOTH,
									DOCA_DPI_PORT, &tmp_rule);
						} else {
							tmp_rule.port_field.start_port =
								atoi(field_str.substr(0,
										idx).c_str());
							tmp_rule.port_field.end_port =
								atoi(field_str.substr(idx
										+ 1).c_str());
							tmp_rule.port_field.side = "server";
						}

					} else {
						/* handle regex rule */
						char *duplicated_regex;

						direction = check_direction(key);
						field = direction == DOCA_DPI_BOTH ? key : key+4;
						if (field2num.count(field) == 0) {
							DOCA_DLOG_ERR("field %s is not supporteddddd\n",
								field);
						}
						/* converting to RXP format */
						if (field_str.at(0) == '/') {
							field_str.replace(0, 1, "//");
							field_str.append("/");
						}
						if (!converter.convert_qosmos_rule_to_rxp(field_str,
								converted_str, converter_err)) {
							cout << "Error: regex rule to rxp converter failed" << endl;
							cout << converter_err << endl;
							cout << "regex is: " << field_str << endl;
							return -1;
						}
						tmp_rule.reg_field.field = field2num.at(field);
						if (tmp_rule.reg_field.field <
									DOCA_DPI_MAX_HTTP_FIELD)
							converted_str.append("i");
						duplicated_regex = strdup(converted_str.c_str());
						tmp_rule.reg_field.regex = duplicated_regex;
						tmp_rule.reg_field.offset = -1;
						doca_dpi_compiler_add_rule(ctx, pdd_id,
								i_out, direction,
								DOCA_DPI_REGEX, &tmp_rule);
						free(duplicated_regex);
					}
				} else if (CHECK_OBJ_INT(val)) {
					int_data = json_object_get_int(val);
					if (!strcmp(key, "port")) {
						tmp_rule.port_field.start_port = int_data;
						tmp_rule.port_field.end_port = int_data;
						tmp_rule.port_field.side = "server";
						doca_dpi_compiler_add_rule(ctx, pdd_id, i_out,
								DOCA_DPI_BOTH, DOCA_DPI_PORT,
								&tmp_rule);

					}
				} else if (CHECK_OBJ_OBJECT(val)) {
					if (handle_stream_offset(val, key, &tmp_rule))
						return -1;
					direction = check_direction(key);
					doca_dpi_compiler_add_rule(ctx, pdd_id, i_out, direction,
									DOCA_DPI_REGEX, &tmp_rule);
					free(tmp_rule.reg_field.regex);
				}
			}
		}
	}
	return 0;
}

static long
parse_pdd_json(const char *data_file, struct doca_dpi_compiler *compiler_ctx)
{
	char *buffer = NULL;
	struct json_object *parsed_json, *apps_list_obj, *app_obj;
	int nb_apps, i_app;
	long res;

	if ((res = read_rules_db(data_file, &buffer)) < 0)
		return res;
	DOCA_DLOG_DBG("start parsing PDD file\n");
	parsed_json = json_tokener_parse(buffer);
	if (parsed_json == NULL) {
		DOCA_DLOG_ERR("json format is broken\n");
		res = -1;
		goto err_free_buffer;
	}
	if (!json_object_object_get_ex(parsed_json, "protocols", &apps_list_obj)) {
		DOCA_DLOG_DBG("the input pdd file isn't an aggregated file.\n");
		if ((res = parse_app_pdd(compiler_ctx, parsed_json)))
			goto err_free_json;
	} else {
		nb_apps = json_object_array_length(apps_list_obj);
		for (i_app = 0 ; i_app < nb_apps ; i_app++) {
			app_obj = json_object_array_get_idx(apps_list_obj, i_app);
			if ((res = parse_app_pdd(compiler_ctx, app_obj)))
				goto err_free_json;
		}
	}
	res = 0;
err_free_json:
	json_object_put(parsed_json);
err_free_buffer:
	free(buffer);
	return res;
}

int qosmos_main(struct in_args *arg) {

	DOCA_DLOG_DBG("compiler begins...\n");
	long res;
	struct doca_dpi_compiler *compiler_ctx;
	struct doca_dpi_compiler_in_args in_arg;

	compiler_ctx = doca_dpi_compiler_alloc();

	if ((res = parse_pdd_json(arg->input_file, compiler_ctx)) < 0 ) {
		DOCA_DLOG_ERR("parsing failed for file %s\n", arg->input_file);
		doca_dpi_compiler_dealloc(compiler_ctx);
		return res;
	}

	in_arg.cdo_file_name = arg->cdo_file_name;
	if (doca_dpi_compiler_save(compiler_ctx, &in_arg)) {
		doca_dpi_compiler_dealloc(compiler_ctx);
		return -1;
	}

	printf("Finish build of CDO json file\n");
	doca_dpi_compiler_dealloc(compiler_ctx);

	printf("Finish clearing all data structures\n");

	return 0;
}

