/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dpi_suricata_parser.hpp"
#include "dpi_compiler.hpp"
#include <doca_version.h>
#include <iostream>
#include <string.h>
#include <list>
#include <set>
#include <map>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <getopt.h>
#include <unistd.h>

using namespace std;

#define MAX_FILE_NAME 255

extern struct suricata_hdr header;
extern struct sig_global_options glob_opt;
extern map <string, int> classtype_map;

const set<int> l3_prot = {
	DOCA_DPI_IP,
	DOCA_DPI_IP6
};

const set<int> l4_prot = {
	DOCA_DPI_TCP,
	DOCA_DPI_UDP,
	DOCA_DPI_ICMP,
};

const set<int> l7_prot = {
	DOCA_DPI_HTTP,
	DOCA_DPI_HTTP2,
	DOCA_DPI_SSL,
	DOCA_DPI_SIP,
	DOCA_DPI_RTMP,
	DOCA_DPI_DNS,
	DOCA_DPI_QUIC,
	DOCA_DPI_STUN,
	DOCA_DPI_TLS
};

void
init_sig_header()
{
	header.action = NULL;
	header.l7_protocol = -1;
	header.l4_protocol = -1;
	header.l3_protocol = -1;
	header.src_ip = NULL;
	header.src_netmask = -1;
	header.src_port.valid = false;
	header.direction = -1;
	header.dst_ip = NULL;
	header.dst_netmask = -1;
	header.dst_port.valid = false;
}

void
init_sig_global_options()
{
	glob_opt.rule_count = 0;
	glob_opt.flow_dir = -1;
	glob_opt.ip_proto = -1;
	glob_opt.itype = -1;
	glob_opt.icode = -1;
	glob_opt.sid = -1;
	glob_opt.gid = -1;
	glob_opt.msg = NULL;
	glob_opt.sticky_field = -1;
	glob_opt.sticky_prot = -1;
	glob_opt.priority = -1;
	glob_opt.classtype = NULL;
	return;
}

static bool
not_hex_char(char c)
{
	/* Check if char is [0,9] , [a-f] or [A-F] */
	if ((c < 58 && c > 47) || (c < 103 && c > 96) || (c < 71 && c > 64))
		return false;
	else
		return true;
}

/* convert the content field to fit RXP syntax.
 * return NULL if fail and converted string otherwise
 */
char *
content_converter(const char *content)
{
	bool hex = false;
	char *converted_content;
	string in_field = content;
	string::iterator itr = in_field.begin();
	string::iterator itr_end = in_field.end();
	string out_field = "/\\Q";
	string hex_str;

	while (itr != itr_end) {
		if (*itr == '|') {
			hex = !hex;
			if (hex)
				out_field += "\\E";
			else
				out_field += "\\Q";
		} else if (hex) {
			while (*itr == ' ')
				itr++;
			if (*itr == '|')
				continue;
			if(not_hex_char(itr[0]) || not_hex_char(itr[1])) {
				printf("Error: char received as hex in %s are illegal\n",
					in_field.c_str());
				return NULL;
			}
			hex_str = "\\x";
			hex_str += itr[0];
			hex_str += itr[1];
			out_field += hex_str;
			itr++;
		} else {
			out_field += *itr;
		}
		itr++;
	}
	out_field += "\\E/";
	converted_content = strdup(out_field.c_str());
	if (converted_content == NULL) {
		printf("Error: memory alloc failed for content convertion\n");
		return NULL;
	}
	return converted_content;
}

void initiate_rule(struct doca_dpi_rule *rule)
{
	glob_opt.rule_count++;
	rule->bitmask = 0;
	rule->anchor = false;
	rule->dollar = false;
	rule->nocase = false;
	rule->offset = -1;
	rule->direction = header.direction;
	rule->bitmask = rule->bitmask | DOCA_DPI_DIRECTION_MASK;
	if (header.l7_protocol >= 0) {
		rule->l7_protocol = header.l7_protocol;
		rule->bitmask = rule->bitmask | DOCA_DPI_L7_PROTOCOL_MASK;
	} else if (header.l4_protocol >= 0) {
		rule->l4_protocol = header.l4_protocol;
		rule->bitmask = rule->bitmask | DOCA_DPI_L4_PROTOCOL_MASK;
		rule->field = DOCA_DPI_L4_STREAM;
		rule->bitmask = rule->bitmask | DOCA_DPI_FIELD_MASK;
	} else if (header.l3_protocol >= 0) {
		rule->l3_protocol = header.l3_protocol;
		rule->bitmask = rule->bitmask | DOCA_DPI_L3_PROTOCOL_MASK;
	}
	if (l7_prot.count(glob_opt.sticky_prot)) {
		rule->l7_protocol = glob_opt.sticky_prot;
		rule->bitmask = rule->bitmask | DOCA_DPI_L7_PROTOCOL_MASK;
		rule->field = glob_opt.sticky_field;
		rule->bitmask = rule->bitmask | DOCA_DPI_FIELD_MASK;
	} else if (l4_prot.count(glob_opt.sticky_prot)) {
		rule->l4_protocol = glob_opt.sticky_prot;
		rule->bitmask = rule->bitmask | DOCA_DPI_L4_PROTOCOL_MASK;
		rule->field = glob_opt.sticky_field;
		rule->bitmask = rule->bitmask | DOCA_DPI_FIELD_MASK;
	}
	return;
}

int update_compiler(doca_dpi_compiler *ctx, list <struct doca_dpi_rule> *rules_list)
{
	struct doca_dpi_rule rule;
	struct doca_dpi_sig_info sig_info;
	union rule_value tmp_rule;
	int ret, gid;
	string tmp_str;

	/* add signature info to compiler's list */
	sig_info.name = glob_opt.msg;
	sig_info.action = header.action;
	if (doca_dpi_compiler_set_sig(ctx, glob_opt.sid, &sig_info)) {
		ret = -1;
		goto error;
	}

	/* add rules to compiler's list */
	gid = glob_opt.gid < 0 ? 0 : glob_opt.gid;

	/* add priority number */
	if ((glob_opt.priority < 0) && (glob_opt.classtype != NULL)) {
		if (classtype_map.count(glob_opt.classtype)) {
			glob_opt.priority = classtype_map[glob_opt.classtype];
		} else {
			printf("Error: classtype value %s is in rule but not exist in given classification file\n",
				glob_opt.classtype);
			ret = -1;
			goto error;
		}
	}
	if (glob_opt.priority == 0) {
		printf("Error: zero is not a legal value for priority field\n");
		ret = -1;
		goto error;
	}
	if (glob_opt.priority > 0) {
		tmp_rule.priority.prior1 = glob_opt.priority;
		tmp_rule.priority.prior2 = glob_opt.priority;
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_PRIORITY, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}

	/* add protocol */
	tmp_rule.primitive = -1;
	if (glob_opt.ip_proto >= 0) {
		tmp_rule.primitive = glob_opt.ip_proto;
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_IP_PROT, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	} else if (header.l4_protocol >= 0) {
		tmp_rule.primitive = header.l4_protocol;
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_IP_PROT, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}
	if (header.l3_protocol >= 0) {
		tmp_rule.primitive = header.l3_protocol;
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_ETH_PROT, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}
	if (header.l7_protocol >= 0) {
		tmp_rule.primitive = header.l7_protocol;
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_L7_PROT, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}

	/* add ip address and port numbers */
	if (header.dst_ip != NULL) {
		tmp_rule.addr_field.ip_address = header.dst_ip;
		tmp_rule.addr_field.ip_mask = header.dst_netmask;
		tmp_rule.addr_field.side = "server";
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_ADDRESS, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}
	if (header.dst_port.valid) {
		tmp_rule.port_field.start_port = header.dst_port.start;
		tmp_rule.port_field.end_port = header.dst_port.end;
		tmp_rule.port_field.side = "server";
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_PORT, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}
	if (header.src_ip != NULL) {
		tmp_rule.addr_field.ip_address = header.src_ip;
		tmp_rule.addr_field.ip_mask = header.src_netmask;
		tmp_rule.addr_field.side = "client";
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_ADDRESS, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}
	if (header.src_port.valid) {
		tmp_rule.port_field.start_port = header.src_port.start;
		tmp_rule.port_field.end_port = header.src_port.end;
		tmp_rule.port_field.side = "client";
		if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
					   DOCA_DPI_PORT, &tmp_rule)) {
			ret = -1;
			goto error;
		}
	}

	/* add regex rules */
	while (!rules_list->empty()) {
		rule = rules_list->back();
		rules_list->pop_back();

		if (rule.bitmask & DOCA_DPI_REGEX_MASK) {
			tmp_rule.reg_field.field = rule.field;
			/* update case insensitive for flag and http protocol */
			tmp_str = rule.regex;
			if ((rule.nocase == false && rule.l7_protocol == DOCA_DPI_HTTP) || rule.nocase == true)
				tmp_str.append("i");
			tmp_rule.reg_field.regex = (char *)tmp_str.c_str();
			free(rule.regex);
			if (rule.offset)
				tmp_rule.reg_field.offset = rule.offset;
			if (doca_dpi_compiler_add_rule(ctx, glob_opt.sid, gid, header.direction,
							DOCA_DPI_REGEX, &tmp_rule)) {
				ret = -1;
				goto error;
			}
		}
	}
	ret = 0;
error:
	free(glob_opt.classtype);
	free(glob_opt.msg);
	free(header.action);
	free(header.dst_ip);
	free(header.src_ip);
	return ret;
}

static int pcre_update_modifier(char mod, string &pcre_regex, struct doca_dpi_rule *rule,
				 bool &newline_check);

/* convert the pcre field to fit RXP syntax.
 */
int
pcre_converter(const char *pcre, struct doca_dpi_rule *rule)
{
	string pcre_regex = pcre;
	string pcre_out;
	bool newline_check = false;
	char modifier;
	size_t mod_idx = 0;

	if (pcre_regex.find('/') > 0) {
		printf("Error: pcre rule %s must start with \"/\"\n", pcre_regex.c_str());
		return -1;
	}

	mod_idx = pcre_regex.rfind('/');
	pcre_out = pcre_regex.substr(0, mod_idx + 1);
	while (++mod_idx < pcre_regex.length()) {
		modifier = pcre_regex[mod_idx];
		if (pcre_update_modifier(modifier, pcre_out, rule, newline_check)) {
			printf("Error: modifier %c in the pcre rule %s is not supported as Suricata syntax\n",
				modifier, pcre_regex.c_str());
			return -1;
		}
	}
	if (!newline_check) {
		/* add (?s) (?-s) to the pcre in order to exclude newline
		 * from the dot metacharacter
		 */
		pcre_out.replace(0, 1, "/(?s)");
		if (pcre_out.at(pcre_out.rfind('/') - 1) == '$')
			pcre_out.replace(pcre_out.rfind('$'), 1, "(?-s)$");
		else
			pcre_out.replace(pcre_out.rfind('/'), 1, "(?-s)/");
	}
	rule->regex = strdup(pcre_out.c_str());
	if (rule->regex == NULL) {
		printf("Error: memory allocation failed for pcre string %s\n", pcre_out.c_str());
		return -1;
	}
	return 0;
}

static int pcre_update_modifier(char mod, string &pcre_regex, struct doca_dpi_rule *rule,
				 bool &newline_check)
{
	if (mod == 'i') {
		rule->nocase = true;
		return 0;
	}
	if (mod == 's') {
		newline_check = true;
		return 0;
	}
	if (mod == 'A') {
		pcre_regex.replace(0, 1, "/^");
		return 0;
	}
	if (mod == 'I') {
		rule->l7_protocol = DOCA_DPI_HTTP;
		rule->field = DOCA_DPI_HTTP_URI;
		return 0;
	}
	if (mod == 'C') {
		rule->l7_protocol = DOCA_DPI_HTTP;
		rule->field = DOCA_DPI_HTTP_COOKIE;
		return 0;
	}
	if (mod == 'V') {
		rule->l7_protocol = DOCA_DPI_HTTP;
		rule->field = DOCA_DPI_HTTP_USER_AGENT;
		return 0;
	}
	if (mod == 'W') {
		rule->l7_protocol = DOCA_DPI_HTTP;
		rule->field = DOCA_DPI_HTTP_HOST;
		return 0;
	}
	if (mod == 'P') {
		rule->l7_protocol = DOCA_DPI_HTTP;
		rule->field = DOCA_DPI_HTTP_CONTENT;
		rule->direction = DOCA_DPI_CTS;
		return 0;
	}
	if (mod == 'Q') {
		rule->l7_protocol = DOCA_DPI_HTTP;
		rule->field = DOCA_DPI_HTTP_CONTENT;
		rule->direction = DOCA_DPI_STC;
		return 0;
	}
	return -1;
}
