/*
 * @file    rxp_qosmos_class.cpp
 * @author  Mark McMullan <mmcmullan@mellanox.com>
 *
 * @section LICENSE
 *
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "rxp_qosmos_class.hpp"
#include "rxp_qosmos_general_utils.hpp"
#include <iostream>
#include <sstream>
#include <map>

using namespace std;

rxp_qosmos_class::rxp_qosmos_class() {}

rxp_qosmos_class::~rxp_qosmos_class() {}


bool rxp_qosmos_class::convert_qosmos_rule_to_rxp(string& in_rule, string& out_rule, string& error_text)
{
    string qosmos_rule = "";
    string modifiers = "";
    bool rule_found = true;
    bool rule_start_anchor;
    bool rule_end_anchor;
    if (in_rule[0] == '/')
    {
        bool modifiers_found = false;
        for (int i = in_rule.length() - 1; i >= 0; i--)
        {
            if (in_rule[i] == '/')
            {
                if (modifiers_found)
                {
                    modifiers = in_rule.substr(i + 1, in_rule.length() - i - 1);
                }
                qosmos_rule = in_rule.substr(1, i - 1);
                if (i == 0)
                {
                    rule_found = false;
                }
                break;
            }
            else
            {
                modifiers_found = true;
            }
        }
        if (!rule_found)
        {
            error_text = "Found '/' at start of rule but not at end";
            return false;
        }
    }
    else
    {
        qosmos_rule = in_rule;
    }

    try
    {
        rule_start_anchor = true;
        rule_end_anchor = true;
        bool rule_is_group = false;

        string in_rule_aux = qosmos_rule;
        string out_rule_aux = qosmos_rule;

        while(in_rule_aux != out_rule)
        {
            in_rule_aux = out_rule_aux;
            out_rule_aux = "";

            bool rule_start_anchor_aux = true;
            bool data_found = false;
            int end_dot_star_index = 0;
            bool end_dot_star_found = false;

            scan_and_convert_qosmos_rule(in_rule_aux, out_rule_aux, data_found, rule_start_anchor_aux, end_dot_star_found,
                                         end_dot_star_index, rule_is_group);

            out_rule = "";
            // If the last data in the rule is a .*, remove it and do not use a $ anchor
            if (!rule_start_anchor_aux)
            {
                rule_start_anchor = false;
            }
            if (end_dot_star_found && !data_found)
            {
                out_rule = out_rule_aux.substr(0,end_dot_star_index - 1);
                out_rule += out_rule_aux.substr(end_dot_star_index + 1, out_rule_aux.length() - (end_dot_star_index +1));

                rule_end_anchor = false;
            }
            else
            {
                out_rule += out_rule_aux;
            }
            // Removing a .* may leave an empty group, rxpc fails if it finds this so remove any that are found
            bool is_escaped = false;
            out_rule_aux = "";
            for (unsigned int i = 0; i < out_rule.length(); i++)
            {
                if (out_rule[i] == '\\')
                {
                    is_escaped = true;
                    out_rule_aux += out_rule[i];
                    continue;
                }

                if (!is_escaped && i < out_rule.length() &&  out_rule[i] == '(' && out_rule[i+1] == ')')
                {
                    i++;
                    continue;
                }

                out_rule_aux += out_rule[i];
            }
        }

        out_rule = "/";
        if (rule_start_anchor)
        {
            out_rule += "^";
        }

        if (!rule_is_group && (rule_start_anchor || rule_end_anchor))
        {
            out_rule += "(";
        }
        out_rule += out_rule_aux;
        if (!rule_is_group  && (rule_start_anchor || rule_end_anchor))
        {
            out_rule += ")";
        }

        if (rule_end_anchor)
        {
            out_rule += "$";
        }

        out_rule += "/" + modifiers;
    }
    catch (string& str)
    {
        error_text = str;
        return false;
    }
    return true;
}

void rxp_qosmos_class::scan_and_convert_qosmos_rule(string& in_rule, string& out_rule, bool& data_found, bool& rule_start_anchor, bool& end_dot_star_found, int& end_dot_star_index, bool& rule_is_group)
{
    int char_pointer = 0;
    unsigned char c;
    string current_literal = "";
    string current_rule = in_rule;
    string current_error = "";

    bool is_quoted = false;
    bool in_case_insensitive = false;
    bool initial_paren = false;

    while(char_pointer < (int)in_rule.length())
    {
        c = get_c(char_pointer, current_rule);

        switch (c)
        {
            case '\\':
                c = get_c(char_pointer,current_rule);
                switch (c)
                {
                    case 'E':
                        is_quoted = false;
                        //Not a literal
                        //Add to rule
                        current_literal += '\\';
                        current_literal += c;
                        break;
                    case 'Q':
                        is_quoted = true;
                        //Not a literal
                        //Add to rule
                        current_literal += '\\';
                        current_literal += c;
                        break;
                    case 'a':
                    case 'c':
                    case 'e':
                    case 'f':
                    case 'n':
                    case 'r':
                    case 't':
                    case 'C':
                    case 'd':
                    case 'D':
                    case 'h':
                    case 'H':
                    case 'i':
                    case 'I':
                    case 's':
                    case 'S':
                    case 'v':
                    case 'V':
                    case 'w':
                    case 'W':
                        //Add to rule and is literal
                        current_literal += '\\';
                        current_literal += c;
                        data_found = true;
                        break;
                    case 'l':
                    case 'L':
                    case 'N':
                    case 'p':
                    case 'P':
                    case 'R':
                    case 'X':
                    case 'b':
                    case 'B':
                    case 'A':
                    case 'Z':
                    case 'z':
                    case 'G':
                    case 'K':
                    case 'k':
                    case 'u':
                    case 'U':
                        //Add to rule
                        current_literal += '\\';
                        current_literal += c;
                        break;
                    case 'g':
                        current_literal += '\\';
                        current_literal += c;
                        if(peek_c(char_pointer, current_rule) == '{')
                        {
                            //\g{n}
                            c = get_c(char_pointer, current_rule);
                            current_literal += c;
                            if(peek_c(char_pointer, current_rule) == '-')
                            {
                                //\g{-n}
                                c = get_c(char_pointer, current_rule);
                                current_literal += c;
                            }
                            if(isdigit(peek_c(char_pointer,current_rule)))
                            {
                                data_found = true;
                            }
                        }
                        break;/* If we found an hex we need to set that \x00 is found */
                    case 'x':
                    {
                        //\x{h}  character with hex code h
                        //\x{hh}  character with hex code hh
                        //\x{hhh}  character with hex code hhh
                        //\x{hhhh}  character with hex code hhhh
                        current_literal += '\\';
                        current_literal += c;

                        string hex = "";
                        string numeric_string = "";
                        if(get_c(char_pointer, current_rule) == '{')
                        {
                            int count_hex_digits = 0;
                            //Check how many hex digits are in the sequence
                           // current_literal += get_c(char_pointer, current_rule);
                            while(c != '}')
                            {
                                c = get_c(char_pointer, current_rule);
                                if(c != '}')
                                {
                                    hex += c;
                                    count_hex_digits++;
                                }
                            }
                            compile_hex_string(hex, current_literal);
                            //Unget all the characters as they have just been counted
                            //And not dealt with properly i.e. numerical value stored
                            unget_c(char_pointer);
                            while(peek_c(char_pointer, current_rule) != '{')
                            {
                                unget_c(char_pointer);
                            }

                            get_c(char_pointer, current_rule);
                            while(peek_c(char_pointer, current_rule) != '}')
                            {
                                numeric_string += compile_numeric_value(char_pointer, current_rule, BASE_HEX, true,
                                        (count_hex_digits%2) ? 1 : 0);
                                count_hex_digits = 0;
                            }
                            get_c(char_pointer, current_rule);
                        }
                        else
                        {
                            unget_c(char_pointer);

                            // Only \xnn is supported so only get the next two characters
                            if (isxdigit(peek_c(char_pointer, current_rule)))
                            {
                                hex += get_c(char_pointer, current_rule);
                            }
                            else
                            {
                                throw message_specific_error("invalid hex character found after \\x", char_pointer);
                            }
                            if (char_pointer < (int)current_rule.size() && isxdigit(peek_c(char_pointer, current_rule)))
                            {
                                hex += get_c(char_pointer, current_rule);
                            }

                            compile_hex_string(hex, current_literal);

                            /*while(isxdigit(peek_c(char_pointer, current_rule)))
                            {
                                hex += get_c(char_pointer, current_rule);
                            }

                            int char_pointer_aux = char_pointer + 1;
                            if (char_pointer >= current_rule.size() ||
                                (peek_c(char_pointer, current_rule) != '\\' && peek_c(char_pointer_aux, current_rule)))
                            {
                                stringstream ss;
                                ss << "Unclosed hexadecimal group" << endl;
                                throw ss.str();
                            }

                            compile_hex_string(hex, current_literal);

                            if (hex.length() > 2 && char_pointer < current_rule.size())
                            {
                                // a hex string longer than 2 means that the construct is \x....\x, the last \x needs to be removed
                                get_c(char_pointer, current_rule);
                                get_c(char_pointer, current_rule);
                            }*/
                        }

                        data_found = true;
                        break;
                    }
                    case '[':
                    {
                        current_literal += "(?i)";
                        in_case_insensitive = true;
                        break;
                    }
                    case ']':
                    {
                        if (in_case_insensitive)
                        {
                            current_literal += "(?-i)";
                            in_case_insensitive = false;
                        }
                        break;
                    }
                    default:
                        //Add to rule and is literal
                        current_literal += '\\';
                        current_literal += c;
                        data_found = true;
                        break;
                }
                break;
                    case '[':
                        current_literal += c;
                        data_found = true;

                        if (!is_quoted)
                        {
                            //Check how many hex digits are in the sequence
                            current_literal += get_c(char_pointer, current_rule);
                            char name_delimiter = ']';
                            capture_until_limit(char_pointer, current_rule, is_quoted, name_delimiter, current_literal);
                        }
                        break;
                    case '(':
                        current_literal += c;
                        if(is_quoted)
                        {
                            data_found = true;
                        }
                        else
                        {
                            if (char_pointer == 1)
                            {
                                initial_paren = true;
                            }
                            c = get_c(char_pointer, current_rule);
                            switch (c)
                            {
                                case '?':
                                {
                                    current_literal += c;
                                    bool carry_on = true;
                                    bool first = true;
                                    while(carry_on)
                                    {
                                        c = get_c(char_pointer, current_rule);
                                        if(first && isdigit(c))
                                        {
                                            current_literal += c;
                                            break;
                                        }
                                        switch (c)
                                        {
                                            case '<':
                                            {
                                                current_literal += c;
                                                if(first)
                                                {
                                                    c = get_c(char_pointer, current_rule);
                                                    switch (c)
                                                    {
                                                        case '=':
                                                        case '!':
                                                        {
                                                            break;
                                                        }
                                                        default:
                                                        {
                                                            current_literal += c;
                                                            char name_delimiter = '>';
                                                            capture_until_limit(char_pointer, current_rule, is_quoted, name_delimiter, current_literal);
                                                            break;
                                                        }
                                                    }
                                                }
                                                carry_on = false;
                                                break;
                                            }
                                            case '\'':
                                            {
                                                current_literal += c;
                                                if(first)
                                                {
                                                    c = get_c(char_pointer, current_rule);
                                                    current_literal += c;
                                                    char name_delimiter = '\'';
                                                    capture_until_limit(char_pointer, current_rule, is_quoted, name_delimiter, current_literal);
                                                }
                                                carry_on = false;
                                                break;
                                            }
                                            case 'P':
                                                current_literal += c;
                                                if(first)
                                                {
                                                    c = get_c(char_pointer, current_rule);
                                                    current_literal += c;
                                                    switch (c)
                                                    {
                                                        case '<':
                                                        {
                                                            char name_delimiter = '>';
                                                            capture_until_limit(char_pointer, current_rule, is_quoted, name_delimiter, current_literal);
                                                            break;
                                                        }
                                                        case '=':
                                                        {
                                                            char name_delimiter = ')';
                                                            capture_until_limit(char_pointer, current_rule, is_quoted, name_delimiter, current_literal);
                                                            break;
                                                        }
                                                        default:
                                                            current_literal += c;
                                                            break;
                                                    }
                                                }
                                                carry_on = false;
                                                break;
                                            case '>':
                                            case '=':
                                            case '!':
                                            case 'R':
                                            case '+':
                                            case '&':
                                            case 'J':
                                            case 'U':
                                            case 'C':
                                                current_literal += c;
                                                carry_on = false;
                                                break;
                                            case 's':
                                            case 'm':
                                            case 'i':
                                            case 'x':
                                                current_literal += c;
                                                break;
                                            case '-':
                                                if(first && isdigit(peek_c(char_pointer, current_rule)))
                                                {
                                                    carry_on = false;
                                                }
                                                current_literal += c;
                                                break;
                                            case ':':
                                                current_literal += c;
                                                carry_on = false;
                                                break;
                                            case '(':
                                            case '#':
                                                current_literal += c;
                                                if(first)
                                                {
                                                    c = get_c(char_pointer, current_rule);
                                                    current_literal += c;
                                                    char name_delimiter = ')';
                                                    capture_until_limit(char_pointer, current_rule, is_quoted, name_delimiter, current_literal);
                                                }
                                                carry_on = false;
                                                break;
                                            default:
                                                current_literal += c;
                                                carry_on = false;
                                                break;
                                        }
                                        first = false;
                                    }
                                    break;
                                }
                                case '*':
                                case ')':
                                    //Add to rule
                                    current_literal += c;
                                    break;
                                default:
                                    //Not a literal
                                    unget_c(char_pointer);
                                    break;
                            }
                        }
                        break;
                    case '{':
                        current_literal += c;
                        data_found = true;

                        if (!is_quoted)
                        {
                            //Check how many hex digits are in the sequence
                            current_literal += get_c(char_pointer, current_rule);
                            char name_delimiter = '}';
                            capture_until_limit(char_pointer, current_rule, is_quoted, name_delimiter, current_literal);
                        }
                        break;
                    case '/':
                    {
                        data_found = true;
                        if (is_quoted)
                        {
                            current_literal += c;
                        }
                        else
                        {
                            current_literal += "\\/";
                        }
                        break;
                    }
                    case '^':
                    case '$':
                    {
                        if (is_quoted)
                        {
                            current_literal += c;
                            data_found = true;
                        }
                        else
                        {
                            throw message_specific_error("non-escaped anchor found in rule", char_pointer);
                        }
                        break;
                    }
                    case ')':
                    {
                        current_literal += c;
                        if(is_quoted)
                        {
                            data_found = true;
                        }
                        else if (initial_paren && char_pointer == (int)in_rule.length())
                        {
                            rule_is_group = true;
                        }
                        break;
                    }
                    case '*':
                    case '+':
                    case '|':
                    case '?':
                        current_literal += c;
                        if(is_quoted)
                        {
                            //Add to rule and is literal
                            data_found = true;
                        }
                        break;
                    case '.':
                    {
                        if(is_quoted)
                        {
                            //Add to rule and is literal
                            current_literal += c;
                            data_found = true;
                        }
                        else if (!data_found && char_pointer < (int)current_rule.length())
                        {
                            // If the initial data in the rule is a .*, it needs to be removed and the
                            // rule should not be anchored
                            if (peek_c(char_pointer, current_rule) == '*')
                            {
                                get_c(char_pointer, current_rule);
                                rule_start_anchor = false;
                                data_found = true;
                            }
                        }
                        else
                        {
                            current_literal += c;

                            if (peek_c(char_pointer, current_rule) == '*')
                            {
                                end_dot_star_found = true;
                                end_dot_star_index = out_rule.length() + 1;
                                data_found = false;
                            }
                        }
                        break;
                    }
                    default:
                        //Add to rule and is literal
                        current_literal += c;
                        data_found = true;
                        break;
        }

        //add the current literal to the rule
        out_rule += current_literal;
        current_literal = "";
    }
}

void rxp_qosmos_class::capture_until_limit(int & char_pointer,
        string & current_rule,
        bool & is_quoted,
        char name_delimiter,
        string & current_literal)
{
    unsigned char c;
    bool is_quoted_one = false;
    bool continue_match = true;
    while(continue_match)
    {
        c = get_c(char_pointer, current_rule);
        current_literal += c;

        if(is_quoted && c == '\\' && peek_c(char_pointer, current_rule) == 'Q')
        {
            is_quoted = false;
        }
        else if(is_quoted_one)
        {
            is_quoted_one = false;
        }
        if(c == '\\' && peek_c(char_pointer, current_rule) == 'E')
        {
            is_quoted = true;
        }
        else if(c == '\\')
        {
            is_quoted_one = true;
        }
        if(c == name_delimiter && !is_quoted_one && !is_quoted)
        {
            continue_match = false;
        }
    }
}


void rxp_qosmos_class::unget_c(int & char_pointer)
{
    if(char_pointer<1)
    {
        throw message_specific_error("character pointer before start of rules", char_pointer);
    }
    char_pointer--;
}

unsigned char rxp_qosmos_class::peek_c(int & char_pointer, string & current_rule)
{
    return current_rule[char_pointer];
}
unsigned char rxp_qosmos_class::get_c(int & char_pointer, std::string & current_rule)
{
    if(char_pointer == (int)current_rule.size())
    {
        throw message_specific_error("unclosed group, character pointer has exceeded the rule length", char_pointer);
    }
    return current_rule[char_pointer++];
}

unsigned char rxp_qosmos_class::compile_numeric_value(int & char_pointer, string & current_rule, rxpc_number_base base, bool bracket_delimited, int override_number_of_characters)
{
    int maximum_number_of_characters, current_index = 0;
    unsigned char c;
    unsigned char result = 0;
    if(override_number_of_characters > 0)
    {
        maximum_number_of_characters = override_number_of_characters;
    }
    else
    {
        if(base == BASE_HEX)
        {
            maximum_number_of_characters = 2;
        }
        else
        {
            maximum_number_of_characters = 3;
        }
    }
    while(current_index < maximum_number_of_characters)
    {
        try
        {
            c = get_c(char_pointer, current_rule);
        }
        catch(string & msg)
        {
            return result;
        }
        if(isxdigit(c))
        {
            if(general_utilities::hex2int(c) >= base)
            {
                if(current_index > 0 && !bracket_delimited)
                {
                    unget_c(char_pointer);
                    return result;
                }
                else
                {
                    throw message_unexpected_char("numeric sequence", c, char_pointer);
                }
            }
            //255 is the maximum byte value
            //The following statement ensures
            //this is not exceeded
            uint32_t value = general_utilities::hex2int(c);
            if((result * base) + value > 255)
            {
                //result -= rxp_general_utilities::hex2int(c);
                unget_c(char_pointer);
                if(base == BASE_DECIMAL && bracket_delimited)
                {
                    throw message_specific_error("numbered reference exceeds 255", char_pointer);
                }
                return result;
            }
            result *= base;
            result += general_utilities::hex2int(c);
            current_index++;
        }
        else if(c == '}' && bracket_delimited)
        {
            unget_c(char_pointer);
            return result;
        }
        else if(c == ')' && bracket_delimited)
        {
            unget_c(char_pointer);
            return result;
        }
        else if(bracket_delimited)
        {
            throw message_unexpected_char("numeric sequence", c, char_pointer);
        }
        else
        {
            if(current_index > 0 && !bracket_delimited)
            {
                unget_c(char_pointer);
                return result;
            }
            else
            {
                throw message_unexpected_char("numeric escape sequence", c, char_pointer);
            }
        }
    }
    return result;
}

void rxp_qosmos_class::compile_hex_string(const std::string& hex_string, std::string& escaped_string)
{
    for (int i = 0; i < (int)hex_string.length(); i++)
    {
        if (i != 0 && i % 2 == 0)
        {
            escaped_string += "\\x";
        }
        escaped_string += hex_string[i];
    }
}

string rxp_qosmos_class::message_specific_error(string message, int & position)
{
    stringstream ss;
    ss << message << endl << "     :position " << position;
    return ss.str();
}

string rxp_qosmos_class::message_unexpected_char(string message, unsigned char & c, int & position)
{
    stringstream ss;
    ss << "unexpected character \"" << c << "\" found in " << message << endl << "     :position " << position;
    return ss.str();
}
