/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <arpa/inet.h>
#include <bsd/string.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/epoll.h>
#include <sys/signalfd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include <doca_argp.h>
#include <doca_log.h>
#include <doca_dev.h>

#include "socket_relay_core.h"

#define CC_MAX_QUEUE_SIZE 128					/* Maximum amount of messages on Comm Channel queue */
#define CC_MAX_EVENTS 2						/* CC maximum number of file descriptors on which epoll_wait can signal */
#define CC_MSG_NO_DATA (sizeof(uint8_t) + sizeof(uint32_t))	/* First 5 bytes are meta information */
#define CC_MSG_TERMINATION 1					/* Send one byte only for termination message */
#define CC_SEND_TIMEOUT 100					/* Comm Channel send timeout */
#define ACCEPT_CONN_MAX_EVENTS 2				/* Accept connections thread maximum number of events */
#define SOCKET_BACKLOG 32					/* Number of incoming connection waiting in the queue */
#define MAX_CONNECTIONS_NUM 512				/* Maximum number of incoming connections */
#define DEFAULT_ENTRIES_NUM 32					/* Default number of entries in fd to fd table */
#define SLEEP_IN_NANOS (10 * 1000)				/* Sample write() every 10 microseconds */
#define HOST_EPOLL_NUM 4					/* Total number of epoll instances needed */

DOCA_LOG_REGISTER(SOCKET_RELAY::CORE);

static bool stop_cc_sender;
static bool termination_due_to_err;
static bool server_disconnect;

/*
 * ARGP Callback - Handle Comm Channel service name parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
cc_name_callback(void *param, void *config)
{
	struct sr_cfg *cfg = (struct sr_cfg *)config;
	const char *cc_name = (char *)param;

	if (strnlen(cc_name, CC_MAX_NAME) == CC_MAX_NAME) {
		DOCA_LOG_ERR("Comm Channel service name exceeding the maximum size of %d", CC_MAX_NAME - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strlcpy(cfg->cc_name, cc_name, CC_MAX_NAME);

	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle server socket path parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
socket_path_callback(void *param, void *config)
{
	struct sr_cfg *cfg = (struct sr_cfg *)config;
	const char *uds_path = (char *)param;

	if (strnlen(uds_path, SOCKADDR_UN_PATH) == SOCKADDR_UN_PATH) {
		DOCA_LOG_ERR("Socket path exceeding the maximum size of %d", SOCKADDR_UN_PATH - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strlcpy(cfg->socket_path, uds_path, SOCKADDR_UN_PATH);

	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle Comm Channel DOCA device PCI address parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
dev_pci_addr_callback(void *param, void *config)
{
	struct sr_cfg *cfg = (struct sr_cfg *)config;
	const char *dev_pci_addr = (char *)param;

	if (strnlen(dev_pci_addr, PCI_ADDR_LEN) == PCI_ADDR_LEN) {
		DOCA_LOG_ERR("Comm Channel DOCA device PCI address exceeding the maximum size of %d", PCI_ADDR_LEN - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}

	strlcpy(cfg->cc_dev_pci_addr, dev_pci_addr, PCI_ADDR_LEN);

	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle Comm Channel DOCA device representor PCI address parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
rep_pci_addr_callback(void *param, void *config)
{
	struct sr_cfg *cfg = (struct sr_cfg *)config;
	const char *rep_pci_addr = (char *)param;

	if (cfg->mode == SR_MODE_DPU) {
		if (strnlen(rep_pci_addr, PCI_ADDR_LEN) == PCI_ADDR_LEN) {
			DOCA_LOG_ERR(
				"Comm Channel DOCA device representor PCI address exceeding the maximum size of %d",
				PCI_ADDR_LEN - 1);
			return DOCA_ERROR_INVALID_VALUE;
		}

		strlcpy(cfg->cc_dev_rep_pci_addr, rep_pci_addr, PCI_ADDR_LEN);
	}

	return DOCA_SUCCESS;
}

/*
 * Change socket configuration to non-blocking mode
 *
 * @fd [in]: File descriptor to make non-blocking
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
make_non_blocking(int fd)
{
	int flags, s;

	flags = fcntl(fd, F_GETFL, 0);
	if (flags == -1) {
		DOCA_LOG_ERR("Failed to retrieve socket status");
		return DOCA_ERROR_IO_FAILED;
	}

	flags |= O_NONBLOCK;
	s = fcntl(fd, F_SETFL, flags);
	if (s == -1) {
		DOCA_LOG_ERR("Failed to set a new socket status");
		return DOCA_ERROR_IO_FAILED;
	}

	return DOCA_SUCCESS;
}

/*
 * Parse string pci address into bdf struct
 *
 * @pci_addr [in]: PCI address string
 * @out_bdf [out]: doca_pci_bdf struct with the parsed string
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
parse_pci_addr(char const *pci_addr, struct doca_pci_bdf *out_bdf)
{
	unsigned int bus_bitmask = 0xFFFFFF00;
	unsigned int dev_bitmask = 0xFFFFFFE0;
	unsigned int func_bitmask = 0xFFFFFFF8;
	uint32_t tmpu;
	char tmps[4];

	if (pci_addr == NULL || strlen(pci_addr) != 7 || pci_addr[2] != ':' || pci_addr[5] != '.')
		return DOCA_ERROR_INVALID_VALUE;

	tmps[0] = pci_addr[0];
	tmps[1] = pci_addr[1];
	tmps[2] = '\0';
	tmpu = strtoul(tmps, NULL, 16);
	if ((tmpu & bus_bitmask) != 0)
		return DOCA_ERROR_INVALID_VALUE;
	out_bdf->bus = tmpu;

	tmps[0] = pci_addr[3];
	tmps[1] = pci_addr[4];
	tmps[2] = '\0';
	tmpu = strtoul(tmps, NULL, 16);
	if ((tmpu & dev_bitmask) != 0)
		return DOCA_ERROR_INVALID_VALUE;
	out_bdf->device = tmpu;

	tmps[0] = pci_addr[6];
	tmps[1] = '\0';
	tmpu = strtoul(tmps, NULL, 16);
	if ((tmpu & func_bitmask) != 0)
		return DOCA_ERROR_INVALID_VALUE;
	out_bdf->function = tmpu;

	return DOCA_SUCCESS;
}

/*
 * Open DOCA device representor
 *
 * @pci_addr [in]: User PCI address input
 * @dev [in]: DOCA device to look at for representors
 * @dev_rep [out]: DOCA device representor that was opened
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
open_doca_device_rep(char *pci_addr, struct doca_dev *dev, struct doca_dev_rep **dev_rep)
{
	doca_error_t result;
	uint32_t dev_count;
	struct doca_devinfo_rep **devinfo_rep_list;
	struct doca_devinfo_rep *devinfo_queried = NULL;
	struct doca_pci_bdf queried_pci_bdf;
	struct doca_pci_bdf user_pci_bdf;
	int it;

	result = doca_devinfo_rep_list_create(dev, DOCA_DEV_REP_FILTER_NET, &devinfo_rep_list, &dev_count);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create DOCA device representors list");
		return result;
	}

	if (dev_count <= 0) {
		DOCA_LOG_ERR("No Hardware devices representors were found");
		doca_devinfo_rep_list_destroy(devinfo_rep_list);
		result = DOCA_ERROR_NOT_FOUND;
		return result;
	}

	result = parse_pci_addr(pci_addr, &user_pci_bdf);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse Comm Channel DOCA device representor PCI address");
		doca_devinfo_rep_list_destroy(devinfo_rep_list);
		return result;
	}

	for (it = 0; it < dev_count; ++it) {
		result = doca_devinfo_rep_get_pci_addr(devinfo_rep_list[it], &queried_pci_bdf);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_WARN("Failed to retrieve PCI address for DOCA device representor");
			continue;
		}

		if (user_pci_bdf.raw != queried_pci_bdf.raw)
			continue;

		DOCA_LOG_DBG("Found a DOCA device representor according to the provided PCI address");

		devinfo_queried = devinfo_rep_list[it];

		break;
	}

	if (devinfo_queried == NULL) {
		DOCA_LOG_ERR("Failed to find DOCA device representor");
		return DOCA_ERROR_NOT_FOUND;
	}

	result = doca_dev_rep_open(devinfo_queried, dev_rep);
	if (result != DOCA_SUCCESS)
		DOCA_LOG_ERR("Failed to open DOCA device representor");

	doca_devinfo_rep_list_destroy(devinfo_rep_list);

	return result;
}

/*
 * Open DOCA device based on PCI address
 *
 * @pci_addr [in]: User PCI address
 * @dev [out]: DOCA device that was opened
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
open_doca_device(char *pci_addr, struct doca_dev **dev)
{
	doca_error_t result;
	uint32_t dev_count;
	struct doca_devinfo **devinfo_list;
	struct doca_devinfo *devinfo_queried = NULL;
	struct doca_pci_bdf dev_pci_bdf;
	struct doca_pci_bdf queried_pci_bdf;
	int it;

	result = doca_devinfo_list_create(&devinfo_list, &dev_count);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create DOCA device list");
		return result;
	}

	if (dev_count <= 0) {
		DOCA_LOG_ERR("No Hardware devices were found");
		doca_devinfo_list_destroy(devinfo_list);
		result = DOCA_ERROR_NOT_FOUND;
		return result;
	}

	result = parse_pci_addr(pci_addr, &dev_pci_bdf);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse Comm Channel DOCA device PCI address");
		doca_devinfo_list_destroy(devinfo_list);
		return result;
	}

	for (it = 0; it < dev_count; ++it) {
		result = doca_devinfo_get_pci_addr(devinfo_list[it], &queried_pci_bdf);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_WARN("Failed to retrieve PCI address for DOCA device");
			continue;
		}

		if (dev_pci_bdf.raw != queried_pci_bdf.raw)
			continue;

		DOCA_LOG_DBG("Found DOCA device according to the provided PCI address");

		devinfo_queried = devinfo_list[it];

		break;
	}

	if (devinfo_queried == NULL) {
		DOCA_LOG_ERR("Failed to find DOCA device according to PCI address:%s", pci_addr);
		doca_devinfo_list_destroy(devinfo_list);
		return DOCA_ERROR_NOT_FOUND;
	}

	result = doca_dev_open(devinfo_queried, dev);
	if (result != DOCA_SUCCESS)
		DOCA_LOG_ERR("Failed to open DOCA device");

	doca_devinfo_list_destroy(devinfo_list);

	return result;
}

/*
 * Set Comm Channel properties
 *
 * @mode [in]: Mode of operation
 * @ep [in]: Comm Channel endpoint
 * @dev [in]: DOCA device object to use
 * @dev_rep [in]: DOCA device representor object to use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
set_cc_properties(enum sr_mode mode, struct doca_comm_channel_ep_t *ep, struct doca_dev *dev,
			struct doca_dev_rep *dev_rep)
{
	doca_error_t result;

	result = doca_comm_channel_ep_set_device(ep, dev);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set DOCA device property");
		return result;
	}

	result = doca_comm_channel_ep_set_max_msg_size(ep, CC_MAX_MSG_SIZE);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set max_msg_size property");
		return result;
	}

	result = doca_comm_channel_ep_set_send_queue_size(ep, CC_MAX_QUEUE_SIZE);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set snd_queue_size property");
		return result;
	}

	result = doca_comm_channel_ep_set_recv_queue_size(ep, CC_MAX_QUEUE_SIZE);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set rcv_queue_size property");
		return result;
	}

	if (mode == SR_MODE_DPU) {
		result = doca_comm_channel_ep_set_device_rep(ep, dev_rep);
		if (result != DOCA_SUCCESS)
			DOCA_LOG_ERR("Failed to set DOCA device representor property");
	}

	return result;
}

/*
 * Kill all running threads by sending a signal
 *
 * @sig [in]: Signal to send
 */
static void
kill_all_threads(int sig)
{
	kill(getpid(), sig);
}

/*
 * Increase/decrease connections counter by one
 *
 * @conn_count [in]: Connection count
 * @is_increase [in]: True if increase is needed, false otherwise
 * @mutex [in]: Connection count mutex
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
change_conn_count(int *conn_count, bool is_increase, pthread_mutex_t *mutex)
{
	int pthread_res;

	pthread_res = pthread_mutex_lock(mutex);
	if (pthread_res != 0) {
		DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
		return DOCA_ERROR_NO_LOCK;
	}
	if (is_increase)
		(*conn_count)++;
	else
		(*conn_count)--;

	pthread_res = pthread_mutex_unlock(mutex);
	if (pthread_res != 0) {
		DOCA_LOG_ERR("Failed to unlock resource, error=%d", errno);
		return DOCA_ERROR_NO_LOCK;
	}

	return DOCA_SUCCESS;
}

/*
 * Block all threads from receiving SIGINT and SIGTERM and generate signal file descriptor
 * which will be added to Comm Channel main epoll to receive user SIGINT (ctrl + c) from the shell
 * to close the application gracefully
 *
 * @mode [in]: Mode of operation
 * @cc_send_intr_fd [out]: Comm Channel send threads interrupt file descriptor
 * @cc_recv_intr_fd [out]: Comm Channel receive threads interrupt file descriptor
 * @uds_clients_intr_fd [out]: UDS clients listener threads interrupt file descriptor
 * @accept_conn_intr_fd [out]: New connection acceptor thread interrupt file descriptor
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
create_signals_fd(enum sr_mode mode, int *cc_send_intr_fd, int *cc_recv_intr_fd, int *uds_clients_intr_fd,
		  int *accept_conn_intr_fd)
{
	sigset_t signal_mask;
	int intr_fd, idx, interrupt_fds = HOST_EPOLL_NUM;
	int *fd_array[HOST_EPOLL_NUM] = {
		cc_send_intr_fd,
		cc_recv_intr_fd,
		uds_clients_intr_fd,
		accept_conn_intr_fd
	};

	sigemptyset(&signal_mask);
	sigaddset(&signal_mask, SIGINT);
	sigaddset(&signal_mask, SIGTERM);

	/* Block all future threads on INT and TER signals in order to perform normal exit */
	if (sigprocmask(SIG_BLOCK, &signal_mask, NULL) != 0) {
		DOCA_LOG_ERR("Failed to create blocked signal mask, error=%d", errno);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	if (mode == SR_MODE_DPU)
		interrupt_fds--;

	/* Create signal file descriptors to catch ctrl + c signal from shell for all relevant threads */
	for (idx = 0; idx < interrupt_fds; idx++) {
		intr_fd = signalfd(-1, &signal_mask, 0);
		if (intr_fd == -1) {
			DOCA_LOG_ERR("Failed to create interrupt file descriptor, error=%d", errno);
			return DOCA_ERROR_IO_FAILED;
		}
		*(fd_array[idx]) = intr_fd;
	}

	return DOCA_SUCCESS;
}

/*
 * Create 2 epoll files descriptor on the DPU and 3 on the Host
 * Comm Channel epoll will be triggered once a new message was received
 * Accept new connection epoll will be triggered when a new client wants to connect
 * UDS clients listener epoll will be triggered if there is a new client message to read
 *
 * @mode [in]: Mode of operation
 * @cc_send_epoll_fd [out]: Comm Channel send epoll file descriptor
 * @cc_recv_epoll_fd [out]: Comm Channel receive epoll file descriptor
 * @uds_clients_epoll_fd [out]: UDS clients epoll file descriptor
 * @accept_conn_epoll_fd [out]: New connection epoll file descriptor
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
create_epoll_fds(enum sr_mode mode, int *cc_send_epoll_fd, int *cc_recv_epoll_fd, int *uds_clients_epoll_fd,
		 int *accept_conn_epoll_fd)
{
	int epoll_fd, idx, epoll_instances = HOST_EPOLL_NUM;
	int *fd_array[HOST_EPOLL_NUM] = {
		cc_send_epoll_fd,
		cc_recv_epoll_fd,
		uds_clients_epoll_fd,
		accept_conn_epoll_fd
	};

	if (mode == SR_MODE_DPU)
		epoll_instances--;

	for (idx = 0; idx < epoll_instances; idx++) {
		epoll_fd = epoll_create1(0);
		if (epoll_fd == -1) {
			DOCA_LOG_ERR("Failed to create epoll instance, error=%d", errno);
			return DOCA_ERROR_IO_FAILED;
		}
		*(fd_array[idx]) = epoll_fd;
	}

	return DOCA_SUCCESS;
}

/*
 * Close all opened file descriptors
 *
 * @mode [in]: Mode of operation
 * @cc_send_epoll_fd [in]: Comm Channel send epoll file descriptor
 * @cc_recv_epoll_fd [in]: Comm Channel receive epoll file descriptor
 * @accept_conn_epoll_fd [in]: UDS clients listener epoll file descriptor
 * @uds_clients_epoll_fd [in]: New connection acceptor epoll file descriptor
 * @cc_send_intr_fd [out]: Comm Channel send threads interrupt file descriptor
 * @cc_recv_intr_fd [out]: Comm Channel receive threads interrupt file descriptor
 * @clients_intr_fd [out]: UDS clients listener threads interrupt file descriptor
 * @accept_conn_intr_fd [out]: New connection acceptor thread interrupt file descriptor
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
close_fds(enum sr_mode mode, int cc_send_epoll_fd, int cc_recv_epoll_fd, int accept_conn_epoll_fd,
	  int uds_clients_epoll_fd, int cc_send_intr_fd, int cc_recv_intr_fd, int clients_intr_fd,
	  int accept_conn_intr_fd)
{
	int idx, fd_count = HOST_EPOLL_NUM;
	int epoll_array[HOST_EPOLL_NUM] = {
		cc_send_epoll_fd,
		cc_recv_epoll_fd,
		uds_clients_epoll_fd,
		accept_conn_epoll_fd
	};
	int intr_array[HOST_EPOLL_NUM] = {
		cc_send_intr_fd,
		cc_recv_intr_fd,
		clients_intr_fd,
		accept_conn_intr_fd
	};

	if (mode == SR_MODE_DPU)
		fd_count--;

	for (idx = 0; idx < fd_count; idx++) {
		close(epoll_array[idx]);
		close(intr_array[idx]);
	}

	return DOCA_SUCCESS;
}

/*
 * Close all opened UDS clients descriptors on Host to DPU or DPU to Host table and free relevant memory
 *
 * @key [in]: Key to access which holds the client file descriptor
 * @value [in]: Not relevant
 * @user_data [in]: Always NULL
 * @return: TRUE on success and FALSE otherwise
 */
static gboolean
close_fd(void *key, void *value, void *user_data)
{
	close(*(int *)key);
	return TRUE;
}

/*
 * Free allocated memory for the application message and message context structures
 *
 * @ctx [in]: Message context to free
 */
static void
free_sr_msg_ctx(void *ctx)
{
	struct sr_message_ctx *msg_ctx = (struct sr_message_ctx *)ctx;

	free(msg_ctx->msg);
	free(msg_ctx);
}

/*
 * Initiate Glib data structures
 *
 * @mode [in]: Mode of operation
 * @host_to_dpu_fd_table [out]: DPU side data structure which hold Host client fd as a key and DPU client fd as a value
 * @dpu_to_host_fd_table [out]: DPU side data structure which hold DPU client fd as a key and Host client fd as a value
 * @host_fd_set [out]: Host side data structure which hold all UDS clients file descriptors
 * @msg_queue [out]: Shared data structure to both sides which hold all message need to be sent on Comm Channel
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
sr_init_glib_structures(enum sr_mode mode, GHashTable **host_to_dpu_fd_table, GHashTable **dpu_to_host_fd_table,
			GHashTable **host_fd_set, GQueue **msg_queue)
{
	GHashTable *table;
	GQueue *queue;

	queue = g_queue_new();
	if (queue == NULL) {
		DOCA_LOG_ERR("Failed to allocate messages queue");
		return DOCA_ERROR_NO_MEMORY;
	}
	*msg_queue = queue;

	if (mode == SR_MODE_DPU) {
		table = g_hash_table_new_full(g_int_hash, g_int_equal, g_free, g_free);
		if (table == NULL) {
			DOCA_LOG_ERR("Failed to allocate Host to DPU file descriptor table");
			return DOCA_ERROR_NO_MEMORY;
		}
		*host_to_dpu_fd_table = table;

		table = g_hash_table_new(g_int_hash, g_int_equal);
		if (table == NULL) {
			DOCA_LOG_ERR("Failed to allocate DPU to Host file descriptor table");
			return DOCA_ERROR_NO_MEMORY;
		}
		*dpu_to_host_fd_table = table;
	} else {
		table = g_hash_table_new_full(g_int_hash, g_int_equal, g_free, NULL);
		if (table == NULL) {
			DOCA_LOG_ERR("Failed to allocate file descriptors set");
			return DOCA_ERROR_NO_MEMORY;
		}
		*host_fd_set = table;
	}

	return DOCA_SUCCESS;
}

/*
 * Initiate pthread relevant data structures
 *
 * @mode [in]: Mode of operation
 * @cc_mutex [out]: Comm Channel mutex
 * @queue_mutex [out]: Comm Channel send queue mutex
 * @table_mutex [out]: DPU to Host file descriptor table mutex
 * @conn_count_mutex [out]: Connections counter mutex
 * @queue_cond [out]: Comm Channel send queue condition variable
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
init_pthread(enum sr_mode mode, pthread_mutex_t *cc_mutex, pthread_mutex_t *queue_mutex,
		pthread_mutex_t *table_mutex, pthread_mutex_t *conn_count_mutex, pthread_cond_t *queue_cond)
{
	if (pthread_mutex_init(cc_mutex, NULL) != 0) {
		DOCA_LOG_ERR("Comm Channel lock initialization failed, error=%d", errno);
		return DOCA_ERROR_NO_LOCK;
	}

	if (pthread_mutex_init(queue_mutex, NULL) != 0) {
		DOCA_LOG_ERR("Comm Channel lock initialization failed, error=%d", errno);
		goto failed_init_queue_mutex;
	}

	if (pthread_mutex_init(table_mutex, NULL) != 0) {
		DOCA_LOG_ERR("Connection count lock initialization failed, error=%d", errno);
		goto failed_init_table_mutex;
	}

	if (pthread_cond_init(queue_cond, NULL) != 0) {
		DOCA_LOG_ERR("Comm Channel messages queue sync init failed, error=%d", errno);
		goto failed_init_cond_var;
	}

	if (mode == SR_MODE_HOST) {
		if (pthread_mutex_init(conn_count_mutex, NULL) != 0) {
			DOCA_LOG_ERR("Connection count lock initialization failed, error=%d", errno);
			goto failed_init_conn_mutex;
		}
	}

	return DOCA_SUCCESS;

failed_init_conn_mutex:
	pthread_cond_destroy(queue_cond);
failed_init_cond_var:
	pthread_mutex_destroy(table_mutex);
failed_init_table_mutex:
	pthread_mutex_destroy(queue_mutex);
failed_init_queue_mutex:
	pthread_mutex_destroy(cc_mutex);

	return DOCA_ERROR_NO_LOCK;
}

/*
 * Allocate memory for message and message context structures
 *
 * @msg_ctx [out]: Message context to allocate
 * @msg [out]: Message to allocate
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
allocate_cc_msg(struct sr_message_ctx **msg_ctx, struct sr_message **msg)
{
	/* Allocate new message context and a new message context */
	*msg_ctx = (struct sr_message_ctx *)malloc(sizeof(struct sr_message_ctx));
	if (*msg_ctx == NULL) {
		DOCA_LOG_ERR("Failed to allocate new message context for new connection");
		return DOCA_ERROR_NO_MEMORY;
	}
	*msg = (struct sr_message *)malloc(sizeof(struct sr_message));
	if (*msg == NULL) {
		DOCA_LOG_ERR("Failed to allocate new message for new connection");
		free(*msg_ctx);
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

/*
 * Send a message to Comm Channel
 *
 * @cc_send_epoll_fd [in]: Comm Channel send epoll file descriptor
 * @cc_send_intr_fd [in]: Comm Channel send interrupt file descriptor
 * @ep [in]: Comm Channel endpoint
 * @peer_addr [in]: Comm Channel peer to send message to
 * @cc_mutex [in]: Comm Channel mutex
 * @msg [in]: Message to send
 * @bytes_to_send [in]: Number of bytes to send
 */
static void
send_msg_to_cc(int cc_send_epoll_fd, int cc_send_intr_fd, struct doca_comm_channel_ep_t *ep, struct doca_comm_channel_addr_t *peer_addr,
	       pthread_mutex_t *cc_mutex, struct sr_message *msg, size_t bytes_to_send)
{
	struct epoll_event events[CC_MAX_EVENTS];
	doca_error_t result;
	int pthread_res, nfds, fd_idx;

	while (1) {
		pthread_res = pthread_mutex_lock(cc_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
			return;
		}
		result = doca_comm_channel_ep_sendto(ep, msg, bytes_to_send, DOCA_CC_MSG_FLAG_NONE, peer_addr);
		pthread_res = pthread_mutex_unlock(cc_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to unlock resource, error=%d", errno);
			return;
		}

		if (result == DOCA_ERROR_AGAIN) {
			result = doca_comm_channel_ep_event_handle_arm_send(ep);
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to arm Comm Channel send event channel, error=%d", errno);
				return;
			}

			nfds = epoll_wait(cc_send_epoll_fd, events, CC_MAX_EVENTS, CC_SEND_TIMEOUT);
			if (nfds == -1) {
				DOCA_LOG_ERR("Failed to wait on epoll instance, error=%d", errno);
				return;
			}

			/* Check if interrupt was received */
			for (fd_idx = 0; fd_idx < nfds; fd_idx++) {
				if (events[fd_idx].data.fd == cc_send_intr_fd)
					return;
			}

			continue;

		} else if (result != DOCA_SUCCESS)
			DOCA_LOG_WARN("Failed to send message on Comm Channel: %s", doca_get_error_string(result));

		if (msg->type == MSG_TYPE_TERMINATION || msg->type == MSG_TYPE_SERVER_DISCONNECT)
			DOCA_LOG_DBG("Message type %d was sent successfully", msg->type);
		else
			DOCA_LOG_DBG("Message type %d from fd %d was sent successfully", msg->type,
				     htonl(msg->client_fd));
		break;
	}
}

/*
 * Send all messages that were left in the CC queue
 *
 * @cc_sender_ctx [in]: CC sender thread ctx
 */
static void
send_all_queue_messages(void *cc_sender_ctx)
{
	struct cc_sender_ctx *ctx = (struct cc_sender_ctx *)cc_sender_ctx;
	struct sr_message_ctx *msg_ctx = NULL;

	while (!g_queue_is_empty(ctx->msg_queue)) {
		msg_ctx = (struct sr_message_ctx *)g_queue_pop_head(ctx->msg_queue);
		send_msg_to_cc(ctx->cc_send_epoll_fd, ctx->cc_send_intr_fd, ctx->cc_ep, *(ctx->cc_peer_addr), ctx->cc_mutex, msg_ctx->msg,
			       msg_ctx->bytes_to_send);
		free_sr_msg_ctx(msg_ctx);
	}
}

/*
 * Push termination message to CC send queue
 *
 * @queue [in]: CC messages queue
 */
static void
push_queue_termination_message(void *queue)
{
	struct sr_message_ctx *msg_ctx = NULL;
	struct sr_message *msg = NULL;
	GQueue *msg_queue = (GQueue *)queue;

	if (allocate_cc_msg(&msg_ctx, &msg) != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to allocate new message context for termination message");
		return;
	}

	msg->type = MSG_TYPE_TERMINATION;
	msg_ctx->bytes_to_send = CC_MSG_TERMINATION;
	msg_ctx->msg = msg;

	g_queue_push_tail(msg_queue, msg_ctx);
}

/*
 * Host side function which configures the UDS socket
 *
 * @sock_path [in]: Socket path to bind
 * @fd [out]: File descriptor to save the new created socket
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
host_config_uds_server_socket(const char *sock_path, int *fd)
{
	int listen_fd, result;
	struct sockaddr_un sock_addr = {0};

	listen_fd = socket(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK, 0);
	if (listen_fd == -1) {
		DOCA_LOG_ERR("Failed to create a socket for incoming connections, error=%d", errno);
		return DOCA_ERROR_IO_FAILED;
	}

	sock_addr.sun_family = AF_UNIX;
	strlcpy(sock_addr.sun_path, sock_path, SOCKADDR_UN_PATH);
	result = bind(listen_fd, (const struct sockaddr *)&sock_addr, sizeof(struct sockaddr_un));
	if (result == -1) {
		DOCA_LOG_ERR("Failed to bind socket for incoming connections on path %s, error=%d", sock_addr.sun_path,
			     errno);
		close(listen_fd);
		return DOCA_ERROR_IO_FAILED;
	}

	result = listen(listen_fd, SOCKET_BACKLOG);
	if (result == -1) {
		DOCA_LOG_ERR("Failed to listen on socket for incoming connections, error=%d", errno);
		close(listen_fd);
		return DOCA_ERROR_IO_FAILED;
	}

	*fd = listen_fd;

	return DOCA_SUCCESS;
}

/*
 * Host clients listener thread
 *
 * @host_clients_listener_ctx [in]: Host clients listener thread context
 * @return: NULL
 */
static void *
host_clients_listener(void *host_clients_listener_ctx)
{
	struct host_clients_listener_ctx *ctx = (struct host_clients_listener_ctx *)host_clients_listener_ctx;
	struct epoll_event interrupt_ev = {
		.events = EPOLLIN,
		.data.fd = ctx->interrupt_fd,
	};
	struct sr_message_ctx *msg_ctx = NULL;
	struct sr_message *msg = NULL;
	struct epoll_event *events = NULL;
	int read_itr, nfds, fd_idx, timeout = -1;
	int pthread_res, client_fd;
	ssize_t bytes_ret;
	bool is_last_msg;

	/* Register to be notified once user interrupt was received */
	if (epoll_ctl(ctx->clients_listener_epoll_fd, EPOLL_CTL_ADD, ctx->interrupt_fd, &interrupt_ev) == -1) {
		DOCA_LOG_ERR("Failed to add interrupt file descriptor to Comm Channel epoll instance, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_clients_listener_resources;
	}

	events = (struct epoll_event *)malloc(sizeof(struct epoll_event) * MAX_CONNECTIONS_NUM);
	if (events == NULL) {
		DOCA_LOG_ERR("Failed allocate epoll events array");
		ctx->thread_result = DOCA_ERROR_NO_MEMORY;
		goto destroy_clients_listener_resources;
	}

	while (1) {
		nfds = epoll_wait(ctx->clients_listener_epoll_fd, events, MAX_CONNECTIONS_NUM, timeout);
		if (nfds == -1) {
			DOCA_LOG_ERR("Failed to wait on epoll instance, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_IO_FAILED;
			goto send_termination_msg;
		}

		for (fd_idx = 0; fd_idx < nfds; fd_idx++) {
			if (events[fd_idx].data.fd == ctx->interrupt_fd) {
				DOCA_LOG_DBG("UDS clients listener thread received exit signal");
				goto destroy_clients_listener_resources;
			}

			client_fd = events[fd_idx].data.fd;

			if (events[fd_idx].events == EPOLLHUP || events[fd_idx].events == EPOLLERR) {
				DOCA_LOG_DBG("Received hang up notification on fd=%d", client_fd);
				epoll_ctl(ctx->clients_listener_epoll_fd, EPOLL_CTL_DEL, client_fd, NULL);
				close(client_fd);
				if (change_conn_count(ctx->conn_count, false, ctx->conn_count_mutex) != DOCA_SUCCESS) {
					ctx->thread_result = DOCA_ERROR_BAD_STATE;
					goto send_termination_msg;
				}
				continue;
			}

			/* Lock messages queue and insert new connection message into it */
			pthread_res = pthread_mutex_lock(ctx->queue_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to lock resource");
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}

			read_itr = 0;
			is_last_msg = false;
			while (!is_last_msg) {
				if (allocate_cc_msg(&msg_ctx, &msg) != DOCA_SUCCESS) {
					DOCA_LOG_ERR("Failed to allocate new message context for new connection");
					pthread_mutex_unlock(ctx->queue_mutex);
					ctx->thread_result = DOCA_ERROR_NO_MEMORY;
					goto send_termination_msg;
				}

				msg->client_fd = htonl((uint32_t)client_fd);
				msg_ctx->bytes_to_send = CC_MSG_NO_DATA;

				bytes_ret = read(client_fd, msg->data, CC_MAX_MSG_SIZE);
				read_itr++;

				if (bytes_ret == -1) {
					if (errno != EAGAIN || errno != EWOULDBLOCK)
						DOCA_LOG_DBG(
							"Failed to read from fd %d and to send the data to CC, error=%d",
							client_fd, errno);
					free(msg_ctx);
					free(msg);
					break;

				} else if (bytes_ret == 0) {

					/* If itr is larger than one, last message size was equal to maximum size */
					if (read_itr > 1)
						break;

					/* Client gracefully disconnected, send disconnect message */
					epoll_ctl(ctx->clients_listener_epoll_fd, EPOLL_CTL_DEL, client_fd, NULL);
					close(client_fd);

					if (change_conn_count(ctx->conn_count, false, ctx->conn_count_mutex) !=
					    DOCA_SUCCESS) {
						pthread_mutex_unlock(ctx->queue_mutex);
						free(msg_ctx);
						free(msg);
						ctx->thread_result = DOCA_ERROR_BAD_STATE;
						goto send_termination_msg;
					}

					DOCA_LOG_DBG("Sending a disconnect message for client with fd=%d", client_fd);

					msg->type = MSG_TYPE_CLIENT_DISCONNECT;

					pthread_res = pthread_mutex_lock(ctx->table_mutex);
					if (pthread_res != 0) {
						DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
						pthread_mutex_unlock(ctx->queue_mutex);
						free(msg_ctx);
						free(msg);
						ctx->thread_result = DOCA_ERROR_NO_LOCK;
						goto send_termination_msg;
					}

					if (!g_hash_table_remove(ctx->host_fd_set, &client_fd)) {
						DOCA_LOG_ERR("Local UDS client removal from table failed");
						pthread_mutex_unlock(ctx->table_mutex);
						pthread_mutex_unlock(ctx->queue_mutex);
						free(msg_ctx);
						free(msg);
						ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
						goto send_termination_msg;
					}

					pthread_res = pthread_mutex_unlock(ctx->table_mutex);
					if (pthread_res != 0) {
						DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
						pthread_mutex_unlock(ctx->queue_mutex);
						free(msg_ctx);
						free(msg);
						ctx->thread_result = DOCA_ERROR_NO_LOCK;
						goto send_termination_msg;
					}

				} else if (bytes_ret <= CC_MAX_MSG_SIZE) {
					msg->type = MSG_TYPE_DATA;
					msg_ctx->bytes_to_send += bytes_ret;
					if (bytes_ret < CC_MAX_MSG_SIZE)
						is_last_msg = true;
				}

				msg_ctx->msg = msg;

				DOCA_LOG_DBG("Pushing message of type %d from fd %d to sender thread", msg->type,
					     client_fd);

				g_queue_push_tail(ctx->msg_queue, msg_ctx);

				pthread_res = pthread_cond_signal(ctx->queue_cond);
				if (pthread_res != 0) {
					DOCA_LOG_ERR(
						"Failed to notify CC send thread that a new connection has been made, error=%d",
						errno);
					pthread_mutex_unlock(ctx->queue_mutex);
					free(msg);
					free(msg_ctx);
					ctx->thread_result = DOCA_ERROR_BAD_STATE;
					goto send_termination_msg;
				}
			}

			pthread_res = pthread_mutex_unlock(ctx->queue_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to unlock resource");
				free(msg);
				free(msg_ctx);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}
		}
	}

send_termination_msg:
	termination_due_to_err = true;

destroy_clients_listener_resources:
	if (events != NULL)
		free(events);
	stop_cc_sender = true;
	pthread_cond_signal(ctx->queue_cond);
	kill_all_threads(SIGINT);
	return NULL;
}

/*
 * Host clients acceptor thread
 *
 * @host_clients_acceptor_ctx [in]: Host clients acceptor thread context
 * @return: NULL
 */
static void *
host_clients_acceptor(void *host_clients_acceptor_ctx)
{
	struct host_clients_acceptor_ctx *ctx = (struct host_clients_acceptor_ctx *)host_clients_acceptor_ctx;
	struct sr_message_ctx *msg_ctx = NULL;
	struct sr_message *msg = NULL;
	struct epoll_event events[ACCEPT_CONN_MAX_EVENTS];
	struct epoll_event listen_event = {
		.events = EPOLLIN,
	};
	struct epoll_event interrupt_ev = {
		.events = EPOLLIN,
		.data.fd = ctx->interrupt_fd,
	};
	int listen_fd = 0, new_conn_fd, result, fd_idx, timeout = -1;
	int conn_count, pthread_res, nfds, max_events_len = MAX_CONNECTIONS_NUM;
	int *alloc_client_fd;

	result = host_config_uds_server_socket(ctx->sock_path, &listen_fd);
	if (result != DOCA_SUCCESS) {
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_accept_conn_resources;
	}

	/* Register to be notified once user interrupt was received */
	if (epoll_ctl(ctx->accept_clients_epoll_fd, EPOLL_CTL_ADD, ctx->interrupt_fd, &interrupt_ev) == -1) {
		DOCA_LOG_ERR("Failed to add interrupt file descriptor to Comm Channel epoll instance, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_accept_conn_resources;
	}

	/* Register to be notified once a new connection was made */
	listen_event.data.fd = listen_fd;
	result = epoll_ctl(ctx->accept_clients_epoll_fd, EPOLL_CTL_ADD, listen_fd, &listen_event);
	if (result == -1) {
		DOCA_LOG_WARN("Failed to add main listen file descriptor to event list, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_accept_conn_resources;
	}

	while (1) {
		struct epoll_event event;

		event.events = EPOLLIN;
		nfds = epoll_wait(ctx->accept_clients_epoll_fd, events, ACCEPT_CONN_MAX_EVENTS, timeout);
		if (nfds == -1) {
			DOCA_LOG_ERR("Failed to wait on epoll instance, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_IO_FAILED;
			goto send_termination_msg;
		}

		for (fd_idx = 0; fd_idx < nfds; fd_idx++) {
			if (events[fd_idx].data.fd == ctx->interrupt_fd) {
				DOCA_LOG_DBG("UDS connections acceptor thread received exit signal");
				goto destroy_accept_conn_resources;
			}
		}

		/* Max amount of connections, can't receive new ones */
		pthread_res = pthread_mutex_lock(ctx->conn_count_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to lock resource");
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto send_termination_msg;
		}

		conn_count = *ctx->conn_count;

		pthread_res = pthread_mutex_unlock(ctx->conn_count_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to unlock resource");
			ctx->thread_result = DOCA_ERROR_BAD_STATE;
			goto send_termination_msg;
		}

		if (conn_count == max_events_len) {
			DOCA_LOG_WARN("Maximum amount of connections already received");
			continue;
		}

		new_conn_fd = accept(listen_fd, NULL, NULL);
		if (new_conn_fd == -1) {
			DOCA_LOG_WARN("Failed to accept new connection, error=%d", errno);
			continue;
		}

		DOCA_LOG_DBG("New connection has been made, fd=%d", new_conn_fd);

		/* Allocate new message context and a new message context */
		msg_ctx = (struct sr_message_ctx *)malloc(sizeof(struct sr_message_ctx));
		if (msg_ctx == NULL) {
			DOCA_LOG_ERR("Failed to allocate new message context for new connection");
			close(new_conn_fd);
			ctx->thread_result = DOCA_ERROR_NO_MEMORY;
			goto send_termination_msg;
		}

		msg = (struct sr_message *)malloc(sizeof(struct sr_message));
		if (msg == NULL) {
			DOCA_LOG_ERR("Failed to allocate new message for new connection");
			free(msg_ctx);
			close(new_conn_fd);
			ctx->thread_result = DOCA_ERROR_NO_MEMORY;
			goto send_termination_msg;
		}

		msg->type = MSG_TYPE_CLIENT_CONNECTED;
		msg->client_fd = htonl(new_conn_fd);

		msg_ctx->bytes_to_send = CC_MSG_NO_DATA;
		msg_ctx->msg = msg;

		/* Lock messages queue and insert new connection message into it */
		pthread_res = pthread_mutex_lock(ctx->queue_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to lock resource");
			free(msg);
			free(msg_ctx);
			close(new_conn_fd);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto send_termination_msg;
		}

		DOCA_LOG_DBG("Pushing message of type %d from fd %d to CC send thread", msg->type, new_conn_fd);

		g_queue_push_tail(ctx->msg_queue, msg_ctx);

		pthread_res = pthread_cond_signal(ctx->queue_cond);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to notify CC send thread that a new connection has been made, error=%d",
				     errno);
			g_queue_pop_tail(ctx->msg_queue);
			pthread_mutex_unlock(ctx->queue_mutex);
			free(msg);
			free(msg_ctx);
			close(new_conn_fd);
			ctx->thread_result = DOCA_ERROR_BAD_STATE;
			goto send_termination_msg;
		}

		pthread_res = pthread_mutex_unlock(ctx->queue_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to unlock resource");
			g_queue_pop_tail(ctx->msg_queue);
			pthread_mutex_unlock(ctx->queue_mutex);
			free(msg);
			free(msg_ctx);
			close(new_conn_fd);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto send_termination_msg;
		}

		/* Add new fd to fd's set */
		alloc_client_fd = (int *)g_malloc(sizeof(int));
		if (alloc_client_fd == NULL) {
			DOCA_LOG_ERR("Failed to allocate new fd for Host client");
			close(new_conn_fd);
			ctx->thread_result = DOCA_ERROR_NO_MEMORY;
			goto send_termination_msg;
		}
		*alloc_client_fd = new_conn_fd;

		pthread_res = pthread_mutex_lock(ctx->table_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			close(new_conn_fd);
			g_free(alloc_client_fd);
			goto send_termination_msg;
		}

		if (!g_hash_table_add(ctx->host_fd_set, (void *)alloc_client_fd)) {
			DOCA_LOG_ERR("Failed to insert new fd to file descriptors set");
			ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
			pthread_mutex_unlock(ctx->table_mutex);
			close(new_conn_fd);
			g_free(alloc_client_fd);
			goto send_termination_msg;
		}

		pthread_res = pthread_mutex_unlock(ctx->table_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto send_termination_msg;
		}

		if (change_conn_count(ctx->conn_count, true, ctx->conn_count_mutex) != DOCA_SUCCESS) {
			ctx->thread_result = DOCA_ERROR_BAD_STATE;
			goto send_termination_msg;
		}

		event.data.fd = new_conn_fd;
		make_non_blocking(new_conn_fd);

		/* Add new fd to listener thread epoll */
		result = epoll_ctl(ctx->clients_listener_epoll_fd, EPOLL_CTL_ADD, new_conn_fd, &event);
		if (result == -1) {
			DOCA_LOG_WARN("Failed to add new connection file descriptor to event list, error=%d", errno);
			goto send_termination_msg;
		}
	}

send_termination_msg:
	termination_due_to_err = true;

destroy_accept_conn_resources:
	g_hash_table_foreach_remove(ctx->host_fd_set, (GHRFunc)close_fd, NULL);
	g_hash_table_destroy(ctx->host_fd_set);

	close(listen_fd);
	remove(ctx->sock_path);
	kill_all_threads(SIGINT);

	return NULL;
}

/*
 * DPU side function which creates a new UDS client and connects it to the user server socket
 *
 * @server_path [in]: Server to connect
 * @epoll_fd [in]: UDS clients listener file descriptor
 * @client_fd [out]: New client file descriptor
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
dpu_connect_new_client_to_server(char *server_path, int epoll_fd, int *client_fd)
{
	struct sockaddr_un addr;
	struct epoll_event ev;
	doca_error_t result;
	int fd;

	fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (fd == -1) {
		DOCA_LOG_ERR("Failed to create client socket, error=%d", errno);
		return DOCA_ERROR_IO_FAILED;
	}

	memset(&addr, 0, sizeof(struct sockaddr_un));
	addr.sun_family = AF_UNIX;
	strlcpy(addr.sun_path, server_path, SOCKADDR_UN_PATH);

	if (connect(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_un)) == -1) {
		DOCA_LOG_ERR("Failed to connect client socket to server on path %s, error=%d", server_path, errno);
		close(fd);
		return DOCA_ERROR_IO_FAILED;
	}

	result = make_non_blocking(fd);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to make socket unblock");
		close(fd);
		return DOCA_ERROR_IO_FAILED;
	}

	ev.events = EPOLLIN;
	ev.data.fd = fd;

	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, fd, &ev) == -1) {
		DOCA_LOG_ERR("Failed to modify DPU clients epoll instance, error=%d", errno);
		close(fd);
		return DOCA_ERROR_IO_FAILED;
	}

	*client_fd = fd;

	return DOCA_SUCCESS;
}

/*
 * DPU clients listener thread
 *
 * @dpu_clients_listener_ctx [in]: DPU clients listener thread context
 * @return: NULL
 */
static void *
dpu_clients_listener(void *dpu_clients_listener_ctx)
{
	struct dpu_clients_listener_ctx *ctx = (struct dpu_clients_listener_ctx *)dpu_clients_listener_ctx;
	struct epoll_event *events = NULL;
	struct epoll_event interrupt_ev = {
		.events = EPOLLIN,
		.data.fd = ctx->interrupt_fd,
	};
	struct sr_message_ctx *msg_ctx = NULL;
	struct sr_message *msg = NULL;
	int nfds, dpu_client_fd, fd_idx, timeout = -1;
	int pthread_res, read_itr;
	int *alloc_host_client_fd = NULL;
	ssize_t bytes_ret;
	bool is_last_msg;

	/* Register to be notified once user interrupt was received */
	if (epoll_ctl(ctx->clients_listener_epoll_fd, EPOLL_CTL_ADD, ctx->interrupt_fd, &interrupt_ev) == -1) {
		DOCA_LOG_ERR("Failed to add interrupt file descriptor to clients epoll instance, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_clients_listener_resources;
	}

	events = (struct epoll_event *)malloc(sizeof(struct epoll_event) * MAX_CONNECTIONS_NUM);
	if (events == NULL) {
		DOCA_LOG_ERR("Failed allocate epoll events array");
		ctx->thread_result = DOCA_ERROR_NO_MEMORY;
		goto destroy_clients_listener_resources;
	}

	while (1) {
		nfds = epoll_wait(ctx->clients_listener_epoll_fd, events, MAX_CONNECTIONS_NUM, timeout);
		if (nfds == -1) {
			DOCA_LOG_ERR("Failed to wait on epoll instance, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_IO_FAILED;
			goto send_termination_msg;
		}

		for (fd_idx = 0; fd_idx < nfds; fd_idx++) {
			if (events[fd_idx].data.fd == ctx->interrupt_fd) {
				DOCA_LOG_DBG("UDS clients listener thread exiting");
				goto destroy_clients_listener_resources;
			}
			dpu_client_fd = events[fd_idx].data.fd;

			is_last_msg = false;

			pthread_res = pthread_mutex_lock(ctx->table_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}
			alloc_host_client_fd = (int *)g_hash_table_lookup(ctx->dpu_to_host_fd_table, &dpu_client_fd);
			if (alloc_host_client_fd == NULL) {
				DOCA_LOG_ERR("Failed to find local UDS client file descriptor %d", dpu_client_fd);
				pthread_mutex_unlock(ctx->table_mutex);
				ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
				goto send_termination_msg;
			}
			pthread_res = pthread_mutex_unlock(ctx->table_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to unlock resource, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}

			/* Lock messages queue and insert new connection message into it */
			pthread_res = pthread_mutex_lock(ctx->queue_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to lock resource");
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}

			read_itr = 0;
			is_last_msg = false;

			while (!is_last_msg) {
				if (allocate_cc_msg(&msg_ctx, &msg) != DOCA_SUCCESS) {
					DOCA_LOG_ERR("Failed to allocate new message context for new connection message");
					ctx->thread_result = DOCA_ERROR_NO_MEMORY;
					goto send_termination_msg;
				}

				msg->client_fd = htonl(*alloc_host_client_fd);
				msg_ctx->bytes_to_send = CC_MSG_NO_DATA;

				bytes_ret = read(dpu_client_fd, msg->data, CC_MAX_MSG_SIZE);
				read_itr++;

				if (bytes_ret == -1) {
					if (errno == EAGAIN || errno == EWOULDBLOCK)
						break;
					DOCA_LOG_ERR(
						"Failed to read from file descriptor %d and to send the data to CC, error=%d",
						dpu_client_fd, errno);
					free(msg_ctx);
					free(msg);
					break;

				} else if (bytes_ret == 0) {
					/* If itr is larger than one, last message size was equal to maximum size */
					if (read_itr > 1)
						break;

					DOCA_LOG_ERR("UDS server disconnected, notifying Host side and exiting...");

					server_disconnect = true;

					msg->type = MSG_TYPE_SERVER_DISCONNECT;
					msg_ctx->msg = msg;

					g_queue_push_tail(ctx->msg_queue, msg_ctx);

					pthread_res = pthread_cond_signal(ctx->queue_cond);
					if (pthread_res != 0) {
						DOCA_LOG_ERR(
							"Failed to notify CC send thread that a new connection has been made, error=%d",
							errno);
						pthread_mutex_unlock(ctx->queue_mutex);
						free(msg);
						free(msg_ctx);
						ctx->thread_result = DOCA_ERROR_BAD_STATE;
						goto send_termination_msg;
					}

					/* unlock queue mutex */
					pthread_res = pthread_mutex_unlock(ctx->queue_mutex);
					if (pthread_res != 0) {
						DOCA_LOG_ERR("Failed to unlock resource");
						free(msg);
						free(msg_ctx);
						ctx->thread_result = DOCA_ERROR_NO_LOCK;
						goto send_termination_msg;
					}

					ctx->thread_result = DOCA_ERROR_NOT_CONNECTED;

					/* signal cc listen thread to finish */
					goto destroy_clients_listener_resources;

				} else if (bytes_ret <= CC_MAX_MSG_SIZE) {
					msg->type = MSG_TYPE_DATA;
					msg_ctx->bytes_to_send += bytes_ret;
					if (bytes_ret < CC_MAX_MSG_SIZE)
						is_last_msg = true;
				}

				msg_ctx->msg = msg;

				DOCA_LOG_DBG("Pushing message of type %d from fd %d to sender thread", msg->type,
					     *alloc_host_client_fd);

				g_queue_push_tail(ctx->msg_queue, msg_ctx);

				pthread_res = pthread_cond_signal(ctx->queue_cond);
				if (pthread_res != 0) {
					DOCA_LOG_ERR(
						"Failed to notify CC send thread that a new connection has been made, error=%d",
						errno);
					pthread_mutex_unlock(ctx->queue_mutex);
					free(msg);
					free(msg_ctx);
					ctx->thread_result = DOCA_ERROR_BAD_STATE;
					goto send_termination_msg;
				}
			}

			pthread_res = pthread_mutex_unlock(ctx->queue_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to unlock resource");
				free(msg);
				free(msg_ctx);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}
		}
	}

send_termination_msg:
	termination_due_to_err = true;

destroy_clients_listener_resources:
	if (events != NULL)
		free(events);
	stop_cc_sender = true;
	pthread_cond_signal(ctx->queue_cond);
	kill_all_threads(SIGINT);
	return NULL;
}

/*
 * Comm Channel listener thread
 *
 * @cc_listener_ctx [in]: Comm Channel listener thread context
 * @return: NULL
 */
static void *
cc_listener(void *cc_listener_ctx)
{
	struct cc_listener_ctx *ctx = (struct cc_listener_ctx *)cc_listener_ctx;
	struct epoll_event events[CC_MAX_EVENTS];
	struct epoll_event interrupt_ev = {
		.events = EPOLLIN,
		.data.fd = ctx->cc_recv_intr_fd
	};
	struct epoll_event recv_ev = {
		.events = EPOLLIN,
		.data.fd = ctx->cc_recv_fd
	};
	struct timespec ts = {
		.tv_sec = 0,
		.tv_nsec = SLEEP_IN_NANOS
	};
	struct sr_message msg = {0};
	int nfds, fd_idx, uds_client_fd, host_client_fd, timeout = -1;
	int *alloc_dpu_client_fd, *alloc_host_client_fd, pthread_res;
	size_t bytes_sent, bytes_to_send = 0, msg_len;
	ssize_t bytes_ret;
	doca_error_t result;

	/* Register to be notified once user interrupt was received */
	if (epoll_ctl(ctx->cc_recv_epoll_fd, EPOLL_CTL_ADD, ctx->cc_recv_intr_fd, &interrupt_ev) == -1) {
		DOCA_LOG_ERR("Failed to add interrupt file descriptor to Comm Channel epoll instance, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_cc_listener_resources;
	}

	/* Register to be notified once Comm Channel read event was received */
	if (epoll_ctl(ctx->cc_recv_epoll_fd, EPOLL_CTL_ADD, ctx->cc_recv_fd, &recv_ev) == -1) {
		DOCA_LOG_ERR("Failed to add CC file descriptor to Comm Channel epoll instance, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_cc_listener_resources;
	}

	while (1) {
		pthread_res = pthread_mutex_lock(ctx->cc_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto send_termination_msg;
		}

		msg_len = sizeof(struct sr_message);
		result = doca_comm_channel_ep_recvfrom(ctx->cc_ep, &msg, &msg_len, DOCA_CC_MSG_FLAG_NONE,

						       &ctx->cc_peer_addr);
		pthread_res = pthread_mutex_unlock(ctx->cc_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to unlock resource, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto send_termination_msg;
		}

		if (result == DOCA_ERROR_AGAIN) {
			result = doca_comm_channel_ep_event_handle_arm_recv(ctx->cc_ep);
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to arm Comm Channel receive event channel, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_IO_FAILED;
				goto send_termination_msg;
			}

			nfds = epoll_wait(ctx->cc_recv_epoll_fd, events, CC_MAX_EVENTS, timeout);
			if (nfds == -1) {
				DOCA_LOG_ERR("Failed to wait on epoll instance, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_IO_FAILED;
				goto send_termination_msg;
			}

			/* Check if interrupt was received */
			for (fd_idx = 0; fd_idx < nfds; fd_idx++) {
				if (events[fd_idx].data.fd == ctx->cc_recv_intr_fd) {
					DOCA_LOG_DBG("Comm Channel listener thread received exit signal");
					goto destroy_cc_listener_resources;
				}
			}

			/* Comm channel event was received, run recvfrom() again */
			continue;

		/* Make sure enough bytes were sent */
		} else if (result != DOCA_SUCCESS || msg_len < CC_MSG_TERMINATION) {
			DOCA_LOG_WARN("Failed to receive channel message: %s", doca_get_error_string(result));
			continue;
		}

		/* Handle termination message */
		if (msg.type == MSG_TYPE_TERMINATION) {
			DOCA_LOG_INFO("Comm Channel listener received termination message");
			goto destroy_cc_listener_resources;

		} else if (ctx->sr_cfg->mode == SR_MODE_HOST && msg.type == MSG_TYPE_SERVER_DISCONNECT) {
			DOCA_LOG_INFO("Comm Channel listener received server disconnected message");
			goto destroy_cc_listener_resources;
		}

		if (msg_len < CC_MSG_NO_DATA) {
			DOCA_LOG_ERR("Failed to receive message, not enough bytes to read");
			continue;
		}

		/* take Host client file descriptor from relevant message bytes */
		host_client_fd = ntohl(msg.client_fd);

		uds_client_fd = host_client_fd;

		/* Handle DPU uds client disconnect */
		if (ctx->sr_cfg->mode == SR_MODE_DPU && msg.type == MSG_TYPE_CLIENT_DISCONNECT) {
			DOCA_LOG_DBG("Client disconnect message received for Host fd %d", host_client_fd);
			alloc_dpu_client_fd =
				(int *)g_hash_table_lookup(ctx->host_to_dpu_fd_table, (void *)&host_client_fd);
			if (alloc_dpu_client_fd == NULL) {
				DOCA_LOG_ERR(
					"Disconnect event received, failed to find UDS client (Host) file descriptor %d",
					host_client_fd);
				ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
				goto send_termination_msg;
			}

			close(*alloc_dpu_client_fd);

			pthread_res = pthread_mutex_lock(ctx->table_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}

			if (!g_hash_table_remove(ctx->dpu_to_host_fd_table, (void *)alloc_dpu_client_fd)) {
				DOCA_LOG_ERR("Failed to delete UDS client (DPU) file descriptor %d", host_client_fd);
				ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
				pthread_mutex_unlock(ctx->table_mutex);
				goto send_termination_msg;
			}

			pthread_res = pthread_mutex_unlock(ctx->table_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to unlock resource, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}

			if (!g_hash_table_remove(ctx->host_to_dpu_fd_table, (void *)&host_client_fd)) {
				DOCA_LOG_ERR("Failed to delete UDS client (Host) file descriptor %d", host_client_fd);
				ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
				goto send_termination_msg;
			}

			continue;

		/* Handle new client message */
		} else if (ctx->sr_cfg->mode == SR_MODE_DPU && msg.type == MSG_TYPE_CLIENT_CONNECTED) {
			alloc_dpu_client_fd = (int *)g_hash_table_lookup(ctx->host_to_dpu_fd_table, &host_client_fd);

			if (alloc_dpu_client_fd != NULL) {
				DOCA_LOG_ERR("Client connect message received, Host client was found in table");
				ctx->thread_result = DOCA_ERROR_BAD_STATE;
				goto send_termination_msg;
			}

			DOCA_LOG_DBG("New client connected, Host side file descriptor received: %d", host_client_fd);
			alloc_dpu_client_fd = (int *)g_malloc(sizeof(int));
			if (alloc_dpu_client_fd == NULL) {
				DOCA_LOG_ERR("Failed to allocate new fd for DPU client");
				ctx->thread_result = DOCA_ERROR_NO_MEMORY;
				goto send_termination_msg;
			}

			result = dpu_connect_new_client_to_server(ctx->sr_cfg->socket_path,
								  ctx->dpu_uds_clients_epoll_fd, alloc_dpu_client_fd);
			if (result != DOCA_SUCCESS) {
				ctx->thread_result = result;
				g_free(alloc_dpu_client_fd);
				goto send_termination_msg;
			}

			alloc_host_client_fd = (int *)g_malloc(sizeof(int));
			if (alloc_host_client_fd == NULL) {
				DOCA_LOG_ERR("Failed to allocate new fd for Host client");
				ctx->thread_result = DOCA_ERROR_NO_MEMORY;
				close(*alloc_dpu_client_fd);
				g_free(alloc_dpu_client_fd);
				goto send_termination_msg;
			}

			*alloc_host_client_fd = host_client_fd;

			if (!g_hash_table_insert(ctx->host_to_dpu_fd_table, (void *)alloc_host_client_fd,
						 (void *)alloc_dpu_client_fd)) {
				DOCA_LOG_ERR("Failed to insert new fd to Host to DPU file descriptors table");
				ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
				close(*alloc_dpu_client_fd);
				g_free(alloc_dpu_client_fd);
				g_free(alloc_host_client_fd);
				goto send_termination_msg;
			}

			pthread_res = pthread_mutex_lock(ctx->table_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}

			if (!g_hash_table_insert(ctx->dpu_to_host_fd_table, (void *)alloc_dpu_client_fd,
						 (void *)alloc_host_client_fd)) {
				DOCA_LOG_ERR("Failed to insert new fd to DPU to Host file descriptors table");
				ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
				pthread_mutex_unlock(ctx->table_mutex);
				close(*alloc_dpu_client_fd);
				g_hash_table_remove(ctx->host_to_dpu_fd_table, (void *)&alloc_host_client_fd);
				goto send_termination_msg;
			}

			pthread_res = pthread_mutex_unlock(ctx->table_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_NO_LOCK;
				goto send_termination_msg;
			}

			continue;
		}

		/* No raw data to read */
		if (msg_len == CC_MSG_NO_DATA) {
			DOCA_LOG_WARN("Message was not received properly, no raw data to read");
			continue;
		}

		if (ctx->sr_cfg->mode == SR_MODE_DPU) {
			alloc_dpu_client_fd = (int *)g_hash_table_lookup(ctx->host_to_dpu_fd_table, &host_client_fd);

			if (alloc_dpu_client_fd == NULL) {
				DOCA_LOG_ERR("Failed to find relevant DPU file descriptor, Host side fd received: %d",
					     host_client_fd);
				ctx->thread_result = DOCA_ERROR_BAD_STATE;
				goto send_termination_msg;
			}
			uds_client_fd = *alloc_dpu_client_fd;
		}

		bytes_to_send = msg_len - CC_MSG_NO_DATA;
		bytes_sent = 0;
		while (bytes_sent < bytes_to_send) {
			bytes_ret = write(uds_client_fd, msg.data, bytes_to_send);
			if (bytes_ret == -1) {
				if (errno == EAGAIN || errno == EWOULDBLOCK) {
					nanosleep(&ts, &ts);
					continue;
				}

				DOCA_LOG_ERR("Failed to send buffer to client side, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_IO_FAILED;
				goto send_termination_msg;
			}
			bytes_sent += bytes_ret;
		}

		bytes_to_send = 0;
	}

send_termination_msg:
	termination_due_to_err = true;

destroy_cc_listener_resources:

	if (ctx->sr_cfg->mode == SR_MODE_DPU) {
		/* Close all clients file descriptors on DPU side */
		g_hash_table_foreach_remove(ctx->dpu_to_host_fd_table, (GHRFunc)close_fd, NULL);
		g_hash_table_destroy(ctx->dpu_to_host_fd_table);

		/* Free all allocated file descriptors pointers */
		g_hash_table_destroy(ctx->host_to_dpu_fd_table);
	}
	kill_all_threads(SIGINT);

	return NULL;
}

/*
 * Comm Channel sender thread
 *
 * @cc_sender_ctx [in]: Comm Channel sender thread context
 * @return: NULL
 */
static void *
cc_sender(void *cc_sender_ctx)
{
	struct cc_sender_ctx *ctx = (struct cc_sender_ctx *)cc_sender_ctx;
	struct sr_message_ctx *msg_ctx = NULL;
	struct epoll_event interrupt_ev = {
		.events = EPOLLIN,
		.data.fd = ctx->cc_send_intr_fd
	};
	struct epoll_event send_ev = {
		.events = EPOLLIN,
		.data.fd = ctx->cc_send_fd
	};
	int pthread_res;

	/* Register to be notified once user interrupt was received */
	if (epoll_ctl(ctx->cc_send_epoll_fd, EPOLL_CTL_ADD, ctx->cc_send_intr_fd, &interrupt_ev) == -1) {
		DOCA_LOG_ERR("Failed to add interrupt file descriptor to Comm Channel epoll instance, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_cc_sender_resources;
	}

	/* Register to be notified once Comm Channel send event was received */
	if (epoll_ctl(ctx->cc_send_epoll_fd, EPOLL_CTL_ADD, ctx->cc_send_fd, &send_ev) == -1) {
		DOCA_LOG_ERR("Failed to add CC file descriptor to Comm Channel epoll instance, error=%d", errno);
		ctx->thread_result = DOCA_ERROR_IO_FAILED;
		goto destroy_cc_sender_resources;
	}

	while (!stop_cc_sender) {
		pthread_res = pthread_mutex_lock(ctx->queue_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to lock resource, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto destroy_cc_sender_resources;
		}

		while (g_queue_is_empty(ctx->msg_queue)) {
			pthread_res = pthread_cond_wait(ctx->queue_cond, ctx->queue_mutex);
			if (pthread_res != 0) {
				DOCA_LOG_ERR("Failed to wait on queue sync variable, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_BAD_STATE;
				goto destroy_cc_sender_resources;
			}

			if (stop_cc_sender) {
				/* Make sure that disconnect/termination message will be sent to other side */
				if (termination_due_to_err || server_disconnect) {
					if (termination_due_to_err) {
						DOCA_LOG_DBG("Pushing termination message to CC message queue");
						push_queue_termination_message(ctx->msg_queue);
					}
					send_all_queue_messages(ctx);
				}
				goto destroy_cc_sender_resources;
			}
		}

		msg_ctx = (struct sr_message_ctx *)g_queue_pop_head(ctx->msg_queue);
		if (msg_ctx == NULL) {
			pthread_res = pthread_mutex_unlock(ctx->queue_mutex);
			if (pthread_res != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to unlock resource, error=%d", errno);
				ctx->thread_result = DOCA_ERROR_INVALID_VALUE;
				goto destroy_cc_sender_resources;
			}
			continue;
		}

		pthread_res = pthread_mutex_unlock(ctx->queue_mutex);
		if (pthread_res != 0) {
			DOCA_LOG_ERR("Failed to unlock resource, error=%d", errno);
			ctx->thread_result = DOCA_ERROR_NO_LOCK;
			goto destroy_cc_sender_resources;
		}

		send_msg_to_cc(ctx->cc_send_epoll_fd, ctx->cc_send_intr_fd, ctx->cc_ep, *(ctx->cc_peer_addr),
			       ctx->cc_mutex, msg_ctx->msg, msg_ctx->bytes_to_send);

		free_sr_msg_ctx(msg_ctx);
	}

destroy_cc_sender_resources:

	DOCA_LOG_DBG("Comm Channel sender thread exiting");
	g_queue_free_full(ctx->msg_queue, free_sr_msg_ctx);
	kill_all_threads(SIGINT);

	return NULL;
}

/*
 * Initiate Comm Channel on both sides
 *
 * @cfg [in]: Configuration structure
 * @ep [out]: Comm Channel endpoint
 * @peer [out]: Comm Channel peer address
 * @dev [out]: DOCA device object to use
 * @dev_rep [out]: DOCA device representor object to use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
init_cc(struct sr_cfg *cfg, struct doca_comm_channel_ep_t **ep, struct doca_comm_channel_addr_t **peer,
	   struct doca_dev **dev, struct doca_dev_rep **dev_rep)
{
	doca_error_t result;
	struct timespec ts = {0};

	result = doca_comm_channel_ep_create(ep);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create Comm Channel endpoint: %s", doca_get_error_string(result));
		return result;
	}

	result = open_doca_device(cfg->cc_dev_pci_addr, dev);
	if (result != DOCA_SUCCESS) {
		doca_comm_channel_ep_destroy(*ep);
		return result;
	}

	/* Open DOCA device representor on DPU side */
	if (cfg->mode == SR_MODE_DPU) {
		result = open_doca_device_rep(cfg->cc_dev_rep_pci_addr, *dev, dev_rep);
		if (result != DOCA_SUCCESS) {
			doca_comm_channel_ep_destroy(*ep);
			doca_dev_close(*dev);
			return result;
		}
	}

	result = set_cc_properties(cfg->mode, *ep, *dev, *dev_rep);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set Comm Channel properties");
		doca_comm_channel_ep_destroy(*ep);
		if (cfg->mode == SR_MODE_DPU)
			doca_dev_rep_close(*dev_rep);
		doca_dev_close(*dev);
		return result;
	}

	if (cfg->mode == SR_MODE_DPU) {
		result = doca_comm_channel_ep_listen(*ep, cfg->cc_name);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to listen on Comm Channel endpoint: %s", doca_get_error_string(result));
			doca_comm_channel_ep_destroy(*ep);
			doca_dev_rep_close(*dev_rep);
			doca_dev_close(*dev);
			return result;
		}
	} else {
		result = doca_comm_channel_ep_connect(*ep, cfg->cc_name, peer);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Couldn't establish a connection with the server node: %s",
				     doca_get_error_string(result));
			doca_comm_channel_ep_destroy(*ep);
			doca_dev_close(*dev);
			return result;
		}

		while ((result = doca_comm_channel_peer_addr_update_info(*peer)) == DOCA_ERROR_CONNECTION_INPROGRESS)
			nanosleep(&ts, &ts);

		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to validate the connection with the DPU: %s", doca_get_error_string(result));
			return result;
		}

		DOCA_LOG_INFO("Connection to DPU was established successfully");
	}

	return result;
}

/*
 * Start tool threads
 *
 * @mode [in]: Mode of operation
 * @host_clients_acceptor_ctx [out]: Host clients acceptor thread context
 * @host_clients_listener_ctx [out]: Host clients listener thread context
 * @dpu_clients_listener_ctx [out]: DPU clients listener thread context
 * @cc_listener_ctx [out]: Comm Channel listener thread context
 * @cc_sender_ctx [out]: Comm Channel sender thread contextl
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
sr_pthread_start(enum sr_mode mode, struct host_clients_acceptor_ctx *host_clients_acceptor_ctx,
		 struct host_clients_listener_ctx *host_clients_listener_ctx,
		 struct dpu_clients_listener_ctx *dpu_clients_listener_ctx, struct cc_listener_ctx *cc_listener_ctx,
		 struct cc_sender_ctx *cc_sender_ctx)
{
	pthread_t cc_listener_thread, cc_sender_thread, clients_acceptor_thread, clients_listener_thread;
	int result;
	void *pthread_ctx = dpu_clients_listener_ctx;
	void *(*pthread_func)(void *) = &dpu_clients_listener;

	result = pthread_create(&cc_listener_thread, NULL, cc_listener, cc_listener_ctx);
	if (result != 0) {
		DOCA_LOG_ERR("Failed to create Comm Channel receive thread");
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	result = pthread_create(&cc_sender_thread, NULL, cc_sender, cc_sender_ctx);
	if (result != 0) {
		DOCA_LOG_ERR("Failed to create Comm Channel send thread");
		kill_all_threads(SIGINT);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	if (mode == SR_MODE_HOST) {
		result = pthread_create(&clients_acceptor_thread, NULL, host_clients_acceptor, host_clients_acceptor_ctx);
		if (result != 0) {
			DOCA_LOG_ERR("Failed to create Connection accepting thread");
			kill_all_threads(SIGINT);
			return DOCA_ERROR_OPERATING_SYSTEM;
		}
		pthread_ctx = host_clients_listener_ctx;
		pthread_func = &host_clients_listener;
	}

	result = pthread_create(&clients_listener_thread, NULL, pthread_func, pthread_ctx);
	if (result != 0) {
		DOCA_LOG_ERR("Failed to create Connection accepting thread");
		kill_all_threads(SIGINT);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	pthread_join(cc_listener_thread, NULL);
	pthread_join(cc_sender_thread, NULL);

	return DOCA_SUCCESS;
}

/*
 * Review threads result
 *
 * @mode [in]: Mode of operation
 * @host_clients_acceptor_ctx [in]: Host clients acceptor thread context
 * @host_clients_listener_ctx [in]: Host clients listener thread context
 * @dpu_clients_listener_ctx [in]: DPU clients listener thread context
 * @cc_listener_ctx [in]: Comm Channel listener thread context
 * @cc_sender_ctx [in]: Comm Channel sender thread contextl
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
threads_results(enum sr_mode mode, struct host_clients_acceptor_ctx *host_clients_acceptor_ctx,
		   struct host_clients_listener_ctx *host_clients_listener_ctx,
		   struct dpu_clients_listener_ctx *dpu_clients_listener_ctx, struct cc_listener_ctx *cc_listener_ctx,
		   struct cc_sender_ctx *cc_sender_ctx)
{
	doca_error_t result;

	result = cc_listener_ctx->thread_result;
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Comm Channel listener thread returned unsuccessfully: %s", doca_get_error_string(result));
		return result;
	}

	result = cc_sender_ctx->thread_result;
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Comm Channel sender thread returned unsuccessfully: %s", doca_get_error_string(result));
		return result;
	}

	if (mode == SR_MODE_DPU) {
		result = dpu_clients_listener_ctx->thread_result;
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("UDS clients listener thread returned unsuccessfully: %s",
				     doca_get_error_string(result));
			return result;
		}
	} else {
		result = host_clients_acceptor_ctx->thread_result;
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("UDS clients acceptor thread returned unsuccessfully: %s",
				     doca_get_error_string(result));
			return result;
		}
		result = host_clients_listener_ctx->thread_result;
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("UDS clients listener thread returned unsuccessfully: %s",
				     doca_get_error_string(result));
		}
	}

	return result;
}

/*
 * Destroy Comm Channel resources
 *
 *
 * @ep [in]: Comm Channel endpoint
 * @peer [in]: Comm Channel peer address
 * @dev [in]: Comm Channel DOCA device
 * @dev_rep [in]: Comm Channel DOCA device representor
 */
static void
destroy_cc(struct doca_comm_channel_ep_t *ep, struct doca_comm_channel_addr_t *peer, struct doca_dev *dev,
	      struct doca_dev_rep *dev_rep)
{
	doca_error_t result;

	if (peer != NULL) {
		result = doca_comm_channel_ep_disconnect(ep, peer);
		if (result != DOCA_SUCCESS)
			DOCA_LOG_ERR("Failed to disconnect from Comm Channel peer address: %s",
				     doca_get_error_string(result));
	}

	result = doca_comm_channel_ep_destroy(ep);
	if (result != DOCA_SUCCESS)
		DOCA_LOG_ERR("Failed to destroy Comm Channel endpoint: %s", doca_get_error_string(result));

	if (dev_rep != NULL) {
		result = doca_dev_rep_close(dev_rep);
		if (result != DOCA_SUCCESS)
			DOCA_LOG_ERR("Failed to close Comm Channel DOCA device representor: %s",
				     doca_get_error_string(result));
	}

	result = doca_dev_close(dev);
	if (result != DOCA_SUCCESS)
		DOCA_LOG_ERR("Failed to close Comm Channel DOCA device: %s", doca_get_error_string(result));
}

/*
 * Destroy pthread structures
 *
 * @cc_mutex [in]: Comm Channel mutex
 * @queue_mutex [in]: Comm Channel send queue mutex
 * @conn_count_mutex [in]: Connections counter mutex
 * @table_mutex [in]: DPU to Host file descriptor table mutex
 * @queue_cond [in]: Comm Channel send queue condition variable
 */
static void
destroy_pthread(pthread_mutex_t *cc_mutex, pthread_mutex_t *queue_mutex, pthread_mutex_t *conn_count_mutex,
		   pthread_mutex_t *table_mutex, pthread_cond_t *queue_cond)
{
	pthread_mutex_destroy(cc_mutex);
	pthread_mutex_destroy(queue_mutex);
	pthread_mutex_destroy(conn_count_mutex);
	pthread_mutex_destroy(table_mutex);
	pthread_cond_destroy(queue_cond);
}

doca_error_t
register_sr_params()
{
	doca_error_t result;
	struct doca_argp_param *socket_path_param, *cc_name_param, *dev_pci_addr_param, *rep_pci_addr_param;

	/* Create and register UDS path (DPU side will connect to it, Host side will bind to it) */
	result = doca_argp_param_create(&socket_path_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(socket_path_param, "s");
	doca_argp_param_set_long_name(socket_path_param, "socket");
	doca_argp_param_set_description(socket_path_param,
					"Unix domain socket path, host side will bind to and DPU connect to");
	doca_argp_param_set_callback(socket_path_param, socket_path_callback);
	doca_argp_param_set_type(socket_path_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(socket_path_param);
	result = doca_argp_register_param(socket_path_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Create and register Comm Channel service name */
	result = doca_argp_param_create(&cc_name_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(cc_name_param, "n");
	doca_argp_param_set_long_name(cc_name_param, "cc-name");
	doca_argp_param_set_description(cc_name_param,
					"Comm Channel service name");
	doca_argp_param_set_callback(cc_name_param, cc_name_callback);
	doca_argp_param_set_type(cc_name_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(cc_name_param);
	result = doca_argp_register_param(cc_name_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Create and register Comm Channel DOCA device PCI address */
	result = doca_argp_param_create(&dev_pci_addr_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(dev_pci_addr_param, "d");
	doca_argp_param_set_long_name(dev_pci_addr_param, "dev-pci");
	doca_argp_param_set_description(dev_pci_addr_param,
					"Comm Channel DOCA device PCI address");
	doca_argp_param_set_callback(dev_pci_addr_param, dev_pci_addr_callback);
	doca_argp_param_set_type(dev_pci_addr_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(dev_pci_addr_param);
	result = doca_argp_register_param(dev_pci_addr_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Create and register Comm Channel DOCA device representor PCI address */
	result = doca_argp_param_create(&rep_pci_addr_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(rep_pci_addr_param, "r");
	doca_argp_param_set_long_name(rep_pci_addr_param, "rep-pci");
	doca_argp_param_set_description(rep_pci_addr_param,
					"Comm Channel DOCA device representor PCI address (needed only on DPU)");
	doca_argp_param_set_callback(rep_pci_addr_param, rep_pci_addr_callback);
	doca_argp_param_set_type(rep_pci_addr_param, DOCA_ARGP_TYPE_STRING);
	result = doca_argp_register_param(rep_pci_addr_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	return DOCA_SUCCESS;
}

doca_error_t
sr_start(struct sr_cfg *sr_cfg, struct cc_listener_ctx *cc_listener_ctx, struct cc_sender_ctx *cc_sender_ctx,
	 struct host_clients_acceptor_ctx *host_accept_clients_ctx,
	 struct host_clients_listener_ctx *host_clients_listener_ctx,
	 struct dpu_clients_listener_ctx *dpu_clients_listener_ctx)
{
	struct doca_comm_channel_ep_t *cc_ep = NULL;
	struct doca_comm_channel_addr_t *cc_peer = NULL;
	struct doca_dev *cc_dev = NULL;
	struct doca_dev_rep *cc_dev_rep = NULL;
	enum sr_mode mode = sr_cfg->mode;
	doca_error_t result;
	int conn_count = 0;

	/* File descriptors */
	int cc_send_intr_fd, cc_recv_intr_fd, clients_intr_fd, accept_clients_intr_fd;
	int cc_send_epoll_fd, cc_recv_epoll_fd, clients_listener_epoll_fd, accept_clients_epoll_fd;

	/* Pthread variables */
	pthread_mutex_t cc_mutex, conn_count_mutex, queue_mutex, table_mutex;
	pthread_cond_t queue_cond;

	/* Glib data structures */
	GHashTable *host_to_dpu_fd_table = NULL, *dpu_to_host_fd_table = NULL;
	GHashTable *host_fd_set = NULL;
	GQueue *msg_queue = NULL;

	result = init_cc(sr_cfg, &cc_ep, &cc_peer, &cc_dev, &cc_dev_rep);
	if (result != DOCA_SUCCESS)
		return result;

	result = create_signals_fd(mode, &cc_send_intr_fd, &cc_recv_intr_fd, &clients_intr_fd, &accept_clients_intr_fd);
	if (result != DOCA_SUCCESS) {
		destroy_cc(cc_ep, cc_peer, cc_dev, cc_dev_rep);
		return result;
	}

	result = create_epoll_fds(mode, &cc_send_epoll_fd, &cc_recv_epoll_fd, &clients_listener_epoll_fd,
				  &accept_clients_epoll_fd);
	if (result != DOCA_SUCCESS) {
		close(cc_send_intr_fd);
		close(cc_recv_intr_fd);
		close(clients_intr_fd);
		close(accept_clients_intr_fd);
		if (sr_cfg->mode == SR_MODE_HOST)
			close(accept_clients_intr_fd);
		destroy_cc(cc_ep, cc_peer, cc_dev, cc_dev_rep);
		return result;
	}

	result = init_pthread(mode, &cc_mutex, &queue_mutex, &table_mutex, &conn_count_mutex, &queue_cond);
	if (result != DOCA_SUCCESS) {
		close_fds(mode, cc_send_epoll_fd, cc_recv_epoll_fd, clients_listener_epoll_fd, accept_clients_epoll_fd,
			  cc_send_intr_fd, cc_recv_intr_fd, clients_intr_fd, accept_clients_intr_fd);
		destroy_cc(cc_ep, cc_peer, cc_dev, cc_dev_rep);
		return result;
	}

	sr_init_glib_structures(sr_cfg->mode, &host_to_dpu_fd_table, &dpu_to_host_fd_table, &host_fd_set, &msg_queue);

	/* Init Comm Channel send thread context */
	cc_sender_ctx->cc_ep = cc_ep;
	cc_sender_ctx->cc_peer_addr = &cc_listener_ctx->cc_peer_addr;
	cc_sender_ctx->cc_send_epoll_fd = cc_send_epoll_fd;
	cc_sender_ctx->cc_send_intr_fd = cc_send_intr_fd;
	cc_sender_ctx->cc_mutex = &cc_mutex;
	cc_sender_ctx->msg_queue = msg_queue;
	cc_sender_ctx->queue_mutex = &queue_mutex;
	cc_sender_ctx->queue_cond = &queue_cond;

	/* Init Comm Channel receive thread context */
	cc_listener_ctx->sr_cfg = sr_cfg;
	cc_listener_ctx->cc_ep = cc_ep;
	cc_listener_ctx->cc_peer_addr = cc_peer;
	cc_listener_ctx->cc_recv_epoll_fd = cc_recv_epoll_fd;
	cc_listener_ctx->cc_recv_intr_fd = cc_recv_intr_fd;
	cc_listener_ctx->cc_mutex = &cc_mutex;
	cc_listener_ctx->host_to_dpu_fd_table = host_to_dpu_fd_table;
	cc_listener_ctx->dpu_to_host_fd_table = dpu_to_host_fd_table;
	cc_listener_ctx->dpu_uds_clients_epoll_fd = clients_listener_epoll_fd;
	cc_listener_ctx->table_mutex = &table_mutex;

	doca_comm_channel_ep_get_event_channel(cc_ep, &cc_sender_ctx->cc_send_fd, &cc_listener_ctx->cc_recv_fd);

	if (mode == SR_MODE_HOST) {
		cc_sender_ctx->cc_peer_addr = &cc_peer;

		/* Init Host UDS clients listener thread context */
		host_clients_listener_ctx->mode = mode;
		host_clients_listener_ctx->host_fd_set = host_fd_set;
		host_clients_listener_ctx->msg_queue = msg_queue;
		host_clients_listener_ctx->clients_listener_epoll_fd = clients_listener_epoll_fd;
		host_clients_listener_ctx->conn_count = &conn_count;
		host_clients_listener_ctx->interrupt_fd = clients_intr_fd;
		host_clients_listener_ctx->conn_count_mutex = &conn_count_mutex;
		host_clients_listener_ctx->queue_mutex = &queue_mutex;
		host_clients_listener_ctx->table_mutex = &table_mutex;
		host_clients_listener_ctx->queue_cond = &queue_cond;

		/* Init Host UDS clients acceptor thread context */
		host_accept_clients_ctx->sock_path = sr_cfg->socket_path;
		host_accept_clients_ctx->clients_listener_epoll_fd = clients_listener_epoll_fd;
		host_accept_clients_ctx->accept_clients_epoll_fd = accept_clients_epoll_fd;
		host_accept_clients_ctx->host_fd_set = host_fd_set;
		host_accept_clients_ctx->msg_queue = msg_queue;
		host_accept_clients_ctx->conn_count = &conn_count;
		host_accept_clients_ctx->interrupt_fd = accept_clients_intr_fd;
		host_accept_clients_ctx->conn_count_mutex = &conn_count_mutex;
		host_accept_clients_ctx->queue_mutex = &queue_mutex;
		host_accept_clients_ctx->table_mutex = &table_mutex;
		host_accept_clients_ctx->queue_cond = &queue_cond;

	} else {

		/* Init DPU UDS clients listener thread context */
		dpu_clients_listener_ctx->msg_queue = msg_queue;
		dpu_clients_listener_ctx->dpu_to_host_fd_table = dpu_to_host_fd_table;
		dpu_clients_listener_ctx->clients_listener_epoll_fd = clients_listener_epoll_fd;
		dpu_clients_listener_ctx->interrupt_fd = clients_intr_fd;
		dpu_clients_listener_ctx->table_mutex = &table_mutex;
		dpu_clients_listener_ctx->queue_mutex = &queue_mutex;
		dpu_clients_listener_ctx->queue_cond = &queue_cond;
	}

	result = sr_pthread_start(mode, host_accept_clients_ctx, host_clients_listener_ctx, dpu_clients_listener_ctx,
				  cc_listener_ctx, cc_sender_ctx);

	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed initialize Unix domain socket routine: %s", doca_get_error_string(result));
		close_fds(mode, cc_send_epoll_fd, cc_recv_epoll_fd, clients_listener_epoll_fd, accept_clients_epoll_fd,
			  cc_send_intr_fd, cc_recv_intr_fd, clients_intr_fd, accept_clients_intr_fd);
		destroy_cc(cc_ep, cc_peer, cc_dev, cc_dev_rep);
		return result;
	}

	result = threads_results(mode, host_accept_clients_ctx, host_clients_listener_ctx, dpu_clients_listener_ctx,
				 cc_listener_ctx, cc_sender_ctx);

	destroy_pthread(&cc_mutex, &queue_mutex, &conn_count_mutex, &table_mutex, &queue_cond);
	close_fds(mode, cc_send_epoll_fd, cc_recv_epoll_fd, clients_listener_epoll_fd, accept_clients_epoll_fd,
		  cc_send_intr_fd, cc_recv_intr_fd, clients_intr_fd, accept_clients_intr_fd);
	destroy_cc(cc_ep, cc_peer, cc_dev, cc_dev_rep);

	DOCA_LOG_INFO("Exiting Socket Relay");

	return result;
}
