/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef SOCKET_RELAY_CORE_H_
#define SOCKET_RELAY_CORE_H_

#include <stdint.h>
#include <stdbool.h>
#include <pthread.h>
#include <glib.h>

#include <doca_comm_channel.h>
#include <doca_dev.h>

#define CC_MAX_MSG_SIZE 4075					/* Comm Channel maximum message size minus 5 meta bytes */
#define CC_MAX_NAME 120						/* Comm Channel service name */
#define SOCKADDR_UN_PATH 108					/* Library array size for socket path */
#define PCI_ADDR_LEN 8						/* PCI address string length */

enum sr_mode {
	SR_MODE_HOST,						/* Socket relay running on the Host */
	SR_MODE_DPU,						/* Socket relay running on the DPU */
};

enum __attribute__((__packed__)) msg_type {
	MSG_TYPE_SERVER_DISCONNECT = 0x01,			/* Server disconnected message */
	MSG_TYPE_CLIENT_CONNECTED = 0x02,			/* Client connected message */
	MSG_TYPE_CLIENT_DISCONNECT = 0x03,			/* Client disconnected message */
	MSG_TYPE_DATA = 0x04,					/* Message contains data to read */
	MSG_TYPE_TERMINATION = 0x05,				/* Termination message */
};

struct sr_cfg {
	enum sr_mode mode;					/* Mode of operation */
	char socket_path[SOCKADDR_UN_PATH];			/* UDS Server socket path on both sides */
	char cc_name[CC_MAX_NAME];				/* Comm Channel service name */
	char cc_dev_pci_addr[PCI_ADDR_LEN];			/* Comm Channel DOCA device PCI address */
	char cc_dev_rep_pci_addr[PCI_ADDR_LEN];			/* Comm Channel DOCA device representor PCI address */
};

struct __attribute__((__packed__)) sr_message {
	enum msg_type type;					/* Message type */
	uint32_t client_fd;					/* Client file descriptor on the Host */
	uint8_t data[CC_MAX_MSG_SIZE];				/* Raw data to send */
};

struct sr_message_ctx {
	struct sr_message *msg;					/* Message to send on Comm Channel */
	size_t bytes_to_send;					/* Number of bytes to send on Comm Channel */
};

/* This structure is used on both sides threads (DPU and Host) */
struct cc_sender_ctx {
	struct doca_comm_channel_ep_t *cc_ep;			/* Comm Channel endpoint */
	struct doca_comm_channel_addr_t **cc_peer_addr;		/* Comm Channel peer address */
	int cc_send_epoll_fd;					/* Comm Channel send epoll instance file descriptor */
	int cc_send_fd;						/* Comm Channel send file descriptor */
	int cc_send_intr_fd;					/* Comm Channel send Interrupt file descriptor */
	pthread_mutex_t *cc_mutex;				/* Comm Channel sendto/recvfrom cc_mutex */
	GQueue *msg_queue;					/* Comm Channel Messages queue */
	pthread_mutex_t *queue_mutex;				/* Queue mutex */
	pthread_cond_t *queue_cond;				/* Messages queue condition variable */
	doca_error_t thread_result;				/* Thread final result */
};

/* This structure is used on both sides threads (DPU and Host) */
struct cc_listener_ctx {
	struct sr_cfg *sr_cfg;					/* Socket Relay configuration structure */
	struct doca_comm_channel_ep_t *cc_ep;			/* Comm Channel endpoint */
	struct doca_comm_channel_addr_t *cc_peer_addr;		/* Comm Channel peer address */
	int cc_recv_epoll_fd;					/* Comm Channel receive epoll instance file descriptor */
	int cc_recv_fd;						/* Comm Channel receive file descriptor */
	GHashTable *host_to_dpu_fd_table;			/* DPU side table of client (Host) fd to client fd (DPU) */
	GHashTable *dpu_to_host_fd_table;			/* DPU side Table of client (DPU) fd to client fd (Host) */
	int dpu_uds_clients_epoll_fd;				/* Epoll fd which holds clients fd's on DPU only */
	int cc_recv_intr_fd;					/* Comm Channel receive Interrupt file descriptor */
	pthread_mutex_t *cc_mutex;				/* Comm Channel sendto/recvfrom mutex */
	pthread_mutex_t *table_mutex;				/* Mutex for DPU fd to Host fd table */
	doca_error_t thread_result;				/* Thread final result */
};

/* This ctx will be used by Host thread */
struct host_clients_acceptor_ctx {
	const char *sock_path;					/* UDS server socket path */
	int clients_listener_epoll_fd;				/* Epoll fd for all clients fd's on Host */
	int accept_clients_epoll_fd;				/* Epoll fd for main listen socket fd's on Host */
	GHashTable *host_fd_set;				/* Set of client file descriptors */
	GQueue *msg_queue;					/* Comm Channel Messages queue */
	int *conn_count;					/* Count number of active connection */
	int interrupt_fd;					/* Interrupt file descriptor */
	pthread_mutex_t *conn_count_mutex;			/* Incoming connections count conn_count_mutex */
	pthread_mutex_t *queue_mutex;				/* Queue mutex */
	pthread_mutex_t *table_mutex;				/* DPU fd to Host Glib table mutex */
	pthread_cond_t *queue_cond;				/* Messages queue condition variable */
	doca_error_t thread_result;				/* Thread final result */
};

/* This ctx is used by Host thread */
struct host_clients_listener_ctx {
	enum sr_mode mode;					/* Running mode */
	GHashTable *host_fd_set;				/* Set of client file descriptors */
	GQueue *msg_queue;					/* Comm Channel Messages queue */
	int clients_listener_epoll_fd;				/* Epoll fd for all clients fd's on Host */
	int *conn_count;					/* Count number of active connection */
	int interrupt_fd;					/* Interrupt file descriptor */
	pthread_mutex_t *conn_count_mutex;			/* Incoming connections count conn_count_mutex */
	pthread_mutex_t *queue_mutex;				/* Queue mutex */
	pthread_mutex_t *table_mutex;				/* DPU fd to Host Glib table mutex */
	pthread_cond_t *queue_cond;				/* Messages queue condition variable */
	doca_error_t thread_result;				/* Thread final result */
};

/* This ctx is used by DPU thread */
struct dpu_clients_listener_ctx {
	GQueue *msg_queue;					/* Comm Channel Messages queue */
	GHashTable *dpu_to_host_fd_table;			/* Table of client (DPU) fd to client fd (Host) */
	int clients_listener_epoll_fd;				/* Epoll fd which holds clients fd's on DPU */
	int interrupt_fd;					/* Interrupt file descriptor */
	pthread_mutex_t *table_mutex;				/* DPU fd to Host Glib table mutex */
	pthread_mutex_t *queue_mutex;				/* Queue mutex */
	pthread_cond_t *queue_cond;				/* Messages queue condition variable */
	doca_error_t thread_result;				/* Thread final result */
};

/*
 * Register application arguments
 *
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t register_sr_params(void);

/*
 * Initiate all needed data structures and start the tool
 *
 * @sr_cfg [in]: Mode of operation
 * @cc_listener_ctx [in]: Comm Channel listener thread context
 * @cc_sender_ctx [in]: Comm Channel sender thread context
 * @host_accept_clients_ctx [in]: Accept connections thread context
 * @host_clients_listener_ctx [in]: Host UDS clients listener thread context
 * @dpu_clients_listener_ctx [in]: DPU UDS clients listener thread context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t sr_start(struct sr_cfg *sr_cfg, struct cc_listener_ctx *cc_listener_ctx,
		      struct cc_sender_ctx *cc_sender_ctx, struct host_clients_acceptor_ctx *host_accept_clients_ctx,
		      struct host_clients_listener_ctx *host_clients_listener_ctx,
		      struct dpu_clients_listener_ctx *dpu_clients_listener_ctx);

#endif /* SOCKET_RELAY_CORE_H_ */
